/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement datetime
open core,termDomain
constants
className = "com/rudstrand/Datetime/datetime".
classVersion = "1.0".
clauses
classInfo(className, classVersion).
%   date format codes: refer to initial.pro (f_date)
%
class predicates
datNoPrior : (integer,integer) determ (i,o).
datAccNo : (integer,integer) determ (i,o).
datDaysInMonth : (integer,integer,integer) procedure (i,i,o).
datMonthDays : (integer,integer,integer) determ (i,i,o).
datMonthForDay : (integer,integer,integer) determ (i,i,o).
clauses
% ----------------------------------------------------------------------------------------------
% datAccLeap
% ----------------------------------------------------------------------------------------------
% returns the number of accumulated leap years for the years [1775,Year]
%
datAccLeap(Y,_) :-
	Y<1775,
	pfeil::syserr("datAccLeap","",[],[Y]),!,fail.
datAccLeap(1800,6) :- !.
datAccLeap(1900,30) :- !.
datAccLeap(1950,42) :- !.
datAccLeap(1990,51) :- !.
datAccLeap(Year,NLeaps) :-
	datLeapYear(Year,N),
	Y=Year-1,
	datAccLeap(Y,M),
	NLeaps=N+M,!.
% ----------------------------------------------------------------------------------------------
% datAccNo
% ----------------------------------------------------------------------------------------------
% determines the number of accumulated days prior to 1 Jan in the given year. Year must be 
% greater or equal to 1775.
%
datAccNo(Y,_) :-
	Y<1775,
	pfeil::syserr("datAccNo","Invalid year",[],[Y]),!,fail.
datAccNo(Y,N) :- 
	datNoPrior(Y,N),!
;
	Ym=Y-1,
	datLeapYear(Ym,Leap),
	datAccNo(Ym,N1),
	N=N1+365+Leap,!.
% ----------------------------------------------------------------------------------------------
% datWeekDay
% ----------------------------------------------------------------------------------------------
datWeekDay(Day,No) :-
	datDMY_to_I(3,1,2000,Base), % 03Jan2000 was a Monday (=1)
	Day>=Base,
	Diff=Day-Base,
	No=(Diff mod 7)+1,
	!
;
	datDMY_to_I(3,1,2000,Base), 
	Diff=Base-Day,
	No=8-(Diff mod 7),
	No<8,
	!
;
	No=1,
	!.	
% ----------------------------------------------------------------------------------------------
% datDayNo
% ----------------------------------------------------------------------------------------------
% D M Y ---> day no in year
%
datDayNo(D,M,Y,N) :-
	datLeapYear(Y,Leap),
	datMonthDays(Leap,M,MDays),
	N=MDays+D,!.
% ----------------------------------------------------------------------------------------------
% datLeapYear
% ----------------------------------------------------------------------------------------------
datLeapYear(Y,1) :- Rem=Y mod 400, Rem=0,!.
datLeapYear(Y,0) :- Rem=Y mod 100, Rem=0,!.
datLeapYear(Y,1) :- Rem=Y mod 4, Rem=0,!.
datLeapYear(_,0) :- !.
% ----------------------------------------------------------------------------------------------
% datDayInMonth
% ----------------------------------------------------------------------------------------------
datDayInMonth(Y,D,M) :-
	datLeapYear(Y,Leap),
	datMonthForDay(Leap,D,M),!.
% -------------------------------------------------------------------------- 
% datMonthNo
% -------------------------------------------------------------------------- 
class predicates
datMonthNoX : (string,integer) determ (i,o) (o,i) (i,i).
clauses
datMonthNo(M,N) :-
	bound(M), 
	MU=string::toUpperCase(M),
	datMonthNoX(MU,N),!
;
	free(M),	
	datMonthNoX(M,N),!.
	
datMonthNoX("JAN",1) :- !.
datMonthNoX("FEB",2) :- !.
datMonthNoX("MAR",3) :- !.
datMonthNoX("APR",4) :- !.
datMonthNoX("MAY",5) :- !.
datMonthNoX("JUN",6) :- !.
datMonthNoX("JUL",7) :- !.
datMonthNoX("AUG",8) :- !.
datMonthNoX("SEP",9) :- !.
datMonthNoX("OCT",10) :- !.
datMonthNoX("NOV",11) :- !.
datMonthNoX("DEC",12) :- !.
datMonthNoX(S,N) :-
	free(S), bound(N),
	pfeil::syserr("datMonthNoX","Invalid month no",[],[N]),!,fail
;
	bound(S), free(N),
	pfeil::syserr("datMonthNoX","Invalid month name",[S]),!,fail.	
% ----------------------------------------------------------------------------------------------
% datDMY_to_I
% ----------------------------------------------------------------------------------------------
% DD MM YYYY --> internal date no (1.1.1775 is 1) Dates must be on or after 1.1.1775
%
datDMY_to_I(D,M,Y,No) :-
	datAccNo(Y,N1),
	datDayNo(D,M,Y,N2),
	No=N1+N2,!
;
	Sd=toString(D), 
	Sm=toString(M),
	Sy=toString(Y),
	sx::concat([Sd,"-",Sm,"-",Sy,"'"],S),
	pfeil::syserr("datDMT_to_I","Invalid date",[S]),
	!,fail.
% ----------------------------------------------------------------------------------------------
% datI_to_DMY
% ----------------------------------------------------------------------------------------------
% internal date no --> DD MM YYYY (1.1.1775 is 1) Dates must be on or after 1.1.1775
%
class predicates
datNoCount : (integer,integer,integer,integer,integer) determ (i,i,o,o,o).
datNoBase : (integer,integer) procedure (i,o).
clauses
datI_to_DMY(No,D,M,Y) :-
	datNoBase(No,BaseYear),
	datNoPrior(BaseYear,N1),
	Rest=No-N1,
	datNoCount(BaseYear,Rest,D,M,Y),!.
datNoCount(Year,Rest,D,M,Y) :-
	datLeapYear(Year,Leap),
	NDays=365+Leap,
	Rest>NDays,
	Yn=Year+1,
	R=Rest-NDays,
	datNoCount(Yn,R,D,M,Y),!
;
	Y=Year,
	datLeapYear(Year,Leap),
	datMonthForDay(Leap,Rest,M),
	datMonthDays(Leap,M,MDays),
	D=Rest-MDays,!.
datNoBase(No,BaseYear) :-
	BaseYear=1990,
	datNoPrior(BaseYear,N),
	N<=No,!
;
	BaseYear=1950,
	datNoPrior(BaseYear,N),
	N<=No,!
;
	BaseYear=1900,
	datNoPrior(BaseYear,N),
	N<=No,!
;
	BaseYear=1800,
	datNoPrior(BaseYear,N),
	N<=No,!
;
	BaseYear=1775,!.
% ------------------------------------------------------------------------------------------------------------
% datNoPrior
% ------------------------------------------------------------------------------------------------------------
datNoPrior(1775,0) :- !.
datNoPrior(1800,9131) :- !.
datNoPrior(1900,45661) :- !.
datNoPrior(1950,63923) :- !.
datNoPrior(1990,78532) :- !.
% ------------------------------------------------------------------------------------------------------------
% datS
% ------------------------------------------------------------------------------------------------------------
% converts to Pro/3 output format (according to preference settings)
%
datS(DD,MM,YY,S) :-
	YY<100,
	globalVars::get_i("f_yyyy")>YY,
	YYYY=1900+YY,
	datS(DD,MM,YYYY,S),
	!
;
	YY<100,
	YYYY=2000+YY,
	datS(DD,MM,YYYY,S),
	!.
datS(DD,MM,YYYY,S) :-
	globalVars::get_i("f_date")=0,
	datDMY_to_I(DD,MM,YYYY,No),
	S=toString(No),
	!
;	
	globalVars::get_i("f_date")=2,
	No=YYYY*10000+MM*100+DD,
	S=toString(No),
	!
;
	globalVars::get_i("f_datm")=0,
	C=globalVars::get_s("f_dats"),
	Sd=toString(DD),
	sx::sxL_Pad(Sd,"0",2,Sdp),
	Md=toString(MM),
	sx::sxL_Pad(Md,"0",2,Smp),
	Sy=toString(YYYY),
	sx::sxL_Pad(Sy,"0",2,Syp),
	sx::sxlJoin([Sdp,Smp,Syp],C,S),!
;		
	globalVars::get_i("f_datm")=1,
	C=globalVars::get_s("f_dats"),
	Sd=toString(DD),
	sx::sxL_Pad(Sd,"0",2,Sdp),
	datMonthNo(MMM,MM),
	Sy=toString(YYYY),
	sx::sxL_Pad(Sy,"0",2,Syp),
	sx::sxlJoin([Sdp,MMM,Syp],C,S),!.
% ------------------------------------------------------------------------------------------------------------
% datS_to_I
% ------------------------------------------------------------------------------------------------------------
datS_to_I(S,No) :-
	globalVars::get_i("f_datm")=0,
	Cs=globalVars::get_s("f_dats"),
	string::frontchar(Cs,C,_),
	sx::searchchar(S,C,P1),
	N1=P1-1,
	sx::frontstr(N1,S,Ds,_),
	sx::frontstr(P1,S,_,S1),
	sx::searchchar(S1,C,P2),
	N2=P2-1,
	sx::frontstr(N2,S1,Ms,_),
	sx::frontstr(P2,S1,_,Ys),
    D=sx::toInt(Ds),
	M=sx::toInt(Ms),
	Y=sx::toInt(Ys),
	datYYYY(Y,YY),
	datDMY_to_I(D,M,YY,No),!
;
	globalVars::get_i("f_datm")=1,
	Cs=globalVars::get_s("f_dats"),
	string::frontchar(Cs,C,_),
	sx::searchchar(S,C,P1),
	N1=P1-1,
	sx::frontstr(N1,S,Ds,_),
	sx::frontstr(P1,S,_,S1),
	sx::frontstr(3,S1,Ms,_),
	sx::frontstr(4,S1,_,Ys),
	D=sx::toInt(Ds),
	datMonthNo(Ms,M),
	Y=sx::toInt(Ys),
	datYYYY(Y,YY),
	datDMY_to_I(D,M,YY,No),!.	
% ------------------------------------------------------------------------------------------------------------
% datStrAny
% ------------------------------------------------------------------------------------------------------------
datStrAny(SDate,DateNo) :-
	% YYYYMMDD
	sx::sxAllDigits(SDate),
	string::length(SDate)=8,
	sx::frontstr(4,SDate,YYYYs,S2),
	YYYY=sx::toInt(YYYYs),
	%
	sx::frontstr(2,S2,MMs,DDs),
	MM=sx::toInt(MMs),
	DD=sx::toInt(DDs),
	datDMY_to_I(DD,MM,YYYY,DateNo),
	!
;	
	% YYMMDD
	sx::sxAllDigits(SDate),
	L=string::length(SDate),
	L=6,
	sx::frontstr(2,SDate,YYs,S2),
	YY=sx::toInt(YYs),
	%
	sx::frontstr(2,S2,MMs,DDs),
	MM=sx::toInt(MMs),
	DD=sx::toInt(DDs),
	datYYYY(YY,YYYY),
	datDMY_to_I(DD,MM,YYYY,DateNo),!
;	
	% DD-MM-YY[YY]
	sx::frontstr(2,SDate,DDs,_),
	sx::sxAllDigits(DDs),
	DD=sx::toInt(DDs),
	%
	sx::frontstr(3,SDate,_,S2),
	sx::frontstr(2,S2,MMs,_),
	MM=sx::toInt(MMs),
	%
	sx::frontstr(3,S2,_,Ys),
	Y=sx::toInt(Ys),
	datYYYY(Y,YYYY),
	%
	datDMY_to_I(DD,MM,YYYY,DateNo),!
;	
	% DD-M-YY[YY]
	sx::frontstr(2,SDate,DDs,_),
	DD=sx::toInt(DDs),
	%
	sx::frontstr(3,SDate,_,S2),
	sx::frontstr(1,S2,MMs,_),
	MM=sx::toInt(MMs),
	%
	sx::frontstr(2,S2,_,Ys),
	Y=sx::toInt(Ys),
	datYYYY(Y,YYYY),
	%
	datDMY_to_I(DD,MM,YYYY,DateNo),!
;
	% DD-MMM-YY[YY]
	sx::frontstr(2,SDate,DDs,_),
	sx::sxAllDigits(DDs),
	DD=sx::toInt(DDs),
	%
	sx::frontstr(3,SDate,_,S2),
	sx::frontstr(3,S2,MMs,_),
	datMonthNo(MMs,MM),
	%
	sx::frontstr(4,S2,_,Ys),
	Y=sx::toInt(Ys),
	datYYYY(Y,YYYY),
	%
	datDMY_to_I(DD,MM,YYYY,DateNo),!
;
	% D-MM-YY[YY]
	sx::frontstr(1,SDate,DDs,_),
	sx::sxAllDigits(DDs),
	DD=sx::toInt(DDs),
	%
	sx::frontstr(2,SDate,_,S2),
	sx::frontstr(2,S2,MMs,_),
	sx::sxAllDigits(MMs),
	MM=sx::toInt(MMs),
	%
	sx::frontstr(3,S2,_,Ys),
	Y=sx::toInt(Ys),
	datYYYY(Y,YYYY),
	%
	datDMY_to_I(DD,MM,YYYY,DateNo),!
;	
	% D-M-YY[YY]
	sx::frontstr(1,SDate,DDs,_),
	sx::sxAllDigits(DDs),
	DD=sx::toInt(DDs),
	%
	sx::frontstr(2,SDate,_,S2),
	sx::frontstr(1,S2,MMs,_),
	sx::sxAllDigits(MMs),
	MM=sx::toInt(MMs),
	%
	sx::frontstr(2,S2,_,Ys),
	Y=sx::toInt(Ys),
	datYYYY(Y,YYYY),
	%
	datDMY_to_I(DD,MM,YYYY,DateNo),!
;
	% D-MMM-YY[YY]
	sx::frontstr(1,SDate,DDs,_),
	sx::sxAllDigits(DDs),
	DD=sx::toInt(DDs),
	%
	sx::frontstr(2,SDate,_,S2),
	sx::frontstr(3,S2,MMs,_),
	datMonthNo(MMs,MM),
	%
	sx::frontstr(3,S2,_,Ys),
	Y=sx::toInt(Ys),
	datYYYY(Y,YYYY),
	%
	datDMY_to_I(DD,MM,YYYY,DateNo),!
;
	pfeil::syserr("datStrAny",SDate),!,fail.
% ------------------------------------------------------------------------------------------------------------
% datYYYY
% ------------------------------------------------------------------------------------------------------------
datYYYY(YYYY,YYYY) :- YYYY>=1000,!.
datYYYY(YY,YYYY) :-
	globalVars::get_i("f_yyyy")>=YY,
	YYYY=2000+YY,!
;
	YYYY=1900+YY,!.	
% -------------------------------------------------------------------------- 
% timCmp                      
% -------------------------------------------------------------------------- 
timCmp(Da,Ta,Db,Tb,C) :-
	Da>Db,
	C=-1,!
;
	Da=Db,
	Ta>Tb,
	C=-1,!
;
	Da=Db,
	Ta=Tb,
	C=0,!
;
	C=1,!.			
% -------------------------------------------------------------------------- 
% timNow                     
% -------------------------------------------------------------------------- 
timNow(Dat,Tim) :-
    Time=time::new(),
    Time:getDate(Y,M,D), 
    Time:getTime(H,Mi,S), %@@ getTimeDetailed could be used if sub-seconds are required 22.8.4
	datDMY_to_I(D,M,Y,Dat),
	Tim=H*360000+Mi*6000+S*100,
	!.
% -------------------------------------------------------------------------- 
% timDurText                
% -------------------------------------------------------------------------- 
timDurText(Days,Hours,Mins,Secs,T) :-
	THours=Hours+Days*24,
	THours<100,
	S2=toString(Secs),
	sx::sxL_Pad(S2,"0",2,S3),
	M2=toString(Mins),
	sx::sxL_Pad(M2,"0",2,M3),
	H2=toString(THours),
	sx::sxL_Pad(H2,"0",2,H3),
	sx::concat([H3,":",M3,":",S3],T),
	!
;
	THours=Hours+Days*24,
	THours>=100,
	S2=toString(Secs),
	sx::sxL_Pad(S2,"0",2,S3),
	M2=toString(Mins),
	sx::sxL_Pad(M2,"0",2,M3),
	sx::concat(["**:",M3,":",S3],T),
	!.
% -------------------------------------------------------------------------- 
% tsCmp                       
% -------------------------------------------------------------------------- 
% succeeds if <T1> <opr> <T2> is true.
%
tsCmp("=",T,T) :- !.
tsCmp("<",[Y,M,D,H,Mi,S,B],[Y2,M2,D2,H2,Mi2,S2,B2]) :-
	Y<Y2,!
;	Y=Y2,M<M2,!
;	Y=Y2,M=M2,D<D2,!
;	Y=Y2,M=M2,D=D2,H<H2,!
;	Y=Y2,M=M2,D=D2,H=H2,Mi<Mi2,!
;	Y=Y2,M=M2,D=D2,H=H2,Mi=Mi2,S<S2,!
;	Y=Y2,M=M2,D=D2,H=H2,Mi=Mi2,S=S2,B<B2,!.
tsCmp(">=",T1,T2) :- not(tsCmp("<",T1,T2)),!.
tsCmp("<>",T1,T2) :- not(tsCmp("=",T1,T2)),!.
tsCmp(">",T1,T2)  :- not(tsCmp("<",T1,T2)),not(tsCmp("=",T1,T2)),!.
tsCmp("<=",T1,T2) :- 
	tsCmp("<",T1,T2),!
;	tsCmp("=",T1,T2),!.
% -------------------------------------------------------------------------- 
% tsDiff
% -------------------------------------------------------------------------- 
class predicates
tsDiff_2 : (integer,integer,integer,integer,integer,integer,integer,integer,integer,integer,integer,integer) procedure (i,i,i,i,i,i,i,i,o,o,o,o).
clauses
% returns the difference of two time stamps (first timestamp must be <= second timestamp)
%
tsDiff([Y2,M2,D2,H2,Mi2,S2,_],[Y3,M3,D3,H3,Mi3,S3,_],D,H,M,S) :-
	Y2=Y3,
	M2=M3,
	tsDiff_2(D2,H2,Mi2,S2,D3,H3,Mi3,S3,D,H,M,S),
	!
;
	datDMY_to_I(D2,M2,Y2,No2),
	datDMY_to_I(D3,M3,Y3,No3),
	tsDiff_2(1,H2,Mi2,S2,0,H3,Mi3,S3,D4,H,M,S),
	D=No3-No2-1+D4-1,
	!.
		
tsDiff_2(D2,H2,M2,S2,D3,H3,M3,S3,D,H,M,S) :-
	Diff=(D3-D2)*86400+(H3-H2)*3600+(M3-M2)*60+S3-S2,
	D=Diff div 86400,
	H=(Diff-D*86400) div 3600,
	M=(Diff-D*86400-H*3600) div 60,
	S=Diff mod 60,
	!.	
% -------------------------------------------------------------------------- 
% tsDiffSecs
% -------------------------------------------------------------------------- 
% returns the difference in whole no. of seconds between first and second timestamp
% Note! CANNOT HANDLE CROSSING OF MONTH
%
tsDiffSecs([Y1,M1,D1,H1,Mi1,S1,_],[Y2,M2,D2,H2,Mi2,S2,_],Diff) :-
	Y1=Y2, M1=M2, D1=D2,
	Diff=(H2-H1)*3600+(Mi2-Mi1)*60+S2-S1+1,
	!
;
	Diff1=(D2-D1)*24+(H2-H1),
	Diff=Diff1*3600+(Mi2-Mi1)*60+S2-S1+1,
	!
;
    vpiCommonDialogs::note("Error in datetime::tsDiffSecs","Cannot compute elapsed time over a month boundary!"),
	Diff=0,
	!.		
% -------------------------------------------------------------------------- 
% tsStrIL                     
% -------------------------------------------------------------------------- 
% converts from timestamp integer list format to string format
%
tsStrIL([Y,M,D,HH,MM,SS,FF],S) :-
	Ys=toString(Y),
	Ms=toString(M),
	Ds=toString(D),
	HHs=toString(HH),
	MMs=toString(MM),
	SSs=toString(SS),
	FFs=toString(FF),
	sx::concat(["[",Ys,",",Ms,",",Ds,",",HHs,",",MMs,",",SSs,",",FFs,"]"],S),!.
% -------------------------------------------------------------------------- 
% tmrSet/Expired            
% -------------------------------------------------------------------------- 
tmrSet(0) :-
	globalVars::set_n("timer",-1.0),
	!.	
tmrSet(Timset) :-
	tsNow(T),
	T=[Y,M,D,HH,MM,SS,_],
	Timer=(Y-2000)*365.25+M*30.4375+D+((HH*3600+MM*60+SS)+Timset)/86400.0,
	globalVars::set_n("timer",Timer),
	!.
	
tmrExpired() :-
	globalVars::get_n("timer")=-1.0,
	!,fail
;
	tsNow(T),
	T=[Y,M,D,HH,MM,SS,_],
	Now=(Y-2000)*365.25+M*30.4375+D+(HH*3600.0+MM*60.0+SS)/86400.0,
	globalVars::get_n("timer")<Now,
	!.	
% -------------------------------------------------------------------------- 
% tsNow                
% -------------------------------------------------------------------------- 
% returns current timestamp
%
tsNow([Y,M,D,H,Mi,S,0]) :-
    Time=time::new(),
    Time:getDate(Y,M,D), 
    Time:getTime(H,Mi,S), %@@ getTimeDetailed could be used if sub-seconds are required 22.8.4
    !.
% -------------------------------------------------------------------------- 
% tsStr
% -------------------------------------------------------------------------- 
% converts between timestamp and date-time text format
%
tsStr(TS,DateTime) :-
	TS=[YYYY,MM,DD,H,M,S,_],
	datS(DD,MM,YYYY,S1),
	sx::sx_str_int(Hs,2,H),
	sx::sx_str_int(Ms,2,M),
	sx::sx_str_int(Ss,2,S),
	sx::concat([S1," ",Hs,":",Ms,":",Ss],DateTime),
	!.

tsXML(TS,DateTime) :-
	TS=[YYYY,MM,DD,H,M,S,_],
	sx::sx_str_int(YYYYs,4,YYYY),
	sx::sx_str_int(MMs,2,MM),
	sx::sx_str_int(DDs,2,DD),
	sx::sx_str_int(Hs,2,H),
	sx::sx_str_int(Ms,2,M),
	sx::sx_str_int(Ss,2,S),
	sx::concat([YYYYs,"-",MMs,"-",DDs,"T",Hs,":",Ms,":",Ss],DateTime),
	!.

tsSQL(TS,DateTime) :-
	TS=[YYYY,MM,DD,H,M,S,_],
	sx::sx_str_int(YYYYs,4,YYYY),
	sx::sx_str_int(MMs,2,MM),
	sx::sx_str_int(DDs,2,DD),
	sx::sx_str_int(Hs,2,H),
	sx::sx_str_int(Ms,2,M),
	sx::sx_str_int(Ss,2,S),
	sx::concat([YYYYs,"-",MMs,"-",DDs," ",Hs,":",Ms,":",Ss],DateTime),
	!.
% ----------------------------------------------------------------------------------------------
% datMonthDays
% ----------------------------------------------------------------------------------------------
% returns the no of days in the year prior to given month no (par#2) par#1 is 1 if leap year,
% else 0
%
datMonthDays(_,1,0) :- !.
datMonthDays(_,2,31) :- !.
datMonthDays(0,3,59) :- !.
datMonthDays(0,4,90) :- !.
datMonthDays(0,5,120) :- !.
datMonthDays(0,6,151) :- !.
datMonthDays(0,7,181) :- !.
datMonthDays(0,8,212) :- !.
datMonthDays(0,9,243) :- !.
datMonthDays(0,10,273) :- !.
datMonthDays(0,11,304) :- !.
datMonthDays(0,12,334) :- !.
datMonthDays(1,3,60) :- !.
datMonthDays(1,4,91) :- !.
datMonthDays(1,5,121) :- !.
datMonthDays(1,6,152) :- !.
datMonthDays(1,7,182) :- !.
datMonthDays(1,8,213) :- !.
datMonthDays(1,9,244) :- !.
datMonthDays(1,10,274) :- !.
datMonthDays(1,11,305) :- !.
datMonthDays(1,12,335) :- !.
% ----------------------------------------------------------------------------------------------
% datMonthForDay
% ----------------------------------------------------------------------------------------------
% returns the month no for a given day no in year. par#1 is 1 if leap year, else 0
%
datMonthForDay(0,D,M) :-
	D<32, M=1,!
;	D<60, M=2,!	
;	D<91, M=3,!	
;	D<121, M=4,!	
;	D<152, M=5,!	
;	D<182, M=6,!	
;	D<213, M=7,!	
;	D<244, M=8,!	
;	D<274, M=9,!	
;	D<305, M=10,!	
;	D<335, M=11,!	
;	M=12,!.
datMonthForDay(1,D,M) :-
	D<32, M=1,!
;	D<61, M=2,!	
;	D<92, M=3,!	
;	D<122, M=4,!	
;	D<153, M=5,!	
;	D<183, M=6,!	
;	D<214, M=7,!	
;	D<245, M=8,!	
;	D<275, M=9,!	
;	D<306, M=10,!	
;	D<336, M=11,!	
;	M=12,!.	
% ----------------------------------------------------------------------------------------------
% datDaysInMonth
% ----------------------------------------------------------------------------------------------
datDaysInMonth(Y,M,Days) :-
	set::member(M,[1,3,5,7,8,10,12]),
	Days=31,
	!
;
	set::member(M,[4,6,9,11]),
	Days=30,
	!
;
	datLeapYear(Y,Leap),
	Days=28+Leap,
	!.
end implement datetime
