/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement dirgraph
open core,termDomain

constants

className = "com/rudstrand/dirgraph/dirgraph".
classVersion = "1.0".

facts

acyclic : integer:=0. % -1=cycles allowed; 1=acyclic
dg : (integer Mark,sterm Type1,sterm Name1,sterm Type2,sterm Name2,sterm Status) nondeterm.
visio : (integer Min,integer Max,sterm Type,string VisioType) nondeterm.
shape : (string,string,string) nondeterm.

predicates

markAll : (integer).
markNode : (integer,sterm,sterm,sterm,sterm).
markNodes : (integer,sterm,sterm).
marked : (integer,sterm,sterm,sterm,sterm) determ.

drawGraph : (outputStream,count).

clauses
% -----------------------------------------------------------------------------------------------
%                                            CLASS PREDICATES
% -----------------------------------------------------------------------------------------------
classInfo(className, classVersion).
% -----------------------------------------------------------------------------------------------
% new
% -----------------------------------------------------------------------------------------------
new() :- new(-1),!.

new(Acyclic) :-
    retractall(dg(_,_,_,_,_,_)),
    Acyclic=1,
    acyclic:=1,
    !
;
    acyclic:=-1,
    !.
% -----------------------------------------------------------------------------------------------
% isAcyclic
% -----------------------------------------------------------------------------------------------
isAcyclic()=Acyclic :- Acyclic=acyclic,!.
% -----------------------------------------------------------------------------------------------
% getName
% -----------------------------------------------------------------------------------------------
getName(Name,VName) :-
    fstrSwitchR(Tf,Name),
    fstr_str(Tf,VName),
    !
;
    VName=toString(Name),
    !.
% -----------------------------------------------------------------------------------------------
% copyGraph
% -----------------------------------------------------------------------------------------------
class predicates
copyGraphX : (dirgraph,dirgraph).
clauses

copyGraph(G,H) :-
    H=dirgraph::new(1),
    copyGraphX(G,H),
    !.
    
copyGraphX(G,H) :-    
    G:get_nd(T1,N1,T2,N2,S),
        H:add(T1,N1,T2,N2,S),
        fail
;
    !.
% -----------------------------------------------------------------------------------------------
%                                          OBJECT PREDICATES
% -----------------------------------------------------------------------------------------------
get_nd(Name1,Name2) :- dg(_,_,Name1,_,Name2,_).
get_nd(Type1,Name1,Type2,Name2) :- dg(_,Type1,Name1,Type2,Name2,_).
get_nd(Type1,Name1,Type2,Name2,Status) :- dg(_,Type1,Name1,Type2,Name2,Status).
delete(Type1,Name1,Type2,Name2) :- retractall(dg(_,Type1,Name1,Type2,Name2,_)),!.
delete(Type1,Name1,Type2,Name2,Status) :- retractall(dg(_,Type1,Name1,Type2,Name2,Status)),!.
deleteAll() :- retractall(dg(_,_,_,_,_,_)),!.
deleteAllWherePred(Type) :- retractall(dg(_,Type,_,_,_,_)),!.
deleteAllWherePred(Type,Name) :- retractall(dg(_,Type,Name,_,_,_)),!.
deleteAllWhereSucc(Type) :- retractall(dg(_,_,_,Type,_,_)),!.
deleteAllWhereSucc(Type,Name) :- retractall(dg(_,_,_,Type,Name,_)),!.
deleteWhereStatus(Status) :- retractall(dg(_,_,_,_,_,Status)),!.
has(Status) :- dg(_,_,_,_,_,Status),!.
has(Name1,Name2) :- dg(_,_,Name1,_,Name2,_),!.
has(Type1,Name1,Type2,Name2) :- dg(_,Type1,Name1,Type2,Name2,_),!.
has(Type1,Name1,Type2,Name2,Status) :- dg(_,Type1,Name1,Type2,Name2,Status),!.
hasAsPred(Type,Name) :- dg(_,Type,Name,_,_,_),!.
hasAsSucc(Type,Name) :- dg(_,_,_,Type,Name,_),!.
% -----------------------------------------------------------------------------------------------
% allNodes
% -----------------------------------------------------------------------------------------------
predicates
allNodesX : (termSet).
clauses

allNodes()=Set :-
    Set=termSet::new(),
    allNodesX(Set),
    !.
    
allNodesX(Set) :-
    get_nd(T1,N1,T2,N2),
    Set:add(cmp("",[T1,N1])),
    Set:add(cmp("",[T2,N2])),
    fail
;
    !.
% -----------------------------------------------------------------------------------------------
% add
% -----------------------------------------------------------------------------------------------
add(N1,N2) :- add(nill,N1,nill,N2,nill),!.
add(T1,N1,T2,N2) :- add(T1,N1,T2,N2,nill),!.

add(T1,N1,T2,N2,S) :- 
    dg(_,T1,N1,T2,N2,_),
    !
;
    acyclic=1,
    % check if new node will violate acyclicity constraint
    Set=succsRec(T2,N2),
    Set:member(cmp("",[T1,N1])),
    !,fail
;
    assert(dg(0,T1,N1,T2,N2,S)),
    !.
% -----------------------------------------------------------------------------------------------
% replace
% -----------------------------------------------------------------------------------------------
replace(FromT,FromN,ToT,ToN) :-
    dg(M,FromT,FromN,T2,N2,S),
        retract(dg(M,FromT,FromN,T2,N2,S)),
        assert(dg(M,ToT,ToN,T2,N2,S)),
        fail
;        
    dg(M,T1,N1,FromT,FromN,S),
        retract(dg(M,T1,N1,FromT,FromN,S)),
        assert(dg(M,T1,N1,ToT,ToN,S)),
        fail
;        
    !.
% -----------------------------------------------------------------------------------------------
% changeType
% -----------------------------------------------------------------------------------------------
changeType(From,To) :-
    dg(M,From,N1,T2,N2,S),
        retract(dg(M,From,N1,T2,N2,S)),
        assert(dg(M,To,N1,T2,N2,S)),
        fail
;        
    dg(M,T1,N1,From,N2,S),
        retract(dg(M,T1,N1,From,N2,S)),
        assert(dg(M,T1,N1,To,N2,S)),
        fail
;        
    !.
% -----------------------------------------------------------------------------------------------
% deleteStartNodes
% -----------------------------------------------------------------------------------------------
deleteStartNodes() :-
    dg(M,A1,A2,B1,B2,S),
        not(dg(_,_,_,A1,A2,_)),
        retract(dg(M,A1,A2,B1,B2,S)),
        fail
;
    !.
% -----------------------------------------------------------------------------------------------
% setStatus
% -----------------------------------------------------------------------------------------------
setStatus(S) :-
    dg(M,T1,N1,T2,N2,Status),
        not(Status=S),
        retract(dg(M,T1,N1,T2,N2,Status)),
        assert(dg(M,T1,N1,T2,N2,S)),
        fail
;
    !.

setStatus(FromStatus,ToStatus) :-
    dg(M,T1,N1,T2,N2,FromStatus),
        retract(dg(M,T1,N1,T2,N2,FromStatus)),
        assert(dg(M,T1,N1,T2,N2,ToStatus)),
        fail
;
    !.

setStatus(T1,N1,T2,N2,S) :-
    dg(M,T1,N1,T2,N2,Status),
        not(Status=S),
        retract(dg(M,T1,N1,T2,N2,Status)),
        assert(dg(M,T1,N1,T2,N2,S)),
        fail
;
    !.

setStatusWherePred(T1,N1,S) :-
    dg(M,T1,N1,T2,N2,Status),
        not(Status=S),
        retract(dg(M,T1,N1,T2,N2,Status)),
        assert(dg(M,T1,N1,T2,N2,S)),
        fail
;
    !.

setStatusWherePred(T1,N1,FromStatus,ToStatus) :-
    dg(M,T1,N1,T2,N2,FromStatus),
        retract(dg(M,T1,N1,T2,N2,FromStatus)),
        assert(dg(M,T1,N1,T2,N2,ToStatus)),
        fail
;
    !.

setStatusWhereSucc(T2,N2,S) :-
    dg(M,T1,N1,T2,N2,Status),
        not(Status=S),
        retract(dg(M,T1,N1,T2,N2,Status)),
        assert(dg(M,T1,N1,T2,N2,S)),
        fail
;
    !.

setStatusWhereSucc(T2,N2,FromStatus,ToStatus) :-
    dg(M,T1,N1,T2,N2,FromStatus),
        retract(dg(M,T1,N1,T2,N2,FromStatus)),
        assert(dg(M,T1,N1,T2,N2,ToStatus)),
        fail
;
    !.
% -----------------------------------------------------------------------------------------------
% deleteNode
% -----------------------------------------------------------------------------------------------
deleteNode(Type,Name) :-
    retractall(dg(_,Type,Name,_,_,_)),
    retractall(dg(_,_,_,Type,Name,_)),
    !.
% -----------------------------------------------------------------------------------------------
% deleteNodeType
% -----------------------------------------------------------------------------------------------
deleteNodeType(Type) :-
    retractall(dg(_,Type,_,_,_,_)),
    retractall(dg(_,_,_,Type,_,_)),
    !.
% -----------------------------------------------------------------------------------------------
% submerge/P/S
% -----------------------------------------------------------------------------------------------
submerge(T,N,Status) :-
    dg(_,Tx,Nx,T,N,_),
       dg(_,T,N,Ta,Tb,_),
            add(Tx,Nx,Ta,Tb,Status),
            fail
;
    retractall(dg(_,T,N,_,_,_)),
    retractall(dg(_,_,_,T,N,_)),
    !.        
            
submergeP(T,N) :-
    dg(_,Tx,Nx,T,N,Status),
       dg(_,T,N,Ta,Tb,_),
            add(Tx,Nx,Ta,Tb,Status),
            fail
;
    retractall(dg(_,T,N,_,_,_)),
    retractall(dg(_,_,_,T,N,_)),
    !.        
            
submergeS(T,N) :-
    dg(_,Tx,Nx,T,N,_),
       dg(_,T,N,Ta,Tb,Status),
            add(Tx,Nx,Ta,Tb,Status),
            fail
;
    retractall(dg(_,T,N,_,_,_)),
    retractall(dg(_,_,_,T,N,_)),
    !.        

subMergeType(T) :- subMergeTypeP(T),!.

submergeType(T,Status) :-
    dg(_,Tx,Nx,T,N,_),
       dg(_,T,N,Ta,Tb,_),
            add(Tx,Nx,Ta,Tb,Status),
            fail
;
    retractall(dg(_,T,_,_,_,_)),
    retractall(dg(_,_,_,T,_,_)),
    !.        
            
submergeTypeP(T) :-
    dg(_,Tx,Nx,T,N,Status),
       dg(_,T,N,Ta,Tb,_),
            add(Tx,Nx,Ta,Tb,Status),
            fail
;
    retractall(dg(_,T,_,_,_,_)),
    retractall(dg(_,_,_,T,_,_)),
    !.        
            
submergeTypeS(T) :-
    dg(_,Tx,Nx,T,N,_),
       dg(_,T,N,Ta,Tb,Status),
            add(Tx,Nx,Ta,Tb,Status),
            fail
;
    retractall(dg(_,T,_,_,_,_)),
    retractall(dg(_,_,_,T,_,_)),
    !.        
% -----------------------------------------------------------------------------------------------
% mark...
% -----------------------------------------------------------------------------------------------
markAll(Mark) :-
    dg(M,T1,N1,T2,N2,S),
        M<>Mark,
        retract(dg(M,T1,N1,T2,N2,S)),
        assert(dg(Mark,T1,N1,T2,N2,S)),
        fail
;
    !.
    
markNode(Mark,T,N,T2,N2) :-
    dg(M,T,N,T2,N2,S),
        M<>Mark,
        retract(dg(M,T,N,T2,N2,S)),
        assert(dg(Mark,T,N,T2,N2,S)),
        fail
;
    !.

markNodes(Mark,T,N) :-
    dg(M,T,N,T2,N2,S),
        M<>Mark,
        retract(dg(M,T,N,T2,N2,S)),
        assert(dg(Mark,T,N,T2,N2,S)),
        fail
;
    !.
    
marked(Mark,T,N,T2,N2) :- dg(Mark,T,N,T2,N2,_),!.    
% -----------------------------------------------------------------------------------------------
% a/cyclicNodes
% -----------------------------------------------------------------------------------------------
predicates
cyclicNodesX : (termSet).
clauses

acyclicNodes()=Set :-
    SetC=cyclicNodes(),
    Set=allNodes(),
    Set:subtract(SetC),
    !.
    
    
cyclicNodes()=Set :-
    Set=termSet::new(),
    cyclicNodesX(Set),
    !.
    
cyclicNodesX(Set) :-
    dg(_,T,N,_,_,_),
    Succs=succsRec(T,N),
    Succs:member(cmp("",[T,N])),
    Set:add(cmp("",[T,N])),
    fail
;
    !.
% -----------------------------------------------------------------------------------------------
% cycle
% -----------------------------------------------------------------------------------------------
predicates
cycleSetX : (termSet,sterml,sterm,sterml,sterml) determ (i,i,i,i,o).
clauses

% fails if T|N is not in a cycle, else
% returns the members in (one of the) cycle(s) where T|N is a member
%
cycleSet(T,N,CycleList) :-
    StartNode=cmp("",[T,N]),
    SuccSet=succsRec(T,N),
    SuccSet:member(StartNode),
    % T|N is in a cycle
    Succs=succs(T,N),
    L=Succs:get(),
    VisitedSet=termSet::new(),
    cycleSetX(VisitedSet,[],StartNode,L,CycleList),
    !
;
    StartNode=cmp("",[T,N]),
    SuccSet=succsRec(T,N),
    SuccSet:member(StartNode),
    N=str(Start),
    S=string::concat("Error: incomplete cycle for  ",Start),
    CycleList=[cmp("",[str("NIL"),str(S)])],
    !.    
    
cycleSetX(_,ListIn,StartNode,[Node|_],[StartNode|ListIn]) :-
    StartNode=Node,
    !. % cycle is completed
cycleSetX(VisitedSet,ListIn,StartNode,[Node|Ns],ListOut) :-
    Node=cmp("",[T,N]), 
    not(set::member(Node,ListIn)),
    % not already part of the cycle being determined
    not(VisitedSet:member(Node)),
    % not yet visited this node
    Succs=succs(T,N),
    L=Succs:get(),
    % continue with all of Node's successors
    VisitedSet:add(Node),
    cycleSetX(VisitedSet,[Node|ListIn],StartNode,L,ListOut),
    !
;
    cycleSetX(VisitedSet,ListIn,StartNode,Ns,ListOut),
    !.    
% -----------------------------------------------------------------------------------------------
% preds
% -----------------------------------------------------------------------------------------------
predicates
predsX : (termSet,sterm,sterm).
predsRecGet : (integer,termSet,sterm,sterm) determ.  
predsRecGetX : (integer,termSet,sterml,sterml) determ.    
clauses

preds(T,N)=Set :-
    Set=termSet::new(),
    predsX(Set,T,N),
    !.
    
predsX(Set,T,N) :-
    dg(_,T2,N2,T,N,_),
    Set:add(cmp("",[T2,N2])),
    fail
;
    !.

% return node T|N's successor set L levels deep
%    
predsRec(T,N)=Set :- Set=predsRec(999999,T,N),!.

predsRec(L,T,N)=Set :-
    Set=termSet::new(),
    predsRecGet(L,Set,T,N),
    !.

predsRecGet(0,_,_,_) :-
    !.
predsRecGet(L,Set,T,N) :-
    Preds=preds(T,N),
    PredsNodes=Preds:get(),
    unpackNodes(PredsNodes,Ts,Ns),
    predsRecGetX(L,Set,Ts,Ns),
    !.
    
predsRecGetX(_,_,[],[]) :-
    !.
predsRecGetX(L,Set,[T|Ts],[N|Ns]) :-
    Set:member(cmp("",[T,N])),
    predsRecGetX(L,Set,Ts,Ns),
    !
;    
    Set:add(cmp("",[T,N])),
    Next=L-1,
    predsRecGet(Next,Set,T,N),
    predsRecGetX(L,Set,Ts,Ns),
    !.
% -----------------------------------------------------------------------------------------------
% succs/Rec
% -----------------------------------------------------------------------------------------------
predicates
succsX : (termSet,sterm,sterm).
succsRecGet : (integer,termSet,sterm,sterm) determ.  
succsRecGetX : (integer,termSet,sterml,sterml) determ.    
unpackNodes : (sterml,sterml,sterml) determ (i,o,o).
clauses

succs(T,N)=Set :-
    Set=termSet::new(),
    succsX(Set,T,N),
    !.
    
succsX(Set,T,N) :-
    dg(_,T,N,T2,N2,_),
    Set:add(cmp("",[T2,N2])),
    fail
;
    !.

% return node T|N's successor set L levels deep
%    
succsRec(T,N)=Set :- Set=succsRec(999999,T,N),!.

succsRec(L,T,N)=Set :-
    Set=termSet::new(),
    succsRecGet(L,Set,T,N),
    !.

succsRecGet(0,_,_,_) :-
    !.
succsRecGet(L,Set,T,N) :-
    Succs=succs(T,N),
    SuccsNodes=Succs:get(),
    unpackNodes(SuccsNodes,Ts,Ns),
    succsRecGetX(L,Set,Ts,Ns),
    !.
    
succsRecGetX(_,_,[],[]) :-
    !.
succsRecGetX(L,Set,[T|Ts],[N|Ns]) :-
    Set:member(cmp("",[T,N])),
    succsRecGetX(L,Set,Ts,Ns),
    !
;    
    Set:add(cmp("",[T,N])),
    Next=L-1,
    succsRecGet(Next,Set,T,N),
    succsRecGetX(L,Set,Ts,Ns),
    !.
    
unpackNodes([],[],[]) :-
    !.
unpackNodes([Node|Nodes],[T|Ts],[N|Ns]) :-
    Node=cmp(_,[T,N]),
    unpackNodes(Nodes,Ts,Ns),
    !.
% -----------------------------------------------------------------------------------------------
% startNodes/NodeNames
% -----------------------------------------------------------------------------------------------
predicates
startNodesX : (termSet).
startNodeNamesX : (termSet).
clauses

startNodes()=TS :-
    TS=termSet::new(),
    startNodesX(TS),
    !.

startNodesX(TS) :-
    dg(_,T1,N1,_,_,_),
    not(dg(_,_,_,T1,N1,_)),
    not(TS:member(cmp("",[T1,N1]))),
    TS:add(cmp("",[T1,N1])),
    fail
;
    !.

startNodeNames()=TS :-
    TS=termSet::new(),
    startNodeNamesX(TS),
    !.

startNodeNamesX(TS) :-
    dg(_,_,N1,_,_,_),
    not(dg(_,_,_,_,N1,_)),
    not(TS:member(N1)),
    TS:add(N1),
    fail
;
    !.
% -----------------------------------------------------------------------------------------------
% endNodes/NodeNames
% -----------------------------------------------------------------------------------------------
predicates
endNodesX : (termSet).
endNodeNamesX : (termSet).
clauses

endNodes()=TS :-
    TS=termSet::new(),
    endNodesX(TS),
    !.

endNodesX(TS) :-
    dg(_,_,_,T2,N2,_),
    not(dg(_,T2,N2,_,_,_)),
    TS:add(cmp("",[T2,N2])),
    fail
;
    !.

endNodeNames()=TS :-
    TS=termSet::new(),
    endNodeNamesX(TS),
    !.

endNodeNamesX(TS) :-
    dg(_,_,_,_,N2,_),
    not(dg(_,_,N2,_,_,_)),
    TS:add(N2),
    fail
;
    !.
% -----------------------------------------------------------------------------------------------
% excludeSet
% -----------------------------------------------------------------------------------------------
excludeSet(Set) :-
    M=Set:get_nd(),
            M=cmp(_,[T,N]),
            deleteNode(T,N),
            fail
;
    !.  
% -----------------------------------------------------------------------------------------------
% includeSet
% -----------------------------------------------------------------------------------------------
includeSet(Set) :-
    C=complementSet(Set),
    excludeSet(C),
    !.  
% -----------------------------------------------------------------------------------------------
% complementSet
% -----------------------------------------------------------------------------------------------
complementSet(Set)=C :-
    C=allNodes(),
    C:subtract(Set),
    !.      
% -----------------------------------------------------------------------------------------------
% serial
% -----------------------------------------------------------------------------------------------
predicates
currentStarters : (termSet,count) determ.
removeStartNode : (sterm,sterm,termSet,count) determ.
clauses

serial()=Set :-
    Set=termSet::new(),
    C=count::newCounter(0),
    currentStarters(Set,C),
    !
;    
    Set=termSet::new(),
    !.
    
currentStarters(Set,C) :-
    not(dg(_,_,_,_,_,_)),
    % finished
    ! 
;    
    dg(_,A1,A2,_,_,_),
        not(dg(_,_,_,A1,A2,_)),
        C:inc(No),
        Ns=toString(No),
        Set:add(cmp(Ns,[A1,A2])),
        removeStartNode(A1,A2,Set,C),
        fail
;
    currentStarters(Set,C),
    !.
    
removeStartNode(T,N,Set,C) :-
    dg(_,T,N,X,Y,_),
        dg(_,X,Y,_,_,_),
        retractall(dg(_,T,N,X,Y,_)),
        fail
;    
    dg(_,T,N,X,Y,_), % X,Y is an end-node
        dg(_,X2,_,X,Y,_),
        not(X2=T),
        retractall(dg(_,T,N,X,Y,_)),
        fail
;    
    dg(_,T,N,X,Y,_), % X,Y is an end-node
        dg(_,_,Y2,X,Y,_),
        not(Y2=N),
        retractall(dg(_,T,N,X,Y,_)),
        fail
;    
    dg(_,T,N,X,Y,_), % X,Y is an end-node
        retractall(dg(_,T,N,X,Y,_)),
        C:inc(No),
        Ns=toString(No),
        Set:add(cmp(Ns,[X,Y])),
        fail
;    
    dg(_,T,N,X,Y,_),
    retractall(dg(_,T,N,X,Y,_)),
    !.
% -----------------------------------------------------------------------------------------------
% cascadeStatusBack
% -----------------------------------------------------------------------------------------------
predicates
cascBack : (sterm,sterm,sterm).
clauses

% cascades back (against the direction of the graph) new status 
%
cascadeStatusBack(NewStat) :-
    markAll(0), % mark as unvisited
    dg(0,A1,A2,B1,B2,NewStat), % all unvisited edges with new status
        retract(dg(0,A1,A2,B1,B2,NewStat)),
        assert(dg(1,A1,A2,B1,B2,NewStat)), % mark as visited
        cascBack(A1,A2,NewStat),
        fail
;
    !.
    
cascBack(B1,B2,NewStat) :-
    dg(0,A1,A2,B1,B2,S), % all unvisited edges with B as succ
        retract(dg(0,A1,A2,B1,B2,S)),
        assert(dg(1,A1,A2,B1,B2,NewStat)), % mark as visited and set new status
        cascBack(A1,A2,NewStat),
        fail
;
    !.
% -----------------------------------------------------------------------------------------------
% cascadeStatusForward
% -----------------------------------------------------------------------------------------------
predicates
cascForward : (sterm,sterm,sterm).
clauses

% cascades forward (with the direction of the graph) new status 
%
cascadeStatusForward(NewStat) :-
    markAll(0), % mark as unvisitied
    dg(0,A1,A2,B1,B2,NewStat), % all unvisited edges with new status
        retract(dg(0,A1,A2,B1,B2,NewStat)),
        assert(dg(1,A1,A2,B1,B2,NewStat)), % mark as visited
        cascForward(B1,B2,NewStat),
        fail
;
    !.
    
cascForward(A1,A2,NewStat) :-
    dg(0,A1,A2,B1,B2,S), % all unvisited edges with A as pred
        retract(dg(0,A1,A2,B1,B2,S)),
        assert(dg(1,A1,A2,B1,B2,NewStat)), % mark as visited
        cascForward(B1,B2,NewStat),
        fail
;
    !.
% -----------------------------------------------------------------------------------------------
%                                                     VISIO DRAWING (NEW)
% -----------------------------------------------------------------------------------------------

% -----------------------------------------------------------------------------------------------
% drwVisio
% -----------------------------------------------------------------------------------------------
predicates
addShape : (string,string,count).
copyHeaders : (inFile,outputStream).
drwGraph : (outputStream,count,visioSymbol_CALLBACK,visioLink_CALLBACK) procedure.
clauses

drwVisio(HeaderFile,OutFile,Note,Scb,Lcb) :-
    IF=inFile::new(HeaderFile,"\n"),
    IF:openInfile,
    OF=outputStream_file::create(OutFile),
    copyHeaders(IF,OF),
    IF:close,
    C=count::newCounter(0),
    retractall(shape(_,_,_)),
    addShape("AN",Note,C),
    drwGraph(OF,C,Scb,Lcb),
    OF:close,
    !.

copyHeaders(IF,OF) :-
    programControl::repeat,
        S=IF:read(),
        OF:write(S),
        IF:fileIsRead,
        !
;
    !.

drwGraph(OF,C,Scb,Lcb) :-
    % for all nodes
    dg(_,Type1,Name1,Type2,Name2,_),
            Scb(Type1,Name1,Symbol1,Title1),
            Scb(Type2,Name2,Symbol2,Title2),
            addShape(Symbol1,Title1,C),
            addShape(Symbol2,Title2,C),
            fail
;
    % for all shapes
    shape(No,Type,Name),
            sx::sxQuote(No,QNo),
            sx::sxQuote(Type,QType),
            sx::sxQuote(Name,QName),
            Shape=string::concat("Shape,",QNo,",",QType,",",QName),
            OF:write(Shape),
            OF:nl,
            fail
;
    % for all edges
    dg(_,Type1,Name1,Type2,Name2,Name3),
            Scb(Type1,Name1,Symbol1,Title1),
            Scb(Type2,Name2,Symbol2,Title2),
            shape(No1,Symbol1,Title1),
            shape(No2,Symbol2,Title2),
            Lcb(Type1,Name1,Type2,Name2,Name3,Title3),
            sx::sxQuote(No1,QNo1),
            sx::sxQuote(No2,QNo2),
            sx::sxQuote(Title3,QTitle3),
            C:inc(N),
            Ns=toString(N),
            Link=string::concatList(["Link,\"N",Ns,"\",\"CN\",",QTitle3,",",QNo1,",",QNo2]),
            OF:write(Link),
            OF:nl,
            fail
;
    !.
    
addShape(Type,Name,C) :-
    shape(_,Type,Name),
    !
;
    C:inc(N),
    Ns=toString(N),
    No=string::concat("N",Ns),
    assert(shape(No,Type,Name)),
    !.

% -----------------------------------------------------------------------------------------------
%                                                     VISIO DRAWING (OLD)
% -----------------------------------------------------------------------------------------------

% -----------------------------------------------------------------------------------------------
% drawGraph
% -----------------------------------------------------------------------------------------------
predicates
getVisioType : (string,sterm,string) procedure (i,i,o).
clauses

drawGraph(OF,C) :-
    dg(_,Type1,Name1,Type2,Name2,_),
    getName(Name1,VName1),
    getVisioType(VName1,Type1,VType1),
    getName(Name2,VName2),
    getVisioType(VName2,Type2,VType2),
    addShape(VType1,VName1,C),
    addShape(VType2,VName2,C),
    fail
;
    shape(Type1,Type2,Name),
    sx::sxQuote(Type1,QType1),
    sx::sxQuote(Type2,QType2),
    sx::sxQuote(Name,QName),
    Shape=string::concat("Shape,",QType1,",",QType2,",",QName),
    OF:write(Shape),
    OF:nl,
    fail
;    
    dg(_,Type1,Name1,Type2,Name2,Name3),
    getName(Name3,VName3),
    sx::sxQuote(VName3,QName3),
    getName(Name1,VName1),
    getVisioType(VName1,Type1,VType1),
    getName(Name2,VName2),
    getVisioType(VName2,Type2,VType2),
    shape(IName1,VType1,VName1),
    shape(IName2,VType2,VName2),
    sx::sxQuote(IName1,QIName1),
    sx::sxQuote(IName2,QIName2),
    C:inc(N),
    Ns=toString(N),
    Link=string::concatList(["Link,\"N",Ns,"\",\"CN\",",QName3,",",QIName1,",",QIName2]),
    OF:write(Link),
    OF:nl,
    fail
;
    !.
    
getVisioType(Name,Type,VType) :-
    L=string::length(Name),
    visio(Min,Max,Type,VType),
    L>=Min, L<=Max,
    !
;
    VType="RE2R",
    !.
% -----------------------------------------------------------------------------------------------
% drawVisio
% -----------------------------------------------------------------------------------------------
drawVisioDetailed(HeaderFile,OutFile) :-
    IF=inFile::new(HeaderFile,"\n"),
    IF:openInfile,
    OF=outputStream_file::create(OutFile),
    copyHeaders(IF,OF),
    IF:close,
    C=count::newCounter(0),
    retractall(shape(_,_,_)),
    drawGraph(OF,C),
    OF:close,
    !.
% -----------------------------------------------------------------------------------------------
% setVisio/resetVisio
% -----------------------------------------------------------------------------------------------
setVisio([],[],[],[]) :-
    !.
setVisio([Min|Mins],[Max|Maxs],[H|T],[Hv|Tv]) :-
    assert(visio(Min,Max,H,Hv)),
    setVisio(Mins,Maxs,T,Tv),
    !.
    
resetVisio() :- retractall(visio(_,_,_,_)),!.    
end implement dirgraph
