/*****************************************************************************

                        Copyright (c) Prolog Development Center A/S

******************************************************************************/

implement editorWindow
open core,vpiDomains,resourceIdentifiers

class facts
lastRect : (integer Left,integer Top).
fileTypes : string_list := []. % list for open/save-file std dialog


constants
        className = "editorWindow/editorWindow".
        classVersion = "".

facts
thisWin : vpiDomains::windowHandle := uncheckedConvert(vpiDomains::windowHandle, 0).
isUnicode : boolean.
fileName : string.

class predicates

clauses
classInfo(className, classVersion).
getFileTypes()=L :- L=fileTypes,!.
setFileTypes(L) :- fileTypes:=L,!.

/* new/open */

        new(Parent, FileName,Wrap):-
            FileName="New",
            file::writeString("New",""),
            ReadOnly = b_false,
            Indent = b_true,
            InitPos = 1,
            isUnicode := true,
            fileName := FileName,
            R=getEditRect(1),
            Menu=taskwindow::getMenu(),
            W=vpiEditor::createWrap(windowType,R,FileName,Menu,Parent,
                 windowFlags,vpi::fontCreate(ff_fixed,[],8), ReadOnly,
                 Indent,Wrap,"",InitPos,eventHandler),
            guix::addWin(0,FileName,W),
            !.
            
        new(Parent, FileName,Wrap):-
            file::existfile(FileName),
            S=file::readString(FileName,Unicode),
            ReadOnly = b_false,
            Indent = b_true,
            InitPos = 1,
            isUnicode := Unicode,
            fileName := FileName,
            R=getEditRect(1),
            Menu=taskwindow::getMenu(),
            W=vpiEditor::createWrap(windowType,R,FileName,Menu,Parent,
                 windowFlags,vpi::fontCreate(ff_fixed,[],8),ReadOnly,
                 Indent,Wrap,S,InitPos,eventHandler),
            guix::addWin(0,FileName,W),
            !.
            
        new(Parent, FileName,Wrap):-
            file::writeString(FileName,""),
            ReadOnly = b_false,
            Indent = b_true,
            InitPos = 1,
            isUnicode := true,
            fileName := FileName,
            R=getEditRect(1),
            Menu=taskwindow::getMenu(),
            W=vpiEditor::createWrap(windowType,R,FileName,Menu,Parent,
                 windowFlags,vpi::fontCreate(ff_fixed,[],8),ReadOnly,
                 Indent,Wrap,"",InitPos,eventHandler),
            guix::addWin(0,FileName,W),
            !.

/* save/save as */

    predicates
        onFileSave : vpiDomains::menuItemHandler.
    clauses
        onFileSave(_MenuTag) = handled(0):-
            Text = vpiEditor::getText(thisWin),
            file::writeString(fileName, Text, isUnicode).

    predicates
        onFileSaveAs : vpiDomains::menuItemHandler.
    clauses
        onFileSaveAs(_MenuTag) = handled(0):-
        Name=guix::getWinName(thisWin),
        Path=fileName::getPath(Name),
        Type=fileName::getExtension(Name),
        Mask=string::concat("*.",Type),
	    File=vpiCommonDialogs::getFileName(Mask,fileTypes,"Save as",[],Path,_),
	    NPath=fileName::getPath(File),
	    directory::existDirectory(NPath),
        S=vpiEditor::getText(thisWin),
        file::writeString(File,S),
        vpi::winSetText(thisWin,File),
        guix::getWinAll(Clas,_,thisWin,SolNo,TabNo,No),
        guix::setWin(Clas,File,thisWin,SolNo,TabNo,No),
        fio::setFileDefaults("E",File),
        !;!.

/* copy/cut/delete/paste/redo/undo/find/replace */

    predicates
        onEditCopy : vpiDomains::menuItemHandler.
    clauses
        onEditCopy(_MenuTag) = handled(0):-
            vpiEditor::copy(thisWin).

    predicates
        onEditCut : vpiDomains::menuItemHandler.
    clauses
        onEditCut(_MenuTag) = handled(0):-
            vpiEditor::cut(thisWin).

    predicates
        onEditDelete : vpiDomains::menuItemHandler.
    clauses
        onEditDelete(_MenuTag) = handled(0):-
            vpiEditor::doDel(thisWin).

    predicates
        onEditPaste : vpiDomains::menuItemHandler.
    clauses
        onEditPaste(_MenuTag) = handled(0):-
            vpiEditor::paste(thisWin).

    predicates
        onEditRedo : vpiDomains::menuItemHandler.
    clauses
        onEditRedo(_MenuTag) = handled(0):-
            vpiEditor::redo(thisWin).

    predicates
        onEditUndo : vpiDomains::menuItemHandler.
    clauses
        onEditUndo(_MenuTag) = handled(0):-
            vpiEditor::undo(thisWin).

    predicates
        onEditFind : vpiDomains::menuItemHandler.
    clauses
        onEditFind(_MenuTag) = handled(0):-
            vpiEditor::searchDialog(thisWin).

    predicates
        onEditFindNext : vpiDomains::menuItemHandler.
    clauses
        onEditFindNext(_MenuTag) = handled(0):-
            vpiEditor::searchAgain(thisWin).

    predicates
        onEditReplace : vpiDomains::menuItemHandler.
    clauses
        onEditReplace(_MenuTag) = handled(0):-
            vpiEditor::replaceDialog(thisWin).

    predicates
        eventHandler : vpiDomains::ehandler.
    clauses
        eventHandler(Win, Event) = Result :-
            Result = generatedEventHandler(Win, Event),
            !.
        eventHandler(_, e_Menu(Id, _)) = _ :-
            vpiEditor::edit_MenuUndo <= Id,
            vpiEditor::edit_MenuFont >=  Id,
            !, 
            fail. %internal editor's events should be kept internally.
        eventHandler(Win, e_Menu(Id, CAS)) = Result :-
            Parent = vpi::winGetParent(Win),
            Result = vpi::winSendEvent(Parent, e_Menu(Id, CAS)).

/* close */

    predicates
        onCloseRequest : vpiDomains::closeRequestHandler.
    clauses
        onCloseRequest() = defaultHandling() :-
            closeEditWin(thisWin),
            !;!.

/* create */

    predicates
        onCreate : vpiDomains::createHandler.
    clauses
        onCreate(_CreationData) = defaultHandling().
        /*  :-
            vpi::menuEnable(thisWin,id_file_save,b_true),
            vpi::menuEnable(thisWin,id_file_save_as,b_true),
            vpi::menuEnable(thisWin,id_edit_copy,b_true),
            vpi::menuEnable(thisWin,id_edit_cut,b_true),
            vpi::menuEnable(thisWin,id_edit_delete,b_true),
            vpi::menuEnable(thisWin,id_edit_paste,b_true),
            vpi::menuEnable(thisWin,id_edit_redo,b_true),
            vpi::menuEnable(thisWin,id_edit_undo,b_true),
            vpi::menuEnable(thisWin,id_edit_find,b_true),
            vpi::menuEnable(thisWin,id_edit_find_next,b_true),
            vpi::menuEnable(thisWin,id_edit_replace,b_true).
            */
    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            thisWin := erroneous.

    predicates
        onSizeChanged : vpiDomains::sizeHandler.
    clauses
        onSizeChanged(_Width, _Height) = defaultHandling().

    % This code is maintained by the VDE. Do not update it manually, 17:15:17-8.5.2003
    constants
        windowType : vpiDomains::wintype = w_TopLevel.
        windowFlags : vpiDomains::wsflags = [wsf_SizeBorder,wsf_TitleBar,wsf_Maximize,wsf_Minimize,wsf_Close,wsf_ClipSiblings,wsf_ClipChildren].
        rectangle : vpiDomains::rct = rct(100,80,752,462).
        menu : vpiDomains::menu = resMenu(id_TaskMenu).
        title : string = "editorWindow".

    predicates
        generatedEventHandler : vpiDomains::ehandler.
    clauses
        generatedEventHandler(Win, e_create(_)) = _ :-
            thisWin := Win,
            fail.
        generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
            handled(Result) = onCreate(CreationData).
        generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
            handled(Result) = onSizeChanged(Width, Height).
        generatedEventHandler(_Win, e_Menu(id_file_save, _)) = Result :-
            handled(Result) = onFileSave(id_file_save).
        generatedEventHandler(_Win, e_Menu(id_file_save_as, _)) = Result :-
            handled(Result) = onFileSaveAs(id_file_save_as).
            /* 
        generatedEventHandler(_Win, e_Menu(id_edit_copy, _)) = Result :-
            handled(Result) = onEditCopy(id_edit_copy).
        generatedEventHandler(_Win, e_Menu(id_edit_cut, _)) = Result :-
            handled(Result) = onEditCut(id_edit_cut).
        generatedEventHandler(_Win, e_Menu(id_edit_delete, _)) = Result :-
            handled(Result) = onEditDelete(id_edit_delete).
        generatedEventHandler(_Win, e_Menu(id_edit_paste, _)) = Result :-
            handled(Result) = onEditPaste(id_edit_paste).
        generatedEventHandler(_Win, e_Menu(id_edit_redo, _)) = Result :-
            handled(Result) = onEditRedo(id_edit_redo).
        generatedEventHandler(_Win, e_Menu(id_edit_undo, _)) = Result :-
            handled(Result) = onEditUndo(id_edit_undo).
        generatedEventHandler(_Win, e_Menu(id_edit_find, _)) = Result :-
            handled(Result) = onEditFind(id_edit_find).
        generatedEventHandler(_Win, e_Menu(id_edit_find_next, _)) = Result :-
            handled(Result) = onEditFindNext(id_edit_find_next).
        generatedEventHandler(_Win, e_Menu(id_edit_replace, _)) = Result :-
            handled(Result) = onEditReplace(id_edit_replace).
            */
        generatedEventHandler(_Win, e_CloseRequest()) = Result :-
            handled(Result) = onCloseRequest().
        generatedEventHandler(_Win, e_Destroy()) = Result :-
            handled(Result) = onDestroy().
    % end of automatic code
% ----------------------------------------------------------------------------------------------
% getMax
% ----------------------------------------------------------------------------------------------
getMax(X,Y) :-
    W=guix::getWinOfClas(1),
    R=vpi::winGetOuterRect(W),
    R=vpiDomains::rct(_,_,Mx,My),
    X=Mx-20,
    Y=My-150,
    !.
% ----------------------------------------------------------------------------------------------
% initEditRect/getEditRect
% ----------------------------------------------------------------------------------------------
class predicates
adjustX : (integer,integer,integer,integer) procedure (i,i,i,o).
adjustY : (integer,integer,integer,integer) procedure (i,i,i,o).
clauses

setEditRect(X,Y) :- retractall(lastRect(_,_)), assert(lastRect(X,Y)),!.
    
getEditRect(Cascade)=EditRect :-
    Cascade=1,
    W=guix::getWinOfClas(1),
    R=vpi::winGetOuterRect(W),
    R=vpiDomains::rct(_,_,Mx,My),
    lastRect(Xr,Yr),
    X=Xr+20,
    Y=Yr+20,
    Wd=Mx div 2,
    H=My div 2,
    adjustX(X,Wd,Mx,Xa),
    adjustY(Y,H,My,Ya),
    X2=Xa+Wd,
    Y2=Ya+H,
    EditRect=vpiDomains::rct(Xa,Ya,X2,Y2),
    setEditRect(Xa,Ya),
    !
;    
    Cascade=-1,
    W=guix::getWinOfClas(1),
    R=vpi::winGetOuterRect(W),
    R=vpiDomains::rct(_,_,Mx,My),
    lastRect(X,Y),
    Wd=Mx div 2,
    H=My div 2,
    adjustX(X,Wd,Mx,Xa),
    adjustY(Y,H,My,Ya),
    X2=Xa+Wd,
    Y2=Ya+H,
    EditRect=vpiDomains::rct(Xa,Ya,X2,Y2),
    setEditRect(Xa,Ya),
    !.
getEditRect(_)=vpiDomains::rct(10,30,110,130) :- !.
    
adjustX(X,W,Mx,Xr) :-
    Xc=X+W div 2,
    Xc<=Mx,
    Xr=X,
    !
;
    Xr=0,
    !.

adjustY(Y,H,My,Yr) :-
    Yc=Y+H div 2,
    Yc<=My,
    Yr=Y,
    !
;
    Yr=0,
    !.
% -------------------------------------------------------------------------------------------------------------
% closeEditWins
% -------------------------------------------------------------------------------------------------------------
class predicates
closeWin : (string,vpiDomains::windowHandle,integer). 
clauses

closeEditWins() :-
    guix::getWinAll(0,Name,W,_,_,_),
    closeEditWin(Name,W),
    fail
;
    !.

closeEditWin(W) :-
            Name=guix::getWinName(W),
            closeEditWin(Name,W),
            !.
closeEditWin(Name,W) :-
            file::existfile(Name),
            Se=vpiEditor::getText(W),
            Sf=file::readString(Name,_),
            Sf=Se,
            guix::dropWin(W),
            !
;
            Name="New",
            P=globalVars::get_s("std_path"),
            File=vpiCommonDialogs::getFileName("*.*",
            ["XML-files","*.xml",
	         "NL-files","*.3nl",
	         "PR-files","*.3pr"],
	         "Save-file",[],P,_),
            closeWin(File,W,vpiCommonDialogs::resp_default),
            !
;
            Name<>"New",
            Mess=string::concat("Do you want to save ",Name,"?"),
            Reply=vpiCommonDialogs::ask("Text is new or has been changed",Mess,["Yes","No","Cancel"]),
            closeWin(Name,W,Reply),
            !.
    
closeWin(Name,W,Reply) :-
            Reply=vpiCommonDialogs::resp_default,
            S=vpiEditor::getText(W),
            file::writeString(Name,S),
            guix::dropWin(W),
            !
;            
            Reply=vpiCommonDialogs::resp_2,
            guix::dropWin(W),
            !
;            
            !.
% ------------------------------------------------------------------------------------------------------------
% getEditorFileDefaults
% ------------------------------------------------------------------------------------------------------------
class predicates
fileDef : (string Mask,string_list Types) procedure (o,o).
clauses

getEditorFileDefaults(Mask,List,Path) :-
  	LP=globalVars::get_s("last_path"),
  	LP="",
  	Path=globalVars::get_s("std_path"),
  	P=string::concat(Path,"*.*"),
  	file::existfile(P),
	fileDef(Mask,List),
	!
;
  	Path=globalVars::get_s("last_path"),
  	P=string::concat(Path,"*.*"),
  	file::existfile(P),
	fileDef(Mask,List),
	!
;
	Path="",
	fileDef(Mask,List),
	!.

fileDef(Mask,List) :-
	L=editorWindow::getFileTypes(),
    Lr=fio::getRecent("E"),
    Lr=[H|_],
    T=filename::getExtension(H),
  	Mask=string::concat("*.",T),
  	set::setGet(L,Ix,Mask),
	Ix2=Ix-1,
	set::setGet(L,Ix2,Text),
	set::setRemSub([Text,Mask],L,L2),
	List=[Text,Mask|L2],
	!
;
	List=editorWindow::getFileTypes(),
	Mask="*.3nl",
	!.
% -------------------------------------------------------------------------------------------------------------
% openFiles
% -------------------------------------------------------------------------------------------------------------
openFiles(L) :- 
    Wrap=globalVars::get_i("f_wrap"),
    openFiles(Wrap,L),
    !.

openFiles(_,[]) :- 
            !.
openFiles(Wrap,[Filename]) :-
  	        fio::setFileDefaults("E",Filename),
  	        openEditor(Wrap,Filename),
            !.
openFiles(Wrap,[Filename|Filenames]) :-
            openEditor(Wrap,FileName),
            openFiles(Filenames),
            !.
% -------------------------------------------------------------------------------------------------------------
% openEditor
% -------------------------------------------------------------------------------------------------------------
openEditor(Wrap,Name) :-
/*
    fio::getSize(Name)>2048.0,
    Mess=string::concat("File ",Name," is too big for the editor!"), 
    vpiCommonDialogs::note(Mess),
    !
;
*/
  	W=guix::getWinOfClas(1),
    set::setGet([-1,0,1],Ix,Wrap),
    set::setGet([0,0,1],Ix,Wrap2),
     _=editorWindow::new(W,Name,Wrap2),
     !.
end implement editorWindow
