/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement fileSeg
open core

constants
className = "com/rudstrand/fileSeg/fileSeg".
classVersion = "1.0".

facts
ifile : inputStream := erroneous.
ofile : outputStream := erroneous.
fileIn : string := "".
fileOut : string := "".
fileType : integer :=0. % 1=single delimiter; 2=double delimiter; 3=in/out stripping
include : integer :=0. % 1=include delimiter in returned segment; -1=do not include
delimStart : string := "". % opening delimiter
delimEnd : string := "". % closing delimiter
delim : string := "". % delimiter
buffer : string := "". % unprocessed part of data read from filef

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% openWrite (constructor)
% ----------------------------------------------------------------------------------------------
openWrite(FileIn,FileOut,DelStart,DelEnd,Include) :-
    fileIn:=FileIn,
    fileOut:=FileOut,
    delimStart:=DelStart,
    delimEnd:=DelEnd,
    include:=Include,
    buffer:="",
    fileType:=3,
    !.
% ----------------------------------------------------------------------------------------------
% openRead (constructor)
% ----------------------------------------------------------------------------------------------
openRead(File,DelStart,DelEnd,Include) :-
    fileIn:=File,
    delimStart:=DelStart,
    delimEnd:=DelEnd,
    include:=Include,
    buffer:="",
    fileType:=2,
    !.

openRead(File,Delim,Include) :-
    fileIn:=File,
    delim:=Delim,
    include:=Include,
    buffer:="",
    fileType:=1,
    !.
% ----------------------------------------------------------------------------------------------
% openReadFile
% ----------------------------------------------------------------------------------------------
openReadFile() :-
    fio::isUnicode(fileIn),
    ifile:=inputStream_file::openFile(fileIn),
    !
; 
    not(fio::isUnicode(fileIn)),
    ifile:=inputStream_file::openFile8(fileIn),
    !.
% ----------------------------------------------------------------------------------------------
% openWriteFile
% ----------------------------------------------------------------------------------------------
openWriteFile() :-
    fio::isUnicode(fileIn),
    ifile:=inputStream_file::openFile(fileIn),
    ofile:=outputStream_file::create(fileOut),
    !
;
    not(fio::isUnicode(fileIn)),
    ifile:=inputStream_file::openFile8(fileIn),
    ofile:=outputStream_file::create8(fileOut),
    !.
% ----------------------------------------------------------------------------------------------
% close
% ----------------------------------------------------------------------------------------------
close() :- 
    fileType=3,
    ifile:close(),
    ofile:close(),
    !
;    
    ifile:close(),
    !.
% ----------------------------------------------------------------------------------------------
% fileEnd
% ----------------------------------------------------------------------------------------------
fileEnd() :- buffer="", ifile:endOfStream().
% ----------------------------------------------------------------------------------------------
% stripInv
% ----------------------------------------------------------------------------------------------
predicates
findEnd : (string) determ.
findStart : (string) determ.
stripFront : (integer,integer,string,string,string,string) determ (i,i,i,i,o,o).
stripToEnd : (integer,integer,string,string,string) determ (i,i,i,i,o).
clauses

stripInv() :-
    ifile:endOfStream(),
    !
;    
    Buffer=ifile:readString(10240),
    findStart(Buffer),
    !.
    
findStart(Buffer) :-
    P=string::search(Buffer,delimStart,string::caseSensitive()),
    stripFront(include,P,delimStart,Buffer,Front,Rest),
    ofile:write(Front),
    findEnd(Rest),
    !
;    
    % no start
    ofile:write(Buffer),
    stripInv,
    !.
    
findEnd(Buffer) :-
    P=string::search(Buffer,delimEnd,string::caseSensitive()),
    stripToEnd(include,P,delimEnd,Buffer,Rest),
    findStart(Rest),
    !
;
    ifile:endOfStream(),
    !
;    
    NBuffer=ifile:readString(10240),
    findEnd(NBuffer),
    !.

stripFront(Include,P,Delim,Buffer,Front,Rest) :-
    Include=-1, % delimiter not included in stripped away part
    L=string::length(Delim),
    Q=P+L,
    sx::frontstr(Q,Buffer,Front,Rest),
    !
;
    Include=1,
    P=0,
    Front="",
    Rest=Buffer,
    !
;
    Include=1,
    sx::frontstr(P,Buffer,Front,Rest),
    !.

stripToEnd(Include,P,Delim,Buffer,Rest) :-
    Include=-1,
    P=0,
    Rest=Buffer,
    !
;
    Include=-1,
    sx::frontstr(P,Buffer,_,Rest),
    !
;
    Include=1,
    L=string::length(Delim),
    Q=P+L,
    sx::frontstr(Q,Buffer,_,Rest),
    !.
% ----------------------------------------------------------------------------------------------
% readNext
% ----------------------------------------------------------------------------------------------
predicates
getSeg : (string,integer,string,string,string) determ (i,i,i,o,o).
getSeg : (string,integer,string,string,string,string) determ (i,i,i,i,o,o).
readToDelim : (string,string,string) determ (i,i,o).
clauses

readNext()=S:-
    fileType=2,
    R=buffer,
    De=delimEnd,
    readToDelim(R,De,T),
    Inc=include,
    Ds=delimStart,
    getSeg(T,Inc,Ds,De,S,Rest),
    buffer:=Rest,    
    !
;
    fileType=2,
    not(ifile:endOfStream()),
    R=buffer,
    Ds=delimStart,
    readToDelim(R,Ds,T),
    P=string::search(T,Ds,string::caseSensitive()),
    sx::frontstr(P,T,_,Rest),
    buffer:=Rest,
    S=readNext(),
    !
;
    fileType=2,
    S="",
    buffer:="",
    !
;
    fileType=1,
    R=buffer,
    D=delim,
    readToDelim(R,D,T),
    Inc=include,
    getSeg(T,Inc,D,S,Rest),
    buffer:=Rest,
    !
;
    fileType=1,
    S="",
    buffer:="",
    !.

readToDelim(R,De,T) :-
    _=string::search(R,De,string::caseSensitive()),
    T=R,
    !
;
    not(ifile:endOfStream()),
    S=ifile:readString(1024),
    Rn=string::concat(R,S),
    readToDelim(Rn,De,T),
    !.

getSeg(S,Inc,D,Seg,R) :-
    Inc=1,
    P=string::search(S,D,string::caseSensitive()),
    L=P+string::length(D),
    sx::frontstr(L,S,Seg,R),
    !
;    
    Inc=-1,
    P=string::search(S,D,string::caseSensitive()),
    P>0,
    Seg=string::subString(S,0,P),
    L=P+string::length(D),
    sx::frontstr(L,S,_,R),
    !
;    
    Inc=-1,
    P=string::search(S,D,string::caseSensitive()),
    P=0,
    Seg="",
    L=string::length(D),
    sx::frontstr(L,S,_,R),
    !.

getSeg(S,Inc,Ds,De,Seg,R) :-
    Inc=1,
    Ps=string::search(S,Ds,string::caseSensitive()),
    Pe=string::search(S,De,string::caseSensitive()),
    L=Pe-Ps+string::length(De),
    L>0,
    Seg=string::subString(S,Ps,L),
    Lf=Ps+L,
    sx::frontstr(Lf,S,_,R),
    !
;    
    Inc=-1,
    Ps=string::search(S,Ds,string::caseSensitive())+string::length(Ds),
    Pe=string::search(S,De,string::caseSensitive()),
    L=Pe-Ps,
    L>0,
    Seg=string::subString(S,Ps,L),
    Lf=Ps+L,
    sx::frontstr(Lf,S,_,R),
    !
;    
    Inc=-1,
    Ps=string::search(S,Ds,string::caseSensitive())+string::length(Ds),
    Pe=string::search(S,De,string::caseSensitive()),
    L=Pe-Ps,
    L=0,
    Seg="",
    Lf=Ps+L,
    sx::frontstr(Lf,S,_,R),
    !.
end implement fileSeg
