/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement fio
open core

constants
className = "com/rudstrand/fio/fio".
classVersion = "1.0".

class facts
recentFiles : (string Clas,string_list Files). % recently opened files 
maxRecent : integer :=20. 

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% getRecent
% ----------------------------------------------------------------------------------------------
class predicates
cleanList : (string_list,string_list) determ (i,o).
clauses

getRecent(Clas)=L :-
    not(recentFiles(Clas,_)),
    L=[],
    !
;    
    recentFiles(Clas,L2), 
    cleanList(L2,L),
    !.

addRecent(Clas,File) :-
    L2=getRecent(Clas),
    set::setRemSub([File],L2,L3),
    set::trunc(maxRecent,[File|L3],L4),
    retractall(recentFiles(Clas,_)),
    assert(recentFiles(Clas,L4)),
    !.

addRecentList(Clas,Files) :-
    L2=getRecent(Clas),
    set::setRemSub(Files,L2,L3),
    set::conc([Files,L3],L4),
    set::trunc(maxRecent,L4,L5),
    retractall(recentFiles(Clas,_)),
    assert(recentFiles(Clas,L5)),
    !.

cleanList([],[]) :-
    !.
cleanList([H|T],L) :-
    E=filename::getExtension(H),
    string::toLowercase(E)="sql",
    cleanList(T,Tc),
    L=[H|Tc],
    !
;
    file::existfile(H),
    cleanList(T,Tc),
    L=[H|Tc],
    !
;
    cleanList(T,L),
    !.
% ------------------------------------------------------------------------------------------------------------
% setFileDefaults
% ------------------------------------------------------------------------------------------------------------
setFileDefaults(Clas,Filename) :-
    file_name(Filename,P,_,_),
    globalVars::set_s("last_path",P),
    addRecent(Clas,FileName),
    !.
% ----------------------------------------------------------------------------------------------
% path
% ----------------------------------------------------------------------------------------------
path("")=P :- P="",!.
path(S)=P :- sx::sxGiveTail(S,"\\",P),!.
% ----------------------------------------------------------------------------------------------
% file_name
% ----------------------------------------------------------------------------------------------
file_name(FileName,Path,Name,Type) :-
	bound(FileName), free(Path), free(Name), free(Type),
	Path=fileName::getPath(FileName),
	Name=fileName::getName(FileName),
	Type=fileName::getExtension(FileName),
	!
;
	bound(Path), bound(Name), bound(Type), free(FileName),
	FileName=string::concat(Path,Name,".",Type),
	!.
% -------------------------------------------------------------------------------------------------------------
% getSize
% -------------------------------------------------------------------------------------------------------------
getSize(Name)=Kb :-
    file::getFileProperties(Name,_,Size,_,_,_),
    Size=unsigned64(L,H),
    Kb=L/1024.0+H*((math::power(2,32)-1)/1024.0),
    !.
% ----------------------------------------------------------------------------------------------
% filCopy
% ----------------------------------------------------------------------------------------------
filCopy(F,G) :-
	file::existfile(F),
	file::copy(F,G),
	!
;
	!.
% ----------------------------------------------------------------------------------------------
% isUnicode
% ----------------------------------------------------------------------------------------------
class predicates
isUni : (string) determ.
clauses

isUnicode(File) :-
    Ext=filename::getExtension(File),
    Extl=string::toLowercase(Ext),
    isUni(Extl),
    !.
    
isUni("3nl") :- globalVars::get_i("uc_3nl")=1,!.    
isUni("3pr") :- globalVars::get_i("uc_3pr")=1,!.    
isUni("3dl") :- globalVars::get_i("uc_3dl")=1,!.    
isUni("xml") :- globalVars::get_i("uc_xml")=1,!.
isUni(Type) :- not(set::member(Type,["3nl","3pr","3dl","xml"])),!.    
% ----------------------------------------------------------------------------------------------
% readFile
% ----------------------------------------------------------------------------------------------
class predicates
readRecs : (inputStream,string_list) procedure (i,o).
clauses

readFile(File)=Recs :-
    not(isUnicode(File)), 
    S=inputStream_file::openFile8(File),
    readRecs(S,Recs),
    S:close(),
    !
;
    isUnicode(File),
    S=inputStream_file::openFile(File),
    readRecs(S,Recs),
    S:close(),
    !.
    
readFile(Unicode,File)=Recs :-
    Unicode=1,
    S=inputStream_file::openFile(File),
    readRecs(S,Recs),
    S:close(),
    !
;    
    S=inputStream_file::openFile8(File),
    readRecs(S,Recs),
    S:close(),
    !.
    
readRecs(S,[]) :-    
     S:endOfStream(),
     !.
readRecs(S,[H|T]) :-    
    H=S:readLine(),
    readRecs(S,T),
    !.
% ----------------------------------------------------------------------------------------------
% writeFile
% ----------------------------------------------------------------------------------------------
class predicates
writeRecs : (integer AddLFs,outputStream,string_list) determ.
clauses

writeFile(AddLFs,File,Type,Recs) :-
    Type=8,
    S=outputStream_file::create8(File),
    writeRecs(AddLFs,S,Recs),
    S:close(),
    !
;
    Type=16,
    S=outputStream_file::create(File),
    writeRecs(AddLFs,S,Recs),
    S:close(),
    !.
    
writeRecs(_,_,[]) :-    
     !.
writeRecs(-1,S,[H|T]) :-    
    S:write(H),
    writeRecs(-1,S,T),
    !.
writeRecs(1,S,[H|T]) :-    
    S:write(H),
    S:nl(),
    writeRecs(1,S,T),
    !.
% ------------------------------------------------------------------------------------------------------------
% getProcFileDefaults
% ------------------------------------------------------------------------------------------------------------
class predicates
fileDef2P : (string Mask,string_list Types) procedure (o,o).
clauses

getProcFileDefaults(Mask,List,Path) :-
  	LP=globalVars::get_s("last_path"),
  	LP="",
  	Path=globalVars::get_s("std_path"),
  	file::existfile(Path),
	fileDef2P(Mask,List),
	!
;
  	Path=globalVars::get_s("last_path"),
  	file::existfile(Path),
	fileDef2P(Mask,List),
	!
;
	Path="",
	fileDef2P(Mask,List),
	!.

fileDef2P(Mask,List) :-
	List=["XML-files","*.xml",
	   "Delimited files","*.3dl",
	   "NL-files","*.3nl",
	   "PR-files","*.3pr"],
	Mask="*.xml",
	!.
end implement fio
