/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement globalVars
open core,termDomain,resourceIdentifiers,globalDomains

class facts
iV : (string,integer) nondeterm.
nV : (string,real) nondeterm.
sV : (string,string) nondeterm.
tV : (string,sterm) nondeterm.
retPars : string_list_list := []. % a stack of string lists where each list is one or more parameters returned from a dialog.

constants
className = "com/rudstrand/globalVars/globalVars".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------	
% popRetPars/pushRetPars
% -------------------------------------------------------------------------------------	
pushRetPars(L) :-
    T=retPars,
    U=[L|T],
    retPars:=U,
    !.
    
popRetPars()=L :-
    T=retPars,
    T=[L|U],
    retPars:=U,
    !
;
    pfeil::syserr("popRetPars","No parameters on stack!"),
    L=[],
    !.
% -------------------------------------------------------------------------------------	
% get_i/_s/_n
% -------------------------------------------------------------------------------------	
get_i(Var) = Val :-
                    profileTime::start("get_i"),
    iV(Var,Val),
                    profileTime::stop("get_i"),
    !
;
    pfeil::syserr("get_i","Unknown integer global variable",[Var]),!,fail.
    
get_s(Var) = Val :-
                    profileTime::start("get_s"),
    sV(Var,Val),
                    profileTime::stop("get_s"),
    !
;
    pfeil::syserr("get_s","Unknown string global variable",[Var]),!,fail.

get_n(Var) = Val :-
    nV(Var,Val),
    !
;
    pfeil::syserr("get_n","Unknown real global variable",[Var]),!,fail.

get_t(Var) = Val :-
    tV(Var,Val),
    !
;
    pfeil::syserr("get_t","Unknown sterm global variable",[Var]),!,fail.
    
getc_i(Def,Var)=V :-
    iV(Var,V),
    !
;
    V=Def,
    !.

getc_s(Def,Var)=V :-
    sV(Var,V),
    !
;
    V=Def,
    !.
% -------------------------------------------------------------------------------------	
% set_i/_s/_n
% -------------------------------------------------------------------------------------	
set_i(Var,Val) :-
    retractall(iV(Var,_)),
    assert(iV(Var,Val)),
    !.
set_n(Var,Val) :-
    retractall(nV(Var,_)),
    assert(nV(Var,Val)),
    !.
set_s(Var,Val) :-
    retractall(sV(Var,_)),
    assert(sV(Var,Val)),
    !.
set_t(Var,Val) :-
    retractall(tV(Var,_)),
    assert(tV(Var,Val)),
    !.
% -----------------------------------------------------------------------------------------------
% incr; decr; add
% -----------------------------------------------------------------------------------------------
incr(Name,N,Ni) :-
    N=get_i(Name),
	Ni=N+1,
	set_i(Name,Ni),
	!.

decr(Name,N,Ni) :-
    N=get_i(Name),
	Ni=N-1,
	set_i(Name,Ni),
	!.
	
add(Name,Add,After) :-
       N=get_i(Name),
       After=N+Add,
       set_i(Name,After),
       !.
% -----------------------------------------------------------------------------------------------
% set_max/min
% -----------------------------------------------------------------------------------------------
setMax_i(Name,V) :-
	get_i(Name)>=V,
	!
;
	set_i(Name,V),
	!.	

setMin_i(Name,V) :-
	get_i(Name)<=V,
	!
;
	set_i(Name,V),
	!.	
% -----------------------------------------------------------------------------------------------
% iVp/sVp
% -----------------------------------------------------------------------------------------------
iVp(Var,Val) :- iV(Var,Val).
sVp(Var,Val) :- sV(Var,Val).
nVp(Var,Val) :- nV(Var,Val).
% -------------------------------------------------------------------------------------	
% drop_t
% -------------------------------------------------------------------------------------	
drop_t(V) :- retractall(tV(V,_)),!.
% -------------------------------------------------------------------------------------	
% save
% -------------------------------------------------------------------------------------	
class predicates
saveX : (lst,string_list).
clauses

save(File,Includes) :-
	List=lst::new(),
	saveX(List,Includes),
	L=List:get(),
	sx::sxlJoin(L,"\n",S),
	file::writeString(File,S),
	!
;
    !.
		
saveX(List,[]) :-
	globalVars::iVp(Var,Val),
	Vs=toString(Val),
    S=string::concat(Var,"|i|",Vs),
	List:add(S),
	fail
;	
	globalVars::nVp(Var,Val),
	Vs=toString(Val),
    S=string::concat(Var,"|n|",Vs),
	List:add(S),
	fail
;	
	globalVars::sVp(Var,Val),
    S=string::concat(Var,"|s|",Val),
	List:add(S),
	fail
;
    !.

saveX(List,Includes) :-
	globalVars::iVp(Var,Val),
	set::member(Var,Includes),
	Vs=toString(Val),
    S=string::concat(Var,"|i|",Vs),
	List:add(S),
	fail
;	
	globalVars::nVp(Var,Val),
	set::member(Var,Includes),
	Vs=toString(Val),
    S=string::concat(Var,"|n|",Vs),
	List:add(S),
	fail
;	
	globalVars::sVp(Var,Val),
	set::member(Var,Includes),
    S=string::concat(Var,"|s|",Val),
	List:add(S),
	fail
;
    !.
% -------------------------------------------------------------------------------------	
% load
% -------------------------------------------------------------------------------------	
class predicates
loadX : (string_list,string_list) determ.
clauses

load(File,Includes) :-
	file::existfile(File),
	S=file::readString(File,_),
	sx::sxUnpack(S,'\n',L),
	loadX(L,Includes),
	!
;
	!.
	
loadX([],_) :-
    !.
loadX([H|T],[]) :-
    sx::sxUnpack(H,'|',L),
    L=[V,Typ,Val],
    Typ="i",
    IVal=sx::toInt(Val),
    globalVars::set_i(V,IVal),
    loadX(T,[]),
    !
;    
    sx::sxUnpack(H,'|',L),
    L=[V,Typ,Val],
    Typ="n",
    IVal=sx::toReal(Val),
    globalVars::set_n(V,IVal),
    loadX(T,[]),
    !
;    
    sx::sxUnpack(H,'|',L),
    L=[V,Typ,Val],
    Typ="s",
    globalVars::set_s(V,Val),
    loadX(T,[]),
    !
;    
    loadX(T,[]),
    !.
loadX([H|T],Includes) :-
    sx::sxUnpack(H,'|',L),
    L=[V,Typ,Val],
    set::member(V,Includes),
    Typ="i",
    IVal=sx::toInt(Val),
    globalVars::set_i(V,IVal),
    loadX(T,Includes),
    !
;    
    sx::sxUnpack(H,'|',L),
    L=[V,Typ,Val],
    set::member(V,Includes),
    Typ="n",
    IVal=sx::toReal(Val),
    globalVars::set_n(V,IVal),
    loadX(T,Includes),
    !
;    
    sx::sxUnpack(H,'|',L),
    L=[V,Typ,Val],
    set::member(V,Includes),
    Typ="s",
    globalVars::set_s(V,Val),
    loadX(T,Includes),
    !
;    
    loadX(T,Includes),
    !.
end implement globalVars
