/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement guix
open core

class facts
win_db : (integer WinClass,string Name,vpiDomains::windowHandle W,integer SolNo,integer TabNo,integer WinNo) nondeterm.  
% WinClass: see hlpWebX
% WinNo is a unique number assigned sequentially - sometimes used to link related windows : see rub::rubTW_Get; rubEW_Get
winCount : integer := 0.
deltax : integer :=100.
deltay : integer :=100.

constants
className = "com/rudstrand/guix/guix".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% -----------------------------------------------------------------------------------------------
% deaultText
% -----------------------------------------------------------------------------------------------
defaultText("",D)=V:-V=D,!.
defaultText(A,_)=V:-V=A,!.
% ----------------------------------------------------------------------------------------------
% addWin
% ----------------------------------------------------------------------------------------------
addWin(Clas,Win) :-
    N=winCount+1,
    winCount:=N,
    assert(win_db(Clas,"",Win,0,0,N)),
    !.
addWin(Clas,Name,Win) :-
    N=winCount+1,
    winCount:=N,
    assert(win_db(Clas,Name,Win,0,0,N)),
    !.
addWin(Clas,Name,Win,SolNo,TabNo) :-
    N=winCount+1,
    winCount:=N,
    assert(win_db(Clas,Name,Win,SolNo,TabNo,N)),
    !.
% ----------------------------------------------------------------------------------------------
% setWin
% ----------------------------------------------------------------------------------------------
setWin(Clas,Name,W,SolNo,TabNo,No) :- 
    win_db(_,_,W,_,_,_),
    retractall(win_db(_,_,W,_,_,_)),
    assert(win_db(Clas,Name,W,SolNo,TabNo,No)),
    !.
% ----------------------------------------------------------------------------------------------
% succ/pred
% ----------------------------------------------------------------------------------------------
succ(W)=Ws :- getWin(W,N), M=N+1, getWin(Ws,M),!.
pred(W)=Wp :- getWin(W,N), M=N-1, getWin(Wp,M),!.
% ----------------------------------------------------------------------------------------------
% existWin
% ----------------------------------------------------------------------------------------------
existWin(Name) :- win_db(_,Name,_,_,_,_),!.
% ----------------------------------------------------------------------------------------------
% getCount
% ----------------------------------------------------------------------------------------------
getCount()=N:-N=winCount,!.
% ----------------------------------------------------------------------------------------------
% getWin;getWinType; getWinAll
% ----------------------------------------------------------------------------------------------
getWinOfClas(Clas)=W :- win_db(Clas,_,W,_,_,_).
getWinOfName(Name)=W :- win_db(_,Name,W,_,_,_).
getWinAll(Clas,Name,W,SolNo,TabNo,No) :- win_db(Clas,Name,W,SolNo,TabNo,No).
getWin(W,N) :- win_db (_,_,W,_,_,N),!.
getWinClas(W)=Clas :- win_db(Clas,_,W,_,_,_),!.
getWinName(W)=File :-win_db(_,File,W,_,_,_),!.
getWinNo(W,Clas,SolNo,TabNo) :- win_db(Clas,_,W,SolNo,TabNo,_),!.
% ----------------------------------------------------------------------------------------------
% dropWin
% ----------------------------------------------------------------------------------------------
dropWin(Win) :-
    retractall(win_db(_,_,Win,_,_,_)),
    !.
% ----------------------------------------------------------------------------------------------
% setPixelAdjust
% ----------------------------------------------------------------------------------------------
setPixelAdjust(X,Y) :-
    deltax:=X,
    deltay:=Y.

getPixelAdjust(X,Y) :-
    X=deltax,
    Y=deltay.
% ----------------------------------------------------------------------------------------------
% pixelsLC/UC/Digits
% ----------------------------------------------------------------------------------------------
pixels(Clas,S)=P :-
    W=getWinOfClas(Clas),
    vpi::winGetTextExtent(W,S,-1,Px,_),
    P=(Px*deltax) div 100,
    !.
pixelsLC(Clas,Width)=P :-
    W=getWinOfClas(Clas),
    vpi::winGetTextExtent(W,"abcdefghijklmnopqrstuvwxyz",-1,Wd,_),
    P=(((Wd*Width) div 20)*deltax) div 100, 
    !.
pixelsUC(Clas,Width)=P :-
    W=getWinOfClas(Clas),
    vpi::winGetTextExtent(W,"ABCDEFGHIJKLMNOPQRSTUVWXYZ",-1,Wd,_),
    P=(((Wd*Width) div 20)*deltax) div 100,
    !.
pixelsDigits(Clas,Width)=P :-
    W=getWinOfClas(Clas),
    vpi::winGetTextExtent(W,"0123456789",-1,Wd,_),
    P=(((Wd*Width) div 8)*deltax) div 100,
    !.
% ----------------------------------------------------------------------------------------------
% lbxDelSet/DelSetS/DelIs/DelS
% ----------------------------------------------------------------------------------------------
class predicates
lbxDelX : (vpiDomains::windowHandle,integer_list) determ.
lbxDelSetX : (vpiDomains::windowHandle,integer_list) determ.
clauses
lbxDelSet(Win,CtlNo,L) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	set::setRev(L,Lr),
	lbxDelSetX(Ctl,Lr),
	!.
lbxDelSetS(_,_,[]) :-
    !.
lbxDelSetS(Win,CtlNo,[S|L]) :-
    lbxDelS(Win,CtlNo,S),
    lbxDelSetS(Win,CtlNo,L),
    !.
    
lbxDelS(Win,CtlNo,S) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	All=vpi::lboxGetAll(Ctl),
	set::setGet(All,No,S),
	Del=No-1,
	vpi::lboxDelete(Ctl,Del),
	!.
	
lbxDelIs(Win,CtlNo,Ix) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	lbxDelX(Ctl,Ix),
	!.
	
lbxDelSetX(_,[]) :-
	!.
lbxDelSetX(C,[M|L]) :-
	vpi::lboxDelete(C,M),
	lbxDelSetX(C,L),
	!.
	
lbxDelX(_,[]) :- 
	!.
lbxDelX(Ctl,[Ix|T]) :-
	vpi::lboxDelete(Ctl,Ix),
	lbxDelX(Ctl,T),
	!.
% ----------------------------------------------------------------------------------------------
% lbxPop
% ----------------------------------------------------------------------------------------------
% populates a list box or combo
%
lbxPop(Win,CtlNo,S) :-
	vpi::cursorSetWait(),
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxClear(Ctl),
	vpi::lboxSuspend(Ctl),
	vpi::lboxAdd(Ctl,-1,S),
	vpi::lboxResume(Ctl),	
	vpi::cursorSet(Win,vpiDomains::cursor_Arrow),!.
% ----------------------------------------------------------------------------------------------
% lbxClear
% ----------------------------------------------------------------------------------------------
lbxClear(Win,CtlNo) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxClear(Ctl),
	!
;
	!.
% ----------------------------------------------------------------------------------------------
% lbxAdd/List
% ----------------------------------------------------------------------------------------------
lbxAdd(Win,CtlNo,S) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	L=vpi::lboxGetAll(Ctl),
	set::member(S,L),
	!
;
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxAdd(Ctl,S),
	!.
	
lbxAddList(Win,CtlNo,SL) :-
	lbxGetAll(Win,CtlNo,L),
	set::conc([SL,L],New),
	set::distinct(New,NewD),
	lbxPop(Win,CtlNo,NewD),
	!.
% ----------------------------------------------------------------------------------------------
% lbxGet/Multi/All
% ----------------------------------------------------------------------------------------------
lbxGet(Win,CtlNo,Sel) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxGetSel(Ctl,Sels,_),
	Sels=[Sel],
	!.
	
lbxGetMulti(Win,CtlNo,Sels) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxGetSel(Ctl,Sels,_),
	!.
	
lbxGetAll(Win,CtlNo,L) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	L=vpi::lboxGetAll(Ctl),
	!.
% ----------------------------------------------------------------------------------------------
% lbtGet; lbtGetF; lbtGetB
% ----------------------------------------------------------------------------------------------
% gets the selected item in a list box - gets first item of no selection
%
lbtGet(Win,CtlNo,Selected) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxGetSel(Ctl,Sels,_),
	Sels=[Selected],
	!
;
	txtGet(Win,CtlNo,Selected),
	!
;
	% unnecessary? (12.6.02)
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	Selected=vpi::lboxGetItem(Ctl,0),
	!	
;
	pfeil::syserr("lbtGet","",[],[CtlNo]),!,fail.	
lbtGetF(Win,CtlNo,Selected) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxGetSel(Ctl,Sels,_),
	Sels=[Selected],
	!.
% ----------------------------------------------------------------------------------------------
% lbtSelectFirst
% ----------------------------------------------------------------------------------------------
lbtSelectFirst(Win,CtlNo) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	All=vpi::lboxGetAll(Ctl),
	All=[H|_],
	vpi::lboxSetSel(Ctl,0,b_true),
	txtSet(Win,CtlNo,H),
	!
;
	!.
% ----------------------------------------------------------------------------------------------
% lbtSel
% ----------------------------------------------------------------------------------------------
% sets an item as selected in list button
%
lbtSel(Win,CtlNo,Selected) :-
	lbtSelF(Win,CtlNo,Selected),
	!
;
	pfeil::syserr("lbtSel","",[Selected],[CtlNo]),!,fail.	

lbtSelF(Win,CtlNo,Selected) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	All=vpi::lboxGetAll(Ctl),
	set::setGet(All,N,Selected),
	M=N-1,
	vpi::lboxSetSel(Ctl,M,b_true),
	txtSet(Win,CtlNo,Selected),
	!.
% ----------------------------------------------------------------------------------------------
% lbxSel/MultiSel
% ----------------------------------------------------------------------------------------------
class predicates
lbxSelS : (string_list,integer,string,vpiDomains::windowHandle) determ.
lbxSelL : (string_list,integer,string_list,vpiDomains::windowHandle).
lbxSelC : (string_list,string_list) determ.	
clauses

lbxSel(Win,CtlNo,Selected) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	All=vpi::lboxGetAll(Ctl),
	lbxSelS(All,0,Selected,Ctl),
	!.

% sets given list control entries as selected (deselects the remaining entries)
% Note! use lbt_sel for list buttons
% parameters:
% 1	in	window
% 2	in	listbox control no
% 3	in	selected members in par#2
% 26.8.96
%
lbxMultiSel(Win,CtlNo,Selected) :-
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	All=vpi::lboxGetAll(Ctl),
	lbxSelC(Selected,All),
	lbxSelL(All,0,Selected,Ctl),
	!.

lbxSelL([],_,_,_) :- 
	!.
lbxSelL([H|T],No,Selected,Ctl) :-
	set::member(H,Selected),
	vpi::lboxSetSel(Ctl,No,b_true),
	M=No+1,
	lbxSelL(T,M,Selected,Ctl),
	!
;	
	vpi::lboxSetSel(Ctl,No,b_false),
	M=No+1,
	lbxSelL(T,M,Selected,Ctl),
	!.
	
lbxSelC([],_) :- 
	!.
lbxSelC([H|T],All) :-
	set::member(H,All),
	lbxSelC(T,All),
	!
;
	pfeil::syserr("lbxSel","",[H|All]),
	!,fail.	

lbxSelS([H|T],No,Selected,Ctl) :-
	H=Selected,
	vpi::lboxSetSel(Ctl,No,b_true),
	!
;	
	M=No+1,
	lbxSelS(T,M,Selected,Ctl),
	!.
% ----------------------------------------------------------------------------------------------
% lbxRem/Reselect
% ----------------------------------------------------------------------------------------------
lbxRem(Win,CtlNo) :-
	lbxGetAll(Win,CtlNo,All),
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxGetSel(Ctl,Sels,_),
	set::setRemSub(Sels,All,Left),
	lbxPop(Win,CtlNo,Left),	
	!.
	
lbxRemReselect(Win,CtlNo) :-
	lbxGetAll(Win,CtlNo,All),
	Ctl=vpi::winGetCtlHandle(Win,CtlNo),
	vpi::lboxGetSel(Ctl,Sels,Indexes),
	Indexes=[First|_],
	set::setRemSub(Sels,All,Left),
	lbxPop(Win,CtlNo,Left),	
	set::size(Left,N),
	N>0,
	Select=First-1,
	set::setMax([0,Select],Sel),
	vpi::lboxSetSel(Ctl,Sel,b_True),
    _=vpi::winSendEvent(Win,vpiDomains::e_Control(CtlNo,vpiDomains::wc_LBox,Ctl,vpiDomains::selchanged())),
    !
;
	!.
% ----------------------------------------------------------------------------------------------
% ctlEnable/Disable/SetState
% ----------------------------------------------------------------------------------------------
ctlEnable(Win,Ctls) :- ctlSetState(Win,Ctls,[vpiDomains::wsf_Enabled]),!.
ctlDisable(Win,Ctls) :- ctlSetState(Win,Ctls,[vpiDomains::wsf_Disabled]),!.
ctlVisible(Win,Ctls) :- ctlSetState(Win,Ctls,[vpiDomains::wsf_Visible]),!.
ctlInvisible(Win,Ctls) :- ctlSetState(Win,Ctls,[vpiDomains::wsf_Invisible]),!.
ctlSetState(Win,[C|Cs],States) :-
	Ctl=vpi::winGetCtlHandle(Win,C),
	vpi::winSetState(Ctl,States),
	ctlSetState(Win,Cs,States),
	!.
ctlSetState(_,_,_) :- 
    !.
% ----------------------------------------------------------------------------------------------
% guiCtlWins
% ----------------------------------------------------------------------------------------------
guiCtlWins(_,[],[]) :- 
    !.
guiCtlWins(W,[C|Cs],[Cw|Cws]) :-
	Cw=vpi::winGetCtlHandle(W,C),
	guiCtlWins(W,Cs,Cws),
	!.
% ----------------------------------------------------------------------------------------------
% guiCB_Get/Set
% ----------------------------------------------------------------------------------------------
class predicates
guiOnOff : (integer,integer) determ (i,o) (o,i).
clauses

guiCB_Get(W,CB,Checked) :-
	Ctl=vpi::winGetCtlHandle(W,CB),
	State=vpi::winIsChecked(Ctl),
	guiOnOff(State,Checked),
	!.
guiCB_Set(W,CB,Val) :-
	Ctl=vpi::winGetCtlHandle(W,CB),
	guiOnOff(VVal,Val),
	vpi::winCheck(Ctl,VVal),
	!.
guiOnOff(vpiDomains::checkbox_on,1) :- 
	!.
guiOnOff(vpiDomains::checkbox_off,-1) :- 
	!.
guiOnOff(vpiDomains::checkbox_undeterm,0) :- 
	!.
% ----------------------------------------------------------------------------------------------
% guiRB_Get/Set
% ----------------------------------------------------------------------------------------------
% gets/sets radio button group
%
guiRB_Get(W,RBs,Nos,N) :- 
	guiRB_GetF(W,RBs,Nos,N),
	!
;	
	pfeil::syserr("guiRB_Get",""),!,fail.
	
guiRB_GetF(W,[RB|RBs],[No|Nos],N) :- 
	Ctl=vpi::winGetCtlHandle(W,RB),
	State=vpi::winIsChecked(Ctl),
	State=vpiDomains::checkbox_on,
	N=No,
	!
;	
	guiRB_GetF(W,RBs,Nos,N),
	!.
guiRB_SetN(W,RBs,Nos,N) :-
	set::setGet(Nos,IndexNo,N),
	set::setGet(RBs,IndexNo,RB),
	set::setRem(RBs,IndexNo,RBsOff),
	guiRB_Set(W,[RB|RBsOff]),
	!
;		
	pfeil::syserr("guiRB_SetX",""),!,fail.
guiRB_Set(W,RBs) :- 
	guiCtlWins(W,RBs,CRBs),
	CRBs=[CRB|_],
	vpi::winCheckRadioButton(CRB,CRBs),
	!.
% ----------------------------------------------------------------------------------------------
% setStyle
% ----------------------------------------------------------------------------------------------
class predicates
ctlSetFont : (vpiDomains::windowHandle W,integer_list Ctls,vpiDomains::font) procedure (i,i,i).
clauses
setStyle(W,Ctls,Bolds) :-
    Bolds=[],
  	Font=vpi::fontCreate(vpiDomains::ff_Helvetica,[],8),
  	vpi::winSetFont(W,Font),
  	ctlSetFont(W,Ctls,Font),
  	!		    
;    
  	Font=vpi::fontCreate(vpiDomains::ff_Helvetica,[],8),
  	vpi::winSetFont(W,Font),
  	ctlSetFont(W,Ctls,Font),		    
  	FontBold=vpi::fontCreate(vpiDomains::ff_Helvetica,[vpiDomains::fs_Bold],8),
  	ctlSetFont(W,Bolds,FontBold),
  	!.		    
ctlSetFont(Win,[C|Ctls],Font) :-  	
	W=vpi::winGetCtlHandle(Win,C),
  	vpi::winSetFont(W,Font),
  	ctlSetFont(Win,Ctls,Font),
  	!.
ctlSetFont(_,_,_) :- 
	!.
% ----------------------------------------------------------------------------------------------
% txtGets; txtGet; txtGetI
% ----------------------------------------------------------------------------------------------
% gets control text(s) - string
txtGet(W,C,S) :-
	Ctl=vpi::winGetCtlHandle(W,C),
	S=vpi::winGetText(Ctl),
	!.	
% (signed) integer
txtGetI(W,C,In) :-
	txtGet(W,C,S),
	In=sx::toInt(S),
	!.
txtGetI_M(W,C,M,I) :-
	txtGetI(W,C,I),
	!
;
	txtGet(W,C,S),
	sx::sxStrip(S,St),
	St<>"",
	sx::concat(["Expected an integer (",M,"), but found '",St,"'"],Mess),
	vpiCommonDialogs::note("VALIDATION ERROR",Mess),
	!,fail
;
	sx::concat(["Expected an integer (",M,"), but field is blank"],Mess),
	vpiCommonDialogs::note("VALIDATION ERROR",Mess),
	!,fail.
% (signed) fix
txtGetFx(W,C,F) :-
	txtGet(W,C,S),
	F=sx::toReal(S),
	!.
% (signed) fix w/error message
txtGetFxM(W,C,M,F) :-
	txtGetFx(W,C,F),
	!
;
	txtGet(W,C,S),
	sx::sxStrip(S,St),
	St<>"",
	sx::concat(["Expected a fixed point number (",M,"), but found '",St,"'"],Mess),
	vpiCommonDialogs::note("VALIDATION ERROR",Mess),
	!,fail
;
	sx::concat(["Expected a fixed point number (",M,"), but field is blank"],Mess),
	vpiCommonDialogs::note("VALIDATION ERROR",Mess),
	!,fail.
% string list
txtGets(W,[C|Cs],[S|Ss]) :-
	txtGet(W,C,S),
	txtGets(W,Cs,Ss),
	!.
txtGets(_,_,[]) :- 
    !.
% ----------------------------------------------------------------------------------------------
% txtSet/I
% ----------------------------------------------------------------------------------------------
txtSet(W,C,S) :-
	Ctl=vpi::winGetCtlHandle(W,C),
	vpi::winSetText(Ctl,S),
	!.
txtSetI(W,C,I) :-
	Ctl=vpi::winGetCtlHandle(W,C),
	S=toString(I),
	vpi::winSetText(Ctl,S),
	!.
end implement guix
