/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement inFile
open core

constants
className = "com/rudstrand/inFile/inFile".
classVersion = "1.0".

facts
eor : string := "".
eof : integer := 0.
buffer : string := "".
istream : inputStream := erroneous.
bufferSize : integer := 0.
fileIn : string := "".
fileMode : integer := 16. % 8/16 for ansi/unicode

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% new
% ----------------------------------------------------------------------------------------------
new(FileName,EndOfRec) :- new(FileName,EndOfRec,1024),!.

new(FileName,EndOfRec,BufferSize) :-
    eor:=EndOfRec,
    buffer:="",
    eof:=-1,
    bufferSize:=BufferSize,
    fileIn:=FileName,
    !.
% ----------------------------------------------------------------------------------------------
% openInfile
% ----------------------------------------------------------------------------------------------
openInfile() :-
    fio::isUnicode(fileIn),
    istream:=inputStream_file::openFile(fileIn),
    fileMode:=16,
    !
; 
    not(fio::isUnicode(fileIn)),
    istream:=inputStream_file::openFile8(fileIn),
    fileMode:=8,
    !.    

openInfile(Unicode) :-
    Unicode=1,
    istream:=inputStream_file::openFile(fileIn),
    fileMode:=16,
    !
; 
    istream:=inputStream_file::openFile8(fileIn),
    fileMode:=8,
    !.    
% ----------------------------------------------------------------------------------------------
% mode
% ----------------------------------------------------------------------------------------------
mode()=Mode :- Mode=fileMode,!.
% ----------------------------------------------------------------------------------------------
% close
% ----------------------------------------------------------------------------------------------
close() :-
    istream:close(),
    !.
% ----------------------------------------------------------------------------------------------
% fileIsRead
% ----------------------------------------------------------------------------------------------
fileIsRead() :-
    eof=1,
    buffer="",
    !.
% ----------------------------------------------------------------------------------------------
% read
% ----------------------------------------------------------------------------------------------
predicates
readMore : ().
readNetX : (string,string) procedure (i,o).
clauses

readNet()=Rec :-
    Rec2=read(),
    readNetX(Rec2,Rec),
    !.
    
readNetX(Rec2,Rec) :-    
    Pos=string::search(Rec2,eor,string::caseSensitive()),
    string::front(Rec2,Pos,Rec,_),
    !
;
    Rec=Rec2,
    !.
    
read()=Rec :-
    % blank record in buffer
    P=string::search(buffer,eor,string::caseSensitive()),
    P=0,
    Q=P+string::length(eor),
    string::front(buffer,Q,_,B),
    buffer:=B,
    Rec=eor,
    !
;
    % non-blank record in buffer
    P=string::search(buffer,eor,string::caseSensitive()),
    P>0,
    Q=P+string::length(eor),
    string::front(buffer,Q,Rec,T),
    buffer:=T,
    !
;    
    % eof
    eof=1,
    Rec=buffer,
    buffer:="",
    !
;
    % read more into buffer
    readMore,
    Rec=read(),
    !.
    
readMore() :-
    not(istream:endOfStream()),
    S=istream:readString(bufferSize),
    B=buffer,
    buffer:=string::concat(B,S),
    !
;
    eof:=1,
    !.
end implement inFile
