/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement lst
open core

facts
list : (integer,string) nondeterm.
count : integer :=0.
unique : integer := 0.

constants
className = "com/rudstrand/lst/lst".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

% ----------------------------------------------------------------------------------------------
% new
% ----------------------------------------------------------------------------------------------   
new() :- new(-1).    

new(Unique):-
    Unique=1,
    retractall(list(_,_)),
    count:=0,
    unique:=Unique,
    !
;
    retractall(list(_,_)),
    count:=0,
    unique:=-1,
    !.
% ----------------------------------------------------------------------------------------------
% deleteAll
% ----------------------------------------------------------------------------------------------
deleteAll() :-
    retractall(list(_,_)),
    count:=0,
    !.
% ----------------------------------------------------------------------------------------------
% add/Unique
% ----------------------------------------------------------------------------------------------
predicates
topX : (integer,string,integer) determ (i,o,o).
clauses

add(S) :-
    unique=1,
    addUnique(S),
    !
;
    N=count+1,
    assert(list(N,S)),
    count:=N,
    !.

addUnique(S) :- 
    list(_,S),
    !
;
    N=count+1,
    assert(list(N,S)),
    count:=N,
    !.

addList([]) :-
    !.
addList([H|T]) :-
    add(H),
    addList(T),
    !.
    
top()=Top :- 
    topX(count,Top,P),
    count:=P,
    !.

topX(0,_,_) :- 
    !,fail.
topX(N,Top,P) :-
    list(N,Top),
    P=N,
    !
;
    M=N-1,
    topX(M,Top,P),
    !.
% ----------------------------------------------------------------------------------------------
% delete/Top
% ----------------------------------------------------------------------------------------------
delete(S) :- retractall(list(_,S)),!.

deleteTop() :-
    not(list(_,_)),
    !,fail.
deleteTop() :-
    topX(count,Top,P),
    count:=P-1,
    retractall(list(P,Top)),
    !.
% ----------------------------------------------------------------------------------------------
% get/Sorted
% ----------------------------------------------------------------------------------------------
predicates
getX : (integer,string_list) procedure (i,o).
clauses

getCount()=Count :- Count=count,!.

getNext(Count,Next,NextCount) :-
    Count=0,
    Next="",
    NextCount=-1,
    !
;    
    list(Count,Next),
    NextCount=Count-1,
    !
;
    N=Count-1,
    getNext(N,Next,NextCount),
    !.

get()=L :- getX(count,L),!.

get_nd()=S :- list(_,S).
   
getSorted()=L :-
    getX(count,L2),
    sx::sxlSort(L2,L),
    !.

getX(0,[]) :-
    !.
getX(N,[H|T]) :-
    list(N,H),
    M=N-1,
    getX(M,T),
    !.
getX(N,T) :-
    M=N-1,
    getX(M,T),
    !.
% ----------------------------------------------------------------------------------------------
% member
% ----------------------------------------------------------------------------------------------
member(M) :- list(_,M),!.
% ----------------------------------------------------------------------------------------------
% size
% ----------------------------------------------------------------------------------------------
predicates
sizeX : (integer,integer) procedure (i,o).
clauses

size()=N :- sizeX(count,N),!.

sizeX(0,0) :-
    !.
sizeX(N,Size) :-
    list(N,_),
    M=N-1,
    sizeX(M,RestSize),
    Size=RestSize+1,
    !
;    
    M=N-1,
    sizeX(M,Size),
    !.
% ----------------------------------------------------------------------------------------------
% sort
% ----------------------------------------------------------------------------------------------
predicates
sortX : (string_list).
clauses

sort() :-
    L=getSorted(),
    retractall(list(_,_)),
    count:=0,
    sortX(L),
    !.
    
sortX([]) :- !.
sortX([H|T]) :- add(H), sortX(T),!.    
    
end implement lst
