/*****************************************************************************

                        Copyright (c) Prolog Development Center A/S

******************************************************************************/

/*****************************************************************************
This scanner breaks a string up into a list of tokens of the following
kinds:

TOK	= lbrack; rbrack;    [ ]
	  lpar; rpar;        ( )
	  lcurly; rcurly;    { }
	  var(STRING);       X,  This, _ok
	  atom(STRING);      hello, :-, :::, fail

	  int(INTEGER);      123, -456
	  str(STRING);       "This is a string"
	  char(CHAR);        `*
	  
	  comma;             ,
	  bar;               |
	  dot                .

A symbol is either a name starting with a lowercase letter, or a sequence
of the following characters:

  +  -  *  /  =  `  :  .  \  ^  <  >  ?  @  #  $  &

******************************************************************************/


implement pie_scan
open core

constants
        className = "com/visual-prolog/Scanner/scanner".
        version = "$JustDate: 11.01.03 $$Revision: 3 $".

clauses
        classInfo(className, version).

class predicates 
  scan_error : (string) determ anyflow.
  maketok : (cursor_POSITION,string,tok,string,string,cursor_POSITION) determ anyflow.
  scan_str : (cursor_POSITION,char,string,string,string) determ anyflow.
  search_ch : (char,string,core::integer32,core::integer32) determ anyflow.
  skipspaces : (string,string,core::integer32) determ anyflow.
  isspace : (char) determ anyflow.
  white_follow : (string) determ anyflow.
  is_symbchar : (char) determ anyflow.
  scan_atom : (string,cursor_POSITION) determ anyflow.
  commentlength : (core::integer32,string,cursor_POSITION,cursor_POSITION) determ anyflow.
  ok_follow_uscore : (char) determ anyflow.
clauses
  scan_error(STRING):-pfeil::syserr("scan_error","",[STRING],[]),!.

  tokl(POS,STR,[t(TOK,POS1)|TOKL]):-
	skipspaces(STR,STR1,NOOFSP),
	STR1<>"",
	POS1=POS+NOOFSP,
	sx::fronttoken(STR1,STRTOK,STR2),!,
	maketok(POS,STRTOK,TOK,STR2,STR3,LEN1),
	LEN=string::length(STRTOK),
	POS2=POS1+LEN+LEN1,
	tokl(POS2,STR3,TOKL).
  tokl(_,_,[]).

  skipspaces(STR,STR2,NOOFSP):-
	string::frontchar(STR,CH,STR1),
	isspace(CH),!,
	skipspaces(STR1,STR2,N1),
	NOOFSP=N1+1.
  skipspaces(STR,STR3,NOOFSP):-
	string::frontchar(STR,C,STR1),
	C='%',
	search_ch('\n',STR1,0,N),
	Nx=N+1,
	string::front(STR1,Nx,DROP,STR2),
	skipspaces(STR2,STR3,N1),
	N2=string::length(DROP),
	NOOFSP=N1+1+N2,
	!.
  skipspaces(STR,STR4,NOOFSP):-
	string::frontchar(STR,C1,STR1),
	C1='/',
	string::frontchar(STR1,C2,STR2),
	C2='*',
	commentlength(0,STR2,0,N),
	string::front(STR2,N,DROP,STR3),
	skipspaces(STR3,STR4,N1),
	N2=string::length(DROP),
	NOOFSP=N1+2+N2,
	!.
  skipspaces(STR,STR,0).

  isspace(' ').  isspace('\t').  isspace('\n').

  white_follow(S):-
	sx::frontchar(S,CH,_),
	not(isspace(CH)),!,fail.
  white_follow(_).

  ok_follow_uscore('_'):-!.
  ok_follow_uscore(CH):-CH>='0',CH<='9',!.
  ok_follow_uscore(CH):-CH>='a',CH<='z',!.
  ok_follow_uscore(CH):-CH>='A',CH<='Z',!.

  maketok(_,"(",lpar,S,S,0):-!.
  maketok(_,")",rpar,S,S,0):-!.
  maketok(_,"[",lbrack,S,S,0):-!.
  maketok(_,"]",rbrack,S,S,0):-!.
  maketok(_,"{",lcurly,S,S,0):-!.
  maketok(_,"}",rcurly,S,S,0):-!.
  maketok(_,"|",bar,S,S,0):-!.

  maketok(_,".",dot,S,S,0):-
	white_follow(S),!.

  maketok(_,",",comma,S,S,0):-!.
  maketok(_,";",atom(";"),S,S,0):-!.
  maketok(_,"!",atom("!"),S,S,0):-!.
  maketok(_,"#",atom("#"),S,S,0):-!. % Pro/3
  maketok(_,Hex,hex(Hex),S,S,0):- % Clan (to support scanning of hex literals in e.g. pie_scan.pro) (21.11.04)
    string::length(Hex)>2,
    string::front(Hex,2,Hx,_),
    Hx="0x",
    !.
/* 13.7.5
  maketok(_,"`",chr(T),S1,S2,1):-!,
	sx::frontchar(S1,T,S2).
*/	
  maketok(_,STRING,atom(SYMB),S,S1,LEN):-
	CH=string::subchar(STRING,0),
	is_symbchar(CH),!,
	scan_atom(S,LEN),
	sx::frontstr(LEN,S,SYMB1,S1),
	sx::frontchar(SYMB,CH,SYMB1).
%	
% Pro/3 change: Changed the two next clauses to avoid backslashes with quotes and apostrophes
%
  maketok(CURSOR,S,str(STR),S1,S2,LEN):-
    string::length(S)=1,
    Cqt=string::subChar(S,0),
    Cqt='"',
	!,scan_str(CURSOR,Cqt,S1,S2,STR),
	LEN1=string::length(STR),
	LEN=LEN1+1.
  maketok(CURSOR,S,str(STR),S1,S2,LEN):-
    S="'",
    !,Cap=string::subChar(S,0),
	scan_str(CURSOR,Cap,S1,S2,STR),
	LEN1=string::length(STR),
	LEN=LEN1+1.
  maketok(_,INTSTR,int(LONG),S,S,0):-
    LONG=sx::toInt(INTSTR),!.

%
% Pro/3 change: to accept single-quoted strings (interpretion is same as double quoted strings)
%
maketok(CURSOR,"'",str(ATOM),S1,S2,LEN):-!,
    Ca=string::subchar("'",0),
	scan_str(CURSOR,Ca,S1,S2,ATOM),
	LEN=string::length(ATOM)+1.
%
% Pro/3 addition: fix-number extension
%
maketok(_,FixStr,fix(Fix),S,S,0) :-
	_=string::searchchar(FixStr,'.',string::caseSensitive()),
	Fix=sx::toReal(FixStr),!.
%
% Pro/3 change: enhancement to permit large integers in text
%
maketok(_,INTSTR,int(INTEGER),S,S,0) :- 
    INTEGER=sx::toInt(INTSTR),!.
maketok(_,INTSTR,fix(F),S,S,0) :- 
    F=sx::toReal(INTSTR),!.

%  maketok(_,REALSTR,real(REAL),S,S,0):-str_real(REALSTR,REAL),!.
  maketok(_,STRING,var(STRING),S,S,0):-
	CH=string::subChar(STRING,0), 
	CH>='A',CH<='Z',
	string::isname(STRING),
	!
;
	CH=string::subChar(STRING,0), 
	CH='_',
	!.
maketok(_,"_",var(NAME),S1,S2,LEN):-
	sx::frontchar(S1,CH,_),ok_follow_uscore(CH),!,
	sx::fronttoken(S1,TOK,S2),
	LEN=string::length(TOK),
	TOKU=string::toUpperCase(TOK),
	sx::frontchar(NAME,'_',TOKU).
  maketok(_,"_",var("_"),S,S,0):-!.
  maketok(_,STRING,atom(Name),S,S,0) :-
        string::isname(STRING),
        !,
        %Pro/3: Name=string::toLowerCase(STRING).
        Name=STRING.
  maketok(Pos,Start,_,Rest,_,_):-
    S2=toString(Pos),
    L=string::length(Rest),
    L>50,
    string::front(Rest,51,S4,_),
    Mess=string::concatList(["Invalid token. Start position=",S2,", character=",Start,", rest=",S4]),
	scan_error(Mess),
	fail
;
    S2=toString(Pos),
    Mess=string::concatList(["Invalid token. Start position=",S2,", character=",Start,", rest=",Rest]),
	scan_error(Mess),
	fail.

  scan_str(_,CH,IN,OUT,STR):-
	search_ch(CH,IN,0,N),
	N=0,
	!,
	STR="",
	OUT1=IN,
	sx::frontchar(OUT1,_,OUT)
;
	search_ch(CH,IN,0,N),
	sx::frontstr(N,IN,STR,OUT1),!,
	sx::frontchar(OUT1,_,OUT).
  scan_str(Pos,StringStart,Rest,_,_):-
    S2=toString(Pos),
    S3=string::charToString(StringStart),
    L=string::length(Rest),
    L>50,
    string::front(Rest,51,S4,_),
    Mess=string::concatList(["String not terminated. Start position=",S2,", delimiter=",S3,", rest=",S4]),
	scan_error(Mess),
	fail
;
    S2=toString(Pos),
    S3=string::charToString(StringStart),
    Mess=string::concatList(["String not terminated. Start position=",S2,", delimiter=",S3,", rest=",Rest]),
	scan_error(Mess),
	fail.
	
  commentlength(0,STR,N1,N):-
	sx::frontchar(STR,'*',STR1),
	sx::frontchar(STR1,'/',_),!,
	N=N1+2.
  commentlength(NEST,STR,N1,N):-
	sx::frontchar(STR,'*',STR1),
	sx::frontchar(STR1,'/',STR2),!,
	NEST1=NEST-1,
	N11=N1+2,
	commentlength(NEST1,STR2,N11,N).
  commentlength(NEST,STR,N1,N):-
	sx::frontchar(STR,'/',STR1),
	sx::frontchar(STR1,'*',STR2),!,
	NEST1=NEST+1,
	N11=N1+2,
	commentlength(NEST1,STR2,N11,N).
  commentlength(NEST,STR,N1,N):-
	sx::frontchar(STR,_,STR1),
	N11=N1+1,
	commentlength(NEST,STR1,N11,N).

  search_ch(CH,STR,N,N):-
	sx::frontchar(STR,CH,_),!.
  search_ch(CH,STR,N,N1):-
	sx::frontchar(STR,_,S1),
	N2=N+1,
	search_ch(CH,S1,N2,N1).

  scan_atom(S1,LEN):-
	sx::frontchar(S1,CH,S2),
	is_symbchar(CH),!,
	scan_atom(S2,LEN1),
	LEN=LEN1+1.
  scan_atom(_,0).

  is_symbchar('+').  is_symbchar('-').  is_symbchar('*').  is_symbchar('/').
  is_symbchar('=').  is_symbchar(':').  is_symbchar('.').  is_symbchar('&').
  is_symbchar('\\'). is_symbchar('^').  is_symbchar('<').  is_symbchar('>').
  is_symbchar('?').  is_symbchar('@').  
  %Pro/3 is_symbchar('#').  
  is_symbchar('$').
  % Pro/3 additions:
  is_symbchar('%').
  is_symbchar('~').
  is_symbchar('`').
end implement pie_scan
