/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement print
open core
domains
printresolution = res(integer HDPI,integer VPIE,integer Hres,integer VRes).
constants
        className = "com/rudstrand/print/print".
        classVersion = "1.0".
class predicates
prnWarnWidth : (integer).
prnWrapText : (integer,string,string) procedure (i,i,o).
  get_line : (string Input,char NewLineChar,string Line,string Rest) determ (i,i,o,o).
clauses
clauses
classInfo(className, classVersion).
/******************************************************
 Return the first line in a string,The returned line
 contains a linefeed character.
******************************************************/
clauses
  get_line(Str,Char,StartStr,Rest1):-
	sx::searchchar(Str,Char,Pos),!,
	Len = Pos - 1,
	sx::frontstr(Len,Str,StartStr,Rest),
	sx::frontchar(Rest,_,Rest1).
  get_line(Str,_,Str,"").
/******************************************************
	Expand TAB's
******************************************************/
class predicates
  expand_tabs : (integer TabSize,string StrWithTabs,string Expanded) procedure (i,i,o).
  expand_tabs : (integer TabSize,integer CurOffset,string StrWithTabs,string Expanded) procedure (i,i,i,o).
clauses
  expand_tabs(TabSize,Str,Expanded):-
	expand_tabs(TabSize,0,Str,Expanded).
/*@@
  expand_tabs(TabSize,Offset,Str,NewExpanded):-
	sx::searchchar(Str,'\t',POS),!,
	Len = Pos - 1,
	sx::frontstr(Len,Str,Start,Rest),
	sx::frontchar(Rest,_,Rest1),
	NoOfSpaces = TabSize - (Offset + Len) mod TabSize,
	Offset1 = Offset + Len + NoOfSpaces,
	expand_tabs(TabSize,Offset1,Rest1,RestExpanded),
	NoOfSpaces=string::length(Spaces),
	format(NewExpanded,"%s%s%s",Start,Spaces,RestExpanded).
*/	
  expand_tabs(_,_,S,S).
/******************************************************
 Print a single line, if the line is longer than global prn_width (Pro/3)
 characters, the line wraps to the next line.
******************************************************/
class predicates
  print_line : (vpiDomains::windowHandle,integer,integer,integer,string,integer,integer,integer,integer,string) determ (i,i,i,i,i,i,i,i,o,o).
clauses
  print_line(_,_,_,_,"",_,_,DELTA1,DELTA2,""):-!,                                       % At end of string
	DELTA2 = DELTA1 + 1.
  print_line(_,_,_,_,RestStrNEW,LineNo,LineNo,DELTA,DELTA,RestStrNEW):-!.       % At end of page
  print_line(PW,SX,SY,LH,NEWStr,Line1,Lines,DELTA1,DELTA,Rest):-
	vpi::drawText(PW,SX,SY,NEWStr),
	print_line(PW,SX,SY,LH,"",Line1,Lines,DELTA1,DELTA,Rest).
/******************************************************
	Print all the lines in a page
******************************************************/
class predicates
  print_PageLines : (vpiDomains::windowHandle,integer StartX,integer StartY,integer LineW,integer LineH,integer CurrLine,integer NoOfLines,string Str,string Rest,integer TabSize) determ (i,i,i,i,i,i,i,i,o,i).
  myconcat : (string,string,string) procedure (i,i,o).
clauses
  print_PageLines(_,_,_,_,_,_,_,"","",_):-!.                            % At end of string
  print_PageLines(_,_,_,_,_,MaxLineNo,MaxLineNo,Str,Str,_):-!.  % At end of page
  print_PageLines(PW,SX,SY,LW,LH,LineNo,MaxLineNo,Str,Rest,TabSize):-
	get_line(Str,'\n',StartStr,RestStr2),
	expand_tabs(TabSize,StartStr,NEWStr),
	print_line(PW,SX,SY,LH,NEWStr,LineNo,MaxLineNo,0,DELTA,RestStr1),
	LineNo1 = LineNo + DELTA,
	SYNEW = SY + LH * DELTA,
	myconcat(RestStr1,RestStr2,RestStr),
	print_PageLines(PW,SX,SYNEW,LW,LH,LineNo1,MaxLineNo,RestStr,Rest,TabSize).
  /*@@
  print_PageLinesF : (vpiDomains::windowHandle,integer StartX,integer StartY,integer LineW,integer LineH,integer CurrLine,integer NoOfLines,integer TabSize,FILE).
  print_PageLinesF(_,_,_,_,_,_,_,_,File):- eof(File),!.
  print_PageLinesF(_,_,_,_,_,MaxLineNo,MaxLineNo,_,_):-!.  % At end of page
  print_PageLinesF(PW,SX,SY,LW,LH,LineNo,MaxLineNo,TabSize,File):-
  	readln(StartStr),
	expand_tabs(TabSize,StartStr,NEWStr),
	print_line(PW,SX,SY,LH,NEWStr,LineNo,MaxLineNo,0,DELTA,_),
	LineNo1 = LineNo + DELTA,
	SYNEW = SY + LH * DELTA,
	print_PageLinesF(PW,SX,SYNEW,LW,LH,LineNo1,MaxLineNo,TabSize,File).
*/
  myconcat("",S,S):-!.
  myconcat(S,"",S):-!.
  myconcat(Str1,Str2,Result):-
	Result=string::concat(Str1,Str2).
/******************************************************
	Print One Single page
******************************************************/
class predicates
  print_Page : (vpiDomains::windowHandle,string,integer PageNo,printresolution,string Inp,string Rest,integer TabSize) determ (i,i,i,i,i,o,i).
clauses
  print_Page(PW,FName,PageNo,res(HRes,VRes,Height,Width),Str,RestStr,TabSize):-
	V_SCR_RES=vpi::getAttrVal(vpiDomains::attr_screen_vres),
	% Pro/3:
	FSize=(8*VRes)/V_SCR_RES,
	IFSize=math::round(FSize),
  	Font=vpi::fontCreate(vpiDomains::ff_Fixed,[],IFSize),
  	vpi::winSetFont(PW,Font),
	%
	vpi::printStartPage(PW),
	% -1/2 inch from left and from right
	LineWidth = Width - HRes,
	StartX = HRes div 2,
	StartY = VRes div 2,
	Header=string::format("%s  Page -%d-",FName,PageNo),
	vpi::drawText(PW,StartX,StartY,Header),     %print header
	% -1/2 inch from top and from bottom
	LineHeight=(15*VRes) div V_SCR_RES,
	Lines = (Height-VRes) div LineHeight,
	StartY1 = StartY + 2 * LineHeight,
	Lines1 = Lines-2,
	print_PageLines(PW,StartX,StartY1,LineWidth,LineHeight,0,Lines1,Str,RestStr,TabSize),
	vpi::printEndPage(PW),!.
/*@@
  print_PageF : (vpiDomains::windowHandle,string,integer PageNo,printresolution,integer TabSize,FILE).
  print_PageF(PW,FName,PageNo,res(HRes,VRes,Height,Width),TabSize,File):-
	V_SCR_RES=vpi::getAttrVal(attr_screen_vres),
  	style_db(printer,Fnt,Style,Size,_,_,_,_,_),
	FSIZE=(Size*VRes)/V_SCR_RES,
  	Font=vpi::fontCreate(Fnt,Style,FSIZE),
  	vpi::winSetFont(PW,Font),
	vpi::printStartPage(PW),
	% -1/2 inch from left and from right
	LineWidth = Width - HRes,
	StartX = HRes div 2,
	StartY = VRes div 2,
	format(Header,"%s  Page -%d-",FName,PageNo),
	vpi::drawText(PW,StartX,StartY,Header),     %print header
	% -1/2 inch from top and from bottom
	LineHeight=(15*VRes)/V_SCR_RES,
	Lines = (Height-VRes) / LineHeight,
	StartY1 = StartY + 2 * LineHeight,
	Lines1 = Lines-2,
	print_PageLinesF(PW,StartX,StartY1,LineWidth,LineHeight,0,Lines1,TabSize,File),
	vpi::printEndPage(PW),!.
*/	
/******************************************************
	Print all the pages in a file
******************************************************/
class predicates
  print_Pages : (vpiDomains::windowHandle,string FName,integer PageNo,printresolution,integer TabSize,string) determ.
clauses
  print_Pages(_,_,_,_,_,""):-!.
  print_Pages(PW,FName,PageNo,PrintRes,TabSize,Str):-
	print_Page(PW,FName,PageNo,PrintRes,Str,RestStr,TabSize),
	PageNo1=PageNo+1,
	print_Pages(PW,FName,PageNo1,PrintRes,TabSize,RestStr).
/*@@
  print_PagesF : (vpiDomains::windowHandle,string FName,integer PageNo,printresolution,integer TabSize,FILE).
  print_PagesF(_,_,_,_,_,File):- eof(File),!.
  print_PagesF(PW,Title,PageNo,PrintRes,TabSize,File):-
	print_PageF(PW,Title,PageNo,PrintRes,TabSize,File),
	PageNo1=PageNo+1,
	print_PagesF(PW,Title,PageNo1,PrintRes,TabSize,File).
*/
/******************************************************
	Print file
******************************************************/
clauses
print_file_str(FName,TabSize,RName):-
	Str=file::readString(FName),
	print_str(RName,Str,TabSize,-1),
	!.
% fourth parameter to indicate wrapping (1/-1)
%	
print_str(Title,Str,TabSize,1) :-	
	vpiCommonDialogs::printSetup(999,_,_,_,_),
	prnWrapText(75,Str,StrW),
	PW = vpi::printStartJob("Printing..."),
	Height = vpi::getAttrVal(vpiDomains::attr_printer_height),
	Width = vpi::getAttrVal(vpiDomains::attr_printer_width),
	HRes = vpi::getAttrVal(vpiDomains::attr_printer_hres),
	VRes = vpi::getAttrVal(vpiDomains::attr_printer_vres),
	PageNo=1,
	print_Pages(PW,Title,PageNo,res(HRes,VRes,Height,Width),TabSize,StrW),
	vpi::printEndJob(PW),
	!
;
    !.
print_str(Title,Str,TabSize,-1) :-	
	vpiCommonDialogs::printSetup(999,_,_,_,_),
	PW = vpi::printStartJob("Printing..."),
	Height = vpi::getAttrVal(vpiDomains::attr_printer_height),
	Width = vpi::getAttrVal(vpiDomains::attr_printer_width),
	HRes = vpi::getAttrVal(vpiDomains::attr_printer_hres),
	VRes = vpi::getAttrVal(vpiDomains::attr_printer_vres),
	PageNo=1,
	print_Pages(PW,Title,PageNo,res(HRes,VRes,Height,Width),TabSize,Str),
	vpi::printEndJob(PW),
	!
;
    !.
/*@@
print_file(RWidth,FileName,Title,TabSize) :-
	prnWarnWidth(RWidth),
	vpiCommonDialogs::printSetup(999,_,_,_,_),
	PW = vpi::printStartJob("Printing..."),
	Height = vpi::getAttrVal(vpiDomains::attr_printer_height),
	Width = vpi::getAttrVal(vpiDomains::attr_printer_width),
	HRes = vpi::getAttrVal(vpiDomains::attr_printer_hres),
	VRes = vpi::getAttrVal(vpiDomains::attr_printer_vres),
	PageNo=1,
	%
	openread(output_file,FileName),
	readdevice(output_file),
	set(out_file_op,1),
	print_PagesF(PW,Title,PageNo,res(HRes,VRes,Height,Width),TabSize,output_file),
	vpi::printEndJob(PW),
	closefile(output_file),
	set(out_file_op,-1),
	readdevice(keyboard),!.
*/
prnWarnWidth(W) :-
	W>80,
	Ws=toString(W),
	sx::concat(["This report is ",Ws," characters wide. Select PORTRAIT or LANDSCAPE as appropriate"],M),
	vpiCommonDialogs::note("Pro/3",M),
	!
;
	!.
prnWrapText(W,S,Sw) :-
	sx::searchchar(S,'\n',P),
	Wm=W+1,
	P>Wm,
	sx::frontstr(W,S,F,T),
	prnWrapText(W,T,St),
	sx::concat([F,"\n",St],Sw),!
;
	sx::searchchar(S,'\n',P),
	sx::frontstr(P,S,F,T),
	prnWrapText(W,T,St),
	sx::concat([F,St],Sw),!
;
	L=string::length(S),
	L>W,
	sx::frontstr(W,S,F,T),
	prnWrapText(W,T,St),
	sx::concat([F,"\n",St],Sw),!
;
	Sw=S,!.
		
end implement print
