/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement rMatrix
open core

constants
className = "com/rudstrand/../Library/rMatrix/rMatrix".
classVersion = "1.0".

facts
mx : (integer,real_list) nondeterm.
mxCols : integer:=0.
sums : real_list := [].
rows : integer_list := [].
maxs : real_list := [].
mins : real_list := [].

clauses
classInfo(className, classVersion).

% ---------------------------------------------------------------------------------------------------------------
% new
% ---------------------------------------------------------------------------------------------------------------
new(Cols) :-
    mxCols:=Cols,
    !.
% ---------------------------------------------------------------------------------------------------------------
% colDeviations
% ---------------------------------------------------------------------------------------------------------------
predicates
colDevsX : (real_list).
colDevsCols : (real_list,real_list,real_list,real_list) determ (i,i,i,o).
clauses

colDeviations(Averages)=ColDevs :-
    set::initialize(mxCols,0.0,L),
    sums:=L,
    colDevsX(Averages),
    ColDevs=sums,
    !.
    
colDevsX(Averages) :-
    mx(_,L),
        colDevsCols(Averages,L,sums,NewSums),
        sums:=NewSums,
        fail
;
    !.
    
colDevsCols([],_,_,[]) :-
    !.
colDevsCols([A|Averages],[H|T],[S|Sums],[N|NewSums]) :-
    Diff=A-H,
    N=S+Diff*Diff,
    colDevsCols(Averages,T,Sums,NewSums),
    !.
% ---------------------------------------------------------------------------------------------------------------
% colMinRow
% ---------------------------------------------------------------------------------------------------------------
predicates
colMinsRowX : ().
colMinsRowCheck : (integer,real_list,integer_list,real_list,real_list,integer_list) determ (i,i,i,i,o,o).
clauses

colMinsRow(Mins,Rows) :-
    mins:=getRow(1),
    set::initialize(mxCols,1,L),
    rows:=L,
    colMinsRowX,
    Mins=mins,
    Rows=rows,
    !.
    
colMinsRowX() :-
    mx(N,L),
        colMinsRowCheck(N,mins,rows,L,Mins,Rows),
        mins:=Mins,
        rows:=Rows,
        fail
;
    !.
        
colMinsRowCheck(_,[],_,_,[],[]) :-
    !.
colMinsRowCheck(N,[Min|Mins],[Row|Rows],[R|Rs],[NewMin|NewMins],[NewRow|NewRows]) :-
    R>=Min,
    NewMin=Min,
    NewRow=Row,
    colMinsRowCheck(N,Mins,Rows,Rs,NewMins,NewRows),
    !
;      
    NewMin=R,
    NewRow=N,
    colMinsRowCheck(N,Mins,Rows,Rs,NewMins,NewRows),
    !.
% ---------------------------------------------------------------------------------------------------------------
% colMaxRow
% ---------------------------------------------------------------------------------------------------------------
predicates
colMaxsRowX : ().
colMaxsRowCheck : (integer,real_list,integer_list,real_list,real_list,integer_list) determ (i,i,i,i,o,o).
clauses

colMaxsRow(Maxs,Rows) :-
    maxs:=getRow(1),
    set::initialize(mxCols,1,L),
    rows:=L,
    colMaxsRowX,
    Maxs=maxs,
    Rows=rows,
    !.
    
colMaxsRowX() :-
    mx(N,L),
        colMaxsRowCheck(N,maxs,rows,L,Maxs,Rows),
        maxs:=Maxs,
        rows:=Rows,
        fail
;
    !.
        
colMaxsRowCheck(_,[],_,_,[],[]) :-
    !.
colMaxsRowCheck(N,[Max|Maxs],[Row|Rows],[R|Rs],[NewMax|NewMaxs],[NewRow|NewRows]) :-
    R<=Max,
    NewMax=Max,
    NewRow=Row,
    colMaxsRowCheck(N,Maxs,Rows,Rs,NewMaxs,NewRows),
    !
;      
    NewMax=R,
    NewRow=N,
    colMaxsRowCheck(N,Maxs,Rows,Rs,NewMaxs,NewRows),
    !.
% ---------------------------------------------------------------------------------------------------------------
% colSums
% ---------------------------------------------------------------------------------------------------------------
predicates
colSumsX : ().
colSumsCol : (real_list,real_list,real_list) determ (i,i,o).
clauses

colSums()=V :-
    set::initialize(mxCols,0.0,Sums),
    sums:=Sums,
    colSumsX,
    V=sums,
    !.
    
colSumsX() :-    
    mx(_,L),
        colSumsCol(sums,L,Sums),
        sums:=Sums,
        fail
;
    !.
        
colSumsCol([],[],[]) :-
    !.
colSumsCol([H|T],[H2|T2],[Hs|Ts]) :- Hs=H+H2, colSumsCol(T,T2,Ts),!.    
% ---------------------------------------------------------------------------------------------------------------
% getCell
% ---------------------------------------------------------------------------------------------------------------
getCell(Row,Col)=V :-
    mx(Row,L),
    set::setGet(L,Col,V),
    !.
% ---------------------------------------------------------------------------------------------------------------
% getRow
% ---------------------------------------------------------------------------------------------------------------
getRow(Row)=V :- mx(Row,V),!.
% ---------------------------------------------------------------------------------------------------------------
% setCell
% ---------------------------------------------------------------------------------------------------------------
setCell(Row,Col,V) :-
    mx(Row,L),
    set::setReplace(L,Col,V,Lm),
    retract(mx(Row,L)),
    assert(mx(Row,Lm)),
    !
;
    set::initialize(mxCols,0.0,L),
    set::setReplace(L,Col,V,Lm),
    assert(mx(Row,Lm)),
    !.    
% ---------------------------------------------------------------------------------------------------------------
% setRow
% ---------------------------------------------------------------------------------------------------------------
setRow(Row,V) :-
    mx(Row,L),
    retract(mx(Row,L)),
    assert(mx(Row,V)),
    !
;
    assert(mx(Row,V)),
    !.
end implement rMatrix
