/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement set
open core,termDomain,globalDomains

constants
className = "com/rudstrand/set/set".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% initialize
% ----------------------------------------------------------------------------------------------
initialize(0,_,[]) :-
    !.
initialize(N,V,[V|L]) :-
    M=N-1,
    initialize(M,V,L),
    !.
% ----------------------------------------------------------------------------------------------
% sort
% ----------------------------------------------------------------------------------------------
class predicates
insSort : (string S,string_list L,string_list Ls) procedure (i,i,o).
insSort : (integer S,integer_list L,integer_list Ls) procedure (i,i,o).
insSort : (real S,real_list L,real_list Ls) procedure (i,i,o).
clauses

% sorts the list ascending
%
sort([],[]) :- !.
sort([H],[H]) :- !.
sort([H|T],Sorted) :-
	sort(T,Ts),
	insSort(H,Ts,Sorted),
	!.
% inserts P into a sorted list
%
insSort(P,[],[P]) :- !.	
insSort(P,[H|T],[P,H|T]) :- P<=H,!.
insSort(P,[H|T],[H|PT]) :- insSort(P,T,PT),!.	
% ----------------------------------------------------------------------------------------------
% setCount/setCountL
% ----------------------------------------------------------------------------------------------
setCount(S,[S|Set],Count) :- setCount(S,Set,C), Count=C+1,!. 
setCount(S,[_|Set],Count) :- setCount(S,Set,Count),!.
setCount(_,_,0) :- !.

setCountL(S,list(S,Set),Count) :- setCountL(S,Set,C), Count=C+1,!. 
setCountL(S,list(_,Set),Count) :- setCountL(S,Set,Count),!.
setCountL(_,_,0) :- !.
% ----------------------------------------------------------------------------------------------
% setNos
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	list
% 2	in	members
% 3	out	members position (1,2,..) in list
% fails if member is not in list
%
setNos(_,[],[]) :-
	!.
setNos(Set,[M|Ms],[N|Ns]) :-
	setGet(Set,N,M),
	setNos(Set,Ms,Ns),
	!.
% ----------------------------------------------------------------------------------------------
% invertToColMajorOrder
% ----------------------------------------------------------------------------------------------
class predicates
invertToCMO_GetCol : (string_list_list,string_list,string_list_list) determ (i,o,o).
invertToCMO_GetCol : (integer_list_list,integer_list,integer_list_list) determ (i,o,o).
invertToCMO_GetCol : (real_list_list,real_list,real_list_list) determ (i,o,o).
invertToCMO_GetCol : (sterml_list,sterml,sterml_list) determ (i,o,o).
clauses

invertToColMajorOrder([[]|_],[]) :-
    !.
invertToColMajorOrder(Rows,[Col|Cols]) :-
    invertToCMO_GetCol(Rows,Col,RowRests),
    invertToColMajorOrder(RowRests,Cols),
    !.
    
invertToCMO_GetCol([],[],[]) :-
    !.
invertToCMO_GetCol([Row|Rows],[RowH|RowHs],[RowRest|RowRests]) :-
    Row=[RowH|RowRest],
    invertToCMO_GetCol(Rows,RowHs,RowRests),
    !.
% ----------------------------------------------------------------------------------------------
% conc
% ----------------------------------------------------------------------------------------------
class predicates
conc_x : (sterml,sterml,sterml) procedure (o,i,i).
conc_x : (string_list,string_list,string_list) procedure (o,i,i).
conc_x : (integer_list,integer_list,integer_list) procedure (o,i,i).
conc_x : (real_list,real_list,real_list) procedure (o,i,i).
conc_x : (vpiTree::tree_list,vpiTree::tree_list,vpiTree::tree_list) procedure (o,i,i).
conc_x : (vpiDomains::menuItem_list,vpiDomains::menuItem_list,vpiDomains::menuItem_list) procedure (o,i,i).
conc_x : (vpiTableed::tableedit_DATA,vpiTableed::tableedit_DATA,vpiTableed::tableedit_DATA) procedure (o,i,i).
conc_x : (vpiTableed::tableedit_COLUMNDATA,vpiTableed::tableedit_COLUMNDATA,vpiTableed::tableedit_COLUMNDATA) procedure (o,i,i).
conc_x : (vpiDomains::windef_list,vpiDomains::windef_list,vpiDomains::windef_list) procedure (o,i,i).
conc_x : (dbLoc_list,dbLoc_list,dbLoc_list) procedure (o,i,i).
clauses

conc([L],L) :- !.
conc([L1,L2],Lc) :- conc_x(Lc,L1,L2),!.
conc([L1|Ls],Lc) :- conc(Ls,Lc1), conc_x(Lc,L1,Lc1),!.
conc(_,[]) :- !.

% conc_x(A,B,C) if A=B||C
%
conc_x([S|A],[],[S|C]) :- conc_x(A,[],C),!.
conc_x([S|A],[S|B],C) :- conc_x(A,B,C),!.		
conc_x([],_,_) :- !.
% ----------------------------------------------------------------------------------------------
% minus
% ----------------------------------------------------------------------------------------------
minus([],_)=[] :- !.
minus(Set1,[])=Set1 :- !.
minus(Set1,[H|T])=Set :-
    Set2=removeElement(H,Set1),    
    Set=minus(Set2,T),
    !.
% ----------------------------------------------------------------------------------------------
% removeElement
% ----------------------------------------------------------------------------------------------
removeElement(_,[])=[] :- !.
removeElement(M,[H|T])=Set :-
    M=H,
    Set=removeElement(M,T),
    !
;
    Set1=removeElement(M,T),
    Set=[H|Set1],
    !.
% ----------------------------------------------------------------------------------------------
% search
% ----------------------------------------------------------------------------------------------
class predicates
searchX : (integer,integer,string_list,string_list,integer) determ (i,i,i,i,o).
searchX : (integer,integer,integer_list,integer_list,integer) determ (i,i,i,i,o).
searchX : (integer,integer,real_list,real_list,integer) determ (i,i,i,i,o).
searchX : (integer,integer,sterml,sterml,integer) determ (i,i,i,i,o).
searchY : (string_list,string_list) determ.
searchY : (integer_list,integer_list) determ.
searchY : (real_list,real_list) determ.
searchY : (sterml,sterml) determ.
clauses

search(Set,Sub)=P:-
    size(Set,N1),
    size(Sub,N2),
    N1>=N2,
    MaxStart=N1-N2,
    searchX(0,MaxStart,Set,Sub,P),
    !.
    
searchX(P,Max,Set,Sub,Pos) :-
    searchY(Set,Sub),
    Pos=P,
    !
;
    Pn=P+1,
    Pn<=Max,
    searchX(Pn,Max,Set,Sub,Pos),
    !.
        
searchY(_,[]) :-
    !.    
searchY([H|T],[H2|T2]) :-
    H=H2,
    searchY(T,T2),
    !.
% ----------------------------------------------------------------------------------------------
% trunc
% ----------------------------------------------------------------------------------------------
trunc(0,S,S) :-
    !.
trunc(_,[],[]) :-
    !.
trunc(N,[H|T],[H|Tt]) :-
    M=N-1,
    trunc(M,T,Tt),
    !.
% ----------------------------------------------------------------------------------------------
% front
% ----------------------------------------------------------------------------------------------
front(0,S,[],S) :-
    !.
front(N,[H|T],[H|Tm],Tail) :-
    N>0,
    M=N-1,
    front(M,T,Tm,Tail),
    !.
% ----------------------------------------------------------------------------------------------
% intersection
% ----------------------------------------------------------------------------------------------
intersection([],_,[]) :-
    !.
intersection([H|T],Set,[H|Ti]) :-
    member(H,Set),
    intersection(T,Set,Ti),
    !.
intersection([_|T],Set,Ti) :-
    intersection(T,Set,Ti),
    !.
% ----------------------------------------------------------------------------------------------
% setEq
% ----------------------------------------------------------------------------------------------
setEq([],[]) :-
    !.
setEq([H|T],[H2|T2]) :-
    H=H2,
    setEq(T,T2),
    !.
% ----------------------------------------------------------------------------------------------
% distinct
% ----------------------------------------------------------------------------------------------
distinct([],[]) :- !.
distinct([M],[M]) :- !.
distinct([H|T],D) :-
    member(H,T),
    distinct(T,D),
    !
;
    distinct(T,Dt),
    D=[H|Dt],
    !.
% ----------------------------------------------------------------------------------------------
% setGet
% ----------------------------------------------------------------------------------------------
class predicates
setGetX : (integer N,string_list SL,integer Ix,string S) determ (i,i,i,o) (i,i,o,i).
setGetX : (integer N ,integer_list IL,integer Ix,integer I) determ (i,i,i,o) (i,i,o,i).
setGetX : (integer N,real_list RL,integer Ix,real R) determ (i,i,i,o) (i,i,o,i).
setGetX : (integer N,sterml RL,integer Ix,sterm R) determ (i,i,i,o) (i,i,o,i).
setGetX : (integer N,dbLoc_list RL,integer Ix,dbLoc R) determ (i,i,i,o) (i,i,o,i).
setGetX : (integer N,windowHandle_list RL,integer Ix,vpiDomains::windowHandle R) determ (i,i,i,o) (i,i,o,i).
clauses
% gets given set member's index in list (1,2,...) or given list index's member fails if index is out of range of list or member doesn't exist in list.
%
setGetI(Set,IndexNo,Member) :- setGet(Set,IndexNo,Member),!. % dummy to suppress domain-errors in certain contexts (23.11.4)

setGet(Set,IndexNo,Member) :- 
	setGetX(1,Set,IndexNo,Member),
	!.

setGetX(HeadNo,[H|_],HeadNo,H) :- 
	!.
setGetX(HeadNo,[_|T],No,S) :- 
	Next=HeadNo+1, 
	setGetX(Next,T,No,S),
	!.
% ----------------------------------------------------------------------------------------------
% sum
% ----------------------------------------------------------------------------------------------
sum([])=Sum:-Sum=0,!.
sum([N|Ns])=S:-S=N+sum(Ns),!.
% ----------------------------------------------------------------------------------------------
% max/min setMax/Min
% ----------------------------------------------------------------------------------------------
% determines the highest/lowest value in set
%
max([N])=Max:-Max=N,!.
max([N|Ns])=N:-N>=max(Ns),!.
max([_|Ns])=Max:-Max=max(Ns),!.

min([N])=Min:-Min=N,!.
min([N|Ns])=N:-N<=min(Ns),!.
min([_|Ns])=Min:-Min=min(Ns),!.

setMax([X],X) :- 
	!.
setMax([X1,X2],X1) :- 
	X1>X2,
	!.
setMax([_,X2],X2) :- 
	!.
setMax([X1|X],X2) :- 
	setMax(X,X2), 
	X2>X1,
	!.
setMax([X1|_],X1) :- 
	!.

setMin([X],X) :- 
	!.
setMin([X1,X2],X1) :- 
	X1<X2,
	!.
setMin([_,X2],X2) :- 
	!.
setMin([X1|X],X2) :- 
	setMin(X,X2), 
	X2<X1,
	!.
setMin([X1|_],X1) :- 
	!.
/*
setMaxFix([X],X) :- 
	!.
setMaxFix([X1,X2],X1) :- 
	X1>X2,
	!.
setMaxFix([_,X2],X2) :- 
	!.
setMaxFix([X1|X],X2) :- 
	setMaxFix(X,X2), 
	X2>X1,
	!.
setMaxFix([X1|_],X1) :- 
	!.
setMaxFix(_,0) :- 
	feil("setMaxFix",[],[],nill),!.

setMinFix([X],X) :- 
	!.
setMinFix([X1,X2],X1) :- 
	X1<X2,
	!.
setMinFix([_,X2],X2) :- 
	!.
setMinFix([X1|X],X2) :- 
	setMinFix(X,X2), 
	X2<X1,
	!.
setMinFix([X1|_],X1) :- 
	!.
setMinFix(_,0) :- 
	feil("setMin",[],[],nill),!.
*/
% ----------------------------------------------------------------------------------------------
% member
% ----------------------------------------------------------------------------------------------
member(Member,[Member|_]) :- 
    !.
member(Member,[_|T]) :- 
    member(Member,T),
    !.
% ----------------------------------------------------------------------------------------------
% setRem
% ----------------------------------------------------------------------------------------------
class predicates
setRemX : (string_list,integer,integer,string_list) determ (i,i,i,o).
setRemX : (integer_list,integer,integer,integer_list) determ (i,i,i,o).
setRemX : (real_list,integer,integer,real_list) determ (i,i,i,o).
setRemX : (sterml,integer,integer,sterml) determ (i,i,i,o).
clauses
% setRem(List1,No,List2) removes member No from List1 to give List2
%
setRem(Lc,No,Lr) :- 
    setRemX(Lc,1,No,Lr),
    !.

setRemX([_|T],No,No,T) :- 
    !.

setRemX([H|T],N,No,[H|L]) :-	
	M=N+1,
	setRemX(T,M,No,L),
	!.	
% ----------------------------------------------------------------------------------------------
% setRemFront
% ----------------------------------------------------------------------------------------------
setRemFront(S,N,S) :-
	N<=0,
	!.
setRemFront([_|S],N,R) :-
	M=N-1,
	setRemFront(S,M,R),
	!.
% ----------------------------------------------------------------------------------------------
% setRemSub
% ----------------------------------------------------------------------------------------------
% setRemSub(List1,List2,List) removes all members of List1 from List2 to give List3
%
setRemSub([],L,L) :- 
    !.
setRemSub(_,[],[]) :- 
    !.
setRemSub(L1,[H|T],L) :-  
    member(H,L1), 
    setRemSub(L1,T,L),
    !.
setRemSub(L1,[H|T],[H|L]) :- 
    setRemSub(L1,T,L),
    !.
% ----------------------------------------------------------------------------------------------
% replaceMember
% ----------------------------------------------------------------------------------------------
replaceMember([],_,_,[]) :- !.
replaceMember([Old|T],Old,New,[New|Tm]) :- replaceMember(T,Old,New,Tm),!.
replaceMember([H|T],Old,New,[H|Tm]) :- replaceMember(T,Old,New,Tm),!.
% ----------------------------------------------------------------------------------------------
% setReplace
% ----------------------------------------------------------------------------------------------
class predicates
setRepX : (sterml,integer,integer,sterm,sterml) determ (i,i,i,i,o).
setRepX : (string_list,integer,integer,string,string_list) determ (i,i,i,i,o).
setRepX : (integer_list,integer,integer,integer,integer_list) determ (i,i,i,i,o).
setRepX : (real_list,integer,integer,real,real_list) determ (i,i,i,i,o).
clauses

% replaces a member of a set (members are numbered 1,2,...)
%
setReplace(List,No,New,ListM) :- setRepX(List,1,No,New,ListM),!.

setRepX([_|T],N,N,New,[New|T]) :- !.
setRepX([H|T],N,No,New,[H|Tm]) :- Next=N+1, setRepX(T,Next,No,New,Tm),!.
% ----------------------------------------------------------------------------------------------
% setRev
% ----------------------------------------------------------------------------------------------
setRev([],[]) :-
	!.
setRev([H|T],Rev) :-
	setRev(T,Tr),
	conc([Tr,[H]],Rev),
	!.	
% ----------------------------------------------------------------------------------------------
% size
% ----------------------------------------------------------------------------------------------
size([],0) :-
    !.
size([_|T],N) :-
    size(T,M),
    N=M+1,
    !.
end implement set
