/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement sx
open core
constants
className = "com/rudstrand/sx/sx".
classVersion = "1.0".
class predicates
sxChop : (string,integer,string_list) determ (i,i,o).
sxChopNL : (string,integer,string_list) determ (i,i,o).
clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% setEOLs
% ----------------------------------------------------------------------------------------------
setEOLs(Max,EOL,S,T) :-
    string::length(S)<=Max,
    T=S,
    !
;
    U=string::subString(S,0,Max),
    L=string::length(S)-Max,
    Rp=Max-1,
    R=string::subString(S,Rp,L),
    setEOLs(Max,"\n",R,V),
    T=string::concat(U,EOL,V),
    !.
% ----------------------------------------------------------------------------------------------
% setFirstLower
% ----------------------------------------------------------------------------------------------
setFirstLower(S)=L:-
    string::length(S)>0,
    frontstr(1,S,F,T),
    FL=string::toLowerCase(F),
    L=string::concat(FL,T),
    !.
    
setFirstUpper(S)=L:-
    string::length(S)>0,
    frontstr(1,S,F,T),
    FU=string::toUpperCase(F),
    L=string::concat(FU,T),
    !.    
% ----------------------------------------------------------------------------------------------
% sxlTrim
% ----------------------------------------------------------------------------------------------
sxlTrim([],[]) :-
    !.
sxlTrim([H|T],[Hs|Ts]) :-
    Hs=string::trim(H),
    sxlTrim(T,Ts),
    !.
% ----------------------------------------------------------------------------------------------
% sxStripFrontMarks
% ----------------------------------------------------------------------------------------------
sxStripFrontMarks(S,[],S) :-
	!.
sxStripFrontMarks(S,[Mark|Marks],Stripped) :-
	searchstring(S,Mark,P),
	P=1,
	L=string::length(Mark),
	frontstr(L,S,_,T),
	sxStripFrontMarks(T,Marks,Stripped),
	!
;
	sxStripFrontMarks(S,Marks,Stripped),
	!.
% ----------------------------------------------------------------------------------------------
% sxBreak
% ----------------------------------------------------------------------------------------------
% a string of the general format 'aaabbbccc' is when given B='bbb' return the front and end parts
% 'aaa' and 'ccc'. The front and end parts can be empty. B cannot be empty and must exist in the
% given string for the predicate to succeed.
% 
sxBreak(S,B,F,T) :-
	searchstring(S,B,P),
	P=1,
	F="",
	L=string::length(B),
	M=string::length(S),
	L=M,
	T="",
	!
;
	searchstring(S,B,P),
	P=1,
	F="",
	L=string::length(B),
	M=string::length(S),
	L<M,
	frontstr(L,S,_,T),
	!
;
	searchstring(S,B,P),
	P>1,
	L=string::length(B),
	Q=P-1,
	frontstr(Q,S,F,U),
	M=string::length(U),
	L=M,
	T="",
	!
;
	searchstring(S,B,P),
	P>1,
	L=string::length(B),
	Q=P-1,
	frontstr(Q,S,F,U),
	frontstr(L,U,_,T),
	!.
% ----------------------------------------------------------------------------------------------
% sxlFrontStr
% ----------------------------------------------------------------------------------------------
sxlFrontStr(_,[],[],[]) :-
	!.
sxlFrontStr(N,[H|Ht],[F|FF],[T|TT]) :-
	frontstr(N,H,F,T),
	sxlFrontStr(N,Ht,FF,TT),
	!.
% ----------------------------------------------------------------------------------------------
% quoteList                                                            
% ----------------------------------------------------------------------------------------------
quoteList([],[]) :-
    !.
quoteList([S|L],[Sq|Lq]) :-
    sxQuote(S,Sq),
    quoteList(L,Lq),
    !.
% ----------------------------------------------------------------------------------------------
% sxl_count                                                                    
% ----------------------------------------------------------------------------------------------
% counts the number of strings in S and determines the length of the longest 
% member string
class predicates
sxl_count_x : (string_list,integer,integer,integer,integer) determ (i,i,o,i,o).
clauses
sxl_count([],0,0) :- !.
sxl_count(S,MaxLen,N) :- sxl_count_x(S,0,MaxLen,0,N),!.
sxl_count_x([],M,M,N,N) :- !.
sxl_count_x([H|T],Mi,Mo,Ni,No) :-
	M=string::length(H),
	M>Mi,
	N=Ni+1,
	sxl_count_x(T,M,Mo,N,No),!
;
	_=string::length(H),
	N=Ni+1,
	sxl_count_x(T,Mi,Mo,N,No),!.
% -------------------------------------------------------------------------------------------------------------
% sxBlockNL
% -------------------------------------------------------------------------------------------------------------
sxBlockNL(Indent,Width,S,Block) :-
	W=Width-Indent,
	sxChopNL(S,W,L2),
	leftAdjustList(L2,L2a),
	sx_ini(Indent,' ',Si),
	sxlAddPrefix(L2a,Si,L3),
	sxlJoin(L3,"\n",Block),
	!.
% ----------------------------------------------------------------------------------------------
% leftAdjust/List
% ----------------------------------------------------------------------------------------------
leftAdjust("","") :-
    !.
leftAdjust(S,T) :-
    string::frontChar(S,C,S2),
    C=' ',
    leftAdjust(S2,T),
    !.
leftAdjust(S,S) :-
    !.
    
leftAdjustList([],[]) :-
    !.
leftAdjustList([H|T],[Ha|Ta]) :-
    leftAdjust(H,Ha),
    leftAdjustList(T,Ta),
    !.
% ----------------------------------------------------------------------------------------------
% sxChop/NL
% ----------------------------------------------------------------------------------------------
class predicates
sxSearchSetRight : (string,string_list,integer) determ (i,i,o).
sxSearchSet : (string,string_list,integer) determ (i,i,o).
clauses

sxChop(S,W,[S]) :-
	string::length(S)<=W,
	!.
sxChop(S,W,[F|SS]) :-
	frontstr(W,S,F,T),
	sxChop(T,W,SS),
	!.

sxChopNL(S,_,[]) :-
	sxStrip(S,Ss),
	string::length(Ss)=0,
	!.
sxChopNL(S,W,[S]) :-
	sxStrip(S,Ss),
	string::length(Ss)<=W,
	!.
sxChopNL(S,W,[F|T]) :-
	frontstr(W,S,Sf,St),
	sxSearchSetRight(Sf,[" ",". ",", ",";","[","]","{","}","|"],P),
	P>0,
	P>=(W*7)/10,
	frontstr(P,Sf,F,Ft),
	S2=string::concat(Ft,St),
	S3=string::trimFront(S2),
	sxChopNL(S3,W,T),
	!
;	
	frontstr(W,S,F,St),
	S2=string::trimFront(St),
	sxChopNL(S2,W,T),
	!.	

sxSearchSetRight(S,Searchs,P) :-
	sx_rev(S,Sr),
	sxRevs(Searchs,SearchsR),
	sxSearchSet(Sr,SearchsR,Pr),
	P=string::length(S)-Pr+1,
	!.
	
sxSearchSet(_,[],0) :-
	!.
sxSearchSet(S,[Search|T],P) :-
	sxSearchSet(S,T,P3),
	P3>0,
	searchstring(S,Search,P2),
	P2<=P3,
	P=P2,
	!
;
	sxSearchSet(S,T,P3),
	P3>0,
	searchstring(S,Search,P2),
	P2>P3,
	P=P3,
	!
;
	sxSearchSet(S,T,P3),
	P3=0,
	searchstring(S,Search,P),
	!
;
	sxSearchSet(S,T,P),
	!.
% ----------------------------------------------------------------------------------------------
% isBlank
% ----------------------------------------------------------------------------------------------
isBlank(S) :-
    St=string::trim(S),
    St="",
    !.
% ----------------------------------------------------------------------------------------------
% replaceAll/C
% ----------------------------------------------------------------------------------------------
replaceAllC(S,From,To)=Sr :-
    FromS=string::charToString(From),
    ToS=string::charToString(To),
    Sr=replaceAll(S,FromS,ToS),
    !.
    
replaceAll(S,From,To)=Sr :-
    S<>"",
    From<>"",
    _=string::search(S,From,string::caseSensitive()),
    Sr=string::replaceAll(S,From,To,string::caseSensitive()),
    !
;
    Sr=S,
    !.
% ----------------------------------------------------------------------------------------------
% sxSetFirstUpper
% ----------------------------------------------------------------------------------------------
sxSetFirstUpper(S,T) :-
	frontchar(S,F,R),
	Fs=string::charToString(F),
	Fu=string::toUpperCase(Fs),
	T=string::concat(Fu,R),
	!.
% ----------------------------------------------------------------------------------------------
% sxFrame
% ----------------------------------------------------------------------------------------------
sxFrame(S,F1,F2,T) :- T=string::concat(F1,S,F2),!.
sxFrame(S,F,T) :- T=string::concat(F,S,F),!.	
% ----------------------------------------------------------------------------------------------
% sxlStripComments
% ----------------------------------------------------------------------------------------------
class predicates
stripToEnd : (string_list,string_list) determ (i,o).
clauses
% Strips of comments of format '/* comment */' and '     %comment' (i.e. NOT %-style comments as line-trailer)
% Comments can not be nested. 
% The source is not parsed with respect to string literals, thus '/*' and '*/' can NOT be used inside literals.
% fails if unmatched '*/'
%
sxlStripComments([],[]) :-
    !.
sxlStripComments([H|T],Ls) :-
    string::trim(H)="",
    sxlStripComments(T,Ls),
    !
;    
    % line: '%xxxxxxxxx' 
    P=string::searchchar(H,'%',string::caseSensitive()),
    P=0,
    sxlStripComments(T,Ls),
    !
;    
    % line: '       %xxxxxxxxx'
    P=string::searchchar(H,'%',string::caseSensitive()),
    P>0,
    frontstr(P,H,Front,_),
    string::trim(Front)="",
    sxlStripComments(T,Ls),
    !
;   
    % line: '/*xxxxxxx'
    P=string::search(H,"/*",string::caseSensitive()),
    P=0,
    stripToEnd([H|T],L),
    sxlStripComments(L,Ls),
    !
;
    % line: 'yyyy/*xxxxxxxx'
    P=string::search(H,"/*",string::caseSensitive()),
    P>0,
    frontstr(P,H,Front,Tail),
    stripToEnd([Tail|T],L),
    sxlStripComments([Front|L],Ls),
    !
;
    sxlStripComments(T,Ls2),
    Ls=[H|Ls2],
    !.
stripToEnd([H|T],Ls) :-
    P=string::search(H,"*/",string::caseSensitive()),
    P2=P+2,
    frontstr(P2,H,_,Tail),
    Ls=[Tail|T],
    !
;    
    stripToEnd(T,Ls),
    !.
% ----------------------------------------------------------------------------------------------
% sxStripComments
% ----------------------------------------------------------------------------------------------
class predicates
stripCommentsA : (string,string) procedure (i,o).
stripCommentsB : (string,string) determ (i,o).
stripComm : (integer,string,string) determ (i,i,o).
stripP_Com : (string_list,string_list) determ (i,o).
clauses
    
% Strips of comments of format '/* comment */' and '     %comment'
% Comments can not be nested. 
% The source is not parsed with respect to string literals, thus '/*' and '*/' can NOT be used inside literals.
% fails if unmatched '*/'
% 3.3.96
%
sxStripComments(Sc,Sr) :-
    stripCommentsA(Sc,S2),
    stripCommentsB(S2,Sr),
    !.
stripCommentsB(Sc,Sr) :-
    sxUnpack(Sc,'\n',L),
    stripP_Com(L,Lr),
    sxlJoin(Lr,"\n",Sr),
    !.
stripP_Com([],[]) :-
    !.
stripP_Com([H|T],L) :-
    Ht=string::trim(H),
    Ht<>"",
    C=string::subChar(Ht,0),
    C='%',
    stripP_Com(T,L),
    !
;
    stripP_Com(T,L2),
    L=[H|L2],
    !.
    
stripCommentsA(Sc,Sr) :-
	searchstring(Sc,"/*",P1),
	P1=1,
	frontstr(2,Sc,_,T1),
	stripComm(1,T1,Rest),
	stripCommentsA(Rest,Sr),
	!
;	
	searchstring(Sc,"/*",P1),
	L=P1-1,
	frontstr(L,Sc,F1,T1),
	frontstr(2,T1,_,T2),
	stripComm(1,T2,Rest),
	stripCommentsA(Rest,S1),
	Sr=string::concat(F1,S1),
	!.
stripCommentsA(S,S) :- 
    !.
stripComm(0,S,S) :- 
    !.
stripComm(EndsLeft,S,R) :-
	searchstring(S,"*/",P1),
	searchstring(S,"/*",P2),
	P2<P1,
	% nested comment
	M=EndsLeft+1,
	L=P2+1,
	frontstr(L,S,_,T),
	stripComm(M,T,R),
	!
;
	searchstring(S,"*/",P1),
	L=P1+1,
	frontstr(L,S,_,T),
	M=EndsLeft-1,
	stripComm(M,T,R),!.
% ----------------------------------------------------------------------------------------------
% sxNormalize
% ----------------------------------------------------------------------------------------------
sxNormalize(S,T) :-
	S2=string::trim(S),
	sxUnpack(S2,' ',L),
	sxlJoin(L," ",T),
	!
;
	T=string::trim(S),
	!.
% ----------------------------------------------------------------------------------------------
% sxTrunc
% ----------------------------------------------------------------------------------------------
sxTrunc(S,N,T) :-
	string::length(S)>N,
	T=string::subString(S,0,N),
	!
;
	T=S,
	!.	
% ----------------------------------------------------------------------------------------------
% sxRemChar/Str
% ----------------------------------------------------------------------------------------------
sxRemChar(S,C,T) :-
    Cs=string::charToString(C),
    sxRemStr(S,Cs,T),
    !.
    
sxRemStr(S,Cs,T) :-
    T=replaceAll(S,Cs,""),
    !.
% ----------------------------------------------------------------------------------------------
% unpackList2/3
% ----------------------------------------------------------------------------------------------
unpackList2([],_,[],[]) :-
    !.
unpackList2([H|T],Delim,[H1|T1],[H2|T2]) :-
    sxUnpackS(H,Delim,M),
    M=[H1,H2],
    unpackList2(T,Delim,T1,T2),
    !
;
    pfeil::syserr("unpackList2","Expected two packed elements",[H]),!,fail.
unpackList3([],_,[],[],[]) :-
    !.
unpackList3([H|T],Delim,[H1|T1],[H2|T2],[H3|T3]) :-
    sxUnpackS(H,Delim,M),
    M=[H1,H2,H3],
    unpackList3(T,Delim,T1,T2,T3),
    !
;
    pfeil::syserr("unpackList3","Expected three packed elements",[H]),!,fail.
unpackList4([],_,[],[],[],[]) :-
    !.
unpackList4([H|T],Delim,[H1|T1],[H2|T2],[H3|T3],[H4|T4]) :-
    sxUnpackS(H,Delim,M),
    M=[H1,H2,H3,H4],
    unpackList4(T,Delim,T1,T2,T3,T4),
    !
;
    pfeil::syserr("unpackList4","Expected four packed elements",[H]),!,fail.
unpackList5([],_,[],[],[],[],[]) :-
    !.
unpackList5([H|T],Delim,[H1|T1],[H2|T2],[H3|T3],[H4|T4],[H5|T5]) :-
    sxUnpackS(H,Delim,M),
    M=[H1,H2,H3,H4,H5],
    unpackList5(T,Delim,T1,T2,T3,T4,T5),
    !
;
    pfeil::syserr("unpackList5","Expected five packed elements",[H]),!,fail.
unpackList6([],_,[],[],[],[],[],[]) :-
    !.
unpackList6([H|T],Delim,[H1|T1],[H2|T2],[H3|T3],[H4|T4],[H5|T5],[H6|T6]) :-
    sxUnpackS(H,Delim,M),
    M=[H1,H2,H3,H4,H5,H6],
    unpackList6(T,Delim,T1,T2,T3,T4,T5,T6),
    !
;
    pfeil::syserr("unpackList6","Expected six packed elements",[H]),!,fail.
unpackList7([],_,[],[],[],[],[],[],[]) :-
    !.
unpackList7([H|T],Delim,[H1|T1],[H2|T2],[H3|T3],[H4|T4],[H5|T5],[H6|T6],[H7|T7]) :-
    sxUnpackS(H,Delim,M),
    M=[H1,H2,H3,H4,H5,H6,H7],
    unpackList7(T,Delim,T1,T2,T3,T4,T5,T6,T7),
    !
;
    pfeil::syserr("unpackList7","Expected seven packed elements",[H]),!,fail.
% ----------------------------------------------------------------------------------------------
% maxLength
% ----------------------------------------------------------------------------------------------
class predicates
maxLen : (integer,string_list) -> integer.
clauses
maxLength([])=0:-!.
maxLength(L)=Max :-
    Max=maxLen(0,L),
    !.
    
maxLen(W,[])=W:-!.
maxLen(W,[H|T])=Max:-
    M=string::length(H),
    M>=W,
    Max=maxLen(M,T),
    !
;
    Max=maxLen(W,T),
    !.
% ----------------------------------------------------------------------------------------------
% sx10C
% ----------------------------------------------------------------------------------------------
% generates the 10s complement of a int or fix string
%
sx10C("","") :- 
	!.
sx10C(S,T) :-
	sxTail(1,S,F,T2),
	T2="%",
	sx10C(F,T),
	!.
sx10C(S,T) :-
	string::frontchar(S,C,S1),
	C='.',
	sx10C(S1,S2),
	Cs=string::charToString(C),
	T=string::concat(Cs,S2),
	!
;	
	string::frontchar(S,C,S1),
	C=',',
	sx10C(S1,S2),
	Cs=string::charToString(C),
	T=string::concat(Cs,S2),
	!
;	
	string::frontchar(S,C,S1),
	N=string::searchchar("0123456789",C,string::caseSensitive()),
	M=9-N,
	Cm=string::subString("0123456789",M,1),
	sx10C(S1,S2),
	T=string::concat(Cm,S2),
	!.
% ----------------------------------------------------------------------------------------------
% conFormatF                                                           
% ----------------------------------------------------------------------------------------------
conFormatF(F,Form,S) :-
	_=string::searchchar(Form,'%',string::caseSensitive()),
	FormM=replaceAll(Form,"%",""),
	Fh=F*100.0,
	conFormatF(Fh,FormM,S1),
	S=string::concat(S1,"%"),
	!
;
	% comma w/decimals
	_=string::searchchar(Form,',',string::caseSensitive()),
	P=string::searchchar(Form,'.',string::caseSensitive()),
	sx::frontstr(P,Form,_,T),
	sx::frontstr(1,T,_,Ds),
	D=sx::toInt(Ds),
	fixFormatR(F,D,1,S),
	!
;	
	% no comma w/decimals
	P=string::searchchar(Form,'.',string::caseSensitive()),
	sx::frontstr(P,Form,_,T),
	sx::frontstr(1,T,_,Ds),
	D=sx::toInt(Ds),
	fixFormatR(F,D,0,S),
	!
;
	% comma no decimals
	_=string::searchchar(Form,',',string::caseSensitive()),
	fixFormatR(F,-1,1,S),
	!
;	
	% no comma no decimal (default format)
	fixFormatR(F,-1,0,S),
	!.
% ----------------------------------------------------------------------------------------------
% fixComma                                                                    
% ----------------------------------------------------------------------------------------------
class predicates
fix_com : (string,integer,string) determ (i,i,o).
clauses
% adds commas to the string number: 1123000.00 ---> 1,123,000.00
%
fixComma(Sc,Sr) :-
	string::frontchar(Sc,Sign,Rest),
	Sign='-',
	N=string::length(Rest),
	fix_com(Rest,N,S1),
	Sr=string::concat("-",S1),
	!
;	
	string::frontchar(Sc,Sign,Rest),
	Sign='+',
	N=string::length(Rest),
	fix_com(Rest,N,Sr),
	!
;	
	N=string::length(Sc),
	fix_com(Sc,N,Sr),
	!.
fix_com(Am,N,Am) :- 
    N<4,
    !.
fix_com(Am,N,AmCom) :-
	sxTail(3,Am,A1,T),
	M=N-3,
	fix_com(A1,M,A2),
	concat([A2,",",T],AmCom),
	!.
% ----------------------------------------------------------------------------------------------
% fixExpConv
% ----------------------------------------------------------------------------------------------
class predicates
fixExpConvX : (string,integer,string) determ (i,i,o).
fixExpConvY : (string,string,integer,string) determ (i,i,i,o).
clauses
% converts from 009999.99900E-9990 to non-exponentiated format
% Note!! no format validation is done
%
fixExpConv(Sc,Sr) :-
	searchchar(Sc,'E',P),
	% E-format
	N=P-1,
	N>0,
	frontstr(N,Sc,Num,S2),
	string::frontchar(S2,_,S3),
	Exp=toInt(S3),
	fixExpConvX(Num,Exp,Sr),
	!
;
	% decimal format
	Sr=Sc,
	!.	
fixExpConvX(Num,Exp,S) :-
	searchchar(Num,'.',P),
	% w/ decimal point
	N=P-1,
	frontstr(N,Num,Int,S1),
	string::frontchar(S1,_,Dec),
	fixExpConvY(Int,Dec,Exp,S),
	!
;	
	% wout decimal point
	fixExpConvY(Num,"0",Exp,S),
	!.
fixExpConvY(Int,Dec,Exp,S) :-
	Exp>0,
	sx_ini(Exp,'0',S1),
	S2=string::concat(Dec,S1),
	frontstr(Exp,S2,S3,S4),
	concat([Int,S3,".",S4],Sx),
	S=string::trim(Sx),
	!
;
	frontstr(1,Int,Sign,Rest),
	Sign="-",
	N=-Exp,
	sx_ini(N,'0',S1),
	S2=string::concat(S1,Rest),
	sxTail(N,S2,S4,S5),
	concat([Sign,S4,".",S5,Dec],Sx),
	S=string::trim(Sx),
	!
;
	frontstr(1,Int,Sign,Rest),
	Sign="+",
	N=-Exp,
	sx_ini(N,'0',S1),
	S2=string::concat(S1,Rest),
	sxTail(N,S2,S4,S5),
	concat([Sign,S4,".",S5,Dec],Sx),
	S=string::trim(Sx),
	!
;
	N=-Exp,
	sx_ini(N,'0',S1),
	S2=string::concat(S1,Int),
	sxTail(N,S2,S4,S5),
	concat([S4,".",S5,Dec],Sx),
	S=string::trim(Sx),
	!.
% ----------------------------------------------------------------------------------------------
% fixFormatR/I/L/S                                                               
% ----------------------------------------------------------------------------------------------
class predicates
fix_adj : (string,integer,integer,string) determ (i,i,i,o).
fix_dec : (string,integer,integer,string) determ (i,i,i,o).
fixExpAdj : (string,integer,string) determ (i,i,o).
fixExpAdj2 : (string,string,integer,string) determ (i,i,i,o).
clauses

% 1. par (in)	S: string input (w/wout decimal point; w/wout E) 
%                   R/I: numeric input -123  123.23  12.067e-12
% 2. par (in)	#decimals desired; or -1 if adapted to argument's precision
% 3. par (in)	1:use commas in the integer portion; 0:not
% 4. par (out)	formatted number (string)
%
fixFormatR(R,NDec,UseComma,FF) :-
    S=toString(R),
    searchchar(S,'e',_),
    sxUnpack(S,'e',L),
    L=[R2,E],
    Exp=toInt(E),
    fixExpAdj(R2,Exp,S2),
    fixFormatS(S2,NDec,UseComma,FF),
    !
;
	S=toString(R),
	searchchar(S,'.',_),
	fixFormatS(S,NDec,UseComma,FF),
	!
;	
	S2=toString(R),
	S3=string::concat(S2,".0"),
	fixFormatS(S3,NDec,UseComma,FF),
	!.
	
fixFormatI(I,NDec,UseComma,FF) :-
    S2=toString(I),
	S3=string::concat(S2,".0"),
	fixFormatS(S3,NDec,UseComma,FF),
	!.
	
fixFormatS(S,NDec,UseComma,FF) :-
	fixExpConv(S,S2),
	fix_adj(S2,NDec,UseComma,FF),
	!.

fix_adj(Am,-1,0,Am) :- 
    !.
fix_adj(Am,0,0,AmAdj) :-
	searchchar(Am,'.',P),
	N=P-1,
	frontstr(N,Am,AmAdj,_),
	!
;	
	AmAdj=Am,
	!.
fix_adj(Am,NDec,0,AmAdj) :-
	searchchar(Am,'.',P),
	N=P-1,
	frontstr(N,Am,AmInt1,Am1),
	frontstr(1,Am1,_,AmDec1),
	N1=string::length(AmDec1),
	fix_dec(AmDec1,N1,NDec,AmDec2),
	concat([AmInt1,".",AmDec2],AmAdj),
	!
;	
	fix_dec("00",2,NDec,Am2),
	concat([Am,".",Am2],AmAdj),
	!.
fix_adj(Am,0,1,AmAdj) :-
	searchchar(Am,'.',P),
	N=P-1,
	frontstr(N,Am,Am1,_),
	fixComma(Am1,AmAdj),
	!
;	
	fixComma(Am,AmAdj),
	!.
fix_adj(Am,-1,1,AmAdj) :-
	searchchar(Am,'.',P),
	N=P-1,
	frontstr(N,Am,AmInt1,Am1),
	frontstr(1,Am1,_,Dec),
	fixComma(AmInt1,AmInt2),
	concat([AmInt2,".",Dec],AmAdj),
	!
;	
	fixComma(Am,AmAdj),
	!.
fix_adj(Am,NDec,1,AmAdj) :-
	searchchar(Am,'.',P),
	N=P-1,
	frontstr(N,Am,AmInt1,Am1),
	frontstr(1,Am1,_,AmDec1),
	N1=string::length(AmDec1),
	fix_dec(AmDec1,N1,NDec,AmDec2),
	fixComma(AmInt1,AmInt2),
	concat([AmInt2,".",AmDec2],AmAdj),
	!
;	
	fixComma(Am,Am1),
	fix_dec("00",2,NDec,Am2),
	concat([Am1,".",Am2],AmAdj),
	!.

fix_dec(Am,L,L,Am) :- 
    !.
fix_dec(AmDec,L,N,AmAdj) :-
	L<N,
	Am2=string::concat(AmDec,"00"),
	M=L+2,
	fix_dec(Am2,M,N,AmAdj),
	!
;
	frontstr(N,AmDec,AmAdj,_),
	!.

% parameters
% 1 in  number-string w/wout sign w/wout dec.point
% 2 in  exponention factor w/wout sign
% 3 out adjusted number string w/ dec point and reflecting exponentiation
%
fixExpAdj(S,Exp,Sr) :-
    searchchar(S,'.',_),
    sxUnpack(S,'.',L),
    L=[Si,Sd],
    fixExpAdj2(Si,Sd,Exp,Sr),
    !
;
    S2=string::concat(S,".0"),
    fixExpAdj(S2,Exp,Sr),
    !.
    
fixExpAdj2(Si,Sd,Exp,Sr) :-
    Exp<-16,
    Sr="0.0",
    !
;
    Exp<0,
    N=-Exp,
    sx_ini(N,'0',Zeros),
    SiZ=string::concat(Zeros,Si),
    sxTail(N,SiZ,SiZF,SiZT),
    Sr=string::concat(SiZF,".",SiZT,Sd),
    !
;
    Exp>0,
    sx_ini(Exp,'0',Zeros),
    SdZ=string::concat(Sd,Zeros),
    frontstr(Exp,SdZ,SdZF,SdZT),
    Sr=string::concat(Si,SdZF,".",SdZT),
    !.
% ----------------------------------------------------------------------------------------------
% frontchar                  
% ----------------------------------------------------------------------------------------------
% VIP5 compatible
frontchar(S,F,T) :-
     bound(S), free(F), free(T), string::frontchar(S,F,T),!
;    free(S), bound(F), bound(T), Fs=string::charToString(F), S=string::concat(Fs,T),!
;    bound(S), bound(F), bound(T), Fs=string::charToString(F), Sx=string::concat(Fs,T), Sx=S,!
;    bound(S), bound(F), free(T), string::frontchar(S,FF,T), F=FF,!.
% ----------------------------------------------------------------------------------------------
% frontstr                            
% ----------------------------------------------------------------------------------------------
% VIP5 compatible
frontstr(N,S,Head,Tail) :-
    N=0,
    Head="",
    Tail=S,
    !
;
    L=string::length(S),
    N<L,
    Head=string::subString(S,0,N),
    Lr=L-N,
    Tail=string::subString(S,N,Lr),
    !
;
    L=string::length(S),
    N=L,
    Head=S,
    Tail="",
    !.
% ----------------------------------------------------------------------------------------------
% fronttoken                
% ----------------------------------------------------------------------------------------------
% VIP5 compatible
fronttoken(S,T,R) :-
    bound(S),free(T),free(R), string::fronttoken(S,T,R),!
    ;free(S), bound(T), bound (R), S=string::concat(T,R),!
    ;bound(S),bound(T),bound(R), Sx=string::concat(T,R), Sx=S,!
    ;bound(S),bound(T),free(R), string::fronttoken(S,Tx,R), Tx=T,!.
% ----------------------------------------------------------------------------------------------
% searchchar  
% ----------------------------------------------------------------------------------------------
% VIP5 compatible (1st pos. is 1 and case sensitive)       
searchchar(S,C,P) :-
    Q=string::searchchar(S,C,string::caseSensitive()),
    P=Q+1,
    !.
% ----------------------------------------------------------------------------------------------
% searchstring
% ----------------------------------------------------------------------------------------------
% VIP5 compatible (1st pos. is 1 and case sensitive)       
searchstring(S,Sub,P) :-
    Q=string::search(S,Sub,string::caseSensitive()),
    P=Q+1,
    !.
% ----------------------------------------------------------------------------------------------
% substring
% ----------------------------------------------------------------------------------------------
% VIP5 compatible (1st pos. is 1 and case sensitive)       
substring(S,Start,Length,Sub) :-
    NStart=Start-1,
    Sub=string::subString(S,NStart,Length),
    !.
% ----------------------------------------------------------------------------------------------
% sxDefault
% ----------------------------------------------------------------------------------------------
sxDefault("",S,S) :- !.
sxDefault(S,_,S) :- !.
% ----------------------------------------------------------------------------------------------
% sxRemoveHead/Tail                                                  
% ----------------------------------------------------------------------------------------------
sxRemoveTail(SE,E,S) :-
    string::length(E)=0,
    S=SE,
    !
;
    string::length(SE)<string::length(E),
    S=SE,
    !
;
    L=string::length(E),
    sxTail(L,SE,H,T),
    T=E,
    S=H,
    !
;
    S=SE,
    !.
sxRemoveHead(ES,E,S) :-
    string::length(E)=0,
    S=ES,
    !
;
    string::length(ES)<string::length(E),
    S=ES,
    !
;
    L=string::length(E),
    frontstr(L,ES,H,T),
    H=E,
    S=T,
    !
;
    S=ES,
    !.
% ----------------------------------------------------------------------------------------------
% sxGiveHead/Tail                                                  
% ----------------------------------------------------------------------------------------------
sxGiveTail(S,E,Se) :-
	string::length(E)=0,
	Se=S,
	!
;	
	L=string::length(E),
	L2=string::length(S),
	L2<L,
	Se=string::concat(S,E),
	!
;	
	L=string::length(E),
	sxTail(L,S,_,E2),
	E=E2,
	Se=S,
	!
;
	Se=string::concat(S,E),
	!.	
sxGiveHead(S,E,Se) :-
	string::length(E)=0,
	Se=S,
	!
;	
	L=string::length(E),
	L2=string::length(S),
	L2<L,
	Se=string::concat(E,S),
	!
;	
	L=string::length(E),
	frontstr(L,S,E2,_),
	E=E2,
	Se=S,
	!
;
	Se=string::concat(E,S),
	!.	
% ----------------------------------------------------------------------------------------------
% tokens
% ----------------------------------------------------------------------------------------------
tokens(S,T) :-
    frontToken(S,Tok,Rem),
    tokens(Rem,Tokens),
    T=[Tok|Tokens],
    !
;
    T=[],
    !.
% ----------------------------------------------------------------------------------------------
% letter/letterS
% ----------------------------------------------------------------------------------------------
isLetter(C) :- _=string::searchchar("ABCDEFGHIJKLMNOPQRSTUVWXYZ",C,string::caseInsensitive()),!.
isLetterS(C) :- 
    string::length(C)=1,
    _=string::search("ABCDEFGHIJKLMNOPQRSTUVWXYZ",C,string::caseInsensitive()),
    !
;
    string::length(C)<>1,
    pfeil::syserr("isLetterS","Invalid parameter",[C]),!,fail.
% ----------------------------------------------------------------------------------------------
% digit/digitS
% ----------------------------------------------------------------------------------------------
isDigit(C) :- _=string::searchchar("0123456789",C,string::caseInsensitive()),!.
isDigitS(C) :- 
    string::length(C)=1,
    _=string::search("0123456789",C,string::caseInsensitive()),
    !
;
    string::length(C)<>1,
    pfeil::syserr("isDigitS","Invalid parameter",[C]),!,fail.
% ----------------------------------------------------------------------------------------------
% sxSubstList
% ----------------------------------------------------------------------------------------------
sxSubstNR_List(S,[],_,S) :-
	!.
sxSubstNR_List(Sc,[C|Cs],[Cm|Cms],Sr) :-
	S2=replaceAll(Sc,C,Cm),
	sxSubstNR_List(S2,Cs,Cms,Sr),
	!.
sxSubstList(S,[],_,S) :-
	!.
sxSubstList(Sc,[C|Cs],[Cm|Cms],Sr) :-
    S2=replaceAll(Sc,C,Cm),
	sxSubstList(S2,Cs,Cms,Sr),
	!.
% ----------------------------------------------------------------------------------------------
% sx_ini
% ----------------------------------------------------------------------------------------------
% generates a string S filled with C and with length L
% 14.5.96
%
sx_ini(0,_,"") :- 
    !.
sx_ini(L,C,S) :- 
	N=L-1,
	sx_ini(N,C,T),
	Cs=string::charToString(C),
	S=string::concat(Cs,T),
	!.
% ----------------------------------------------------------------------------------------------
% sxUnquote; sxQuote
% ----------------------------------------------------------------------------------------------
sxUnquote(S,T) :-
	string::frontchar(S,C,T1),
	C='"',
	sxTail(1,T1,T,_),
	!
;
	T=S,
	!.	
	
sxQuote(S,Q) :-
    S="",
    Q="\"\"",
    !
;    
	string::subChar(S,0)='"',
	Q=S,
	!
;
	Q=string::concat("\"",S,"\""),
	!.	
/*	
sxApost(S,A) :-
	frontstr(1,S,F,_),
	F="'",
	A=S,
	!
;
	concat(["'",S,"'"],A),
	!.		
	
sxUnapost(A,S) :-
	frontstr(1,A,F,T),
	F="'",
	sxTail(1,T,S,_),
	!
;
	S=A,
	!.		
*/	
% ----------------------------------------------------------------------------------------------
% sx_rev
% ----------------------------------------------------------------------------------------------
% reverses a string
%
sx_rev("","") :- !.
sx_rev(S,R) :-
	sxTail(1,S,F,T),
	sx_rev(F,Fr),
	R=string::concat(T,Fr),
	!.
sxRevs([],[]) :-
	!.
sxRevs([H|T],[Hr|Tr]) :-
	sx_rev(H,Hr),
	sxRevs(T,Tr),
	!.	
% ----------------------------------------------------------------------------------------------
% sx_str_int                                                       
% ----------------------------------------------------------------------------------------------
% converts an integer to string with given length (adds leading 0s if required)
%
class predicates
sx_str_int_x: (string,integer,string) procedure (i,i,o).
clauses
sx_str_int(S,L,N) :-
	S1=toString(N),
	sx_str_int_x(S1,L,S),
	!.
sx_str_int_x(Sc,L,Sr) :-
	string::length(Sc)<L,
	S1=string::concat("0",Sc),
	sx_str_int_x(S1,L,Sr),
	!.
sx_str_int_x(S,_,S) :- 
    !.
% ----------------------------------------------------------------------------------------------
% sxStrip
% ----------------------------------------------------------------------------------------------
sxStrip(Sc,Sr) :- Sr=string::trim(Sc),!.
/*
sxStripClean(Sc,Sr) :-
	sxStrip(Sc,S2),
	sxSubstChar(S2,'\n',' ',S3),
	sxSubst(S3,"  "," ",Sr),
	!.
	
sxlStrip([],[]) :-
	!.
sxlStrip([H|T],[Hs|Ts]) :-
	sxStrip(H,Hs),
	sxlStrip(T,Ts),
	!.
*/
% ----------------------------------------------------------------------------------------------
% sxTail
% ----------------------------------------------------------------------------------------------
% gets a tail substring - fails if requested tail is longer than string
%
sxTail(TailLength,S,Sfront,Stail) :-
   N=string::length(S),
   N>=TailLength,
   FrontLength=N-TailLength,
   frontstr(FrontLength,S,Sfront,Stail),
   !.
/*
sxTailC(S,C,F) :-
	bound(S), free(C), free(F),
  	sxTail(1,S,F,Cs),
  	str_char(Cs,C),
  	!
;
	bound(S), bound(C), free(F),
  	sxTail(1,S,F,T),
  	str_char(T,Tc),
  	C=Tc,
  	!
;
	bound(S), bound(C), bound(F),
  	sxTail(1,S,Fx,T),
  	str_char(T,Cx),
  	F=Fx,
  	C=Cx,
  	!
;
	free(S), bound(C), bound(F),
  	str_char(Cs,C),
  	concat(F,Cs,S),
  	!.
 */
% ----------------------------------------------------------------------------------------------
% sxAllDigits                                                                
% ----------------------------------------------------------------------------------------------
% @@ note that this predicate originally failed on zero-length strings.note sure if this is the case now (22.8.4).
sxAllDigits(S) :-
    string::hasDecimalDigits(S),
    !.
% ----------------------------------------------------------------------------------------------
% sxL_Pad
% ----------------------------------------------------------------------------------------------
% left-pads with given character to given length
%
sxL_Pad(Sc,Pad,Len,Sr) :-
	M=string::length(Sc),
	Diff=Len-M,
	Diff>0,
	CPad=string::subChar(Pad,0),
	sx_ini(Diff,CPad,S1),
	Sr=string::concat(S1,Sc),
	!.
sxL_Pad(S,_,_,S) :- 
	!.
% ----------------------------------------------------------------------------------------------
% sxUnpack/S;sxlUnpackS
% ----------------------------------------------------------------------------------------------
/*
sxUnpackScan(S,'!',L) :-
	sxSubst(S,"!!","~~",S2),
	sxUnpack(S2,'!',L2),
	sxlSubst(L2,"~~","!",L),
	!.
sxUnpackScan(S,'?',L) :-
	sxSubst(S,"??","~~",S2),
	sxUnpack(S2,'?',L2),
	sxlSubst(L2,"~~","?",L),
	!.
*/	
unpack(S,Cs)=L :- sxUnpackS(S,Cs,L),!.

sxUnpack(S,C,[F|L]) :-
	searchchar(S,C,P),
	P=1,
	F="",
	frontstr(1,S,_,T),
	sxUnpack(T,C,L),
	!
;
	searchchar(S,C,P),
	N=P-1,
	frontstr(N,S,F,_),
	frontstr(P,S,_,T),
	sxUnpack(T,C,L),
	!.
sxUnpack(S,_,[S]) :- 
	!.		
sxUnpackS(S,C,[F|L]) :-
	searchstring(S,C,P),
	P=1,
	F="",
	Cn=string::length(C),
	frontstr(Cn,S,_,T3),
	sxUnpackS(T3,C,L),
	!
;
	searchstring(S,C,P),
	N=P-1,
	frontstr(N,S,F,T2),
	Cn=string::length(C),
	frontstr(Cn,T2,_,T3),
	sxUnpackS(T3,C,L),
	!.
sxUnpackS(S,_,[S]) :- 
	!.		
	
unpackList([],_,[]) :- 
    !.
unpackList([H|T],Cs,[HL|TL]) :-
    sxUnpackS(H,Cs,HL),
    unpackList(T,Cs,TL),
    !.
/*
sxUnpackSB("",_,[]) :- 
	!.		
sxUnpackSB(S,C,L) :-
	searchstring(S,C,P),
	P=1,
	str_len(C,Cn),
	frontstr(Cn,S,_,T3),
	sxUnpackSB(T3,C,L),
	!.
sxUnpackSB(S,C,[F|L]) :-
	searchstring(S,C,P),
	N=P-1,
	frontstr(N,S,F,T2),
	str_len(C,Cn),
	frontstr(Cn,T2,_,T3),
	sxUnpackSB(T3,C,L),
	!.
sxUnpackSB(S,_,[S]) :- 
	!.		
sxlUnpackS([],_,[],[]) :-
	!.
sxlUnpackS([H|T],C,[HL|Tm],[Hmm|Tmm]) :-
	sxUnpackS(H,C,L),
	L=[HL|TL],	
	sxlJoin(TL,C,Hmm),
	sxlUnpackS(T,C,Tm,Tmm),
	!.
*/
% ----------------------------------------------------------------------------------------------
% concat                                                                    
% ----------------------------------------------------------------------------------------------
% concatenates the members of a string list
%
concat([],"") :- !. 
concat([H|T],S) :- concat(T,S1), S=string::concat(H,S1),!.
% ----------------------------------------------------------------------------------------------
% sxlJoin/sxlJoinB
% ----------------------------------------------------------------------------------------------
% joins members in list with a join-string in between
% sxlJoin	ignores blank strings
% sxlJoinB	does not ignore blank strings
%
join([],_)="" :- !.
join([H],_)=H :- !.
join([H|T],J)=S :- S2=join(T,J), S=string::concat(H,J,S2),!.

joinIgnoreBlanks(L,J)=S :- sxlRemBlanks(L,Ln), S=join(Ln,J),!.

sxlJoin(L,J,S) :- S=joinIgnoreBlanks(L,J),!.
sxlJoinB(L,J,S) :- S=join(L,J),!.
% ----------------------------------------------------------------------------------------------
% pack
% ----------------------------------------------------------------------------------------------
pack([],_)="" :-!.
pack([H],_)=H :- !.
pack([H,M|T],P)=S :- Hm=string::concat(H,P,M), S=pack([Hm|T],P),!.
% ----------------------------------------------------------------------------------------------
% sxlRemBlanks
% ----------------------------------------------------------------------------------------------
% removes blank strings from list
%
sxlRemBlanks([],[]) :- !.
sxlRemBlanks([""|T],L) :- sxlRemBlanks(T,L),!.
sxlRemBlanks([H|T],[H|L]) :- sxlRemBlanks(T,L),!.
% ----------------------------------------------------------------------------------------------
% sxlSort
% ----------------------------------------------------------------------------------------------
class predicates
sxlInsSort : (string S,string_list L,string_list Ls) procedure (i,i,o).
clauses
% sorts the list ascending
%
sxlSort([],[]) :- !.
sxlSort([H],[H]) :- !.
sxlSort([H|T],Sorted) :-
	sxlSort(T,Ts),
	sxlInsSort(H,Ts,Sorted),
	!.
% inserts P into a sorted list
%
sxlInsSort(P,[],[P]) :- !.	
sxlInsSort(P,[H|T],[P,H|T]) :- P<=H,!.
sxlInsSort(P,[H|T],[H|PT]) :- sxlInsSort(P,T,PT),!.	
% ----------------------------------------------------------------------------------------------
% sxlSubst/List/NR_List
% ----------------------------------------------------------------------------------------------
sxlSubst([],_,_,[]) :- !.
sxlSubst([H|T],Fr,To,[Ht|Tt]) :- Ht=replaceAll(H,Fr,To),sxlSubst(T,Fr,To,Tt),!.
sxlSubstList([],_,_,[]) :- !.
sxlSubstList([H|T],Fr,To,[Ht|Tt]) :- sxSubstList(H,Fr,To,Ht),sxlSubstList(T,Fr,To,Tt),!.
sxlSubstNR_List([],_,_,[]) :- !.
sxlSubstNR_List([H|T],Fr,To,[Ht|Tt]) :- sxSubstNR_List(H,Fr,To,Ht),sxlSubstNR_List(T,Fr,To,Tt),!.
% ----------------------------------------------------------------------------------------------
% sxRPad
% ----------------------------------------------------------------------------------------------
% right-pads with given character to given length
%
sxR_Pad(Sc,Pad,Len,Sr) :-
	Diff=Len-string::length(Sc),
	Diff>0,
	CPad=string::subchar(Pad,0),
	sx_ini(Diff,CPad,S1),
	Sr=string::concat(Sc,S1),
	!.
sxR_Pad(S,_,_,S) :- 
	!.
% ----------------------------------------------------------------------------------------------
% sxl_il                                               
% ----------------------------------------------------------------------------------------------
sxl_il([],[]) :- !.
sxl_il([H|T],[Hi|Ti]) :- Hi=toInt(H), sxl_il(T,Ti),!.
sxl_ilR([],[]) :- !.
sxl_ilR([H|T],[Hi|Ti]) :- H=toString(Hi), sxl_ilR(T,Ti),!.
% ----------------------------------------------------------------------------------------------
% toInt/List
% ----------------------------------------------------------------------------------------------
class predicates
intProper : (string,integer,integer) determ (i,o,o).
clauses

toIntList([],[]) :- !.
toIntList([H|T],[Hi|Ti]) :- Hi=toInt(H), toIntList(T,Ti),!.

toInt(S)=Int :-
    St=string::trim(S),
    string::frontchar(St,C,Sp),
    C='-',
    string::hasDecimalDigits(Sp),
    intProper(Sp,_,Iu),
    Int=(-1)*Iu,
    !
;
    St=string::trim(S),
    string::frontchar(St,C,Sp),
    C='+',
    string::hasDecimalDigits(Sp),
    intProper(Sp,_,Int),
    !
;
    St=string::trim(S),
    string::hasDecimalDigits(St),
    intProper(St,_,Int),
    !.
intProper(S,Multi,Int) :-
    string::frontchar(S,F,R),
    R="",
    Int=string::searchchar("0123456789",F,string::caseSensitive()),
    Multi=10,
    !.
intProper(S,Multi,Int) :-
    string::frontchar(S,F,R),
    intProper(R,M,I),
    Int=I+M*string::searchchar("0123456789",F,string::caseSensitive()),
    Multi=10*M,
    !.    
% ----------------------------------------------------------------------------------------------
% sxReal
% ----------------------------------------------------------------------------------------------
% converts from real to string and makes sure that the string result has a decimal point 
% (and at least one digit right of the point)
%
sxReal(S,R) :-
	S=toString(R),
	_=string::searchchar(S,'.',string::caseSensitive()),
	!
;
	S1=toString(R),
	S=string::concat(S1,".0"),
	!.	
% ----------------------------------------------------------------------------------------------
% toReal
% ----------------------------------------------------------------------------------------------
toReal(S)=R :- 
    sxRemChar(S,',',Sr),
    conversion5x::str_real(Sr,R),
    !.
% ----------------------------------------------------------------------------------------------
% movePoint
% ----------------------------------------------------------------------------------------------
class predicates
movePointS : (integer,string,string) determ (i,i,o).    
clauses
movePoint(N,F,Fm) :-
    N=0,
    Fm=F,
    !
;    
    F<0.0,
    Fa=-F,
    movePoint(N,Fa,F2),
    Fm=-F2,
    !
;  
    S=toString(F),
    _=string::searchchar(S,'.',string::caseSensitive()),
    Na=math::abs(N),
    sx_ini(Na,'0',Z),
    S2=string::concat(Z,S,Z),
    movePointS(N,S2,S3),
    Fm=toReal(S3),
    !
;    
    S=toString(F),
    Na=math::abs(N),
    sx_ini(Na,'0',Z),
    S2=string::concat(Z,S,".0",Z),
    movePointS(N,S2,S3),
    Fm=toReal(S3),
    !.
movePointS(N,S,Sm) :-
    N<0,
    Na=-N,
    P=string::searchchar(S,'.',string::caseSensitive()),
    string::front(S,P,F,T),
    sx::sxTail(Na,F,F2,T2),
    string::front(T,1,_,T3),
    Sm=string::concat(F2,".",T2,T3),
    !
;
    N>0,
    P=string::searchchar(S,'.',string::caseSensitive()),
    string::front(S,P,F,T),
    string::front(T,1,_,T2),
    string::front(T2,N,F2,T3),
    Sm=string::concat(F,F2,".",T3),
    !.
% ----------------------------------------------------------------------------------------------
% sxlAddPrefix
% ----------------------------------------------------------------------------------------------
sxlAddPrefix([H|T],Prefix,[Ha|Ta]) :-
	Ha=string::concat(Prefix,H),
	sxlAddPrefix(T,Prefix,Ta),!.
sxlAddPrefix(_,_,[]) :-
	!.
% ----------------------------------------------------------------------------------------------
% sxlAddSuffix
% ----------------------------------------------------------------------------------------------
sxlAddSuffix([H|T],Suff,[Ha|Ta]) :-
	Ha=string::concat(H,Suff),
	sxlAddSuffix(T,Suff,Ta),!.
sxlAddSuffix(_,_,[]) :-
	!.
% ----------------------------------------------------------------------------------------------
% sxGetInBrack/S
% ----------------------------------------------------------------------------------------------
sxGetInBrack(S,B1,B2,Sub) :-
	searchchar(S,B1,P),
	frontstr(P,S,_,S2),
	searchchar(S2,B2,P2),
	N=P2-1,
	frontstr(N,S2,Sub,_),
	!.

sxGetInBrackS(S,B1,B2,Sub) :-
	P1=string::search(S,B1),
	L1=P1+string::length(B1),
	frontstr(L1,S,_,S2),
	P2=string::search(S2,B2),
	frontstr(P2,S2,Sub,_),
	!.
% ----------------------------------------------------------------------------------------------
% sxSignedInteger/Fix
% ----------------------------------------------------------------------------------------------
class predicates
sxIntegerTail : (string) determ (i).
clauses
sxInteger(S,I) :-
	sxUnsignedInteger(S,I),
	!
;
	sxSignedInteger(S,I),
	!.
	
sxSignedInteger(S,I) :-
	sxStrip(S,S2),
	frontchar(S2,F,S3),
	F='+',
	sxUnsignedInteger(S3,I),
	!
;
	sxStrip(S,S2),
	frontchar(S2,F,S3),
	F='-',
	sxUnsignedInteger(S3,I2),
	I=-I2,
	!.
	
sxUnsignedInteger(S,I) :-
	sxStrip(S,S2),
	frontchar(S2,F,S3),
	Fs=string::charToString(F),
	string::hasDecimalDigits(Fs),
	sxIntegerTail(S3),
	S4=replaceAll(S2,",",""),
	I=toInt(S4),
	!.
	
sxIntegerTail("") :-
	!.
sxIntegerTail(S) :-
	frontchar(S,F,S2),
	Fs=string::charToString(F),
	string::hasDecimalDigits(Fs),
	sxIntegerTail(S2),
	!
;
	frontchar(S,F,S2),
	F=',',
	sxUnsignedInteger(S2,_),
	!.
sxFix(S,I) :-
	sxUnsignedFix(S,I),
	!
;
	sxSignedFix(S,I),
	!.
		
sxSignedFix(S,I) :-
	sxStrip(S,S2),
	frontchar(S2,F,S3),
	F='+',
	sxUnsignedFix(S3,I),
	!
;
	sxStrip(S,S2),
	frontchar(S2,F,S3),
	F='-',
	sxUnsignedFix(S3,I2),
	I=-I2,
	!.
	
sxUnsignedFix(S,R) :-
	not(searchchar(S,'.',_)),
	sxStrip(S,S2),
	sxUnsignedInteger(S2,I),
	R=I,
	!
;
	sxStrip(S,S2),
	frontchar(S2,F,S3),
	F='.',
	not(searchchar(S3,',',_)),
	sxUnsignedInteger(S3,_),
	frontchar(S4,'0',S3),
	R=toReal(S4),
	!
;
	sxStrip(S,S2),
	sxTail(1,S2,S3,T2),
	T2=".",
	sxTail(1,S3,_,T3),
	T3<>" ",
	sxUnsignedInteger(S3,I),
	R=I,
	!
;	
	sxStrip(S,S2),
	sxUnpack(S2,'.',L),
	L=[Si,Sr],
	sxTail(1,Si,_,T2),
	T2<>" ",
	not(frontchar(Sr,' ',_)),
	not(searchchar(Sr,',',_)),
	sxUnsignedInteger(Si,I),
	I>=0,
	sxUnsignedInteger(Sr,_),
	SiR=toString(I),
	concat([SiR,".",Sr],S3),
	R=toReal(S3),
	!
;
	sxStrip(S,S2),
	sxUnpack(S2,'.',L),
	L=[Si,Sr],
	sxTail(1,Si,_,T2),
	T2<>" ",
	not(frontchar(Sr,' ',_)),
	not(searchchar(Sr,',',_)),
	sxUnsignedInteger(Si,I),
	I<0,
	sxUnsignedInteger(Sr,_),
	SiR=toString(I),
	concat([SiR,".",Sr],S3),
	R2=toReal(S3),
	R=-R2,
	!.
% ----------------------------------------------------------------------------------------------
% joinIntegers
% ----------------------------------------------------------------------------------------------
joinIntegers([],_,"") :- 
    !.
joinIntegers([I|It],Sep,S) :-
	Sf=toString(I),
	joinIntegers(It,Sep,S1),
	S=string::concat(Sf,Sep,S1),
	!.
% ----------------------------------------------------------------------------------------------
% getInt
% ----------------------------------------------------------------------------------------------
class predicates
getInt2 : (string,string,string) procedure (i,o,o).
clauses

getInt(S,N,T) :-
    L=string::length(S),
    L>=3,
    string::front(S,1,Sign,Rest),
    Sign="-",
    getInt(Rest,Na,T),
    N=-Na,
    !
;
    L=string::length(S),
    L>=2,
    string::front(S,1,F,_),
    string::hasDecimalDigits(F),
    getInt2(S,Ns,T),
    N=toInt(Ns),
    !.
    
getInt2(S,Ns,T) :-
    S<>"",
    string::front(S,1,F,Rest),
    string::hasDecimalDigits(F),
    getInt2(Rest,Ns2,T),
    Ns=string::concat(F,Ns2),
    !
;
    T=S,
    Ns="",
    !.
% -------------------------------------------------------------------------------------------------------------
% makeIdentifier
% -------------------------------------------------------------------------------------------------------------
class predicates
makeRest : (integer,integer,string,string,string) procedure (i,i,i,i,o).
makeRestBreak : (integer,integer,string,string,string) procedure (i,i,i,i,o).
setCase : (integer,string,string) determ (i,i,o).
clauses

% makes an identifier from a string e.g. 12 High_ street --> highStreet
% parameters:
% 1 in first letter in identifier set to case: 1=as given (mixed), 2=lower,3=upper
% 2 in after breaks                         do.
% 2 in remaining letters                   do.
% 3 in NL text
% 4 in max length of generated identifier
% 5 in replacement string for blanks (e.g. "_" or "")
% 6 out generated identifier
%
makeIdentifier(FirstCase,BreakCase,Case,NL,Max,BlankReplace,Id) :-
    NL="",
    !,fail
;    
    string::front(NL,1,F,Rest),
    sx::isLetterS(F),
    setCase(FirstCase,F,Fc),
    makeRest(Case,BreakCase,Rest,BlankReplace,Rid),
    Lid=string::concat(Fc,Rid),
    sxTrunc(Lid,Max,Id2),
    sx::sxRemoveTail(Id2,BlankReplace,Id),
    !
;
    string::front(NL,1,_,Rest),
    makeIdentifier(FirstCase,BreakCase,Case,Rest,Max,BlankReplace,Id),
    !.

makeRest(Case,BreakCase,NL,BlankReplace,Id) :-
    NL="",
    Id=NL,
    !
;    
    string::front(NL,1,F,Rest),
    sx::isLetterS(F),
    setCase(Case,F,Fc),
    makeRest(Case,BreakCase,Rest,BlankReplace,Rid),
    Id=string::concat(Fc,Rid),
    !
;    
    string::front(NL,1,F,Rest),
    sx::isDigitS(F),
    makeRest(Case,BreakCase,Rest,BlankReplace,Rid),
    Id=string::concat(F,Rid),
    !
;    
    string::front(NL,1,F,Rest),
    F=" ",
    makeRestBreak(Case,BreakCase,Rest,BlankReplace,Rid),
    Id=string::concat(BlankReplace,Rid),
    !
;
    string::front(NL,1,_,Rest),
    makeRestBreak(BreakCase,Case,Rest,BlankReplace,Id),
    !.
    
makeRestBreak(Case,BreakCase,NL,BlankReplace,Id) :-
    NL="",
    Id=NL,
    !
;    
    string::front(NL,1,F,Rest),
    sx::isLetterS(F),
    setCase(BreakCase,F,Fc),
    makeRest(Case,BreakCase,Rest,BlankReplace,Rid),
    Id=string::concat(Fc,Rid),
    !
;    
    string::front(NL,1,F,Rest),
    sx::isDigitS(F),
    makeRestBreak(Case,BreakCase,Rest,BlankReplace,Rid),
    Id=string::concat(F,Rid),
    !
;
    string::front(NL,1,_,Rest),
    makeRestBreak(Case,BreakCase,Rest,BlankReplace,Id),
    !.
    
setCase(1,C,C) :- !.
setCase(2,C,CL) :- CL=string::toLowercase(C),!.    
setCase(3,C,CU) :- CU=string::toUppercase(C),!.    
end implement sx
