/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement ted
open core,termDomain,globalDomains

constants
className = "com/rudstrand/ted/ted".
classVersion = "1.0".

class facts
twh : (string_list) determ. % headers
tww : (integer_list) determ. % pixel widths
twa : (integer_list) determ. % alignments (-1:L, 0:C, 1:R)
twr : (integer_list) determ. % read-only flags (0/1)
twb : (string_list) determ. % buttons (possibly empty)
tedCount : integer := 0.

class predicates
tableed_class_handler : vpiDomains::ehandler.
%no_Callback : vpiTableed::tableedit_CALLBACK .
makeDataOfStrings : (string_list_list,color_list,vpiTableed::tableedit_DATA) determ (i,i,o).
rectangle : (integer,integer,integer_list,string_list,vpiDomains::rct) determ (i,i,i,i,o).
getCols : () -> vpiTableed::tableedit_COLUMNLIST determ.

clauses
classInfo(className, classVersion).

getCount()=N:-N=tedCount.
getButtons()=L:-twb(L).
% ----------------------------------------------------------------------------------------------
% eventhandler for table window class
% ----------------------------------------------------------------------------------------------
class predicates
getColAlign : (integer,vpiTableed::tableedit_COLUMNPROPERTY) determ (i,o).
clauses
 
tableed_class_handler(W,vpiDomains::e_Create(_))=0 :-
    Cols=getCols(),
    vpiTableed::tableedit_init(W,[],Cols,[]),
    !.
tableed_class_handler(W,Event)=0 :-
    vpiTableed::tableedit_handleevent(W,Event),
    !.

getCols()=Cols:-
    twr(ReadOnly),
    twh(Headers),
    tww(Widths),
    twa(Alignments),
    getColsX(ReadOnly,Headers,Widths,Alignments,Cols),
    !.
    
getColsX([],_,_,_,[]) :-
    !.
getColsX([R|Rs],[H|Hs],[W|Ws],[A|As],[Col|Cols]) :-    
    getColAlign(A,Align),
    Col=vpiTableed::column([vpiTableed::title(H),vpiTableed::readonly(R),vpiTableed::width(W),Align]),
    getColsX(Rs,Hs,Ws,As,Cols),
    !.
    
getColAlign(-1,vpiTableed::alignment(vpiGrid::grid_left_align())).    
getColAlign(1,vpiTableed::alignment(vpiGrid::grid_right_align())).    
getColAlign(0,vpiTableed::alignment(vpiGrid::grid_center())).    
% ----------------------------------------------------------------------------------------------
% create table window
% ----------------------------------------------------------------------------------------------
createTableWin(TW,Title,Rows,ReadOnly,Headers,Widths,Alignments,Data,Buttons,Sort,EH,Cascade)=W :-
    retractall(twr(_)),
    retractall(twh(_)),
    retractall(tww(_)),
    retractall(twa(_)),
    retractall(twb(_)),
    set::size(ReadOnly,NCols),
    assert(twr(ReadOnly)),
    assert(twh(Headers)),
    assert(tww(Widths)),
    assert(twa(Alignments)),
    assert(twb(Buttons)),
    N=tedCount,
    M=N+1,
    tedCount:=M,
    Ns=toString(M),
    TableClass=string::concat("tableed_class",Ns),
    vpi::classCreate(TableClass,tableed_class_handler),   
    rectangle(Cascade,Rows,Widths,Buttons,R),
    W=vpi::winCreate(vpiDomains::w_TopLevel,R,Title,vpiDomains::noMenu(),TW,
        [vpiDomains::wsf_TitleBar,vpiDomains::wsf_Minimize,vpiDomains::wsf_BlackRect,vpiDomains::wsf_Close,
         vpiDomains::wsf_ClipSiblings,vpiDomains::wsf_SizeBorder],EH,0),
    C=guix::succ(W),
   	Font=vpi::fontCreate(vpiDomains::ff_Helvetica,[],8),
  	vpi::winSetFont(W,Font),
  	vpi::winSetFont(C,Font),
  	setColProp(C,NCols,Sort),
    vpiTableed::tableedit_setalldata(C,Data),
    !.

setColProp(_,0,_) :-
    !.
setColProp(C,NCols,Sort) :-
    Sort=1,
    vpiTableed::tableedit_setcolumnproperties(C,NCols,[vpiTableed::sortmode(vpiTableed::ascending([]))]),  	
    Next=NCols-1,
    setColProp(C,Next,Sort),
    !
;   
    Sort<>1,
    vpiTableed::tableedit_setcolumnproperties(C,NCols,[vpiTableed::sortmode(vpiTableed::none)]),  	
    Next=NCols-1,
    setColProp(C,Next,Sort),
    !.
% ----------------------------------------------------------------------------------------------
% makeDataOfPackedStrings/Inv
% ----------------------------------------------------------------------------------------------
makeDataOfPackedStrings(PSL,Colors,Data) :-
    sx::unpackList(PSL,"||",RowList),
    makeDataOfStrings(RowList,Colors,Data),
    !.

makeDataOfPackedStringsInv(PSL,Colors,Data) :-
    sx::unpackList(PSL,"||",RowList),
    set::invertToColMajorOrder(RowList,ColList),
    makeDataOfStrings(ColList,Colors,Data),
    !.
% ----------------------------------------------------------------------------------------------
% makeDataOfCmp
% ----------------------------------------------------------------------------------------------
class predicates
makeDataOfTerms : (sterml,vpiTableed::tableedit_COLUMNDATA) determ (i,o).
clauses

makeDataOfCmp([],[]) :-
    !.
makeDataOfCmp([cmp(_,Cols)|Rows],[Data|Datas]) :-
    makeDataOfTerms(Cols,Data),
    makeDataOfCmp(Rows,Datas),
    !.

makeDataOfTerms([],[]) :-
    !.
makeDataOfTerms([str(V)|Terms],[vpiTableed::s(V)|Cells]) :- makeDataOfTerms(Terms,Cells),!.
makeDataOfTerms([int(V)|Terms],[vpiTableed::i(V)|Cells]) :- makeDataOfTerms(Terms,Cells),!.
makeDataOfTerms([fix(V)|Terms],[vpiTableed::r(V)|Cells]) :- makeDataOfTerms(Terms,Cells),!.
% ----------------------------------------------------------------------------------------------
% makeDataOfStrings
% ----------------------------------------------------------------------------------------------
class predicates
makeDataOfStringsCol : (string_list,color_list,vpiTableed::tableedit_COLUMNDATA) determ (i,i,o).
clauses

makeDataOfStrings([],_,[]) :-
    !.
makeDataOfStrings([Col|Cols],Colors,[DCol|DCols]) :-
    makeDataOfStringsCol(Col,Colors,DCol),
    makeDataOfStrings(Cols,Colors,DCols),
    !.

makeDataOfStringsCol([],_,[]) :-
    !.
makeDataOfStringsCol([Cell|Cells],[Color|Colors],[vpiTableed::sc(Cell,Color)|DCells]) :-
    makeDataOfStringsCol(Cells,Colors,DCells),
    !.
% ----------------------------------------------------------------------------------------------
% rectangle
% ----------------------------------------------------------------------------------------------
rectangle(Cascade,Rows,Widths,Buttons,R) :-
    editorWindow::getMax(Mx,My),
    guix::getPixelAdjust(_,Dy),
    Ry=((Rows*18+37)*Dy) div 100,
    Ay=set::min([My,Ry]),
    Rx=set::sum(Widths)+70,
    Ax=set::min([Mx,Rx]),
    Re=editorWindow::getEditRect(Cascade),
    Re=vpiDomains::rct(X,Y,_,_),
    set::size(Buttons,N),
    set::setMin([N,1],Nm), 
    Hx=X+Ax,
    Hy=Y+Ay+Nm*20,
    R=vpiDomains::rct(X,Y,Hx,Hy),
    !.
% ----------------------------------------------------------------------------------------------
% makeCell
% ----------------------------------------------------------------------------------------------
makeCell(int(_),NL,vpiTableed::sc(NL,Col)) :- Col=globalVars::get_i("col_int"),!.
makeCell(fix(_),NL,vpiTableed::sc(NL,Col)) :- Col=globalVars::get_i("col_fix"),!.
makeCell(str(_),NL,vpiTableed::sc(NLu,Col)) :- Col=globalVars::get_i("col_str"),sx::sxUnquote(NL,NLu),!.
makeCell(atom(_),NL,vpiTableed::sc(NL,Col)) :- Col=globalVars::get_i("col_atom"),!.
makeCell(nill,_,vpiTableed::sc("",Col)) :- Col=globalVars::get_i("col_str"),!.
makeCell(_,NL,vpiTableed::sc(NL,Col)) :- Col=globalVars::get_i("col_str"),!.

makeCell_s(S,vpiTableed::sc(Su,Col)) :- Col=globalVars::get_i("col_str"),sx::sxUnquote(S,Su),!.
makeCell_i(I,vpiTableed::ic(I,Col)) :- Col=globalVars::get_i("col_int"),!.
end implement ted
