/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement termDomain
open core

constants
className = "com/rudstrand/termDomain/termDomain".
classVersion = "1.0".
brA="{A".
brS="{S".
brF="{F".
brN="{N".
brI="{I".
brV="{V".
brC="{C".
brL="{L".
brT="{T".
br0="{0".
br="{".
bre="}".
cbr='{'.
cbre='}'.

class predicates
fstrSwitch : (sterm,sterm) determ (i,o). % fstr(I,"p3_din",...) -> int(I)

clauses
classInfo(className, classVersion).
% -----------------------------------------------------------------------------------------------
% isBasicTerm/F
% -----------------------------------------------------------------------------------------------
class predicates
isBasicTermX : (sterml) determ.
clauses

isBasicTerm(T) :-
    isBasicTermF(T),
    !
;
    S=toString(T),
	pfeil::syserr("isBasicTerm","Invalid basic term",[S]),!,fail.	
    
isBasicTermF(T) :-
    T=cmp(_,L),
    isBasicTermX(L),
    !.
    
isBasicTermX([]) :- !.
isBasicTermX([str(_)|T]) :- isBasicTermX(T),!.    
isBasicTermX([atom(_)|T]) :- isBasicTermX(T),!.    
isBasicTermX([int(_)|T]) :- isBasicTermX(T),!.    
isBasicTermX([fix(_)|T]) :- isBasicTermX(T),!.    
% -----------------------------------------------------------------------------------------------
% isTermString/Indirect
% -----------------------------------------------------------------------------------------------
isValidTermString(S) :- _=fromTermStringF(S),!.

isTermString(S) :-
    N=string::length(S),
    N>2,
    string::frontchar(S,F,_),
    F=cbr,
    P=N-1,
    string::subString(S,P,1)=bre,
    !.

isTermStringIndirect(S) :-
    string::length(S)>8,
    string::front(S,2,Tag,R2),
    Tag=brT,
    string::front(R2,1,LcLetter,R3),
    string::hasAlpha(LcLetter),
    string::isLowercase(LcLetter),
    Be=string::concat(br,LcLetter,bre),
    Pos=string::search(R3,Be,string::caseSensitive),
    string::front(R3,Pos,_,R4),
    string::length(R4)=3,
    !.
% -----------------------------------------------------------------------------------------------
% fromTermString/Indirect
% -----------------------------------------------------------------------------------------------
class predicates
ftsElementF : (string,string,sterm,string) determ (i,i,o,o).
ftsTermList : (string,sterml,string) determ (i,o,o).
ftsList : (string,sterm,string) determ (i,o,o).   
clauses

fromTermStringIndirect(S)=Term :-
    Term=fromTermStringIndirectF(S),
    !
;
	pfeil::syserr("fromTermStringIndirect","Invalid indirect term-string",[S]),!,fail.	
    
fromTermStringIndirectF(S)=Term :-
    string::length(S)>8,
    string::front(S,2,Tag,R2),
    Tag=brT,
    string::front(R2,1,LcLetter,R3),
    string::hasAlpha(LcLetter),
    string::isLowercase(LcLetter),
    Be=string::concat(br,LcLetter,bre),
    Pos=string::search(R3,Be,string::caseSensitive),
    string::front(R3,Pos,S2,R4),
    string::length(R4)=3,
    Term=fromTermStringF(S2),
    !.

fromTermString(S)=_ :-
   string::length(S)<3,
   pfeil::syserr("fromTermString","Term-strings must be at least three characters long",[S]),!,fail.	
fromTermString(S)=Term :- 
    Term=fromTermStringF(S),
    !
;
    pfeil::syserr("fromTermString","Invalid term-string",[S]),!,fail.	
    
fromTermStringF(S)=Term :- 
    string::front(S,2,Tag,R2),
    ftsElementF(Tag,R2,Term,R),
    R="",
    !.

ftsElementF(brS,S,str(V),R) :-
    Pos=string::searchchar(S,cbre,string::caseSensitive),
    string::front(S,Pos,V,R2),
    string::frontchar(R2,_,R),
    !.
ftsElementF(brA,S,atom(V),R) :-
    Pos=string::searchchar(S,cbre,string::caseSensitive),
    string::front(S,Pos,V,R2),
    string::frontchar(R2,_,R),
    !.
ftsElementF(brV,S,var(V),R) :-
    Pos=string::searchchar(S,cbre,string::caseSensitive),
    string::front(S,Pos,V,R2),
    string::frontchar(R2,_,R),
    !.
ftsElementF(brI,S,int(V),R) :-
    Pos=string::searchchar(S,cbre,string::caseSensitive),
    string::front(S,Pos,Vs,R2),
    V=toTerm(Vs),
    string::frontchar(R2,_,R),
    !.
ftsElementF(brN,S,fix(V),R) :-
    Pos=string::searchchar(S,cbre,string::caseSensitive),
    string::front(S,Pos,Vs,R2),
    V=toTerm(Vs),
    string::frontchar(R2,_,R),
    !.
ftsElementF(brC,S,cmp(H,T),R) :-
    string::front(S,2,Tag,R2),
    Tag=brS,
    ftsElementF(Tag,R2,Ht,R3),
    Ht=str(H),
    ftsTermList(R3,T,R), % scans up to and including the closing bracket for the compound term
    !.
ftsElementF(brL,S,list(H,T),R) :-
    string::front(S,2,Tag,R2),
    ftsElementF(Tag,R2,H,R3),
    ftsList(R3,T,R),
    !.
ftsElementF(brT,S,str(T),R) :-
    string::front(S,1,LcLetter,_),
    string::hasAlpha(LcLetter),
    string::isLowercase(LcLetter),
    Be=string::concat(br,LcLetter,bre),
    Pos=string::search(S,Be,string::caseSensitive),
    string::front(S,Pos,T2,R3),
    string::front(R3,3,_,R),
    T=string::concat(brT,T2,br,LcLetter,bre),
    !.
ftsElementF(brF,S,fstr(Vs,Vd,Vf,Vi),R) :-
    Pos=string::searchchar(S,cbre,string::caseSensitive),
    string::front(S,Pos,Vs,R2),
    string::frontchar(R2,_,R3),
    string::front(R3,2,Tag3,R4),
    ftsElementF(Tag3,R4,Td,R5),
    Td=str(Vd),
    string::front(R5,2,Tag4,R6),
    ftsElementF(Tag4,R6,Tf,R7),
    Tf=str(Vf),
    string::front(R7,2,Tag5,R8),
    ftsElementF(Tag5,R8,Ti,R),
    Ti=int(Vi),
    !.
ftsElementF(br0,S,nill,R) :-
    string::frontchar(S,B,R),
    B=cbre,
    !.
    
ftsTermList(S,[],R) :-
    string::frontchar(S,B,R),
    B=cbre,
    !.
ftsTermList(S,[H|T],R) :-
   string::front(S,2,Tag,R2),
   ftsElementF(Tag,R2,H,R3),
   ftsTermList(R3,T,R),
   !.  

ftsList(S,nill,R) :-
    string::frontchar(S,B,R),
    B=cbre,
    !.
ftsList(S,list(H,T),R) :-
   string::front(S,2,Tag,R2),
   ftsElementF(Tag,R2,H,R3),
   ftsList(R3,T,R),
   !.     
% -----------------------------------------------------------------------------------------------
% toTermString/Indirect
% -----------------------------------------------------------------------------------------------
class predicates
toTermStringX : (sterm) -> string determ.
toTsTermList : (sterml,string) determ (i,o).
toTsList : (sterm,string) determ (i,o).
clauses

toTermStringIndirect(Letter,Term)=S :-
    string::length(Letter)=1,
    S1=toTermStringX(Term),
    Lc=string::toLowercase(Letter),
    S=string::concat("{T",Lc,S1,"{",Lc,"}"),
    %checkLength(S),
    !.

toTermString(T)=S :- 
    S=toTermStringX(T),
    %checkLength(S),
    !
;
    pfeil::syserr("toTermString","Error in term-string creation"),!,fail.	

toTermStringX(atom(T))=S :- 
    S=string::concat(brA,T,bre),
    !.
toTermStringX(str(T))=S :- 
    string::length(T)>2,
    string::front(T,2,F,_),
    F=brT,
    S=T,
    !
;    
    S=string::concat(brS,T,bre),
    !.
toTermStringX(var(T))=S :- 
    S=string::concat(brV,T,bre),
    !.
toTermStringX(nill)=S :- 
    S=string::concat(br0,bre),
    !.
toTermStringX(int(T))=S :- 
    Ts=toString(T),
    S=string::concat(brI,Ts,bre),
    !.
toTermStringX(fix(T))=S :- 
    Ts=toString(T),
    S=string::concat(brN,Ts,bre),
    !.
toTermStringX(cmp(H,T))=S :- 
    toTsTermList(T,Ts), 
    S=string::concat(brC,brS,H,bre,Ts,bre),
    !.   
toTermStringX(list(H,T))=S :-
    Hs=toTermStringX(H),
    toTsList(T,Ts),
    S=string::concat(brL,Hs,Ts,bre),
    !.
toTermStringX(fstr(T,D,F,W))=S :-
    Ws=toString(W),
    S=string::concatList([brF,T,bre,brS,D,bre,brS,F,bre,brI,Ws,bre]),
    !.

toTsTermList([],"") :-
    !.
toTsTermList([H|T],S) :- 
    Hs=toTermStringX(H),
    toTsTermList(T,Ts),
    S=string::concat(Hs,Ts),
    !.

toTsList(nill,"") :-
    !.
toTsList(list(H,T),S) :-
    Hs=toTermStringX(H),
    toTsList(T,Ts),
    S=string::concat(Hs,Ts),
    !.
% -----------------------------------------------------------------------------------------------
% fstr_str
% -----------------------------------------------------------------------------------------------
fstr_str(fstr(S,_,_,_),S) :- !.
% -----------------------------------------------------------------------------------------------
% fSwitch
% -----------------------------------------------------------------------------------------------
class predicates
fSwitchList : (sterml,sterm) determ (i,o).
clauses

fSwitch(Cf,C) :- 
  Cf=fstr(_,_,_,_), fstrSwitch(Cf,C), !
; Cf=fint(V,_), C=int(V), !
; Cf=ffix(V,_), C=fix(V), !
; Cf=fvar(V,_), C=var(V), !
; Cf=cmp(_,ListMembers), fSwitchList(ListMembers,C), !.

fSwitchList([],nill) :- !.
fSwitchList([H|T],list(Hs,Ts)) :- fSwitch(H,Hs), fSwitchList(T,Ts),!.
% -----------------------------------------------------------------------------------------------
% fstrSwitch/R
% -----------------------------------------------------------------------------------------------
% switches from fstr...-representation to domain-specific representation
%
fstrSwitch(fstr(S,"p3_din",_,_),int(V)) :- V=sx::toInt(S),!.
fstrSwitch(fstr(S,"p3_did",_,_),int(V)) :- V=sx::toInt(S),!.
fstrSwitch(fstr(S,"p3_dnu",_,_),fix(V)) :- V=sx::toReal(S),!.
fstrSwitch(fstr(S,"p3_dinc",_,_),fix(V)) :- V=sx::toReal(S),!.
fstrSwitch(fstr(S,"p3_ddec",_,_),fix(V)) :- V=sx::toReal(S),!.
fstrSwitch(fstr(S,"p3_dpr",_,_),fix(V)) :- V=sx::toReal(S),!.
fstrSwitch(fstr(S,"p3_dcf",_,_),fix(V)) :- V=sx::toReal(S),!.
fstrSwitch(fstr(S,"p3_dit",_,_),fix(V)) :- V=sx::toReal(S),!.
fstrSwitch(fstr(S,"p3_dif",_,_),atom(S)) :- !.
fstrSwitch(fstr(S,"p3_dli",_,_),str(S)) :- !.
fstrSwitch(fstr(S,"p3_dsy",_,_),str(S)) :- !.
fstrSwitch(fstr(S,"p3_dkb",_,_),str(S)) :- !.
fstrSwitch(fstr(S,"p3_dxt",_,_),str(S)) :- !.
fstrSwitch(fstr(S,"p3_dxd",_,_),str(S)) :- !.
fstrSwitch(fstr(S,"p3_dxdt",_,_),str(S)) :- !.
fstrSwitch(fstr(S,"p3_dxdr",_,_),str(S)) :- !.
fstrSwitchR(fstr(S,"p3_din","",L),int(V)) :- S=toString(V),L=string::length(S),!.
fstrSwitchR(fstr(S,"p3_dnu","",L),fix(V)) :- S=toString(V),L=string::length(S),!.
fstrSwitchR(fstr(S,"p3_dif","",L),atom(S)) :- L=string::length(S),!.
fstrSwitchR(fstr(S,"p3_dli","",L),str(S)) :- L=string::length(S),!.
% -----------------------------------------------------------------------------------------------
% fterm
% -----------------------------------------------------------------------------------------------
fterm(str(S),D,F,W,fstr(S,D,F,W)) :- !.
fterm(int(I),D,F,W,fstr(S,D,F,W)) :- S=toString(I),!.
fterm(fix(R),D,F,W,fstr(S,D,F,W)) :- S=toString(R),!.
fterm(atom(A),D,F,W,fstr(A,D,F,W)) :- !.
% --------------------------------------------------------------------------
% listBR
% --------------------------------------------------------------------------
listBR(nill,[]) :- !.
listBR(list(H,T),[H|Tb]) :- listBR(T,Tb),!.
% -----------------------------------------------------------------------------------------------
% cl.._list..
% -----------------------------------------------------------------------------------------------
% clause list conversion
%
cl_list([],nill) :- 
	!.
cl_list([H|T],list(H,Tm)) :- 
	cl_list(T,Tm),!.
cl_listR([],nill) :- 
	!.
cl_listR([H|T],list(H,Tm)) :- 
	cl_listR(T,Tm),!.
cls_list([],nill) :- 
	!.
cls_list([H|T],list(str(H),Tm)) :- 
	cls_list(T,Tm),!.
cls_listR([],nill) :- 
	!.
cls_listR([H|T],list(str(H),Tm)) :- 
	cls_listR(T,Tm),!.
cla_list([],nill) :- 
	!.
cla_list([H|T],list(atom(H),Tm)) :- 
	cla_list(T,Tm),!.
cla_listR([],nill) :- 
	!.
cla_listR([H|T],list(atom(H),Tm)) :- 
	cla_listR(T,Tm),!.
cli_listR([],nill) :- 
	!.
cli_listR([H|T],list(int(H),Tm)) :- 
	cli_listR(T,Tm),!.
% ----------------------------------------------------------------------------------------------
% list_conc
% ----------------------------------------------------------------------------------------------
list_conc(nill,"") :- 
	!.
list_conc(list(str(S),nill),S) :- 
	!.
list_conc(list(str(S1),T),S) :- 
	list_conc(T,S2), 
	sx::concat([S1," ",S2],S),!.
% -----------------------------------------------------------------------------------------------
% str_SLIST; str_SLIST_t; var_SLIST
% -----------------------------------------------------------------------------------------------
str_slist([],[]) :- 
	!.
str_slist([str(S)|T],[S|Tm]) :- 
	str_slist(T,Tm),
	!.
str_slist([fstr(S,"p3_dany","",0)|T],[S|Tm]) :- 
	str_slist(T,Tm),
	!.
str_slist(L,[S|Tm]) :- 
	L=[fstr(S,_,_,_)|T],
	str_slist(T,Tm),
	!.
str_slist(_,_) :-
	pfeil::syserr("termDomain::str_slist","Other error"),
	!,fail.
var_SLIST([],[]) :- 
	!.
var_SLIST([var(S)|T],[S|Tm]) :- 
	var_SLIST(T,Tm),!.
var_SLIST([fvar(S,"p3_dany")|T],[S|Tm]) :- 
	var_SLIST(T,Tm),!.
var_SLIST(_,_) :-
	pfeil::syserr("var_SLIST","Other error"),
	!,fail.
var_SLIST_R([],[]) :- 
	!.
var_SLIST_R([var(S)|T],[S|Tm]) :- 
	var_SLIST_R(T,Tm),!.
var_SLIST_R([fvar(S,"p3_dany")|T],[S|Tm]) :- 
	var_SLIST_R(T,Tm),!.
var_SLIST_R(_,_) :-
	pfeil::syserr("termDomain::var_SLIST","Other error"),
	!,fail.
% -----------------------------------------------------------------------------------------------
% sterm_strR
% -----------------------------------------------------------------------------------------------
% converts between sterm and string format
%
sterm_strR(T,S) :-
	S="true",
	T=atom(S),
	!
;	
	S="false",
	T=atom(S),
	!
;	
	T=toTerm(S),
	!.
% -----------------------------------------------------------------------------------------------
% realType
% -----------------------------------------------------------------------------------------------
realType(fix(F),F) :- 
	!.
realType(int(I),F) :- 
	F=I,
	!.
realType(str(S),F) :-
    F=sx::toReal(S),
	!.
	
intType(int(I),I) :- 
	!.
intType(str(S),I) :-
	I=sx::toInt(S),
	!.
/*
% -----------------------------------------------------------------------------------------------
% setTsMaxLength
% -----------------------------------------------------------------------------------------------
class facts
tsMaxLength : integer:=10000. % max. length of term-strings 

setTsMaxLength : (integer).

setTsMaxLength(N) :- tsMaxLength:=N,!.

checkLength : (string) determ.
checkLength(S) :-
    string::length(S)<=tsMaxLength,
    !
;
    M=string::length(S),
    tsMaxLength<M,
    Ns=toString(tsMaxLength),
    Ms=toString(M),
    Mess=string::concat("Term-string with length ",Ms," exceeds max. ",Ns,"!"),
    vpiCommonDialogs::ask(Mess,"Do you want to increase maximum?",["Yes","No"])=vpiCommonDialogs::resp_default,
    tsMaxLength:=M,
    !.
*/    
end implement termDomain
