/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement cluster
open core,termDomain

constants
className = "com/rudstrand/Cluster/cluster".
classVersion = "1.0".

class predicates
createNewMember : () determ.
invokeKb : (string Qkb) determ.
processRecord : (odbcConnection,string,count,integer,integer,string,string) determ.
streamIsEmpty : (string) determ.
deriveCluster : (string,string,string,string) determ.    

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% startCluster
% ----------------------------------------------------------------------------------------------
class predicates
continueCluster : (string) determ.
clauses

startCluster(Cluster,Kb,Net,Share,Stream) :-
    datetime::tsNow(T1),
    programControl::repeat,
        readStream(Stream),
        continueCluster(Stream),
    dlgIF::msgS("Knowledge stream is processed!"),
    deriveCluster(Kb,Cluster,Net,Share),
    datetime::tsNow(T2),
    datetime::tsDiff(T1,T2,DD,HH,MM,SS),
	datetime::timDurText(DD,HH,MM,SS,Sd),
    invokeKb(Kb),
	Mess=string::concat("Cluster processing was completed during ",Sd,"!"),
    dlgIF::dlgNote(0,Mess),
    !
;
    Cluster=getCluster(),
    Kb=getMember(Cluster,0),
    dlgIF::dlgNote(0,"Cluster processing failed!"),
    invokeKb(Kb),
    !.
       
continueCluster(Stream) :-
    streamIsEmpty(Stream),
    !
;        
    createNewMember,
    !,fail.
% ----------------------------------------------------------------------------------------------
% deriveCluster
% ----------------------------------------------------------------------------------------------
class predicates
nextKb : (string,count,integer) -> string determ. 
deriveClusterX : (string,string,string,string,integer).
subDerPub : (string,string,string) determ.
checkShareChanges : (string,string,string,string,count) determ.
continueNext : (string,string,string,string,string,count) determ.    
clauses

deriveCluster(Kb,Cluster,Net,Share) :-
    Con=sqlBase::getConnection(),
    Cluster=getCluster(),
    Sql1=string::concat("select max(member_no) from p3publications.cluster_member where cluster_name='",Cluster,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,    
    Max=Select1:getI(1,1),
    deriveClusterX(Kb,Cluster,Net,Share,Max),
    !.
    
deriveClusterX(Kb,Cluster,Net,Share,Max) :-    
    invokeKb(Kb),
    subDerPub(Kb,Cluster,Net),
    % main loop
    C=count::new(0),
    Cd=count::new(1),
    programControl::repeat,
        KbNext=nextKb(Cluster,C,Max),
        invokeKb(KbNext),
        checkShareChanges(KbNext,Cluster,Net,Share,Cd),
        !
;
    dlgIF::dlgNote(3,"Initial processing of member 0 failed!"),
    !.
    
subDerPub(Kb,Cluster,Net) :-        
    Con=sqlBase::getConnection(),
    sqlSt::subscribeOpen(Net),
            Rep=rep::newReport(4,"Derivation of All Sentences",70),
	der::rederiveAll(Rep,Nt,Ns),
            Nts=toString(Nt),
            Nss=toString(Ns),
            Mess=string::concatList(["Generated ",Nss," sentences of ",Nts," types."]),
            dlgIF::msgS(Mess),
            dlgIF::msgStatus(Mess),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    Sql1=string::concatList(["update p3publications.cluster_member set last_derivation='",Time,"' where kb='",Kb,"' and cluster_name='",Cluster,"'"]),
    sqlTools::executeDirect(Con,Sql1),    
    sqlSt::publishOpen(Net),
	datetime::tsNow(Ts2),
    datetime::tsSQL(Ts2,Time2),
    Sql2=string::concatList(["update p3publications.cluster_member set last_share='",Time2,"' where kb='",Kb,"' and cluster_name='",Cluster,"'"]),
    sqlTools::executeDirect(Con,Sql2),  
    !
;
    Mess=string::concat("Derivation-step failed for KB: ",Kb,"!"),
    dlgIF::dlgNote(3,Mess),!,fail.

nextKb(Cluster,C,Max)=Kb :-
    C:inc(Next),
    Next>Max,
    C:set(0),
    Kb=getMember(Cluster,0),
    !
;
    Next=C:get(),
    Next<=Max,
    Kb=getMember(Cluster,Next),
    !.  

checkShareChanges(Kb,Cluster,Net,Share,Cd) :-
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select last_derivation from p3publications.cluster_member where cluster_name='",Cluster,"' and kb='",Kb,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,    
    Td=Select1:getDT(1,1),
    Sql2=string::concat("select max(created) from ",Share),
    Select2=sqlSelect::new(Con,Sql2),
    N2=Select2:getNRows(),
    N2>0,    
    Ts=Select2:getDT(1,1),
    !,continueNext(Kb,Cluster,Net,Td,Ts,Cd),
    M=Cd:get(),
    Ms=toString(M),
    Mess=string::concat("Cluster processing involved ",Ms," derivation steps!"),
    dlgIF::msgS(Mess),
    !
;
    Mess=string::concat("Cluster open share changes check failed for KB: ",Kb,"!"),
    dlgIF::dlgNote(3,Mess),
    !.    

% succeeds if all members are complete (or error), else fails such that processing of next member is triggered
%
continueNext(Kb,Cluster,Net,Td,Ts,Cd) :-
    Td<Ts,
    Con=sqlBase::getConnection(),
    Sql2=string::concatList(["update p3publications.cluster_member set complete=0 where cluster_name='",Cluster,"'"]),
    sqlTools::executeDirect(Con,Sql2),    
    subDerPub(Kb,Cluster,Net),
            Cd:inc(N),
            Ns=toString(N),
            Mess=string::concat("Completed derivation ",Ns," based on KB: ",Kb),
            dlgIF::msgS(Mess),
    !,fail
;    
    Td>=Ts,
    Con=sqlBase::getConnection(),
    Sql2=string::concatList(["update p3publications.cluster_member set complete=1 where cluster_name='",Cluster,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql2),   
    Sql3=string::concat("select kb from p3publications.cluster_member where complete=0 and cluster_name='",Cluster,"'"),
    Select3=sqlSelect::new(Con,Sql3),
    !,Select3:getNRows()=0,
    !
;
    Mess=string::concat("Cluster processing failed to continue for KB: ",Kb,"!"),
    dlgIF::dlgNote(3,Mess),
    !.
% ----------------------------------------------------------------------------------------------
% getStream
% ----------------------------------------------------------------------------------------------
getStream()=Stream :-
    Con=sqlBase::getConnection(),
    Cluster=getCluster(),
    Sql1=string::concat("select knowledge_stream from p3publications.cluster where cluster_name='",Cluster,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,    
    Stream=Select1:getS(1,1),
    !.

getStream(Cluster)=Stream :-
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select knowledge_stream from p3publications.cluster where cluster_name='",Cluster,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,    
    Stream=Select1:getS(1,1),
    !.
% ----------------------------------------------------------------------------------------------
% getNet
% ----------------------------------------------------------------------------------------------
getNet()=Net :-
    Con=sqlBase::getConnection(),
    Cluster=getCluster(),
    Sql1=string::concat("select net_name from p3publications.cluster where cluster_name='",Cluster,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,    
    Net=Select1:getS(1,1),
    !.
% ----------------------------------------------------------------------------------------------
% getCluster
% ----------------------------------------------------------------------------------------------
getCluster()=Cluster :-
    Con=sqlBase::getConnection(),
    Kb=sqlBase::getDbKb(),
    Sql1=string::concat("select cluster_name from p3publications.cluster_member where kb='",Kb,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,
    Cluster=Select1:getS(1,1),
    !.
% ----------------------------------------------------------------------------------------------
% getMember
% ----------------------------------------------------------------------------------------------
getMember(Cluster,MemberNo)=Qkb :-
    Con=sqlBase::getConnection(),
    Ns=toString(MemberNo),
    Sql1=string::concat("select kb from p3publications.cluster_member where member_no=",Ns," and cluster_name='",Cluster,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,
    Qkb=Select1:getS(1,1),
    !.
% ----------------------------------------------------------------------------------------------
% readStream
% ----------------------------------------------------------------------------------------------
class predicates
getKb : () -> string determ.
clauses

readStream(Stream) :-
    C=count::new(),
    Kb=getKb(),
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select k_format,k_type,knowledge from ",Stream," where processed=0"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    Row=programControl::sequence_nd(1,N1,1),
        F=Select1:getI(Row,1),
        T=Select1:getI(Row,2),
        K=Select1:getS(Row,3),
        processRecord(Con,Stream,C,F,T,K,Kb),
        fail
;
    !.

getKb()=Kb :-
    dbBase::kbIsSql,
    Kb=sqlBase::getDbKb(),
    !
;
    Kb=dbBase::dbName(),
    !.
% ----------------------------------------------------------------------------------------------
% streamIsEmpty
% ----------------------------------------------------------------------------------------------
streamIsEmpty(Stream) :-
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select k_format from ",Stream," where processed=0"),
    Select1=sqlSelect::new(Con,Sql1),
    Select1:getNRows()=0,
    !.
% ----------------------------------------------------------------------------------------------
% invokeKb
% ----------------------------------------------------------------------------------------------
invokeKb(Kb) :-
    dbOpen::closeKB(0,-1,-1,-1),
    Kbf=string::concat(Kb,".sql"),
    dbOpen::openKB_Step3(1,Kbf),
    !.
% ----------------------------------------------------------------------------------------------
% createNewMember
% ----------------------------------------------------------------------------------------------
class predicates
createClusterMember : (string) -> string determ.
clauses

createNewMember() :-
    Cluster=getCluster(),
    % ensure caches are uptodate for init of new KB
    cache::loadTmConditionally,
    cache::loadSmConditionally,
    dbOpen::closeKB(0,-1,-1,-1),
    NewKb=createClusterMember(Cluster),
    KbName=string::concat(NewKb,".sql"),
    dbOpen::create(KbName),
    proc::loadStandard(2),
    % save caches from invoking KB to new KB
    cache::saveSm,
    cache::saveTm,
    !.

createClusterMember(Cluster)=Qkb :-
    Con=sqlBase::getConnection(),
    % generate new DB name 
    Sql1=string::concat("select max(member_no) from p3publications.cluster_member where cluster_name='",Cluster,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    Select1:getNRows()>0,
    M=Select1:getI(1,1),
    N=M+1,
    Ns=toString(N),
    Sql2=string::concat("select kb from p3publications.cluster_member where cluster_name='",Cluster,"' and member_no=0"),
    Select2=sqlSelect::new(Con,Sql2),
    Select2:getNRows()>0,
    Qkb0=Select2:getS(1,1),
    L=sx::unpack(Qkb0,"."),
    L=[Db,Kb],
    NewDb=string::concat(Db,"_",Ns),
    Qkb=string::concat(NewDb,".",Kb),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,T),
    Sql3=string::concatList(["insert p3publications.cluster_member (cluster_name,kb,member_no,last_share,last_derivation,complete) values ('",Cluster,"','",Qkb,"',",Ns,",'",T,"','",T,"',0)"]),
    sqlTools::executeDirect(Con,Sql3),
    % document that new Kb publishes and subscribes to the open share
    Kb0=getMember(Cluster,0),
    Sql7=string::concatList(["select net_name from p3publications.cluster where cluster_name='",Cluster,"'"]),
    Select7=sqlSelect::new(Con,Sql7),
    Select7:getNRows()>0,
    Net=Select7:getS(1,1),   
    Sql4=string::concatList(["select mode from p3publications.publication where kb='",Kb0,"' and net='",Net,"' and st=''"]),
    Select4=sqlSelect::new(Con,Sql4),
    Select4:getNRows()>0,
    Mode=Select4:getI(1,1),
    ModeS=toString(Mode),
    Sql5=string::concatList(["insert p3publications.subscription (net,kb,mode,last_sub,st) values ('",Net,"','",Qkb,"',",ModeS,",'",T,"','')"]),           
    sqlTools::executeDirect(Con,Sql5),
    Sql6=string::concatList(["insert p3publications.publication (net,kb,mode,last_pub,st,st_format,table_name) values ('",Net,"','",Qkb,"',",ModeS,",'",T,"','',0,'')"]),            
    sqlTools::executeDirect(Con,Sql6),
            tt::tt_s(1,"Created new cluster member",Qkb),
    !
;
    dlgIF::dlgNote(1,"Failed to create new cluster member!"),
    !,fail.
% ----------------------------------------------------------------------------------------------
% reset
% ----------------------------------------------------------------------------------------------
class predicates
reset_1 : (string).
reset_2 : (integer).
clauses

reset(Cluster,Share) :-
    reset_1(Cluster),
    reset_2(Share),
    Con=sqlBase::getConnection(),
    Sql1=string::concatList(["delete from p3publications.cluster_member where cluster_name='",Cluster,"' and member_no>0"]),
    sqlTools::executeDirect(Con,Sql1),    
    Stream=getStream(Cluster),
    Sql2=string::concatList(["update ",Stream," set processed=0,kb=''"]),
    sqlTools::executeDirect(Con,Sql2),
    !
;
    !.
    
reset_1(Cluster) :-    
    Con=sqlBase::getConnection(),
    Sql1=string::concatList(["select kb from p3publications.cluster_member where cluster_name='",Cluster,"' and member_no>0"]),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    Row=programControl::sequence_nd(1,N1,1),
        Kb=Select1:getS(Row,1),
        Sql2=string::concatList(["delete from p3publications.publication where kb='",Kb,"'"]),
        sqlTools::executeDirect(Con,Sql2),
        Sql3=string::concatList(["delete from p3publications.subscription where kb='",Kb,"'"]),
        sqlTools::executeDirect(Con,Sql3),
        L=sx::unpack(Kb,"."),
        L=[Db|_],
        Sql4=string::concatList(["drop database if exists ",Db]),
        sqlTools::executeDirect(Con,Sql4),
        fail
;
    !.   
    
reset_2(0) :-
    Net=getNet(),
    sqlSt::getShareTable(Net,Table,_),
    Con=sqlBase::getConnection(),
    Sql2=string::concatList(["delete from ",Table]),
    sqlTools::executeDirect(Con,Sql2),
    !.
reset_2(_) :- !.
% ----------------------------------------------------------------------------------------------
% processRecord
% ----------------------------------------------------------------------------------------------
processRecord(Con,Stream,C,Format,Type,S,Kb) :-
    Type=2,
    sqlSt::transSentence(Format,S,Lt),
    C:inc(N),
    dbBase::recordInsert(Type,"$STREAM",N,0,Lt,atom("true"),_),
    FormatS=toString(Format),
    TypeS=toString(Type),
    Sql=string::concatList(["update ",Stream," set processed=1,kb='",Kb,"' where k_format=",FormatS," and k_type=",TypeS," and knowledge='",S,"'"]),
    sqlTools::executeDirect(Con,Sql),
    !
;
    Type=3,
    sx::sxGiveTail(S,"!",S2),
    transnl::parTranslate(S2,Term,TermType),
    TermType="r",
    Term=cmp(_,[Lt,Rt]),
	Lt=cmp(PID,_),	
	C:inc(N),
	Ns=toString(N),
	sx::sxlJoin([PID,"STREAM",Ns,"0"],"|",Rule),
	dbExtension::updateDependencies(1,Type,Lt,Rt,Rule,Acyclic),
	Acyclic=0,
	dbExtension::updateDependencies(3,Type,Lt,Rt,Rule,_),
	C:dec(_),
	!
;	
    Type=3,
    sx::sxGiveTail(S,"!",S2),
    transnl::parTranslate(S2,Term,TermType),
    TermType="r",
    Term=cmp(_,[Lt,Rt]),
	C:inc(N),
    dbBase::recordInsert(Type,"$STREAM",N,0,Lt,Rt,_),
    FormatS=toString(Format),
    TypeS=toString(Type),
    Sql=string::concatList(["update ",Stream," set processed=1,kb='",Kb,"' where k_format=",FormatS," and k_type=",TypeS," and knowledge='",S,"'"]),
    sqlTools::executeDirect(Con,Sql),
    !
;
    Type=4,
    sx::sxGiveTail(S,"!",S2),
    transnl::parTranslate(S2,Term,TermType),
    TermType="f",
    Term=cmp(_,[Lt,Rt]),
	Lt=cmp(PID,_),	
	C:inc(N),
	Ns=toString(N),
	sx::sxlJoin([PID,"STREAM",Ns,"0"],"|",Rule),
	dbExtension::updateDependencies(1,Type,Lt,Rt,Rule,Acyclic),
	Acyclic=0,
	dbExtension::updateDependencies(3,Type,Lt,Rt,Rule,_),
	C:dec(_),
	!
;	
    Type=4,
    sx::sxGiveTail(S,"!",S2),
    transnl::parTranslate(S2,Term,TermType),
    TermType="f",
    Term=cmp(_,[Lt,Rt]),
	C:inc(N),
    dbBase::recordInsert(Type,"$STREAM",N,0,Lt,Rt,_),
    FormatS=toString(Format),
    TypeS=toString(Type),
    Sql=string::concatList(["update ",Stream," set processed=1,kb='",Kb,"' where k_format=",FormatS," and k_type=",TypeS," and knowledge='",S,"'"]),
    sqlTools::executeDirect(Con,Sql),
    !
;
    Type=6,
    dlgIF::dlgNote(1,"Inexact rules not yet supported in knowledge streams!"), %@@
    !,fail.
end implement cluster
