/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement sqlUtility
open core,termDomain

constants
className = "com/rudstrand/SqlUtility/sqlUtility".
classVersion = "1.0".

class predicates
deleteKbRecord : (lst,integer,string,integer) determ.
checkChainTable : (lst,count,count,count,count,integer,string).
checkFound : (integer,lst,count,integer,string,integer,string,integer) determ.
checkI : (string,lst,count,integer,string,integer,integer) determ.
checkS : (string,lst,count,integer,string,string,string) determ.
checkCondition : (string,lst,count,integer,integer,sterm,string).
checkChainTableRecord : (lst,count,count,count,integer,string,integer,integer,string,integer,string) determ.
checkChainTables : (lst,count,count,count,count,count,integer).
checkKbTable : (lst,count,integer).
checkRec : (odbcConnection,lst,count,integer,integer,string,string) determ.
checkKbRec : (odbcConnection,lst,count,integer,string,integer) determ.
checkChain : (lst,count,integer,string).

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------------------------------
% kbRestore
% -------------------------------------------------------------------------------------------------------------
class predicates
insertRec : (odbcConnection,string,string).
kbRestoreX : (infile,count,integer).
kbRestoreContinue : (integer,infile,string,integer).
readRecord : (infile,count,integer,odbcConnection,string,integer).    
clauses

kbRestore(Name) :-
    File=inFile::new(Name,"{}",10240),
    File:openInfile,
    Time=File:readNet(),
    _=File:readNet(), % always "SQL"
    Version=File:readNet(),
    NRecS=File:readNet(),
    NRec=sx::toInt(NRecS),
    KbFile=File:readNet(),
    Mess2=string::concatList(["Backup of ",KbFile," version ",Version," with ",NRecS,"  records dated ",Time]),
    Reply=vpiCommonDialogs::ask("Please confirm KB restore!",Mess2,["Yes","No"]),
    kbRestoreContinue(Reply,File,KbFile,NRec),
    File:close(),
    !.

kbRestoreContinue(Reply,File,KbFile,NRec) :-
    Reply=vpiCommonDialogs::resp_default,
    _=File:readNet(), % Db
    Kb=File:readNet(),
    TermFormatS=File:readNet(),
    ChPrefix=File:readNet(),
    StPrefix=File:readNet(),
    StDb=File:readNet(),
    Blank=File:readNet(),
    MaxNameS=File:readNet(),
    CaseS=File:readNet(),
    MaxTextS=File:readNet(),
    sqlBase::deleteKb,
    MaxName=sx::toInt(MaxNameS),
    Case=sx::toInt(CaseS),
    MaxText=sx::toInt(MaxTextS),
    TermFormat=sx::toInt(TermFormatS),
    sqlBase::setParameters(MaxName,ChPrefix,StPrefix,Blank,Case,StDb,TermFormat,MaxText),
    W=guix::getWinOfClas(1),
    Dlg=kbinit::new(),
    Dlg:show(W,1),
    Ps=globalVars::popRetPars(),
    Ps=["1"],   
    sqlBase::createKb(Kb),
    globalVars::set_s("kb_file",KbFile),
    Count=count::new(),
    kbRestoreX(File,Count,NRec),
    sqlBase::setRef,
    N=Count:get(),
    Ns=toString(N),
    Mess=string::concat(Ns," records restored!"),
    dlgIF::dlgNote(0,Mess),
    !
;    
    !.
    
kbRestoreX(File,Count,NRec) :-
    Step=NRec div 100,
    Con=sqlBase::getConnection(),
    MainTable=sqlBase::getMainTable(),
    sqlBase::initRef,
    programControl::repeat,
        globalVars::get_i("quit_flag")=0,
        readRecord(File,Count,NRec,Con,MainTable,Step),
        File:fileIsRead(),
        !
;
    dlgIF::dlgNote(0,"KB restore aborted by user!"),
    !.

readRecord(File,Count,NRec,Con,MainTable,Step) :-
    Chain=File:readNet(),
    Chain<>"",
    RecTypeS=File:readNet(),
    Date=File:readNet(),
    Realm=File:readNet(),
    RecNo=File:readNet(),
    SubNo=File:readNet(),
    St=File:readNet(),
    _=File:readNet(), % Ref
    Lt=File:readNet(),
    Rt=File:readNet(),
    RecType=sx::toInt(RecTypeS),
    dbBase::recordType(RecType,Basetype,ActiveX,SuspX),
    jmod::swapI(SuspX,-1,0,Susp),
    jmod::swapI(ActiveX,-1,0,Active),
    ActiveS=toString(Active),
    SuspS=toString(Susp),
    BasetypeS=toString(Basetype),
    sqlBase::checkIE(Rectype,IE),
    IEs=toString(IE),
    Ref=sqlBase::getRef(Chain),
    RefS=toString(Ref),
    Table=sqlBase::chainTable(Chain),
    Sql1=string::concatList(["insert ",Table," values (",RefS,",",BasetypeS,",'",St,"',",IEs,",'",Lt,"','",Rt,"')"]),
    Sql2=string::concatList(["insert ",MainTable," values (",RecTypes,",",BasetypeS,",",ActiveS,",",SuspS,",'",Date,"','",Realm,"',",RecNo,",",SubNo,",'",St,"','",Chain,"',",RefS,")"]), 
    insertRec(Con,Sql1,Sql2),
    Count:inc(NTot),
    dlgIF::msgProgress("restore at",NTot,Step,NRec),
    !
;
    !.

insertRec(Con,Sql1,Sql2) :-
    sqlBase::start,
    sqlTools::executeDirect(Con,Sql1),
    sqlTools::executeDirect(Con,Sql2),
    sqlBase::commit,
    !
;
    sqlBase::rollback,
    Mess=string::concatList(["SQL record insert failed: ",Sql2]),
    dlgIF::dlgNote(3,Mess),
    !.
% -------------------------------------------------------------------------------------------------------------
% kbBackup
% -------------------------------------------------------------------------------------------------------------
class predicates
kbBackupX : (odbcConnection,outFile,integer,integer,string,integer).
kbBackupSelect : (odbcConnection,outFile,string,string,string,integer,count,integer).    
getCount : (odbcConnection,string,integer) -> integer determ.    
clauses

% parameters:
% 1 in destination file name
% 2 in 0 include derived knowledge, else exclude
%
kbBackup(Name,IncludeDerived) :-
    File=outFile::new(Name),
    sqlBase::getParameters(MaxName,ChainPrefix,StPrefix,Usc,Case,StDb,TermFormat,TermLength),
    Db=sqlBase::getDb(),
    Kb=sqlBase::getKb(),
    KbFile=globalVars::get_s("kb_file"),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    Con=sqlBase::getConnection(),
    MainTable=sqlBase::getMainTable(),
    NRec=getCount(Con,MainTable,IncludeDerived),
    NRecS=toString(NRec),
	dbBase::getProgramBuild(PV,_,_,_,_),
	PVF=PV/100.0,
	sx::fixFormatR(PVF,2,0,Version),
    File:write(Time),
    File:write("{}"),
    File:write("SQL"),
    File:write("{}"),
    File:write(Version),
    File:write("{}"),
    File:write(NRecS),
    File:write("{}"),
    File:write(KbFile),
    File:write("{}"),
    File:write(Db),
    File:write("{}"),
    File:write(Kb),
    File:write("{}"),
    TermFormatS=toString(TermFormat),
    File:write(TermFormatS),
    File:write("{}"),
    File:write(ChainPrefix),
    File:write("{}"),
    File:write(StPrefix),
    File:write("{}"),
    File:write(StDb),
    File:write("{}"),
    File:write(Usc),
    File:write("{}"),
    S2=toString(MaxName),
    File:write(S2),
    File:write("{}"),
    S3=toString(Case),
    File:write(S3),
    File:write("{}"),
    S4=toString(TermLength),
    File:write(S4),
    File:write("{}"),
    kbBackupX(Con,File,IncludeDerived,TermLength,MainTable,NRec),
    File:close,
    Mess=string::concat("Backup of ",NRecS," records written to ",Name,"!"),
    dlgIF::dlgNote(0,Mess),
    !.

getCount(Con,MainTable,IncludeDerived)=N :-
    IncludeDerived=0,
    Sql=string::concat("select count(*) from ",MainTable),
    Select=sqlSelect::new(Con,Sql),
    Select:getNRows()=1,
    N=Select:getI(1,1),
    !
;
    IncludeDerived<>0,
    Sql=string::concat("select count(*) from ",MainTable," where base_type <> 1 and base_type <> 5"),
    Select=sqlSelect::new(Con,Sql),
    Select:getNRows()=1,
    N=Select:getI(1,1),
    !
;
    dlgIF::dlgNote(3,"Unable to count records to be backed up!"),
    !,fail.
    
kbBackupX(Con,File,IncludeDerived,Max,MainTable,NRec) :-
    IncludeDerived=0,
    Count=count::new(),
    sqlBase::chains(Chain,_,Table),
        Sql=string::concat("select record_type,record_date,realm,recno,subno,st,ref from ",MainTable," where chain='",Chain,"' order  by ref"),
        kbBackupSelect(Con,File,Sql,Table,Chain,Max,Count,NRec),
        fail
;
    IncludeDerived<>0,
    Count=count::new(),
    sqlBase::chains(Chain,_,Table),
        Sql=string::concat("select record_type,record_date,realm,recno,subno,st,ref from ",MainTable," where chain='",Chain,"' and base_type <> 1 and base_type <> 5 order  by ref"),
        kbBackupSelect(Con,File,Sql,Table,Chain,Max,Count,NRec),
        fail
;
    !.

kbBackupSelect(Con,File,Sql,Table,Chain,Max,Count,NRec) :-
    Step=NRec div 100,
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        RecType=Select:getI(Row,1),
        Date=Select:getDT(Row,2),
        Realm=Select:getS(Row,3),
        RecNo=Select:getI(Row,4),
        SubNo=Select:getI(Row,5),
        St=Select:getS(Row,6),
        Ref=Select:getI(Row,7),
        RefS=toString(Ref),
        Sql2=string::concat("select lterm,rterm from ",Table," where ref=",RefS),
        Select2=sqlSelect::new(Con,Sql2,Max),
        Select2:getNRows()=1,
        Lt=Select2:getS(1,1),
        Rt=Select2:getS(1,2),
        %
        File:write(Chain),
        File:write("{}"),
        S1=toString(RecType),
        File:write(S1),
        File:write("{}"),
        File:write(Date),
        File:write("{}"),
        File:write(Realm),
        File:write("{}"),
        S2=toString(RecNo),
        File:write(S2),
        File:write("{}"),
        S3=toString(SubNo),
        File:write(S3),
        File:write("{}"),
        File:write(St),
        File:write("{}"),
        File:write(RefS),
        File:write("{}"),
        File:write(Lt),
        File:write("{}"),
        File:write(Rt),
        File:write("{}"),
        Count:inc(NTot),
        dlgIF::msgProgress("backup at",NTot,Step,NRec),
        fail
;
    !.    
% -------------------------------------------------------------------------------------------------------------
% checkKb
% -------------------------------------------------------------------------------------------------------------
checkKb(Repair) :-
    S=sqlBase::getKb(),
    Mess=string::concat("Integrity Errors in ",S),
    Report=rep::newReport(Mess,80),
    not(Report:cancelled),
	vpi::cursorSetWait(),
    List=lst::new(),
    Err=count::new(),
    Size=count::new(),
    Biggest=count::new(),
    Records=count::new(),
    NChains=count::new(),
    checkChainTables(List,Err,Size,Biggest,NChains,Records,Repair),
    checkKbTable(List,Err,Repair),
    %
    L=List:getSorted(),
    Nc=NChains:get(),
    Nr=Records:get(),
    ErrI=Err:get(),
    Size2=Size:get(),
    SizeI=Size2 div 1024,
    BigI=Biggest:get(),
    ErrS=toString(Erri),
    SizeS=toString(SizeI),
    BigS=toString(BigI),
    Ncs=toString(Nc),
    Nrs=toString(Nr),
    p3Rep::header(Report),   
    Report:repLine(' '),
    Report:repWL(0,"#Errors           = ",ErrS),
    Report:repWL(0,"#Chains           = ",Ncs),
    Report:repWL(0,"#Records          = ",Nrs),
    Report:repWL(0,"Kb total term size= ",SizeS," Kb"),
    Max=sqlBase::getMaxLength(),
    MaxS=toString(Max),
    Report:repWL(0,"Biggest term      = ",BigS," characters (max. set at ",MaxS,")"),
    Report:repLine('-'),
    Report:repList(0,L),
    Report:close,
    !.
% -------------------------------------------------------------------------------------------------------------
% checkKbTable
% -------------------------------------------------------------------------------------------------------------
% parameters:
% 1 in report
% 2 in -1/1 repair discrepancies
%
checkKbTable(List,Err,Repair) :-
    sqlBase::chains(Chain),
        checkChain(List,Err,Repair,Chain),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% checkChain
% -------------------------------------------------------------------------------------------------------------
checkChain(List,Err,Repair,Chain) :-
    Table=sqlBase::chainTable(Chain),
    Con=sqlBase::getConnection(),
    Sql=string::concat("select ref from ",Table),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    Ns=toString(N),
    S=string::concat("CHAIN ",Chain," has ",Ns," unique ref-values in chain-table"),
    List:add(S),
    Row=programControl::sequence_nd(1,N,1),
        Ref=Select:getI(Row,1),
        checkKbRec(Con,List,Err,Repair,Chain,Ref),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% checkKbRec
% -------------------------------------------------------------------------------------------------------------
checkKbRec(Con,List,Err,Repair,Chain,Ref) :-
    MainTable=sqlBase::getMainTable(),
    RefS=toString(Ref),
    Sql=string::concatList(["select count(*) from ",MainTable," where ref=",RefS," and chain='",Chain,"'"]),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getI(1,1),
    checkRec(Con,List,Err,Repair,N,Chain,RefS),
    !.
% -------------------------------------------------------------------------------------------------------------
% checkRec
% -------------------------------------------------------------------------------------------------------------   
checkRec(Con,List,Err,Repair,N,Chain,RefS) :-
    N=1,
    !
;
    N>1,
    Ns=toString(N),
    S=string::concatList(["ERROR in chain ",Chain,": ",Ns," records for ref ",RefS," in chain ",Chain,"!"]),
    List:add(S),
    Err:inc,
    !
;
    Repair=1,
    Table=sqlBase::chainTable(Chain),
    MainTable=sqlBase::getMainTable(),
    Sql=string::concatList(["select  base_type,st,lterm,rterm from ",Table," where ref=",RefS]),
    Max=sqlBase::getMaxLength(),
    Select=sqlSelect::new(Con,Sql,Max),
    Select:getNRows()=1,
    Base2=Select:getI(1,1),
    LTermS=Select:getS(1,3),
    LTerm=sqlBase::sqlToTerm(LTerms),
    LTerm=cmp(Chain,_),
    BaseS=toString(Base2),
    dbBase::kbTermST_proc(LTerm,ST),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    Sql2=string::concatList(["insert ",MainTable," values (",BaseS,",",BaseS,",0,0,'",Time,"','$REPAIRED',0,0,'",ST,"','",Chain,"',",RefS,")"]), 
    sqlTools::executeDirect(Con,Sql2),
    S=string::concatList(["REPAIR in chain ",Chain,": Added missing KB-table record for ref ",RefS," with base-type ",BaseS," and sentence-type ",ST,"!"]),
    List:add(S),
    Err:inc,
    !
;
    Repair=-1,
    Table=sqlBase::chainTable(Chain),
    Sql=string::concatList(["select  base_type,st,lterm,rterm from ",Table," where ref=",RefS]),
    Max=sqlBase::getMaxLength(),
    Select=sqlSelect::new(Con,Sql,Max),
    Select:getNRows()=1,
    Base2=Select:getI(1,1),
    LTermS=Select:getS(1,3),
    LTerm=sqlBase::sqlToTerm(LTerms),
    LTerm=cmp(Chain,_),
    BaseS=toString(Base2),
    dbBase::kbTermST_proc(LTerm,ST),
    S=string::concatList(["ERROR in chain ",Chain,": Missing KB-table record for ref ",RefS," with base-type ",BaseS," and sentence-type ",ST,"!"]),
    List:add(S),
    Err:inc,
    !.
% -------------------------------------------------------------------------------------------------------------
% checkChainTables
% -------------------------------------------------------------------------------------------------------------
% parameters:
% 1 in list
% 2 in -1/1 repair discrepancies
%
checkChainTables(List,Err,Size,Biggest,NChains,Records,Repair) :-
    sqlBase::chains(Chain),
        NChains:inc,
        checkChainTable(List,Err,Size,Biggest,Records,Repair,Chain),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% checkChainTable
% -------------------------------------------------------------------------------------------------------------
checkChainTable(List,Err,Size,Biggest,Records,Repair,Chain) :-
    MainTable=sqlBase::getMainTable(),
    Con=sqlBase::getConnection(),
    Sql=string::concat("select  ref,base_type,st,record_type from ",MainTable," where chain='",Chain,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    Ns=toString(N),
    Records:add(N),
    S=string::concatList(["CHAIN ",Chain," has ",Ns," records in KB-table."]),
    List:add(S),
    Row=programControl::sequence_nd(1,N,1),
        Ref=Select:getI(Row,1),
        Base=Select:getI(Row,2),
        St=Select:getS(Row,3),
        Rt=Select:getI(Row,4),
        sqlBase::checkIE(Rt,Ie),
        Table=sqlBase::chainTable(Chain),
        checkChainTableRecord(List,Err,Size,Biggest,Repair,Chain,Ref,Base,St,Ie,Table),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% checkChainTableRecord
% -------------------------------------------------------------------------------------------------------------
checkChainTableRecord(List,Err,Size,Biggest,Repair,Chain,Ref,Base,St,Ie,Table) :-
    Con=sqlBase::getConnection(),
    SRef=toString(Ref),
    Sql=string::concatList(["select  base_type,st,ie,lterm,rterm from ",Table," where ref=",SRef]),
    Max=sqlBase::getMaxLength(),
    Select=sqlSelect::new(Con,Sql,Max),
    N=Select:getNRows(),
    checkFound(N,List,Err,Repair,Chain,Base,St,Ref),
    N>0,
    Base2=Select:getI(1,1),
    St2=Select:getS(1,2),
    Ie2=Select:getI(1,3),
    LTermS=Select:getS(1,4),
    RTermS=Select:getS(1,5),
    L1=string::length(LTermS),
    L2=string::length(RTermS),
    Size:add(L1),
    Size:add(L2),
    Biggest:setMax(L1),
    Biggest:setMax(L2),
    checkI(Chain,List,Err,Ref,"record base-types",Base,Base2),
    checkI(Chain,List,Err,Ref,"record inference-code",Ie,Ie2),
    checkS(Chain,List,Err,Ref,"record sentence types",St,St2),
    RTerm=sqlBase::sqlToTerm(RTerms),
    checkCondition(Chain,List,Err,Ref,Base2,RTerm,RTerms),
    !.
% -------------------------------------------------------------------------------------------------------------
% checkCondition
% -------------------------------------------------------------------------------------------------------------
checkCondition(Chain,List,Err,Ref,Base,RTerm,RTerms) :-
    set::member(Base,[1,2]),
    RTerm=atom("true"),
    !
;    
    set::member(Base,[1,2]),
    RefS=toString(Ref),
    BaseS=toString(Base),
    S=string::concatList(["ERROR in chain ",Chain,": Record ",RefS," with condition ",RTerms," has invalid base-type ",BaseS,"!"]),
    List:add(S),
    Err:inc,
    !    
;    
    Base=6,
    RTerm=nill,
    !
;
    Base=6,
    RefS=toString(Ref),
    BaseS=toString(Base),
    S=string::concatList(["ERROR in chain ",Chain,": Record ",RefS," with condition ",RTerms," has invalid base-type ",BaseS,"!"]),
    List:add(S),
    Err:inc,
    !    
;  
    set::member(Base,[3,4,5]),
    !
;
    RefS=toString(Ref),
    BaseS=toString(Base),
    S=string::concatList(["ERROR in chain ",Chain,": Record ",RefS," has invalid base-type ",BaseS,"!"]),
    List:add(S),
    Err:inc,
    !.
% -------------------------------------------------------------------------------------------------------------
% checkI/S
% -------------------------------------------------------------------------------------------------------------
checkI(Chain,List,Err,Ref,S,V,V2) :-
    V=V2,
    !
;
    RefS=toString(Ref),
    Vs=toString(V),
    V2s=toString(Vs),
    S=string::concatList(["ERROR in chain ",Chain,": Record ",RefS," has mis-matched ",S," ",Vs," and ",V2s,"!"]),
    List:add(S),
    Err:inc,
    !.    

checkS(Chain,List,Err,Ref,S,V,V2) :-
    V=V2,
    !
;
    RefS=toString(Ref),
    S=string::concatList(["ERROR in chain ",Chain,": Record ",RefS," has mis-matched ",S," ",V," and ",V2,"!"]),
    List:add(S),
    Err:inc,
    !.    
% -------------------------------------------------------------------------------------------------------------
% checkFound
% -------------------------------------------------------------------------------------------------------------
checkFound(N,List,Err,Repair,Chain,Base,St,Ref) :-
    N=1,
    !
;
    N=0,
    BaseS=toString(Base),
    RefS=toString(Ref),
    S=string::concatList(["ERROR in chain ",Chain,": Record ",RefS," of base-type ",BaseS," and sentence type ",St," not found in chain-table!"]),
    List:add(S),
    Err:inc,
    deleteKbRecord(List,Repair,Chain,Ref),
    !
;
    N>1,
    BaseS=toString(Base),
    Ns=toString(N),
    RefS=toString(Ref),
    S=string::concatList(["ERROR in chain ",Chain,": Record ",RefS," of base-type ",BaseS," and sentence type ",St," has ",Ns," chain-table records!"]),
    List:add(S),
    Err:inc,
    !.
% -------------------------------------------------------------------------------------------------------------
% deleteKbRecord
% -------------------------------------------------------------------------------------------------------------
deleteKbRecord(List,Repair,Chain,Ref) :-
    Repair=-1,
    !
;
    MainTable=sqlBase::getMainTable(),
    Con=sqlBase::getConnection(),
    SRef=toString(Ref),
    Sql=string::concat("delete from ",MainTable," where chain='",Chain,"' and ref=",SRef),
    sqlTools::executeDirect(Con,Sql),
    sqlBase::checkForEmptyChain(Chain),
    S=string::concatList(["REPAIR in chain ",Chain,": Missing KB-table record deleted!"]),
    List:add(S),
    !. 
end implement sqlUtility
