/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement stTables
open core, vpiDomains, resourceIdentifiers

constants
className = "com/rudstrand/TaskWindow/stTables".
classVersion = "1.0".

predicates
displayKbNet : (integer,integer,integer,string_list,string_list).
updateOpenPub : () determ.
updateOpenSub : () determ.

clauses
classInfo(className, classVersion).

facts
thisWin : vpiDomains::windowHandle := erroneous.

clauses
new(Parent):-
        _ = vpi::winCreate(windowType, rectangle, title, menu, Parent,
                           windowFlags, eventHandler, 0).

predicates
eventHandler : vpiDomains::ehandler.
clauses
eventHandler(Win, Event) = Result :-
    Result = generatedEventHandler(Win, Event),
    !.
eventHandler(Win, e_Menu(Id, CAS)) = Result :-
    Parent = vpi::winGetParent(Win),
    Result = vpi::winSendEvent(Parent, e_Menu(Id, CAS)).

/* create pub db */

predicates
    onControlidc_create_db : vpiDomains::controlHandler.
clauses
    onControlidc_create_db(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    vpiCommonDialogs::resp_default=vpiCommonDialogs::ask("Please confirm!","Documentation of existing publishing networks (if any) will be deleted!",["Yes","No"]),
    sqlSt::createDb,
    guix::ctlDisable(thisWin,[idc_create_db]),
    dlgIF::dlgNote(0,"(Re-)Created p3bridge database!"),
    !;!.

/* open pub db */

predicates
    onControlidc_open_db : vpiDomains::controlHandler.
clauses
    onControlidc_open_db(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,"use p3publications"),
    L=sqlSt::getNets(),
    guix::lbxPop(thisWin,idc_nets,L),
    guix::ctlDisable(thisWin,[idc_create_db,idc_open_db]),
    guix::ctlEnable(thisWin,[idct_publishing_networks,idc_create_net,idc_nets,idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1,idc_open_share,idct_def_db,idc_def_db]),
    !;!.

/* select pub net */

predicates
    onControlidc_nets : vpiDomains::controlHandler.
clauses
    onControlidc_nets(_CtrlID, _CtrlType, _CtrlWin,CtrlInfo) = handled(0) :-
    CtrlInfo=vpiDomains::selchanged(),
    guix::lbtGet(thisWin,idc_nets,Net),
    sqlSt::getNet(Net,OpenShare,_,_,_),
    OpenShare=0,
    % existing net used by KB
    sqlSt::getKbNetUse(Net,O1,O2,_,_),
    O3=O1+O2,
    O3>0,
    Mess="KB open share codes do not match 'no open share' in p3publications!",
    dlgIF::dlgNote(2,Mess),
    !
;    
    CtrlInfo=vpiDomains::selchanged(),
    guix::lbtGet(thisWin,idc_nets,Net),
    sqlSt::getNet(Net,OpenShare,_,Pubs2,DefDb),
    guix::txtSet(thisWin,idc_def_db,DefDb),
    OpenShare=0,
    guix::lbxPop(thisWin,idcx_pubs,Pubs2),
    sqlSt::getKbNetUse(Net,_,_,Pubs,Subs),
    displayKbNet(0,3,3,Pubs,Subs),
    !
;
    CtrlInfo=vpiDomains::selchanged(),
    guix::lbtGet(thisWin,idc_nets,Net),
    sqlSt::getNet(Net,OpenShare,_,_,DefDb),
    OpenShare>0,
    guix::txtSet(thisWin,idc_def_db,DefDb),
    sqlSt::getKbNetUse(Net,_,_,Pubs,Subs),
    set::size(Pubs,N1),
    set::size(Subs,N2),
    N=N1+N2,
    N>0,
    Mess="KB subscriptions/publications do not match 'open share' in p3publications!",
    dlgIF::dlgNote(2,Mess),
    !
;
    CtrlInfo=vpiDomains::selchanged(),
    guix::lbtGet(thisWin,idc_nets,Net),
    sqlSt::getNet(Net,OpenShare,Table,_,DefDb),
    OpenShare>0,
    % existing open share used by KB
    guix::txtSet(thisWin,idc_def_db,DefDb),
    guix::txtSet(thisWin,idc_share_table,Table),
    sqlSt::getKbNetUse(Net,Op,Os,_,_),
    displayKbNet(OpenShare,Op,Os,[],[]),
    !
;
    CtrlInfo=vpiDomains::modified(),
    % new net
    guix::lbtGet(thisWin,idc_nets,Net),
    sqlSt::getNet(Net,_,_,_,_),
    dlgIF::dlgNote(0,"Network name is already used!"),
    !
;
    CtrlInfo=vpiDomains::modified(),
    % new net
    guix::lbtGet(thisWin,idc_nets,Net),
    not(sqlSt::getNet(Net,_,_,_,_)),
    guix::ctlEnable(thisWin,[idc_open_share]),
    !
;
    !.

predicates
    onControlidc_open_share : vpiDomains::controlHandler.
clauses
    onControlidc_open_share(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    guix::guiCB_Get(thisWin,idc_open_share,Open),
    Open=1,
    guix::ctlEnable(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1,idc_share_table,idct_share_table]),
    guix::guiRB_SetN(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1],[1,9,10,11],1),
    !
;    
    guix::guiCB_Get(thisWin,idc_open_share,Open),
    Open=1,
    guix::ctlDisable(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1]),
    !;!.
    
predicates
    onControlidc_delete_net : vpiDomains::controlHandler.
clauses
    onControlidc_delete_net(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
    guix::lbtGetF(thisWin,idc_nets,Net),
    sqlSt::deleteKbNet(Net),
    sqlSt::deleteNet(Net),
    Mess=string::concat("Deleted publishing network ",Net,"!"),
    dlgIF::dlgNote(0,Mess),
    !;!.

predicates
    onControlidc_create_net : vpiDomains::controlHandler.
clauses
    onControlidc_create_net(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),	
    guix::guiCB_Get(thisWin,idc_open_share,Open),
    Open=1,
	guix::txtGet(thisWin,idc_share_table,Table),
	Table="",
	dlgIF::dlgNote(0,"Specify an open share table!"),
	!
;    
	guix::txtGet(thisWin,idc_def_db,Db),
	Db="",
	dlgIF::dlgNote(0,"Specify a default publishing SQL database!"),
	!
;    
    guix::guiCB_Get(thisWin,idc_open_share,Open),
    Open=1,
	guix::txtGet(thisWin,idc_share_table,Table),
    guix::guiRB_Get(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1],[1,9,10,11],OpenShare),
    guix::txtGet(thisWin,idc_nets,Net),
    guix::txtGet(thisWin,idc_def_db,Db),
    sqlSt::createNet(Net,OpenShare,Table,Db),
    guix::ctlDisable(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1,idc_open_share]),
    guix::lbxPop(thisWin,idc_publications,[]),
    guix::lbxPop(thisWin,idc_subscriptions,[]),
    guix::ctlEnable(thisWin,[idc_pubi,idc_pubr,idc_pubn,idc_subi,idc_subr,idc_subn]),
    guix::ctlDisable(thisWin,[idc_publications,idc_sts,idc_tablename,idc_incremental,idc_replacement,idc_adhoc_publication,idc_pub_now,idc_nl,idc_pr,idc_xmlnl,idc_xmlpr,
                                      idc_table,idc_subscriptions,idc_subs,idc_incremental1,idc_replacement1,idc_adhoc_subscription,idc_sub_now,idc_add1,idc_remove1]),
    Mess=string::concat("Created open share publishing network ",Net,"!"),
    dlgIF::dlgNote(0,Mess),
    !
;    
    guix::guiCB_Get(thisWin,idc_open_share,Open),
    Open=-1,
    guix::txtGet(thisWin,idc_nets,Net),
    Net<>"",
    guix::txtGet(thisWin,idc_def_db,Db),
    sqlSt::createNet(Net,Db),
    guix::ctlDisable(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1,idc_open_share]),
    guix::lbxPop(thisWin,idc_publications,[]),
    guix::lbxPop(thisWin,idc_subscriptions,[]),
    guix::lbxGetAll(thisWin,idc_stall,AllSts),
    guix::lbxPop(thisWin,idc_sts,AllSts),
    guix::ctlEnable(thisWin,[idc_publications,idc_sts,idct_table,idc_tablename,idc_incremental,idc_replacement,idc_adhoc_publication,idc_pub_now,idc_nl,idc_pr,idc_xmlnl,idc_xmlpr,idc_table]),
    guix::ctlDisable(thisWin,[idc_subscriptions,idc_subs,idc_incremental1,idc_replacement1,idc_adhoc_subscription,idc_sub_now,idc_add1,idc_remove1]),
    Mess=string::concat("Created publishing network ",Net,"!"),
    dlgIF::dlgNote(0,Mess),
    !
;    
    !;!.

/* open share pub/sub */

predicates
    onControlidc_pubi : vpiDomains::controlHandler.
clauses
    onControlidc_pubi(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    updateOpenPub(),
    !;!.
predicates
    onControlidc_pubn : vpiDomains::controlHandler.
clauses
    onControlidc_pubn(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    updateOpenPub(),
    !;!.

predicates
    onControlidc_pubr : vpiDomains::controlHandler.
clauses
    onControlidc_pubr(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    updateOpenPub(),
    !;!.

predicates
    onControlidc_subi : vpiDomains::controlHandler.
clauses
    onControlidc_subi(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    updateOpenSub(),
    !;!.

predicates
    onControlidc_subn : vpiDomains::controlHandler.
clauses
    onControlidc_subn(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    updateOpenSub(),
    !;!.

predicates
    onControlidc_subr : vpiDomains::controlHandler.
clauses
    onControlidc_subr(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    updateOpenSub(),
    !;!.

/* add publication */    
    
predicates
    onControlidc_add : vpiDomains::controlHandler.
clauses
    onControlidc_add(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
    guix::lbtGetF(thisWin,idc_nets,Net),
	Net<>"",
	guix::lbtGetF(thisWin,idc_sts,St),
	St<>"",
	guix::txtGet(thisWin,idc_tablename,TableName),
	TableName<>"",
	%
    guix::guiRB_Get(thisWin,[idc_nl,idc_table,idc_pr,idc_xmlnl,idc_xmlpr],[1,8,9,10,11],Format),
    guix::guiRB_Get(thisWin,[idc_incremental,idc_replacement],[1,2],Mode),
    sqlSt::addPub(Net,St,TableName,Format,Mode,PubS),
    guix::lbxAdd(thisWin,idc_publications,PubS),
    guix::lbxSel(thisWin,idc_publications,PubS),
    guix::lbxDelS(thisWin,idc_sts,St),
    Mess=string::concat("Defined publication of '",St,"' to '",TableName,"!"),
	dlgIF::dlgNote(0,Mess),
	!
;    
	guix::lbtGetF(thisWin,idc_sts,St),
	St="",
	dlgIF::dlgNote(0,"Select a sentence type!"),
	!
;	
	guix::txtGet(thisWin,idc_tablename,TableName),
	TableName="",
	dlgIF::dlgNote(0,"Enter a table name!"),
	!;!.
    
/* remove pub */

predicates
    onControlidc_remove : vpiDomains::controlHandler.
clauses
    onControlidc_remove(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    guix::lbtGetF(thisWin,idc_nets,Net),
    guix::lbxGet(thisWin,idc_publications,P),
    sx::sxUnpack(P,'|',L),
    L=[St|_],
    Kb=dbBase::dbName(),
    sqlSt::remPub(Kb,Net,St),
    guix::lbxRemReselect(thisWin,idc_publications),
    guix::lbxAdd(thisWin,idc_sts,St),
    !;!.

/* publish now */

predicates
    onControlidc_pub_now : vpiDomains::controlHandler.
clauses
    onControlidc_pub_now(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
	guix::lbtGetF(thisWin,idc_publications,Pub),
	L=sx::unpack(Pub,"|"),
	L=[St|_],
    sqlSt::publishSt(St),
    Mess=string::concat("Sentence type '",St,"' published!"),
	dlgIF::dlgNote(0,Mess),
	!
;    
	guix::lbtGetF(thisWin,idc_publications,St),
	St="",
	dlgIF::dlgNote(0,"Select a published sentence type!"),
	!;!.

/* ad-hoc publication */

predicates
    onControlidc_adhoc_publication : vpiDomains::controlHandler.
clauses
    onControlidc_adhoc_publication(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
	guix::lbtGetF(thisWin,idc_sts,St),
	St<>"",
	guix::txtGet(thisWin,idc_tablename,TableName),
    guix::lbtGetF(thisWin,idc_nets,Net),
	TableName<>"",
	%
    guix::guiRB_Get(thisWin,[idc_nl,idc_table,idc_pr,idc_xmlnl,idc_xmlpr],[1,8,9,10,11],Format),
    sqlSt::publishSt(St,Net,TableName,Format,2),
    Mess=string::concat("Sentence type '",St,"' published to table ",TableName,"!"),
	dlgIF::dlgNote(0,Mess),
	!
;    
	guix::lbtGetF(thisWin,idc_sts,St),
	St="",
	dlgIF::dlgNote(0,"Select a sentence type!"),
	!
;	
	guix::txtGet(thisWin,idc_tablename,TableName),
	TableName="",
	dlgIF::dlgNote(0,"Enter a table name!"),
	!;!.

/* add sub */

predicates
    onControlidc_add1 : vpiDomains::controlHandler.
clauses
    onControlidc_add1(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
    guix::lbtGetF(thisWin,idc_nets,Net),
	guix::lbtGetF(thisWin,idc_subs,Sub),
	Sub<>"",
	sx::sxUnpack(Sub,'|',L),
	L=[St|_],
	%
    guix::guiRB_Get(thisWin,[idc_incremental1,idc_replacement1],[1,2],Mode),
    sqlSt::addSub(Net,St,Mode),
    L2=sx::unpack(Sub,"|"),
    L2=[St2,Form2,Mode2,Table2|_],
    P=sx::pack([St2,Form2,Mode2,Table2],"|"),
    guix::lbxAdd(thisWin,idc_subscriptions,P),
    guix::lbxSel(thisWin,idc_subscriptions,P),
    guix::lbxDelS(thisWin,idc_subs,Sub),
    Mess=string::concat("Defined subscription of '",St,"!"),
	dlgIF::dlgNote(0,Mess),
	!
;    
	guix::lbtGetF(thisWin,idc_subs,Subs),
	Subs="",
	dlgIF::dlgNote(0,"Select a subscription!"),
	!;!.

/* rem sub */

predicates
    onControlidc_remove1 : vpiDomains::controlHandler.
clauses
    onControlidc_remove1(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    guix::lbtGetF(thisWin,idc_nets,Net),
    guix::lbxGet(thisWin,idc_subscriptions,Sub),
    sx::sxUnpack(Sub,'|',L),
    L=[St|_],
    sqlSt::remSub(Net,St),
    guix::lbxRemReselect(thisWin,idc_subscriptions),
    guix::lbxAdd(thisWin,idc_subs,Sub),
    !;!.

/* subscribe now */

predicates
    onControlidc_sub_now : vpiDomains::controlHandler.
clauses
    onControlidc_sub_now(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
    guix::lbtGetF(thisWin,idc_nets,Net),
	guix::lbtGetF(thisWin,idc_subscriptions,Sub),
	Sub<>"",
	%
	sx::sxUnpack(Sub,'|',L),
	L=[St,FormatN,ModeN,Table],
	dbBase::kbID(ModeN,ModeK,_,_),
	sqlSt::modeCode(ModeK,Mode),
	sqlSt::formatCode(FormatN,Format),
    sqlSt::subscribeSt(St,Net,Table,Format,Mode),
    Mess=string::concat("Sentence type '",St,"' subscribed from table ",Table,"!"),
	dlgIF::dlgNote(0,Mess),
	!
;    
	guix::lbtGet(thisWin,idc_subs,Sub),
	Sub="",
	dlgIF::dlgNote(0,"Select a published sentence type!"),
	!;!.

/* adhoc sub */

predicates
    onControlidc_adhoc_subscription : vpiDomains::controlHandler.
clauses
    onControlidc_adhoc_subscription(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
    guix::lbtGetF(thisWin,idc_nets,Net),
	guix::lbtGetF(thisWin,idc_subs,Sub),
	Sub<>"",
	%
	sx::sxUnpack(Sub,'|',L),
	L=[St,SFormat,Table],
	Format=sx::toInt(SFormat),
    sqlSt::subscribeSt(St,Net,Table,Format,2),
    Mess=string::concat("Sentence type '",St,"' subscribed from table ",Table,"!"),
	dlgIF::dlgNote(0,Mess),
	!
;    
	guix::lbtGetF(thisWin,idc_subs,Sub),
	Sub="",
	dlgIF::dlgNote(0,"Select a published sentence type!"),
	!;!.

/* update default db */

predicates
    onControlidc_update : vpiDomains::controlHandler.
clauses
    onControlidc_update(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	sqlBase::getParameters(A,B,C,D,E,_,F,G),
	guix::txtGet(thisWin,idc_sql,Db),
	sqlBase::setParameters(A,B,C,D,E,Db,F,G),
	!;!.

predicates
    onCreate : vpiDomains::createHandler.
clauses
    onCreate(_CreationData) = defaultHandling() :-
	guix::addWin(11,thisWin),
	guix::setStyle(thisWin,[idc_create_db,idc_open_db,idc_open_share,idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1,idct_share_table,idc_share_table,
	                                idct_publishing_networks,idc_delete_net,idc_create_net,idc_pubi,idc_pubr,idc_pubn,idc_subr,idc_subi,idc_subn,idc_nets,
	                                idc_publications,idc_sts,idct_table,idc_tablename,idc_incremental,idc_replacement,idc_adhoc_publication,idc_pub_now,idc_add,idc_remove,
	                                idc_nl,idc_pr,idc_xmlnl,idc_xmlpr,idc_table,idc_sql,idc_update,idct_sql,idct_def_db,idc_def_db,
	                                idc_subscriptions,idc_subs,idc_incremental1,idc_replacement1,idc_adhoc_subscription,idc_sub_now,idc_add1,idc_remove1],[]),
    guix::ctlDisable(thisWin,[idc_open_share,idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1,idct_share_table,idc_share_table,idc_nets,
	                                  idct_publishing_networks,idc_delete_net,idc_create_net,idc_pubi,idc_pubr,idc_pubn,idc_subr,idc_subi,idc_subn,idct_def_db,idc_def_db,
	                                  idc_publications,idc_sts,idct_table,idc_tablename,idc_incremental,idc_replacement,idc_adhoc_publication,idc_pub_now,idc_add,idc_remove,
	                                  idc_nl,idc_pr,idc_xmlnl,idc_xmlpr,idc_table,
	                                  idc_subscriptions,idc_subs,idc_incremental1,idc_replacement1,idc_adhoc_subscription,idc_sub_now,idc_add1,idc_remove1]),
	vpi::cursorSetWait(),
    guix::guiRB_SetN(thisWin,[idc_nl,idc_table,idc_pr,idc_xmlnl,idc_xmlpr],[1,8,9,10,11],8),
    guix::guiRB_SetN(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1],[1,9,10,11],1),
    guix::guiRB_SetN(thisWin,[idc_incremental,idc_replacement],[1,2],2),
    guix::guiRB_SetN(thisWin,[idc_incremental1,idc_replacement1],[1,2],2),
    STs=pop::getSTs(),
	dbBase::nlSings(Sts,Stsn),
	guix::lbxPop(thisWin,idc_stall,Stsn),
	sqlBase::getParameters(_,_,_,_,_,Db,_,_),
	guix::txtSet(thisWin,idc_sql,Db),
	sqlBase::openConnection,
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

predicates
    onDestroy : vpiDomains::destroyHandler.
clauses
    onDestroy() = defaultHandling() :-
        thisWin := erroneous.

predicates
    onSizeChanged : vpiDomains::sizeHandler.
clauses
    onSizeChanged(_Width, _Height) = defaultHandling():-
        vpiToolbar::resize(thisWin).

predicates
    onControlidc_incremental1 : vpiDomains::controlHandler.
clauses
    onControlidc_incremental1(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        guix::guiRB_SetN(thisWin,[idc_incremental1,idc_replacement1],[1,2],1),
        !;!.

predicates
    onControlidc_replacement1 : vpiDomains::controlHandler.
clauses
    onControlidc_replacement1(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        guix::guiRB_SetN(thisWin,[idc_incremental1,idc_replacement1],[1,2],2),
        !;!.

% This code is maintained by the VDE do not update it manually, 23:46:29-19.6.2006
constants
    windowType : vpiDomains::wintype = w_TopLevel.
    windowFlags : vpiDomains::wsflags = [wsf_SizeBorder,wsf_TitleBar,wsf_Maximize,wsf_Minimize,wsf_Close,wsf_ClipSiblings,wsf_ClipChildren].
    rectangle : vpiDomains::rct = rct(10,50,976,623).
    menu : vpiDomains::menu = noMenu.
    title : string = "Publications and Subscriptions".

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        _ = vpi::winCreateControl(wc_LBox,rct(8,132,552,348),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight],idc_publications),
        _ = vpi::winCreateControl(wc_LBoxEdit,rct(560,132,960,600),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort],idc_sts),
        _ = vpi::winCreateControl(wc_Edit,rct(648,160,960,184),"",Win,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll],idc_tablename),
        _ = vpi::winCreateControl(wc_Edit,rct(560,296,848,320),"",Win,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll],idc_sql),
        _ = vpi::winCreateControl(wc_Text,rct(560,272,848,292),"Default ad-hoc publication database:",Win,[wsf_AlignLeft],idct_sql),
        _ = vpi::winCreateControl(wc_Text,rct(560,160,640,180),"Table name:",Win,[wsf_AlignLeft],idct_table),
        _ = vpi::winCreateControl(wc_PushButton,rct(560,324,848,348),"Update default ad-hoc publication database",Win,[wsf_Group,wsf_TabStop],idc_update),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,4,248,28),"(Re-)Create p3publications database",Win,[wsf_Group,wsf_TabStop],idc_create_db),
        _ = vpi::winCreateControl(wc_Text,rct(256,4,387,24),"Publishing network",Win,[wsf_AlignLeft],idct_publishing_networks),
        _ = vpi::winCreateControl(wc_LBoxEdit,rct(400,4,688,472),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort],idc_nets),
        _ = vpi::winCreateControl(wc_PushButton,rct(696,32,904,56),"Create new publishing network",Win,[wsf_Group,wsf_TabStop],idc_create_net),
        _ = vpi::winCreateControl(wc_PushButton,rct(696,4,904,28),"Delete publishing network",Win,[wsf_Group,wsf_TabStop],idc_delete_net),
        _ = vpi::winCreateControl(wc_CheckBox,rct(128,60,232,80),"Open share",Win,[wsf_Auto,wsf_Group,wsf_TabStop],idc_open_share),
        _ = vpi::winCreateControl(wc_Text,rct(8,112,960,128),"Publications",Win,[wsf_AlignCenter],idct_publications),
        _ = vpi::winCreateControl(wc_RadioButton,rct(560,188,660,208),"Incremental",Win,[wsf_Auto,wsf_TabStop,wsf_Group],idc_incremental),
        _ = vpi::winCreateControl(wc_RadioButton,rct(664,188,773,208),"Replacement",Win,[wsf_Auto,wsf_TabStop],idc_replacement),
        _ = vpi::winCreateControl(wc_PushButton,rct(560,216,656,240),"Add",Win,[wsf_Group,wsf_TabStop],idc_add),
        _ = vpi::winCreateControl(wc_PushButton,rct(560,244,656,268),"Remove",Win,[wsf_Group,wsf_TabStop],idc_remove),
        _ = vpi::winCreateControl(wc_RadioButton,rct(664,220,736,240),"NL",Win,[wsf_Auto,wsf_TabStop,wsf_Group],idc_nl),
        _ = vpi::winCreateControl(wc_RadioButton,rct(664,248,736,268),"PR",Win,[wsf_Auto,wsf_TabStop],idc_pr),
        _ = vpi::winCreateControl(wc_RadioButton,rct(744,220,816,240),"XML/NL",Win,[wsf_Auto,wsf_TabStop],idc_xmlnl),
        _ = vpi::winCreateControl(wc_RadioButton,rct(744,248,816,268),"XML/PR",Win,[wsf_Auto,wsf_TabStop],idc_xmlpr),
        _ = vpi::winCreateControl(wc_RadioButton,rct(824,248,896,268),"Table",Win,[wsf_Auto,wsf_TabStop],idc_table),
        _ = vpi::winCreateControl(wc_RadioButton,rct(392,60,464,80),"XML/NL",Win,[wsf_Auto,wsf_TabStop],idc_xmlnl1),
        _ = vpi::winCreateControl(wc_RadioButton,rct(240,60,312,80),"NL",Win,[wsf_Auto,wsf_TabStop,wsf_Group],idc_nl1),
        _ = vpi::winCreateControl(wc_RadioButton,rct(320,60,392,80),"PR",Win,[wsf_Auto,wsf_TabStop],idc_pr1),
        _ = vpi::winCreateControl(wc_RadioButton,rct(480,60,552,80),"XML/PR",Win,[wsf_Auto,wsf_TabStop],idc_xmlpr1),
        _ = vpi::winCreateControl(wc_PushButton,rct(824,216,959,240),"Ad-hoc Publication",Win,[wsf_Group,wsf_TabStop],idc_adhoc_publication),
        _ = vpi::winCreateControl(wc_PushButton,rct(824,432,959,456),"Ad-hoc Subscription",Win,[wsf_Group,wsf_TabStop],idc_adhoc_subscription),
        _ = vpi::winCreateControl(wc_PushButton,rct(560,460,656,484),"Remove",Win,[wsf_Group,wsf_TabStop],idc_remove1),
        _ = vpi::winCreateControl(wc_PushButton,rct(560,432,656,456),"Add",Win,[wsf_Group,wsf_TabStop],idc_add1),
        _ = vpi::winCreateControl(wc_RadioButton,rct(664,404,773,424),"Replacement",Win,[wsf_TabStop],idc_replacement1),
        _ = vpi::winCreateControl(wc_RadioButton,rct(560,404,660,424),"Incremental",Win,[wsf_TabStop,wsf_Group],idc_incremental1),
        _ = vpi::winCreateControl(wc_LBoxEdit,rct(560,376,960,844),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort],idc_subs),
        _ = vpi::winCreateControl(wc_Text,rct(8,356,960,372),"Subscriptions",Win,[wsf_AlignCenter],idct_subscriptions),
        _ = vpi::winCreateControl(wc_LBox,rct(8,376,552,568),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight],idc_subscriptions),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,32,248,56),"Open p3publications database",Win,[wsf_Group,wsf_TabStop],idc_open_db),
        _ = vpi::winCreateControl(wc_LBox,rct(8,68,32,96),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight,wsf_Invisible],idcx_pubs),
        _ = vpi::winCreateControl(wc_LBox,rct(40,68,64,96),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight,wsf_Invisible],idcx_subs),
        _ = vpi::winCreateControl(wc_LBox,rct(96,68,120,96),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight,wsf_Invisible],idc_stall),
        _ = vpi::winCreateControl(wc_Text,rct(128,84,232,104),"Share Table",Win,[wsf_AlignCenter],idct_share_table),
        _ = vpi::winCreateControl(wc_Edit,rct(240,84,544,108),"",Win,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll],idc_share_table),
        _ = vpi::winCreateControl(wc_Text,rct(256,36,440,56),"Default publishing database",Win,[wsf_AlignLeft],idct_def_db),
        _ = vpi::winCreateControl(wc_Edit,rct(448,32,688,56),"",Win,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll],idc_def_db),
        _ = vpi::winCreateControl(wc_PushButton,rct(824,188,959,212),"Publish now!",Win,[wsf_Group,wsf_TabStop],idc_pub_now),
        _ = vpi::winCreateControl(wc_PushButton,rct(824,404,959,428),"Subscribe Now!",Win,[wsf_Group,wsf_TabStop],idc_sub_now),
        _ = vpi::winCreateControl(wc_RadioButton,rct(704,64,832,84),"Publish (incr)",Win,[wsf_Auto,wsf_TabStop],idc_pubi),
        _ = vpi::winCreateControl(wc_RadioButton,rct(568,64,696,84),"Publish (repl.)",Win,[wsf_Auto,wsf_TabStop,wsf_Group],idc_pubr),
        _ = vpi::winCreateControl(wc_RadioButton,rct(840,64,960,84),"No publication",Win,[wsf_Auto,wsf_TabStop],idc_pubn),
        _ = vpi::winCreateControl(wc_RadioButton,rct(568,88,696,108),"Subscribe (repl.)",Win,[wsf_Auto,wsf_TabStop,wsf_Group],idc_subr),
        _ = vpi::winCreateControl(wc_RadioButton,rct(704,88,832,108),"Subscribe (incr.)",Win,[wsf_Auto,wsf_TabStop],idc_subi),
        _ = vpi::winCreateControl(wc_RadioButton,rct(840,88,960,108),"No subscription",Win,[wsf_Auto,wsf_TabStop],idc_subn),
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
        handled(Result) = onSizeChanged(Width, Height).
    generatedEventHandler(_Win, e_Control(idc_update, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_update(idc_update, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_create_db, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_create_db(idc_create_db, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_open_db, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_open_db(idc_open_db, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_nets, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_nets(idc_nets, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_delete_net, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_delete_net(idc_delete_net, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_create_net, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_create_net(idc_create_net, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_adhoc_publication, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_adhoc_publication(idc_adhoc_publication, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_add, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_add(idc_add, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_remove, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_remove(idc_remove, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_add1, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_add1(idc_add1, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_remove1, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_remove1(idc_remove1, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_adhoc_subscription, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_adhoc_subscription(idc_adhoc_subscription, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_pub_now, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_pub_now(idc_pub_now, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_sub_now, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_sub_now(idc_sub_now, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_open_share, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_open_share(idc_open_share, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_pubi, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_pubi(idc_pubi, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_pubn, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_pubn(idc_pubn, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_pubr, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_pubr(idc_pubr, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_subi, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_subi(idc_subi, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_subn, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_subn(idc_subn, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_subr, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_subr(idc_subr, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_incremental1, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_incremental1(idc_incremental1, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_replacement1, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_replacement1(idc_replacement1, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
% end of automatic code

% -------------------------------------------------------------------------------------------------------------
% displayKbNet
% -------------------------------------------------------------------------------------------------------------
predicates
pubSts : (string_list,string_list,string_list) determ (i,i,o).
newSubs : (string_list,string_list,string_list) determ (i,i,o).
removeSub : (string,string_list,string_list) determ (i,i,o).    
clauses

displayKbNet(OpenShare,Op,Os,Pubs,Subs) :-
    OpenShare=0,
    guix::ctlDisable(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1]),
    guix::guiCB_Set(thisWin,idc_open_share,-1),
    guix::lbxPop(thisWin,idc_publications,Pubs),
    guix::lbxGetAll(thisWin,idc_stall,Sts),
    pubSts(Pubs,Sts,Stsp),
    guix::lbxPop(thisWin,idc_sts,Stsp),
    guix::lbxPop(thisWin,idc_subscriptions,Subs),
    guix::lbxGetAll(thisWin,idcx_pubs,PubsAll),
    newSubs(Subs,PubsAll,Nsubs),
    guix::lbxPop(thisWin,idc_subs,Nsubs),
    guix::ctlDisable(thisWin,[idc_pubi,idc_pubr,idc_pubn,idc_subr,idc_subi,idc_subn,idc_def_db,idct_def_db]),
    guix::ctlEnable(thisWin,[idc_subscriptions,idc_subs,idc_incremental1,idc_replacement1,idc_adhoc_subscription,idc_add1,idc_remove1,idc_add,idc_remove]),
    guix::ctlEnable(thisWin,[idc_publications,idc_sts,idct_table,idc_tablename,idc_incremental,idc_replacement,idc_adhoc_publication,idc_sub_now,idc_pub_now,idc_nl,idc_pr,idc_xmlnl,idc_xmlpr,idc_table]),
    !
;
    set::member(OpenShare,[1,9,10,11]),
    guix::guiRB_SetN(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1],[1,9,10,11],OpenShare),
    guix::lbxPop(thisWin,idc_publications,[]),
    guix::lbxPop(thisWin,idc_subscriptions,[]),
    guix::guiCB_Set(thisWin,idc_open_share,1),
    guix::guiRB_SetN(thisWin,[idc_pubi,idc_pubr,idc_pubn],[1,2,3],Op),
    guix::guiRB_SetN(thisWin,[idc_subi,idc_subr,idc_subn],[1,2,3],Os),
    guix::ctlEnable(thisWin,[idct_def_db,idc_def_db,idct_table,idc_tablename,idc_pubi,idc_pubr,idc_pubn,idc_subr,idc_subi,idc_subn]),
    guix::ctlDisable(thisWin,[idc_open_share,idc_publications,idc_sts,idc_incremental,idc_replacement,idc_adhoc_publication,idc_pub_now,idc_nl,idc_pr,idc_xmlnl,idc_xmlpr,idc_table,idc_add,idc_remove,idc_add1,idc_remove1]),
    guix::ctlDisable(thisWin,[idc_subscriptions,idc_subs,idc_incremental1,idc_replacement1,idc_adhoc_subscription,idc_sub_now,idc_add1,idc_remove1]),
    guix::ctlDisable(thisWin,[idc_nl1,idc_pr1,idc_xmlnl1,idc_xmlpr1,idc_def_db,idct_def_db,idc_tablename]),
    !
;
    S1=toString(OpenShare),
    Mess=string::concat("Invalid open share format: ",S1,"!"),
    dlgIF::dlgNote(2,Mess),
    !.
    
pubSts([],Sts,Sts) :- !.
pubSts([P|Pubs],StsAll,Sts) :-
    sx::sxUnpack(P,'|',L),
    L=[St|_],
    Sts2=set::removeElement(St,StsAll),
    pubSts(Pubs,Sts2,Sts),
    !.
    
newSubs([],L,L) :- !.
newSubs([S|Subs],PubsAll,NewSubs) :-
    sx::sxUnpack(S,'|',L),
    L=[St|_],
    removeSub(St,PubsAll,Pubs2),
    newSubs(Subs,Pubs2,NewSubs),
    !.

removeSub(St,[P|Ps],Pubs) :-
    L=sx::unpack(P,"|"),
    L=[St|_],
    Pubs=Ps,
    !
;
    removeSub(St,Ps,Pubs2),
    Pubs=[P|Pubs2],
    !.
% -------------------------------------------------------------------------------------------------------------
% updateOpenPub/Sub
% -------------------------------------------------------------------------------------------------------------
updateOpenPub() :-
    guix::guiRB_Get(thisWin,[idc_pubi,idc_pubr,idc_pubn],[1,2,3],ModeN),
    ModeN=3,
    guix::txtGet(thisWin,idc_nets,Net),
    Con=sqlBase::getConnection(),
    Kb=sqlBase::getDbKb(),
    Sql2=string::concatList(["delete from p3publications.publication where net='",Net,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql2),
    !
;
    guix::guiRB_Get(thisWin,[idc_pubi,idc_pubr,idc_pubn],[1,2,3],ModeN),
    ModeN<3,
    guix::txtGet(thisWin,idc_nets,Net),
    ModeS=toString(ModeN),
    Con=sqlBase::getConnection(),
    Kb=sqlBase::getDbKb(),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    Sql2=string::concatList(["delete from p3publications.publication where net='",Net,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql2),
    Sql4=string::concatList(["insert p3publications.publication (net,kb,mode,last_pub,st,st_format,table_name) values ('",Net,"','",Kb,"',",ModeS,",'",Time,"','',0,'')"]),            
    sqlTools::executeDirect(Con,Sql4),
    !.    
    
updateOpenSub() :-
    guix::guiRB_Get(thisWin,[idc_subi,idc_subr,idc_subn],[1,2,3],ModeN),
    ModeN=3,
    guix::txtGet(thisWin,idc_nets,Net),
    Con=sqlBase::getConnection(),
    Kb=sqlBase::getDbKb(),
    Sql2=string::concatList(["delete from p3publications.subscription where net='",Net,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql2),
    !
;
    guix::guiRB_Get(thisWin,[idc_subi,idc_subr,idc_subn],[1,2,3],ModeN),
    ModeN<3,
    guix::txtGet(thisWin,idc_nets,Net),
    ModeS=toString(ModeN),
    Con=sqlBase::getConnection(),
    Kb=sqlBase::getDbKb(),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    Sql2=string::concat("delete from p3publications.subscription where net='",Net,"' and kb='",Kb,"'"),
    sqlTools::executeDirect(Con,Sql2),
    Sql4=string::concatList(["insert p3publications.subscription (net,kb,mode,last_sub,st) values ('",Net,"','",Kb,"',",ModeS,",'",Time,"','')"]),            
    sqlTools::executeDirect(Con,Sql4),
    !.    
    
end implement stTables
