/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement taskWindow
open core,vpiDomains,resourceIdentifiers,termDomain,globalDomains

constants
className = "com/rudstrand/TaskWindow/taskWindow".
classVersion = "1.0".
mdiProperty : integer = core::b_true.

class facts
customMenu : vpiDomains::menu := erroneous.

facts
thisWin : vpiDomains::windowHandle := erroneous.

predicates
preCheck : () determ.
preCheckX : (integer,integer,integer,integer) determ.
menuDisable : ().
eventHandler : vpiDomains::ehandler.
onCreate : vpiDomains::createHandler.
     
clauses
classInfo(className, classVersion).

getMenu()=Menu :- Menu=customMenu,!.

show():-
    vpi::setAttrVal(attr_win_mdi, mdiProperty),
    vpi::setErrorHandler(vpi::dumpErrorHandler),
    vpi::init(windowFlags, eventHandler, menu, "", title).

eventHandler(Win, Event) = Result :-
    Result = generatedEventHandler(Win, Event).

    onCreate(_CreationData) = defaultHandling() :-
	vpi::cursorSetWait(),
	_=vpiStartup::create(thisWin,"pro3_spl.bmp"),
    CL=mainExe::getCommandLine(),
	init::globals,
	init::exeDir,
	init::loadP3_Prefs,
	init::exeDir,
 	guix::addWin(1,thisWin),
 	editorWindow::setEditRect(10,30),
 	editorWindow::setFileTypes(["XML-files","*.xml","NL-files","*.3nl","PR-files","*.3pr","Delimited files","*.3dl",
	                                          "Report-files","*.3rp","Trace-files","*.3tc","Format-files","*.3ff","Note-files","*.txt"]),
    guix::setStyle(thisWin,[],[]),
    _=assistant::new(thisWin),
	vpiStartup::close("pro3_spl.bmp"),
	vpi::cursorSet(thisWin,cursor_Arrow),
	dbOpen::openKB(CL),
	_=vpiStartup::create(thisWin,"pro3_spl.bmp"),
	DBs=pop::getDB_Infos(),
	AW=guix::getWinOfClas(4),
	Nodes=pop::getNodesNL(),
	guix::lbxPop(AW,idc_nodes,Nodes),
	guix::lbxPop(AW,idc_dbs,DBs),
	X22=globalVars::get_i("awx"),
	Y22=globalVars::get_i("awy"),
	H22x=X22+773,
	H22y=Y22+373,
	vpi::winMove(AW,rct(X22,Y22,H22x,H22y)),
	cache::loadTmConditionally,
	cache::loadSmConditionally,
	datetime::tsNow(Time),
	dateTime::tsStr(Time,S),
	tt::tt_s(1,"execution time",S),
	Menu=vpi::menuGet(thisWin),
	dbExtension::kbQueryMenuTree(Menu,QMenu),
	vpi::menuSet(thisWin,QMenu),
	menuDisable,
	customMenu:=vpi::menuGet(thisWin),
	dlgIF::msgS("Pro/3 ready."),
	dlgIF::reportState,
	vpiStartup::close("pro3_spl.bmp"),
	globalVars::set_i("assready",1),
    TraceLevel=tt::getLevel(),
    STraceLevel=toString(TraceLevel),
	vpiToolbar::setValue(thisWin,idt_tracelevel,vpiToolbar::text_value(STraceLevel)),
	vpi::cursorSet(thisWin,cursor_Arrow),
	!
;
    !.
               
    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            thisWin := erroneous.
    predicates
        onSizeChanged : vpiDomains::sizeHandler.
    clauses
        onSizeChanged(_Width, _Height) = defaultHandling():-
            vpiToolbar::resize(thisWin).
            %vpiMessage::resize(thisWin).

/* query menu */

    predicates
        onquery : vpiDomains::menuItemHandler.
    clauses
        onquery(_MenuTag) = handled(0).
        
/*  F I L E */

/* recent files */

    predicates
        onFileRecentFiles : vpiDomains::menuItemHandler.
    clauses
        onFileRecentFiles(_MenuTag) = handled(0) :-
        L=fio::getRecent("E"),
        Ok=vpiCommonDialogs::listSelect("Recently used files",L,0,File,_),
        Ok=b_true,
        file::existFile(File),
        editorWindow::openFiles([File]),
        fio::addRecent("E",File),
        !;!.
        
/* New/Open/Save/Save as/Exit */

    predicates
        onFileNew : vpiDomains::menuItemHandler.
    clauses
        onFileNew(_MenuTag) = handled(0) :-
        Wr=globalVars::get_i("f_wrap"),
        set::setGet([-1,0,1],Ix,Wr),
        set::setGet([0,0,1],Ix,Wrap),
        _=editorwindow::new(thisWin,"New.3nl",Wrap),
        !;!.

    predicates
        onFileOpen : vpiDomains::menuItemHandler.
    clauses
        onFileOpen(_MenuTag) = handled(0) :-
  	        editorWindow::getEditorFileDefaults(Mask,List,Path),
	        _=vpiCommonDialogs::getFileName(Mask,List,"Open Files",[dlgfn_MultiSel],Path,Filenames),
            editorWindow::openFiles(FileNames),
            fio::addRecentList("E",FileNames),
            !;!.

    predicates
        onFileSave : vpiDomains::menuItemHandler.
    clauses
        onFileSave(_MenuTag) = handled(0).

    predicates
        onFileSaveAs : vpiDomains::menuItemHandler.
    clauses
        onFileSaveAs(_MenuTag) = handled(0).
        
/* print */        

    predicates
        onFilePrint : vpiDomains::menuItemHandler.
    clauses
        onFilePrint(_MenuTag) = handled(0) :-
        W=vpi::winGetActiveWindow(),
        guix::getWinClas(W)=0, % editor text
        Name=guix::getWinName(W),
        S=vpiEditor::getText(W),
        print::print_str(Name,S,8,-1),
        !
;
        W=vpi::winGetActiveWindow(),
        guix::getWinClas(W)=201, % sentence window
        Name=guix::getWinName(W),
        tedButton::getAddress(W,Loc),
        table::sentencePrint(Loc,Name),
        sdb::sdbSetStatus,
        !
;
        W=vpi::winGetActiveWindow(),
        guix::getWinClas(W)=203, % sentence table
        guix::getWinAll(_,Name,W,SolNo,TabNo,_),
        Dlg = tableprint::new(),
        Dlg:show(thisWin,Name,SolNo,TabNo),
        sdb::sdbSetStatus,
        !
;
        W=vpi::winGetActiveWindow(),
        guix::getWinClas(W)=40, % rule editor (rule tree is 41)
        rub::rubTW_Get(W,TrW),
        _=vpi::winSendEvent(TrW,e_Menu(id_file_print,0)),
        sdb::sdbSetStatus,
        !
;
        W=vpi::winGetActiveWindow(),
        Clas=guix::getWinClas(W),
        set::member(Clas,[44,45,50]),
        _=vpi::winSendEvent(W,e_Menu(id_file_print,0)),
        sdb::sdbSetStatus,
        !
;
        !.
            
/* Process window */

    predicates
        onFileProcess : vpiDomains::menuItemHandler.
    clauses
        onFileProcess(_MenuTag) = handled(0) :-
        vpi::cursorSetWait(),
        W=vpi::winGetActiveWindow(),
                        tt::startProfile("Process Window"),
        proc::processWin(W),
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
                        tt::stopProfile,
	    sdb::sdbSetStatus,
        !;!.

/* Process all windows */

    predicates
        onFileProcessAll : vpiDomains::menuItemHandler.
    clauses
        onFileProcessAll(_MenuTag) = handled(0) :-
        vpi::cursorSetWait(),
        proc::processWins,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	    sdb::sdbSetStatus,
        !;!.

/* Process files */

    predicates
        onFileProcessFile : vpiDomains::menuItemHandler.
    clauses
        onFileProcessFile(_MenuTag) = handled(0):-
        fio::getProcFileDefaults(Mask,List,Path),
	    _=vpiCommonDialogs::getFileName(Mask,List,"Process Files",[dlgfn_MultiSel],Path,Filenames),
        vpi::cursorSetWait(),
        proc::delRealm(FileNames,L),
        dep::massSentenceUpdateStart,
        proc::processFiles(L),
        dep::massSentenceUpdateEnd,
        set::size(L,N),
        S=toString(N),
        Mess=string::concat(S," files processed!"),
        dlgIF::dlgNote(0,Mess),
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	    sdb::sdbSetStatus,
        !;!.

/* display sr in rule editor */

    predicates
        onFileDisplaySentenceRule : vpiDomains::menuItemHandler.
    clauses
        onFileDisplaySentenceRule(_MenuTag) = handled(0):-
	L=globalVars::get_i("error_flag")+globalVars::get_i("quit_flag"),
	L>0,
	dlgIF::dlgNote(1,"Cannot process input while Error or Quit-flag is ON. Reset flags ([X]-button in the toolbar) or exit and re-start Pro/3"),
	!
;
	W=vpi::winGetFocus(),
	guix::getWinAll(WType,FileName,W,_,_,_),
	WType=0,
	Type=fileName::getExtension(FileName),
	LType=string::toLowercase(Type),
	LType="xml",
	vpi::cursorSetWait(),
	S=vpiEditor::getText(W),
	transnl::scnText("xml",S,T,_),
	ruleeditor::rtrDisplayXML(T),
    vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!
;
	W=vpi::winGetFocus(),
	guix::getWinAll(WType,FileName,W,_,_,_),
	WType=0,
	Type=fileName::getExtension(FileName),
	LType=string::toLowercase(Type),
	LType="3nl",
	vpi::cursorSetWait(),
	S=vpiEditor::getText(W),
	transnl::scnTextBlock("3nl",S,L,_),
	ruleeditor::rtrDisplayNLs(FileName,1,L),
    vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

/* reload std knowledge */
    
    predicates
        onFileReloadStandardKnowledge : vpiDomains::menuItemHandler.
    clauses
        onFileReloadStandardKnowledge(_MenuTag) = handled(0) :-
	Reply=vpiCommonDialogs::ask("Pro/3 standard knowledge files","You normally load ALL files (except in some multi-database configurations where only a few of the files might be needed).",["All","All ex. functions","Minimum"]),
	proc::loadStandard(Reply),
	!;!.

/* KNOWLEDGE STREAM */

predicates
    onFileCreateKnowledgeStream : vpiDomains::menuItemHandler.
clauses
    onFileCreateKnowledgeStream(_MenuTag) = handled(0) :-
        Dlg=knowledgestream::new(),
        Dlg:show(thisWin),
  	sdb::sdbSetStatus,
  	!;!.

/* file scanning */

    predicates
        onFileScanFilesSelected : vpiDomains::menuItemHandler.
    clauses
        onFileScanFilesSelected(_MenuTag) = handled(0):-
  	P=globalVars::get_s("scan_path"),
	FileName=vpiCommonDialogs::getFileName("*.*",["","*.*"],"Select Files",[dlgfn_MultiSel],P,Files),
	FileName<>"",
	P2=fileName::getPath(FileName),
	globalVars::set_s("scan_path",P2),
	filescan::fscFileScan(Files),
  	sdb::sdbSetStatus,
    !;!.
    
/* change trace level */

    predicates
        onChangeTrace : vpiDomains::menuItemHandler.
    clauses
        onChangeTrace(_MenuTag) = handled(0):-
        N=tt::getLevel(),
        N<5,
        M=N+1,
        tt::setLevel(M),
        Ls=toString(M),
        vpiToolbar::setValue(thisWin,idt_tracelevel,vpiToolbar::text_value(Ls)),
        !
;        
        tt::setLevel(1),
        vpiToolbar::setValue(thisWin,idt_tracelevel,vpiToolbar::text_value("1")),
        !;!.

/* reset trace */
        
    predicates
        onFileResetTrace : vpiDomains::menuItemHandler.
    clauses
        onFileResetTrace(_MenuTag) = handled(0):-
        tt::reset,
	    sdb::sdbSetStatus,
    !;!.

/* display trace */
    predicates
        onFileDisplayTrace : vpiDomains::menuItemHandler.
    clauses
        onFileDisplayTrace(_MenuTag) = handled(0):-
	tt::tt_close,
	F=tt::getFile(),
	editorWindow::openFiles([F]),
  	sdb::sdbSetStatus,
  	!;!.

/* reset quit and error flags */

    predicates
        onFileResetQuitAndErrorFlags : vpiDomains::menuItemHandler.
    clauses
        onFileResetQuitAndErrorFlags(_MenuTag) = handled(0):-
        dep::massSentenceUpdateAbort,
        dlgIF::setFlag("quit_flag",0),
        dlgIF::setFlag("error_flag",0),
        ir::setBreak(0),
        dlgIF::dbaStatuses,
        dlgIF::resetMessages,
	    sdb::sdbSetStatus,
        !;!.

/* close all windows */

    predicates
        onFileCloseAllWindows : vpiDomains::menuItemHandler.
    clauses
        onFileCloseAllWindows(_MenuTag) = handled(0) :-
        editorWindow::closeEditWins,
        guix::getWinAll(Clas,_,W,_,_,_),
        Clas<>1,
        Clas<>4,
        vpi::winDestroy(W),
        guix::dropWin(W),
	    fail;!.
    predicates
        onFileDisplayAssistantwindowOnTop : vpiDomains::menuItemHandler.
    clauses
        onFileDisplayAssistantwindowOnTop(_MenuTag) = handled(0):-
        W=guix::getWinOfClas(4),
        vpi::winBringToTop(W),
        !;!.

/* sql backup */

predicates
    onFileSqlBackup : vpiDomains::menuItemHandler.
clauses
    onFileSqlBackup(_MenuTag) = handled(0) :-
    Reply=vpiCommonDialogs::ask("Backup scope","Do you want to include derived knowledge in the backup?",["Yes","No"]),
    Path=globalVars::get_s("std_path"),
	File=vpiCommonDialogs::getFileName("*.3bc",["*.3bc","SQL backup files","*.*","All files"],"Select an existing or a new backup file",[dlgfn_MultiSel],Path,_),
	File<>"",
	sqlUtility::kbBackup(File,Reply),
  	sdb::sdbSetStatus,
    !;!.
    

/* sql restore */

predicates
    onFileSqlRestore : vpiDomains::menuItemHandler.
clauses
    onFileSqlRestore(_MenuTag) = handled(0) :-
    Path=globalVars::get_s("std_path"),
	File=vpiCommonDialogs::getFileName("*.3bc",["*.3bc","SQL backup files","*.*","All files"],"Select backup file",[dlgfn_MultiSel],Path,_),
	File<>"",
	sqlUtility::kbRestore(File),
  	sdb::sdbSetStatus,
    !;!.

/* EDIT */

/* cut/copy/paste/delete */

    predicates
        onEditCut : vpiDomains::menuItemHandler.
    clauses
        onEditCut(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::cut(EW),
 	!;!.
    predicates
        onEditCopy : vpiDomains::menuItemHandler.
    clauses
        onEditCopy(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::copy(EW),
 	!;!.
    predicates
        onEditPaste : vpiDomains::menuItemHandler.
    clauses
        onEditPaste(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::paste(EW),
 	!;!.
    predicates
        onEditDelete : vpiDomains::menuItemHandler.
    clauses
        onEditDelete(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::delete(EW),
 	!;!.
 	
 	/* undo/redo */
 	
    predicates
        onEditUndo : vpiDomains::menuItemHandler.
    clauses
        onEditUndo(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::undo(EW),
 	!;!.
    predicates
        onEditRedo : vpiDomains::menuItemHandler.
    clauses
        onEditRedo(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::redo(EW),
 	!;!.
    /* find etc. */
    
    predicates
        onEditFind : vpiDomains::menuItemHandler.
    clauses
        onEditFind(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::searchDialog(EW),
 	!;!.
    predicates
        onEditFindNext : vpiDomains::menuItemHandler.
    clauses
        onEditFindNext(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::searchAgain(EW),
 	!;!.
    predicates
        onEditReplace : vpiDomains::menuItemHandler.
    clauses
        onEditReplace(_MenuTag) = handled(0):-
  	sdb::sdbSetStatus,
  	EW=vpi::winGetFocus(), 
 	vpiEditor::replaceDialog(EW),
 	!;!.

/* tree up/down/left/right */

    predicates
        onDirUp : vpiDomains::menuItemHandler.
    clauses
        onDirUp(_MenuTag) = handled(0):-
	W=vpi::winGetActiveWindow(),
	_=vpi::winSendEvent(W,e_Menu(id_dir_up,0)),
	    sdb::sdbSetStatus,
	!;!.
    predicates
        onDirDown : vpiDomains::menuItemHandler.
    clauses
        onDirDown(_MenuTag) = handled(0):-
	W=vpi::winGetActiveWindow(),
	_=vpi::winSendEvent(W,e_Menu(id_dir_down,0)),
	    sdb::sdbSetStatus,
	!;!.
    predicates
        onDirRight : vpiDomains::menuItemHandler.
    clauses
        onDirRight(_MenuTag) = handled(0):-
	W=vpi::winGetActiveWindow(),
	_=vpi::winSendEvent(W,e_Menu(id_dir_right,0)),
	    sdb::sdbSetStatus,
	!;!.
    predicates
        onDirLeft : vpiDomains::menuItemHandler.
    clauses
        onDirLeft(_MenuTag) = handled(0):-
	W=vpi::winGetActiveWindow(),
	_=vpi::winSendEvent(W,e_Menu(id_dir_left,0)),
	    sdb::sdbSetStatus,
	!;!.

/* SERVER */

    predicates
        onserver : vpiDomains::menuItemHandler.
    clauses
        onserver(_MenuTag) = handled(0):-
            Dlg = servermode::new(),
            Dlg:show(thisWin),
            sdb::sdbSetStatus,
        !;!.

/* K B */

/* initialization parameters */

    predicates
        onKnowledgeBaseInitializationParameters : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseInitializationParameters(_MenuTag) = handled(0) :-
        preCheck,
        Dlg = kbinit::new(),
        Dlg:show(thisWin,2),
        _=globalVars::popRetPars(),
        sdb::sdbSetStatus,
        !;!.

/* re-init KB */

    predicates
        onKnowledgeBaseReinitialize : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseReinitialize(_MenuTag) = handled(0):-
        preCheck,
        dbOpen::reinitialize,
        sdb::sdbSetStatus,
        !;!.

/* realms/chains */

    predicates
        onKnowledgeBaseRealmschains : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseRealmschains(_MenuTag) = handled(0) :-
        preCheck,
        _=records::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* record lookup */

    predicates
        onKnowledgeBaseRecordLookup : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseRecordLookup(_MenuTag) = handled(0) :-
        preCheck,
        _=lookup::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* suspended records */

    predicates
        onKnowledgeBaseSuspendedRecords : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseSuspendedRecords(_MenuTag) = handled(0):-
        preCheck,
        _=suspended::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* create cluster */

predicates
    onKnowledgeBaseCreateCluster : vpiDomains::menuItemHandler.
clauses
    onKnowledgeBaseCreateCluster(_MenuTag) = handled(0) :-
        Dlg=createcluster::new(),
        Dlg:show(thisWin),
        sdb::sdbSetStatus,
        !;!.

/* public/external chains */

    predicates
        onKnowledgeBasePublicChains : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBasePublicChains(_MenuTag) = handled(0):-
        preCheck,
        _=publicchains::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

    predicates
        onKnowledgeBaseExternalChains : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseExternalChains(_MenuTag) = handled(0):-
        preCheck,
        _=externalchains::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.
        
/* update graphs */        

    predicates
        onKnowledgeBaseUpdateDependencyGraphsInKb : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseUpdateDependencyGraphsInKb(_MenuTag) = handled(0) :-
        preCheck,
        vpi::cursorSetWait(),
        dep::saveGraphs,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.
        

/* regeneration of indexes and graphs */

    predicates
        onKnowledgeBaseRegenerateIndexes : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseRegenerateIndexes(_MenuTag) = handled(0):-
        preCheck,
        vpi::cursorSetWait(),
        dbBase::indexRegen,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.
    
    predicates
        onKnowledgeBaseRegenerateSentenceModelGraph : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseRegenerateSentenceModelGraph(_MenuTag) = handled(0):-
        preCheck,
        vpi::cursorSetWait(),
        dep::depSMM_Regen,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.

    predicates
        onKnowledgeBaseRegenerateDependencyGraph : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseRegenerateDependencyGraph(_MenuTag) = handled(0):-
        preCheck,
        vpi::cursorSetWait(),
        dep::regenDependencyGraph,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.

    predicates
        onKnowledgeBaseRegenerateDerivationGraph : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseRegenerateDerivationGraph(_MenuTag) = handled(0):-
        preCheck,
        vpi::cursorSetWait(),
        der::genDerivationGraph,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.

    predicates
        onKnowledgeBaseRegenerateIndexesAndGraphs : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseRegenerateIndexesAndGraphs(_MenuTag) = handled(0):-
        preCheck,
        vpi::cursorSetWait(),
        dbBase::indexRegen,
        globalVars::get_i("quit_flag")=0,
        dep::depSMM_Regen,
        globalVars::get_i("quit_flag")=0,
        dep::regenDependencyGraph,
        globalVars::get_i("quit_flag")=0,
        der::genDerivationGraph,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.

/* check integrity */

predicates
    onKnowledgeBaseCheckDatabaseIntegrity : vpiDomains::menuItemHandler.
clauses
    onKnowledgeBaseCheckDatabaseIntegrity(_MenuTag) = handled(0) :-
        preCheck,
        vpi::cursorSetWait(),
        sqlUtility::checkKb(-1),
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.

predicates
    onKnowledgeBaseCheckAndRepairDatabaseIntegrity : vpiDomains::menuItemHandler.
clauses
    onKnowledgeBaseCheckAndRepairDatabaseIntegrity(_MenuTag) = handled(0) :-
        preCheck,
        vpi::cursorSetWait(),
        sqlUtility::checkKb(1),
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.

/* source backup */

    predicates
        onFileSourceBackup : vpiDomains::menuItemHandler.
    clauses
        onFileSourceBackup(_MenuTag) = handled(0):-
        preCheck,
        _=backup::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* garbage collection */

    predicates
        onKnowledgeBaseGarbageCollection : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseGarbageCollection(_MenuTag) = handled(0):-
        preCheck,
        globalVars::get_i("quit_flag")=0,
        globalVars::get_i("error_flag")=0,
        vpi::cursorSetWait(),
        dbaBase::database(DB,_,_,0),
        DB:db_garbagecollect(),
        globalVars::set_i("knowopen",2),
        dlgIF::dlgNote(0,"Garbage collected!"),
        sdb::sdbSetStatus,
        !;!.

/* make copy */
        
    predicates
        onKnowledgeBaseMakeCopy : vpiDomains::menuItemHandler.
    clauses
        onKnowledgeBaseMakeCopy(_MenuTag) = handled(0) :-
        preCheck,
        globalVars::get_i("quit_flag")=0,
        globalVars::get_i("error_flag")=0,
        vpi::cursorSetWait(),
        S=globalVars::get_s("kb_file"),
        fio::file_name(S,P,F,_),
        Fc=string::concat(P,"copy of ",F,".3kb"),
        dbaBase::database(KB,_,_,0),
        dep::saveGraphs,
        KB:db_flush,
        KB:db_copy(Fc,chainDB::in_file),
        Mess=string::concat("Database copied to '",Fc,"'!"),
        dlgIF::dlgNote(0,Mess),
        sdb::sdbSetStatus,
        !;!.

/* open SQL connection (for 3KB type database using subs/pubs) */

predicates
    onKnowledgeBaseOpenSqlConnection : vpiDomains::menuItemHandler.
clauses
    onKnowledgeBaseOpenSqlConnection(_MenuTag) = handled(0) :-
    sqlBase::openConnection,
    !;!.

/* export to SQL */

predicates
    onKnowledgeBaseExportToSql : vpiDomains::menuItemHandler.
clauses
    onKnowledgeBaseExportToSql(_MenuTag) = handled(0) :-
        preCheck,
        globalVars::get_i("quit_flag")=0,
        globalVars::get_i("error_flag")=0,
        Dlg=export::new(),
        K=globalVars::get_s("kb_file"),
        N=fileName::getName(K),
        Dlg:show(thisWin,N),
        sdb::sdbSetStatus,
        !;!.
        
/* S T R U C T U R E */
    
/* context */
    
    predicates
        onStructureContext : vpiDomains::menuItemHandler.
    clauses
        onStructureContext(_MenuTag) = handled(0) :-
        preCheck,
        D=context::new(),
        D:show(thisWin),
        !;!.

/* rename segment */

predicates
    onStructureRenameSegment : vpiDomains::menuItemHandler.
clauses
    onStructureRenameSegment(_MenuTag) = handled(0):-
        preCheck,
        D=renameSegment::new(),            
        D:show(thisWin),
        !;!.

/* current terminology */

    predicates
        onStructureCurrentTerminology : vpiDomains::menuItemHandler.
    clauses
        onStructureCurrentTerminology(_MenuTag) = handled(0):-
        preCheck,
        _=language::new(thisWin),
	    sdb::sdbSetStatus,
	    !;!.

/* kb names */

    predicates
        onStructureKbNames : vpiDomains::menuItemHandler.
    clauses
        onStructureKbNames(_MenuTag) = handled(0):-
        preCheck,
        _=kbname::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* nl names */

    predicates
        onStructureNlNames : vpiDomains::menuItemHandler.
    clauses
        onStructureNlNames(_MenuTag) = handled(0):-
        preCheck,
        _=nlname::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* application identifiers */

    predicates
        onStructureApplicationIdentifiers : vpiDomains::menuItemHandler.
    clauses
        onStructureApplicationIdentifiers(_MenuTag) = handled(0):-
        preCheck,
        _=identifiers::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* st */

predicates
    onStructureSentenceType : vpiDomains::menuItemHandler.
clauses
    onStructureSentenceType(_MenuTag) = handled(0) :-
        preCheck,
        _=sm::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* sm element */

    predicates
        onStructureSentenceModelElement : vpiDomains::menuItemHandler.
    clauses
        onStructureSentenceModelElement(_MenuTag) = handled(0):-
        preCheck,
        _=sentencemodel::new(thisWin,-1),
	    sdb::sdbSetStatus,
        !;!.
    
/* annotations */

    predicates
        onStructureAnnotations : vpiDomains::menuItemHandler.
    clauses
        onStructureAnnotations(_MenuTag) = handled(0):-
        preCheck,
        _=annotation::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* table layout window */

    predicates
        onSentencesTableLayout : vpiDomains::menuItemHandler.
    clauses
        onSentencesTableLayout(_MenuTag) = handled(0):-
        W=vpi::winGetActiveWindow(),
        guix::getWinAll(Clas,ST,W,SolNo,TabNo,_),
        Clas=203, % table window
        preCheck,
        layout::display(ST,SolNo,TabNo),
        !
;        
        preCheck,
        _=tablelayout::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* validate table layouts */

    predicates
    onStructureValidateTableLayouts : vpiDomains::menuItemHandler.
clauses
    onStructureValidateTableLayouts(_MenuTag) = handled(0) :-
        preCheck,
        vpi::cursorSetWait(),
        layout::validate,
	    sdb::sdbSetStatus,
	    vpiCommonDialogs::note("Table layouts validated!"),
        !;!.
    
predicates
    ontmc : vpiDomains::menuItemHandler.
clauses
    ontmc(_MenuTag) = handled(0) :-
        preCheck,
        vpi::cursorSetWait(),
        cache::toggleTm,
        sdb::sdbSetStatus,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        !;!.

predicates
    onsmc : vpiDomains::menuItemHandler.
clauses
    onsmc(_MenuTag) = handled(0) :-
        preCheck,
        vpi::cursorSetWait(),
        cache::toggleSm,
        sdb::sdbSetStatus,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        !;!.

/* R U L E S */

/* sentence rules */

    predicates
        onRulesSentenceRules : vpiDomains::menuItemHandler.
    clauses
        onRulesSentenceRules(_MenuTag) = handled(0) :-
        _=rules::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* inexact rules */

    predicates
        onRulesInexactReasoningRules : vpiDomains::menuItemHandler.
    clauses
        onRulesInexactReasoningRules(_MenuTag) = handled(0):-
        _=guix::getWinOfClas(60),
        dlgIF::dlgNote(1,"Only one instance of the inexact rule window can be opened!"),
        !
;        
        _=inexrules::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* functions */

    predicates
        onRulesFunctions : vpiDomains::menuItemHandler.
    clauses
        onRulesFunctions(_MenuTag) = handled(0):-
        _=function::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* graphs */

    predicates
        onRulesKnowledgeDependencyGraphs : vpiDomains::menuItemHandler.
    clauses
        onRulesKnowledgeDependencyGraphs(_MenuTag) = handled(0):-
        _=knowdeps::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* delete */

predicates
    onRulesDeleteAllSentenceRules : vpiDomains::menuItemHandler.
clauses
    onRulesDeleteAllSentenceRules(_MenuTag) = handled(0):-
	Reply=vpiCommonDialogs::ask("Please confirm","Do you want to delete all sentence rules?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	vpi::cursorSetWait(),
	dbBase::recBasetypeLocsAll(3,L3),
	dbBase::recBasetypeLocsAll(5,L5),
	dbBase::recordsDelete(L3),
	dbBase::recordsDelete(L5),
	dlgIF::msgS("All sentence rules deleted!"),
	sdb::sdbSetStatus,
	!;!.

predicates
    onRulesDeleteAllInexactRules : vpiDomains::menuItemHandler.
clauses
    onRulesDeleteAllInexactRules(_MenuTag) = handled(0):-
	Reply=vpiCommonDialogs::ask("Please confirm","Do you want to delete all inexact rules?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	vpi::cursorSetWait,
    dbBase::irDeleteAll,
	dlgIF::msgS("All inexact rules deleted!"),
	sdb::sdbSetStatus,
	!;!.

    predicates
        onRulesActivateAllSentenceRules : vpiDomains::menuItemHandler.
    clauses
        onRulesActivateAllSentenceRules(_MenuTag) = handled(0):-
	vpi::cursorSetWait(),
	dep::depSetRuleStatusAll(3),
	vpi::cursorSet(thisWin,cursor_Arrow),
  	sdb::sdbSetStatus,
	!;!.

    predicates
        onRulesDeactivateAllSentenceRules : vpiDomains::menuItemHandler.
    clauses
        onRulesDeactivateAllSentenceRules(_MenuTag) = handled(0):-
	vpi::cursorSetWait(),
	dep::depSetRuleStatusAll(-3),
	vpi::cursorSet(thisWin,cursor_Arrow),
  	sdb::sdbSetStatus,
	!;!.

    predicates
        onRulesSetAllDependenciesAsGenerated : vpiDomains::menuItemHandler.
    clauses
        onRulesSetAllDependenciesAsGenerated(_MenuTag) = handled(0):-
	vpi::cursorSetWait(),
	dep::setStatus("p3_pDP",int(0),int(1)),
	dlgIF::msgS("The entire dependency graph set with status=generated."),
	vpi::cursorSet(thisWin,cursor_Arrow),
	sdb::sdbSetStatus,
	!;!.

    predicates
        onRulesSetAllDependenciesAsUngenerated : vpiDomains::menuItemHandler.
    clauses
        onRulesSetAllDependenciesAsUngenerated(_MenuTag) = handled(0):-
	vpi::cursorSetWait(),
	dep::setStatus("p3_pDP",int(1),int(0)),
	dlgIF::msgS("The entire dependency graph set with status=ungenerated."),
	vpi::cursorSet(thisWin,cursor_Arrow),
	sdb::sdbSetStatus,
	!;!.

/* S E N T E N C E S */

/* list sentences */

    predicates
        onSentencesListSentences : vpiDomains::menuItemHandler.
    clauses
        onSentencesListSentences(_MenuTag) = handled(0) :-
        vpi::cursorSetWait(),
        _=sentences::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* ST tables */

predicates
    onSentencesSttables : vpiDomains::menuItemHandler.
clauses
    onSentencesSttables(_MenuTag) = handled(0) :-
        vpi::cursorSetWait(),
        _=stTables::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* inexact rule reasoning */

    predicates
        onSentencesInexactRuleReasoning : vpiDomains::menuItemHandler.
    clauses
        onSentencesInexactRuleReasoning(_MenuTag) = handled(0):-
        _=reasoning::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.
    predicates
        onSentencesDeleteInexactRuleReasoningSentences : vpiDomains::menuItemHandler.
    clauses
        onSentencesDeleteInexactRuleReasoningSentences(_MenuTag) = handled(0):-
	dbBase::accessErrorChain("r"),
	!
;	
	Reply=vpiCommonDialogs::ask("Please confirm","Do you want to delete all inexact rule reasoning sentences?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	vpi::cursorSetWait(),
	dbBase::chainDelete("r"),
	dlgIF::msgS("All inexact rule reasoning records deleted!"),
	sdb::sdbSetStatus,
	!;!.

/* derivations */

    predicates
        onSentencesSentenceDerivation : vpiDomains::menuItemHandler.
    clauses
        onSentencesSentenceDerivation(_MenuTag) = handled(0):-
        _=derivation::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* derivation */

predicates
    onSentencesDeriveSentences : vpiDomains::menuItemHandler.
clauses
    onSentencesDeriveSentences(_MenuTag) = handled(0) :-
	sdb::sdbSetStatus,
	tt::getLevel()>1,
	vpiCommonDialogs::resp_default<>vpiCommonDialogs::ask("Please confirm","Do you really want to perform this operation with trace level greater than 1?",["Yes","No"]),
    !
;    
                    tt::startProfile("Rederivation"),
	der::rederive(),
                    tt::stopProfile,
	    sdb::sdbSetStatus,
	!;!.

/* full rederivation */

    predicates
        onSentencesCompleteSentenceRederivation : vpiDomains::menuItemHandler.
    clauses
        onSentencesCompleteSentenceRederivation(_MenuTag) = handled(0):-
	sdb::sdbSetStatus,
	tt::getLevel()>1,
	vpiCommonDialogs::resp_default<>vpiCommonDialogs::ask("Please confirm","Do you really want to perform this operation with trace level greater than 1?",["Yes","No"]),
    !
;    
        vpi::cursorSetWait(),
			datetime::tsNow(Ts),
	Rep=rep::newReport(2,"Derivation of All Sentences",70),
                    tt::startProfile("Full Rederivation"),
	der::rederiveAll(Rep,Nt,Ns),
                    tt::stopProfile,
	Rep:close,
			datetime::tsNow(Te),
			datetime::tsDiff(Ts,Te,D,H,M,S),
			datetime::timDurText(D,H,M,S,Time),
			Nts=toString(Nt),
			Nss=toString(Ns),
	sx::concat(["Generated ",Nss," sentences of ",Nts," types. Elapsed time ",Time,"."],Mess),
	dlgIF::msgS(Mess),
	dlgIF::msgStatus(Mess),
	!;!.

/* process cluster */

predicates
    onSentencesProcessCluster : vpiDomains::menuItemHandler.
clauses
    onSentencesProcessCluster(_MenuTag) = handled(0) :-
        vpi::cursorSetWait(),
        Cluster=cluster::getCluster(),
        Kb=cluster::getMember(Cluster,0),
        Net=cluster::getNet(),
        Stream=cluster::getStream(Cluster),
        sqlSt::getShareTable(Net,Share,_),
        cluster::startCluster(Cluster,Kb,Net,Share,Stream),
	    sdb::sdbSetStatus,
	    !
;	    
        not(_=cluster::getCluster()),
        dlgIF::dlgNote(0,"Cluster not defined or erroneous!"),
        !
;	    
        Cluster=cluster::getCluster(),
        not(_=cluster::getMember(Cluster,0)),
        dlgIF::dlgNote(0,"Originating cluster member can not be found!"),
        !
;	    
        not(_=cluster::getNet()),
        dlgIF::dlgNote(0,"Cluster publishing network not defined or erroneous!"),
        !
;	    
        Cluster=cluster::getCluster(),
        not(_=cluster::getStream(Cluster)),
        dlgIF::dlgNote(0,"Knowledge stream not defined or erroneous!"),
        !
;	    
        Net=cluster::getNet(),
        not(sqlSt::getShareTable(Net,_,_)),
        dlgIF::dlgNote(0,"Open share table not defined or erroneous!"),
        !;!.

/* multi-DB rederivation */

    predicates
        onSentencesMultidbRederivation : vpiDomains::menuItemHandler.
    clauses
        onSentencesMultidbRederivation(_MenuTag) = handled(0):-
        _=multikb::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

/* delete all derived sentences */

    predicates
        onSentencesDeleteAllDerivedSentences : vpiDomains::menuItemHandler.
    clauses
        onSentencesDeleteAllDerivedSentences(_MenuTag) = handled(0):-
	vpi::cursorSetWait(),
	dbBase::sentenceDerivedDelete(-1),
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
    sdb::sdbSetStatus,
	!;!.

/* refresh subscribed STs */

predicates
    onSentencesRefreshImportedSentenceTypes : vpiDomains::menuItemHandler.
clauses
    onSentencesRefreshImportedSentenceTypes(_MenuTag) = handled(0) :-
	vpi::cursorSetWait(),
	sqlBase::openConnection,
	sqlSt::subscribeAll,
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
    sdb::sdbSetStatus,
	!;!.

/* refresh published STs */

predicates
    onSentencesRefreshExportedSentenceTypes : vpiDomains::menuItemHandler.
clauses
    onSentencesRefreshExportedSentenceTypes(_MenuTag) = handled(0) :-
	vpi::cursorSetWait(),
	sqlBase::openConnection,
	sqlSt::publishAll,
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
    sdb::sdbSetStatus,
	!;!.

/* REPORTS */

    predicates
        onReportsAllAbout : vpiDomains::menuItemHandler.
    clauses
        onReportsAllAbout(_MenuTag) = handled(0):-
        _=allabout::new(thisWin),
	    sdb::sdbSetStatus,
        !;!.

    predicates
        onReportsTerminology : vpiDomains::menuItemHandler.
    clauses
        onReportsTerminology(_MenuTag) = handled(0):-
        p3rep::repToken(1),
	    sdb::sdbSetStatus,
        !;!.
        
    predicates
        onReportsTerminology2 : vpiDomains::menuItemHandler.
    clauses
        onReportsTerminology2(_MenuTag) = handled(0):-
        p3rep::repToken(2),
	    sdb::sdbSetStatus,
        !;!.

    predicates
        onReportsSentenceTypesAndGroups : vpiDomains::menuItemHandler.
    clauses
        onReportsSentenceTypesAndGroups(_MenuTag) = handled(0):-
        D=sentencerep::new(),
        D:show(thisWin),
        sdb::sdbSetStatus,
        !;!.

predicates
    onReportsInexactRules : vpiDomains::menuItemHandler.
clauses
    onReportsInexactRules(_MenuTag) = handled(0) :-
        vpi::cursorSetWait(),
        irReport::reportRules,
        vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
        sdb::sdbSetStatus,
        !;!.
        
predicates
    onReportsSentenceRulesWhereNoSentencesAreDerived : vpiDomains::menuItemHandler.
clauses
    onReportsSentenceRulesWhereNoSentencesAreDerived(_MenuTag) = handled(0) :-
        p3rep::repNonDerived,
	    sdb::sdbSetStatus,
        !;!.

    predicates
        onReportsUnreferencedKnowledge : vpiDomains::menuItemHandler.
    clauses
        onReportsUnreferencedKnowledge(_MenuTag) = handled(0) :-
        p3rep::unrefKnowledge,
        sdb::sdbSetStatus,
        !;!.

    predicates
        onReportsUnreferencedStructureObjects : vpiDomains::menuItemHandler.
    clauses
        onReportsUnreferencedStructureObjects(_MenuTag) = handled(0):-
        p3rep::unrefStrucKnowledge,
        sdb::sdbSetStatus,
        !;!.

predicates
    onReportsReferentialIntegrityProblems : vpiDomains::menuItemHandler.
clauses
    onReportsReferentialIntegrityProblems(_MenuTag) = handled(0):-
        p3rep::integrityProblems,
        sdb::sdbSetStatus,
        !;!.


/* P R E F E R E N C E S */

/* directories */

    predicates
        onPreferencesDirectoriesAndFiles : vpiDomains::menuItemHandler.
    clauses
        onPreferencesDirectoriesAndFiles(_MenuTag) = handled(0) :-
            Dlg = directories::new(),
            Dlg:show(thisWin),
            sdb::sdbSetStatus,
            !;!.

/* output */

    predicates
        onPreferencesOutputFormat : vpiDomains::menuItemHandler.
    clauses
        onPreferencesOutputFormat(_MenuTag) = handled(0) :-
            Dlg = outputpreferences::new(),
            Dlg:show(thisWin),
            sdb::sdbSetStatus,
            !;!.

/* processing */

    predicates
        onPreferencesProcessing : vpiDomains::menuItemHandler.
    clauses
        onPreferencesProcessing(_MenuTag) = handled(0) :-
            Dlg = processingpreferences::new(),
            Dlg:show(thisWin),
            sdb::sdbSetStatus,
            !;!.

    predicates
        onPreferencesFilescannerPreferences : vpiDomains::menuItemHandler.
    clauses
        onPreferencesFilescannerPreferences(_MenuTag) = handled(0):-
            Dlg = filescannerpref::new(),
            Dlg:show(thisWin),
            sdb::sdbSetStatus,
            !;!.

/* save current preferences */

    predicates
        onPreferencesSaveCurrentPreferences : vpiDomains::menuItemHandler.
    clauses
        onPreferencesSaveCurrentPreferences(_MenuTag) = handled(0) :-
        init::savePrefs,
        sdb::sdbSetStatus,
        !;!.

    predicates
        onPreferencesResetAllflag : vpiDomains::menuItemHandler.
    clauses
        onPreferencesResetAllflag(_MenuTag) = handled(0):-
        globalVars::set_i("all_flag",-1),
        sdb::sdbSetStatus,
        !;!.

    predicates
        onPreferencesResetWindowCascadeposition : vpiDomains::menuItemHandler.
    clauses
        onPreferencesResetWindowCascadeposition(_MenuTag) = handled(0):-
        editorWindow::setEditRect(10,30),
        sdb::sdbSetStatus,
        !;!.

/* H E L P */

    predicates
        onHelpAbout : vpiDomains::menuItemHandler.
    clauses
        onHelpAbout(_MenuTag) = handled(0) :-
            AboutDialog = aboutDialog::new(),
            AboutDialog:show(thisWin),
            sdb::sdbSetStatus,
            !;!.

    predicates
        onHelpGuide : vpiDomains::menuItemHandler.
    clauses
        onHelpGuide(_MenuTag) = handled(0) :-
            File=fioExt::getPath("exe","pro3_Guide.xml"),
            file::existfile(File),
            Dlg=guide::new(),
            Dlg:show(thisWin,0),
            sdb::sdbSetStatus,
            !;!.

    predicates
        onHelpTableOfContents : vpiDomains::menuItemHandler.
    clauses
        onHelpTableOfContents(_MenuTag) = handled(0) :-
        dlgIF::hlpWeb("Documentation"),
	    sdb::sdbSetStatus,
        !;!.

    predicates
        onHelpActiveWindow : vpiDomains::menuItemHandler.
    clauses
        onHelpActiveWindow(_MenuTag) = handled(0) :-
        dlgIF::hlpWeb(""),
	    sdb::sdbSetStatus,
        !;!.

    predicates
        onHelpLicenseAgreement : vpiDomains::menuItemHandler.
    clauses
        onHelpLicenseAgreement(_MenuTag) = handled(0):-
        Dlg=license::new(),
        Dlg:show(thisWin),
	    sdb::sdbSetStatus,
        !;!.

predicates
    onHelpAboutTheKnowledgeBase : vpiDomains::menuItemHandler.
clauses
    onHelpAboutTheKnowledgeBase(_MenuTag) = handled(0) :-
	KBid=globalVars::get_s("kb_id"),
    sentencemodel::dmAnnoDisplayAsst("p3_iDB",Kbid),
    !;!.

/* exit */

    predicates
        onCloseRequest : vpiDomains::closeRequestHandler.
    clauses
        onCloseRequest() = defaultHandling() :-
	init::savePrefs,
	dbOpen::closeKB(1,1,1,1),
    editorWindow::closeEditWins,
	tt::tt_close(),
	vpi::winDestroy(thisWin),
	!
;	
	dbOpen::closeKB(1,1,1,1),
    editorWindow::closeEditWins,
    vpiCommonDialogs::ask("Cannot close all files - KB could be come corrupted","Do you want to exit Pro/3 anyway?",["Yes","No"])=vpiCommonDialogs::resp_default,
    vpi::winDestroy(thisWin),
	!;!.

    predicates
        onexit : vpiDomains::menuItemHandler.
    clauses
        onexit(_MenuTag) = handled(0) :-
	_=vpiStartup::create(thisWin,"pro3_sple.bmp"),
	init::savePrefs,
	dbOpen::closeKB(1,1,1,1),
    editorWindow::closeEditWins,
	tt::tt_close(),
	vpiStartup::close("pro3_sple.bmp"),
	vpi::winDestroy(thisWin),
	!;!.

/* needs to be reinserted after change in task window menu
    generatedEventHandler(_Win, e_Menu(MenuID, _)) = Result :-
            MenuID>=7000,
            MenuID<9000,
            proc::menuQueryRun(MenuID),
            sdb::sdbSetStatus,
            handled(Result) = onquery(id_query),!.
*/            

% This code is maintained by the VDE do not update it manually, 14:44:17-26.11.2006
constants
    windowFlags : vpiDomains::wsflags = [wsf_SizeBorder,wsf_TitleBar,wsf_Maximize,wsf_Minimize,wsf_ClipSiblings,wsf_Maximized,wsf_Close].
    menu : vpiDomains::menu = resMenu(id_TaskMenu).
    title : string = "Pro/3".

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        projectToolbar::create(thisWin),
        statusLine::create(thisWin),
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
        handled(Result) = onSizeChanged(Width, Height).
    generatedEventHandler(_Win, e_Menu(MenuID, _)) = Result :-
            MenuID>=7000,
            MenuID<9000,
            proc::menuQueryRun(MenuID),
            sdb::sdbSetStatus,
            handled(Result) = onquery(id_query),!.
    generatedEventHandler(_Win, e_Menu(id_help_about, _)) = Result :-
            handled(Result) = onHelpAbout(id_help_about).
    generatedEventHandler(_Win, e_Menu(id_file_new, _)) = Result :-
            handled(Result) = onFileNew(id_file_new).
    generatedEventHandler(_Win, e_Menu(id_file_open, _)) = Result :-
            handled(Result) = onFileOpen(id_file_open).
    generatedEventHandler(_Win, e_Menu(id_file_save, _)) = Result :-
            handled(Result) = onFileSave(id_file_save).
    generatedEventHandler(_Win, e_Menu(id_file_save_as, _)) = Result :-
            handled(Result) = onFileSaveAs(id_file_save_as).
    generatedEventHandler(_Win, e_Menu(id_Preferences_directories_and_files, _)) = Result :-
            handled(Result) = onPreferencesDirectoriesAndFiles(id_Preferences_directories_and_files).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_initialization_parameters, _)) = Result :-
            handled(Result) = onKnowledgeBaseInitializationParameters(id_Knowledge_base_initialization_parameters).
    generatedEventHandler(_Win, e_Menu(id_help_guide, _)) = Result :-
            handled(Result) = onHelpGuide(id_help_guide).
    generatedEventHandler(_Win, e_Menu(id_file_process, _)) = Result :-
            handled(Result) = onFileProcess(id_file_process).
    generatedEventHandler(_Win, e_Menu(id_file_process_all, _)) = Result :-
            handled(Result) = onFileProcessAll(id_file_process_all).
    generatedEventHandler(_Win, e_Menu(id_file_close_all_windows, _)) = Result :-
            handled(Result) = onFileCloseAllWindows(id_file_close_all_windows).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_realmschains, _)) = Result :-
            handled(Result) = onKnowledgeBaseRealmschains(id_Knowledge_base_realmschains).
    generatedEventHandler(_Win, e_Menu(id_help_table_of_contents, _)) = Result :-
            handled(Result) = onHelpTableOfContents(id_help_table_of_contents).
    generatedEventHandler(_Win, e_Menu(id_help_active_window, _)) = Result :-
            handled(Result) = onHelpActiveWindow(id_help_active_window).
    generatedEventHandler(_Win, e_Menu(id_help_license_agreement, _)) = Result :-
            handled(Result) = onHelpLicenseAgreement(id_help_license_agreement).
    generatedEventHandler(_Win, e_Menu(id_exit, _)) = Result :-
            handled(Result) = onexit(id_exit).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_record_lookup, _)) = Result :-
            handled(Result) = onKnowledgeBaseRecordLookup(id_Knowledge_base_record_lookup).
    generatedEventHandler(_Win, e_Menu(id_file_reload_standard_knowledge, _)) = Result :-
            handled(Result) = onFileReloadStandardKnowledge(id_file_reload_standard_knowledge).
    generatedEventHandler(_Win, e_Menu(id_Sentences_list_sentences, _)) = Result :-
            handled(Result) = onSentencesListSentences(id_Sentences_list_sentences).
    generatedEventHandler(_Win, e_Menu(id_Structure_context, _)) = Result :-
            handled(Result) = onStructureContext(id_Structure_context).
    generatedEventHandler(_Win, e_Menu(id_query, _)) = Result :-
            handled(Result) = onquery(id_query).
    generatedEventHandler(_Win, e_Menu(id_Preferences_output_format, _)) = Result :-
            handled(Result) = onPreferencesOutputFormat(id_Preferences_output_format).
    generatedEventHandler(_Win, e_Menu(id_file_recent_files, _)) = Result :-
            handled(Result) = onFileRecentFiles(id_file_recent_files).
    generatedEventHandler(_Win, e_Menu(id_file_print, _)) = Result :-
            handled(Result) = onFilePrint(id_file_print).
    generatedEventHandler(_Win, e_Menu(id_Preferences_processing, _)) = Result :-
            handled(Result) = onPreferencesProcessing(id_Preferences_processing).
    generatedEventHandler(_Win, e_Menu(id_Preferences_save_current_preferences, _)) = Result :-
            handled(Result) = onPreferencesSaveCurrentPreferences(id_Preferences_save_current_preferences).
    generatedEventHandler(_Win, e_Menu(id_Rules_sentence_rules, _)) = Result :-
            handled(Result) = onRulesSentenceRules(id_Rules_sentence_rules).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_regenerate_indexes, _)) = Result :-
            handled(Result) = onKnowledgeBaseRegenerateIndexes(id_Knowledge_base_regenerate_indexes).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_regenerate_sentence_model_graph, _)) = Result :-
            handled(Result) = onKnowledgeBaseRegenerateSentenceModelGraph(id_Knowledge_base_regenerate_sentence_model_graph).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_regenerate_dependency_graph, _)) = Result :-
            handled(Result) = onKnowledgeBaseRegenerateDependencyGraph(id_Knowledge_base_regenerate_dependency_graph).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_regenerate_derivation_graph, _)) = Result :-
            handled(Result) = onKnowledgeBaseRegenerateDerivationGraph(id_Knowledge_base_regenerate_derivation_graph).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_regenerate_indexes_and_graphs, _)) = Result :-
            handled(Result) = onKnowledgeBaseRegenerateIndexesAndGraphs(id_Knowledge_base_regenerate_indexes_and_graphs).
    generatedEventHandler(_Win, e_Menu(id_Sentences_sentence_derivation, _)) = Result :-
            handled(Result) = onSentencesSentenceDerivation(id_Sentences_sentence_derivation).
    generatedEventHandler(_Win, e_Menu(id_file_reset_trace, _)) = Result :-
            handled(Result) = onFileResetTrace(id_file_reset_trace).
    generatedEventHandler(_Win, e_Menu(id_file_reset_quit_and_error_flags, _)) = Result :-
            handled(Result) = onFileResetQuitAndErrorFlags(id_file_reset_quit_and_error_flags).
    generatedEventHandler(_Win, e_Menu(id_Sentences_delete_all_derived_sentences, _)) = Result :-
            handled(Result) = onSentencesDeleteAllDerivedSentences(id_Sentences_delete_all_derived_sentences).
    generatedEventHandler(_Win, e_Menu(id_Structure_sentence_model_element, _)) = Result :-
            handled(Result) = onStructureSentenceModelElement(id_Structure_sentence_model_element).
    generatedEventHandler(_Win, e_Menu(id_Sentences_table_layout, _)) = Result :-
            handled(Result) = onSentencesTableLayout(id_Sentences_table_layout).
    generatedEventHandler(_Win, e_Menu(id_Reports_terminology, _)) = Result :-
            handled(Result) = onReportsTerminology(id_Reports_terminology).
    generatedEventHandler(_Win, e_Menu(id_Reports_terminology_2, _)) = Result :-
            handled(Result) = onReportsTerminology2(id_Reports_terminology_2).
    generatedEventHandler(_Win, e_Menu(id_change_trace, _)) = Result :-
            handled(Result) = onChangeTrace(id_change_trace).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_public_chains, _)) = Result :-
            handled(Result) = onKnowledgeBasePublicChains(id_Knowledge_base_public_chains).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_external_chains, _)) = Result :-
            handled(Result) = onKnowledgeBaseExternalChains(id_Knowledge_base_external_chains).
    generatedEventHandler(_Win, e_Menu(id_Sentences_multidb_rederivation, _)) = Result :-
            handled(Result) = onSentencesMultidbRederivation(id_Sentences_multidb_rederivation).
    generatedEventHandler(_Win, e_Menu(id_Rules_inexact_reasoning_rules, _)) = Result :-
            handled(Result) = onRulesInexactReasoningRules(id_Rules_inexact_reasoning_rules).
    generatedEventHandler(_Win, e_Menu(id_file_process_file, _)) = Result :-
            handled(Result) = onFileProcessFile(id_file_process_file).
    generatedEventHandler(_Win, e_Menu(id_file_source_backup, _)) = Result :-
            handled(Result) = onFileSourceBackup(id_file_source_backup).
    generatedEventHandler(_Win, e_Menu(id_Sentences_inexact_rule_reasoning, _)) = Result :-
            handled(Result) = onSentencesInexactRuleReasoning(id_Sentences_inexact_rule_reasoning).
    generatedEventHandler(_Win, e_Menu(id_Sentences_delete_inexact_rule_reasoning_sentences, _)) = Result :-
            handled(Result) = onSentencesDeleteInexactRuleReasoningSentences(id_Sentences_delete_inexact_rule_reasoning_sentences).
    generatedEventHandler(_Win, e_Menu(id_Sentences_complete_sentence_rederivation, _)) = Result :-
            handled(Result) = onSentencesCompleteSentenceRederivation(id_Sentences_complete_sentence_rederivation).
    generatedEventHandler(_Win, e_Menu(id_dir_up, _)) = Result :-
            handled(Result) = onDirUp(id_dir_up).
    generatedEventHandler(_Win, e_Menu(id_dir_down, _)) = Result :-
            handled(Result) = onDirDown(id_dir_down).
    generatedEventHandler(_Win, e_Menu(id_dir_right, _)) = Result :-
            handled(Result) = onDirRight(id_dir_right).
    generatedEventHandler(_Win, e_Menu(id_dir_left, _)) = Result :-
            handled(Result) = onDirLeft(id_dir_left).
    generatedEventHandler(_Win, e_Menu(id_file_display_sentence_rule, _)) = Result :-
            handled(Result) = onFileDisplaySentenceRule(id_file_display_sentence_rule).
    generatedEventHandler(_Win, e_Menu(id_file_scan_files_selected, _)) = Result :-
            handled(Result) = onFileScanFilesSelected(id_file_scan_files_selected).
    generatedEventHandler(_Win, e_Menu(id_file_display_trace, _)) = Result :-
            handled(Result) = onFileDisplayTrace(id_file_display_trace).
    generatedEventHandler(_Win, e_Menu(id_Rules_functions, _)) = Result :-
            handled(Result) = onRulesFunctions(id_Rules_functions).
    generatedEventHandler(_Win, e_Menu(id_Rules_knowledge_dependency_graphs, _)) = Result :-
            handled(Result) = onRulesKnowledgeDependencyGraphs(id_Rules_knowledge_dependency_graphs).
    generatedEventHandler(_Win, e_Menu(id_Structure_kb_names, _)) = Result :-
            handled(Result) = onStructureKbNames(id_Structure_kb_names).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_suspended_records, _)) = Result :-
            handled(Result) = onKnowledgeBaseSuspendedRecords(id_Knowledge_base_suspended_records).
    generatedEventHandler(_Win, e_Menu(id_Structure_current_terminology, _)) = Result :-
            handled(Result) = onStructureCurrentTerminology(id_Structure_current_terminology).
    generatedEventHandler(_Win, e_Menu(id_Structure_nl_names, _)) = Result :-
            handled(Result) = onStructureNlNames(id_Structure_nl_names).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_reinitialize, _)) = Result :-
            handled(Result) = onKnowledgeBaseReinitialize(id_Knowledge_base_reinitialize).
    generatedEventHandler(_Win, e_Menu(id_Structure_application_identifiers, _)) = Result :-
            handled(Result) = onStructureApplicationIdentifiers(id_Structure_application_identifiers).
    generatedEventHandler(_Win, e_Menu(id_file_display_assistantwindow_on_top, _)) = Result :-
            handled(Result) = onFileDisplayAssistantwindowOnTop(id_file_display_assistantwindow_on_top).
    generatedEventHandler(_Win, e_Menu(id_Reports_sentence_types_and_groups, _)) = Result :-
            handled(Result) = onReportsSentenceTypesAndGroups(id_Reports_sentence_types_and_groups).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_garbage_collection, _)) = Result :-
            handled(Result) = onKnowledgeBaseGarbageCollection(id_Knowledge_base_garbage_collection).
    generatedEventHandler(_Win, e_Menu(id_Structure_annotations, _)) = Result :-
            handled(Result) = onStructureAnnotations(id_Structure_annotations).
    generatedEventHandler(_Win, e_Menu(id_Rules_activate_all_sentence_rules, _)) = Result :-
            handled(Result) = onRulesActivateAllSentenceRules(id_Rules_activate_all_sentence_rules).
    generatedEventHandler(_Win, e_Menu(id_Rules_deactivate_all_sentence_rules, _)) = Result :-
            handled(Result) = onRulesDeactivateAllSentenceRules(id_Rules_deactivate_all_sentence_rules).
    generatedEventHandler(_Win, e_Menu(id_Rules_set_all_dependencies_as_generated, _)) = Result :-
            handled(Result) = onRulesSetAllDependenciesAsGenerated(id_Rules_set_all_dependencies_as_generated).
    generatedEventHandler(_Win, e_Menu(id_Rules_set_all_dependencies_as_ungenerated, _)) = Result :-
            handled(Result) = onRulesSetAllDependenciesAsUngenerated(id_Rules_set_all_dependencies_as_ungenerated).
    generatedEventHandler(_Win, e_Menu(id_Preferences_reset_allflag, _)) = Result :-
            handled(Result) = onPreferencesResetAllflag(id_Preferences_reset_allflag).
    generatedEventHandler(_Win, e_Menu(id_Preferences_reset_window_cascadeposition, _)) = Result :-
            handled(Result) = onPreferencesResetWindowCascadeposition(id_Preferences_reset_window_cascadeposition).
    generatedEventHandler(_Win, e_Menu(id_Preferences_filescanner_preferences, _)) = Result :-
            handled(Result) = onPreferencesFilescannerPreferences(id_Preferences_filescanner_preferences).
    generatedEventHandler(_Win, e_Menu(id_Reports_all_about, _)) = Result :-
            handled(Result) = onReportsAllAbout(id_Reports_all_about).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_update_dependency_graphs_in_kb, _)) = Result :-
            handled(Result) = onKnowledgeBaseUpdateDependencyGraphsInKb(id_Knowledge_base_update_dependency_graphs_in_kb).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_make_copy, _)) = Result :-
            handled(Result) = onKnowledgeBaseMakeCopy(id_Knowledge_base_make_copy).
    generatedEventHandler(_Win, e_Menu(id_Reports_unreferenced_knowledge, _)) = Result :-
            handled(Result) = onReportsUnreferencedKnowledge(id_Reports_unreferenced_knowledge).
    generatedEventHandler(_Win, e_Menu(id_Reports_unreferenced_structure_objects, _)) = Result :-
            handled(Result) = onReportsUnreferencedStructureObjects(id_Reports_unreferenced_structure_objects).
    generatedEventHandler(_Win, e_Menu(id_server, _)) = Result :-
            handled(Result) = onserver(id_server).
    generatedEventHandler(_Win, e_Menu(id_Structure_validate_table_layouts, _)) = Result :-
            handled(Result) = onStructureValidateTableLayouts(id_Structure_validate_table_layouts).
    generatedEventHandler(_Win, e_Menu(id_Reports_referential_integrity_problems, _)) = Result :-
            handled(Result) = onReportsReferentialIntegrityProblems(id_Reports_referential_integrity_problems).
    generatedEventHandler(_Win, e_Menu(id_Structure_rename_segment, _)) = Result :-
            handled(Result) = onStructureRenameSegment(id_Structure_rename_segment).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_export_to_sql, _)) = Result :-
            handled(Result) = onKnowledgeBaseExportToSql(id_Knowledge_base_export_to_sql).
    generatedEventHandler(_Win, e_Menu(idt_tmc, _)) = Result :-
            handled(Result) = ontmc(idt_tmc).
    generatedEventHandler(_Win, e_Menu(idt_smc, _)) = Result :-
            handled(Result) = onsmc(idt_smc).
    generatedEventHandler(_Win, e_Menu(id_Rules_delete_all_sentence_rules, _)) = Result :-
            handled(Result) = onRulesDeleteAllSentenceRules(id_Rules_delete_all_sentence_rules).
    generatedEventHandler(_Win, e_Menu(id_Rules_delete_all_inexact_rules, _)) = Result :-
            handled(Result) = onRulesDeleteAllInexactRules(id_Rules_delete_all_inexact_rules).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_check_database_integrity, _)) = Result :-
            handled(Result) = onKnowledgeBaseCheckDatabaseIntegrity(id_Knowledge_base_check_database_integrity).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_check_and_repair_database_integrity, _)) = Result :-
            handled(Result) = onKnowledgeBaseCheckAndRepairDatabaseIntegrity(id_Knowledge_base_check_and_repair_database_integrity).
    generatedEventHandler(_Win, e_Menu(id_Sentences_refresh_imported_sentence_types, _)) = Result :-
            handled(Result) = onSentencesRefreshImportedSentenceTypes(id_Sentences_refresh_imported_sentence_types).
    generatedEventHandler(_Win, e_Menu(id_Sentences_refresh_exported_sentence_types, _)) = Result :-
            handled(Result) = onSentencesRefreshExportedSentenceTypes(id_Sentences_refresh_exported_sentence_types).
    generatedEventHandler(_Win, e_Menu(id_Reports_sentence_rules_where_no_sentences_are_derived, _)) = Result :-
            handled(Result) = onReportsSentenceRulesWhereNoSentencesAreDerived(id_Reports_sentence_rules_where_no_sentences_are_derived).
    generatedEventHandler(_Win, e_Menu(id_file_sql_backup, _)) = Result :-
            handled(Result) = onFileSqlBackup(id_file_sql_backup).
    generatedEventHandler(_Win, e_Menu(id_file_sql_restore, _)) = Result :-
            handled(Result) = onFileSqlRestore(id_file_sql_restore).
    generatedEventHandler(_Win, e_Menu(id_Structure_sentence_type, _)) = Result :-
            handled(Result) = onStructureSentenceType(id_Structure_sentence_type).
    generatedEventHandler(_Win, e_Menu(id_help_about_the_knowledge_base, _)) = Result :-
            handled(Result) = onHelpAboutTheKnowledgeBase(id_help_about_the_knowledge_base).
    generatedEventHandler(_Win, e_Menu(id_Reports_inexact_rules, _)) = Result :-
            handled(Result) = onReportsInexactRules(id_Reports_inexact_rules).
    generatedEventHandler(_Win, e_Menu(id_Sentences_sttables, _)) = Result :-
            handled(Result) = onSentencesSttables(id_Sentences_sttables).
    generatedEventHandler(_Win, e_Menu(id_file_create_knowledge_stream, _)) = Result :-
            handled(Result) = onFileCreateKnowledgeStream(id_file_create_knowledge_stream).
    generatedEventHandler(_Win, e_Menu(id_Sentences_process_cluster, _)) = Result :-
            handled(Result) = onSentencesProcessCluster(id_Sentences_process_cluster).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_create_cluster, _)) = Result :-
            handled(Result) = onKnowledgeBaseCreateCluster(id_Knowledge_base_create_cluster).
    generatedEventHandler(_Win, e_Menu(id_Sentences_derive_sentences, _)) = Result :-
            handled(Result) = onSentencesDeriveSentences(id_Sentences_derive_sentences).
    generatedEventHandler(_Win, e_Menu(id_Knowledge_base_open_sql_connection, _)) = Result :-
            handled(Result) = onKnowledgeBaseOpenSqlConnection(id_Knowledge_base_open_sql_connection).
    generatedEventHandler(_Win, e_CloseRequest()) = Result :-
        handled(Result) = onCloseRequest().
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
    % end of automatic code
    
menuDisable() :-
    dbBase::kbIsSql,
    !,vpi::menuEnable(thisWin,id_Knowledge_base_public_chains,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_external_chains,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_regenerate_indexes_and_graphs,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_garbage_collection,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_make_copy,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_export_to_sql,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_regenerate_indexes,b_false),
    vpi::menuEnable(thisWin,id_Sentences_multidb_rederivation,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_open_sql_connection,b_false),
    !
;
    vpi::menuEnable(thisWin,id_Knowledge_base_check_database_integrity,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_check_and_repair_database_integrity,b_false),
    %vpi::menuEnable(thisWin,id_Sentences_refresh_exported_sentence_types,b_false),
    %vpi::menuEnable(thisWin,id_Sentences_refresh_imported_sentence_types,b_false),
    %vpi::menuEnable(thisWin,id_file_create_knowledge_stream,b_false),
    vpi::menuEnable(thisWin,id_file_sql_restore,b_false),
    vpi::menuEnable(thisWin,id_file_sql_backup,b_false),
    vpi::menuEnable(thisWin,id_Knowledge_base_create_cluster,b_false),
    %vpi::menuEnable(thisWin,id_Sentences_sttables,b_false),
    vpi::menuEnable(thisWin,id_Sentences_process_cluster,b_false),
    !.

preCheck() :-
    Q=globalVars::get_i("quit_flag"),
    E=globalVars::get_i("error_flag"),
    _=cache::tmCacheIsLoaded(),
    cache::smCacheIsLoaded,
    preCheckX(Q,E,0,0),
    !
;    
    Q=globalVars::get_i("quit_flag"),
    E=globalVars::get_i("error_flag"),
    _=cache::tmCacheIsLoaded(),
    preCheckX(Q,E,0,1),
    !
;    
    Q=globalVars::get_i("quit_flag"),
    E=globalVars::get_i("error_flag"),
    cache::smCacheIsLoaded(),
    preCheckX(Q,E,1,0),
    !
;    
    Q=globalVars::get_i("quit_flag"),
    E=globalVars::get_i("error_flag"),
    preCheckX(Q,E,1,1),
    !.

preCheckX(Q,E,T,S) :-
    S=Q+E+T+S,
    S>0,
    vpi::alarm(vpiDomains::mesbox_iconexclamation),
    fail
;
    T=1,
    dlgIF::msgS("TERMINOLOGY CACHE IS OFF!"),
    fail
;
    S=1,
    dlgIF::msgS("SENTENCE MODEL CACHE IS OFF!"),
    fail
;
    Q=1,
    dlgIF::msgS("QUIT HAS BEEN REQUESTED - NO COMMANDS PROCESSED UNTIL FLAG IS TURNED OFF (USE THE X-BUTTON)!"),
    fail
;
    E=1,
    dlgIF::msgS("ERROR(S) HAVE BEEN ENCOUNTERED - NO COMMANDS PROCESSED UNTIL FLAG IS TURNED OFF (USE THE X-BUTTON)!"),
    fail
;
    Q=0,
    E=0,
    !.
/*
    generatedEventHandler(_Win, e_Menu(id_edit_cut, _)) = Result :-
            handled(Result) = onEditCut(id_edit_cut).
    generatedEventHandler(_Win, e_Menu(id_edit_copy, _)) = Result :-
            handled(Result) = onEditCopy(id_edit_copy).
    generatedEventHandler(_Win, e_Menu(id_edit_paste, _)) = Result :-
            handled(Result) = onEditPaste(id_edit_paste).
    generatedEventHandler(_Win, e_Menu(id_edit_delete, _)) = Result :-
            handled(Result) = onEditDelete(id_edit_delete).
    generatedEventHandler(_Win, e_Menu(id_edit_undo, _)) = Result :-
            handled(Result) = onEditUndo(id_edit_undo).
    generatedEventHandler(_Win, e_Menu(id_edit_redo, _)) = Result :-
            handled(Result) = onEditRedo(id_edit_redo).
    generatedEventHandler(_Win, e_Menu(id_edit_find, _)) = Result :-
            handled(Result) = onEditFind(id_edit_find).
    generatedEventHandler(_Win, e_Menu(id_edit_find_next, _)) = Result :-
            handled(Result) = onEditFindNext(id_edit_find_next).
    generatedEventHandler(_Win, e_Menu(id_edit_replace, _)) = Result :-
            handled(Result) = onEditReplace(id_edit_replace).
            */
end implement taskWindow
