/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement export
open core, vpiDomains, resourceIdentifiers

constants
className = "com/rudstrand/TaskWindow/export/export".
classVersion = "1.0".

facts
thisWin : vpiDomains::windowHandle := erroneous.
sqlName : string := "".

predicates
checkGraphs : (integer) determ.

clauses
classInfo(className, classVersion).

show(Parent,Name):-
    sqlName:=Name,
    _ = vpi::winCreateDynDialog(Parent, controlList, eventHandler, 0).

predicates
    eventHandler : vpiDomains::ehandler.
clauses
    eventHandler(Win, Event) = Result :-
        Result = generatedEventHandler(Win, Event).

predicates
    onCreate : vpiDomains::createHandler.
clauses
    onCreate(_CreationData) = defaultHandling():-
    guix::txtSet(thisWin,idc_sql,sqlName),
    guix::txtSet(thisWin,idc_kb,sqlName),
	guix::guiCB_Set(thisWin,idc_create,1),
	guix::guiCB_Set(thisWin,idc_add,1),
    !;!.

predicates
    onDestroy : vpiDomains::destroyHandler.
clauses
    onDestroy() = defaultHandling() :-
        thisWin := erroneous.

predicates
    onControlOK : vpiDomains::controlHandler.
clauses
    onControlOK(_Ctrl, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        guix::txtGet(thisWin,idc_kb,Name),
       	L=string::length(Name),
       	L<1,
       	dlgIF::dlgNote(0,"KB name cannot be blank!"),
       	!
    ;
        guix::txtGet(thisWin,idc_kb,Name),
        not(sqlTools::validSqlName(Name)),
       	dlgIF::dlgNote(0,"Invalid KB name format!"),
       	!
    ;
        guix::txtGet(thisWin,idc_sql,Name),
       	L=string::length(Name),
       	L<1,
       	dlgIF::dlgNote(0,"SQL DB name cannot be blank!"),
       	!
    ;
        guix::txtGet(thisWin,idc_sql,Name),
        not(sqlTools::validSqlName(Name)),
       	dlgIF::dlgNote(0,"Invalid SQL DB name format!"),
       	!
    ;
       	guix::guiCB_Get(thisWin,idc_add,KB_Create), 
        guix::txtGet(thisWin,idc_sql,Sql),
       	guix::guiCB_Get(thisWin,idc_create,DB_Create),
        guix::txtGet(thisWin,idc_kb,Name),
       	guix::guiCB_Get(thisWin,idc_include,IncludeDerived),
       	expsql::export(Sql,DB_Create,Name,KB_Create,IncludeDerived),
       	checkGraphs(KB_Create),
        vpi::winDestroy(thisWin),
        !;!.

predicates
    onControlidc_help : vpiDomains::controlHandler.
clauses
    onControlidc_help(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
        dlgIF::hlpWeb("w_export"),
        !.

% This code is maintained by the VDE do not update it manually, 08:52:05-22.11.2005
constants
    dialogType : vpiDomains::wintype = wd_modal.
    title : string = "Export to SQL".
    rectangle : vpiDomains::rct = rct(50,40,389,134).
    dialogFlags : vpiDomains::wsflags = [wsf_Close,wsf_TitleBar].
    dialogFont = "Arial".
    dialogFontSize = 8.

constants
    controlList : vpiDomains::windef_list =
        [
        dlgFont(wdef(dialogType, rectangle, title, u_DlgBase),
                dialogFont, dialogFontSize, dialogFlags),
        ctl(wdef(wc_PushButton,rct(88,76,144,90),"&OK",u_DlgBase),idc_ok,[wsf_Border,wsf_Default,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(180,76,236,90),"&Help",u_DlgBase),idc_help,[wsf_Border,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_CheckBox,rct(4,56,136,66),"Include derived knowledge",u_DlgBase),idc_include,[wsf_Auto,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_Edit,rct(228,28,336,40),"",u_DlgBase),idc_kb,[wsf_AlignCenter,wsf_Group,wsf_TabStop,wsf_AutoHScroll]),
        ctl(wdef(wc_CheckBox,rct(4,44,136,54),"(Re-)create the SQL knowledge base",u_DlgBase),idc_add,[wsf_Auto,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_Edit,rct(228,2,336,14),"",u_DlgBase),idc_sql,[wsf_AlignCenter,wsf_Group,wsf_TabStop,wsf_AutoHScroll]),
        ctl(wdef(wc_CheckBox,rct(4,14,136,24),"(Re-)create the SQL database",u_DlgBase),idc_create,[wsf_Auto,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_Text,rct(4,30,224,40),"SQL knowledge base",u_DlgBase),idct_sql_knowledge_base_name,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(4,2,224,12),"SQL database where the SQL knowledge base will be stored",u_DlgBase),idct_sql_database_where_the_sql_knowledge_base_will_be_stored,[wsf_AlignLeft])
        ].

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Control(idc_ok, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlOK(idc_ok, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_help, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_help(idc_help, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
% end of automatic code

checkGraphs(1) :-
    dlgIF::dlgNote(0,"Export completed!"),
    !.
checkGraphs(-1) :-
    dlgIF::dlgNote(0,"Export completed! Note that the knowledge dependency graphs (in the generated SQL KB), need to be regenerated when two or more 3KB-databases are exported into the same SQL knowledge base!"),
    !.
end implement export
