/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement externalchains
open core,vpiDomains,resourceIdentifiers,termDomain

constants
className = "com/rudstrand/TaskWindow/externalchains".
classVersion = "1.0".

class facts
newExtDB : (string,string) nondeterm.

class predicates
dbuExtChainsIns : (chainDB,string_list) determ.
remChains : (string_list,string_list,string_list) determ (i,i,o).
wwAddChainsE : (string,string_list,vpiDomains::windowHandle,vpiDomains::windowHandle) determ.
wwSetDB : (string,vpiDomains::windowHandle) determ.
toggle : (string_list,string_list) determ (i,o).

clauses
classInfo(className, classVersion).
    facts
        thisWin : vpiDomains::windowHandle := erroneous.
    clauses
        new(Parent):-
            _ = vpi::winCreate(windowType, rectangle, title, menu, Parent,
                               windowFlags, eventHandler, 0).
    predicates
        eventHandler : vpiDomains::ehandler.
    clauses
        eventHandler(Win, Event) = Result :-
            Result = generatedEventHandler(Win, Event),
            !.
        eventHandler(Win, e_Menu(Id, CAS)) = Result :-
            Parent = vpi::winGetParent(Win),
            Result = vpi::winSendEvent(Parent, e_Menu(Id, CAS)).
/* select external DB */
    predicates
        onControlSelectExternalDatabase : vpiDomains::controlHandler.
    clauses
        onControlSelectExternalDatabase(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxPop(thisWin,idc_chains,[]),
	guix::txtSet(thisWin,idc_id,""),
	guix::txtSet(thisWin,idc_external_db,""),
  	P=globalVars::get_s("std_path"),
	File=vpiCommonDialogs::getFileName("*.3kb",["KB files","*.3kb"],"External KB",[],P,_),
	File<>"",
	fioExt::filValidKB_Name(1,File,_),	
	wwSetDB(File,thisWin),
	!;!.
/* add external chains */
    predicates
        onControlselect : vpiDomains::controlHandler.
    clauses
        onControlselect(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::txtGet(thisWin,idc_id,DbId),
	DbId="",
	dlgIF::dlgNote(1,"A unique one-letter ID must be specified for the external database"),
	!
;	
	guix::txtGet(thisWin,idc_id,Id),
	Id2=globalVars::get_s("kb_id"),
	Id=Id2,
	dlgIF::dlgNote(1,"The specified ID is already in use for the main database"),
	!
;	
	guix::txtGet(thisWin,idc_id,Id),
	guix::txtGet(thisWin,idc_external_db,File),
	dbaBase::database(_,F,Id,_),
	F<>File,
	dlgIF::dlgNote(1,"The specified ID is already in use for another external database"),
	!
;	
	guix::txtGet(thisWin,idc_id,Id),
	dbaBase::database(_,_,Id,_),
	% an existing open external DB
	guix::lbxGetMulti(thisWin,idc_chains,Chains),
	Cc=vpi::winGetCtlHandle(thisWin,idc_chains),
	Cs=vpi::winGetCtlHandle(thisWin,idc_schains),
	wwAddChainsE(Id,Chains,Cc,Cs),
	guix::lbxGetAll(thisWin,idc_chains,All),
	set::setRemSub(Chains,All,Rest),
	guix::lbxPop(thisWin,idc_chains,Rest),
	!
;	
	guix::txtGet(thisWin,idc_id,Id),
	not(dbaBase::database(_,_,Id,_)),
	newExtDB(_,Id),
	% a previously specified new external DB
	Cc=vpi::winGetCtlHandle(thisWin,idc_chains),
	guix::lbxGetMulti(thisWin,idc_chains,Chains),
	Cs=vpi::winGetCtlHandle(thisWin,idc_schains),
	wwAddChainsE(Id,Chains,Cc,Cs),
	guix::lbxGetAll(thisWin,idc_chains,All),
	set::setRemSub(Chains,All,Rest),
	guix::lbxPop(thisWin,idc_chains,Rest),
	!
;
	guix::txtGet(thisWin,idc_id,Id),
	not(dbaBase::database(_,_,Id,_)),
	not(newExtDB(_,Id)),
	% a new external DB
	guix::txtGet(thisWin,idc_external_db,File),
	assert(newExtDB(File,Id)),
	Cc=vpi::winGetCtlHandle(thisWin,idc_chains),
	guix::lbxGetMulti(thisWin,idc_chains,Chains),
	Cs=vpi::winGetCtlHandle(thisWin,idc_schains),
	wwAddChainsE(Id,Chains,Cc,Cs),
	guix::lbxGetAll(thisWin,idc_chains,All),
	set::setRemSub(Chains,All,Rest),
	guix::lbxPop(thisWin,idc_chains,Rest),
	!;!.
/* remove external chains */
    predicates
        onControlremove : vpiDomains::controlHandler.
    clauses
        onControlremove(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetMulti(thisWin,idc_schains,Ls),
	guix::lbxGetAll(thisWin,idc_schains,La),
	set::setRemSub(Ls,La,Lm),
	guix::lbxPop(thisWin,idc_schains,Lm),
	!;!.

/* read/write toggle */

    predicates
        onControltoggle : vpiDomains::controlHandler.
    clauses
        onControltoggle(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetMulti(thisWin,idc_schains,Ls),
	toggle(Ls,Ls2),
	guix::lbxGetAll(thisWin,idc_schains,La),
	set::setRemSub(Ls,La,Lm),
	set::conc([Lm,Ls2],Lm2),
	guix::lbxPop(thisWin,idc_schains,Lm2),
	!;!.

/* update external chains */

    predicates
        onControlok : vpiDomains::controlHandler.
    clauses
        onControlok(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	dbBase::chainDelete("p3_pIxdb"),
	guix::lbxGetAll(thisWin,idc_schains,Chains),
    dbaBase::database(DB,_,_,0),
	dbuExtChainsIns(DB,Chains),
	vpiCommonDialogs::ask("Please confirm!","Changes to external chains should be followed by regenerating indexes and graphs!",["Regenerate","Skip"])=vpiCommonDialogs::resp_default,
    vpi::cursorSetWait(),
	dbBase::indexRegen,
	globalVars::get_i("quit_flag")=0,
	dep::depSMM_Regen,
	globalVars::get_i("quit_flag")=0,
	dep::regenDependencyGraph,
	globalVars::get_i("quit_flag")=0,
    der::genDerivationGraph,	
    vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	vpi::winDestroy(thisWin),
	!
;
	vpi::winDestroy(thisWin),
    !.

/* create, destroy etc */

    predicates
        onCreate : vpiDomains::createHandler.
    clauses
        onCreate(_CreationData) = defaultHandling():-
   	guix::setStyle(thisWin,[idc_chains,idc_schains,idc_select,idc_ok,idc_toggle,idct_external_chains,
  	                   idct_public_chains,idc_id,idc_select_external_database,idc_external_db,idc_remove,idct_db_id],[]),
    guix::addWin(31,thisWin),
	guix::ctlDisable(thisWin,[idc_id]),
	L=pop::getChainsExtAccessNL(),
	guix::lbxPop(thisWin,idc_schains,L),
	guix::addWin(113,thisWin),
	!;!.
    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            guix::dropWin(thisWin),
            thisWin := erroneous.
    predicates
        onSizeChanged : vpiDomains::sizeHandler.
    clauses
        onSizeChanged(_Width, _Height) = defaultHandling():-
            vpiToolbar::resize(thisWin).
    predicates
        onCloseRequest : vpiDomains::closeRequestHandler.
    clauses
        onCloseRequest() = defaultHandling().
    % This code is maintained by the VDE do not update it manually, 20:05:09-23.1.2005
    constants
        windowType : vpiDomains::wintype = w_TopLevel.
        windowFlags : vpiDomains::wsflags = [wsf_TitleBar,wsf_Minimize,wsf_Close,wsf_ClipSiblings,wsf_ClipChildren].
        rectangle : vpiDomains::rct = rct(100,80,714,569).
        menu : vpiDomains::menu = noMenu.
        title : string = "External Chains".
    predicates
        generatedEventHandler : vpiDomains::ehandler.
    clauses
        generatedEventHandler(Win, e_create(_)) = _ :-
            thisWin := Win,
            _ = vpi::winCreateControl(wc_PushButton,rct(8,4,176,28),"Select external database",Win,[wsf_Group,wsf_TabStop],idc_select_external_database),
            _ = vpi::winCreateControl(wc_Edit,rct(264,4,609,28),"",Win,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll,wsf_Disabled],idc_external_db),
            _ = vpi::winCreateControl(wc_LBox,rct(9,68,300,422),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight,wsf_MultiSelect],idc_chains),
            _ = vpi::winCreateControl(wc_LBox,rct(314,68,605,422),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight,wsf_MultiSelect],idc_schains),
            _ = vpi::winCreateControl(wc_Text,rct(9,47,301,65),"Public chains",Win,[wsf_AlignCenter],idct_public_chains),
            _ = vpi::winCreateControl(wc_Text,rct(312,44,603,62),"External chains",Win,[wsf_AlignCenter],idct_external_chains),
            _ = vpi::winCreateControl(wc_PushButton,rct(248,440,375,475),"Update",Win,[wsf_Group,wsf_TabStop],idc_ok),
            _ = vpi::winCreateControl(wc_PushButton,rct(8,428,211,452),"Add selected external chain(s)",Win,[wsf_Group,wsf_TabStop],idc_select),
            _ = vpi::winCreateControl(wc_Edit,rct(227,4,258,28),"",Win,[wsf_UpperCase,wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll],idc_id),
            _ = vpi::winCreateControl(wc_PushButton,rct(389,428,605,452),"Remove selected external chains",Win,[wsf_Group,wsf_TabStop],idc_remove),
            _ = vpi::winCreateControl(wc_Text,rct(183,6,223,26),"DB ID",Win,[wsf_AlignLeft],idct_db_id),
            _ = vpi::winCreateControl(wc_PushButton,rct(389,460,605,484),"READ/WRITE Toggle",Win,[wsf_Group,wsf_TabStop],idc_toggle),
            fail.
        generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
            handled(Result) = onCreate(CreationData).
        generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
            handled(Result) = onSizeChanged(Width, Height).
        generatedEventHandler(_Win, e_Control(idc_select_external_database, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlSelectExternalDatabase(idc_select_external_database, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_select, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlselect(idc_select, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_remove, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlremove(idc_remove, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_ok, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlok(idc_ok, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_toggle, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControltoggle(idc_toggle, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Destroy()) = Result :-
            handled(Result) = onDestroy().
    % end of automatic code
% --------------------------------------------------------------------------
% wwSetDB
% --------------------------------------------------------------------------
wwSetDB(File,W) :-
	dbaBase::database(_,F,_,0),
	dbaBase::sameDB(File,F),
	dlgIF::dlgNote(1,"The selected database is the main database!"),
	guix::txtSet(W,idc_external_db,""),
	guix::txtSet(W,idc_id,""),
	guix::ctlDisable(W,[idc_id]),
	!
;	
	dbaBase::database(EDB,F,Id,No),
	No>0,
	dbaBase::sameDB(File,F),
	!, % existing external DB
	Pubs=pop::getChainsPubDB_AccessNL(EDB),
	Chains=pop::getChains(),
	remChains(Chains,Pubs,PubsNet),
	guix::txtSet(W,idc_external_db,File),
	guix::txtSet(W,idc_id,Id),
	guix::ctlDisable(W,[idc_id]),
	guix::lbxPop(W,idc_chains,PubsNet),
	!
;
    % new external DB
	guix::ctlEnable(W,[idc_id]),
	guix::txtSet(W,idc_external_db,File),
  	dbaUtil::dbaOpenAuxDB(File,EDB),
	Pubs=pop::getChainsPubDB_AccessNL(EDB),
	dbaUtil::dbaCloseAux(EDB),
	Chains=pop::getChains(),
	remChains(Chains,Pubs,PubsNet),
	guix::lbxPop(W,idc_chains,PubsNet),
	!.
% --------------------------------------------------------------------------
% wwAddChainsE
% --------------------------------------------------------------------------
wwAddChainsE(_,[],_,_) :-
	!.
wwAddChainsE(Id,[Chain|Chains],Cp,Ce) :-
	sx::concat([Id," ",Chain],S),
	vpi::lboxAdd(Ce,S),
	wwAddChainsE(Id,Chains,Cp,Ce),
	!.
% --------------------------------------------------------------------------
% remChains
% --------------------------------------------------------------------------
remChains([],Pubs,Pubs) :-
    !.
remChains([C|Chains],Pubs,PubsNet) :-
    dbBase::nlSing(C,Cn),
    Cx=string::concat("READ:  ",Cn),
    Cy=string::concat("WRITE: ",Cn),
    set::setRemSub([Cx,Cy],Pubs,Pubs2),
    remChains(Chains,Pubs2,PubsNet),
    !.
% ----------------------------------------------------------------------------------------------
% dbuExtChainsIns
% ----------------------------------------------------------------------------------------------
dbuExtChainsIns(_,[]) :-
	!.
dbuExtChainsIns(KB,[Chain|Chains]) :-
	sx::frontstr(1,Chain,DbId,_),
	dbaBase::database(EDB,_,DbId,_), % existing ext. db
	sx::frontstr(2,Chain,_,S2),
	sx::frontstr(7,S2,Acc,Ch),
	dbBase::kbID(Ch,ChK,_,_),
	set::setGet(["READ:  ","WRITE: "],N,Acc),
	set::setGet(["p3_iRo","p3_iRw"],N,AccK),
	dbaBase::dbaInsertRecord(2,"$SYS",0,0,cmp("p3_pIxdb",[cmp("p3_eChan",[atom(ChK),atom(AccK)]),termDomain::str(DbId)]),atom("true"),_),
	dbaBase::dropExtChain(ChK),
	Access=N-1,
	dbaBase::addExtChain(Chk,EDB,Access),
	dbuExtChainsIns(KB,Chains),
	!
;	
	sx::frontstr(1,Chain,DbId,_),
	not(dbaBase::database(_,_,DbId,_)),
	newExtDB(File,DbId),
	sx::frontstr(2,Chain,_,S2),
	sx::frontstr(7,S2,Acc,Ch),
	dbBase::kbID(Ch,ChK,_,_),
	set::setGet(["READ:  ","WRITE: "],N,Acc),
	set::setGet(["p3_iRo","p3_iRw"],N,AccK),
	dbaBase::dbaInsertRecord(2,"$SYS",0,0,cmp("p3_pIxdb",[cmp("p3_eChan",[atom(ChK),atom(AccK)]),termDomain::str(DbId)]),atom("true"),_),
	dbaBase::dropExtChain(ChK),
    dbaExt::openExternalDB(File,DbId,EDB),
	Access=N-1,
	dbaBase::addExtChain(Chk,EDB,Access),
	dbuExtChainsIns(KB,Chains),
	!.
% ----------------------------------------------------------------------------------------------
% toggle
% ----------------------------------------------------------------------------------------------
toggle([],[]) :-
    !.
toggle([H|T],[Hm|Tm]) :-
    sx::frontstr(2,H,Id,H2),
    sx::frontstr(7,H2,Acc,Rest),
    Acc="WRITE: ", % change to read
    Hm=string::concat(Id,"READ:  ",Rest),
    toggle(T,Tm),
    !
;    
    % change to write (if permissible)
    sx::frontstr(1,H,Id,_),
    dbaBase::database(DB,_,Id,_),
    L=pop::getChainsPubDB_AccessNL(DB),
    sx::frontstr(2,H,Rest,_),
    set::member(Rest,L),
    % not permissible
    Hm=H,
    toggle(T,Tm),
    !
;
    % change to write
    sx::frontstr(2,H,Id,H2),
    sx::frontstr(7,H2,_,Rest),
    Hm=string::concat(Id,"WRITE: ",Rest),
    toggle(T,Tm),
    !.
end implement externalchains
