/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement knowledgestream
    open core, vpiDomains, resourceIdentifiers

constants
    className = "com/rudstrand/TaskWindow/knowledgestream".
    classVersion = "1.0".

clauses
    classInfo(className, classVersion).

facts
    thisWin : vpiDomains::windowHandle := erroneous.

clauses
    show(Parent):-
        _ = vpi::winCreateDynDialog(Parent, controlList, eventHandler, 0).

predicates
eventHandler : vpiDomains::ehandler.
setupTable : (string,integer,string) procedure (i,i,o).

clauses
    eventHandler(Win, Event) = Result :-
        Result = generatedEventHandler(Win, Event).

predicates
    onCreate : vpiDomains::createHandler.
clauses
    onCreate(_CreationData) = defaultHandling().

predicates
    onDestroy : vpiDomains::destroyHandler.
clauses
    onDestroy() = defaultHandling() :-
        thisWin := erroneous.

predicates
    onControlOK : vpiDomains::controlHandler.
clauses
    onControlOK(_Ctrl, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        guix::txtGet(thisWin,idc_file,File),
        File="",
        dlgIF::dlgNote(0,"Enter a knowledge stream source file!"),
        !
;        
        guix::txtGet(thisWin,idc_file,File),
        not(file::existExactFile(File)),
        dlgIF::dlgNote(0,"File doesn't exist!"),
        !
;    
        guix::txtGet(thisWin,idc_file,File),
        Ext=fileName::getExtension(File),
        ExtL=string::toLowercase(Ext),
        not(set::member(ExtL,["3nl","3pr","xml"])),
        dlgIF::dlgNote(0,"Invalid file extension!"),
        !
;        
        guix::txtGet(thisWin,idc_table,File),
        File="",
        dlgIF::dlgNote(0,"Enter a SQL table name!"),
        !
;          
        guix::txtGet(thisWin,idc_file,File),
        Ext=fileName::getExtension(File),
        ExtL=string::toLowercase(Ext),
        guix::guiRB_Get(thisWin,[idc_s,idc_r,idc_i,idc_f],[2,3,6,4],Type),
        guix::guiCB_Get(thisWin,idc_recreate,Recreate),
        guix::txtGet(thisWin,idc_table,Table),
        sqlBase::openConnection(),
        setupTable(Table,Recreate,Qtable),
        proc::processFile(Type,ExtL,File,Qtable),
        dlgIF::dlgNote(0,"Knowledge written to stream!"),
        !;!.
        
predicates
    onControlidc_help : vpiDomains::controlHandler.
clauses
    onControlidc_help(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	dlgIF::hlpWeb("KnowledgeStreams"),
	!.

predicates
    onCloseRequest : vpiDomains::closeRequestHandler.
clauses
    onCloseRequest() = defaultHandling() :-
        vpi::winDestroy(thisWin).

predicates
    onControlidc_select_source_file : vpiDomains::controlHandler.
clauses
    onControlidc_select_source_file(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
  	        editorWindow::getEditorFileDefaults(Mask,_,Path),
  	        set::member(Mask,["*.xml","*.3nl","*.3pr","*.3dl"]),
  	        List=["XML-files","*.xml","NL-files","*.3nl","PR-files","*.3pr","Delimited files","*.3dl"],
	        _=vpiCommonDialogs::getFileName(Mask,List,"Open Files",[],Path,Filenames),
	        FileNames=[File],
	        guix::txtSet(thisWin,idc_file,File),
	        !
;	        
  	        editorWindow::getEditorFileDefaults(Mask,_,Path),
  	        not(set::member(Mask,["*.xml","*.3nl","*.3pr","*.3dl"])),
  	        List=["XML-files","*.xml","NL-files","*.3nl","PR-files","*.3pr","Delimited files","*.3dl"],
	        _=vpiCommonDialogs::getFileName("*.3nl",List,"Open Files",[],Path,Filenames),
	        FileNames=[File],
	        guix::txtSet(thisWin,idc_file,File),
	        !;!.

% This code is maintained by the VDE do not update it manually, 12:10:08-28.5.2006
constants
    dialogType : vpiDomains::wintype = wd_modal.
    title : string = "Create/write to Knowledge Stream".
    rectangle : vpiDomains::rct = rct(50,40,366,121).
    dialogFlags : vpiDomains::wsflags = [wsf_Close,wsf_TitleBar].
    dialogFont = "Arial".
    dialogFontSize = 8.

constants
    controlList : vpiDomains::windef_list =
        [
        dlgFont(wdef(dialogType, rectangle, title, u_DlgBase),
                dialogFont, dialogFontSize, dialogFlags),
        ctl(wdef(wc_PushButton,rct(80,66,164,78),"&OK",u_DlgBase),idc_ok,[wsf_Border,wsf_Default,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(176,66,260,78),"&Help",u_DlgBase),idc_help,[wsf_Border,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_CheckBox,rct(4,34,76,44),"(Re-)create table",u_DlgBase),idc_recreate,[wsf_Auto,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_RadioButton,rct(4,48,76,58),"Sentences",u_DlgBase),idc_s,[wsf_Auto,wsf_TabStop,wsf_Group]),
        ctl(wdef(wc_RadioButton,rct(80,48,152,58),"Sentence rules",u_DlgBase),idc_r,[wsf_Auto,wsf_TabStop]),
        ctl(wdef(wc_RadioButton,rct(156,48,232,58),"Inexact rules",u_DlgBase),idc_i,[wsf_Auto,wsf_TabStop]),
        ctl(wdef(wc_RadioButton,rct(236,48,312,58),"Function definitions",u_DlgBase),idc_f,[wsf_Auto,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(4,2,76,14),"Select source file",u_DlgBase),idc_select_source_file,[wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_Edit,rct(80,2,312,14),"",u_DlgBase),idc_file,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll]),
        ctl(wdef(wc_Edit,rct(80,18,312,30),"",u_DlgBase),idc_table,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll]),
        ctl(wdef(wc_Text,rct(4,20,76,30),"Table",u_DlgBase),idct_table,[wsf_AlignLeft])
        ].

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Control(idc_ok, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlOK(idc_ok, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_help, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_help(idc_help, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_select_source_file, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_select_source_file(idc_select_source_file, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_CloseRequest()) = Result :-
        handled(Result) = onCloseRequest().
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
% end of automatic code


setupTable(Table,Recreate,Qtable) :-
    not(_=string::searchchar(Table,'.',string::caseSensitive)),
    Db=sqlBase::getDb(),
    Qtable2=string::concat(Db,".",Table),
    setupTable(Qtable2,Recreate,Qtable),
    !
;
    _=string::searchchar(Table,'.',string::caseSensitive),
    L=sx::unpack(Table,"."),
    L=[D,_],
    Sql1=string::concat("create database if not exists ",D),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,Sql1),
    Recreate=1,
    Sql2=string::concat("drop table if exists ",Table),
    sqlTools::executeDirect(Con,Sql2),
    Sql3=string::concat("create table if not exists ",Table," (k_format tinyint unsigned not null,k_type tinyint unsigned not null,knowledge text not null,created datetime not null,processed tinyint unsigned not null,kb varchar(128) not null,source varchar(64) not null)"),
    sqlTools::executeDirect(Con,Sql3),
    Qtable=Table,
    !
;
    Qtable=Table,
    !.
    
end implement knowledgestream
