/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement multikb
open core,vpiDomains,resourceIdentifiers,termDomain

constants
className = "com/rudstrand/TaskWindow/multikb".
classVersion = "1.0".

class predicates
getDBs : (sterml,string_list) determ (i,o) (o,i).

clauses
classInfo(className, classVersion).

    facts
        thisWin : vpiDomains::windowHandle := erroneous.

    clauses
        new(Parent):-
            _ = vpi::winCreate(windowType, rectangle, title, menu, Parent,
                               windowFlags, eventHandler, 0).

    predicates
        eventHandler : vpiDomains::ehandler.
    clauses
        eventHandler(Win, Event) = Result :-
            Result = generatedEventHandler(Win, Event),
            !.
        eventHandler(Win, e_Menu(Id, CAS)) = Result :-
            Parent = vpi::winGetParent(Win),
            Result = vpi::winSendEvent(Parent, e_Menu(Id, CAS)).

/* load/save script */

    predicates
        onControlload : vpiDomains::controlHandler.
    clauses
        onControlload(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	F=vpiCommonDialogs::getFileName("*.xml",["Re-derivation script files","*.xml"],"Select re-derivation script file",[],"",_),
	F<>"",
	file::existFile(F),
	X=file::readString(F),
	xmlGeneric::remHeader(X,X2),
	XML=xmlTree::newTree(X2),
	Tree=XML:getSubTree("rederivation"),
    Tree=cmp("rederivation",[nill|DBs]),
    getDBs(DBs,L),
	guix::lbxPop(thisWin,idc_kbs,L),
	!;!.

    predicates
        onControlsave : vpiDomains::controlHandler.
    clauses
        onControlsave(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetAll(thisWin,idc_kbs,L),
	set::size(L,N),
	N>1,
	getDBs(DBs,L),
	Tree=cmp("rederivation",[nill|DBs]),
	xmlP3::scriptWrap("PR",[Tree],Xt),
	XML=xmlTree::newTreeXML(Xt),
	X=XML:getX(5),
	xmlGeneric::xmlHeader(X,Xh),
	F=vpiCommonDialogs::getFileName("*.xml",["Re-derivation script files","*.xml"],"Select re-derivation script file",[],"",_),
	F<>"",
	file::writeString(F,Xh),
	dlgIF::dlgNote(0,"Re-derivation script has been saved"),
	!
;
	dlgIF::dlgNote(1,"The re-derivation script should include min. two databases"),
	!;!.	

/* add, remove */

    predicates
        onControladd : vpiDomains::controlHandler.
    clauses
        onControladd(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	_=vpiCommonDialogs::getFileName("*.3kb",["DB files","*.3kb"],"Select databases to bed added",[dlgfn_MultiSel],"",L),
	guix::lbxAddList(thisWin,idc_kbs,L),
	!;!.

    predicates
        onControlremove : vpiDomains::controlHandler.
    clauses
        onControlremove(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
  	guix::lbxRemReselect(thisWin,idc_kbs),
	!;!.

/* move up; down */

predicates
    onControlidc_down : vpiDomains::controlHandler.
clauses
    onControlidc_down(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	C=vpi::winGetCtlHandle(thisWin,idc_kbs),
	vpi::lboxGetSel(C,DEs,Nos),
	DEs=[DE],
	Nos=[No],
	DEa=vpi::lboxGetAll(C),
	set::size(DEa,N),
	No<(N-1),
	vpi::lboxDelete(C,No),
	M=No+1,
	vpi::lboxAdd(C,M,DE),
	vpi::lboxSetSel(C,M,b_true),
	!;!.

predicates
    onControlidc_up : vpiDomains::controlHandler.
clauses
    onControlidc_up(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	C=vpi::winGetCtlHandle(thisWin,idc_kbs),
	vpi::lboxGetSel(C,DEs,Nos),
	DEs=[DE],
	Nos=[No],
	No>0,
	vpi::lboxDelete(C,No),
	M=No-1,
	vpi::lboxAdd(C,M,DE),
	vpi::lboxSetSel(C,M,b_true),
	!;!.

/* delete all derived */

    predicates
        onControlDeleteAll : vpiDomains::controlHandler.
    clauses
        onControlDeleteAll(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetAll(thisWin,idc_kbs,L),
	set::size(L,N),
	N<2,
	dlgIF::dlgNote(1,"The database list to be rederived must include at least two databases!"),
	!
;        
	guix::lbxGetAll(thisWin,idc_kbs,L),
	set::size(L,N),
	N>1,
	vpi::cursorSetWait(),
	dbaExt::dbaMultiDeleteDerived(1,L),
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

/* regenerate graphs */

    predicates
        onControlRegenOnly : vpiDomains::controlHandler.
    clauses
        onControlRegenOnly(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetAll(thisWin,idc_kbs,L),
	set::size(L,N),
	N<2,
	dlgIF::dlgNote(1,"The database list to be rederived must include at least two databases!"),
	!
;        
	guix::lbxGetAll(thisWin,idc_kbs,L),
	vpi::cursorSetWait(),
	dbaExt::dbaMultiDeleteDerived(2,L),
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

/* delete derived + regenerate graphs */

    predicates
        onControlregen : vpiDomains::controlHandler.
    clauses
        onControlregen(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetAll(thisWin,idc_kbs,L),
	set::size(L,N),
	N<2,
	dlgIF::dlgNote(1,"The database list to be rederived must include at least two databases!"),
	!
;        
	guix::lbxGetAll(thisWin,idc_kbs,L),
	vpi::cursorSetWait(),
	dbaExt::dbaMultiDeleteDerived(3,L),
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

/* run derivation cycles */

    predicates
        onControlrun : vpiDomains::controlHandler.
    clauses
        onControlrun(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetAll(thisWin,idc_kbs,L),
	set::size(L,N),
	N<2,
	dlgIF::dlgNote(1,"The database list to be rederived must include at least two databases!"),
	!
;        
	guix::lbxGetAll(thisWin,idc_kbs,L),
	vpi::cursorSetWait(),
	guix::txtGetI(thisWin,idc_cycles,N),
	N>0,
	dbaExt::dbaMultiRederive(L,N),
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

/* delete derived + regenerate graphs + run cycles */

    predicates
        onControlRegenRun : vpiDomains::controlHandler.
    clauses
        onControlRegenRun(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGetAll(thisWin,idc_kbs,L),
	set::size(L,N),
	N<2,
	dlgIF::dlgNote(1,"The database list to be rederived must include at least two databases!"),
	!
;        
	guix::lbxGetAll(thisWin,idc_kbs,L),
	vpi::cursorSetWait(),
	guix::txtGetI(thisWin,idc_cycles,N),
	N>0,
	dbaExt::dbaMultiDeleteDerived(3,L),
	dbaExt::dbaMultiRederive(L,N),
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

    predicates
        onControlclose : vpiDomains::controlHandler.
    clauses
        onControlclose(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
            guix::dropWin(thisWin),
            vpi::winDestroy(thisWin).

/* create, destroy etc */

    predicates
        onCreate : vpiDomains::createHandler.
    clauses
        onCreate(_CreationData) = defaultHandling():-
   	guix::setStyle(thisWin,[idc_kbs,idc_load,idc_save,idc_remove,idc_add,idc_run,idc_close,idc_delete_all,idct_kbs,idct_cycles,idc_cycles,idc_regen,idc_regen_only,
   	                                idc_regen_run,idc_up,idc_down],[]),
  	guix::txtSet(thisWin,idc_cycles,"1"),
	guix::addWin(112,thisWin),
	!;!.

    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            guix::dropWin(thisWin),
            thisWin := erroneous.

    predicates
        onSizeChanged : vpiDomains::sizeHandler.
    clauses
        onSizeChanged(_Width, _Height) = defaultHandling():-
            vpiToolbar::resize(thisWin).

    predicates
        onCloseRequest : vpiDomains::closeRequestHandler.
    clauses
        onCloseRequest() = defaultHandling():-
            guix::dropWin(thisWin),
            vpi::winDestroy(thisWin).

% This code is maintained by the VDE do not update it manually, 13:18:33-3.6.2005
constants
    windowType : vpiDomains::wintype = w_TopLevel.
    windowFlags : vpiDomains::wsflags = [wsf_TitleBar,wsf_Minimize,wsf_Close,wsf_ClipSiblings,wsf_ClipChildren].
    rectangle : vpiDomains::rct = rct(10,50,486,543).
    menu : vpiDomains::menu = noMenu.
    title : string = "Multi-DB Sentence Derivation".

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        _ = vpi::winCreateControl(wc_LBox,rct(8,60,384,304),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_NoIntegralHeight],idc_kbs),
        _ = vpi::winCreateControl(wc_Text,rct(8,36,471,56),"Databases",Win,[wsf_AlignCenter],idct_kbs),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,4,224,28),"Load re-derivation script",Win,[wsf_Group,wsf_TabStop],idc_load),
        _ = vpi::winCreateControl(wc_PushButton,rct(248,4,472,28),"Save as re-derivation script",Win,[wsf_Group,wsf_TabStop],idc_save),
        _ = vpi::winCreateControl(wc_PushButton,rct(392,64,472,88),"Add",Win,[wsf_Group,wsf_TabStop],idc_add),
        _ = vpi::winCreateControl(wc_PushButton,rct(392,96,472,120),"Remove",Win,[wsf_Group,wsf_TabStop],idc_remove),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,316,471,340),"Step 1 - Delete all databases' derived sentences",Win,[wsf_Group,wsf_TabStop],idc_delete_all),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,372,330,396),"Step 3 - Run re-derivation cycles",Win,[wsf_Group,wsf_TabStop],idc_run),
        _ = vpi::winCreateControl(wc_PushButton,rct(168,464,320,488),"Close",Win,[wsf_Group,wsf_TabStop],idc_close),
        _ = vpi::winCreateControl(wc_Text,rct(336,372,424,396),"No. of cycles",Win,[wsf_AlignLeft],idct_cycles),
        _ = vpi::winCreateControl(wc_Edit,rct(432,372,464,396),"",Win,[wsf_Group,wsf_TabStop,wsf_AutoHScroll,wsf_AlignLeft],idc_cycles),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,400,471,424),"Step 1 followed by 2",Win,[wsf_Group,wsf_TabStop],idc_regen),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,344,471,368),"Step 2 - Regenerate all databases' dependency graphs",Win,[wsf_Group,wsf_TabStop],idc_regen_only),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,428,471,452),"Step 1 followed by 2 and 3",Win,[wsf_Group,wsf_TabStop],idc_regen_run),
        _ = vpi::winCreateControl(wc_PushButton,rct(392,132,472,156),"Move Up",Win,[wsf_Group,wsf_TabStop],idc_up),
        _ = vpi::winCreateControl(wc_PushButton,rct(392,164,472,188),"Move Down",Win,[wsf_Group,wsf_TabStop],idc_down),
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
        handled(Result) = onSizeChanged(Width, Height).
    generatedEventHandler(_Win, e_Control(idc_add, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControladd(idc_add, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_close, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlclose(idc_close, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_delete_all, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlDeleteAll(idc_delete_all, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_load, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlload(idc_load, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_regen, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlregen(idc_regen, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_regen_only, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlRegenOnly(idc_regen_only, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_regen_run, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlRegenRun(idc_regen_run, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_remove, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlremove(idc_remove, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_run, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlrun(idc_run, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_save, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlsave(idc_save, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_down, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_down(idc_down, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_up, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_up(idc_up, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_CloseRequest()) = Result :-
        handled(Result) = onCloseRequest().
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
    % end of automatic code

getDBs([],[]) :-
    !.
getDBs([D|Ds],[F|Fs]) :-
    D=cmp("externalDatabaseFile",[nill,termDomain::str(F)]),
    getDBs(Ds,Fs),
    !
;
    bound(D),free(F),
    S=toString(D),
    M=string::concat("Script error - invalid externalDatabaseFile-tag: ",S,"!"),
    dlgIF::dlgNote(1,M),!,fail.
end implement multikb
