/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement rules
open core,vpiDomains,resourceIdentifiers,termDomain,globalDomains

constants
className = "com/rudstrand/TaskWindow/rules".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

facts
thisWin : vpiDomains::windowHandle := erroneous.

clauses
new(Parent):-
        _ = vpi::winCreate(windowType, rectangle, title, menu, Parent,
                               windowFlags, eventHandler, 0).

predicates
eventHandler : vpiDomains::ehandler.

class predicates
rdgReloadAll : (vpiDomains::windowHandle) determ.
rdgAllRules : (dbLoc_list,string_list,string_list,string_list) determ (i,o,o,o).
rdgDRules : (vpiDomains::windowHandle,string) determ. 
rdgDeleteRules : (string_list) determ.
rdgDisplayRules : (integer,vpiDomains::windowHandle,string,string) determ.
rdgDisplayRulesChain : (integer,vpiDomains::windowHandle,string) determ.
getRule : (string,dbLoc) determ (i,o).
rdgGetRealm : (string,string,string) determ (i,o,o).

clauses
        eventHandler(Win, Event) = Result :-
            Result = generatedEventHandler(Win, Event),
            !.
        eventHandler(Win, e_Menu(Id, CAS)) = Result :-
            Parent = vpi::winGetParent(Win),
            Result = vpi::winSendEvent(Parent, e_Menu(Id, CAS)).

/* by realm/chain */

    predicates
        onControlByChain : vpiDomains::controlHandler.
    clauses
        onControlByChain(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	Ctl=vpi::winGetCtlHandle(thisWin,idc_rules),
	vpi::lboxClear(Ctl),
  	guix::lbxGetAll(thisWin,idc_all_chains,S),
  	guix::lbxPop(thisWin,idc_rule_realms,S),
  	!;!.

    predicates
        onControlByRealm : vpiDomains::controlHandler.
    clauses
        onControlByRealm(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	Ctl=vpi::winGetCtlHandle(thisWin,idc_rules),
	vpi::lboxClear(Ctl),
  	guix::lbxGetAll(thisWin,idc_all_realms,S),
  	guix::lbxPop(thisWin,idc_rule_realms,S),
  	!;!.

/* realm/chain selected */

    predicates
        onControlRuleRealms : vpiDomains::controlHandler.
    clauses
        onControlRuleRealms(_CtrlID, _CtrlType, _CtrlWin,CtrlInfo) = handled(0):-
    CtrlInfo=vpiDomains::selchanged(),
	guix::lbxGet(thisWin,idc_rule_realms,RC),
  	rdgDRules(thisWin,RC),
	!;!.

/* rule selected */

    predicates
        onControlrules : vpiDomains::controlHandler.
    clauses
        onControlrules(_CtrlID, _CtrlType, _CtrlWin,CtrlInfo) = handled(0) :-
    CtrlInfo=vpiDomains::selchanged(),
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,No,SNo,_,Lt,_),
    Lt=cmp(Chain,_),
	NoS=toString(No),
	SNoS=toString(SNo),
    dbBase::nlSing(Chain,ChainNL),
	sx::sxlJoin([ChainNL,Realm,NoS,SNoS],"|",Rule),
	sentencemodel::dmAnnoDisplayAsst("p3_iSR",Rule),
  	guix::ctlEnable(thisWin,[idc_view,idc_view_record,idc_delete,idc_copy_rule]),
  	!
;  	
    CtrlInfo=vpiDomains::activated(),
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
	ruleeditor::rtrTreeDisplayRec(Loc),
	!;!.

/* view */

    predicates
        onControlview : vpiDomains::controlHandler.
    clauses
        onControlview(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
    globalVars::get_i("error_flag")=0,    
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
	ruleeditor::rtrTreeDisplayRec(Loc),
	!
;
    dlgIF::dlgNote(0,"Cannot open the rule-editor when ERROR-flag is ON!"),
    !.

/* view record */

    predicates
        onControlViewRecord : vpiDomains::controlHandler.
    clauses
        onControlViewRecord(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
	W=guix::getWinOfClas(1),
	RecordWindow=record::new(W),
	Wr=RecordWindow:getWin(),
	record::kbaWinDisplay(Wr,Loc),
	!;!.

/* delete */

    predicates
        onControldelete : vpiDomains::controlHandler.
    clauses
        onControldelete(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
	dbBase::recordDelete(Loc),
	guix::lbxDelS(thisWin,idc_rules,S2),
	%
	guix::lbxGet(thisWin,idc_rule_realms,RC),
  	rdgDRules(thisWin,RC),
  	% select first entry
    guix::lbxGetAll(thisWin,idc_rules,L),
    L=[H|_],
    guix::lbxSel(thisWin,idc_rules,H),
	!;!.		

/* delete all in realm/chain */

    predicates
        onControlDeleteAll : vpiDomains::controlHandler.
    clauses
        onControlDeleteAll(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	vpi::cursorSetWait(),
	guix::lbxGetAll(thisWin,idc_rules,Rules),
	rdgDeleteRules(Rules),
	guix::lbxGet(thisWin,idc_rule_realms,Realm),
	guix::lbxDelS(thisWin,idc_rule_realms,Realm),
	Ctl=vpi::winGetCtlHandle(thisWin,idc_rules),
	vpi::lboxClear(Ctl),
	vpi::cursorSet(thisWin,vpiDomains::cursor_Arrow),
	!;!.

/* enter new rule */

    predicates
        onControlnew : vpiDomains::controlHandler.
    clauses
        onControlnew(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
  	guix::guiRB_Get(thisWin,[idc_by_realm,idc_by_chain],[1,2],N),
  	N=1, % by realm
	guix::lbxGet(thisWin,idc_rule_realms,RC),
    Dlg = newrule::new(),
    Dlg:show(thisWin,RC),
    !
;    
    Dlg = newrule::new(),
    Dlg:show(thisWin,""),
	!;!.

/* refresh window */

    predicates
        onControlrefresh : vpiDomains::controlHandler.
    clauses
        onControlrefresh(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	vpi::cursorSetWait(),
  	guix::ctlDisable(thisWin,[idc_view,idc_view_record,idc_delete,idc_copy_rule,idc_delete_all]),
	guix::lbxGet(thisWin,idc_rule_realms,RC),
  	rdgReloadAll(thisWin),
	guix::lbxSel(thisWin,idc_rule_realms,RC),
  	rdgDRules(thisWin,RC),
  	% select first entry
    guix::lbxGetAll(thisWin,idc_rules,L),
    L=[H|_],
    guix::lbxSel(thisWin,idc_rules,H),
    !
;    
  	rdgReloadAll(thisWin),
	!;!.

/* copy rule */

    predicates
        onControlCopyRule : vpiDomains::controlHandler.
    clauses
        onControlCopyRule(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
    Dlg=saveasrealm::new(),
    Dlg:setPars(Loc),
    Dlg:show(thisWin),
	guix::lbxGet(thisWin,idc_rule_realms,RC),
	rdgReloadAll(thisWin),
	guix::lbxSel(thisWin,idc_rule_realms,RC),
  	rdgDRules(thisWin,RC),
  	% select first entry
    guix::lbxGetAll(thisWin,idc_rules,L),
    L=[H|_],
    guix::lbxSel(thisWin,idc_rules,H),
	!;!.

/* activate/inactivate suspend/unsuspend */

predicates
    onControlidc_activate : vpiDomains::controlHandler.
clauses
    onControlidc_activate(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
	dbBase::getRecord(Loc,R,_,_,_,_,_,_,_,_,_,_),
	Rm=-R,
	dbBase::recordSetType(Loc,Rm),
	guix::lbxGet(thisWin,idc_rule_realms,RC),
  	rdgDRules(thisWin,RC),
  	% select first entry
    guix::lbxGetAll(thisWin,idc_rules,L),
    L=[H|_],
    guix::lbxSel(thisWin,idc_rules,H),
	!;!.

predicates
    onControlidc_suspend : vpiDomains::controlHandler.
clauses
    onControlidc_suspend(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
	dbBase::getRecord(Loc,R,_,_,S,_,_,_,_,_,_,_),
	S=1,
	Rm=R div 10,
	dbBase::recordSetType(Loc,Rm),
	guix::lbxGet(thisWin,idc_rule_realms,RC),
  	rdgDRules(thisWin,RC),
  	% select first entry
    guix::lbxGetAll(thisWin,idc_rules,L),
    L=[H|_],
    guix::lbxSel(thisWin,idc_rules,H),
    !
;
	guix::lbxGet(thisWin,idc_rules,S2),
	getRule(S2,Loc),
	dbBase::getRecord(Loc,R,_,_,S,_,_,_,_,_,_,_),
	S=-1,
	Rm=R*10,
	dbBase::recordSetType(Loc,Rm),
	guix::lbxGet(thisWin,idc_rule_realms,RC),
  	rdgDRules(thisWin,RC),
  	% select first entry
    guix::lbxGetAll(thisWin,idc_rules,L),
    L=[H|_],
    guix::lbxSel(thisWin,idc_rules,H),
    !;!.
    
/* include/exclude suspended */

predicates
    onControlidc_include_suspended : vpiDomains::controlHandler.
clauses
    onControlidc_include_suspended(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	vpi::cursorSetWait(),
  	%guix::guiRB_SetN(thisWin,[idc_by_realm,idc_by_chain],[1,2],1),
  	guix::ctlDisable(thisWin,[idc_view,idc_view_record,idc_delete,idc_copy_rule,idc_delete_all]),
  	rdgReloadAll(thisWin),
  	!;!.

/* create/destroy */

    predicates
        onCreate : vpiDomains::createHandler.
    clauses
        onCreate(_CreationData) = defaultHandling() :-
	guix::addWin(42,thisWin),
    guix::setStyle(thisWin,[idc_rule_realms,idc_delete_all,idc_refresh,idc_suspend,
  	                                idc_by_realm,idc_by_chain,idc_activate,idc_include_suspended,
  	                                idc_rules,idc_new,idc_view,idc_view_record,idc_delete,idc_copy_rule],[]),
	vpi::cursorSetWait(),
  	guix::guiRB_SetN(thisWin,[idc_by_realm,idc_by_chain],[1,2],1),
	guix::guiCB_Set(thisWin,idc_include_suspended,-1),
  	guix::ctlDisable(thisWin,[idc_view,idc_view_record,idc_delete,idc_copy_rule,idc_delete_all]),
  	rdgReloadAll(thisWin),
  	!;!.

    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            guix::dropWin(thisWin),
            thisWin := erroneous.
    predicates
        onSizeChanged : vpiDomains::sizeHandler.
    clauses
        onSizeChanged(_Width, _Height) = defaultHandling():-
            vpiToolbar::resize(thisWin).

% This code is maintained by the VDE do not update it manually, 13:01:12-4.12.2005
constants
    windowType : vpiDomains::wintype = w_TopLevel.
    windowFlags : vpiDomains::wsflags = [wsf_TitleBar,wsf_Minimize,wsf_Close,wsf_ClipSiblings,wsf_ClipChildren].
    rectangle : vpiDomains::rct = rct(100,50,761,589).
    menu : vpiDomains::menu = noMenu.
    title : string = "Sentence Rules".

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        _ = vpi::winCreateControl(wc_LBox,rct(208,4,544,268),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight],idc_rule_realms),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,244,200,268),"Delete Rules In Realm/Chain",Win,[wsf_Group,wsf_TabStop],idc_delete_all),
        _ = vpi::winCreateControl(wc_PushButton,rct(552,178,656,210),"Delete Rule",Win,[wsf_Group,wsf_TabStop],idc_delete),
        _ = vpi::winCreateControl(wc_PushButton,rct(552,236,656,268),"View Record",Win,[wsf_Group,wsf_TabStop],idc_view_record),
        _ = vpi::winCreateControl(wc_PushButton,rct(552,62,656,94),"View/Edit Rule",Win,[wsf_Group,wsf_TabStop],idc_view),
        _ = vpi::winCreateControl(wc_PushButton,rct(552,4,656,36),"New Rule",Win,[wsf_Group,wsf_TabStop],idc_new),
        _ = vpi::winCreateControl(wc_LBox,rct(8,272,656,508),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight],idc_rules),
        _ = vpi::winCreateControl(wc_PushButton,rct(552,120,656,152),"Copy Rule",Win,[wsf_Group,wsf_TabStop],idc_copy_rule),
        _ = vpi::winCreateControl(wc_LBox,rct(472,396,503,423),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_NoIntegralHeight,wsf_Disabled,wsf_Invisible],idc_all_rules),
        _ = vpi::winCreateControl(wc_LBox,rct(512,396,543,423),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_NoIntegralHeight,wsf_Disabled,wsf_Invisible],idc_all_realms),
        _ = vpi::winCreateControl(wc_LBox,rct(488,428,519,455),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_NoIntegralHeight,wsf_Disabled,wsf_Invisible],idc_all_chains),
        _ = vpi::winCreateControl(wc_RadioButton,rct(8,4,200,24),"List rule realms",Win,[wsf_Auto,wsf_TabStop,wsf_Group],idc_by_realm),
        _ = vpi::winCreateControl(wc_RadioButton,rct(8,28,200,48),"List rule chains",Win,[wsf_Auto,wsf_TabStop],idc_by_chain),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,216,200,240),"Refresh",Win,[wsf_Group,wsf_TabStop],idc_refresh),
        _ = vpi::winCreateControl(wc_CheckBox,rct(8,72,192,92),"Include suspended rules",Win,[wsf_Auto,wsf_Group,wsf_TabStop],idc_include_suspended),
        _ = vpi::winCreateControl(wc_PushButton,rct(8,512,167,536),"Activate/Inactivate",Win,[wsf_Group,wsf_TabStop],idc_activate),
        _ = vpi::winCreateControl(wc_PushButton,rct(176,512,335,536),"Suspend/Unsuspend",Win,[wsf_Group,wsf_TabStop],idc_suspend),
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
        handled(Result) = onSizeChanged(Width, Height).
    generatedEventHandler(_Win, e_Control(idc_by_chain, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlByChain(idc_by_chain, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_by_realm, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlByRealm(idc_by_realm, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_rule_realms, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlRuleRealms(idc_rule_realms, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_rules, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlrules(idc_rules, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_view, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlview(idc_view, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_view_record, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlViewRecord(idc_view_record, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_delete, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControldelete(idc_delete, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_delete_all, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlDeleteAll(idc_delete_all, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_new, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlnew(idc_new, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_refresh, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlrefresh(idc_refresh, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_copy_rule, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlCopyRule(idc_copy_rule, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_activate, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_activate(idc_activate, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_suspend, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_suspend(idc_suspend, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_include_suspended, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_include_suspended(idc_include_suspended, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
    % end of automatic code
% ----------------------------------------------------------------------------------------------
% rdgReloadAll
% ----------------------------------------------------------------------------------------------
class predicates
rdgReloadX : (integer,dbLoc_list) procedure (i,o).
clauses

rdgReloadAll(W) :-
  	guix::guiRB_Get(W,[idc_by_realm,idc_by_chain],[1,2],N),
  	N=1,
  	guix::guiCB_Get(W,idc_include_suspended,Susp),
  	rdgReloadX(Susp,RC),
 	rdgAllRules(RC,_,ChainsN,RealmsN),
 	set::distinct(ChainsN,Chains),
 	set::distinct(RealmsN,Realms),
	guix::lbxPop(W,idc_rule_realms,Realms),
	guix::lbxPop(W,idc_all_realms,Realms),
	guix::lbxPop(W,idc_all_chains,Chains),
	Ctl=vpi::winGetCtlHandle(W,idc_rules),
	vpi::lboxClear(Ctl),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!
;	
  	guix::guiRB_Get(W,[idc_by_realm,idc_by_chain],[1,2],N),
  	N=2,
  	guix::guiCB_Get(W,idc_include_suspended,Susp),
  	rdgReloadX(Susp,RC),
 	rdgAllRules(RC,_,ChainsN,RealmsN),
 	set::distinct(ChainsN,Chains),
 	set::distinct(RealmsN,Realms),
	guix::lbxPop(W,idc_rule_realms,Chains),
	guix::lbxPop(W,idc_all_realms,Realms),
	guix::lbxPop(W,idc_all_chains,Chains),
	Ctl=vpi::winGetCtlHandle(W,idc_rules),
	vpi::lboxClear(Ctl),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.

rdgReloadX(Susp,RC) :-
    Susp=1,
	dbBase::recTypeLocs(3,RCa),
	dbBase::recTypeLocs(-3,RCb),
	dbBase::recTypeLocs(30,RCc),
	dbBase::recTypeLocs(-30,RCd),
	set::conc([RCa,RCb,RCc,RCd],RC),
	!
;	
    Susp=-1,
	dbBase::recTypeLocs(3,RCa),
	dbBase::recTypeLocs(-3,RCb),
	set::conc([RCa,RCb],RC),
	!
;	
	RC=[],
	!.
% ----------------------------------------------------------------------------------------------
% rdgAllRules
% ----------------------------------------------------------------------------------------------
rdgAllRules([],[],[],[]) :-
	!.
rdgAllRules([Loc|Locs],[RCm|RCsm],[ChainNL|Chains],[KRealm|Realms]) :-
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,LT,_),
    LT=cmp(Chain,_),
	dbBase::nlSing(Chain,ChainNL),
	KRealm=dbExtension::realmPrefix(Chain,Realm),
	SLoc=dbBase::dbRef(Loc),
	SLoc2=string::concat("[",SLoc,"]"),
	sx::sxlJoin([SLoc2,ChainNL,KRealm],"|",RCm),
	rdgAllRules(Locs,RCsm,Chains,Realms),
	!.
% ----------------------------------------------------------------------------------------------
% rdgDRules
% ----------------------------------------------------------------------------------------------
rdgDRules(W,RC) :-
  	guix::guiRB_Get(W,[idc_by_realm,idc_by_chain],[1,2],N),
  	N=1, % by realm
  	Suffix=dbBase::dbId(), %@@ to be removed with phase out of multi-KB
  	Realm=string::concat(RC,"|",Suffix),
	sentencemodel::dmAnnoDisplayAsst("p3_iRelm",Realm),
  	guix::guiCB_Get(W,idc_include_suspended,Susp),
  	dbExtension::realmRemovePrefix(RC,R,Id),
	rdgDisplayRules(Susp,W,Id,R),
	guix::ctlEnable(W,[idc_delete_all]),  	
	guix::ctlDisable(W,[idc_view,idc_view_record,idc_delete,idc_copy_rule]),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!
;	
  	guix::guiRB_Get(W,[idc_by_realm,idc_by_chain],[1,2],N),
  	N=2, % by chain
	sentencemodel::dmAnnoDisplayAsst("p3_iPT",RC),
  	guix::guiCB_Get(W,idc_include_suspended,Susp),
  	dbBase::kbID(RC,Chain,_,_),
	rdgDisplayRulesChain(Susp,W,Chain),
	guix::ctlEnable(W,[idc_delete_all]),  	
  	guix::ctlDisable(W,[idc_view,idc_view_record,idc_delete,idc_copy_rule]),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.
% ----------------------------------------------------------------------------------------------
% rdgDisplayRules/Chain
% ----------------------------------------------------------------------------------------------
class predicates
addRuleInRealm : (vpiDomains::windowHandle,dbLoc_list,string,string) determ.
addRuleInChain : (vpiDomains::windowHandle,dbLoc_list,string) determ.
stGroup : (sterm,string) determ (i,o).	
argE : (sterm,string) determ (i,o).
clauses

% parameters:
% 1 in -1/1 excl/incl susp rules
% 2 in win
% 3	in	KB id
% 4	in	realm name
%
rdgDisplayRules(Susp,Win,Id,Realm) :-
    Susp=1,
	C=vpi::winGetCtlHandle(Win,idc_rules),
	vpi::lboxClear(C),
	dbBase::recTypeLocs(3,L2),
	dbBase::recTypeLocs(-3,L3),
	dbBase::recTypeLocs(30,L4),
	dbBase::recTypeLocs(-30,L5),
	set::conc([L2,L3,L4,L5],L),
	addRuleInRealm(C,L,Id,Realm),
	!
;	
    Susp=-1,
	C=vpi::winGetCtlHandle(Win,idc_rules),
	vpi::lboxClear(C),
	dbBase::recTypeLocs(3,L2),
	dbBase::recTypeLocs(-3,L3),
	set::conc([L2,L3],L),
	addRuleInRealm(C,L,Id,Realm),
	!.
	
% parameters:
% 3	in	chain name
%
rdgDisplayRulesChain(Susp,Win,Chain) :-
    Susp=1,
	C=vpi::winGetCtlHandle(Win,idc_rules),
	vpi::lboxClear(C),
	dbBase::recTypeLocs(3,L2),
	dbBase::recTypeLocs(-3,L3),
	dbBase::recTypeLocs(30,L4),
	dbBase::recTypeLocs(-30,L5),
	set::conc([L2,L3,L4,L5],L),
	addRuleInChain(C,L,Chain),
	!
;	
    Susp=-1,
	C=vpi::winGetCtlHandle(Win,idc_rules),
	vpi::lboxClear(C),
	dbBase::recTypeLocs(3,L2),
	dbBase::recTypeLocs(-3,L3),
	set::conc([L2,L3],L),
	addRuleInChain(C,L,Chain),
	!.

addRuleInRealm(_,[],_,_) :-
	!.
addRuleInRealm(Ctl,[Loc|Locs],Id,Realm) :-
    Id=dbBase::dbId(Loc),
    dbBase::getRecord(Loc,RT,_,_,_,_,RealmX,SentNo,SubNo,_,LT,_),
    RealmX=Realm,
	S2=toString(SentNo),
	S3=toString(SubNo),
	dbExtension::dbuRT_Text(-1,RT,St),
    dbExtension::dbuSR_Type(Loc,TypeNo),
    rub::rubRuleType(_,TypeNo,TypeK),
    dbBase::nlSing(TypeK,Type),
    stGroup(LT,Stn),
	SRef=dbBase::dbRef(Loc),
	sx::concat([Realm,"|",S2,"|",S3,": ",Stn," (",Type," - ",St,") [",SRef,"]"],S),
	vpi::lboxAdd(Ctl,S),
	addRuleInRealm(Ctl,Locs,Id,Realm),
	!
;	
	addRuleInRealm(Ctl,Locs,Id,Realm),
	!.

addRuleInChain(_,[],_) :-
	!.
addRuleInChain(Ctl,[Loc|Locs],Chain) :-
    dbBase::getRecord(Loc,RT,_,_,_,_,Realm,SentNo,SubNo,_,LT,_),
    LT=cmp(ChainX,_),
    ChainX=Chain,
	S2=toString(SentNo),
	S3=toString(SubNo),
	dbExtension::dbuRT_Text(-1,RT,St),
    dbExtension::dbuSR_Type(Loc,TypeNo),
    rub::rubRuleType(_,TypeNo,TypeK),
    dbBase::nlSing(TypeK,Type),
    stGroup(LT,Stn),
	SRef=dbBase::dbRef(Loc),
	sx::concat([Realm,"|",S2,"|",S3,": ",Stn," (",Type," - ",St,") [",SRef,"]"],S),
	vpi::lboxAdd(Ctl,S),
	addRuleInChain(Ctl,Locs,Chain),
	!
;	
	addRuleInChain(Ctl,Locs,Chain),
	!.

stGroup(LT,St) :-
    LT=cmp(Sg,[A1,A2|_]),
    dbBase::smClassDefF(Sg,"p",Card,_,_,_,_),
    Card=2,
    argE(A1,A1n),
    argE(A2,A2n),
    dbBase::nlSing(Sg,Sgn),
    St=string::concat(A1n," ",Sgn," ",A2n),
    !
;    
    LT=cmp(Sg,[A1|_]),
    dbBase::smClassDefF(Sg,"p",Card,_,_,_,_),
    Card=1,
    argE(A1,A1n),
    dbBase::nlSing(Sg,Sgn),
    St=string::concat(A1n," ",Sgn),
    !
;
    dlgIF::dlgNote(1,"Unrecognizable sentence rule record!"),
    !,fail.
    
argE(var(X),Xn) :- sx::sxQuote(X,Xn),!.
argE(cmp(E,_),En) :- dbBase::nlSing(E,En),!.        
% ----------------------------------------------------------------------------------------------
% getRule
% ----------------------------------------------------------------------------------------------
getRule(S,Loc) :-
    sx::sxGetInBrack(S,'[',']',SLoc),
    Loc=dbBase::dbRefR(SLoc),
    !.
% ----------------------------------------------------------------------------------------------
% rdgGetRealm
% ----------------------------------------------------------------------------------------------
rdgGetRealm(S,Realm,SNo) :-
	% SG-rule
	sx::sxUnpackS(S,": ",Lr),
	Lr=[RealmData,S5],
	dbExtension::dbuSentenceR(Realm,SentNo,SubNo,RealmData),
	sx::searchchar(S5,'[',P5),
	N5=P5-1,
	sx::frontstr(N5,S5,Pn,_),
	dbBase::kbID_F(Pn,_,Clas,_),
	Clas="p3_iPT",
	SubNo>0,
	No=SentNo*1000+SubNo,
	SNo=toString(No),
	!
;	
	% SG-rule
	sx::sxUnpackS(S,": ",Lr),
	Lr=[RealmData,S5],
	dbExtension::dbuSentenceR(Realm,SentNo,SubNo,RealmData),
	sx::searchchar(S5,'[',P5),
	N5=P5-1,
	sx::frontstr(N5,S5,Pn,_),
	dbBase::kbID_F(Pn,_,Clas,_),
	Clas="p3_iPT",
	SubNo=0,
	SNo=toString(SentNo),
	!
;	
	% ST-rule
	sx::sxUnpackS(S,": ",Lr),
	Lr=[RealmData,_],
	dbExtension::dbuSentenceR(Realm,SentNo,SubNo,RealmData),
	SubNo>0,
	No=SentNo*1000+SubNo,
	SNo=toString(No),
	!
;	
	% ST-rule
	sx::sxUnpackS(S,": ",Lr),
	Lr=[RealmData,_],
	dbExtension::dbuSentenceR(Realm,SentNo,SubNo,RealmData),
	SubNo=0,
	SNo=toString(SentNo),
	!.
% ----------------------------------------------------------------------------------------------
% rdgDeleteRules
% ----------------------------------------------------------------------------------------------
rdgDeleteRules([]) :-
	!.
rdgDeleteRules([R|Rules]) :-
	getRule(R,Loc),
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,SNo,_,_,_),
	SNo<>0,
	dbBase::recordDelete(Loc),
	rdgDeleteRules(Rules),
	!
;
	getRule(R,Loc),
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,No,SNo,_,_,_),
	SNo=0,
	dbBase::realmDeleteNo(Realm,No),
	rdgDeleteRules(Rules),
	!.
end implement rules
