/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement servermode
open core,vpiDomains,resourceIdentifiers,termDomain

constants
className = "com/rudstrand/TaskWindow/servermode/servermode".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

class facts
nLoops : integer := 0. % # processing loops
nRequests : integer := 0. % #requests completed
fOutq : integer := 0. % temporary storage for global f_outq
crLog : integer :=0. % temporary storage for global crlog
crDet : integer :=0. % temporary storage for global crdet
kbName : string := "". % KB-name (net of path/type)
waitTimer : integer := -1.
serverError : integer := -1. % set to 1 if an error which impairs the server occurs
stream : integer :=0. % -1/1 knowledge stream is being used
subs : integer :=0. % -1/1 subscription is being used
stList : string_list := []. % ST derivation list on the format returned by depList
facts
thisWin : vpiDomains::windowHandle := erroneous.

predicates
checkStream : ().
checkStreamChange : ().
execRequest : (sqlSelect) determ.
processReport : (string,integer) determ.
processRequest : (string,integer,string,integer) determ.
processSubRequest : (string,integer,string).
setServerStatus : (integer) determ.
setServerStatus : (integer,string) determ.
addLog : (string).
processNext : ().
connectServer : () determ.
dropRespTables : (string).
exportQueries : () determ.
processCluster : (integer,integer,integer,string,string_list) determ.
fatalError : (string).
cleanDb : ().
removeStartedRequests : ().
serverStatus : ().       
setStream : ().
subscribe : () determ.
optimizeSubs : () determ.
     
clauses
show(Parent):-
     _ = vpi::winCreateDynDialog(Parent, controlList, eventHandler, 0).

predicates
eventHandler : vpiDomains::ehandler.

clauses
        eventHandler(Win, Event) = Result :-
            Result = generatedEventHandler(Win, Event).

    predicates
        onTimer : vpiDomains::timerHandler.
    clauses
        onTimer(TimerID) = defaultHandling() :-
        TimerID=waitTimer,
        nLoops:=nLoops+1,
        nLoops mod 600=0,
        tt::getLevel()>0,
        tt::stopProfile(),
        xmlGeneric::timeNowXML(T),
        tt::memory(1,T),
        tt::startProfile("processNext"),
        processNext,
        !
;        
        TimerID=waitTimer,
        nLoops:=nLoops+1,
        processNext,
        !;!.
        
/* connect */

    predicates
        onControlOK : vpiDomains::controlHandler.
    clauses
        onControlOK(_Ctrl, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        connectServer,
        setStream,
        guix::lbtSel(thisWin,idc_status,"SERVER CONNECTED"),
        guix::ctlVisible(thisWin,[idct_3,idc_started]),
        xmlGeneric::timeNowXML(T),
        guix::txtSet(thisWin,idc_started,T),
        setStream,
        guix::ctlDisable(thisWin,[idc_ok,idc_cancel]),
        guix::ctlEnable(thisWin,[idc_stream,idc_process_requests,idc_disconnect,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall]),
        guix::ctlVisible(thisWin,[idc_processed,idct_5,idc_count]),
        !;!.

/* process */

    predicates
        onControlProcessRequests : vpiDomains::controlHandler.
    clauses
        onControlProcessRequests(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
        serverError=1,
        vpi::timerKill(waitTimer),
        guix::ctlDisable(thisWin,[idc_pause]),
        guix::ctlEnable(thisWin,[idc_process_requests,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall,idc_disconnect]),
        guix::lbtSel(thisWin,idc_status,"PAUSE REQUESTED"),
        setServerStatus(2),
        removeStartedRequests,
        ir::setBreak(0),
        guix::txtSet(thisWin,idc_waiting,"PAUSED"),
        !
;        
        sqlSt::hasSubscriptions,
        guix::txtGet(thisWin,idc_stream,Stream),
        Stream="",
        stream:=-1,
        setServerStatus(1,Stream),
        guix::lbtSel(thisWin,idc_status,"PROCESSING"),
        guix::ctlDisable(thisWin,[idc_stream,idc_process_requests,idc_disconnect,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall]),
        guix::ctlEnable(thisWin,[idc_pause]),
        cleanDb,
        optimizeSubs,
        tt::startProfile("processNext"),
        processNext,
        !
;        
        sqlSt::hasSubscriptions,
        guix::txtGet(thisWin,idc_stream,Stream),
        Stream<>"",
        setServerStatus(1,Stream),
        guix::lbtSel(thisWin,idc_status,"PROCESSING"),
        guix::ctlDisable(thisWin,[idc_stream,idc_process_requests,idc_disconnect,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall]),
        guix::ctlEnable(thisWin,[idc_pause]),
        cleanDb,
        subs:=1,
        checkStream,
        tt::startProfile("processNext"),
        processNext,
        !
;        
        not(sqlSt::hasSubscriptions),
        subs:=-1,
        guix::txtGet(thisWin,idc_stream,Stream),
        Stream="",
        stream:=-1,
        setServerStatus(1,Stream),
        guix::lbtSel(thisWin,idc_status,"PROCESSING"),
        guix::ctlDisable(thisWin,[idc_stream,idc_process_requests,idc_disconnect,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall]),
        guix::ctlEnable(thisWin,[idc_pause]),
        cleanDb,
        tt::startProfile("processNext"),
        processNext,
        !
;        
        not(sqlSt::hasSubscriptions),
        subs:=-1,
        guix::txtGet(thisWin,idc_stream,Stream),
        setServerStatus(1,Stream),
        guix::lbtSel(thisWin,idc_status,"PROCESSING"),
        guix::ctlDisable(thisWin,[idc_stream,idc_process_requests,idc_disconnect,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall]),
        guix::ctlEnable(thisWin,[idc_pause]),
        cleanDb,
        checkStream,
        tt::startProfile("processNext"),
        processNext,
        !;!.

/* pause */
        
    predicates
        onControlpause : vpiDomains::controlHandler.
    clauses
        onControlpause(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        vpi::timerKill(waitTimer),
        guix::ctlDisable(thisWin,[idc_pause]),
        guix::ctlEnable(thisWin,[idc_stream,idc_process_requests,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall,idc_disconnect]),
        guix::lbtSel(thisWin,idc_status,"PAUSE REQUESTED"),
        setServerStatus(2),
        guix::txtSet(thisWin,idc_waiting,"PAUSED"),
        removeStartedRequests,
        guix::lbtSel(thisWin,idc_status,"SERVER CONNECTED"),
        !;!.

/* disconnect */

    predicates
        onControldisconnect : vpiDomains::controlHandler.
    clauses
        onControldisconnect(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
        dbBase::kbIsSql,
        guix::txtGet(thisWin,idc_stream,Stream),
        setServerStatus(0,Stream),
        guix::lbtSel(thisWin,idc_status,"SERVER NOT CONNECTED"),
        guix::ctlDisable(thisWin,[idc_process_requests,idc_pause,idc_disconnect,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall]),
        guix::ctlEnable(thisWin,[idc_ok,idc_cancel]),
        guix::ctlInvisible(thisWin,[idc_processed,idct_5,idc_count]),
        !
;        
        not(dbBase::kbIsSql),
        guix::txtGet(thisWin,idc_stream,Stream),
        setServerStatus(0,Stream),
        sqlBase::closeConnection,
        guix::lbtSel(thisWin,idc_status,"SERVER NOT CONNECTED"),
        guix::ctlDisable(thisWin,[idc_process_requests,idc_pause,idc_disconnect,idc_delete,idc_resetall,idc_reset_dispatcher,idc_export,idc_reset_rtall]),
        guix::ctlEnable(thisWin,[idc_ok,idc_cancel]),
        guix::ctlInvisible(thisWin,[idc_processed,idct_5,idc_count]),
        !;!.

/* export queries */
        
    predicates
    onControlidc_export : vpiDomains::controlHandler.
clauses
    onControlidc_export(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    exportQueries,
    addLog("Exported KB-queries."),
    !;!.

/* reset response table */        

    predicates
        onControldelete : vpiDomains::controlHandler.
    clauses
        onControldelete(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    vpiCommonDialogs::ask("Please confirm!","Do you want to reset the dispatcher response table for this KB?",["Yes","No"])=vpiCommonDialogs::resp_default,
    dropRespTables(kbName),        
    Con=sqlBase::getConnection(),
    S3=string::concatList(["delete from p3bridge.p3response_column where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S3),
    S4=string::concatList(["delete from p3bridge.p3response_table where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S4),
    S5=string::concatList(["delete from p3bridge.p3response where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S5),
    S6=string::concatList(["delete from p3bridge.p3question where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S6),
    S7=string::concatList(["delete from p3bridge.p3answer where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S7),
    S8=string::concatList(["delete from p3bridge.p3input_facility where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S8),
                Mess=string::concat("Reset response-tables for ",kbName,"."),
                addLog(Mess),
    !;!.

    predicates
        onControlResetRtall : vpiDomains::controlHandler.
    clauses
        onControlResetRtall(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
    vpiCommonDialogs::ask("Please confirm!","Do you want to reset the dispatcher response table for all KBs?",["Yes","No"])=vpiCommonDialogs::resp_default,        
    dropRespTables(""),        
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3response_table"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3response_column"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3response"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3question"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3input_facility"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3answer"),
                addLog("Reset response-tables for all KBs."),
    !;!.

/* reset dispatcher */

    predicates
        onControlResetDispatcher : vpiDomains::controlHandler.
    clauses
        onControlResetDispatcher(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    vpiCommonDialogs::ask("Please confirm!","Do you want to reset the dispatcher tables for this KB?",["Yes","No"])=vpiCommonDialogs::resp_default,        
    dropRespTables(kbName),        
    S6=string::concatList(["delete from p3bridge.p3question where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S6),
    S7=string::concatList(["delete from p3bridge.p3answer where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S7),
    S1=string::concatList(["delete from p3bridge.p3response_column where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S1),
    S4=string::concatList(["delete from p3bridge.p3response_table where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S4),
    S5=string::concatList(["delete from p3bridge.p3response where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S5),
    S3=string::concatList(["delete from p3bridge.p3request where kb='",kbName,"'"]),
    sqlTools::executeDirect(Con,S3),
    S2=string::concatList(["delete from p3bridge.p3server where kb='",kbName,"'"]),
    sqlTools::executeDirect(Con,S2),
    S8=string::concatList(["delete from p3bridge.p3input_facility where request_no in (select request_no from p3bridge.p3request where kb='",kbName,"')"]),
    sqlTools::executeDirect(Con,S8),
    S21=string::concat("delete from p3bridge.p3alternative where user_name='Pro/3' and kb='",kbName,"'"),
    sqlTools::executeDirect(Con,S21),
    S22=string::concat("delete from p3bridge.p3cluster where user_name='Pro/3' and kb='",kbName,"'"),
    sqlTools::executeDirect(Con,S22),
    S23=string::concat("delete from p3bridge.p3query where user_name='Pro/3' and kb='",kbName,"'"),
    sqlTools::executeDirect(Con,S23),
    S24=string::concat("delete from p3bridge.p3query_cluster where user_name='Pro/3' and kb='",kbName,"'"),
    sqlTools::executeDirect(Con,S24),
    S25=string::concat("delete from p3bridge.p3query_parameter where user_name='Pro/3' and kb='",kbName,"'"),
    sqlTools::executeDirect(Con,S25),
    S26=string::concat("delete from p3bridge.p3parameter where user_name='Pro/3' and kb='",kbName,"'"),
    sqlTools::executeDirect(Con,S26),
    setServerStatus(2),
                Mess=string::concat("Reset P3BRIDGE for ",kbName,"."),
                addLog(Mess),
    !;!.

    predicates
        onControlresetall : vpiDomains::controlHandler.
    clauses
        onControlresetall(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    vpiCommonDialogs::ask("Please confirm!","Do you want to reset the dispatcher for all KBs?",["Yes","No"])=vpiCommonDialogs::resp_default,        
     dropRespTables(""),       
    Con=sqlBase::getConnection(), 
    sqlTools::executeDirect(Con,"delete from p3bridge.p3request"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3server"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3response_table"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3response_column"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3response"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3question"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3input_facility"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3answer"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3alternative where user_name='Pro/3'"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3cluster where user_name='Pro/3'"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3query where user_name='Pro/3'"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3query_cluster where user_name='Pro/3'"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3query_parameter where user_name='Pro/3'"),
    sqlTools::executeDirect(Con,"delete from p3bridge.p3parameter where user_name='Pro/3'"),
    %
    sqlTools::executeDirect(Con,"delete from p3bridge.p3admin"),
    setServerStatus(2),
    Sql2=string::concatList(["insert p3bridge.p3admin (last_used_no) values (0)"]),
    sqlTools::executeDirect(Con,Sql2),
                addLog("Reset P3BRIDGE for all KBs."),
    !;!.

/* help */

    predicates
        onControlhelp : vpiDomains::controlHandler.
    clauses
        onControlhelp(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        dlgIF::hlpWeb("Documentation"),
        !;!.

/* exit */
                        
    predicates
        onControlCancel : vpiDomains::controlHandler.
    clauses
        onControlCancel(_Ctrl, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
            removeStartedRequests,
            cleanDb,
            ir::setBreak(0),
            guix::dropWin(thisWin),
            vpi::winDestroy(thisWin).

/* create etc */

    predicates
        onCreate : vpiDomains::createHandler.
    clauses
        onCreate(_CreationData) = defaultHandling() :-
        F=globalVars::get_s("kb_file"),
        F="",
        fatalError("Blank KB-name!"),
        !
;        
        F=globalVars::get_s("kb_file"),
        kbName:=fileName::getName(F),
        guix::lbxPop(thisWin,idc_status,["PROCESSING","SERVER CONNECTED","PAUSE REQUESTED","SERVER NOT CONNECTED"]),
        guix::ctlDisable(thisWin,[idc_stream,idc_process_requests,idc_pause,idc_disconnect,idc_delete]),
        guix::lbtSel(thisWin,idc_status,"SERVER NOT CONNECTED"),
        guix::ctlInvisible(thisWin,[idct_3,idc_started,idc_processed,idct_5,idc_count]),
        fOutq:=globalVars::get_i("f_outq"),
        crLog:=globalVars::get_i("crlog"),
        globalVars::set_i("crlog",-1),
        crDet:=globalVars::get_i("crdet"),
        globalVars::set_i("crdet",-1),
        guix::addWin(5,thisWin),
        globalVars::set_i("server",1),
        serverError:=-1,
        !;!.
        
    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            guix::dropWin(thisWin),
            % restore format globals
            globalVars::set_i("f_outq",fOutq),
            globalVars::set_i("crlog",crLog),
            globalVars::set_i("crdet",crDet),
            globalVars::set_i("server",-1),
            thisWin := erroneous.

    predicates
        onCloseRequest : vpiDomains::closeRequestHandler.
    clauses
        onCloseRequest() = defaultHandling() :-
        !.

% This code is maintained by the VDE do not update it manually, 21:54:33-11.6.2006
constants
    dialogType : vpiDomains::wintype = wd_modal.
    title : string = "Server Mode".
    rectangle : vpiDomains::rct = rct(10,40,494,335).
    dialogFlags : vpiDomains::wsflags = [wsf_Close,wsf_TitleBar].
    dialogFont = "MS Sans Serif".
    dialogFontSize = 8.

constants
    controlList : vpiDomains::windef_list =
        [
        dlgFont(wdef(dialogType, rectangle, title, u_DlgBase),
                dialogFont, dialogFontSize, dialogFlags),
        ctl(wdef(wc_PushButton,rct(4,4,80,16),"Connect",u_DlgBase),idc_ok,[wsf_Default,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(324,4,400,16),"Exit server mode",u_DlgBase),idc_cancel,[wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(404,4,480,16),"&Help",u_DlgBase),idc_help,[wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(84,4,160,16),"Process requests",u_DlgBase),idc_process_requests,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_PushButton,rct(164,4,240,16),"Pause",u_DlgBase),idc_pause,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_PushButton,rct(244,4,320,16),"Disconnect",u_DlgBase),idc_disconnect,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_LBoxButton,rct(84,68,200,186),"",u_DlgBase),idc_status,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_Disabled]),
        ctl(wdef(wc_LBox,rct(4,84,480,292),"",u_DlgBase),idc_processed,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_NoIntegralHeight]),
        ctl(wdef(wc_PushButton,rct(4,36,80,48),"Reset Response Table",u_DlgBase),idc_delete,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_PushButton,rct(204,36,280,48),"Reset Dispatcher",u_DlgBase),idc_reset_dispatcher,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_PushButton,rct(84,36,200,48),"Reset Response Table ALL KBs",u_DlgBase),idc_reset_rtall,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_PushButton,rct(284,36,400,48),"Reset Dispatcher ALL KBs",u_DlgBase),idc_resetall,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_PushButton,rct(404,36,480,48),"Export queries",u_DlgBase),idc_export,[wsf_Group,wsf_TabStop,wsf_Disabled]),
        ctl(wdef(wc_Edit,rct(84,20,240,32),"",u_DlgBase),idc_stream,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll]),
        ctl(wdef(wc_Text,rct(204,52,280,64),"#requests processed:",u_DlgBase),idct_5,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(284,52,316,64),"",u_DlgBase),idc_count,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(4,68,80,80),"Status:",u_DlgBase),idct_2,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(4,52,80,64),"Connected since:",u_DlgBase),idct_3,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(84,52,200,64),"",u_DlgBase),idc_started,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(204,68,364,80),"",u_DlgBase),idc_waiting,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(320,52,480,64),"",u_DlgBase),idc_rstatus,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(368,68,480,80),"",u_DlgBase),idc_currtime,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(4,22,80,32),"Knowledge stream",u_DlgBase),idct_stream,[wsf_AlignLeft])
        ].

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Control(idc_ok, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlOK(idc_ok, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_cancel, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlCancel(idc_cancel, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_help, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlhelp(idc_help, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_pause, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlpause(idc_pause, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_process_requests, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlProcessRequests(idc_process_requests, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_disconnect, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControldisconnect(idc_disconnect, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_delete, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControldelete(idc_delete, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_reset_dispatcher, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlResetDispatcher(idc_reset_dispatcher, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_reset_rtall, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlResetRtall(idc_reset_rtall, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_resetall, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlresetall(idc_resetall, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_export, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_export(idc_export, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_CloseRequest()) = Result :-
        handled(Result) = onCloseRequest().
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
    generatedEventHandler(_Win, e_Timer(TimerID)) = Result :-
        handled(Result) = onTimer(TimerID).
    % end of automatic code
% ----------------------------------------------------------------------------------------------
% connectServer
% ----------------------------------------------------------------------------------------------
% refer to XP's ODBC Data Source Administrator, which may or may not preset
% sql user ID, password and db to be used.
%
connectServer() :-
    sqlBase::openConnection,
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,"use p3bridge"),
    setServerStatus(2),
    guix::txtSet(thisWin,idc_count,"0"),
    sqlSt::createDbIfnotexists,
    !
;
    fatalError("Cannot connect to server!"),
    !,fail.
% ----------------------------------------------------------------------------------------------
% removeStartedRequests
% ----------------------------------------------------------------------------------------------
% sets requests (should be at most one) which were left in started-mode last
% time the server was processing, as in error
%
removeStartedRequests() :-
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select request_no from p3bridge.p3request where kb='",kbName,"' and request_status=2"),
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Request=Select:getI(Nc,1), 
        SRequest=toString(Request),
        S9=string::concatList(["update p3bridge.p3request set request_status=-10 where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S9),
        Mess=string::concat("Removed started request #",SRequest,"!"),
        addLog(Mess),
        fail
;       
    !.
% ----------------------------------------------------------------------------------------------
% cleanDb
% ----------------------------------------------------------------------------------------------
predicates
cleanTables : (odbcConnection,string).
clauses

cleanDb() :-
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select request_no,request_status from p3bridge.p3request where kb='",kbName,"' and request_status in (4,5)"),
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Request=Select:getI(Nc,1), 
        Status=Select:getI(Nc,2),
        SRequest=toString(Request),
        cleanTables(Con,SRequest),
        S3=string::concatList(["delete from p3bridge.p3response_column where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S3),
        S4=string::concatList(["delete from p3bridge.p3response_table where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S4),
        S5=string::concatList(["delete from p3bridge.p3response where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S5),
        S6=string::concatList(["delete from p3bridge.p3question where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S6),
        S7=string::concatList(["delete from p3bridge.p3answer where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S7),
        S8=string::concatList(["delete from p3bridge.p3input_facility where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S8),
        NewStatus=Status+2,
        SNewStatus=toString(NewStatus),
        S9=string::concatList(["update p3bridge.p3request set request_status=",SNewStatus," where request_no=",SRequest]),
        sqlTools::executeDirect(Con,S9),
        Mess=string::concat("Removed records and tables from request ",SRequest,"!"),
        addLog(Mess),
        fail
;       
    !.

cleanTables(Con,SRequest) :-
    Sql1=string::concat("select table_name from p3bridge.p3response_table where request_no=",SRequest),
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Table=Select:getS(Nc,1), 
        Sql2=string::concatList(["drop table if exists p3bridge.",Table]),
        sqlTools::executeDirect(Con,Sql2),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% processNext
% ----------------------------------------------------------------------------------------------
% called when server connects and when timer expires - selects next request for processing
% always succeeds
%
processNext() :-
    stream=1,
    S=string::concat("select request_no,response_format,query_type,query_text,user_name from p3bridge.p3request where request_no=",
                             "(select min(request_no) from p3bridge.p3request where (kb='",kbName,"' or query_type=0) and request_status=1)"),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    Sql:getNRows()>0,
    % remove leftover-states from previous request
    globalVars::set_i("quit_flag",0),
    removeStartedRequests,
    ir::setBreak(0),
                    datetime::tsNow(Start),
    checkStreamChange,
    execRequest(Sql),
                    datetime::tsNow(End),
                    datetime::tsDiff(Start,End,T1,T2,T3,T4),
                    datetime::timDurText(T1,T2,T3,T4,Diff),
                    Mess=string::concat("Elapsed time:",Diff),
                    addLog(Mess),
    nRequests:=nRequests+1,
                guix::txtSet(thisWin,idc_waiting,""),
                guix::txtSetI(thisWin,idc_count,nRequests),
    !
;	
    stream=-1,
    subs=1,
    S=string::concat("select request_no,response_format,query_type,query_text,user_name from p3bridge.p3request where request_no=",
                             "(select min(request_no) from p3bridge.p3request where (kb='",kbName,"' or query_type=0) and request_status=1)"),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    Sql:getNRows()>0,
    % remove leftover-states from previous request
    globalVars::set_i("quit_flag",0),
    removeStartedRequests,
    ir::setBreak(0),
                    datetime::tsNow(Start),
    subscribe,
    execRequest(Sql),
                    datetime::tsNow(End),
                    datetime::tsDiff(Start,End,T1,T2,T3,T4),
                    datetime::timDurText(T1,T2,T3,T4,Diff),
                    Mess=string::concat("Elapsed time:",Diff),
                    addLog(Mess),
    nRequests:=nRequests+1,
                guix::txtSet(thisWin,idc_waiting,""),
                guix::txtSetI(thisWin,idc_count,nRequests),
    !
;	
    stream=-1,
    subs=-1,
    S=string::concat("select request_no,response_format,query_type,query_text,user_name from p3bridge.p3request where request_no=",
                             "(select min(request_no) from p3bridge.p3request where (kb='",kbName,"' or query_type=0) and request_status=1)"),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    Sql:getNRows()>0,
    % remove leftover-states from previous request
    globalVars::set_i("quit_flag",0),
    removeStartedRequests,
    ir::setBreak(0),
                    datetime::tsNow(Start),
    execRequest(Sql),
                    datetime::tsNow(End),
                    datetime::tsDiff(Start,End,T1,T2,T3,T4),
                    datetime::timDurText(T1,T2,T3,T4,Diff),
                    Mess=string::concat("Elapsed time:",Diff),
                    addLog(Mess),
    nRequests:=nRequests+1,
                guix::txtSet(thisWin,idc_waiting,""),
                guix::txtSetI(thisWin,idc_count,nRequests),
    !
;
    % no waiting requests
    waitTimer<>-1,
    vpi::timerKill(waitTimer),
                serverStatus,
                guix::txtSet(thisWin,idc_waiting,"NO REQUESTS IN QUEUE"),
                cleanDb,
    waitTimer:=vpi::timerSet(thisWin,100),
    %waitTimer:=vpi::timerSet(thisWin,1000),
    !
;
    % no waiting requests
                serverStatus,
                guix::txtSet(thisWin,idc_waiting,"NO REQUESTS IN QUEUE"),
                cleanDb,
    waitTimer:=vpi::timerSet(thisWin,1000),
    !.
% ----------------------------------------------------------------------------------------------
% execRequest
% ----------------------------------------------------------------------------------------------   
predicates
checkRequest : (integer,integer,string,integer) procedure (i,i,i,o).
invalidQueryType : (integer) determ.
clauses

% executes given request
% always succeeds
%
execRequest(Sql) :-    
    Next=Sql:getI(1,1),
    QType=Sql:getI(1,3),
    QType=0, 
    % REDERIVATION
    globalVars::set_i("request_no",Next),
    SNext=toString(Next),
                Mess=string::concat("PROCESSING REDERIVATION REQUEST ",SNext),
                guix::txtSet(thisWin,idc_waiting,Mess),
    xmlGeneric::timeNowXML(T),
                SType=toString(QType),
                Mess1=string::concat("#",SNext," (",SType,")"),
                addLog(Mess1),
                Mess2=string::concat("Started:",T,"."),
                addLog(Mess2),
    checkRequest(0,QType,"",Error),
    processRequest(SNext,QType,"",Error),
    !
;
    Next=Sql:getI(1,1),
    RespFormat=Sql:getI(1,2),
    QType=Sql:getI(1,3),
    QType>50, 
    QType<60,
    % DB QUERY
    QueryType=QType-50,
    Query2=Sql:getS(1,4),
    Query3=string::trim(Query2),
    sx::sxGiveTail(Query3,"?",Query),
    globalVars::set_i("request_no",Next),
    SNext=toString(Next),
                Mess=string::concat("PROCESSING DB QUERY REQUEST ",SNext),
                guix::txtSet(thisWin,idc_waiting,Mess),
    xmlGeneric::timeNowXML(T),
                SType=toString(QueryType),
                Mess1=string::concat("#",SNext," (",SType,") ",Query),
                addLog(Mess1),
                Mess2=string::concat("Started:",T,"."),
                addLog(Mess2),
    checkRequest(RespFormat,QueryType,Query,Error),
    RespM=-RespFormat, % use negative value since server-mode works more similar to 'output to one file'
    globalVars::set_i("f_outq",RespM),
    processRequest(SNext,QueryType,Query,Error),
    !
;
    Next=Sql:getI(1,1),
    RespFormat=Sql:getI(1,2),
    QType=Sql:getI(1,3),
    QType>60, 
    QType<70,
    % CLUSTER QUERY
    QueryType=QType-60,
    Query2=Sql:getS(1,4),
    Query=string::trim(Query2),
    globalVars::set_i("request_no",Next),
    SNext=toString(Next),
                Mess=string::concat("PROCESSING CLUSTER REQUEST ",SNext),
                guix::txtSet(thisWin,idc_waiting,Mess),
    xmlGeneric::timeNowXML(T),    
                Mess1=string::concat("#",SNext," (51) ",Query,"."),
                addLog(Mess1),
                Mess2=string::concat("Started:",T,"."),
                addLog(Mess2),
    RespM=-RespFormat, % use negative value since server-mode works more similar to 'output to one file'
    globalVars::set_i("f_outq",RespM),
    sx::sxUnpack(Query,'?',QueriesB),
    sx::sxlRemBlanks(QueriesB,Queries),
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S2=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=2 where request_no=",SNext]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S2),
    globalVars::set_i("request_sols",0),
    processCluster(RespFormat,QueryType,1,SNext,Queries),
    !
;
    Next=Sql:getI(1,1),
    QueryType=Sql:getI(1,3),
    QueryType>100, 
    % REPORT
    globalVars::set_i("request_no",Next),
    SNext=toString(Next),
                Mess=string::concat("PROCESSING REQUEST ",SNext),
                guix::txtSet(thisWin,idc_waiting,Mess),
    xmlGeneric::timeNowXML(T),    
                SType=toString(QueryType),
                Mess1=string::concat("#",SNext," (",SType,")."),
                addLog(Mess1),
                Mess2=string::concat("Started:",T,"."),
                addLog(Mess2),
    processReport(SNext,QueryType),
    !
;    
    Next=Sql:getI(1,1),
    RespFormat=Sql:getI(1,2),
    QueryType=Sql:getI(1,3),
    QueryType>0,
    QueryType<10,
    % DIRECT QUERY
    Query2=Sql:getS(1,4),
    Query3=string::trim(Query2),
    sx::sxGiveTail(Query3,"?",Query),
    globalVars::set_i("request_no",Next),
    SNext=toString(Next),
                Mess=string::concat("PROCESSING REQUEST ",SNext),
                guix::txtSet(thisWin,idc_waiting,Mess),
    xmlGeneric::timeNowXML(T),
                SType=toString(QueryType),
                Mess1=string::concat("#",SNext," (",SType,") ",Query),
                addLog(Mess1),
                Mess2=string::concat("Started:",T,"."),
                addLog(Mess2),
    checkRequest(RespFormat,QueryType,Query,Error),
    RespM=-RespFormat, % use negative value since server-mode works more similar to 'output to one file'
    globalVars::set_i("f_outq",RespM),
    processRequest(SNext,QueryType,Query,Error),
    !
;    
    QueryType=Sql:getI(1,3),
    invalidQueryType(QueryType),
    Next=Sql:getI(1,1),
    SNext=toString(Next),
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_end=",QNow,",request_status=-13 where request_no=",SNext]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
                    Mess1=string::concat("Invalid request type! Ended:",QNow,"."),
                    addLog(Mess1),
    !.
    
invalidQueryType(Q) :- Q<=0,!;
                                  Q>=10, Q<=50,!;
                                  Q=60,!;
                                  Q>=70, Q<=100,!.
                                  
checkRequest(RespFormat,QueryType,Query,Error) :-
    QueryType=0,
    subs=1,
    stream=-1,
    Error=0,
    !
;
    QueryType=0,
    Error=-200,
    !
;
    not(set::member(RespFormat,[1,8,9,10,11])),
    Error=-2,
    !
;
    not(set::member(QueryType,[1,9])),
    Error=-3,
    !
;
    Query="?", % blank query
    Error=0,
    !
;    
	transnl::parSynTree(Query,SynTr),
    SynTr=termDomain::cmp("ABOUT",_), % all about query
    Error=0,
    !
;    
    QueryType=1,
    not(transnl::transQuery(Query,_,_)),
    Error=-4,
    !
;
    QueryType=9,
    sx::sxRemoveTail(Query,"?",QueryT),
	not(transpr::proPR_Term(QueryT,_)),
	Error=-4,
    !
;
    Error=0,
    !.
% ----------------------------------------------------------------------------------------------
% processReport
% ----------------------------------------------------------------------------------------------
predicates
procReport : (integer,string,integer) determ (i,i,o).
sqlList : (integer,string_list,string,integer) determ (i,i,i,o).
clauses

processReport(SNo,Report) :-
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=2 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    procReport(Report,SNo,N),
    Ns=toString(N),
    xmlGeneric::timeNowXML(Now2),
    QNow2=sqlTools::xmlToDatetime(Now2),
    S2=string::concatList(["update p3bridge.p3request set request_end=",QNow2,",request_status=3,no_of_solutions=",Ns," where request_no=",SNo]),
    sqlTools::executeDirect(Con,S2),
                    Mess1=string::concat("Completed report at ",QNow,"."),
                    addLog(Mess1),
    !
;
    S=string::concat("select request_status from p3bridge.p3request where request_no=",SNo),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    Sql:getNRows()=1,
    Status=Sql:getI(1,1),
    Status<0,
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S2=string::concatList(["update p3bridge.p3request set request_end=",QNow," where request_no=",SNo]),
    sqlTools::executeDirect(Con,S2),
                    SStatus=toString(Status),
                    Mess1=string::concat("Aborted by error ",SStatus," at ",QNow,"."),
                    addLog(Mess1),
    !
;
    S=string::concat("select request_status from p3bridge.p3request where request_no=",SNo),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    Sql:getNRows()=0,
                    xmlGeneric::timeNowXML(Now),
                    QNow=sqlTools::xmlToDatetime(Now),
                    Mess1=string::concat("*** CANNOT FIND REQUEST RECORD #",SNo," at ",QNow," ***"),
                    addLog(Mess1),
    !
;    
    globalVars::get_i("quit_flag")=1,
    globalVars::set_i("quit_flag",0),
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_end=",QNow,",request_status=-10 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
                    Mess1=string::concat("Aborted by system (quit) at ",QNow,"."),
                    addLog(Mess1),
    !
;
    % failure
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=-10 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
                    Mess1=string::concat("Aborted by system error at ",QNow,"."),
                    addLog(Mess1),
    !.
        
procReport(RNo,SNo,N) :-
    set::member(RNo,[101,102]),
    Form=RNo-100,
    p3Rep::repToken(Form),
    P=globalVars::get_s("std_path"),
    File=string::concat(P,"\\",kbName,"_Syntagms.3rp"),
    L=fio::readFile(File),
    sqlList(1,L,SNo,N),
    !.
procReport(110,SNo,N) :-
    p3Rep::repSent(-1,-1,-1,-1,-1,-1,-1),
    P=globalVars::get_s("std_path"),
    File=string::concat(P,"\\",kbName,"_Sentences.3rp"),
    L=fio::readFile(File),
    sqlList(1,L,SNo,N),
    !.
procReport(111,SNo,N) :-
    p3Rep::repSent(-1,-1,-1,1,-1,1,-1),
    P=globalVars::get_s("std_path"),
    File=string::concat(P,"\\",kbName,"_Sentences.3rp"),
    L=fio::readFile(File),
    sqlList(1,L,SNo,N),
    !.
procReport(120,SNo,N) :-
    irReport::reportRules,
    P=globalVars::get_s("std_path"),
    File=string::concat(P,"\\",kbName,"_InexactRules.3rp"),
    L=fio::readFile(File),
    sqlList(1,L,SNo,N),
    !.

sqlList(N,[],_,N) :-
    !.
sqlList(N,[H|T],SNo,M) :-
    Ns=toString(N),
    S=string::concatList(["insert p3bridge.p3response (request_no,sub_no,line_no,line) values (",SNo,",1,",Ns,",'",H,"')"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),   
    Next=N+1,
    sqlList(Next,T,SNo,M),
    !.
% ----------------------------------------------------------------------------------------------
% processRequest
% ----------------------------------------------------------------------------------------------
% parameters:
% 1 in request#
% 2 in request type
% 3 in query
% 4 in result of pre-run request validation
% always succeeds unless special system error
%
processRequest(SNo,QueryType,Query,Error) :-
    Error<0,
    SError=toString(Error),
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=",SError," where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
                    Mess1=string::concat("Error ",SError," in request at ",QNow,"."),
                    addLog(Mess1),
    !
;
    globalVars::get_i("quit_flag")=1, 
    globalVars::set_i("quit_flag",0), 
    % system error
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=-10 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
                    Mess1=string::concat("Error in request (quit) at ",QNow,"."),
                    addLog(Mess1),
    !
;
    Error=0,
    QueryType=0, % rederivation
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=2 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    globalVars::get_i("quit_flag")=0,
    xmlGeneric::timeNowXML(Now2b),
    QNow2=sqlTools::xmlToDatetime(Now2b),
    S2=string::concatList(["update p3bridge.p3request set request_end=",QNow2,",request_status=6 where request_no=",SNo]),
    sqlTools::executeDirect(Con,S2),
                    Mess1=string::concatList(["Completed at ",QNow2,"."]),
                    addLog(Mess1),
    !
;
    Error=0,
    QueryType=1, % NL-query
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=2 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    globalVars::set_i("request_subno",1),
    globalVars::set_i("request_sols",0),
    proc::runQueryNL(Con,Query),
    globalVars::get_i("quit_flag")=0,
    xmlGeneric::timeNowXML(Now2b),
    QNow2=sqlTools::xmlToDatetime(Now2b),
    Sols=globalVars::get_i("request_sols"),
    SSols=toString(Sols),
    S2=string::concatList(["update p3bridge.p3request set request_end=",QNow2,",request_status=3,no_of_solutions=",SSols," where request_no=",SNo]),
    sqlTools::executeDirect(Con,S2),
                    Mess1=string::concatList(["Completed at ",QNow2," with ",SSols," solutions."]),
                    addLog(Mess1),
    !
;    
    Error=0,
    QueryType=9, % PR-query
    sx::sxRemoveTail(Query,"?",Query2),
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=2 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    globalVars::set_i("request_subno",1),
    globalVars::set_i("request_sols",0),
    proc::runQueryPR(Con,Query2),
    globalVars::get_i("quit_flag")=0,
    xmlGeneric::timeNowXML(Now2b),
    QNow2=sqlTools::xmlToDatetime(Now2b),
    Sols=globalVars::get_i("request_sols"),
    SSols=toString(Sols),
    S2=string::concatList(["update p3bridge.p3request set request_end=",QNow2,",request_status=3,no_of_solutions=",SSols," where request_no=",SNo]),
    sqlTools::executeDirect(Con,S2),
                    Mess1=string::concatList(["Completed at ",QNow," with ",SSols," solutions."]),
                    addLog(Mess1),
    !
;
    S=string::concat("select request_status from p3bridge.p3request where request_no=",SNo),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    Sql:getNRows()=1,
    Status=Sql:getI(1,1),
    Status<0,
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    S2=string::concatList(["update p3bridge.p3request set request_end=",QNow," where request_no=",SNo]),
    sqlTools::executeDirect(Con,S2),
                    SStatus=toString(Status),
                    Mess1=string::concat("Aborted by error ",SStatus," at ",QNow,"."),
                    addLog(Mess1),
    !
;
    S=string::concat("select request_status from p3bridge.p3request where request_no=",SNo),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    Sql:getNRows()=0,
                    xmlGeneric::timeNowXML(Now),
                    QNow=sqlTools::xmlToDatetime(Now),
                    Mess1=string::concat("*** CANNOT FIND REQUEST RECORD #",SNo," at ",QNow," ***"),
                    addLog(Mess1),
    !
;    
    globalVars::get_i("quit_flag")=1,
    globalVars::set_i("quit_flag",0),
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    Sols=globalVars::get_i("request_sols"),
    SSols=toString(Sols),
    S=string::concatList(["update p3bridge.p3request set request_end=",QNow,",request_status=-10,no_of_solutions=",SSols," where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
                    Mess1=string::concat("Aborted by system (quit) at ",QNow,"."),
                    addLog(Mess1),
    !
;
    % failure
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    Sols=globalVars::get_i("request_sols"),
    SSols=toString(Sols),
    S=string::concatList(["update p3bridge.p3request set request_start=",QNow,",request_status=-10,no_of_solutions=",SSols," where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
                    Mess1=string::concat("Aborted by system error at ",QNow,"."),
                    addLog(Mess1),
    !.
% ----------------------------------------------------------------------------------------------
% setServerStatus
% ----------------------------------------------------------------------------------------------
predicates
replaceServer : (integer,string) determ.
clauses

setServerStatus(Status,Stream) :-
    replaceServer(Status,Stream),
    !.
 
setServerStatus(Status) :- 
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select knowledge_stream from p3bridge.p3server where kb='",kbName,"'"),
    Select=sqlSelect::new(Con,Sql1),
    Select:getNRows()>0,
    Stream=Select:getS(1,1), 
    replaceServer(Status,Stream),
    !
;
    replaceServer(Status,""),
    !.

replaceServer(Status,Stream) :-
    Sql2=string::concat("delete from p3bridge.p3server where kb='",kbName,"'"),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,Sql2),
    SStatus=toString(Status),
    xmlGeneric::timeNowXML(Now),
    QNow=sqlTools::xmlToDatetime(Now),
    Sql3=string::concatList(["insert p3bridge.p3server (status,status_time,kb,knowledge_stream) values (",SStatus,",",QNow,",'",kbName,"','",Stream,"')"]),
    sqlTools::executeDirect(Con,Sql3),
    !.
% ----------------------------------------------------------------------------------------------
% addLog
% ----------------------------------------------------------------------------------------------
addLog(S) :-
    guix::lbxGetAll(thisWin,idc_processed,L),
    L2=[S|L],
    set::trunc(50,L2,L3),
    guix::lbxPop(thisWin,idc_processed,L3),
    !.
% ----------------------------------------------------------------------------------------------
% processSubRequest
% ----------------------------------------------------------------------------------------------
processSubRequest(Sub,QueryType,Query) :-
    QueryType=1,
    NSub=sx::toInt(Sub),
    globalVars::set_i("request_subno",NSub),
    Con=sqlBase::getConnection(),
    proc::runQueryNL(Con,Query),
                    Mess1=string::concat("Completed ",Sub,". sub-request."),
                    addLog(Mess1),
    !
;
    QueryType=9,
    sx::sxRemoveTail(Query,"?",Query2),
    NSub=sx::toInt(Sub),
    globalVars::set_i("request_subno",NSub),
    Con=sqlBase::getConnection(),
    proc::runQueryPR(Con,Query2),
                    Mess1=string::concat("Completed ",Sub,". sub-request."),
                    addLog(Mess1),
    !
;    
                    Mess1=string::concat("Processing of ",Sub,". sub-request failed!"),
                    addLog(Mess1),
    !.
% ----------------------------------------------------------------------------------------------
% dropRespTables
% ----------------------------------------------------------------------------------------------
predicates
dropRespX : (sqlSelect,integer) determ.
clauses

dropRespTables("") :-
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,"select table_name from p3bridge.p3response_table"),
    dropRespX(Sql,1),
    !
;
    !.
dropRespTables(KB) :-
    S=string::concatList(["select table_name from p3bridge.p3response_table where request_no in (select request_no from p3bridge.p3request where kb='",KB,"')"]),
    Con=sqlBase::getConnection(),
    Sql=sqlSelect::new(Con,S),
    dropRespX(Sql,1),
    !
;
    !.
    
dropRespX(Sql,Row) :-    
    Table=Sql:getS(Row,1),
    S=string::concat("drop table ",Table),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    Next=Row+1,
    dropRespX(Sql,Next),
    !.
% ----------------------------------------------------------------------------------------------
% exportQueries
% ----------------------------------------------------------------------------------------------
predicates
expAlts : (integer,string_list,string,string) determ.    
expParameters : (string,string_list) determ.
expQueries : (string,string_list,string_list,string_list) determ.    
expCluster : (string,string,string_list,string_list,string_list) determ.    
clusterQuery : (string,string,string,string_list,string_list) determ.
getMenuQualifier : (string,string_list,string_list,string) determ (i,i,i,o).  
cluster : (string,string_list) determ (i,o).
insertQueryPars : (string,string,string).
quoted : (string,string) procedure (i,o).
clauses

exportQueries() :-
    Lang=globalVars::get_s("lang"),
    Qs=pop::getQueries(Lang),
    Ms=pop::getMenus(Lang),
    Ps=pop::getParameters(Lang),
    S1=string::concat("delete from p3bridge.p3query_parameter where parameter_name in (select parameter_name from p3bridge.p3parameter where user_name='Pro/3' and kb='",kbName,"')"),
    S2=string::concat("delete from p3bridge.p3parameter where user_name='Pro/3' and kb='",kbName,"'"),
    S3=string::concat("delete from p3bridge.p3alternative where user_name='Pro/3' and kb='",kbName,"'"),
    S4=string::concat("delete from p3bridge.p3query where user_name='Pro/3' and kb='",kbName,"'"),
    S5=string::concat("delete from p3bridge.p3cluster where user_name='Pro/3' and kb='",kbName,"'"),
    S6=string::concat("delete from p3bridge.p3query_cluster where user_name='Pro/3' and kb='",kbName,"'"),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S1),
    sqlTools::executeDirect(Con,S2),
    sqlTools::executeDirect(Con,S3),
    sqlTools::executeDirect(Con,S4),
    sqlTools::executeDirect(Con,S5),
    sqlTools::executeDirect(Con,S6),
    expParameters(kbName,Ps),
    expQueries(kbName,Qs,Qs,Ms),
    !.

expCluster(_,_,[],_,_) :-
    !.
expCluster(Cluster,KB,[Q|Qs],Queries,Menus) :-
    clusterQuery(Cluster,KB,Q,Queries,Menus),
    expCluster(Cluster,KB,Qs,Queries,Menus),
    !.

clusterQuery(Cluster,_,Id,[],_) :-
    Mess=string::concat("Cluster: ",Cluster," Query: ",Id),
    vpiCommonDialogs::note("CANNOT FIND CLUSTERED KB-QUERY",Mess),
    !. 
clusterQuery(Cluster,KB,Id,[Q|Queries],Menus) :-
    sx::sxUnpackS(Q,"||",L),
    L=[Qid,_,Name,_,_,Menu],
    Id=Qid,
    getMenuQualifier(Menu,Menus,Menus,Qual),
    QName=string::concat(Qual," - ",Name),
    TName=string::trim(QName),
    S=string::concatList(["insert p3bridge.p3query_cluster (query_name,user_name,cluster_name,kb) values ('",TName,"','Pro/3','",Cluster,"','",KB,"')"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    !
;    
    clusterQuery(Cluster,KB,Id,Queries,Menus),
    !.

expQueries(_,[],_,_) :-
    !.
expQueries(KB,[H|T],AllQs,Menus) :-
    sx::sxUnpackS(H,"||",L),
    L=[Id,_,Name,Format,Query,Menu],
    sx::frontstr(2,Id,F,_),
    IdN=sx::toInt(F),
    IdN>=75,
    cluster(Query,Queries),
    getMenuQualifier(Menu,Menus,Menus,Qual),
    QName=string::concat(Qual," - ",Name),
    TName=string::trim(QName),
    S=string::concatList(["insert p3bridge.p3cluster (cluster_name,user_name,kb,query_format) values ('",TName,"','Pro/3','",KB,"','",Format,"')"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    expCluster(TName,KB,Queries,AllQs,Menus),
    expQueries(KB,T,AllQs,Menus),
    !
;
    sx::sxUnpackS(H,"||",L),
    L=[Id,_,Name,Format,Query,Menu],
    sx::frontstr(2,Id,F,_),
    IdN=sx::toInt(F),
    IdN>=75,
    not(cluster(Query,_)),
    getMenuQualifier(Menu,Menus,Menus,Qual),
    QName=string::concat(Qual," - ",Name),
    TName=string::trim(QName),
    sx::sxGiveTail(Query,"?",Query2),
    Query3=sx::replaceAllC(Query2,'`','"'),
    S=string::concatList(["insert p3bridge.p3query (query_name,query_text,user_name,kb,query_format) values ('",
                                                            TName,"','",Query3,"','Pro/3','",KB,"','",Format,"')"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    insertQueryPars(KB,TName,Query3),
    expQueries(KB,T,AllQs,Menus),
    !
;
    expQueries(KB,T,AllQs,Menus),
    !.
    
expParameters(_,[]) :-
    !.
expParameters(KB,[H|T]) :-
    sx::sxUnpackS(H,"||",L),
    L=[Name,_,Dom,Title,Alts],
    Alts<>"",
    sx::sxUnpackS(Alts,";;",As),
    quoted(Dom,Quoted),
    S=string::concatList(["insert p3bridge.p3parameter (parameter_name,kb,user_name,description,domain,quoted) values ('",
                                                                   Name,"','",KB,"','Pro/3','",Title,"','",Dom,"',",Quoted,")"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    expAlts(1,As,Name,KB),
    expParameters(KB,T),
    !
;
    sx::sxUnpackS(H,"||",L),
    L=[Name,_,Dom,Title],
    quoted(Dom,Quoted),
    S=string::concatList(["insert p3bridge.p3parameter (parameter_name,kb,user_name,description,domain,quoted) values ('",
                                                                   Name,"','",KB,"','Pro/3','",Title,"','",Dom,"',",Quoted,")"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    expParameters(KB,T),
    !.

quoted(Dom,Quoted) :-
    dbBase::kbID(Dom,D,_,_),
    dom::domStr(D),
    Quoted="1",
    !
;
    Quoted="0",
    !.

expAlts(_,[],_,_) :-
    !.
expAlts(N,[A|As],Name,KB) :-
    Ns=toString(N),
    S=string::concatList(["insert p3bridge.p3alternative (parameter_name,kb,user_name,alternative,alternative_no) values ('",
                                                                   Name,"','",KB,"','Pro/3','",A,"',",Ns,")"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    Next=N+1,
    expAlts(Next,As,Name,KB),
    !.

getMenuQualifier("7000",_,_,"") :-
    !. % root menu
getMenuQualifier(Menu,[],_,"") :-
    Mess=string::concat("Menu: ",Menu),
    vpiCommonDialogs::note("CANNOT FIND QUERY-MENU",Mess),
    !.
getMenuQualifier(Menu,[M|Menus],AllMenus,Qual) :-
    sx::sxUnpackS(M,"||",L),
    L=[MenuX,_,Name,MenuNext],
    MenuX=Menu,
    getMenuQualifier(MenuNext,AllMenus,AllMenus,Q2),
    Qual=string::concat(Q2," ",Name),
    !
;    
    getMenuQualifier(Menu,Menus,AllMenus,Qual),
    !.    
    
cluster(S,Qs) :-
    S="",
    !,fail
;
    sx::searchstring(S,";",_),
    sx::sxUnpack(S,';',Qs),
    !.

insertQueryPars(KB,QName,Q) :-
    sx::sxGetInBrack(Q,'[',']',Par),
    P=string::concat("[",Par,"]"),
    Q2=sx::replaceAll(Q,P,""),
    S=string::concatList(["insert p3bridge.p3query_parameter (parameter_name,kb,query_name,user_name) values ('",Par,"','",KB,"','",QName,"','Pro/3')"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S),
    insertQueryPars(KB,QName,Q2),
    !
;
    !.
% ----------------------------------------------------------------------------------------------
% processCluster
% ----------------------------------------------------------------------------------------------
processCluster(_,_,_,SNo,[]) :-
    xmlGeneric::timeNowXML(Now2b),
    QNow2=sqlTools::xmlToDatetime(Now2b),
    Sols=globalVars::get_i("request_sols"),
    SSols=toString(Sols),
    S3=string::concatList(["update p3bridge.p3request set request_end=",QNow2,",request_status=3,no_of_solutions=",SSols," where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S3),
                    Mess1=string::concat("Completed cluster:",Now2b," with ",SSols," solution(s)."),
                    addLog(Mess1),
    !.
processCluster(RespFormat,Format,N,SNo,[Query|Queries]) :-
    Sub=toString(N),
    sx::sxGiveTail(Query,"?",Q),
    checkRequest(RespFormat,Format,Q,Error),
    Error=0,
    processSubRequest(Sub,Format,Q),
    Next=N+1,
    processCluster(RespFormat,Format,Next,SNo,Queries),
    !
;    
    % error occurred stop processing
    xmlGeneric::timeNowXML(Now2b),
    QNow2=sqlTools::xmlToDatetime(Now2b),
    S3=string::concatList(["update p3bridge.p3request set request_end=",QNow2,",request_status=-10,no_of_solutions=0 where request_no=",SNo]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,S3),
                    Mess1=string::concat("Aborted cluster due to error:",QNow2,"."),
                    addLog(Mess1),
    !.
% ----------------------------------------------------------------------------------------------
% serverStatus
% ----------------------------------------------------------------------------------------------
serverStatus() :-
    C1=globalVars::get_i("solref"),
    C2=globalVars::get_i("solint"),
    C1s=toString(C1),
    C2s=toString(C2),
	tt::freeMemory(Af,Bf,Cf,Df,_,_,_,_),
	As=toString(Af),
	Bs=toString(Bf),
	Cs=toString(Cf),
	Ds=toString(Df),
	sx::fixFormatS(As,1,0,Asf),
	sx::fixFormatS(Bs,1,0,Bsf),
	sx::fixFormatS(Cs,1,0,Csf),
	sx::fixFormatS(Ds,1,0,Dsf),
	dlgIF::dbaStatusesX("knowopen",2,"+",S1,_),
	dlgIF::dbaStatusesX("derflag",0,"!",S2,_),
	dlgIF::dbaStatusesX("quit_flag",1,"Q",S3,_),
	dlgIF::dbaStatusesX("error_flag",1,"E",S4,_),
	Ix=globalVars::get_i("all_del")+2,
	set::setGet(["S","","D"],Ix,S5),
	S=string::concatList(["#",C1s,"(",C2s,") ",S1,S2,S3,S4,"A",S5," ",Asf,"% ",Bsf,"% ",Csf,"% ",Dsf,"%"]),
    guix::txtSet(thisWin,idc_rstatus,S),
    xmlGeneric::timeNowXML(T),
    guix::txtSet(thisWin,idc_currTime,T),
	!
;
    !.
% ----------------------------------------------------------------------------------------------
% fatalError
% ----------------------------------------------------------------------------------------------
fatalError(Message) :-
    globalVars::set_i("server",-1),
    Mess=string::concat("Fatal server error: ",Message),
    dlgIF::dlgNote(2,Mess),
    globalVars::set_i("server",1),
    serverError:=1,
    !.
% ----------------------------------------------------------------------------------------------
% fatalServerError
% ----------------------------------------------------------------------------------------------
fatalServerError(Message) :-
    globalVars::set_i("server",-1),
    Mess=string::concat("Fatal server error: ",Message),
    dlgIF::dlgNote(2,Mess),
    globalVars::set_i("server",1),
    serverError:=1,
    !.
% ----------------------------------------------------------------------------------------------
% checkStream/Change
% ----------------------------------------------------------------------------------------------
checkStream() :-
    guix::txtGet(thisWin,idc_stream,Stream),
    Stream<>"",
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select kb from ",Stream," where processed=0"),
    Select1=sqlSelect::new(Con,Sql1),
    N1=Select1:getNRows(),
    N1>0,    
    stream:=1,
    !
;
    stream:=-1,
   !.
   
/* @@optimizations:
1. should only (re)derive if anything was read from the stream'
2. should do selective derivation instead of rederivation
*/
checkStreamChange() :-
    stream=1,
    guix::txtGet(thisWin,idc_stream,Stream),
    Stream<>"",
    cluster::readStream(Stream),
	der::rederive,  
    !
;
   !.
% ----------------------------------------------------------------------------------------------
% setStream
% ----------------------------------------------------------------------------------------------
setStream() :-        
        Con=sqlBase::getConnection(),
        Sql1=string::concat("select knowledge_stream from p3bridge.p3server where kb='",kbName,"'"),
        Select=sqlSelect::new(Con,Sql1),
        Select:getNRows()>0,
        Stream=Select:getS(1,1),
        guix::txtSet(thisWin,idc_stream,Stream),
        !
;
        guix::txtSet(thisWin,idc_stream,""),
        !.
% ----------------------------------------------------------------------------------------------
% optimizeSubs
% ----------------------------------------------------------------------------------------------
optimizeSubs() :-
    dbBase::chainDelete("r"),
			dlgIF::msgStatus("Deleting all derived sentences..."),
	dbBase::sentenceDerivedDelete(-1),
			dlgIF::msgStatus("Subscribing sentences..."),
	sqlSt::subscribeAll,
	dep::regenDependencyGraph,
	der::depList(-1,L,_,_,_),
			dlgIF::msgStatus(""),
    stList:=L,
    subs:=1,
    !.
% ----------------------------------------------------------------------------------------------
% subscribe
% ----------------------------------------------------------------------------------------------
subscribe() :-
/* 11.11.2006 optimization
	sqlSt::subscribeAll,
			Rep=rep::newReport(5,"",0),
	der::depDeleteSTs(Rep,stList),
	der::depDeriveSTs(Rep,stList,_,_),
			Rep:close(),
*/			
	sqlSt::subscribeAll,
	der::depDeleteSTs(stList),
	der::depDeriveSTs(stList),
    !.
end implement servermode
