/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement suspended
open core,vpiDomains,resourceIdentifiers,globalDomains,termDomain

constants
className = "com/rudstrand/TaskWindow/suspended".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

facts
thisWin : vpiDomains::windowHandle := erroneous.

clauses
new(Parent):-
    _ = vpi::winCreate(windowType, rectangle, title, menu, Parent,
                        windowFlags, eventHandler, 0).

class predicates
dbuUnsuspends : (string_list) determ.
dbuDeleteSuspends : (string_list) determ.
recListFormat : (dbLoc_list,string_list) determ (i,o).

predicates
eventHandler : vpiDomains::ehandler.

clauses
eventHandler(Win, Event) = Result :-
    Result = generatedEventHandler(Win, Event),
    !.
eventHandler(Win, e_Menu(Id, CAS)) = Result :-
    Parent = vpi::winGetParent(Win),
    Result = vpi::winSendEvent(Parent, e_Menu(Id, CAS)).

predicates
    onCreate : vpiDomains::createHandler.
    clauses
        onCreate(_CreationData) = defaultHandling():-
	vpi::cursorSetWait(),
   	guix::setStyle(thisWin,[idc_suspended_records,idc_remove_selected,idc_remove_all,idc_view_record,idc_del_sel,idc_del_all],[]),
	guix::addWin(33,thisWin),
	dbBase::recTypeLocs(30,L2),
	dbBase::recTypeLocs(40,L3),
	dbBase::recTypeLocs(50,L4),
	dbBase::recTypeLocs(60,L5),
	dbBase::recTypeLocs(-30,L6),
	dbBase::recTypeLocs(-40,L7),
	dbBase::recTypeLocs(-50,L8),
	dbBase::recTypeLocs(-60,L9),
	dbBase::recTypeLocs(20,L10),
	set::conc([L2,L3,L4,L5,L6,L7,L8,L9,L10],List2),
	recListFormat(List2,List),
	guix::lbxPop(thisWin,idc_suspended_records,List),
	vpi::cursorSet(thisWin,cursor_Arrow),
	!;!.

    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            guix::dropWin(thisWin),
            thisWin := erroneous.

    predicates
        onSizeChanged : vpiDomains::sizeHandler.
    clauses
        onSizeChanged(_Width, _Height) = defaultHandling():-
            vpiToolbar::resize(thisWin).

    predicates
        onCloseRequest : vpiDomains::closeRequestHandler.
    clauses
        onCloseRequest() = defaultHandling():-
        guix::dropWin(thisWin),
        vpi::winDestroy(thisWin).

    predicates
        onControlDelSel : vpiDomains::controlHandler.
    clauses
        onControlDelSel(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	vpi::cursorSetWait(),
  	C=vpi::winGetCtlHandle(thisWin,idc_suspended_records),
  	vpi::lboxGetSel(C,Sels,L),
  	dbuDeleteSuspends(Sels),
  	guix::lbxDelSet(thisWin,idc_suspended_records,L),
	vpi::cursorSet(thisWin,cursor_Arrow),
	!;!.

    predicates
        onControlDelAll : vpiDomains::controlHandler.
    clauses
        onControlDelAll(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	vpi::cursorSetWait(),
  	C=vpi::winGetCtlHandle(thisWin,idc_suspended_records),
  	Sels=vpi::lboxGetAll(C),
  	dbuDeleteSuspends(Sels),
  	vpi::lboxClear(C),
	vpi::cursorSet(thisWin,cursor_Arrow),
	!;!.

    predicates
        onControlViewRecord : vpiDomains::controlHandler.
    clauses
        onControlViewRecord(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
  	C=vpi::winGetCtlHandle(thisWin,idc_suspended_records),
  	vpi::lboxGetSel(C,Sels,_),
  	Sels=[Sel|_],
	sx::sxGetInBrack(Sel,'[',']',SLoc),
	Loc=dbBase::dbRefR(SLoc),
	W=guix::getWinOfClas(1),
	RecordWindow=record::new(W),
	Wr=RecordWindow:getWin(),
	record::kbaWinDisplay(Wr,Loc),
	!;!.

    predicates
        onControlRemoveAll : vpiDomains::controlHandler.
    clauses
        onControlRemoveAll(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	vpi::cursorSetWait(),
  	C=vpi::winGetCtlHandle(thisWin,idc_suspended_records),
  	Sels=vpi::lboxGetAll(C),
  	dbuUnsuspends(Sels),
	vpi::cursorSet(thisWin,cursor_Arrow),
	vpi::lboxClear(C),
	!;!.

    predicates
        onControlRemoveSelected : vpiDomains::controlHandler.
    clauses
        onControlRemoveSelected(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
	vpi::cursorSetWait(),
  	C=vpi::winGetCtlHandle(thisWin,idc_suspended_records),
  	vpi::lboxGetSel(C,Sels,L),
  	dbuUnsuspends(Sels),
  	guix::lbxDelSet(thisWin,idc_suspended_records,L),
	vpi::cursorSet(thisWin,cursor_Arrow),
	!;!.

    % This code is maintained by the VDE do not update it manually, 15:46:29-30.1.2005
    constants
        windowType : vpiDomains::wintype = w_TopLevel.
        windowFlags : vpiDomains::wsflags = [wsf_TitleBar,wsf_Minimize,wsf_Close,wsf_ClipSiblings,wsf_ClipChildren].
        rectangle : vpiDomains::rct = rct(100,80,822,525).
        menu : vpiDomains::menu = noMenu.
        title : string = "Suspended Records".

    predicates
        generatedEventHandler : vpiDomains::ehandler.
    clauses
        generatedEventHandler(Win, e_create(_)) = _ :-
            thisWin := Win,
            _ = vpi::winCreateControl(wc_LBox,rct(8,4,716,413),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort,wsf_NoIntegralHeight,wsf_MultiSelect],idc_suspended_records),
            _ = vpi::winCreateControl(wc_PushButton,rct(100,418,263,442),"Remove All Suspensions",Win,[wsf_Group,wsf_TabStop],idc_remove_all),
            _ = vpi::winCreateControl(wc_PushButton,rct(266,418,546,442),"Remove Suspensions of Selected Records",Win,[wsf_Group,wsf_TabStop],idc_remove_selected),
            _ = vpi::winCreateControl(wc_PushButton,rct(9,418,97,442),"View Record",Win,[wsf_Group,wsf_TabStop],idc_view_record),
            _ = vpi::winCreateControl(wc_PushButton,rct(648,418,717,442),"Delete All",Win,[wsf_Group,wsf_TabStop],idc_del_all),
            _ = vpi::winCreateControl(wc_PushButton,rct(549,418,645,442),"Delete Selected",Win,[wsf_Group,wsf_TabStop],idc_del_sel),
            fail.
        generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
            handled(Result) = onCreate(CreationData).
        generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
            handled(Result) = onSizeChanged(Width, Height).
        generatedEventHandler(_Win, e_Control(idc_del_sel, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlDelSel(idc_del_sel, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_del_all, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlDelAll(idc_del_all, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_view_record, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlViewRecord(idc_view_record, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_remove_all, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlRemoveAll(idc_remove_all, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_Control(idc_remove_selected, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlRemoveSelected(idc_remove_selected, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_CloseRequest()) = Result :-
            handled(Result) = onCloseRequest().
        generatedEventHandler(_Win, e_Destroy()) = Result :-
            handled(Result) = onDestroy().
    % end of automatic code
% ----------------------------------------------------------------------------------------------
% dbuUnsuspend/s
% ----------------------------------------------------------------------------------------------
dbuUnsuspends([]) :-
	!.
dbuUnsuspends([H|T]) :-
	sx::sxGetInBrack(H,'[',']',SLoc),
	Loc=dbBase::dbRefR(SLoc),
	dbBase::recordUnsuspend(Loc),
	dbuUnsuspends(T),
	!.
% ----------------------------------------------------------------------------------------------
% dbuDeleteSuspends
% ----------------------------------------------------------------------------------------------
class predicates
dbuDeleteSuspend : (string).
clauses

dbuDeleteSuspends([]) :-
	!.
dbuDeleteSuspends([H|T]) :-
	dbuDeleteSuspend(H),
	dbuDeleteSuspends(T),
	!.
	
dbuDeleteSuspend(S) :-
	sx::sxGetInBrack(S,'[',']',SLoc),
	Loc=dbBase::dbRefR(SLoc),
	dbBase::recordDelete(Loc),
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
% recListFormat
% ----------------------------------------------------------------------------------------------
class predicates
recFormat : (dbLoc,string) determ (i,o).
clauses

recListFormat([],[]) :-
	!.
recListFormat([H|T],[Hf|Tf]) :-
    recFormat(H,Hf),
    recListFormat(T,Tf),
	!.

recFormat(Loc,S) :-
	dbBase::getRecord(Loc,RecTyp,Base,_,_,_,Realm,_,_,_,LT,_),
	Base=6,
	STyp=toString(RecTyp),
	LT=cmp(CRT,[atom(CR)|_]),
	dbBase::nlSing(CRT,S2),
	dbBase::nlSing(CR,S3),
	SRef=dbBase::dbRef(Loc),
	sx::concat([STyp," ",Realm," {",S2,":",S3,"} [",SRef,"]"],S),
	!
;	
	dbBase::getRecord(Loc,RecTyp,Base,_,_,_,Realm,No,Sub,_,_,_),
	Base<>6,
	STyp=toString(RecTyp),
	SNo=toString(No),
	SSub=toString(Sub),
	SRef=dbBase::dbRef(Loc),
	sx::concat([STyp," ",Realm,"|",SNo,"|",SSub," [",SRef,"]"],S),
	!.
end implement suspended
