/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement tablelayout
    open core, vpiDomains, resourceIdentifiers

    constants
        className = "com/rudstrand/TaskWindow/tablelayout".
        classVersion = "1.0".

    clauses
        classInfo(className, classVersion).

    facts
        thisWin : vpiDomains::windowHandle := erroneous.

    clauses
        new(Parent):-
            _ = vpi::winCreate(windowType, rectangle, title, menu, Parent,
                               windowFlags, eventHandler, 0).

    predicates
        eventHandler : vpiDomains::ehandler.
    clauses
        eventHandler(Win, Event) = Result :-
            Result = generatedEventHandler(Win, Event),
            !.
        eventHandler(Win, e_Menu(Id, CAS)) = Result :-
            Parent = vpi::winGetParent(Win),
            Result = vpi::winSendEvent(Parent, e_Menu(Id, CAS)).

    predicates
        onCreate : vpiDomains::createHandler.
    clauses
        onCreate(_CreationData) = defaultHandling():-
        vpi::cursorSetWait(),
        guix::setStyle(thisWin,[idct_sentence_type,idc_st],[]),
        L=pop::getSentenceSTs([]),
        dbBase::nlSings(L,Ln),
        guix::lbxPop(thisWin,idc_st,Ln),
        guix::addWin(102,thisWin),
        vpi::cursorSet(thisWin,cursor_Arrow),
  	    !;!.

    predicates
        onDestroy : vpiDomains::destroyHandler.
    clauses
        onDestroy() = defaultHandling() :-
            guix::dropWin(thisWin),
            thisWin := erroneous.

    predicates
        onSizeChanged : vpiDomains::sizeHandler.
    clauses
        onSizeChanged(_Width, _Height) = defaultHandling():-
            vpiToolbar::resize(thisWin).

    predicates
        onControlst : vpiDomains::controlHandler.
    clauses
        onControlst(_CtrlID, _CtrlType, _CtrlWin, CtrlInfo) = handled(0):-
        CtrlInfo=vpiDomains::selchanged,
        guix::lbtGet(thisWin,idc_st,ST),
        layout::display(ST,0,0),
        !;!.

    predicates
        onCloseRequest : vpiDomains::closeRequestHandler.
    clauses
        onCloseRequest() = defaultHandling():-
            guix::dropWin(thisWin),
            vpi::winDestroy(thisWin),
            !.

    % This code is maintained by the VDE do not update it manually, 13:47:26-26.1.2005
    constants
        windowType : vpiDomains::wintype = w_TopLevel.
        windowFlags : vpiDomains::wsflags = [wsf_TitleBar,wsf_Minimize,wsf_Close,wsf_ClipSiblings,wsf_ClipChildren].
        rectangle : vpiDomains::rct = rct(100,80,416,138).
        menu : vpiDomains::menu = noMenu.
        title : string = "Table Layout".

    predicates
        generatedEventHandler : vpiDomains::ehandler.
    clauses
        generatedEventHandler(Win, e_create(_)) = _ :-
            thisWin := Win,
            _ = vpi::winCreateControl(wc_LBoxButton,rct(8,24,312,464),"",Win,[wsf_Group,wsf_TabStop,wsf_VScroll,wsf_Sort],idc_st),
            _ = vpi::winCreateControl(wc_Text,rct(8,8,312,23),"Sentence type",Win,[wsf_AlignCenter],idct_sentence_type),
            fail.
        generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
            handled(Result) = onCreate(CreationData).
        generatedEventHandler(_Win, e_Size(Width, Height)) = Result :-
            handled(Result) = onSizeChanged(Width, Height).
        generatedEventHandler(_Win, e_Control(idc_st, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlst(idc_st, CtrlType, CtrlWin, CtlInfo).
        generatedEventHandler(_Win, e_CloseRequest()) = Result :-
            handled(Result) = onCloseRequest().
        generatedEventHandler(_Win, e_Destroy()) = Result :-
            handled(Result) = onDestroy().
    % end of automatic code
end implement tablelayout
