/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement renameSegment
open core,vpiDomains,resourceIdentifiers,globalDomains,termDomain

constants
    className = "com/rudstrand/TaskWindow/tinyDialogs/renameSegment".
    classVersion = "1.0".

clauses
    classInfo(className, classVersion).

facts
    thisWin : vpiDomains::windowHandle := erroneous.

clauses
    show(Parent):-
        _ = vpi::winCreateDynDialog(Parent, controlList, eventHandler, 0).

predicates
eventHandler : vpiDomains::ehandler.
renameSegment : (string,string,count).
renameRealm : (string,string,integer) procedure (i,i,o).

clauses
    eventHandler(Win, Event) = Result :-
        Result = generatedEventHandler(Win, Event).

predicates
    onCreate : vpiDomains::createHandler.
clauses
    onCreate(_CreationData) = defaultHandling() :-
  	guix::guiRB_SetN(thisWin,[idc_realm,idc_segment],[1,2],1),
  	!;!.

predicates
    onDestroy : vpiDomains::destroyHandler.
clauses
    onDestroy() = defaultHandling() :-
        thisWin := erroneous.

predicates
    onControlOK : vpiDomains::controlHandler.
clauses
    onControlOK(_Ctrl, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        guix::guiRB_Get(thisWin,[idc_realm,idc_segment],[1,2],B),
        B=1,
        guix::txtGet(thisWin,idc_present,P),
        guix::txtGet(thisWin,idc_new,N),
        N<>"",
        P<>"",
        N<>P,
        string::frontchar(N,F1,_),
        F1<>'$',
        string::frontchar(P,F2,_),
        F2<>'$',
        vpi::cursorSetWait(),
        renameRealm(P,N,M),
        S=toString(M),
        vpi::cursorSet(thisWin,cursor_Arrow),
        Mess=string::concat("Renamed realm-name in ",S," records. Annotation for the renamed realm is not changed!"),
        dlgIF::dlgNote(0,Mess),
        vpi::winDestroy(thisWin),
        !
;        
        guix::guiRB_Get(thisWin,[idc_realm,idc_segment],[1,2],B),
        B=2,
        guix::txtGet(thisWin,idc_present,P),
        guix::txtGet(thisWin,idc_new,N),
        N<>"",
        P<>"",
        N<>P,
        vpi::cursorSetWait(),
        Context=dbBase::getContext(),
        C=count::new(0),
        renameSegment(P,N,C),
        M=C:get(),
        S=toString(M),
        vpi::cursorSet(thisWin,cursor_Arrow),
        Mess=string::concat("Renamed segment in ",S," records!"),
        set::replaceMember(Context,P,N,NewContext),
        dbBase::setContext(NewContext),
        dlgIF::dlgNote(0,Mess),
        vpi::winDestroy(thisWin),
        !
;
        dlgIF::dlgNote(0,"Names cannot be blank.  Present and new name must be different. Realm names cannot start with $!"),
        !.

predicates
    onControlidc_help : vpiDomains::controlHandler.
clauses
    onControlidc_help(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0):-
    dlgIF::hlpWeb("w_renameseg"),
    !.

predicates
    onControlidc_realm : vpiDomains::controlHandler.
clauses
    onControlidc_realm(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
  	guix::guiRB_SetN(thisWin,[idc_realm,idc_segment],[1,2],1),
  	!;!.

predicates
    onControlidc_segment : vpiDomains::controlHandler.
clauses
    onControlidc_segment(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
  	guix::guiRB_SetN(thisWin,[idc_realm,idc_segment],[1,2],2),
  	!;!.

% This code is maintained by the VDE do not update it manually, 15:45:38-12.11.2005
constants
    dialogType : vpiDomains::wintype = wd_modal.
    title : string = "Rename Realm/Segment".
    rectangle : vpiDomains::rct = rct(50,40,237,117).
    dialogFlags : vpiDomains::wsflags = [wsf_Close,wsf_TitleBar].
    dialogFont = "Arial".
    dialogFontSize = 8.

constants
    controlList : vpiDomains::windef_list =
        [
        dlgFont(wdef(dialogType, rectangle, title, u_DlgBase),
                dialogFont, dialogFontSize, dialogFlags),
        ctl(wdef(wc_PushButton,rct(24,62,92,74),"&OK",u_DlgBase),idc_ok,[wsf_Border,wsf_Default,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(100,62,168,74),"&Help",u_DlgBase),idc_help,[wsf_Border,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_Edit,rct(56,28,184,40),"",u_DlgBase),idc_present,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll]),
        ctl(wdef(wc_Edit,rct(56,44,184,56),"",u_DlgBase),idc_new,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll]),
        ctl(wdef(wc_RadioButton,rct(4,2,52,12),"Realm",u_DlgBase),idc_realm,[wsf_TabStop]),
        ctl(wdef(wc_RadioButton,rct(4,14,52,24),"Segment",u_DlgBase),idc_segment,[wsf_TabStop]),
        ctl(wdef(wc_Text,rct(4,28,52,38),"Present name",u_DlgBase),idct_present_name,[wsf_AlignLeft]),
        ctl(wdef(wc_Text,rct(4,44,52,54),"New name",u_DlgBase),idct_new_name,[wsf_AlignLeft])
        ].

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Control(idc_ok, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlOK(idc_ok, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_help, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_help(idc_help, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_realm, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_realm(idc_realm, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Control(idc_segment, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_segment(idc_segment, CtrlType, CtrlWin, CtlInfo).
    generatedEventHandler(_Win, e_Destroy()) = Result :-
        handled(Result) = onDestroy().
% end of automatic code

renameRealm(Pres,New,N) :-
    dbBase::realmRename(Pres,New,N),
    !.
    
renameSegment(Pres,New,Count) :-
    dbBase::accessErrorChain("q"),
    !
;    
	dbBase::chainSentences("q",LTerm,Loc),
        LTerm=cmp(H,[Arg,termDomain::str(Pres)]),
        NTerm=cmp(H,[Arg,termDomain::str(New)]),
        dbBase::getRecord(Loc,RT,_,_,_,_,Realm,No,SubNo,_,_,_),
        dbBase::recordDeleteOutright(Loc),
        dbBase::recordInsertOutright(RT,Realm,No,SubNo,NTerm,atom("true"),_),
        dbBase::chainHasChangedChain("q"),
        Count:inc(_),
        fail
;    
	dbBase::chainSentences("v",LTerm,Loc),
        LTerm=cmp(H,[Arg,termDomain::str(Pres)]),
        NTerm=cmp(H,[Arg,termDomain::str(New)]),
        dbBase::getRecord(Loc,RT,_,_,_,_,Realm,No,SubNo,_,_,_),
        dbBase::recordDeleteOutright(Loc),
        dbBase::recordInsertOutright(RT,Realm,No,SubNo,NTerm,atom("true"),_),
        dbBase::chainHasChangedChain("v"),
        Count:inc(_),
        fail
;
       !.   
end implement renameSegment
