/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement questioncluster
open core, vpiDomains, resourceIdentifiers

constants
className = "com/rudstrand/TaskWindow/winputfacility/questioncluster".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

facts
thisWin : vpiDomains::windowHandle := erroneous.
rules : string_list := [].

predicates
popAlts : (string_list,integer) determ.
getAnswers : (string_list,integer,string_list,string_list) determ (i,i,o,o).
displayAnno : (integer,string_list) determ.

clauses
show(Parent,Ctls,Prompt,X,Y,Rules):-
        rules:=Rules,
        Y2=Y+50,
        Rectangle=rct(50,40,X,Y2),
        C=dlgFont(wdef(dialogType,Rectangle,Prompt,u_DlgBase),dialogFont, dialogFontSize, dialogFlags),
        L=controlList,
        L=[_|T],
        set::conc([[C|T],Ctls],Cs),
        _ = vpi::winCreateDynDialog(Parent,Cs,eventHandler,0),
        !.

predicates
    eventHandler : vpiDomains::ehandler.
clauses
    eventHandler(Win, Event) = Result :-
        Result = generatedEventHandler(Win, Event).

predicates
    onCreate : vpiDomains::createHandler.
clauses
    onCreate(_CreationData) = defaultHandling() :-
        popAlts(rules,6004),
        !;!.

predicates
    onDestroy : vpiDomains::destroyHandler.
clauses
    onDestroy() = defaultHandling() :-
        thisWin := erroneous.

predicates
    onControlOK : vpiDomains::controlHandler.
clauses
    onControlOK(_Ctrl, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
        getAnswers(rules,6004,Questions,Answers),
        globalVars::pushRetPars(Questions),
        globalVars::pushRetPars(Answers),
        globalVars::pushRetPars(["1"]),
        vpi::winDestroy(thisWin),
        !;!.

predicates
    onControlCancel : vpiDomains::controlHandler.
clauses
    onControlCancel(_Ctrl, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
    vpiCommonDialogs::ask("Please confirm","Do you really want to stop evaluation of the inexact rule network (and related query)?",["Yes","No"])=vpiCommonDialogs::resp_default,
    ir::setBreak(1),
  	globalVars::pushRetPars(["0"]),
    vpi::winDestroy(thisWin),
    !;!.

predicates
    onControlidc_help : vpiDomains::controlHandler.
clauses
    onControlidc_help(_CtrlID, _CtrlType, _CtrlWin, _CtrlInfo) = handled(0) :-
	dlgIF::hlpNo(60),
	!;!.

% This code is maintained by the VDE do not update it manually, 11:15:29-13.7.2005
constants
    dialogType : vpiDomains::wintype = wd_modal.
    title : string = "questioncluster".
    rectangle : vpiDomains::rct = rct(50,40,353,97).
    dialogFlags : vpiDomains::wsflags = [wsf_Close,wsf_TitleBar].
    dialogFont = "Arial".
    dialogFontSize = 8.

constants
    controlList : vpiDomains::windef_list =
        [
        dlgFont(wdef(dialogType, rectangle, title, u_DlgBase),
                dialogFont, dialogFontSize, dialogFlags),
        ctl(wdef(wc_PushButton,rct(4,2,88,14),"&OK",u_DlgBase),idc_ok,[wsf_Border,wsf_Default,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(112,2,196,14),"Stop",u_DlgBase),idc_cancel,[wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_PushButton,rct(216,2,300,14),"&Help",u_DlgBase),idc_help,[wsf_Border,wsf_Group,wsf_TabStop]),
        ctl(wdef(wc_Edit,rct(4,16,300,54),"",u_DlgBase),idc_anno,[wsf_AlignLeft,wsf_Group,wsf_TabStop,wsf_AutoHScroll,wsf_MultiLine,wsf_Disabled])
        ].

predicates
    generatedEventHandler : vpiDomains::ehandler.
clauses
    generatedEventHandler(Win, e_create(_)) = _ :-
        thisWin := Win,
        fail.
    generatedEventHandler(_Win, e_Create(CreationData)) = Result :-
        handled(Result) = onCreate(CreationData).
    generatedEventHandler(_Win, e_Control(idc_ok, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlOK(idc_ok, CtrlType, CtrlWin, CtlInfo),!.
    generatedEventHandler(_Win, e_Control(idc_cancel, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlCancel(idc_cancel, CtrlType, CtrlWin, CtlInfo),!.
    generatedEventHandler(_Win, e_Control(idc_help, CtrlType, CtrlWin, CtlInfo)) = Result :-
            handled(Result) = onControlidc_help(idc_help, CtrlType, CtrlWin, CtlInfo),!.
    generatedEventHandler(_Win, e_Destroy()) = Result :-
            handled(Result) = onDestroy().
% end of automatic code

generatedEventHandler(_Win, e_Control(No,_,_,_)) = 1 :-
            N=No-6001,
            R=N mod 5,
            R=0,
            !,
            M=(N+5) div 5,
            displayAnno(M,rules),
            !.

popAlts([],_) :-
    !.
popAlts([R|Rs],No) :-
    dbExtension::getQuestion(R,Question,QDom,Default),
	nl::conValX(Default,QDom,"",DefaultS),
    irbase::unpackQuestion(Question,QDom,_,AltsNL),  
    set::size(AltsNL,N),
    N>0,
    guix::lbxPop(thisWin,No,AltsNL),
    guix::lbtSel(thisWin,No,DefaultS),
    Next=No+5,
    popAlts(Rs,Next),
    !
;    
    Next=No+5,
    popAlts(Rs,Next),
    !.

getAnswers([],_,[],[]) :-
    !.
getAnswers([R|Rs],No,[Q|Qs],[A|As]) :-
    dbExtension::getQuestion(R,Question,QDom,_),
    irbase::unpackQuestion(Question,QDom,_,AltsNL),  
    set::size(AltsNL,N),
    N=0,
    !,
    M=No-1,
    guix::txtGet(thisWin,M,Q),
    guix::txtGet(thisWin,No,A),
    NextNo=No+5,
    getAnswers(Rs,NextNo,Qs,As),
    !
;    
    M=No-1,
    guix::txtGet(thisWin,M,Q),
    guix::lbtGet(thisWin,No,A),
    NextNo=No+5,
    getAnswers(Rs,NextNo,Qs,As),
    !.

displayAnno(No,[R|Rs]) :-
    No=1,
    dbExtension::kbGetAnnotationBlank("p3_iIR",R,Note),
    guix::txtSet(thisWin,idc_anno,Note),
    !
;
    Next=No-1,
    displayAnno(Next,Rs),
    !.    
end implement questioncluster
