/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement bck
open core,termDomain,globalDomains

constants
className = "com/rudstrand/bck/bck".
classVersion = "1.0".

class predicates
bckAnno : (integer,integer,integer) determ (i,i,i).
bckSM : (integer,integer,integer) determ (i,i,i).
bckSM_X : (outFile,integer,integer,string,count).
bckFunc : (integer,integer) determ.
bckQuery : (integer) determ (i).
bckInpfac : (integer) determ (i).
bckSents : (integer,integer,integer) determ (i,i,i).
bckTL : (integer,integer,integer) determ (i,i,i).
bckCH : (integer,integer,integer) determ (i,i,i).
bckTerm : (integer,integer,integer) determ (i,i,i).
bckCR : (integer,integer) determ (i,i).
bckSR : (integer) determ (i).
bckSentenceRecord : (outFile,integer,integer,count,dbLoc) failure.
bckFileHead : (integer,integer,string,outFile) determ (i,i,i,o).
bckFileTail : (outFile,integer) determ.
dmFuncSentences : (string_list,string_list) determ (i,o).
realmType : (string,integer) procedure (i,o).

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------
% bckRun                                                         
% -------------------------------------------------------------------------------------
bckRun(Btm,Txml,Tnl,Bsm,Sxml,Snl,Ban,ANxml,ANnl,Bfu,Fxml,_,Bas,ASxml,ASnl,Bsr,Bqr,Bcr,Bcrnl,Btl,TLxml,TLnl,Bch,CHxml,CHnl,Inpfac) :-
	bckTerm(Btm,Txml,Tnl),
	bckSM(Bsm,Sxml,Snl),
	bckAnno(Ban,ANxml,ANnl),
	bckSents(Bas,ASxml,ASnl),
	bckFunc(Bfu,Fxml),
	bckSR(Bsr),
	bckQuery(Bqr),
	bckCR(Bcr,Bcrnl),
	bckTL(Btl,TLxml,TLnl),
	bckCH(Bch,CHxml,CHnl),
	bckInpfac(Inpfac),
	!.
% -------------------------------------------------------------------------------------
% bckCH                                     
% -------------------------------------------------------------------------------------
class predicates
bckCH_X : (outFile,integer,integer,count).
clauses

bckCH(-1,_,_) :- 
	!.
bckCH(1,XML,NL) :-
	vpi::cursorSetWait(),
	bckFileHead(XML,NL,"BackupPublicChains",File),
	Count=count::newCounter(0),
	bckCH_X(File,XML,NL,Count),
	    N1=Count:get(),
	    S1=toString(N1),
		sx::concat(["backed up ",S1," public chain declarations"],M1),
		dlgIF::msgS(M1),
	% close
	bckFileTail(File,XML),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.

bckCH_X(File,XML,NL,Count) :-
    dbBase::chainSentences("p3_pIpch",LT,Loc),
	LT=cmp(_,[cmp("p3_eChan",_)]),
	bckSentenceRecord(File,XML,NL,Count,Loc),
	Count:inc(_),
	fail
;
	!.
% -------------------------------------------------------------------------------------
% bckTL                                                
% -------------------------------------------------------------------------------------
class predicates
bckTL_X : (outFile,integer,integer,count).
clauses

bckTL(_,_,_) :- 
    dbBase::chainIsExternal("y"),
    !.
bckTL(-1,_,_) :- 
	!.
bckTL(1,XML,NL) :-
	vpi::cursorSetWait(),
	bckFileHead(XML,NL,"BackupTableLayouts",File),
	Count=count::newCounter(0),
	bckTL_X(File,XML,NL,Count),
	    N1=Count:get(),
	    S1=toString(N1),
		sx::concat(["backed up ",S1," table layouts"],M1),
		dlgIF::msgS(M1),
	% close
	bckFileTail(File,XML),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.

bckTL_X(File,XML,NL,Count) :-
    dbBase::chainSentences("y",LT,Loc),
	LT=cmp(_,[cmp("p3_eTabLay",_)]),
	bckSentenceRecord(File,XML,NL,Count,Loc),
	Count:inc(_),
	fail
;
	!.
% -------------------------------------------------------------------------------------
% bckAnno                                                     
% -------------------------------------------------------------------------------------
class predicates
bckAnnoX : (outFile,integer,integer,count).
clauses

bckAnno(_,_,_) :- 
    dbBase::chainIsExternal("p3_pIntd"),
    !.
bckAnno(-1,_,_) :- 
	!.
bckAnno(1,XML,NL) :-
	vpi::cursorSetWait(),
	bckFileHead(XML,NL,"BackupAnnotations",File),
	Count=count::newCounter(0),
	bckAnnoX(File,XML,NL,Count),
	    N1=Count:get(),
	    S1=toString(N1),
		sx::concat(["backed up ",S1," annotations"],M1),
		dlgIF::msgS(M1),
	% close
	bckFileTail(File,XML),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.
bckAnnoX(File,XML,NL,Count) :-
    dbBase::chainSentences("p3_pIntd",LT,Loc),
	LT=cmp(_,[cmp("p3_eAnno",_)]),
	bckSentenceRecord(File,XML,NL,Count,Loc),
	Count:inc(_),
	fail
;
	!.
% -------------------------------------------------------------------------------------
% bckFileHead/Tail                                    
% -------------------------------------------------------------------------------------
bckFileHead(XML,NL,Name,File) :-
	XML=1,
	FileName=globalVars::get_s("kb_file"),
	fio::file_name(FileName,Path,Nm,_),
	NameX=string::concat(Nm,"_",Name),
    fioExt::filGenName(0,Path,NameX,"xml",FF),
	File=outFile::new(FF),
	xmlP3::xmlKnowA(A),
	File:write(A),
	!
;
	XML=-1,
	Ix=NL+2,
	set::setGet(["3pr","","3nl"],Ix,Typ),
	FileName=globalVars::get_s("kb_file"),
	fio::file_name(FileName,Path,Nm,_),
	NameX=string::concat(Nm,"_",Name),
    fioExt::filGenName(0,Path,NameX,Typ,FF),
	File=outFile::new(FF),
	!.
bckFileTail(File,XML) :-
	XML=1,
	xmlP3::xmlKnowB(B),
	File:write(B),
	Name=File:getName(),
	File:close,
	sx::concat(["Backup stored in ",Name,"."],S),	
	dlgIF::msgS(S),
	!
;
	XML=-1,
	Name=File:getName(),
	File:close,
	sx::concat(["Backup stored in ",Name,"."],S),	
	dlgIF::msgS(S),
	!.
% -------------------------------------------------------------------------------------
% bckFunc                                          
% -------------------------------------------------------------------------------------
class predicates
bckFuncX : (outFile,integer).
bckFuncY : (integer,string,sterm,string) determ (i,i,i,o).
clauses

bckFunc(-1,_) :- 
	!.
bckFunc(1,XML) :-
	vpi::cursorSetWait(),
	bckFileHead(XML,1,"BackupFunctions",File),
	bckFuncX(File,XML),
	% close
	bckFileTail(File,XML),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.
bckFuncX(File,XML) :-
	Count=count::newCounter(0),
	bckSM_X(File,XML,1,"f",Count),
	bckSM_X(File,XML,1,"g",Count),
	dbBase::chainSentences("q",LT,Loc),
        dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
        realmType(Realm,RealmTyp),
        RealmTyp>1,
        LT=cmp(_,[cmp(Clas,_)|_]),
        set::member(Clas,["f","g"]),
        bckFuncY(XML,Realm,LT,S),
        File:writenl(S),
        fail
;
	!.
bckFuncY(_,_,Term,_) :-
	Term=cmp(_,[cmp(_,[atom(F)|_])|_]),
    dbBase::chainIsExternal(F),
    !,fail.
bckFuncY(-1,_,Term,S) :-
	Term=cmp(_,[cmp(Clas,[atom(F)|_])|_]),
	set::member(Clas,["f","g"]),
	dbExtension::functionBodies(F,Bodies1),
	dmFuncSentences(Bodies1,Bodies2),
	sx::sxlJoin(Bodies2,"\n",S),
	!.
bckFuncY(1,Realm,Term,X) :-
	Term=cmp("q",[cmp(_,[atom(F)|_])|_]),
	xmlP3::xmlFromFunc("NL",Realm,F,X),
	!.
% -------------------------------------------------------------------------------------
% bckInpfac                                              
% -------------------------------------------------------------------------------------
class predicates
bckInpfacX : (outFile,count).
clauses

bckInpfac(_) :- 
    dbBase::chainIsExternal("p3_pIRinp"),
    !.
bckInpfac(-1) :- 
	!.
bckInpfac(1) :-
	vpi::cursorSetWait(),
	bckFileHead(-1,1,"BackupInputFacilities",File),
	Count=count::newCounter(0),
	bckInpfacX(File,Count),
		N1=Count:get(),
		S1=toString(N1),
		sx::concat(["backed up ",S1," input facilities"],M1),
		dlgIF::msgS(M1),
	bckFileTail(File,-1),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.
	
bckInpfacX(File,Count) :-
    dbBase::chainSentences("p3_pIRinp",LT,Loc),
        dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
        realmType(Realm,RealmTyp),
        RealmTyp>1,
        nl::conSentence(LT,S2),
        S=string::concat(S2,"!"),
        File:writenl(S),
        Count:inc(_),
        fail
;
	!.
% -------------------------------------------------------------------------------------
% bckQuery                                                
% -------------------------------------------------------------------------------------
class predicates
bckQueryX : (outFile,count,string).
clauses

bckQuery(_) :-
    dbBase::chainIsExternal("y"),
    !.
bckQuery(-1) :- 
	!.
bckQuery(1) :-
	vpi::cursorSetWait(),
	bckFileHead(-1,1,"BackupQueries",File),
	Count=count::newCounter(0),
	bckQueryX(File,Count,"p3_eKbm"),
		N1=Count:get(),
		S1=toString(N1),
		sx::concat(["backed up ",S1," query menus"],M1),
		dlgIF::msgS(M1),
	Count:set(0),
	bckQueryX(File,Count,"p3_eKbp"),
		N3=Count:get(),
		S3=toString(N3),
		sx::concat(["backed up ",S3," query parameters"],M3),
		dlgIF::msgS(M3),
	Count:set(0),
	bckQueryX(File,Count,"p3_eKbq"),
		N2=Count:get(),
		S2=toString(N2),
		sx::concat(["backed up ",S2," queries"],M2),
		dlgIF::msgS(M2),
	% close
	bckFileTail(File,-1),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.

bckQueryX(File,Count,Ex) :-
	dbBase::chainSentences("y",LT,Loc),
        dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
        realmType(Realm,RealmTyp),
        RealmTyp>1,
        LT=cmp(_,[cmp(E,_)]),
        E=Ex,
        nl::conSentence(LT,S2),
        S=string::concat(S2,"!"),
        File:writenl(S),
        Count:inc(_),
        fail
;
	!.
% -------------------------------------------------------------------------------------
% bckSents                                                   
% -------------------------------------------------------------------------------------
class predicates
bckAppX : (outFile,integer,integer,count).
bckAppChain : (string) determ.
clauses

bckSents(-1,_,_) :- 
	!.
bckSents(1,XML,NL) :-
	vpi::cursorSetWait(),
	bckFileHead(XML,NL,"BackupSentences",File),
	Count=count::newCounter(0),
	bckAppX(File,XML,NL,Count),
		N1=Count:get(),
		S1=toString(N1),
		sx::concat(["backed up ",S1," application sentences"],M1),
		dlgIF::msgS(M1),
	% close
	bckFileTail(File,XML),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.

bckAppX(File,XML,NL,Count) :-
    dbBase::chains_nd(-1,Chain),
        not(dbBase::chainIsExternal(Chain)),
        bckAppChain(Chain),
        dbBase::chainSentences(Chain,_,Loc),
            bckSentenceRecord(File,XML,NL,Count,Loc), % always fails
            !
;
	!.

bckAppChain(Chain) :- 
	string::length(Chain)=1,
	!,fail
;
	sx::frontstr(3,Chain,F,_),
	F="p3_",
	!,fail
;
	!.		
% -------------------------------------------------------------------------------------
% bckSM                                                   
% -------------------------------------------------------------------------------------
bckSM(_,_,_) :-
   dbBase::chainIsExternal("q"),
   !.
bckSM(-1,_,_) :- 
	!.
bckSM(1,XML,NL) :-
	vpi::cursorSetWait(),
	bckFileHead(XML,NL,"BackupSentenceModel",File),
	Count=count::newCounter(0),
	%
	File:nl,
	bckSM_X(File,XML,NL,"a",Count),
		N1=Count:get(),
		S1=toString(N1),
		sx::concat(["backed up ",S1," data element types"],M1),
		dlgIF::msgS(M1),
	% 
	Count:set(0),
	File:nl,
	bckSM_X(File,XML,NL,"e",Count),
		N2=Count:get(),
		S2=toString(N2),
		sx::concat(["backed up ",S2," entity types"],M2),
		dlgIF::msgS(M2),
	% 
	File:nl,
	Count:set(0),
	bckSM_X(File,XML,NL,"p",Count),
		N4=Count:get(),
		S4=toString(N4),
		sx::concat(["backed up ",S4," predicate types"],M4),
		dlgIF::msgS(M4),
	% 
	File:nl,
	Count:set(0),
	bckSM_X(File,XML,NL,"f",Count),
		N5=Count:get(),
		S5=toString(N5),
		sx::concat(["backed up ",S5," deterministic function declarations"],M5),
		dlgIF::msgS(M5),
	% 
	File:nl,
	Count:set(0),
	bckSM_X(File,XML,NL,"g",Count),
		N6=Count:get(),
		S6=toString(N6),
		sx::concat(["backed up ",S6," non-deterministic function declarations"],M6),
		dlgIF::msgS(M6),
	% 
	File:nl,
	Count:set(0),
	bckSM_X(File,XML,NL,"h",Count),
		N7=Count:get(),
		S7=toString(N7),
		sx::concat(["backed up ",S7," inexact rule declarations"],M7),
		dlgIF::msgS(M7),
	% close
	bckFileTail(File,XML),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.
% -------------------------------------------------------------------------------------
% bckSM_X                                      
% -------------------------------------------------------------------------------------
bckSM_X(File,XML,NL,Clas,Count) :-
    not(dbBase::chainIsExternal("q")),
	dbBase::chainSentences("q",Term,Loc),
	Term=cmp(_,[cmp(C,_),_]),
	C=Clas,
	bckSentenceRecord(File,XML,NL,Count,Loc),
	!
;
	!.
% -------------------------------------------------------------------------------------
% bckCR                                         
% -------------------------------------------------------------------------------------
bckCR(-1,_) :- 
	!.
bckCR(1,NL) :-
	vpi::cursorSetWait(),
	Rules=pop::getCRs(),
	bckFileHead(1,NL,"BackupInexactRules",File),
	Count=count::newCounter(0),
	Ix=NL+2,
	set::setGet(["PR","","NL"],Ix,Form),
	bckCRs(File,Form,Rules,Count),
	N=Count:get(),
	%
	bckFileTail(File,1),
			Ns=toString(N),
			sx::concat(["Backed up ",Ns," inexact reasoning rules."],Mess),
			dlgIF::msgS(Mess),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.
	
bckCRs(File,Form,Rules,Count) :-
	set::size(Rules,N),
	Next=programControl::sequence_nd(1,N,1),
		globalVars::get_i("quit_flag")=0,
		set::setGet(Rules,Next,Rule),
		dbBase::getIrDef(Rule,Body,Loc),
		Body=cmp(Chain,_),
		not(dbBase::chainIsExternal(Chain)),
        dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
		realmType(Realm,RealmTyp),
		RealmTyp>1,
				sx::concat(["Backing up inexact reasoning rule ",Rule,"."],Mess),
				dlgIF::msgS(Mess),
		xmlP3::xmlFromCR(Form,Realm,Body,X),
		File:writenl(X),
		Count:inc(_),
		fail
;
	!.	
% -------------------------------------------------------------------------------------
% bckSentenceRecord                                           
% -------------------------------------------------------------------------------------
bckSentenceRecord(File,-1,NL,Count,Loc) :-	
	NL=1,
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,Term,_),
	realmType(Realm,RealmTyp),
	RealmTyp>1,
	nl::conSentence(Term,S2),
	S=string::concat(S2,"!"),
	File:writenl(S),
	Count:inc(_),
	!,fail
;	
	NL=-1,
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,Term,_),
	realmType(Realm,RealmTyp),
	RealmTyp>1,
	nl::cprTerm(Term,PR),
	File:write(PR),
	File:writenl("!"),
	Count:inc(_),
	!,fail.
bckSentenceRecord(File,1,NL,Count,Loc) :-	
	NL=1,
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,No,_,_,Term,_),
	realmType(Realm,RealmTyp),
	RealmTyp>1,
	xmlP3::xmlFromSentence("NL",Realm,No,Term,X),
	File:writenl(X),
	Count:inc(_),
	!,fail
;	
	NL=-1,
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,No,_,_,Term,_),
	realmType(Realm,RealmTyp),
	RealmTyp>1,
	xmlP3::xmlFromSentence("PR",Realm,No,Term,X),
	File:writenl(X),
	Count:inc(_),
	!,fail.
% -------------------------------------------------------------------------------------
% bckSR                                             
% -------------------------------------------------------------------------------------
class predicates
bckRules : (outFile,dbLoc_list,count).
clauses

bckSR(-1) :- 
	!.
bckSR(1) :-
	vpi::cursorSetWait(),
	globalVars::set_i("inter",-1),
	dbBase::recTypeLocs(3,RCa),
	dbBase::recTypeLocs(-3,RCb),
	dbBase::recTypeLocs(30,RCc),
	dbBase::recTypeLocs(-30,RCd),
	set::conc([RCa,RCb,RCc,RCd],Rules),
	bckFileHead(1,-1,"BackupSentenceRules",File),
	%
	Count=count::newCounter(0),
	bckRules(File,Rules,Count),
	N=Count:get(),
	bckFileTail(File,1),
			Ns=toString(N),
			sx::concat(["Backed up ",Ns," sentence rules."],Mess),
			dlgIF::msgS(Mess),
	W=guix::getWinOfClas(1),
	globalVars::set_i("inter",1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.

bckRules(File,Rules,Count) :-
	set::size(Rules,N),
	Next=programControl::sequence_nd(1,N,1),
		globalVars::get_i("quit_flag")=0,
		set::setGet(Rules,Next,Loc),
		dbBase::getRecord(Loc,RT,_,_,_,_,Realm,No,SubNo,_,LTerm,RTerm),
		LTerm=cmp(Chain,_),
		not(dbBase::chainIsExternal(Chain)),
		SNo=toString(No),
				SSubNo=toString(SubNo),
				sx::concat([Realm,"|",SNo,"|",SSubNo],SRealm),
				sx::concat(["Backing up ",SRealm,"."],Mess),
				dlgIF::msgS(Mess),
		rt::rcnTree(RT,cmp(":-",[LTerm,RTerm]),Rule,_,_),
		Rule=cmp(Typ,_),
		rub::rubRuleType(Typ,Type,_),
		SType=toString(Type),
		xmlP3::xmlSentenceRule(SType,Realm,SNo,SSubNo,Rule,X),
		File:writenl(X),
		Count:inc(_),
		fail
;
	!.	
% -------------------------------------------------------------------------------------
% bckTerm                                                         
% -------------------------------------------------------------------------------------
class predicates
bckTermX : (outFile,integer,integer,string,count).
clauses

bckTerm(_,_,_) :-
   dbBase::chainIsExternal("v"),
	!.
bckTerm(-1,_,_) :- 
	!.
bckTerm(1,XML,NL) :-
	vpi::cursorSetWait(),
	bckFileHead(XML,NL,"BackupTerminology",File),
	Count=count::newCounter(0),
	% syntagms
	File:nl,
	bckTermX(File,XML,NL,"t",Count),
		N1=Count:get(),
		S1=toString(N1),
		sx::concat(["backed up ",S1," syntagms"],M1),
		dlgIF::msgS(M1),
	% synonyms
	File:nl,
	Count:set(0),
	bckTermX(File,XML,NL,"s",Count),
		N2=Count:get(),
		S2=toString(N2),
		sx::concat(["backed up ",S2," synonyms"],M2),
		dlgIF::msgS(M2),
	% close
	bckFileTail(File,XML),
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	!.

% succeeds when all all index entries have been processed
%
bckTermX(File,XML,NL,Clas,Count) :-
    not(dbBase::chainIsExternal("v")),
	dbBase::chainSentences("v",Term,Loc),
	Term=cmp(_,[cmp(Cls,_)|_]),
	Cls=Clas,
	bckSentenceRecord(File,XML,NL,Count,Loc), % always fails
	!
;
	!.
% ----------------------------------------------------------------------------------------------
% realmType
% ----------------------------------------------------------------------------------------------
% parameters:
% 2 out 1=PRO3...  2=$...  3=else
%
realmType(Realm,Typ) :-
	Realm="",
	Typ=3,
	!
;
	sx::frontchar(Realm,'$',_),
	Typ=2,
	!
;
	L=string::length(Realm),
	L<4,
	Typ=3,
	!
;
	sx::frontstr(4,Realm,F,_),
	F="PRO3",
	Typ=1,
	!
;
	Typ=3,
	!.
% --------------------------------------------------------------------------------------------
% dmFuncSentences
% --------------------------------------------------------------------------------------------
dmFuncSentences([],[]) :-
	!.
dmFuncSentences([B|Bs],L) :-
    dbBase::nlSing("p3_iIdefb",IsDef),
	sx::sxUnpackS(B,IsDef,L2),
	L2=[H,T],
	sx::sxUnpack(T,';',L3),
	Hs=string::concat(H,IsDef),
	sx::sxlRemBlanks(L3,L4),
	sx::sxlTrim(L4,L4t),
	sx::sxlAddSuffix(L4t,";",L5),
	set::setRev(L5,L6),
	L6=[H6|T6],
	sx::sxTail(1,H6,H7,_),
	set::setRev([H7|T6],L7),
	dmFuncSentences(Bs,Ls),
	set::conc([[Hs],L7,[""],Ls],L),
	!
;
	dmFuncSentences(Bs,L),
	!.
end implement bck
