/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement cache
open core,termDomain,resourceIdentifiers

constants
className = "com/rudstrand/cache/cache".
classVersion = "1.0".

class facts
smc : integer :=0. % 0/1 sm cache loaded
tmc : integer :=0. % 0/1 tm cache loaded
lang : string :="". % cached language (blank if no cached language)
synt : (string KbName,string Clas,string Form,string Type,string_list Tokens,string Segment,string NlName) nondeterm.
sm : (string KbName,string Type,integer Card,string_list DEs,string Domain,string Form,integer Width,string Segment,string IrChainType,string IrSubType) nondeterm.

class predicates
setFlags : () determ.

clauses
classInfo(className, classVersion).

% ----------------------------------------------------------------------------------------------
%                                                CACHES GENERAL
% ----------------------------------------------------------------------------------------------
check(Chain) :-
    Chain="q",
    offSm,
    !
;    
    Chain="v",
    offTm,
    !
;
    !.
% ----------------------------------------------------------------------------------------------
%                                      TERMINOLOGY CACHE
% ----------------------------------------------------------------------------------------------

% ----------------------------------------------------------------------------------------------
% saveTm
% ----------------------------------------------------------------------------------------------
class predicates
saveTmX : (count).
clauses

saveTm() :-
    C=count::new(),
    saveTmX(C),
    N=C:get(),
    Ns=toString(N),
    Mess=string::concat("Saved terminology cache with ",Ns," elements!"),
    dlgIF::msgS(Mess),
    !.
    
saveTmX(C) :-    
    getSyntagms_nd(KbName,Clas,Form,Type,Tokens,Segment,_),
        termDomain::cls_list(Tokens,Toks),
        Lt=cmp("v",[cmp(Type,[atom(lang),str(KbName),atom(Clas),atom(Form),Toks]),str(Segment)]),
        C:inc(N),
        dbBase::recordInsertOutrightIgnoreCache(2,"TM",N,0,Lt,atom("true"),_),
        fail
;   
    !.
% ----------------------------------------------------------------------------------------------
% getNl
% ----------------------------------------------------------------------------------------------
getNl(KbName)=Nl :- 
    synt(KbName,_,_,"t",_,_,Nl),
    !
;    
    synt(KbName,_,_,"s",_,_,Nl),
    !.
getNl(KbName,Form)=Nl :- 
    synt(KbName,_,Form,"t",_,_,Nl),
    !
;    
    synt(KbName,_,Form,"s",_,_,Nl),
    !.
% ----------------------------------------------------------------------------------------------
% getSyntagm..
% ----------------------------------------------------------------------------------------------
getSyntagm(Nl,KbName,Clas,Form) :-
    synt(KbName,Clas,Form,"t",_,_,Nl),
    !
;    
    synt(KbName,Clas,Form,"s",_,_,Nl),
    !
;    
	pscan::pscNL(Nl,Toks),
	pscan::str_tokens(Toks,SToks1),
	termDomain::str_slist(SToks1,Tokens),
	synt(KbName,Clas,Form,"t",Tokens,_,_),
	!
;    
	pscan::pscNL(Nl,Toks),
	pscan::str_tokens(Toks,SToks1),
	termDomain::str_slist(SToks1,Tokens),
	synt(KbName,Clas,Form,"s",Tokens,_,_),
	!.

getSyntagms_nd(KbName,Clas,Form,Type,Tokens,Segment,Nl) :- synt(KbName,Clas,Form,Type,Tokens,Segment,Nl).
% ----------------------------------------------------------------------------------------------
% kbNameExists
% ----------------------------------------------------------------------------------------------
kbNameExists(KbName) :- synt(KbName,_,_,_,_,_,_),!.
% ----------------------------------------------------------------------------------------------
% loadTm/Conditonally
% ----------------------------------------------------------------------------------------------
loadTmConditionally(Language) :-
    tmc=1,
    lang=Language,
    !
;
    loadTm(Language),
    !
;
    !.
    
loadTmConditionally() :-
    Lang=globalVars::get_s("lang"),
    loadTmConditionally(Lang),
    !
;
    !.
    
loadTm(Language) :-
    dlgIF::msgStatus("Loading terminology cache..."),
    retractall(synt(_,_,_,_,_,_,_)),
    dbBase::syntagm_nd(Language,KbName,Clas,Form,Type,Tokens,Segment),
        sx::sxlJoin(Tokens," ",Nl),
        assert(synt(KbName,Clas,Form,Type,Tokens,Segment,Nl)),
        fail
;
    lang:=Language,
    tmc:=1,
    setFlags,
    dbBase::nlTry(Language,Ln),
    Mess=string::concat("Loaded ",Ln," terminology cache!"),
    dlgIF::msgS(Mess),
    dlgIF::msgStatus(""),
    !.
% ----------------------------------------------------------------------------------------------
% tmCacheIsLoaded
% ----------------------------------------------------------------------------------------------
tmCacheIsLoaded()=L :- tmc=1, L=lang,!.
% ----------------------------------------------------------------------------------------------
% removeTm
% ----------------------------------------------------------------------------------------------
removeTm() :-
    retractall(synt(_,_,_,_,_,_,_)),
    lang:="",
    tmc:=0,
    setFlags,
    dlgIF::msgS("Removed terminology cache!"),
    !.
% ----------------------------------------------------------------------------------------------
% toggleTm
% ----------------------------------------------------------------------------------------------
toggleTm() :-
    tmc=1,
    removeTm(),
    !
;
    Lang=globalVars::get_s("lang"),
    loadTm(Lang),
    !.
% ----------------------------------------------------------------------------------------------
% offTm
% ----------------------------------------------------------------------------------------------
offTm() :-
    tmc=1,
    removeTm(),
    !
;
    !.
% ----------------------------------------------------------------------------------------------
%                                      SENTENCE MODEL CACHE
% ----------------------------------------------------------------------------------------------


% ----------------------------------------------------------------------------------------------
% saveSm
% ----------------------------------------------------------------------------------------------
class predicates
saveSmX : (count).
clauses

saveSm() :-
    C=count::new(),
    saveSmX(C),
    N=C:get(),
    Ns=toString(N),
    Mess=string::concat("Saved sentence model cache with ",Ns," elements!"),
    dlgIF::msgS(Mess),
    !.
    
saveSmX(C) :-    
    getSmClass_nd("a",KbName,_,_,Domain,Form,Width,Segment,_,_),
        Lt=cmp("q",[cmp("a",[atom(KbName),atom(Domain),str(Form),int(Width)]),str(Segment)]),
        C:inc(N),
        dbBase::recordInsertOutrightIgnoreCache(2,"SM",N,0,Lt,atom("true"),_),
        fail
;        
    getSmClass_nd("e",KbName,_,DEs,_,_,_,Segment,_,_),
        termDomain::cla_list(DEs,Ds),
        Lt=cmp("q",[cmp("e",[atom(KbName),Ds]),str(Segment)]),
        C:inc(N),
        dbBase::recordInsertOutrightIgnoreCache(2,"SM",N,0,Lt,atom("true"),_),
        fail
;        
    getSmClass_nd("p",KbName,Card,DEs,_,_,_,Segment,_,_),
        termDomain::cla_list(DEs,Ds),
        Lt=cmp("q",[cmp("p",[atom(KbName),int(Card),Ds]),str(Segment)]),
        C:inc(N),
        dbBase::recordInsertOutrightIgnoreCache(2,"SM",N,0,Lt,atom("true"),_),
        fail
;        
    getSm_nd(KbName,Clas,_,DEs,Domain,_,_,Segment),
        set::member(Clas,["f","g","h","i"]),
        termDomain::cla_list(DEs,Ds),
        Lt=cmp("q",[cmp(Clas,[atom(KbName),atom(Domain),Ds]),str(Segment)]),
        C:inc(N),
        dbBase::recordInsertOutrightIgnoreCache(2,"SM",N,0,Lt,atom("true"),_),
        fail
;   
    !.
% ----------------------------------------------------------------------------------------------
% getSm
% ----------------------------------------------------------------------------------------------
getSm(KbName,Clas,Card,DEs,Domain,Form,Width,Segment) :- sm(KbName,Clas,Card,DEs,Domain,Form,Width,Segment,_,_),!.
getSm(KbName,Clas,Card,DEs,Domain,Form,Width,Segment,ChainType,SubType) :- sm(KbName,Clas,Card,DEs,Domain,Form,Width,Segment,ChainType,SubType),!.

getSm_nd(KbName,Clas,Card,DEs,Domain,Form,Width,Segment) :- sm(KbName,Clas,Card,DEs,Domain,Form,Width,Segment,_,_).

getSmClass_nd(Clas,KbName,Card,DEs,Domain,Form,Width,Segment,ChainType,Subtype) :- sm(KbName,Clas,Card,DEs,Domain,Form,Width,Segment,ChainType,Subtype).
% ----------------------------------------------------------------------------------------------
% loadSm/Conditionally
% ----------------------------------------------------------------------------------------------
class predicates
inexType : (string,string,string,string) procedure (i,i,o,o).
clauses

loadSmConditionally() :-
    smc=0,
    loadSm,
    !
;
    !.
    
loadSm() :-
    dlgIF::msgStatus("Loading sentence model cache..."),
    retractall(sm(_,_,_,_,_,_,_,_,_,_)),
    dbBase::chainSentences("q",LT),
        LT=cmp(_,[cmp(Clas,[atom(KbName)|T]),str(Segment)]),
        dbBase::smDef2(Clas,T,Card,DEs,Domain,Form,Width),
        inexType(Clas,KbName,ChainType,SubType),
        assert(sm(KbName,Clas,Card,DEs,Domain,Form,Width,Segment,ChainType,SubType)),
        fail
;
    smc:=1,
    setFlags,
    dlgIF::msgS("Loaded sentence model cache!"),
    dlgIF::msgStatus(""),
    !.

inexType("h",K,ChainType,SubType) :- dbBase::irTypeF(K,ChainType,SubType),!.
inexType(_,_,"","") :- !.
% ----------------------------------------------------------------------------------------------
% smCacheIsLoaded
% ----------------------------------------------------------------------------------------------
smCacheIsLoaded() :- smc=1,!.
% ----------------------------------------------------------------------------------------------
% removeSm
% ----------------------------------------------------------------------------------------------
removeSm() :-
    retractall(sm(_,_,_,_,_,_,_,_,_,_)),
    smc:=0,
    setFlags,
    dlgIF::msgS("Removed sentence model cache!"),
    !.
% ----------------------------------------------------------------------------------------------
% toggleSm
% ----------------------------------------------------------------------------------------------
toggleSm() :-
    smc=1,
    removeSm(),
    !
;
    loadSm,
    !.
% ----------------------------------------------------------------------------------------------
% offSm
% ----------------------------------------------------------------------------------------------
offSm() :-
    smc=1,
    removeSm(),
    !
;
    !.
% ----------------------------------------------------------------------------------------------
%                                                LOCAL
% ----------------------------------------------------------------------------------------------
% ----------------------------------------------------------------------------------------------
% setFlags
% ----------------------------------------------------------------------------------------------
setFlags() :-
    T=tmc+1,
    set::setGet(["","T"],T,S1),
    S=smc+1,
    set::setGet(["","S"],S,S2),
    Flags=string::concat(S1,S2),
    W=guix::getWinOfClas(1),
	vpiToolbar::setValue(W,idt_cache,vpiToolbar::text_value(Flags)),
	!.	
end implement cache
