/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement conversion
open core,termDomain,globalDomains

constants
    className = "com/rudstrand/conversion/conversion".
    classVersion = "1.0".

clauses
classInfo(className, classVersion).
% -----------------------------------------------------------------------------------------------
% toTree;fromTree
% -----------------------------------------------------------------------------------------------
class predicates
toTrees : (sterml,vpiTree::tree_list) determ (i,o).
fromTrees : (vpiTree::tree_list,sterml) determ (i,o).
clauses

fromTree(Tree)=Term :-
    Tree=vpiTree::tree(Selector,vpiTree::marked,Trees,Arrow),
    fromTrees(Trees,T),
    S=termDomain::toTermStringIndirect("s",str(Selector)),
    Term=cmp("",[str(S),atom("marked"),int(Arrow)|T]),
    !
;    
    Tree=vpiTree::tree(Selector,vpiTree::unmarked,Trees,Arrow),
    fromTrees(Trees,T),
    S=termDomain::toTermStringIndirect("s",str(Selector)),
    Term=cmp("",[str(S),atom("unmarked"),int(Arrow)|T]),
    !.
    
fromTrees([],[]) :-
    !.
fromTrees([H|T],[Ht|Tt]) :-
    Ht=fromTree(H),
    fromTrees(T,Tt),
    !.
    
toTree(Term)=Tree :-
    Term=cmp(_,[str(S),atom("marked"),int(Arrow)|T]),
    toTrees(T,Trees),
    Ts=termDomain::fromTermStringIndirect(S),
    Ts=str(Selector),
    Tree=vpiTree::tree(Selector,vpiTree::marked,Trees,Arrow),
    !
;    
    Term=cmp(_,[str(S),atom("unmarked"),int(Arrow)|T]),
    toTrees(T,Trees),
    Ts=termDomain::fromTermStringIndirect(S),
    Ts=str(Selector),
    Tree=vpiTree::tree(Selector,vpiTree::unmarked,Trees,Arrow),
    !.    

toTrees([],[]) :-
    !.
toTrees([H|T],[Ht|Tt]) :-
    Ht=toTree(H),
    toTrees(T,Tt),
    !.
% -----------------------------------------------------------------------------------------------
% bool
% -----------------------------------------------------------------------------------------------
bool(1)=b_true.
bool(-1)=b_false.
% -----------------------------------------------------------------------------------------------
% term_sterm; sterm_term
% -----------------------------------------------------------------------------------------------
class predicates
stl : (sterml,globalRefDomains::terml) determ (i,o).
tsl : (globalRefDomains::terml,sterml) determ (i,o).
clauses

term_sterm(globalRefDomains::str(S),str(S)) :- !.
term_sterm(globalRefDomains::atom(S),atom(S)) :- !.
term_sterm(globalRefDomains::int(S),int(S)) :- !.
term_sterm(globalRefDomains::fix(S),fix(S)) :- !.
term_sterm(globalRefDomains::nill,nill) :- !.
term_sterm(globalRefDomains::cmp(S,L),cmp(S,Lm)) :- tsl(L,Lm),!.
term_sterm(globalRefDomains::list(Ha,Ta),list(Hb,Tb)) :- term_sterm(Ha,Hb), term_sterm(Ta,Tb),!.

term_sterm(T,_) :-
	S=toString(T),
	dlgIF::feil("term_sterm",["unrecognizable term:",S],[],nill),!,
	fail.

sterm_term(str(S),globalRefDomains::str(S)) :- !.
sterm_term(atom(S),globalRefDomains::atom(S)) :- !.
sterm_term(int(S),globalRefDomains::int(S)) :- !.
sterm_term(fix(S),globalRefDomains::fix(S)) :- !.
sterm_term(cmp(S,L),globalRefDomains::cmp(S,Lm)) :- stl(L,Lm),!.
sterm_term(nill,globalRefDomains::nill) :- !.
sterm_term(list(Ha,Ta),globalRefDomains::list(Hb,Tb)) :- sterm_term(Ha,Hb), sterm_term(Ta,Tb),!.

sterm_term(T,_) :-
	S=toString(T),
	dlgIF::feil("sterm_term",["unrecognizable term:",S],[],nill),!,
	fail.

tsl([],[]) :- !.
tsl([H|T],[Hm|Tm]) :- term_sterm(H,Hm), tsl(T,Tm),!.
stl([],[]) :- !.
stl([H|T],[Hm|Tm]) :- sterm_term(H,Hm), stl(T,Tm),!.
% --------------------------------------------------------------------------
% 
% --------------------------------------------------------------------------
realToIntT(R)=I:- R2=math::trunc(R), I=convert(integer,R2),!.
realToIntR(R)=I:-R2=math::round(R), I=convert(integer,R2),!.
intToReal(I)=R:- R=convert(real,I),!.

intRefType(int(I),I) :- 
	!.
intRefType(str(S),I) :-
	I=sx::toInt(S),
	!.

strRefType(str(I),I) :- 
	!.
realRefType(fix(F),F) :- 
	!.
realRefType(int(I),F) :- 
	F=I,
	!.
realRefType(str(S),F) :-
    F=sx::toReal(S),
	!.
	
realsRefType([],[]) :-
	!.
realsRefType([H|T],[Hr|Tr]) :- 
	realRefType(H,Hr),
	realsRefType(T,Tr),
	!.

intsRefType([],[]) :-
	!.
intsRefType([H|T],[Hr|Tr]) :- 
	intRefType(H,Hr),
	intsRefType(T,Tr),
	!.

strsRefType([],[]) :-
	!.
strsRefType([H|T],[Hr|Tr]) :- 
	strRefType(H,Hr),
	strsRefType(T,Tr),
	!.
% --------------------------------------------------------------------------
% toStrings
% --------------------------------------------------------------------------
toStrings([],[]) :- !.
toStrings([str(H)|T],[H|Ts]) :- toStrings(T,Ts),!.     
toStrings([atom(H)|T],[H|Ts]) :- toStrings(T,Ts),!.     

toStringsR([],[]) :- !.
toStringsR([str(H)|T],[H|Ts]) :- toStringsR(T,Ts),!.     
end implement conversion
