/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

class dbBase
open core,termDomain,globalDomains

predicates
classInfo : core::classInfo.

toLoc : (string,dbLoc) procedure (i,o).
toLocs : (string_list,dbLoc_list) determ (i,o).
%
% database main
%
setSql : ().
kbIsSql : () determ.
getContext : () -> string_list.
getCurrentSegment : () -> string determ.
setContext : (string_list).
dbName : () -> string determ.
dbFile : (dbloc) -> string determ.
dbId : () -> string determ.
dbId : (dbLoc) -> string determ.
dbIdChain : (string Chain) -> string determ.
dbRef : (dbloc) -> string determ.
dbRefR : (string) -> dbLoc determ.
dbLocChain : (string Chain,string Ref) -> dbLoc determ.
%
% chain-related
%
chainDelete : (string Chain) determ.
chainDeleteOutright : (string Chain) determ. % deletes chain with no integrity checks or graph update (except write-access)
chainDeleteSentences : (integer Type,string Chain,integer N) determ (i,i,o). % type: 0=base_type 1+2,1=base_type 1,2=base_type 2
chainSuspend : (string Chain) determ.
%
chainAccess : (string Chain) -> integer determ.
chains_nd : (integer InclExtChains,string Chain) nondeterm (i,o).
chainCount : (string Chain) -> integer determ.
chainSentences : (string Chain,sterm LT) nondeterm (i,o).
chainSentences : (string Chain,sterm LT,dbLoc Loc) nondeterm (i,o,o).
chainTerms : (string Chain,sterm LT,sterm RT) nondeterm (i,o,o).
chainTerms : (string Chain,sterm LT,sterm RT,dbLoc Loc) nondeterm (i,o,o,o).
chainTermsAll : (string Chain,sterm LT,sterm RT) nondeterm (i,o,o).
chainTermsAll : (string Chain,sterm LT,sterm RT,dbLoc Loc) nondeterm (i,o,o,o).
chainReadOnlyWarning : (string Chain) determ.
chainHasChangedTerm : (sterm) determ.
chainHasChangedChain : (string) determ.
publicChain : (string Chain,integer Access) determ (i,o).
getExtChains : (string_list R,string_list W) procedure (o,o).
chainType : (string Chain,string DbId,integer Type,integer Access) determ (i,o,o,o).
chainExists : (string Chain) determ.
chainIsExternal : (string Chain) determ.
%
% realm-related
%
realmDelete : (string Realm) determ.
realmDelete : (string KbId,string Realm) determ.
realmDeleteNo : (string Realm,integer No) determ.
realmDeleteSubno : (string Realm,integer No,integer Subno) determ.
realmKbDelete : (string_list KbFiles,string Realm) determ.
realmRename : (string From,string To,integer N) procedure (i,i,o).
%
realms : () -> string_list.
realmCount : (string DbId,string Realm) -> integer determ.
realmExists : (string Realm,string_list Files) determ (i,o).
realmNoCount : (string Realm,integer No) -> integer determ.
realmRecords_nd : (string KbId,string Realm) -> dbLoc nondeterm.
realmSubnoCount : (string Realm,integer No,integer Subno) -> integer determ.
%
% SR-related
%
srDerivedDelete : (integer) determ.
srDelete : (string Realm,integer No,integer Sub) determ.
%
getSr : (string Realm,integer No,integer Sub) -> dbLoc_list determ.
srLoc : (string Chain,string Realm,integer No,integer Sub) -> dbLoc determ.
srLocF : (string Chain,string Realm,integer No,integer Sub) -> dbLoc determ.
srLocId : (string KbId,string Realm,integer No,integer Sub) -> dbLoc determ.
srLocIdF : (string KbId,string Realm,integer No,integer Sub) -> dbLoc determ.
%
% IR-related
%
irReplace : (string Realm,sterm Declaration,sterm Definition) determ.
irDeleteAll : (). % deletes all IRs
%
irChain_nd : (string) multi (o).
getIrDef : (string Rule,sterm LTerm,dbLoc Loc) determ (i,o,o).
getIrDefF : (string Rule,sterm LTerm,dbLoc Loc) determ (i,o,o).
irType : (string KbName,string ChainType,string SubType) determ (i,o,o).
irTypeF : (string KbName,string ChainType,string SubType) determ (i,o,o).
getIrs : () -> lst.
getIrsOfSubtype : (string SubType) -> lst.
% 
% Sentence-related
%
% 
% Record-related
%
recordInsert : (integer,string,integer,integer,sterm,sterm) determ.
recordInsert : (integer,string,integer,integer,sterm,sterm,dbLoc) determ (i,i,i,i,i,i,o).
recordInsertOutright : (integer,string,integer,integer,sterm,sterm,dbLoc) determ (i,i,i,i,i,i,o).
recordInsertOutrightIgnoreCache : (integer,string,integer,integer,sterm,sterm,dbLoc) determ (i,i,i,i,i,i,o).
recordDelete : (dbLoc) determ.
recordsDelete : (dbLoc_list) determ.
recordDeleteOutright : (dbLoc) determ.
recordSuspend : (dbLoc) determ.
recordSetType : (dbloc,integer RecType) determ.
recordUnsuspend : (dbLoc) determ.
%
getRecord : (dbLoc,integer RT,integer BT,integer Active,integer Suspended,string Time,string Realm,integer No,integer SubNo,string ST,sterm LTerm,sterm RTerm) determ (i,o,o,o,o,o,o,o,o,o,o,o).
recordIsSuspended : (dbLoc) determ.
recordType : (integer FullType,integer BaseType,integer Active,integer Suspended) determ (i,o,o,o).
% 
% Record set-related
%
sentenceDerivedDelete : (integer) determ.
%
typeCount : (integer RecordType) -> integer determ.
recTypeRefs : (integer) -> lst. % returns a list of all records of given type (format: Ref|Chain)
recTypeRefList : (integer,string_list) procedure (i,o). % returns a list of all records of given type (format: Ref|Chain)
recTypeLocs : (integer RecType,dbLoc_list L) determ (i,o).
recBasetypeLocs : (integer RecType,dbLoc_list L) determ (i,o).
recBasetypeLocsAll : (integer RecType,dbLoc_list L) determ (i,o).
%
% ST-related 
%
stTypeDelete : (integer Basetype,string ST,integer N) determ (i,i,o).
%
stCount : (string ST,integer NType1,integer NType2) determ (i,o,o).
stList : () -> string_list.
stBasetypes : (integer Basetype) -> string_list.
stSentence_nd : (string St,sterm Lt,dbLoc Loc) nondeterm (i,o,o).
stSentence_next : (string St,dbLoc Loc,dbLoc Next) determ (i,i,o).
stSentence_prev : (string St,dbLoc Loc,dbLoc Next) determ (i,i,o).
stNL_KB : (string,string) determ (i,o).
stKB_NL : (string,string) procedure (i,o).
stNL : (string STk,string NL) determ (i,o).
stNL_F : (string STk,string NL) determ (i,o).
stComps : (string ST,integer Card,string Pred,string Ent1,string Ent2) determ (i,o,o,o,o).
kbSTX_NL_KB : (string,string) determ (i,o).
kbSTX_NL : (string STk,string NL) determ (i,o).
stExists : (integer,string) determ (i,i). % succeeds if record-type and ST exists
kbST_DEs : (string,string_list) determ (i,o).
% returns a term's relation argument list
termSt : (sterm,string) determ (i,o).
termStF : (sterm,string) determ (i,o).
querySt : (sterm,string) determ (i,o).
queryStF : (sterm,string) determ (i,o).
querySgNl : (sterm,string) determ (i,o). % returns a SG-query's SG-name in NL
/*
kbST_KB : (sterm,string) determ (i,o).
kbST_KB_Q : (sterm,string) determ (i,o).
kbST_KB_Q_F : (sterm,string) determ (i,o).
kbTermSTQ : (sterm,string_list) determ (i,o).
kbTermSTQ_F : (sterm,string_list) determ (i,o).
*/
kbTermST : (sterm,string_list) determ (i,o).
kbTermSTp : (sterm,string) determ (i,o).
kbTermST_proc : (sterm,string) procedure (i,o).
% 
% SM-related
%
smDef : (string K,string Clas,integer Card,string_list DEs,string Domain,string Format,integer Width) determ (i,o,o,o,o,o,o).
smDef : (string K,string Clas,integer Card,string_list DEs,string Domain,string Format,integer Width,dbLoc Loc) determ (i,o,o,o,o,o,o,o).
smDefF : (string K,string Clas,integer Card,string_list DEs,string Domain,string Format,integer Width) determ (i,o,o,o,o,o,o).
smDefF : (string K,string Clas,integer Card,string_list DEs,string Domain,string Format,integer Width,dbLoc Loc) determ (i,o,o,o,o,o,o,o).
smDef2 : (string Clas,sterml RestBody,integer Card,string_list DEs,string Domain,string Format,integer Width) determ (i,i,o,o,o,o,o).
smClassDef_nd : (string Clas,string KbName,integer Card,string_list DEs,string Domain,string Format,integer Width,string Segment,string ChainType,string Subtype) nondeterm (i,o,o,o,o,o,o,o,o,o).
smClassDef : (string K,string Clas,integer Card,string_list DEs,string Domain,string Format,integer Width) determ (i,i,o,o,o,o,o).
smClassDefF : (string K,string Clas,integer Card,string_list DEs,string Domain,string Format,integer Width) determ (i,i,o,o,o,o,o).
smDeDefs : (string_list,sterml) determ (i,o).
smDefTree : (string,string,sterm) determ (i,o,o).
smDefTreeF : (string,string,sterm) determ (i,o,o).
smClassDefTree : (string,string,sterm) determ (i,i,o).
smClassDefTreeF : (string,string,sterm) determ (i,i,o).
smRef : (string KbName) -> dbLoc Loc determ.
smRefF : (string KbName) -> dbLoc Loc determ.
smSegment : (string KbName) -> string Segment determ.
smStDesSql : (string St,string_list DEs) determ (i,o). % returns list of ST's DEs as SQL column names
smStDes : (string St,string_list DEs) determ (i,o). % returns list of ST's DEs on the format 'et|de' or 'et|cde|mde'
kbDE_Defs : (sterm,sterml) determ (i,o).
kbDE_DefsL : (string_list,sterml) determ (i,o).
kbEntityDEs : (string,string_list,sterml) determ (i,o,o).
kbPredDEs : (string,string_list,sterml) determ (i,o,o).
% 
% Terminology-related
%
syntagmDelete : (string KbName).
%
kbID : (string,string,string,string) determ (i,o,o,o).
kbIDs : (string_list,string_list) determ (i,o).
kbID_F : (string,string,string,string) determ (i,o,o,o).
kbIdLoc : (string) -> dbLoc determ. 
kbSym_to_IF : (string,string) determ (i,o).
kbIF_to_Sym : (string,string) procedure (i,o).
kbIF_to_Syms : (string_list,string_list) determ (i,o).
nlTrys : (string_list KBnames,string_list NLnames) procedure (i,o).
nlSings : (string_list KBnames,string_list NLnames) procedure (i,o).
nlTry : (string KBname,string NLname) procedure (i,o).
nlForm : (string Kname,string Form,string NL) determ (i,i,o).
nlFormF : (string Kname,string Form,string NL) determ (i,i,o).
nlSing : (string Kname,string NL) procedure (i,o).
nlSingF : (string Kname,string NL) determ (i,o).
nlSingPrefixed : (string,string,string) procedure (i,i,o).
nlSingsPrefixed : (string,string_list,string_list) procedure (i,i,o).
syntagm_nd : (string Language,string KbName,string Clas,string Form,string Type,string_list Tokens,string Segment) nondeterm (i,o,o,o,o,o,o) (i,i,o,o,o,o,o).
dbaReservedKB_Id : (string) determ.
makeNewIdentifier : (string,string,string) determ (i,i,o).
kbNameExists : (string KbName) determ. 
mainFormOfSyno : (string SynoNL,string MainNL) determ (i,o).
%
% other
%
indexRegen : () determ.
getTimingEvent : (string,integer,integer) procedure (i,o,o).
insertTimingEvent : (string,integer,integer) determ.
accessErrorChain : (string) determ.
accessErrorTerm : (sterm) determ.
getProgramBuild : (integer VR,integer V,integer R,integer B,string Status) determ (o,o,o,o,o).
isValidId : (string) determ.
flagsClear : () determ.
getDomainsNL : (string_list DEs,string_list Doms) determ (i,o).
joinNL : (string_list L,string Res) procedure (i,o).
joinB_NL : (string_list L,string Res) procedure (i,o).
end class dbBase