/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement dbBase
open core,termDomain,globalDomains

constants
className = "com/rudstrand/dbBase/dbBase".
classVersion = "1.0".

class facts
sql : integer:=0. % 1=sql kb, 0=3kb kb
context : string_list := []. % current context

class predicates        
locFromKba : (string_list,dbLoc_list) determ (i,o).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% makeNewIdentifier
% ----------------------------------------------------------------------------------------------
class predicates
makeNewIdentifierX : (integer,string,integer,string,string) determ (i,i,i,i,o).
clauses

makeNewIdentifier(NL,BlankReplace,Id) :-
    Max=globalVars::get_i("max_kbn"),
    makeNewIdentifierX(1,NL,Max,BlankReplace,Id),
    !.
    
makeNewIdentifierX(N,NL,Max,BlankReplace,Id) :-
    sx::makeIdentifier(2,3,2,NL,Max,"",Id),
    not(kbNameExists(Id)),
    !
;
    Ns=toString(N),
    NL2=string::concat(NL,"_",Ns),
    M=N+1,
    makeNewIdentifierX(M,NL2,Max,BlankReplace,Id),
    !.
% ----------------------------------------------------------------------------------------------
% smStDes
% ----------------------------------------------------------------------------------------------
class predicates
addDe : (integer,string,string_list,string_list,integer) procedure (i,i,i,o,i).
desSql : (string_list,string_list,integer,string,integer) determ (i,o,i,i,i).
clauses

smStDes(St,Des) :-
    sx::sxUnpack(St,'|',L),
    L=[P,Ea,Eb],
    smClassDef(P,"p",_,D1,_,_,_),
    smClassDef(Ea,"e",_,D2,_,_,_),
    smClassDef(Eb,"e",_,D3,_,_,_),
    Pf1=string::concat(P,"|"),
    Pf2=string::concat(Ea,"|"),
    Pf3=string::concat(Eb,"|"),
    sx::sxlAddPrefix(D1,Pf1,D1p),
    sx::sxlAddPrefix(D2,Pf2,D2p),
    sx::sxlAddPrefix(D3,Pf3,D3p),
    set::conc([D2p,D3p,D1p],Des),
    !
;
    sx::sxUnpack(St,'|',L),
    L=[P,Ea],
    smClassDef(P,"p",_,D1,_,_,_),
    smClassDef(Ea,"e",_,D2,_,_,_),
    Pf1=string::concat(P,"|"),
    Pf2=string::concat(Ea,"|"),
    sx::sxlAddPrefix(D1,Pf1,D1p),
    sx::sxlAddPrefix(D2,Pf2,D2p),
    set::conc([D2p,D1p],Des),
    !.
    
smStDesSql(St,Des) :-
    smStDes(St,Ds),
    sqlBase::getParameters(Max,_,_,Usc,Case,_,_,_),
    desSql(Ds,Des,Max,Usc,Case),
    !.
    
desSql([],[],_,_,_) :-
    !.
desSql([H|T],Ts,Max,Usc,Case) :-
    sx::sxUnpack(H,'|',L),
    L=[_,D],
    desSql(T,Ts2,Max,Usc,Case),
    nlSing(D,Dn),
    sx::makeIdentifier(Case,Case,Case,Dn,Max,Usc,K2),
    sqlTools::makeSqlName(K2,Max,K),
    addDe(2,K,Ts2,Ts,Max),
    !
;    
    sx::sxUnpack(H,'|',L),
    L=[_,C,D],
    desSql(T,Ts2,Max,Usc,Case),
    nlSing(C,Cn),
    nlSing(D,Dn),
    Dm=string::concat(Cn,"_",Dn),
    sx::makeIdentifier(Case,Case,Case,Dm,Max,Usc,K2),
    sqlTools::makeSqlName(K2,Max,K),
    addDe(2,K,Ts2,Ts,Max),
    !.

addDe(N,D,S,T,Max) :-
    set::member(D,S),
    string::length(D)=Max,
    Mx=Max-1,
    string::front(D,Mx,D2,_),
    Ns=toString(N),
    D3=string::concat(D2,Ns),
    Next=N+1,
    addDe(Next,D3,S,T,Max),
    !
;
    set::member(D,S),
    Ns=toString(N),
    D2=string::concat(D,Ns),
    Next=N+1,
    addDe(Next,D2,S,T,Max),
    !
;
    T=[D|S],
    !.
% ----------------------------------------------------------------------------------------------
% irReplace
% ----------------------------------------------------------------------------------------------
irReplace(Realm,Declaration,Definition) :-
	Definition=cmp(_,[atom(Rule)|_]),
    smDefF(Rule,Clas,_,_,_,_,_),
    Clas<>"h",
    Mess=string::concat("Inexact rule KB-name is already used for a sentence model element of type: ",Clas,"!"),
    dlgIF::dlgNote(1,Mess),!,fail
;    
    % delete def. if exists
	Definition=cmp(_,[atom(Rule)|_]),
	getIrDefF(Rule,_,Loc),
	recordDelete(Loc),
	fail
;	
    % delete decl. if exists
	Definition=cmp(_,[atom(Rule)|_]),
    Loc=smRefF(Rule),
	recordDelete(Loc),
	fail
;
	recordInsert(2,Realm,0,0,Declaration,atom("true"),_),   
	recordInsert(6,Realm,0,0,Definition,nill,_),
	!.
% ----------------------------------------------------------------------------------------------
% realmRename
% ----------------------------------------------------------------------------------------------
realmRename(From,To,N) :-
    kbIsSql,
    !, 
    FromU=string::toUpperCase(From),
    ToU=string::toUpperCase(To),
    sqlBase::realmRename(FromU,ToU,N),
    globalVars::set_i("knowopen",2),
    !
;
    FromU=string::toUpperCase(From),
    ToU=string::toUpperCase(To),
    dbaLib::realmRename(FromU,ToU,N),
    globalVars::set_i("knowopen",2),
    !.
% ----------------------------------------------------------------------------------------------
% recordIsSuspended
% ----------------------------------------------------------------------------------------------
recordIsSuspended(Loc) :-
    getRecord(Loc,_,_,_,Suspended,_,_,_,_,_,_,_),
    Suspended=1,
    !.
% ----------------------------------------------------------------------------------------------
% toLoc/s
% ----------------------------------------------------------------------------------------------
toLoc(S,L) :- L=toTerm(S),!.

toLocs([],[]) :-
    !.
toLocs([H|T],[Hx|Tx]) :-
    Hx=toTerm(H),
    toLocs(T,Tx),
    !.
% ----------------------------------------------------------------------------------------------
% setSql; knIsSql
% ----------------------------------------------------------------------------------------------
setSql() :- sql:=1,!.
kbIsSql() :- sql=1,!.
% context
%
% ----------------------------------------------------------------------------------------------
% getContext/setContext/getCurrentSegment
% ----------------------------------------------------------------------------------------------
getContext()=L:-L=context.
setContext(L):-context:=L.
getCurrentSegment()=S :- context=[S|_],!.
% -------------------------------------------------------------------------------------
% dbFile
% -------------------------------------------------------------------------------------
dbFile(sloc(_,_))=File :- File=globalVars::get_s("kb_file"),!.
dbFile(cloc(Kb,_))=File :- dbaBase::database(Kb,File,_,_),!.
% -------------------------------------------------------------------------------------
% dbId/Chain
% -------------------------------------------------------------------------------------
dbId()=Id :- 
    kbIsSql,
    !,Id="X",
    !
;
    dbaBase::database(_,_,Id,0),
    !.
    
dbId(sloc(_,_))=Id :- Id="X",!.
dbId(cloc(Kb,_))=Id :- dbaBase::database(Kb,_,Id,_),!.

dbIdChain(Chain)=Id :-
    kbIsSql,
    !,Id="X",
    !
;
    dbaBase::dbaMap(Chain,KB,_),
    dbaBase::database(KB,_,Id,_),
    !.
% -------------------------------------------------------------------------------------
% dbRef
% -------------------------------------------------------------------------------------
dbRef(sloc(Chain,Ref))=RefS :- 
    S=toString(Ref), 
    sx::sxL_Pad(S,"0",9,S2), 
    RefS=string::concat(Chain,"|",S2),
    !.
dbRef(cloc(KB,Ref))=RefS :- 
    dbaBase::dba_address(Ref,S2), 
    dbaBase::database(KB,_,Id,_),
    RefS=string::concat(Id,":",S2),
    !.
    
dbRefR(S)=Loc :-
    string::subChar(S,1)=':',
    sx::sxUnpack(S,':',L),
    L=[Id,SRef],
    dbaBase::database(KB,_,Id,_),
    dbaBase::dba_address(Ref,SRef), 
    Loc=cloc(KB,Ref),
    !
;
    sx::sxUnpack(S,'|',L),
    L=[Chain,SRef],
    Ref=sx::toInt(SRef),
    Loc=sloc(Chain,Ref),
    !.
% -------------------------------------------------------------------------------------
% dbLocChain
% -------------------------------------------------------------------------------------
dbLocChain(Chain,SRef)=Loc :-
    kbIsSql,
    !,Ref=sx::toInt(SRef),
    Loc=sloc(Chain,Ref),
    !
;
    dbaBase::dba_address(Ref,SRef),
    dbaBase::dbaMap(Chain,KB,_),
    Loc=cloc(KB,Ref),
    !.
% -------------------------------------------------------------------------------------
% recordDelete
% -------------------------------------------------------------------------------------
recordDelete(sloc(Chain,Ref)) :-
    cache::check(Chain),
    sqlBase::recordDelete(Chain,Ref),
    !.
recordDelete(cloc(Db,Ref)) :- 
	Db:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,LTerm,_),
    accessErrorTerm(LTerm),
    !,fail
;
	Db:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,LTerm,_),
    LTerm=cmp(Chain,_),
    cache::check(Chain),
    dbaBase::dbaDeleteRecord(Db,Ref),
    !.

recordDeleteOutright(sloc(Chain,Ref)) :- 
    cache::check(Chain),
    sqlBase::recordDeleteOutright(Chain,Ref),
    !.
recordDeleteOutright(cloc(KB,Ref)) :- 
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,LTerm,_),
    LTerm=cmp(Chain,_),
    cache::check(Chain),
    KB:term_delete(Chain,Ref),
    !.
% -------------------------------------------------------------------------------------
% recordInsert
% -------------------------------------------------------------------------------------
recordInsert(RecType,Realm,SentNo,SubNo,LTerm,RTerm) :- recordInsert(RecType,Realm,SentNo,SubNo,LTerm,RTerm,_),!.

recordInsert(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Loc) :-
    kbIsSql,
    !,
	LTerm=cmp(Chain,_),
    cache::check(Chain),
    sqlBase::recordInsert(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref),
    Loc=sloc(Chain,Ref),
    !
;    
	accessErrorTerm(LTerm),
	!,fail
;     
	LTerm=cmp(Chain,_),
    cache::check(Chain),
	dbaBase::dbaInsertRecord(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref),
	dbaBase::dbaMap(Chain,KB,_),
    Loc=cloc(KB,Ref),
    !.
    
recordInsertOutright(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Loc) :-
    kbIssql,
    !,
	LTerm=cmp(Chain,_),
    cache::check(Chain),
    sqlBase::recordInsertOutright(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref),
    Loc=sloc(Chain,Ref),
    !
;    
    LTerm=cmp(Chain,_),
    dbaBase::dbaMap(Chain,KB,_),
	datetime::tsNow(Ts),
    cache::check(Chain),
	Rec=clause(RecType,Ts,SentNo,SubNo,Realm,LTerm,RTerm),
	KB:chain_insertz(Chain,Rec,Ref),
    Loc=cloc(KB,Ref),
    !.

recordInsertOutrightIgnoreCache(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Loc) :-
    kbIssql,
    !,
    LTerm=cmp(Chain,_),
    sqlBase::recordInsertOutright(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref),
    Loc=sloc(Chain,Ref),
    !
;    
    LTerm=cmp(Chain,_),
    dbaBase::dbaMap(Chain,KB,_),
	datetime::tsNow(Ts),
	Rec=clause(RecType,Ts,SentNo,SubNo,Realm,LTerm,RTerm),
	KB:chain_insertz(Chain,Rec,Ref),
    Loc=cloc(KB,Ref),
    !.
% ----------------------------------------------------------------------------------------------
% recordUnsuspend
% ----------------------------------------------------------------------------------------------
recordUnsuspend(Loc) :-
    kbIsSql,
	!, 
	dbBase::getRecord(Loc,_,Base,_,_,_,Realm,No,Sub,_,LT,RT),
	dbBase::recordDeleteOutright(Loc),
	dbBase::recordInsert(Base,Realm,No,Sub,LT,RT,_),
	!
;
	dbBase::getRecord(Loc,_,Base,_,_,_,Realm,No,Sub,_,LT,RT),
	dbBase::recordDelete(Loc),
	dbBase::recordInsert(Base,Realm,No,Sub,LT,RT,_),
	!.	
% ----------------------------------------------------------------------------------------------
% chainDelete/Outright
% ----------------------------------------------------------------------------------------------
chainDeleteOutright(Chain) :-
    kbIsSql,
    !,
    cache::check(Chain),
    sqlBase::chainDeleteOutright(Chain),
    !
;
    chainExists(Chain),
	dbaBase::dbaMap(Chain,KB,_),
    cache::check(Chain),
    KB:chain_delete(Chain),
    !
;
    !.
    
chainDelete(Chain) :-
    kbIsSql,
	!,
    cache::check(Chain),
	sqlBase::chainDelete(Chain),
	!
;
    chainReadOnlyWarning(Chain),
    !,fail
;
	dbaBase::dbaMap(Chain,KB,_),
    cache::check(Chain),
	dbaUtil::dbaDeleteChain(KB,Chain),
	!.
% ----------------------------------------------------------------------------------------------
% sentenceDerivedDelete
% ----------------------------------------------------------------------------------------------
sentenceDerivedDelete(Externals) :-
    kbIsSql,
    !,sqlBase::deleteBasetypeOutright(1),
    % regen. graphs
	dep::setStatus("p3_pDP",int(1),int(0)),
	dep::depSMM_DelDerived,
	!
;
    dbaBase::dbaDeleteDerivedFacts(Externals),
    % regen. graphs
	dep::setStatus("p3_pDP",int(1),int(0)),
	dep::depSMM_DelDerived,
    !.
% ----------------------------------------------------------------------------------------------
% srDerivedDelete
% ----------------------------------------------------------------------------------------------
srDerivedDelete(Mode) :-
    kbIsSql,
    !,sqlBase::deleteBasetypeOutright(5),
	!
;
    dbaBase::dbaDeleteDerivedRules(Mode),
    !.
% ----------------------------------------------------------------------------------------------
% stTypeDelete
% ----------------------------------------------------------------------------------------------
stTypeDelete(Basetype,ST,N) :-
    Basetype=0,
    stTypeDelete(1,ST,N1),
    stTypeDelete(2,ST,N2),
    N=N1+N2,
    !
; 
    Basetype>0,
    kbIsSql,
    sqlBase::stTypeDelete(Basetype,ST,N),
	!
;
    Basetype>0,
    Basetype<=2,
    dbaBase::dbaDeleteST(Basetype,ST,N),
    !
;
    Basetype>2,
    not(kbIsSql),
    S=toString(Basetype),
    Mess=string::concat("Attempt to delete ST for records type ",S," in 3KB-type database!"),
    dlgIF::dlgNote(4,Mess),
    !,fail.
% -------------------------------------------------------------------------------------
% chainSentences/Terms...
% -------------------------------------------------------------------------------------
chainSentences(Chain,LT) :- 
    kbIsSql, 
    sqlBase::chainSentencesDirect(Chain,LT)
;
    not(kbIsSql),
    dbaLib::chainTerms(Chain,LT,RT), 
    RT=atom("true").

chainSentences(Chain,LT,Loc) :- 
    kbIsSql,
    sqlBase::chainSentencesDirect(Chain,LT,Ref),
    Loc=sloc(Chain,Ref)
;
    not(kbIsSql),
    dbaLib::chainTerms(Chain,LT,RT,Loc), 
    RT=atom("true").

chainTerms(Chain,LT,RT) :- chainTerms(Chain,LT,RT,_).

chainTerms(Chain,LT,RT,Loc) :-
    kbIsSql,
    sqlBase::chainTermsDirect(Chain,LT,RT,Ref),
    Loc=sloc(Chain,Ref)
;
    not(kbIsSql),
    dbaLib::chainTerms(Chain,LT,RT,Loc).

chainTermsAll(Chain,LT,RT) :- chainTermsAll(Chain,LT,RT,_).

chainTermsAll(Chain,LT,RT,Loc) :-
    kbIsSql,
    sqlBase::chainTermsDirectAll(Chain,LT,RT,Ref),
    Loc=sloc(Chain,Ref)
;
    not(kbIsSql),
    dbaLib::chainTermsAll(Chain,LT,RT,Loc).
% -------------------------------------------------------------------------------------
% syntagm_nd
% -------------------------------------------------------------------------------------
syntagm_nd(Language,KbName,Clas,Form,Type,Tokens,Segment) :-
    kbIsSql,
    sqlBase::chainSentencesDirect("v",LT),
        LT=cmp(_,[cmp(Type,[atom(Ln),str(KbName),atom(Clas),atom(Form),Toks]),str(Segment)]),
        Ln=Language,
        termDomain::listBR(Toks,L),
        conversion::toStrings(L,Tokens)
; 
    not(kbIsSql),
    dbaBase::dbaMap("v",KB,_),
    KB:chain_terms("v",Rec,_),
        Rec=clause(2,_,_,_,_,LT,_),
        LT=cmp(_,[cmp(Type,[atom(Ln),str(KbName),atom(Clas),atom(Form),Toks]),str(Segment)]),
        Ln=Language,
        termDomain::listBR(Toks,L),
        conversion::toStrings(L,Tokens).
% -------------------------------------------------------------------------------------
% kbID/IDs/_F
% -------------------------------------------------------------------------------------
% finds the KB name of a NL string (un-tokenized NL string) representing a syntagm.
% fails if no internal representation exists
%
kbIDs([],[]) :-
	!.
kbIDs([H|T],[Hi|Ti]) :-
	kbID(H,Hi,_,_),
	kbIDs(T,Ti),
	!.
		
kbID(NL,KbId,Clas,Form) :- 
	kbID_F(NL,KbId,Clas,Form),
	!
;
    Mess=string::concat("Unable to translate: ",NL),
    dlgIF::dlgNote(2,Mess),!,fail.

kbID_F(NL,KbId,Clas,Form) :- 
	Lng=globalVars::get_s("lang"),
	Lng=cache::tmCacheIsLoaded(),
    !,cache::getSyntagm(NL,KbId,Clas,Form),
    !
;    
    kbIsSql,
	!,Lng=globalVars::get_s("lang"),
	pscan::pscNL(NL,Toks),
	pscan::str_tokens(Toks,SToks1),
	termDomain::str_slist(SToks1,SToks2),
	termDomain::cls_list(SToks2,Tokens),
    sqlBase::chainSentencesDirect("v",LT),
	LT=cmp(_,[cmp(Typ,[atom(Lng),str(KbId),atom(Clas),atom(Form),TokensX]),_]),
	set::member(Typ,["t","s"]),
	Tokens=TokensX,
	!
;	
	kbBase::kbID_RefF(NL,KbId,Clas,Form,_,_),
	!.
	
kbIdLoc(NL)=Loc :-
    kbIsSql,
    !,Lng=globalVars::get_s("lang"),
    pscan::pscNL(NL,Toks),
    pscan::str_tokens(Toks,SToks1),
    termDomain::str_slist(SToks1,SToks2),
    termDomain::cls_list(SToks2,Tokens),
    sqlBase::chainSentencesDirect("v",LT,Ref),
    LT=cmp(_,[cmp(Typ,[atom(Lng),_,_,_,TokensX]),_]),
    set::member(Typ,["t","s"]),
    Tokens=TokensX,
    Loc=sloc("v",Ref),
    !
;
    kbBase::kbID_Ref(NL,_,_,_,Ref,_),
    dbaBase::dbaMap("v",KB,_),
    Loc=cloc(KB,Ref),
    !.
% ----------------------------------------------------------------------------------------------
% kbSym_to_IF; kbIF_to_Sym/s
% ----------------------------------------------------------------------------------------------
class predicates
kbSym_to_IFx : (string_list,string_list) determ (i,o).
kbIF_SymX : (string_list,string_list) determ (i,o).
clauses

% parameters:
% 1	in	inputted symbol w or wout quotes
% 2	out	IF of symbol
%
kbSym_to_IF(Sym,Sif) :-
	sx::sxUnquote(Sym,S2),
	sx::sxUnpack(S2,'|',L),
	kbSym_to_IFx(L,Li),
	sx::sxlJoin(Li,"|",Sif),
	!.

kbSym_to_IFx([],[]) :-
	!.
kbSym_to_IFx([H|T],[Hi|Ti]) :-
	kbID_F(H,Hi,_,_),
	kbSym_to_IFx(T,Ti),
	!
;
    kbIsSql,
	stNL_KB(H,Hi),
	kbSym_to_IFx(T,Ti),
	!
;	
    not(kbIsSql),
	stNL_KB(H,Hi),
	kbSym_to_IFx(T,Ti),
	!
;	
	Hi=H,
	kbSym_to_IFx(T,Ti),
	!.
	
kbIF_to_Sym(Sif,Sym) :-			
	% interpretation: ST (or SG)
	sx::searchchar(Sif,'|',_),
	kbIsSql,
	SymU=sqlBase::nlST_F(Sif),
	sx::sxQuote(SymU,Sym),
	!
;
	% interpretation: ST (or SG)
	sx::searchchar(Sif,'|',_),
	not(kbIsSql),
	stNL_F(Sif,SymU),
	sx::sxQuote(SymU,Sym),
	!
;
	% interpretation: SR
	sx::sxUnpack(Sif,'|',L),
	L=[STk,Realm,Nos,SubNos],
	nlSingF(STk,ST),
	_=sx::toInt(Nos),
	_=sx::toInt(SubNos),
	sx::sxlJoin([ST,Realm,Nos,SubNos],"|",Sym),
	!
;
	% interpretation: concatenated symbols
	sx::searchchar(Sif,'|',_),
	sx::sxUnpack(Sif,'|',L),
	kbIF_SymX(L,Ls),
	sx::sxlJoin(Ls,"|",SymU),
	sx::sxQuote(SymU,Sym),
	!
;
	% interpretation: single identifier
	nlSingF(Sif,SymU),
	sx::sxQuote(SymU,Sym),
	!
;
	sx::sxQuote(Sif,Sym),
	!.
	
% parameters:
% 1	in	(a) list of symbols (IF)
%               (b) list of NODECLASS|NODENAME (IF)
% 2	out	(a) list of symbols (NL)
%               (b) list of NODECLASS: NODENAME (NL)
% NODENAME is a symbol
% NODECLASS is an identifier
%	
kbIF_to_Syms([],[]) :-
	!.
kbIF_to_Syms([H|T],[Hn|Tn]) :-
	sx::searchchar(H,'{',P),
	N=P-1,
	sx::frontstr(N,H,Clas,S),
	sx::frontchar(S,_,Node),
	nlSing(Clas,ClasN),
	kbIF_to_Sym(Node,NodeN),
	sx::sxlJoin([ClasN,NodeN],": ",Hn),
	kbIF_to_Syms(T,Tn),
	!
;
	kbIF_to_Sym(H,Hn),
	kbIF_to_Syms(T,Tn),
	!.
		
kbIF_SymX([],[]) :-
	!.
kbIF_SymX([H|T],[Hs|Ts]) :-
	nlTry(H,Hs),
	kbIF_SymX(T,Ts),
	!.
% -----------------------------------------------------------------------------------------------
% kbNLTry/Trys/Sings
% -----------------------------------------------------------------------------------------------
class predicates
clauses

nlTrys([],[]) :- 
    !.
nlTrys([H|T],[Hn|Tn]) :-
    nlTry(H,Hn),
    nlTrys(T,Tn),
    !.

nlSings([],[]) :- 
    !.
nlSings([H|T],[Hn|Tn]) :-
    nlSing(H,Hn),
    nlSings(T,Tn),
    !.
    
nlTry(K,NL) :-
    _=cache::tmCacheIsLoaded(),
    NL=cache::getNl(K),
    !
;
    dbBase::nlSingF(K,NL),
    !
;
    sx::sxUnpack(K,'|',L),
    L=[P,A1],
    A1="_",
    nlSingF(P,Pn),
    NL=string::concat("X ",Pn),
    !
;
    sx::sxUnpack(K,'|',L),
    L=[P,A1,A2],
    A1="_",
    A2="_",
    nlSingF(P,Pn),
    NL=string::concat("X ",Pn," Y"),
    !
;
    sx::sxUnpack(K,'|',L),
    L=[P,A1,A2],
    A1="_",
    nlSingF(P,Pn),
    nlSingF(A2,An),
    NL=string::concat("X ",Pn," ",An),
    !
;
    sx::sxUnpack(K,'|',L),
    L=[P,A1,A2],
    A2="_",
    nlSingF(P,Pn),
    nlSingF(A1,An),
    NL=string::concat(An," ",Pn," X"),
    !
;
    sx::sxUnpack(K,'|',L),
    L=[P,A1],
    nlSingF(P,Pn),
    nlSingF(A1,A1n),
    NL=string::concat(A1n," ",Pn),
    !
;
    sx::sxUnpack(K,'|',L),
    L=[P,A1,A2],
    nlSingF(P,Pn),
    nlSingF(A1,A1n),
    nlSingF(A2,A2n),
    NL=string::concat(A1n," ",Pn," ",A2n),
    !
;
    kbIF_to_Sym(K,NL),
    !.
% -----------------------------------------------------------------------------------------------
% nlForm 
% -----------------------------------------------------------------------------------------------
% parameters:
% 1 in		KB name
% 2 in		Form
% 3 out		NL (with given form)
%
nlForm(Rep,Form,NL) :- 
    nlFormF(Rep,Form,NL),
    !
;
    Mess=string::concat("No NL-name for: ",Rep," (",Form,")!"),
    dlgIF::dlgNote(2,Mess),!,fail.

nlFormF(Rep,Form,NL) :- 
    _=cache::tmCacheIsLoaded(),
    !,
    NL=cache::getNl(Rep,Form),
    !
;    
    kbIsSql,
    !,
    NL=sqlBase::nlFormF(Rep,Form),
    !
;    
	kbBase::kbNL_RefF(Rep,Form,NL,_),
	!.
% -----------------------------------------------------------------------------------------------
% nlSing/F
% -----------------------------------------------------------------------------------------------
% parameters:
% 1 in		KB name
% 2 out		NL (with pref. singular form or nill)
%
nlSing(Rep,NL) :-
	nlSingF(Rep,NL),
	!
;
	dlgIF::feil("nlSing",[Rep],[],nill),
	NL="",
	!.
	
nlSingF("","") :-
	!.
nlSingF(Rep,NL) :-
	vars::varFuncParR(Func,DE,Rep),
	Func="p3_iThis",
	nlSingF(Func,S2),
	nlSingF(DE,S4),
	sx::concat([S2," ",S4],NL),
	!
;
	vars::varFuncParR(Func,DE,Rep),
	nlSingF(Func,S2),
	nlSingF(DE,S4),
	sx::concat([S2,":",S4],NL),
	!
;
	sx::searchchar(Rep,'|',_),
	stNL_F(Rep,NL),
	!
;
	nlFormF(Rep,"p3_iSnu",NL),
	!
;
	nlFormF(Rep,"p3_iSnd",NL),
	!
;
	nlFormF(Rep,"p3_iNIL",NL),
	!.
% -----------------------------------------------------------------------------------------------
% nlSingPrefixed
% -----------------------------------------------------------------------------------------------
% parameters:
% 1 in		prefix
% 2 in		KB name with or without prefix p#1
% 3 out		NL (with pref. singular form or nill) prefixed with p#1
%
nlSingPrefixed(Prefix,H,NL) :-
	sx::searchstring(H,Prefix,P),
	P=1,
	L=string::length(Prefix),
	sx::frontstr(L,H,_,Ht),
	nlSing(Ht,Hm),
	NL=string::concat(Prefix,Hm),
	!
;
	nlSing(H,NL),
	!	
;
	dlgIF::feil("nlSingPrefixed",[H],[],nill),
	NL="",
	!.
nlSingsPrefixed(_,[],[]) :- 
	!.
nlSingsPrefixed(Prefix,[H|T],[Hn|Tn]) :-
	nlSingPrefixed(Prefix,H,Hn),
	nlSingsPrefixed(Prefix,T,Tn),
	!.
% ----------------------------------------------------------------------------------------------
% stNL_KB
% ----------------------------------------------------------------------------------------------
class predicates
kbSTx : (pie_scan::wrds,string) determ (i,o).
clauses
% parameters:
% 1	in	ST (NL)
% 2	out	ST (IF)
% fails if unknown input
%
stNL_KB(NL,K) :-
	transnl::scnNL(NL,Wrds),
	kbSTx(Wrds,K),
	!.

stKB_NL(K,NL) :-
    nlSing(K,NL),
	!.

kbSTx(Ws,K) :-
	Ws=[pie_scan::w(_,Eka,"p3_iET",_),pie_scan::w(_,P,"p3_iPT",_),pie_scan::w(_,Ekb,"p3_iET",_)],
	sx::sxlJoin([P,Eka,Ekb],"|",K),
	!
;	
	Ws=[pie_scan::w(_,Eka,"p3_iET",_),pie_scan::w(_,P,"p3_iPT",_)],
	sx::sxlJoin([P,Eka],"|",K),
	!.
% ----------------------------------------------------------------------------------------------
% stNL
% ----------------------------------------------------------------------------------------------
% converts either (a) ST (PRED|ENT1[|ENT2]) or (b) SG to NL
%	
stNL(ST,NL) :-
	stNL_F(ST,NL),
	!
;
	dlgIF::feil("stNL",[ST],[],nill),!,fail.
		
stNL_F(ST,NL) :-
	not(_=string::searchchar(ST,'|',string::caseSensitive())),
	nlSing(ST,NL),
	!
;	
	stComps(ST,Card,P,E1,E2),
	Card=2,
	nlSingF(P,S1),
	nlSingF(E1,S2),
	nlSingF(E2,S3),
	sx::concat([S2," ",S1," ",S3],NL),
	!
;
	stComps(ST,Card,P,E1,_),
	Card=1,
	nlSingF(P,S1),
	nlSingF(E1,S2),
	sx::concat([S2," ",S1],NL),
	!.
	
kbSTX_NL(ST,NL) :-
    sx::sxUnpack(ST,'|',L),
    L=[P,E,V],
    V="_",
    nlSingF(P,Pn),
    nlSingF(E,En),
    NL=string::concat(En," ",Pn," X"),
    !
;    
    sx::sxUnpack(ST,'|',L),
    L=[P,V,E],
    V="_",
    nlSingF(P,Pn),
    nlSingF(E,En),
    NL=string::concat("X ",Pn," ",En),
    !.
% ----------------------------------------------------------------------------------------------
% stComps
% ----------------------------------------------------------------------------------------------
% unpacks an ST text (PRED|ENT1[|ENT2])
%	
stComps(ST,Card,P,E1,E2) :-
	sx::sxUnpack(ST,'|',L),
	L=[P,E1,E2],
	Card=2,
	!
;	
	sx::sxUnpack(ST,'|',L),
	L=[P,E1],
	E2="",
	Card=1,
	!.
% ----------------------------------------------------------------------------------------------
% kbSTX_NL_KB
% ----------------------------------------------------------------------------------------------
class predicates
kbSTXx : (pie_scan::wrds,string) determ (i,o).
clauses
% parameters:
% 1	in	ST with open variable (NL)
% 2	out	ST (IF)
% fails if unknown input
%
kbSTX_NL_KB(NL,K) :-
	transnl::scnNL(NL,Wrds),
	kbSTXx(Wrds,K),
	!.
kbSTXx(Ws,K) :-
	Ws=[pie_scan::w(_,_,"p3_dvar",_),pie_scan::w(_,P,"p3_iPT",_),pie_scan::w(_,Ekb,"p3_iET",_)],
	sx::sxlJoin([P,"_",Ekb],"|",K),
	!
;	
	Ws=[pie_scan::w(_,Ekb,"p3_iET",_),pie_scan::w(_,P,"p3_iPT",_),pie_scan::w(_,_,"p3_dvar",_)],
	sx::sxlJoin([P,Ekb,"_"],"|",K),
	!.
% ----------------------------------------------------------------------------------------------
% stCount
% ----------------------------------------------------------------------------------------------
stCount(ST,N1,N2) :-
    kbIsSql,
    !,N1=sqlBase::getST_Count(ST,1),
    N2=sqlBase::getST_Count(ST,2),
    !
;
	stComps(ST,Card,P,E1,E2),
	C1=count::newCounter(0),
	C2=count::newCounter(0),
	dbaBase::dbaMapST(ST,KB,_),
	kbBase::kbST_CountX(KB,Card,P,E1,E2,C1,C2),
	N1=C1:get(),
	N2=C2:get(),
    !.   
% ----------------------------------------------------------------------------------------------
% stExists
% ----------------------------------------------------------------------------------------------
% succeeds if records of ST (facts) exist, else fails
% parameters:
% 1	in	1=only consider derived facts; 2=only consider source facts; 3=1+2
% 2	in	ST
%
stExists(RecType,ST) :-
    RecType=1,
    kbIsSql,
    !, _=sqlBase::existStD(ST),
    !
;    
    RecType=2,
    kbIsSql,
    !, _=sqlBase::existStI(ST),
    !
;    
    RecType=3,
    kbIsSql,
    !, _=sqlBase::existSt(ST),
    !
;    
	stComps(ST,Card,P,E1,E2),
	dbaBase::dbaMapST(ST,KB,_),
	kbBase::kbST_Exists3kb(KB,RecType,Card,P,E1,E2),
	!.
% ----------------------------------------------------------------------------------------------
% smClassDef_nd
% ----------------------------------------------------------------------------------------------
class predicates
smClassDetails : (string,string,integer,string_list,string,string,integer,string,string) determ (i,i,o,o,o,o,o,o,o).
clauses

smClassDef_nd(Clas,KbName,Card,DEs,Domain,Format,Width,Segment,ChainType,Subtype) :-
    cache::smCacheIsLoaded,
    cache::getSmClass_nd(Clas,KbName,Card,DEs,Domain,Format,Width,Segment,ChainType,Subtype)
;
    not(cache::smCacheIsLoaded),
    chainSentences("q",LT),
        LT=cmp(_,[cmp(Clas,[atom(KbName)|_]),str(Segment)]),
        smClassDetails(Clas,KbName,Card,DEs,Domain,Format,Width,ChainType,Subtype).        
    
smClassDetails(Clas,KbName,Card,DEs,Domain,Format,Width,ChainType,Subtype) :-  
    Clas="h",
    smClassDef(KbName,Clas,Card,DEs,Domain,Format,Width),
    irTypeF(KbName,ChainType,SubType),
    !
;        
    smClassDef(KbName,Clas,Card,DEs,Domain,Format,Width),
    ChainType="",
    Subtype="",
    !.  
% ----------------------------------------------------------------------------------------------
% smDeDefs
% ----------------------------------------------------------------------------------------------
class predicates
totalWidth : (sterml,integer) determ (i,o).    
clauses

smDeDefs([],[]) :-
    !.
smDeDefs([D|Ds],[Def|Defs]) :-
    smClassDefF(D,"a",_,_,Domain,Format,Width),
    Def=fstr(D,Domain,Format,Width),
    smDeDefs(Ds,Defs),
    !
;
    Mess=string::concat("Error in fetching DE definition for: ",D),
    dlgIF::dlgNote(1,Mess),
    !,fail.
    
totalWidth([],0) :-
    !.
totalWidth([fstr(_,_,_,W)|T],SumW) :-
    totalWidth(T,W2),
    SumW=W2+W,
    !.
% ----------------------------------------------------------------------------------------------
% smClassDefTree/smDefTree
% ----------------------------------------------------------------------------------------------
smDefTree(K,Type,Def) :-
    smDefTreeF(K,Type,Def),
    !
;
    Mess=string::concat("Unknown SM-object: ",K),
    dlgIF::dlgNote(2,Mess),
    !,fail.
    
smDefTreeF(K,Type,Def) :-
    smDefF(K,Type,_,_,Domain,Format,Width),
    Type="a",
    Def=fstr(K,Domain,Format,Width),
    !
; 
    smDefF(K,Type,_,DEs,_,_,_),
    Type="e",
    smDeDefs(DEs,DE_Defs),
    Def=cmp("ENT_TYPE",[cmp("REP",[atom(K)]),cmp("ATTRS",DE_Defs)]),
    !
;    
    smDefF(K,Type,_,DEs,Domain,_,_),
    set::member(Type,["f","g","h","i"]),
    smDeDefs(DEs,DE_Defs),
    Def=cmp("REL_TYPE",
	             [cmp("REP",[atom(K)]),
	              cmp("DOM",[atom(Domain)]),
	              cmp("FPAR",DE_Defs)]),
     !
;    
    smDefF(K,Type,Card,DEs,_,_,_),
    Type="p",
    DEs=[],
    !,
    Def=cmp("REL_TYPE",
	             [cmp("REP",[atom(K)]),
	              cmp("CARD",[int(Card)])]),
    !
;    
    smDefF(K,Type,Card,DEs,_,_,_),
    Type="p",
    smDeDefs(DEs,DE_Defs),
    Def=cmp("REL_TYPE",
	             [cmp("REP",[atom(K)]),
	              cmp("CARD",[int(Card)]),
	              cmp("ATTRS",DE_Defs)]),
    !.

smClassDefTree(K,Type,Def) :-
    smClassDefTreeF(K,Type,Def),
    !
;
    Mess=string::concat("Unknown SM-object: ",K," of type: ",Type),
    dlgIF::dlgNote(2,Mess),
    !,fail.
    
smClassDefTreeF(K,Type,Def) :-
    Type="a",
    smClassDefF(K,"a",_,_,Domain,Format,Width),
    Def=fstr(K,Domain,Format,Width),
    !
; 
    Type="e",
    smClassDefF(K,"e",_,DEs,_,_,_),
    smDeDefs(DEs,DE_Defs),
    Def=cmp("ENT_TYPE",[cmp("REP",[atom(K)]),cmp("ATTRS",DE_Defs)]),
    !
;    
    set::member(Type,["f","g","h","i"]),
    smClassDefF(K,Type,_,DEs,Domain,_,_),
    smDeDefs(DEs,DE_Defs),
    Def=cmp("REL_TYPE",
	             [cmp("REP",[atom(K)]),
	              cmp("DOM",[atom(Domain)]),
	              cmp("FPAR",DE_Defs)]),
     !
;    
    Type="p",
    smClassDefF(K,"p",Card,DEs,_,_,_),
    DEs=[],
    !,
    Def=cmp("REL_TYPE",
	             [cmp("REP",[atom(K)]),
	              cmp("CARD",[int(Card)])]),
    !
;    
    Type="p",
    smClassDefF(K,"p",Card,DEs,_,_,_),
    smDeDefs(DEs,DE_Defs),
    Def=cmp("REL_TYPE",
	             [cmp("REP",[atom(K)]),
	              cmp("CARD",[int(Card)]),
	              cmp("ATTRS",DE_Defs)]),
    !.
% ----------------------------------------------------------------------------------------------
% smRef
% ----------------------------------------------------------------------------------------------
smRef(K)=Loc :-
    Loc=smRefF(K),
    !
;
    Mess=string::concat("Unknown sentence model object: ",K,"!"),
    dlgIF::dlgNote(2,Mess),!,fail.
    
smRefF(K)=Loc :-
    kbIsSql,
    !,sqlBase::chainSentencesDirect("q",LT,Ref),
	LT=cmp(Chain,[cmp(_,[atom(K)|_]),_]),
    Loc=sloc(Chain,Ref),
    !
;    
    kbBase::kbSM_RefF(K,KB,Ref),
    Loc=cloc(KB,Ref),
    !.
% ----------------------------------------------------------------------------------------------
% smClassDef
% ----------------------------------------------------------------------------------------------
smClassDef(K,Clas,Card,DEs,Domain,Format,Width) :-
    smClassDefF(K,Clas,Card,DEs,Domain,Format,Width),
    !
;
    Mess=string::concat("Unknown ",Clas,"-type sentence model object: ",K),
    dlgIF::dlgNote(2,Mess),!,fail.
    
smClassDefF(K,Clas,Card,DEs,Domain,Format,Width) :-
    cache::smCacheIsLoaded,
    !,cache::getSm(K,Clas,Card,DEs,Domain,Format,Width,_),
    !
;    
    kbIsSql,
    !,sqlBase::chainSentencesDirect("q",LT),
	LT=cmp(_,[cmp(Cls,[atom(Rep)|T]),_]),
	K=Rep,
	Clas=Cls,
	smDef2(Clas,T,Card,DEs,Domain,Format,Width),
	!
;   
    Clas="a",
    kbBase::kbDE_DefF(K,Def,_),
    Def=fstr(_,Domain,Format,Width),
    Card=0,
    DEs=[],
    !
;
    Clas="e",
    kbBase::kbEntityDefF(K,_,DEs,_),
	set::size(DEs,Width),
	Card=0,
	Domain="",
	Format="",
    !
;
    kbBase::funcDefF(K,Cls,Domain,_,DEs,_),
    Cls=Clas,
	set::size(DEs,Width),
	Card=0,
	Format="",
    !    
;
    Clas="h",
    kbBase::irDeclarationF(K,Domain,_,DEs,_),
	set::size(DEs,Width),
	Card=0,
	Format="",
    !
;
    Clas="i",
    kbBase::kbPredDefF(K,Cls,_,Def,DEs,_),
    Cls=Clas,
    Def=cmp("REL_TYPE",[_,cmp(_,[atom(Domain)]),_]),
	set::size(DEs,Width),
	Card=0,
	Format="",
	!
;
    Clas="p",
    kbBase::kbPredDefF(K,Cls,Card,_,DEs,_),
    Cls=Clas,
	set::size(DEs,Width),
	Domain="",
	Format="",
	!.
% ----------------------------------------------------------------------------------------------
% smDef
% ----------------------------------------------------------------------------------------------    
smDef(K,Clas,Card,DEs,Domain,Format,Width) :-
    smDefF(K,Clas,Card,DEs,Domain,Format,Width),
    !
;
    Mess=string::concat("Unknown sentence model object: ",K),
    dlgIF::dlgNote(2,Mess),!,fail.
    
smDef(K,Clas,Card,DEs,Domain,Format,Width,Loc) :-
    smDefF(K,Clas,Card,DEs,Domain,Format,Width,Loc),
    !
;
    Mess=string::concat("Unknown sentence model object: ",K),
    dlgIF::dlgNote(2,Mess),!,fail.
    
smDefF(K,Clas,Card,DEs,Domain,Format,Width) :-
    cache::smCacheIsLoaded,
    cache::getSm(K,Clas,Card,DEs,Domain,Format,Width,_),
    !
;    
    smDefF(K,Clas,Card,DEs,Domain,Format,Width,_),
    !.
smDefF(K,Clas,Card,DEs,Domain,Format,Width,Loc) :-
    kbIsSql,
    !,sqlBase::chainSentencesDirect("q",LT,Ref),
	LT=cmp(_,[cmp(Clas,[atom(Rep)|T]),_]),
	K=Rep,
	smDef2(Clas,T,Card,DEs,Domain,Format,Width),
	Loc=sloc("q",Ref),
	!
;   
    kbBase::kbDE_DefF(K,Def,Ref),
    Clas="a",
    Def=fstr(_,Domain,Format,Width),
    Card=0,
    DEs=[],
    dbaBase::dbaMap("q",KB,_),
    Loc=cloc(KB,Ref),
    !
;
    kbBase::kbEntityDefF(K,_,DEs,Ref),
    Clas="e",
	set::size(DEs,Width),
	Card=0,
	Domain="",
	Format="",
    dbaBase::dbaMap("q",KB,_),
    Loc=cloc(KB,Ref),
    !
;
    kbBase::funcDefF(K,_,Domain,_,DEs,Ref),
    Clas="f",
	set::size(DEs,Width),
	Card=0,
	Format="",
    dbaBase::dbaMap("q",KB,_),
    Loc=cloc(KB,Ref),
    !    
;
    kbBase::funcDefF(K,_,Domain,_,DEs,Ref),
    Clas="g",
	set::size(DEs,Width),
	Card=0,
	Format="",
    dbaBase::dbaMap("q",KB,_),
    Loc=cloc(KB,Ref),
    !
;
    kbBase::irDeclarationF(K,Domain,_,DEs,Ref),
    Clas="h",
	set::size(DEs,Width),
	Card=0,
	Format="",
    dbaBase::dbaMap("q",KB,_),
    Loc=cloc(KB,Ref),
    !
;
    kbBase::kbPredDefF(K,_,Card,_,DEs,Ref),
    Clas="p",
	set::size(DEs,Width),
	Domain="",
	Format="",
    dbaBase::dbaMap("q",KB,_),
    Loc=cloc(KB,Ref),
	!.

smDef2(Clas,T,Card,DEs,Domain,Format,Width) :-
    Clas="a",
    T=[atom(Domain),str(Format),int(Width)|_],
    Card=0,
    DEs=[],
    !
;
    Clas="e",
    T=[Ds|_],
	termDomain::cla_listR(DEs,Ds),
	set::size(DEs,Width),
	Card=0,
	Domain="",
	Format="",
    !
;
    Clas="f",
    T=[atom(Domain),Ds|_],
	termDomain::cla_listR(DEs,Ds),
	set::size(DEs,Width),
	Card=0,
	Format="",
    !
;
    Clas="g",
    T=[atom(Domain),Ds|_],
	termDomain::cla_listR(DEs,Ds),
	set::size(DEs,Width),
	Card=0,
	Format="",
    !
;
    Clas="h",
    T=[atom(Domain),Ds|_],
	termDomain::cla_listR(DEs,Ds),
	set::size(DEs,Width),
	Card=0,
	Format="",
    !
;
    Clas="i",
    T=[atom(Domain),Ds|_],
	termDomain::cla_listR(DEs,Ds),
	set::size(DEs,Width),
	Card=0,
	Format="",
    !
;
    Clas="p",
    T=[int(Card),Ds|_],
	termDomain::cla_listR(DEs,Ds),
	set::size(DEs,Width),
	Domain="",
	Format="",
    !.
% ----------------------------------------------------------------------------------------------
% sxlJoinB_NL; sxlJoinNL
% ----------------------------------------------------------------------------------------------
% joins members in list according to the general list syntax (A,B,C and D)
%
joinNL(L,S) :-
	sx::sxlRemBlanks(L,L1),
	joinB_NL(L1,S),
	!.
	
joinB_NL([],"") :- 
    !.
joinB_NL([S],S) :- 
    !.
joinB_NL([S1,S2],S) :- 
	nlSing("p3_iAnd",Sand),
	sx::concat([S1," ",Sand," ",S2],S),
	!.
joinB_NL([H|T],S) :- 
    joinB_NL(T,S1), 
    sx::concat([H,", ",S1],S),
    !.
% -----------------------------------------------------------------------------------------------
% getDomainsNL
% -----------------------------------------------------------------------------------------------
getDomainsNL([],[]) :-
    !.
getDomainsNL([DE|Ds],[Dom|Doms]) :-
    sx::sxUnpack(DE,'|',L),
    L=[_,MDE],
    smClassDef(MDE,"a",_,_,D,_,_),
    nlSing(D,Dom),
    getDomainsNL(Ds,Doms),
    !
;
    smClassDef(DE,"a",_,_,D,_,_),
    nlSing(D,Dom),
    getDomainsNL(Ds,Doms),
    !.
% ----------------------------------------------------------------------------------------------
% flagsClear
% ----------------------------------------------------------------------------------------------
flagsClear() :-
    globalVars::get_i("error_flag")=1,
    dlgIF::dlgNote(1,"Updates are not allowed when the ERROR flag is ON!"),
    
    !,fail
;    
    globalVars::get_i("quit_flag")=1,
    dlgIF::dlgNote(1,"Updates are not allowed when the QUIT flag is ON!"),
    !,fail
;
    !.
% ----------------------------------------------------------------------------------------------
% dbaReservedKB_Id
% ----------------------------------------------------------------------------------------------
% succeeds if the given syntagm/synonym is RESERVED
% fails if NOT RESERVED
%
dbaReservedKB_Id(Id) :-
	string::length(Id)=1,
	sx::concat(["All one-letter KB names are restricted to PRO3xxxx realms. You can not use '",Id,"' as requested."],S),
	dlgIF::feil("dbaInsertRecord",[S],[],nill),!
;	
	sx::frontstr(3,Id,F,_),
	F="p3_",
	sx::concat(["All KB names starting with 'p3_' are restricted to PRO3xxx-realms. You can not use '",Id,"' as requested."],S),
	dlgIF::feil("dbaInsertRecord",[S],[],nill),
	!.

% -------------------------------------------------------------------------------------
% kbST_DEs                                                         
% -------------------------------------------------------------------------------------
kbST_DEs(ST,Ds) :-
    sx::sxUnpack(ST,'|',L),
    L=[P,Ea,Eb],
    smClassDefF(Ea,"e",_,DEs1,_,_,_),
    smClassDefF(Eb,"e",_,DEs2,_,_,_),
    smClassDefF(P,"p",_,DEs3,_,_,_),
    set::conc([DEs1,DEs2,DEs3],Ds),
    !
;    
    sx::sxUnpack(ST,'|',L),
    L=[P,Ea],
    smClassDefF(Ea,"e",_,DEs1,_,_,_),
    smClassDefF(P,"p",_,DEs3,_,_,_),
    set::conc([DEs1,DEs3],Ds),
    !.
% ----------------------------------------------------------------------------------------------
% kbDE_Defs/DE_DefsL
% ----------------------------------------------------------------------------------------------
kbDE_Defs(nill,[]) :- 
	!.
kbDE_Defs(list(atom(Rep),Reps),[Def|Defs]) :-
	smDefTree(Rep,Type,Def),
	Type="a",
	kbDE_Defs(Reps,Defs),
	!.

kbDE_DefsL([],[]) :-
	!.
kbDE_DefsL([Rep|Reps],[Def|Defs]) :-
	smDefTree(Rep,Type,Def),
	Type="a",
	kbDE_DefsL(Reps,Defs),
	!.
% -------------------------------------------------------------------------------------
% kbEntityDEs                                                
% -------------------------------------------------------------------------------------
kbEntityDEs(K,DEs,DE_Defs) :-
	smClassDef(K,"e",_,DEs,_,_,_),
	smClassDefTree(K,"e",Def),
	Def=cmp(_,[_,cmp("ATTRS",DE_Defs)]),
	!.
% -------------------------------------------------------------------------------------
% kbPredDEs                                                
% -------------------------------------------------------------------------------------
kbPredDEs(K,DEs,Defs) :-
	smClassDef(K,"p",_,DEs,_,_,_),
	smClassDefTree(K,"p",Def),
	Def=cmp(_,[_,_,cmp("ATTRS",Defs)]),
	!.
kbPredDEs(K,[],[]) :-
	smClassDef(K,"p",_,Ds,_,_,_),
	Ds=[],
	!.
kbPredDEs(K,_,_) :-
	dlgIF::feil("kbPredDEs",[K],[],nill),!,fail.
% -------------------------------------------------------------------------------------
% kbTermST                                                         
% -------------------------------------------------------------------------------------
% determines the predicate argument combination (sentence type) of a record.
%
% parameters:
% 1	in	term
% 2	out	rel.arg.combination as a string list [Rel,Arg1], [Rel,Arg1,Arg2]
%               KB names are used
kbTermST_proc(T,ST) :-
    kbTermSTp(T,ST),
    !
;
    ST="",
    !.

kbTermSTp(T,ST) :- kbTermST(T,L), sx::sxlJoin(L,"|",ST),!.

kbTermST(cmp(R,[cmp(E1,_),cmp(E2,_)|_]),[R,E1,E2]) :- 
	!.
kbTermST(cmp(R,[cmp(E,_)|_]),[R,E]) :-
	!.

termSt(T,St) :-
       termStF(T,St),
       !
;       
    dlgIF::dlgNote(1,"Term's sentence type cannot be determined!"),
    !,fail.

termStF(T,St) :-
    T=cmp(R,[cmp(E1,_),cmp(E2,_)|_]),
    St=sx::join([R,E1,E2],"|"),
    !
;    
    T=cmp(R,[cmp(E1,_)|_]),
    St=sx::join([R,E1],"|"),
    !.
    
querySt(Q,St) :-
    queryStF(Q,St),
    !
;
    dlgIF::dlgNote(1,"Query's sentence type cannot be determined!"),
    !,fail.
    
queryStF(Q,St) :-
    Q=cmp(",",[H|_]),
    queryStF(H,St),
    !
;    
    Q=cmp("p3_iNdrf",[_,_,str(Qy)]),
	Qt=termDomain::fromTermStringIndirect(Qy),
    queryStF(Qt,ST),
    !
;    
    Q=cmp("p3_iDdrf",[_,_,str(Qy)]),
	Qt=termDomain::fromTermStringIndirect(Qy),
    queryStF(Qt,ST),
    !
;
    Q=cmp("p3_iSSS",[_,str(Qy)]),
	Qt=termDomain::fromTermStringIndirect(Qy),
    queryStF(Qt,ST),
    !
;
	Q=cmp("p3_iExis",[str(Qy)]),
	Qt=termDomain::fromTermStringIndirect(Qy),
    queryStF(Qt,ST),
    !
;
    termStF(Q,St),
    !.
    
querySgNl(Q,SgNl) :-
    Q=cmp(",",[H|_]),
    querySgNl(H,SgNl),
    !
;    
    Q=cmp(P,[cmp(E1,_),cmp(E2,_)|_]),
    nlSingF(P,Pn),
    nlSingF(E1,E1n),
    nlSingF(E2,E2n),
    SgNl=sx::pack([E1n,Pn,E2n]," "),
    !
;    
    Q=cmp(P,[var(X1),var(X2)|_]),
    nlSingF(P,Pn),
    SgNl=sx::pack([X1,Pn,X2]," "),
    !
;    
    Q=cmp(P,[cmp(E1,_),var(X2)|_]),
    nlSingF(P,Pn),
    nlSingF(E1,E1n),
    SgNl=sx::pack([E1n,Pn,X2]," "),
    !
;    
    Q=cmp(P,[var(X1),cmp(E2,_)|_]),
    nlSingF(P,Pn),
    nlSingF(E2,E2n),
    SgNl=sx::pack([X1,Pn,E2n]," "),
    !
;    
    Q=cmp(P,[cmp(E1,_)|_]),
    nlSingF(P,Pn),
    nlSingF(E1,E1n),
    SgNl=sx::pack([E1n,Pn]," "),
    !
;    
    Q=cmp(P,[var(X1)|_]),
    nlSingF(P,Pn),
    SgNl=sx::pack([X1,Pn]," "),
    !
;    
    dlgIF::dlgNote(1,"Cannot determine SG-combination!"),
    !,fail.
% -------------------------------------------------------------------------------------
% kbST_KB                                                              
% -------------------------------------------------------------------------------------
% determines the predicate argument combination (sentence type) of a sentence or sentence-rule term
% fails if function or CR
% parameters:
% 1	in	term
% 2	out	ST: P|E1[|E2]
%
/*
kbST_KB(Term,ST) :-
	kbTermST(Term,List),
	sx::sxlJoin(List,"|",ST),
	!.
*/	
% ----------------------------------------------------------------------------------------------
% chainHasChangedTerm/Chain                                                        
% ----------------------------------------------------------------------------------------------
chainHasChangedTerm(nill) :-
	globalVars::get_i("knowopen")=2,
	!
;
	globalVars::set_i("knowopen",2),		     
	dlgIF::dbaStatuses,
	!.
chainHasChangedTerm(LTerm) :-
	globalVars::get_i("knowopen")=2,
	!,
	LTerm=cmp(Ch,_),
	dbaBase::dbaUpdatedCh(Ch),
	!
;
	globalVars::set_i("knowopen",2),		     
	LTerm=cmp(Ch,_),
	dbaBase::dbaUpdatedCh(Ch),
	dlgIF::dbaStatuses,
	!.

chainHasChangedChain(Ch) :-
	globalVars::get_i("knowopen")=2,
	!,
	dbaBase::dbaUpdatedCh(Ch),
    !
;
	globalVars::set_i("knowopen",2),		     
	dbaBase::dbaUpdatedCh(Ch),
	dlgIF::dbaStatuses,
	!.
% ----------------------------------------------------------------------------------------------
% recordType                                                            
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	record type
% 2	out	basic type 1-7
% 3	out	active/inactive 1/-1
% 4	out	suspended/not 1/-1
%
recordType(1,1,1,-1) :- !.
recordType(2,2,1,-1) :- !.
recordType(3,3,1,-1) :- !.
recordType(4,4,1,-1) :- !.
recordType(5,5,1,-1) :- !.
recordType(6,6,1,-1) :- !.
recordType(-3,3,-1,-1) :- !.
recordType(-5,5,-1,-1) :- !.
recordType(20,2,1,1) :- !. % 11.4.4 susp. sentences were not allowed before
recordType(30,3,1,1) :- !.
recordType(40,4,1,1) :- !.
recordType(50,5,1,1) :- !.
recordType(60,6,1,1) :- !.
recordType(-30,3,-1,1) :- !.
recordType(-50,5,-1,1) :- !.
recordType(RT,_,_,_) :- dlgIF::feil("recordType",[],[RT],nill),!,fail.	
% ----------------------------------------------------------------------------------------------
% recTypeRefs/List                                                            
% ----------------------------------------------------------------------------------------------
class predicates
dbuRTR_2 : (chainDB::bt_selector,string,lst) determ.
recTyperefsX : (lst,integer).
clauses

recTypeRefList(RT,RefChains) :- 
    L=recTypeRefs(RT),
    RefChains=L:get(),
    !.

recTypeRefs(RT)=L :-
	L=lst::new(-1),
	recTypeRefsX(L,RT),
	!.

recTypeRefsX(L,RT) :-
    kbIsSql,
    sqlBase::typeRefs(RT,Chain,Ref),
        RefS=toString(Ref),
        S=string::concat(RefS,"|",Chain),
        L:add(S),
        fail
;
    not(kbIsSql),
	Ix=dbaBase::getIx('C',"RTYP"),
	RTs=toString(RT),
	Key=string::concat(RTs,"|"),
	dbaBase::database(MKB,_,_,0),
	not(MKB:key_search(Ix,Key,_)),
	programControl::repeat,
	   dbuRTR_2(Ix,RTs,L),
	!
;
    !.
    
dbuRTR_2(Ix,RTs,List) :-  
	dbaBase::database(MKB,_,_,0),
	MKB:key_current(Ix,FKey,Ref),
	sx::sxUnpack(FKey,'|',L),
	L=[SRT,Chain],
	SRT=RTs,
	dbaBase::dba_address(Ref,SRef),
	sx::sxlJoin([SRef,Chain],"|",RefChain),
	List:add(RefChain),
	MKB:key_next(Ix,_),
	!,fail
;
    !.
% ----------------------------------------------------------------------------------------------
% chainReadOnlyWarning
% ----------------------------------------------------------------------------------------------
% fails if chain has write access
%
chainReadOnlyWarning(Chain) :-
    not(kbIsSql),
	dbaBase::dbaMap(Chain,_,Access),
	Access=0,
	sx::concat(["Attempt to write to the external read-only chain '",Chain,"'"],M),
	dlgIF::dlgNote(1,M),
	!.
% ----------------------------------------------------------------------------------------------
% getSr
% ----------------------------------------------------------------------------------------------
getSr(Realm,No,Sub)=L :-
    kbIsSql,
    !,L=sqlBase::getSrs(Realm,No,Sub),
    !
;
    Sub=0,
	dbaBase::dbaRealmNoGet(Realm,No,L),
	!
;
    Sub<>0,
	dbaBase::dbaRealmSubNoGet(Realm,No,Sub,L),
	!.
% ----------------------------------------------------------------------------------------------
% srDelete
% ----------------------------------------------------------------------------------------------
srDelete(Realm,No,Sub) :-
    kbIsSql,
    L=sqlBase::getSrs(Realm,No,Sub),
    sqlBase::recordDeletes(L),
    !
;
    not(kbIsSql),
    Sub=0,
    realmDeleteNo(Realm,No),
    !
;    
    not(kbIsSql),
    Sub<>0,
    realmDeleteSubno(Realm,No,Sub),
    !.
% ----------------------------------------------------------------------------------------------
% locFromKba
% ----------------------------------------------------------------------------------------------
locFromKba([],[]) :-
    !.
locFromKba([H|T],[HL|TL]) :-
    sx::sxTail(1,H,RefS,Id),
    dbaBase::dba_address(Ref,RefS),
  	dbaBase::database(KB,_,Id,_),
  	HL=cloc(KB,Ref),
    locFromKba(T,TL),
    !.   
% ----------------------------------------------------------------------------------------------
% accessErrorTerm/Chain                                                            
% ----------------------------------------------------------------------------------------------
accessErrorChain(Chain) :-
    kbIsSql,
    !,fail
;    
	dbaBase::dbaMap(Chain,_,Access),
	Access=0,
	dlgIF::feil("accessErrorChain",["Attempt to change external read-only chain",Chain],[],nill),!.

accessErrorTerm(Term) :-
    kbIsSql,
    !,fail
;    
	Term=cmp(Chain,_),
	dbaBase::dbaMap(Chain,_,Access),
	Access=0,
	dlgIF::feil("accessErrorTerm",["Attempt to change external read-only chain",Chain],[],Term),!.
% ----------------------------------------------------------------------------------------------
% typeCount                                                    
% ----------------------------------------------------------------------------------------------
typeCount(RT)=N :-
    kbIsSql,
    N=sqlBase::typeCount(RT),
    !
;
    set::member(RT,[2,3,4,6,-3,30,40,60,-30]),
    RTs=toString(RT),
    F=string::concat(RTs,"|"),
	dbaUtil::dbuIxCountFront("RTYP",F,N),
	!
;
    RTs=toString(RT),
    Mess=string::concat("Attempt to count record type ",RTs," in a 3kb-database!"),
    dlgIF::dlgNote(3,Mess),!,fail.
% ----------------------------------------------------------------------------------------------
% getIrDef
% ----------------------------------------------------------------------------------------------
getIrDef(Rule,Body,Loc) :-
    getIrDefF(Rule,Body,Loc),
    !
;
    Mess=string::concat("Unknown inexact rule definition: ",Rule,"!"),
    dlgIF::dlgNote(2,Mess),!,fail.
    
getIrDefF(Rule,Body,Loc) :-
    kbIsSql,
    !,sqlBase::getIrDefF(Rule,Body,Chain,Ref),
    Loc=sloc(Chain,Ref),
    !
;    
    dbaUtil::dbuCR_GetRefF(Rule,Body,_,KB,Ref,_,_),
    Loc=cloc(KB,Ref),
    !.
% ----------------------------------------------------------------------------------------------
% realms
% ----------------------------------------------------------------------------------------------
class predicates
getRealmsX : (lst) procedure.
clauses

realms()=L :-
    kbIsSql,
    !,L=sqlBase::realms(),
    !
;
    List=lst::new(),
    getRealmsX(List),
    L=List:get(),
    !.

getRealmsX(List) :-
	Ix=dbaBase::getIx('C',"REALM"),
	dbaBase::database(DB,_,_,0),
	DB:key_last(Ix,_),
	DB:key_current(Ix,LastKey,_),
	DB:key_first(Ix,_),
    DB:key_current(Ix,FirstKey,_),
    List:add(FirstKey),
    FirstKey<>LastKey,
    programControl::repeat,
        DB:key_next(Ix,_),
        DB:key_current(Ix,Key,_),
        Key<>List:top(),
        List:add(Key),
        Key=LastKey,
    !
;
    !.    
% ----------------------------------------------------------------------------------------------
% syntagmDelete
% ----------------------------------------------------------------------------------------------
syntagmDelete(KbName) :-
    Language=globalVars::get_s("lang"),
    chainSentences("v",LT,Loc),
        LT=cmp(_,[cmp(_,[atom(Language),str(KbName)|_]),_]),
        recordDelete(Loc),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% mainFormOfSyno
% ----------------------------------------------------------------------------------------------
mainFormOfSyno(S,M) :-
    kbID(S,K,_,_),
    nlSing(K,M),
    !.
% ----------------------------------------------------------------------------------------------
% kbNameExists
% ----------------------------------------------------------------------------------------------
kbNameExists(KbName) :-
	Lng=globalVars::get_s("lang"),
	Lng=cache::tmCacheIsLoaded(),
	!,cache::kbNameExists(KbName),
	!
;	
    kbIsSql,
	!,Lng=globalVars::get_s("lang"),
    chainSentences("v",LT,_),
    LT=cmp(_,[cmp(_,[atom(Lng),str(KbName)|_]),_]),
    !
;
	Lng=globalVars::get_s("lang"),
	dbaBase::database(DB,_,_,0),
	Ix=dbaBase::getIx('C',"SYN1"),
	Indx=string::concat(Lng,"|",KbName),
	DB:key_search(Ix,Indx,_),
    !.
% ----------------------------------------------------------------------------------------------
% srLoc/Id
% ----------------------------------------------------------------------------------------------
srLoc(Chain,Realm,No,SubNo)=Loc :-
    Loc=srLocF(Chain,Realm,No,SubNo),
    !
;    
    S1=toString(No),
    S2=toString(SubNo),
    sx::sxlJoin([Chain,Realm,S1,S2],"|",S3),
    Mess=string::concat("Unknown sentence rule (by Chain): ",S3),
    dlgIF::dlgNote(2,Mess),!,fail.
    
srLocF(Chain,Realm,No,SubNo)=Loc :-
    kbIsSql,
    !,sqlBase::getRuleRecord(Chain,Realm,No,SubNo,Ref),
    Loc=sloc(Chain,Ref),
    !
;
    SNo=toString(No),
    SSub=toString(SubNo),
    sx::sxlJoin([Chain,Realm,SNo,SSub],"|",Rule),
    dbaUtil::dbuRuleRef(Rule,KB,Ref),
    Loc=cloc(KB,Ref),
    !.

srLocId(Id,Realm,No,SubNo)=Loc :-
    Loc=srLocIdF(Id,Realm,No,SubNo),
    !
;    
    S1=toString(No),
    S2=toString(SubNo),
    sx::sxlJoin([Id,Realm,S1,S2],"|",S3),
    Mess=string::concat("Unknown sentence rule (by DB Id): ",S3),
    dlgIF::dlgNote(2,Mess),!,fail.
    
srLocIdF(Id,Realm,No,SubNo)=Loc :-
    kbIsSql,
    !,sqlBase::getRuleRecordChain(Realm,No,SubNo,Chain,Ref),
    Loc=sloc(Chain,Ref),
    !
;
	dbaUtil::dbuGetInRealm(Id,Realm,No,SubNo,KB,Ref),
    Loc=cloc(KB,Ref),
    !.
% ----------------------------------------------------------------------------------------------
% chainAccess
% ----------------------------------------------------------------------------------------------
chainAccess(Chain)=A :-
    kbIsSql,
    !, A=1,
    !
;
    dbaBase::dbaMap(Chain,_,A),
    !.
% ----------------------------------------------------------------------------------------------
% realmRecords_nd
% ----------------------------------------------------------------------------------------------
realmRecords_nd(Id,Realm)=Loc :-
    kbIssql,
    sqlBase::realmRefsAll(Realm,Chain,Ref),
    Loc=sloc(Chain,Ref)
;    
    not(kbIsSql),
	dbaBase::database(KB,_,Id,_),
	Ix=dbaBase::getIx('C',"REALM"),
	sx::concat([Realm,"|",Id],Key1),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Key1,Ref),
	Loc=cloc(KB,Ref)
;    
    not(kbIsSql),
	dbaBase::database(KB,_,Id,_),
	Ix=dbaBase::getIx('C',"REALM"),
	sx::concat([Realm,"|",Id],Key1),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Key1,_),
	DB:key_current(Ix,_,_),
	!,programControl::repeat,
        DB:key_next(Ix,Ref),
        DB:key_current(Ix,Key3,_),
        Key3=Key1,
        Loc=cloc(KB,Ref).
% ----------------------------------------------------------------------------------------------
% getRecord
% ----------------------------------------------------------------------------------------------
getRecord(Loc,RT,BT,Active,Suspended,Time,Realm,No,SubNo,ST,LTerm,RTerm) :-
    Loc=sloc(Chain,Ref),
    sqlBase::getRecord(Ref,Chain,RT,BT,ActiveX,SuspX,Time,Realm,No,SubNo,ST,LTerm,RTerm),
    jmod::swapI(SuspX,0,-1,Suspended),
    jmod::swapI(ActiveX,0,-1,Active),
    !
;
    Loc=cloc(KB,Ref),
    KB:ref_term(Ref,Rec),
	Rec=clause(RT,Ts,No,SubNo,Realm,LTerm,RTerm),
    datetime::tsSQL(Ts,Time),
    recordType(RT,BT,Active,Suspended),
    kbTermST_proc(LTerm,ST),
    !.
% ----------------------------------------------------------------------------------------------
% chainIsExternal
% ----------------------------------------------------------------------------------------------
chainIsExternal(Chain) :-
    kbIsSql,
    !,fail
;
    dbaBase::extChain(Chain,_,_),
    !.
% ----------------------------------------------------------------------------------------------
% chains_nd
% ----------------------------------------------------------------------------------------------
chains_nd(ExternalChains,Chain) :-
    kbIsSql,
    !,sqlBase::chains(Chain)
;
    dbaBase::database(DB,_,_,0),
	DB:db_chains(Chain)
;
    ExternalChains=1,
    dbaBase::database(DB,_,_,N),
    N>0,
	DB:db_chains(Chain),
	chainIsExternal(Chain).
% ----------------------------------------------------------------------------------------------
% chainCount
% ----------------------------------------------------------------------------------------------
chainCount(Chain)=N :-
    kbIsSql,
    !,N=sqlBase::chainCount(Chain),
    !
;
    dbaUtil::dbaCountChain(Chain,N),
    !.
% ----------------------------------------------------------------------------------------------
% irType
% ----------------------------------------------------------------------------------------------
class predicates
irSubType : (sterm,string,string) determ (i,o,o).
clauses

irType(KbName,ChainType,SubType) :-
    irTypeF(KbName,ChainType,SubType),
    !
;
    Mess=string::concat("Unknown inexact rule definition: ",KbName,"!"),
    dlgIF::dlgNote(2,Mess),!,fail.

irTypeF(KbName,ChainType,SubType) :-
    getIrDefF(KbName,LTerm,_),
    irSubType(LTerm,ChainType,SubType),
    !.

irSubType(LT,ChainType,SubType) :-
	LT=cmp(ChainType,[_,cmp(_,_),_,_,_,_,_]),
	ChainType="p3_iCswi",
    SubType="p3_iCswc",
	!
;	
	LT=cmp(ChainType,[_,atom(_),_,_,_,_,_]),
	ChainType="p3_iCswi",
    SubType="p3_iCswp",
	!
;    
	LT=cmp(ChainType,[_,cmp("p3_iQry",_),_,_,_,_,_]),
	ChainType="p3_iCdat",
    SubType="p3_iCqry",
	!
;    
	LT=cmp(ChainType,[_,cmp("p3_iQst",_),_,_,_,_,_]),
	ChainType="p3_iCdat",
    SubType="p3_iCqst",
    !
;    
	LT=cmp(ChainType,_),
	SubType=ChainType,
	!.
% ----------------------------------------------------------------------------------------------
% irDeleteAll
% ----------------------------------------------------------------------------------------------
irDeleteAll() :-
    irChain_nd(Chain),
        chainDelete(Chain),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% getIrs
% ----------------------------------------------------------------------------------------------
class predicates
getIrsX : (lst).
clauses

 getIrs()=L :-
    L=lst::new(-1),
    getIrsX(L),
    !.
    
getIrsX(L) :-
    smClassDef_nd("h",KbName,_,_,_,_,_,_,_,_),
        L:add(KbName),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% getIrsOfSubType
% ----------------------------------------------------------------------------------------------
class predicates
getIrsOfSubtypeX : (lst,string).
clauses

 getIrsOfSubtype(SubType)=L :-
    L=lst::new(-1),
    getIrsOfSubtypeX(L,SubType),
    !.
    
getIrsOfSubtypeX(L,SubType) :-
    smClassDef_nd("h",KbName,_,_,_,_,_,_,_,Sub),
        Sub=SubType,
        L:add(KbName),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% smSegment
% ----------------------------------------------------------------------------------------------
smSegment(KbName)=Segment :-
    Loc=smRefF(KbName),
    getRecord(Loc,_,_,_,_,_,_,_,_,_,LTerm,_),
    LTerm=cmp(_,[_,str(Segment)]),
    !.
% ----------------------------------------------------------------------------------------------
% irChain_nd                                                            
% ----------------------------------------------------------------------------------------------
irChain_nd("p3_iCmap"). 
irChain_nd("p3_iCdat"). 
irChain_nd("p3_iCbay"). 
irChain_nd("p3_iCpar"). 
irChain_nd("p3_iCand"). 
irChain_nd("p3_iCor"). 
irChain_nd("p3_iCcmb").
irChain_nd("p3_iCnot"). 
irChain_nd("p3_iCswi"). 
irChain_nd("p3_iZand"). 
irChain_nd("p3_iZor"). 
irChain_nd("p3_iZnot"). 
irChain_nd("p3_iZowa"). 
% ----------------------------------------------------------------------------------------------
% publicChain                                    
% ----------------------------------------------------------------------------------------------
publicChain(Chain,Access) :-
    kbIsSql,
    !,fail
;
    dbaBase::pubChain(Chain,Access),
    !.
% ----------------------------------------------------------------------------------------------
% stSentence_nd                             
% ----------------------------------------------------------------------------------------------
stSentence_nd(St,Lt,Loc) :- 
    kbIsSql,
    sx::sxUnpack(St,'|',L),
    L=[Chain|_],
    sqlBase::stSentence_nd(St,LT,_,Ref),
        Loc=sloc(Chain,Ref)
;
    not(kbIsSql),
    sx::sxUnpack(St,'|',L),
    L=[Chain|_],
    dbaBase::dbaMap(Chain,KB,_),
	KB:chain_terms(Chain,Rec,Ref),
        Rec=clause(Type,_,_,_,_,LT,_),
        AType=math::abs(Type),
        AType<=2,
        kbTermSTp(LT,Str),
        St=Str,
        Loc=cloc(KB,Ref).
% ----------------------------------------------------------------------------------------------
% stSentence_next
% ----------------------------------------------------------------------------------------------
stSentence_next(St,Loc,Next) :-
    Prev=dbLocation::new(sloc("",0)),
    stSentence_nd(St,_,Next),
        PrevLoc=Prev:get(),
        Prev:set(Next),
        Loc=PrevLoc,
        !.
% ----------------------------------------------------------------------------------------------
% stSentence_prev
% ----------------------------------------------------------------------------------------------
stSentence_prev(St,Loc,Previous) :-
    Prev=dbLocation::new(sloc("",0)),
    stSentence_nd(St,_,Next),
        Previous=Prev:get(),
        Prev:set(Next),
        Next=Loc,
        not(Previous=sloc("",0)),
        !.
% ----------------------------------------------------------------------------------------------
% chainType                      
% ----------------------------------------------------------------------------------------------
chainType(Chain,DbId,Type,Access) :-
    kbIsSql,
    !,DbId="X",
    Type=0,
    Access=1,
    !
;
    dbaBase::dbaChainType(Chain,DbId,Type,Access),
    !.
% ----------------------------------------------------------------------------------------------
% dbName
% ----------------------------------------------------------------------------------------------
dbName()=Name :-
    N=globalVars::get_s("kb_file"),
	Name=fileName::getName(N),
	!.
% ----------------------------------------------------------------------------------------------
% getExtChains
% ----------------------------------------------------------------------------------------------
getExtChains(R,W) :-
    kbIsSql,
    !,R=[],
    W=[],
    !
;
    dbaBase::getExtChains(R,W),
    !.
% ----------------------------------------------------------------------------------------------
% chainDeleteSentences
% ----------------------------------------------------------------------------------------------
chainDeleteSentences(Type,Chain,N) :-
    kbIsSql,
    !,
    cache::check(Chain),
    dep::massSentenceUpdateStart,
    sqlBase::chainDeleteSentences(Type,Chain,N),
    dep::massSentenceUpdateEnd,
    !
;
    chainReadOnlyWarning(Chain),
    !,fail
;    
    dep::massSentenceUpdateStart,
    cache::check(Chain),
    dbaBase::dbaDeleteSG(Type,Chain,N),
    dep::massSentenceUpdateEnd,
    !.
% ----------------------------------------------------------------------------------------------
% chainSuspend
% ----------------------------------------------------------------------------------------------
chainSuspend(Chain) :-
    kbIsSql,
    sqlBase::chainBasetypeRefs(3,Chain,Ref),
        sqlBase::recordSuspend(Chain,Ref),
        fail
;
    kbIsSql,
    sqlBase::chainBasetypeRefs(4,Chain,Ref),
        sqlBase::recordSuspend(Chain,Ref),
        fail
;
    kbIsSql,
    sqlBase::chainBasetypeRefs(5,Chain,Ref),
        sqlBase::recordSuspend(Chain,Ref),
        fail
;
    kbIsSql,
    sqlBase::chainBasetypeRefs(6,Chain,Ref),
        sqlBase::recordSuspend(Chain,Ref),
        fail
;
    chainReadOnlyWarning(Chain),
    !,fail
;
    not(kbIsSql),
    dbaBase::dbaSuspendChain(Chain),
    !.
% ----------------------------------------------------------------------------------------------
% recordSuspend
% ----------------------------------------------------------------------------------------------
recordSuspend(sloc(Chain,Ref)) :- 
    sqlBase::recordSuspend(Chain,Ref),
    !.
recordSuspend(cloc(KB,Ref)) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,LTerm,_),
    LTerm=cmp(Chain,_),
    chainReadOnlyWarning(Chain),
    !,fail
;    
    dbaBase::dbaSuspendRecord(KB,Ref),
    !.
% ----------------------------------------------------------------------------------------------
% chainExists
% ----------------------------------------------------------------------------------------------
chainExists(Chain) :-
    kbIsSql,
    !,sqlBase::existChain(Chain),
    !
;
    dbaBase::chainExists(Chain),
    !.
% ----------------------------------------------------------------------------------------------
% realmExists
% ----------------------------------------------------------------------------------------------
realmExists(Realm,Files) :-
    kbIsSql,
    !,sqlBase::realmCount(Realm)>0,
    F=globalVars::get_s("kb_file"),
    Files=[F],
    !
;
    dbaBase::dbaRealmExists(Realm,Files),
    !.
% ----------------------------------------------------------------------------------------------
% indexRegen                                            
% ----------------------------------------------------------------------------------------------
indexRegen() :-
    kbIsSql,
    !
;
	dlgIF::msgStatus("Regenerating all indexes..."),
	dbaUtil::dbaIxRegenAll,
	dlgIF::msgStatus(""),
    !.
% ----------------------------------------------------------------------------------------------
% getTimingEvent                                                        
% ----------------------------------------------------------------------------------------------
getTimingEvent(Event,Dat,Tim) :-
    chainSentences("x",LT),
	LT=cmp(_,[cmp("p3_eKtim",[atom(Event),int(Dat),int(Tim)])]),
    !
;
	Dat=0,
	Tim=0,
	!.		
% ----------------------------------------------------------------------------------------------
% insertTimingEvent                                                        
% ----------------------------------------------------------------------------------------------
insertTimingEvent(Event,Dat,Tim) :-
    chainSentences("x",LTerm,Loc),
	LTerm=cmp("x",[cmp("p3_eKtim",[atom(Event),_,_])]),
	NTerm=cmp("x",[cmp("p3_eKtim",[atom(Event),int(Dat),int(Tim)])]),
	recordDeleteOutright(Loc),
    recordInsertOutright(2,"$SYS",0,0,NTerm,atom("true"),_),
	chainHasChangedTerm(NTerm),
	!
;
	NTerm=cmp("x",[cmp("p3_eKtim",[atom(Event),int(Dat),int(Tim)])]),
    recordInsertOutright(2,"$SYS",0,0,NTerm,atom("true"),_),
	chainHasChangedTerm(NTerm),
	!.
% -------------------------------------------------------------------------------------------------------------
% getProgramBuild
% -------------------------------------------------------------------------------------------------------------
% returns program version VRR and version status (text)
%
getProgramBuild(VR,V,R,B,Stat) :-
	V=globalVars::get_i("p3v"),
	R=globalVars::get_i("p3r"),
	B=globalVars::get_i("p3b"),
	VR=V*100+R,
	Stat=globalVars::get_s("p3s"),
	!.
% -------------------------------------------------------------------------------------------------------------
% realmDelete
% -------------------------------------------------------------------------------------------------------------
realmDelete(Realm) :-
    kbIsSql,
    !,sqlBase::realmDelete(Realm,_),
    !
;
	dbaBase::database(KB,_,_,_),
	dbaUtil::dbaDeleteRealmKB(KB,Realm),
	!.

realmDelete(Id,Realm) :-
    kbIsSql,
    !,sqlBase::realmDelete(Realm,_),
    !
;
	dbaBase::database(KB,_,Id,_),
	dbaUtil::dbaDeleteRealmKB(KB,Realm),
    !.
% -------------------------------------------------------------------------------------------------------------
% realmKbDelete
% -------------------------------------------------------------------------------------------------------------
realmKbDelete(Files,Realm) :-
    kbIsSql,
    !,sqlBase::realmDelete(Realm,_),
    !
;
	dbaUtil::dbaDeleteKB_Realm(Files,Realm),
    !.
% -------------------------------------------------------------------------------------------------------------
% recordsDelete
% -------------------------------------------------------------------------------------------------------------
recordsDelete([]) :-
    !.
recordsDelete([Loc|Locs]) :-
    recordDelete(Loc),
    recordsDelete(Locs),
    !.
% -------------------------------------------------------------------------------------------------------------
% realmDeleteNo/Subno
% -------------------------------------------------------------------------------------------------------------
realmDeleteNo(Realm,No) :-
    kbIsSql,
    !,sqlBase::realmDeleteNo(Realm,No),
    !
;
    dbaBase::dbaRealmNoGet(Realm,No,Locs),
    recordsDelete(Locs),
    !.
    
realmDeleteSubno(Realm,No,Subno) :-
    kbIsSql,
    !,sqlBase::realmDeleteSubno(Realm,No,Subno),
    !
;
    dbaBase::dbaRealmSubNoGet(Realm,No,SubNo,Locs),
    recordsDelete(Locs),
    !.
% -------------------------------------------------------------------------------------------------------------
% realmNoCount/realmSubnoCount
% -------------------------------------------------------------------------------------------------------------
realmNoCount(Realm,No)=N :-
    kbIsSql,
    !,N=sqlBase::realmNoCount(Realm,No),
    !
;
	dbaBase::dbaRealmNoGet(Realm,No,L),
	set::size(L,N),
    !.
    
realmSubnoCount(Realm,No,Subno)=N :-
    kbIsSql,
    !,N=sqlBase::realmSubnoCount(Realm,No,Subno),
    !
;
    dbaBase::dbaRealmSubNoGet(Realm,No,Subno,SRefs),
	set::size(SRefs,N),
    !.    
% -------------------------------------------------------------------------------------------------------------
% recTypeLocs
% -------------------------------------------------------------------------------------------------------------
recTypeLocs(RecType,List) :-
    kbIsSql,
    !,sqlBase::recTypeLocs(RecType,List),
    !
;    
    set::member(RecType,[1,2]),
    S=toString(RecType),
    Mess=string::concat("Attempt to call recTypeLocs in 3KB-database with record type:",S,"!"),
    dlgIF::dlgNote(4,Mess),
    !,fail
;    
    dbaUtil::dbuGetRecTypes(RecType,List),
    !.
% -------------------------------------------------------------------------------------------------------------
% recBasetypeLocsAll
% -------------------------------------------------------------------------------------------------------------
recBasetypeLocsAll(BaseType,List) :-
    kbIsSql,
    !,sqlBase::recBasetypeLocsAll(BaseType,List),
    !
;    
    Type2=-Basetype,
    Type3=Basetype*10,
    Type4=-Type3,
    recTypeLocs(Basetype,L1),
    recTypeLocs(Type2,L2),
    recTypeLocs(Type3,L3),
    recTypeLocs(Type4,L4),
    set::conc([L1,L2,L3,L4],List),
    !.

recBasetypeLocs(BaseType,List) :-
    kbIsSql,
    !,sqlBase::recBasetypeLocs(BaseType,List),
    !
;    
    Type2=-Basetype,
    recTypeLocs(Basetype,L1),
    recTypeLocs(Type2,L2),
    set::conc([L1,L2],List),
    !.
% -------------------------------------------------------------------------------------------------------------
% recordSetType
% -------------------------------------------------------------------------------------------------------------
recordSetType(Loc,RecType) :-
    Loc=sloc(Chain,Ref),
    !,sqlBase::recordSetType(Chain,Ref,RecType),
    !
;
    Loc=cloc(KB,Ref),
	dbaUtil::dbuChangeRecType(KB,Ref,RecType),
    !.
% -------------------------------------------------------------------------------------------------------------
% realmCount
% -------------------------------------------------------------------------------------------------------------
realmCount(Id,Realm)=N :-
    kbIsSql,
    !,N=sqlBase::realmCount(Realm),
    !
;
    dbaBase::database(KB,_,Id,_),
	dbaUtil::dbaCountRealm(KB,Realm,N),
    !.
% -------------------- --------------------------------------------------------------------------
% isValidId
% ----------------------------------------------------------------------------------------------
isValidId(S) :-
	string::isname(S),
	string::frontchar(S,F,_),
	F<>'_',
	string::length(S)<=globalVars::get_i("max_kbn"),!.
% -------------------------------------------------------------------------------------------------------------
% stList
% -------------------------------------------------------------------------------------------------------------
stList()=L :-
    kbIsSql,
    !,
    L=sqlBase::stList(),
    !
;  
    L=dbaLib::stList(),  
    !.   
% -------------------------------------------------------------------------------------------------------------
% stBasetypes
% -------------------------------------------------------------------------------------------------------------
stBasetypes(Basetype)=L :-
    kbIsSql,
    !,
    L=sqlBase::stBasetypes(Basetype),
    !
;
    L=dbaLib::stBasetypes(Basetype),
    !.
end implement dbBase
