/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement dbExtension
open core,termDomain,globalDomains

constants
className = "com/rudstrand/dbExtension/dbExtension".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

% ----------------------------------------------------------------------------------------------
% multiDb
% ----------------------------------------------------------------------------------------------
multiDb() :-
    dbBase::kbIsSql,
    !,fail
;
    dbaBase::multiDb,
    !.
% ----------------------------------------------------------------------------------------------
% realmPrefix
% ----------------------------------------------------------------------------------------------
realmPrefix(Chain,Realm)=PrefixedRealm :-
    not(dbBase::kbIsSql),
    multiDb,
	Id=dbBase::dbIdChain(Chain),
	PrefixedRealm=string::concat(Id,":",Realm),
	!
;	
    PrefixedRealm=Realm,
    !.
% ----------------------------------------------------------------------------------------------
% realmRemovePrefix
% ----------------------------------------------------------------------------------------------
realmRemovePrefix(S,Realm,Id) :-
    string::length(S)>2,
    string::front(S,1,Id,S2),
    string::front(S2,1,Col,Realm),
    Col=":",
    !
;
    dbBase::kbIsSql,
    Realm=S,
    Id="X",
    !
;    
    Realm=S,
    dbaBase::database(_,_,Id,0),
    !.
% ----------------------------------------------------------------------------------------------
% tableLayoutLoc                    
% ----------------------------------------------------------------------------------------------
tableLayoutLoc(ST)=Loc :-
    Loc=tableLayoutLocF(ST),
    !
;
    Mess=string::concat("Unknown table layour for sentence type: ",ST),
    dlgIF::dlgNote(3,Mess),!,fail.

tableLayoutLocF(ST)=Loc :-
    dbBase::chainSentences("y",LT,Loc),
    LT=cmp(_,[cmp("p3_eTabLay",[str(ST)|_])]),
    !.
% ----------------------------------------------------------------------------------------------
% updateDependencies                                           
% ----------------------------------------------------------------------------------------------
% performs the required updates of the dependency graph implied by record inserts/deletions.
% parameters:
% 1	in	1=insert; 3=delete
% 2	in	record type
% 3	in 	L term to be inserted/deleted
% 4 in	R term		do.
% 5	in	sentence rule key (if sentence rule)
% 6 out 1=acycliclicy maintained 0=acyclicity violated
%
updateDependencies(Action,RecType,LTerm,RTerm,Rule,Acyclic) :-
	dbBase::recordType(RecType,_,_,Susp),
	Susp=1,
	% ignore deletions/insertions of suspended records
	Acyclic=1,
	!
;	
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType<>4, % all record types except function definitions
	dep::updateDP_Graph(Action,RecType,LTerm,RTerm,Rule,Acyclic),
	!
;
    dbBase::kbIsSql,
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	% since there are generally several definitions of the same function, it is necessary to re-analyze and reflect the dependencies of all definitions
	LTerm=cmp(Func,_),
	% delete and then reinsert all dependencies where the function def is the dependent
	dep::deleteWhereDep("p3_pDP",cmp("n",[int(19),str(Func)])),
    sqlBase::chainTermsDirect(Func,LT,RT),
        dep::updateDP_Graph(1,RecType,LT,RT,"",Acyclic),
        Acyclic=0, 
        % no need to continue since cycle detected
        !
;    
    not(dbBase::kbIsSql),
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	% since there are generally several definitions of the same function, it is necessary to re-analyze and reflect the dependencies of all definitions
	LTerm=cmp(Func,_),
	% delete and then reinsert all dependencies where the function def is the dependent
	dep::deleteWhereDep("p3_pDP",cmp("n",[int(19),str(Func)])),
	dbaBase::dbaMap(Func,KB,_,_),
	KB:chain_terms(Func,Rec,_),
        Rec=clause(_,_,_,_,_,LT,RT),
        dep::updateDP_Graph(1,RecType,LT,RT,"",Acyclic),
        Acyclic=0, 
        % no need to continue since cycle detected
        fail
;
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	Acyclic=1,
	!    
;
	dbBase::recordType(RecType,BaseType,_,_),
    not(set::member(BaseType,[3,4,5,6])),
    Acyclic=1,
    !.
% -------------------------------------------------------------------------------------------------------------
% dbaDuplicate
% -------------------------------------------------------------------------------------------------------------
% succeeds if a fact exists already in the KB
%
dbaDuplicate(LTerm) :-
    dbBase::kbIsSql,
    !,_=sqlBase::existSentence(LTerm),
    !
;    
	LTerm=cmp(Chain,_),
	dbaBase::dbaMap(Chain,KB,_),
	KB:chain_terms(Chain,Rec,_),
	Rec=clause(_,_,_,_,_,LT,atom("true")),
	LTerm=LT,
	!.
% -----------------------------------------------------------------------------------------------
% kbST_RuleCount
% -----------------------------------------------------------------------------------------------
class predicates
kbST_RC : (string,count) procedure (i,i).
clauses

kbST_RuleCount(ST,N) :-
	C=count::newCounter(0),
	kbST_RC(ST,C),
	N=C:get(),
	!.

kbST_RC(ST,C) :-
	dep::get_nd("p3_pDR",A1,_,B1,B2),
	   A1=int(3),
	   B1=int(2),
	   B2=str(ST),
	   C:inc(_),
	   fail
;
	!.
% ----------------------------------------------------------------------------------------------
% syntagmSets
% ----------------------------------------------------------------------------------------------
class predicates
get_syntagms_x : (chainDB,string,string,sterml,chainDB::bt_selector,sterml) procedure (i,i,i,i,i,o).
tail_match : (sterm,sterml,sterm,sterml,sterml) procedure (i,i,i,i,o).
clauses

% 1	in	first token 
% 2	in	tail of tokens
% 3	out	syntagm sets (possibly empty)
%
syntagmSets(Key,Tail,SynSets) :-
    dbBase::kbIsSql,
    sqlBase::getSyntagmSet([str(Key)|Tail],SynSets),
    !    
;    
    not(dbBase::kbIsSql),
	Lng=globalVars::get_s("lang"),
	dbaBase::database(DB,_,_,0),
	Ix=dbaBase::getIx('C',"TOKEN"),
	sx::concat([Lng,"|",Key],Indx),
	DB:key_search(Ix,Indx,Ref),
	dbaBase::dbaMap("v",KB,_),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,Term,atom("true")),
	Term=cmp("v",[cmp(Ent,[atom(Lng),_,_,_,list(str(Key),T)]),_]),
	set::member(Ent,["t","s"]),
	get_syntagms_x(DB,Lng,Key,Tail,Ix,Terms),
	tail_match(T,Tail,Term,Terms,SynSets),
	!.
		
get_syntagms_x(DB,Lng,Key,Tail,Ix,SynSets) :-
	DB:key_next(Ix,Ref),
	dbaBase::dbaMap("v",KB,_),	
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,Term,atom("true")),
	Term=cmp("v",[cmp(Ent,[atom(Lng),_,_,_,list(str(Key),T)]),_]),
	set::member(Ent,["t","s"]),
	get_syntagms_x(DB,Lng,Key,Tail,Ix,Terms),
	tail_match(T,Tail,Term,Terms,SynSets),
	!.
get_syntagms_x(_,_,_,_,_,[]) :- 
	!.	
% the resulting syntagm set is set to [Set|Sets] if the token list tail from the KB 
% matches the scanned words, else the resulting set is [Sets].
% always succeeds.
%
tail_match(nill,_,Set,Sets,[Set|Sets]) :- 
	!.
tail_match(list(H,T),[H|Tokens],Set,Sets,SynSets) :- 
	tail_match(T,Tokens,Set,Sets,SynSets),
	!.
tail_match(_,_,_,Sets,Sets) :- 
	!.
% ----------------------------------------------------------------------------------------------
% dbuSR_Type
% ----------------------------------------------------------------------------------------------
dbuSR_Type(Loc,TypeNo) :-
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,_,_,_,Cond),
	Cond=cmp(Proc,_),
	rub::rubRuleType(_,TypeNo,Proc),
	!
;
	TypeNo=1,
	!.
% -------------------------------------------------------------------------------------
% kbGetAnno...
% -------------------------------------------------------------------------------------
kbGetAnnotationNL_Blank(Type,Node,Note) :-
	kbGetAnnoNL_F(Type,Node,Note),
	!
;
	Note="",
	!.

kbGetAnnotationBlank(Type,Node,Note) :-
	kbGetAnnoF(Type,Node,Note),
	!
;
	Note="",
	!.

kbGetAnnoF(Type,Node,Note) :-
    dbBase::kbIsSql,
    !,sqlBase::chainSentencesDirect("p3_pIntd",LT),
    LT=cmp("p3_pIntd",[cmp("p3_eAnno",[atom(Type),str(Node),str(Nt)])]),
    Note=Nt,
	!
;
	dbaBase::dbaMap("p3_pIntd",KB,_),
	KB:chain_terms("p3_pIntd",Cl,_),
	Cl=clause(_,_,_,_,_,cmp("p3_pIntd",[cmp("p3_eAnno",[atom(Type),str(Node),str(Nt)])]),_),
	Note=Nt,
	!.

kbGetAnnoNL(Type,Node,Note) :-
    kbGetAnnoNL_F(Type,Node,Note),
	!
;
	Mess=string::concat("Unknown annotaion: ",Node),
	dlgIF::dlgNote(2,Mess),!,fail.
	
kbGetAnnoNL_F(Type,Node,Note) :-
	dbBase::kbID(Type,Kt,_,_),
	dbBase::kbSym_to_IF(Node,Kn),
	kbGetAnnoF(Kt,Kn,Note),
	!.
% -------------------------------------------------------------------------------------
% kbQuery                                        
% -------------------------------------------------------------------------------------
% retrieves a query given its menu ID.
%
kbQuery(MenuId,Name,Frmat,Query,Loc) :-
    kbQueryF(MenuId,Name,Frmat,Query,Loc),
    !
;	
	S1=toString(MenuId),
	Mess=string::concat("Cannot find KB query #",S1,"!"),
	dlgIF::dlgNote(2,Mess),
	!,fail.
    
kbQueryF(MenuId,Name,Frmat,Query,Loc) :-
    dbBase::kbIsSql,
	!,Lng=globalVars::get_s("lang"),
	sqlBase::chainSentencesDirect("y",Term,Ref),
    Term=cmp(_,[cmp("p3_eKbq",[int(MenuId),atom(Lng),str(Name),atom(Frmat),str(Query),_])]),
    Loc=sloc("y",Ref),
	!
;    
	Lng=globalVars::get_s("lang"),
	dbaBase::dbaMap("y",KB,_),
	KB:chain_terms("y",clause(_,_,_,_,_,Term,RT),Ref),
	RT=atom("true"),
    Term=cmp("y",[cmp("p3_eKbq",[int(MenuId),atom(Lng),str(Name),atom(Frmat),str(Query),_])]),
    Loc=cloc(KB,Ref),
	!.
% -------------------------------------------------------------------------------------
% kbQueryMenuLoc
% -------------------------------------------------------------------------------------
kbQueryMenuLoc(Menu)=Loc :-
    Loc=kbQueryMenuLocF(Menu),
    !
;
	S1=toString(Menu),
	Mess=string::concat("Cannot find KB query menu #",S1,"!"),
	dlgIF::dlgNote(2,Mess),
	!,fail.

kbQueryMenuLocF(Menu)=Loc :-
    dbBase::chainSentences("y",LT,Loc),
    LT=cmp(_,[cmp("p3_eKbm",[int(Menu)|_])]),
    !.
% -------------------------------------------------------------------------------------
% kbQueryMenuTree
% -------------------------------------------------------------------------------------
class predicates
getMenus : (lst,vpiDomains::menuTag,string).
getQueryEntries : (lst,vpiDomains::menuTag,string).
makeMenus : (string_list,vpiDomains::menuItem_list) determ (i,o).   
kbQueryEntries : (vpiDomains::menuTag,vpiDomains::menuItem_list) determ (i,o).
kbMenuTree : (vpiDomains::menuTag,vpiDomains::menuItem_list) determ (i,o).	
kbMenuEntries : (vpiDomains::menuTag,vpiDomains::menuItem_list) determ (i,o).	
kbMenuTrees : (vpiDomains::menuItem_list,vpiDomains::menuItem_list) determ (i,o).	
clauses

kbQueryMenuTree(vpiDomains::dynMenu(L),DynMenu) :-
    kbQueryMenuTreeF(vpiDomains::dynMenu(L),DynMenu),
    !
;	
	dlgIF::dlgNote(2,"Cannot find KBquery menu!"),
	!,fail.
    
kbQueryMenuTreeF(vpiDomains::dynMenu(L),vpiDomains::dynMenu([Q|T])) :-
    L=[vpiDomains::txt(Id,Menu,Acc,F1,F2,_)|T],
	kbMenuTree(7000,Menus),
	Q=vpiDomains::txt(Id,Menu,Acc,F1,F2,Menus),
	!.

kbMenuTree(Id,Menus) :-
	kbQueryEntries(Id,Entries),
	kbMenuEntries(Id,Submenus1),
	kbMenuTrees(SubMenus1,Submenus),
	set::conc([Entries,SubMenus],Menus),
	!.

kbMenuTrees([],[]) :- 
    !.
kbMenuTrees([vpiDomains::txt(Id,A,B,C,D,_)|T],[vpiDomains::txt(Id,A,B,C,D,Menus)|Ms]) :-	
	kbMenuTree(Id,Menus),
	kbMenuTrees(T,Ms),
	!.
	
% gets all sub-menus which belong to menu 'Id'.
% never fails.
%
kbMenuEntries(Id,Menus) :-
	Lng=globalVars::get_s("lang"),
    List=lst::new(),
    getMenus(List,Id,Lng),
    L=List:get(),
    makeMenus(L,Menus),
    !.

getMenus(List,Id,Lng) :-    
    dbBase::kbIsSql,
    sqlBase::chainSentencesDirect("y",LT),
        LT=cmp(_,[cmp("p3_eKbm",[int(MenuId),atom(Lng),str(Name),int(Id)])]),
        MenuIDs=toString(MenuId),
        S=string::concat(MenuIDs,"||",Name),
        List:add(S),
        fail    
;    
    not(dbBase::kbIsSql),
	dbaBase::dbaMap("y",KB,_),
	KB:chain_terms("y",clause(RT,_,_,_,_,Term,_),_),
        Term=cmp(_,[cmp("p3_eKbm",[int(MenuId),atom(Lng),str(Name),int(Id)])]),
        dbBase::recordType(RT,RTb,_,_),
        RTb<=2,
        MenuIDs=toString(MenuId),
        S=string::concat(MenuIDs,"||",Name),
        List:add(S),
        fail
;        
	!.

makeMenus([],[]) :-
    !.
makeMenus([S|L],[Menu|Menus]) :-
    sx::sxUnpackS(S,"||",L2),
    L2=[Id,Name],
    Idn=sx::toInt(Id),
	Menu=vpiDomains::txt(Idn,Name,vpiDomains::noAccelerator(),b_true,b_false,[]),
    makeMenus(L,Menus),
    !.
    
% gets all menu entries which belong to menu 'Id'
% never fails.
%
kbQueryEntries(Id,Entries) :-
	Lng=globalVars::get_s("lang"),
    List=lst::new(),
    getQueryEntries(List,Id,Lng),
    L=List:get(),
    makeMenus(L,Entries),
    !.

getQueryEntries(List,Id,Lng) :-    
    dbBase::kbIsSql,
    sqlBase::chainSentencesDirect("y",LT),
        LT=cmp(_,[cmp("p3_eKbq",[int(MenuId),atom(Lng),str(Name),_,_,int(Id)])]),
        MenuIDs=toString(MenuId),
        S=string::concat(MenuIDs,"||",Name),
        List:add(S),
        fail
;
    not(dbBase::kbIsSql),
	dbaBase::dbaMap("y",KB,_),
	KB:chain_terms("y",clause(RT,_,_,_,_,Term,_),_),
        Term=cmp(_,[cmp("p3_eKbq",[int(MenuId),atom(Lng),str(Name),_,_,int(Id)])]),
        dbBase::recordType(RT,RTb,_,_),
        RTb<=2,
        MenuIDs=toString(MenuId),
        S=string::concat(MenuIDs,"||",Name),
        List:add(S),
        fail
;        
	!.
% -------------------------------------------------------------------------------------
% kbQueryPar                                               
% -------------------------------------------------------------------------------------
% retrieves a kb query parameter given its parameter no.
%
kbQueryPar(Par,Dom,Title,Alts,Loc) :-
    kbQueryParF(Par,Dom,Title,Alts,Loc),
    !
;	
	sx::concat(["Cannot find query parameter ",Par],S2),
	dlgIF::feil("kbQueryPar",[S2],[],nill),
	!,fail.
    
kbQueryParF(Par,Dom,Title,Alts,Loc) :-
    dbBase::kbIsSql,
	!,Lng=globalVars::get_s("lang"),
    sqlBase::chainSentencesDirect("y",LT,Ref),
    LT=cmp(_,[cmp("p3_eKbp",[str(Par),atom(Lng),atom(Dom),str(Title),A])]),
    cls_listR(Alts,A),
    Loc=sloc("y",Ref),
    !
;
	Lng=globalVars::get_s("lang"),
	dbaBase::dbaMap("y",KB,_),
	KB:chain_terms("y",clause(RT,_,_,_,_,Term,_),Ref),
    Term=cmp("y",[cmp("p3_eKbp",[str(Par),atom(Lng),atom(Dom),str(Title),A])]),
    cls_listR(Alts,A),
    dbBase::recordType(RT,RTb,_,_),
    RTb<=2,
    Loc=cloc(KB,Ref),
	!.
% ----------------------------------------------------------------------------------------------
% kbScanTerm
% ----------------------------------------------------------------------------------------------
class predicates
kbBasicTerm : (sterm,sterm) determ (i,o).
kbScanTermX : (string,pie_scan::wrds,sterm) determ (i,i,o).
clauses

kbScanTerm(D,V,Term) :-
	pscan::pscNL(V,Toks),
	pscan::str_tokens(Toks,SToks),
	transnl::scnSyntagmSets(Toks,SToks,Sets),
	transnl::nl_tok_scan(Toks,SToks,Sets,Wrds),
	kbScanTermX(D,Wrds,Term),
	!.
	
kbScanTerm(D,V,Term) :-
	pscan::pscNL(V,Toks),
	pscan::str_tokens(Toks,SToks),
	transnl::scnSyntagmSets(Toks,SToks,Sets),
	transnl::nl_tok_scan(Toks,SToks,Sets,Wrds),
	syn::synDE_Value(D,Wrds,WrdsLeft,Term1),
	kbBasicTerm(Term1,Term),
	WrdsLeft=[],
	!.
	
kbScanTermX(_,[pie_scan::w(_,S,"p3_dli",_)],str(S)) :- 
	!.		
kbScanTermX(_,[pie_scan::w(_,S,"p3_din",_)],int(I)) :- 
	I=sx::toInt(S),
	!.		
kbScanTermX(_,[pie_scan::w(_,S,"p3_dnu",_)],fix(F)) :- 
	F=sx::toReal(S),
	!.
kbScanTermX(_,[pie_scan::w(_,S,Dom,_)],fix(F)) :- 
	set::member(Dom,["p3_dcf","p3_dpr","p3_dmg","p3_dinc","p3_ddec"]),
	F=sx::toReal(S),
	dom::domInRange(Dom,F),
	!
;
	F=sx::toReal(S),
	dbBase::nlSing(Dom,DomNL),
	sx::concat([DomNL," value out of range:",S],Mess),
	dlgIF::feil("kbScanTermX",[Mess],[],nill),
	!,fail.
kbScanTermX(_,[pie_scan::w(_,S,_,_)],atom(S)) :- 
	!.
kbBasicTerm(fstr(S,"p3_dli",_,_),str(S)) :- 
	!.
kbBasicTerm(fstr(S,"p3_dif",_,_),atom(S)) :- 
	!.
kbBasicTerm(fint(I,_),int(I)) :- 
	!.
kbBasicTerm(ffix(I,_),fix(I)) :- 
	!.
% -------------------------------------------------------------------------------------
% kbDefName                                                             
% -------------------------------------------------------------------------------------
class predicates
kbDefZ : (string_list,string_list) determ (i,o).
kbDefX : (integer,string_list,integer,integer,integer,integer,string_list,integer,string_list) determ (i,i,i,i,i,i,i,i,o).
kbDefY : (integer,integer,string,string) procedure (i,i,i,o).
clauses
% generates a default and non-existing KB name for given NL name. returns blank if unable to generate a non-existing name
%
kbDefName(Prefix,NL,Name) :-
	sx::sxStrip(NL,S2),
	sx::sxUnpack(S2,' ',Syllables),
	sx::sxl_count(Syllables,Longest,N_Syllables),
	Max2=globalVars::get_i("max_kbn"),
	Max=Max2-1,
	kbDefZ(Syllables,Name2),
	kbDefX(1,Syllables,1,N_Syllables,Longest,Max,Name2,0,Name3),
	sx::concat([Prefix|Name3],Name4),
	kbDefY(1,Max,Name4,Name),
	!
;
	Name="",
	!.
kbDefZ([],[]) :-
	!.
kbDefZ([_|T],[""|Tm]) :-
	kbDefz(T,Tm),
	!.
	
kbDefX(Round,Syllables,SyllableNo,N_Syllables,SyllableWidth,MaxLen,NameIn,NameLength,NameOut) :-
	NameLength=MaxLen,
	NameOut=NameIn,
	!
;
	Round>SyllableWidth,
	NameOut=NameIn,
	!
;
	SyllableNo>N_Syllables,
	NextRound=Round+1,
	kbDefX(NextRound,Syllables,1,N_Syllables,SyllableWidth,MaxLen,NameIn,NameLength,NameOut),
	!
;
	set::setGet(Syllables,SyllableNo,S),
	string::length(S)<Round,
	NextSyllableNo=SyllableNo+1,
	kbDefX(Round,Syllables,NextSyllableNo,N_Syllables,SyllableWidth,MaxLen,NameIn,NameLength,NameOut),
	!
;
	set::setGet(Syllables,SyllableNo,S),
	sx::frontstr(Round,S,F,_),
	sx::sxTail(1,F,_,Cs),
	C=string::subChar(Cs,0),
	not(sx::isLetter(C)),
	not(sx::isDigit(C)),
	NextSyllableNo=SyllableNo+1,
	kbDefX(Round,Syllables,NextSyllableNo,N_Syllables,SyllableWidth,MaxLen,NameIn,NameLength,NameOut),
	!
;	
	Round=1,
	set::setGet(Syllables,SyllableNo,S),
	sx::frontstr(Round,S,F,_),
	sx::sxTail(1,F,_,Cs),
    Cu=string::toUpperCase(Cs),
	set::setGet(NameIn,SyllableNo,T),
	Tm=string::concat(T,Cu),
	set::setReplace(NameIn,SyllableNo,Tm,NewNameIn),
	NextSyllableNo=SyllableNo+1,
	NewLength=NameLength+1,
	kbDefX(Round,Syllables,NextSyllableNo,N_Syllables,SyllableWidth,MaxLen,NewNameIn,NewLength,NameOut),
	!
;	
	Round>1,
	set::setGet(Syllables,SyllableNo,S),
	sx::frontstr(Round,S,F,_),
	sx::sxTail(1,F,_,Cs),
	set::setGet(NameIn,SyllableNo,T),
	Tm=string::concat(T,Cs),
	set::setReplace(NameIn,SyllableNo,Tm,NewNameIn),
	NextSyllableNo=SyllableNo+1,
	NewLength=NameLength+1,
	kbDefX(Round,Syllables,NextSyllableNo,N_Syllables,SyllableWidth,MaxLen,NewNameIn,NewLength,NameOut),
	!.
kbDefY(N,Max,NameIn,NameOut) :-
	not(dbBase::kbID_F(NameIn,_,_,_)),
	NameOut=NameIn,
	!
;
	N=9,
	kbDefY(0,Max,NameIn,NameOut),
	!
;
	string::length(NameIn)=Max,
	sx::frontchar(NameIn,F,T2),
	Fs=string::charToString(F),
	sx::frontchar(T2,_,T3),
	Ns=toString(N),
	sx::concat([Fs,T3,Ns],Name),
	Next=N+1,
	kbDefY(Next,Max,Name,NameOut),
	!
;
	NameOut=NameIn,
	!.
% -------------------------------------------------------------------------------------
% kbPrintSent
% -------------------------------------------------------------------------------------
class predicates
kbPrintAnno : (integer,string,string,string) procedure (i,i,i,o).
kbPS_DEs : (integer,sterml,sterml,string) determ (i,i,i,o).
kbPS_Body : (integer,string,string,string,sterm,string) determ (i,i,i,i,i,o).
kbPS_W : (sterml,integer,integer) determ (i,i,o).
clauses
% parameters:
% 1	in	0/1 header
% 2	in	0/1 trailer
% 3 in  0/1 annotation
%
kbPrintSent(1,1,PrintAnn,Loc,ST,S) :-
    dbBase::getRecord(Loc,_,R,_,_,T,Realm,SNo,SubNo,STk,LT,_),
	set::setGet(["Derived","Source"],R,Typ),
	SRef=dbBase::dbRef(Loc),
	S2=toString(SNo),
	S3=toString(SubNo),
	sx::concat([Typ," sentence ",SRef,"\n",Realm,"|",S2,"|",S3," ",T,"\n\n",ST],Shead),
	dbBase::stComps(STk,Card,P,E1,E2),
	kbPrintAnno(PrintAnn,"p3_iST",STk,Anno),
	kbPS_Body(Card,P,E1,E2,LT,Sbody),
	sx::concat([Shead,"\n\n",Anno,Sbody],S),
	!.
kbPrintSent(1,-1,PrintAnn,Loc,_,S) :-
    dbBase::getRecord(Loc,_,R,_,_,T,Realm,SNo,SubNo,STk,LT,_),
	set::setGet(["Derived","Source"],R,Typ),
	SRef=dbBase::dbRef(Loc),
	S2=toString(SNo),
	S3=toString(SubNo),
	sx::concat([Typ," sentence ",SRef,"\n",Realm,"|",S2,"|",S3," ",T],Shead),
	dbBase::stComps(STk,Card,P,E1,E2),
	kbPrintAnno(PrintAnn,"p3_iST",STk,Anno),
	kbPS_Body(Card,P,E1,E2,LT,Sbody),
	sx::concat([Shead,"\n\n",Anno,Sbody],S),
	!.
kbPrintSent(-1,1,PrintAnn,Loc,ST,S) :-
    dbBase::getRecord(Loc,_,R,_,_,_,_,_,_,STk,LT,_),
	set::setGet(["Derived","Source"],R,Typ),
	dbBase::stComps(STk,Card,P,E1,E2),
	kbPS_Body(Card,P,E1,E2,LT,Sbody),
	kbPrintAnno(PrintAnn,"p3_iST",STk,Anno),
	sx::concat([Typ," ",ST,"\n\n",Anno,Sbody],S),
	!.
kbPrintSent(-1,-1,_,Loc,_,S) :-
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,_,STk,LT,_),
	dbBase::stComps(STk,Card,P,E1,E2),
	kbPS_Body(Card,P,E1,E2,LT,S),
	!.
% generates print layout text-format for one sentence
% parameters:
% 1	in	cardinality
%
kbPS_Body(2,P,Ea,Eb,LT,S) :-
	dbBase::smClassDefTree(P,"p",PDef),
	PDef=cmp(_,[_,_,cmp("ATTRS",PDEs)]),
	dbBase::smClassDefTree(Ea,"e",EaDef),
	EaDef=cmp(_,[_,cmp("ATTRS",EaDEs)]),
	dbBase::smClassDefTree(Eb,"e",EbDef),
	EbDef=cmp(_,[_,cmp("ATTRS",EbDEs)]),
	LT=cmp(_,[cmp(_,Eavs),cmp(_,Ebvs)|Pvs]),
	set::conc([EaDEs,EbDEs,PDEs],DEs),
	set::conc([Eavs,Ebvs,Pvs],Vs),
	kbPS_W(DEs,0,W),
	kbPS_DEs(W,DEs,Vs,S),
	!
;	
	dbBase::smClassDefTree(Ea,"e",EaDef),
	EaDef=cmp(_,[_,cmp("ATTRS",EaDEs)]),
	dbBase::smClassDefTree(Eb,"e",EbDef),
	EbDef=cmp(_,[_,cmp("ATTRS",EbDEs)]),
	LT=cmp(_,[cmp(_,Eavs),cmp(_,Ebvs)]),
	set::conc([EaDEs,EbDEs],DEs),
	set::conc([Eavs,Ebvs],Vs),
	kbPS_W(DEs,0,W),
	kbPS_DEs(W,DEs,Vs,S),
	!.
kbPS_Body(1,P,Ea,_,LT,S) :-
	dbBase::smClassDefTree(P,"p",PDef),
	PDef=cmp(_,[_,_,cmp("ATTRS",PDEs)]),
	dbBase::smClassDefTree(Ea,"e",EaDef),
	EaDef=cmp(_,[_,cmp("ATTRS",EaDEs)]),
	LT=cmp(_,[cmp(_,Eavs)|Pvs]),
	set::conc([EaDEs,PDEs],DEs),
	set::conc([Eavs,Pvs],Vs),
	kbPS_W(DEs,0,W),
	kbPS_DEs(W,DEs,Vs,S),
	!
;	
	dbBase::smClassDefTree(Ea,"e",EaDef),
	EaDef=cmp(_,[_,cmp("ATTRS",EaDEs)]),
	LT=cmp(_,[cmp(_,Eavs)]),
	kbPS_W(EaDEs,0,W),
	kbPS_DEs(W,EaDEs,Eavs,S),
	!.

% generates print layout text-format for one DE/value pair
% parameters:
% 1	in	field width
% 2	in	list of DE definitions
% 3	in	list of values corr. to p#2
% 4	out	text
%
kbPS_DEs(_,[],_,"") :-
	!.
kbPS_DEs(FW,[fstr(D,Dom,Form,_)|Ds],[V|Vs],S) :-
	nl::conValX(V,Dom,Form,Vnq),
	sx::sxUnquote(Vnq,Vn),
	dbBase::nlSing(D,Dn),
	sx::sxR_Pad(Dn," ",FW,Dnp),
	sx::concat([Dnp,": ",Vn],Sm),
	kbPS_DEs(FW,Ds,Vs,Sr),
	sx::concat([Sm,"\n",Sr],S),
	!.

% returns the widest DE header text
%	
kbPS_W([],W,W) :-
	!.
kbPS_W([fstr(D,_,_,_)|Ds],CW,MW) :-
	dbBase::nlSing(D,Dn),
	W=string::length(Dn),
	set::setMax([CW,W],NW),
	kbPS_W(Ds,NW,MW),
	!.			

kbPrintAnno(1,Type,Node,Anno) :-
	kbGetAnnotationF(Type,Node,Note,_),
	Note2=sx::replaceAll(Note,"\n"," "),
	Note3=sx::replaceAll(Note2,"  "," "),
	sx::sxBlockNL(0,60,Note3,Note4),
	Anno=string::concat(Note4,"\n\n"),
	!.
kbPrintAnno(_,_,_,"") :-
	!.
% -------------------------------------------------------------------------------------
% kbGetAnno...                                            
% -------------------------------------------------------------------------------------
kbGetAnnotation(Type,Node,Note,Loc) :-
	kbGetAnnotationF(Type,Node,Note,Loc),
	!
;
	sx::concat([Type," ",Node],S),
	dlgIF::feil("kbGetAnnotation",[S,"Unknown annotation"],[],nill),
	!,fail.
	
kbGetAnnotationF(Type,Node,Note,Loc) :-
	dbBase::chainSentences("p3_pIntd",LT,Loc),
	LT=cmp("p3_pIntd",[cmp("p3_eAnno",[atom(Type),str(Node),str(Nt)])]),
	Note=Nt,
	!.
% ----------------------------------------------------------------------------------------------
% kbSG_Count
% ----------------------------------------------------------------------------------------------
class predicates
kbSG_CountX : (chainDB,string,count,count).
kbSG_CountY : (integer,count,count).	
clauses

% counts the number of type-1 and type-2 record of given SG
%
kbSG_Count(SG,N1,N2) :-
    dbBase::kbIsSql,
    !,N1=sqlBase::chainTypeCount(1,SG),
    N2=sqlBase::chainTypeCount(2,SG),
    !
;    
	C1=count::newCounter(0),
	C2=count::newCounter(0),
	dbaBase::dbaMap(SG,KB,_),
	kbSG_CountX(KB,SG,C1,C2),
	N1=C1:get(),
	N2=C2:get(),
	!.
	
kbSG_CountX(KB,P,C1,C2) :-
	KB:chain_terms(P,Cl,_),
	Cl=clause(Typ,_,_,_,_,_,atom("true")),
	AZ=math::abs(Typ), AZ<10,
	kbSG_CountY(Typ,C1,C2),
	fail
;
	!.
	
kbSG_CountY(1,C,_) :- C:inc(_),!.		
kbSG_CountY(2,_,C) :- C:inc(_),!.		
kbSG_CountY(-1,C,_) :- C:inc(_),!.		
kbSG_CountY(-2,_,C) :- C:inc(_),!.		
kbSG_CountY(Typ,_,_) :- dlgIF::feil("kbSG_CountY",[],[Typ],nill),!.		
% -------------------------------------------------------------------------------------
% kbFuncBody/L                                                            
% -------------------------------------------------------------------------------------
class predicates
kbFBX : (lst,string).
kbFuncBodiesX : (string_list,string_list) determ (i,o).
kbFuncBodiesL : (string,string_list,string_list_list) determ (i,i,o).
clauses

% parameters:
% 1	in	function (KB)
% 2	out	function bodies (NL format) for a given function
%
functionBodies(Func,NLs) :-
	Arr=lst::new(1),
	kbFBX(Arr,Func),
	Locs=Arr:get(),
	kbFuncBodiesX(Locs,NLs),
	!.

kbFBX(Arr,Func) :-
	dbBase::chainTerms(Func,_,_,Loc),
	    SLoc=toString(Loc),
        Arr:add(SLoc),
        fail
;
	!.	
	
kbFuncBodiesX([],[]) :- 
	!.	
kbFuncBodiesX([SLoc|SLocs],[Body|Bodies]) :-
    Loc=toTerm(SLoc),
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,_,_,LT,RT),
	nl::conFunction(cmp(":-",[LT,RT]),Body),
	kbFuncBodiesX(SLocs,Bodies),
	!.

% parameters:
% 1	in	NL/PR
% 2	in	function (KB)
% 3	out	function bodies (NL) as "list[bodies] of list[statements]"
%
functionStatements(Form,Func,Bodies) :-
	Arr=lst::new(1),
	kbFBX(Arr,Func),
	Locs=Arr:get(),
	kbFuncBodiesL(Form,Locs,Bodies),
	!.

kbFuncBodiesL(_,[],[]) :- 
	!.	
kbFuncBodiesL(Form,[SLoc|SLocs],[Body|Bodies]) :-
	Form="NL",	
    Loc=toTerm(SLoc),
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,_,_,LT,RT),
	nl::conFunctionL(cmp(":-",[LT,RT]),Body),
	kbFuncBodiesL(Form,SLocs,Bodies),
	!
;
	Form="PR",	
    Loc=toTerm(SLoc),
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,_,_,LT,RT),
	nl::cprRecord(LT,RT,B),
	Body=[B],
	kbFuncBodiesL(Form,SLocs,Bodies),
	!.
% ----------------------------------------------------------------------------------------------
% dbuDeleteST_Reason/All
% ----------------------------------------------------------------------------------------------
class predicates
dbuDelSTR : (string,count).
clauses

dbuDeleteReason(N) :-
            dlgIF::msgStatus("Deleting all inexact rule reasoning sentences..."),
    vpi::cursorSetWait(),
    dbBase::stTypeDelete(2,"r|u",N1),
    dbBase::stTypeDelete(2,"r|w",N2),
            N=N1+N2,
            dlgIF::msgStatus(""),
	!.

dbuDeleteST_Reason(ST,N) :-
            vpi::cursorSetWait(),
	Count=count::newCounter(0),
	dbuDelSTR(ST,Count),
	N=Count:get(),
            dlgIF::msgStatus(""),
	!.
	
dbuDelSTR(ST,Count) :-
    dbBase::kbIsSql,
    sqlBase::chainSentencesDirect("r",LT,Ref),
        LT=cmp("r",[cmp(E,[str(STr)|_])]),
        set::member(E,["u","w"]),
        ST=STr,
        sqlBase::recordDelete("r",Ref),
        fail    
;
    not(dbBase::kbIsSql),
	dbaBase::dbaMap("r",KB,_),
	KB:chain_terms("r",Term,Ref),
        Term=clause(_,_,_,_,_,cmp("r",[cmp(E,[str(STr)|_])]),atom("true")),
        set::member(E,["u","w"]),
        ST=STr,
        dbaBase::dbaDeleteRecord(KB,Ref),
        Count:inc(_),
        fail
;
	!.
% ----------------------------------------------------------------------------------------------
% dbuChainType
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	chain
% 2	out	1=internal; 2=public; 3=external
% 3	out	0=read-only; 1=read-write
% 4	out	KB Id
%
dbuChainType(Chain,Typ,RW,Id) :-
    dbBase::kbIsSql,
    !,Typ=1,
    RW=1,
    Id="X",
    !
;    
	dbaBase::pubChain(Chain,RW),
	Typ=2,
	Id=globalVars::get_s("kb_id"),
	!
;
	dbaBase::extChain(Chain,KB,RW),
	dbaBase::database(KB,_,Id,_),
	Typ=3,
	!
;
	Typ=1,
	RW=1,
	Id=globalVars::get_s("kb_id"),
	!.
% ----------------------------------------------------------------------------------------------
% dbuCR_GetTD
% ----------------------------------------------------------------------------------------------
dbuCR_GetTD(Rule,Typ,Dom) :-
    dbBase::kbIsSql,
	string::frontchar(Rule,F,R),
	F='*',
    sqlBase::getIrDefF(R,Body,_,_),
	Body=cmp(Typ,_),
	dbBase::smClassDefF(R,"h",_,_,Dom,_,_),
	!
;
    dbBase::kbIsSql,
    sqlBase::getIrDefF(Rule,Body,_,_),
	Body=cmp(Typ,_),
	dbBase::smClassDefF(Rule,"h",_,_,Dom,_,_),
	!
;
    dbBase::kbIsSql,
	dbBase::smClassDefF(Rule,"h",_,_,Dom,_,_),
	dlgIF::feil("dbuCR_GetTD",["Unknown or invalid inexact rule: ",Rule],[],nill),!,fail
;
    not(dbBase::kbIsSql),
	string::frontchar(Rule,F,R),
	F='*',
	dbuCR_GetF(R,Body,_,_),
	Body=cmp(Typ,_),
	kbBase::irDeclarationF(Rule,Dom,_,_,_),
	!
;
    not(dbBase::kbIsSql),
	dbuCR_GetF(Rule,Body,_,_),
	Body=cmp(Typ,_),
	kbBase::irDeclarationF(Rule,Dom,_,_,_),
	!
;
    not(dbBase::kbIsSql),
	kbBase::irDeclarationF(Rule,Dom,_,_,_),
	dlgIF::feil("dbuCR_GetTD",["Unknown or invalid inexact rule: ",Rule],[],nill),!,fail
;
	dlgIF::feil("dbuCR_GetTD",["No declaration of inexact rule: ",Rule],[],nill),!,fail.
% ----------------------------------------------------------------------------------------------
% dbuCR_Get
% ----------------------------------------------------------------------------------------------
dbuCR_Get(Rule,Body,Params,RT) :-
	string::frontchar(Rule,F,R),
	F='*',
	dbuCR_GetF(R,Body,Params,RT),
	!
;
	dbuCR_GetF(Rule,Body,Params,RT),
	!
;
	dlgIF::feil("dbuCR_Get",["Unknown inexact rule",Rule,"!"],[],nill),
	!,fail.	
		
dbuCR_GetF(Rule,Body,Params,RT) :-
    dbBase::getIrDefF(Rule,Body,Loc),
    dbBase::getRecord(Loc,RT,_,_,_,_,_,_,_,_,_,_),
    dbBase::smClassDefTreeF(Rule,"h",Def),
    Def=cmp("REL_TYPE", [_,_,cmp("FPAR",Params)]),
	!.
% ----------------------------------------------------------------------------------------------
% getQuestion
% ----------------------------------------------------------------------------------------------
getQuestion(Rule,Question,QuestionDomain,Default) :-
	dbuCR_GetF(Rule,Body,_,_),
	Body=cmp("p3_iCdat",[_,cmp(_,[atom(QuestionDomain),str(Question)]),_,_,_,Default,_]),
	!
;
    Mess=string::concat("Unknown question-rule '",Rule,"'!"),
    dlgIF::dlgNote(1,Mess),!,fail.   
% ----------------------------------------------------------------------------------------------
% dbuCR_SwitchValDom
% ----------------------------------------------------------------------------------------------
dbuCR_SwitchValDom(Rule,Dom) :- 
	dbuCR_GetF(Rule,LT,_,_),
	LT=cmp("p3_iCswi",[_,atom(DE)|_]), 
	dbBase::smClassDef(DE,"a",_,_,Dom,_,_),
	!
;	
	dbuCR_GetF(Rule,LT,_,_),
	LT=cmp("p3_iCswi",[_,cmp(Rc,_)|_]),
	dbBase::smClassDef(Rc,"h",_,_,Dom,_,_),
	!
;
    Mess=string::concat("Cannot determine switch-value domain for rule: ",Rule),
	dlgIF::dlgNote(3,Mess),!,fail.
% ----------------------------------------------------------------------------------------------
% dbuCheckSTD
% ----------------------------------------------------------------------------------------------
dbuCheckSTD(Open) :-
	dbBase::stExists(2,"x|p3_eVers"),
	dbBase::stExists(2,"x|p3_eTerm"),
	dbBase::stExists(2,"q|a"),
	dbBase::stExists(2,"q|e"),
	%dbBase::stExists(2,"q|d"),
	dbBase::stExists(2,"q|p"),
	dbBase::stExists(2,"v|t"),
	!
;	
    Open=1,
    dlgIF::dlgNote(0,"Cannot find all meta/system administration chains!"),
	!
;	
    Open=-1,
	Reply=vpiCommonDialogs::ask("Cannot find all meta/system administration chains!","Do you to still want to close the KB?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	!.
% ----------------------------------------------------------------------------------------------
% dbuDE_Defs
% ----------------------------------------------------------------------------------------------
dbuDE_Defs([],[]) :-
	!.
dbuDE_Defs([atom(D)|Ds],[Def|Defs]) :-
	dbBase::smClassDefTree(D,"a",Def),
	dbuDE_Defs(Ds,Defs),
	!.
% ----------------------------------------------------------------------------------------------
% dbuRT_Text
% ----------------------------------------------------------------------------------------------
class predicates
dbuRT_A : (integer,string) determ (i,o).
clauses

dbuRT_Text(UC,RT,S) :-
	UC=1,
	dbuRT_A(RT,S2),
	S=string::toUpperCase(S2),
	!
;	
	dbuRT_A(RT,S),
	!.
	
dbuRT_A(RT,S) :-
	dbBase::recordType(RT,_,Act,Susp),
	%set::setGet(["derived sentence","sentence","sentence rule","function","derived rule","certainty rule"],BType,Sa),
	B=Act+2,
	set::setGet(["inactive","","active"],B,Sb),
	C=Susp+2,
	set::setGet(["","","suspended"],C,Sc),
    %sx::sxlJoin([Sb,Sc,Sa]," ",S),
    sx::sxlJoin([Sb,Sc]," ",S),
	!.
% ----------------------------------------------------------------------------------------------
% dbuSentenceR
% ----------------------------------------------------------------------------------------------
dbuSentenceR(R,N,M,S) :-
	sx::sxUnpack(S,'|',La),
	La=[R,Ns,Ms],
	N=sx::toInt(Ns),
	M=sx::toInt(Ms),
	!.
% ----------------------------------------------------------------------------------------------
% dbaDeleteRule/SubRule                                                
% ----------------------------------------------------------------------------------------------
% deletes all rule records with given Realm and No
%
dbaDeleteRule(Realm,No) :-
    dbBase::flagsClear,
    dbBase::realmDeleteNo(Realm,No),
	!.

% deletes all rule records with given Realm, No and SubNo
%	
dbaDeleteSubRule(Realm,No,SubNo) :-
    dbBase::flagsClear,
    dbBase::realmDeleteSubno(Realm,No,SubNo),
	!.
% ----------------------------------------------------------------------------------------------
% dbaCheckRuleReplace
% ----------------------------------------------------------------------------------------------
dbaCheckRuleReplace(Realm,No) :-
	globalVars::get_i("all_flag")=1,
	vpi::cursorSetWait(),
	dbBase::realmDeleteNo(Realm,No),
	!	
;
	dbBase::realmNoCount(Realm,No)>0,
	!,SNo=toString(No),
	sx::concat(["Rule record ",Realm,"|",SNo," exists"],Mess1),
	Reply=vpiCommonDialogs::ask(Mess1,"Do you want to replace it?",["Yes","Yes ALL","No"]),
	Reply<>vpiCommonDialogs::resp_3,
	set::setGet([0,1],Flag,Reply),
	set::setGet([-1,1],Flag,FlagX),
	globalVars::set_i("all_flag",FlagX),
	vpi::cursorSetWait(),
	dbBase::realmDeleteNo(Realm,No),
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
% dbaCheckSubRuleReplace
% ----------------------------------------------------------------------------------------------
dbaCheckSubRuleReplace(Realm,No,SubNo) :-
	globalVars::get_i("all_flag")=1,
	vpi::cursorSetWait(),
	dbBase::realmDeleteSubno(Realm,No,SubNo),
	!	
;
	dbBase::realmSubnoCount(Realm,No,SubNo)>0,
	!,
	SNo=toString(No),
	SSubNo=toString(SubNo),
	sx::concat(["Rule record ",Realm,"|",SNo,"|",SSubNo," exists"],Mess1),
	Reply=vpiCommonDialogs::ask(Mess1,"Do you want to replace it?",["Yes","Yes ALL","No"]),
	Reply<>vpiCommonDialogs::resp_3,
	set::setGet([0,1],Flag,Reply),
	set::setGet([-1,1],Flag,FlagX),
	globalVars::set_i("all_flag",FlagX),
	vpi::cursorSetWait(),
	dbBase::realmDeleteSubno(Realm,No,SubNo),
	!
;
	!.	
end implement dbExtension
