/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement dbOpen
open core,termDomain,resourceIdentifiers,globalDomains

constants
className = "com/rudstrand/dbOpen/dbOpen".
classVersion = "1.0".

class predicates
deleteDB_Version : ().
setCurrentTerminology : () procedure.
setFileNames : (string) determ.
setSegmentContext : () determ.
versionUpgrade : (integer).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% checkKB_Files                                                     
% ----------------------------------------------------------------------------------------------
checkKB_Files(KB) :-
	KB="",
	!,fail
;
    sqlBase::names(KB,_,_,_),
    !
;
	file::existFile(KB),
	!
;		
	fioExt::filChangeType(KB,"3kc",Fa),
	file::existFile(Fa),
	file::delete(Fa),
	fail
;
	fioExt::filChangeType(KB,"3eq",Fb),
	file::existFile(Fb),
	file::delete(Fb),
	!
;
    !.
% -------------------------------------------------------------------------------------------------------------
% existKB
% -------------------------------------------------------------------------------------------------------------
existKB(File) :-
    sqlBase::names(File,DB,KB,_),
    !,sqlBase::openConnection(),
    sqlBase::openDb(DB),
    sqlBase::existKb(KB)>0,
    !
;    
    file::existFile(File),
    !.
% -------------------------------------------------------------------------------------------------------------
% deleteDB_Version
% -------------------------------------------------------------------------------------------------------------
deleteDB_Version() :-
	dbBase::chainSentences("x",Term,Loc),
	Term=cmp(_,[cmp("p3_eVers",[_,str(Version)])]),
	set::member(Version,["DB VERSION","KB","KB compatibility"]),
	dbBase::recordDeleteOutright(Loc),
	!
;
	!.
% -------------------------------------------------------------------------------------------------------------
% versionCheck
% -------------------------------------------------------------------------------------------------------------
class predicates
versionComp : (string,integer,integer) determ.
versionCompX : (string,integer,integer) procedure.
clauses

versionCheck() :-
    % version 1 KB
    V=globalVars::get_i("p3v"),
    Vs=toString(V),
    R=globalVars::get_i("p3r"),
    Rs=toString(R),
    B=globalVars::get_i("p3b"),
    Bs=toString(B),
    VRB=string::concat(Vs,".",Rs,".",Bs),
    S=globalVars::get_s("p3s"),
    D=globalVars::get_s("p3d"),
	sx::concat(["Program version ",VRB," ",D," ",S],M1),
	dlgIF::msgS(M1),
	dbBase::getProgramBuild(PV,_,_,_,_),
	dbOpen::getDB_Version(KV),
	KVF=KV/100.0,
	KV<=PV,
	sx::fixFormatR(KVF,2,0,S2),
	sx::concat(["KB version ",S2],M2),
	dlgIF::msgS(M2),
	globalVars::set_i("kbVersion",KV),
	versionComp(VRB,PV,KV),
	!
;
	dbBase::getProgramBuild(PV,_,_,_,_),
	dbOpen::getDB_Version(KV),
	KV>PV,
	KV<200,
    dlgIF::dlgNote(1,"This knowledge base was created under a more recent version of PRO3.EXE, and can only be opened under this or later versions!"),
	tt::tt_close(),
	TW=guix::getWinOfClas(1),
	vpi::winDestroy(TW),
	!,fail
;
	tt::tt_close(),
	TW=guix::getWinOfClas(1),
	vpi::winDestroy(TW),
	!,fail.	

% succeeds if the given version is compatible with the program version
%
versionComp(_,_,V) :-
    V>199,
    dlgIF::dlgNote(1,"This version of PRO3.EXE cannot use pre version 5 databases!"),
	tt::tt_close(),
	TW=guix::getWinOfClas(1),
	vpi::winDestroy(TW),
	!,fail.	
versionComp(_,V,V) :-
	globalVars::set_i("knowopen",1),
	globalVars::set_i("p3up",-1),
	!.
versionComp(PV,P,K) :-
	versionCompX(PV,P,K),
	dlgIF::dlgExp(17),
	globalVars::set_i("p3up",1),
	globalVars::set_i("knowopen",2),
	!.

versionCompX(_,V,V) :-
	!.
versionCompX(PV,V,100) :-
	dlgIF::dlgNote(0,"Upgrade of KB from version 1.0 to 1.1. No changes required!"),
	versionCompX(PV,V,101),
	!.
versionCompX(_,_,_) :-
	!.
% -------------------------------------------------------------------------------------------------------------
% newVersion
% -------------------------------------------------------------------------------------------------------------
newVersion() :-
	globalVars::get_i("p3up")=-1,
	dep::loadGraphs,
	! 
;	
	% KB was upgraded during opening
	dbBase::getProgramBuild(PV,_,_,_,_),
	deleteDB_Version,
	dbBase::recordInsert(2,"$SYS",0,0,cmp("x",[cmp("p3_eVers",[int(PV),str("DB VERSION")])]),atom("true"),_),
	W=guix::getWinOfClas(1), 
    Dlg = directories::new(),
    Dlg:show(W),
	versionUpgrade(PV),
	globalVars::set_i("p3up",-1),
	!.

versionUpgrade(_) :-
	dep::loadGraphs,
	dlgIF::dlgExp(11),
	Reply=vpiCommonDialogs::ask("New Pro/3 Version","Do you want to reload standard knowledge files? (recommended!)",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	proc::loadStandardConf,
	!;!.	
% ----------------------------------------------------------------------------------------------
% reinitialize                                                 
% ----------------------------------------------------------------------------------------------
% initializes the open KB (after confirmation)
%
reinitialize() :-
    dbBase::kbIsSql,
    !,Reply=vpiCommonDialogs::ask("Re-initializing KB","All contents of the main database will be lost. Do you want to re-initialize?",["Yes","No"]),
    Reply=vpiCommonDialogs::resp_default, 
    sqlBase::deleteKb,
	F=globalVars::get_s("kb_file"),
    openKB_Step2(F),
	!
;
    Reply=vpiCommonDialogs::ask("Re-initializing KB","All contents of the main database will be lost. Do you want to re-initialize?",["Yes","No"]),
    Reply=vpiCommonDialogs::resp_default, 
    globalVars::set_i("knowopen",1),
	closeKB(1,1,1,1),
	F=globalVars::get_s("kb_file"),
	file::delete(F),
    openKB_Step2(F),
	!.
% ----------------------------------------------------------------------------------------------
% storeSettings                                                         
% ----------------------------------------------------------------------------------------------
storeSettings () :-
	dbBase::stTypeDelete(2,"x|p3_eConf",_),
	TOK=globalVars::get_i("max_tok"),
	RLN=globalVars::get_i("max_rln"),
	KBN=globalVars::get_i("max_kbn"),
	dbBase::recordInsert(2,"$SYS",0,0,cmp("x",[cmp("p3_eConf",[atom("p3_iTokn"),int(TOK)])]),atom("true"),_),
	dbBase::recordInsert(2,"$SYS",0,0,cmp("x",[cmp("p3_eConf",[atom("p3_iMxRN"),int(RLN)])]),atom("true"),_),
	dbBase::recordInsert(2,"$SYS",0,0,cmp("x",[cmp("p3_eConf",[atom("p3_iMxKN"),int(KBN)])]),atom("true"),_),
	!.
% -------------------------------------------------------------------------------------------------------------
% create
% -------------------------------------------------------------------------------------------------------------
% creates and closes a new KB
%
create(FileName) :-
    sqlBase::names(FileName,D,K,_),
    % SQL-type
    !,sqlBase::openConnection,
    sqlBase::openDb(D),
    sqlBase::createKb(K),
    dbBase::setSql,
	globalVars::set_s("kb_id","X"),
	dbBase::setContext([]),
	dbBase::getProgramBuild(VR,_,_,_,_),
    dbBase::recordInsert(2,"$SYS",0,0,cmp("x",[cmp("p3_eVers",[int(VR),str("DB VERSION")])]),atom("true"),_),
	storeDBID("X"),
	storeSettings,
	dlgIF::dbaStatuses,
	dlgIF::msgS("KB initialized."),
    !
;
	% 3KB-type
	DB=chainDB::db_create(FileName,chainDB::in_file),
	Id=globalVars::get_s("kb_id"),
	dbaBase::dropTables,
    dbaBase::addDatabase(DB,FileName,Id),
	DB:db_reuserefs(1),
	dbaBase::dbaIndexesOps(DB,'C',2), % create
	dbBase::getProgramBuild(VR,_,_,_,_),
    dbaBase::dbaInsertRecord(2,"$SYS",0,0,cmp("x",[cmp("p3_eVers",[int(VR),str("DB VERSION")])]),atom("true"),_),
	storeDBID(Id),
	storeSettings,
	dbaUtil::dbaCloseBasic(DB,'C'),
	dlgIF::dbaStatuses,
	dlgIF::msgS("KB initialized."),
	!.
% -------------------------------------------------------------------------------------------------------------
% dbLog
% -------------------------------------------------------------------------------------------------------------
dbLog(KB_File,FL) :-
	file::existfile(FL),
	datetime::tsNow(T),
	datetime::tsStr(T,Ts),
	S=file::readString(FL),
	S2=string::concat(S,"\n",KB_File," ",Ts),
	file::writeString(FL,S2),
	!
;
    W=guix::getWinOfClas(1),
    Dlg=license::new(),
    Dlg:show(W),
	datetime::tsNow(T),
	datetime::tsStr(T,Ts),
	S2=string::concat("Initialized ",Ts),
	file::writeString(FL,S2),
	dbLog(KB_File,FL),
	!.
% -------------------------------------------------------------------------------------------------------------
% getDB_Version
% -------------------------------------------------------------------------------------------------------------
getDB_Version(VR) :-
    dbBase::kbIsSql,
    !,sqlBase::chainSentencesDirect("x",LT),
    LT=cmp(_,[cmp("p3_eVers",[int(VR),str("DB VERSION")])]),
    !
;    
    dbaBase::database(DB,_,_,0),
	DB:chain_terms("x",Term,_),
	Term=clause(_,_,_,_,_,cmp(_,[cmp("p3_eVers",[int(VR),str("DB VERSION")])]),_),
	!.
% ----------------------------------------------------------------------------------------------
% storeDBID                                                            
% ----------------------------------------------------------------------------------------------
storeDBID(Id) :-
	dbBase::chainSentences("x",LTerm,Loc),
        LTerm=cmp(_,[cmp("p3_eDbas",_)]),
        dbBase::recordDelete(Loc),
        fail
;
	dbBase::recordInsert(2,"$SYS",0,0,cmp("x",[cmp("p3_eDbas",[str(Id)])]),atom("true"),_),
	!.
% ----------------------------------------------------------------------------------------------
% closeKB                                                              
% ----------------------------------------------------------------------------------------------
class predicates
closeKB_X : (integer I1,integer I2,integer I3) determ.
closeKB_Step2 : (integer I1,integer I2,integer I3) determ.
closeTrace : (integer I1) procedure.
dbaUpdatePublic : (integer I1,integer I2) procedure.
clauses

% parameters:
% 1	in	1=close when exiting Pro/3; 0=close during session with Pro/3
% 2	in	1=check for STD-files; -1=not
% 3	in	1=questions asked; -1=no questions asked
% 4	in	-1/1 close trace
%
closeKB(Close,Check,Ask,CloseTrace) :-
	Check=1,
	globalVars::get_i("knowopen")>0,
	not(dbExtension::dbuCheckSTD(-1)),
	!,fail
;
    dep::graphsAreChanged,
    globalVars::set_i("knowopen",2),
	closeKB_X(Close,Ask,CloseTrace),
	!
;
    dbBase::kbIsSql,
    !,closeKB_X(Close,-1,CloseTrace),
    !
;
	closeKB_X(Close,Ask,CloseTrace),
	!.
		
closeKB_X(1,Ask,CloseTrace) :-
    N=globalVars::get_i("solref"),
	tt::tt_i(1,"External solution set counter",N),
	M=globalVars::get_i("solref"),
	tt::tt_i(1,"Internal solution set counter",M),
	globalVars::get_i("knowopen")=0,
    dlgIF::dbaStatuses,
	closeTrace(CloseTrace),
	!
;
	globalVars::get_i("knowopen")=1,
	% no changes - close
	datetime::timNow(Dat,Tim),
	dbaExt::writeLocations,
	closeKB_Step2(0,Dat,Tim),
	closeTrace(CloseTrace),
	!
;	
	Ask=1,
	globalVars::get_i("backup_active")=1,
	globalVars::get_i("knowopen")=2,
	globalVars::get_i("knowerr")=1,
	% changes and errors
	dlgIF::dlgExp(1),
	Reply=vpiCommonDialogs::ask("Save changes in KB?","Error(s) have occurred during your session. Do you still want to save the changes to file? Since you are using the backup option you can also restore the opening state.",["Save","Restore","Cancel"]),
	!,
	Reply<>2,
	dbaExt::writeLocations,
	Action=Reply+1,
	datetime::timNow(Dat,Tim),
	closeKB_Step2(Action,Dat,Tim),
	closeTrace(CloseTrace),
	!	
;	
	Ask=-1,
	globalVars::get_i("backup_active")=1,
	globalVars::get_i("knowopen")=2,
	globalVars::get_i("knowerr")=1,
	% changes and errors
	dbaExt::writeLocations,
	datetime::timNow(Dat,Tim),
	closeKB_Step2(1,Dat,Tim),
	closeTrace(CloseTrace),
	!	
;
	Ask=-1,
	globalVars::get_i("backup_active")=1,
	globalVars::get_i("knowopen")=2,
	% changes and no errors
	dbaExt::writeLocations,
    datetime::timNow(Dat,Tim),
	closeKB_Step2(1,Dat,Tim),
	closeTrace(CloseTrace),
	!
;
	Ask=1,
	globalVars::get_i("backup_active")=1,
	globalVars::get_i("knowopen")=2,
	% changes and no errors
	dlgIF::dlgExp(2),
	Reply=vpiCommonDialogs::ask("Save changes in KB?","The KB has been changed. Since you are using the backup option you can abandon these changes and restore the opening state.",["Save","Restore","Cancel Close"]),
	!,
	Reply<>2,
	dbaExt::writeLocations,
	Action=Reply+1,
	datetime::timNow(Dat,Tim),
	closeKB_Step2(Action,Dat,Tim),
	closeTrace(CloseTrace),
	!
;
	globalVars::get_i("backup_active")=-1,
	globalVars::get_i("knowopen")=2,
	dbaExt::writeLocations,
	datetime::timNow(Dat,Tim),
	closeKB_Step2(1,Dat,Tim),
	closeTrace(CloseTrace),
	!.

closeKB_X(0,Ask,CloseTrace) :-
	N=globalVars::get_i("solref"),
	tt::tt_i(1,"External solution set counter",N),
	M=globalVars::get_i("solref"),
	tt::tt_i(1,"Internal solution set counter",M),
	globalVars::get_i("knowopen")=0,
	dlgIF::dbaStatuses,
	closeTrace(CloseTrace),
	!
;
	globalVars::get_i("knowopen")=1,
	% no changes - close
	datetime::timNow(Dat,Tim),
	closeKB_Step2(0,Dat,Tim),
	closeTrace(CloseTrace),
	!
;	
	Ask=1,
	globalVars::get_i("knowopen")=2,
	globalVars::get_i("knowerr")=1,
	% changes and errors
	dlgIF::dlgExp(1),
	Reply=vpiCommonDialogs::ask("Save changes in KB?","Error(s) have occurred during your session. Do you still want to save the changes to file? Since you are using the backup option you can also restore the opening state.",["Save","Restore","Cancel"]),
	!,
	Reply<>2,
	Action=Reply+1,
    datetime::timNow(Dat,Tim),
	closeKB_Step2(Action,Dat,Tim),
	closeTrace(CloseTrace),
	!	
;
	Ask=1,
	globalVars::get_i("knowopen")=2,
	% changes and no errors
	dlgIF::dlgExp(2),
	Reply=vpiCommonDialogs::ask("Save changes in KB?","The KB has been changed. Since you are using the backup option you can abandon these changes and restore the opening state.",["Save","Restore","Cancel"]),
	!,
	Reply<>2,
	Action=Reply+1,
	datetime::timNow(Dat,Tim),
	closeKB_Step2(Action,Dat,Tim),
	closeTrace(CloseTrace),
	!
;	
	Ask=-1,
	globalVars::get_i("knowopen")=2,
	globalVars::get_i("knowerr")=1,
	% changes and errors
    datetime::timNow(Dat,Tim),
	closeKB_Step2(1,Dat,Tim),
	closeTrace(CloseTrace),
	!	
;
	Ask=-1,
	globalVars::get_i("knowopen")=2,
	% changes and no errors
	datetime::timNow(Dat,Tim),
	closeKB_Step2(1,Dat,Tim),
	closeTrace(CloseTrace),
	!.

closeTrace(1) :-
	tt::tt_close,
	!.
closeTrace(_) :-
	!.

% parameters:
% 1	in	0=close/no changes; 1=save/changes; 2=restore
%
closeKB_Step2(Action,Dat,Tim) :-
	Action=0,
	dbBase::kbIsSql,
    sqlBase::closeKb,
	File=globalVars::get_s("kb_file"),
			sx::concat(["Closed KB ",File,"."],S3),
			dlgIF::msgS(S3),
	globalVars::set_i("knowopen",0),
	sdb::sdbClose,
	dlgIF::dbaStatuses,
	!
;
	Action=1,
	dbBase::kbIsSql,
    dep::saveGraphs,
    sqlBase::closeKb,
	File=globalVars::get_s("kb_file"),
			sx::concat(["Saved and closed KB ",File,"."],S3),
			dlgIF::msgS(S3),
	globalVars::set_i("knowopen",0),
	sdb::sdbClose,
	dlgIF::dbaStatuses,
	!
;
	Action=0,
	not(dbBase::kbIsSql),
	dbaBase::database(DB,_,_,0),
	dbaUtil::dbaCloseBasic(DB,'C'), 
	File=globalVars::get_s("kb_file"),
			sx::concat(["Closed KB ",File,"."],S3),
			dlgIF::msgS(S3),
	dbaExt::dbaCloseExternals(Action,Dat,Tim),
	dbaUtil::dbaFlagSet(File,"3op",0),
	globalVars::set_i("knowopen",0),
	sdb::sdbClose,
	dlgIF::dbaStatuses,
	!
;
	Action=1,
	not(dbBase::kbIsSql),
    dep::saveGraphs,
	dbaUpdatePublic(Dat,Tim),
	dbaBase::database(DB,_,_,0),
	dbaUtil::dbaCloseBasic(DB,'C'),
	File=globalVars::get_s("kb_file"),
			sx::concat(["Saved and closed KB ",File,"."],S3),
			dlgIF::msgS(S3),
	dbaExt::dbaCloseExternals(Action,Dat,Tim),
	dbaUtil::dbaFlagSet(File,"3op",0),
	dbaUtil::dbaFlagSet(File,"3eq",0),
	globalVars::set_i("knowopen",0),
	sdb::sdbClose,
	dlgIF::dbaStatuses,
	!
;
	Action=2,
	FileName=globalVars::get_s("kb_file"),
	fioExt::filChangeType(FileName,"3kc",OldFile),
	file::existFile(OldFile),
	vpi::cursorSetWait(),
	dbaBase::database(DB,_,_,0),
	dbaUtil::dbaCloseBasic(DB,'C'),
	dbaExt::dbaCloseExternals(Action,Dat,Tim),
	File=globalVars::get_s("kb_file"),
	dbaUtil::dbaFlagSet(File,"3op",0),
	file::copy(OldFile,FileName),
	dbaUtil::dbaFlagSet(FileName,"3eq",1),
			S2=globalVars::get_s("kb_file"),
			sx::concat(["Closed and restored KB ",S2,"."],S3),
			dlgIF::msgS(S3),
	globalVars::set_i("knowopen",0),
	sdb::sdbClose,
	dlgIF::dbaStatuses,
	!
;
	Action=2,
	vpi::cursorSetWait(),
	dbaBase::database(DB,_,_,0),
	dbaUtil::dbaCloseBasic(DB,'C'),
	dbaExt::dbaCloseExternals(Action,Dat,Tim),
	File=globalVars::get_s("kb_file"),
	dbaUtil::dbaFlagSet(File,"3op",0),
	FileName=globalVars::get_s("kb_file"),
	dbaUtil::dbaFlagSet(FileName,"3eq",1),
			S2=globalVars::get_s("kb_file"),
			sx::concat(["Closed and restored KB ",S2,"."],S3),
			dlgIF::msgS(S3),
	globalVars::set_i("knowopen",0),
	sdb::sdbClose,
	dlgIF::dbaStatuses,
	!.
	
dbaUpdatePublic(Dat,Tim) :-
	% set update time for changed public changes	
	dbaBase::pubChain(Chain,_),
	dbaBase::changedChain(Chain),
    dbBase::insertTimingEvent(Chain,Dat,Tim),
	fail
;
	% set index regen. time
	globalVars::get_i("ixrg")=-1,
    dbBase::insertTimingEvent("p3_iVal",Dat,Tim),
	!
;
	globalVars::get_i("ixrg")=-1,
	dlgIF::dlgNote(0,"Index tables need regeneration"),
	!	
;	
	!.		
% -------------------------------------------------------------------------------------------------------------
% getSettings                                                        
% -------------------------------------------------------------------------------------------------------------
getSettings(TOK,KBN,RLN) :-
    dbBase::kbIsSql,
    sqlBase::chainSentencesDirect("x",LT1),
	LT1=cmp(_,[cmp("p3_eConf",[atom("p3_iTokn"),T])]),
	T=int(TOK),
    sqlBase::chainSentencesDirect("x",LT2),
	LT2=cmp(_,[cmp("p3_eConf",[atom("p3_iMxKN"),K])]),
	K=int(KBN),
    sqlBase::chainSentencesDirect("x",LT3),
	LT3=cmp(_,[cmp("p3_eConf",[atom("p3_iMxRN"),R])]),
	R=int(RLN),
	!
;
    not(dbBase::kbIsSql),
    dbaBase::database(DB,_,_,0),
	DB:chain_terms("x",clause(_,_,_,_,_,C,_),_),
	C=cmp("x",[cmp("p3_eConf",[atom("p3_iTokn"),T])]),
	T=int(TOK),
	DB:chain_terms("x",clause(_,_,_,_,_,C2,_),_),
	C2=cmp("x",[cmp("p3_eConf",[atom("p3_iMxKN"),K])]),
	K=int(KBN),
	DB:chain_terms("x",clause(_,_,_,_,_,C3,_),_),
	C3=cmp("x",[cmp("p3_eConf",[atom("p3_iMxRN"),R])]),
	R=int(RLN),
	!
;
	dlgIF::dlgNote(0,"Database configuration settings not found - defaults assumed"),
	TOK=globalVars::get_i("max_tok"),
	RLN=globalVars::get_i("max_rln"),
	KBN=globalVars::get_i("max_kbn"),
	dbOpen::storeSettings,
	!.
% -------------------------------------------------------------------------------------------------------------
% openKB
% -------------------------------------------------------------------------------------------------------------
class predicates
openKB_X : (string) determ.
clauses

% parameters:
% 1	in	KB file (or blank)
%
openKB(File) :- 
    File="",
    Lc=fio::getRecent("K"),
    set::size(Lc,N),
    N>0,
    Ok=vpiCommonDialogs::listSelect("Recently used KBs - (Cancel to select another/enter new KB!)",Lc,0,File2,_),
    Ok=b_true,
    openKB_X(File2),
    !
;            
    File="",
	KB=vpiCommonDialogs::getFileName("*.3kb",["KB files","*.3kb"],"Select existing KB (or enter name of new - SQL-based KBs are named 'dbase.kbase.sql')",[],"",_),
	fioExt::filValidKB_Name(1,KB,_),
	dbOpen::checkKB_Files(KB),
	openKB_Step2(KB),
	!
;
    File="",
	Reply=vpiCommonDialogs::ask("You cannot run Pro/3 without a KB","Do you want to launch the help system before Pro/3 exits?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	dlgIF::hlpWeb("pro3exe"),
	W=guix::getWinOfClas(1),
	_=vpi::winSendEvent(W,vpiDomains::e_CloseRequest),
	!	
;
    File="",
	W=guix::getWinOfClas(1),
	_=vpi::winSendEvent(W,vpiDomains::e_CloseRequest),
	!
;
    % SQL
    sqlBase::names(File,_,_,_),
	!,openKB_Step2(File),
    sqlSt::createDbIfnotexists,
    !
;    
    % 3KB
	fioExt::filCheckFile(1,"Knowledge Base",File,KB),
	openKB_Step2(KB),
	!.

openKB_X(File) :-    
    dbOpen::existKB(File),
    openKB(File),
    !
;
    sqlBase::names(File,_,_,_),
    % non-existing SQL KB - initialize it
	openKB_Step3(0,File),
	dlgIF::dlgExp(22),
	W=vpi::getTaskWindow(),
     _=vpi::winSendEvent(W,vpiDomains::e_Menu(id_Preferences_directories_and_files,0)),
    !.
% --------------------------------------------------------------------------
% openKB_Step2
% --------------------------------------------------------------------------
openKB_Step2(FileName) :-
    %
    % SQL
    %
    sqlBase::names(FileName,Db,Kb,_),
    sqlBase::openConnection,
    sqlBase::openDb(Db),
    sqlBase::existKb(Kb)>0,
	openKB_Step3(1,FileName),
	dbExtension::dbuCheckSTD(1),
    !
;    
    sqlBase::names(FileName,_,_,_),
	openKB_Step3(0,FileName),
	dlgIF::dlgExp(22),
	W=guix::getWinOfClas(1),
     _=vpi::winSendEvent(W,vpiDomains::e_Menu(id_Preferences_directories_and_files,0)),
    !
;   %
    % 3KB
    %    
	P=globalVars::get_s("exe_path"),
	LF=string::concat(P,"\\exe\\Pro3.log"),
	dbOpen::dbLog(FileName,LF),
	globalVars::get_i("backup")=1,
	dbaUtil::dbaFlagON(FileName,"3op"),
	dlgIF::dlgExp(6),
	sx::concat(["Restore with DB backup?"],Mess),
	Reply=vpiCommonDialogs::ask(Mess,"Main DB was not properly closed last time it was used and cannot be opened. Restore with DB backup?",["Yes","No","Cancel"]),
	dbaUtil::dbaOpenRestore(Reply,FileName), % normally fails 
	!
;	
	dbaUtil::dbaFlagON(FileName,"3op"),
	dlgIF::dlgExp(7),
	dlgIF::feil("openKB_Step2",["DB was not properly closed and cannot be opened!"],[],nill),
	!
;	
	file::existFile(FileName),
	openKB_Step3(1,FileName),
	dbExtension::dbuCheckSTD(1),
	!
;	
    % initialization of a new DB
	not(file::existFile(FileName)),
	openKB_Step3(0,FileName),
	dlgIF::dlgExp(22),
	W=guix::getWinOfClas(1),
     _=vpi::winSendEvent(W,vpiDomains::e_Menu(id_Preferences_directories_and_files,0)),
	!.
% --------------------------------------------------------------------------
% openKB_Step3
% --------------------------------------------------------------------------
% opens the main DB (maindb)
% parameters:
% 1	in	0=new KB; 1=existing KB
% 2	in	KB file name
%
openKB_Step3(Existing,FileName) :-
    %
    % SQL
    %
    Existing=0,
	globalVars::get_i("error_flag")=0,
    dbaBase::dropTables,
    sqlBase::names(FileName,DB,KB,_),
	globalVars::set_i("knowopen",0),
	W=guix::getWinOfClas(1),
    Dlg=kbinit::new(),
    Dlg:show(W,1),
    Ps=globalVars::popRetPars(),
    Ps=["1"],
	dlgIF::dbaStatuses,
	KbName=string::concat(DB,".",KB,".sql"),
	create(KbName),
	globalVars::set_s("kb_id","X"),
	setFileNames(FileName),
	fio::addRecent("K",KbName),
    tt::tt_open,
	sqlBase::openDb(DB),
    sqlBase::createKb(KB),
	sdb::sdbClose,
	init::loadKB_Prefs,
	init::adjustDirs,
    sdb::sdbOpen,
	dep::loadGraphs,
	proc::loadStandardConf,
    setSegmentContext,
	setCurrentTerminology,
	cache::loadSm,
	dlgIF::setFlag("knowerr",0),
	dlgIF::dbaStatuses,
	dlgIF::msgS("DB opened."),
	!
;
    Existing=1,
    sqlBase::names(FileName,Db,Name,_),
	globalVars::get_i("error_flag")=0,
	vpi::cursorSetWait(),
	sqlBase::openDb(Db),
	sqlBase::openKb(Name),
	globalVars::set_s("kb_id","X"),
	setFileNames(FileName),
	fio::addRecent("K",FileName),
    tt::tt_open,
    dbBase::setSql,
	dbaBase::dropTables(),
	dbOpen::versionCheck, 
	init::loadKB_Prefs,
	init::adjustDirs,
	sdb::sdbOpen,
	dbOpen::newVersion, 
    setSegmentContext,
	setCurrentTerminology,
	cache::loadSm,
	dlgIF::setFlag("knowerr",0),
	dlgIF::dbaStatuses,
    !
;
    %
    % 3KB
    %
    Existing=0,
	globalVars::get_i("error_flag")=0,
    dbaBase::dropTables,
	globalVars::set_i("knowopen",0),
	W=guix::getWinOfClas(1),
    Dlg=kbinit::new(),
    Dlg:show(W,0),
    Ps=globalVars::popRetPars(),
    Ps=["1"],
	dlgIF::dbaStatuses,
	create(FileName),
	DB=chainDB::db_open(FileName,chainDB::in_file),
	setFileNames(FileName),
	fio::addRecent("K",FileName),
    tt::tt_open,
	Id=globalVars::get_s("kb_id"),
	dbaBase::dropTables(),
	dbaBase::addDatabase(DB,FileName,Id),
	dbaBase::dbaIndexesOps(DB,'C',1), % open
	sdb::sdbClose,
	init::loadKB_Prefs,
	init::adjustDirs,
    sdb::sdbOpen,
	dep::loadGraphs,
	proc::loadStandardConf,
    dbaExt::externalDBs,
    setSegmentContext,
	dlgIF::setFlag("knowerr",0),
	dlgIF::dbaStatuses,
	dlgIF::msgS("DB opened."),
	!
;    
    Existing=1,
	globalVars::get_i("error_flag")=0,
	vpi::cursorSetWait(),
    dbaBase::dropTables,
	dbaExt::readLocations(FileName),
	dbaUtil::dbaBackup(Existing,FileName),
	DB=chainDB::db_open(FileName,chainDB::in_file),
	setFileNames(FileName),
	fio::addRecent("K",FileName),
    tt::tt_open,
	Id=globalVars::get_s("kb_id"),
	dbaBase::dropTables(),
	dbaBase::addDatabase(DB,FileName,Id),
	dbaBase::dbaIndexesOps(DB,'C',1), % open
	dbOpen::versionCheck,
	dbaUtil::dbaCheckDB_Id(KBid),
	globalVars::set_s("kb_id",KBid),
	dbOpen::getSettings(TOK,KBN,RLN),
	globalVars::set_i("max_tok",TOK),
	globalVars::set_i("max_kbn",KBN),
	globalVars::set_i("max_rln",RLN),
	init::dbaIxSetLengths,
	dbaUtil::dbaFlagSet(FileName,"3op",1),
	dbBase::getTimingEvent("p3_iVal",Dat,Tim),
	dbaExt::dbaPublicChains(Dat,Tim),
	dbaExt::openExternalDBs(Existing,Dat,Tim),
	dbaUtil::dbaOpenRegenIx,
	init::loadKB_Prefs,
	init::adjustDirs,
	sdb::sdbOpen,
	dbOpen::newVersion, 
    setSegmentContext,
	setCurrentTerminology,
	cache::loadSm,
	dlgIF::setFlag("knowerr",0),
	dlgIF::dbaStatuses,
	!.
	
% ----------------------------------------------------------------------------------------------
% setFileNames
% ----------------------------------------------------------------------------------------------
setFileNames(FileName) :-
    sqlBase::names(FileName,DB,KB,Path),
	!,KbName=string::concat(Path,"\\",DB,".",KB,".sql"),
	globalVars::set_s("kb_file",KbName),
    globalVars::set_s("std_path",Path),
	tt::setFile(KbName),
	!
;
    Path2=filename::getPath(FileName),
    Path2="",
	Path=globalVars::get_s("exe_path"),
	KbName=string::concat(Path,"\\",FileName),
	globalVars::set_s("kb_file",KbName),
    globalVars::set_s("std_path",Path),
	tt::setFile(KbName),
	!   
;    
    Path=filename::getPath(FileName),
    Path<>"",
    sx::sxRemoveTail(Path,"\\",Path2),
	globalVars::set_s("kb_file",FileName),
    globalVars::set_s("std_path",Path2),
	tt::setFile(FileName),
	!.   
% ----------------------------------------------------------------------------------------------
% setCurrentTerminology                                                            
% ----------------------------------------------------------------------------------------------
setCurrentTerminology () :-
	Terms=pop::getTerminologies(),
	set::member("p3_iEngl",Terms),
	globalVars::set_s("lang","p3_iEngl"),
	cache::loadTm("p3_iEngl"),
		dbBase::nlSingF("p3_iEngl",L),
		sx::concat(["Current terminology is ",L,"."],Mess),
		dlgIF::msgS(Mess),
	!
;
	Terms=pop::getTerminologies(),
	Terms=[T|_],
	globalVars::set_s("lang",T),
	cache::loadTm(T),
		dbBase::nlSingF(T,L),
		sx::concat(["Current terminology is ",L,"."],Mess),
		dlgIF::msgS(Mess),
	!
;
	Terms=pop::getTerminologies(),
	Terms=[T|_],
	globalVars::set_s("lang",T),
	        dlgIF::dlgExp(1002),
		sx::concat(["Cannot translate current terminology KB name '",T,"'. This is possibly caused by missing terminology, external DBs not yet opened or old index table."],Mess),
		dlgIF::dlgNote(1,Mess),
		sx::concat(["Current terminology is '",T,"'."],Mess2),
		dlgIF::msgS(Mess2),
	!
;
	dlgIF::dlgNote(0,"No terminologies found in the KB. Reload Pro/3 standard files if this does not resolve the problem."),
	!.	
% ----------------------------------------------------------------------------------------------
% setSegmentContext                                                            
% ----------------------------------------------------------------------------------------------
setSegmentContext() :-
    Context=pop::getSegments(),
	set::size(Context,Nc),
	Nc>1,
	set::setRemSub(["Pro/3"],Context,NewContext),
	dbBase::setContext(NewContext),
	!
;
    Context=pop::getSegments(),
	set::size(Context,Nc),
	Nc=1,
	Context=[S],
	S<>"Pro/3",
	!
;	
	% create a segment called after the KB name (use upper case only) by inserting an identifier for the
	% KB name (also upper case) - make this segment current
	F=globalVars::get_s("kb_file"),
    sqlBase::names(F,_,Name,_),
	!,
	UName=string::toUpperCase(Name),
	LName=string::toLowerCase(Name),
	Lng=globalVars::get_s("lang"),
	pscan::pscNL(UName,Toks1),
	pscan::str_tokens(Toks1,Toks2),
	termDomain::cl_list(Toks2,Toks),
	L=globalVars::get_i("max_kbn"),
	sx::sxTrunc(LName,L,LNameT),
	dbBase::recordInsert(2,"$TRM",0,0,cmp("v",[cmp("t",[atom(Lng),str(LNameT),atom("p3_dif"),atom("p3_iNIL"),Toks]),str(UName)]),atom("true"),_),
	dbBase::setContext([UName]),
    !
;	
	% create a segment called after the KB name (use upper case only) by inserting an identifier for the
	% KB name (also upper case) - make this segment current
	F=globalVars::get_s("kb_file"),
	fio::file_name(F,_,Name,_),
	UName=string::toUpperCase(Name),
	LName=string::toLowerCase(Name),
	not(dbaBase::extChain("v",_,0)),
	Lng=globalVars::get_s("lang"),
	pscan::pscNL(UName,Toks1),
	pscan::str_tokens(Toks1,Toks2),
	termDomain::cl_list(Toks2,Toks),
	L=globalVars::get_i("max_kbn"),
	sx::sxTrunc(LName,L,LNameT),
	dbBase::recordInsert(2,"$TRM",0,0,cmp("v",[cmp("t",[atom(Lng),str(LNameT),atom("p3_dif"),atom("p3_iNIL"),Toks]),str(UName)]),atom("true"),_),
	dbBase::setContext([UName]),
    !.
end implement dbOpen
