/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement dbaBase
open core,termDomain,globalDomains

class facts
% there is one kbs-fact per database. the No's are assigned sequentially, however 0 is always the main db while neagtive numbers are used for aux DBs.
% external DBs thus have numbers > 0. All file names are kept in lower case letters in the tables, while KB Ids are kept in upper case.
% all but aux DBs have a one-letter identifier.
%
kbCount : integer :=0.
kbCountAux : integer :=-1.
kbs : (chainDB KB,string KBfile,string KBid,integer No) nondeterm.  % database table 	
kbmap : (string ExternalChain,chainDB KB,integer Access) nondeterm. % chain mapping table (external chains only)
kbpub : (string PublicChain,integer Access) nondeterm. % public chain table (this is loaded once upon opening KB since changes in session are unimportant)
kbcc : (string Chain) nondeterm. % a table of external read-write chains and public chains which have been changed during session
btrsel : (char Ix,string BTreeName,chainDB::bt_selector Selector). % Ix='C' for maindb and 'Y' for copydb

constants
className = "com/rudstrand/dbaBase/dbaBase".
classVersion = "1.0".

class predicates
dbaIxAll_DelCommon : (integer,string,chainDB,string,integer,integer,chainDB::ref) determ.
dbaIxAll_InsCommon : (integer,string,chainDB,char,string,integer,integer,chainDB::ref) determ.

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------------------------------
% class fact predicates
% -------------------------------------------------------------------------------------------------------------
% the database table
%
class predicates
getDatabasesL : (lst).
getDatabasesX : (string_list,chainDB_list,string_list,integer_list) determ (i,o,o,o).
clauses

database(DB,File,Id,No) :- kbs(DB,File,Id,No).

addDatabase(DB,File,Id) :-
    No=kbCount,
    not(kbs(DB,_,_,_)),
    not(kbs(_,File,_,_)),
    not(kbs(_,_,Id,_)),
    not(kbs(_,_,_,No)),
    assert(kbs(DB,File,Id,No)),
    kbCount:=No+1,
    !
;
    No=kbCount,
    dlgIF::feil("addDatabase",["Database exists already",File,Id],[No],nill),!,fail.

addDatabaseAux(DB,File) :-
    No=kbCountAux,
    not(kbs(DB,_,_,_)),
    not(kbs(_,File,_,_)),
    not(kbs(_,_,_,No)),
    assert(kbs(DB,File,"",No)),
    kbCountAux:=No-1,
    !
;
    No=kbCount,
    dlgIF::feil("addDatabaseAux",["Database exists already",File],[No],nill),!,fail.

dropDatabase(DB) :-
    retractall(kbs(DB,_,_,_)),
    !.

getDatabases(KBs,Files,Ids,Nos) :-
    L=lst::new(),
    getDatabasesL(L),
    Ids=L:get(),
    getDatabasesX(Ids,KBs,Files,Nos),
    !.
    
getDatabasesL(L) :-
    kbs(_,_,Id,_),
    L:add(Id),
    fail
;
    !.
    
getDatabasesX([],[],[],[]) :-
    !.
getDatabasesX([Id|Ids],[KB|KBs],[File|Files],[No|Nos]) :-
    kbs(KB,File,Id,No),
    getDatabasesX(Ids,KBs,Files,Nos),
    !.
    
% the external chain table
%    
class predicates
getExtChainsX : (lst,lst) procedure.
getDB_ExtChainsX : (chainDB,lst,lst) procedure.
clauses

extChain(ExternalChain,DB,Access) :- kbmap(ExternalChain,DB,Access).

addExtChain(ExternalChain,DB,Access) :- assert(kbmap(ExternalChain,DB,Access)).

dropExtChain(Chain) :- retractall(kbmap(Chain,_,_)).

getExtChains(Read,Write) :-
    ListR=lst::new(),
    ListW=lst::new(),
    getExtChainsX(ListR,ListW),
    Read=ListR:get(),
    Write=ListW:get(),
    !.
 
getExtChainsX(Lr,Lw) :-
    kbmap(C,_,0),
    Lr:add(C),
    fail
;
    kbmap(C,_,1),
    Lw:add(C),
    fail
;
    !.

getDB_ExtChains(DB,Read,Write) :-
    ListR=lst::new(),
    ListW=lst::new(),
    getDB_ExtChainsX(DB,ListR,ListW),
    Read=ListR:get(),
    Write=ListW:get(),
    !.
 
getDB_ExtChainsX(DB,Lr,Lw) :-
    kbmap(C,DB,0),
    Lr:add(C),
    fail
;
    kbmap(C,DB,1),
    Lw:add(C),
    fail
;
    !.

% the public chain table
%
pubChain(PublicChain,Access) :- kbpub(PublicChain,Access).

addPubChain(PublicChain,Access) :- assert(kbpub(PublicChain,Access)),!.

% the changed chain table
%
changedChain(Chain) :- kbcc(Chain). 

% misc
%
dropTables() :-
	retractall(kbs(_,_,_,_)),	
	retractall(btrsel(_,_,_)),	
	retractall(kbmap(_,_,_)),	
	retractall(kbpub(_,_)),	
	retractall(kbcc(_)),
	dbBase::setContext([]),
	kbCount:=0,
    !.

sameDB(F,G) :-
    Fl=string::toLowerCase(F),
    Gl=string::toLowerCase(G),
    Fl=Gl,
    !.
% ----------------------------------------------------------------------------------------------
% multiDb
% ----------------------------------------------------------------------------------------------
multiDb() :- kbs(_,_,_,N), N>0,!.   
% ----------------------------------------------------------------------------------------------
% dbaChainType
% ----------------------------------------------------------------------------------------------
% returns chain characteristics
% parameters:
% 1	in 	chain
% 2	out	DB ID
% 3	out	0=internal  1=public  2=external
% 4	out	0=readonly  1=write  -1=n/a
%
dbaChainType(Chain,DbId,Type,Access) :-
	dbaMap(Chain,KB,Access),
	database(Kb,_,DbId,No),
	No>0,
	Type=2,
	!
;	
    kbpub(Chain,Access),
	Type=1,
	dbaMap(Chain,KB,_),
	database(Kb,_,DbId,_),
	!
;
	dbaMap(Chain,KB,_),
	database(Kb,_,DbId,_),
	Type=0,	
	Access=-1,
	!.
% ----------------------------------------------------------------------------------------------
% kbaAddress/R
% ----------------------------------------------------------------------------------------------
kbaAddress(DB,Ref,Addr) :-
	kbs(DB,_,Id,_),
	dbaBase::dba_address(Ref,SRef),
	Addr=string::concat(SRef,Id),
	!
;
	dlgIF::feil("kbaAddress",[],[],nill),!,fail.
	
kbaAddressR(DB,Ref,Addr) :-
	sx::sxTail(1,Addr,SRef,Id),
	kbs(DB,_,Id,_),
	dbaBase::dba_address(Ref,SRef),
	!	
;
	dlgIF::feil("kbaAddressR",[Addr],[],nill),!,fail.
% --------------------------------------------------------------------------
% dbaRealmNoGet/SubNoGet
% --------------------------------------------------------------------------
class predicates
dbaRealmNoGetX : (chainDB::bt_selector,string,integer,dbLoc_list,integer) procedure (i,i,i,o,o).
dbaRealmNoGetY : (integer,chainDB::bt_selector,string,integer,dbLoc_list) determ (i,i,i,i,o).
dbaRealmNoGetZ : (integer,chainDB::bt_selector,string,integer,dbLoc_list) procedure (i,i,i,i,o).
dbaRSNG : (integer,dbLoc_list,dbLoc_list) determ (i,i,o).
clauses

% parameters:
% 1	in	realm name
% 2	in	realm no (or -1 if all in realm)
% 3	out	list of KB addresses
%
dbaRealmNoGet(R,N,L) :-
	Ru=string::toUpperCase(R),
	Ix=getIx('C',"REALM"),
	Key=string::concat(Ru,"|"),
	kbs(DB,_,_,0),
	DB:key_search(Ix,Key,_), % will always fail
	fail
;
	Ru=string::toUpperCase(R),
	Ix=getIx('C',"REALM"),
	dbaRealmNoGetY(1,Ix,Ru,N,L),
	!.

dbaRealmNoGetY(1,Ix,R,N,L) :-
	dbaRealmNoGetX(Ix,R,N,L2,Continue),
	dbaRealmNoGetZ(Continue,Ix,R,N,L3),
	set::conc([L2,L3],L),
	!.
dbaRealmNoGetY(-1,_,_,_,[]) :-
	!.
	
dbaRealmNoGetZ(1,Ix,R,N,L) :-
	kbs(DB,_,_,0),
	DB:key_next(Ix,_),
	dbaRealmNoGetY(1,Ix,R,N,L),
	!.	
dbaRealmNoGetZ(_,_,_,_,[]) :-
	!.

dbaRealmNoGetX(Ix,Ru,N,[cloc(KB,Ref)],1) :-
	N=-1,
	kbs(DB,_,_,0),
	DB:key_current(Ix,Key,Ref),
	sx::sxUnpack(Key,'|',L),
	L=[Realm,Id],
	Realm=Ru,
	kbs(KB,_,Id,_),
	!
;
	N<>-1,
	kbs(DB,_,_,0),
	DB:key_current(Ix,Key,Ref),
	sx::sxUnpack(Key,'|',L),
	L=[Realm,Id],
	Realm=Ru,
	kbs(KB,_,Id,_),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,SentNo,_,_,_,_),
	N=SentNo,
	!.

dbaRealmNoGetX(Ix,Ru,_,[],1) :-
	kbs(DB,_,_,0),
	DB:key_current(Ix,Key,_),
	sx::sxUnpack(Key,'|',L),
	L=[Realm,_],
	Realm=Ru,
	!.
dbaRealmNoGetX(_,_,_,[],-1) :-
	!.	

dbaRealmSubNoGet(R,N,M,SRefs) :-
	dbaRealmNoGet(R,N,SRefs2),
	dbaRSNG(M,SRefs2,SRefs),
	!.

dbaRSNG(_,[],[]) :-
	!.
dbaRSNG(SubNo,[cloc(KB,Ref)|SRefs],SRefsR) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,SNo,_,_,_),
	SubNo=SNo,
	dbaRSNG(SubNo,SRefs,SRefs2),
	SRefsR=[cloc(KB,Ref)|SRefs2],
	!
;
	dbaRSNG(SubNo,SRefs,SRefsR),
	!.
% ----------------------------------------------------------------------------------------------
% dbaChains
% ----------------------------------------------------------------------------------------------
% corresponds to db_chains in a multi-KB environment
%
dbaChains(DB,Chain) :-
    kbs(DB,_,_,0),
	% returns all maindb chains
	DB:db_chains(Chain)
;
	% returns all external chains 
	kbmap(Chain,DB,_).
% --------------------------------------------------------------------------
% dba_address
% --------------------------------------------------------------------------
dba_address(Ref,Addr) :-
	bound(Ref), free(Addr),
	Addr2=toString(Ref),
	sx::sxL_Pad(Addr2,"0",12,Addr),
	!
;
	bound(Addr), free(Ref),
	Addr2=string::trimFront(Addr),
	string::frontchar(Addr2,F,_),
	F='~',
	pfeil::syserr("dba_address","Invalid DB address",[Addr]),!,fail
;
	bound(Addr), free(Ref),
	Addr2=string::trimFront(Addr),
	sx::sxL_Pad(Addr2,"0",12,Addr3),
	Ref=toTerm(Addr3),
	!.
% ----------------------------------------------------------------------------------------------
% dbaMap; dbaMapST; dbaMapTerm                                                      
% ----------------------------------------------------------------------------------------------
dbaMap(Chain,KB,Access) :-
    kbmap(Chain,KB,Access),
	!
;
    database(KB,_,_,0),
    Access=1,
	!.
	
dbaMap(Chain,KB,Access,No) :-
    kbmap(Chain,KB,Access),
    database(KB,_,_,No),
	!
;
    database(KB,_,_,0),
    No=0,
    Access=1,
	!.
	
dbaMapST(ST,KB,Access) :-
	sx::sxUnpack(ST,'|',L),
	L=[P|_],
	dbaMap(P,KB,Access),
	!		
;
	dlgIF::feil("dbaMapST",[ST],[],nill),!,fail.			
	
dbaMapTerm(Term,KB,Access) :-
	Term=cmp(Chain,_),
	dbaMap(Chain,KB,Access),
	!
;
	dlgIF::feil("dbaMapTerm",[],[],Term),!,fail.			
% ----------------------------------------------------------------------------------------------
% dbaUpdatedCh                                                                
% ----------------------------------------------------------------------------------------------
dbaUpdatedCh(Ch) :-
	kbcc(Ch),
	!
;	
	kbmap(Ch,_,_), 
	assert(kbcc(Ch)),
	!
;
	kbpub(Ch,_), % public
	assert(kbcc(Ch)),
	!
;	
	kbmap(Ch,_,_), 
	assert(kbcc(Ch)),
	!
;
	!.
% --------------------------------------------------------------------------
% getIx/dropIx
% --------------------------------------------------------------------------
getIx(Clas,Name)=Selector:-
    btrsel(Clas,Name,Selector),
    !.
% -------------------------------------------------------------------------------------------------------------
% dbaIndexesOps; dbaIxOps
% -------------------------------------------------------------------------------------------------------------
class predicates
dbaIxOps : (chainDB,char Clas,integer Action,string IxName) determ.
dbaIxOpsX : (chainDB,char,integer,string) determ.
clauses

% opens, creates, closes, deletes index tables
% parameters:
% 1 in DB
% 2	in	par#1=maindb:Y  copydb:C
% 3	in	1=open, 2=create, 0=close; 3=delete (closes first)
%
dbaIndexesOps(DB,C,Action) :-
	dbaIxOps(DB,C,Action,"SYN1"),
	dbaIxOps(DB,C,Action,"TOKEN"),
	dbaIxOps(DB,C,Action,"RULE"),
	dbaIxOps(DB,C,Action,"RTYP"),
	dbaIxOps(DB,C,Action,"REALM"),
	!.

dbaIxOps(DB,KBix,Ops,Ix) :-
			C=string::charToString(KBix),
			tt::tt_s(2,"ix class",C),
			tt::tt_i(2,"ix ops",Ops),
			tt::tt_s(2,"ix",Ix),
	dbaIxOpsX(DB,KBix,Ops,Ix),
	!.

dbaIxOpsX(DB,KBix,2,"TOKEN") :-
	retractall(btrsel(KBix,"TOKEN",_)),
	L=globalVars::get_i("max_kbn")+globalVars::get_i("max_tok")+1,
	globalVars::set_i("ix_token",L),
	DB:bt_create("TOKEN",Ix,L,16,1),
	assert(btrsel(KBix,"TOKEN",Ix)),
	!.
dbaIxOpsX(DB,KBix,2,"SYN1") :-
	retractall(btrsel(KBix,"SYN1",_)),
	L=globalVars::get_i("max_kbn")*2+1,
	globalVars::set_i("ix_syn1",L),
	DB:bt_create("SYN1",Ix,L,16,1),
	assert(btrsel(KBix,"SYN1",Ix)),
	!.
dbaIxOpsX(DB,KBix,2,"RULE") :-
	retractall(btrsel(KBix,"RULE",_)),
	L=globalVars::get_i("max_kbn")+32+2+3+3,
	globalVars::set_i("ix_rule",L),
	DB:bt_create("RULE",Ix,L,8,0),
	assert(btrsel(KBix,"RULE",Ix)),
	!.
dbaIxOpsX(DB,KBix,2,"RTYP") :-
	retractall(btrsel(KBix,"RTYP",_)),
	L=3+globalVars::get_i("max_kbn")+1,
	globalVars::set_i("ix_rtyp",L),
	DB:bt_create("RTYP",Ix,L,32,1),
	assert(btrsel(KBix,"RTYP",Ix)),
	!.
dbaIxOpsX(DB,KBix,2,"REALM") :-
	retractall(btrsel(KBix,"REALM",_)),
	L=globalVars::get_i("max_rln")+1+1,
	globalVars::set_i("ix_realm",L),
	DB:bt_create("REALM",Ix,L,32,1),
	assert(btrsel(KBix,"REALM",Ix)),
	!.
dbaIxOpsX(_,_,2,Ix) :-
	dlgIF::feil("dbaIxOps",[Ix],[2],nill),
	!,fail.
	
dbaIxOpsX(DB,KBix,1,Ix) :-
	DB:bt_open(Ix,Ix1),
	assert(btrsel(KBix,Ix,Ix1)),!
;	
	dlgIF::feil("dbaIxOps",[Ix],[1],nill),
	!,fail.
dbaIxOpsX(DB,KBix,0,Ix) :-
	btrsel(KBix,Ix,Ix1),
		tt::tt_s(3,"Close index",Ix),
	DB:bt_close(Ix1),
	retract(btrsel(KBix,Ix,Ix1)),
	!
;	
	dlgIF::feil("dbaIxOps",[Ix],[0],nill),
	!,fail.
dbaIxOpsX(DB,KBix,3,Ix) :-
	dbaIxOpsX(DB,KBix,0,Ix),
	DB:bt_delete(Ix),
	!	
;	
	dlgIF::feil("dbaIxOps",[Ix],[3],nill),
	!,fail.
% ----------------------------------------------------------------------------------------------
% dbaSuspendRecord                                                              
% ----------------------------------------------------------------------------------------------
% suspends a record (unless already suspended)
%
dbaSuspendRecord(KB,Ref) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,LTerm,_),
	dbBase::accessErrorTerm(LTerm),
	!
;
	KB:ref_term(Ref,Rec),
	Rec=clause(RT,_,_,_,_,_,_),
	dbBase::recordType(RT,_,_,Susp),
	Susp=1,
	!
;
    dbBase::flagsClear,
	KB:ref_term(Ref,Rec),
	Rec=clause(RT,_,SentNo,SubNo,Realm,LTerm,RTerm),
	dbaDeleteRecord(KB,Ref),
	NRT=RT*10,
	dbaInsertRecord(NRT,Realm,SentNo,SubNo,LTerm,RTerm,_),
	!.	
/*
% --------------------------------------------------------------------------
% kbaDeleteRecs
% --------------------------------------------------------------------------
kbaDeleteRecs : (string_list) determ (i).
kbaDeleteRecs([]) :-
	!.
kbaDeleteRecs([R|Rs]) :-
    dbBase::flagsClear,
	dbaBase::kbaAddressR(DB,Ref,R),
	dbaDeleteRecord(DB,Ref),
	kbaDeleteRecs(Rs),
	!.
*/	
% ----------------------------------------------------------------------------------------------
% dbaInsertRecord/Term                                                            
% ----------------------------------------------------------------------------------------------
class predicates
dbaInsRecX : (chainDB,integer,string,integer,integer,sterm,sterm,chainDB::ref) determ (i,i,i,i,i,i,i,o).
dbaInvalidSyntagm : (string,string,string,string,string,string,string,integer) determ.
otherRecInsChecks : (integer,string,integer,integer,sterm,sterm) determ.
clauses

% inserts a record into the db (if validated)
% fails if not validated.
% parameters:
% 1	in	record type
% 2	in	source file
% 3	in	sentence no
% 4	in	sentence sub no
% 5	in	left-side term
% 6	in	right-side term
% 7	out	db reference
%
dbaInsertRecord(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref) :-
    not(dbBase::flagsClear),
    !,fail
;
	dbBase::accessErrorTerm(LTerm),
	!,fail
;	
	LTerm=cmp(PID,_),	
	dbaBase::dbaMap(PID,KB,_),
	dbaInsRecX(KB,RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref),
	!.

dbaInsertTerm(RecType,S,C,SubNo,cmp(":-",[LTerm,RTerm]),Ref) :- dbaInsertRecord(RecType,S,C,SubNo,LTerm,RTerm,Ref),!.
dbaInsertTerm(RecType,S,C,SubNo,LTerm,Ref) :- dbaInsertRecord(RecType,S,C,SubNo,LTerm,atom("true"),Ref),!.  	

dbaInsRecX(KB,RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref) :-
    % check: DUPLICATE
	globalVars::get_i("dup_chk")=1,
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType>=1,
	BaseType<=2,
	sx::frontstr(4,Realm,File,_),
	not(set::member(File,["PRO3","$SYS","$IRR","$GRAPH"])),
	dbExtension::dbaDuplicate(LTerm),
            nl::cprTerm(LTerm,S),
            Mess=string::concat("Filtered duplicate sentence: ",S,"!"),
            dbaBase::dba_address(Ref,"0000000"), % dummy reference
            dlgIF::msgS(Mess),
	!
;
    % derived sentences can be inserted without checks
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=1,
	RealmU=string::toUpperCase(Realm),
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	Rec=clause(RecType,Ts,SentNo,SubNo,RealmU,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref),
    !
; 
    % check: BLANK REALM
	Realm="",
	dlgIF::feil("dbaInsRecX",["Blank realm name is not allowed"],[],nill),
	!,fail
;
    % check: INVALID RECORD-TYPE
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType>7, 
	dlgIF::feil("dbaInsRecX",["Invalid record type"],[RecType],nill),
	!,fail
; 
    % check: OTHER
    globalVars::get_i("oth_chk")=1,
	dbBase::recordType(RecType,BaseType,_,_),
    otherRecInsChecks(BaseType,Realm,SentNo,SubNo,LTerm,RTerm),
    !,fail
; 
	% insert non-$GRAPH and non-$SYS sentence record
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2, 
	Realm<>"$SYS",
	Realm<>"$GRAPH",
	!,
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	RealmU=string::toUppercase(Realm),
	Rec=clause(RecType,Ts,SentNo,SubNo,RealmU,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref),
	dbBase::chainHasChangedTerm(LTerm),
    dbaIxAll(KB,1,RecType,RealmU,SentNo,SubNo,LTerm,Ref),
	dep::smGraphUpdate(1,RecType,LTerm),
	dbExtension::updateDependencies(1,RecType,LTerm,RTerm,"",_),
	!
;
    % insert $GRAPH sentences
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2,
	Realm="$GRAPH",
	!,
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	Rec=clause(RecType,Ts,SentNo,SubNo,Realm,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref), 
	% no need to mark db as updated as this is done separately for graph changes
    !
;
    % insert $SYS sentences
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2,
	Realm="$SYS",
	!,
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	Rec=clause(RecType,Ts,SentNo,SubNo,Realm,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref),
    !
;
	% insert rule record
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=3,
	LTerm=cmp(PID,_),	
	SSent=toString(SentNo),
	SSub=toString(SubNo),
	sx::sxlJoin([PID,Realm,SSent,SSub],"|",Rule),
	dbExtension::updateDependencies(1,RecType,LTerm,RTerm,Rule,Acyclic),
	Acyclic=0,
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,Rule,_),
	Mess=string::concat("Assert of sentence rule: ",Rule," failed due to cycle!"),
	dlgIF::msgS(Mess),
	!,fail
;
	% insert rule record
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=3,
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	RealmU=string::toUppercase(Realm),
	Rec=clause(RecType,Ts,SentNo,SubNo,RealmU,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref),
	dbBase::chainHasChangedTerm(LTerm),
    dbaIxAll(KB,1,RecType,RealmU,SentNo,SubNo,LTerm,Ref),
	!
;
	% insert derived rule record
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=5,
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	RealmU=string::toUppercase(Realm),
	Rec=clause(RecType,Ts,SentNo,SubNo,RealmU,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref),
	dbBase::chainHasChangedTerm(LTerm),
    dbaIxAll(KB,1,RecType,RealmU,SentNo,SubNo,LTerm,Ref),
	!
;
	% insert function def
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	dbExtension::updateDependencies(1,RecType,LTerm,RTerm,"",Acyclic),
	Acyclic=0,
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,"",_),
	LTerm=cmp(F,_),
	Mess=string::concat("Assert of function definition: ",F," failed due to cycle!"),
	dlgIF::msgS(Mess),
	!,fail
;
	% insert function def
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	RealmU=string::toUppercase(Realm),
	Rec=clause(RecType,Ts,SentNo,SubNo,RealmU,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref),
	dbBase::chainHasChangedTerm(LTerm),
    dbaIxAll(KB,1,RecType,RealmU,SentNo,SubNo,LTerm,Ref),
	!
;
	% insert inexact rule
	LTerm=cmp(_,[atom(Rule)|_]),	
	dbExtension::updateDependencies(1,RecType,LTerm,nill,Rule,Acyclic),
	Acyclic=0,
	dbExtension::updateDependencies(3,RecType,LTerm,nill,Rule,_),
	Mess=string::concat("Assert of inexact rule: ",Rule," failed due to cycle!"),
	dlgIF::msgS(Mess),
	!,fail
;
	% insert inexact rule
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=6,
	LTerm=cmp(PID,_),	
	datetime::tsNow(Ts),
	RealmU=string::toUppercase(Realm),
	Rec=clause(RecType,Ts,SentNo,SubNo,RealmU,LTerm,RTerm),
	KB:chain_insertz(PID,Rec,Ref),
	dbBase::chainHasChangedTerm(LTerm),
    dbaIxAll(KB,1,RecType,RealmU,SentNo,SubNo,LTerm,Ref),
	!
;
    SRec=toString(RecType),
    S1=toString(SentNo),
    S2=toString(SubNo),
    Mess=string::concatList(["Record type: ",SRec," realm: ",Realm,"|",S1,"|",S2," NOT inserted!"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.	        

% performs other record-insert checks
%
otherRecInsChecks(BaseType,Realm,SentNo,SubNo,LTerm,RTerm) :-
	BaseType=2, 
	Realm="PRO3-B-BASE.3PR",
	LTerm=cmp("x",[cmp("p3_eVers",[int(MV),str("PRO/3 SYSTEM SENTENCES VERSION")])]),
	dbBase::getProgramBuild(PV,_,_,_,_),
	PV<>MV,
	PVF=PV/100.0,
	MVF=MV/100.0,
	sx::fixFormatR(PVF,2,0,PVs),
	sx::fixFormatR(MVF,2,0,MVs),
	sx::concat(["You are storing PRO3-B-BASE.3PR v.",MVs," under PRO3.EXE v.",PVs,". This can cause errors. Compatibility issues are documented under help page 'PRO3.EXE'. You can get updated files from www.ligaya.net/p3!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!
;
	BaseType=2, 
	LTerm=cmp("v",[cmp(T,[atom(Lng),str(Rep),atom(Clas),atom(Form),_]),_]),
	dbaInvalidSyntagm(Realm,T,Lng,Rep,Clas,Form,Realm,SentNo),
	!
;
	BaseType=2, 
	not(RTerm=atom("true")),
	dlgIF::feil("otherRecInsChecks",["Attempt to insert a type-2 record with invalid right-side"],[],RTerm),
	!
;
	BaseType=2, 
	LTerm=cmp("p3_pIpch",[cmp("p3_eChan",[atom(Chain),_])]),
	set::member(Chain,["x","z","p3_pDP","p3_pDR","p3_pSM","p3_pIpch","p3_pIxdb"]),
	Mess=string::concat("Attempt to define system chain as public",Chain,"!"),
	dlgIF::dlgNote(1,Mess),
	!
;
    SentNo<0,
	dlgIF::dlgNote(1,"Attempt to insert a record with negative record number!"),
	!
;
    SubNo<0,
	dlgIF::dlgNote(1,"Attempt to insert a record with negative record sub-number!"),
	!.

% succeeds if the given syntagm/synonym is INVALID
% fails if VALID
%
dbaInvalidSyntagm(Realm,T,Lng,Rep,Clas,Form,Realm,No) :-
    _=string::search("+-*/^=><,;:()[]{}",Rep),
    !,fail
; 
    set::member(Rep,["...","<=",">=","<>"]),
    !,fail
;
	not(string::isName(Rep)),
	sx::concat(["'",Rep,"' is not a valid KB name"],Mess),
	dlgIF::feil("dbaInvalidSyntagm",[Mess],[],nill),
	!	
;
	sx::frontstr(1,Rep,F,_),
    Fu=string::toUppercase(F),
	Fu=F,
	sx::concat(["'",Rep,"' is not a valid KB name, which must start with a lower case letter"],Mess),
	dlgIF::feil("dbaInvalidSyntagm",[Mess],[],nill),
	!	
;
	sx::frontchar(Rep,F,_),
	not(sx::isLetter(F)),
	sx::concat(["'",Rep,"' is not a valid KB name, which must start with a lower case letter"],Mess),
	dlgIF::feil("dbaInvalidSyntagm",[Mess],[],nill),
	!	
;
	sx::frontstr(4,Realm,F1,_),
	F1<>"PRO3",
	dbBase::dbaReservedKB_Id(Rep),
	!
;
	T="s",
	kbBase::kbNL_RefF2(Rep,Clas,Form,_,_),
	!,fail % ok - the main form exists
;
	T="s",	
	SNo=toString(No),
	Mess=string::concatList(["Invalid synonym for (",Rep,") in sentence ",SNo," in realm ",Realm,"!"]),
	dlgIF::dlgNote(1,Mess),
	!
;	
	sx::frontstr(4,Realm,F1,_),
	F1<>"PRO3",
	not(dbBase::isValidId(Rep)),
	SNo=toString(No),
	Mess=string::concatList(["Invalid KB-name (",Rep,") in sentence ",SNo," in realm ",Realm,"!"]),
	dlgIF::dlgNote(1,Mess),
	!
;
	sx::sx_rev(Rep,S1),
	sx::searchchar(S1,'_',P),
	N=P-1,
	sx::frontstr(N,S1,F,_),
	string::hasDecimalDigits(F),
	SNo=toString(No),
	sx::concat(["KB name '",Rep,"' in sentence ",SNo," in realm ",Realm," ends with '_digit(s)'"],Mess),
	dlgIF::feil("dbaInvalidSyntagm",[Mess],[],nill),!
;
	sx::concat([Lng,"|",Rep],Key),
	Ix=dbaBase::getIx('C',"SYN1"),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Key,Ref),
	dbaBase::dbaMap("v",KB,_),
	KB:ref_term(Ref,Rec),
	Rec=clause(2,_,_,_,_,Term,atom("true")),
	Term=cmp("v",[cmp("t",[atom(Lng),str(Rep),atom(Clas),atom(Form),_]),_]),
	SNo=toString(No),
	sx::concat(["Token '",Lng,";",Rep,";",Clas,";",Form,"' in sentence ",SNo," in realm ",Realm," exists already"],Mess),
	dlgIF::feil("dbaInvalidSyntagm",[Mess],[],nill),
	!
;
	sx::concat([Lng,"|",Rep],Indx),	
	Ix=dbaBase::getIx('C',"SYN1"),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Indx,Ref1),
	dbaBase::dbaMap("v",KB,_),
	KB:ref_term(Ref1,Rec),
	Rec=clause(_,_,_,_,_,Term,_),
	Term=cmp("v",[cmp("t",[atom(Lng),str(Rep),atom(Clas2),_,_]),_]),
	Clas<>Clas2,
	SNo=toString(No),
	sx::concat(["Token '",Lng,";",Rep,";",Clas,"' in sentence ",SNo," in realm ",Realm," exists already as syntactic class '",Clas2,"'"],Mess),
	dlgIF::feil("dbaInvalidSyntagm",[Mess],[],nill),
	!
;
	Ix=dbaBase::getIx('C',"SYN1"),
	dbaBase::database(DB,_,_,0),
	DB:key_next(Ix,Ref1),
	dbaBase::dbaMap("v",KB,_),
	KB:ref_term(Ref1,Rec),
	Rec=clause(_,_,_,_,_,Term,_),
	Term=cmp("v",[cmp("t",[atom(Lng),str(Rep),atom(Clas2),_,_]),_]),
	Clas<>Clas2,
	SNo=toString(No),
	sx::concat(["Token '",Lng,";",Rep,";",Clas,"' in sentence ",SNo," in realm ",Realm," exists already as syntactic class '",Clas2,"'"],Mess),
	dlgIF::feil("dbaInvalidSyntagm",[Mess],[],nill),
	!.
% ----------------------------------------------------------------------------------------------
% dbaDeleteRecord                                                              
% ----------------------------------------------------------------------------------------------
class predicates
dbaDelRecX : (chainDB,chainDB::ref,integer,integer,integer,string,sterm,sterm) determ.
clauses

% deletes a record from the database. 
%
dbaDeleteRecord(KB,Ref) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(RecType,_,SentNo,SubNo,SourceFile,LTerm,RTerm),
	dbaDelRecX(KB,Ref,RecType,SentNo,SubNo,SourceFile,LTerm,RTerm),
	!.

dbaDelRecX(_,_,RecType,_,_,_,LTerm,_) :-
    not(dbBase::flagsClear),
    !
;    
	dbBase::accessErrorTerm(LTerm),
	!
;
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType>7, 
	dlgIF::feil("dbaDelRecX",["Unknown record type"],[RecType],nill),
	!,fail.
dbaDelRecX(KB,Ref,RecType,SentNo,SubNo,SourceFile,LTerm,RTerm) :-
	% derived fact
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=1,
	LTerm=cmp(Chain,_),
	KB:term_delete(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	% suspended
	dbBase::recordType(RecType,_,_,Susp),
	Susp=1,
	LTerm=cmp(Chain,_),
	KB:term_delete(Chain,Ref),
	dbaIxAll(KB,0,RecType,SourceFile,SentNo,SubNo,LTerm,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2,
	% fact
	SourceFile<>"$SYS",
	SourceFile<>"$GRAPH",
	!,
	dep::smGraphCheckDelete(BaseType,LTerm,RTerm),
	LTerm=cmp(Chain,_),
	KB:term_delete(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbaIxAll(KB,0,RecType,SourceFile,SentNo,SubNo,LTerm,Ref),
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,"",_),
	dep::smGraphUpdate(3,RecType,LTerm),
	!
;
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2,
	SourceFile="$GRAPH",
	!,
	LTerm=cmp(Chain,_),
	KB:term_delete(Chain,Ref),
	!
;
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2,
	SourceFile="$SYS",
	!,
	LTerm=cmp(Chain,_),
	KB:term_delete(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	% function body
	LTerm=cmp(F,_),
	KB:term_delete(F,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbaIxAll(KB,0,RecType,SourceFile,SentNo,SubNo,LTerm,Ref),
	dbExtension::updateDependencies(3,RecType,LTerm,atom("true"),"",_),
	!
;
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=6,
	% certainty rule
	LTerm=cmp(Chain,[atom(Rule)|_]),
	!,
	KB:term_delete(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbaIxAll(KB,0,RecType,SourceFile,SentNo,SubNo,LTerm,Ref),
	dbExtension::updateDependencies(3,RecType,LTerm,nill,Rule,_),
	dep::smGraphUpdate(3,RecType,LTerm),
	!
;
	% rule
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=3,
	LTerm=cmp(Chain,_),
	dbuRuleKey(Chain,Ref,Rule),
	KB:term_delete(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbaIxAll(KB,0,RecType,SourceFile,SentNo,SubNo,LTerm,Ref),
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,Rule,_),
	!
;
	% rule
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=5,
	LTerm=cmp(Chain,_),
	dbuRuleKey(Chain,Ref,Rule),
	KB:term_delete(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbaIxAll(KB,0,RecType,SourceFile,SentNo,SubNo,LTerm,Ref),
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,Rule,_),
	!
;	
	dbaBase::dba_address(Ref,SRef),
	dlgIF::feil("dbaDelRecX",[SRef,SourceFile],[RecType,SentNo,SubNo],cmp("",[LTerm,RTerm])),
	!,fail.
% ----------------------------------------------------------------------------------------------
% dbaIxAll
% ----------------------------------------------------------------------------------------------
% inserts/deletes all relevant indexes for given record
% parameters:
% 1	in	KB (of given Ref)
% 2	in	1=insert; 0=delete
% 3	in	rec.type
%
dbaIxAll(KB,1,Z,Realm,SentNo,SubNo,Term,Ref) :-	
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=1,
	% derived fact - no indexes 
	!
;	
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=2,
	Term=cmp("v",[cmp("t",[atom(Lng),str(Rep),_,_,list(str(T),_)]),_]),
	sx::concat([Lng,"|",T],Indx),
	dbaIxIns("TOKEN","ix_token",Indx,Ref),
	sx::concat([Lng,"|",Rep],Indx2),
	dbaIxIns("SYN1","ix_syn1",Indx2,Ref),
	Term=cmp(Chain,_),
	dbaIxAll_InsCommon(Z,Chain,KB,'C',Realm,SentNo,SubNo,Ref),!
;		
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=2,
	% synonym-type of record
	Term=cmp("v",[cmp("s",[atom(Lng),_,_,_,list(str(T),_)]),_]),
	sx::concat([Lng,"|",T],Indx),
	dbaIxIns("TOKEN","ix_token",Indx,Ref),
	Term=cmp(Chain,_),
	dbaIxAll_InsCommon(Z,Chain,KB,'C',Realm,SentNo,SubNo,Ref),!
;
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=2,
	Term=cmp(Chain,_),
	dbaIxAll_InsCommon(Z,Chain,KB,'C',Realm,SentNo,SubNo,Ref),!
;
	dbBase::recordType(Z,BaseType,_,_),
	set::member(BaseType,[3,4,5,7]),
	% rule	or function
	Term=cmp(Chain,_),
	dbaIxAll_InsCommon(Z,Chain,KB,'C',Realm,SentNo,SubNo,Ref),!
;
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=6,
	% certainty rule
	Term=cmp(Chain,_),
	dbaIxAll_InsCommon(Z,Chain,KB,'C',Realm,SentNo,SubNo,Ref),
	!.
dbaIxAll(KB,0,Z,SourceFile,SentNo,SubNo,Term,Ref) :-
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=1,
	% derived fact
	!
;
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=2,
	% syntagm-type of record
	Term=cmp("v",[cmp("t",[atom(Lng),str(Rep),_,_,list(str(T),_)]),_]),
	Ix0=dbaBase::getIx('C',"TOKEN"),
	sx::concat([Lng,"|",T],Indx),
	dbaBase::database(DB,_,_,0),
	DB:key_delete(Ix0,Indx,Ref),	
	Ix2=dbaBase::getIx('C',"SYN1"),
	sx::concat([Lng,"|",Rep],Indx2),
	DB:key_delete(Ix2,Indx2,Ref),	
	Term=cmp(Chain,_),
	dbaIxAll_DelCommon(Z,Chain,KB,SourceFile,SentNo,SubNo,Ref),!
;		
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=2,
	% synonym-type of record
	Term=cmp("v",[cmp("s",[atom(Lng),_,_,_,list(str(T),_)]),_]),
	Ix0=dbaBase::getIx('C',"TOKEN"),
	sx::concat([Lng,"|",T],Indx),
	dbaBase::database(DB,_,_,0),
	DB:key_delete(Ix0,Indx,Ref),	
	Term=cmp(Chain,_),
	dbaIxAll_DelCommon(Z,Chain,KB,SourceFile,SentNo,SubNo,Ref),!
;	
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=2,
	% fact
	Term=cmp(Chain,_),
	dbaIxAll_DelCommon(Z,Chain,KB,SourceFile,SentNo,SubNo,Ref),!
;
	dbBase::recordType(Z,BaseType,_,_),
	set::member(BaseType,[3,4,5]),
	% rule or function
	Term=cmp(Chain,_),
	dbaIxAll_DelCommon(Z,Chain,KB,SourceFile,SentNo,SubNo,Ref),!
;
	dbBase::recordType(Z,BaseType,_,_),
	BaseType=6,
	% certainty rule
	Term=cmp(Chain,_),
	dbaIxAll_DelCommon(Z,Chain,KB,SourceFile,SentNo,SubNo,Ref),
	!.
dbaIxAll(_,A,B,SourceFile,SentNo,SubNo,Term,_) :- 
	dlgIF::feil("dbaIxAll",[SourceFile],[A,B,SentNo,SubNo],Term),!,fail.

% inserts REALM/RTYP/RULE/DBRF indexes (as relevant for the rec.type) for given record
%
dbaIxAll_InsCommon(RT,Chain,KB,_,Realm,SentNo,SubNo,Ref) :-
	RT=20, % suspended sentences
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaIxIns("REALM","ix_realm",Key,Ref),
	SRT=toString(RT),
	sx::concat([SRT,"|",Chain],Key7),
	dbaIxIns("RTYP","ix_rtyp",Key7,Ref),
	!
;
	dbBase::recordType(RT,BaseType,_,_),
	BaseType=6,
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaIxIns("REALM","ix_realm",Key,Ref),
	SRT=toString(RT),
	sx::concat([SRT,"|",Chain],Key7),
	dbaIxIns("RTYP","ix_rtyp",Key7,Ref),
	!
;
	dbBase::recordType(RT,BaseType,_,_),
	set::member(BaseType,[3,5]),
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaIxIns("REALM","ix_realm",Key,Ref),
	SRT=toString(RT),
	sx::concat([SRT,"|",Chain],Key7),
	dbaIxIns("RTYP","ix_rtyp",Key7,Ref),
	SSub=toString(SubNo),
	SNo=toString(SentNo),
	sx::sxlJoin([Chain,Realm,SNo,SSub],"|",Key5),
	dbaIxIns("RULE","ix_rule",Key5,Ref),
	!
;
	dbBase::recordType(RT,BaseType,_,_),
	BaseType=4,
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaIxIns("REALM","ix_realm",Key,Ref),
	SRT=toString(RT),
	sx::concat([SRT,"|",Chain],Key7),
	dbaIxIns("RTYP","ix_rtyp",Key7,Ref),
	!
;	
	% non-suspended sentences
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaIxIns("REALM","ix_realm",Key,Ref),
	!.

% deletes REALM/RTYP/RULE/DBRF indexes (as relevant for the rec.type) for given record
%
dbaIxAll_DelCommon(RT,Chain,KB,Realm,SentNo,SubNo,Ref) :-	
	% system realms
	sx::frontchar(Realm,'$',_),
	IxR=dbaBase::getIx('C',"REALM"),
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaBase::database(DB,_,_,0),
	DB:key_delete(IxR,Key,Ref),		
	!
;
	dbBase::recordType(RT,BaseType,_,_),
	BaseType=6,
	IxR=dbaBase::getIx('C',"REALM"),
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaBase::database(DB,_,_,0),
	DB:key_delete(IxR,Key,Ref),		
	Ix7=dbaBase::getIx('C',"RTYP"),
	SRT=toString(RT),
	sx::concat([SRT,"|",Chain],Key7),
	DB:key_delete(Ix7,Key7,Ref),	
	!
;
	dbBase::recordType(RT,BaseType,_,_),
	set::member(BaseType,[3,5,7]),
	IxR=dbaBase::getIx('C',"REALM"),
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaBase::database(DB,_,_,0),
	DB:key_delete(IxR,Key,Ref),		
	Ix7=dbaBase::getIx('C',"RTYP"),
	SRT=toString(RT),
	sx::concat([SRT,"|",Chain],Key7),
	DB:key_delete(Ix7,Key7,Ref),	
	Ix6=dbaBase::getIx('C',"RULE"),
	SSub=toString(SubNo),
	SNo=toString(SentNo),
	sx::sxlJoin([Chain,Realm,SNo,SSub],"|",Key5),
	DB:key_delete(Ix6,Key5,Ref),	
	!
;
	dbBase::recordType(RT,BaseType,_,_),
	BaseType=4,
	IxR=dbaBase::getIx('C',"REALM"),
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaBase::database(DB,_,_,0),
	DB:key_delete(IxR,Key,Ref),		
	Ix7=dbaBase::getIx('C',"RTYP"),
	SRT=toString(RT),
	sx::concat([SRT,"|",Chain],Key7),
	DB:key_delete(Ix7,Key7,Ref),	
	!
;	
	% non-system realm sentences
	IxR=dbaBase::getIx('C',"REALM"),
    dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaBase::database(DB,_,_,0),
	DB:key_delete(IxR,Key,Ref),		
	!.
% -------------------------------------------------------------------------------------------------------------
% dbaIxIns
% -------------------------------------------------------------------------------------------------------------
dbaIxIns(IxName,GV,Key,Ref) :-
	N=string::length(Key),
	L=globalVars::get_i(GV),
	N<=L,
	Ix1=dbaBase::getIx('C',IxName),
	dbaBase::database(DB,_,_,0),
	DB:key_insert(Ix1,Key,Ref),
	!
;	
	N=string::length(Key),
	L=globalVars::get_i(GV),
	N>L,
	dbaBase::dba_address(Ref,SRef),
	SMax=toString(L),
	sx::concat(["Index '",IxName,"' key '",Key,"' longer than max",SMax," in record ",SRef,". Record inserted with truncated key!"],Mess),
	dlgIF::dlgNote(1,Mess),
	sx::frontstr(L,Key,TKey,_),
	dbaIxIns(IxName,GV,TKey,Ref),
	!.
% ----------------------------------------------------------------------------------------------
% dbuRuleKey
% ----------------------------------------------------------------------------------------------
dbuRuleKey(Chain,Ref,Key) :-
	dbaBase::dbaMap(Chain,KB,_),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,SentNo,SubNo,Realm,_,_),
	SSent=toString(SentNo),
	SSub=toString(SubNo),
	sx::sxlJoin([Chain,Realm,SSent,SSub],"|",Key),
	!.
% ----------------------------------------------------------------------------------------------
% dbaDeleteDerivedRules                                                             
% ----------------------------------------------------------------------------------------------
class predicates
dbaDelDR : (integer,string_list) determ.
clauses

% parameters:
% 1	in	0=error if access violation; 1=skip access violations
%
dbaDeleteDerivedRules(Mode) :-
	dbBase::recTypeRefList(5,R2),
	dbBase::recTypeRefList(-5,R3),
	dbBase::recTypeRefList(50,R4),
	dbBase::recTypeRefList(-50,R5),
	set::conc([R2,R3,R4,R5],RefChains),
	dbaDelDR(Mode,RefChains),
	!.
	
dbaDelDR(_,[]) :-
	!.
dbaDelDR(Mode,[R|Rs]) :-
	sx::sxUnpack(R,'|',L),
	L=[SRef,Chain],
	dbaBase::dbaMap(Chain,KB,Acc),
	Acc=1,
	dbaBase::dba_address(Ref,SRef),
	dbaBase::dbaDeleteRecord(KB,Ref),
	dbaDelDR(Mode,Rs),
	!
;
	Mode=1,
	sx::sxUnpack(R,'|',L),
	L=[_,Chain],
	dbaBase::dbaMap(Chain,_,Acc),
	Acc=0,
	dbaDelDR(Mode,Rs),
	!
;
	Mode=0,
	sx::sxUnpack(R,'|',L),
	L=[_,Chain],
	dbaBase::dbaMap(Chain,_,Acc),
	Acc=0,
	dlgIF::feil("dbaDelDR",["Attempt to write to external read-only chain",Chain],[],nill),!.
% ----------------------------------------------------------------------------------------------
% dbaDeleteDerivedFacts                                                             
% ----------------------------------------------------------------------------------------------
% parameters:
% 1 in -1/1 delete also facts in external DBs
%
dbaDeleteDerivedFacts(Externals) :-
	Externals=-1,
	dbaBase::database(DB,_,_,0),
	DB:db_chains(Chain),
	DB:chain_terms(Chain,clause(Z,_,_,_,_,_,_),Ref),
	set::member(Z,[1,8,10,-1,-8,-10,80,-80]),
	DB:term_delete(Chain,Ref),
	fail
;
	Externals=1,
	dbaBase::dbaChains(KB,Chain),
	dbaBase::dbaMap(Chain,_,Acc),
	Acc=1, % write access
	KB:chain_terms(Chain,clause(Z,_,_,_,_,_,_),Ref),
	set::member(Z,[1,8,10,-1,-8,-10,80,-80]),
	KB:term_delete(Chain,Ref),
	fail
;	
	!.
% --------------------------------------------------------------------------
% dbaDeleteSG
% --------------------------------------------------------------------------
class predicates
dbaDelSG_X : (chainDB,integer,string,count).
dbaDelSG_Y : (integer,string_list,integer) determ (i,i,o).
clauses

% deletes an SG
% parameters:
% 1	in	record type to be deleted (1/2), or 0 (both type 1 and 2)
% 2	in	SG
% 3	out	#records deleted
%
dbaDeleteSG(RecType,SG,N) :-
	dbBase::chainReadOnlyWarning(SG),
	N=0,
	!
;	
	STs=pop::getSTs(SG,[]),
	dep::massSentenceUpdateStart,
	dbaDelSG_Y(RecType,STs,N),
	dep::massSentenceUpdateEnd,
	!.

dbaDelSG_Y(_,[],0) :- 
    !.
dbaDelSG_Y(RecType,[ST|STs],N) :-
	dbaDeleteST(RecType,ST,N1),
	dbaDelSG_Y(RecType,STs,N2),
	N=N1+N2,
	!.
	
dbaDelSG_X(KB,RecType,SG,Count) :-
	RecType>0,
	KB:chain_terms(SG,Term,Ref),
	Term=clause(RT,_,_,_,_,_,atom("true")),
	dbBase::recordType(RT,RTb,_,_),
	RecType=RTb,
	dbaBase::dbaDeleteRecord(KB,Ref),
	Count:inc(_),
	fail
;
	RecType>0,
	!		
;
	RecType=0,	
	KB:chain_terms(SG,Term,Ref),
	Term=clause(_,_,_,_,_,_,atom("true")),
	dbaBase::dbaDeleteRecord(KB,Ref),
	Count:inc(_),
	fail
;
	!.		
% --------------------------------------------------------------------------
% dbaDeleteST
% --------------------------------------------------------------------------
class predicates
dbaDelST_X : (chainDB,integer,integer,string,string,string,count) determ.
clauses

% deletes ST 
% parameters:
% 1	in	record type to be deleted (1/2), or 0 (both type 1 and 2)
% 2	in	ST
% 3	out	#records deleted
%
dbaDeleteST(RecType,ST,N) :-
	dbBase::stComps(ST,_,Chain,_,_),
	dbBase::chainReadOnlyWarning(Chain),
	N=0,
	!
;
	dbBase::stComps(ST,Card,P,E1,E2),
	Count=count::newCounter(0),
	dbaBase::dbaMapST(ST,KB,_),
	dep::massSentenceUpdateStart,
	dbaDelST_X(KB,RecType,Card,P,E1,E2,Count),
	dep::massSentenceUpdateEnd,
	N=Count:get(),
	!.

dbaDelST_X(KB,RecType,1,P,E1,_,Count) :-
	RecType>0,
	KB:chain_terms(P,Term,Ref),
	Term=clause(RT,_,_,_,_,cmp(P,[cmp(E1,_)|_]),atom("true")),
	dbBase::recordType(RT,RTb,_,_),
	RecType=RTb,
	dbBase::recordDelete(cloc(KB,Ref)),
	Count:inc(_),
	fail
;
	RecType>0,
	!
;
	RecType=0,		
	KB:chain_terms(P,Term,Ref),
	Term=clause(_,_,_,_,_,cmp(P,[cmp(E1,_)|_]),atom("true")),
	dbBase::recordDelete(cloc(KB,Ref)),
	Count:inc(_),
	fail
;
	!.		
dbaDelST_X(KB,RecType,2,P,E1,E2,Count) :-
	RecType>0,
	KB:chain_terms(P,Term,Ref),
	Term=clause(RT,_,_,_,_,cmp(P,[cmp(E1,_),cmp(E2,_)|_]),atom("true")),
	dbBase::recordType(RT,RTb,_,_),
	RecType=RTb,
	dbBase::recordDelete(cloc(KB,Ref)),
	Count:inc(_),
	fail
;
	RecType>0,
	!
;
	RecType=0,	
	KB:chain_terms(P,Term,Ref),
	Term=clause(_,_,_,_,_,cmp(P,[cmp(E1,_),cmp(E2,_)|_]),atom("true")),
	dbBase::recordDelete(cloc(KB,Ref)),
	Count:inc(_),
	fail
;
	!.
% ----------------------------------------------------------------------------------------------
% dbaSuspendChain                                                              
% ----------------------------------------------------------------------------------------------
dbaSuspendChain(C) :-
	dbaBase::dbaMap(C,KB,_),
	KB:chain_terms(C,Cls,Ref),
	Cls=clause(_,_,_,_,_,_,_),
	dbaBase::dbaSuspendRecord(KB,Ref),
	fail
;
	!.
% ----------------------------------------------------------------------------------------------
% chainExists
% ----------------------------------------------------------------------------------------------
chainExists(Chain) :-
	dbaBase::dbaMap(Chain,KB,_),
	KB:chain_first(Chain,_),
	!.
% --------------------------------------------------------------------------
% dbaRealmExists
% --------------------------------------------------------------------------
class predicates
dbaRealmExistsX : (lst,string).
addRealm : (string,string,lst,string) determ.
clauses

% parameters:
% 1	in	realm name
% 2	out	file names for KBs where realm exists
%
dbaRealmExists(R,Files) :-
    List=lst::new(),
	Ru=string::toUpperCase(R),
    dbaRealmExistsX(List,Ru),
    Files=List:get(),
    !.

dbaRealmExistsX(List,R) :-
	Ix=dbaBase::getIx('C',"REALM"),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,R,_), % always fails
	fail
;	
	Ix=dbaBase::getIx('C',"REALM"),
	dbaBase::database(DB,_,_,0),
	DB:key_current(Ix,Key,_),
	sx::sxUnpack(Key,'|',L),
	L=[R,Id],
	dbaBase::database(_,File,Id,_),
	List:addUnique(File),
	fail
;	
	Ix=dbaBase::getIx('C',"REALM"),
	dbaBase::database(DB,_,_,0),
  	DB:key_next(Ix,_),
	DB:key_current(Ix,Key,_),
	sx::sxUnpack(Key,'|',L),
	L=[Rx,Id],
	dbaBase::database(_,File,Id,_),
	addRealm(R,Rx,List,File),
    !
;
    !.
    
addRealm(R,Rx,List,File) :-    
    R=Rx,
	List:addUnique(File),
	!,fail
;
    R<>Rx,
    !.
end implement dbaBase
