/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement dbaExt
open core,termDomain,globalDomains
class facts
exDb : (string FileName,string Id) nondeterm.
exCh : (string Chain,string Id,integer Access) nondeterm.
constants
className = "com/rudstrand/dbaExt/dbaExt".
classVersion = "1.0".
class predicates
dbaExtDB_Status : (chainDB DB,integer Status) procedure (i,o).
dbaVerifySettings : (chainDB,integer,integer,integer).
clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------------------------------
% openExternalDBs
% -------------------------------------------------------------------------------------------------------------
class predicates
checkExternalChain : (chainDB,string,integer,integer,integer) determ.
checkChainUpdateTime : (string,integer,integer).
dbaHasAcc : (string,chainDB,integer) determ (i,i,i).
checkExternalDB : (string,string,string) determ (i,i,o).
clauses
% opens external DBs and chains on the basis of the exDb and exCh tables 
%
openExternalDBs(0,_,_) :-
	!.
openExternalDBs(1,Dat,Tim) :-
	exDb(FileOrig,Id),
	checkExternalDB(Id,FileOrig,File),
	openExternalDB(File,Id,DB),
	exCh(Chain,Id,Access),
	checkExternalChain(DB,Chain,Access,Dat,Tim),
			N=Access+1,
			set::setGet(["read-only","read-write"],N,S2),
			sx::concat(["Noted external ",S2," chain ",Chain," in DB ",Id,"."],S),
			dlgIF::msgS(S),
	fail
;
	!.
		
openExternalDB(File,Id,DB) :-
	globalVars::get_i("backup")=1,
	dbaUtil::dbaFlagON(File,"3op"),
	dlgIF::dlgExp(6),
	sx::concat(["External DB ",File," was not properly closed"],Mess),
	Reply=vpiCommonDialogs::ask(Mess,"Restore with DB backup?",["Yes","No","Cancel"]),
	dbaUtil::dbaOpenRestore(Reply,File),
	% cancelled opening
	!,fail
;
	globalVars::get_i("backup")=1,
	dbaUtil::dbaBackup(1,File),
	DB=chainDB::db_open(File,chainDB::in_file),
	dbaBase::addDatabase(DB,File,Id),
			tt::tt_s(1,"File",File),
			sx::concat(["Opened external DB ID=",Id," in file ",File],Mess),
			dlgIF::msgS(Mess),
	dbOpen::getSettings(T,K,R),
	dbaVerifySettings(DB,T,K,R),
	dbaUtil::dbaFlagSet(File,"3op",1),
	!
;
	DB=chainDB::db_open(File,chainDB::in_file),
	dbaBase::addDatabase(DB,File,Id),
			tt::tt_s(1,"File",File),
			sx::concat(["Opened external DB ID=",Id," in file ",File],Mess),
			dlgIF::msgS(Mess),
	dbOpen::getSettings(T,K,R),
	dbaVerifySettings(DB,T,K,R),
	dbaUtil::dbaFlagSet(File,"3op",1),
	!.
	
% checks that external chains are valid and accessible and if indexes are up-to-date
%
checkExternalChain(DB,Chain,Access,Dat,Tim) :-
    dbaBase::extChain(Chain,_,_),
    Mess=string::concat("External chain '",Chain,"' located in more than one external database!"),
	dlgIF::dlgNote(1,Mess),
	!,fail
;
    dbaBase::dataBase(MainDB,_,_,0),
    MainDB:chain_first(Chain,_),
    Mess=string::concat("External chain '",Chain,"' also exists in the main database!"),
	dlgIF::dlgNote(1,Mess),
	!,fail
;
	not(dbaHasAcc(Chain,DB,Access)),
	N=Access+1,
	set::setGet(["read-only","read-write"],N,S2),
	sx::concat(["No ",S2," access to external chain '",Chain,"'. CHAIN IS IGNORED!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail
;	
	checkChainUpdateTime(Chain,Dat,Tim),
	dbaBase::addExtChain(Chain,DB,Access),
	!.
dbaHasAcc(Chain,KB,1) :-
	KB:chain_terms("p3_pIpch",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp("p3_pIpch",[cmp("p3_eChan",[atom(Chain),atom("p3_iRw")])]),
	!.
dbaHasAcc(Chain,KB,0) :-
	KB:chain_terms("p3_pIpch",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp("p3_pIpch",[cmp("p3_eChan",[atom(Chain),_])]),
	!.
% checks whether an external DB file exists - and if not requests for another file name
%
checkExternalDB(Id,File,_) :-
    Id2=globalVars::get_s("kb_id"),
    Id=Id2,
    Mess=string::concat("Database identifier character '",Id,"' is used for the main DB and cannot be used for external database!"),
    dlgIF::dlgNote(1,Mess),!,fail
;    
    File2=globalVars::get_s("kb_file"),
    dbaBase::sameDB(File,File2),
    Mess=string::concat("Database file '",Id,"' is already opened as the main DB and cannot be opened as external database!"),
    dlgIF::dlgNote(1,Mess),!,fail
;    
    dbaBase::database(_,File2,_,_),
    dbaBase::sameDB(File,File2),
    Mess=string::concat("Database file '",File,"' is already declared as an external database!"),
    dlgIF::dlgNote(1,Mess),!,fail
;    
    dbaBase::database(_,_,Id,_),
    Mess=string::concat("Database identifier character '",Id,"' is already used!"),
    dlgIF::dlgNote(1,Mess),!,fail.
checkExternalDB(KBid,FileC,FileR) :-
	not(fioExt::filValidKB_Name(-1,FileC,_)),
	dlgIF::dlgExp(8),
	sx::concat(["'",FileC,"' is not a valid KB file name"],Mess),
	Reply=vpiCommonDialogs::ask(Mess,"Do you want to specify another file",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	File=vpiCommonDialogs::getFileName("*.3kb",["KB files","*.3kb"],"External KB",[],"",_),
	retractall(exDb(_,KBid)),
	assert(exDb(File,KBid)),
	checkExternalDB(KBid,File,FileR),
	!.
checkExternalDB(_,File,File) :-
	file::existfile(File),
	!.
checkExternalDB(KBid,FileC,FileR) :-
	dlgIF::dlgExp(9),
	sx::concat(["External KB '",FileC,"' does not exist"],S2),
	Reply=vpiCommonDialogs::ask(S2,"Do you want to specify another file",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	File=vpiCommonDialogs::getFileName("*.3kb",["KB files","*.3kb"],"External KB",[],"",_),
	retractall(exDb(_,KBid)),
	assert(exDb(File,KBid)),
	checkExternalDB(KBid,File,FileR),
	!.
	
% sets index regen. flag if external chain has been updated after given date/time
%
checkChainUpdateTime(Chain,Dat,Tim) :-
	globalVars::get_i("ixrg")=1,
	!
;
	dbBase::getTimingEvent(Chain,D,T),
	datetime::timCmp(Dat,Tim,D,T,Cmp),
	Cmp=1,
	globalVars::set_i("ixrg",1),
	!
;
	!.		
% ----------------------------------------------------------------------------------------------
% dbaExtDB_Status                                                        
% ----------------------------------------------------------------------------------------------
dbaExtDB_Status(KB,Status) :-
	dbaBase::extChain(Chain,KB,1),
	dbaBase::changedChain(Chain),
	% DB has been updated
	Status=2,
	!
;
	dbaBase::database(KB,_,_,No),
	No>0,
	% DB is open, but not updated
	Status=1,
	!
;
	Status=0,
	!.	
% -------------------------------------------------------------------------------------------------------------
% dbaMultiRederive
% -------------------------------------------------------------------------------------------------------------
class predicates
dbaMultiDeriveStep2 : (rep,string_list,integer) determ (i,i,o).
dbaMultiDeriveStep1 : (rep,string_list,integer).
clauses

% parameters:
% 1 in list of KBs
% 2 in #cycles (min. 1)
%
dbaMultiRederive(L,N) :-
	F=globalVars::get_s("kb_file"),
	dbOpen::closeKB(0,-1,-1,-1),
			globalVars::set_i("msgon",-1),
			dlgIF::msgOffS(""),
    Rep=rep::newReport(2,"MultiRedrive",80),
	dbaMultiDeriveStep1(Rep,L,N),
			dlgIF::msgOffS("Multi-DB rederivation completed!"), 
	Rep:close(),
	dbOpen::openKB(F),
			globalVars::set_i("msgon",1),
	!.

dbaMultiDeriveStep1(Rep,L,NoCycles) :-
   N=programControl::sequence_nd(1,NoCycles,1),
			Sn=toString(N),
			dlgIF::msgOffS("Re-derivation cycle "),
			dlgIF::msgOffA(Sn),
        Rep:repLine('-'),
        Mess=string::concat("CYCLE #",Sn),
        Rep:repWL(0,Mess),
        Rep:repLine('-'),
        dbaMultiDeriveStep2(Rep,L,Nu),
			Nus=toString(Nu),
			Mess2=string::concat("Derived a total of ",Nus," sentences in cycle."),
			dlgIF::msgOffS(Mess2), 
			Rep:repWL(0,Mess2),
        fail
;
   !.

dbaMultiDeriveStep2(_,[],0) :-
	!.
dbaMultiDeriveStep2(Rep,[KB|KBs],N) :-
	dbOpen::openKB(KB),
			dlgIF::msgOffS("Re-deriving "),
			dlgIF::msgOffA(KB),
			dlgIF::msgOffA(": "),
            Mess=string::concat("Re-deriving ",KB),
            Rep:repWL(0,""),
            Rep:repWL(0,Mess),			
	der::rederiveAll(Rep,Ntypes,Nsents),
			St=toString(Ntypes),
			Ss=toString(Nsents),
			dlgIF::msgOffA(Ss),
			dlgIF::msgOffA(" sentences of "),
			dlgIF::msgOffA(St),
			dlgIF::msgOffA(" types."), 
	dbOpen::closeKB(0,-1,-1,-1),
	dbaMultiDeriveStep2(Rep,KBs,M),
	N=M+NSents,
	!.
% -------------------------------------------------------------------------------------------------------------
% dbaMultiDeleteDerived
% -------------------------------------------------------------------------------------------------------------
class predicates
dbaMultiDeleteDerivedX : (integer,string_list) determ.
clauses

% parameters:
% 1 in 1=delete derived, 2=regen dep graph, 3=1+2
% 2 in list of KBs
%
dbaMultiDeleteDerived(Graphs,KBs) :-
	F=globalVars::get_s("kb_file"),
	dbOpen::closeKB(0,-1,-1,-1),
	dbaMultiDeleteDerivedX(Graphs,KBs),
	dbOpen::openKB(F),
	!.
dbaMultiDeleteDerivedX(_,[]) :-
	!.
dbaMultiDeleteDerivedX(1,[KB|KBs]) :-
	% 1 - delete derived facts
	dbOpen::openKB(KB),
	dbaBase::dbaDeleteDerivedFacts(-1),
	dbOpen::closeKB(0,-1,-1,-1),
	dbaMultiDeleteDerivedX(1,KBs),
	!.
dbaMultiDeleteDerivedX(2,[KB|KBs]) :-
	% 2 - regenerate dep graph
	dbOpen::openKB(KB),
	dep::regenDependencyGraph,
	dbOpen::closeKB(0,-1,-1,-1),
	dbaMultiDeleteDerivedX(2,KBs),
	!.
dbaMultiDeleteDerivedX(3,[KB|KBs]) :-
	% 1 + 2
	dbOpen::openKB(KB),
	dbaBase::dbaDeleteDerivedFacts(-1),
	dep::regenDependencyGraph,
	dbOpen::closeKB(0,-1,-1,-1),
	dbaMultiDeleteDerivedX(3,KBs),
	!.
% ----------------------------------------------------------------------------------------------
% dbaPublicChains
% ----------------------------------------------------------------------------------------------
class predicates
dbaPubIx : (integer,string,integer,integer).
clauses
dbaPublicChains(Dat,Tim) :-
    dbaBase::database(DB,_,_,0),
	DB:chain_terms("p3_pIpch",Rec,_),
	   Rec=clause(2,_,_,_,_,E,_),
	   E=cmp("p3_pIpch",[cmp("p3_eChan",[atom(Chain),atom(AccCode)])]),
	   set::setGet(["p3_iRo","p3_iRw"],N,AccCode),	
	   Acc=N-1,
	   dbaBase::addPubChain(Chain,Acc),
  	   dbaPubIx(Acc,Chain,Dat,Tim),
			set::setGet(["read-only","read-write"],N,S2),
			sx::concat(["Noted public ",S2," chain ",Chain,"."],S),
			dlgIF::msgS(S),
	fail
;
	!.	
dbaPubIx(Acc,Chain,Dat,Tim) :-
	Acc=0,
	!
;	
	dbBase::getTimingEvent(Chain,D,T),
	datetime::timCmp(Dat,Tim,D,T,Cmp),
	Cmp=1,
	globalVars::set_i("ixrg",1),
	!
;
	!.		
% ----------------------------------------------------------------------------------------------
% readLocations                                                            
% ----------------------------------------------------------------------------------------------
class predicates
readLocs : (sterml) determ.
readChains : (string,sterml) determ.    
clauses

readLocations(FileName) :-
    sqlBase::names(FileName,_,_,_),
    !
;    
    retractall(exDb(_,_)),
    retractall(exCh(_,_,_)),
	fioExt::filChangeType(FileName,"xml",File),
	file::existFile(File),
	X=file::readString(File),
	xmlGeneric::remHeader(X,X2),
	XML=xmlTree::newTree(X2),
	Tree=XML:getSubTree("databaseLocations"),
	Tree=cmp(_,
	               [_,
	                cmp("mainDatabaseId",[_,str(Id)])
	                |ExtDBs]),
    globalVars::set_s("kb_id",Id),
    readLocs(ExtDBs),       
	!
;
	!.
	
readLocs([]) :-
       !.
readLocs([X|Xs]) :-       
    X=cmp("externalDatabase",
              [_,
               cmp("externalDatabaseId",[nill,str(I)]),
               cmp("externalDatabaseFile",[nill,str(F)])
               |L]),
    assert(exDb(F,I)),
    readChains(I,L),
    readLocs(Xs),
    !
;
    S=toString(X),
    Mess=string::concat("XML-error: Invalid externalDatabase-tag: ",S,"!"),
    dlgIF::dlgNote(1,Mess),
    !,fail.
readChains(_,[]) :-
    !.
readChains(Id,[X,Y|Xs]) :-
    X=cmp("access",[_,str(A)]),
    Y=cmp("chain",[_,str(C)]),
    set::setGet(["READ","WRITE"],N,A),
    Ai=N-1,
    Ai>=0,
    Ai<=1,
    assert(exCh(C,Id,Ai)),
    readChains(Id,Xs),
    !
;
    S2=toString(X),
    S3=toString(Y),
    Mess=string::concat("XML-error: Invalid access- or chain-tag: ",S2," ",S3,"!"),
    dlgIF::dlgNote(1,Mess),
    !,fail.
% ----------------------------------------------------------------------------------------------
% writeLocations                                         
% ----------------------------------------------------------------------------------------------
class predicates
extLocs : (string_list,string_list,integer_list,sterml) determ (i,i,i,o).
extChains : (string,string_list,sterml) determ (i,i,o).    
clauses

% (re-)creates the main DB's location file on the basis of external chains declared in the DB
% 
writeLocations() :-
    dbBase::kbIsSql,
    !
;    
	File=globalVars::get_s("kb_file"),
	fioExt::filChangeType(File,"xml",File2),
	fioExt::filChangeType(File,"xmlc",File3),
	fio::filCopy(File2,File3),
    dbaBase::getDatabases(_,Files,Ids,Nos),
	extLocs(Files,Ids,Nos,ExternalDBs),
	Id=globalVars::get_s("kb_id"),
	Tree=cmp("databaseLocations",
	               [nill,
	                cmp("mainDatabaseId",[nill,str(Id)])
	                |ExternalDBs]),
	xmlP3::scriptWrap("PR",[Tree],Xt),
	XML=xmlTree::newTreeXML(Xt),
	X=XML:getX(5),
	xmlGeneric::xmlHeader(X,Xh),
	file::writeString(File2,Xh),
	!.
	
extLocs(_,_,[],[]) :-
    !.
extLocs([F|Files],[I|Ids],[N|Nos],[X|Xs]) :-
    N>0,
    dbaBase::database(DB,_,I,_),
    dbaBase::getDB_ExtChains(DB,Cr,Cw),
    extChains("READ",Cr,Lr),
    extChains("WRITE",Cw,Lw),
    set::conc([Lr,Lw],L),
    X=cmp("externalDatabase",
              [nill,
               cmp("externalDatabaseId",[nill,str(I)]),
               cmp("externalDatabaseFile",[nill,str(F)])
               |L]),
    extLocs(Files,Ids,Nos,Xs),
    !.
extLocs([_|Files],[_|Ids],[_|Nos],Xs) :-
    extLocs(Files,Ids,Nos,Xs),
    !.
extChains(_,[],[]) :-
    !.
extChains(A,[C|Cs],[cmp("access",[nill,str(A)]),cmp("chain",[nill,str(C)])|Ts]) :-
    extChains(A,Cs,Ts),
    !.
% -------------------------------------------------------------------------------------------------------------
% dbaCloseExternals
% -------------------------------------------------------------------------------------------------------------
class predicates
dbaUpdateExternals : (chainDB DB,integer,integer) procedure.
dbaCloseExt_2 : (integer,integer,integer,chainDB,string,integer,integer) determ.	
clauses
% parameters:
% 1	in	0=close/no changes; 1=save changes; 2=restore (either explicit restore or restore
%               necessitated by RAM chains (always require restore even with no changes)
%
dbaCloseExternals(Context,Dat,Tim) :-
    dbaBase::database(KB,File,Id,No),
    No>0,
			tt::tt_s(2,"Closing external DB",File),
			tt::tt_s(2,"Closing external DB",Id),
    dbaExtDB_Status(KB,Status),
            tt::tt_i(2,"Status",Status),
	Backup=globalVars::get_i("backup_active"),
			tt::tt_i(2,"Backup",Backup),
	dbaCloseExt_2(Backup,Context,Status,KB,File,Dat,Tim),
			sx::concat(["Closed KB ",File,"."],S3),
			dlgIF::msgS(S3),
	fail
;	
			tt::tt_ss(2,"All external DBs closed"),
	%dbaBase::dropMap(),
	!.
dbaCloseExt_2(1,Context,Status,KB,File,Dat,Tim) :-
	Status=2, 
	Context=1,
	% updated - save
	dbaUpdateExternals(KB,Dat,Tim),
	KB:db_flush(),
	KB:db_close(),
			S=toString(KB),
			tt::tt_s(1,"Closed",S),
	dbaUtil::dbaFlagSet(File,"3op",0),
	dbaUtil::dbaFlagSet(File,"3eq",0),
	!
;	
	Status=2, 
	Context=2,
	% updated - restore
	KB:db_close(),
			S=toString(KB),
			tt::tt_s(1,"Closed",S),
	fioExt::filChangeType(File,"3kc",BFile),
	file::copy(BFile,File),
	dbaUtil::dbaFlagSet(File,"3op",0),
	dbaUtil::dbaFlagSet(File,"3eq",1),
	!
;	
	Status=1,
	% no change
	KB:db_close(),
			S=toString(KB),
			tt::tt_s(1,"Closed",S),
	dbaUtil::dbaFlagSet(File,"3op",0),
	dbaUtil::dbaFlagSet(File,"3eq",1),
	!
;
	S1=toString(Context),
	S2=toString(Status),
	sx::concat(["Invalid close parameters (1,",S1,",",S2,") for external DB ",File],Mess),
	dlgIF::dlgNote(2,Mess),
	!.		
dbaCloseExt_2(-1,Context,Status,KB,File,Dat,Tim) :-
	Status=2,
	dbaUpdateExternals(KB,Dat,Tim),
	KB:db_flush(),
	KB:db_close(),
			S=toString(KB),
			tt::tt_s(1,"Closed",S),
	!
;	
	Status=1,
	dbaUpdateExternals(KB,Dat,Tim),
	KB:db_flush(),
	KB:db_close(),
			S=toString(KB),
			tt::tt_s(1,"Closed",S),
	!
;	
	S1=toString(Context),
	S2=toString(Status),
	sx::concat(["Invalid close parameters (0,",S1,",",S2,") for external DB ",File],Mess),
	dlgIF::dlgNote(2,Mess),
	!.
			
% sets update-time for external read-write chains
%	
dbaUpdateExternals(KB,Dat,Tim) :-
	dbaBase::changedChain(Chain),
	dbBase::insertTimingEvent(Chain,Dat,Tim),
	Mess=string::concat("Read-only chain '",Chain,"' marked as changed!"),
	dlgIF::dlgNote(1,Mess),
	fail
;
	dbaBase::changedChain(Chain),
	dbaBase::extChain(Chain,KB,1),
	dbBase::insertTimingEvent(Chain,Dat,Tim),
	fail
;
	!.		
% ----------------------------------------------------------------------------------------------
% dbaVerifySettings                                                           
% ----------------------------------------------------------------------------------------------
class predicates
dbaVF_a : (chainDB,integer).
dbaVF_b : (chainDB,integer).
dbaVF_c : (chainDB,integer).
clauses

dbaVerifySettings(DB,T,K,R) :-
	dbaVF_a(DB,T),
	dbaVF_b(DB,K),
	dbaVF_c(DB,R),!.
	
dbaVF_a(DB,T) :-	
	TT=globalVars::get_i("max_tok"),
	TT<T,
	dbaBase::database(DB,File,_,_),
	sx::concat(["Max. token length in DB ",File," exceeds that of main DB. This can give index integrity problems."],Mess),
	dlgIF::dlgNote(1,Mess),!
;
	!.
dbaVF_b(DB,K) :-	
	TT=globalVars::get_i("max_kbn"),
	TT<K,
	dbaBase::database(DB,File,_,_),
	sx::concat(["Max. KB-name length in DB ",File," exceeds that of main DB. This can give index integrity problems."],Mess),
	dlgIF::dlgNote(1,Mess),!
;
	!.
dbaVF_c(DB,R) :-	
	TT=globalVars::get_i("max_rln"),
	TT<R,
	dbaBase::database(DB,File,_,_),
	sx::concat(["Max. realm name length in DB ",File," exceeds that of main DB. This can give index integrity problems."],Mess),
	dlgIF::dlgNote(1,Mess),!
;
	!.
% --------------------------------------------------------------------------
% externalDBs
% --------------------------------------------------------------------------
class predicates
extTerm : ().
extSM : ().
clauses

% opens external DBs during DB initialization if so requested
%
externalDBs() :-
    dbaBase::dbaMap("v",DB1,_),
    dbaBase::dbaMap("q",DB2,_),
    DB1:chain_first("v",_),
    DB2:chain_first("q",_),
	Reply=vpiCommonDialogs::ask("External databases?","Do you want to access (more) external databases?",["Yes","No"]),
    Reply=vpiCommonDialogs::resp_default,
    W=guix::getWinOfClas(1),
    _=externalchains::new(W),
	dbaUtil::dbaIxRegenAll,
    !
;    
    dbaBase::dbaMap("v",DB,_),
    not(DB:chain_first("v",_)), % no terminology
    extTerm,
	dbaUtil::dbaIxRegenAll,
	externalDBs,
    !
;    
    dbaBase::dbaMap("q",DB,_),
    not(DB:chain_first("q",_)), % no SM
    dlgIF::dlgNote(0,"You must now select the external DB with the sentence model (chain q)"),
    W=guix::getWinOfClas(1),
    _=externalchains::new(W),
	dbaUtil::dbaIxRegenAll,
    !
;
    !.

extTerm() :-
	File=vpiCommonDialogs::getFileName("*.3kb",["KB files","*.3kb"],"Select the external DB with the terminology (chain v)",[],"",_),
	fioExt::filValidKB_Name(1,File,_),
    dbaUtil::dbaOpenAuxDB(File,DB2),
    L=pop::getChainsPubDB(DB2,-1),
    set::member("v",L),
    dbaUtil::dbaCloseAux(DB2),
    Id=vpiCommonDialogs::getString("External database identifier","Enter the one-letter identifier used for the terminology DB within current KB",""),
    Id2=string::toUpperCase(Id),
    string::length(Id2)=1,
    openExternalDB(File,Id2,DB),
    dbaBase::addExtChain("v",DB,0),
    !
;
    !.

extSM() :-
    !
;
    !.
    
end implement dbaExt
