/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement dbaUtil
open core,termDomain,globalDomains

constants
className = "com/rudstrand/dbaUtil/dbaUtil".
classVersion = "1.0".

class predicates
dbaRepairID : (string Id) determ (o).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% dbuCR_GetRef
% ----------------------------------------------------------------------------------------------           
dbuCR_GetRef(Rule,Body,Params,KB,Ref,Realm,RT) :-
	dbuCR_GetRefF(Rule,Body,Params,KB,Ref,Realm,RT),
	!
;	
	dlgIF::feil("dbuCR_GetRef",["Can't find certainty rule",Rule],[],nill),
	!,fail.	

dbuCR_GetRefF(Rule,Body,DE_Defs,KB,Ref,Realm,RT) :-
    dbBase::irChain_nd(Chain),
            dbaBase::dbaMap(Chain,KB,_),
            KB:chain_terms(Chain,Rec,Ref),
                    Rec=clause(RT,_,_,_,Realm,Body,_),
                    Body=cmp(_,[atom(Rule)|_]),
                    kbBase::irDeclarationF(Rule,_,DE_Defs,_,_), 
                    !
;                    
    dbBase::irChain_nd(Chain),
            dbaBase::dbaMap(Chain,KB,_),
            KB:chain_terms(Chain,Rec,Ref),
                    Rec=clause(RT,_,_,_,Realm,Body,_),
                    Body=cmp(_,[atom(Rule)|_]),
                    not(kbBase::irDeclarationF(Rule,_,_,_,_)),
                    Mess=string::concat("No declaration found for inexact rule: ",Rule), 
                    dlgIF::dlgNote(3,Mess), 
                    !,fail.                    
% ----------------------------------------------------------------------------------------------
% dbuChangeRecType
% ----------------------------------------------------------------------------------------------
class predicates
dbuCRT_Valid : (integer,integer,integer) determ (i,i,i).
dbuCRT_Valid2 : (integer,integer) determ (i,i).
clauses

% changes a record's type - validates that the change is logical - uses db update if graphs/indexes are
% unaffected
%
dbuChangeRecType(KB,Ref,NewType) :-
	% validate 1
	dbBase::recordType(NewType,Base,Act,Susp),
	not(dbuCRT_Valid(Base,Act,Susp)),
	KB:ref_term(Ref,clause(CurrType,_,SentNo,SubNo,SourceFile,LTerm,RTerm)),
	dlgIF::feil("dbuChangeRecType",[SourceFile],[SentNo,SubNo,CurrType,NewType],cmp("",[LTerm,RTerm])),
	!,fail
;
	% validate 1
	dbBase::recordType(NewType,Base,Act,Susp),
	not(dbuCRT_Valid(Base,Act,Susp)),
	dlgIF::feil("dbuChangeRecType",[],[NewType],nill),
	!,fail
;
	% validate 2
	KB:ref_term(Ref,clause(CurrType,_,_,_,_,_,_)),
	dbBase::recordType(CurrType,Base,Act,Susp),
	not(dbuCRT_Valid(Base,Act,Susp)),
	!,fail
;
	% validate 3
	KB:ref_term(Ref,clause(CurrType,_,_,_,_,_,_)),
	dbBase::recordType(CurrType,Base,_,_),
	dbBase::recordType(NewType,NBase,_,_),
	not(dbuCRT_Valid2(Base,NBase)),
	!,fail
;
	% no change
	KB:ref_term(Ref,clause(CurrType,_,_,_,_,_,_)),
	CurrType=NewType,
	!
;
	% temporary change between 1 and 8
	KB:ref_term(Ref,clause(CurrType,Ts,SentNo,SubNo,SourceFile,LTerm,RTerm)),
	dbBase::recordType(CurrType,Base,_,_),
	dbBase::recordType(NewType,NBase,_,_),
	Base<>NBase,
	KB:term_replace(Ref,clause(NewType,Ts,SentNo,SubNo,SourceFile,LTerm,RTerm)),
	!
;	
	% suspension/desuspension
	dbBase::recordType(NewType,_,_,Susp),
	KB:ref_term(Ref,clause(CurrType,_,SentNo,SubNo,Realm,LTerm,RTerm)),
	dbBase::recordType(CurrType,_,_,CSusp),
	Susp<>CSusp,
	not(dbBase::accessErrorTerm(LTerm)),
	dbaBase::dbaDeleteRecord(KB,Ref),
	dbaBase::dbaInsertRecord(NewType,Realm,SentNo,SubNo,LTerm,RTerm,_),
	!
;
	% activation/deactivation
	dbBase::recordType(NewType,_,Act,_),
	KB:ref_term(Ref,clause(CurrType,Ts,SentNo,SubNo,SourceFile,LTerm,RTerm)),
	dbBase::recordType(CurrType,_,CAct,_),
	Act<>CAct,
	not(dbBase::accessErrorTerm(LTerm)),
	KB:term_replace(Ref,clause(NewType,Ts,SentNo,SubNo,SourceFile,LTerm,RTerm)),
	LTerm=cmp(Chain,_),
	S2=toString(CurrType),
	sx::sxlJoin([S2,Chain],"|",CKey),
	Ix=dbaBase::getIx('C',"RTYP"),
	dbaBase::database(MKB,_,_,0),
	MKB:key_search(Ix,CKey,Ref),
	MKB:key_delete(Ix,CKey,Ref),
	S3=toString(NewType),
	sx::sxlJoin([S3,Chain],"|",NKey),
	dbaBase::dbaIxIns("RTYP","ix_rtyp",NKey,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	KB:ref_term(Ref,clause(CurrType,_,SentNo,SubNo,SourceFile,LTerm,RTerm)),
	dlgIF::feil("dbuChangeRecType",[SourceFile],[SentNo,SubNo,CurrType,NewType],cmp("",[LTerm,RTerm])),
	!
;
	dlgIF::feil("dbuChangeRecType",[],[NewType],nill),
	!,fail.

dbuCRT_Valid(B,A,S) :-
	B<1,
	!,fail
;
	B>8,
	!,fail
;
	A=-1,
	not(set::member(B,[3,5,7])),
	!,fail
;
	S=1,
	not(set::member(B,[3,4,5,6,7])),
	!,fail
;
	!.
dbuCRT_Valid2(Ba,Bb) :-
	Ba=Bb,
	!
;
	Ba=1,
	Bb=8,
	!
;
	Ba=8,
	Bb=1,
	!
;			
	!,fail.
% ----------------------------------------------------------------------------------------------
% dbuRealm/C/No
% ----------------------------------------------------------------------------------------------
dbuRealmC(Chain,Ref,KB,Realm) :-
	dbuRealmNoC(Chain,Ref,KB,Realm,_,_),
	!.
dbuRealmNo(KB,Ref,Realm,No,SNo) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,No,SNo,Realm,_,_),
	!.
dbuRealmNoC(Chain,Ref,KB,Realm,No,SNo) :-
	dbaBase::dbaMap(Chain,KB,_),
	dbuRealmNo(KB,Ref,Realm,No,SNo),
	!.
% ----------------------------------------------------------------------------------------------
% getCR_TermF
% ----------------------------------------------------------------------------------------------
getCR_TermF(Rule,LT,Ref) :-
    dbBase::irChain_nd(Chain),
            dbaBase::dbaMap(Chain,KB,_),
            KB:chain_terms(Chain,Rec,Ref),
                    Rec=clause(_,_,_,_,_,LT,_),
                    LT=cmp(_,[atom(Rule)|_]),
                    !.
            
% ----------------------------------------------------------------------------------------------
% dbuDBRF
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	DBRF-key
% 2	out	KB 
% 3	out	Ref
%
dbuDBRF(Key,KB,Ref) :-
	sx::frontstr(12,Key,SRef,Id),
	dbaBase::dba_address(Ref,SRef),
	dbaBase::database(KB,_,Id,_),
	!
;
	dlgIF::feil("dbuDBRF",[Key],[],nill),!,fail.
% ----------------------------------------------------------------------------------------------
% dbuPubChainsIns
% ----------------------------------------------------------------------------------------------
dbuPubChainsIns(_,_,[]) :-
	!.
dbuPubChainsIns(KB,Access,[Chain|Chains]) :-
	dbaBase::dbaInsertRecord(2,"$SYS",0,0,cmp("p3_pIpch",[cmp("p3_eChan",[atom(Chain),atom(Access)])]),atom("true"),_),
	dbuPubChainsIns(KB,Access,Chains),
	!.
% ----------------------------------------------------------------------------------------------
% dbuRuleRef
% ----------------------------------------------------------------------------------------------
dbuRuleRef(Rule,KB,Ref) :-
	sx::sxUnpack(Rule,'|',L),
	L=[Chain|_],
	dbaBase::dbaMap(Chain,KB,_),
	Ix=dbaBase::getIx('C',"RULE"),
	dbaBase::database(MKB,_,_,0),
	MKB:key_search(Ix,Rule,Ref),
	!.
% ----------------------------------------------------------------------------------------------
% dbuGetInRealm
% ----------------------------------------------------------------------------------------------
class predicates
dbuGIR : (chainDB,chainDB::bt_selector,string,integer,integer,chainDB::ref,integer) determ (i,i,i,i,i,o,o).
dbuGIR_Check : (string,integer,integer,string,integer,integer,integer) determ (i,i,i,i,i,i,o).
clauses

dbuGetInRealm(KbId,Realm,N,M,KB,Ref) :-
	dbaBase::database(KB,_,KbId,_),
	Ix=dbaBase::getIx('C',"REALM"),
	sx::sxlJoin([Realm,KbId],"|",Key),
	dbaBase::database(MKB,_,_,0),
	MKB:key_search(Ix,Key,_), 
	MKB:key_current(Ix,FKey,Ref),
	Key=FKey,
	KB:ref_term(Ref,clause(_,_,NN,MM,_,_,_)),
	N=NN,
	M=MM,
	!
;	
	dbaBase::database(KB,_,KbId,_),
	Ix=dbaBase::getIx('C',"REALM"),
	sx::sxlJoin([Realm,KbId],"|",Key),
	dbaBase::database(MDB,_,_,0),
	MDB:key_search(Ix,Key,_), 
	dbuGIR(KB,Ix,Key,N,M,Ref,Match),
	Match=1,
	!.	
dbuGIR(KB,Ix,Key,N,M,Ref,Match) :-	
	dbaBase::database(MKB,_,_,0),
    programControl::repeat,
	   MKB:key_next(Ix,_),
	   MKB:key_current(Ix,FKey,Ref),
	   KB:ref_term(Ref,clause(_,_,FN,FM,_,_,_)),
	   dbuGIR_Check(Key,N,M,FKey,FN,FM,Match),
	   !.
dbuGIR_Check(Key,N,M,FKey,FN,FM,Match) :-
	Key<>FKey,
	Match=0,
	!
;
	N=FN,
	M=FM,
	Match=1,
	!.
% ----------------------------------------------------------------------------------------------
% dbuGetRecTypes
% ----------------------------------------------------------------------------------------------
class predicates	
dbuGetRT_2 : (chainDB::bt_selector,lst,string,integer) determ.
dbuGetRT_3 : (chainDB::bt_selector,lst,string,integer) determ.
clauses

% returns a list of all records (DBRF key format) of given record type
%
dbuGetRecTypes(RecType,List) :-
	S2=toString(RecType),
	Key=string::concat(S2,"|"),
	Ix=dbaBase::getIx('C',"RTYP"),
	Arr=lst::new(-1),
	L=string::length(Key),
	dbuGetRT_2(Ix,Arr,Key,L),
	List2=Arr:get(),
	dbBase::toLocs(List2,List),
	!.

dbuGetRT_2(Ix,Arr,Key,L) :-
    dbaBase::database(MKB,_,_,0),
	MKB:key_search(Ix,Key,_), 
	!,
	programControl::repeat,
	   dbuGetRT_3(Ix,Arr,Key,L),
	!
;	
	programControl::repeat,
	   dbuGetRT_3(Ix,Arr,Key,L),
	!.
		
dbuGetRT_3(Ix,Arr,Key,L) :-
    dbaBase::database(MKB,_,_,0),
	MKB:key_current(Ix,FKey,Ref),
	sx::frontstr(L,FKey,CKey,_),
	Key=CKey,
	%dbaBase::dba_address(Ref,SRef),
	sx::sxUnpack(FKey,'|',L2),
	L2=[_,Chain],
	dbaBase::dbaMap(Chain,KB,_),
	%dbaBase::database(KB,_,Id,_),
	%S=string::concat(SRef,Id),
	S=toString(cloc(KB,Ref)),
	Arr:add(S),
	MKB:key_next(Ix,_),
	!,fail
;
	!.	
% ----------------------------------------------------------------------------------------------
% dbuIxCountFront
% ----------------------------------------------------------------------------------------------
class predicates
dbuIxCountFrontX : (chainDB,chainDB::bt_selector,string,integer,count) determ.
clauses

dbuIxCountFront(Index,Key,N) :-
	Ix=dbaBase::getIx('C',Index),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Key,_), 
	!,
	Cnt=count::newCounter(0),
	L=string::length(Key),
	programControl::repeat,
	   dbuIxCountFrontX(DB,Ix,Key,L,Cnt),
	N=Cnt:get(),
	!
;
	Ix=dbaBase::getIx('C',Index),
	dbaBase::database(DB,_,_,0),
	not(DB:key_search(Ix,Key,_)), 
	!,
	Cnt=count::newCounter(0),
	L=string::length(Key),
	programControl::repeat,
	   dbuIxCountFrontX(DB,Ix,Key,L,Cnt),
	N=Cnt:get(),
	!
;
	N=0,
	!.

dbuIxCountFrontX(DB,Ix,Key,L,Cnt) :-  
	DB:key_current(Ix,FKey,_),
	sx::frontstr(L,FKey,CKey,_),
	Key=CKey,
	Cnt:inc(_),
	DB:key_next(Ix,_),
	!,fail
;
	!.	
% -------------------------------------------------------------------------------------------------------------
% dbaBackup
% -------------------------------------------------------------------------------------------------------------
dbaBackup(0,_) :-
	!.
dbaBackup(_,_) :-
	globalVars::get_i("backup")=-1,
	globalVars::set_i("backup_active",-1),
	!.
dbaBackup(1,KB) :-
	dbaFlagON(KB,"3eq"),
	fioExt::filChangeType(KB,"3kc",Backup),
	file::existFile(Backup),
	globalVars::set_i("backup_active",1),
	!
;
	dbaFlagON(KB,"3eq"),
	fioExt::filChangeType(KB,"3kc",Backup),
	not(file::existFile(Backup)),
	file::copy(KB,Backup),
	dbaFlagSet(KB,"3eq",1),
	globalVars::set_i("backup_active",1),
	!
;	
	not(dbaFlagON(KB,"3eq")),
	fioExt::filChangeType(KB,"3kc",Backup),
	file::copy(KB,Backup),
	dbaFlagSet(KB,"3eq",1),
	globalVars::set_i("backup_active",1),
	!.
% ----------------------------------------------------------------------------------------------
% dbaCheckDB_Id                                                    
% ----------------------------------------------------------------------------------------------
class predicates
dbaGetDB_Id : (string) determ (o).
dbaCheckId : (string) determ.
clauses

dbaCheckDB_Id(Id) :-
	dbaGetDB_Id(Id),
	dbaCheckId(Id),
	!
;
	Id=globalVars::get_s("kb_idf"),
	Id<>"",
	dbOpen::storeDBID(Id),
	!
;
	dbaRepairID(Id),
	!.
				
dbaGetDB_Id(KBi) :-
    dbaBase::database(DB,_,_,0),
	DB:chain_terms("x",clause(_,_,_,_,_,LTerm,_),_),
	LTerm=cmp("x",[cmp("p3_eDbas",[str(KBi)])]),
	!.	
dbaCheckId(Id) :-
	globalVars::get_s("kb_idf")="",
	!
;
	Idf=globalVars::get_s("kb_idf"),
	Idf<>Id,
	dlgIF::feil("dbaCheckId",["DB ID mismatch",Id,Idf],[],nill),!,fail
;
	!.		
% ----------------------------------------------------------------------------------------------
% dbaCloseBasic                                                             
% ----------------------------------------------------------------------------------------------
% parameters:
% 1 in  Kb  DB
% 2 in  C index character (Y/C)
%
dbaCloseBasic(Kb,C) :-
    dbaBase::dbaIndexesOps(Kb,C,0),
	Kb:db_flush(),
	Kb:db_close(),
	dbaBase::database(Kb,File,Id,No),
	SNo=toString(No),
	Mess=string::concatList(["Closed KB ",File," [",Id," #",SNo,"]"]),
    dlgIF::msgS(Mess),
	!
;
	dlgIF::feil("dbaCloseBasic",[],[],nill),!,fail.	
% ----------------------------------------------------------------------------------------------
% dbaCloseAux                                                       
% ----------------------------------------------------------------------------------------------
% parameters:
% 1 in  DB
%
dbaCloseAux(DB) :-
	DB:db_flush(),
	DB:db_close(),
	dbaBase::database(DB,File,Id,_),
	dbaBase::dropDatabase(DB),
	Mess=string::concatList(["Closed auxilliary DB '",File,"' [",Id,"]"]),
    dlgIF::msgS(Mess),
	!
;
	dlgIF::feil("dbaCloseAux",[],[],nill),!,fail.	
% ----------------------------------------------------------------------------------------------
% dbaDeleteChain                                                              
% ----------------------------------------------------------------------------------------------
dbaDeleteChain(KB,Chain) :-
	globalVars::get_i("quit_flag")=1,
	dlgIF::dlgNote(1,"Cannot delete a chain when QUIT has been requested"),
	!,fail
;
    dbBase::chainReadonlyWarning(Chain),
    !,fail
;	   
    dep::massSentenceUpdateStart,
	KB:chain_terms(Chain,Rec,Ref),
	Rec=clause(_,_,_,_,_,_,_),
    dbaBase::dbaDeleteRecord(KB,Ref),
	fail
;
    dep::massSentenceUpdateEnd,
	!.	
% ----------------------------------------------------------------------------------------------
% dbaDeleteRealm                                                              
% ----------------------------------------------------------------------------------------------
class predicates
deleteRealmKey : (chainDB,string,string) determ.
deleteRealmX : (chainDB,chainDB,chainDB::bt_selector,string) determ.
clauses

dbaDeleteRealmKB(KB,Realm) :-
	globalVars::get_i("quit_flag")=1,
	dlgIF::dlgExp(3),
	dlgIF::dlgNote(1,"Cannot delete a realm when QUIT has been requested"),
	!,fail
;	 
	set::member(Realm,["$SYS"]),
	dlgIF::dlgNote(1,"System realm $SYS cannot be deleted!"),
	!,fail
;
	dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dep::massSentenceUpdateStart,
	deleteRealmKey(KB,"REALM",Key),
	dep::massSentenceUpdateEnd,
	!,globalVars::get_i("quit_flag")=0,
	!.

deleteRealmKey(KB,Index,Key) :-
	Ix=dbaBase::getIx('C',Index),
	dbaBase::database(MDB,_,_,0),
	programControl::repeat,
	   deleteRealmX(KB,MDB,Ix,Key),
	   !.

deleteRealmX(KB,MDB,Ix,Key) :-
	% delete current record and make next record in index current
	globalVars::get_i("quit_flag")=0,
	MDB:key_search(Ix,Key,Ref),
	dbBase::recordDelete(cloc(KB,Ref)),
	!,fail
;
	!.	
% ----------------------------------------------------------------------------------------------
% dbaDeleteKB_Realm                                                              
% ----------------------------------------------------------------------------------------------
dbaDeleteKB_Realm([],_) :-
	!.
dbaDeleteKB_Realm([File|Files],Realm) :-
	dbaBase::database(KB,File,_,_),
	dbaDeleteRealmKB(KB,Realm),
	dbaDeleteKB_Realm(Files,Realm),
	!.
% ----------------------------------------------------------------------------------------------
% dbaFlagON                                                        
% ----------------------------------------------------------------------------------------------
% checks if the given flag is ON
%
dbaFlagON(FileName,Flag) :-
	fioExt::filChangeType(FileName,Flag,F3),
	file::existFile(F3),
	!
;	
	fioExt::filChangeType(FileName,Flag,F3),
	not(file::existFile(F3)),
	!,fail
;
	dlgIF::feil("dbaFlagON",[Flag],[],nill),!,fail.	
% -------------------------------------------------------------------------------------------------------------
% dbaFlagSet
% -------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	flag: 3OP (KB is not properly closed) 3eq (3kc = 3kb)
% 2	in	1:set flag ON;  0:set flag OFF
%
dbaFlagSet(KBF,Flag,1) :-
	globalVars::get_i("backup_active")=-1,
	!
;		
	fioExt::filChangeType(KBF,"3fg",F2),
	file::existFile(F2),
	fioExt::filChangeType(KBF,Flag,F3),
	file::copy(F2,F3),
	!
;
	% std flag file template doesn't exist - create it and call again!
	fioExt::filChangeType(KBF,"3fg",F2),
	not(file::existFile(F2)),
	file::writeString(F2,"#"),
	dbaFlagSet(KBF,Flag,1),
	!
;
	dlgIF::feil("dbaFlag",[Flag],[1],nill),!,fail.		
dbaFlagSet(FileName,Flag,0) :-	
	fioExt::filChangeType(FileName,Flag,F3),
	file::existFile(F3),
	file::delete(F3),
	!
;
	!.	
% --------------------------------------------------------------------------
% dbaOpenRestore
% --------------------------------------------------------------------------
% attempts to copy 3KC to 3KB because 3KB was not properly closed (only called if backup system used)
% NOTE! succeeds only if input is Cancel or error encountered.
%	
dbaOpenRestore(vpiCommonDialogs::resp_default,FileName) :-
	fioExt::filChangeType(FileName,"3kc",Backup),
	not(file::existfile(Backup)),
	sx::concat(["Backup of database ",FileName," does not exist!"],Mess),
	dlgIF::dlgNote(1,Mess),
	dlgIF::setFlag("error_flag",1),
	dlgIF::setFlag("quit_flag",1),
	!
;
	fioExt::filChangeType(FileName,"3kc",BackFile),
	file::copy(BackFile,FileName),
	dbaFlagSet(FileName,"3eq",1),
	dbaFlagSet(FileName,"3op",0),
	!,fail.
dbaOpenRestore(vpiCommonDialogs::resp_3,_) :- 
	dlgIF::setFlag("quit_flag",1),
	!.		
% -------------------------------------------------------------------------------------------------------------
% dbaOpenRegenIx
% -------------------------------------------------------------------------------------------------------------
dbaOpenRegenIx() :-
	globalVars::get_i("ixrg")=1,
                dlgIF::msgStatus("regenerating indexes..."),
	vpi::cursorSetWait(),
	dbaIxRegenAll,
	W=guix::getWinOfClas(1),
	vpi::cursorSet(W,vpiDomains::cursor_Arrow),
	dbBase::chainHasChangedTerm(nill),
                dlgIF::msgStatus(""),
                dlgIF::msgS("Regenerated indexes."),
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
% dbaIxRegenAll
% ----------------------------------------------------------------------------------------------
class predicates
dbaIxGenX : (count,count).
clauses

% regenerates all KB indexes for all KB records
%
dbaIxRegenAll() :-
	C1=count::newCounter(0),
	C2=count::newCounter(0),
	dbaIxGenX(C1,C2),
	dbBase::chainHasChangedTerm(nill),
	globalVars::set_i("ixrg",-1),
        N1=C1:get(),
        S1=toString(N1),
        N2=C2:get(),
        S2=toString(N2),
		sx::concat(["Generated indexes for ",S2," records in ",S1," chains."],Mess),
		dlgIF::msgS(Mess),
	!.

dbaIxGenX(C1,C2) :-
	vpi::cursorSetWait(),
	dbaBase::database(DB,_,_,0),
	dbaBase::dbaIndexesOps(DB,'C',3),
	dbaBase::dbaIndexesOps(DB,'C',2),
	dbaBase::dbaChains(KB,Chain),
	    dbaBase::database(KB,Name,Id,_),
		sx::concat(["Generates indexes for chain ",Chain," in ",Name," (",Id,")..."],Mess),
		dlgIF::msgS(Mess),
	    C1:inc(_),
	    KB:chain_terms(Chain,Rec,Ref),
	    C2:inc(_),
	    dbaLib::ixsInsert(KB,Ref,Rec),
	fail
;
	!.
% ----------------------------------------------------------------------------------------------
% dbaRepairID                                             
% ----------------------------------------------------------------------------------------------
dbaRepairID(Id) :-
	Id=vpiCommonDialogs::getString("The DB has no one-letter DB-identifier","Please specify an identifier for the main DB",""),
    string::length(Id)=1,
	!,
	globalVars::set_s("kb_id",Id),
    dbaBase::dbaInsertRecord(2,"$SYS",0,0,cmp("x",[cmp("p3_eDbas",[str(Id)])]),atom("true"),_),
	!
;
	dbaRepairID(Id),
	!.
% -------------------------------------------------------------------------------------------------------------
% dbaOpenAuxDB
% -------------------------------------------------------------------------------------------------------------
% formerly dbaOpenCopyDB
%
dbaOpenAuxDB(FileName,DB) :-
	vpi::cursorSetWait(),
	file::existfile(FileName),
	DB=chainDB::db_open(FileName,chainDB::in_file),
	dbaBase::addDatabaseAux(DB,FileName),
	%dbaBase::dbaIndexesOps(KB,'Y',1), % should not need indexes in aux. databases 8.1.5
	sx::concat(["Opened auxilliary DB ",FileName],Mess),
	dlgIF::msgS(Mess),
	!
;
	not(file::existfile(FileName)),
	dbOpen::create(FileName),
	dbaOpenAuxDB(FileName,DB),
	!
;
	file::existfile(FileName),
	sx::concat(["Cannot open auxilliary DB ",FileName],Mess),
	dlgIF::dlgNote(3,Mess),
	!,fail.
% ----------------------------------------------------------------------------------------------
% dbaCountChain                                                             
% ----------------------------------------------------------------------------------------------
class predicates
dbaCCX : (chainDB,string,count).
clauses

dbaCountChain(Chain,N) :-
    C=count::newCounter(0),
	dbaBase::dbaMap(Chain,KB,_),
	dbaCCX(KB,Chain,C),
	N=C:get(),
	!.
	
dbaCCX(KB,Chain,C) :-	
	KB:chain_terms(Chain,Rec,_),
	Rec=clause(_,_,_,_,_,_,_),
	C:inc(_),
	fail
;
	!.	
% ----------------------------------------------------------------------------------------------
% dbaCountRealm                                                             
% ----------------------------------------------------------------------------------------------
class predicates
dbaCR : (chainDB,chainDB::bt_selector,string,count) determ.
clauses

dbaCountRealm(KB,Realm,N) :-
	Ix=dbaBase::getIx('C',"REALM"),
	dbaBase::database(KB,_,Id,_),
	sx::concat([Realm,"|",Id],Key),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Key,_), 
	Cnt=count::newCounter(0),
	programControl::repeat,
	   dbaCR(DB,Ix,Key,Cnt),
	N=Cnt:get(),
	!
;
	N=0,
	!.
		   
dbaCR(DB,Ix,Realm,Cnt) :-  
	DB:key_current(Ix,Key,_),
	Key=Realm,
	Cnt:inc(_),
	DB:key_next(Ix,_),
	!,fail
;
	!.	
end implement dbaUtil
