/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement dep
open core,termDomain,globalDomains

constants
className = "com/rudstrand/dep/dep".
classVersion = "1.0".

class facts
dpGraph : dirgraph := erroneous.
drGraph : dirgraph := erroneous.
smGraph : dirgraph := erroneous.
dpChanged : integer :=0. % set if the graph has been regenerated during the session
drChanged : integer :=0. % set if the graph has been regenerated during the session
smChanged : integer :=0. % set if the graph has been regenerated during the session

massSentenceUpdate : integer := 0.                     % mass sentence updating on-going (0/1)
changedST : (integer Action,string ST) nondeterm. % action: 1=type-2 sentence inserted, 3=type-2 sentence deleted (during mass-update)
checkIntegrity : integer := 1. % -1/1 check ref. integrity upon KB change
updateSmGraph : integer := 1. % -1/1 update SM graph

class predicates
kNode : (string Chain,string Entity,integer NodeType) determ (i,i,o).

smGraphInsertChainQ2 : (sterm).
smGraphInsertChainQ3 : (sterm).
smGraphInsertChainQ4 : (sterm).
smGraphInsertChainY1 : (sterm).
smGraphInsertAnno : (sterm).
smGraphInsertInpFac : (sterm).
smGraphInsertKbQuery : (sterm).
smGraphInsertKbQueryCluster : (sterm).

chkGetAllDepsX : (sterml,sterml) determ (i,o).
getAllDeps : (string,integer,string) -> termSet determ. 
insertFuncDecs : (string_list) determ.
insertIR_Decs : (string_list) determ.
insPars : (sterm,integer,string) determ.	
insertAll : (string).
isCR : (string) determ (i).
nodeToNL : (integer,sterm,string) determ (i,i,o).
nodeToNLs : (integer,sterml,string_list) determ (i,i,o).
depSMM_Exists : (sterm,sterm) determ (i,i).
depDepsOfType : (dirgraph,integer_list,sterml,sterml) determ (i,i,i,o).
depDetsOfType : (dirgraph,integer_list,sterml,sterml) determ (i,i,i,o).
depDepsOfTypeRec : (dirgraph,integer_list,sterml,sterml) determ (i,i,i,o).
depListStrToTerm : (string_list,sterml) determ (i,o).
chkGetSTs : (dirgraph,sterml,sterml) determ (i,i,o).
insertIR_Deps : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depDERF : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depInsDeterminants : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depExists : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depFuncBody : (string,integer,string,integer,sterm,integer) procedure (i,i,i,i,i,o).
depFuncCall : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depPred : (sterm,integer,string) determ (i,o,o).
depPredCond : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depProc : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depSSS : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depVarCond : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depValue : (string,integer,string,integer,sterm,integer) procedure (i,i,i,i,i,o).
smGraphInsertCR : (string) determ (i).
smGraphInsertFU : (string) determ (i).
smGraphInsertDEs : (string_list,sterm) determ.
smGraphInsertET : (string) determ.
smGraphInsertPT : (string) determ.

deleteSM_Deps : (sterm) determ.
deleteSM_Dets : (sterm) determ.

addChangedST : (integer,string).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% set/getIntegrityCheck
% ----------------------------------------------------------------------------------------------
setIntegrityCheck(Check) :- checkIntegrity:=Check,!.
getIntegrityCheck()=Check :- Check=checkIntegrity,!.
% ----------------------------------------------------------------------------------------------
% setSmGraphUpdate
% ----------------------------------------------------------------------------------------------
setSmGraphUpdate(V) :- updateSmGraph:=V,!.
% ----------------------------------------------------------------------------------------------
% drawDeps
% ----------------------------------------------------------------------------------------------
drawDeps(Title,Note,"p3_pDP",GraphType,ST,SR,CR,FU,XP3F,Node,Depth) :- 
    dirgraph::copyGraph(dpGraph,G),
    draw::drawGraph(Title,Note,G,GraphType,ST,SR,CR,FU,XP3F,Node,Depth),
    !.     
drawDeps(Title,Note,"p3_pDR",GraphType,ST,SR,CR,FU,XP3F,Node,Depth) :- 
    dirgraph::copyGraph(drGraph,G),
    draw::drawGraph(Title,Note,G,GraphType,ST,SR,CR,FU,XP3F,Node,Depth),
    !.     
% ----------------------------------------------------------------------------------------------
% graphsAreChanged
% ----------------------------------------------------------------------------------------------
graphsAreChanged() :-
    dpChanged=1,
    !
;    
    drChanged=1,
    !
;    
    smChanged=1,
    !.
% ----------------------------------------------------------------------------------------------
% copy/restore
% ----------------------------------------------------------------------------------------------
copy("p3_pDP",C) :- dirgraph::copyGraph(dpGraph,C),!.
copy("p3_pDR",C) :- dirgraph::copyGraph(drGraph,C),!.
copy("p3_pSM",C) :- dirgraph::copyGraph(smGraph,C),!.

restore("p3_pDP",C) :- dirgraph::copyGraph(C,G),dpGraph:=G,!.
restore("p3_pDR",C) :- dirgraph::copyGraph(C,G),drGraph:=G,!.
restore("p3_pSM",C) :- dirgraph::copyGraph(C,G),smGraph:=G,!.
% ----------------------------------------------------------------------------------------------
% initGraphs
% ----------------------------------------------------------------------------------------------
initGraphs() :-
    dpGraph:=dirgraph::new(1),
    drGraph:=dirgraph::new(1),
    smGraph:=dirgraph::new(1),
    dpChanged:=-1,
    drChanged:=-1,
    smChanged:=-1,
    !.
% ----------------------------------------------------------------------------------------------
% loadGraphs
% ----------------------------------------------------------------------------------------------
class predicates
loadGraph : (string) determ.
loadGraphX : (string,dirgraph).
clauses

loadGraphs() :-
    loadGraph("p3_pDP"),
    loadGraph("p3_pDR"),
    loadGraph("p3_pSM"),
    !.
    
loadGraph(Graph) :-
    Graph="p3_pDP",
    dpGraph:=dirgraph::new(1),
    loadGraphX(Graph,dpGraph),
    dpChanged:=-1,
    !
;    
    Graph="p3_pDR",
    drGraph:=dirgraph::new(1),
    loadGraphX(Graph,drGraph),
    drChanged:=-1,
    !
;    
    Graph="p3_pSM",
    smGraph:=dirgraph::new(1),
    loadGraphX(Graph,smGraph),
    smChanged:=-1,
    !.  
    
loadGraphX(Graph,DG) :-
	Graph="p3_pSM",
	dbBase::chainSentences(Graph,LTerm),
	LTerm=cmp(_,[cmp(_,[N1,S1]),cmp(_,[N2,S2])]),
	DG:add(N1,S1,N2,S2,int(0)),
	fail
;
	Graph<>"p3_pSM",
	dbBase::chainSentences(Graph,LTerm),
	LTerm=cmp(_,[cmp(_,[N1,S1]),cmp(_,[N2,S2]),Status]),
	DG:add(N1,S1,N2,S2,Status),
	fail
;
    !.
% ----------------------------------------------------------------------------------------------
% kNode
% ----------------------------------------------------------------------------------------------
kNode("q","p",1) :- !.
kNode("q","a",13) :- !.
kNode("q","e",14) :- !.
kNode("q","h",16) :- !.
kNode("q","f",17) :- !.
kNode("q","g",17) :- !.
kNode("y","p3_eTabLay",27) :- !.
kNode("y","p3_eKbq",28) :- !.
kNode("y","p3_eKbq",29) :- !.
kNode("y","p3_eKbm",30) :- !.
kNode("y","p3_eKbp",31) :- !.
kNode("p3_pIRinp","p3_eQueSer",25) :- !.
kNode("p3_pIRinp","p3_eQueClu",26) :- !.
kNode("p3_pIntd","p3_eAnno",0) :- !.
% ----------------------------------------------------------------------------------------------
% saveGraphs
% ----------------------------------------------------------------------------------------------
class predicates
saveGraphSM : ().
saveGraph : (string,dirgraph).
clauses

saveGraphs() :-
    saveGraph("p3_pDP",dpGraph),
    saveGraph("p3_pDR",drGraph),
    saveGraphSM,
    dpChanged:=-1,
    drChanged:=-1,
    smChanged:=-1,
    !.
    
saveGraphSM() :-  
    dbBase::chainDeleteOutright("p3_pSM"),
	smGraph:get_nd(N1,S1,N2,S2,_),
        LTerm=cmp("p3_pSM",[cmp("n",[N1,S1]),cmp("n",[N2,S2])]),
        dbBase::recordInsert(2,"$GRAPH",0,0,LTerm,atom("true")),
        fail
;
    !.
    
saveGraph(Graph,DG) :-    
    dbBase::chainDeleteOutright(Graph),
    DG:cascadeStatusForward(int(0)),
	DG:get_nd(N1,S1,N2,S2,Status),
        LTerm=cmp(Graph,[cmp("n",[N1,S1]),cmp("n",[N2,S2]),Status]),
        dbBase::recordInsert(2,"$GRAPH",0,0,LTerm,atom("true")),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% chainGraphSetChanged/IsChanged
% ----------------------------------------------------------------------------------------------
chainGraphSetChanged("p3_pDP") :- dpChanged:=1,!.
chainGraphSetChanged("p3_pDR") :- drChanged:=1,!.
chainGraphSetChanged("p3_pSM") :- smChanged:=1,!.
chainGraphIsChanged("p3_pDP") :- dpChanged=1,!.
chainGraphIsChanged("p3_pDR") :- drChanged=1,!.
chainGraphIsChanged("p3_pSM") :- smChanged=1,!.
% ----------------------------------------------------------------------------------------------
% changeType
% ----------------------------------------------------------------------------------------------
changeType("p3_pDP",From,To) :- dpGraph:changeType(int(From),int(To)),!.
changeType("p3_pDR",From,To) :- drGraph:changeType(int(From),int(To)),!.
changeType("p3_pSM",From,To) :- smGraph:changeType(int(From),int(To)),!.
% ----------------------------------------------------------------------------------------------
% deleteAll/WhereDep
% ----------------------------------------------------------------------------------------------
deleteAll("p3_pDP") :- dpGraph:deleteAll.
deleteAll("p3_pDR") :- drGraph:deleteAll.
deleteAll("p3_pSM") :- smGraph:deleteAll.

deleteNode("p3_pDP",cmp(_,[T,N])) :- dpGraph:deleteNode(T,N),!.
deleteNode("p3_pDR",cmp(_,[T,N])) :- drGraph:deleteNode(T,N),!.
deleteNode("p3_pSM",cmp(_,[T,N])) :- smGraph:deleteNode(T,N),!.

deleteWhereDep("p3_pDP",cmp(_,[T,N])) :- dpGraph:deleteAllWhereSucc(T,N),!.
deleteWhereDep("p3_pDR",cmp(_,[T,N])) :- drGraph:deleteAllWhereSucc(T,N),!.
deleteWhereDep("p3_pSM",cmp(_,[T,N])) :- smGraph:deleteAllWhereSucc(T,N),!.
% ----------------------------------------------------------------------------------------------
% add
% ----------------------------------------------------------------------------------------------
add("p3_pDP",T1,N1,T2,N2,S) :- dpGraph:add(T1,N1,T2,N2,S),!.
add("p3_pDR",T1,N1,T2,N2,S) :- drGraph:add(T1,N1,T2,N2,S),!.
add("p3_pSM",T1,N1,T2,N2,S) :- smGraph:add(T1,N1,T2,N2,S),!.
% ----------------------------------------------------------------------------------------------
% get_nd
% ----------------------------------------------------------------------------------------------
get_nd("p3_pDP",T1,N1,T2,N2) :- dpGraph:get_nd(T1,N1,T2,N2).
get_nd("p3_pDR",T1,N1,T2,N2) :- drGraph:get_nd(T1,N1,T2,N2).
get_nd("p3_pSM",T1,N1,T2,N2) :- smGraph:get_nd(T1,N1,T2,N2).

get_nd("p3_pDP",T1,N1,T2,N2,S) :- dpGraph:get_nd(T1,N1,T2,N2,S).
get_nd("p3_pDR",T1,N1,T2,N2,S) :- drGraph:get_nd(T1,N1,T2,N2,S).
get_nd("p3_pSM",T1,N1,T2,N2,S) :- smGraph:get_nd(T1,N1,T2,N2,S).
% ----------------------------------------------------------------------------------------------
% subMergeTypes
% ----------------------------------------------------------------------------------------------
subMergeTypes(_,[]) :-
       !.
subMergeTypes("p3_pDP",[H|T]) :- dpGraph:subMergeTypeP(int(H)), subMergeTypes("p3_pDP",T),!.
subMergeTypes("p3_pDR",[H|T]) :- drGraph:subMergeTypeP(int(H)), subMergeTypes("p3_pDR",T),!.
subMergeTypes("p3_pSM",[H|T]) :- smGraph:subMergeTypeP(int(H)), subMergeTypes("p3_pSM",T),!.
% ----------------------------------------------------------------------------------------------
% endNodes
% ----------------------------------------------------------------------------------------------
endNodes("p3_pDP",L) :- Set=dpGraph:endNodes(), L=Set:get(),!.
endNodes("p3_pDR",L) :- Set=drGraph:endNodes(), L=Set:get(),!.
endNodes("p3_pSM",L) :- Set=smGraph:endNodes(), L=Set:get(),!.
% ----------------------------------------------------------------------------------------------
% hasSTofSG
% ----------------------------------------------------------------------------------------------
hasSTofSG(SG) :-
    dpGraph:get_nd(T1,N1,_,_,_),
        set::member(T1,[int(2),int(5)]),
        N1=str(N),
        sx::sxUnpack(N,'|',L),
        L=[P|_],
        SG=P,
        !
;        
    dpGraph:get_nd(_,_,T1,N1,_),
        set::member(T1,[int(2),int(5)]),
        N1=str(N),
        sx::sxUnpack(N,'|',L),
        L=[P|_],
        SG=P,
        !.
% ----------------------------------------------------------------------------------------------
% hasET
% ----------------------------------------------------------------------------------------------
hasET(ET) :-
    dpGraph:get_nd(T1,N1,_,_,_),
        set::member(T1,[int(2),int(5)]),
        N1=str(N),
        sx::sxUnpack(N,'|',L),
        set::member(ET,L),
        !
;        
    dpGraph:get_nd(_,_,T1,N1,_),
        set::member(T1,[int(2),int(5)]),
        N1=str(N),
        sx::sxUnpack(N,'|',L),
        set::member(ET,L),
        !.
% ----------------------------------------------------------------------------------------------
% hasNode
% ----------------------------------------------------------------------------------------------
hasNode(Graph,15,Name) :- 
    hasNode(Graph,16,Name), % IR defs are not in graph, but IR defs are
    !.
hasNode("p3_pDP",Type,Name) :-
    dpGraph:get_nd(T1,N1,_,_,_),
        T1=int(Type),
        N1=str(Name),
        !
;             
    dpGraph:get_nd(_,_,T1,N1,_),
        T1=int(Type),
        N1=str(Name),
        !.
hasNode("p3_pDR",Type,Name) :-
    drGraph:get_nd(T1,N1,_,_,_),
        T1=int(Type),
        N1=str(Name),
        !
;             
    drGraph:get_nd(_,_,T1,N1,_),
        T1=int(Type),
        N1=str(Name),
        !.
hasNode("p3_pSM",Type,Name) :-
    smGraph:get_nd(T1,N1,_,_),
        T1=int(Type),
        N1=str(Name),
        !
;             
    smGraph:get_nd(_,_,T1,N1),
        T1=int(Type),
        N1=str(Name),
        !.
% ----------------------------------------------------------------------------------------------
% isDeterminant/Dependent
% ----------------------------------------------------------------------------------------------
isDeterminant("p3_pDP",Type,Name) :-
    dpGraph:get_nd(T1,N1,_,_,_),
        T1=int(Type),
        N1=str(Name),
        !.      
isDeterminant("p3_pDR",Type,Name) :-
    drGraph:get_nd(T1,N1,_,_,_),
        T1=int(Type),
        N1=str(Name),
        !.      
isDeterminant("p3_pSM",Type,Name) :-
    smGraph:get_nd(T1,N1,_,_),
        T1=int(Type),
        N1=str(Name),
        !.      

isDependent("p3_pDP",Type,Name) :-
    dpGraph:get_nd(_,_,T1,N1,_),
        T1=int(Type),
        N1=str(Name),
        !.      
isDependent("p3_pDR",Type,Name) :-
    drGraph:get_nd(_,_,T1,N1,_),
        T1=int(Type),
        N1=str(Name),
        !.      
isDependent("p3_pSM",Type,Name) :-
    smGraph:get_nd(_,_,T1,N1,_),
        T1=int(Type),
        N1=str(Name),
        !.      
% ----------------------------------------------------------------------------------------------
% depSMM_DelDerived
% ----------------------------------------------------------------------------------------------
% deletes all dependencies involving derived facts
%
depSMM_DelDerived() :- 
    smGraph:deleteNodeType(int(11)),
    !.
% ----------------------------------------------------------------------------------------------
% depSMM_Regen
% ----------------------------------------------------------------------------------------------
class predicates
insertSts : (integer,string_list) determ.
clauses

% regenerates the entire sentence model graph based on the facts (and sentence model)
%
depSMM_Regen() :-
	vpi::cursorSetWait(),
            datetime::tsNow(T1),
            dlgIF::msgS("Regenerates sentence model graph..."),
    smGraph:deleteAll,		
    smChanged:=1,	
            dlgIF::msgStatus("Regenerates sentence model graph (identifying sentence types)..."),
    L1=dbBase::stBasetypes(1),
    insertSts(11,L1),
    L2=dbBase::stBasetypes(2),
    insertSts(12,L2),
            dlgIF::msgStatus("Regenerates sentence model graph (data element dependents)..."),
	insertAll("q"),
            dlgIF::msgStatus("Regenerates sentence model graph (table layouts and KB query/menus/parameters)..."),
	insertAll("y"),
            dlgIF::msgStatus("Regenerates sentence model graph (input facilities)..."),
	insertAll("p3_pIRinp"), % input facilities
            dlgIF::msgStatus("Regenerates sentence model graph (annotations)..."),
	insertAll("p3_pIntd"), % annotations
            datetime::tsNow(T2),
            datetime::tsDiffSecs(T1,T2,Diff),
            SDiff=toString(Diff),
            Mess=string::concatList(["Regenerated sentence model graph during ",SDiff," secs!"]),
            dlgIF::msgS(Mess),
            dlgIF::msgStatus(""),
	TW=guix::getWinOfClas(1),
	vpi::cursorSet(TW,vpiDomains::cursor_Arrow),
	!.

insertSts(_,[]) :-
    !.
insertSts(Type,[H|T]) :-
    smGraphInsert(Type,H),
    insertSts(Type,T),
    !.
% ----------------------------------------------------------------------------------------------
% insertAll                                         
% ----------------------------------------------------------------------------------------------
insertAll("q") :-
    dbBase::chainSentences("q",LTerm),
        globalVars::get_i("quit_flag")=0,
        smGraphInsertChainQ2(LTerm),
        fail
;
    dbBase::chainSentences("q",LTerm),
        globalVars::get_i("quit_flag")=0,
        smGraphInsertChainQ3(LTerm),
        fail
;
    dbBase::chainSentences("q",LTerm),
        globalVars::get_i("quit_flag")=0,
        smGraphInsertChainQ4(LTerm),
        fail
;
    !.
insertAll("y") :-
    dbBase::chainSentences("y",LTerm),
        globalVars::get_i("quit_flag")=0,
        smGraphInsertChainY1(LTerm),
        fail
;
    dbBase::chainSentences("y",LTerm),
        globalVars::get_i("quit_flag")=0,
        smGraphInsertKbQuery(LTerm),
        fail
;
    dbBase::chainSentences("y",LTerm),
        globalVars::get_i("quit_flag")=0,
        smGraphInsertKbQueryCluster(LTerm),
        fail
;
    !.
insertAll(Chain) :-
    dbBase::chainSentences(Chain,LTerm),
        globalVars::get_i("quit_flag")=0,
        smGraphInsertInpFac(LTerm),
        smGraphInsertAnno(LTerm),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% smGraphUpdate                                                            
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	1=insert; 3=delete
% 2	in	record type
% 3	in	left side term (fact)
%
smGraphUpdate(_,RecType,_) :-
    updateSmGraph=-1,
    !
;    
	dbBase::recordType(RecType,_,_,Susp),
	Susp=1,
	!.
%
% insert/sentence
%	
smGraphUpdate(1,2,LTerm) :-
	LTerm=cmp("q",_),
	% only one of the predicates is relevant for a given LTerm
    smGraphInsertChainQ2(LTerm),
    smGraphInsertChainQ3(LTerm),
    smGraphInsertChainQ4(LTerm),
    !
;    
	LTerm=cmp("y",_),
    smGraphInsertChainY1(LTerm),
    smGraphInsertKbQueryCluster(LTerm),
    smGraphInsertKbQuery(LTerm),
    !
;
	LTerm=cmp("p3_pIRinp",_),
    smGraphInsertInpFac(LTerm),
    !
;
	LTerm=cmp("p3_pIntd",_),
    smGraphInsertAnno(LTerm),
    !
;    
	% insert other inputted sentence
	massSentenceUpdate=1,
	dbBase::termSt(LTerm,ST),
	addChangedST(1,ST),
	!
;	
	% insert other inputted sentence
	massSentenceUpdate=0,
	dbBase::termSt(LTerm,ST),
	smGraphInsert(12,ST),
	!.
%
% delete/sentence
%	
smGraphUpdate(3,2,LTerm) :-
	LTerm=cmp("q",[cmp(E,[atom(A)|_])|_]),
	set::member(E,["a","e","p","f","g","h"]),
	kNode("q",E,N),
	deleteSM_Deps(cmp("n",[int(N),str(A)])),
	deleteSM_Dets(cmp("n",[int(N),str(A)])),
	!
;	
	LTerm=cmp("p3_pIRinp",[cmp("p3_eQueSer",[_,str(R)|_]),_]),
	kNode("p3_pIRinp","p3_eQueSer",N),
	deleteSM_Deps(cmp("n",[int(N),str(R)])),
	deleteSM_Dets(cmp("n",[int(N),str(R)])),
	!
;	
	LTerm=cmp("p3_pIRinp",[cmp("p3_eQueClu",[str(Pr)|_]),atom(R)]),
	kNode("p3_pIRinp","p3_eQueClu",N),
	P=string::concat(R,"|",Pr),
	deleteSM_Deps(cmp("n",[int(N),str(P)])),
	deleteSM_Dets(cmp("n",[int(N),str(P)])),
	!
;	
	LTerm=cmp("p3_pIntd",[cmp("p3_eAnno",[_,str(Name)|_])]),
	kNode("p3_pIntd","p3_eAnno",N),
	deleteSM_Deps(cmp("n",[int(N),str(Name)])),
	deleteSM_Dets(cmp("n",[int(N),str(Name)])),
	!
;	
	LTerm=cmp("y",[cmp("p3_eTabLay",[str(ST)|_])]),
	kNode("y","p3_eTabLay",N),
	deleteSM_Deps(cmp("n",[int(N),str(ST)])),
	deleteSM_Dets(cmp("n",[int(N),str(ST)])),
	!
;	
	LTerm=cmp("y",[cmp("p3_eKbq",[int(Id)|_])]),
	SId=toString(Id),
	% is actually either 28 or 29, but deletes both instead of checking (no side-effect)
	deleteSM_Deps(cmp("n",[int(28),str(SId)])),
	deleteSM_Dets(cmp("n",[int(28),str(SId)])),
	deleteSM_Deps(cmp("n",[int(29),str(SId)])),
	deleteSM_Dets(cmp("n",[int(29),str(SId)])),
	!
;	
	LTerm=cmp("y",[cmp(E,[int(Id)|_])]),
	set::member(E,["p3_eKbp","p3_eKbm"]),
	kNode("y",E,N),
	SId=toString(Id),
	deleteSM_Deps(cmp("n",[int(N),str(SId)])),
	deleteSM_Dets(cmp("n",[int(N),str(SId)])),
	!
;
	% delete inputted fact
	massSentenceUpdate=1,
	dbBase::termSt(LTerm,ST),
	addChangedST(3,ST),
	!
;
	% delete inputted fact
	massSentenceUpdate=0,
	dbBase::termSt(LTerm,ST),
	not(dbBase::stExists(2,ST)),
	deleteSM_Deps(cmp("n",[int(12),str(ST)])),
	deleteSM_Dets(cmp("n",[int(12),str(ST)])),
	!
;
    !.
%
% insert/delete non-sentence
%    
smGraphUpdate(Action,RecType,LTerm) :-
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	LTerm=cmp(Func,_),
	dbBase::chainCount(Func)>0,
    % func def exists
    dbBase::smDefF(Func,_,_,_,_,_,_),
    % func decl exists
	dep::insertDependency("p3_pSM",cmp("n",[int(17),str(Func)]),cmp("n",[int(19),str(Func)]),0,_),
    !
;    
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	LTerm=cmp(Func,_),
	% func dec or def (or both missing)
	dep::deleteDependency("p3_pSM",cmp("n",[int(17),str(Func)]),cmp("n",[int(19),str(Func)])),
	!
;
    Action=1,
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=6,
	LTerm=cmp(_,[atom(Rule)|_]),
    dbBase::smClassDefF(Rule,"h",_,_,_,_,_),
	dep::insertDependency("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(15),str(Rule)]),0,_),
    !
;    
    Action=1,
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=6,
	LTerm=cmp(_,[atom(Rule)|_]),
	dep::deleteDependency("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(15),str(Rule)])),
    !
;
    Action=3,
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=6,
	LTerm=cmp(_,[atom(Rule)|_]),
	dep::deleteDependency("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(15),str(Rule)])),
    !
;
    S=toString(RecType),
    Mess=string::concat("Unexpected record type in SM-graph update: ",S),
    dlgIF::dlgNote(3,Mess),!,fail.
% ----------------------------------------------------------------------------------------------
% smGraphInsertChainQ2/3/4
% ----------------------------------------------------------------------------------------------
% updates SM-graph on the basis of an insert to the q-chain
%
smGraphInsertChainQ2(LTerm) :-
	LTerm=cmp(_,[cmp("e",[atom(E)|_])|_]),
	% insert ET
	smGraphInsertET(E),
	!
;
	LTerm=cmp(_,[cmp("p",[atom(P)|_])|_]),
	% insert PT
	smGraphInsertPT(P),
	!
;
    !.

smGraphInsertChainQ3(LTerm) :-
	LTerm=cmp(_,[cmp("f",[atom(F)|_]),_]),
	% insert function decl
	smGraphInsertFU(F),
	!
;	
	LTerm=cmp(_,[cmp("g",[atom(F)|_]),_]),
	% insert function decl
	smGraphInsertFU(F),
	!
;
    !.

smGraphInsertChainQ4(LTerm) :-
	LTerm=cmp(_,[cmp("h",[atom(F)|_]),_]),
	% insert ir decl
	smGraphInsertCR(F),
	!
;	
	!.
% ----------------------------------------------------------------------------------------------
% smGraphInsertInpFac
% ----------------------------------------------------------------------------------------------
class predicates
insIR_Rules : (sterm,integer,string) determ.	
clauses

smGraphInsertInpFac(Body) :-
    Body=cmp("p3_pIRinp",[cmp("p3_eQueSer",[_,atom(Rule),_,atom("p3_iPR"),str(Q),Pars]),atom(LocRule)]),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(LocRule)]),cmp("n",[int(25),str(Rule)]),0),
	insPars(Pars,25,LocRule),
	transpr::proPR_Term(Q,Q2),
	Q2=cmp("p3_iDdrf",[_,_,str(Qa)]),
	Qt=termDomain::fromTermStringIndirect(Qa),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(25),str(Rule)]),0),
    !
;    
    Body=cmp("p3_pIRinp",[cmp("p3_eQueSer",[_,atom(Rule),_,atom("p3_iPR"),str(Q),Pars]),atom(LocRule)]),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(LocRule)]),cmp("n",[int(25),str(Rule)]),0),
	insPars(Pars,25,LocRule),
	transpr::proPR_Term(Q,Q2),
	Q2=cmp("p3_iDdrf",[_,_,str(Qa)]),
	Qt=termDomain::fromTermStringIndirect(Qa),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1,E2],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E2)]),cmp("n",[int(25),str(Rule)]),0),
    !
;   
    Body=cmp("p3_pIRinp",[cmp("p3_eQueSer",[_,atom(Rule),_,atom("p3_iNL"),str(Q),Pars]),atom(LocRule)]),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(LocRule)]),cmp("n",[int(25),str(Rule)]),0),
	insPars(Pars,25,LocRule),
    irBase::wcrTransQuery(Q,Q2,_),
	Q2=cmp("p3_iDdrf",[_,_,str(Qa)]),
	Qt=termDomain::fromTermStringIndirect(Qa),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(25),str(Rule)]),0),
    !
;    
    Body=cmp("p3_pIRinp",[cmp("p3_eQueSer",[_,atom(Rule),_,atom("p3_iNL"),str(Q),Pars]),atom(LocRule)]),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(LocRule)]),cmp("n",[int(25),str(Rule)]),0),
	insPars(Pars,25,LocRule),
    irBase::wcrTransQuery(Q,Q2,_),
	Q2=cmp("p3_iDdrf",[_,_,str(Qa)]),
	Qt=termDomain::fromTermStringIndirect(Qa),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1,E2],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(25),str(Rule)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E2)]),cmp("n",[int(25),str(Rule)]),0),
    !
;    
    Body=cmp("p3_pIRinp",[cmp("p3_eQueClu",[str(Prompt),Calls,Pars]),atom(Rule)]),
	P=string::concat(Rule,"|",Prompt),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(26),str(P)]),0),
	insPars(Pars,26,P),
	insIR_Rules(Calls,26,P),
    !
;    
    !.
    
insIR_Rules(nill,_,_) :-
    !.
insIR_Rules(list(atom(R),L),N,S) :-    
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(16),str(R)]),cmp("n",[int(N),str(S)]),0),
	insIR_Rules(L,N,S),
	!.
% ----------------------------------------------------------------------------------------------
% insPars
% ----------------------------------------------------------------------------------------------
insPars(nill,_,_) :-
    !.
insPars(list(atom(P),L),N,S) :-    
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(13),str(P)]),cmp("n",[int(N),str(S)]),0),
	insPars(L,N,S),
	!.
% ----------------------------------------------------------------------------------------------
% smGraphInsertAnno
% ----------------------------------------------------------------------------------------------
smGraphInsertAnno(Body) :-
	Body=cmp("p3_pIntd",[cmp("p3_eAnno",[atom(Type)|_])]),
	nodIF_No(Type,N),
	set::member(N,[13,21,22,24]),
	!
;
	Body=cmp("p3_pIntd",[cmp("p3_eAnno",[atom("p3_iFunc"),str(Name)|_])]),
	pie::builtInFunc(Name),
	!
;
	Body=cmp("p3_pIntd",[cmp("p3_eAnno",[atom(Type),str(Name)|_])]),
	nodIF_No(Type,No),
	No=3,
	insertDependencyCheckDet("p3_pDP",cmp("n",[int(No),str(Name)]),cmp("n",[int(0),str(Name)]),0),
	!
;
	Body=cmp("p3_pIntd",[cmp("p3_eAnno",[atom(Type),str(Name)|_])]),
	nodIF_No(Type,No),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(No),str(Name)]),cmp("n",[int(0),str(Name)]),0),
	!
;
    !.
% ----------------------------------------------------------------------------------------------
% smGraphInsertChainY1/KB_Query/KB_QueryCluster
% ----------------------------------------------------------------------------------------------
class predicates
insQueryPars : (string_list,string) determ.
insClusteredQuery : (string_list,string) determ.
clauses

% updates SM-graph on the basis of an insert to the y
%
smGraphInsertChainY1(LTerm) :-
	LTerm=cmp(_,[cmp("p3_eKbm",[int(Id),_,_,int(Menu)])]),
	SMenu=toString(Menu),
	SId=toString(Id),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(30),str(SMenu)]),cmp("n",[int(30),str(SId)]),0),
	!
;	
	LTerm=cmp(_,[cmp("p3_eTabLay",[str(ST)|_])]),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(27),str(ST)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(27),str(ST)]),0),
	!
;	
	LTerm=cmp(_,[cmp("p3_eTabLay",[str(ST)|_])]),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1,E2],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(27),str(ST)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(27),str(ST)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E2)]),cmp("n",[int(27),str(ST)]),0),
	!
;	
    !.

smGraphInsertKbQueryCluster(LTerm) :-
	LTerm=cmp(_,[cmp("p3_eKbq",[int(Id),_,_,_,str(QueryList),int(Menu)])]),
	sx::sxUnpack(QueryList,';',Queries),
	proc::checkClusteredQueries(Queries,_,_),
	set::size(Queries,N),
	N>0,
	SId=toString(Id),
	insClusteredQuery(Queries,SId),
	SMenu=toString(Menu),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(30),str(SMenu)]),cmp("n",[int(29),str(SId)]),0),
	!
;	
    !.

smGraphInsertKbQuery(LTerm) :-
	LTerm=cmp(_,[cmp("p3_eKbq",[_,_,_,_,str(QueryList),_])]),
	sx::sxUnpack(QueryList,';',Queries),
	proc::checkClusteredQueries(Queries,_,_),
	% query clusters are processed separately
	!
;	
	LTerm=cmp(_,[cmp("p3_eKbq",[int(Id),_,_,atom("p3_iPR"),str(Query),int(Menu)])]),
	proc::getPars(Query,"_X",Q,Pars),
	SId=toString(Id),
	SMenu=toString(Menu),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(30),str(SMenu)]),cmp("n",[int(28),str(SId)]),0),
	insQueryPars(Pars,SId),
	transpr::proPR_Term(Q,Qt),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(28),str(SId)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(28),str(SId)]),0),
	!
;	
	LTerm=cmp(_,[cmp("p3_eKbq",[int(Id),_,_,atom("p3_iPR"),str(Query),int(Menu)])]),
	proc::getPars(Query,"_X",Q,Pars),
	SId=toString(Id),
	SMenu=toString(Menu),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(30),str(SMenu)]),cmp("n",[int(28),str(SId)]),0),
	insQueryPars(Pars,SId),
	transpr::proPR_Term(Q,Qt),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1,E2],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(28),str(SId)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(28),str(SId)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E2)]),cmp("n",[int(28),str(SId)]),0),
	!
;	
	LTerm=cmp(_,[cmp("p3_eKbq",[int(Id),_,_,atom("p3_iNL"),str(Query),int(Menu)])]),
	proc::getPars(Query,"_X",Q,Pars),
	SId=toString(Id),
	SMenu=toString(Menu),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(30),str(SMenu)]),cmp("n",[int(28),str(SId)]),0),
	insQueryPars(Pars,SId),
	sx::sxGiveTail(Q,"?",Q2),
	transnl::transQuery(Q2,Qt,_),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(28),str(SId)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(28),str(SId)]),0),
	!
;	
	LTerm=cmp(_,[cmp("p3_eKbq",[int(Id),_,_,atom("p3_iNL"),str(Query),int(Menu)])]),
	proc::getPars(Query,"_X",Q,Pars),
	SId=toString(Id),
	SMenu=toString(Menu),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(30),str(SMenu)]),cmp("n",[int(28),str(SId)]),0),
	insQueryPars(Pars,SId),
	sx::sxGiveTail(Q,"?",Q2),
	transnl::transQuery(Q2,Qt,_),
	dbBase::queryStF(Qt,ST),
	sx::sxUnpack(ST,'|',L),
	L=[P,E1,E2],
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(28),str(SId)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(28),str(SId)]),0),
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(14),str(E2)]),cmp("n",[int(28),str(SId)]),0),
	!
;
    !.

insQueryPars([],_) :-
       !.
insQueryPars([P|Ps],SId) :-       
    dbExtension::kbQueryParF(P,_,_,_,_) ,
	insertDependency("p3_pSM",cmp("n",[int(31),str(P)]),cmp("n",[int(28),str(SId)]),0,_),
	insQueryPars(Ps,Sid),
	!
;	
    nodIF_No(K,28),
    dbBase::nlTry(K,NL),
	Mess=string::concatList(["KB query parameter '",P,"' referred by ",NL," ",SId," does not exist!"]),
	dlgIF::dlgNote(0,Mess),
	insQueryPars(Ps,Sid),
	!.    

insClusteredQuery([],_) :-
       !.
insClusteredQuery([P|Ps],SId) :-       
	insertDependencyCheckDet("p3_pSM",cmp("n",[int(28),str(P)]),cmp("n",[int(29),str(SId)]),0),
	insClusteredQuery(Ps,Sid),
	!.
% ----------------------------------------------------------------------------------------------
% insertFuncDecs                                        
% ----------------------------------------------------------------------------------------------
insertFuncDecs([]) :-
    !.
insertFuncDecs([Func|T]) :-    
    dbBase::chainExists(Func),
    % func decl exists
	dep::insertDependency("p3_pSM",cmp("n",[int(17),str(Func)]),cmp("n",[int(19),str(Func)]),0,_),
    insertFuncDecs(T),
	!
;
    insertFuncDecs(T),
    !.
% ----------------------------------------------------------------------------------------------
% insertIR_Decs                                        
% ----------------------------------------------------------------------------------------------
insertIR_Decs([]) :-
    !.
insertIR_Decs([Rule|T]) :-    
    dbBase::getIrDefF(Rule,_,_),
	dep::insertDependency("p3_pSM",cmp("n",[int(16),str(Rule)]),cmp("n",[int(15),str(Rule)]),0,_),
    insertIR_Decs(T),
	!
;
    insertIR_Decs(T),
    !.
% ----------------------------------------------------------------------------------------------
% depSMM_Exists                                                    
% ----------------------------------------------------------------------------------------------
depSMM_Exists(Det,Dep) :-
    Det=cmp(_,[A1,A2]),
    Dep=cmp(_,[B1,B2]),
    smGraph:has(A1,A2,B1,B2),
    !.
% ----------------------------------------------------------------------------------------------
% smGraphInsert                                                    
% ----------------------------------------------------------------------------------------------
% updates the SM graph on the basis of a fact insertion (given by the fact's ST)
% parameters:
% 1	in	11 derived fact insertion; 12 input fact insertion
% 2	in	ST
%
smGraphInsert(N,ST) :-	
	dbBase::stComps(ST,Card,P,E1,_),
	Card=1,
	insertDependency("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(N),str(ST)]),0,_),
	insertDependency("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(N),str(ST)]),0,_),
	!
;	
	dbBase::stComps(ST,Card,P,E1,E2),
	Card=2,
	insertDependency("p3_pSM",cmp("n",[int(14),str(E1)]),cmp("n",[int(N),str(ST)]),0,_),
	insertDependency("p3_pSM",cmp("n",[int(14),str(E2)]),cmp("n",[int(N),str(ST)]),0,_),
	insertDependency("p3_pSM",cmp("n",[int(1),str(P)]),cmp("n",[int(N),str(ST)]),0,_),
	!
;
	dlgIF::feil("smGraphInsert",[ST],[N],nill),!,fail.	
% ----------------------------------------------------------------------------------------------
% smGraphInsertCR
% ----------------------------------------------------------------------------------------------
smGraphInsertCR(C) :-
	dbBase::smClassDef(C,"h",_,DEs,_,_,_),
	Node=cmp("n",[int(16),str(C)]),
	smGraphInsertDEs(DEs,Node),
	!.
% ----------------------------------------------------------------------------------------------
% smGraphInsertDEs
% ----------------------------------------------------------------------------------------------
smGraphInsertDEs([],_) :- 
	!.
smGraphInsertDEs([DE|DEs],Node) :-
	insertDependency("p3_pSM",cmp("n",[int(13),str(DE)]),Node,0,_),
	smGraphInsertDEs(DEs,Node),
	!.
% ----------------------------------------------------------------------------------------------
% smGraphInsertET
% ----------------------------------------------------------------------------------------------
smGraphInsertET(E) :-
	dbBase::smClassDef(E,"e",_,DEs,_,_,_),
	Node=cmp("n",[int(14),str(E)]),
	smGraphInsertDEs(DEs,Node),!.
% ----------------------------------------------------------------------------------------------
% smGraphInsertFU
% ----------------------------------------------------------------------------------------------
smGraphInsertFU(F) :-
	dbBase::smDef(F,Clas,_,DEs,_,_,_),
	set::member(Clas,["f","g"]),
	Node=cmp("n",[int(17),str(F)]),
	smGraphInsertDEs(DEs,Node),
	!.
% ----------------------------------------------------------------------------------------------
% smGraphInsertPT
% ----------------------------------------------------------------------------------------------
smGraphInsertPT(P) :-
	dbBase::smClassDef(P,"p",_,DEs,_,_,_),
	Node=cmp("n",[int(1),str(P)]),
	smGraphInsertDEs(DEs,Node),!.
% ----------------------------------------------------------------------------------------------
% depSMM_ResetST
% ----------------------------------------------------------------------------------------------
% redetermines SMM nodes Referring to given ST
%
depSMM_ResetST(ST) :-
    smGraph:get_nd(A1,A2,B1,B2,S),
        B2=str(ST),
        B1=int(N),
        N>=11,
        N<=12,
        T=N-10,
        not(dbBase::stExists(T,ST)),
        smGraph:delete(A1,A2,B1,B2,S),
        fail
;
	!.
% ----------------------------------------------------------------------------------------------
% regenDependencyGraph
% ----------------------------------------------------------------------------------------------
class predicates
regenDP : ().
regenDPX : (dbLoc_list) determ (i).
clauses

% regenerates the entire dependency graph based on the rules in the KB
%
regenDependencyGraph() :-
	vpi::cursorSetWait(),
    datetime::tsNow(T1),
    N2=dbBase::typeCount(-3),
    N3=dbBase::typeCount(3),
    Rules=N2+N3,
    Funcs=dbBase::typeCount(4),
    CRs=dbBase::typeCount(6),
    S1=toString(Rules),
    S2=toString(Funcs),
    S3=toString(CRs),
    Mess1=string::concatList(["Regenerating dependency graph for ",S1," sentence rules, ",S3," inexact rules and ",S2," functions..."]),
    dlgIF::msgStatus(Mess1),
    dpGraph:deleteAll,
	dpChanged:=1,
	regenDP,
	dpGraph:setStatus(int(0)),
    datetime::tsNow(T2),
    datetime::tsDiffSecs(T1,T2,Diff),
    dlgIF::msgElapsed("Regenerated dependency graph during",Diff),
    dlgIF::msgStatus(""),
	TW=guix::getWinOfClas(1),
	vpi::cursorSet(TW,vpiDomains::cursor_Arrow),
    !.

regenDP() :-
    dbBase::recTypeLocs(-3,L),
    regenDPX(L),
    fail
;        
    dbBase::recTypeLocs(3,L),
    regenDPX(L),
    fail
;        
    dbBase::recTypeLocs(4,L),
    regenDPX(L),
    fail
;        
    dbBase::recTypeLocs(6,L),
    regenDPX(L),
    fail
;    
    !.

regenDPX([Loc|Locs]) :-
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,_,_,LTerm,_),
    LTerm=cmp(Chain,_),
    dbBase::chainIsExternal(Chain),
	% the determinants of external chains are not included in the dependency graph
    regenDPX(Locs),
	!
;
    dbBase::getRecord(Loc,RcT,Base,_,_,_,Realm,N,M,_,LT,RT),
    Base=3,
    Ns=toString(N),
    Ms=toString(M),
    LT=cmp(Chain,_),
	sx::sxlJoin([Chain,Realm,Ns,Ms],"|",Rule),
	ART=math::abs(RcT),
	updateDP_Graph(1,ART,LT,RT,Rule,_),
    regenDPX(Locs),
	!
;
    dbBase::getRecord(Loc,RcT,Base,_,_,_,_,_,_,_,LT,RT),
    Base<>3,
	ART=math::abs(RcT),
	updateDP_Graph(1,ART,LT,RT,"",_),
    regenDPX(Locs),
	!.
% ----------------------------------------------------------------------------------------------
% depDetsOfType/DepsOfType
% ----------------------------------------------------------------------------------------------
class predicates
depDOT : (dirgraph,integer,integer_list,sterml,lst) determ (i,i,i,i,i).
clauses

% parameters:
% 1	in	graph
% 2	in	nodetypes or [] if all types
% 3	in	nodes
% 4	out	nodes of type p#2 determining p#3
%
depDetsOfType(Graph,NodeTypes,Nodes,Dets) :-
	Arr=lst::new(),
	depDOT(Graph,1,NodeTypes,Nodes,Arr),
	L=Arr:get(),
	depListStrToTerm(L,Dets),
	!.

depDepsOfType(Graph,NodeTypes,Nodes,Deps) :-
	Arr=lst::new(),
	depDOT(Graph,2,NodeTypes,Nodes,Arr),
	L=Arr:get(),
	depListStrToTerm(L,Deps),
	!.

depDepsOfTypeRec(Graph,NodeTypes,Nodes,Dets) :-
	depDepsOfType(Graph,NodeTypes,Nodes,Dets2),
	set::conc([Nodes,Dets2],NewNodes),
	set::distinct(NewNodes,NewNodesD),
	depDepsOfType(Graph,NodeTypes,NewNodesD,Dets3),
	set::size(Dets2,N2),
	set::size(Dets3,N3),
	N3>N2,
	depDepsOfTypeRec(Graph,NodeTypes,NewNodesD,Dets),
	!
;
	depDepsOfType(Graph,NodeTypes,Nodes,Dets),
	!.

depDOT(DG,1,NodeTypes,Nodes,Arr) :-
	% find determinants
	set::size(NodeTypes,NT),
	NT>0,
	DG:get_nd(Det1,Det2,Dep1,Dep2),
        Det1=int(Tp),
        set::member(Tp,NodeTypes),
        set::member(cmp("n",[Dep1,Dep2]),Nodes),
        S=toString(cmp("n",[Det1,Det2])),
        Arr:add(S),
        fail
;
	% find determinants
	set::size(NodeTypes,NT),
	NT=0,
	DG:get_nd(Det1,Det2,Dep1,Dep2),
        set::member(cmp("n",[Dep1,Dep2]),Nodes),
        S=toString(cmp("n",[Det1,Det2])),
        Arr:add(S),
        fail
;
	!.
depDOT(DG,2,NodeTypes,Nodes,Arr) :-
	% find dependents
	set::size(NodeTypes,NT),
	NT>0,
	DG:get_nd(Det1,Det2,Dep1,Dep2),
        Dep1=int(Tp),
        set::member(Tp,NodeTypes),
        set::member(cmp("n",[Det1,Det2]),Nodes),
        S=toString(cmp("n",[Dep1,Dep2])),
        Arr:add(S),
        fail
;
	% find dependents
	set::size(NodeTypes,NT),
	NT=0,
	DG:get_nd(Det1,Det2,Dep1,Dep2),
        set::member(cmp("n",[Det1,Det2]),Nodes),
        S=toString(cmp("n",[Dep1,Dep2])),
        Arr:add(S),
        fail
;
	!.
% ----------------------------------------------------------------------------------------------
% depIsDependent
% ----------------------------------------------------------------------------------------------
% succeeds if Dep is a dependent (direct/indirect) of Node in the given graph
%
depIsDependent("p3_pDP",Node,Dep) :-
    Node=cmp(_,[T,N]),
    Succs=dpGraph:succsRec(T,N),
    Dep=cmp(_,[T2,N2]),
    Succs:member(cmp("",[T2,N2])),
    !.
depIsDependent("p3_pDR",Node,Dep) :-
    Node=cmp(_,[T,N]),
    Succs=drGraph:succsRec(T,N),
    Dep=cmp(_,[T2,N2]),
    Succs:member(cmp("",[T2,N2])),
    !.
depIsDependent("p3_pSM",Node,Dep) :-
    Node=cmp(_,[T,N]),
    Succs=smGraph:succsRec(T,N),
    Dep=cmp(_,[T2,N2]),
    Succs:member(cmp("",[T2,N2])),
    !.
% ----------------------------------------------------------------------------------------------
% depListStrToTerm                                                        
% ----------------------------------------------------------------------------------------------
depListStrToTerm([],[]) :- 
	!.
depListStrToTerm([H|T],[Ht|Tt]) :-
	Ht=toTerm(H),
	depListStrToTerm(T,Tt),
	!.	
% ----------------------------------------------------------------------------------------------
% updateDP_Graph                                                            
% ----------------------------------------------------------------------------------------------
% updates the dependency graph on the basis of an inserted or deleted knowledge node
% note:
% sentence rules: no update if rule's conclusion is in an external chain
% parameters:
% 1 in	1=insert; 3=delete
% 2	in	record type
% 3	in	conclusion
% 4	in	condition
% 5	in 	rule-key (if sentence rule)
% 6 out -1/1 if disallowed cyclicity was introduced
%
updateDP_Graph(ChangeType,RT,LTerm,_,_,1) :-
% 
% FACT
%
    RT>=1, RT<=2,
	dbBase::kbTermST(LTerm,SentType),
	sx::sxlJoin(SentType,"|",ST),
	sentenceChanged(ChangeType,ST),
	!.
% 
% SR
%
updateDP_Graph(1,RT,LTerm,RTerm,Rule,Acyclic) :-
	dbBase::recordType(RT,BType,_,_),
	BType=3,
	LTerm=cmp(Chain,_),
	dbBase::chainIsExternal(Chain),
	Acyclic=1,
	!
;
	dbBase::recordType(RT,BType,_,_),
	BType=3,
	% insert dependencies for all the rule's determinants 
	depInsDeterminants("p3_pDP",3,Rule,0,RTerm,Acyclic1),
	% insert the rule's dependent (set status 0)
	depPred(LTerm,Type1,Node1),
	insertDependency("p3_pDP",cmp("n",[int(3),str(Rule)]),cmp("n",[int(Type1),str(Node1)]),0,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
updateDP_Graph(3,RT,LTerm,_,Rule,1) :-
	dbBase::recordType(RT,BType,_,_),
	BType=3,
	LTerm=cmp(Chain,_),
	dbBase::chainIsExternal(Chain),
	!
;
	dbBase::recordType(RT,BType,_,_),
	BType=3,
	setUnderived("p3_pDP",int(3),str(Rule)),
	dpGraph:deleteNode(int(3),str(Rule)),
	!.
%	
% derived rule (these updates concern the derivation graph only)
%
updateDP_Graph(1,RT,LTerm,RTerm,Rule,Acyclic) :-
	dbBase::recordType(RT,BType,_,_),
	BType=5,
	LTerm=cmp(Chain,_),
	dbBase::chainIsExternal(Chain),
	Acyclic=1,
	!
;
	dbBase::recordType(RT,BType,_,_),
	BType=5,
	% insert dependencies for all the rule's determinants
	depInsDeterminants("p3_pDR",3,Rule,0,RTerm,Acyclic1), 
	% insert the rule's dependent (set status 0)
	depPred(LTerm,Type1,Node1),
	insertDependency("p3_pDR",cmp("n",[int(3),str(Rule)]),cmp("n",[int(Type1),str(Node1)]),0,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
updateDP_Graph(3,RT,LTerm,_,Rule,1) :-
	dbBase::recordType(RT,BType,_,_),
	BType=5,
	LTerm=cmp(Chain,_),
	dbBase::chainIsExternal(Chain),
	!
;
	dbBase::recordType(RT,BType,_,_),
	BType=5,
	drGraph:deleteNode(int(3),str(Rule)),
	!.
%	
% FUNC DEF 
%
updateDP_Graph(1,4,cmp(Func,_),Body,_,Acyclic) :-
	% insert dependencies for function def's determinants
	depFuncBody("p3_pDP",19,Func,0,Body,Acyclic),
	setStatusWhereDeterminant("p3_pDP",0,cmp("n",[int(19),str(Func)])),
	setUnderived("p3_pDP",int(19),str(Func)),
	!.
%	
% IR DEF
%
updateDP_Graph(1,6,Term,_,Rule,Acyclic) :-
	% insert dependencies for all the IR's determinants 
	insertIR_Deps("p3_pDP",15,Rule,0,Term,Acyclic),
	!.
updateDP_Graph(3,6,_,_,Rule,1) :-
	setUnderived("p3_pDP",int(15),str(Rule)),
	dpGraph:deleteAllWhereSucc(int(15),str(Rule)),
	setStatusWhereDeterminant("p3_pDP",0,cmp("n",[int(15),str(Rule)])),
	!.
% ----------------------------------------------------------------------------------------------
% depSetRuleStatusAll                                                            
% ----------------------------------------------------------------------------------------------
class predicates
depRuleStatusX : (integer,dbLoc_list) determ (i,i).
clauses

% activates/deactivates ALL rules
%
depSetRuleStatusAll(NewStatus) :-
	OldStatus=-NewStatus,	
	vpi::cursorSetWait(),
	dbBase::recTypeLocs(OldStatus,Locs),
	depRuleStatusX(NewStatus,Locs),
	!.
	
depRuleStatusX(_,[]) :-
	!.
depRuleStatusX(RT,[Loc|Locs]) :-
    dbBase::getRecord(Loc,_,_,_,_,_,_,_,_,_,LTerm,_),
    LTerm=cmp(Chain,_),
	dbBase::chainAccess(Chain)=0,
	depRuleStatusX(RT,Locs),
	!
;	
	dbBase::recordSetType(Loc,RT),
	depRuleStatusX(RT,Locs),
	!.
% ----------------------------------------------------------------------------------------------
% deleteSM_Deps
% ----------------------------------------------------------------------------------------------
% deletes all edges Node->X
%
deleteSM_Deps(Node) :-
    Node=cmp(_,[T,N]),
    smGraph:deleteAllWherePred(T,N),
	!.	
% ----------------------------------------------------------------------------------------------
% deleteSM_Deps
% ----------------------------------------------------------------------------------------------
% deletes all edges X->Node
%
deleteSM_Dets(Node) :-
    Node=cmp(_,[T,N]),
    smGraph:deleteAllWhereSucc(T,N),
    smGraph:setStatusWherePred(T,N,int(0)),
	!.	
% ----------------------------------------------------------------------------------------------
% depDERF
% ----------------------------------------------------------------------------------------------
depDERF(Matrix,Type,Node,Stat,cmp(C,[_,_,str(PredCond)]),Acyclic) :-	
	set::member(C,["p3_iNdrf","p3_iDdrf"]),
    Term=termDomain::fromTermStringIndirect(PredCond),
	depInsDeterminants(Matrix,Type,Node,Stat,Term,Acyclic),
	!.
depDERF(Matrix,_,_,_,cmp(C,B),_) :- 
	set::member(C,["p3_iNdrf","p3_iDdrf"]),
	dlgIF::feil("depDERF",[Matrix],[],cmp(C,B)),!,fail.
% ----------------------------------------------------------------------------------------------
% insertIR_Deps          
% ----------------------------------------------------------------------------------------------
class predicates
dep_CR_Calls : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
dep_CR_Call : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
depCR_X : (string,integer,string,integer,sterml,integer) determ (i,i,i,i,i,o).
depCR_Query : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
clauses

insertIR_Deps(Matrix,Type,Node,Stat,Term,Acyclic) :-
	% fuzzy set
	Term=cmp(CR,[atom(Rule),Calls|_]),
	set::member(CR,["p3_iZnot","p3_iZand","p3_iZor","p3_iZowa"]),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Calls,Acyclic),
	!
;
	% bay/and/or
	Term=cmp(CR,[atom(Rule),Calls,_,_,Con|_]),
	set::member(CR,["p3_iCbay","p3_iCand","p3_iCor","p3_iCcmb"]),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic1),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Calls,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;	
	% parameter switch
	Term=cmp("p3_iCswi",[atom(Rule),atom(_),Calls,_,_,_,Con|_]),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic1),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Calls,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;	
	% call switch
	Term=cmp("p3_iCswi",[atom(Rule),Call,Calls,_,_,_,Con|_]),
	dep_CR_Call(Matrix,Type,Rule,Stat,Call,Acyclic1),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic2),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Calls,Acyclic3),
	Acyclic=Acyclic1*Acyclic2*Acyclic3,
	!
;	
	% parameter
	Term=cmp("p3_iCpar",[atom(Rule),Call,Calls,_,_,Con|_]),
	dep_CR_Call(Matrix,Type,Rule,Stat,Call,Acyclic1),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic2),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Calls,Acyclic3),
	Acyclic=Acyclic1*Acyclic2*Acyclic3,
	!
;	
	% not
	Term=cmp("p3_iCnot",[atom(Rule),Call,_,_,Con|_]),
	dep_CR_Call(Matrix,Type,Rule,Stat,Call,Acyclic1),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;	
	% map
	Term=cmp("p3_iCmap",[atom(Rule),Call,_,_,_,_,Con|_]),
	dep_CR_Call(Matrix,Type,Rule,Stat,Call,Acyclic1),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;	
	% question
	Term=cmp("p3_iCdat",[atom(Rule),cmp("p3_iQst",_),_,_,_,_,Con|_]),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic),
	!
;	
	% query
	Term=cmp("p3_iCdat",[atom(Rule),cmp("p3_iQry",[_,Query]),_,_,_,_,Con|_]),
	depCR_Query(Matrix,Type,Rule,Stat,Query,Acyclic1),
	dep_CR_Calls(Matrix,Type,Rule,Stat,Con,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	dlgIF::feil("insertIR_Deps",[Matrix,Node],[Type,Stat],Term),!,fail.

dep_CR_Calls(_,_,_,_,nill,1) :-
	!.
dep_CR_Calls(Matrix,Type,Node,Stat,cmp(_,L),Acyclic) :-
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic),
	!.
dep_CR_Call(Matrix,Type,Node,Stat,SCall,Acyclic) :-
	depCR_X(Matrix,Type,Node,Stat,[SCall],Acyclic),
	!.

depCR_X(_,_,_,_,[],1) :-
	!.
depCR_X(Matrix,Type,Node,Stat,[H|L],Acyclic) :-
	% not, criterion, map
	H=cmp(R,[cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	% owa with weight call
	H=cmp(R,[cmp(Ri,_),cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	insertDependency(Matrix,cmp("n",[int(15),str(Ri)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic2),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic3),
	Acyclic=Acyclic1*Acyclic2*Acyclic3,
	!
;
	% parameter; owa w/out weight call
	H=cmp(R,[_,cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	% bay with inc and dec
	H=cmp(R,[cmp(Ri,_),cmp(Rd,_),cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	insertDependency(Matrix,cmp("n",[int(15),str(Ri)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic2),
	insertDependency(Matrix,cmp("n",[int(15),str(Rd)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic3),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic4),
	Acyclic=Acyclic1*Acyclic2*Acyclic3*Acyclic4,
	!
;
	% bay with inc
	H=cmp(R,[cmp(Ri,_),_,cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	insertDependency(Matrix,cmp("n",[int(15),str(Ri)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic2),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic3),
	Acyclic=Acyclic1*Acyclic2*Acyclic3,
	!
;
	% bay with dec
	H=cmp(R,[_,cmp(Rd,_),cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	insertDependency(Matrix,cmp("n",[int(15),str(Rd)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic2),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic3),
	Acyclic=Acyclic1*Acyclic2*Acyclic3,
	!
;
	% bay w/out inc/dec; context
	H=cmp(R,[_,_,cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	% switch; context
	H=cmp(R,[_,_,_,_,cmp("p3_iAps",_)]),
	insertDependency(Matrix,cmp("n",[int(15),str(R)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic1),
	depCR_X(Matrix,Type,Node,Stat,L,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	dlgIF::feil("depCR_X",[Matrix,Node],[Type,Stat],cmp("",[H|L])),!,fail.	
	
depCR_Query(Matrix,Type,Node,Stat,Q,Acyclic) :-
	Q=cmp("p3_iDdrf",[_,_,str(Qa)]),
    Qt=termDomain::fromTermStringIndirect(Qa),
	dbBase::querySt(Qt,ST),
	insertDependency(Matrix,cmp("n",[int(2),str(ST)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!
;
	Q=cmp("p3_iSSS",[_,str(Qa)]),
    Qt=termDomain::fromTermStringIndirect(Qa),
	dbBase::querySt(Qt,ST),
	insertDependency(Matrix,cmp("n",[int(2),str(ST)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!
;
	Q=cmp("p3_iExis",[str(Qa)]),
    Qt=termDomain::fromTermStringIndirect(Qa),
	dbBase::querySt(Qt,ST),
	insertDependency(Matrix,cmp("n",[int(2),str(ST)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!
;
	Q=cmp(Func,_),
	dbBase::smDefF(Func,Clas,_,_,_,_,_),
	set::member(Clas,["f","g"]),
	insertDependency(Matrix,cmp("n",[int(19),str(Func)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!
;
	dlgIF::feil("depCR_Query",[Matrix,Node],[Type,Stat],Q),!,fail.	
% ----------------------------------------------------------------------------------------------
% depExists
% ----------------------------------------------------------------------------------------------
depExists(Matrix,Type,Node,Stat,cmp("p3_iExis",[str(PredCond)]),Acyclic) :-	
    Term=termDomain::fromTermStringIndirect(PredCond),
	depInsDeterminants(Matrix,Type,Node,Stat,Term,Acyclic),
	!.
% ----------------------------------------------------------------------------------------------
% depFuncBody          
% ----------------------------------------------------------------------------------------------
class predicates
depFuncStatement : (string,integer,string,integer,sterm,integer) procedure (i,i,i,i,i,o).
clauses

depFuncBody(Matrix,Type,Node,Stat,cmp(",",[B1,B2]),Acyclic) :-
	depFuncBody(Matrix,Type,Node,Stat,B1,Acyclic1),
	depFuncBody(Matrix,Type,Node,Stat,B2,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
depFuncBody(Matrix,Type,Node,Stat,B,Acyclic) :- 
	depFuncStatement(Matrix,Type,Node,Stat,B,Acyclic),
	!.
depFuncBody(Matrix,Type,Node,Stat,C,_) :- 
	dlgIF::feil("depFuncBody",[Matrix,Node],[Type,Stat],C),!,fail.

depFuncStatement(Matrix,Type,Node,Stat,cmp(Func,_),Acyclic) :- 
	dbBase::smDefF(Func,Clas,_,_,_,_,_),
	set::member(Clas,["f","g"]),
	% function call
	insertDependency(Matrix,cmp("n",[int(19),str(Func)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!.
depFuncStatement(Matrix,Type,Node,Stat,cmp(Derf,C),Acyclic) :- 
	set::member(Derf,["p3_iNdrf","p3_iDdrf"]),
	depDERF(Matrix,Type,Node,Stat,cmp(Derf,C),Acyclic),
	!.
depFuncStatement(Matrix,Type,Node,Stat,cmp("p3_iSSS",C),Acyclic) :- 
	depSSS(Matrix,Type,Node,Stat,cmp("p3_iSSS",C),Acyclic),
	!.
depFuncStatement(Matrix,Type,Node,Stat,cmp("p3_iExis",C),Acyclic) :- 
	depExists(Matrix,Type,Node,Stat,cmp("p3_iExis",C),Acyclic),
	!.
% other types of statements do not have any relevance for the dependency graph
depFuncStatement(_,_,_,_,_,1) :- 
	!.
% ----------------------------------------------------------------------------------------------
% depFuncCall
% ----------------------------------------------------------------------------------------------
class predicates
depFuncCallPars : (string,integer,string,integer,sterml,sterml,integer) determ (i,i,i,i,i,i,o).
depFuncCallListPars1 : (string,integer,string,integer,string,sterm,integer) determ (i,i,i,i,i,i,o).
depFuncCallListPars2 : (string,integer,string,integer,string,sterml,integer) procedure (i,i,i,i,i,i,o).
clauses

depFuncCall(Matrix,Type,Node,Stat,Cond,Acyclic) :-
	Cond=cmp("p3_iIR",[atom(CR)|Pars]),
	dbBase::smClassDefTree(CR,"h",Def),
	Def=cmp(_,[_,_,cmp("FPAR",ParDEs)]),
	depFuncCallPars(Matrix,Type,Node,Stat,ParDEs,Pars,Acyclic1),
	insertDependency(Matrix,cmp("n",[int(15),str(CR)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	Cond=cmp(Func,[_,_|Pars]),
	dbBase::smClassDefF(Func,"f",_,_,_,_,_),
	dbBase::smClassDefTree(Func,"f",Def),
	Def=cmp(_,[_,_,cmp("FPAR",ParDEs)]),
	depFuncCallPars(Matrix,Type,Node,Stat,ParDEs,Pars,Acyclic1),
	insertDependency(Matrix,cmp("n",[int(19),str(Func)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	Cond=cmp(Func,[_|Pars]),
	dbBase::smClassDefF(Func,"g",_,_,_,_,_),
	dbBase::smClassDefTree(Func,"g",Def),
	Def=cmp(_,[_,_,cmp("FPAR",ParDEs)]),
	depFuncCallPars(Matrix,Type,Node,Stat,ParDEs,Pars,Acyclic1),
	insertDependency(Matrix,cmp("n",[int(19),str(Func)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	Cond=cmp("p3_iDtFn",_),
	dlgIF::feil("depFuncCall",[],[],Cond),!,fail
;
	Cond=cmp(Func,_),
	dbBase::smClassDefF(Func,"p",_,_,_,_,_),
	dlgIF::feil("depFuncCall",[],[],Cond),!,fail.

depFuncCallPars(_,_,_,_,[],[],1) :- 
	!.
depFuncCallPars(Matrix,Type,Node,Stat,[fstr(_,D,_,_)|Ds],[cmp(_,Par)|Ps],Acyclic) :- 
	% list-parameters represented as list(...)
	Par=[list(P,Pt)],
	sx::sxTail(2,D,Dom,T),
	T="_l",
	depFuncCallListPars1(Matrix,Type,Node,Stat,Dom,list(P,Pt),Acyclic1),
	depFuncCallPars(Matrix,Type,Node,Stat,Ds,Ps,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;	
	Par=[P],
	depValue(Matrix,Type,Node,Stat,P,Acyclic1),
	depFuncCallPars(Matrix,Type,Node,Stat,Ds,Ps,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;
	% list-parameters represented as [...]
	sx::sxTail(2,D,Dom,T),
	T="_l",
	depFuncCallListPars2(Matrix,Type,Node,Stat,Dom,Par,Acyclic1),
	depFuncCallPars(Matrix,Type,Node,Stat,Ds,Ps,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
depFuncCallPars(Matrix,Type,Node,Stat,[fstr(_,D,_,_)|Ds],[Par|Ps],Acyclic) :- 
	% list-parameters represented as list(...)
	Par=list(P,Pt),
	sx::sxTail(2,D,Dom,T),
	T="_l",
	depFuncCallListPars1(Matrix,Type,Node,Stat,Dom,list(P,Pt),Acyclic1),
	depFuncCallPars(Matrix,Type,Node,Stat,Ds,Ps,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!
;	
	depValue(Matrix,Type,Node,Stat,Par,Acyclic1),
	depFuncCallPars(Matrix,Type,Node,Stat,Ds,Ps,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.

depFuncCallListPars1(_,_,_,_,_,nill,1) :- 
	!.
depFuncCallListPars1(Matrix,Type,Node,Stat,D,list(P,Ps),Acyclic) :- 
	depValue(Matrix,Type,Node,Stat,P,Acyclic1),
	depFuncCallListPars1(Matrix,Type,Node,Stat,D,Ps,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
	
depFuncCallListPars2(_,_,_,_,_,[],1) :- 
	!.
depFuncCallListPars2(Matrix,Type,Node,Stat,D,[P|Ps],Acyclic) :- 
	depValue(Matrix,Type,Node,Stat,P,Acyclic1),
	depFuncCallListPars2(Matrix,Type,Node,Stat,D,Ps,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
% ----------------------------------------------------------------------------------------------
% getSTs
% ----------------------------------------------------------------------------------------------
class predicates
getSTsX : (string,string,lst).
getSTsY : (lst,string,sterm,sterm) determ.
clauses

getSTs(Graph,SG,STs) :-
	Arr=lst::new(),
	getSTsX(Graph,SG,Arr),
	STs2=Arr:get(),
	set::distinct(STs2,STs),
	!.

getSTsX(Graph,SG,Arr) :-
    get_nd(Graph,T1,N1,T2,N2),
        getSTsY(Arr,SG,T1,N1),
        getSTsY(Arr,SG,T2,N2),
        fail
;
	!.

getSTsY(Arr,"",int(N),str(ST)) :-
    set::member(N,[2,11,12]),
    Arr:add(ST),
    !
;
    !.
getSTsY(Arr,SG,int(N),str(ST)) :-
    set::member(N,[2,11,12]),
    sx::sxUnpack(ST,'|',L),
    L=[SG|_],
    Arr:add(ST),
    !
;
    !.
% ----------------------------------------------------------------------------------------------
% deleteDependecy
% ----------------------------------------------------------------------------------------------
deleteDependency("p3_pSM",cmp(_,[int(T1),str(N1)]),cmp(_,[int(T2),str(N2)])) :- smGraph:delete(int(T1),str(N1),int(T2),str(N2)),!.
deleteDependency("p3_pDP",cmp(_,[int(T1),str(N1)]),cmp(_,[int(T2),str(N2)])) :- dpGraph:delete(int(T1),str(N1),int(T2),str(N2)),!.
deleteDependency("p3_pDR",cmp(_,[int(T1),str(N1)]),cmp(_,[int(T2),str(N2)])) :- drGraph:delete(int(T1),str(N1),int(T2),str(N2)),!.    
% ----------------------------------------------------------------------------------------------
% insertDependencyCheckDet           
% ----------------------------------------------------------------------------------------------
insertDependencyCheckDet(Graph,Det,Dep,Status) :-
    Det=cmp(_,[int(N),_]),
    N=2,
    !
;    
    Graph="p3_pSM",
    Det=cmp(_,[int(30),str("7000")]),
    % 7000 is the Pro/3 root menu and does not exist in the graph
    insertDependency(Graph,Det,Dep,Status,_),    
    !    
;    
    Det=cmp(_,[int(N),str(S)]),
    hasNode(Graph,N,S),
    insertDependency(Graph,Det,Dep,Status,_),    
    !
;    
    Det=cmp(_,[int(N),str(S)]),
    Dep=cmp(_,[int(N2),str(S2)]),
    N=N2,
    nodIF_No(K,N),
    dbBase::nlTry(K,NL),
    nodIF_No(K2,N2),
    dbBase::nlTry(K2,NL2),
    set::setGet(["p3_pSM","p3_pDP","p3_pDR"],Ix,Graph),
    set::setGet(["sentence model","dependency","derivation"],Ix,G),    
    Mess=string::concatList([NL," ",S," referred in ",NL2," ",S2," does not exist in the ",G," graph - this can simply be a knowledge loading sequencing issue!"]),
    dlgIF::msgS(Mess),
    insertDependency(Graph,Det,Dep,Status,_),    
    !
;    
    Det=cmp(_,[int(N),str(S)]),
    set::member(N,[1,16,17]),
    dbBase::smDefF(S,Clas,_,_,_,_,_),
    set::member(Clas,["p","f","g","h"]),
    insertDependency(Graph,Det,Dep,Status,_),    
    !
;    
    Det=cmp(_,[int(N),str(S)]),
    nodIF_No(K,N),
    dbBase::nlTry(K,NL),
    Dep=cmp(_,[int(N2),str(S2)]),
    nodIF_No(K2,N2),
    dbBase::nlTry(K2,NL2),
    set::setGet(["p3_pSM","p3_pDP","p3_pDR"],Ix,Graph),
    set::setGet(["sentence model","dependency","derivation"],Ix,G),
    Mess=string::concatList([NL," ",S," referred in ",NL2," ",S2," does not exist in the ",G," graph!"]),
    dlgIF::dlgNote(0,Mess),
	!
;    
    Det=cmp(_,[int(N),str(S)]),
    nodIF_No(K,N),
    dbBase::nlTry(K,NL),
    Dep=cmp(_,[int(N2),str(S2)]),
    nodIF_No(K2,N2),
    dbBase::nlTry(K2,NL2),
    set::setGet(["p3_pSM","p3_pDP","p3_pDR"],Ix,Graph),
    set::setGet(["sentence model","dependency","derivation"],Ix,G),
    Mess=string::concatList([NL," ",S," referred in ",NL2," ",S2," does not exist in the ",G," graph,!"]),
    dlgIF::dlgNote(0,Mess),
	!.    
% ----------------------------------------------------------------------------------------------
% insertDependency           
% ----------------------------------------------------------------------------------------------
% inserts a dependency if not already exists and it does not cause a cycle in the graph (in which case predicate fails)
% if exists then updates status as required (recursively in case of dependency graph)
% parameters:
% 1	in	graph
% 2	in	determinant
% 3	in	dependent
% 4	in	dependency status
%
insertDependency("p3_pSM",Det,Dep,_,Acyclic) :-
	depSMM_Exists(Det,Dep),
    Acyclic=1,
	!
;	
	% insert 
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
  	smGraph:add(A1,A2,B1,B2),
    Acyclic=1,
  	!
;
	Dep=cmp(_,[int(I1),str(N1)]),	
	Det=cmp(_,[int(I2),str(N2)]),
	S1=toString(I1),
	S2=toString(I2),
	sx::concat(["SM-graph dependency is cyclic: ",N2,"/",S2,"->",N1,"/",S1],Mess),
	dlgIF::feil("insertDependency",[Mess],[],nill),
    Acyclic=0,
	!.
insertDependency(Matrix,Det,Dep,Status,Acyclic) :-
            tt::tt_s(2,"Graph",Matrix),
			tt::tt_term(2,"dependent",Dep),
			tt::tt_term(2,"determinant",Det),
    Matrix="p3_pDP",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    dpGraph:has(A1,A2,B1,B2,int(Status)), % exists with same Status
    Acyclic=1,
    !
;
    Status=0,
    Matrix="p3_pDP",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    dpGraph:has(A1,A2,B1,B2), % exists with different Status
    dpGraph:setStatus(A1,A2,B1,B2,int(Status)), 
    setUnderived("p3_pDP",B1,B2),
    Acyclic=1,
    !
;
    Status=1,
    Matrix="p3_pDP",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    dpGraph:has(A1,A2,B1,B2), % exists with different Status
    dpGraph:setStatus(A1,A2,B1,B2,int(Status)), 
    setDerived("p3_pDP",A1,A2),
    Acyclic=1,
    !
;    
	% dependency is a direct recursive function call - no need to insert
	Dep=cmp(_,[int(17),str(N1)]),	
	Det=cmp(_,[int(17),str(N2)]),
	N1=N2,
	sx::concat(["Direct recursive function call: ",N2,"->",N1," - ignored in graph (",Matrix,")!"],Mess),
	dlgIF::msgS(Mess),
    Acyclic=1,
	!
;
	Dep=cmp("n",[int(17),str(N1)]),
	Det=cmp("n",[int(17),str(N2)]),
	depIsDependent(Matrix,Dep,Det),
	sx::concat(["Indirect recursive function call: ",N2,"->",N1," - ignored in graph!"],Mess),
	dlgIF::msgS(Mess),
    Acyclic=1,
	!
;	
    Status=0,
    Matrix="p3_pDP",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    dpGraph:add(A1,A2,B1,B2,int(Status)),
    setUnderived("p3_pDP",B1,B2),
 			tt::tt_i(2,"inserted",Status),
    Acyclic=1,
 	!
;	
    Status=1,
    Matrix="p3_pDP",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    dpGraph:add(A1,A2,B1,B2,int(Status)),
    setDerived("p3_pDP",A1,A2),
 			tt::tt_i(2,"inserted",Status),
     Acyclic=1,
 	!
;
    Matrix="p3_pDR",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    drGraph:has(A1,A2,B1,B2,int(Status)), % exists with same Status
    Acyclic=1,
    !
;
    Matrix="p3_pDR",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    drGraph:has(A1,A2,B1,B2), % exists with different Status
    drGraph:setStatus(A1,A2,B1,B2,int(Status)), 
    Acyclic=1,
    !
;	
    Matrix="p3_pDR",
  	Det=cmp(_,[A1,A2]),
  	Dep=cmp(_,[B1,B2]),
    drGraph:add(A1,A2,B1,B2,int(Status)),
 			tt::tt_i(2,"inserted",Status),
    Acyclic=1,
 	!
;
	Dep=cmp("n",[int(T1),str(N1)]),
	Det=cmp("n",[int(T2),str(N2)]),
	depIsDependent(Matrix,Dep,Det),
	T2s=toString(T2),
	T1s=toString(T1),
    sx::concat(["Dependency is cyclic: ",T2s,"|",N2,"->",T1s,"|",N1," (",Matrix,")!"],Mess),
	dlgIF::msgS(Mess),
    Acyclic=0,
    !.
% ----------------------------------------------------------------------------------------------
% depInsDeterminants
% ----------------------------------------------------------------------------------------------
% inserts all determinants of a node
% parameters:
% 1	in	graph
% 2	in	node-type
% 3	in	node-name
% 4	in	insertion status
% 5	in	node's determinant terms
%
depInsDeterminants(Matrix,Type,Node,Stat,cmp(",",[Cond1,Cond2]),Acyclic) :-
	depInsDeterminants(Matrix,Type,Node,Stat,Cond1,Acyclic1),
	depInsDeterminants(Matrix,Type,Node,Stat,Cond2,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
depInsDeterminants(Matrix,Type,Node,Stat,cmp("not",[Cond]),Acyclic) :-
	depInsDeterminants(Matrix,Type,Node,Stat,Cond,Acyclic),
	!.
depInsDeterminants(Matrix,Type,Node,Stat,Cond,Acyclic) :- 
	depProc(Matrix,Type,Node,Stat,Cond,Acyclic),
	!
;	
	depSSS(Matrix,Type,Node,Stat,Cond,Acyclic),
	!
;
	depVarCond(Matrix,Type,Node,Stat,Cond,Acyclic),
	!
;
	depPredCond(Matrix,Type,Node,Stat,Cond,Acyclic),
	!
;				
	depPred(Cond,Type1,Node1),
	insertDependency(Matrix,cmp("n",[int(Type1),str(Node1)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!.
% ----------------------------------------------------------------------------------------------
% depPred
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	predicate term
% 2	out	1:sentence group; 2:sentence type
% 3	out	#p2=1: predicate; #p2=2: RAC
%
depPred(Cond,Type,Node) :-
	Cond=cmp(R,[cmp(E,_)|_]),
	dbBase::smClassDefF(R,"p",Card,_,_,_,_),
    Card=1,
	Type=2,
	sx::concat([R,"|",E],Node),
	!
;
	Cond=cmp(R,[var(_)|_]),
	dbBase::smClassDefF(R,"p",Card,_,_,_,_),
	Card=1,
	Type=1,
	Node=R,
	!
;
	Cond=cmp(R,[cmp(E1,_),cmp(E2,_)|_]),
	dbBase::smClassDefF(R,"p",Card,_,_,_,_),
	Card=2,
	Type=2,
	sx::concat([R,"|",E1,"|",E2],Node),
	!
;
	Cond=cmp(R,[var(_),cmp(E2,_)|_]),
	dbBase::smClassDefF(R,"p",Card,_,_,_,_),
	Card=2,
	Type=5,
	sx::concat([R,"|_|",E2],Node),
	!
;
	Cond=cmp(R,[cmp(E2,_),var(_)|_]),
	dbBase::smClassDefF(R,"p",Card,_,_,_,_),
	Card=2,
	Type=5,
	sx::concat([R,"|",E2,"|_"],Node),
	!
;
	Cond=cmp(R,[var(_),var(_)|_]),
	dbBase::smClassDefF(R,"p",Card,_,_,_,_),
	Card=2,
	Type=1,
	Node=R,
	!.
% ----------------------------------------------------------------------------------------------
% depPredCond
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	predicate term
% 2	out	1:sentence group; 2:sentence type
% 3	out	#p2=1: predicate; #p2=2: RAC
%
depPredCond(Matrix,Type,Node,_,Cond,_) :-
	Cond=cmp("REL_COND",_),
	dlgIF::feil("depPredCond",["Unexpected parameter",Matrix,Node],[Type],Cond),!,fail.
depPredCond(Matrix,Type,Node,Stat,Cond,Acyclic) :-
	Cond=cmp(P,[cmp(Ea,_),cmp(Eb,_)|_]),
	dbBase::smClassDefF(P,"p",Card,_,_,_,_),
	Card=2,
        sx::sxlJoin([P,Ea,Eb],"|",ST),
	insertDependency(Matrix,cmp("n",[int(2),str(ST)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!.
depPredCond(Matrix,Type,Node,Stat,Cond,Acyclic) :-
	Cond=cmp(P,[var(_),cmp(Eb,_)|_]),
	dbBase::smClassDefF(P,"p",Card,_,_,_,_),
	Card=2,
	sx::concat([P,"|_|",Eb],STx),
	insertDependency(Matrix,cmp("n",[int(5),str(STx)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!.
depPredCond(Matrix,Type,Node,Stat,Cond,Acyclic) :-
	Cond=cmp(P,[cmp(Eb,_),var(_)|_]),
	dbBase::smClassDefF(P,"p",Card,_,_,_,_),
	Card=2,
	sx::concat([P,"|",Eb,"|_"],STx),
	insertDependency(Matrix,cmp("n",[int(5),str(STx)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!.
depPredCond(Matrix,Type,Node,Stat,Cond,Acyclic) :-
	Cond=cmp(P,[cmp(Ea,_)|_]),
	dbBase::smClassDefF(P,"p",Card,_,_,_,_),
	Card=1,
    sx::sxlJoin([P,Ea],"|",ST),
	insertDependency(Matrix,cmp("n",[int(2),str(ST)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!.
depPredCond(Matrix,Type,Node,Stat,Cond,Acyclic) :-
	Cond=cmp(P,[var(_)|_]),
	dbBase::smClassDefF(P,"p",Card,_,_,_,_),
	Card=1,
	insertDependency(Matrix,cmp("n",[int(1),str(P)]),cmp("n",[int(Type),str(Node)]),Stat,Acyclic),
	!.
% ----------------------------------------------------------------------------------------------
% depProc
% ----------------------------------------------------------------------------------------------
class predicates
depAssCond : (string,integer,string,integer,sterm,integer) determ (i,i,i,i,i,o).
clauses

depProc(Matrix,Type,Node,Stat,cmp(Rule,[str(Q1),str(D),str(A1)|_]),Acyclic) :-
    termDomain::isTermStringIndirect(Q1),
    termDomain::isTermStringIndirect(D),
    termDomain::isTermStringIndirect(A1),
	rub::rubRuleType(_,_,Rule),
	% sentence type condition
	Q2=termDomain::fromTermStringIndirect(Q1),
	depInsDeterminants(Matrix,Type,Node,Stat,Q2,Acyclic1),
	% implied DE assignment expression conditions
	A2=termDomain::fromTermStringIndirect(A1),
	depAssCond(Matrix,Type,Node,Stat,A2,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.   
	
depAssCond(_,_,_,_,nill,1) :- 
	!.
depAssCond(Matrix,Type,Node,Stat,list(Cond,T),Acyclic) :-
	depValue(Matrix,Type,Node,Stat,Cond,Acyclic1),
	depAssCond(Matrix,Type,Node,Stat,T,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
depAssCond(Matrix,Type,Node,Stat,C,_) :- 
	dlgIF::feil("depAssCond",[Matrix,Node],[Type,Stat],C),
	!,fail.
% ----------------------------------------------------------------------------------------------
% depSSS
% ----------------------------------------------------------------------------------------------
depSSS(Matrix,Type,Node,Stat,cmp("p3_iSSS",[_,str(PredCond)]),Acyclic) :-
    Term=termDomain::fromTermStringIndirect(PredCond),
	depInsDeterminants(Matrix,Type,Node,Stat,Term,Acyclic),
	!.
% ----------------------------------------------------------------------------------------------
% setStatus
% ----------------------------------------------------------------------------------------------
% changes all dependencies in the given graph with status 'From' to status 'To'
%
setStatus("p3_pDP",From,To) :- 
    dpGraph:setStatus(From,To),
	!.	
setStatus("p3_pDR",From,To) :- 
    drGraph:setStatus(From,To),
	!.	
% ----------------------------------------------------------------------------------------------
% setDerived
% ----------------------------------------------------------------------------------------------
setDerived("p3_pDP",T,N) :-
    dpGraph:get_nd(X,Y,T,N,int(0)),
        dpGraph:setStatus(X,Y,T,N,int(1)),
        setDerived("p3_pDP",X,Y),
        fail
;
    !.
setDerived("p3_pDR",T,N) :-
    drGraph:get_nd(X,Y,T,N,int(0)),
        drGraph:setStatus(X,Y,T,N,int(1)),
        setDerived("p3_pDR",X,Y),
        fail
;
    !.

setUnderived("p3_pDP",T,N) :-
    dpGraph:get_nd(T,N,X,Y,int(1)),
        dpGraph:setStatus(T,N,X,Y,int(0)),
        setUnderived("p3_pDP",X,Y),
        fail
;
    !.
setUnderived("p3_pDR",T,N) :-
    drGraph:get_nd(T,N,X,Y,int(1)),
        drGraph:setStatus(T,N,X,Y,int(0)),
        setUnderived("p3_pDR",X,Y),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% setStatusOpenSTs                                        
% ----------------------------------------------------------------------------------------------
setStatusOpenSTs(Status,SG) :-
    dpGraph:get_nd(_,_,int(5),STo),
        STo=str(STop),
        sx::sxUnpack(STop,'|',L),
        L=[SG|_],
        setStatusWhereDependent("p3_pDP",Status,cmp("n",[int(5),STo])),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% setStatusWhereDependent                                               
% ----------------------------------------------------------------------------------------------
% sets given status to all edges where Node is the dependent
%
setStatusWhereDependent("p3_pDP",Status,Node) :-
    Status=0,
    Node=cmp(_,[A1,A2]),
    setUnderived("p3_pDP",A1,A2),
    dpGraph:setStatusWhereSucc(A1,A2,int(Status)),
	!
;	
    Status=1,
    Node=cmp(_,[A1,A2]),
    setDerived("p3_pDP",A1,A2),
    dpGraph:setStatusWhereSucc(A1,A2,int(Status)),
	!.

setStatusWhereDependent("p3_pDR",Status,Node) :-
    Status=0,
    Node=cmp(_,[A1,A2]),
    setUnderived("p3_pDR",A1,A2),
    drGraph:setStatusWhereSucc(A1,A2,int(Status)),
	!
;	
    Status=1,
    Node=cmp(_,[A1,A2]),
    setDerived("p3_pDR",A1,A2),
    drGraph:setStatusWhereSucc(A1,A2,int(Status)),
	!.
% ----------------------------------------------------------------------------------------------
% setStatusWhereDeterminant                                               
% ----------------------------------------------------------------------------------------------
% sets given status to all edges where Node is the dependent
%
setStatusWhereDeterminant("p3_pDP",Status,Node) :-
    Status=0,
    Node=cmp(_,[A1,A2]),
    dpGraph:setStatusWherePred(A1,A2,int(Status)),
    setUnderived("p3_pDP",A1,A2),
	!
;	
    Status=1,
    Node=cmp(_,[A1,A2]),
    dpGraph:setStatusWherePred(A1,A2,int(Status)),
    setDerived("p3_pDP",A1,A2),
	!.

setStatusWhereDeterminant("p3_pDR",Status,Node) :-
    Status=0,
    Node=cmp(_,[A1,A2]),
    drGraph:setStatusWherePred(A1,A2,int(Status)),
    setUnderived("p3_pDR",A1,A2),
	!
;	
    Status=1,
    Node=cmp(_,[A1,A2]),
    drGraph:setStatusWherePred(A1,A2,int(Status)),
    setDerived("p3_pDR",A1,A2),
	!.
% ----------------------------------------------------------------------------------------------
% sentenceChanged
% ----------------------------------------------------------------------------------------------
% updates the dependency graph on the basis of a changed sentence of given type. However, if mass update is on-going,
% then only remembers the sentence type.
%
sentenceChanged(Change,ST) :-
    massSentenceUpdate=1,
    changedST(Change,ST), % already noted
    !
;    
	sx::sxUnpack(ST,'|',L),
	L=[P|_],
	dbBase::chainIsExternal(P),
	% ignore external chain change
	!
;
    massSentenceUpdate=1,
    assert(changedST(Change,ST)),
    !
;
	% change status for all dependencies where the fact's ST or SG is the determinant
	setStatusWhereDeterminant("p3_pDP",0,cmp("n",[int(2),str(ST)])),
	sx::sxUnpack(ST,'|',L),
	L=[P|_],
	setStatusWhereDeterminant("p3_pDP",0,cmp("n",[int(1),str(P)])),
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
% depValue
% ----------------------------------------------------------------------------------------------
depValue(Matrix,Type,Node,Stat,cmp(Proc,[P1]),Acyclic) :-
	dbBase::smClassDefF(Proc,"i",_,_,_,_,_),
	depValue(Matrix,Type,Node,Stat,P1,Acyclic),
	!.
depValue(Matrix,Type,Node,Stat,cmp(Proc,[P1,P2]),Acyclic) :-
	dbBase::smClassDefF(Proc,"i",_,_,_,_,_),
	depValue(Matrix,Type,Node,Stat,P1,Acyclic1),
	depValue(Matrix,Type,Node,Stat,P2,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
depValue(Matrix,Type,Node,Stat,cmp(Opr,[Opn1,Opn2]),Acyclic) :-
	% expression
	opr::opr_any(Opr,2),
	depValue(Matrix,Type,Node,Stat,Opn1,Acyclic1),
	depValue(Matrix,Type,Node,Stat,Opn2,Acyclic2),
	Acyclic=Acyclic1*Acyclic2,
	!.
depValue(Matrix,Type,Node,Stat,cmp(Opr,[Opn]),Acyclic) :-
	% expression
	opr::opr_any(Opr,1),
	depValue(Matrix,Type,Node,Stat,Opn,Acyclic),
	!.
depValue(Matrix,Type,Node,Stat,C,Acyclic) :- 
	depDERF(Matrix,Type,Node,Stat,C,Acyclic), 
	!.
depValue(Matrix,Type,Node,Stat,C,Acyclic) :- 
	depFuncCall(Matrix,Type,Node,Stat,C,Acyclic), 
	!.
depValue(_,_,_,_,_,1) :- 
	!.
% ----------------------------------------------------------------------------------------------
% depVarCond
% ----------------------------------------------------------------------------------------------
depVarCond(Matrix,Type,Node,Stat,cmp(Opr,[Var,Exp]),Acyclic) :-
	opr::opr_any(Opr,2),
	Var=var(_),
	depValue(Matrix,Type,Node,Stat,Exp,Acyclic),
	!
;	
	opr::opr_any(Opr,2), 
	Var=fvar(_,_),
	depValue(Matrix,Type,Node,Stat,Exp,Acyclic),
	!.
% ----------------------------------------------------------------------------------------------
% smGraphCheckDelete
% ----------------------------------------------------------------------------------------------
class predicates
checkDelete : (sterm) determ.
clauses

% succeeds if KE is willing to delete or suspend all dependent knowledge of the record under deletion (or there is no such dependent knowledge),
% parameters:
% 1 in record base type
% 2 in left side term
% 3 in right side term
%
smGraphCheckDelete(B,_,_) :-
   B<>2,
	!.
smGraphCheckDelete(_,_,_) :- 
	checkIntegrity=-1,
	!.
smGraphCheckDelete(2,LTerm,atom("true")) :-
	LTerm=cmp("q",[cmp(E,[atom(A)|_])|_]),
	set::member(E,["a","e","f","g","h","p"]),
	kNode("q",E,N),
	!,checkDelete(cmp("n",[int(N),str(A)])),
	!
;	
	LTerm=cmp("p3_pIRinp",[cmp("p3_eQueSer",[_,str(R)|_]),_]),
	kNode("p3_pIRinp","p3_eQueSer",N),
	checkDelete(cmp("n",[int(N),str(R)])),
	!
;	
	LTerm=cmp("p3_pIRinp",[cmp("p3_eQueClu",[str(Pr)|_]),atom(R)]),
	kNode("p3_pIRinp","p3_eQueClu",N),
	P=string::concat(R,"|",Pr),
	!,checkDelete(cmp("n",[int(N),str(P)])),
	!
;	
	LTerm=cmp("p3_pIntd",[cmp("p3_eAnno",[_,str(Name)|_])]),
	kNode("p3_pIntd","p3_eAnno",N),
	!,checkDelete(cmp("n",[int(N),str(Name)])),
	!
;	
	LTerm=cmp("y",[cmp("p3_eTabLay",[str(ST)|_])]),
	kNode("y","p3_eTabLay",N),
	!,checkDelete(cmp("n",[int(N),str(ST)])),
	!
;	
	LTerm=cmp("y",[cmp("p3_eKbq",[int(Id)|_])]),
	SId=toString(Id),
	% is actually either 28 or 29, but deletes both instead of checking (no side-effect)
	!,checkDelete(cmp("n",[int(28),str(SId)])),
	checkDelete(cmp("n",[int(29),str(SId)])),
	!
;	
	LTerm=cmp("y",[cmp(E,[int(Id)|_])]),
	set::member(E,["p3_eKbp","p3_eKbm"]),
	kNode("y",E,N),
	SId=toString(Id),
	!,checkDelete(cmp("n",[int(N),str(SId)])),
	!
;
    !.

checkDelete(Node) :-
    Node=cmp(_,[int(N),str(Name)]),
	chkGetAllDeps(Node,_,SDeps),
	dbBase::nlTry(Name,NL),
	nodIF_No(TypK,N),
	dbBase::nlSing(TypK,TypN),
	sx::concat([TypN," ",NL],S),
    Dlg=dependents::new(),
    Dlg:setPars(SDeps,S,1),
    W=guix::getWinOfClas(1),
    Dlg:show(W),
	globalVars::popRetPars()=["0"], 
	!, % delete!
    globalVars::get_i("quit_flag")=0,
	globalVars::get_i("error_flag")=0,
	!.
% ----------------------------------------------------------------------------------------------
% chkGetAllDeps
% ----------------------------------------------------------------------------------------------
class predicates
getST_Deps : (sterml,sterml) determ (i,o).
clauses

chkGetAllDirectDeps(Node,Deps,DepsNL) :-	
	Node=cmp(_,[int(1),str(P)]), % PT
	Set1=getAllDeps("p3_pSM",1,P),
	Set2=getAllDeps("p3_pDP",1,P),
	Set1:union(Set2),
	Deps1=Set1:get(),
	chkGetSTs(smGraph,[cmp("n",[int(1),str(P)])],ST2), 
	chkGetSTs(dpGraph,[cmp("n",[int(1),str(P)])],ST3),
	set::conc([ST2,ST3],STs), % all STs of PT
	getST_Deps(STs,Deps2),
	set::conc([Deps1,Deps2],Deps3),
	set::distinct(Deps3,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
	Node=cmp(_,[int(14),str(E)]), % ET
	Set1=getAllDeps("p3_pSM",14,E),
	Deps1=Set1:get(),
	chkGetSTs(smGraph,[cmp("n",[int(14),str(E)])],ST2), 
	chkGetSTs(dpGraph,[cmp("n",[int(14),str(E)])],ST3),
	set::conc([ST2,ST3],STs), % all STs of ET
	getST_Deps(STs,Deps2),
	set::conc([Deps1,Deps2],Deps3),
	set::distinct(Deps3,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
	Node=cmp(C,[int(15),B]),
    chkGetAllDirectDeps(cmp(C,[int(16),B]),Deps,DepsNL),
    !
;	
	Node=cmp(_,[int(16),str(Name)]), % ir decl
	Set1=getAllDeps("p3_pSM",16,Name),
	Set2=getAllDeps("p3_pDP",15,Name),
	Set1:union(Set2),
	Deps2=Set1:get(),
	set::distinct(Deps2,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
	Node=cmp(_,[int(17),str(Name)]), % func decl
	Set1=getAllDeps("p3_pSM",17,Name),
	Set2=getAllDeps("p3_pDP",19,Name),
	Set1:union(Set2),
	Deps2=Set1:get(),
	set::distinct(Deps2,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
	Node=cmp(_,[int(N),str(Name)]),
	set::member(N,[13,28,29,30]),
	Set1=getAllDeps("p3_pSM",N,Name),
	Deps2=Set1:get(),
	set::distinct(Deps2,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
    Deps=[],
    DepsNL=[],
    !.

chkGetAllDeps(Node,Deps,DepsNL) :-	
	Node=cmp(_,[int(1),str(P)]), % PT
	Set1=getAllDeps("p3_pSM",1,P),
	Set2=getAllDeps("p3_pDP",1,P),
	Set1:union(Set2),
	Deps1=Set1:get(),
	chkGetSTs(smGraph,[cmp("n",[int(1),str(P)])],ST2), 
	chkGetSTs(dpGraph,[cmp("n",[int(1),str(P)])],ST3),
	set::conc([ST2,ST3],STs), % all STs of PT
	getST_Deps(STs,Deps2),
	set::conc([Deps1,Deps2],Deps3),
	set::distinct(Deps3,Deps4),
	chkGetAllDepsX(Deps4,Deps5),
	set::conc([Deps4,Deps5],Deps6),
	set::distinct(Deps6,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
	Node=cmp(_,[int(14),str(E)]), % ET
	Set1=getAllDeps("p3_pSM",14,E),
	Deps1=Set1:get(),
	chkGetSTs(smGraph,[cmp("n",[int(14),str(E)])],ST2), 
	chkGetSTs(dpGraph,[cmp("n",[int(14),str(E)])],ST3),
	set::conc([ST2,ST3],STs), % all STs of ET
	getST_Deps(STs,Deps2),
	set::conc([Deps1,Deps2],Deps3),
	set::distinct(Deps3,Deps4),
	chkGetAllDepsX(Deps4,Deps5),
	set::conc([Deps4,Deps5],Deps6),
	set::distinct(Deps6,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;	
	Node=cmp(C,[int(15),B]),
    chkGetAllDeps(cmp(C,[int(16),B]),Deps,DepsNL),
    !
;
	Node=cmp(_,[int(16),str(Name)]), % ir decl
	Set1=getAllDeps("p3_pSM",16,Name),
	Set2=getAllDeps("p3_pDP",15,Name),
	Set1:union(Set2),
	Deps2=Set1:get(),
	chkGetAllDepsX(Deps2,Deps3),
	set::conc([Deps2,Deps3],Deps4),
	set::distinct(Deps4,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
	Node=cmp(_,[int(17),str(Name)]), % func decl
	Set1=getAllDeps("p3_pSM",17,Name),
	Set2=getAllDeps("p3_pDP",19,Name), 
	Set1:union(Set2),
	Deps2=Set1:get(),
	chkGetAllDepsX(Deps2,Deps3),
	set::conc([Deps2,Deps3],Deps4),
	set::distinct(Deps4,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
	Node=cmp(_,[int(N),str(Name)]),
	set::member(N,[13,28,29,30]),
	Set1=getAllDeps("p3_pSM",N,Name),
	Deps2=Set1:get(),
	chkGetAllDepsX(Deps2,Deps3),
	set::conc([Deps2,Deps3],Deps4),
	set::distinct(Deps4,Deps),
	nodeToNLs(1,Deps,DepsNL),
	!
;
    Deps=[],
    DepsNL=[],
    !.

chkGetAllDepsX([],[]) :-
    !.
chkGetAllDepsX([H|T],Ds) :-
    chkGetAllDeps(H,Ds2,_),
    chkGetAllDepsX(T,Ds3),
    set::conc([Ds2,Ds3],Ds),
    !.
    
% gets all nodes which will have to be deleted or suspended if the given node is deleted
%
getST_Deps(STs,Deps) :-
    set::distinct(STs,STsD),
	depDetsOfType(dpGraph,[3],STsD,Deps5), % SRs concluding these STs
	depDetsOfType(dpGraph,[3,16,17],STsD,Deps6), % STs' dependent SRs,CRs,Fs
	set::conc([Deps5,Deps6,STsD],Deps9),
	set::distinct(Deps9,Deps),
    !.   
% ----------------------------------------------------------------------------------------------
% chkGetSTs 
% ----------------------------------------------------------------------------------------------
class predicates
chkGetST_2 : (lst,dirgraph,sterml) determ (i,i,i).
chkGetST_3 : (lst,dirgraph,sterm).
chkGetST_4 : (lst,string,sterm).
clauses

% parameters:
% 1	in	graph
% 2	in	list of PT/ET nodes
% 3	in	list ST nodes Referring to p#2
%
chkGetSTs(Graph,Nodes,STs) :-
	Arr=lst::new(),
	chkGetST_2(Arr,Graph,Nodes),
	L=Arr:get(),
	depListStrToTerm(L,STs),
	!.

chkGetST_2(_,_,[]) :-
	!.
chkGetST_2(Arr,Graph,[Node|Nodes]) :-
	chkGetST_3(Arr,Graph,Node),
	chkGetST_2(Arr,Graph,Nodes),
	!.

chkGetST_3(Arr,DG,Node) :-
	Node=cmp(_,[_,str(N)]),
	DG:get_nd(A1,A2,B1,B2),
        chkGetST_4(Arr,N,cmp("n",[A1,A2])),
        chkGetST_4(Arr,N,cmp("n",[B1,B2])),
	fail
;
	!.

chkGetST_4(Arr,E,Node) :-
	Node=cmp("n",[int(2),str(ST)]),
	dbBase::stComps(ST,_,P,E1,E2),
	set::member(E,[P,E1,E2]),
	T=toString(Node),
	Arr:add(T),
	!.
chkGetST_4(_,_,_) :- 
	!.
% ----------------------------------------------------------------------------------------------
% getDeps
% ----------------------------------------------------------------------------------------------
getAllDeps("p3_pSM",Type,Name)=Set :- Set=smGraph:succs(int(Type),str(Name)),!.
getAllDeps("p3_pDR",Type,Name)=Set :- Set=drGraph:succs(int(Type),str(Name)),!.
getAllDeps("p3_pDP",Type,Name)=Set :- Set=dpGraph:succs(int(Type),str(Name)),!.
% ----------------------------------------------------------------------------------------------
% nodIF_No/NL
% ----------------------------------------------------------------------------------------------
% relates node type name (IF) to node type number used in the graphs
% 
%							IN GRAPHS:	IDENTIFYING NAME:
nodIF_No("p3_iAnno",0) :- !. % annotation - (no name)
nodIF_No("p3_iPT",1) :- !.  % SG / predicate type	c		IF
nodIF_No("p3_iST",2) :- !.  % ST 			c,m		IF
nodIF_No("p3_iSR",3) :- !.  % sentence rule 		c,m		CHAIN|REALM|SENTNO|SUBNO
nodIF_No("p3_iNIL",4) :- !.  % function generic		IF
nodIF_No("p3_iSTO",5) :- !. % open ST 
nodIF_No("p3_iNIL",6) :- !. % formerly: IR generic
nodIF_No("p3_iNIL",7) :- !.  % formerly: function generic
nodIF_No("p3_iNIL",8) :- !. 
nodIF_No("p3_iNIL",9) :- !. 
nodIF_No("p3_iNIL",10) :- !. 
nodIF_No("p3_iDST",11) :- !. % derived ST		h		IF
nodIF_No("p3_iIST",12) :- !. % inputted ST		h		IF
nodIF_No("p3_iDET",13) :- !. % DE			h		IF
nodIF_No("p3_iET",14) :- !. % entity type		h		IF
nodIF_No("p3_iIRdef",15) :- !. % inexact rule definition
nodIF_No("p3_iIRdec",16) :- !.  % inexact rule declaration
nodIF_No("p3_iFncD",17) :- !. % function declaration   		h		IF (function)   
nodIF_No("p3_iFncB",19) :- !. % function definition  		h		IF (function)  
nodIF_No("p3_iNIL",20) :- !. 
nodIF_No("p3_iRelm",21) :- !. % realm			-		realm name
nodIF_No("p3_iDB",22) :- !.  % database			-		database ID
nodIF_No("p3_iNIL",23) :- !. 
nodIF_No("p3_iSegm",24) :- !. % segment			-		segment name
nodIF_No("p3_iInpSer",25) :- !. % series input facility - inexact rule
nodIF_No("p3_iInpClu",26) :- !. % cluster input facility - prompt
nodIF_No("p3_iTabLay",27) :- !. % table layout - ST
nodIF_No("p3_iKQry",28) :- !. % KB query - id no
nodIF_No("p3_iKQryC",29) :- !. % KB query cluster - id no
nodIF_No("p3_iKQryM",30) :- !. % KB query menu - id no
nodIF_No("p3_iKQryP",31) :- !. % KB query parameter - name

nodIF_NoNL(NL,Typ) :-
    nodIF_No(K,Typ),
    dbBase::nlSing(K,NL),
    !.
% ----------------------------------------------------------------------------------------------
% nodTypesIF/NL
% ----------------------------------------------------------------------------------------------
class predicates
nodTypesX : (integer,string_list,string_list) determ (i,o,o).
clauses

nodTypes(K,N) :-
	nodTypesX(1,K,N),
	!.

nodTypesX(32,[],[]) :-
	!.
nodTypesX(No,[K|Ks],[N|Ns]) :-
	nodIF_No(K,No),
	dbBase::nlSing(K,N),
	Next=No+1,
	nodTypesX(Next,Ks,Ns),
	!.	
nodTypesX(No,Ks,Ns) :-
	Next=No+1,
	nodTypesX(Next,Ks,Ns),
	!.	
% ----------------------------------------------------------------------------------------------
% isCR
% ----------------------------------------------------------------------------------------------
isCR("p3_iCswi") :- !.
isCR("p3_iCbay") :- !.
isCR("p3_iCdat") :- !.
isCR("p3_iCand") :- !.
isCR("p3_iCor") :- !.
isCR("p3_iCnot") :- !.
isCR("p3_iCcmb") :- !.
isCR("p3_iCmap") :- !.
isCR("p3_iCpar") :- !.
isCR("p3_iZnot") :- !.
isCR("p3_iZowa") :- !.
isCR("p3_iZand") :- !.
isCR("p3_iZor") :- !.
% -------------------------------------------------------------------------------------
% nodeToNL/s                                                             
% -------------------------------------------------------------------------------------
% parameters:
% 1 in 1=output format 'NN-NODETYPE:NODENAME' else 'NN-NODENAME'
% 2 in node term
% 3 out node (NL)
%
nodeToNL(Prefixed,Node,SNode) :-
	Prefixed=1,
	Node=cmp(_,[int(T),str(N)]),
	nodIF_No(K,T),
	dbBase::nlSing(K,S2),
	dbBase::kbIF_to_Sym(N,S3),
	S4=toString(T),
	sx::concat([S4,"-",S2,": ",S3],SNode),
	!
;
	Node=cmp(_,[int(T),str(N)]),
	dbBase::kbSym_to_IF(N,S2),
	S4=toString(T),
	sx::concat([S4,"-",S2],SNode),
	!.
nodeToNLs(_,[],[]) :-
	!.
nodeToNLs(Prefixed,[Node|Nodes],[SNode|SNodes]) :-
	nodeToNL(Prefixed,Node,SNode),
	nodeToNLs(Prefixed,Nodes,SNodes),
	!.
% ----------------------------------------------------------------------------------------------
% derivationStatus
% ----------------------------------------------------------------------------------------------
derivationStatus(DerFlag) :-
    dpChanged=0,
    DerFlag=1,
    !
;    
    dpGraph:get_nd(_,_,int(2),_,int(0)),
    DerFlag=0,
    !
;
    dpGraph:get_nd(_,_,int(1),_,int(0)),
    DerFlag=0,
    !
;
    DerFlag=1,
    !.
% ----------------------------------------------------------------------------------------------
% cascadeStatusBack
% ----------------------------------------------------------------------------------------------
cascadeStatusBack("p3_pDP",NewStatus) :- dpGraph:cascadeStatusBack(int(NewStatus)),!.
cascadeStatusBack("p3_pDR",NewStatus) :- drGraph:cascadeStatusBack(int(NewStatus)),!.
% ----------------------------------------------------------------------------------------------
% cascadeStatusForward
% ----------------------------------------------------------------------------------------------
cascadeStatusForward("p3_pDP",NewStatus) :- dpGraph:cascadeStatusForward(int(NewStatus)),!.
cascadeStatusForward("p3_pDR",NewStatus) :- drGraph:cascadeStatusForward(int(NewStatus)),!.
% ----------------------------------------------------------------------------------------------
% massSentenceUpdateStart/End
% ----------------------------------------------------------------------------------------------
massSentenceUpdateStart() :-
    massSentenceUpdate=1,
    pfeil::syserr("massSentenceUpdateStart","Mass update cannot be started twice!"),
    !,fail
;    
    retractall(changedST(_,_)),
    massSentenceUpdate:=1,
    !.
    
massSentenceUpdateEnd() :-
    massSentenceUpdate=0,
    dlgIF::dlgNote(3,"Attempting to end non-started mass sentence update!"),
    !,fail
;    
    massSentenceUpdate:=0,
    changedST(1,ST),
        sentenceChanged(1,ST),
        smGraphInsert(12,ST),
        fail
;            
    changedST(3,ST),
        sentenceChanged(3,ST),
        not(dbBase::stExists(2,ST)),
        deleteSM_Deps(cmp("n",[int(12),str(ST)])),
        deleteSM_Dets(cmp("n",[int(12),str(ST)])),
        fail
;            
    !.    
    
massSentenceUpdateStarted() :- massSentenceUpdate=1,!.   

massSentenceUpdateAbort() :-
    massSentenceUpdate=1,
    globalVars::get_i("error_flag")=0,
    globalVars::get_i("quit_flag")=0,
    changedST(_,_),
    massSentenceUpdateEnd,
    !
;
    massSentenceUpdate:=0,
    !.
% ----------------------------------------------------------------------------------------------
% addChangedST
% ----------------------------------------------------------------------------------------------
addChangedST(Action,ST) :-
    changedST(Action,ST),
    !
;
    assert(changedST(Action,ST)),
    !.
% ----------------------------------------------------------------------------------------------
% deleteWhereStatus
% ----------------------------------------------------------------------------------------------
deleteWhereStatus("p3_pDP",S) :- dpGraph:deleteWhereStatus(int(S)),!.
deleteWhereStatus("p3_pDR",S) :- drGraph:deleteWhereStatus(int(S)),!.
% ----------------------------------------------------------------------------------------------
% serial
% ----------------------------------------------------------------------------------------------
serial("p3_pDP")=ST_Terms :- ST_Terms=dpGraph:serial(),!.
serial("p3_pDR")=ST_Terms :- ST_Terms=drGraph:serial(),!.
serial("p3_pSM")=ST_Terms :- ST_Terms=smGraph:serial(),!.
% ----------------------------------------------------------------------------------------------
% deleteST_StartNodes
% ----------------------------------------------------------------------------------------------
deleteST_StartNodes() :-
    drGraph:get_nd(int(2),A2,B1,B2,Status),
        not(drGraph:hasAsSucc(int(2),A2)),
        % A1/A2 is an ST start-node i.e. an entered or external ST
        drGraph:delete(int(2),A2,B1,B2),
        Status=int(0),
        % move the underived status to the successors
        drGraph:setStatusWherePred(B1,B2,int(0)),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% realmRename
% ----------------------------------------------------------------------------------------------
realmRename(From,To) :-
    dpGraph:get_nd(int(3),A2,_,_,_),
        A2=str(S),
        sx::sxUnpack(S,'|',L),
        L=[Chain,From,No,Sub],
        not(dbBase::chainIsExternal(Chain)),
        Lm=[Chain,To,No,Sub],
        sx::sxlJoin(Lm,"|",Sm),
        dpGraph:replace(int(3),str(S),int(3),str(Sm)),
        fail
;
    !.
end implement dep
