/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement dlgIF
open core, resourceIdentifiers

constants
className = "com/rudstrand/dlgIF/dlgIF".
classVersion = "1.0".

class facts
message : (string) nondeterm. % list of suppressed messages

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------- 
% resetMessages
% -------------------------------------------------------------------------- 
resetMessages() :-
    retractall(message(_)),
    !.
% -------------------------------------------------------------------------- 
% timPeriodText                       
% -------------------------------------------------------------------------- 
timPeriodText(Secs,T) :-
	Secs<60,
	S=toString(Secs),
	sx::concat([S," seconds"],T),
	!
;
	Secs<3600,
	Mins=Secs div 60,
	Secs2=Secs-Mins*60,
	S=toString(Secs2),
	M=toString(Mins),
	sx::concat([M,":",S," minutes"],T),
	dlgIF::msgS(T),
	!
;		
	Mins=Secs div 60,
	Secs2=Secs-Mins*60,
	Hours=Mins div 60,
	Mins2=Mins-Hours*60,
	S=toString(Secs2),
	M=toString(Mins2),
	H=toString(Hours),
	sx::concat([H,":",M,":",S," hours"],T),
	dlgIF::msgS(T),
	!.
% ------------------------------------------------------------------------------------------------------------
% reportState
% ------------------------------------------------------------------------------------------------------------
class predicates
userRealms : (string_list).
clauses

reportState():-
    globalVars::get_i("explain")=1,
    Rs=dbBase::realms(),
    userRealms(Rs),
    !
;
    !.
    
userRealms([]) :-
    dlgIF::dlgExp(26),
    !.
userRealms([H|T]) :-
    string::length(H)>4,
    S=string::subString(H,0,4),
    S="PRO3",
    userRealms(T),
    !
;
    S=string::subString(H,0,1),
    S="$",
    userRealms(T),
    !
;
    !.
% ------------------------------------------------------------------------------------------------------------
% msgElapsed
% ------------------------------------------------------------------------------------------------------------
msgElapsed(Mess,Secs) :-
	timPeriodText(Secs,S),
	sx::concat([Mess," ",S,"."],T),
	msgS(T),
	!.
% ------------------------------------------------------------------------------------------------------------
% dlgExp
% ------------------------------------------------------------------------------------------------------------
dlgExp(N) :-
    File=fioExt::getPath("exe","pro3_Guide.xml"),
    not(file::existfile(File)),
    !
;
    guide::alreadyDisplayed(N),
    !
;
	globalVars::get_i("explain")=-1, 
    globalVars::get_i("assready")=-1,
    !
;
	globalVars::get_i("explain")=-1, 
    globalVars::get_i("assready")=1,
	W=guix::getWinOfClas(4),
	dlgExpBlock(N,48,Title,Conts),
  	guix::ctlInvisible(W,[idc_nodes]),
  	guix::ctlVisible(W,[idc_topics]),
  	guix::txtSet(W,idc_annotation,"Note!"),
    guix::lbtSel(W,idc_topics,Title),
  	guix::txtSet(W,idc_edit,Conts),
  	!
;  	
    globalVars::get_i("assready")=-1,
	globalVars::get_i("explain")=1,
	W=guix::getWinOfClas(1),
	guide::setDisplayed(N),
    Dlg=guide::new(),
    Dlg:show(W,N),
	!;!.

dlgExpGet(N,Title,Note) :-
    File=fioExt::getPath("exe","pro3_Guide.xml"),
    file::existFile(File),
    SNo=toString(N),
    S=file::readString(File,_),
    X=xmlTree::newTree(S),
    Title=X:getTagValueAfterTag("no",SNo,"title"),
    Note=X:getTagValueAfterTag("no",SNo,"content"),
    !.
    
dlgExpBlock(N,Width,Title,Note) :-
    dlgExpGet(N,Title,S),
	Note2=sx::replaceAll(S,"\n"," "),
	Note3=sx::replaceAll(Note2,"  "," "),
	sx::sxBlockNL(0,Width,Note3,Note),
	!.
% ----------------------------------------------------------------------------------------------
% dbaStatuses
% ----------------------------------------------------------------------------------------------
class predicates
dbaFlagText : (string,string) determ (i,o).	
clauses

% the toolbar-displayed flags include
% !	there are (potentially) un-derived facts
% +	KB has been changed since last open/save
% Q	quit flag ON
% E	error flag ON
%
dbaStatuses () :- 
	globalVars::get_i("all_flag")=1,
	dbaStatusesX("knowopen",2,"+",S1,_),
	dbaStatusesX("derflag",0,"!",S2,_),
	dbaStatusesX("quit_flag",1,"Q",S3,_),
	dbaStatusesX("error_flag",1,"E",S4,_),
	M=globalVars::get_i("all_del")+2,
	set::setGet(["S","","D"],M,S5),
	sx::concat([S1,S2,S3,S4,"A",S5],Flags),
	W=guix::getWinOfClas(1),  
	vpiToolbar::setValue(W,idt_flags,vpiToolbar::text_value(Flags)),
    /*
	AW=guix::getWinOfClas(4),
	set::setGet(["suspend","delete"],M,S51),
	S52=string::concat("ALL=",S51),
	set::distinct([S11,S12,S13,S14,S52],L2),
	set::setRemSub([""],L2,L),
	guix::lbxPop(AW,idc_set_flags,L),
	*/
	!
;	
	globalVars::get_i("all_flag")=-1,
	dbaStatusesX("knowopen",2,"+",S1,_),
	dbaStatusesX("derflag",0,"!",S2,_),
	dbaStatusesX("quit_flag",1,"Q",S3,_),
	dbaStatusesX("error_flag",1,"E",S4,_),
	sx::concat([S1,S2,S3,S4],Flags),
	W=guix::getWinOfClas(1),  
	vpiToolbar::setValue(W,idt_flags,vpiToolbar::text_value(Flags)),
	/*
	AW=guix::getWinOfClas(4),
	set::distinct([S11,S12,S13,S14],L2),
	set::setRemSub([""],L2,L),
	guix::lbxPop(AW,idc_set_flags,L),
	*/
	!.
	
dbaStatusesX(Flag,Val,SFlag,SFlag,FlagText) :- 
    Flag="derflag",
	dep::derivationStatus(V),
	V=Val,
	dbaFlagText(Flag,FlagText),
	!
;
    Flag<>"derflag",
	Val=globalVars::get_i(Flag),
	dbaFlagText(Flag,FlagText),
	!.	
dbaStatusesX(_,_,_,"","") :- 
	!.
		
dbaFlagText("knowopen","KB has been updated!") :- !.	
dbaFlagText("derflag","Sentences are not (fully) derived!") :- !.	
dbaFlagText("quit_flag","Quit has been requested!") :- !.	
dbaFlagText("error_flag","Error has occurred!") :- !.	
% ------------------------------------------------------------------------------------------------------------
% notesOff/On
% ------------------------------------------------------------------------------------------------------------
notesOff() :-
    message("*"),
    !
;
    assert(message("*")),
    !.

notesOn() :-
    message("*"),
    retract(message("*")),
    !
;
    !.
% ------------------------------------------------------------------------------------------------------------
% dlgNote
% ------------------------------------------------------------------------------------------------------------
class predicates
dlgNoteX : (string,string).
clauses

% N
% 0 notification, errors without consequence for KB
% 1 user error possibly with consequence for KB
% 2 error in KB content
% 3 system/program error
% 4 unsupported functionality
% 5 function call error
%
dlgNote(N,S) :-
    globalVars::get_i("server")=1,
    !,
    tt::tt_i(1,"*SERVER* dlgNote/N",N),
    tt::tt_s(1,"*SERVER* dlgNote/S",S),
    !
;    
    N=0,
    tt::tt_i(1,"dlgNote/N",N),
    tt::tt_s(1,"dlgNote/S",S),
    message("*"), 
    !
;    
    N=0,
    tt::tt_i(1,"dlgNote/N",N),
    tt::tt_s(1,"dlgNote/S",S),
    message(S), 
    !
;    
    N>0,
    globalVars::set_i("error_flag",1),
	dbaStatuses,
    tt::tt_i(1,"dlgNote/N",N),
    tt::tt_s(1,"dlgNote/S",S),
    message("*"), 
    !
;    
    N>0,
    globalVars::set_i("error_flag",1),
	dbaStatuses,
    tt::tt_i(1,"dlgNote/N",N),
    tt::tt_s(1,"dlgNote/S",S),
    message(S), 
    !
;
    N=0,
    W=guix::getWinOfClas(1),
    Dlg = note::new(),
    Dlg:show(W,"For Information",S),
    L=globalVars::popRetPars(),
    L=[Par],
    dlgNoteX(Par,S),
    !
;
    N<6,
    globalVars::set_i("error_flag",1),
	dbaStatuses,
    Ix=N+1,
    set::setGet(["For Information","Error","KB Error","System Error","Unsupported Functionality","Function Call Failed"],Ix,Type),
    W=guix::getWinOfClas(1),
    Dlg = note::new(),
    Dlg:show(W,Type,S),
    L=globalVars::popRetPars(),
    L=[Par],
    dep::massSentenceUpdateAbort,
    dlgNoteX(Par,S),
    !
;    
    N>=6,
    globalVars::set_i("error_flag",1),
	dbaStatuses,
    W=guix::getWinOfClas(1),
    Dlg = note::new(),
    Dlg:show(W,"Note!",S),
    L=globalVars::popRetPars(),
    L=[Par],
    dlgNoteX(Par,S),
    !
;
    !.

dlgNoteX("no",S) :- assert(message(S)),!.
dlgNoteX("no_any",_) :- assert(message("*")),!.
dlgNoteX("quit",_) :- globalVars::set_i("quit_flag",1),!.
dlgNoteX(_,_) :- !.
% ------------------------------------------------------------------------------------------------------------
% feil
% ------------------------------------------------------------------------------------------------------------
feil(S1,Strings,Integers,Tree) :-
    dep::massSentenceUpdateAbort,
	globalVars::get_i("error_flag")=1,
	!
;
    globalVars::get_i("server")=1,
	Mess2=string::concat("*SERVER* E R R O R  in predicate ",S1),
	tt::tt_header(1,Mess2),
	tt::tt_strings(1,"",Strings),
	tt::tt_is(1,"",Integers),
	tt::tt_term(1,"Tree",Tree),
	dbaStatuses,
	!
;
    globalVars::get_i("server")<>1,
    setFlag("error_flag",1),
	setFlag("knowerr",1),
	setFlag("quit_flag",1),
	S2=toString(Tree),
	sx::sxlJoin(Strings,"\n",S),
	sx::joinIntegers(Integers,"; ",Is),
	sx::concat([S,"\n",Is,"\nTree=",S2,"\n\n(in predicate ",S1,")"],S3),
	vpiCommonDialogs::note("Error",S3),
	Mess=string::concat("\nERROR IN:",S1,"\n"),
	msgS(Mess),
	Mess2=string::concat("E R R O R  in predicate ",S1),
	tt::tt_header(1,Mess2),
	tt::tt_strings(1,"",Strings),
	tt::tt_is(1,"",Integers),
	tt::tt_term(1,"Tree",Tree),
	dbaStatuses,
	!.
feil(_,_,_,_) :- 
    !.
% ------------------------------------------------------------------------------------------------------------
% feil_pie
% ------------------------------------------------------------------------------------------------------------
feil_pie(S1,Strings,Integers,Tree) :-
    globalVars::get_i("server")=1,
	tt::tt_header(1,"*SERVER* I N F E R E N C E   ERROR"),
	N1=globalVars::get_i("piercl"),
	N2=globalVars::get_i("solref"),
	N2<>0,
	sol::solSetHeader(N2,N3,_,_,Q,_),
	tt::tt_i(1,"Error recursion level=",N1),
	tt::tt_i(1,"Solution reference=",N2),
	tt::tt_i(1,"Solution counter=",N3),
	tt::tt_term(1,"query",Q),
	sol::tt_sors(1,"solutions",N2),
	feil(S1,Strings,Integers,Tree),
	!
;
    globalVars::get_i("server")<>1,
	msgS("INFERENCE ERROR"),
	tt::tt_header(1,"I N F E R E N C E   ERROR"),
	N1=globalVars::get_i("piercl"),
	N2=globalVars::get_i("solref"),
	N2<>0,
	sol::solSetHeader(N2,N3,_,_,Q,_),
	tt::tt_i(1,"Error recursion level=",N1),
	tt::tt_i(1,"Solution reference=",N2),
	tt::tt_i(1,"Solution counter=",N3),
	tt::tt_term(1,"query",Q),
	sol::tt_sors(1,"solutions",N2),
	feil(S1,Strings,Integers,Tree),
	!
;
	feil(S1,Strings,Integers,Tree),
	!.
% ------------------------------------------------------------------------------------------------------------
% msgProgress
% ------------------------------------------------------------------------------------------------------------
msgProgress(T,N,Step,Max) :-
    N=Max,
    S=string::concat(T," 100%..."),
	W=guix::getWinOfClas(1),
	vpiToolbar::setValue(W,idt_help_line,vpiToolbar::text_value(S)),
	!
;
    R=N mod Step,
    R=0,
    P=(N*100) div Max,
    Ps=toString(P),
    S=string::concat(T," ",Ps,"%..."),
	W=guix::getWinOfClas(1),
	vpiToolbar::setValue(W,idt_help_line,vpiToolbar::text_value(S)),
	!
;
    !.
% ------------------------------------------------------------------------------------------------------------
% msgStatus
% ------------------------------------------------------------------------------------------------------------
msgStatus(S) :-
	W=guix::getWinOfClas(1),
	vpiToolbar::setValue(W,idt_help_line,vpiToolbar::text_value(S)),
	!.
% ------------------------------------------------------------------------------------------------------------
% msgS; msgA
% ------------------------------------------------------------------------------------------------------------
msgOffS(S) :-
	MW=guix::getWinOfClas(4),
	C=vpi::winGetCtlHandle(MW,idc_msg),
	vpi::lboxCountAll(C)<500,
	vpi::lboxAdd(C,0,S),
	vpi::winSetFocus(C),
	vpi::winUpdate(C),
	MS=string::concat("Message: ",S),
	tt::tt_ss(1,MS),
	!
;
	MW=guix::getWinOfClas(4),
	C=vpi::winGetCtlHandle(MW,idc_msg),
	vpi::lboxDelete(C,499),
	vpi::lboxAdd(C,0,S),
	vpi::winSetFocus(C),
	vpi::winUpdate(C),
	MS=string::concat("Message: ",S),
	tt::tt_ss(1,MS),
	!
;
	!.
msgOffA(S) :-
	MW=guix::getWinOfClas(4),
	C=vpi::winGetCtlHandle(MW,idc_msg),
	S2=vpi::lboxGetItem(C,0),
	S3=string::concat(S2,S),
	vpi::lboxDelete(C,0),
	vpi::lboxAdd(C,0,S3),
	vpi::winSetFocus(C),
	vpi::winUpdate(C),
	MS=string::concat("Message: ",S),
	tt::tt_ss(1,MS),
	!
;
	!.
msgA(S) :-
	globalVars::get_i("msgon")=1,
	msgOffA(S),
	!
;
	!.	
msgS(S) :-
	globalVars::get_i("msgon")=1,
	msgOffS(S),
	!
;
	!.	
% ------------------------------------------------------------------------------------------------------------
% confCont
% ------------------------------------------------------------------------------------------------------------
confCont(N,THold,S) :-
	globalVars::get_i("quit_flag")=1,
	!,fail
;
    globalVars::get_i("server")=1,
    N=THold,
    dlgIF::dlgNote(1,"Aborted syntax analysis on re-scan threshold"),
    !,fail
;    
    globalVars::get_i("server")=-1,
	Rem=N mod THold,
	Rem=0,
	dlgExp(20),
	S1=toString(N),
	sx::concat([S1," ",S,", Do you want to quit?"],Mess),
	Reply=vpiCommonDialogs::ask("Pro/3",Mess,["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	setFlag("quit_flag",1),
    dbaStatuses,
	!
;
	!.	
% ------------------------------------------------------------------------------------------------------------
% hlpWeb
% ------------------------------------------------------------------------------------------------------------
class predicates
hlpWebX : (integer,string) procedure (i,o).
hlpCompleteFile : (string,string) determ (i,o).
clauses

hlpWeb(Fil) :-
	Fil="",
	W=vpi::winGetActiveWindow(),
	Type=guix::getWinClas(W),
	hlpWebX(Type,S),
	hlpWeb(S),
	!
;	
	Fil="",
	hlpWeb("Documentation"),
	!
;	
	globalVars::get_i("wwwhlp")=-1,
	hlpCompleteFile(Fil,H),
	B=globalVars::get_s("browser"),
	fioExt::filCheckFile(-1,"Documentation web browser",B,Br),
	NullWin=uncheckedConvert(vpiDomains::windowHandle, 0),
	_=vpi::createProcess(NullWin,Br,H,vpiDomains::wsf_Maximized),
	!
;
	globalVars::get_i("wwwhlp")=1,
	string::length(Fil)>2,
	sx::frontstr(2,Fil,FileH,_),
	FileH="w_",
	sx::concat(["http://www.ligaya.net/p3/wins/",Fil,".html"],H),
	B=globalVars::get_s("browser"),
	fioExt::filCheckFile(-1,"Documentation web browser",B,Br),
	NullWin=uncheckedConvert(vpiDomains::windowHandle, 0),
	_=vpi::createProcess(NullWin,Br,H,vpiDomains::wsf_Maximized),
	!
;
	globalVars::get_i("wwwhlp")=1,
	sx::concat(["http://www.ligaya.net/p3/",Fil,".html"],H),
	B=globalVars::get_s("browser"),
	fioExt::filCheckFile(-1,"Documentation web browser",B,Br),
	NullWin=uncheckedConvert(vpiDomains::windowHandle, 0),
	_=vpi::createProcess(NullWin,Br,H,vpiDomains::wsf_Maximized),
	!
;	
	!.

hlpNo(N) :-
    hlpWebX(N,F),
    hlpWeb(F),
    !.
    
% returns the help topic corresponding to the given window type
%	
% not supported:
% 0 edit
% 1 task
% 2 table (in table window)
% 3 table window
% 5 server window
%
% lines with comments have updated html help page
%
hlpWebX(4,"w_ke-asst") :- !. % assistant window
hlpWebX(10,"w_sentences") :- !. % sentences window
hlpWebX(11,"SentenceTypeTables") :- !. % sentences window
hlpWebX(13,"w_sentence") :- !.
hlpWebX(30,"w_kbrecs") :- !. % KB records window
hlpWebX(31,"w_kbrec") :- !. % KB record window
hlpWebX(32,"w_SentenceDerivation") :- !. % Derivation window
hlpWebX(33,"w_susp") :- !.
hlpWebX(34,"w_allabout") :- !. 
hlpWebX(35,"w_tableprint") :- !.
hlpWebX(40,"rule_editor") :- !. % rule editor window
hlpWebX(42,"w_sr") :- !. % sentence rule window
hlpWebX(46,"w_draw") :- !.
hlpWebX(60,"InexactRulesUsage") :- !.
hlpWebX(61,"InexactRuleTypeDetails") :- !.
hlpWebX(62,"w_function") :- !.
hlpWebX(63,"InexactRuleEvaluation") :- !.
hlpWebX(70,"QueriesInKB") :- !.
hlpWebX(101,"w_sm") :- !. % sentence model element
hlpWebX(102,"w_layout") :- !. 
hlpWebX(106,"w_currseg") :- !.
hlpWebX(107,"w_currterm") :- !.
hlpWebX(108,"w_syntagm") :- !.
hlpWebX(109,"w_application") :- !.
hlpWebX(110,"w_kbid") :- !.
hlpWebX(111,"w_anno") :- !.
hlpWebX(112,"w_multi") :- !. % multi-DB rederivation
hlpWebX(113,"w_external") :- !.
hlpWebX(114,"w_LookUp") :- !.
hlpWebX(115,"w_public") :- !.
hlpWebX(116,"SourceBackup") :- !.
hlpWebX(117,"w_sm") :- !.
hlpWebX(118,"w_st") :- !.
hlpWebX(119,"InexactRuleEvaluation") :- !.
hlpWebX(62,"CR_Evaluation") :- !.
% tree windows
hlpWebX(41,"rule_editor") :- !. % rule-tree window
hlpWebX(44,"InexactRulesUsage") :- !. % IR-tree window
hlpWebX(45,"InexactRulesUsage") :- !. % IR reasoning-tree window
hlpWebX(50,"w_draw") :- !. % dependency graph tree window
% grid windows
hlpWebX(201,"w_sentence") :- !. % sentence window 
hlpWebX(202,"w_tablelayout") :- !. % table layout window
hlpWebX(203,"w_table") :- !. % table window
% 204 is used for table-control inside a table window - the table window is created first 
hlpWebX(205,"w_table") :- !. % input facility window
% 206 is used for table-control inside a input facility window - the input facility window is created first 
hlpWebX(_,"Documentation") :- !.

hlpCompleteFile(Fil,CFil) :-
	Fil="",
	!, hlpCompleteFile("Documentation",CFil),
	!
;
	Dir=globalVars::get_s("exe_path"),
	sx::concat([Dir,"\\doc\\",Fil,".html"],CFil),
	file::existfile(CFil),
	!
;
	Dir=globalVars::get_s("exe_path"),
	sx::concat([Dir,"\\doc\\wins\\",Fil,".html"],CFil),
	file::existfile(CFil),
	!
;
	Dir=globalVars::get_s("exe_path"),
	sx::concat([Dir,"\\doc\\Documentation.html"],CFil),
	file::existfile(CFil),
	!.
% ------------------------------------------------------------------------------------------------------------
% setFlag
% ------------------------------------------------------------------------------------------------------------
setFlag(Flag,V) :-
	globalVars::set_i(Flag,V),
	dbaStatuses,
	!;!.
% ------------------------------------------------------------------------------------------------------------
% msgPC
% ------------------------------------------------------------------------------------------------------------
msgPC(Mess,PC) :-
	S1=toString(PC),
	sx::sxL_Pad(S1,"0",3,S2),
	sx::concat([Mess," ",S2,"%"],S),
	msgStatus(S),
	!.
end implement dlgIF
