/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement dom
open core,termDomain

constants
className = "com/rudstrand/dom/dom".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% domAdjustTerm
% ----------------------------------------------------------------------------------------------
% adjust a term to comply with its domain - fails if this cannot be done
domAdjustTerm(V,Dom,V) :-
                    Dom="p3_dany",!
;                   V=int(_),domInt(Dom),!
;			        V=fix(_),domFix(Dom),!
;			        V=str(_),domStr(Dom),!
;			        V=atom(_),domAtom(Dom),!.
domAdjustTerm(V,Dom,Va) :-	
                V=int(I),domFix(Dom),R=conversion::intToReal(I),Va=fix(R),!
;				V=int(I),domStr(Dom),S=toString(I),Va=str(S),!
;			    V=fix(R),domInt(Dom),I=conversion::realToIntT(R),Va=int(I),!
;               V=fix(R),domStr(Dom),S=toString(R),Va=str(S),!
;				V=atom(A),domStr(Dom),Va=str(A),!.
% ----------------------------------------------------------------------------------------------
% domAll
% ----------------------------------------------------------------------------------------------
domAll("p3_din").
domAll("p3_dnu").
domAll("p3_dli").
domAll("p3_dts").
domAll("p3_dif").
domAll("p3_dit").
domAll("p3_did").
domAll("p3_dsy").
domAll("p3_dcf").
domAll("p3_dmg").
domAll("p3_dinc").
domAll("p3_ddec").
domAll("p3_dpr").
domAll("p3_dany").
domAll("p3_dxt").
domAll("p3_dxd").
domAll("p3_dxdt").
domAll("p3_dxdr").
domAll("p3_din_l").
domAll("p3_dnu_l").
domAll("p3_dli_l").
domAll("p3_dts_l").
domAll("p3_dif_l").
domAll("p3_dit_l").
domAll("p3_did_l").
domAll("p3_dsy_l").
domAll("p3_dcf_l").
domAll("p3_dmg_l").
domAll("p3_dinc_l").
domAll("p3_ddec_l").
domAll("p3_dpr_l").
domAll("p3_dany_l").
domAll("p3_dxt_l").
domAll("p3_dxd_l").
domAll("p3_dxdt_l").
domAll("p3_dxdr_l").
domAllSuffix("IN").
domAllSuffix("NU").
domAllSuffix("LI").
domAllSuffix("TS").
domAllSuffix("IF").
domAllSuffix("IT").
domAllSuffix("ID").
domAllSuffix("SY").
domAllSuffix("CF").
domAllSuffix("MG").
domAllSuffix("KB").
domAllSuffix("INC").
domAllSuffix("DEC").
domAllSuffix("PR").
domAllSuffix("ANY").
domAllSuffix("XT").
domAllSuffix("XD").
domAllSuffix("XDT").
domAllSuffix("XDR").
domAllSuffix("IN_L").
domAllSuffix("NU_L").
domAllSuffix("LI_L").
domAllSuffix("TS_L").
domAllSuffix("IF_L").
domAllSuffix("IT_L").
domAllSuffix("ID_L").
domAllSuffix("SY_L").
domAllSuffix("CF_L").
domAllSuffix("MG_L").
domAllSuffix("KB_L").
domAllSuffix("INC_L").
domAllSuffix("DEC_L").
domAllSuffix("PR_L").
domAllSuffix("ANY_L").
domAllSuffix("XT_L").
domAllSuffix("XD_L").
domAllSuffix("XDT_L").
domAllSuffix("XDR_L").
% ----------------------------------------------------------------------------------------------
% domAtom/Str/Int/Fix
% ----------------------------------------------------------------------------------------------
domAtom(Dom) :-
	Dom="p3_dif",
	!.
domStr(Dom) :-
	set::member(Dom,["p3_dli","p3_dts","p3_dsy","p3_dxd","p3_dxt","p3_dxdt","p3_dxdr"]),
	!.
domInt(Dom) :-
	set::member(Dom,["p3_din","p3_dit","p3_did"]),
	!.
domFix(Dom) :-
	set::member(Dom,["p3_dnu","p3_dcf","p3_dinc","p3_ddec","p3_dpr","p3_dmg"]),
	!.
domXmlDt(Dom) :-
	set::member(Dom,["p3_dxd","p3_dxdt"]),
	!.
domNum(Dom) :-
	domInt(Dom),
	!
;
	domFix(Dom),
	!.
listDomain(Dom,Domain) :-
       sx::sxTail(2,Dom,Domain,T),
       T="_l",
       !.
% ----------------------------------------------------------------------------------------------
% domStatProc
% ----------------------------------------------------------------------------------------------
domStatProc(Dom) :-
    domNum(Dom),
    !
;
    domXmlDt(Dom),
    !.
% ----------------------------------------------------------------------------------------------
% domCompat
% ----------------------------------------------------------------------------------------------
class predicates
domSubset : (string,string_list) procedure (i,o).
domSubsetLax : (string,string_list) determ (i,o).
domSubset2 : (string,string_list) determ (i,o).
clauses
% checks domain compatibility (i.e. if Dom1=Dom2 type of assignments are valid (this is always valid when the the second domain 
% is a logical sub-set of the first domain)
%
domCompat(Dest,Src) :-
	domCompatF(Dest,Src),
	!
;	
	M=globalVars::get_i("domlax")+2,
	set::setGet(["under strict domain compatibility checking","","under relaxed domain compatibility checking"],M,Pref),
	sx::concat(["Attempt to assign ",Src," to ",Dest," (",Pref,")!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail.
domCompatF(Dest,Src) :-
	Dest=Src,
	!
;
	Dest="p3_dany",
	!
;
	globalVars::get_i("domlax")=1,
	domSubsetLax(Dest,L),
	set::member(Src,L),
	!
;
	globalVars::get_i("domlax")=-1,
	domSubset(Dest,L),
	set::member(Src,L),
	!.
	
domSubset("p3_dany",L) :- 
	domListNL(Lm),
	Lm=[_|L],
	!.
domSubset("p3_din",["p3_did"]) :- !.
domSubset("p3_dnu",["p3_din","p3_dcf","p3_dmg","p3_dpr","p3_dinc","p3_ddec","p3_dit","p3_did"]) :- !.
domSubset("p3_dli",["p3_dxd","p3_dxt","p3_dxdt","p3_dxdr","p3_dts"]) :- !.
domSubSet(_,[]) :- !.
domSubsetLax(D,L) :-
	domSubset(D,L2),
	domSubset2(D,L3),
	set::conc([L2,L3],L4),
	set::distinct(L4,L5),
	set::setRemSub([D],L5,L),
	!.
domSubset2("p3_did",["p3_din"]) :- 
	!.	
domSubset2("p3_dcf",["p3_dnu"|L]) :- 
	domSubset("p3_dnu",L2), 
	set::setRemSub(["p3_did","p3_dit"],L2,L),
	!.	
domSubset2("p3_dinc",["p3_dnu"|L]) :- 
	domSubset("p3_dnu",L2), 
	set::setRemSub(["p3_did","p3_dit"],L2,L),
	!.	
domSubset2("p3_dpr",["p3_dnu"|L]) :- 
	domSubset("p3_dnu",L2),
	set::setRemSub(["p3_din","p3_did","p3_dit","p3_dinc"],L2,L),
	!.	
domSubset2("p3_dmg",["p3_dnu"|L]) :- 
	domSubset("p3_dnu",L2),
	set::setRemSub(["p3_din","p3_did","p3_dit","p3_dinc"],L2,L),
	!.	
domSubset2("p3_ddec",["p3_dnu"|L]) :- 
	domSubset("p3_dnu",L2),
	set::setRemSub(["p3_din","p3_did","p3_dit","p3_dinc"],L2,L),
	!.	
% -------------------------------------------------------------------------------------------------------------
% domInRange
% -------------------------------------------------------------------------------------------------------------
domInRange(Dom,R) :- 	Dom="p3_dcf",!,R>=-1.0,R<=1.0,!
;			Dom="p3_dpr",!,R>=0.0,R<=1.0,!
;			Dom="p3_dmg",!,R>=0.0,R<=1.0,!
;			Dom="p3_dinc",!,R>=1.0,!
;			Dom="p3_ddec",!,R>=0.0,R<=1.0,!
;			!.
% ----------------------------------------------------------------------------------------------
% domList/NL
% ----------------------------------------------------------------------------------------------
domListNL(["p3_dany","p3_din","p3_dnu","p3_dit","p3_did","p3_dcf","p3_dmg","p3_dinc","p3_ddec","p3_dpr",
         "p3_dli","p3_dts","p3_dif","p3_dsy","p3_dxd","p3_dxt",
         "p3_dxdt","p3_dxdr"]) :- !.
domList(["p3_din","p3_dnu","p3_dit","p3_did","p3_dcf","p3_dmg","p3_dinc","p3_ddec","p3_dpr",
         "p3_dli","p3_dif","p3_dsy","p3_dany","p3_dxd","p3_dxt",
         "p3_dxdt","p3_dxdr","p3_dts",
         "p3_din_l","p3_dnu_l","p3_dit_l","p3_did_l","p3_dcf_l","p3_dmg_l","p3_dinc_l","p3_ddec_l","p3_dpr_l",
         "p3_dli_l","p3_dif_l","p3_dsy_l","p3_dany_l","p3_dxd_l","p3_dxt_l",
         "p3_dxdt_l","p3_dxdr_l","p3_dts_l"]) :-!.
% ----------------------------------------------------------------------------------------------
% nlList
% ----------------------------------------------------------------------------------------------
nlList()=L :-
    domListNL(DomsNl),
    domList(DomsAll),
    set::setRemSub(DomsNl,DomsAll,DomsL),
    dbBase::nlSings(DomsNl,NDomsNl),
    dbBase::nlSings(DomsL,NDomsL),
    set::sort(NDomsNl,NDomsNlS),
    set::sort(NDomsL,NDomsLs),
    set::conc([NDomsNlS,NDomsLs],L),
    !.
% ----------------------------------------------------------------------------------------------
% domShortLong/F
% ----------------------------------------------------------------------------------------------
domShortLong(S,L) :-
	domShortLongF(S,L),
	!
;
	bound(S), 
	free(L),
	dlgIF::feil("domShortLong",["Short",S],[],nill),!,fail
;
	bound(L), 
	free(S),
	dlgIF::feil("domShortLong",["Long",L],[],nill),!,fail.
domShortLongF(S,L) :-
	bound(S),bound(L),
	domAll(L),
	SL=string::toLowerCase(S),
	D=string::concat("p3_d",SL),
	D=L,
	!
;
	bound(S), 
	free(L),
	SL=string::toLowerCase(S),
	L=string::concat("p3_d",SL),
	domAll(L),
	!
;
	free(S),
	bound(L),
	domAll(L),
	sx::frontstr(4,L,F,T),
	F="p3_d",
	S=string::toUpperCase(T),
	!.
% ----------------------------------------------------------------------------------------------
% domValCompat/F
% ----------------------------------------------------------------------------------------------
domValCompat(Dom,Val) :-
	domValCompatF(Dom,Val),
	!
;
	termDomain::fstrSwitchR(F,Val),
	F=fstr(S,_,_,_),
	M=globalVars::get_i("domlax")+2,
	set::setGet(["strict domain compatibility checking","","relaxed domain compatibility checking"],M,Pref),
	sx::concat(["Value ",S," not compatible with ",Dom," (",Pref,")!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail.
	
domValCompatF("p3_dany",_) :- !.
domValCompatF("p3_did",int(V)) :- !,V>=0,!.
domValCompatF("p3_dit",fix(V)) :- !,V>=0.0,!.
domValCompatF("p3_dcf",int(V)) :- !,V>=-1,V<=1,!.
domValCompatF("p3_dcf",fix(V)) :- !,V>=-1.0,V<=1.0,!.
domValCompatF("p3_dmg",fix(V)) :- !,V>=0.0,V<=1.0,!.
domValCompatF("p3_dmg",int(V)) :- !,V>=0, V<=1,!.
domValCompatF("p3_dpr",fix(V)) :- !,V>=0.0,V<=1.0,!.
domValCompatF("p3_dpr",int(V)) :- !,V>=0, V<=1,!.
domValCompatF("p3_dinc",fix(V)) :- !,V>=1.0,!.
domValCompatF("p3_dinc",int(V)) :- !,V>=1,!.
domValCompatF("p3_ddec",fix(V)) :- !,V>0.0,V<1.0,!.
domValCompatF("p3_dxd",str(_)) :- !.
domValCompatF("p3_dxt",str(_)) :- !.
domValCompatF("p3_dxdt",str(_)) :- !.
domValCompatF("p3_dxdr",str(_)) :- !.
domValCompatF("p3_dxd",str(S)) :- !,xmlGeneric::xmlValidDate(S),!.
domValCompatF("p3_dxt",str(S)) :- !,xmlGeneric::xmlValidTime(S),!.
domValCompatF("p3_dxdt",str(S)) :- !,xmlGeneric::xmlValidDatetime(S),!.
domValCompatF("p3_dxdr",str(S)) :- !,xmlGeneric::xmlValidDuration(S),!.
domValCompatF(Dom,Val) :-
	termDomain::fstrSwitchR(F,Val),
	F=fstr(_,MainDom,_,_),
	Dom=MainDom,
	!
;
	termDomain::fstrSwitchR(F,Val),
	F=fstr(_,MainDom,_,_),
	domSubset(MainDom,L),
	set::member(Dom,L),
	!.
end implement dom
