/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement expsql
open core,termDomain,globalDomains

constants
className = "com/rudstrand/expsql/expql".
classVersion = "1.0".

class predicates
sqlChainTables : (integer,integer).
saveTable : (chainDB,string,integer).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% export
% ----------------------------------------------------------------------------------------------
class predicates
sqlDatabase : (string,integer,integer) determ.
sqlMainTable : (integer,string,integer) determ.
initRef : (integer,string) determ.
clauses

export(DbName,DbCreate,KbName,KbCreate,IncludeDerived) :-
    W=guix::getWinOfClas(1),
    Dlg=kbinit::new(),
    Dlg:show(W,1),
    Ps=globalVars::popRetPars(),
    Ps=["1"],   
    sqlBase::getParameters(Max,_,_,_,_,_,_,_),
    Id=globalVars::get_s("kb_id"), % exporting DB's id
    dbOpen::storeDBID("X"),
    sqlBase::openConnection,
    sqlDatabase(DbName,DbCreate,Max),
    initRef(KbCreate,KbName),
    sqlMainTable(KbCreate,KbName,Max),
    sqlChainTables(KbCreate,IncludeDerived),
    dbOpen::storeDBID(Id),
    !
;
    dlgIF::dlgNote(1,"Failed to export knowledge base to SQL!"),
    !,fail.

sqlDatabase(Name,Create,Max) :-
    string::length(Name)>Max,
    S=string::concat("Database name '",Name,"' is too long!"),
    dlgIF::dlgNote(0,S),
    !,fail
;
    Create=1,
    sqlBase::createDb(Name),
    !
;
    Create=-1,
    sqlBase::openDb(Name),
    !
;
    dlgIF::dlgNote(1,"Failed to create/use SQL database!"),
    !,fail.   

sqlMainTable(KbCreate,Name,Max) :-
    L=string::length(Name)+5,
    L>Max,
    S=string::concat("Main table name '",Name,"' is too long!"),
    dlgIF::dlgNote(0,S),
    !,fail
;
    KbCreate=1,
    sqlBase::createKb(Name),
    !
;
    KbCreate=-1,
    !.

initRef(-1,Name) :- 
    _=sqlBase::existKb(Name), 
    sqlBase::openKb(Name),
    sqlBase::setRef,
    !    
;
    S=string::concat("SQL KB ",Name," does not exist!"),
    dlgIF::dlgNote(0,S),
    !,fail. 
initRef(1,_) :- 
    sqlBase::initRef,
    !.
% ----------------------------------------------------------------------------------------------
% sqlChainTables
% ----------------------------------------------------------------------------------------------
class predicates
checkTable : (integer,string) determ.
clauses

sqlChainTables(KbCreate,IncludeDerived) :-
    KbCreate=1,
    dbaBase::database(KB,_,_,0),
    KB:db_chains(Chain),
        checkTable(KbCreate,Chain),
        saveTable(KB,Chain,IncludeDerived),
        fail
;
    KbCreate=-1,
    dbaBase::database(KB,_,_,0),
    KB:db_chains(Chain),
        not(set::member(Chain,["x","p3_pDP","p3_pSM","p3_pDR"])),
        checkTable(KbCreate,Chain),
        saveTable(KB,Chain,IncludeDerived),
        fail
;
    !.
    
checkTable(-1,_) :-
    !.
checkTable(1,Chain) :-
    sqlBase::chainDeleteOutright(Chain),
    !.
% ----------------------------------------------------------------------------------------------
% saveTable
% ----------------------------------------------------------------------------------------------
saveTable(KB,Chain,IncludeDerived) :-
    IncludeDerived=1,
                                    tt::tt_s(2,"Chain",Chain),
    KB:chain_terms(Chain,C,Ref),
        C=clause(RTy,_,SentNo,SubNo,Realm,LT,RT),
                                    dbaBase::dba_address(Ref,SRef),
                                    tt::tt_s(2,"Old ref",SRef),
                                    RTyS=toString(RTy),
                                    tt::tt_s(2,"RT",RTyS),
        sqlBase::recordInsertOutright(RTy,Realm,SentNo,SubNo,LT,RT,_),
                                    N=sqlBase::currentRef(Chain),
                                    tt::tt_i(2,"New ref",N),
                                    Ns=toString(N),
                                    Mess=string::concatList(["Exported ",Chain," ",RTyS," ",SRef," to ",Ns,"!"]),
                                    dlgIF::msgS(Mess),
        fail
;
    IncludeDerived=-1,
                                    tt::tt_s(2,"Chain",Chain),
    KB:chain_terms(Chain,C,Ref),
        C=clause(RTy,_,SentNo,SubNo,Realm,LT,RT),
        dbBase::recordType(RTy,Base,_,_),
        Base<>1,
        Base<>5,
                                    dbaBase::dba_address(Ref,SRef),
                                    tt::tt_s(2,"Old ref",SRef),
                                    RTyS=toString(RTy),
                                    tt::tt_s(2,"RT",RTyS),
        sqlBase::recordInsertOutright(RTy,Realm,SentNo,SubNo,LT,RT,_),
                                    N=sqlBase::currentRef(Chain),
                                    tt::tt_i(2,"New ref",N),
                                    Ns=toString(N),
                                    Mess=string::concatList(["Exported ",Chain," ",RTyS," ",SRef," to ",Ns,"!"]),
                                    dlgIF::msgS(Mess),
        fail
;
    !.
end implement expsql
