/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement fioExt
open core,termDomain

constants
className = "com/rudstrand/fioExt/fioExt".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% makeRealmName
% ----------------------------------------------------------------------------------------------
 makeRealmName(File)=Name:-
    N=fileName::getName(File),
    E=filename::getExtension(File),
    S2=string::concat(N,".",E),
    sx::sxRemChar(S2,' ',S3),
    S4=string::toUpperCase(S3),
    L=globalVars::get_i("max_rln"),
    sx::sxTrunc(S4,L,Name),
    !.
% ----------------------------------------------------------------------------------------------
% filDirEnd
% ----------------------------------------------------------------------------------------------
filDirEnd(Di,Dr) :-
	Di="",
	!,fail
;
	sx::sxTail(1,Di,_,T),
	T="\\",
	Dr=Di,
	!
;
	Dr=string::concat(Di,"\\"),
	!.	
% ----------------------------------------------------------------------------------------------
% filChangeType
% ----------------------------------------------------------------------------------------------
filChangeType(OldFile,NewType,NewFile) :-
	fio::file_name(OldFile,P,N,_),
	NewFile=string::concat(P,N,".",NewType),
	!.
% ------------------------------------------------------------------------------------------------------------
% setDefaultPath
% ------------------------------------------------------------------------------------------------------------
setDefaultPath(File)=PathFile :-
    P=filename::getPath(File),
    P="",
    DP=globalVars::get_s("std_path"),
    PathFile=string::concat(DP,"\\",File),
    !
;
    PathFile=File,
    !.
% ------------------------------------------------------------------------------------------------------------
% setType
% ------------------------------------------------------------------------------------------------------------
% Type may or may not have a leading '.'
%
setType(Name,Type)=FileName :-
    _=string::searchchar(Name,'.',string::caseSensitive()),
    FileName=Name,
    !
;
    string::subChar(Type,0)='.',
    FileName=string::concat(Name,Type),
    !
;
    FileName=string::concat(Name,".",Type),
    !.
% ------------------------------------------------------------------------------------------------------------
% inPathList
% ------------------------------------------------------------------------------------------------------------
% parameters:
% Path      in      a path ending with \
% PathList  in      a list of paths ending with \
% Succeeds if Path is in PathList or if a sub-directory of one of the directories in PathList
%
inPathList(Path,[P|Ps]) :-
    Path=P,
    !
;    
    L=string::length(P),
    string::length(Path)>=L,
    S=string::subString(Path,0,L),
    S=P,
    !
;    
    inPathList(Path,Ps),
    !.
% -------------------------------------------------------------------------------------------------------------
% makeFileName
% -------------------------------------------------------------------------------------------------------------
makeFileName(NL,Type,FileName) :-
    sx::makeIdentifier(1,3,2,NL,50,"",Name),
    FileName=string::concat(Name,".",Type),
    !.
% -------------------------------------------------------------------------------------------------------------
% file_type
% -------------------------------------------------------------------------------------------------------------
file_type(N,T) :- N<0,M=-N,file_type(M,T),!.
file_type(1,"3nl") :- !.	
file_type(2,"3nc") :- !.	
file_type(3,"3rp") :- !.	
file_type(4,"3rp") :- !.	
file_type(5,"3dl") :- !.	
file_type(6,"3dl") :- !.
% type 8 is table window
file_type(9,"3pr") :- !.
file_type(10,"xml") :- !.
file_type(11,"xml") :- !.
file_type(_,"333") :- !.
% -------------------------------------------------------------------------------------------------------------
% filGenName
% -------------------------------------------------------------------------------------------------------------
class predicates
filGen : (integer,string,string,string,integer,string) determ (i,i,i,i,i,o).
filOk : (integer,string) determ (i,i).
clauses

filGenName(Checks,Path,Name,Type,File) :-
    Path="",
    File=string::concat(Name,".",Type),
    filOk(Checks,File),
    !
;
    Path<>"",
    sx::sxGiveTail(Path,"\\",Path2),
    File=string::concat(Path2,Name,".",Type),
    filOk(Checks,File),
    !
;
    filGen(Checks,Path,Name,Type,1,File),
    !.
filGen(Checks,Path,Name,Type,No,File) :-
    SNo=toString(No),
    sx::sxGiveTail(Path,"\\",Path2),
    File=string::concatList([Path2,Name,"_",SNo,".",Type]),
    filOk(Checks,File),
    !
;    
    Next=No+1,
    filGen(Checks,Path,Name,Type,Next,File),
    !.
    
filOk(Check,File) :-
    Check=3,
    not(file::existFile(File)),
    not(guix::existWin(File)),
    !
;
    Check=2,
    not(guix::existWin(File)),
    !
;    
    Check=1,
    not(file::existFile(File)),
    !
;
    Check=0,
    !.
% -------------------------------------------------------------------------------------------------------------
% getPath
% -------------------------------------------------------------------------------------------------------------
getPath(Dir,File)=D:-
    Dir="xml",
    Dp=globalVars::get_s("exe_path"),
    D=string::concat(Dp,"\\doc\\xml\\",File),
    !
;
    Dir="exe",
    Dp=globalVars::get_s("exe_path"),
    D=string::concat(Dp,"\\exe\\",File),
    !
;
    Dir="doc",
    Dp=globalVars::get_s("exe_path"),
    D=string::concat(Dp,"\\doc\\",File),
    !
;
    Dir="std",
    Dp=globalVars::get_s("exe_path"),
    D=string::concat(Dp,"\\std\\",File),
    !
;
    dlgIF::feil("fioExt::getPath",["Unknown directory",Dir,File],[],nill),!,fail.
% -------------------------------------------------------------------------------------------------------------
% filValidKB_Name
% -------------------------------------------------------------------------------------------------------------
filValidKB_Name(Warn,F,Name) :-
    sqlBase::names(F,_,Name,_),
    !
;
	fio::file_name(F,_,Name,T),
	string::isname(Name),
	Tu=string::toLowerCase(T),
	Tu="3kb",
	!
;
	Warn=1,
	Mess=string::concat("'",F,"' is not a valid database name"),
	dlgIF::dlgNote(1,Mess),
	!,fail.
% ----------------------------------------------------------------------------------------------
% filCheckFile
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	KE cancels: 1=abort execution; 0=return blank file; -1=fail
% 2	in	File description
% 3	in	file (at call)
% 4	out	file (at return)
% 
filCheckFile(Context,S,Fc,Fr) :-
	file::existfile(Fc),
	Fr=Fc,
	!
;
	Context=1,
	sx::frontstr(1,Fc,D,_),
	SDir=string::concat(D,":\\"),
	Mess=string::concat(S," ",Fc," does not exist. Specify another file or location or Cancel to quit Pro/3."),
	dlgIF::dlgNote(0,Mess),
	fio::file_name(Fc,_,Fn,Typ),
	sx::concat([Fn,".",Typ],Sn),
	sx::concat([S," ",Sn],Sh),
	Fr=vpiCommonDialogs::getFileName(Sn,["",Sn],Sh,[],SDir,_),
	Fr<>"",
	!	
;
	Context<>1,
	sx::frontstr(1,Fc,D,_),
	SDir=string::concat(D,":\\"),
	sx::concat([S," ",Fc," does not exist. Specify another file or location."],Mess),
	dlgIF::dlgNote(0,Mess),
	fio::file_name(Fc,_,Fn,Typ),
	sx::concat([Fn,".",Typ],Sn),
	sx::concat([S," ",Sn],Sh),
	Fr=vpiCommonDialogs::getFileName(Sn,["",Sn],Sh,[],SDir,_),
	Fr<>"",
	!	
;
	Context=-1,
	!,fail
;
	Context=0,
	Fr="",
	!	
;
	N=globalVars::get_i("knowopen"),
	N<>0,
	Reply=vpiCommonDialogs::ask("Mandatory file is unspecified/missing","Do you want to attempt to close the KB before Pro/3 quits?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	dbOpen::closeKB(1,1,1,1),
    W=guix::getWinOfClas(1),
    _=vpi::winSendEvent(W,vpiDomains::e_CloseRequest),
	!,fail
;
	Reply=vpiCommonDialogs::ask("Mandatory file is unspecified/missing","Do you want to launch Help-web before Pro/3 quits?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
    dlgIF::hlpWeb(""),
    W=guix::getWinOfClas(1),
    _=vpi::winSendEvent(W,vpiDomains::e_CloseRequest),
	!,fail
;
    W=guix::getWinOfClas(1),
    _=vpi::winSendEvent(W,vpiDomains::e_CloseRequest),
	!,fail.	
end implement fioExt
