/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement gra
open core,termDomain,globalDomains

constants
className = "com/rudstrand/gra/gra".
classVersion = "1.0".

facts
graphChain : string := "".
graph : dirgraph := erroneous.

clauses
classInfo(className, classVersion).

predicates
removeP3F : (integer,integer).

clauses
% ----------------------------------------------------------------------------------------------
% new
% ----------------------------------------------------------------------------------------------
newGraph(Graph):-
    graphChain:=Graph,
    dep::copy(Graph,Copy),
    graph:=Copy,
    !
;
    !.
% ----------------------------------------------------------------------------------------------
% get                                                  
% ----------------------------------------------------------------------------------------------
predicates
getX : (trm).
clauses

get()=Nodes:-    
	L=trm::newList(),
	getX(L),
	Nodes=L:get(),
	!.

getX(L) :-
	graph:get_nd(A1,A2,B1,B2),
        L:add(cmp(graphChain,[cmp("n",[A1,A2]),cmp("n",[B1,B2])])),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% getType
% ----------------------------------------------------------------------------------------------
predicates
getTypeX : (lst,integer).
clauses

getType(Type)=Nodes :-
    L=lst::new(),
    getTypeX(L,Type),
    Nodes=L:get(),
    !.
 
getTypeX(L,Type) :-
	graph:get_nd(A1,A2,_,_),
        A1=int(Type),
        A2=str(Node),
        L:add(Node),
        fail
;
	graph:get_nd(_,_,B1,B2),
        B1=int(Type),
        B2=str(Node),
        L:add(Node),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% changeType
% ----------------------------------------------------------------------------------------------
changeType(From,To) :- graph:changeType(int(From),int(To)),!.
% ----------------------------------------------------------------------------------------------
% filter
% ----------------------------------------------------------------------------------------------
predicates
filterX : (integer,integer) determ.
clauses

% removes elements of the graph according to the selection flags (1=include -1=exclude)
%
filter(ST,SR,CR,FU,XP3F) :-
    removeP3F(XP3F,FU),
	filterX(1,ST),
	filterX(2,ST),
	filterX(3,SR),
	filterX(5,SR),
	filterX(15,CR),
	filterX(19,FU),
	!.

filterX(_,1) :- !.
filterX(Type,-1) :- graph:subMergeType(int(Type)),!.
% ----------------------------------------------------------------------------------------------
% removeP3F                                          
% ----------------------------------------------------------------------------------------------
predicates
removeP3F_X : ().
hasP3F : () determ.
clauses

% removes Pro/3 functions from the graph if opted for	
%	
removeP3F(FU,XP3F) :-
    FU=-1, 
    % All functions will be removed
    !
;    
    XP3F=1,
    hasP3F,
    removeP3F_X,
    removeP3F(FU,XP3F), % recursive removal required
    !
;
    !.
    
removeP3F_X() :-    
    graph:get_nd(A1,A2,_,_),
        A1=int(19),
        A2=str(Fu),
        sx::frontstr(4,Fu,F,_),
        F="p3_f",
        graph:subMergeP(A1,str(Fu)),
        fail
;
    graph:get_nd(_,_,A1,A2),
        A1=int(19),
        A2=str(Fu),
        sx::frontstr(4,Fu,F,_),
        F="p3_f",
        graph:subMergeP(A1,str(Fu)),
        fail
;
    !.
    
hasP3F() :-
    graph:has(int(19),A2,_,_),
    A2=str(Fu),
	sx::frontstr(4,Fu,F,_),
	F="p3_f",
    !
;    
    graph:has(_,_,int(19),B2),
    B2=str(Fu),
	sx::frontstr(4,Fu,F,_),
	F="p3_f",
	!.
% ----------------------------------------------------------------------------------------------
% getSGs/STs
% ----------------------------------------------------------------------------------------------
predicates
getSGs : () -> string_list.
getSTs : (string) -> string_list.
getSGsX : (lst).
getSTsX : (lst,string).
clauses

getSGs()=SGs :-
    L=lst::new(),
    getSGsX(L),
    SGs=L:get(),
    !.

getSTs(SG)=STs :-
    L=lst::new(),
    getSTsX(L,SG),
    STs=L:get(),
    !.

getSTsX(L,SG) :-
    graph:get_nd(A1,A2,_,_),
        A1=int(2),
        A2=str(N),
        sx::sxUnpack(N,'|',L2),
        L2=[P|_],
        SG=P,
        L:addUnique(N),
        fail
;
    graph:get_nd(_,_,A1,A2),
        A1=int(2),
        A2=str(N),
        sx::sxUnpack(N,'|',L2),
        L2=[P|_],
        SG=P,
        L:addUnique(N),
        fail
;
    !.
    
getSGsX(L) :-
    graph:get_nd(A1,A2,_,_),
        A1=int(1),
        A2=str(N),
        L:addUnique(N),
        fail
;
    graph:get_nd(_,_,A1,A2),
        A1=int(1),
        A2=str(N),
        L:addUnique(N),
        fail
;
    !.
% ----------------------------------------------------------------------------------------------
% dtNT_C                                      
% ----------------------------------------------------------------------------------------------
dtNT_C(-1,_,_,"") :- 
	!.
dtNT_C(_,N,_,"") :- 
	N>2,
	!.
dtNT_C(1,1,SG,S) :-
	dbExtension::kbSG_Count(SG,N1,N2),
	S1=toString(N1),
	S2=toString(N2),
	sx::concat([" [",S2,"+",S1,"]"],S),
	!.
dtNT_C(1,2,ST,S) :-
	dbBase::stCount(ST,N1,N2),
	S1=toString(N1),
	S2=toString(N2),
	sx::concat([" [",S2,"+",S1,"]"],S),
	!.
% ----------------------------------------------------------------------------------------------
% graTree                                                            
% ----------------------------------------------------------------------------------------------
class predicates
graInsSubTrees : (sterml,integer,sterm,integer,sterml) determ (i,i,i,i,o).
graInsSubTreesX : (sterml,integer,sterml,integer,sterml) determ (i,i,i,i,o).
graTreeX : (sterm,sterml,sterm) determ (i,i,o).
graSubNodes : (sterml,integer,sterm,sterml) determ (i,i,i,o).
clauses

% the generated tree has the format
%
% a tree	:	cmp(Node,[Sub-trees])   or
%                       str(Node)
% a node	:	"Type|Name"	Type:1,2,3,5,15,19; Name=node name in dep.graph
% a sub-tree	:	list of trees
%
% parameters:
% 1	in	graph (as a list)
% 2	in	1=tree of determinants; 2=tree of dependents
% 3	in	root node of tree
% 4	in	no. of levels in tree (incl. root i.e. 2 is minimum); 0:all levels
% 5	out	tree
%
graTree(Graphs,TreeType,Root,0,Tree) :- 
	graTree(Graphs,TreeType,Root,32000,Tree),
	!.
graTree(_,_,Root,1,str(S)) :-
	p3Tree::dtConvNodeR(S,Root), 
	!.
graTree(Graphs,TreeType,Root,Levels,Tree) :-
	RemLevels=Levels-1,
	graInsSubTrees(Graphs,TreeType,Root,RemLevels,Trees),
	graTreeX(Root,Trees,Tree),
	!.
graTreeX(Root,[],str(S)) :- 
	p3Tree::dtConvNodeR(S,Root), 
	!.
graTreeX(Root,Trees,cmp(S,Trees)) :- 
	p3Tree::dtConvNodeR(S,Root), 
	!.
graInsSubTrees(Graphs,TreeType,Node,Levels,Trees) :-
	graSubNodes(Graphs,TreeType,Node,Nodes),
	graInsSubTreesX(Graphs,TreeType,Nodes,Levels,Trees),
	!
;
	dlgIF::feil("graInsSubTrees",[],[TreeType,Levels],Node),!,fail.
		
graInsSubTreesX(_,_,[],_,[]) :- 
	!.
graInsSubTreesX(Graphs,TreeType,[Node|Nodes],Levels,[Tree|Trees]) :-
	graTree(Graphs,TreeType,Node,Levels,Tree),
	graInsSubTreesX(Graphs,TreeType,Nodes,Levels,Trees),
	!
;
	dlgIF::feil("graInsSubTreesX",[],[TreeType,Levels],Node),!,fail.
	
% parameters:
% 1	in	graph as list
% 2	in	1=dets; 2=deps
% 3	in	node
% 4	in	array representing all direct dets or deps of node
%	
graSubNodes([],_,_,[]) :-	
	!.
graSubNodes([G|Gs],TreeType,N,[Node|Nodes]) :-	
	TreeType=1, % determinants
	N=cmp(_,[int(T),str(S)]),
	G=cmp(_,[cmp(H2,[int(T2),str(S2)]),cmp(_,[int(T3),str(S3)])]),
	T=T3,
	S=S3,
	Node=cmp(H2,[int(T2),str(S2)]),
	graSubNodes(Gs,TreeType,N,Nodes),
	!
;	
	TreeType=2, % dependents
	N=cmp(_,[int(T),str(S)]),
	G=cmp(_,[cmp(_,[int(T2),str(S2)]),cmp(H3,[int(T3),str(S3)])]),
	T=T2,
	S=S2,
	Node=cmp(H3,[int(T3),str(S3)]),
	graSubNodes(Gs,TreeType,N,Nodes),
	!.
graSubNodes([_|Gs],TreeType,N,Nodes) :-	
	graSubNodes(Gs,TreeType,N,Nodes),
	!.
end implement gra
