/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement ir
open core,termDomain,globalDomains

constants
className = "com/rudstrand/ir/ir".
classVersion = "1.0".

class facts
break : integer:=0. % 0:no break set, 1=user break, 2=timeout break
rootRule : (integer,string,string,string,string,sterm) nondeterm. 
	% 1	sol.ref.
	% 2	ST (k)
	% 3	Argument (k)
	% 4	DE (k)
	% 5	root rule (k) (from v.3.11)
	% 6	actual parameters to root rule (from v.3.11)
loadedRule : (string,string,real,sterm,sterm,sterml,sterml,sterm,integer,sterm,real) nondeterm.
evaluation : (string,string,real,sterm,sterm,sterml,sterml,sterm,integer,sterm,real) nondeterm.
	%       evaluation(RuleName,Type,Prior,OutContext,Default,Pars,APars,Query,Eval,CF,PF)
	% 1	rulename
	% 2	type
	% 3	type=p3_iC*: prior, else: 0
	% 4	type=p3_iC*: out of context cf, else: nill
	% 5	p3_iCdat/swi/map: default value, else: nill 
	% 6	formal parameters: [fstr(DE,...),...]
	% 7	actual parameters: [Val,...] e.g. [atom(DEx),...]
	% 8	p3_iCdat:			cmp(QUERY/QUESTION,[atom(RuleReturnDomain),atom(QueryReturnDomain),Query/Question)])
	%       p3_iCmap: 			cmp(RDom,[Map])
	%       p3_iCswi/call: 			atom(RDom)
	%       p3_iCswi/par: 			cmp(RDom,[atom(DE)])
	%	p3_iCbay/and/or/not/cmb/par:	atom(RDom)
	%	p3_iZ*:				atom(RDom)
	% 9	evaluated: 0=no, 1=yes
	% 10 returned value
	% 11 if returned value is a cf, then the corresponding probability
	% 	   else if the returned value is a numeric, then the numeric itself
	%	   else 0
calledRule : (string,string,integer,string,sterm,sterm,string,string,sterml) nondeterm.
	% calledRule(Typ,Rule,N,CRule,Inc,Dec,OpA,OpB,APars)
	% 1	call-type:C=con L=and/or/cmb N=not R=generic call S=swi X=crit M=map PAR/CAL	
	%			B=bay I/IC=inc D/DC=dec
	%	for rules of type p3_iCpar:	PAR R C
	%	for rules of type p3_iZ* (except Zowa) :	Z
	%	for rules of type p3_iZowa: OWA W/WC	  		
	% 2	calling rulename (I/D: bay-rule; IC/DC: arg-rule)
	% 3	call no (for given 1-3)
	% 4	called rulename
	% 5	switch/context: value-1
	%	min (fix(C)); inc (fix(B)); operand-1 (can be fix/int/str/atom(S))
	%	PAR:	var(V)
	%       OWA:	fix(R) or atom(S) in case of weight rule
	% 6	switch/context: value-2 or nill
	%	max (fix(C)); dec (fix(B)); operand-2 (           do.            )
	%	PAR,W:	nill
	% 7	switch/context:	operator 1 (S)
	% 8	switch/context: operator 2 (S)
	% 9	call parameter assignments: [cmp(EqOpr,[fvar(FPar,Dom),RSide]),...]
	%    which states how actual parameters are determined 
	%    e.g. cmp(isIF,[fvar(DEa,"pd_if"),atom(DEb)])
numericMap : (integer,string,real_list,real_list) nondeterm.
	% 1	sol.ref.
	% 2	rulename
	% 3	values
	% 4	mapped values
discreteMap : (integer,string,sterml,sterml) nondeterm.
	% 1	sol.ref.
	% 2	rulename
	% 3	values
	% 4	mapped values
qCluster : (string CallingRule,string Prompt,string_list ClusteredRules,string_list SPars,sterml APars,integer Called) nondeterm.
% CallingRule: rule where input facility is to be activated
% Prompt: prompt text possibly with embedded parampeters
% ClusteredRules: list of rules in the cluster
% SPars: parameters to the facility itself. these must be a (sub)set of the CallingRule's parameters, i.e. CallingRule-parameter's actual value is used
% APars: actual parameters for which the facility has been executed
% Called: 0=never called, 1=executed
qSeries : (string CallingRule,string Prompt,string QstRule,string QstPar,string Format,string Query,string_list SPars,sterml APars,integer Called) nondeterm.
% CallingRule: rule where input facility is to be activated
% Prompt: prompt text possibly with embedded parampeters
% QstRule: related question rule
% QstPar: the parameter name (DE KB-name) embedded in the question text
% Format: p3_iPR or p3_iNL (query text's format)
% Query: query text for the NDERF-query used to fetch actual values (the series) for QstPar
% SPars: parameters to the facility itself. these must be a (sub)set of the CallingRule's parameters, i.e. CallingRule-parameter's actual value is used
% APars: actual parameters for which the facility has been executed
% Called: 0=never called, 1=executed
qAnswer : (string QstRule,string Prompt,string Answer) nondeterm.
% QstRule : question rule
% Prompt : fully resolved question text (with all embedded parameters replaced)
% Answer : answer to the question
%
class predicates
checkInputFac : (string,sterml,sterml).

evaluate : (integer,integer,string,sterml,vpiTree::tree) determ (i,i,i,i,o).
evaluateContext : (integer,integer,string,integer,sterml,sterml,vpiTree::tree_list,integer) procedure (i,i,i,i,i,i,o,o).
evaluateCalledRules : (integer,integer,string,sterml,sterml,vpiTree::tree_list) procedure (i,i,i,i,i,o).
evaluateRule : (integer,integer,string,string,sterml,sterm,real,vpiTree::tree) determ (i,i,i,i,i,o,o,o).

loadSubnet : (integer,string) determ (i,i).
loadRuleCalls : (string,integer,string,integer,sterml) procedure (i,i,i,i,i).
loadInputFac : ().

crtDom : (string,string) determ (i,o).
crtEvalBay : (integer,integer,string,sterml,sterml,real,integer,real) procedure (i,i,i,i,i,i,i,o).
crtEvalOWA : (integer,integer,string,sterml,sterml,real) determ (i,i,i,i,i,o).
crtEvalLog : (string,integer,integer,string,sterml,sterml,integer,real,real) determ (i,i,i,i,i,i,i,o,o).
crtEvalFZ : (string,integer,integer,string,sterml,sterml,integer,real) determ (i,i,i,i,i,i,i,o).
crtEvalSwi : (integer,integer,string,sterml,sterml,sterm,vpiTree::tree) determ (i,i,i,i,i,o,o).
crtEvalParRule : (integer,integer,string,sterml,sterml,sterm,vpiTree::tree) determ (i,i,i,i,i,o,o).
crtLoadMap : (integer,string,sterm) determ (i,i,i).
crtMap : (integer,string,sterm,sterm) determ (i,i,i,o).
crtEvalMapRule : (integer,string,sterml,sterml,sterm,vpiTree::tree) determ (i,i,i,i,o,o).
crtQueryRun : (integer,integer,string,sterml,sterml,sterm,sterm,vpiTree::tree) determ (i,i,i,i,i,i,o,o).
crtQuestionAsk : (integer,string,sterml,string,string,sterm,vpiTree::tree) determ (i,i,i,i,i,o,o).
crtReasonRec : (string,sterml,string,sterm,sterm,string) procedure (i,i,i,i,i,i).
		
runQuestionSeries : (string,string,sterm,string,string,termSet,string,string) determ.
runQuestionCluster : (string,string_list,string_list,sterml) determ.

clauses
classInfo(className, classVersion).
clauses
% ----------------------------------------------------------------------------------------------
% evaluateNetwork               
% ----------------------------------------------------------------------------------------------
class predicates
evaluateNetwork2 : (integer,integer,string,string,integer,string,string,sterml,sterml,sterm) determ (i,i,i,i,i,i,i,i,i,o).
crtCreateRulesCopy : ().
crtInitInstance : (sterm) determ (i).
clauses

evaluateNetwork(SolRef,SolNo,ST,Arg,ArgNo,DE,Rule,Pars,APars,Ct) :-
        break=0,
        evaluateNetwork2(SolRef,SolNo,ST,Arg,ArgNo,DE,Rule,Pars,APars,Ct),
        !.
    
evaluateNetwork2(SolRef,SolNo,ST,Arg,ArgNo,DE,Rule,Pars,APars,Ct) :-
	rootRule(SolRef,ST,Arg,DE,_,_),
	!,
	% rulenetwork is already loaded - initialize new net instance
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(1,"CR - new rule-net instance",Sc),
	termDomain::listBR(AParsL,APars),
	crtInitInstance(AParsL),
	globalVars::set_i("crcnt",0),
	% evaluate
	evaluate(SolRef,SolNo,Rule,APars,Tree),
	evaluation(Rule,_,_,_,_,_,APars,_,_,Ct,Pr),
				tt::tt_s(1,"CR (conclusion)",Sc),
				tt::tt_term(1,"Value 1",Ct),
				tt::tt_n(1,"Value 2",Pr),
	irbase::crtSaveTree(ST,Arg,DE,APars,Tree),
	!
;
	% load network
	crtRetract,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(1,"CR-net loading",Sc),
	loadSubnet(SolRef,Rule),
	loadInputFac,
	termDomain::listBR(AParsL,APars),
	assert(rootRule(SolRef,ST,Arg,DE,Rule,AParsL)),
	crtCreateRulesCopy,
	evaluateNetwork2(SolRef,SolNo,ST,Arg,ArgNo,DE,Rule,Pars,APars,Ct),
	!.
       
crtCreateRulesCopy() :-
	evaluation(A,B,C,D,E,F,G,H,I,J,K),
	assert(loadedRule(A,B,C,D,E,F,G,H,I,J,K)),
	fail
;
	!.
	
% initiates the rule net for given sol.no. from the loadedRule(..) facts, however all evaluated data rules
% are retained (to avoid repeated questions - and also repeated queries). these data rules will in
% any case be re-used only if parameters match. note that the retained data rules then will have at least
% two facts in evaluation(..)
%
crtInitInstance(AParsRoot) :-
	evaluation(Rule,Type,Prior,OC,Def,Pars,APars,Q,Eval,CF,PF),
	Type<>"p3_iCdat",
	retract(evaluation(Rule,Type,Prior,OC,Def,Pars,APars,Q,Eval,CF,PF)),
	fail
;
	evaluation(Rule,Type,Prior,OC,Def,Pars,APars,Q,Eval,CF,PF),
	Eval=0,
	retract(evaluation(Rule,Type,Prior,OC,Def,Pars,APars,Q,Eval,CF,PF)),
	fail
;
	loadedRule(A,B,C,D,E,F,G,H,I,J,K),
	assert(evaluation(A,B,C,D,E,F,G,H,I,J,K)),
	fail
;	
	rootRule(SolRef,ST,Arg,DE,Rule,_),
	retractall(rootRule(_,_,_,_,_,_)),
	assert(rootRule(SolRef,ST,Arg,DE,Rule,AParsRoot)),
	!.	
% ----------------------------------------------------------------------------------------------
% evaluate             
% ----------------------------------------------------------------------------------------------
class predicates
crtCheckEval : (string,sterm) determ (i,i).
crtEvalCon : (string,sterm,string,sterm,sterm,integer) procedure (i,i,i,i,i,o).
crtHasContext : (string) determ (i).
crtRevert : (integer,vpiTree::tree_list,vpiTree::tree_list) determ (i,i,o).	
makeEvaluation : (integer,vpiTree::tree_list,integer,integer,string,sterml,vpiTree::tree) determ (i,i,i,i,i,i,o).
clauses

% evaluates the sub-network (and actual parameters) with given Rule as root (if not yet evaluated), or else if evaluated, returns the evaluation tree
%
evaluate(SolRef,SolNo,Rule,ParVals,Tree) :-
    break>0,
	Tree=vpiTree::tree("** user break **",vpiTree::unmarked,[],vpiTree::tree_DirRight),
    !
;    
	% rule has not yet been called with given actual parameters
	not(evaluation(Rule,_,_,_,_,_,ParVals,_,_,_,_)),
	evaluation(Rule,Type,Prior,OC,Def,Pars,_,Q,_,_,_),
	assert(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,fix(0.0),0.0)),
	evaluate(SolRef,SolNo,Rule,ParVals,Tree),
	!
;	
	% rule is already evaluated for the given actual parameters
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,1,V,_),
	irbase::crtReasonTree(atom(Rule),V,"",Pars,ParVals,[],Tree),
	!
;
	% rule has been called, but not yet valuated for, the given actual parameters - check if out of context
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,0,_,_),
	crtHasContext(Rule),
	!,
	evaluateContext(SolRef,SolNo,Rule,1,Pars,ParVals,ConTrees,InContext),
	break=0,
                    profileTime::start("makeEvaluation"),
	makeEvaluation(InContext,ConTrees,SolRef,SolNo,Rule,ParVals,Tree),
                    profileTime::stop("makeEvaluation"),
	!
;
                    profileTime::start("makeEvaluation"),
	makeEvaluation(-1,[],SolRef,SolNo,Rule,ParVals,Tree),
                    profileTime::stop("makeEvaluation"),
	!.

% parameters:
% 1 in -1:no context, 0:out of context, 1:in context
%
makeEvaluation(InContext,ConTrees,SolRef,SolNo,Rule,ParVals,Tree) :-
	InContext=0, % out of context
	!,
	evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px),
    crtEvalProbability(Rule,ParVals,OC,OP),
	retract(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px)),
	assert(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,1,OC,OP)),
	crtCheckEval(Rule,OC),
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR out of context",Sc),
				tt::tt_term(3,"CR out of context value",OC),
	crtDom(Rule,Dom),				
	crtReasonRec(Rule,ParVals,Dom,OC,nill,"p3_iOutc"),
	irbase::crtReasonTree(atom(Rule),OC,"*",Pars,ParVals,ConTrees,Tree),
	!
;
	evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px),
	Type="p3_iCswi",
	calledRule("X",Rule,_,_,_,_,_,_,_),
	!, % call-type switch rule has been called, but not yet evaluated for the given actual parameters
	checkInputFac(Rule,Pars,ParVals),
                    profileTime::start("evaluateCalledRules"),
	evaluateCalledRules(SolRef,SolNo,Rule,Pars,ParVals,T2s), % criterion
                    profileTime::stop("evaluateCalledRules"),
	% evaluate the rule
	evaluateRule(SolRef,SolNo,Type,Rule,ParVals,C,P,Tswi),	
	% mark as evaluated
	retract(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px)),
	assert(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,1,C,P)),
	crtCheckEval(Rule,C),
	irbase::crtTreePrefixes("CRIT::",T2s,T3s),
	set::conc([ConTrees,T3s],SubTrees),
	irbase::crtReasonTreeAdd(Tswi,SubTrees,Tree),	
	!
;
	evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px),
	Type="p3_iCpar",
	!, % parameterized rule has been called, but not yet evaluated for the given actual parameters
	checkInputFac(Rule,Pars,ParVals),
                    profileTime::start("evaluateCalledRules"),
	evaluateCalledRules(SolRef,SolNo,Rule,Pars,ParVals,T2s), % parameter calls
                    profileTime::stop("evaluateCalledRules"),
	% evaluate the rule
	evaluateRule(SolRef,SolNo,Type,Rule,ParVals,C,P,Tpar),	
	% mark as evaluated
	retract(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px)),
	assert(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,1,C,P)),
	crtCheckEval(Rule,C),
	irbase::crtTreePrefixes("PAR::",T2s,T3s),
	set::conc([ConTrees,T3s,[Tpar]],SubTrees),
	irbase::crtReasonTree(atom(Rule),C,"",Pars,ParVals,[],Tmain),
	irbase::crtReasonTreeAdd(Tmain,SubTrees,Tree),	
	!
;
	evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px),
	% rule has been called, but not yet evaluated for the given actual parameters
	checkInputFac(Rule,Pars,ParVals),
                    profileTime::start("evaluateCalledRules"),
	evaluateCalledRules(SolRef,SolNo,Rule,Pars,ParVals,T2s), 
                    profileTime::stop("evaluateCalledRules"),
	% evaluate the rule
	evaluateRule(SolRef,SolNo,Type,Rule,ParVals,C,P,T2),	
	% mark as evaluated
	retract(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,0,Cx,Px)),
	assert(evaluation(Rule,Type,Prior,OC,Def,Pars,ParVals,Q,1,C,P)),
	crtCheckEval(Rule,C),
	set::conc([ConTrees,T2s],SubTrees),
	irbase::crtReasonTreeAdd(T2,SubTrees,Tree),
	!.

crtHasContext(Rule) :-
	calledRule("C",Rule,_,_,_,_,_,_,_),
	!.
% ----------------------------------------------------------------------------------------------
% evaluateContext             
% ----------------------------------------------------------------------------------------------
class predicates
evaluateContextX : (integer,integer,string,integer,sterml,sterml,integer,vpiTree::tree,vpiTree::tree_list,integer) determ (i,i,i,i,i,i,i,i,o,o).
clauses
% evaluates context (all context rules called by Rule)
% parameters:
% 1	in	sol.ref.
% 2	in	sol.no.
% 3	in	rule
% 4	in	context rule no
% 5	in	calling rule's formal parameters
% 6	in	calling rule's actual parameter values
% 7	out	reasoning tree
% 8	out	in context=1, else 0
%
evaluateContext(SolRef,SolNo,Rule,N,Pars,ParVals,Trees,InContext) :-
	calledRule("C",Rule,N,CRule,Op1,Op2,Opn1,Opn2,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,Tree),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,Val,_),
	crtEvalCon(Opn1,Op1,Opn2,Op2,Val,InCon),
	Next=N+1,
	evaluateContextX(SolRef,SolNo,Rule,Next,Pars,ParVals,InCon,Tree,TreesM,InContext),
	crtRevert(InContext,TreesM,Trees),
	!.
evaluateContext(_,_,_,_,_,_,[],1) :-
	!.
	
evaluateContextX(SolRef,SolNo,Rule,N,Pars,ParVals,InCon,Tree,Trees,InContext) :-
	InCon=0,
	irbase::crtTreePrefix("OUT OF CONTEXT::",Tree,TreeM),
	Trees=[TreeM],
	InContext=0,
	!
;
	InCon=1,
	irbase::crtTreePrefix("IN CONTEXT::",Tree,TreeM),
	Next=N+1,
	evaluateContext(SolRef,SolNo,Rule,Next,Pars,ParVals,TreesM,InContext),
	Trees=[TreeM|TreesM],
	!.

crtEvalCon(Opn1,Op1,Opn2,Op2,Val,InCon) :-
	irbase::evaluateCondition(Opn1,Op1,Opn2,Op2,Val),
	InCon=1,
	!
;
	InCon=0,
	!.

crtRevert(1,Trees,Trees) :-
	!.
crtRevert(0,Trees,[H]) :-
	set::setRev(Trees,TreesR),
	TreesR=[H|_],
	!.

crtCheckEval(Rule,Val) :-
	crtDom(Rule,Dom),
	Dom="p3_dpr",
	dom::domValCompat("p3_dcf",Val),
	!
;
	crtDom(Rule,Dom),
	Dom<>"p3_dpr",
	dom::domValCompat(Dom,Val),
	!
;
	crtDom(Rule,Dom),
	dlgIF::feil_pie("crtCheckEval",["Invalid inexact rule return value",Rule,Dom],[],Val),!,fail
;
	dlgIF::feil_pie("crtCheckEval",["Cannot find inexact rule",Rule],[],Val),!,fail.	
% ----------------------------------------------------------------------------------------------
% evaluateCalledRules
% ----------------------------------------------------------------------------------------------
class predicates
evaluateCalledRulesX : (integer,integer,string,string,integer,sterml,sterml,vpiTree::tree_list) procedure (i,i,i,i,i,i,i,o).
crtEvalCall_IC : (integer,integer,string,string,sterml,sterml,vpiTree::tree_list) determ (i,i,i,i,i,i,o).
crtEvalCall_W : (integer,integer,string,string,sterml,sterml,vpiTree::tree_list) determ (i,i,i,i,i,i,o).
clauses

% evaluates all rules called by Rule except:
% - context rules (which eventually are evaluated prior to the rule itself)
% - branches in switch rules (which are evaluated after the appropriate branch has been determined)
% - parameterized calls in parameter rules (which are evaluated after the parameter-calls have determined the actual parameters)
% the following call types are used:
	% 1	call-type	C=con	L=and/or/cmb	N=not	R=generic call	S=swi	X=crit	M=map
	%			B=bay	I/IC=inc	D/DC=dec
	%	p3_iCpar:	PAR R C
	%	p3_iZ* (except Zowa) :	Z
	%	p3_iZowa: OWA W/WC	  		
% parameters:
% 1	in	sol.ref.
% 2	in	sol.no.
% 3	in	rule
% 4	in	calling rule's formal parameters
% 5	in	calling rule's actual parameter values
% 6	out	rule trees
%
evaluateCalledRules(SolRef,SolNo,Rule,Pars,ParVals,Trees) :-
	evaluateCalledRulesX(SolRef,SolNo,Rule,"L",1,Pars,ParVals,T3s),
	evaluateCalledRulesX(SolRef,SolNo,Rule,"N",1,Pars,ParVals,T4s),
	evaluateCalledRulesX(SolRef,SolNo,Rule,"B",1,Pars,ParVals,T2s),
	evaluateCalledRulesX(SolRef,SolNo,Rule,"X",1,Pars,ParVals,T5s),
	evaluateCalledRulesX(SolRef,SolNo,Rule,"M",1,Pars,ParVals,T6s),
	evaluateCalledRulesX(SolRef,SolNo,Rule,"PAR",1,Pars,ParVals,T7s),
	evaluateCalledRulesX(SolRef,SolNo,Rule,"Z",1,Pars,ParVals,T8s),
	evaluateCalledRulesX(SolRef,SolNo,Rule,"OWA",1,Pars,ParVals,T9s),
	set::conc([T2s,T3s,T4s,T5s,T6s,T7s,T8s,T9s],Trees),
	!.
evaluateCalledRulesX(_,_,_,_,_,_,_,[]) :-
    break>0,
    !.
evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,N,Pars,ParVals,[T]) :-
	set::member(CallType,["N","X","M"]),
	calledRule(CallType,Rule,N,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,T),
	!.	
evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,N,Pars,ParVals,[T|Ts]) :-
	set::member(CallType,["L","Z","PAR"]),
	calledRule(CallType,Rule,N,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,T),
	Next=N+1,
	evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,Next,Pars,ParVals,Ts),
	!.
evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,N,Pars,ParVals,Trees) :-
	CallType="B",
	calledRule(CallType,Rule,N,CRule,_,_,_,_,CPars),
	crtEvalCall_IC(SolRef,SolNo,Rule,CRule,Pars,ParVals,Tid),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,Trule),
	irbase::crtReasonTreeAdd(Trule,Tid,T),
	Next=N+1,
	evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,Next,Pars,ParVals,Ts),
	Trees=[T|Ts],
	!
;
	CallType="B",
	calledRule(CallType,Rule,N,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,T),
	Next=N+1,
	evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,Next,Pars,ParVals,Ts),
	Trees=[T|Ts],
	!
;
	CallType="OWA",
	calledRule(CallType,Rule,N,CRule,_,_,_,_,CPars),
	crtEvalCall_W(SolRef,SolNo,Rule,CRule,Pars,ParVals,Tid),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,Trule),
	irbase::crtReasonTreeAdd(Trule,Tid,T),
	Next=N+1,
	evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,Next,Pars,ParVals,Ts),
	Trees=[T|Ts],
	!
;
	CallType="OWA",
	calledRule(CallType,Rule,N,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,T),
	Next=N+1,
	evaluateCalledRulesX(SolRef,SolNo,Rule,CallType,Next,Pars,ParVals,Ts),
	Trees=[T|Ts],
	!.
evaluateCalledRulesX(_,_,_,_,_,_,_,[]) :-
	!.

crtEvalCall_IC(SolRef,SolNo,Rule,CRule,Pars,ParVals,Trees) :-
	calledRule("I",Rule,_,IRule,_,_,_,_,CPars),
	calledRule("IC",CRule,_,IRule,_,_,_,_,CPars),
	calledRule("D",Rule,_,DRule,_,_,_,_,CPars2),
	calledRule("DC",CRule,_,DRule,_,_,_,_,CPars2),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,IRule,CParVals,T2),
	irbase::crtSubstActPars(CPars2,Pars,ParVals,CParVals2),
	evaluate(SolRef,SolNo,DRule,CParVals2,T3),
	irbase::crtTreePrefix("INC::",T2,T2M),
	irbase::crtTreePrefix("DEC::",T3,T3M),
	Trees=[T2M,T3M],
	!
;
	calledRule("I",Rule,_,IRule,_,_,_,_,CPars),
	calledRule("IC",CRule,_,IRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,IRule,CParVals,T2),
	irbase::crtTreePrefix("INC::",T2,T2M),
	Trees=[T2M],
	!
;
	calledRule("D",Rule,_,DRule,_,_,_,_,CPars2),
	calledRule("DC",CRule,_,DRule,_,_,_,_,CPars2),
	irbase::crtSubstActPars(CPars2,Pars,ParVals,CParVals2),
	evaluate(SolRef,SolNo,DRule,CParVals2,T3),
	irbase::crtTreePrefix("INC::",T3,T3M),
	Trees=[T3M],
	!.

crtEvalCall_W(SolRef,SolNo,Rule,CRule,Pars,ParVals,Trees) :-
	calledRule("W",Rule,_,DRule,_,_,_,_,CPars2),
	calledRule("WC",CRule,_,DRule,_,_,_,_,CPars2),
	irbase::crtSubstActPars(CPars2,Pars,ParVals,CParVals2),
	evaluate(SolRef,SolNo,DRule,CParVals2,T3),
	irbase::crtTreePrefix("WEIGHT::",T3,T3M),
	Trees=[T3M],
	!.
% ----------------------------------------------------------------------------------------------
% evaluateRule          
% ----------------------------------------------------------------------------------------------
class predicates
evaluateRuleX : (integer,integer,string,string,sterml,sterm,vpiTree::tree) determ (i,i,i,i,i,o,o).
crtEvalProbability : (string,sterml,sterm,real) procedure (i,i,i,o).
clauses

% evaluates an unevaluated rule - only called after all called rules have been evaluated
%
evaluateRule(SolRef,SolNo,Type,Rule,ParVals,C,P,Tree) :-
    break=0,
                    profileTime::start("evaluateRule"),
	evaluateRuleX(SolRef,SolNo,Type,Rule,ParVals,C,Tree),
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"IR result of evaluation of rule",Sc),
				tt::tt_term(3,"IR value",C),
	crtEvalProbability(Rule,ParVals,C,P),
                    profileTime::stop("evaluateRule"),
	!.
	
crtEvalProbability(Rule,ParVals,V,P) :-	
	crtDom(Rule,Dom),
	Dom="p3_dcf",
	termDomain::realType(V,C),
	evaluation(Rule,_,Prior,_,_,_,ParVals,_,_,_,_),
	irbase::certtoprob(C,Prior,P),
	!
;
	crtDom(Rule,Dom),
	Dom="p3_dpr",
	termDomain::realType(V,P),
	!
;
	V=fix(Pf),
	P=Pf,
	!
;
	V=int(Pi),
	P=Pi,
	!
;
	P=0.0,
	!.
	
evaluateRuleX(SolRef,SolNo,Type,Rule,ParVals,V,Tree) :-
	Type="p3_iCdat",
	evaluation(Rule,_,_,_,_,Pars,ParVals,cmp("p3_iQry",[_,_,Query]),_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR query rule evaluation",Sc),
				nl::conQuery(Query,QNL),
				tt::tt_s(3,"CR query",QNL),
	crtQueryRun(SolRef,SolNo,Rule,Pars,ParVals,Query,V,Tree),
	!
;
	Type="p3_iCdat",
	evaluation(Rule,_,_,_,_,Pars,ParVals,cmp("p3_iQst",[_,atom(QDom),str(Quest)]),_,_,_),
	Pos=string::search(Quest,"?",string::caseSensitive),
	Posn=Pos+1,
	string::front(Quest,Posn,Qf,Qt),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR question rule evaluation",Sc),
				irBase::wcrQuestionAltsToNL(Qt,Qtn),
	irbase::setQuestionPars(Rule,Qf,Pars,ParVals,Qfn),
    Qn=string::concat(Qfn," ",Qtn),   
				tt::tt_s(3,"CR question",Qn),
	crtQuestionAsk(SolRef,Rule,ParVals,QDom,Qn,V,Tree),
	!
;
	Type="p3_iCdat",
	evaluation(Rule,_,_,_,_,Pars,ParVals,cmp("p3_iQst",[_,atom(QDom),str(Quest)]),_,_,_),
	not(_=string::search(Quest,"?",string::caseSensitive)),
	!, % no question mark, thus no answer alternatives as per syntax rule
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR question rule evaluation",Sc),
				tt::tt_s(3,"CR question",Quest),
	irbase::setQuestionPars(Rule,Quest,Pars,ParVals,Qn2),
	Qn3=string::trim(Qn2),
	sx::sxGiveTail(Qn3,"?",Qn),
				tt::tt_s(3,"CR question",Qn),
	crtQuestionAsk(SolRef,Rule,ParVals,QDom,Qn,V,Tree),
	!
;
	Type="p3_iCmap",
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR map rule evaluation",Sc),
	crtEvalMapRule(SolRef,Rule,Pars,ParVals,V,Tree),
	!
;
	Type="p3_iCbay",
	crtDom(Rule,RDom),
	RDom="p3_dcf",
	evaluation(Rule,_,Prior,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR bayesian rule evaluation",Sc),
	crtEvalBay(SolRef,SolNo,Rule,Pars,ParVals,Prior,1,P),
	irbase::probtocert(P,Prior,C),
	V=fix(C),
	crtReasonRec(Rule,ParVals,"p3_dcf",V,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),V,"|: BAY",Pars,ParVals,[],Tree),
	!
;
	Type="p3_iCbay",
	crtDom(Rule,RDom),
	RDom="p3_dpr",
	evaluation(Rule,_,Prior,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR bayesian rule evaluation",Sc),
	crtEvalBay(SolRef,SolNo,Rule,Pars,ParVals,Prior,1,P),
	V=fix(P),
	crtReasonRec(Rule,ParVals,"p3_dpr",V,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),V,"|: BAY",Pars,ParVals,[],Tree),
	!
;
	Type="p3_iCswi",
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR switch rule evaluation",Sc),
	crtEvalSwi(SolRef,SolNo,Rule,Pars,ParVals,V,Tree),
	!
;
	Type="p3_iCpar",
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR parameterized rule evaluation",Sc),
	crtEvalParRule(SolRef,SolNo,Rule,Pars,ParVals,V,Tree),
	!
;
	crtDom(Rule,RDom),
	RDom="p3_dcf",
	set::member(Type,["p3_iCand","p3_iCor","p3_iCnot","p3_iCcmb"]),
	set::setGet(["p3_iCand","p3_iCor","p3_iCnot","p3_iCcmb"],TypeNo,Type),
	set::setGet(["|: AND","|: OR","|: NOT","|: CMB"],TypeNo,Log),
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR logical combination rule evaluation",Sc),
	crtEvalLog(Type,SolRef,SolNo,Rule,Pars,ParVals,1,C,_),
	V=fix(C),
	crtReasonRec(Rule,ParVals,"p3_dcf",V,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),V,Log,Pars,ParVals,[],Tree),
	!
;
	crtDom(Rule,RDom),
	RDom="p3_dpr",
	set::member(Type,["p3_iCand","p3_iCor","p3_iCnot","p3_iCcmb"]),
	set::setGet(["p3_iCand","p3_iCor","p3_iCnot","p3_iCcmb"],TypeNo,Type),
	set::setGet(["|: AND","|: OR","|: NOT","|: CMB"],TypeNo,Log),
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"CR logical combination rule evaluation",Sc),
	crtEvalLog(Type,SolRef,SolNo,Rule,Pars,ParVals,1,_,P),
	V=fix(P),
	crtReasonRec(Rule,ParVals,"p3_dpr",V,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),V,Log,Pars,ParVals,[],Tree),
	!
;
	Type="p3_iZowa",
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"FZ OWA rule evaluation",Sc),
	crtEvalOWA(SolRef,SolNo,Rule,Pars,ParVals,G),
	V=fix(G),
	crtReasonRec(Rule,ParVals,"p3_dmg",V,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),V,"|: OWA",Pars,ParVals,[],Tree),
	!
;
	set::member(Type,["p3_iZsand","p3_iZaand","p3_iZband","p3_iZdand","p3_iZsor","p3_iZaor","p3_iZbor","p3_iZdor","p3_iZnot"]),
	set::setGet(["p3_iZsand","p3_iZaand","p3_iZband","p3_iZdand","p3_iZsor","p3_iZaor","p3_iZbor","p3_iZdor","p3_iZnot"],TypeNo,Type),
	set::setGet(["::IS","::ISa","::ISb","::ISd","::UN","::UNa","::UNb","::UNd","::CMPL"],TypeNo,Log),
	evaluation(Rule,_,_,_,_,Pars,ParVals,_,_,_,_),
	!,
				Sa=toString(SolRef),
				Sb=toString(SolNo),
				sx::concat([Sa,"|",Sb,"|",Rule],Sc),
				tt::tt_s(3,"FZ combination rule evaluation",Sc),
	crtEvalFZ(Type,SolRef,SolNo,Rule,Pars,ParVals,1,G),
	V=fix(G),
	crtReasonRec(Rule,ParVals,"p3_dmg",V,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),V,Log,Pars,ParVals,[],Tree),
	!
;
    break=0,
	dlgIF::feil_pie("evaluateRuleX",[Type,Rule],[SolRef,SolNo],cmp("",ParVals)),
	!,fail.	
% ----------------------------------------------------------------------------------------------
% crtEvalBay   
% ----------------------------------------------------------------------------------------------
class predicates
crtEvalInc : (string,sterm,string,string,sterml,sterml,real) determ (i,i,i,i,i,i,o).
clauses

crtEvalBay(SolRef,SolNo,Rule,Pars,ParVals,Prior,No,P) :-
	calledRule("B",Rule,No,CRule,VInc,VDec,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,Prior2,_,_,_,CParVals,_,1,_,P2),
	crtEvalInc("I",VInc,Rule,CRule,Pars,ParVals,Inc),
	crtEvalInc("D",VDec,Rule,CRule,Pars,ParVals,Dec),
	Next=No+1,
	crtEvalBay(SolRef,SolNo,Rule,Pars,ParVals,Prior,Next,P3),
					Sa=toString(SolRef),
					Sb=toString(SolNo),
					sx::concat([Sa,"|",Sb,"|",CRule],Sc),
					tt::tt_s(3,"In Bayesian combination",Sc),
	irbase::crtCombineBay(Prior2,P2,Inc,Dec,P3,P), 
	!
;
	P=Prior,
	!.	
	
crtEvalInc(_,fix(F),_,_,_,_,F) :- 
	!.
crtEvalInc(IncDec,atom(IncDecRule),BayRule,ArgRule,Pars,ParVals,V) :-
	calledRule(IncDec,BayRule,_,IncDecRule,_,_,_,_,CPars),
	CC=string::concat(IncDec,"C"),
	calledRule(CC,ArgRule,_,IncDecRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(IncDecRule,_,_,_,_,_,CParVals,_,1,ID,_),
	termDomain::realType(ID,V),
	!
;
	dlgIF::feil_pie("crtEvalInc",[IncDec,IncDecRule,BayRule,ArgRule],[],cmp("",[cmp("",Pars),cmp("",ParVals)])),!,fail.
% ----------------------------------------------------------------------------------------------
% crtEvalOWA
% ----------------------------------------------------------------------------------------------
class predicates
crtEvalW : (string,sterm,string,string,sterml,sterml,real) determ (i,i,i,i,i,i,o).
crtEvalOWA_Weights : (integer,integer,string,sterml,sterml,integer,real) procedure (i,i,i,i,i,i,o).
crtEvalOWA_MG : (integer,integer,string,sterml,sterml,integer,real,real) procedure (i,i,i,i,i,i,i,o).
clauses

crtEvalOWA(SolRef,SolNo,Rule,Pars,ParVals,MGrade) :-
	crtEvalOWA_Weights(SolRef,SolNo,Rule,Pars,ParVals,1,SumW),
	SumW>0.0,
	crtEvalOWA_MG(SolRef,SolNo,Rule,Pars,ParVals,1,SumW,MGrade),
	!
;
	crtEvalOWA_Weights(SolRef,SolNo,Rule,Pars,ParVals,1,_),
	sx::concat(["Sum of weights in OWA fuzzy set rule '",Rule,"' zero or negative!"],Mess),
	dlgIF::dlgNote(1,Mess),
	dlgIF::feil_pie("crtEvalOWA",[Mess],[],nill),!,fail.

crtEvalOWA_Weights(SolRef,SolNo,Rule,Pars,ParVals,No,SumW) :-
	calledRule("OWA",Rule,No,CRule,Weight,_,_,_,_),
	crtEvalW("W",Weight,Rule,CRule,Pars,ParVals,W),
	Next=No+1,
	crtEvalOWA_Weights(SolRef,SolNo,Rule,Pars,ParVals,Next,W2),
	SumW=W+W2,
	!
;
	SumW=0.0,
	!.	

crtEvalOWA_MG(SolRef,SolNo,Rule,Pars,ParVals,No,SumWeights,MGrade) :-
	calledRule("OWA",Rule,No,CRule,Weight,_,_,_,CPars),
	crtEvalW("W",Weight,Rule,CRule,Pars,ParVals,W), % W is the actual weight factor
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,MG),
	Next=No+1,
	crtEvalOWA_MG(SolRef,SolNo,Rule,Pars,ParVals,Next,SumWeights,MG2),
					Sa=toString(SolRef),
					Sb=toString(SolNo),
					sx::concat([Sa,"|",Sb,"|",CRule],Sc),
					tt::tt_s(3,"In OWA combination",Sc),
	MGrade=MG2+(W/SumWeights)*MG,
	!
;
	MGrade=0.0,
	!.	

% returns the OWA-weight (W) as a real whether by constant or by call
%
crtEvalW(_,fix(W),_,_,_,_,W) :- 
	!.
crtEvalW(CT,atom(WRule),OWA_Rule,ArgRule,Pars,ParVals,W) :-
	calledRule(CT,OWA_Rule,_,WRule,_,_,_,_,CPars),
	CC=string::concat(CT,"C"),
	calledRule(CC,ArgRule,_,WRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(WRule,_,_,_,_,_,CParVals,_,1,Val,_),
	termDomain::realType(Val,W),
	!
;
	dlgIF::feil_pie("crtEvalW",[WRule,OWA_Rule,ArgRule],[],cmp("",[cmp("",Pars),cmp("",ParVals)])),!,fail.
% ----------------------------------------------------------------------------------------------
% crtEvalFZ
% ----------------------------------------------------------------------------------------------
class predicates
crtDrastic : (real,real,real,real) determ (i,i,i,o).
clauses

crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,No,G) :-
	Opr="p3_iZnot",
	calledRule("Z",Rule,No,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	G=1.0-Gc,
	!
;
	Opr="p3_iZsand",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	set::setMin([Gc,G2],G),
	!
;
	Opr="p3_iZaand",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	G=Gc*G2,
	!
;
	Opr="p3_iZband",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	G3=Gc+G2-1.0,
	set::setMax([G3,0.0],G),
	!
;
	Opr="p3_iZdand",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	crtDrastic(1.0,Gc,G2,G),
	!
;
	Opr="p3_iZsor",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	set::setMax([Gc,G2],G),
	!
;
	Opr="p3_iZaor",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	G=Gc+G2-Gc*G2,
	!
;
	Opr="p3_iZbor",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	G3=Gc+G2,
	set::setMin([G3,1.0],G),
	!
;
	Opr="p3_iZdor",
	calledRule("Z",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("Z",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,Gc),
	crtEvalFZ(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,G2),
	crtDrastic(0.0,Gc,G2,G),
	!
;
	% the last call
	calledRule("Z",Rule,No,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,G),
	!
;
	dlgIF::feil_pie("crtEvalFZ",[Opr,Rule],[SolRef,SolNo,No],cmp("",ParVals)),!,fail.

crtDrastic(D,Fa,Fb,C) :-
	Fa=D,
	C=Fb,
	!
;
	Fb=D,
	C=Fa,
	!
;
	D=1.0,
	C=0.0,
	!
;
	D=0.0,
	C=1.0,
	!.
% ----------------------------------------------------------------------------------------------
% crtEvalLog
% ----------------------------------------------------------------------------------------------
crtEvalLog(Opr,SolRef,SolNo,Rule,Pars,ParVals,No,C,P) :-
	Opr="p3_iCnot",
	calledRule("N",Rule,No,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,PR),
	P=1.0-PR,
	evaluation(Rule,_,Prior,_,_,_,_,_,_,_,_),
	irbase::probtocert(P,Prior,C),
	!
;
	Opr="p3_iCcmb",
	crtDom(Rule,RDom),
	RDom="p3_dcf",
	calledRule("L",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("L",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,CF,_),
	termDomain::realType(CF,C2),
	crtEvalLog(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,C3,_),
	irbase::crtCombine(C2,C3,C),
	evaluation(Rule,_,Prior,_,_,_,_,_,_,_,_),
	irbase::certtoprob(C,Prior,P),
	!
;
	Opr="p3_iCand",
	crtDom(Rule,RDom),
	RDom="p3_dcf",
	calledRule("L",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("L",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,CF,_),
	termDomain::realType(CF,C2),
	crtEvalLog(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,C3,_),
	set::setMin([C2,C3],C),
	evaluation(Rule,_,Prior,_,_,_,_,_,_,_,_),
	irbase::certtoprob(C,Prior,P),
	!
;
	Opr="p3_iCand",
	crtDom(Rule,RDom),
	RDom="p3_dpr",
	calledRule("L",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("L",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,PR),
	crtEvalLog(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,_,PR3),
	set::setMin([PR,PR3],P),
	evaluation(Rule,_,Prior,_,_,_,_,_,_,_,_),
	irbase::probtocert(P,Prior,C),
	!
;
	Opr="p3_iCor",
	crtDom(Rule,RDom),
	RDom="p3_dcf",
	calledRule("L",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("L",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,CF,_),
	termDomain::realType(CF,C2),
	crtEvalLog(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,C3,_),
	set::setMax([C2,C3],C),
	evaluation(Rule,_,Prior,_,_,_,_,_,_,_,_),
	irbase::certtoprob(C,Prior,P),
	!
;
	Opr="p3_iCor",
	crtDom(Rule,RDom),
	RDom="p3_dpr",
	calledRule("L",Rule,No,CRule,_,_,_,_,_),
	Next=No+1,
	calledRule("L",Rule,Next,_,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,_,PR),
	crtEvalLog(Opr,SolRef,SolNo,Rule,Pars,ParVals,Next,_,PR3),
	set::setMax([PR,PR3],P),
	evaluation(Rule,_,Prior,_,_,_,_,_,_,_,_),
	irbase::probtocert(P,Prior,C),
	!
;
	% the last call
	set::member(Opr,["p3_iCor","p3_iCand","p3_iCcmb"]),
	calledRule("L",Rule,No,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,CF,P),
	termDomain::realType(CF,C),
	!
;
	dlgIF::feil_pie("crtEvalLog",[Opr,Rule],[SolRef,SolNo,No],cmp("",ParVals)),!,fail.
% ----------------------------------------------------------------------------------------------
% crtEvalMapRule
% ----------------------------------------------------------------------------------------------
crtEvalMapRule(SolRef,Rule,Pars,ParVals,C,Tree) :-
	calledRule("M",Rule,1,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,Val,_),
	crtMap(SolRef,Rule,Val,C),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,ParVals,Dom,C,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),C,"|: MAP",Pars,ParVals,[],Tree),
	!
;	
	% mapping failed
	calledRule("M",Rule,1,CRule,_,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,Def,_,CParVals,_,1,_,_),
	C=Def,
	crtDom(Rule,Dom),
	crtReasonRec(Rule,ParVals,Dom,C,nill,"p3_iDeft"),
	irbase::crtReasonTree(atom(Rule),C,"+|: MAP",Pars,ParVals,[],Tree),
	!.
% ----------------------------------------------------------------------------------------------
% crtEvalParRule
% ----------------------------------------------------------------------------------------------
class predicates
crtEvalPR_GetPars : (integer,integer,string,sterml,sterml,integer,sterml,sterml) procedure (i,i,i,i,i,i,o,o).
clauses

crtEvalParRule(SolRef,SolNo,Rule,Pars,ParVals,C,Tree) :-
	% get the variables used as parameters and their values
	crtEvalPR_GetPars(SolRef,SolNo,Rule,Pars,ParVals,1,RPars,RVals),
	% find the main rule
	calledRule("R",Rule,_,CRule,_,_,_,_,CParAss),
	set::conc([Pars,RPars],AllPars),
	set::conc([ParVals,RVals],AllVals),
	irbase::crtSubstActPars(CParAss,AllPars,AllVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,Tree),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,C,_),
	!.

crtEvalPR_GetPars(SolRef,SolNo,Rule,Pars,ParVals,No,[RPar|RPars],[RVal|RVals]) :-
	calledRule("PAR",Rule,No,CRule,RPar,_,_,_,CPars),
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,RVal,_),
	Next=No+1,
	crtEvalPR_GetPars(SolRef,SolNo,Rule,Pars,ParVals,Next,RPars,RVals),
	!.
crtEvalPR_GetPars(_,_,_,_,_,_,[],[]) :-
	!.
% ----------------------------------------------------------------------------------------------
% crtEvalSwi
% ----------------------------------------------------------------------------------------------
class predicates
crtEvalFindBranch : (integer,integer,string,sterml,sterml,integer,sterm,sterm,vpiTree::tree) determ (i,i,i,i,i,i,i,o,o).
crtEvalFindBranchDef : (integer,integer,string,sterml,sterml,sterm,vpiTree::tree) determ (i,i,i,i,i,o,o).
crtGetParVal : (string,sterml,sterml,sterm) determ (i,i,i,o).
clauses

crtEvalSwi(SolRef,SolNo,Rule,Pars,ParVals,C,Tree) :-
	evaluation(Rule,_,_,_,_,_,_,cmp(_,[atom(DE)]),_,_,_),
	crtGetParVal(DE,Pars,ParVals,Val),
	crtEvalFindBranch(SolRef,SolNo,Rule,Pars,ParVals,1,Val,C,Tree2),
	Tree2=vpiTree::tree(Head2,Mark2,[vpiTree::tree(Head3,Mark3,Trees3,No3)],No2),
	dbBase::nlSing(DE,DEn),
	sx::concat(["criterion parameter ",DEn],S),
	irbase::crtReasonTree(str(S),Val,"|P",[],[],[],Tree3),
	Tree=vpiTree::tree(Head2,Mark2,[vpiTree::tree(Head3,Mark3,[Tree3|Trees3],No3)],No2),
	!
;
	evaluation(Rule,_,_,_,Def,_,_,cmp(_,[atom(DE)]),_,_,_),
	% parameter-switch failed - use default
	crtEvalFindBranch(SolRef,SolNo,Rule,Pars,ParVals,1,Def,C,Tree2),
	Tree2=vpiTree::tree(Head2,Mark2,[vpiTree::tree(Head3,Mark3,Trees3,No3)],No2),
	dbBase::nlSing(DE,DEn),
	sx::concat(["PARAMETER: ",DEn],S),
	irbase::crtReasonTree(str(S),Def,"+|P",[],[],[],Tree3),
	Tree=vpiTree::tree(Head2,Mark2,[vpiTree::tree(Head3,Mark3,[Tree3|Trees3],No3)],No2),
	!
;
	evaluation(Rule,_,_,_,_,_,_,atom(_),_,_,_),
	% rule-switch
	calledRule("X",Rule,_,CRule,_,_,_,_,CParExps),
	irbase::crtSubstActPars(CParExps,Pars,ParVals,CParVals),
	evaluation(CRule,_,_,_,_,_,CParVals,_,1,Cx,_),
	crtEvalFindBranch(SolRef,SolNo,Rule,Pars,ParVals,1,Cx,C,Tree),
	!	
;
	evaluation(Rule,_,_,_,Def,_,_,atom(_),_,_,_),
	% rule-switch failed - use default-value to switch
	crtEvalFindBranch(SolRef,SolNo,Rule,Pars,ParVals,1,Def,C,Tree),
	!	
;
	dlgIF::feil_pie("crtEvalSwi",["Switch-rule without default-branch and with no matching branch for default value",Rule],[SolRef,SolNo],nill),!,fail.

crtEvalFindBranch(SolRef,SolNo,Rule,Pars,ParVals,No,Cin,C,Tree) :-
	not(calledRule("S",Rule,No,_,_,_,_,_,_)),
	crtEvalFindBranchDef(SolRef,SolNo,Rule,Pars,ParVals,C,T2),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,ParVals,Dom,C,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),C,"|: SWI",Pars,ParVals,[T2],Tree),
	!
;
	not(calledRule("S",Rule,No,_,_,_,_,_,_)),
	!,fail
;
	calledRule("S",Rule,No,CRule,Op1,Op2,Opn1,Opn2,CPars),
	Opn1<>"",
	irbase::evaluateCondition(Opn1,Op1,Opn2,Op2,Cin),
	% this is the appropriate switch-branch
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,T2),
	evaluation(CRule,_,_,_,_,_,CParValsX,_,1,C,_),
	irbase::crtMatchPars(CParVals,CParValsX),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,ParVals,Dom,C,nill,"p3_iEval"),
	irbase::crtReasonTree(atom(Rule),C,"|: SWI",Pars,ParVals,[T2],Tree),
	!
;
	Next=No+1,
	crtEvalFindBranch(SolRef,SolNo,Rule,Pars,ParVals,Next,Cin,C,Tree),
	!.
	
crtEvalFindBranchDef(SolRef,SolNo,Rule,Pars,ParVals,C,Tree) :-
	calledRule("S",Rule,_,CRule,_,_,"","",CPars), % default branch
	irbase::crtSubstActPars(CPars,Pars,ParVals,CParVals),
	evaluate(SolRef,SolNo,CRule,CParVals,Tree),
	evaluation(CRule,_,_,_,_,_,CParValsX,_,1,C,_),
	irbase::crtMatchPars(CParVals,CParValsX),
	irbase::crtMatchPars(ParVals,CParVals),
	!.
crtGetParVal(DE,[],_,_) :-
	dlgIF::feil_pie("crtGetParVal",["Cannot find switch criterion parameter",DE],[],nill),!,fail.
crtGetParVal(DE,[fstr(D,_,_,_)|Pars],[PV|ParVals],Val) :-
	DE=D,
	Val=PV,
	!
;
	crtGetParVal(DE,Pars,ParVals,Val),
	!.
% ----------------------------------------------------------------------------------------------
% crtMakeTree
% ----------------------------------------------------------------------------------------------
class predicates
crtMakeTreeSwi : (string,sterm,string,sterm,string) determ (i,i,i,i,o).
crtMakeTreeX : (string,string,string,integer,string,vpiTree::tree) determ (i,i,i,i,i,o).	
crtMakeTreeXs : (string,string,integer,string,integer,vpiTree::tree_list) procedure (i,i,i,i,i,o).	
crtNodeName : (string,string,string,string,integer,string) determ (i,i,i,i,o,o).
crtSubTree : (string,string,string,string,vpiTree::tree_list,vpiTree::tree) determ (i,i,i,i,i,o).
assName : (string,string,string) procedure (i,i,o).
clauses

% converts the inexact rule network in the database and starting with given Rule to tree-
% package tree
%
crtMakeTree(Rule,AssPar,Tree) :-
	evaluation(Rule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX("",Type,Rule,AssPar,"",Tree),
	!.

crtMakeTreeX(Prefix,Type,Rule,AssPar,Suffix,Tree) :-
	Type="p3_iCbay",
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts6),
	crtMakeTreeXs("","B",1,Rule,AssPar,Ts2),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts6,Ts2],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"::BAY",Ts,Tree),
	!
;
	Type="p3_iCbay",
	crtMakeTreeXs("","B",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,"::BAY",Ts,Tree),
	!
;
	Type="p3_iCswi",
	evaluation(Rule,_,_,_,_,_,_,cmp(_,[atom(DE)]),_,_,_),
	dbBase::nlSing(DE,_),
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts6),
	crtMakeTreeXs("","S",1,Rule,AssPar,Ts3),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts6,Ts3],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"::SWI",Ts,Tree),
	!
;
	Type="p3_iCswi",
	evaluation(Rule,_,_,_,_,_,_,cmp(_,[atom(DE)]),_,_,_),
	dbBase::nlSing(DE,_),
	crtMakeTreeXs("","S",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,"::SWI",Ts,Tree),
	!
;
	Type="p3_iCswi",
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts6),
	crtMakeTreeXs("CRIT::","X",1,Rule,AssPar,Ts2),
	crtMakeTreeXs("","S",1,Rule,AssPar,Ts3),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts6,Ts2,Ts3],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"::SWI",Ts,Tree),
	!
;
	Type="p3_iCswi",
	crtMakeTreeXs("CRIT::","X",1,Rule,AssPar,Ts2),
	crtMakeTreeXs("","S",1,Rule,AssPar,Ts3),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts2,Ts3],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"::SWI",Ts,Tree),
	!
;
	Type="p3_iCpar",
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts6),
	crtMakeTreeXs("","R",1,Rule,AssPar,Ts2),
	crtMakeTreeXs("","PAR",1,Rule,AssPar,Ts3),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts6,Ts3,Ts2],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"",Ts,Tree),
	!
;
	Type="p3_iCpar",
	crtMakeTreeXs("","R",1,Rule,AssPar,Ts2),
	crtMakeTreeXs("","PAR",1,Rule,AssPar,Ts3),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts3,Ts2],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"",Ts,Tree),
	!
;
	Type="p3_iCnot",
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts6),
	crtMakeTreeXs("","N",1,Rule,AssPar,Ts2),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts6,Ts2],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"::NOT",Ts,Tree),
	!
;
	Type="p3_iCnot",
	crtMakeTreeXs("","N",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,"::NOT",Ts,Tree),
	!
;
	Type="p3_iCmap",
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts6),
	crtMakeTreeXs("","M",1,Rule,AssPar,Ts2),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	set::conc([Ts6,Ts2],Ts),
	crtSubTree(Prefix,NRuleA,Suffix,"::MAP",Ts,Tree),
	!
;
	Type="p3_iCmap",
	crtMakeTreeXs("","M",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,"::MAP",Ts,Tree),
	!
;
	set::member(Type,["p3_iCand","p3_iCor","p3_iCcmb"]),
	set::setGet(["p3_iCand","p3_iCor","p3_iCcmb"],Nt,Type),
	set::setGet(["::AND","::OR","::COMB"],Nt,RTyp),
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts6),
	crtMakeTreeXs("","L",1,Rule,AssPar,Ts2),
	dbBase::nlSing(Rule,NRule),
	set::conc([Ts6,Ts2],Ts),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,RTyp,Ts,Tree),
	!
;
	set::member(Type,["p3_iCand","p3_iCor","p3_iCcmb"]),
	set::setGet(["p3_iCand","p3_iCor","p3_iCcmb"],Nt,Type),
	set::setGet(["::AND","::OR","::COMB"],Nt,RTyp),
	crtMakeTreeXs("","L",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,RTyp,Ts,Tree),
	!
;
	set::member(Type,["p3_iZsand","p3_iZaand","p3_iZband","p3_iZdand","p3_iZsor","p3_iZaor","p3_iZbor","p3_iZdor","p3_iZnot"]),
	set::setGet(["p3_iZsand","p3_iZaand","p3_iZband","p3_iZdand","p3_iZsor","p3_iZaor","p3_iZbor","p3_iZdor","p3_iZnot"],TypeNo,Type),
	set::setGet(["::IS","::ISa","::ISb","::ISd","::UN","::UNa","::UNb","::UNd","::CMPL"],TypeNo,RTyp),
	crtMakeTreeXs("","Z",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,RTyp,Ts,Tree),
	!
;
	Type="p3_iZowa",
	crtMakeTreeXs("","OWA",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix,"::OWA",Ts,Tree),
	!
;
	Type="p3_iCdat",
	crtHasContext(Rule),
	crtMakeTreeXs("CONTEXT::","C",1,Rule,AssPar,Ts),
	dbBase::nlSing(Rule,NRule),
	dbBase::irType(Rule,_,Typ),
	set::setGet(["p3_iCqry","p3_iCqst"],Ix,Typ),
	set::setGet(["::Q","::?"],Ix,Suffix2),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	crtSubTree(Prefix,NRuleA,Suffix2,"",Ts,Tree),
	!
;
	Type="p3_iCdat",
	dbBase::nlSing(Rule,NRule),
	inexgui::getAssociation(Rule,AssPar,Ass),
	assName(NRule,Ass,NRuleA),
	dbBase::irType(Rule,_,Typ),
	set::setGet(["p3_iCqry","p3_iCqst"],Ix,Typ),
	set::setGet(["::Q","::?"],Ix,Suffix2),
	crtSubTree(Prefix,NRuleA,Suffix2,"",[],Tree),
	!
;
	dlgIF::feil("crtMakeTreeX",[Type,Rule,Suffix],[],nill),!,fail.	

crtMakeTreeXs(Prefix,"N",No,Rule,AssPar,[Tree|Lt]) :-
	calledRule("N",Rule,No,CRule,_,_,_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Tree),
	Next=No+1,		
	crtMakeTreeXs(Prefix,"N",Next,Rule,AssPar,Lt),
	!.
crtMakeTreeXs(Prefix,"B",No,Rule,AssPar,[Tree|Lt]) :-
	calledRule("B",Rule,No,CRule,fix(_),fix(_),_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Tree),
	Next=No+1,		
	crtMakeTreeXs(Prefix,"B",Next,Rule,AssPar,Lt),
	!
;	
	calledRule("B",Rule,No,CRule,fix(_),atom(DRule),_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Trule),
	evaluation(DRule,DType,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,DType,DRule,AssPar,"",Tdec),
	Trule=vpiTree::tree(AA,BB,CC,DD),
	Tree=vpiTree::tree(AA,BB,[Tdec|CC],DD),
	Next=No+1,		
	crtMakeTreeXs(Prefix,"B",Next,Rule,AssPar,Lt),
	!
;	
	calledRule("B",Rule,No,CRule,atom(IRule),fix(_),_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Trule),
	evaluation(IRule,IType,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,IType,IRule,AssPar,"",Tinc), 
	Trule=vpiTree::tree(AA,BB,CC,DD),
	Tree=vpiTree::tree(AA,BB,[Tinc|CC],DD),
	Next=No+1,		
	crtMakeTreeXs(Prefix,"B",Next,Rule,AssPar,Lt),
	!
;	
	calledRule("B",Rule,No,CRule,atom(IRule),atom(DRule),_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Trule),
	evaluation(IRule,IType,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,IType,IRule,AssPar,"",Tinc),
	evaluation(DRule,DType,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,DType,DRule,AssPar,"",Tdec),
	Trule=vpiTree::tree(AA,BB,CC,DD),
	Tree=vpiTree::tree(AA,BB,[Tinc,Tdec|CC],DD),
	Next=No+1,		
	crtMakeTreeXs(Prefix,"B",Next,Rule,AssPar,Lt),
	!.
crtMakeTreeXs(Prefix,"OWA",No,Rule,AssPar,[Tree|Lt]) :-
	calledRule("OWA",Rule,No,CRule,fix(Wg),_,_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	Sw=toString(Wg),
	Pref=string::concat(Sw,"::"),
	crtMakeTreeX(Pref,Type,CRule,AssPar,"",Tree),
	Next=No+1,		
	crtMakeTreeXs(Prefix,"OWA",Next,Rule,AssPar,Lt),
	!
;	
	calledRule("OWA",Rule,No,CRule,atom(DRule),_,_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Trule),
	evaluation(DRule,DType,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX("WEIGHT::",DType,DRule,AssPar,"",Tdec),
	Trule=vpiTree::tree(AA,BB,CC,DD),
	Tree=vpiTree::tree(AA,BB,[Tdec|CC],DD),
	Next=No+1,		
	crtMakeTreeXs(Prefix,"OWA",Next,Rule,AssPar,Lt),
	!.
crtMakeTreeXs(Prefix,CallType,No,Rule,AssPar,[Tree|Lt]) :-
	set::member(CallType,["S","C"]),
	calledRule(CallType,Rule,No,CRule,Inc,Dec,Op1,Op2,_),
	crtMakeTreeSwi(Op1,Inc,Op2,Dec,Crit),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	sx::sxRemoveTail(Prefix,"::",Prefix2),
	Prefix3=string::concat(Prefix2,Crit),
	sx::sxGiveTail(Prefix3,"::",Prefix4),
	crtMakeTreeX(Prefix4,Type,CRule,AssPar,"",Tree),
	Next=No+1,		
	crtMakeTreeXs(Prefix,CallType,Next,Rule,AssPar,Lt),
	!.
crtMakeTreeXs(Prefix,CT,No,Rule,AssPar,[Tree|Lt]) :-
	set::member(CT,["L","Z"]),
	calledRule(CT,Rule,No,CRule,_,_,_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Tree),
	Next=No+1,		
	crtMakeTreeXs(Prefix,CT,Next,Rule,AssPar,Lt),
	!
;
	CT="PAR",
	calledRule(CT,Rule,No,CRule,_,_,_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX("PAR::",Type,CRule,AssPar,"",Tree),
	Next=No+1,		
	crtMakeTreeXs(Prefix,CT,Next,Rule,AssPar,Lt),
	!.
crtMakeTreeXs(Prefix,CT,No,Rule,AssPar,[Tree]) :-
	set::member(CT,["R","N","X","M"]),
	calledRule(CT,Rule,No,CRule,_,_,_,_,_),
	evaluation(CRule,Type,_,_,_,_,_,_,_,_,_),
	crtMakeTreeX(Prefix,Type,CRule,AssPar,"",Tree),
	!.
crtMakeTreeXs(_,_,_,_,_,[]) :-
	!.

crtMakeTreeSwi("",_,_,_,"default") :-
	!.
crtMakeTreeSwi(Op1,Inc,"",_,S) :-
	nl::conValX(Inc,"p3_dany","",S2),
	sx::concat([Op1,S2],S),
	!.
crtMakeTreeSwi(Op1,Inc,Op2,Dec,S) :-
	nl::conValX(Inc,"p3_dany","",S2),
	nl::conValX(Dec,"p3_dany","",S3),
	sx::concat([Op1,S2,"&",Op2,S3],S),
	!.

crtSubTree(Prefix,S,Suffix,Type,Trees,Tree) :-
	crtNodeName(S,Suffix,Prefix,Type,N2,S2),
	Tree=vpiTree::tree(S2,vpiTree::unmarked,Trees,N2),
	!.

crtNodeName(A,B,Prefix,Type,N,S) :-
	B="",
	globalVars::incr("crcnt",_,N),
	Ns=toString(N),
	sx::concat([Prefix,A,Type,"{",Ns],S),
	!
;
	sx::frontchar(B,'@',R),	
	globalVars::incr("crcnt",_,N),
	Ns=toString(N),
	sx::concat([Prefix,R,A,Type,"{",Ns],S),
	!
;	
	globalVars::incr("crcnt",_,N),
	Ns=toString(N),
	sx::concat([Prefix,A,B,Type,"{",Ns],S),
	!.
	
assName(NRule,Ass,NRuleA) :-
    Ass="",
    NRuleA=NRule,
    !
;
    NRuleA=string::concat(NRule,"[",Ass,"]"),
    !.
% ----------------------------------------------------------------------------------------------
% crtMap
% ----------------------------------------------------------------------------------------------
% maps value V into Vm
% might fail if discrete map does not have mapping entry
%
crtMap(SolRef,Rule,V,Vm) :-
	numericMap(SolRef,Rule,Rs,Ms),
	!,irbase::crtMapN(V,Rs,Ms,C),
	Vm=fix(C),
			tt::tt_s(3,"CR",Rule),
			tt::tt_term(3,"CR value",V),
			tt::tt_term(3,"CR mapped to",Vm),
	!
;
	discreteMap(SolRef,Rule,Ds,Ms),
	!,irbase::crtMapD(V,Ds,Ms,Vm),
			tt::tt_s(3,"CR",Rule),
			tt::tt_term(3,"CR value",V),
			tt::tt_term(3,"CR mapped to",Vm),
	!
;
	not(discreteMap(SolRef,Rule,_,_)),
	Vm=V,
	!.	
% ----------------------------------------------------------------------------------------------
% crtQueryRun
% ----------------------------------------------------------------------------------------------
class predicates
crtSubstFuncPars : (sterml,string,sterml,sterml,sterml) determ (i,i,i,i,o).
crtSubsFP_Get : (string,sterml,sterml,sterm) determ (i,i,i,o).
clauses

% executes a data rule query
%
crtQueryRun(SolRef,SolNo,Rule,Pars,APars,Query,C,Tree) :-
	Query=cmp("p3_iDdrf",[_,str(DERF),str(Qa)]),
	Term=termDomain::fromTermStringIndirect(Qa),
	termDomain::str_slist(Pars,SPars),
	irbase::crtSetQueryPars(Rule,Term,SPars,APars,Qp),
	piex::plcDDERF(DERF,Qp,Val),
	vars::varDE_Dom(DERF,Dom,Frm,Wid),
	termDomain::fterm(Val,Dom,Frm,Wid,FVal),
	crtMap(SolRef,Rule,Val,C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,RDom),
	crtReasonRec(Rule,APars,RDom,C,Val,"p3_iEval"),
	irbase::crtReasonTreeMap(atom(Rule),FVal,C,"",Pars,APars,[],Tree),
	!
;
	Query=cmp("p3_iDdrf",[_,str(DERF),_]),
	% failed query
	evaluation(Rule,_,_,_,Def,_,_,_,_,_,_),
	vars::varDE_Dom(DERF,Dom,Frm,Wid),
	termDomain::fterm(Def,Dom,Frm,Wid,FVal),
	crtMap(SolRef,Rule,Def,C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,RDom),
	crtReasonRec(Rule,APars,RDom,C,Def,"p3_iDeft"),
	irbase::crtReasonTreeMap(atom(Rule),FVal,C,"+",Pars,APars,[],Tree),
	!
;
	Query=cmp("p3_iSSS",[_,str(Qa)]),
	Term=termDomain::fromTermStringIndirect(Qa),
	termDomain::str_slist(Pars,SPars),
	irbase::crtSetQueryPars(Rule,Term,SPars,APars,Qp),
	piex::plcSSS(Qp,Val),
	crtMap(SolRef,Rule,Val,C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,APars,Dom,C,Val,"p3_iEval"),
	irbase::crtReasonTreeMap(atom(Rule),Val,C,"",Pars,APars,[],Tree),
	!
;
	Query=cmp("p3_iSSS",_),
	% failed query
	evaluation(Rule,_,_,_,Def,_,_,_,_,_,_),
	crtMap(SolRef,Rule,Def,C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,APars,Dom,C,Def,"p3_iDeft"),
	irbase::crtReasonTreeMap(atom(Rule),Def,C,"+",Pars,APars,[],Tree),
	!
;
	Query=cmp("p3_iExis",[str(Qa)]),
	Term=termDomain::fromTermStringIndirect(Qa),
	termDomain::str_slist(Pars,SPars),
	irbase::crtSetQueryPars(Rule,Term,SPars,APars,Qp),
	piex::plcExists(Qp),
	crtMap(SolRef,Rule,atom("p3_iYes"),C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,APars,Dom,C,atom("p3_iYes"),"p3_iEval"),
	irbase::crtReasonTreeMap(atom(Rule),atom("p3_iYes"),C,"",Pars,APars,[],Tree),
	!
;
	Query=cmp("p3_iExis",_),
	% do not exist
	crtMap(SolRef,Rule,atom("p3_iNo"),C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,APars,Dom,C,atom("p3_iNo"),"p3_iDeft"),
	irbase::crtReasonTreeMap(atom(Rule),atom("p3_iNo"),C,"+",Pars,APars,[],Tree),
	!
;
	% det function
	Query=cmp(Func,[atom("p3_iDtFn"),var(VRet)|FPars]),	
	dbBase::smDefF(Func,Clas,_,_,_,_,_),
	set::member(Clas,["f","g"]),
	crtSubstFuncPars(FPars,Rule,Pars,APars,AFPars),
	AQuery=cmp(Func,[atom("p3_iDtFn"),var(VRet)|AFPars]),	
	piex::plcFunc(VRet,AQuery,Val),
	crtMap(SolRef,Rule,Val,C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,APars,Dom,C,Val,"p3_iEval"),
	irbase::crtReasonTreeMap(atom(Rule),Val,C,"",Pars,APars,[],Tree),
	!
;
	% non-det function (call it as deterministic as non-det calls do not make sense in CR queries)
	Query=cmp(Func,[var(VRet)|FPars]),	
	dbBase::smDefF(Func,Clas,_,_,_,_,_),
	set::member(Clas,["f","g"]),
	crtSubstFuncPars(FPars,Rule,Pars,APars,AFPars),
	AQuery=cmp(Func,[atom("p3_iDtFn"),var(VRet)|AFPars]),	
	piex::plcFunc(VRet,AQuery,Val),
	crtMap(SolRef,Rule,Val,C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,APars,Dom,C,Val,"p3_iEval"),
	irbase::crtReasonTreeMap(atom(Rule),Val,C,"",Pars,APars,[],Tree),
	!
;
	Query=cmp(Func,_),	
	dbBase::smDefF(Func,Clas,_,_,_,_,_),
	set::member(Clas,["f","g"]),
	% failed query
	evaluation(Rule,_,_,_,Def,_,_,_,_,_,_),
	crtMap(SolRef,Rule,Def,C),
	%irbase::crtFormat(Rule,C,Cf),
	crtDom(Rule,Dom),
	crtReasonRec(Rule,APars,Dom,C,Def,"p3_iDeft"),
	irbase::crtReasonTreeMap(atom(Rule),Def,C,"+",Pars,APars,[],Tree),
	!
;
	dlgIF::feil_pie("crtQueryRun",[Rule],[SolRef,SolNo],cmp("",[Query|APars])),!,fail.

% parameters:
% 1	in	list of actual parameters in function call query
% 2	in	calling rule (k)
% 3	in	calling rule's formal parameters (fstr(... )
% 4	in	calling rule's actual parameters
% 5	out	substituted actual parameters corr. to par#1
%
crtSubstFuncPars([],_,_,_,[]) :-
	!.
crtSubstFuncPars([var(FPar)|FPars],Rule,Pars,APars,[V|Vs]) :-
	vars::varFuncParR(R,D,FPar),
	set::member(R,[Rule,"p3_iThis"]),
	crtSubsFP_Get(D,Pars,APars,V),
	crtSubstFuncPars(FPars,Rule,Pars,APars,Vs),
	!.
crtSubstFuncPars([FPar|FPars],Rule,Pars,APars,[FPar|Vs]) :-
	crtSubstFuncPars(FPars,Rule,Pars,APars,Vs),
	!.
crtSubsFP_Get(D,[fstr(FP,_,_,_)|FPs],[P|Ps],V) :-
	D=FP,
	V=P,
	!
;
	crtSubsFP_Get(D,FPs,Ps,V),
	!.
% ----------------------------------------------------------------------------------------------
% crtQuestionAsk
% ----------------------------------------------------------------------------------------------
class predicates
crtQuestionValidate : (integer,string,sterml,string,string,string_list,string,vpiTree::tree,sterm) determ (i,i,i,i,i,i,i,o,o).
clauses

% fails if break state occurs
%
crtQuestionAsk(SolRef,Rule,ParVals,QDom,Q,C,Tree) :-
    irbase::unpackQuestion(Q,QDom,Question,AltsNL),
    set::size(AltsNL,N),
    N>0,
    qAnswer(Rule,Question,Ans), % answer already via input facility
	crtQuestionValidate(SolRef,Rule,ParVals,QDom,Q,AltsNL,Ans,Tree,C),
    !
;
    irbase::unpackQuestion(Q,QDom,Question,AltsNL),
    set::size(AltsNL,N),
    N=0,
    qAnswer(Rule,Question,Ans), % answer already via input facility
	crtQuestionValidate(SolRef,Rule,ParVals,QDom,Question,[],Ans,Tree,C),
    !
;
    globalVars::get_i("server")=1,
    irbase::unpackQuestion(Q,QDom,Question,AltsNL),
    set::size(AltsNL,N),
    N>0,
	evaluation(Rule,_,_,_,Def,_,_,_,_,_,_),
	nl::conValX(Def,QDom,"",DefS),
	dbBase::nlSing(QDom,Dom),
	proc::processQuestion(Rule,Question,Dom,AltsNL,DefS,Ans),
	break=0,
	crtQuestionValidate(SolRef,Rule,ParVals,QDom,Q,AltsNL,Ans,Tree,C),
	!
;	
    globalVars::get_i("server")=1,
    break=0,
    irbase::unpackQuestion(Q,QDom,Question,AltsNL),
    set::size(AltsNL,N),
    N=0,
	dbBase::nlSing(QDom,Dom),
	evaluation(Rule,_,_,_,Def,_,_,_,_,_,_),
	nl::conValX(Def,QDom,"",DefS),
	proc::processQuestion(Rule,Question,Dom,[],DefS,Ans),
	break=0,
	crtQuestionValidate(SolRef,Rule,ParVals,QDom,Question,[],Ans,Tree,C),
	!
;
    globalVars::get_i("server")<>1,
    irbase::unpackQuestion(Q,QDom,Question,AltsNL),
    set::size(AltsNL,N),
    N>0,
	evaluation(Rule,_,_,_,Def,_,_,_,_,_,_),
	nl::conValX(Def,QDom,"",DefS),
	dbBase::nlSing(QDom,Dom),
    W=guix::getWinOfClas(1),
    Dlg=inexQuestion::new(),
    Dlg:show(W,Question,Rule,AltsNL,Dom,DefS),
    RetPars=globalVars::popRetPars(),
    RetPars=[AnsType,Ans],
    set::member(AnsType,["1","2"]),
	crtQuestionValidate(SolRef,Rule,ParVals,QDom,Q,AltsNL,Ans,Tree,C),
	!
;	
    globalVars::get_i("server")<>1,
    irbase::unpackQuestion(Q,QDom,Question,AltsNL),
    set::size(AltsNL,N),
    N=0,
	dbBase::nlSing(QDom,Dom),
	evaluation(Rule,_,_,_,Def,_,_,_,_,_,_),
	nl::conValX(Def,QDom,"",DefS),
    W=guix::getWinOfClas(1),
    Dlg=inexQuestion::new(),
    Dlg:show(W,Question,Rule,[],Dom,DefS),
    RetPars=globalVars::popRetPars(),
    RetPars=[AnsType,Ans],
    set::member(AnsType,["1","2"]),
	crtQuestionValidate(SolRef,Rule,ParVals,QDom,Question,[],Ans,Tree,C),
    !.

crtQuestionValidate(SolRef,Rule,ParVals,Dom,Q,Alts,S,Tree,V) :-
    set::size(Alts,N),
    N>0,
    not(set::member(S,Alts)),
	Mess=string::concatList(["The answer [",S,"] is not among the prescribed alternatives"]),
	dlgIF::dlgNote(1,Mess),
    crtQuestionAsk(SolRef,Rule,ParVals,Dom,Q,V,Tree),
    !
;
	dom::domNum(Dom),
	inexgui::wcrCheckVal("answer",Dom,S,PreMap),
	crtMap(SolRef,Rule,PreMap,V),
	%irbase::crtFormat(Rule,V,Vf),
	evaluation(Rule,_,_,_,_,Pars,_,_,_,_,_),
	irbase::crtReasonTreeMap(atom(Rule),PreMap,V,"",Pars,ParVals,[],Tree),
	crtDom(Rule,RDom),				
	crtReasonRec(Rule,ParVals,RDom,PreMap,V,"p3_iEval"),
	!
;
	Dom="p3_dli",
	sx::sxUnquote(S,Cm),
	PreMap=str(Cm),
	crtMap(SolRef,Rule,PreMap,V),
	%irbase::crtFormat(Rule,V,Vf),
	evaluation(Rule,_,_,_,_,Pars,_,_,_,_,_),
	irbase::crtReasonTreeMap(atom(Rule),PreMap,V,"",Pars,ParVals,[],Tree),
	crtDom(Rule,RDom),				
	crtReasonRec(Rule,ParVals,RDom,PreMap,V,"p3_iEval"),
	!
;
	Dom="p3_dif",
	dbBase::kbID_F(S,Cm,_,_),
	PreMap=atom(Cm),
	crtMap(SolRef,Rule,PreMap,V),
	%irbase::crtFormat(Rule,V,Vf),
	evaluation(Rule,_,_,_,_,Pars,_,_,_,_,_),
	irbase::crtReasonTreeMap(atom(Rule),PreMap,V,"",Pars,ParVals,[],Tree),
	crtDom(Rule,RDom),				
	crtReasonRec(Rule,ParVals,RDom,PreMap,V,"p3_iEval"),
	!
;
	dbBase::nlSing(Dom,DomNL),
	sx::concat(["This answer is invalid or unmappable or out of range for domain ",DomNL],Mess),
	dlgIF::dlgNote(1,Mess),
    crtQuestionAsk(SolRef,Rule,ParVals,Dom,Q,V,Tree),
	!.
% ----------------------------------------------------------------------------------------------
% checkInputFac
% ----------------------------------------------------------------------------------------------
class predicates
actPars : (string,string_list,sterml,sterml,sterml) determ (i,i,i,i,o).
getVal : (string,string,sterml,sterml,sterm) determ (i,i,i,i,o).
clauses

% parameters:
% 1 in rule to be checked for input facility
% 2 in rule's format parameters ([fstr(...),...])
% 3 in rule's actual parameters (e.g. [str(...),...]
%
checkInputFac(Rule,Pars,ParVals) :-
    break>0,
    !
;    
    qSeries(Rule,Prompt,QstRule,QstPar,Format,Query,SPars,A,0), % facility never called
    actPars(Rule,SPars,Pars,ParVals,APars),
    irbase::setInpFacPars(Rule,Prompt,SPars,APars,Prompt2),
    irbase::setInpFacPars(Rule,Query,SPars,APars,Query2),
    Set=irbase::queryInpFac(Query2,Format),
    loadedRule(QstRule,_,_,_,Default,_,_,cmp(_,[_,atom(QDom),str(Question)]),_,_,_),
    dbBase::smClassDef(QstPar,"a",_,_,Dom,_,_),
    runQuestionSeries(Prompt2,QstRule,Default,QstPar,Dom,Set,Question,QDom),    
    retract(qSeries(Rule,Prompt,QstRule,QstPar,Format,Query,SPars,A,0)),    
    assert(qSeries(Rule,Prompt,QstRule,QstPar,Format,Query,SPars,APars,1)),    
    !
;
    qSeries(Rule,_,_,_,_,_,SPars,ActPars,1),
    actPars(Rule,SPars,Pars,ParVals,APars),
    ActPars=APars, % already called with same parameters
    !
;    
    qSeries(Rule,Prompt,QstRule,QstPar,Format,Query,SPars,_,1), % facility not called with current parameters
    actPars(Rule,SPars,Pars,ParVals,APars),
    irbase::setInpFacPars(Rule,Prompt,SPars,APars,Prompt2),
    irbase::setInpFacPars(Rule,Query,SPars,APars,Query2),
    Set=irbase::queryInpFac(Query2,Format),
    loadedRule(QstRule,_,_,_,Default,_,_,cmp(_,[_,atom(QDom),str(Question)]),_,_,_),
    dbBase::smClassDef(QstPar,"a",_,_,Dom,_,_),
    runQuestionSeries(Prompt2,QstRule,Default,QstPar,Dom,Set,Question,QDom),    
    assert(qSeries(Rule,Prompt,QstRule,QstPar,Format,Query,SPars,APars,1)),    
    !
;
    qCluster(Rule,Prompt,Rules,SPars,A,0), % facility never called
    actPars(Rule,SPars,Pars,ParVals,APars),
    irbase::setInpFacPars(Rule,Prompt,SPars,APars,Prompt2),
    runQuestionCluster(Prompt2,Rules,SPars,APars),    
    retract(qCluster(Rule,Prompt,Rules,SPars,A,0)),    
    assert(qCluster(Rule,Prompt,Rules,SPars,APars,1)),    
    !
;
    qCluster(Rule,_,_,SPars,ActPars,1),
    actPars(Rule,SPars,Pars,ParVals,APars),
    ActPars=APars, % already called with same parameters
    !
;
    qCluster(Rule,Prompt,Rules,SPars,_,1), % facility not called with current parameters
    actPars(Rule,SPars,Pars,ParVals,APars),
    irbase::setInpFacPars(Rule,Prompt,SPars,APars,Prompt2),
    runQuestionCluster(Prompt2,Rules,SPars,APars),    
    assert(qCluster(Rule,Prompt,Rules,SPars,APars,1)),    
    !
;
    !.

actPars(_,[],_,_,[]) :-
    !.
actPars(Rule,[Par|SPars],Pars,ParVals,[APar|APars]) :-
    getVal(Rule,Par,Pars,ParVals,APar),
    actPars(Rule,SPars,Pars,ParVals,APars),
    !.

getVal(Rule,Par,[],_,_) :-
    Mess=string::concatList(["Cannot find value for parameter '",Par,"' in inexact rule '",Rule,"'"]),
    dlgIF::dlgNote(3,Mess),!,fail.
getVal(_,Par,[fstr(Par,_,_,_)|_],[V|_],V) :- 
    !.
getVal(Rule,Par,[_|Pars],[_|Vals],APar) :-
    getVal(Rule,Par,Pars,Vals,APar),
    !.
% ----------------------------------------------------------------------------------------------
% runQuestionSeries/Cluster
% ----------------------------------------------------------------------------------------------
class predicates
assertAnswers : (string,string_list,string_list) determ.
assertAnswers2 : (string_list,string_list,string_list) determ.
clauses

% parameters:
% 0 in series prompt
% 1 in question rule
% 2 in default answer
% 3 in question parameter to substituted
% 4 in question parameter domain
% 5 in set of substitution values
% 6 in question text (unprocessed)
%  
runQuestionSeries(Prompt,QstRule,Default,QstPar,Dom,Set,Question,QDom) :-
    irbase::unpackQuestion(Question,QDom,QuestionU,AltsNL),
    inputfacility::executeSeries(Prompt,QstRule,Default,QstPar,Dom,Set,QuestionU,AltsNL,QDom,Questions,Answers),
    assertAnswers(QstRule,Questions,Answers),
    !.
    
% parameters:
% 0 in series prompt
% 1 in question rules
% 2 in facility formal parameters (names)
% 3 in facility actual parameter values
%  
runQuestionCluster(Prompt,Rules,SPars,APars) :-    
    inputfacility::executeCluster(Prompt,Rules,SPars,APars,Questions,Answers),
    assertAnswers2(Rules,Questions,Answers),
    !.

assertAnswers(_,[],[]) :-
    !.
assertAnswers(Rule,[Q|Qs],[A|As]) :-
    assert(qAnswer(Rule,Q,A)),
    assertAnswers(Rule,Qs,As),
    !.

assertAnswers2([],[],[]) :-
    !.
assertAnswers2([Rule|Rules],[Q|Qs],[A|As]) :-
    assert(qAnswer(Rule,Q,A)),
    assertAnswers2(Rules,Qs,As),
    !.
% ----------------------------------------------------------------------------------------------
% crtReasonRec
% ----------------------------------------------------------------------------------------------
class predicates
crtReasonRec2 : (string,string,sterm,sterm,sterml,sterm,sterml) determ (i,i,i,i,i,i,o).
crtReasonPars : (sterml,sterm,sterml) determ (i,i,o).
clauses

crtReasonRec(Rule,ParVals,Dom,V,MappedFrom,EvalType) :-
	globalVars::get_i("crdet")=1,
	termDomain::listBR(APars,ParVals),
	evaluation(Rule,RuleType,_,_,_,_,_,_,_,_,_),
	rootRule(_,ST,Arg,DE,RootRule,AParsRoot),
	LT=cmp("r",[cmp("w",[str(ST),
	                      atom(Arg),
	                      atom(DE),
	                      atom(RootRule),
	                      AParsRoot,
	                      atom(RuleType),
	                      atom(Rule),
	                      APars,
	                      atom(Dom),
	                      V,
	                      MappedFrom,
	                      atom(EvalType)])]),
	dbBase::recordInsert(1,"$IRR",0,0,LT,atom("true")),
	!
;
	!.

crtReasonRec2(Dom,Rule,V,MappedFrom,Pars,APars,Eval) :-
	Dom="p3_dcf",
	MappedFrom=nill,
	crtReasonPars(Pars,APars,Params),
	Eval=[atom(Rule),atom("p3_iWith"),atom("p3_iCert"),V,atom("p3_iGivn")|Params],
	!
;
	Dom="p3_dcf",
	crtReasonPars(Pars,APars,Params),
	Eval=[atom(Rule),atom("p3_iWith"),atom("p3_iCert"),V,str("("),MappedFrom,str(")"),atom("p3_iGivn")|Params],
	!
;
	MappedFrom=nill,
	crtReasonPars(Pars,APars,Params),
	Eval=[atom(Rule),atom("p3_iWith"),atom("p3_iResu"),V,atom("p3_iGivn")|Params],
	!
;
	crtReasonPars(Pars,APars,Params),
	Eval=[atom(Rule),atom("p3_iWith"),atom("p3_iResu"),V,str("("),MappedFrom,str(")"),atom("p3_iGivn")|Params],
	!.
	
crtReasonPars([],nill,[]) :-
	!.
crtReasonPars([fstr(DE,_,_,_)|DEs],list(P,Ps),[atom(DE),P|Pars]) :-
	crtReasonPars(DEs,Ps,Pars),
	!.
crtReasonPars(D,P,_) :-
	dlgIF::feil_pie("crtReasonPars",[],[],cmp("*",[cmp("*",D),cmp("*",[P])])),!,fail.
% ----------------------------------------------------------------------------------------------
% crtRetract
% ----------------------------------------------------------------------------------------------
crtRetract() :-
	retractall(rootRule(_,_,_,_,_,_)),
	retractall(evaluation(_,_,_,_,_,_,_,_,_,_,_)),
	retractall(loadedRule(_,_,_,_,_,_,_,_,_,_,_)),
	retractall(calledRule(_,_,_,_,_,_,_,_,_)),
	retractall(numericMap(_,_,_,_)),
	retractall(discreteMap(_,_,_,_)),
    retractall(qAnswer(_,_,_)),
    retractall(qCluster(_,_,_,_,_,_)),
    retractall(qSeries(_,_,_,_,_,_,_,_,_)),
	!.
% ----------------------------------------------------------------------------------------------
% crtTree
% ----------------------------------------------------------------------------------------------
% creates a inexact rule network window
%
crtTree(Rule,AssPar,W) :-
    dbBase::getIrDef(Rule,Body,_),
	crtRetract,
	loadRule(0,Body),
	globalVars::set_i("crcnt",0),
	crtMakeTree(Rule,AssPar,Tree),
	dbBase::nlSing(Rule,NRule),
	S=string::concat("Inexact rule: ",NRule),
	p3Tree::drwOpenTree(S,44,Tree,vpiTree::tree_DirRight,W),
	crtRetract,
	!.
% ----------------------------------------------------------------------------------------------
% crtDom   
% ----------------------------------------------------------------------------------------------
crtDom(Rule,Dom) :-
	evaluation(Rule,"p3_iCdat",_,_,_,_,_,D,_,_,_),
	D=cmp(_,[atom(Dom)|_]),
	!
;
	evaluation(Rule,_,_,_,_,_,_,D,_,_,_),
	D=atom(Dom),
	!
;
	evaluation(Rule,_,_,_,_,_,_,D,_,_,_),
	D=cmp(Dom,_),
	!
;
	dlgIF::feil_pie("crtDom",["Invalid 'rul'-structure",Rule],[],nill),!,fail.
% ----------------------------------------------------------------------------------------------
% setBreak
% ----------------------------------------------------------------------------------------------
setBreak(V) :- 
    break:=V,
    !.
% ----------------------------------------------------------------------------------------------
% loadSubnet          
% ----------------------------------------------------------------------------------------------
% performs the inital loading of a rule network
% 
loadSubnet(SolRef,Rule) :-
	evaluation(Rule,_,_,_,_,_,_,_,_,_,_), % already loaded
	!
;	
    dbBase::getIrDefF(Rule,_,Loc),
    dbBase::recordIsSuspended(Loc),
    Mess=string::concat("Called inexact rule: ",Rule," is suspended!"),
    dlgIF::dlgNote(2,Mess),!,fail
;	
    dbBase::getIrDefF(Rule,Body,_),
	loadRule(SolRef,Body),
	!
;
    Mess=string::concat("Called inexact rule: ",Rule," is unknown!"),
    dlgIF::dlgNote(2,Mess),!,fail.
% ----------------------------------------------------------------------------------------------
% loadRule    
% ----------------------------------------------------------------------------------------------
class predicates
crtContext : (sterm,sterml) determ (i,o).
clauses	

% asserts rule body and loads all called rules
% (this predicate serves as master templates for the various IR term formats)
%
loadRule(SolRef,Body) :-
	Body=cmp("p3_iCmap",
	        [atom(Rule),
	         SCall,
	         Map,
	         fix(Prior),
	         OC,
	         Def,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,Dom,_,_),
	dbBase::smClassDefTree(Rule,"h",DefR),
	DefR=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,"p3_iCmap",Prior,OC,Def,Pars,[],cmp(Dom,[Map]),0,fix(0.0),0.0)),
	loadRuleCalls("M",SolRef,Rule,1,[SCall]),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	crtLoadMap(SolRef,Rule,Map),
	!
;
	Body=cmp("p3_iCdat",
	        [atom(Rule),
	         cmp(Q,QT),
	         Map,
	         fix(Prior),
	         OC,
	         Def,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,RDom,_,_),
	dbBase::smClassDefTree(Rule,"h",DefR),
	DefR=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,"p3_iCdat",Prior,OC,Def,Pars,[],cmp(Q,[atom(RDom)|QT]),0,fix(0.0),0.0)),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	crtLoadMap(SolRef,Rule,Map),
	!
;
	Body=cmp("p3_iCbay",
	        [atom(Rule),
	         cmp(_,Bays),
	         fix(Prior),
	         OC,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,Dom,_,_),
	dbBase::smClassDefTree(Rule,"h",Def),
	Def=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,"p3_iCbay",Prior,OC,fix(0.0),Pars,[],atom(Dom),0,fix(0.0),0.0)),
	loadRuleCalls("B",SolRef,Rule,1,Bays),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	!
;
	Body=cmp("p3_iCswi",
	        [atom(Rule),
	         atom(DE), % parameter-type
	         cmp(_,Swis),
	         fix(Prior),
	         OC,
	         Def,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,RDom,_,_),
	dbBase::smClassDefTree(Rule,"h",DefR),
	DefR=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,"p3_iCswi",Prior,OC,Def,Pars,[],cmp(RDom,[atom(DE)]),0,fix(0.0),0.0)),
	loadRuleCalls("S",SolRef,Rule,1,Swis),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	!
;
	Body=cmp("p3_iCswi", % call-type
	        [atom(Rule),
	         SCall,
	         cmp(_,Swis),
	         fix(Prior),
	         OC,
	         Def,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,RDom,_,_),
	dbBase::smClassDefTree(Rule,"h",DefR),
	DefR=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,"p3_iCswi",Prior,OC,Def,Pars,[],atom(RDom),0,fix(0.0),0.0)),
	loadRuleCalls("X",SolRef,Rule,1,[SCall]),
	loadRuleCalls("S",SolRef,Rule,1,Swis),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	!
;
	Body=cmp("p3_iCpar", 
	        [atom(Rule),
	         SCall,
	         cmp(_,PCals),
	         fix(Prior),
	         OC,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,RDom,_,_),
	dbBase::smClassDefTree(Rule,"h",Def),
	Def=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,"p3_iCpar",Prior,OC,nill,Pars,[],atom(RDom),0,fix(0.0),0.0)),
	loadRuleCalls("R",SolRef,Rule,1,[SCall]),
	loadRuleCalls("PAR",SolRef,Rule,1,PCals),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	!
;
	Body=cmp("p3_iCnot",
	        [atom(Rule),
	         SCall,
	         fix(Prior),
	         OC,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,Dom,_,_),
	dbBase::smClassDefTree(Rule,"h",Def),
	Def=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,"p3_iCnot",Prior,OC,fix(0.0),Pars,[],atom(Dom),0,fix(0.0),0.0)),
	loadRuleCalls("N",SolRef,Rule,1,[SCall]),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	!
;
	Body=cmp(LogTyp,
	        [atom(Rule),
	         cmp(_,Logs),
	         fix(Prior),
	         OC,
	         Con]),
	dbBase::smClassDef(Rule,"h",_,_,Dom,_,_),
	dbBase::smClassDefTree(Rule,"h",Def),
	Def=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	set::member(LogTyp,["p3_iCand","p3_iCor","p3_iCcmb"]),
	assert(evaluation(Rule,LogTyp,Prior,OC,fix(0.0),Pars,[],atom(Dom),0,fix(0.0),0.0)),
	loadRuleCalls("L",SolRef,Rule,1,Logs),
	crtContext(Con,Cons),
	loadRuleCalls("C",SolRef,Rule,1,Cons),
	!
;
	Body=cmp(Typ,
	        [atom(Rule),
	         cmp(FZ,Cals)]),
	dbBase::smClassDef(Rule,"h",_,_,Dom,_,_),
	dbBase::smClassDefTree(Rule,"h",Def),
	Def=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	set::member(Typ,["p3_iZand","p3_iZor","p3_iZnot"]),
	assert(evaluation(Rule,FZ,0.0,nill,nill,Pars,[],atom(Dom),0,nill,0.0)),
	loadRuleCalls("Z",SolRef,Rule,1,Cals),
	!
;
	Body=cmp(Typ,
	        [atom(Rule),
	         cmp(FZ,Cals)]),
	Typ="p3_iZowa",
	dbBase::smClassDef(Rule,"h",_,_,Dom,_,_),
	dbBase::smClassDefTree(Rule,"h",Def),
	Def=cmp("REL_TYPE",[_,_,cmp("FPAR",Pars)]),
	assert(evaluation(Rule,FZ,0.0,nill,nill,Pars,[],atom(Dom),0,nill,0.0)),
	loadRuleCalls("OWA",SolRef,Rule,1,Cals),
	!
;
	dlgIF::feil_pie("loadRule",["Cannot recognize inexact rule"],[],Body),
	!,fail.
crtContext(nill,[]) :-
	!.
crtContext(cmp(_,Cons),Cons) :-
	!.
% ----------------------------------------------------------------------------------------------
% loadRuleCalls
% ----------------------------------------------------------------------------------------------
loadRuleCalls(_,_,_,_,[]) :-
	!.
loadRuleCalls(CallType,SolRef,Rule,CallNo,[cmp(CRule,Arg)|Calls]) :-
	set::member(CallType,["S","C"]),
	Arg=[nill,nill,nill,nill,cmp(_,APars2)],
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule(CallType,Rule,CallNo,CRule,fix(0.0),fix(0.0),"","",APars)),
	Next=CallNo+1,
	loadRuleCalls(CallType,SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	!
;
	set::member(CallType,["S","C"]),
	Arg=[atom(Oper),Opn,nill,nill,cmp(_,APars2)],
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule(CallType,Rule,CallNo,CRule,Opn,nill,Oper,"",APars)),
	Next=CallNo+1,
	loadRuleCalls(CallType,SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	!
;
	set::member(CallType,["S","C"]),
	Arg=[atom(Oper),Opn,atom(Oper2),Opn2,cmp(_,APars2)],
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule(CallType,Rule,CallNo,CRule,Opn,Opn2,Oper,Oper2,APars)),
	Next=CallNo+1,
	loadRuleCalls(CallType,SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	!.
loadRuleCalls("OWA",SolRef,Rule,CallNo,[cmp(CRule,[fix(Wg),cmp(_,APars2)])|Calls]) :-
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule("OWA",Rule,CallNo,CRule,fix(Wg),nill,"","",APars)),
	Next=CallNo+1,
	loadRuleCalls("OWA",SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	!.	
loadRuleCalls("OWA",SolRef,Rule,CallNo,[cmp(CRule,[cmp(DR,B),cmp(_,APars2)])|Calls]) :-
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule("OWA",Rule,CallNo,CRule,atom(DR),nill,"","",APars)),
	Next=CallNo+1,
	loadRuleCalls("OWA",SolRef,Rule,Next,Calls),
	loadRuleCalls("W",SolRef,Rule,1,[cmp(DR,B)]),
	loadRuleCalls("WC",SolRef,CRule,1,[cmp(DR,B)]),
	loadSubnet(SolRef,CRule),
	!.	
loadRuleCalls("B",SolRef,Rule,CallNo,[cmp(CRule,[fix(Min),fix(Max),cmp(_,APars2)])|Calls]) :-
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule("B",Rule,CallNo,CRule,fix(Min),fix(Max),"","",APars)),
	Next=CallNo+1,
	loadRuleCalls("B",SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	!.	
loadRuleCalls("B",SolRef,Rule,CallNo,[cmp(CRule,[fix(Min),cmp(DR,B),cmp(_,APars2)])|Calls]) :-
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule("B",Rule,CallNo,CRule,fix(Min),atom(DR),"","",APars)),
	Next=CallNo+1,
	loadRuleCalls("B",SolRef,Rule,Next,Calls),
	loadRuleCalls("D",SolRef,Rule,1,[cmp(DR,B)]),
	loadRuleCalls("DC",SolRef,CRule,1,[cmp(DR,B)]),
	loadSubnet(SolRef,CRule),
	!.	
loadRuleCalls("B",SolRef,Rule,CallNo,[cmp(CRule,[cmp(IR,B),fix(Min),cmp(_,APars2)])|Calls]) :-
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule("B",Rule,CallNo,CRule,atom(IR),fix(Min),"","",APars)),
	Next=CallNo+1,
	loadRuleCalls("B",SolRef,Rule,Next,Calls),
	loadRuleCalls("I",SolRef,Rule,1,[cmp(IR,B)]),
	loadRuleCalls("IC",SolRef,CRule,1,[cmp(IR,B)]),
	loadSubnet(SolRef,CRule),
	!.	
loadRuleCalls("B",SolRef,Rule,CallNo,[cmp(CRule,[cmp(IR,BI),cmp(DR,BD),cmp(_,APars2)])|Calls]) :-
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule("B",Rule,CallNo,CRule,atom(IR),atom(DR),"","",APars)),
	Next=CallNo+1,
	loadRuleCalls("B",SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	loadRuleCalls("I",SolRef,Rule,1,[cmp(IR,BI)]),
	loadRuleCalls("IC",SolRef,CRule,1,[cmp(IR,BI)]),
	loadRuleCalls("D",SolRef,Rule,1,[cmp(DR,BD)]),
	loadRuleCalls("DC",SolRef,CRule,1,[cmp(DR,BD)]),
	!.	
loadRuleCalls("PAR",SolRef,Rule,CallNo,[cmp(CRule,Arg)|Calls]) :-
	Arg=[var(V),cmp(_,APars2)],
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule("PAR",Rule,CallNo,CRule,var(V),nill,"","",APars)),
	Next=CallNo+1,
	loadRuleCalls("PAR",SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	!.
loadRuleCalls(CType,SolRef,Rule,CallNo,[cmp(CRule,[cmp(_,APars2)])|Calls]) :-
	irBase::wcrAdjustCallParsR(APars,APars2),
	assert(calledRule(CType,Rule,CallNo,CRule,nill,nill,"","",APars)),
	Next=CallNo+1,
	loadRuleCalls(CType,SolRef,Rule,Next,Calls),
	loadSubnet(SolRef,CRule),
	!.	
loadRuleCalls(CType,SolRef,Rule,CallNo,[Call|_]) :-
	dlgIF::feil_pie("crtLoadCalls",[Rule,CType],[SolRef,CallNo],Call),!.
% ----------------------------------------------------------------------------------------------
% crtLoadMap
% ----------------------------------------------------------------------------------------------
class predicates
crtLoadMapN : (integer,string,sterml,sterml,real_list,real_list) determ (i,i,i,i,o,o).
crtSortMap : (real_list,real_list,real_list,real_list) determ (i,i,o,o).
crtSortMapX : (real_list,real_list,real_list,real_list) determ (i,i,i,o).
fixInsSort : (real,real_list,real_list) determ (i,i,o).
fixSort : (real_list,real_list) determ (i,o).
clauses

% loads a rule map
%
crtLoadMap(SolRef,Rule,Map) :-
	Map=cmp("p3_iNmap",[cmp(_,Vs),cmp(_,Ms)]),
	crtLoadMapN(SolRef,Rule,Vs,Ms,FVs2,FMs2),
	crtSortMap(FVs2,FMs2,FVs,FMs),
	assert(numericMap(SolRef,Rule,FVs,FMs)),
	!	
;
	Map=cmp("p3_iDmap",[cmp(_,Vs),cmp(_,Ms)]),
	assert(discreteMap(SolRef,Rule,Vs,Ms)),
	!
;
	Map=nill,
	!	
;
	dlgIF::feil_pie("crtLoadMap",["Invalid map for rule",Rule],[SolRef],Map),
	!,fail.	

crtLoadMapN(_,_,[],[],[],[]) :-
	!.
crtLoadMapN(SolRef,Rule,[fix(V)|Vs],[fix(M)|Ms],[V|Vms],[M|Mms]) :-
	crtLoadMapN(SolRef,Rule,Vs,Ms,Vms,Mms),
	!.
crtLoadMapN(SolRef,Rule,[int(V)|Vs],[fix(M)|Ms],[Vr|Vms],[M|Mms]) :-
	Vr=V,
	crtLoadMapN(SolRef,Rule,Vs,Ms,Vms,Mms),
	!.
crtLoadMapN(SolRef,Rule,[fix(V)|Vs],[int(M)|Ms],[V|Vms],[Mr|Mms]) :-
	Mr=M,
	crtLoadMapN(SolRef,Rule,Vs,Ms,Vms,Mms),
	!.
crtLoadMapN(SolRef,Rule,[int(V)|Vs],[int(M)|Ms],[Vr|Vms],[Mr|Mms]) :-
    Vr=V,
	Mr=M,
	crtLoadMapN(SolRef,Rule,Vs,Ms,Vms,Mms),
	!.
crtLoadMapN(SolRef,Rule,Vs,Ms,_,_) :-
	dlgIF::feil_pie("crtLoadMapN",["Invalid map for rule",Rule],[SolRef],cmp("",[cmp("Values",Vs),cmp("Mapped",Ms)])),	
	!,fail.	

% sorts the numeric map in ascending sequence (for the mapping to work)
%	
crtSortMap(F,T,Fs,Ts) :-
	fixSort(F,Fs),
	crtSortMapX(Fs,F,T,Ts),
	!.
crtSortMapX([],_,_,[]) :-
	!.
crtSortMapX([Fh|Ft],F,T,[Th|Tt]) :-
	set::setGet(F,N,Fh),
	set::setGet(T,N,Th),
	crtSortMapX(Ft,F,T,Tt),
	!.
	
% sorts the list ascending
%
fixSort([],[]) :- !.
fixSort([H],[H]) :- !.
fixSort([H|T],Sorted) :-
	fixSort(T,Ts),
	fixInsSort(H,Ts,Sorted),!.

% inserts P into a sorted list
%
fixInsSort(P,[],[P]) :- !.	
fixInsSort(P,[H|T],[P,H|T]) :- P<=H,!.
fixInsSort(P,[H|T],[H|PT]) :- fixInsSort(P,T,PT),!.	
% ----------------------------------------------------------------------------------------------
% loadInputFac
% ----------------------------------------------------------------------------------------------
class predicates
insertIF : (string,sterml,string) determ.
isRelevant : (string_list) determ.
clauses

% loads all input facilities where all rules related to the facility exist in the current network
%
loadInputFac() :-
    dbBase::chainSentences("p3_pIRinp",LT),
        LT=cmp(_,[cmp(Type,Body),atom(CallingRule)]),
        evaluation(CallingRule,_,_,_,_,_,_,_,_,_,_),
        insertIF(Type,Body,CallingRule),
        fail
;
    !.
    
insertIF(Type,Body,CallingRule) :-
    Type="p3_eQueSer",
    Body=[str(Prompt),atom(QRule),atom(QPar),atom(Format),str(Query),Params],
	evaluation(QRule,_,_,_,_,_,_,_,_,_,_),
    termDomain::listBR(Params,ParamsL),
    conversion::toStrings(ParamsL,SPars),
    assert(qSeries(CallingRule,Prompt,QRule,QPar,Format,Query,SPars,[],0)),
    !
;    
    Type="p3_eQueClu",
    Body=[str(Prompt),QRules,Params],
    termDomain::listBR(QRules,QRulesL),
    conversion::toStrings(QRulesL,SRules),
    isRelevant(SRules),
    termDomain::listBR(Params,ParamsL),
    conversion::toStrings(ParamsL,SPars),
    assert(qCluster(CallingRule,Prompt,SRules,SPars,[],0)),
    !
;
    set::member(Type,["p3_eQueSer","p3_eQueClu"]),
    Mess=string::concat("Invalid IR input facility sentence of type: ",Type),
    dlgIF::dlgNote(2,Mess),!,fail.
    
isRelevant([]) :-
    !.
isRelevant([H|T]) :-     
    evaluation(H,_,_,_,_,_,_,_,_,_,_),
    isRelevant(T),
    !.
end implement ir
