/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement irBase
open core,termDomain

constants
className = "com/rudstrand/irBase/irBase".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% seriesLoc
% ----------------------------------------------------------------------------------------------
seriesLoc(Rule)=Loc :-
    Loc=seriesLocF(Rule),
    !
;
    Mess=string::concat("Unknown question series input facility: ",Rule,"!"),
    dlgIF::dlgNote(3,Mess),!,fail.
    
seriesLocF(Rule)=Loc :-
    dbBase::chainSentences("p3_pIRinp",LT,Loc),
    LT=cmp(_,[cmp("p3_eQueSer",[_,atom(Rule)|_])|_]),
    !.
% ----------------------------------------------------------------------------------------------
% clusterLoc
% ----------------------------------------------------------------------------------------------
clusterLoc(Rule,Prompt)=Loc :-
    Loc=clusterLocF(Rule,Prompt),
    !
;
    Mess=string::concat("Unknown question cluster input facility: ",Rule,"(",Prompt,")!"),
    dlgIF::dlgNote(3,Mess),!,fail.
    
clusterLocF(Rule,Prompt)=Loc :-
    dbBase::chainSentences("p3_pIRinp",LT,Loc),
    LT=cmp(_,[cmp("p3_eQueClu",[str(Prompt)|_]),atom(Rule)]),
    !.
% ----------------------------------------------------------------------------------------------
% wcrAdjustCallParsR
% ----------------------------------------------------------------------------------------------
wcrAdjustCallParsR([],[]) :-
	!.
wcrAdjustCallParsR([H|T],[Hm|Tm]) :-
	Hm=cmp(A,[atom(D)|R]),
	dbBase::smClassDef(D,"a",_,_,Dom,_,_),
	H=cmp(A,[fvar(D,Dom)|R]),
	wcrAdjustCallParsR(T,Tm),
	!.
% ----------------------------------------------------------------------------------------------
% isSubType
% ----------------------------------------------------------------------------------------------
isSubType(Type,CR) :-
    Type="p3_iCswc",
	CR=cmp(_,[_,cmp(_,_),_,_,_,_,_]),
	!
;	
    Type="p3_iCswp",
	CR=cmp(_,[_,nill,_,_,_,_,_]),
	!
;    
    Type="p3_iCqry",
	CR=cmp(_,[_,cmp("p3_iQry",_),_,_,_,_,_]),
	!
;    
    Type="p3_iCqst",
	CR=cmp(_,[_,cmp("p3_iQst",_),_,_,_,_,_]),
    !.
% ----------------------------------------------------------------------------------------------
% wcrQuestionToIF/NL
% ----------------------------------------------------------------------------------------------
class predicates
wcrQuestionToIF_Pars : (string,string) determ (i,o).
wcrQuestionToIF_Choices : (string_list,string_list) determ (i,o).
wcrQuestionToNL_Pars : (string,string) determ (i,o).
wcrQuestionToNL_Choices : (string_list,string_list) determ (i,o).
clauses

% translates the question from NL to IF - this involves translating parameters and anwser alternatives in the question text from NL to IF
%
wcrQuestionToIF(Q,Qt) :-
	Pos=string::searchchar(Q,'?',string::caseSensitive),
	string::front(Q,Pos,Qproper,Qalts),
	sx::sxGetInBrack(Qalts,'[',']',Qt3),
	!, % with answer alternatives
	wcrQuestionToIF_Pars(Qproper,Qproper2),
	sx::sxUnpack(Qt3,';',Choices),
	wcrQuestionToIF_Choices(Choices,Qt4),
	sx::sxlJoin(Qt4,";",Qt5),
	sx::concat([Qproper2,"?[",Qt5,"]"],Qt),
	!
;	
	% without answer alternatives
	wcrQuestionToIF_Pars(Q,Qt),
	!.

wcrQuestionToIF_Pars(Q,Qt) :-
	sx::searchstring(Q,"[",P2),
	P2p=P2-1,
	sx::frontstr(P2p,Q,F,T2),
	sx::frontstr(1,T2,_,T2n),
	sx::searchstring(T2n,"]",P3),
	P4=P3-1,
	P4>1,
	sx::frontstr(P4,T2n,Par2,_),
	sx::sxStrip(Par2,Par),
	transnl::parCompSynTree(Par,Tree),
	Tree=fvar(PVar,_),
	P5=P4+1,
	sx::frontstr(P5,T2n,_,T3),
	wcrQuestionToIF_Pars(T3,Tt),
	sx::concat([F,"[",PVar,"]",Tt],Qt),
	!
;
	_=string::search(Q,"["),
	Mess=string::concat("Invalid parameter in data rule question: ",Q,"!"),
	dlgIF::dlgNote(1,Mess),!,fail
;
	Qt=Q,
	!.	
		
wcrQuestionToIF_Choices([],[]) :-
	!.
wcrQuestionToIF_Choices([H|T],[Hk|Tk]) :-
	sx::sxStrip(H,Hs),
	sx::frontchar(Hs,'"',_),
	sx::sxTail(1,Hs,L,_),
	Lc=string::subChar(L,0),
	Lc='"',
	!,Hk=Hs,
	wcrQuestionToIF_Choices(T,Tk),
	!
;
	sx::sxStrip(H,Hs),
	_=sx::toReal(Hs),
	!,Hk=Hs,
	wcrQuestionToIF_Choices(T,Tk),
	!
;
	sx::sxStrip(H,Hs),
	dbBase::kbID_F(Hs,Hk,_,_),
	!,wcrQuestionToIF_Choices(T,Tk),
	!	
;
	sx::sxStrip(H,Hs),
	Mess=string::concat("Invalid answer choice: ",Hs,"!"),
	dlgIF::dlgNote(1,Mess),
	!,fail.		
		
% translates the question alternatives part to NL
%
wcrQuestionAltsToNL(Q,Qt) :-
    _=string::searchchar(Q,'[',string::caseSensitive),
	sx::sxGetInBrack(Q,'[',']',Qt3),
	sx::sxUnpack(Qt3,';',Choices),
	wcrQuestionToNL_Choices(Choices,Qt4),
	sx::sxlJoin(Qt4,";",Qt5),
	sx::concat(["[",Qt5,"]"],Qt),
	!
;	
    Qt=Q,
	!.
	
% translates the question from IF to NL - this involves translating parameters and anser alternatives in the question text from IF to NL
%
wcrQuestionToNL(Q,Qt) :-
    Pos=string::searchchar(Q,'?',string::caseSensitive),
    string::front(Q,Pos,Qproper,Qtail),
	sx::sxGetInBrack(Qtail,'[',']',_),
	!, % with answer alternatives
    wcrQuestionAltsToNL(Qtail,Qt5),
	wcrQuestionToNL_Pars(Qproper,Qp),
	sx::concat([Qp,"? ",Qt5],Qt),
	!
;	
	wcrQuestionToNL_Pars(Q,Qt2),
	sx::sxGiveTail(Qt2,"?",Qt),
	!.
	
wcrQuestionToNL_Pars(Q,Qt) :-
	sx::searchchar(Q,'[',P2),
	sx::frontstr(P2,Q,F,T2),
	sx::searchchar(T2,']',P3),
	P4=P3-1,
	P4>1,
	sx::frontstr(P4,T2,Par2,T3),
	sx::sxStrip(Par2,Par),
	dbBase::nlSingF(Par,NL),
	sx::frontchar(T3,_,T),
	wcrQuestionToNL_Pars(T,Tt),
	sx::sxTail(1,F,FF,_),
	sx::concat([FF,"[",NL,"]",Tt],Qt),
	!
;
	sx::searchchar(Q,'[',_),
	Mess=string::concat("Invalid parameter in data rule question: ",Q,"!"),
	dlgIF::dlgNote(2,Mess),!,fail
;
	Qt=Q,
	!.		

wcrQuestionToNL_Choices([],[]) :-
	!.
wcrQuestionToNL_Choices([Hs|T],[Hk|Tk]) :-
	sx::frontchar(Hs,'"',_),
	sx::sxTail(1,Hs,L,_),
	string::subChar(L,0)='"',
	Hk=Hs,
	wcrQuestionToNL_Choices(T,Tk),
	!
;
	_=sx::toReal(Hs),
	Hk=Hs,
	wcrQuestionToNL_Choices(T,Tk),
	!
;
	sx::sxStrip(Hs,Hs2),
	dbBase::nlSingF(Hs2,Hk),
	wcrQuestionToNL_Choices(T,Tk),
	!	
;
	Mess=string::concat("Invalid answer choice: ",Hs,"!"),
	dlgIF::dlgNote(1,Mess),
	!,fail.		
% --------------------------------------------------------------------------------------------
% wcrSetType
% --------------------------------------------------------------------------------------------
wcrSetType(S,SetType) :-
	set::member(S,["p3_iZsand","p3_iZaand","p3_iZband","p3_iZdand"]),
	SetType="p3_iZand",
	!
;
	set::member(S,["p3_iZsor","p3_iZaor","p3_iZbor","p3_iZdor"]),
	SetType="p3_iZor",
	!
;
	SetType=S,
	!.
% ----------------------------------------------------------------------------------------------
% wcrTransQuery/Question
% ----------------------------------------------------------------------------------------------
wcrTransQuestion(S,Query) :-
	sx::sxStrip(S,Q2),
	Q2<>"",
	wcrQuestionToIF(Q2,Q3),
	Query=str(Q3),
	!
;
	sx::sxStrip(S,Q2),
	Q2<>"",
	dlgIF::dlgNote(1,"Invalid question format"),
	!,fail
;
	dlgIF::dlgNote(1,"No question entered"),
	!,fail.
		
wcrTransQuery(Qu,Query,QDom) :-
	sx::sxStrip(Qu,Q2),
	sx::sxGiveTail(Q2,"?",Q),
	dbBase::nlSing("p3_iIsTr",NL),
	sx::searchstring(Q,NL,Pos),
	Pos=1,
	% is it true that ... query
	!,transnl::transQuery(Q,Query,QDom),
	!
;
	sx::sxStrip(Qu,Q2),
	sx::sxGiveTail(Q2,"?",Q),
	dbBase::nlForm("p3_iWha","p3_iNIL",What),
	sx::concat([What,"=",Q],Qq),
	transnl::transQuery(Qq,Query,QDom),
	!.
% ----------------------------------------------------------------------------------------------
% queryInpFac
% ----------------------------------------------------------------------------------------------
queryInpFac(QueryText,Format)=Set :-
    Format="p3_iNL",
    irBase::wcrTransQuery(QueryText,Query,_),
	Query=cmp("p3_iNdrf",[_,str(DERF),str(Qa)]),
	Term=termDomain::fromTermStringIndirect(Qa),
	Set=piex::plcNDERF_Set(DERF,Term),
	!
;	
    Format="p3_iPR",
	transpr::proPR_Term(QueryText,Query),
	Query=cmp("p3_iNdrf",[_,str(DERF),str(Qa)]),
	Term=termDomain::fromTermStringIndirect(Qa),
	Set=piex::plcNDERF_Set(DERF,Term),
	!
;
    Mess=string::concatList(["Failed to translate/run question-series query '",QueryText,"'"]),
    dlgIF::dlgNote(3,Mess),!,fail.
% ----------------------------------------------------------------------------------------------
% unpackQuestion
% ----------------------------------------------------------------------------------------------
class predicates
crtQuestionAlts : (string,string_list,string_list) determ (i,i,o).
clauses

unpackQuestion(Q,Dom,Question,AltsNL) :-
    Pos=string::searchchar(Q,'?',string::caseSensitive),
    Posn=Pos+1,
    string::front(Q,Posn,Question,Qt),
	sx::sxGetInBrack(Qt,'[',']',Alts),
	sx::sxUnpack(Alts,';',AltsX),
	crtQuestionAlts(Dom,AltsX,AltsNL),
	!
;	
	Question3=string::trim(Q),
	sx::sxGiveTail(Question3,"?",Question),
    AltsNL=[],
    !.

% translates the question alternatives to NL
%
crtQuestionAlts(_,[],[]) :-
	!.
crtQuestionAlts(Dom,[Alt|Alts],[AltNL|AltsNL]) :-
	Dom="p3_dif",
	dbBase::nlSingF(Alt,AltNL),
	crtQuestionAlts(Dom,Alts,AltsNL),
	!
;	
	AltNL=Alt,
	crtQuestionAlts(Dom,Alts,AltsNL),
	!.	
% ----------------------------------------------------------------------------------------------
% setQuestionPars/InpFacPars
% ----------------------------------------------------------------------------------------------
setQuestionPars(Rule,Q,[],_,Qp) :-
	sx::searchchar(Q,'[',_),
	Mess=string::concatList(["Cannot determine actual value for embedded parameter in question rule: ",Rule," with text: ",Q,"!"]),
	dlgIF::dlgNote(3,Mess),!,fail
;
	Qp=Q,
	!.
setQuestionPars(Rule,Q,[fstr(Par,Dom,Form,_)|Pars],[APar|APars],Qp) :-
	vars::varFuncPar(Rule,Par,Param),
	sx::concat(["[",Param,"]"],P),
	_=string::search(Q,P,string::caseSensitive()),
	nl::conValX_F(APar,Dom,Form,V),
	Qm=sx::replaceAll(Q,P,V),
	setQuestionPars(Rule,Qm,Pars,APars,Qp),
	!
;
	sx::concat(["[P3_iThis__",Par,"]"],P),
	nl::conValX_F(APar,Dom,Form,V),
	Qm=sx::replaceAll(Q,P,V),
	setQuestionPars(Rule,Qm,Pars,APars,Qp),
	!.
	
setInpFacPars(Rule,Q,[],_,Qp) :-
	sx::searchchar(Q,'[',_),
	Mess=string::concatList(["Cannot determine actual value for embedded parameter in inexact rule input facility data: ",Rule," with text: ",Q,"!"]),
	dlgIF::dlgNote(3,Mess),!,fail
;
	Qp=Q,
	!.

setInpFacPars(Rule,Q,[Par|Pars],[APar|APars],Qp) :-
	sx::concat(["[",Par,"]"],P),
	_=string::search(Q,P,string::caseSensitive()),
	dbBase::smClassDef(Par,"a",_,_,Dom,Form,_),
	nl::conValX_F(APar,Dom,Form,V),
	Qm=sx::replaceAll(Q,P,V),
	setInpFacPars(Rule,Qm,Pars,APars,Qp),
	!
;	
	sx::concat(["[P3_iThis__",Par,"]"],P),
	_=string::search(Q,P,string::caseSensitive()),
	dbBase::smClassDef(Par,"a",_,_,Dom,Form,_),
	nl::conValX_F(APar,Dom,Form,V),
	Qm=sx::replaceAll(Q,P,V),
	setInpFacPars(Rule,Qm,Pars,APars,Qp),
	!.
% ----------------------------------------------------------------------------------------------
% crtMapN/D
% ----------------------------------------------------------------------------------------------
crtMapN(int(F),Fs,Cs,C) :-
	FF=F,
	crtMapN(fix(FF),Fs,Cs,C),
	!.
crtMapN(fix(F),Fs,Cs,C) :-
	Fs=[Fa|_],
	F<=Fa,
	Cs=[Ca|_],
	C=Ca,
	!
;
	Fs=[Fa,Fb|_],
	F<Fb,	
	Cs=[Ca,Cb|_],
	C=Ca+((F-Fa)/(Fb-Fa))*(Cb-Ca),
	!
;
	Fs=[_],
	Cs=[Ca],
	C=Ca,
	!
;
	Fs=[_|Ft],
	Cs=[_|Ct],
	crtMapN(fix(F),Ft,Ct,C),
	!
;
	dlgIF::feil_pie("crtMapN",[],[],nill),!,fail.		

crtMapD(V,[V|_],[C|_],C) :-
	!.
crtMapD(fix(F),[int(I)|_],[C|_],C) :-
	F=I,
	!.
crtMapD(int(I),[fix(F)|_],[C|_],C) :-
	I=F,
	!.
crtMapD(V,[_|Rs],[_|Ms],C) :-
	crtMapD(V,Rs,Ms,C),
	!.
% ----------------------------------------------------------------------------------------------
% evaluateCondition
% ----------------------------------------------------------------------------------------------
class predicates	
evaluateConditionF : (string,real,string,real,real) determ (i,i,i,i,i).
evaluateConditionS : (string,string,string) determ (i,i,i).
clauses

evaluateCondition(Oper1,str(Op1),_,_,str(C)) :-
	!, 
	evaluateConditionS(Oper1,Op1,C),
	!.
evaluateCondition(Oper1,atom(Op1),_,_,atom(C)) :-
	!, 
	evaluateConditionS(Oper1,Op1,C),
	!.
evaluateCondition(Oper1,Operator1,_,nill,Val) :-
	!,
	irbase::crtToFix(Operator1,Op1),
	irbase::crtToFix(Val,C),
	evaluateConditionF(Oper1,Op1,"",0.0,C),
	!.
evaluateCondition(Oper1,Operator1,Oper2,Operator2,Val) :-
	!,
	irbase::crtToFix(Operator1,Op1),
	irbase::crtToFix(Operator2,Op2),
	irbase::crtToFix(Val,C),
	evaluateConditionF(Oper1,Op1,Oper2,Op2,C),
	!.
evaluateConditionS("=",Op1,C) :- C=Op1,!.
evaluateConditionS("<>",Op1,C) :- C<>Op1,!.
	
evaluateConditionF("=",Op1,_,_,C) :- C=Op1,!.
evaluateConditionF("<>",Op1,_,_,C) :- C<>Op1,!.
evaluateConditionF(">",Op1,"",_,C) :- C>Op1,!.
evaluateConditionF(">=",Op1,"",_,C) :- C>=Op1,!.
evaluateConditionF("<",Op1,"",_,C) :- C<Op1,!.
evaluateConditionF("<=",Op1,"",_,C) :- C<=Op1,!.
evaluateConditionF(">",Op1,">",Op2,C) :- C>Op1, C>Op2,!.
evaluateConditionF(">=",Op1,">",Op2,C) :- C>=Op1, C>Op2,!.
evaluateConditionF("<",Op1,">",Op2,C) :- C<Op1, C>Op2,!.
evaluateConditionF("<=",Op1,">",Op2,C) :- C<=Op1, C>Op2,!.
evaluateConditionF(">",Op1,">=",Op2,C) :- C>Op1, C>=Op2,!.
evaluateConditionF(">=",Op1,">=",Op2,C) :- C>=Op1, C>=Op2,!.
evaluateConditionF("<",Op1,">=",Op2,C) :- C<Op1, C>=Op2,!.
evaluateConditionF("<=",Op1,">=",Op2,C) :- C<=Op1, C>=Op2,!.
evaluateConditionF(">",Op1,"<",Op2,C) :- C>Op1, C<Op2,!.
evaluateConditionF(">=",Op1,"<",Op2,C) :- C>=Op1, C<Op2,!.
evaluateConditionF("<",Op1,"<",Op2,C) :- C<Op1, C<Op2,!.
evaluateConditionF("<=",Op1,"<",Op2,C) :- C<=Op1, C<Op2,!.
evaluateConditionF(">",Op1,"<=",Op2,C) :- C>Op1, C<=Op2,!.
evaluateConditionF(">=",Op1,"<=",Op2,C) :- C>=Op1, C<=Op2,!.
evaluateConditionF("<",Op1,"<=",Op2,C) :- C<Op1, C<=Op2,!.
evaluateConditionF("<=",Op1,"<=",Op2,C) :- C<=Op1, C<=Op2,!.
evaluateConditionF(">",Op1,"=",Op2,C) :- C>Op1, C=Op2,!.
evaluateConditionF(">=",Op1,"=",Op2,C) :- C>=Op1, C=Op2,!.
evaluateConditionF("<",Op1,"=",Op2,C) :- C<Op1, C=Op2,!.
evaluateConditionF("<=",Op1,"=",Op2,C) :- C<=Op1, C=Op2,!.
% ----------------------------------------------------------------------------------------------
% crtReasonTree/Map
% ----------------------------------------------------------------------------------------------
crtReasonTreeAdd(T,Ts,Tree) :-
	T=vpiTree::tree(A,B,[],D),
	Tree=vpiTree::tree(A,B,Ts,D),
	!
;
	T=vpiTree::tree(A,B,C,D),
	set::conc([Ts,C],TsM),
	Tree=vpiTree::tree(A,B,TsM,D),
	!.

% parameters
% 1	in	node name (rule name or other)
% 2	in	rule's value (or blank if n/a)
% 3	in	|-separated list with 0-3 components
%        blank
%                 redundant tree node
%        prefix|suffix|type
%        |type
%        suffix|type
%                +|... mapping failed
%                +|P  criterion parameter failed
%        suffix
%                 * ooc
%                 + query failed
% 4	in	rule formal parameters (if relevant)
% 5	in	rule actual parameters (if relevant)
% 6	in	sub-trees
% 7	out	tree
%
crtReasonTree(Node,Val,S,Pars,APars,Trees,Tree) :-
	sx::sxUnpack(S,'|',L),
	L=[Pref,Suf,Typ],
	crtReasonName(Node,Val,nill,Pref,Suf,Typ,Pars,APars,Nm,N),
	Tree=vpiTree::tree(Nm,vpiTree::unmarked,Trees,N),
	!
;
	sx::frontchar(S,'|',T),
	crtReasonName(Node,Val,nill,"","",T,Pars,APars,Nm,N),
	Tree=vpiTree::tree(Nm,vpiTree::unmarked,Trees,N),
	!
;
	sx::sxUnpack(S,'|',L),
	L=[Suf,Typ],
	crtReasonName(Node,Val,nill,"",Suf,Typ,Pars,APars,Nm,N),
	Tree=vpiTree::tree(Nm,vpiTree::unmarked,Trees,N),
	!
;
	crtReasonName(Node,Val,nill,"",S,"",Pars,APars,Nm,N),
	Tree=vpiTree::tree(Nm,vpiTree::unmarked,Trees,N),
	!.
	
crtReasonTreeMap(Node,MapFrom,MapTo,Suffix,Pars,APars,Trees,Tree) :-
	crtReasonName(Node,MapTo,MapFrom,"",Suffix,"",Pars,APars,Nm,N),
	Tree=vpiTree::tree(Nm,vpiTree::unmarked,Trees,N),
	!.
% ----------------------------------------------------------------------------------------------
% crtReasonName
% ----------------------------------------------------------------------------------------------
% Prefix: 
% Suffix:
% Type: 
crtReasonName(Rule,Val,MapFrom,Prefix,Suffix,Typ,Pars,APars,Node,N) :-
	Rule=atom(R),
	globalVars::incr("crcnt",_,N),
	dbBase::smClassDef(R,"h",_,_,Dom,_,_),
	termDomain::listBR(ParsL,Pars),
	termDomain::listBR(AParsL,APars),
	Term=cmp("",[Rule,Val,MapFrom,atom(Dom),str(Typ),ParsL,AParsL,str(Prefix),str(Suffix),int(N)]),
	Node=termDomain::toTermStringIndirect("n",Term),
	!
;
	globalVars::incr("crcnt",_,N),
	termDomain::listBR(ParsL,Pars),
	termDomain::listBR(AParsL,APars),
	Term=cmp("",[Rule,Val,MapFrom,atom("p3_dany"),str(Typ),ParsL,AParsL,str(Prefix),str(Suffix),int(N)]),
	Node=termDomain::toTermStringIndirect("n",Term),
	!.
% ----------------------------------------------------------------------------------------------
% crtTreePrefix
% ----------------------------------------------------------------------------------------------
crtTreePrefix(Prefix,vpiTree::tree(N,B,C,D),Tm) :-
    Term=termDomain::fromTermStringIndirect(N),
    Term=cmp(H,[atom(Rule)|T]),
    dbBase::nlSingF(Rule,RuleN),
    R=string::concat(Prefix,RuleN),
    TermM=cmp(H,[str(R)|T]),
    Nm=termDomain::toTermStringIndirect("n",TermM),
	Tm=vpiTree::tree(Nm,B,C,D),
	!
;	
    % @@ not sure if this clause is required 11.4.6
    Term=termDomain::fromTermStringIndirect(N),
    Term=cmp(H,[_|T]),
    TermM=cmp(H,[str(Prefix)|T]),
    Nm=termDomain::toTermStringIndirect("n",TermM),
	Tm=vpiTree::tree(Nm,B,C,D),
	!.
	
crtTreePrefixes(_,[],[]) :-
	!.
crtTreePrefixes(Prefix,[T|Ts],[Tm|TmS]) :-
	crtTreePrefix(Prefix,T,Tm),
	crtTreePrefixes(Prefix,Ts,TmS),
	!.
% ----------------------------------------------------------------------------------------------
% crtSaveTree
% ----------------------------------------------------------------------------------------------
crtSaveTree(ST,Arg,DE,APars,Tree) :-
	globalVars::get_i("crlog")=1,
	termDomain::listBR(AParsL,APars),
	AParsT=termDomain::toTermStringIndirect("p",AParsL),
	TreeTerm=conversion::fromTree(Tree),
	TermString=termDomain::toTermStringIndirect("t",TreeTerm),
	LT=cmp("r",
	      [cmp("u",
	          [str(ST),
	           atom(Arg),
	           atom(DE),
	           str(AParsT),
	           str(TermString)])]),
	dbBase::recordInsert(1,"$IRR",0,0,LT,atom("true")),
	!
;
	!.
% ----------------------------------------------------------------------------------------------
% crtSetQueryPars
% ----------------------------------------------------------------------------------------------
crtSetQueryPars(_,Query,[],_,Query) :-
	!.
crtSetQueryPars(Rule,Query,[Par|Pars],[APar|APars],QueryM) :-
	vars::varFuncPar(Rule,Par,V),
	struc::treReplace(Query,var(V),APar,Q2),
	vars::varFuncPar("p3_iThis",Par,V2),
	struc::treReplace(Q2,var(V2),APar,Q3),
	crtSetQueryPars(Rule,Q3,Pars,APars,QueryM),
	!.
% ----------------------------------------------------------------------------------------------
% crtSubstActPars
% ----------------------------------------------------------------------------------------------
class predicates
crtEvalPar : (sterm,sterml,sterml,sterm) determ (i,i,i,o).
crtIsPar : (string,sterml,sterml,sterm) determ (i,i,i,o).
clauses

% parameters:
% 1	in	actual parameter assignments in called rule
% 2	in	formal parameters in calling rule + variables in parameter-rule
% 3	in	values corresponding to par#2
% 4	out	values corresponding to par#1
%
crtSubstActPars([],_,_,[]) :-
	!.
crtSubstActPars([APar|APars],Pars,Vals,[Value|Values]) :-
	APar=cmp(_,[_,RS]),
	crtEvalPar(RS,Pars,Vals,Value),
	crtSubstActPars(APars,Pars,Vals,Values),
	!.
crtEvalPar(var(Par),Pars,Vals,Val) :-
	vars::varFuncParR(_,DE,Par),
	crtIsPar(DE,Pars,Vals,Val),
	!
;
	% variable used in parameter-rule
	crtIsPar(Par,Pars,Vals,Val),
	!.
crtEvalPar(V,_,_,Vb) :-
	piex::plcBasicValue(V,Vb),
	!.
crtEvalPar(RS,_,_,_) :-
	dlgIF::feil_pie("crtEvalPar",["Cannot determine value of actual parameter"],[],RS),!,fail.
crtIsPar(DE,[fstr(DE,_,_,_)|_],[Val|_],Val) :-
	!.
crtIsPar(Var,[var(Var)|_],[Val|_],Val) :-
	!.
crtIsPar(DE,[_|Pars],[_|Vals],Val) :-
	crtIsPar(DE,Pars,Vals,Val),
	!.
% ----------------------------------------------------------------------------------------------
% crtMatchPars
% ----------------------------------------------------------------------------------------------
crtMatchPars(_,[]) :-
	!. % the called rule in question is parameter-less
crtMatchPars([P],[Q]) :-
	P=Q,
	!
;
	!,fail.
crtMatchPars([P|Ps],[Q|Qs]) :-
	crtMatchPars([P],[Q]),
	crtMatchPars(Ps,Qs),
	!.		
/*
% ----------------------------------------------------------------------------------------------
% crtFormat   
% ----------------------------------------------------------------------------------------------

crtFormat : (string,sterm,sterm) determ (i,i,o).
class predicates
crtFormatX : (string,sterm,sterm) procedure (i,i,o).
clauses

crtFormat(Rule,V,Vf) :-
	dbExtension::dbuCR_GetTD(Rule,_,Dom),
	crtFormatX(Dom,V,Vf),
	!.
crtFormatX("p3_din",int(I),fint(I,"")) :- !.
crtFormatX("p3_dnu",fix(F),ffix(F,"")) :- !.
crtFormatX("p3_dnu",int(I),ffix(F,"")) :- F=I,!.
crtFormatX("p3_dcf",fix(F),ffix(F,".3")) :- !.
crtFormatX("p3_dinc",fix(F),ffix(F,".3")) :- !.
crtFormatX("p3_ddec",fix(F),ffix(F,".3")) :- !.
crtFormatX("p3_dpr",fix(F),ffix(F,".3")) :- !.
crtFormatX("p3_dcf",int(I),ffix(F,".3")) :- F=I,!.
crtFormatX("p3_dinc",int(I),ffix(F,".3")) :- F=I,!.
crtFormatX("p3_ddec",int(I),ffix(F,".3")) :- F=I,!.
crtFormatX("p3_dpr",int(I),ffix(F,".3")) :- F=I,!.
crtFormatX(_,V,V) :- !.
*/
% ----------------------------------------------------------------------------------------------
% crtCombineBay
% ----------------------------------------------------------------------------------------------
crtCombineBay(ClPrior,ClVal,Inc,Dec,RestVal,P) :-
					tt::tt_n(4,"CR Bay ClPrior",ClPrior),
					tt::tt_n(4,"CR Bay ClVal",ClVal),
					tt::tt_n(4,"CR Bay Inc",Inc),
					tt::tt_n(4,"CR Bay Dec",Dec),
					tt::tt_n(4,"CR Bay RestVal",RestVal),
	irbase::probtoodds(RestVal,RestOdds),
	irbase::bayesfactor(ClVal,ClPrior,Dec,Inc,F),
	Odds=F*RestOdds,
	irbase::oddstoprob(Odds,P),
					tt::tt_n(4,"CR Bay RestOdds",RestOdds),
					tt::tt_n(4,"CR Bay F",F),
					tt::tt_n(4,"CR Bay RestOdds",RestOdds),
					tt::tt_n(4,"CR Bay Odds",Odds),
					tt::tt_n(4,"CR Bay P",P),
	!.
% ----------------------------------------------------------------------------------------------
% crtCombine
% ----------------------------------------------------------------------------------------------
crtCombine(C2,C3,C) :-
	C2a=math::abs(C2),
	C2a=1.0,
	C=C2+C3,
	C=0.0,
	!
;
	C2>0.0,
	C3>0.0,
	C=C2+C3-C2*C3,
	!
;
	C2<0.0,
	C3<0.0,
	C=C2+C3+C2*C3,
	!
;
	C2a=math::abs(C2),
	C3a=math::abs(C3),
	C2a<C3a,
	C=(C2+C3)/(1.0-C2a),
	!
;
	C3a=math::abs(C3),
	C=(C2+C3)/(1.0-C3a),
	!.
% ----------------------------------------------------------------------------------------------
% crtToFix
% ----------------------------------------------------------------------------------------------
crtToFix(fix(F),F) :- !.
crtToFix(int(I),F) :- F=I,!.
crtToFix(V,_) :- dlgIF::feil_pie("crtToFix",["Expected a fix(..) or an int(..) term"],[],V),!,fail.
% ----------------------------------------------------------------------------------------------
% direct from pexpert.pro
% ----------------------------------------------------------------------------------------------
bayesfactor(Prob,Prior,Dec,_,F):-
	Prior >= Prob,!,
	probtocert(Prob,Prior,Cert),
	F = ((1.0 - Dec) * (Cert + 1.0) / 1.0) + Dec,
	!.
bayesfactor(Prob,Prior,_,Inc,F):-
	probtocert(Prob,Prior,Cert),
	F = 1.0 + (Inc-1.0) * Cert / 1.0,
	!.

certtoprob(Cert,Prior,Value):-
	Cert <= 0.0, 
	Value = Prior * (Cert + 1.0) / 1.0,
			tt::tt_n(4,"CR Certainty",Cert),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Probability",Value),
	!
;
	Value = Prior + (1.0 - Prior) * Cert / 1.0,
			tt::tt_n(4,"CR Certainty",Cert),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Probability",Value),
	!.

probtocert(_,0.0,-1.0) :-
			tt::tt_ss(4,"CR Prior 0.0"),
			tt::tt_ss(4,"CR Certainty -1.0"),
	!.
probtocert(_,1.0,1.0) :- 
			tt::tt_ss(4,"CR Prior 1.0"),
			tt::tt_ss(4,"CR Certainty 1.0"),
	!.
probtocert(Prob,Prior,Cert) :-
	Prob <= Prior, 
	Cert = 1.0 * Prob / Prior - 1.0,
			tt::tt_n(4,"CR Probability",Prob),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Certainty",Cert),
	!
;
	Cert = 1.0 * (Prob - Prior) / (1.0 - Prior),
			tt::tt_n(4,"CR Probability",Prob),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Certainty",Cert),
	!.

oddstoprob(Odds,Prob) :-
	Prob=Odds/(1.0+Odds)
	.
probtoodds(1.0,10000.0) :- 
	!.
probtoodds(Prob,Odds):-
	Odds=Prob/(1.0-Prob)
	.
/* original:
bayesfactor(Prob,Prior,Dec,_,F):-
	Prior >= Prob,!,
	probtocert(Prob,Prior,Cert),
	F = ((1.0 - Dec) * (Cert + 5.0) / 5.0) + Dec,
	!.
bayesfactor(Prob,Prior,_,Inc,F):-
	probtocert(Prob,Prior,Cert),
	F = 1.0 + (Inc-1.0) * Cert / 5.0,
	!.
certtoprob(Cert,Prior,Value):-
	Cert <= 0.0, 
	Value = Prior * (Cert + 5.0) / 5.0,
			tt::tt_n(4,"CR Certainty",Cert),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Probability",Value),
	!
;
	Value = Prior + (1.0 - Prior) * Cert / 5.0,
			tt::tt_n(4,"CR Certainty",Cert),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Probability",Value),
	!.
probtocert(_,0.0,-5.0) :-
			tt::tt_ss(4,"CR Prior 0.0"),
			tt::tt_ss(4,"CR Certainty -5.0"),
	!.
probtocert(_,1.0,5.0) :- 
			tt::tt_ss(4,"CR Prior 1.0"),
			tt::tt_ss(4,"CR Certainty 5.0"),
	!.
probtocert(Prob,Prior,Cert) :-
	Prob <= Prior, 
	Cert = 5.0 * Prob / Prior - 5.0,
			tt::tt_n(4,"CR Probability",Prob),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Certainty",Cert),
	!
;
	Cert = 5.0 * (Prob - Prior) / (1.0 - Prior),
			tt::tt_n(4,"CR Probability",Prob),
			tt::tt_n(4,"CR Prior",Prior),
			tt::tt_n(4,"CR Certainty",Cert),
	!.
*/
end implement irBase
