/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement irReport
open core,termDomain

constants
className = "com/rudstrand/irReport/irReport".
classVersion = "1.0".

class predicates
reasoning : (rep,string,vpiTree::tree) determ.
rsRule : (rep,string,sterm,sterm,sterm,string,vpiTree::tree_list) determ.
rsParameters : (rep,string_list,sterm) determ.
rsType : (rep,integer,string,sterm,vpiTree::tree_list) determ.
rsCalls : (rep,integer,vpiTree::tree_list) determ.
getWeight : (string,sterml,vpiTree::tree_list,string) determ (i,i,i,o).
callSuffix : (string,string) procedure (i,o).
callPars : (sterm,sterm,string) determ (i,i,o).
rsOwaCalls : (rep,integer,sterml,vpiTree::tree_list,vpiTree::tree_list) determ.
rsSimpleCall : (rep,integer,string,vpiTree::tree) determ.    
rsCall : (vpiTree::tree,string,string) procedure (i,i,o).    
rsContext : (rep,vpiTree::tree_list,integer,vpiTree::tree_list) determ (i,i,o,o).
rsRemoveCalls : (vpiTree::tree_list,vpiTree::tree_list) determ (i,o).
rsSwitchCalls : (rep,vpiTree::tree_list) determ.
rsParCalls : (rep,vpiTree::tree_list) determ.
weightCall : (string,sterml,vpiTree::tree_list,string) determ (i,i,i,o).
ruleSequence : (lst,vpiTree::tree) determ.
getSubtree : (rep,string,string,vpiTree::tree) determ.

adjustAnno : (integer,string,string) determ (i,i,o).
annotation : (rep,string,string,integer) determ.
irActualParameterList : (sterml,sterml,string) determ (i,o,o).
irAssociation : (rep,string).
irBasic : (sterm,string) determ (i,o).
irBasicValue : (sterm,string) determ (i,o).
irCall : (rep,integer,string,sterm) determ.
irContext : (rep,sterml) determ.
irHeading : (rep,string,sterm,string) determ.
irList0 : (rep,string,sterml,sterml) determ (i,i,i,o).
irList1 : (rep,string,sterml,sterml) determ (i,i,i,o).
irList1b : (rep,string,sterml,sterml) determ (i,i,i,o).
irList1Parameter : (rep,string,sterm,sterm,sterml) determ.
irList2 : (rep,string,sterml,sterml) determ (i,i,i,o).
irMap : (rep,sterml).
irParameter : (rep,string,sterm) determ.
irParameterList : (rep,sterml,sterml) determ (i,i,o).
irQuery : (rep,sterm) determ.
irType : (rep,sterm,string).
ruleText : (rep,string) determ.

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------
% reportReasoning
% -------------------------------------------------------------------------------------
class predicates
reasonRule : (rep,lst,vpiTree::tree).
clauses

reportReasoning(Tree) :-
    Kb=dbBase::dbName(),
    File=string::concat(Kb,"_InexactRuleReasoning"),
    Rep=rep::newReport(File,100),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	p3Rep::header(Rep),
	List=lst::new(1),
	ruleSequence(List,Tree),
	List:sort,
	reasonRule(Rep,List,Tree),
	Rep:close(),
	!.

reasonRule(Rep,List,Tree) :-
	Rule=List:get_nd(),
        dbBase::kbID(Rule,K,_,_),
        getSubtree(Rep,Rule,K,Tree),
        fail
;
    !.
% -------------------------------------------------------------------------------------
% ruleSequence
% -------------------------------------------------------------------------------------
class predicates
ruleSequenceX : (lst,vpiTree::tree_list) determ.
clauses

ruleSequenceX(_,[]) :- 
    !.
ruleSequenceX(List,[H|T]) :- 
    ruleSequence(List,H), 
    ruleSequenceX(List,T),
    !.

ruleSequence(List,vpiTree::tree(S,_,L,_)) :-
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule)|_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[_,RuleN],
	List:add(RuleN),
	ruleSequenceX(List,L),
	!
;
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[atom(Rule)|_]),
	dbBase::nlSing(Rule,RuleN),
	List:add(RuleN),
	ruleSequenceX(List,L),
	!.
% -------------------------------------------------------------------------------------
% getSubtree
% -------------------------------------------------------------------------------------
class predicates
getSubtreeX : (rep,string,string,vpiTree::tree_list) determ.
clauses

getSubtree(Rep,Rn,Rk,Tree) :-
    Tree=vpiTree::tree(S,_,L,_),
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule)|_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[_,Rn],
    reasoning(Rep,Rk,Tree),
    getSubtreeX(Rep,Rn,Rk,L),
	!
;
    Tree=vpiTree::tree(S,_,L,_),
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[atom(Rk)|_]),
    reasoning(Rep,Rk,Tree),
    getSubtreeX(Rep,Rn,Rk,L),
	!
;
    Tree=vpiTree::tree(_,_,L,_),
    getSubtreeX(Rep,Rn,Rk,L),
    !.
    
getSubtreeX(Rep,Rn,Rk,[H|T]) :-
    getSubtree(Rep,Rn,Rk,H),
    !
;
    getSubtreeX(Rep,Rn,Rk,T),
    !.

% -------------------------------------------------------------------------------------
% reasoning
% -------------------------------------------------------------------------------------
reasoning(Rep,Rk,vpiTree::tree(S,_,L,_)) :-
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[_,Val,MapFrom,_,_,_,Apars,_,termDomain::str(Suffix),_]),
	rsRule(Rep,Rk,Val,MapFrom,Apars,Suffix,L),
	!.
% -------------------------------------------------------------------------------------
% rsRule
% -------------------------------------------------------------------------------------
class predicates
mapping : (sterm,string) determ (i,o).
clauses

rsRule(Rep,Rule,Val,MapFrom,Apars,_,SubTrees) :-
	dbBase::getIrDefF(Rule,Body,_),
    dbBase::irType(Rule,_,SubType),
    dbBase::smClassDef(Rule,"h",_,DEs,Dom,_,_),
    dbBase::nlSing(Rule,RuleN),
                    dlgIF::msgS(RuleN),
    dbBase::nlSing(SubType,TypeN),
    nl::conValX(Val,Dom,"",Vn),
    mapping(MapFrom,Map),
    dbBase::nlSing(Dom,DomN),
	Rep:repW(0,RuleN,"=",Vn," (",DomN,")"),
	Rep:repWL(0,Map," [",TypeN,"]"),
	rsContext(Rep,SubTrees,Context,SubTrees2),
	rsRemoveCalls(SubTrees2,SubTrees3),
	rsParameters(Rep,DEs,Apars),
 	annotation(Rep,"p3_iIRdec",Rule,5),
	rsType(Rep,Context,SubType,Body,SubTrees3),
    irAssociation(Rep,Rule),
    Rep:repWL(0,""),
	!.

mapping(nill,"") :- !.
mapping(V,S) :- nl::conValX(V,"p3_dany","",Vs), S=string::concat(" MAPPED FROM ",Vs),!.
% -------------------------------------------------------------------------------------
% rsRemoveCalls
% -------------------------------------------------------------------------------------
rsRemoveCalls([],[]) :-
    !.
rsRemoveCalls([Tree|Trees],TreesM) :-
    Tree=vpiTree::tree(S,_,_,_),
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule)|_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[S2,_],
	S2="WEIGHT",
	rsRemoveCalls(Trees,TreesM),
	!
;	
	rsRemoveCalls(Trees,Trees2),
	TreesM=[Tree|Trees2],
	!.
% -------------------------------------------------------------------------------------
% rsContext
% -------------------------------------------------------------------------------------
class predicates
rsContextOut : (rep,vpiTree::tree_list) determ.
rsContextIn : (rep,vpiTree::tree_list,integer,integer,vpiTree::tree_list) determ (i,i,i,o,o).
clauses

rsContext(Rep,Trees,Context,NcTrees) :-
    rsContextOut(Rep,Trees),
    Context=-1,
    NcTrees=[],
    !
;
    rsContextIn(Rep,Trees,0,Context,NcTrees),
    !.

rsContextOut(_,[]) :- 
    !,fail.
rsContextOut(Rep,[vpiTree::tree(S,_,_,_)|Trees]) :-
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule),Val,_,atom(Dom),_,Pars,Apars,_,termDomain::str(Suffix),_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[S2,RuleN],
	S2="OUT OF CONTEXT",
    nl::conValX(Val,Dom,"",ValS),
    callSuffix(Suffix,CallSuffix),
    callPars(Pars,Apars,Ps),
	T=string::concatList([RuleN,Ps,"=",Vals," ",CallSuffix]),
	Rep:repWL(5,S2,": ",T),
	!
;
    rsContextOut(Rep,Trees),
    !.

rsContextIn(_,[],C,C,[]) :-
    !.
rsContextIn(Rep,[Tree|Trees],ContextIn,ContextOut,NcTrees) :-
    Tree=vpiTree::tree(S,_,_,_),
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule),Val,_,atom(Dom),_,Pars,Apars,_,termDomain::str(Suffix),_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[S2,RuleN],
	S2="IN CONTEXT",
    nl::conValX(Val,Dom,"",ValS),
    callSuffix(Suffix,CallSuffix),
    callPars(Pars,Apars,Ps),
	T=string::concatList([RuleN,Ps,"=",Vals," ",CallSuffix]),
	Rep:repWL(5,S2,": ",T),
	rsContextIn(Rep,Trees,1,_,Trees2),
	NcTrees=Trees2,
	ContextOut=1,
	!
;	
	rsContextIn(Rep,Trees,ContextIn,ContextOut,Trees2),
	NcTrees=[Tree|Trees2],
	!.
% -------------------------------------------------------------------------------------
% rsParameters
% -------------------------------------------------------------------------------------
class predicates
rsParametersX : (rep,integer,string_list,sterm) determ.
clauses

rsParameters(Rep,[],_) :-
    Rep:repWL(5,"NO PARAMETERS"),
    !.
rsParameters(Rep,[H|T],list(V,L)) :-
    dbBase::nlSing(H,Hn),
    dbBase::smClassDef(H,"a",_,_,Dom,F,_),
    nl::conValX(V,Dom,F,Vn),
    Rep:repWL(5,"PARAMETERS: ",Hn,"=",Vn),
    rsParametersX(Rep,17,T,L),
    !.
    
rsParametersX(_,_,[],_) :-
    !.
rsParametersX(Rep,Ind,[H|T],list(V,L)) :-
    dbBase::nlSing(H,Hn),
    dbBase::smClassDef(H,"a",_,_,Dom,F,_),
    nl::conValX(V,Dom,F,Vn),
    Rep:repWL(Ind,Hn,"=",Vn),
    rsParametersX(Rep,Ind,T,L),    
    !.
% -------------------------------------------------------------------------------------
% rsType
% -------------------------------------------------------------------------------------
rsType(Rep,Context,Type,Body,SubTrees) :-
    Context=-1,
    !
;    
    Type="p3_iCnot",
    SubTrees=[T],
    rsSimpleCall(Rep,5,"COMPLEMENT OF: ",T),
    !
;    
    Type="p3_iCmap",
    SubTrees=[T],
    rsSimpleCall(Rep,5,"MAPPED FROM: ",T),
    !
;
    set::member(Type,["p3_iCmb","p3_iCand","p3_iCor"]),
    set::setGet(["p3_iCmb","p3_iCand","p3_iCor"],Ix,Type),
    set::setGet(["COMBINED RULES:","AND'ed RULES:","OR'ed RULES:"],Ix,Heading),
    Rep:repWL(5,Heading),
    rsCalls(Rep,5,SubTrees),
    !
;
    set::member(Type,["p3_iCbay"]),
    !
;
    Type="p3_iZnot",
    SubTrees=[T],
    rsSimpleCall(Rep,5,"COMPLEMENT OF: ",T),
    !
;
    set::member(Type,["p3_iZand","p3_iZor"]),
    Body=cmp(_,[_,cmp(SubType,_)|_]),
    set::member(SubType,["p3_iZsand","p3_iZaand","p3_iZdand","p3_iZband","p3_iZsor","p3_iZaor","p3_iZbor","p3_iZdor"]),
    dbBase::nlSing(SubType,Set),
    S=string::concat(Set," OF:"),
    Rep:repWL(5,S),
    rsCalls(Rep,5,SubTrees),
    !
;
    set::member(Type,["p3_iZowa"]),
    Rep:repWL(5,"ORDERED WEIGHTED AVERAGING SET OF:"),
    Body=cmp(_,[_,cmp(_,Weights)]),
    rsOwaCalls(Rep,5,Weights,SubTrees,SubTrees),
    !
;
    Type="p3_iCqry",
    Body=cmp(_,[_,cmp(_,[_,Q|_])|_]),
	nl::conQuery(Q,NL),
    irQuerySt(Rep,NL),
    !
;    
    Type="p3_iCqst",
	Body=cmp(_,[_,cmp(_,[_,str(Qs)]),_,_,_,_,_]),
	irBase::wcrQuestionToNL(Qs,NL),
    Rep:textBlock(5,"QUESTION: ",NL),
    !
;
    Type="p3_iCswc",
    rsSwitchCalls(Rep,SubTrees),
    !
;
    Type="p3_iCpar",
    rsParCalls(Rep,SubTrees),
    !
;
    Type="p3_iCswp",
    !.    
% -------------------------------------------------------------------------------------
% rsParCalls
% -------------------------------------------------------------------------------------
rsParCalls(_,[]) :-
    !.
rsParCalls(Rep,[Tree|Trees]) :-
    Tree=vpiTree::tree(S,_,_,_),
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule)|_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[S1,_],
	S1="PAR",
    rsSimpleCall(Rep,5,"PARAMETER CALL       : ",Tree),
    rsParCalls(Rep,Trees),
    !
;    
    rsSimpleCall(Rep,5,"PARAMETERIZED CALL: ",Tree),
    rsParCalls(Rep,Trees),
    !.
% -------------------------------------------------------------------------------------
% rsSwitchCalls
% -------------------------------------------------------------------------------------
rsSwitchCalls(Rep,[]) :-
    Rep:repWL(5,"** INVALID SWITCH RULE **"),
    !.
rsSwitchCalls(Rep,[_]) :-
    Rep:repWL(5,"** INVALID SWITCH RULE **"),
    !.
rsSwitchCalls(Rep,[Tree1,Tree2|Trees]) :-
    Tree1=vpiTree::tree(S,_,_,_),
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule)|_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[S1,_],
	S1="CRIT",
    rsSimpleCall(Rep,5,"SWITCHED CALL RETURN   : ",Tree2),
    rsSimpleCall(Rep,5,"SELECTED FROM CRITERION: ",Tree1),
    !
;    
    rsSwitchCalls(Rep,[Tree2|Trees]),
    !.
% -------------------------------------------------------------------------------------
% weightCall
% -------------------------------------------------------------------------------------
class predicates
weightCallX : (string,sterml,vpiTree::tree_list,string) determ (i,i,i,o).
matchPars : (sterml,sterm) determ.
clauses

weightCall(_,_,[],Qual) :-
    Qual="** INVALID WEIGHT CALL **",
    !.
weightCall(Wrule,Apars,[Tree|Trees],Qual) :-
    Tree=vpiTree::tree(_,_,L,_),
    weightCallX(Wrule,Apars,L,Qual),
    !
;    
    weightCall(Wrule,Apars,Trees,Qual),
    !.

weightCallX(Wrule,Apars,[Tree|SubTrees],Qual) :-
    Tree=vpiTree::tree(S,_,_,_),
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule),_,_,_,_,_,Apars2,_,_,_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[S1,Rule],
	S1="WEIGHT",
	dbBase::kbID(Rule,RuleK,_,_),
	Wrule=RuleK,
	matchPars(Apars,Apars2),
	rsCall(Tree," <-- ",Qual),
    !
;    
    weightCallX(Wrule,Apars,SubTrees,Qual),
    !.
    
matchPars([],nill) :-
    !.
matchPars([P|Ps],list(Px,L)) :-
    P=Px,
    matchPars(Ps,L),
    !.
% -------------------------------------------------------------------------------------
% rsSimpleCall
% -------------------------------------------------------------------------------------\
rsSimpleCall(Rep,Indent,Header,Tree) :-
    rsCall(Tree," = ",S),
    Rep:textBlock(Indent,Header,S),
    !.
% -------------------------------------------------------------------------------------
% rsCall
% -------------------------------------------------------------------------------------
rsCall(vpiTree::tree(S,_,_,_),Oper,T) :-
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule),Val,_,atom(Dom),_,Pars,Apars,_,termDomain::str(Suffix),_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[_,RuleN],
    nl::conValX(Val,Dom,"",ValS),
    callSuffix(Suffix,CallSuffix),
    callPars(Pars,Apars,Ps),
	T=string::concatList([Vals,Oper,RuleN,Ps,CallSuffix]),
	!
;
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[atom(Rule),Val,_,atom(Dom),_,Pars,Apars,_,termDomain::str(Suffix),_]),
    dbBase::nlSing(Rule,RuleN),
    nl::conValX(Val,Dom,"",ValS),
    callSuffix(Suffix,CallSuffix),
    callPars(Pars,Apars,Ps),
	T=string::concatList([Vals,Oper,RuleN,Ps,CallSuffix]),
	!
;
    T="** INVALID CALL SUB-TREE **",
    !.
% -------------------------------------------------------------------------------------
% rsCalls
% -------------------------------------------------------------------------------------
rsCalls(_,_,[]) :-
    !.
rsCalls(Rep,Indent,[Call|Calls]) :-
    rsCall(Call," = ",S),
	Rep:repWL(Indent,S),
	rsCalls(Rep,Indent,Calls),
	!.
% -------------------------------------------------------------------------------------
% callPars
% -------------------------------------------------------------------------------------
class predicates
callParsX : (sterm,sterm,string) determ (i,i,o).
clauses

callPars(Pars,Apars,S) :-
    callParsX(Pars,Apars,S),
    S="",
    !
;
    callParsX(Pars,Apars,S2),
    S=string::concat(" [",S2,"]"),
    !.
    
callParsX(nill,_,"") :-
    !.
callParsX(list(Par,nill),list(Apar,_),S) :-
    Par=fstr(P,Dom,F,_),
    dbBase::nlSing(P,Pn),
    nl::conValX(Apar,Dom,F,Vn),
    S=string::concat(Pn,"=",Vn),
    !.
callParsX(list(Par,Pars),list(Apar,Apars),S) :-
    Par=fstr(P,Dom,F,_),
    dbBase::nlSing(P,Pn),
    nl::conValX(Apar,Dom,F,Vn),
    callParsX(Pars,Apars,S2),
    S=string::concat(Pn,"=",Vn,", ",S2),
    !.
% -------------------------------------------------------------------------------------
% callSuffix
% -------------------------------------------------------------------------------------
callSuffix(S,T) :-
       _=string::search(S,"*"),
       T=" IS OUT OF CONTEXT",
       !
;
       _=string::search(S,"+"),
       T=" FAILURE - DEFAULT VALUE USED",
       !
;
       T="",
       !.
% -------------------------------------------------------------------------------------
% rsOwaCalls
% -------------------------------------------------------------------------------------
rsOwaCalls(_,_,_,_,[]) :-
    !.
rsOwaCalls(Rep,Indent,DefCalls,SubTrees,[vpiTree::tree(S,_,_,_)|Calls]) :-
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[str(PfRule),Val,_,atom(Dom),_,Pars,Apars,_,termDomain::str(Suffix),_]),
	sx::sxUnpackS(PfRule,"::",L2),
	L2=[_,RuleN],
	dbBase::kbID(RuleN,Rule,_,_),
    nl::conValX(Val,Dom,"",ValS),
	getWeight(Rule,DefCalls,SubTrees,Qual),
    callSuffix(Suffix,CallSuffix),
    callPars(Pars,Apars,Ps),
    sx::fixFormatS(Vals,2,0,Vals2),
    sx::sxL_Pad(Vals2," ",4,Vals3),
	S2=string::concatList([Vals3," (weight ",Qual,")=",RuleN,Ps,CallSuffix]),
	Rep:repWL(Indent,S2),
	rsOwaCalls(Rep,Indent,DefCalls,SubTrees,Calls),
	!
;
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[atom(Rule),Val,_,atom(Dom),_,Pars,Apars,_,termDomain::str(Suffix),_]),
    dbBase::nlSing(Rule,RuleN),
    nl::conValX(Val,Dom,"",ValS),
    callSuffix(Suffix,CallSuffix),
    callPars(Pars,Apars,Ps),
    /*
    Rule is the current rule in owa-set
    DefCalls are the owa-call siblings as per ir def
    SubTrees are the owa sub-tree siblings
    */
	getWeight(Rule,DefCalls,SubTrees,Qual), 
    sx::fixFormatS(Vals,2,0,Vals2),
    sx::sxL_Pad(Vals2," ",4,Vals3),
	S2=string::concatList([Vals3," (weight ",Qual,")=",RuleN,Ps,CallSuffix]),
	Rep:repWL(Indent,S2),
	rsOwaCalls(Rep,Indent,DefCalls,SubTrees,Calls),
	!
;
    Rep:repWL(Indent,"** UNRECOGNIZED OWA CALL **"),
	rsOwaCalls(Rep,Indent,DefCalls,SubTrees,Calls),
    !.
% -------------------------------------------------------------------------------------
% getWeight
% -------------------------------------------------------------------------------------
getWeight(_,[],_,S) :- 
    S="** CANNOT FIND WEIGHT **",
    !.
getWeight(Rule,[cmp(Rule,[Q|_])|_],SubTrees,Qual) :-
    Q=fix(W),
    Ws=toString(W),
    sx::fixFormatS(Ws,3,0,Ws2),
    sx::sxL_Pad(Ws2," ",7,Qual),
    !
;
    Q=cmp(Wrule,[cmp(_,Apars)]),
    weightCall(Wrule,Apars,SubTrees,Qual),
    !.
getWeight(Rule,[_|T],SubTrees,Qual) :-
    getWeight(Rule,T,SubTrees,Qual),
    !.
% -------------------------------------------------------------------------------------
% reportRule
% -------------------------------------------------------------------------------------
class predicates
repRules : (rep,lst).
clauses

reportRules() :-
    Kb=dbBase::dbName(),
    File=string::concat(Kb,"_InexactRules"),
    Rep=rep::newReport(File,100),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	p3Rep::header(Rep),
    L=pop::getCRsNL(),
    sx::sxlSort(L,Ls),
    List=lst::new(),
    List:addList(Ls),
    repRules(Rep,List),
	Rep:close(),
	!.
	
repRules(Rep,L) :-
       R=L:get_nd(),
            dbBase::kbID(R,K,_,_),
            ruleText(Rep,K),
            Rep:repWL(0,""),
            fail
;
    !.
% -------------------------------------------------------------------------------------
% reportRule
% -------------------------------------------------------------------------------------
reportRule(Rule) :-
    dbBase::nlTry(Rule,NL),
    Rep=rep::newReport(2,NL,100),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	ruleText(Rep,Rule),
	Rep:close(),
	!.
% -------------------------------------------------------------------------------------
% ruleText
% -------------------------------------------------------------------------------------
ruleText(Rep,Rule) :-
	dbBase::getIrDefF(Rule,Body,Loc),
	dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
	xmlP3::xmlTreeFromCR("NL",Realm,Body,Tree),
	xmlTree::removeAmpersands(Tree,Tree2),
	Tree2=cmp("inexactRule",[_,T]),
	irType(Rep,T,Rule),
    !.
% -------------------------------------------------------------------------------------
% irType
% -------------------------------------------------------------------------------------
irType(Rep,T,Rule) :-
    T=cmp(Log,[_,Head|Calls]),
    set::setGet(["andRule","orRule","notRule"],Ix,Log),
    set::setGet(["AND'ed RULES:","OR'ed RULES:","INVERTED RULE:"],Ix,Oper),
    set::setGet(["AND-RULE","OR-RULE","NOT-RULE"],Ix,Type),
    irHeading(Rep,Type,Head,Rule),
    irList0(Rep,Oper,Calls,_),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("bayRule",[_,Head|Calls]),
    irHeading(Rep,"BAYESIAN RULE",Head,Rule),
    irList2(Rep,"BAYESIAN COMBINATION OF RULES:",Calls,_),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("combinationRule",[_,Head|Calls]),
    irHeading(Rep,"COMBINATION RULE",Head,Rule),
    irList0(Rep,"COMBINATION OF RULES:",Calls,_),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("fuzzySet",[_,Set,Operation|T2]),
    irBasic(Set,S1),
    irBasic(Operation,S2),
    Rep:repWL(0,S1," [FUZZY SET]"),
	annotation(Rep,"p3_iIRdec",Rule,5),
    irParameterList(Rep,T2,T3),
    S3=string::concat(S2," OF SET(S):"),
    irList0(Rep,S3,T3,_),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("fuzzyOWA_Set",[_,Set|T2]),
    irBasic(Set,S1),
    Rep:repWL(0,S1," [FUZZY SET]"),
	annotation(Rep,"p3_iIRdec",Rule,5),
    irParameterList(Rep,T2,T3),
    irList1(Rep,"ORDERED WEIGHTED AVERAGING OF SET(S):",T3,_),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("queryRule",[_,Heading,Default,Query|Map]),
    irHeading(Rep,"QUERY-RULE",Heading,Rule),
    irBasicValue(Default,S2),
    Rep:repWL(5,"DEFAULT VALUE: ",S2),
    irQuery(Rep,Query),
    irMap(Rep,Map),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("questionRule",[_,Heading,Default,Domain,Question|Map]),
    irHeading(Rep,"QUESTION-RULE",Heading,Rule),
    irBasic(Domain,S3),
    irBasic(Question,S4),
    S5=string::concat(S4," --> ",S3),
    Rep:textBlock(5,"QUESTION: ",S5),
    irBasicValue(Default,S2),
    Rep:repWL(5,"DEFAULT VALUE: ",S2),
    irMap(Rep,Map),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("switchRule",[_,Heading,Default,CritPar|Calls]),
    irHeading(Rep,"CALL-SWITCH RULE",Heading,Rule),
    irBasicValue(Default,DefS),
    Rep:repWL(5,"DEFAULT VALUE: ",DefS),
    irBasic(CritPar,ParS),
    Rep:repWL(5,"CRITERION PARAMETER: ",ParS),
    Rep:repWL(5,"SWITCH-CALLS:"),
    irContextCalls(Rep,Calls),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("switchRule",[_,Heading,Default,CritCall|Calls]),
    irHeading(Rep,"CALL-SWITCH RULE",Heading,Rule),
    irBasicValue(Default,DefS),
    Rep:repWL(5,"DEFAULT VALUE: ",DefS),
    irCall(Rep,5,"CRITERION CALL: ",CritCall),
    Rep:repWL(5,"SWITCH-CALLS:"),
    irContextCalls(Rep,Calls),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("parameterRule",[_,Heading,Call|ParCalls]),
    irHeading(Rep,"PARAMETER-RULE",Heading,Rule),
    irCall(Rep,5,"PARAMETERIZED CALL: ",Call),
    irList1b(Rep,"PARAMETER CALL(S):",ParCalls,_),
    irAssociation(Rep,Rule),
    !
;    
    T=cmp("mapRule",[_,Heading,Default,Call]),
    irHeading(Rep,"MAP-RULE",Heading,Rule),
    irBasicValue(Default,DefS),
    Rep:repWL(5,"DEFAULT VALUE: ",DefS),
    irCall(Rep,5,"MAPPED CALL: ",Call),
    irAssociation(Rep,Rule),
    !
;
    Mess=string::concat("Unrecognized inexact rule: ",Rule,"!"),
    dlgIF::dlgNote(3,Mess),
    !.
% -------------------------------------------------------------------------------------
% irHeading
% -------------------------------------------------------------------------------------
class predicates
irPrior : (rep,string,sterm).
clauses

irHeading(Rep,Type,cmp("heading",[_,Rule,Domain|T2]),RuleK) :-
    irBasic(Rule,S1),
    irBasic(Domain,RetDom),
    Rep:repWL(0,S1," [",Type," --> ",RetDom,"]"),
    irParameterList(Rep,T2,T3),
    T3=[Prior|Context],
 	annotation(Rep,"p3_iIRdec",RuleK,5),
 	irPrior(Rep,RetDom,Prior),
 	irContext(Rep,Context),
   !.
   
irPrior(Rep,RetDom,Prior) :-  
    dbBase::kbID(RetDom,Dom,_,_),
    set::member(Dom,["p3_dpr","p3_dcf"]),
    irBasic(Prior,S),
    Rep:repWL(5,"PRIOR VALUE: ",S),
    !
;
    !.
% -------------------------------------------------------------------------------------
% irAssociation
% -------------------------------------------------------------------------------------
class predicates
irAss : (rep,integer,string) determ.
clauses

irAssociation(Rep,Rule) :-
    C=count::new(),
    dbBase::chainSentences("x",LT,_),
        LT=cmp(_,[cmp("p3_eIrAs",[atom(Rule),AssVal,atom(Dom),str(Format),int(AssPar)])]),
        nl::conValX_F(AssVal,Dom,Format,Ass),
        AssParS=toString(AssPar),
        S=string::concatList([AssParS,": ",Ass]),
        C:inc(N),
        irAss(Rep,N,S),
        fail
;        
    !.

irAss(Rep,1,S) :- Rep:repWL(5,"ASSOCIATION(S): ",S),!.
irAss(Rep,_,S) :- Rep:repWL(21,S),!.
% -------------------------------------------------------------------------------------
% annotation
% -------------------------------------------------------------------------------------
annotation(Rep,Type,Name,Indent) :-
  	dbExtension::kbGetAnnotationBlank(Type,Name,Note),
	Note2=sx::replaceAll(Note,"\n"," "),
	Note3=sx::replaceAll(Note2,"  "," "),
	adjustAnno(Indent,Note3,Anno),
	Rep:repW(0,Anno),
	!.
% -------------------------------------------------------------------------------------
% adjustAnno
% -------------------------------------------------------------------------------------
adjustAnno(Indent,Si,S) :-
    Si="",
    S=Si,
    !
;
    sx::sxBlockNL(Indent,100,Si,S2),
    S=string::concat(S2,"\n"),
    !.
% -------------------------------------------------------------------------------------
% irQuery
% -------------------------------------------------------------------------------------
class predicates
repSample : (rep,string) determ.
irQuerySt : (rep,string).
clauses

irQuery(Rep,Q) :-
    irBasic(Q,Qy),
    irQuerySt(Rep,Qy),
    !.
    
irQuerySt(Rep,Q) :-
    irBase::wcrTransQuery(Q,Qk,Dom),
    dbBase::nlSing(Dom,DomNl),
    S=string::concat(Q," --> ",DomNl),
    Rep:textBlock(5,"QUERY: ",S),
    %
    dbBase::queryStF(Qk,StK),
    dbBase::nlSing(StK,St),
    Rep:repWL(12,"BASED ON SENTENCE TYPE: ",St),
    annotation(Rep,"p3_iST",StK,12),
    p3Rep::sentenceSample(StK,Sample),
    repSample(Rep,Sample),
    !
;
    !.
    
repSample(Rep,S) :-
    S="",
    !
;
    Rep:textBlock(12,"EXAMPLE: ",S),
    !.
% -------------------------------------------------------------------------------------
% irCall
% -------------------------------------------------------------------------------------
irCall(Rep,Indent,Prefix,Call) :-
    Call=cmp(_,[_,R|Apars]),
    irBasic(R,Rule),
    dbBase::kbID(Rule,K,_,_),
    dbBase::irType(K,_,SubType),
    dbBase::nlSing(SubType,TypeNL),
    irActualParameterList(APars,_,Ps),
    S=string::concat(Rule," (",TypeNL,") ",Ps),
    Rep:textBlock(Indent,Prefix,S),
    !.
% -------------------------------------------------------------------------------------
% irList0
% -------------------------------------------------------------------------------------
class predicates
irList0X : (rep,sterml,sterml) procedure (i,i,o).
clauses

irList0(Rep,Oper,T,R) :-
    Rep:repWL(5,Oper),
    irList0X(Rep,T,R),
    !.
    
irList0X(_,[],[]) :-
    !.
irList0X(Rep,[H|T],Rest) :-
    H=cmp(_,[_,Set|T2]),
    irBasic(Set,Rule),
    dbBase::kbID(Rule,K,_,_),
    dbBase::irType(K,_,SubType),
    dbBase::nlSing(SubType,TypeNL),
    irActualParameterList(T2,_,Ps),
    S=string::concat(Rule," (",TypeNL,") ",Ps),
    Rep:textBlock(5,"",S),
    irList0X(Rep,T,Rest),
    !
;
    Rest=[H|T],
    !.
% -------------------------------------------------------------------------------------
% irList1
% -------------------------------------------------------------------------------------
class predicates
irList1X : (rep,sterml,sterml) procedure (i,i,o).
clauses

% handles lists of rule reference with one numeric/call qualifier
%
irList1(Rep,Oper,T,R) :-
    Rep:repWL(5,Oper),
    irList1X(Rep,T,R),
    !.
    
irList1X(_,[],[]) :-
    !.
irList1X(Rep,[H|T],Rest) :-
    H=cmp(Type,[_,Parameter,Rule|ActualParameters]),
    set::setGet(["OWA","ruleCall"],Ix,Type),
    set::setGet(["Weight ","Parameter "],Ix,Prefix),
    irList1Parameter(Rep,Prefix,Parameter,Rule,ActualParameters),
    irList1X(Rep,T,Rest),
    !
;
    Rest=[H|T],
    !.
% -------------------------------------------------------------------------------------
% irList1b
% -------------------------------------------------------------------------------------
class predicates
irList1bX : (rep,sterml,sterml) determ (i,i,o).
clauses

% handles lists of rule reference with one numeric/call qualifier
%
irList1b(Rep,Oper,T,R) :-
    Rep:repWL(5,Oper),
    irList1bX(Rep,T,R),
    !.
    
irList1bX(_,[],[]) :-
    !.
irList1bX(Rep,[Par,Call|T],Rest) :-
    irBasic(Par,Var),
    Pf=string::concat(Var,"="),
    irCall(Rep,5,Pf,Call),
    irList1bX(Rep,T,Rest),
    !
;
    Rest=[Par,Call|T],
    !.
% -------------------------------------------------------------------------------------
% irList2
% -------------------------------------------------------------------------------------
class predicates
irList2X : (rep,sterml,sterml) procedure (i,i,o).
clauses

% handles lists of rule reference with one numeric/call qualifier
%
irList2(Rep,Oper,T,R) :-
    Rep:repWL(5,Oper),
    irList2X(Rep,T,R),
    !.
    
irList2X(_,[],[]) :-
    !.
irList2X(Rep,[H|T],Rest) :-
    H=cmp(_,[_,cmp(_,[_,Par1]),cmp(_,[_,Par2]),Rule]),
    irCall(Rep,5,"",Rule),
    irParameter(Rep,"INCREMENT FACTOR ",Par1),
    irParameter(Rep,"DECREMENT FACTOR ",Par2),
    irList2X(Rep,T,Rest),
    !
;
    Rest=[H|T],
    !.
% -------------------------------------------------------------------------------------
% irParameter
% -------------------------------------------------------------------------------------
irParameter(Rep,Prefix,Parameter) :-
    irBasic(Parameter,Val),
    sx::fixFormatS(Val,3,0,Val2),
    sx::sxL_Pad(Val2," ",7,Value),
    Rep:repWL(15,Prefix,Value),
    !
;
    Parameter=cmp("call",[_,CallRule|CallParameters]),
    irBasic(CallRule,CallRuleS),
    dbBase::kbID(CallRuleS,CallK,_,_),
    dbBase::irType(CallK,_,CallSubType),
    dbBase::nlSing(CallSubType,CallTypeNL),
    irActualParameterList(CallParameters,_,Ps1),
    Pf1=string::concat(Prefix,"by calling: "),
    S1=string::concat(CallRuleS," (",CallTypeNL,") ",Ps1),
    Rep:textBlock(15,Pf1,S1), 
    !.
% -------------------------------------------------------------------------------------
% irList1Parameter
% -------------------------------------------------------------------------------------
irList1Parameter(Rep,Prefix,Parameter,Rule,ActualParameters) :-
    irBasic(Parameter,Val),
    sx::fixFormatS(Val,3,0,Val2),
    sx::sxL_Pad(Val2," ",7,Value),
    irBasic(Rule,RuleS),
    dbBase::kbID(RuleS,K,_,_),
    dbBase::irType(K,_,SubType),
    dbBase::nlSing(SubType,TypeNL),
    irActualParameterList(ActualParameters,_,Ps),
    S=string::concat(RuleS," (",TypeNL,") ",Ps),
    Pf=string::concat(Prefix,Value," "),
    Rep:textBlock(5,Pf,S),
    !
;
    Parameter=cmp("call",[_,CallRule|CallParameters]),
    irBasic(CallRule,CallRuleS),
    dbBase::kbID(CallRuleS,CallK,_,_),
    dbBase::irType(CallK,_,CallSubType),
    dbBase::nlSing(CallSubType,CallTypeNL),
    irActualParameterList(CallParameters,_,Ps1),
    Pf1=string::concat(Prefix," by calling: "),
    S1=string::concat(CallRuleS," (",CallTypeNL,") ",Ps1),
    Rep:textBlock(5,Pf1,S1), 
    %
    irBasic(Rule,RuleS),
    dbBase::kbID(RuleS,K,_,_),
    dbBase::irType(K,_,SubType),
    dbBase::nlSing(SubType,TypeNL),
    irActualParameterList(ActualParameters,_,Ps2),
    S2=string::concat(RuleS," (",TypeNL,") ",Ps2),
    Ind=6+string::length(Prefix),
    Rep:textBlock(Ind,"",S2), 
    !.
% -------------------------------------------------------------------------------------
% irActualParameterList
% -------------------------------------------------------------------------------------
class predicates
irActualParameterListX : (sterml,sterml,string_list) determ (i,o,o).
clauses

irActualParameterList(T,R,S) :-
    irActualParameterListX(T,R,L),
    S2=sx::join(L,","),
    S=string::concat("[",S2,"]"),
    !.
    
irActualParameterListX([],[],[]) :-
    !.
irActualParameterListX([H,A|T],Rest,L) :-
    H=cmp("parameter",[_,str(P)]),
    dbBase::kbID(P,K,_,_),
    dbBase::smClassDef(K,"a",_,_,Domain,_,_),
    dbBase::nlSing(Domain,Dn),
    A=cmp("varVal",[_,str(V)]),
    S=string::concat(P," (",Dn,")=",V),
    irActualParameterListX(T,Rest,L2),
    L=[S|L2],
    !
;
    Rest=[H,A|T],
    L=[],
    !.
% -------------------------------------------------------------------------------------
% irParameterList
% -------------------------------------------------------------------------------------
class predicates
irParameterListX : (rep,sterml,sterml) procedure (i,i,o).
clauses

irParameterList(Rep,T,R) :-
    T=[cmp("parameter",[_,str(P)])|T2],
    dbBase::kbID(P,K,_,_),
    dbBase::smClassDef(K,"a",_,_,Domain,_,_),
    dbBase::nlSing(Domain,Dn),
    Rep:repWL(5,"PARAMETER(S): ",P," (",Dn,")"),
    irParameterListX(Rep,T2,R),
    !
;
    Rep:repWL(5,"NO PARAMETERS"),
    R=T,
    !.
    
irParameterListX(_,[],[]) :-
    !.
irParameterListX(Rep,[H|T],Rest) :-
    H=cmp("parameter",[_,str(P)]),
    dbBase::kbID(P,K,_,_),
    dbBase::smClassDef(K,"a",_,_,Domain,_,_),
    dbBase::nlSing(Domain,Dn),
    Rep:repWL(19,P," (",Dn,")"),
    irParameterListX(Rep,T,Rest),
    !
;
    Rest=[H|T],
    !.
% -------------------------------------------------------------------------------------
% irContext
% -------------------------------------------------------------------------------------
class predicates
irContextCalls : (rep,sterml) determ.
clauses

irContext(_,[]) :-
    !.
irContext(Rep,[Context]) :-
    Context=cmp(_,[_,Out|Calls]),
    irBasicValue(Out,OutS),
    Rep:repWL(5,"CONTEXT: (out-of-context value: ",OutS,")"),
    irContextCalls(Rep,Calls),
    !.
    
irContextCalls(_,[]) :-
    !.
irContextCalls(Rep,[C|Cs]) :-
    C=cmp("scCall",[_,cmp("expression",[_,Oper,Value,Call])]),
    irBasic(Oper,OperS),
    irBasic(Value,ValueS),
    Prefix=string::concat(OperS,ValueS,": "),
    irCall(Rep,14,Prefix,Call),
    irContextCalls(Rep,Cs),
    !
;    
    C=cmp("scCall",[_,Call]),
    Call=cmp("defaultCall",_),
    irCall(Rep,14,"default-call: ",Call),
    irContextCalls(Rep,Cs),
    !
;    
    C=cmp("scCall",[_,cmp("range",[_,Oper1,Value1,Oper2,Value2,Call])]),
    irBasic(Oper1,Oper1S),
    irBasic(Value1,Value1S),
    irBasic(Oper2,Oper2S),
    irBasic(Value2,Value2S),
    Prefix=string::concat(Oper1S,Value1S," & ",Oper2S,Value2S,": "),
    irCall(Rep,14,Prefix,Call),
    irContextCalls(Rep,Cs),
    !.
% -------------------------------------------------------------------------------------
% irMap
% -------------------------------------------------------------------------------------
class predicates
irMapX : (rep,integer,sterml) determ.
clauses

irMap(Rep,Map) :-
    Map=[cmp("map",[_,MapClass,F,T|Maps])],
    irBasic(MapClass,S3),
    irBasic(F,S1),
    irBasic(T,S2),
    Rep:repWL(5,S3," MAP: ",S1," --> ",S2),
    Indent=5+string::length(S3)+6,
    irMapX(Rep,Indent,Maps),
    !
;
    !.
    
irMapX(_,_,[]) :-
    !.
irMapX(Rep,Indent,[F,T|Maps]) :-
    irBasic(F,S1),
    irBasic(T,S2),
    Rep:repWL(Indent,S1," --> ",S2),
    irMapX(Rep,Indent,Maps),
    !.
% -------------------------------------------------------------------------------------
% irBasic
% -------------------------------------------------------------------------------------
irBasic(cmp(_,[_,str(S)]),S) :- !.
% -------------------------------------------------------------------------------------
% irBasicValue
% -------------------------------------------------------------------------------------
irBasicValue(cmp(_,[_,cmp(_,[_,str(S)])]),S) :- !.
end implement irReport
