/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement kbBase
open core,termDomain,globalDomains

constants
className = "com/rudstrand/kbBase/kbBase".
classVersion = "1.0".

class predicates
kbNL_Next : (chainDB,chainDB::bt_selector,chainDB::ref,string,string,string,string,chainDB::ref) determ (i,i,i,i,i,i,o,o).
kbNL_Next2 : (chainDB,chainDB::bt_selector,chainDB::ref,string,string,string,string,string,chainDB::ref) determ (i,i,i,i,i,i,i,o,o).
kbNL_Uneq : (string Klang,string Kname,string Klang2,string Kname2) determ.

clauses
classInfo(className, classVersion).
% -----------------------------------------------------------------------------------------------
% kbID_Delete
% -----------------------------------------------------------------------------------------------
kbID_Delete(NL) :-
	kbID_RefF(NL,_,_,_,Ref,_),
	dbaBase::dbaMap("v",KB,_),
	dbaBase::dbaDeleteRecord(KB,Ref),
	kbID_Delete(NL),
	!
;
	!.
% -------------------------------------------------------------------------------------
% deleteKB_Name                                                    
% -------------------------------------------------------------------------------------
deleteKB_Name(Id) :-
    dbBase::accessErrorChain("v"),
    !,fail
;    
    dbaBase::dbaMap("v",KB,_),
    KB:chain_terms("v",Rec,Ref),
    Rec=clause(_,_,_,_,_,cmp(_,[cmp(_,[_,str(Id),_,_,_]),_]),_),
    dbaBase::dbaDeleteRecord(KB,Ref),
    fail
;
    !.
% -------------------------------------------------------------------------------------
% kbSM_Ref                                                              
% -------------------------------------------------------------------------------------
kbSM_Ref(Rep,KB,Ref) :-
    kbSM_RefF(Rep,KB,Ref),
    !
;
    dlgIF::feil("kbSM_Ref",["Unknown sentence model object",Rep],[],nill),!,fail.
    
kbSM_RefF(Rep,KB,Ref) :-
	dbaBase::dbaMap("q",KB,_),
	KB:chain_terms("q",Rec,Ref),
        Rec=clause(RT,_,_,_,_,cmp(_,[cmp(_,[atom(Rep)|_]),_]),atom("true")),
        dbBase::recordType(RT,RTb,_,_),
        RTb<=2,
	!.

kbSM_ClassRef(Rep,Clas,KB,Ref) :-
    kbSM_ClassRefF(Rep,Clas,KB,Ref),
    !
;
    dlgIF::feil("kbSM_ClassRef",["Unknown sentence model object",Rep],[],nill),!,fail.
    
kbSM_ClassRefF(Rep,Clas,KB,Ref) :-
	dbaBase::dbaMap("q",KB,_),
	KB:chain_terms("q",Rec,Ref),
        Rec=clause(RT,_,_,_,_,cmp(_,[cmp(Clas,[atom(Rep)|_]),_]),atom("true")),
        dbBase::recordType(RT,RTb,_,_),
        RTb<=2,
	!.
% -------------------------------------------------------------------------------------
% kbDE_List                                             
% -------------------------------------------------------------------------------------
kbDE_List([],[]) :-
	!.
kbDE_List([fstr(D,_,_,_)|Defs],[D|Ds]) :-
	kbDE_List(Defs,Ds),
	!.
% ----------------------------------------------------------------------------------------------
% kbST_Rec
% ----------------------------------------------------------------------------------------------
% returns the ST of sentences and rules
%
kbST_Rec(KB,Ref,ST) :-
	KB:ref_term(Ref,T),
	T=clause(_,_,_,_,_,LT,_),
	dbBase::termSt(LT,ST),
	!.
% ----------------------------------------------------------------------------------------------
% kbST_CountX
% ----------------------------------------------------------------------------------------------
class predicates
kbST_CountY : (integer,count,count) determ.
clauses

kbST_CountX(KB,2,P,E1,E2,C1,C2) :-
	KB:chain_terms(P,Cl,_),
	Cl=clause(Typ,_,_,_,_,cmp(_,[cmp(E1,_),cmp(E2,_)|_]),atom("true")),
	kbST_CountY(Typ,C1,C2),
	fail
;
	!.
kbST_CountX(KB,1,P,E1,_,C1,C2) :-
	KB:chain_terms(P,Cl,_),
	Cl=clause(Typ,_,_,_,_,cmp(_,[cmp(E1,_)|_]),atom("true")),
	kbST_CountY(Typ,C1,C2),
	fail
;
	!.
	
kbST_CountY(1,C,_) :- C:inc(_),!.		
kbST_CountY(2,_,C) :- C:inc(_),!.		
kbST_CountY(10,C,_) :- C:inc(_),!.		
kbST_CountY(20,_,C) :- C:inc(_),!.		
% -------------------------------------------------------------------------------------
% kbSM                                                     
% -------------------------------------------------------------------------------------
class predicates
kbSM_Def : (string,sterm,sterm) determ (i,i,o).
clauses
% parameters:
% 1	in	KB-name of sentence model object
% 2	out	SM class (f/g/p/e/a/b)
% 3	out	segment
% 4	out	definition
%
kbSM(Name,Clas,Segment,Def,KB,Ref) :-
	kbSM_F(Name,Clas,Segment,Def,KB,Ref),
	!
;
	dlgIF::feil("kbSM",["Unknown sentence model object",Name],[],nill),!,fail.
	
kbSM_F(Name,Clas,Segment,Def,KB,Ref) :-
	dbaBase::dbaMap("q",KB,_),
	KB:chain_terms("q",Rec,Ref),
	Rec=clause(2,_,_,_,_,LT,_),
	LT=cmp("q",[cmp(Clas,[atom(KB_Name)|_]),str(Segment)]),
	Name=KB_Name,
	kbSM_Def(Clas,LT,Def),
	!.
	
kbSM_Def(Clas,LT,Def) :-
	Clas="a",
	LT=cmp("q",[cmp(Clas,[atom(K),atom(Dom),str(Frm),int(W)]),_]),
	Def=fstr(K,Dom,Frm,W),
	!
;
	Clas="e",
	LT=cmp("q",[cmp(Clas,[atom(K),Ds]),_]),
	kbSMs(Ds,_,_,DEs,_,_),
	Def=cmp("ET",[cmp("KB",[atom(K)]),
	              cmp("DEs",DEs)]),
	!	                    
;
	Clas="p",
	LT=cmp("q",[cmp(Clas,[atom(K),int(Card),Ds]),_]),
	kbSMs(Ds,_,_,DEs,_,_),
	Def=cmp("PT",[cmp("KB",[atom(K)]),
	              cmp("CARD",[int(Card)]),
	              cmp("DEs",DEs)]),
	!	                 
;
	Clas="f",
	LT=cmp("q",[cmp(Clas,[atom(K),atom(Dom),Ds]),_]),
	kbSMs(Ds,_,_,DEs,_,_),
	Def=cmp("FD",[cmp("REP",[atom(K)]),
	              cmp("DOM",[atom(Dom)]),
	              cmp("DEs",DEs)]),
	!	              
;
	Clas="g",
	LT=cmp("q",[cmp(Clas,[atom(K),atom(Dom),Ds]),_]),
	kbSMs(Ds,_,_,DEs,_,_),
	Def=cmp("FN",[cmp("REP",[atom(K)]),
	              cmp("DOM",[atom(Dom)]),
	              cmp("DEs",DEs)]),
	!.
	
kbSMs(nill,[],[],[],[],[]) :- 
	!.
kbSMs(list(atom(K),Ks),[Clas|Cs],[Segment|Ss],[Def|Ds],[KB|KBs],[Ref|RefS]) :-
	kbSM(K,Clas,Segment,Def,KB,Ref),
	kbSMs(Ks,Cs,Ss,Ds,KBs,RefS),
	!.
	
kbSMsL([],[],[],[],[],[]) :- 
	!.
kbSMsL([K|Ks],[Clas|Cs],[Segment|Ss],[Def|Ds],[KB|KBs],[Ref|RefS]) :-
	kbSM(K,Clas,Segment,Def,KB,Ref),
	kbSMsL(Ks,Cs,Ss,Ds,KBs,RefS),
	!.
% -------------------------------------------------------------------------------------
% kbClass; kbClassF                                                   
% -------------------------------------------------------------------------------------
% returns the syntactic class of an identifer
%
kbClass(Id,Clas) :-
	kbClassF(Id,Clas),
	!
;
	dlgIF::feil("kbClass",["Unknown KB name",Id],[],nill),!,fail.
	
kbClassF(Id,Clas) :-
	Lng=globalVars::get_s("lang"),
	dbaBase::dbaMap("v",KB,_),		
	Ix=dbaBase::getIx('C',"SYN1"),
	sx::concat([Lng,"|",Id],Indx),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Indx,Ref),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("v",[cmp("t",[atom(Lng),str(Id),atom(Clas),_,_]),_]),atom("true")),
	!.
% ----------------------------------------------------------------------------------------------
% kbDE_Def;DefF
% ----------------------------------------------------------------------------------------------
class predicates
kbDE_Get : (chainDB,string,chainDB::ref,sterm) determ (i,i,i,o).
clauses

kbDE_Def(Rep,Def) :-
	kbDE_DefF(Rep,Def,_),
	!
;
	dlgIF::feil("kbDE_Def",[Rep],[],nill),
	Def=nill,
	!.		
kbDE_DefF(Rep,Def,Ref) :-
    kbSM_ClassRefF(Rep,"a",KB,Ref),
	kbDE_Get(KB,Rep,Ref,Def),
	!.

kbDE_Get(KB,Rep,Ref,Def) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("a",[atom(Rep),atom(Domain),str(Frm),int(W)]),_]),atom("true")),
	Def=fstr(Rep,Domain,Frm,W),
	!.
% -------------------------------------------------------------------------------------
% kbEntityDef/F                                                              
% -------------------------------------------------------------------------------------
class predicates
kbED_Get : (chainDB,string,chainDB::ref,sterm,string_list) determ (i,i,i,o,o).
clauses

kbEntityDef(Rep,EntDef,DEs) :-
	kbEntityDefF(Rep,EntDef,DEs,_),
	!
;
	sx::concat(["Unknown entity type: '",Rep,"'"],M),
	dlgIF::feil("kbEntityDef",[M],[],nill),
	!,fail.		

kbEntityDefF(Rep,EntDef,DEs,Ref) :-
    kbSM_ClassRefF(Rep,"e",KB,Ref),
	kbED_Get(KB,Rep,Ref,EntDef,DEs),
	!.

kbED_Get(KB,Rep,Ref,EntDef,DEs) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("e",[atom(Rep),Attrs]),_]),atom("true")),
	termDomain::cla_listR(DEs,Attrs),
	dbBase::kbDE_Defs(Attrs,AttDefs),
	EntDef=cmp("ENT_TYPE",
	          [cmp("REP",[atom(Rep)]),
	           cmp("ATTRS",AttDefs)]),
	!.
% -------------------------------------------------------------------------------------
% kbID_Ref/Refs
% -------------------------------------------------------------------------------------
class predicates
kbID_RefX1 : (chainDB,sterm,string,string,string,string,string,chainDB::ref,string) determ (i,i,i,i,o,o,o,o,o).
kbID_RefX2 : (chainDB,sterm,string,string,chainDB::bt_selector,string,string,string,chainDB::ref,string) determ (i,i,i,i,i,o,o,o,o,o).
kbID_RefX3 : (chainDB,sterm,string,string,string,string,string,chainDB::ref,integer,string) determ (i,i,i,i,o,o,o,i,o,o).	
clauses

kbID_Ref(NL,KbId,Clas,Form,Ref,Typ) :-
	kbID_RefF(NL,KbId,Clas,Form,Ref,Typ),!
;
	dlgIF::feil("kbID_Ref",["unknown syntagm:",NL],[],nill),
	!,fail.	

kbID_RefF(NL,KbId,Clas,Form,Ref,Typ) :-
	Lng=globalVars::get_s("lang"),
	dbaBase::dbaMap("v",KB,_),
	pscan::pscNL(NL,Toks),
	pscan::str_tokens(Toks,SToks1),
	termDomain::str_slist(SToks1,SToks2),
	termDomain::cls_list(SToks2,Tokens),
	SToks2=[Key|_],
	kbID_RefX1(KB,Tokens,Lng,Key,KbId,Clas,Form,Ref,Typ),
	!.
	
kbID_RefX1(KB,Tokens,Lng,Key,KbId,Clas,Form,Ref,Typ) :-
	Ix=dbaBase::getIx('C',"TOKEN"),
	sx::concat([Lng,"|",Key],Indx),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Indx,Ref),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,Term,atom("true")),
	Term=cmp("v",[cmp(Typ,[atom(Lng),str(KbId),atom(Clas),atom(Form),Tokens]),_]),
	!
;
	Ix=dbaBase::getIx('C',"TOKEN"),
	sx::concat([Lng,"|",Key],Indx),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Indx,_),
	kbID_RefX2(KB,Tokens,Lng,Key,Ix,KbId,Clas,Form,Ref,Typ),
	!.
kbID_RefX2(KB,Tokens,Lng,Key,Ix,KbId,Clas,Form,Ref,Typ) :-
	dbaBase::database(DB,_,_,0),
	DB:key_next(Ix,Ref),
	kbID_RefX3(KB,Tokens,Lng,Key,KbId,Clas,Form,Ref,Out,Typ),
	!,
	Out=0,
	!
;
	kbID_RefX2(KB,Tokens,Lng,Key,Ix,KbId,Clas,Form,Ref,Typ),
	!.
	
kbID_RefX3(KB,Tokens,Lng,Key,KbId,Clas,Form,Ref,Out,Typ) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,Term,atom("true")),
	Term=cmp("v",[cmp(Typ,[atom(Lng),str(KbId),atom(Clas),atom(Form),Tokens]),_]),
	Out=0,
	!
;	
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,Term,atom("true")),
	Term=cmp("v",[cmp(_,[_,_,_,_,list(str(Key),_)]),_]),
	!,fail
;
	KbId="",
	Clas="",
	Form="",
	Typ="",
	Out=1,
	!.	

kbID_Refs([],[],[],[],[]) :- 
	!.
kbID_Refs([NL|NLs],[Id|Ids],[Clas|Clases],[Form|Forms],[Ref|Refs]) :-
	kbID_Ref(NL,Id,Clas,Form,Ref,_),
	kbID_Refs(NLs,Ids,Clases,Forms,Refs),
	!.
kbID_RefsF([],[],[],[],[]) :- 
	!.
kbID_RefsF([NL|NLs],[Id|Ids],[Clas|Clases],[Form|Forms],[Ref|Refs]) :-
	kbID_RefF(NL,Id,Clas,Form,Ref,_),
	kbID_RefsF(NLs,Ids,Clases,Forms,Refs),
	!.
% -----------------------------------------------------------------------------------------------
% smSegment
% -----------------------------------------------------------------------------------------------
smSegment(Clas,Id,Segment) :-
    set::member(Clas,["p","g","f"]),
    kbPredDefF(Id,_,_,_,_,Ref),
	dbaBase::dbaMap("q",KB,_),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp(_,[_,str(Segment)]),_),
	!
;	
    Clas="e",
    kbEntityDefF(Id,_,_,Ref),
	dbaBase::dbaMap("q",KB,_),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp(_,[_,str(Segment)]),_),
	!
;
    Clas="a",
    kbDE_DefF(Id,_,Ref),
	dbaBase::dbaMap("q",KB,_),
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp(_,[_,str(Segment)]),_),
	!
;    
    Segment="",
    !.
% -------------------------------------------------------------------------------------
% kbPredDef;DefF                                                             
% -------------------------------------------------------------------------------------
class predicates
kbPD_Get : (chainDB,string,chainDB::ref,string,integer,sterm,string_list) determ (i,i,i,i,o,o,o).
kbPD : (string,string,integer,string,sterml,sterm) determ (i,i,i,i,i,o).
clauses

% returns definition info for predicates, functions or procedures
% parameters:
% 1 	in		rep
% 2 	out		class (p/f/g/i/h)
% 3 	out		p:cardinality f/g/i:#parameters
% 4 	out		definition tree
% 5	out		DE list (KB) (DEs or parameters)
%
kbPredDef(Rep,Clas,Card,RelDef,DEs) :-
	kbPredDefF(Rep,Clas,Card,RelDef,DEs,_),
	!
;
	sx::concat(["Unknown predicate, function or procedure '",Rep,"'"],M),
	dlgIF::feil("kbPredDef",[M],[],nill),!,fail.				

kbPredDefF(Rep,Clas,Card,RelDef,DEs,Ref) :-
    Clas="p",
    kbSM_ClassRefF(Rep,Clas,KB,Ref),
	kbPD_Get(KB,Rep,Ref,Clas,Card,RelDef,DEs),
	!
;
    Clas="f",
    kbSM_ClassRefF(Rep,Clas,KB,Ref),
	kbPD_Get(KB,Rep,Ref,Clas,Card,RelDef,DEs),
	!
;
    Clas="g",
    kbSM_ClassRefF(Rep,Clas,KB,Ref),
	kbPD_Get(KB,Rep,Ref,Clas,Card,RelDef,DEs),
	!
;
    Clas="h",
    kbSM_ClassRefF(Rep,Clas,KB,Ref),
	kbPD_Get(KB,Rep,Ref,Clas,Card,RelDef,DEs),
	!
;
    Clas="i",
    kbSM_ClassRefF(Rep,Clas,KB,Ref),
	kbPD_Get(KB,Rep,Ref,Clas,Card,RelDef,DEs),
	!.

kbPD_Get(KB,Rep,Ref,"p",Card,RelDef,DEs) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("p",[atom(Rep),int(Card),nill]),_]),atom("true")),
	DEs=[],
	kbPD(Rep,"p",Card,"",[],RelDef),
	!
;
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("p",[atom(Rep),int(Card),RelPars]),_]),atom("true")),
	termDomain::cla_listR(DEs,RelPars),
	kbPD(Rep,"p",Card,"",[RelPars],RelDef),
	!.
kbPD_Get(KB,Rep,Ref,"f",Card,RelDef,DEs) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("f",[atom(Rep),atom(Dom),RelPars]),_]),atom("true")),
	termDomain::cla_listR(DEs,RelPars),
	set::size(DEs,Card),
	kbPD(Rep,"f",Card,Dom,[RelPars],RelDef),
	!.
kbPD_Get(KB,Rep,Ref,"g",Card,RelDef,DEs) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("g",[atom(Rep),atom(Dom),RelPars]),_]),atom("true")),
	termDomain::cla_listR(DEs,RelPars),
	set::size(DEs,Card),
	kbPD(Rep,"g",Card,Dom,[RelPars],RelDef),
	!.
kbPD_Get(KB,Rep,Ref,"h",Card,RelDef,DEs) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("h",[atom(Rep),atom(Dom),RelPars]),_]),atom("true")),
	termDomain::cla_listR(DEs,RelPars),
	set::size(DEs,Card),
	kbPD(Rep,"h",Card,Dom,[RelPars],RelDef),
	!.
kbPD_Get(KB,Rep,Ref,"i",Card,RelDef,DEs) :-
	KB:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("q",[cmp("i",[atom(Rep),atom(Dom),RelPars]),_]),atom("true")),
	termDomain::cla_listR(DEs,RelPars),
	set::size(DEs,Card),
	kbPD(Rep,"i",Card,Dom,[RelPars],RelDef),
	!.
		
kbPD(Rep,"p",Card,_,[],cmp("REL_TYPE",
	                  [cmp("REP",[atom(Rep)]),
	                   cmp("CARD",[int(Card)])])) :- !.
kbPD(Rep,"p",Card,_,[Attrs],cmp("REL_TYPE",
	                               [cmp("REP",[atom(Rep)]),
	                                cmp("CARD",[int(Card)]),
	                                cmp("ATTRS",AttDefs)])) :- 
	dbBase::kbDE_Defs(Attrs,AttDefs),
	!.
kbPD(Rep,Fnk,_,Dom,[DEs],cmp("REL_TYPE",
	                             [cmp("REP",[atom(Rep)]),
	                              cmp("DOM",[atom(Dom)]),
	                              cmp("FPAR",DE_Defs)])) :- 
	set::member(Fnk,["f","g","i","h"]),
	dbBase::kbDE_Defs(DEs,DE_Defs),
	!.
% ----------------------------------------------------------------------------------------------
% irDeclaration
% ----------------------------------------------------------------------------------------------
irDeclaration(Rule,Dom,DE_Defs,DEs,Ref) :- 
    irDeclarationF(Rule,Dom,DE_Defs,DEs,Ref),
    !
;
    Mess=string::concat("Unknown inexact rule (declaration): ",Rule),
    dlgIF::dlgNote(1,Mess),
    !,fail.
    
irDeclarationF(Rule,Dom,DE_Defs,DEs,Ref) :- 
    kbPredDefF(Rule,Clas,_,Def,DEs,Ref),
    Clas="h",
    Def=cmp("REL_TYPE",
	             [_,
	              cmp("DOM",[atom(Dom)]),
	              cmp("FPAR",DE_Defs)]),
    !.
% ----------------------------------------------------------------------------------------------
% funcDef
% ----------------------------------------------------------------------------------------------
funcDef(Rule,Clas,Dom,DE_Defs,DEs,Ref) :- 
    funcDefF(Rule,Clas,Dom,DE_Defs,DEs,Ref),
    !
;
    Mess=string::concat("Unknown function (declaration): ",Rule),
    dlgIF::dlgNote(1,Mess),
    !,fail.
    
funcDefF(Rule,Clas,Dom,DE_Defs,DEs,Ref) :- 
    kbPredDefF(Rule,Clas,_,Def,DEs,Ref),
    set::member(Clas,["f","g"]),
    Def=cmp("REL_TYPE",
	             [_,
	              cmp("DOM",[atom(Dom)]),
	              cmp("FPAR",DE_Defs)]),
    !.
% ----------------------------------------------------------------------------------------------
% kbNL_Ref2
% ----------------------------------------------------------------------------------------------
kbNL_RefF2(Rep,Class,Form,NL,Ref) :-
	Lng=globalVars::get_s("lang"),
	Chain="v",
	dbaBase::dbaMap(Chain,Kb,_),
	Ix=dbaBase::getIx('C',"SYN1"),
	sx::concat([Lng,"|",Rep],Indx),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Indx,Ref),
	Kb:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("v",[cmp("t",[atom(Ln),str(Rp),atom(Cl),atom(Fm),Tokens]),_]),atom("true")),
	Ln=Lng,
	Rp=Rep,
	Fm=Form,
	Cl=Class,
	termDomain::list_conc(Tokens,NL),
	!
;	
	Lng=globalVars::get_s("lang"),
	Chain="v",
	dbaBase::dbaMap(Chain,Kb,_),
	Ix=dbaBase::getIx('C',"SYN1"),
	dbaBase::database(DB,_,_,0),
	DB:key_next(Ix,Rf),
	kbNL_Next2(Kb,Ix,Rf,Lng,Rep,Class,Form,NL,Ref),
	!.

kbNL_Next2(Kb,Ix,Ref,Lng,Rep,Class,Form,NL,NewRef) :-			
	Kb:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("v",[cmp("t",[atom(L),str(R),_,_,_]),_]),atom("true")),
	kbNL_Uneq(Lng,Rep,L,R),
	!,fail
;
	Kb:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("v",[cmp("t",[atom(Ln),str(Rp),atom(Cl),atom(Fm),Tokens]),_]),atom("true")),
	Ln=Lng,
	Rp=Rep,
	Fm=Form,
	Cl=Class,
	termDomain::list_conc(Tokens,NL),
	NewRef=Ref,
	!
;	
	dbaBase::database(DB,_,_,0),
	DB:key_next(Ix,Ref2),
	kbNL_Next2(Kb,Ix,Ref2,Lng,Rep,Class,Form,NL,NewRef),
	!.
% -----------------------------------------------------------------------------------------------
%                                                 SQL & 3KB KB
% -----------------------------------------------------------------------------------------------
% -----------------------------------------------------------------------------------------------
% kbNL_Ref/F
% -----------------------------------------------------------------------------------------------
kbNL_Ref(Rep,Form,NL,Ref) :-
	kbNL_RefF(Rep,Form,NL,Ref),
	!
;
	Lng=globalVars::get_s("lang"),
	sx::concat(["No syntagm for '",Lng,"|",Rep,"|",Form,"'"],S),
	dlgIF::feil("kbNL_Ref",[S],[],nill),
	!,fail.
	
kbNL_RefF(Rep,Form,NL,Ref) :-
    not(dbBase::kbIsSql),
	Lng=globalVars::get_s("lang"),
	Chain="v",
	dbaBase::dbaMap(Chain,Kb,_),
	Ix=dbaBase::getIx('C',"SYN1"),
	sx::concat([Lng,"|",Rep],Indx),
	dbaBase::database(DB,_,_,0),
	DB:key_search(Ix,Indx,Ref),
	Kb:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("v",[cmp("t",[atom(Ln),str(Rp),_,atom(Fm),Tokens]),_]),atom("true")),
	Ln=Lng,
	Rp=Rep,
	Fm=Form,
	termDomain::list_conc(Tokens,NL),
	!
;	
    not(dbBase::kbIsSql),
	Lng=globalVars::get_s("lang"),
	Chain="v",
	dbaBase::dbaMap(Chain,Kb,_),
	Ix=dbaBase::getIx('C',"SYN1"),
	dbaBase::database(DB,_,_,0),
	DB:key_next(Ix,Rf),
	kbNL_Next(Kb,Ix,Rf,Lng,Rep,Form,NL,Ref),
	!.

kbNL_Next(Kb,Ix,Ref,Lng,Rep,Form,NL,NewRef) :-			
	Kb:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("v",[cmp("t",[atom(L),str(R),_,_,_]),_]),atom("true")),
	kbNL_Uneq(Lng,Rep,L,R),
	!,fail
;
	Kb:ref_term(Ref,Rec),
	Rec=clause(_,_,_,_,_,cmp("v",[cmp("t",[atom(Ln),str(Rp),_,atom(Fm),Tokens]),_]),atom("true")),
	Ln=Lng,
	Rp=Rep,
	Fm=Form,
	termDomain::list_conc(Tokens,NL),
	NewRef=Ref,
	!
;	
	dbaBase::database(DB,_,_,0),
	DB:key_next(Ix,Ref2),
	kbNL_Next(Kb,Ix,Ref2,Lng,Rep,Form,NL,NewRef),
	!.

kbNL_Uneq(Lng,Rep,L,R) :-
	Lng<>L,
	!
;
	Rep<>R,
	!.	
% ----------------------------------------------------------------------------------------------
% kbST_Exists3kb
% ----------------------------------------------------------------------------------------------
kbST_Exists3kb(KB,3,Card,P,E1,E2) :-
	Card=1,
	KB:chain_terms(P,Cl,_),
	Cl=clause(Typ,_,_,_,_,cmp(_,[cmp(E1,_)|_]),atom("true")),
	Typ>=1,
	Typ<=2,
	!
;	
	Card=2,
	KB:chain_terms(P,Cl,_),
	Cl=clause(Typ,_,_,_,_,cmp(_,[cmp(E1,_),cmp(E2,_)|_]),atom("true")),
	Typ>=1,
	Typ<=2,
	!
;
	!,fail.	
kbST_Exists3kb(KB,RecType,Card,P,E1,E2) :-
	Card=1,
	KB:chain_terms(P,Cl,_),
	Cl=clause(RecType,_,_,_,_,cmp(_,[cmp(E1,_)|_]),atom("true")),
	!
;	
	Card=2,
	KB:chain_terms(P,Cl,_),
	Cl=clause(RecType,_,_,_,_,cmp(_,[cmp(E1,_),cmp(E2,_)|_]),atom("true")),
	!.
end implement kbBase
