/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement layout
open core,termDomain,globalDomains

constants
className = "com/rudstrand/layout/layout".
classVersion = "1.0".

class predicates
newLayout : (string STnl) determ.
storedLayout : (string,integer_list,integer_list,string_list,integer,integer) determ.
newData : (string STk,vpiTableed::tableedit_DATA Data) determ (i,o).
entRows : (string E,vpiTableed::tableedit_DATA Data) determ (i,o).
predRows : (string E,vpiTableed::tableedit_DATA Data) determ (i,o).
deRows : (string En,sterml DEs,vpiTableed::tableedit_DATA Rows) determ (i,i,o).
storedData : (string,integer_list,integer_list,string_list,vpiTableed::tableedit_DATA) determ (i,i,i,i,o).
entStRows : (string,integer_list,integer_list,string_list,vpiTableed::tableedit_DATA,integer) determ (i,i,i,i,o,o).
predStRows : (string,integer_list,integer_list,string_list,vpiTableed::tableedit_DATA,integer) determ (i,i,i,i,o,o).
deStRows : (string,sterml,integer_list,integer_list,string_list,vpiTableed::tableedit_DATA) determ (i,i,i,i,i,o).

clauses
classInfo(className, classVersion).
% ---------------------------------------------------------------------------------------------------------------
% validate
% ---------------------------------------------------------------------------------------------------------------
class predicates
validateSTs : (string_list) determ.
validateLayouts : (lst,string_list).
check : (lst,string) determ.    
clauses

validate() :-
    STs=pop::getSTs([]),
    validateSTs(STs),
    List=lst::new(),
    validateLayouts(List,STs),
    !.
    
validateSTs([]) :-
    !.
validateSTs([ST|STs]) :-
    load(ST,_,_,_,_,_),
    fail
;
    validateSTs(STs),
    !.

validateLayouts(L,STs) :-
    dbBase::chainSentences("y",LT,Loc),
        LT=cmp("y",[cmp("p3_eTabLay",[str(STk)|_])]),
        not(set::member(STk,STs)),
        Mess=string::concat("Layout exists for possibly non-existing sentence type: ",STk),
        vpiCommonDialogs::ask("Do you want to delete layout?",Mess,["Yes","No"])=0,
        dbBase::recordDelete(Loc),
        fail
;
    dbBase::chainSentences("y",LT,Loc),
        LT=cmp("y",[cmp("p3_eTabLay",[str(STk)|_])]),
        check(L,STk),
        dbBase::recordDelete(Loc),
        fail
;        
    !.
            
check(L,ST) :-
    L:member(ST),
    Mess=string::concat("There are more than one layout for sentence type: ",ST),
    vpiCommonDialogs::ask("Do you want to delete duplicate layout?",Mess,["Yes","No"])=0,
	!
;
    L:member(ST),
    !,fail
;
    L:add(ST),
    !,fail.
% ---------------------------------------------------------------------------------------------------------------
% display
% ---------------------------------------------------------------------------------------------------------------
% opens layout-window for given ST (whether layout exists or not)
%
display(ST,SolNo,TabNo) :-
    dbBase::stNL_KB(ST,STk),
    load(STk,_,Sorts,Ws,Fs,_),
    storedLayout(STk,Sorts,Ws,Fs,SolNo,TabNo),
    !
;
    newLayout(ST),
    !.
% ---------------------------------------------------------------------------------------------------------------
% insert
% ---------------------------------------------------------------------------------------------------------------
insert(STk,PR) :-
    load(STk,_,_,_,_,Loc),
    !, % layout exists already - delete it
    dbBase::recordDelete(Loc),
    transpr::proPR_Term(PR,T),
    dbBase::recordInsert(2,"$SYS",0,0,T,atom("true")),
            dbBase::nlSing(STk,ST),
            Mess=string::concatList(["Table layout for '",ST,"' updated!"]),
            dlgIF::msgS(Mess),
    !
;    
    % new layout
    transpr::proPR_Term(PR,T),
    dbBase::recordInsert(2,"$SYS",0,0,T,atom("true")),
            dbBase::nlSing(STk,ST),
            Mess=string::concatList(["Table layout for '",ST,"' stored!"]),
            dlgIF::msgS(Mess),
    !.
% ---------------------------------------------------------------------------------------------------------------
% load
% ---------------------------------------------------------------------------------------------------------------
class predicates
validateDEs : (dbLoc,string,string_list,string_list) determ.
clauses

% returns layout for given ST  - fails if no layout exists or if layout is obsolete
%
load(STk,DEs,Sorts,Ws,Fs,Loc) :-
    dbBase::chainSentences("y",LT,Loc),
    LT=cmp("y",[cmp("p3_eTabLay",[str(STk),D,S,W,F])]),
    termDomain::cla_listR(DEs,D),
    termDomain::cli_listR(Sorts,S),
    termDomain::cli_listR(Ws,W),
    termDomain::cls_listR(Fs,F),
    dbBase::kbST_DEs(STk,Ds),
    !,
    validateDEs(Loc,STk,DEs,Ds),
    !.

validateDEs(_,_,[],[]) :-
    !.
validateDEs(Loc,STk,[DE|DEs],[D|Ds]) :-
    DE=D,
    !,
    validateDEs(Loc,STk,DEs,Ds),
    !
;
    dbBase::nlTry(STk,ST),
    Mess=string::concat("Invalid table layout for '",ST,"'. You are recommended to delete it!"),
    vpiCommonDialogs::resp_default=vpiCommonDialogs::ask("Please confirm!",Mess,["Yes","No"]),
    dbBase::recordDelete(Loc),
    !,fail.
validateDEs(Loc,STk,DEs,Ds) :-
    set::size(DEs,N1),
    set::size(Ds,N2),
    N1<>N2,
    dbBase::nlTry(STk,ST),
    Mess=string::concat("Invalid table layout for '",ST,"'. You are recommended to delete it!"),
    vpiCommonDialogs::resp_default=vpiCommonDialogs::ask("Please confirm!",Mess,["Yes","No"]),
    dbBase::recordDelete(Loc),
    !,fail.
% ---------------------------------------------------------------------------------------------------------------
% storedLayout
% ---------------------------------------------------------------------------------------------------------------
% creates a layout window for an existing layout
%
storedLayout(STk,Sorts,Ws,Fs,SolNo,TabNo) :-
    W=guix::getWinOfClas(1),
    ColTitles=["Argument","Data Element","Domain","Sorting","Width","Format"],
    ReadOnly=[1,1,1,0,0,0],
    Widths=[150,200,100,40,40,40],
    Aligns=[-1,-1,0,0,0,0],
    dbBase::nlSing(STk,ST),
    storedData(STk,Sorts,Ws,Fs,Data),
    set::size(Sorts,NRows),
    Title=string::concat("Layout of '",ST,"' tables"),
    Wt=ted::createTableWin(W,Title,NRows,ReadOnly,ColTitles,Widths,Aligns,Data,["Save","Delete"],-1,tedP3::ted_window_eh,1),
    % modify win_db entry to include ST
    guix::getWinAll(_,_,Wt,_,_,No),
    guix::setWin(202,ST,Wt,SolNo,TabNo,No),
    !.
% ---------------------------------------------------------------------------------------------------------------
% newLayout
% ---------------------------------------------------------------------------------------------------------------
% creates a layout window for a new layout
%
newLayout(ST) :-
    W=guix::getWinOfClas(1),
    ColTitles=["Argument","Data Element","Domain","Sorting","Width","Format"],
    ReadOnly=[1,1,1,0,0,0],
    Widths=[150,200,100,40,40,40],
    Aligns=[-1,-1,0,0,0,0],
    dbBase::stNL_KB(ST,STk),
    newData(STk,Data),
    dbBase::kbST_DEs(STk,DEs),
    set::size(DEs,NRows),
    Title=string::concat("Layout of '",ST,"' tables"),
    Wt=ted::createTableWin(W,Title,NRows,ReadOnly,ColTitles,Widths,Aligns,Data,["Save","Delete"],-1,tedP3::ted_window_eh,1),
    % modify win_db entry to include ST
    guix::getWinAll(_,_,Wt,_,_,No),
    guix::setWin(202,ST,Wt,0,0,No),
    !.
% ---------------------------------------------------------------------------------------------------------------
% storedData
% ---------------------------------------------------------------------------------------------------------------
storedData(STk,Sorts,Ws,Fs,Data) :-
    sx::sxUnpack(STk,'|',L),
    L=[P,Ea,Eb],
    entStRows(Ea,Sorts,Ws,Fs,DataEa,Na),
    set::front(Na,Sorts,_,Sorts2),
    set::front(Na,Ws,_,Ws2),
    set::front(Na,Fs,_,Fs2),
    entStRows(Eb,Sorts2,Ws2,Fs2,DataEb,Nb),
    set::front(Nb,Sorts2,_,Sorts3),
    set::front(Nb,Ws2,_,Ws3),
    set::front(Nb,Fs2,_,Fs3),
    predStRows(P,Sorts3,Ws3,Fs3,DataP,_),
    set::conc([DataEa,DataEb,DataP],Data),
    !
;    
    sx::sxUnpack(STk,'|',L),
    L=[P,Ea],
    entStRows(Ea,Sorts,Ws,Fs,DataEa,Na),
    set::front(Na,Sorts,_,Sorts2),
    set::front(Na,Ws,_,Ws2),
    set::front(Na,Fs,_,Fs2),
    predStRows(P,Sorts2,Ws2,Fs2,DataP,_),
    set::conc([DataEa,DataP],Data),
    !.
% ---------------------------------------------------------------------------------------------------------------
% entStRows
% ---------------------------------------------------------------------------------------------------------------
entStRows(E,Sorts,Ws,Fs,Data,N) :-
    dbBase::kbEntityDEs(E,_,DE_Defs),
 	dbBase::nlSing(E,En),
    deStRows(En,DE_Defs,Sorts,Ws,Fs,Data),
    set::size(DE_Defs,N),
    !.
% ---------------------------------------------------------------------------------------------------------------
% predRows
% ---------------------------------------------------------------------------------------------------------------
predStRows(P,Sorts,Ws,Fs,Data,N) :-
    dbBase::kbPredDEs(P,_,DE_Defs),
 	dbBase::nlSing(P,Pn),
    deStRows(Pn,DE_Defs,Sorts,Ws,Fs,Data),
    set::size(DE_Defs,N),
    !.
% ---------------------------------------------------------------------------------------------------------------
% deStRows
% ---------------------------------------------------------------------------------------------------------------
deStRows(_,[],_,_,_,[]) :-
    !.
deStRows(En,[DE|DEs],[Sort|Sorts],[W|Ws],[F|Fs],[RowData|RowsData]) :-
    DE=fstr(D,Dom,_,_),
    ted::makeCell_s(En,C3),
    dbBase::nlSing(D,Dn),
    ted::makeCell_s(Dn,C4),
    dbBase::nlSing(Dom,DomN),
    ted::makeCell_s(DomN,C5),
    C6=vpiTableed::ic(Sort,vpiDomains::color_Linen),
    ted::makeCell_i(W,C7),
    ted::makeCell_s(F,C8),
    RowData=[C3,C4,C5,C6,C7,C8],
    deStRows(En,DEs,Sorts,Ws,Fs,RowsData),
    !.
%
% ---------------------------------------------------------------------------------------------------------------
% newData
% ---------------------------------------------------------------------------------------------------------------
% creates the table-data for a new layout for given ST
%
newData(STk,Data) :-
    sx::sxUnpack(STk,'|',L),
    L=[P,Ea,Eb],
    entRows(Ea,DataEa),
    entRows(Eb,DataEb),
    predRows(P,DataP),
    set::conc([DataEa,DataEb,DataP],Data),
    !
;    
    sx::sxUnpack(STk,'|',L),
    L=[P,Ea],
    entRows(Ea,DataEa),
    predRows(P,DataP),
    set::conc([DataEa,DataP],Data),
    !.
% ---------------------------------------------------------------------------------------------------------------
% entRows
% ---------------------------------------------------------------------------------------------------------------
entRows(E,Data) :-
    dbBase::kbEntityDEs(E,_,DE_Defs),
 	dbBase::nlSing(E,En),
    deRows(En,DE_Defs,Data),
    !.
% ---------------------------------------------------------------------------------------------------------------
% predRows
% ---------------------------------------------------------------------------------------------------------------
predRows(P,Data) :-
    dbBase::kbPredDEs(P,_,DE_Defs),
 	dbBase::nlSing(P,Pn),
    deRows(Pn,DE_Defs,Data),
    !.
% ---------------------------------------------------------------------------------------------------------------
% deRows
% ---------------------------------------------------------------------------------------------------------------
class predicates
sumWidth : (sterml,integer) determ (i,o).    
clauses

deRows(_,[],[]) :-
    !.
deRows(En,[DE|DEs],[RowData|RowsData]) :-
    DE=fstr(D,Dom,F,W),
    ted::makeCell_s(En,C3),
    dbBase::nlSing(D,Dn),
    ted::makeCell_s(Dn,C4),
    dbBase::nlSing(Dom,DomN),
    ted::makeCell_s(DomN,C5),
    C6=vpiTableed::ic(0,vpiDomains::color_Linen),
    ted::makeCell_i(W,C7),
    ted::makeCell_s(F,C8),
    RowData=[C3,C4,C5,C6,C7,C8],
    deRows(En,DEs,RowsData),
    !.

sumWidth([],0) :-
    !.
sumWidth([fstr(_,_,_,W)|Ds],SumW) :-
    sumWidth(Ds,W2),
    SumW=W+W2,
    !.
end implement layout
