/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement nl
open core,termDomain

constants
className = "com/rudstrand/nl/nl".
classVersion = "1.0".

class predicates
conFormatD : (integer,string,string) determ (i,i,o).
conFormatF : (real,string,string) determ (i,i,o).
conFormatI : (integer,string,string) determ (i,i,o).
conArg : (sterm,sterm,string) determ (i,i,o).
conCheckTaut : (string,string,string,string) procedure (i,i,i,o).
conCondition : (sterm,sterm,string_list) determ (i,i,o).
conDE_Cond : (sterm,sterm,string_list) procedure (i,i,o).
conDERF_Var : (string,string) determ (i,o).
conDE_Vars : (sterml,sterm,string_list) determ (i,i,o).
conDERF : (sterm,string) determ (i,o).
conDERF_Det : (string,string) determ (i,o).
conEntCond : (sterml,sterml,string_list) determ (i,i,o).
conEntities : (sterm,string) determ (i,o).
conEntity : (integer,integer,sterm,string) determ (i,i,i,o).
conExpression : (sterm,string) determ (i,o).
conFuncBodyL : (string,string,sterm,string_list) determ (i,i,i,o).
conFuncCallPars : (sterml,sterml,string_list) determ (i,i,o).
conFuncCallRS : (string,sterm,string) determ (i,i,o).
conIndirect : (integer,integer,string,string,string) determ (i,i,i,i,o).
conSentenceCond : (sterm,sterm,string) determ (i,i,o).
conQueryCondition : (string,string) determ (i,o).
conSelConds : (sterml,sterm,sterml) determ (i,i,o).
conSolArg : (integer,integer,integer,integer,sterm,string) determ (i,i,i,o,i,o).
conSolDEs : (integer,integer,integer,integer,integer,sterml,string_list) determ (i,i,i,i,i,i,o).
conSolDQ : (integer,integer,integer,sterm,string) determ (i,i,i,i,o).
conSolEQ : (integer,integer,integer,sterm,string) determ (i,i,i,i,o).
conSolFQ : (integer,integer,integer,sterm,string) determ (i,i,i,i,o).
conSolSQ : (integer,integer,string) determ (i,i,o).
conValue : (sterm,string) determ (i,o).
conVarCond : (sterm,string) determ (i,o).
conWithTail : (string,string_list,string) procedure (i,i,o).
cprSolEQ : (integer,integer,sterm,string) determ (i,i,i,o).
nlTermString : (string,string,string) procedure (i,i,o).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% conDEs
% ----------------------------------------------------------------------------------------------
% Returns the NL translation of a complete list pair of attribute values and 
% attribute definitions.
%
% 1. par in	solution reference# (or 0 if not solution based NL construction)
% 2. par in	solution no (if query solution)
% 3. par in	values
% 4. par in	attribute definitions
%
class predicates
conDE_Value : (string,string,string,string) procedure (i,i,i,o).
conDEsTail : (integer,integer,sterml,sterml,string) determ (i,i,i,i,o).
clauses

conDEs(_,_,[],[],"") :- 
	!.
conDEs(SolRef,No,[Val],[fstr(Name,Dom,Frm,_)],AttVal) :-
	conVal(SolRef,No,Val,Dom,Frm,NLVal,_),
	conDE_Value(Dom,Name,NLVal,AttVal),
	!.
conDEs(SolRef,No,[Val|Vals],[fstr(Name,Dom,Frm,_)|Defs],NL) :-
	conVal(SolRef,No,Val,Dom,Frm,NLVal,_),
	conDE_Value(Dom,Name,NLVal,AttVal),
	conDEsTail(SolRef,No,Vals,Defs,NLs),	
	NL=string::concat(AttVal,NLs),
	!.
conDEs(SolRef,No,Vals,Defs,S) :-
	% error
	conDEsTail(SolRef,No,Vals,Defs,S),
	!.

conDEsTail(_,_,[],[],"") :- 
	!.
conDEsTail(SolRef,No,[Val],[fstr(Name,Dom,Frm,_)],NL) :-
	conVal(SolRef,No,Val,Dom,Frm,NLVal,_),
	conDE_Value(Dom,Name,NLVal,AttVal),
	dbBase::nlForm("p3_iAnd","p3_iNIL",SAnd),
	sx::concat([" ",SAnd," ",AttVal],NL),
	!.
conDEsTail(SolRef,No,[Val|Vals],[fstr(Name,Dom,Frm,_)|Defs],NL) :-
	conVal(SolRef,No,Val,Dom,Frm,NLVal,_),
	conDE_Value(Dom,Name,NLVal,AttVal),
	conDEsTail(SolRef,No,Vals,Defs,NLs),	
	sx::concat([", ",AttVal,NLs],NL),
	!.
conDEsTail(SolRef,No,Vals,Defs,_) :-
	dlgIF::feil("conDEsTail",["NL attribute construction failed"],[],nill),
	tt::tt_i(1,"SolRef",SolRef),
	tt::tt_i(1,"SolNo",No),
	tt::tt_terms(1,"Values",Vals),
	tt::tt_terms(1,"Definitions",Defs),
	!,fail.

% parameters:
% 1	in	domain
% 2	in	DE name (k)
% 3	in	NL value
% 4	out	DE value pair
%
conDE_Value(Dom,Name,"[]",NL) :-
	sx::sxTail(2,Dom,_,T),
	T="_l",
	dbBase::nlSing("p3_iNon",No),
	dbBase::nlSing(Name,NameNL),
	sx::concat([No," ",NameNL],NL),
	!.
conDE_Value(_,Name,Val,NL) :- 
	dbBase::nlSing(Name,NameNL),
	sx::concat([NameNL," ",Val],NL),
	!.
% -----------------------------------------------------------------------------------------------
% conVal
% -----------------------------------------------------------------------------------------------
% constructs NL from the given value and domain
% parameters:
% 1	in	solution reference# (or 0 if the NL isn't based on a solution)
% 2	in	solution no
% 3	in	variable
% 4	in	domain (from the entity type definition)
% 5	in	format
% 6	out	NL value
% 7	out	solution component no
%
conVal(0,_,STerm,Dom,Frm,S,0) :- 
	conValX_F(STerm,Dom,Frm,S),
	!.
conVal(SolRef,SolNo,var(Var),Dom,Frm,S,CompNo) :-
	sol::solVarNo(SolRef,Var,CompNo),
	sol::solCompNo(SolRef,SolNo,CompNo,T),
	conValX_F(T,Dom,Frm,S),
	!.
conVal(_,_,STerm,Dom,Frm,S,0) :- 
	conValX_F(STerm,Dom,Frm,S),
	!.
% -----------------------------------------------------------------------------------------------
% conValX/_F
% -----------------------------------------------------------------------------------------------
class predicates
conValSwitch : (sterm,sterm) determ (i,o).
conValX_Fm : (integer,string,sterml,sterml,string_list) determ (i,i,i,i,o).
clauses
% parameters:
% 1	in	value (term)
% 2	in	value domain
% 3	in	value format
% 4	out	value (NL)
% the count procedure
% @@incomplete??
%
conValX(atom(Proc),_,_,S) :-
	Proc="p3_rCnt",
	dbBase::nlForm(Proc,"p3_iNIL",S),
	!
;
	Proc="p3_rCnt",
	dlgIF::feil("conValX",["No NL for count-procedure (p3_rCnt)."],[],nill),!,fail.
conValX(V,D,F,NL) :-
	conValX_F(V,D,F,NL),
	!
;
	dlgIF::feil("conValX",[D,F],[],V),!,fail.

% value: variable
%
conValX_F(var(Var),_,_,Val) :-
	vars::varFuncParR(Func,DE,Var),
	Func="p3_iThis",
	dbBase::nlSing("p3_iThis",Ths),
	dbBase::nlSing(DE,DEn),
	sx::concat([Ths," ",DEn],Val),	
	!
;
	vars::varFuncParR(Func,DE,Var),
	dbBase::nlSing(Func,FuncN),
	dbBase::nlSing(DE,DEn),
	sx::concat([FuncN,":",DEn],Val),	
	!
;
	vars::varDE_Var(Var,AttName), 
	dbBase::nlForm(AttName,"p3_iSnu",Val),
	!
;
	Val=Var,
	!.	
% value: symbol
%
conValX_F(str(Si),"p3_dsy",_,S) :-
	dbBase::kbIF_to_Sym(Si,S),
	!.
conValX_F(str(Sa),"p3_dts",Form,S) :-
    nlTermString(Sa,Form,S),
    !.
% value: string literal
% @@@ revised 14.4.6: all literal within literal conversions removed
conValX_F(str(Sa),_,_,S) :-  
    sx::sxQuote(Sa,S),
    !.
/*
    globalVars::get_i("f_outq")=7,
	Sb=sx::replaceAll(Sa,"'","\""),
	sx::concat(["'",Sb,"'"],S),
	!
;
    globalVars::get_i("f_outq")=7,
	sx::concat(["'",Sa,"'"],S),
	!
;
	Sb=sx::replaceAll(Sa,"\"","'"),
	sx::concat(["\"",Sb,"\""],S),
	!
;
	sx::concat(["\"",Sa,"\""],S),
	!.
*/	
conValX_F(V,_,_,S) :- 
	V=fstr(_,Dom,Form,_),
	conValSwitch(V,Vs),
	conValX_F(Vs,Dom,Form,S),
	!.
% value: atom
%
conValX_F(atom(A),_,_,S) :-
	dbBase::nlTry(A,S),
	!. 
% value: integer
%
conValX_F(int(I),"p3_din",Form,S) :- 
	conFormatI(I,Form,S),
	!.
conValX_F(int(I),"p3_dnu",Form,S) :- 
	conFormatI(I,Form,S),
	!.
conValX_F(int(I),"p3_did",Form,S) :- 
	conFormatD(I,Form,S),
	!.
conValX_F(int(I),"p3_dli",_,S)  :- 
	S=toString(I),
	!.
conValX_F(int(I),"p3_dany",Form,S)  :- 
	conFormatI(I,Form,S),
	!.
conValX_F(fint(F,_),Dom,Form,S) :- 
	conValX_F(int(F),Dom,Form,S),
	!.
% value: fix
%
conValX_F(fix(F),"p3_dnu",Form,S) :- 
	conFormatF(F,Form,S),
	!.
conValX_F(fix(F),"p3_dli",_,S) :- 
	S=toString(F),
	!.
conValX_F(fix(F),"p3_dany",Form,S) :- 
	conFormatF(F,Form,S),
	!.
conValX_F(ffix(F,Form),Dom,_,S) :-
	Form<>"", 
	conValX_F(fix(F),Dom,Form,S),
	!.
conValX_F(ffix(F,_),Dom,Form,S) :- 
	conValX_F(fix(F),Dom,Form,S),
	!.
conValX_F(fix(F),"p3_din",Form,S) :- 
	conFormatF(F,Form,S),
	!.
% cf/pr/d
%	
conValX_F(fix(F),Dom,Form,S) :- 
	set::member(Dom,["p3_dcf","p3_dmg","p3_dpr"]),
	Form="",
	conValX_F(fix(F),"p3_dnu",".3",S),
	!
;
	set::member(Dom,["p3_dcf","p3_dmg","p3_dpr"]),
	conValX_F(fix(F),"p3_dnu",Form,S),
	!.
conValX_F(V,Dom,Form,S) :- 
	set::member(Dom,["p3_dcf","p3_dmg","p3_dpr"]),
	conValX_F(V,"p3_dany",Form,S),
	!.
% increment/decrement
%	
conValX_F(fix(F),Dom,Form,S) :- 
	set::member(Dom,["p3_dinc","p3_ddec"]),
	Form="",
	conValX_F(fix(F),"p3_dnu",".2",S),
	!
;
	set::member(Dom,["p3_dinc","p3_ddec"]),
	conValX_F(fix(F),"p3_dnu",Form,S),
	!.
conValX_F(V,Dom,Form,S) :- 
	set::member(Dom,["p3_dinc","p3_ddec"]),
	conValX_F(V,"p3_dany",Form,S),
	!.
% value: internal time
%
conValX_F(fint(Time,_),"p3_dit",F,S) :- 
	conValX_F(int(Time),"p3_dit",F,S),
	!.
conValX_F(int(Time),"p3_dit",_,S) :-
	H=Time div 360000,
	L2=Time-H*360000,
	M=L2 div 6000,
	L3=L2-M*6000,
	Sc=L3 div 100,
	F=L3-Sc*100,
	Sh=toString(H),
	sx::sxL_Pad(Sh,"0",2,Sh2),	 
	Sm=toString(M),
	sx::sxL_Pad(Sm,"0",2,Sm2),	 
	Ss=toString(Sc),
	sx::sxL_Pad(Ss,"0",2,Ss2),	 
	Sf=toString(F),
	sx::sxL_Pad(Sf,"0",2,Sf2),	 
	sx::sxlJoin([Sh2,Sm2,Ss2,Sf2],":",S),
	!.
% value: internal date
% p3_did is called as p3_din in expressions to avoid the translation of e.g. integer constants 
% to a date
%	
conValX_F(cmp(Opr,[Opn1]),"p3_did",Form,S) :-
	opr::opr_any(Opr,_),
	conValX_F(Opn1,"p3_din",Form,S1),
	sx::concat(["(",Opr," ",S1,")"],S),
	!.
conValX_F(cmp(Opr,[Opn1,Opn2]),"p3_did",Form,S) :-
	opr::opr_any(Opr,_),
	conValX_F(Opn1,"p3_din",Form,S1),
	conValX_F(Opn2,"p3_din",Form,S2),
	sx::concat(["(",S1," ",Opr," ",S2,")"],S),
	!.
% value: expression
%
conValX_F(cmp(Opr,[Opn1]),Dom,Form,S) :-
	opr::opr_any(Opr,_),
	conValX_F(Opn1,Dom,Form,S1),
	sx::concat(["(",Opr," ",S1,")"],S),
	!.
conValX_F(cmp(Opr,[Opn1,Opn2]),Dom,Form,S) :-
	opr::opr_any(Opr,_),
	conValX_F(Opn1,Dom,Form,S1),
	conValX_F(Opn2,Dom,Form,S2),
	sx::concat(["(",S1," ",Opr," ",S2,")"],S),
	!.
% value: list
%
conValX_F(nill,_,_,"[]") :- 
	!.  % list is empty
conValX_F(list(List,Tail),ListDom,Form,S) :-
	sx::sxTail(2,ListDom,Dom,T),
	T="_l",
	conValL(list(List,Tail),Dom,Form,Ss),
	dbBase::joinB_NL(Ss,S),
	!.
conValL(nill,_,_,[]) :- 
	!.
conValL(list(Val,List),Dom,Form,[S|Ss]) :- 
	conValX_F(Val,Dom,Form,S),
	conValL(List,Dom,Form,Ss),
	!.
% parameters:
% 1	in	1=with MDE names; else w/out
% 2	in	blank or "-"
%
conValX_Fm(_,_,[],[],[]) :-
	!.
conValX_Fm(0,D,[Comp|Comps],[fstr(_,Dom,Form,_)|DEs],[S2|S3]) :-
	conValX_F(Comp,Dom,Form,S2),
	conValX_Fm(0,D,Comps,DEs,S3),
	!.
conValX_Fm(1,D,[Comp|Comps],[fstr(DE,Dom,Form,_)|DEs],[S5|S3]) :-
	conValX_F(Comp,Dom,Form,S2),
	dbBase::nlSing(DE,S4),
	sx::concat([S4,D,S2],S5),
	conValX_Fm(1,D,Comps,DEs,S3),
	!.
conValX_Fm(F,D,Comps,DEs,_) :-
	dlgIF::feil("conValX_F",["Error in complex data element valuation",D],[F],cmp("",[cmp("",Comps),cmp("",DEs)])),!,fail.
conValSwitch(fstr(V,"p3_dif",_,_),atom(V)) :- 
	dbBase::nlSingF(V,_),
	!.
conValSwitch(fstr(V,Dom,_,_),int(I)) :- 
	set::member(Dom,["p3_din","p3_did","p3_dit"]),
	I=sx::toInt(V),
	!.
conValSwitch(fstr(V,Dom,_,_),fix(F)) :- 
	set::member(Dom,["p3_dnu","p3_dpr","p3_dcf","p3_dmg","p3_dinc","p3_ddec"]),
	F=sx::toReal(V),
	!.
conValSwitch(fstr(V,_,_,_),str(V)) :- 
	!.
% -----------------------------------------------------------------------------------------------
% nlTermString
% -----------------------------------------------------------------------------------------------
class predicates
nlTermStringX : (sterml,string_list,string_list) determ (i,i,o).
nlTermStringBasic : (sterm,string) determ (i,o).
nlTermStringBasicX : (sterml,string) determ (i,o).
nlTermStringList : (sterm,string) determ (i,o).
clauses

 nlTermString(TermString,Format,Nl) :-
    Format="B",
    Term=termDomain::fromTermStringIndirectF(TermString),
    nlTermStringBasic(Term,S),
    sx::sxQuote(S,Nl),
    !
;
    not(Format="B"),
    _=string::search(Format,";"),
    sx::sxUnpack(Format,';',L),
    Term=termDomain::fromTermStringIndirectF(TermString),
    termDomain::isBasicTerm(Term),
    Term=cmp(H,T),
    nlTermStringX(T,L,L2),
    sx::sxlJoin(L2,";",S2),
    S3=string::concat(H,"(",S2,")"),
    sx::sxQuote(S3,Nl),
    !
;
    not(Format="B"),
    Format<>"",
    Term=termDomain::fromTermStringIndirectF(TermString),
    termDomain::isBasicTerm(Term),
    Term=cmp(H,T),
    set::size(T,N),
    set::initialize(N,Format,Formats),
    nlTermStringX(T,Formats,L2),
    sx::sxlJoin(L2,";",S2),
    S3=string::concat(H,"(",S2,")"),
    sx::sxQuote(S3,Nl),
    !
;    
    sx::sxQuote(TermString,Nl),
    !.
    
nlTermStringBasic(cmp(H,T),Nl) :-
    nlTermStringBasicX(T,S),
    Nl=string::concat(H,"(",S,")"),
    !.
nlTermStringBasic(list(H,T),Nl) :-
   nlTermStringBasic(H,Hs),
   nlTermStringList(T,S),
   Nl=string::concat("[",Hs,";",S,"]"),
   !.

nlTermStringBasic(str(S),S) :- !.
nlTermStringBasic(atom(S),S) :- !.
nlTermStringBasic(var(S),S) :- !.
nlTermStringBasic(int(S),Si) :- Si=toString(S),!.
nlTermStringBasic(fix(S),Si) :- Si=toString(S),!.
nlTermStringBasic(fstr(S,_,_,_),S) :- !.
nlTermStringBasic(nill,"nill") :- !.
    
nlTermStringBasicX([],"") :-
   !.
nlTermStringBasicX([H],S) :-
   nlTermStringBasic(H,S),
   !.
nlTermStringBasicX([H|T],S) :-
   nlTermStringBasic(H,S1),
   nlTermStringBasicX(T,S2),
   S=string::concat(S1,";",S2),
   !.

nlTermStringList(nill,"") :-
   !.
nlTermStringList(list(H,nill),S) :-
   nlTermStringBasic(H,S),
   !.
nlTermStringList(list(H,T),S) :-
   nlTermStringBasic(H,S1),
   nlTermStringList(T,S2),
   S=string::concat(S1,";",S2),
   !.
   
nlTermStringX([],[],[]) :-
 !.
nlTermStringX([str(V)|T],[F|Fs],[S|Ss]) :- conValX(str(V),"p3_dli",F,S), nlTermStringX(T,Fs,Ss),!.
nlTermStringX([atom(V)|T],[F|Fs],[S|Ss]) :- conValX(atom(V),"p3_dif",F,S), nlTermStringX(T,Fs,Ss),!.
nlTermStringX([int(V)|T],[F|Fs],[S|Ss]) :- conValX(int(V),"p3_din",F,S), nlTermStringX(T,Fs,Ss),!.
nlTermStringX([fix(V)|T],[F|Fs],[S|Ss]) :- conValX(fix(V),"p3_dnu",F,S), nlTermStringX(T,Fs,Ss),!.   
% -----------------------------------------------------------------------------------------------
% subst_var
% -----------------------------------------------------------------------------------------------
% Note that subst_var with variables come in two varities for different 
% format of the 3rd parameter. Would it be possible to call subst-var with 
% a uniform format for this parameter?
% parameters:
% 1	in	left operand
% 2	in	right operand
% 3	in	complete list of entity's attribute variables			
% 4	out	nl rep of (1) 
% 5	out	nl rep of (2)
% 23.5.95
%
subst_var(str(S1),str(S2),_,S1,S2) :- !.
subst_var(str(S),int(I),_,S,S2)    :- S2=toString(I),   !.
subst_var(str(S),fix(F),_,S,S2)    :- S2=toString(F),  !.
subst_var(int(I),str(S),_,S2,S)    :- S2=toString(I),   !.
subst_var(fix(F),str(S),_,S2,S)    :- S2=toString(F),  !.
subst_var(int(I1),int(I2),_,S1,S2) :- S1=toString(I1),  S2=toString(I2),!.
subst_var(fix(F1),fix(F2),_,S1,S2) :- S1=toString(F1), S2=toString(F2),!.
subst_var(int(I),fix(F),_,S1,S2)   :- S1=toString(I),   S2=toString(F),!.
subst_var(fix(F),int(I),_,S1,S2)   :- S2=toString(I),   S1=toString(F),!.
subst_var(Val,var(Var),[cmp("VAR",[var(Var),fstr(Att,Dom,Form,_)])|_],S,AttRep) :- 
	dbBase::nlForm(Att,"p3_iSnu",AttRep),
	conValX_F(Val,Dom,Form,S),!.
subst_var(var(Var),Val,[cmp("VAR",[var(Var),fstr(Att,Dom,Form,_)])|_],AttRep,S) :-
	dbBase::nlForm(Att,"p3_iSnu",AttRep),
	conValX_F(Val,Dom,Form,S),
	!.
subst_var(Val,var(Var),[var(Var)|_],S,AttRep) :- 
	vars::varDE_Var(Var,AttName), 
	dbBase::nlForm(AttName,"p3_iSnu",AttRep),
	dbBase::smDef(AttName,Clas,_,_,Dom,Form,_),
	Clas="a",
	conValX_F(Val,Dom,Form,S),
	!.
subst_var(var(Var),Val,[var(Var)|_],AttRep,S) :-
	vars::varDE_Var(Var,AttName), 
	dbBase::nlForm(AttName,"p3_iSnu",AttRep),
	dbBase::smClassDef(AttName,"a",_,_,Dom,Form,_),
	conValX_F(Val,Dom,Form,S),
	!.
subst_var(var(Var),Val,[_|AttVars],S1,S2) :- 
	subst_var(var(Var),Val,AttVars,S1,S2),!.
subst_var(V1,V2,V3,_,_) :- 
	dlgIF::feil("subst_var",[],[],cmp("xxx",[cmp("V1",[V1]),cmp("V2",[V2]),cmp("V3",V3)])),!,fail.
% -----------------------------------------------------------------------------------------------
% conFormatD/F/I/T
% -----------------------------------------------------------------------------------------------
conFormatI(I,Form,S) :-
	F=I,
	FormM=string::concat(Form,".0"),
	conFormatF(F,FormM,S),!.
	
conFormatD(I,_,S) :- 
	globalVars::get_i("f_date")=0,
	S=toString(I),
	!
;
	globalVars::get_i("f_date")=2,	
	datetime::datI_to_DMY(I,D,M,Y),
	Dt=Y*10000+M*100+D,
	S=toString(Dt),
	!
;	
	datetime::datI_to_DMY(I,D,M,Y),
	datetime::datS(D,M,Y,S),
	!.
conFormatF(F,Form,S) :-
	sx::searchchar(Form,'%',_),
	FormM=sx::replaceAll(Form,"%",""),
	Fh=F*100.0,
	conFormatF(Fh,FormM,S1),
	S=string::concat(S1,"%"),!
;
	% comma w/decimals
	sx::searchchar(Form,',',_),
	sx::searchchar(Form,'.',P),
	sx::frontstr(P,Form,_,T),
	sx::frontstr(1,T,Ds,_),
    D=sx::toInt(Ds),
	sx::fixFormatR(F,D,1,S),!
;	
	% no comma w/decimals
	sx::searchchar(Form,'.',P),
	sx::frontstr(P,Form,_,T),
	sx::frontstr(1,T,Ds,_),
    D=sx::toInt(Ds),
	sx::fixFormatR(F,D,0,S),!
;
	% comma no decimals
	sx::searchchar(Form,',',_),
	sx::fixFormatR(F,-1,1,S),!
;	
	% no comma no decimal (default format)
	sx::fixFormatR(F,-1,0,S),!.
% ----------------------------------------------------------------------------------------------
% conSentence
% ----------------------------------------------------------------------------------------------
class predicates
conParVars : (string,string_list,string_list) determ (i,i,o).	
reformat : (sterml,sterm) procedure (i,o).
rel_attrs : (integer,integer,sterml,sterml,string) determ (i,i,i,i,o).
clauses
	
conSentence(T,S) :-
	% unary 
	T=cmp(H,_),
	H<>":-",
	globalVars::set_i("pre_cnt",0),
	T=cmp(Rel,[Arg|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(1)])|AttDefs]),
	!,
	conEntity(0,0,Arg,S1),
	dbBase::nlForm(Rel,"p3_iSnu",S2),
	rel_attrs(0,0,AttDefs,Attrs,S3),
	sx::concat([S1," ",S2," ",S3],S),
	!
;
	% unary
	T=cmp(":-",[F,atom("true")]),
	globalVars::set_i("pre_cnt",0),
	F=cmp(Rel,[Arg|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(1)])|AttDefs]),
	!,
	conEntity(0,0,Arg,S1),
	dbBase::nlForm(Rel,"p3_iSnu",S2),
	rel_attrs(0,0,AttDefs,Attrs,S3),
	sx::concat([S1," ",S2," ",S3],S),
	!
;	
	% binary
	T=cmp(H,_),
	H<>":-",
	globalVars::set_i("pre_cnt",0),
	T=cmp(Rel,[Arg1,Arg2|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(2)])|AttDefs]),
	!,
	conEntity(0,0,Arg1,S1),
	dbBase::nlForm(Rel,"p3_iSnu",S2),
	conEntity(0,0,Arg2,S3),
	rel_attrs(0,0,AttDefs,Attrs,S4),
	sx::concat([S1," ",S2," ",S3," ",S4],S),
	!
;	
	% binary
	T=cmp(":-",[F,atom("true")]),
	globalVars::set_i("pre_cnt",0),
	F=cmp(Rel,[Arg1,Arg2|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(2)])|AttDefs]),
	!,
	conEntity(0,0,Arg1,S1),
	dbBase::nlForm(Rel,"p3_iSnu",S2),
	conEntity(0,0,Arg2,S3),
	rel_attrs(0,0,AttDefs,Attrs,S4),
	sx::concat([S1," ",S2," ",S3," ",S4],S),
	!.

rel_attrs(_,_,[],_,"") :- 
    !.
rel_attrs(_,_,[cmp("ATTRS",[])],_,"") :-
    !.
rel_attrs(SolRef,SolNo,[cmp("ATTRS",AttDefs)],AttVals,S) :-
	conDEs(SolRef,SolNo,AttVals,AttDefs,S1),
	dbBase::nlForm("p3_iWith","p3_iNIL",S2),
	sx::concat([" ",S2," ",S1],S),!.
	
rel_attrs(SolRef,SolNo,L1,L2,_) :- 
	set::conc([L1,L2],L),
	dlgIF::feil("rel_attrs",[],[SolRef,SolNo],cmp("error",L)),!,fail.	
	
% transforms between pure list-form to cmp(",",[Head,Tail]) form
% Date		09May95
% Status	Ok
%
reformat([],nill) :- !.
reformat([H],H) :- !.
reformat([H|T],cmp(",",[H,Tf])) :- reformat(T,Tf),!.
	
conParVars(_,[],[]) :-
	!.
conParVars(Prefix,[DE|DEs],[ParVar|ParVars]) :-
	ParVar=string::concat(Prefix,DE),
	conParVars(Prefix,DEs,ParVars),!.
% ----------------------------------------------------------------------------------------------
% conFunction/L
% ----------------------------------------------------------------------------------------------
conFunctionL(T,L) :-
	% function definition
	T=cmp(":-",[cmp(Func,[var(FVal)|_]),Body]),
	dbBase::smDefF(Func,Clas,_,_,Domain,_,_),
	set::member(Clas,["f","g"]),
	conFuncBodyL(Domain,FVal,Body,S),
	dbBase::nlSing(Func,FName),
	dbBase::nlSing("p3_iIdefb",DefBy),
	sx::concat([FName," ",DefBy," "],Sh),
	L=[Sh|S],
	!
;
	L=["** Pro/3 is not able to provide an NL translation of the function **"],
	!.
	
conFunction(T,S) :-
	conFunctionL(T,L),
	sx::sxlJoin(L,"",S),
	!.
% ----------------------------------------------------------------------------------------------
% conQuery
% ----------------------------------------------------------------------------------------------
conQuery(T,S) :-
	T=cmp("p3_iDdrf",[_,str(DERF),str(Q)]),
	conQueryCondition(Q,S3),
	dbBase::nlSing("p3_iFst",First),
	dbBase::nlSing("p3_iVal",ValueOf),
	conDERF_Var(DERF,DEn),
	dbBase::nlSing("p3_iWhr",Where),
	sx::concat([First," ",ValueOf," ",DEn," ",Where," ",S3],St),
	S=string::trim(St),
	!
;
	T=cmp("p3_iNdrf",[_,str(DERF),str(Q)]),
	conQueryCondition(Q,S3),
	dbBase::nlSing("p3_iVal",ValueOf),
	conDERF_Var(DERF,DEn),
	dbBase::nlSing("p3_iWhr",Where),
	sx::concat([ValueOf," ",DEn," ",Where," ",S3],St),
	S=string::trim(St),
	!
;
	T=cmp("p3_iSSS",[_,str(Q)]),
	conQueryCondition(Q,S3),
	dbBase::nlSing("p3_iSSS",NumberOf),
	sx::concat([NumberOf," ",S3],St),
	S=string::trim(St),
	!
;
	T=cmp("p3_iExis",[str(Q)]),
	conQueryCondition(Q,S3),
	dbBase::nlSing("p3_iIsTr",IsItTrue),
	sx::concat([IsItTrue," ",S3],St),
	S=string::trim(St),
	!
;
	T=cmp(F,[atom("p3_iDtFn")|Pars]),
	conQuery(cmp(F,Pars),St),
	S=string::trim(St),
	!
;
	T=cmp(F,[_|Pars]),
	dbBase::smDefTreeF(F,_,Def),
	Def=cmp(_,[_,_,cmp("FPAR",FormalPars)]),
	dbBase::nlSing(F,S1),
	conFuncCallPars(FormalPars,Pars,S2),
	dbBase::joinNL(S2,S3),
	sx::concat([S1," [",S3,"]"],St),
	S=string::trim(St),
	!
;
	S="** Pro/3 is not able to provide an NL translation of the query **",
	!.
% ----------------------------------------------------------------------------------------------
% conRecord
% ----------------------------------------------------------------------------------------------
conRecord(RecType,Term,S) :-
	dbBase::recordType(RecType,Z,_,_),
	Z<3,
	conSentence(Term,S2),
	S=string::concat(S2,"!"),
	!
;	
	dbBase::recordType(RecType,Z,_,_),
	Z=4,
	conFunction(Term,S2),
	S=string::concat(S2,"!"),
	!
;	
	dbBase::recordType(RecType,Z,_,_),
	set::member(Z,[3,5]),
	S="** Pro/3 is not able to provide an NL translation **",
	%conRule(Term,S),
	!
;	
	Z=math::abs(RecType),
	Z=6,
	S="** Pro/3 is not able to provide an NL translation **",
	!
;	
	dlgIF::feil("conRecord",[],[RecType],Term),
	!,fail.
% ----------------------------------------------------------------------------------------------
% conRule
% ----------------------------------------------------------------------------------------------
conRule(_,S) :-
	S="** NO SENTENCE RULE TRANSLATION PROVIDED **",
	!.
% ----------------------------------------------------------------------------------------------
% conSol
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	fq=function call query; dq=derf query; sq=sss query xq=existence query eq=other query
% 2	in	solution reference# 
% 3	in	output text option (same as f_outq)
% 4	in	-2 query with variables - no solution
%    		-1 query without variables (yes/no)
%    		N if solution N is to be constructed
% 5	in	term to be translated to NL
% 6	out	NL
%
conSol("eq",SolRef,Opt,Case,Term,S) :- 
    math::abs(Opt)=9,
	cprSolEQ(SolRef,Case,Term,S),
	!
;
    math::abs(Opt)=9,
    Opt=math::abs(9),
	dlgIF::feil("conSol",["PR construction failed"],[],nill),
	tt::tt_term(1,"conSol",Term),
	tt::tt_is(1,"conSol",[SolRef,Case]),
	S="** Pro/3 is not able to provide a translation in the requested format **",
	!.
conSol("eq",SolRef,Option,Case,Term,S) :- 
	conSolEQ(SolRef,Option,Case,Term,S),
	!.
conSol("fq",SolRef,Option,Case,Term,S) :- 
	conSolFQ(SolRef,Option,Case,Term,S),
	!.
conSol("dq",SolRef,Option,Case,Term,S) :- 
	conSolDQ(SolRef,Option,Case,Term,S),
	!.
conSol("sq",SolRef,Option,_,_,S) :- 
	conSolSQ(SolRef,Option,S),
	!.
conSol("xq",SolRef,Option,_,Term,S) :-
    Option=2,
    sol::solSetSize(SolRef,N),
    set::setMax([0,N],M),
    P=M+1,
    conQuery(Term,St),
    set::setGet(["p3_iNo","p3_iYes"],P,Ans),
    dbBase::nlSing(Ans,AnsN),
    S=string::concat(AnsN," (",St,")"),
    !
;
    Option=1,
    sol::solSetSize(SolRef,N),
    set::setMax([0,N],M),
    P=M+1,
    conQuery(Term,_),
    set::setGet(["p3_iNo","p3_iYes"],P,Ans),
    dbBase::nlSing(Ans,S),
	!.
conSol(Type,SolRef,Option,Case,Term,S) :-
	dlgIF::feil("conSol",["NL construction failed"],[],nill),
	tt::tt_term(1,"conSol",Term),
	tt::tt_s(1,"conSol",Type),
	tt::tt_is(1,"conSol",[SolRef,Option,Case]),
	S="** Pro/3 is not able to provide a translation in the requested format **",
	!.
% ----------------------------------------------------------------------------------------------
% cprRecord
% ----------------------------------------------------------------------------------------------
cprRecord(LT,RT,PR) :-
	cprTerm(LT,PL),
	cprTerm(RT,RL),
	sx::concat([PL,":-",RL,"."],PR),
	!.
% ----------------------------------------------------------------------------------------------
% conArg
% ----------------------------------------------------------------------------------------------
% constructs NL for a predicate argument
%
conArg(var(A),_,A) :-	
	!.
conArg(cmp(E,DEs),Cond,S) :-
	conDE_Vars(DEs,Cond,S2),
	conWithTail("p3_iWith",S2,S3),
	dbBase::nlSing(E,S4),
	sx::concat([S4,S3],S),
	!.
% ----------------------------------------------------------------------------------------------
% conCheckTaut
% ----------------------------------------------------------------------------------------------
% constructs NL for opn-opr-opn expression unless the expression is a tautologi (e.g. X_1=X_1)
% always succeeds
%
conCheckTaut(Eq,S,S,"") :- 
    opr::oprEq(Eq),
    !.
conCheckTaut(Opr,S1,S2,NL) :- 
	dbBase::nlSing(Opr,So),
	sx::concat([S1," ",So," ",S2],NL),
	!.
% ----------------------------------------------------------------------------------------------
% conCondition
% ----------------------------------------------------------------------------------------------
% constructs NL for a compound condition
% parameters:
% 1	in	(sub-)condition
% 2	in	original total condition
% 3	out	NL (list of sub-conditions)
%
conCondition(cmp(",",[Cond1,Cond2]),Cond,[S2|S3]) :-
	conCondition(Cond1,Cond,S1),
	S1=[S2],
	conCondition(Cond2,Cond,S3),
	!.	
conCondition(cmp(",",[Cond1,Cond2]),Cond,S3) :-
	conCondition(Cond1,Cond,S1),
	S1=[], % ref. next clause
	conCondition(Cond2,Cond,S3),
	!.	
conCondition(Cond,_,[]) :- 
	Cond=cmp(_,[var(V)|_]),
	vars::varComps(V,_,_,_), % DE conds have been processed already under conPred
	!.
conCondition(Cond,OrgCond,[S]) :- 
	conVarCond(Cond,S),
	!			
;				
	conSentenceCond(Cond,OrgCond,S),
	!.	
% ----------------------------------------------------------------------------------------------
% conDE_Cond
% ----------------------------------------------------------------------------------------------
class predicates
conDE_Conds : (sterm,sterml,string_list) determ (i,i,o).
clauses

conDE_Cond(V,cmp(RangeType,[V,cmp(_,[str(SubType),atom(Dom),Low,High])]),[S]) :-	
	V=var(W),
	set::member(RangeType,["inside","outside"]),
	dbBase::nlSing(RangeType,S3),
	string::front(SubType,1,Ba,Bb),
	vars::varComps(W,_,_,DEw),
	dbBase::nlSing(DEw,S4),
	conValX(Low,Dom,"",S5),
	conValX(High,Dom,"",S6),
	sx::concat([S4," ",S3," ",Ba,S5,", ",S6,Bb],S),
	!.
conDE_Cond(V,cmp(Op,[V,R]),[S]) :-
	V=var(Var),
	opr::oprEq(Op),
	vars::varComps(Var,_,_,DE),
	dbBase::smClassDefF(DE,"a",_,_,Dom,Form,_),
	dbBase::nlSing(DE,S2),
	conValX(R,Dom,Form,S4),
	sx::concat([S2,"=",S4],S),
	!.
conDE_Cond(V,cmp(Op,[V,R]),[S]) :-
	V=var(Var),
	opr::opr_any(Op,_),
	vars::varComps(Var,_,_,DE),
	dbBase::smDefF(DE,Clas,_,_,Dom,Form,_),
	Clas="a",
	dbBase::nlSing(DE,S2),
	dbBase::nlSing(Op,S3),
	conValX(R,Dom,Form,S4),
	sx::sxlJoin([S2,S3,S4]," ",S),
	!.
conDE_Cond(V,list(H,T),S) :-
	conDE_Cond(V,H,S2),
	conDE_Cond(V,T,S3),
	set::conc([S2,S3],S),
	!.
conDE_Cond(V,cmp(_,T),S) :-
	conDE_Conds(V,T,S),
	!.
conDE_Cond(_,_,[]) :-
	!.
	
conDE_Conds(_,[],[]) :-
	!.	
conDE_Conds(V,[H|T],S) :-
	conDE_Cond(V,H,S2),
	conDE_Conds(V,T,S3),
	set::conc([S2,S3],S),
	!.	
% ----------------------------------------------------------------------------------------------
% conDE_Vars
% ----------------------------------------------------------------------------------------------
conDE_Vars([],_,[]) :- 
	!.	
conDE_Vars([var("_")|DE_Vars],Cond,T) :-
	conDE_Vars(DE_Vars,Cond,T),
	!.
conDE_Vars([V|DE_Vars],Cond,T) :-
	conDE_Cond(V,Cond,S2),
	conDE_Vars(DE_Vars,Cond,T2),
	set::conc([S2,T2],T),
	!.
% ----------------------------------------------------------------------------------------------
% conDERF
% ----------------------------------------------------------------------------------------------
conDERF(cmp(Derf,[Var,_,str(Drf)]),NL) :-
	set::member(Derf,["p3_iDdrf","p3_iNdrf"]),
	Var=var(B),
	conDERF_Var(B,S1),
	conQueryCondition(Drf,S2),
	dbBase::nlForm("=","p3_iNIL",S_is),
	dbBase::nlForm("p3_iVal","p3_iNIL",S_val),
	dbBase::nlForm("p3_iWhr","p3_iNIL",S_where),
	conDERF_Det(Derf,First),
	sx::concat([B," ",S_is," ",First,S_val," ",S1," ",S_where," ",S2],NL),
	!.
	
conDERF_Var(V,NL) :-
	vars::varComps(V,_,_,DE),
	dbBase::nlSing(DE,NL),
	!.
% ----------------------------------------------------------------------------------------------
% conDERF_Det
% ----------------------------------------------------------------------------------------------
conDERF_Det("p3_iDdrf",S) :- 
	dbBase::nlForm("p3_iFst","p3_iNIL",S1), 
	S=string::concat(S1," "),
	!.
conDERF_Det("p3_iNdrf","") :- 
	!.
conDERF_Det("p3_iDdrf",S) :- 
	dbBase::nlForm("p3_iFst","p3_iNIL",S1), 
	S=string::concat(S1," "),
	!.
conDERF_Det("p3_iNdrf","") :- 
	!.
% ----------------------------------------------------------------------------------------------
% conEntCond
% ----------------------------------------------------------------------------------------------
class predicates
conEntCondT : (sterml,sterml,string_list) determ (i,i,o).
clauses
% generates an entity condition NL text starting with ' with ...'
% parameters:
% 1	in	list of conditions which refers to entity's attribute variables 
% 2	in	complete list of entity's attribute variables			
% 3	out	entity condition NL text	
% Date		23May95
% Status	Ok
%
conEntCond([nill],_,[""]) :- !.
conEntCond([cmp(",",Conds)],VarAtts,S) :- 
	conEntCond(Conds,VarAtts,S),!.
conEntCond(Conds,Atts,[S]) :- 
	conEntCondT(Conds,Atts,Ss), 
	dbBase::joinNL(Ss,S1),
	dbBase::nlForm("p3_iWith","p3_iNIL",With),
	sx::concat([" ",With," ",S1],S),!.
conEntCondT([],_,[]) :- !.
conEntCondT([cmp(",",Conds)],VarAtts,S) :- 
	conEntCondT(Conds,VarAtts,S),!.
conEntCondT([cmp(Oper,[Opn1,Opn2])|Conds],VarAtts,[S|Ss]) :-
	subst_var(Opn1,Opn2,VarAtts,S1,S2),
	dbBase::nlSing(Oper,SOper),
	sx::concat([S1," ",SOper," ",S2],S),
	conEntCondT(Conds,VarAtts,Ss),!.
% -------------------------------------------------------------------------------------
% conEntities                                                           
% -------------------------------------------------------------------------------------
% constructs NL for entities with or without conditions
%
conEntities(T,S) :-
	% conditional
	T=cmp(",",[cmp(EntName,Vars)|Conds]),
	dbBase::smClassDefTree(EntName,"e",EntDef),
	EntDef=cmp(_,[_,cmp(_,Atts)]),
	vars::var_atts(Vars,Atts,VarAtts),
	dbBase::nlForm(EntName,"p3_iPlu",S1),
	conEntCond(Conds,VarAtts,S2),
	S2=[S3],
	S=string::concat(S1,S3),
	!
;	
	% unconditional
	T=cmp(ObjName,_),
	dbBase::smClassDefF(ObjName,"e",_,_,_,_,_), % purpose?
	dbBase::nlForm(ObjName,"p3_iPlu",S),!
;
	dlgIF::feil("conEntities",["Unrecognized entity format"],[],T),!,fail.	
% ----------------------------------------------------------------------------------------------
% conEntity
% ----------------------------------------------------------------------------------------------
% generates NL translation of an entity
%
class predicates
join_sub : (string,string,string) determ (i,i,o).
clauses
conEntity(SolRef,SolNo,cmp(Entity,Attrs),S) :-
	dbBase::nlSing(Entity,EntNL),
	dbBase::smClassDefTree(Entity,"e",EntDef),
	EntDef=cmp(_,[_,cmp(_,AttDefs)]),
	conDEs(SolRef,SolNo,Attrs,AttDefs,AttNL),
	join_sub(EntNL,AttNL,S),
	!.

join_sub(EntNL,"",EntNL) :-
    !.
join_sub(EntNL,AttNL,S) :-
	dbBase::nlForm("p3_iWith","p3_iNIL",With),
	sx::sxlJoin([EntNL,With,AttNL]," ",S),
	!.
% ----------------------------------------------------------------------------------------------
% conExpression
% ----------------------------------------------------------------------------------------------
conExpression(cmp(Opr,[Opn1,Opn2]),S) :-
	opr::opr_any(Opr,2),
	conValue(Opn1,S1),
	conValue(Opn2,S2),
	dbBase::nlSing(Opr,OprS),
	sx::concat(["(",S1," ",OprS," ",S2,")"],S),
	!.
conExpression(cmp(Opr,[Opn1]),S) :-
	opr::opr_any(Opr,1),
	conValue(Opn1,S1),
	dbBase::nlSing(Opr,OprS),
	sx::concat(["(",OprS," ",S1,")"],S),
	!.
% ----------------------------------------------------------------------------------------------
% conFuncBodyL
% ----------------------------------------------------------------------------------------------
class predicates
conFuncBodyL2 : (string,string,sterm,string_list) determ (i,i,i,o).
conFuncVar : (string,sterm,string) determ (i,i,o).
conFuncStatement : (string,string,sterm,string) determ (i,i,i,o).
conStExpr : (string,sterm,string) determ (i,i,o).
clauses

% constructs NL for a list of function statements
%
conFuncBodyL(Domain,FVal,T,L) :-
	conFuncBodyL2(Domain,FVal,T,L2),
	set::setRev(L2,L3),
	L3=[Last|L4],
	sx::sxTail(1,Last,F,_),
	Fm=string::concat(F,"!"),
	L5=[Fm|L4],
	set::setRev(L5,L),
	!.
	
conFuncBodyL2(Domain,FVal,cmp(",",[B1,B2]),L) :-
	conFuncBodyL2(Domain,FVal,B1,L1),
	conFuncBodyL2(Domain,FVal,B2,L2),
	set::conc([L1,L2],L),
	!.
conFuncBodyL2(Domain,FVal,B,L) :- 
	conFuncStatement(Domain,FVal,B,NL),
	S=string::concat(NL,";"),
	L=[S],
	!.
	
conFuncStatement(Domain,FVal,cmp("not",[cmp("not",[St])]),NL) :-
	conFuncStatement(Domain,FVal,St,S1),
	dbBase::nlForm("p3_iTrue","p3_iNIL",S_not),
	sx::concat([S_not,"(",S1,")"],NL),
	!.
conFuncStatement(Domain,FVal,cmp("not",[St]),NL) :-
	conFuncStatement(Domain,FVal,St,S1),
	dbBase::nlForm("p3_iFals","p3_iNIL",S_not),
	sx::concat([S_not,"(",S1,")"],NL),
	!.
conFuncStatement(_,_,cmp("p3_iExis",[str(PredCond)]),NL) :- 
	conQueryCondition(PredCond,NL),
	!.
conFuncStatement(_,FVal,cmp(Derf,[Var,str(DE_Var),str(Drf)]),NL) :-
	% DERF
	set::member(Derf,["p3_iDdrf","p3_iNdrf"]),
	conFuncVar(FVal,Var,V),
	vars::varComps(DE_Var,_,_,DE),
	dbBase::nlForm(DE,"p3_iSnu",S1),
	conQueryCondition(Drf,S2),
	dbBase::nlForm("=","p3_iNIL",S_is),
	dbBase::nlForm("p3_iVal","p3_iNIL",S_val),
	dbBase::nlForm("p3_iWhr","p3_iNIL",S_where),
	conDERF_Det(Derf,First),
	sx::concat([V," ",S_is," ",First,S_val," ",S1," ",S_where," ",S2],NL),
	!.
conFuncStatement(_,FVal,cmp("p3_iSSS",[Var,str(PredCond)]),NL) :-
	% SSS
	conFuncVar(FVal,Var,V),
	conQueryCondition(PredCond,S2),
	dbBase::nlForm("=","p3_iNIL",S_is),
	dbBase::nlForm("p3_iSSS","p3_iNIL",S_ino),
	sx::concat([V," ",S_is," ",S_ino," ",S2],NL),
	!.
conFuncStatement(_,FVal,T,NL) :- 
	conFuncCallRS(FVal,T,NL),
	!.
conFuncStatement(Domain,FVal,cmp(Opr,[var(FVal),Exp]),NL) :-
	opr::oprEq(Opr),
	dbBase::nlForm("p3_iFRet","p3_iNIL",S1),
	conStExpr(FVal,Exp,S2),
	dom::domShortLong(Dom,Domain),
	sx::concat([S1," =",Dom," ",S2],NL),
	!
;
	not(opr::oprEq(Opr)),
	Mess=string::concat("Invalid operator in function return-statement: ",Opr,"!"),
	dlgIF::dlgNote(3,Mess),!,fail.	
conFuncStatement(_,FVal,cmp(Opr,[Opn,Exp]),NL) :-
	opr::opr_cmpass(Opr),
	conFuncVar(FVal,Opn,S1),
	conStExpr(FVal,Exp,S2),
	conCheckTaut(Opr,S1,S2,NL),
	!.
	
/* 4.2.6 handled in conFuncStatement
conFuncVar(FVal,var(FVal),S) :- 
	dbBase::nlForm("p3_iFRet","p3_iNIL",S),
	!.
*/	
conFuncVar(FVal,var(V),Par) :-
	vars::varFuncParR(Func,DE,V), 
	Func="p3_iThis", % 30.5.5 not sure if p3_iThis will be returned here??
	dbBase::nlSing("p3_iThis",Ths),
	dbBase::nlSing(DE,DEn),
	sx::concat([Ths," ",DEn],Par),
	!
;
	vars::varFuncParR(Func,DE,V), 
	LFunc=sx::setFirstLower(FVal),
	Func=LFunc,
	dbBase::nlSing("p3_iThis",Ths),
	dbBase::nlSing(DE,DEn),
	sx::concat([Ths," ",DEn],Par),
	!
;
	vars::varFuncParR(Func,DE,V), 
	dbBase::nlSing(Func,FuncN),
	dbBase::nlSing(DE,DEn),
	sx::concat([FuncN,":",DEn],Par),
	!.
conFuncVar(_,var(S),S) :- 
	!.
	
conStExpr(FVal,cmp(Opr,[Opn]),NL) :-
   	opr::opr_any(Opr,1),
	dbBase::nlSing(Opr,S1),
	conStExpr(FVal,Opn,S2),
	sx::concat([S1," ",S2],NL),
	!.
conStExpr(FVal,cmp(Opr,[Opn1,Opn2]),NL) :-
   	opr::opr_any(Opr,2),
	dbBase::nlSing(Opr,S1),
	conStExpr(FVal,Opn1,S2),
	conStExpr(FVal,Opn2,S3),
	sx::concat(["(",S2," ",S1," ",S3,")"],NL),
	!.
conStExpr(FVal,C,NL) :- 
	conFuncCallRS(FVal,C,NL),
	!.
conStExpr(_,atom(A),NL)    :- 
	dbBase::nlSing(A,NL),
	!
;
	sx::concat(["unknown identifier '",A,"'"],Mess),
	dlgIF::feil("conStExpr",[Mess],[],nill),
	!,fail.	
conStExpr(_,str(V),Vq)     :- 
    sx::sxQuote(V,Vq),
	!.
conStExpr(_,int(I),V)     :- 
	V=toString(I),
	!.
conStExpr(_,fix(F),V)     :- 
	V=toString(F),
	!.
conStExpr(FVal,var(V),NL) :- 
	conFuncVar(FVal,var(V),NL),
	!.	
% ----------------------------------------------------------------------------------------------
% conCallParList
% ----------------------------------------------------------------------------------------------
class predicates
conCallPars : (sterml,string_list) determ (i,o).
clauses
% cmp("p3_iAps",[cmp("isLI",[fvar("dTicker","p3_dli"),atom("dTicker")])])
conCallParList(C,NL) :-
	C=cmp("p3_iAps",Pars),
	conCallPars(Pars,NLs),
	dbBase::joinNL(NLs,S),
	sx::concat(["[",S,"]"],NL),
	!
;
	dlgIF::feil("conCallParList",[],[],C),!,fail.
		
conCallPars([],[]) :-
	!.
conCallPars([Par|Pars],[NL|NLs]) :-
	Par=cmp(Opr,[fvar(DE,Dom),RS]),
	dbBase::nlSing(Opr,OprNL),
	dbBase::nlSing(DE,DEn),
	conValX(RS,Dom,"",Vs),
	sx::concat([DEn," ",OprNL," ",Vs],NL),
	conCallPars(Pars,NLs),
	!
;
	dlgIF::feil("conCallPars",[],[],Par),!,fail.
% ----------------------------------------------------------------------------------------------
% conFuncCallPars
% ----------------------------------------------------------------------------------------------
class predicates
conFuncCallListPars1 : (string,string,sterm,string_list) determ (i,i,i,o).
clauses
conFuncCallPars([],[],[]) :- 
	!.
conFuncCallPars([fstr(DE,Dom,Form,_)|FPars],[Par|Pars],[NL|NLs]) :- 
	% list-parameters represented as list(...)
	Par=list(P,Pt),
	sx::sxTail(2,Dom,D,T),
	T="_l",
	conFuncCallListPars1(D,Form,list(P,Pt),NL2),
	dbBase::joinNL(NL2,NL3),
	dbBase::nlSing(DE,S3),
	sx::concat([S3,"=",NL3],NL),
	conFuncCallPars(FPars,Pars,NLs),
	!
;	
	conValX(Par,Dom,Form,S2),
	dbBase::nlSing(DE,S3),
	sx::concat([S3,"=",S2],NL),
	conFuncCallPars(FPars,Pars,NLs),
	!.
conFuncCallListPars1(_,_,nill,[]) :- 
	!.
conFuncCallListPars1(Dom,Form,list(P,Ps),[NL|NLs]) :- 
	conValX(P,Dom,Form,NL),
	conFuncCallListPars1(Dom,Form,Ps,NLs),
	!.
% ----------------------------------------------------------------------------------------------
% conFuncCallRS
% ----------------------------------------------------------------------------------------------
class predicates
conFuncCallParsRS : (string,sterml,sterml,string_list) determ (i,i,i,o).
conFuncCallListParsRS : (string,string,sterm,string_list) determ (i,i,i,o).
clauses
% 'V is FuncCall' type of statement
%
conFuncCallRS(FVal,cmp(Func,[_,var(Var)|Pars]),NL) :-
	dbBase::smDefTreeF(Func,_,Def),
	Def=cmp(_,[_,_,cmp("FPAR",DEs)]),
	dbBase::nlSing(Func,S1),
	conFuncCallParsRS(FVal,DEs,Pars,S2),
	dbBase::joinNL(S2,S3),
	sx::concat([Var,"=",S1," [",S3,"]"],NL),!.
conFuncCallRS(FVal,cmp(Func,[var(Var)|Pars]),NL) :-
	dbBase::smDefTreeF(Func,_,Def),
	Def=cmp(_,[_,_,cmp("FPAR",DEs)]),
	dbBase::nlSing(Func,S1),
	conFuncCallParsRS(FVal,DEs,Pars,S2),
	dbBase::joinNL(S2,S3),
	sx::concat([Var,"=",S1," [",S3,"]"],NL),!.
conFuncCallParsRS(_,[],[],[]) :- 
	!.
conFuncCallParsRS(FVal,[fstr(FP,D,F,_)|Ds],[P|Ps],[NL|NLs]) :- 
	dbBase::nlSing(FP,FPn),
	conFuncVar(FVal,P,AP),
	sx::concat([FPn,"=",AP],NL),
	conFuncCallParsRS(FVal,Ds,Ps,NLs),
	!
;
	sx::sxTail(2,D,Dom,T),
	T="_l",
	conFuncCallListParsRS(Dom,F,P,NL1),
	dbBase::joinNL(NL1,AP),
	dbBase::nlSing(FP,FPn),
	sx::concat([FPn,"=",AP],NL),
	conFuncCallParsRS(FVal,Ds,Ps,NLs),
	!
;	
	conValX(P,D,F,AP),
	dbBase::nlSing(FP,FPn),
	sx::concat([FPn,"=",AP],NL),
	conFuncCallParsRS(FVal,Ds,Ps,NLs),
	!.
	
conFuncCallListParsRS(_,_,nill,[]) :- 
	!.
conFuncCallListParsRS(D,F,list(P,Ps),[NL|NLs]) :- 
	conValX(P,D,F,NL),
	conFuncCallListParsRS(D,F,Ps,NLs),!.
% ----------------------------------------------------------------------------------------------
% conIndirect
% ----------------------------------------------------------------------------------------------
conIndirect(1,_,_,PR,NL) :-
	% SSS
	conQueryCondition(PR,S3),
	dbBase::nlForm("p3_iSSS","p3_iNIL",S_ino),
	sx::concat([S_ino," ",S3],NL),
	!.
conIndirect(2,1,Var,PR,NL) :-
	% det. DERF
	vars::varComps(Var,_,_,DE),
	dbBase::nlForm(DE,"p3_iSnu",S1),
	conQueryCondition(PR,S3),
	dbBase::nlForm("p3_iVal","p3_iNIL",S_val),
	dbBase::nlForm("p3_iWhr","p3_iNIL",S_where),
	dbBase::nlForm("p3_iFst","p3_iNIL",S_first),
	sx::concat([S_first," ",S_val," ",S1," ",S_where," ",S3],NL),
	!.
conIndirect(2,-1,Var,PR,NL) :-
	% non-det. DERF
	vars::varComps(Var,_,_,DE),
	dbBase::nlForm(DE,"p3_iSnu",S1),
	conQueryCondition(PR,S3),
	dbBase::nlForm("p3_iVal","p3_iNIL",S_val),
	dbBase::nlForm("p3_iWhr","p3_iNIL",S_where),
	sx::concat([S_val," ",S1," ",S_where," ",S3],NL),
	!.
% ----------------------------------------------------------------------------------------------
% conSentenceCond
% ----------------------------------------------------------------------------------------------
% constructs NL for a predicate (and arguments) in a query
% parameters:
% 1	in	possible predicate component
% 2	in	total condition
% 3	out	NL
%
conSentenceCond(Pred,Cond,S) :- 
	Pred=cmp(P,[Arg|DEs]),
	dbBase::smClassDefF(P,"p",Card,_,_,_,_),
	Card=1,
	% unary
	conDE_Vars(DEs,Cond,T2),
	conWithTail("p3_iWith",T2,S3),
	dbBase::nlSing(P,S4),
	conArg(Arg,Cond,S5),
	sx::concat([S5," ",S4," ",S3],S),
	!
;
	Pred=cmp(P,[Arg1,Arg2|DEs]),
	dbBase::smClassDefF(P,"p",Card,_,_,_,_),
	Card=2,
	% binary
	conDE_Vars(DEs,Cond,T2),
	conWithTail("p3_iWith",T2,S3),
	dbBase::nlSing(P,S4),
	conArg(Arg1,Cond,S5),
	conArg(Arg2,Cond,S6),
	sx::concat([S5," ",S4," ",S6,S3],S),
	!.
% ----------------------------------------------------------------------------------------------
% conQueryCondition
% ----------------------------------------------------------------------------------------------
conQueryCondition(Q,S) :-
    Term=termDomain::fromTermStringIndirect(Q),
	conCondition(Term,Term,S2),
	dbBase::joinNL(S2,S),
	!.
% ----------------------------------------------------------------------------------------------
% conSolArg
% ----------------------------------------------------------------------------------------------
class predicates
conSolEntity : (integer,integer,integer,sterm,string) determ (i,i,i,i,o).
clauses
% constructs a sentence argument from the solution db
% parameters:
% 1	in	sol.ref
% 2	in	sol.no
% 3	in	first sol.comp no to be read
% 4	out	next	"
% 5	in	argument	
% 6	out	argument NL
%
conSolArg(SolRef,SolNo,ArgNoC,ArgNoR,Arg,S) :-	
	Arg=var(V),
	sol::solCompVar(SolRef,SolNo,V,Val),
	conEntity(SolRef,SolNo,Val,S),
	ArgNoR=ArgNoC+1,
	!
;
	Arg=cmp(Ent,Vars),
	conSolEntity(SolRef,SolNo,ArgNoC,cmp(Ent,Vars),S),
	set::size(Vars,N),
	ArgNoR=ArgNoC+N,
	!.
conSolEntity(SolRef,SolNo,CompNo,cmp(Entity,_),S) :-
	globalVars::get_i("f_atom")=1,
	dbBase::nlSing(Entity,EntNL),
	dbBase::smClassDefTree(Entity,"e",EntDef),
	EntDef=cmp(_,[_,cmp(_,DEs)]),
	DEs=[_],
	!,
	conSolDEs(1,-1,SolRef,SolNo,CompNo,DEs,DEsNL2),
	dbBase::joinNL(DEsNL2,DEsNL),
	sx::sxlJoin([EntNL,DEsNL]," ",S),
	!
;	
	globalVars::get_i("f_atom")=2,
	dbBase::nlSing(Entity,EntNL),
	dbBase::smClassDefTree(Entity,"e",EntDef),
	EntDef=cmp(_,[_,cmp(_,DEs)]),
	DEs=[_],
	!,
	conSolDEs(2,-1,SolRef,SolNo,CompNo,DEs,DEsNL2),
	dbBase::joinNL(DEsNL2,DEsNL),
	sx::sxlJoin([DEsNL,EntNL]," ",S),
	!
;	
	dbBase::nlSing(Entity,EntNL),
	dbBase::smClassDefTree(Entity,"e",EntDef),
	EntDef=cmp(_,[_,cmp(_,DEs)]),
	Eq=globalVars::get_i("f_deeq"),
	conSolDEs(3,Eq,SolRef,SolNo,CompNo,DEs,DEsNL2),
	dbBase::joinNL(DEsNL2,DEsNL),
	dbBase::nlSing("p3_iWith",With),
	sx::sxlJoin([EntNL,With,DEsNL]," ",S),
	!.
% ----------------------------------------------------------------------------------------------
% conSelConds
% ----------------------------------------------------------------------------------------------
% selects the conditions from 2nd parameter which has the left side attribute
% variable in the given list of attribute variables (1. par).
%
conSelConds(Atts,cmp(",",[cmp(Opr,[LSide,RSide]),Rest]),[cmp(Opr,[LSide,RSide])|Conds]) :-	
	set::member(LSide,Atts),
	conSelConds(Atts,Rest,Conds),!.
conSelConds(Atts,cmp(",",[_,Rest]),Conds) :- conSelConds(Atts,Rest,Conds),!.
conSelConds(Atts,cmp(Opr,[LSide,RSide]),[cmp(Opr,[LSide,RSide])]) :- 
	set::member(LSide,Atts),!.
conSelConds(_,cmp(_,[_,_]),[]) :- !.
% ----------------------------------------------------------------------------------------------
% conSolDEs
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	ref f_atom
% 2	in	ref. f_deeq
% 3	in	solution reference# 
% 4	in	solution no 
% 5	in	component no
% 6	in	DE definitions
% 7	out	NL
%
conSolDEs(_,_,_,_,_,[],[]) :- 
	!.
conSolDEs(Form,Eq,SolRef,SolNo,CompNo,[fstr(_,Dom,Frm,_)|DEs],[NL|NLs]) :-
	Form<3,
	sol::solCompNo(SolRef,SolNo,CompNo,T),
	conValX_F(T,Dom,Frm,NL),
	Next=CompNo+1,
	conSolDEs(Form,Eq,SolRef,SolNo,Next,DEs,NLs),
	!.
conSolDEs(3,Eq,SolRef,SolNo,CompNo,[fstr(Name,Dom,Frm,_)|DEs],[NL|NLs]) :-
	sol::solCompNo(SolRef,SolNo,CompNo,T),
	conValX_F(T,Dom,Frm,SVal),
	dbBase::nlSing(Name,NameNL),
	Iq=Eq+2,
	set::setGet([" ","","="],Iq,Op),
	sx::concat([NameNL,Op,SVal],NL),
	Next=CompNo+1,
	conSolDEs(3,Eq,SolRef,SolNo,Next,DEs,NLs),
	!.
% ----------------------------------------------------------------------------------------------
% conSolDQ
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	solution reference# 
% 2	in	output text option (same as f_outq)
% 3	in	-2 query with variables - no solution
%    		-1 query without variables (yes/no)
%    		N if solution N is to be constructed
% 4	in	term to be translated to NL
% 5	out	NL
%
conSolDQ(_,0,_,_,"") :- 
	!. % no output option
conSolDQ(SolRef,_,SolNo,Term,S) :- 
			tt::tt_term(2,"Term for construction",Term),
			tt::tt_i(2,"SolNo",SolNo),
	Term=cmp(_,[var(V),str(DE_Var),_]),
	conQuery(Term,Sq),
	vars::varComps(DE_Var,_,_,DE),
	dbBase::smDefF(DE,Clas,_,_,Dom,Form,_),
	Clas="a",
	sol::solCompVar(SolRef,SolNo,V,V2),
	conValX(V2,Dom,Form,Val),
	dbBase::nlSing("=",Is),	
	sx::sxlJoin([Val,Is,Sq]," ",S),
	!.
% ----------------------------------------------------------------------------------------------
% conSolEQ
% ----------------------------------------------------------------------------------------------
class predicates
conArgNo : (sterml,string) determ (i,o).
clauses
% parameters:
% 1	in	solution reference# 
% 2	in	output text option (same as f_outq)
% 3	in	-2 query with variables - no solution
%    		-1 query without variables (yes/no)
%    		N if solution N is to be constructed
% 4	in	term to be translated to NL
% 5	out	NL
%
conSolEQ(_,0,_,_,"") :- 
	!. % no output option
conSolEQ(_,_,-2,Term,S) :- 
	% negative: unconditional unary relation
	Term=cmp(Rel,[var(_)]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iSnu",S1),
	dbBase::nlForm("p3_iNon","p3_iNIL",None),
	sx::concat([None," ",S1,"! "],S),
	!
;	
	% negative: unary relation with entity type given
	Term=cmp(Rel,[cmp(Ent,_)]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iPlu",S1),
	string::isName(Ent),
	dbBase::nlForm(Ent,"p3_iPlu",S2),
	dbBase::nlForm("p3_iNo","p3_iNIL",No),
	sx::concat([No," ",S2," ",S1,"! "],S),
	!
;
	% negative: conditional unary relation
	Term=cmp(",",[cmp(Rel,[Ent])|Conds]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iPlu",S1),
	conEntities(cmp(",",[Ent|Conds]),S2),
	dbBase::nlForm("p3_iNo","p3_iNIL",No),
	sx::concat([No," ",S2," ",S1,"! "],S),
	!
;
	% negative: binary relation (conditions for entity attributes) 
	Term=cmp(",",
	        [cmp(Rel,
	            [cmp(Ent1,Vars1),
	             cmp(Ent2,Vars2)]),
	         Conds]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iPlu",S1),
	%
	conArgNo([cmp(Ent1,Vars1)],S8),
	% negative: select the conditions in Conds relevant for Ent1
	conSelConds(Vars1,Conds,CondsList1),
	reformat(CondsList1,CondsCmp1),
	conEntCond([CondsCmp1],Vars1,S3),
	S3=[S4],
	%
	string::isName(Ent2),
	dbBase::nlForm(Ent2,"p3_iPlu",S9),
	% select the conditions in Conds relevant for Ent2
	conSelConds(Vars2,Conds,CondsList2),
	reformat(CondsList2,CondsCmp2),
	conEntCond([CondsCmp2],Vars2,S6),
	S6=[S7],
	%	
	sx::concat([S8,S4," ",S1," ",S9,S7,"! "],S),
	!	
;
	% negative: binary relation (conditions for entity attributes) 
	Term=cmp(",",
	        [cmp(Rel,
	            [Arg1,
	             cmp(Ent2,Vars2)]),
	         Conds]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iPlu",S1),
	%
	conArgNo([Arg1],S2),
	%
	string::isName(Ent2),
	dbBase::nlForm(Ent2,"p3_iPlu",S9),
	% select the conditions in Conds relevant for Ent2
	conSelConds(Vars2,Conds,CondsList2),
	reformat(CondsList2,CondsCmp2),
	conEntCond([CondsCmp2],Vars2,S6),
	S6=[S7],
	%	
	sx::concat([S2," ",S1," ",S9,S7,"! "],S),
	!	
;
	% negative: binary relation (conditions for entity attributes) 
	Term=cmp(",",
	        [cmp(Rel,
	            [cmp(Ent2,Vars2),
	             Arg1]),
	         Conds]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iPlu",S1),
	%
	conArgNo([Arg1],S2),
	%
	string::isName(Ent2),
	dbBase::nlForm(Ent2,"p3_iPlu",S9),
	% select the conditions in Conds relevant for Ent2
	conSelConds(Vars2,Conds,CondsList2),
	reformat(CondsList2,CondsCmp2),
	conEntCond([CondsCmp2],Vars2,S6),
	S6=[S7],
	%	
	sx::concat([S9,S7," ",S1," ",S2,"! "],S),
	!	
;
	% negative: binary relation (no conditions for entity attributes) 
	Term=cmp(Rel,[Arg1,Arg2]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iPlu",S1),
	conArgNo([Arg1],S2),
	conArgNo([Arg2],S3),
	sx::concat([S2," ",S1," ",S3,"! "],S),
	!.	
conSolEQ(SolRef,_,SolNo,Term,S) :- 
			tt::tt_term(2,"Term for construction",Term),
			tt::tt_i(2,"SolNo",SolNo),
	% UR 
	Term=cmp(Rel,[Arg]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iSnu",S1),
	conSolArg(SolRef,SolNo,1,_,Arg,S2),
	sx::concat([S2," ",S1,"! "],S),
	!
;
	% URA 
	Term=cmp(Rel,[Arg|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(1)]),cmp("ATTRS",AttDefs)]),
	dbBase::nlForm(Rel,"p3_iSnu",S1),
	conSolArg(SolRef,SolNo,1,_,Arg,S2),
	conDEs(SolRef,SolNo,Attrs,AttDefs,S4),
	dbBase::nlForm("p3_iWith","p3_iNIL",With),
	sx::concat([S2," ",S1," ",With," ",S4,"! "],S),
	!
;
	% BR 
	Term=cmp(Rel,[Arg1,Arg2]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iSnu",S1),
	conSolArg(SolRef,SolNo,1,NextArg,Arg1,S2),
	conSolArg(SolRef,SolNo,NextArg,_,Arg2,S3),
	sx::concat([S2," ",S1," ",S3,"! "],S),
	!
;
	% BR 
	Term=cmp(",",[cmp(Rel,[Arg1,Arg2])|_]),
	string::isName(Rel),
	dbBase::nlForm(Rel,"p3_iSnu",S1),
	conSolArg(SolRef,SolNo,1,NextComp,Arg1,S2),
	conSolArg(SolRef,SolNo,NextComp,_,Arg2,S3),
	sx::concat([S2," ",S1," ",S3,"! "],S),
	!
;
	% BRA
	Term=cmp(Rel,[Arg1,Arg2|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(2)]),cmp("ATTRS",AttDefs)]),
	dbBase::nlForm(Rel,"p3_iSnu",S1),
	conSolArg(SolRef,SolNo,1,NextComp,Arg1,S2),
	conSolArg(SolRef,SolNo,NextComp,_,Arg2,S3),
	conDEs(SolRef,SolNo,Attrs,AttDefs,S4),
	dbBase::nlForm("p3_iWith","p3_iNIL",With),
	sx::concat([S2," ",S1," ",S3," ",With," ",S4,"! "],S),
	!.
% constructs NL for an argument in a relational query with no solutions
% parameters:
% 1	in	argument
% 2	out	NL text
% Date	15May95
%
conArgNo([var(_)],"none") :- !.
conArgNo([cmp(Ent,_)],S) :-
	string::isName(Ent),
	dbBase::nlForm(Ent,"p3_iPlu",S1),
	dbBase::nlForm("p3_iNo","p3_iNIL",No),
	sx::concat([No," ",S1],S),!.
% ----------------------------------------------------------------------------------------------
% conSolFQ
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	solution reference# 
% 2	in	output text option (same as f_outq)
% 3	in	-2 query with variables - no solution
%    		-1 query without variables (yes/no)
%    		N if solution N is to be constructed
% 4	in	term to be translated to NL
% 5	out	NL
%
conSolFQ(_,0,_,_,"") :- 
	!. % no output option
conSolFQ(SolRef,Opt,SolNo,Term,S) :- 
			tt::tt_term(2,"Term for construction",Term),
			tt::tt_i(2,"SolNo",SolNo),
	Term=cmp(Func,[atom("p3_iDtFn")|Pars]),
	conSolFQ(SolRef,Opt,SolNo,cmp(Func,Pars),S),
	!
;	
	Term=cmp(F,[var(V)|_]),
	dbBase::smDefTreeF(F,_,Def),
	Def=cmp(_,[_,cmp(_,[atom(Dom)])|_]),
	sol::solCompVar(SolRef,SolNo,V,Val),
	conValX(Val,Dom,"",S3),
	conQuery(Term,S2),
	sx::concat([S3,"=",S2],S),
	!.
% ----------------------------------------------------------------------------------------------
% conSolSQ
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	solution reference# 
% 2	in	output text option (same as f_outq)
% 5	out	NL
%
conSolSQ(_,0,"") :- 
	!. % no output option
conSolSQ(SolRef,_,S) :- 
	sol::solCompNo(SolRef,1,1,Val),
	Val=int(N),
	Ns=toString(N),
	S=string::concat(Ns,"!"),
	!.
% ----------------------------------------------------------------------------------------------
% conValue
% ----------------------------------------------------------------------------------------------
conValue(Var,S) :- 
	Var=var(S),!
;
	Var=fvar(S,_),!.	
conValue(C,S)   :- 
	conDERF(C,S), 
	!.
conValue(C,S) :- 
	conExpression(C,S), 
	!.
conValue(T,S)   :- 
	conValX(T,"p3_dany","",S), 
	!.
% ----------------------------------------------------------------------------------------------
% conVarCond
% ----------------------------------------------------------------------------------------------
class predicates
constr_expr : (string,sterm,string,string,string) determ (i,i,i,i,o).
clauses
conVarCond(cmp(F,[_,_,V,_,_,_]),"") :-	
	set::member(F,["p3_fIRan","p3_fORan"]),
	V=var(W),
	vars::varComps(W,_,_,_),
	!.
conVarCond(cmp(F,[_,_,V,str(Br),Low,High]),S) :-	
	V=var(W),
	set::member(F,["p3_fIRan","p3_fORan"]),
	set::setGet(["p3_fIRan","p3_fORan"],N,F),
	set::setGet(["p3_iIRn","p3_iORn"],N,S2),
	dbBase::nlSing(S2,S3),
	sx::frontstr(1,Br,Ba,Bb),
	conValX(Low,"p3_dnu","",S5),
	conValX(High,"p3_dnu","",S6),
	sx::concat([W," ",S3," ",Ba,S5,", ",S6,Bb],S),
	!.
conVarCond(cmp(Opr,[Var,Exp]),S) :-
	Var=var(V),
	opr::opr_cmpass(Opr),
	vars::varDE_Var(V,_),
	conValue(Exp,S1),
	constr_expr(Opr,Exp,V,S1,S),
	!
;	
	Var=fvar(V,_),
	opr::opr_cmpass(Opr),
	vars::varDE_Var(V,_),
	conValue(Exp,S1),
	constr_expr(Opr,Exp,V,S1,S),
	!
;
	Var=var(V),
	opr::opr_cmpass(Opr),
	conValue(Exp,S1),
	constr_expr(Opr,Exp,V,S1,S),
	!.
	
constr_expr(Opr,cmp(H,_),Ls,Rs,S) :-
	% composite right side
	opr::opr_any(Opr,2),
	H<>"PROC",
	dbBase::nlSing(Opr,S1),
	sx::concat([Ls," ",S1," (",Rs,")"],S),!.
constr_expr(Opr,_,Ls,Rs,S) :-
   	opr::opr_any(Opr,2),
	conCheckTaut(Opr,Ls,Rs,S),!
;
   	opr::opr_any(Opr,2),
	conCheckTaut(Opr,Ls,Rs,S),!.	
% ----------------------------------------------------------------------------------------------
% conWithTail
% ----------------------------------------------------------------------------------------------
conWithTail(_,[],"") :- 
	!.
conWithTail(WithFor,S,T) :-
	dbBase::nlSing(WithFor,With),
	dbBase::joinNL(S,S1),
	sx::concat([" ",With," ",S1],T),
	!.
% ----------------------------------------------------------------------------------------------
% cprTerm
% ----------------------------------------------------------------------------------------------
class predicates
cprTerms : (sterml,string_list) determ (i,o).
cprTermList : (sterm,string_list) determ (i,o).
clauses

cprTerm(cmp(H,T),PR) :-
	H=",",
	cprTerms(T,P2),
	sx::sxlJoin(P2,",",PR),
	!
;
	opr::opr_any(H,2),
	T=[L,R],
	cprTerm(L,Lp),
	cprTerm(R,Rp),
	sx::concat([Lp," ",H," ",Rp],PR),
	!
;
	cprTerms(T,P2),
	sx::sxlJoin(P2,",",P3),
	sx::concat([H,"(",P3,")"],PR),
	!.
cprTerm(atom(A),A) :- 
	!.
cprTerm(var(V),V) :- 
	!.
cprTerm(str(S),PR) :- 
	sx::concat(["\"",S,"\""],PR),
	!.
cprTerm(int(I),PR) :- 
	PR=toString(I),
	!.
cprTerm(fix(F),PR) :- 
	PR=toString(F),
	!.
cprTerm(nill,"[]") :- 
	!.
cprTerm(list(H,T),PR) :- 
	cprTermList(list(H,T),L),
	sx::sxlJoin(L,",",Ls),
	sx::concat(["[",Ls,"]"],PR),
	!.
cprTerm(Term,_) :-
	dlgIF::feil("cprTerm",[],[],Term),!,fail.
		
cprTermList(nill,[]) :- 
	!.
cprTermList(list(H,T),[Sh|St]) :-
	cprTerm(H,Sh),
	cprTermList(T,St),
	!.	
	
cprTerms([],[]) :-
	!.
cprTerms([H|T],[Hm|Tm]) :-
	cprTerm(H,Hm),
	cprTerms(T,Tm),
	!.
% ----------------------------------------------------------------------------------------------
% cprSolEQ
% ----------------------------------------------------------------------------------------------
class predicates
cprSolDEs : (integer,integer,sterml,string_list) determ (i,i,i,o).
cprEntity : (integer,integer,sterm,string) determ (i,i,i,o).
cprSolArg : (integer,integer,integer,integer,sterm,string) determ (i,i,i,o,i,o).
cprVal : (integer,integer,sterm,string) determ (i,i,i,o).
cprValX : (sterm,string) determ (i,o).
cprValXs : (sterml,string_list) determ (i,o).
cprValX_L : (sterm,string_list) determ (i,o).	
clauses
% parameters:
% 1	in	solution reference# 
% 2	in	N if solution N is to be constructed
% 3	in	term to be translated to NL
% 4	out	PR
%
cprSolEQ(SolRef,SolNo,Term,S) :- 
			tt::tt_term(2,"Term for construction",Term),
			tt::tt_i(2,"SolNo",SolNo),
	Term=cmp(Rel,[Arg|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(1)]),cmp("ATTRS",_)]),
	% URA 
	cprSolArg(SolRef,SolNo,1,_,Arg,S2),
	cprSolDEs(SolRef,SolNo,Attrs,S4s),
	sx::sxlJoin(S4s,",",S4),
	sx::concat([Rel,"(",S2,",",S4,")!"],S),
	!
;
	Term=cmp(Rel,[Arg]),
	string::isName(Rel),
	cprSolArg(SolRef,SolNo,1,_,Arg,S2),
	% UR 
	sx::concat([Rel,"(",S2,")!"],S),
	!
;
	Term=cmp(Rel,[Arg1,Arg2|Attrs]),
	string::isName(Rel),
	dbBase::smClassDefTree(Rel,"p",RelDef),
	RelDef=cmp(_,[_,cmp("CARD",[int(2)]),cmp("ATTRS",_)]),
	% BRA
	cprSolArg(SolRef,SolNo,1,NextComp,Arg1,S2),
	cprSolArg(SolRef,SolNo,NextComp,_,Arg2,S3),
	cprSolDEs(SolRef,SolNo,Attrs,S4s),
	sx::sxlJoin(S4s,",",S4),
	sx::concat([Rel,"(",S2,",",S3,",",S4,")!"],S),
	!
;
	% BR 
	Term=cmp(Rel,[Arg1,Arg2]),
	string::isName(Rel),
	cprSolArg(SolRef,SolNo,1,NextArg,Arg1,S2),
	cprSolArg(SolRef,SolNo,NextArg,_,Arg2,S3),
	sx::concat([Rel,"(",S2,",",S3,")!"],S),
	!
;
	% BR 
	Term=cmp(",",[cmp(Rel,[Arg1,Arg2])|_]),
	string::isName(Rel),
	cprSolArg(SolRef,SolNo,1,NextComp,Arg1,S2),
	cprSolArg(SolRef,SolNo,NextComp,_,Arg2,S3),
	sx::concat([Rel,"(",S2,",",S3,")!"],S),
	!.
% constructs PR for a relation argument from the solution db
% parameters:
% 1	in	sol.ref
% 2	in	sol.no
% 3	in	first sol.comp no to be read
% 4	out	next	"
% 5	in	argument	
% 6	out	argument PR
%
cprSolArg(SolRef,SolNo,ArgNoC,ArgNoR,Arg,S) :-	
	Arg=var(V),
	sol::solCompVar(SolRef,SolNo,V,Val),
	cprEntity(SolRef,SolNo,Val,S),
	ArgNoR=ArgNoC+1,
	!
;
	Arg=cmp(Ent,Vars),
	cprEntity(SolRef,SolNo,cmp(Ent,Vars),S),
	set::size(Vars,N),
	ArgNoR=ArgNoC+N,
	!.
cprEntity(SolRef,SolNo,cmp(Entity,Attrs),S) :-
	cprSolDEs(SolRef,SolNo,Attrs,DEs),
	sx::sxlJoin(DEs,",",DE),
	sx::concat([Entity,"(",DE,")"],S),
	!.	
cprSolDEs(_,_,[],[]) :- 
	!.
cprSolDEs(SolRef,SolNo,[Val|Vals],[PR|PRs]) :-
	cprVal(SolRef,SolNo,Val,PR),
	cprSolDEs(SolRef,SolNo,Vals,PRs),
	!.
			
cprVal(SolRef,SolNo,var(Var),Ts) :-
	sol::solVarNo(SolRef,Var,CompNo),
	sol::solCompNo(SolRef,SolNo,CompNo,T),
	cprValX(T,Ts),
	!.
cprVal(_,_,Val,Ts) :-
	cprValX(Val,Ts),
	!.
cprValX(int(I),S) :- 
	S=toString(I),
	!.
cprValX(fix(F),S) :- 
	S=toString(F),
	!.
cprValX(atom(A),A) :- 
	!.
cprValX(str(S),Sq) :- 
	sx::concat(["\"",S,"\""],Sq),
	!.
cprValX(nill,"[]") :- 
	!.
cprValX(list(H,T),S) :-
	cprValX_L(list(H,T),Ss),
	sx::sxlJoin(Ss,",",S1),
	sx::concat(["[",S1,"]"],S),
	!.
cprValX(T,_) :- 
	dlgIF::feil("cprValX",[],[],T),!,fail.	
	
cprValX_L(nill,[]) :-
	!.	
cprValX_L(list(H,T),[S|Ss]) :-
	cprValX(H,S),
	cprValX_L(T,Ss),
	!.	
cprValXs([],[]) :-
	!.
cprValXs([V|Vs],[H|T]) :-
	cprValX(V,H),
	cprValXs(Vs,T),
	!.
end implement nl
