implement p3Tree
open core,termDomain,resourceIdentifiers,vpiDomains,globalDomains
constants
className = "com/rudstrand/p3Tree/p3Tree".
classVersion = "Version 1.0".
clauses
% current class description
classInfo(className, classVersion).

class predicates
setSize : (vpiTree::tree_list Set,integer Size) procedure (i,o).
drwCompress : (vpiTree::tree,vpiTree::tree) determ (i,o).
drwGetCR : (integer,string,string,string) determ (i,i,o,o).
drwGraphComp : (string) determ (i).
drwRightClickSingle : (vpiDomains::windowHandle) determ (i).
drwRightClickDouble : (vpiDomains::windowHandle) determ (i).
drwRuleComp : (string,integer_list,string) determ (i,i,i).
dtNodeTexts : (integer,integer,integer,integer,integer,sterm,sterm) determ (i,i,i,i,i,i,o).
dtCompact : (sterm,sterm) determ (i,o).
dtConvert : (sterm,integer,integer,vpiTree::tree) determ (i,i,o,o).

clauses
% ----------------------------------------------------------------------------------------------
% dtNodeText
% ----------------------------------------------------------------------------------------------
class predicates
dtNT : (integer,string,string) determ (i,i,o).
dtNTs : (integer,integer,integer,integer,integer,sterml,sterml) determ (i,i,i,i,i,i,o).
dtNT_T : (integer,string) determ (i,o).
dtNT_S : (integer,string,string) procedure (i,i,o).
clauses
% parameters:
% 1	in	1=no type text; 2=* function indicator; 3=ST/SG/RL/FN
% 2	in	1=no counts; 2=counts
% 3	in	rule name: 1=address; 2=realm; 3=both
% 4	in	function name: 1=KB; 2=NL; 3=both
% 5	in	ST/SG name: 1=KB; 2=NL; 3=both
% 6	in 	tree
% 7	out	tree
%
dtNodeTexts(C,_,_,_,_,termDomain::str(Node),termDomain::str(NodeM)) :- 
	dtNT(C,Node,NodeM),
	!.
dtNodeTexts(C,SR,CR,FU,XP3F,cmp(Node,Nodes),cmp(NodeM,NodesM)) :-
	dtNT(C,Node,NodeM),
	dtNTs(C,SR,CR,FU,XP3F,Nodes,NodesM),
	!.
dtNT(C,S,Sm) :-
	dtConvNode(S,N),
	N=cmp(_,[int(Type),termDomain::str(Name)]),
	dtNT_T(Type,S1),
	gra::dtNT_C(C,Type,Name,S2),
	dtNT_S(Type,Name,S5),
	sx::concat([S1,S5,S2],Sm),
	!.
	
dtNTs(_,_,_,_,_,[],[]) :- 
	!.
dtNTs(C_Tx,SR,CR,FU,XP3F,[Node|Nodes],[NodeM|NodesM]) :-
	dtNodeTexts(C_Tx,SR,CR,FU,XP3F,Node,NodeM),
	dtNTs(C_Tx,SR,CR,FU,XP3F,Nodes,NodesM),
	!.

dtNT_T(1,"") :- !.
dtNT_T(2,"") :- !.
dtNT_T(3,"") :- !.
dtNT_T(5,"") :- !.
dtNT_T(15,"[C] ") :- !.
dtNT_T(19,"[F] ") :- !.

dtNT_S(Type,K,S) :-
    set::member(Type,[1,15,19]),
	dbBase::nlSing(K,S),
	!.
dtNT_S(2,ST,S) :- 
	dbBase::stNL(ST,S),
	!.
dtNT_S(Type,Rule,S) :-
    set::member(Type,[3,5]),
	sx::sxUnpack(Rule,'|',L),
	L=[P|T],
	Id=dbBase::dbIdChain(P),
	sx::sxlJoin([Id|T],"|",S),
	!.
dtNT_S(_,_,"") :- 
	!.
% ----------------------------------------------------------------------------------------------
% dtDepTree/Open
% ----------------------------------------------------------------------------------------------
class predicates
dtDir : (integer,integer) determ (i,o).
clauses

dtDepTreeOpen(Matrix,TreeType,ST,SC,SR,CR,FU,XP3F,Node,"ALL") :- 
	dtDepTreeOpen(Matrix,TreeType,ST,SC,SR,CR,FU,XP3F,Node,"0"),
	!.
dtDepTreeOpen(Matrix,TreeType,STx,SC,SR,CR,FU,XP3F,Node,Depth) :-
	vpi::cursorSetWait(),
	set::setGet(["p3_pDP","p3_pDR"],Matrix,Graph),
	GR=gra::newGraph(Graph),
	GR:filter(STx,SR,CR,FU,XP3F),
	Graphs=GR:get(),
	%
	Levels=sx::toInt(Depth),
	gra::graTree(Graphs,TreeType,Node,Levels,Tree1),
	dtNodeTexts(SC,SR,CR,FU,XP3F,Tree1,Tree2),
	dtCompact(Tree2,Tree3),
	dtConvert(Tree3,1,_,Tree),
	Tree2=cmp(Sn,_),
	dtDepTreeTitle(Matrix,TreeType,Sn,Title),
	dtDir(TreeType,Dir),
	drwOpenTree(Title,50,Tree,Dir,_),
	!.
dtDir(2,vpiTree::tree_DirLeft) :- 
	!.
dtDir(1,vpiTree::tree_DirRight) :- 
	!.
% ----------------------------------------------------------------------------------------------
% dtConvert
% ----------------------------------------------------------------------------------------------
class predicates
dtConvertX : (sterml,integer,integer,vpiTree::tree_list) determ (i,i,o,o).
clauses
% converts from a STERM-tree to a tree-package tree
%
dtConvert(cmp(S,[]),N,M,vpiTree::tree(T,vpiTree::unmarked,[],0)) :- 
	Ns=toString(N),
	sx::concat([S,"{",Ns],T),
	M=N+1,
	!.
dtConvert(cmp(H,B),N,M,vpiTree::tree(HH,vpiTree::unmarked,Bt,0)) :- 
	Ns=toString(N),
	sx::concat([H,"{",Ns],HH),
	Next=N+1,
	dtConvertX(B,Next,M,Bt),
	!.
dtConvert(termDomain::str(S),N,M,vpiTree::tree(T,vpiTree::unmarked,[],0)) :- 
	Ns=toString(N),
	sx::concat([S,"{",Ns],T),
	M=N+1,
	!.
dtConvert(T,N,_,_) :- 
	dlgIF::feil("dtConvert",[],[N],T),!,fail.
dtConvertX([],N,N,[]) :- 
	!.
dtConvertX([H|T],N,M,[Ht|Tt]) :- 
	dtConvertX(T,N,N2,Tt),
	dtConvert(H,N2,M,Ht),
	!.
% ----------------------------------------------------------------------------------------------
% dtCompact
% ----------------------------------------------------------------------------------------------
class predicates
dtCompactX : (sterml,sterml) determ (i,o).
dtCompactY : (sterml,sterml) determ (i,o).
clauses
dtCompact(cmp(H,T),cmp(H,Tm)) :-
	dtCompactX(T,T2),
	dtCompactY(T2,Tm),
	!.
	
dtCompactX([],[]) :-
	!.
dtCompactX([nill|T],Tm) :-
	dtCompactX(T,Tm),
	!.
dtCompactX([cmp("",Tn)|T],Tm) :- % used to be cmp(nill ??
	set::conc([Tn,T],Tj),
	dtCompactX(Tj,Tm),
	!.
dtCompactX([cmp(H,Tn)|T],[cmp(H,Tnm)|Tm]) :-
	dtCompactX(Tn,Tnm),
	dtCompactX(T,Tm),
	!.
dtCompactX([H|T],[H|Tm]) :-
	dtCompactX(T,Tm),
	!.
dtCompactY([],[]) :-
	!.
dtCompactY([H|T],Tm) :-
	set::member(H,T),
	dtCompactY(T,Tm),
	!.
dtCompactY([cmp(H,Ht)|T],[cmp(H,Htm)|Tm]) :-
	dtCompactY(Ht,Htm),
	dtCompactY(T,Tm),
	!.
dtCompactY([H|T],[H|Tm]) :-
	dtCompactY(T,Tm),
	!.
% ----------------------------------------------------------------------------------------------
% dtDepTreeTitle/Long
% ----------------------------------------------------------------------------------------------
dtDepTreeTitle(1,1,S,Title) :- sx::concat(["Determinants of ",S],Title),!.	
dtDepTreeTitle(1,2,S,Title) :- sx::concat(["Dependents of ",S],Title),!.	
dtDepTreeTitle(2,1,_,Title) :- sx::concat(["Determinants (Derivation Graph)"],Title),!.	
dtDepTreeTitle(2,2,_,Title) :- sx::concat(["Dependents (Derivation Graph)"],Title),!.	
dtDepTreeTitle(1,0,_,Title) :- sx::concat(["Dependency Graph"],Title),!.	
dtDepTreeTitle(2,0,_,Title) :- sx::concat(["Derivation Graph"],Title),!.	
dtDepTreeTitlelong(A,B,N,Title) :-
	dtDepTreeTitle(A,B,N,S),
	FP=globalVars::get_s("kb_file"),
	fio::file_name(FP,_,F,_),
	datetime::tsNow(L),
	datetime::tsStr(L,S2),
	sx::concat(["D",S2," ",S," ",F],Title),
	!.
% ----------------------------------------------------------------------------------------------
% dtConvNode/s
% ----------------------------------------------------------------------------------------------
% converts between node KB format and 'n|xxx' format used in tree drawing
%
dtConvNodes([],[]) :- 
	!.
dtConvNodes([S|Ss],[T|Ts]) :-
	dtConvNode(S,T),
	dtConvNodes(Ss,Ts),
	!.
dtConvNode(S,T) :-
	sx::frontstr(2,S,SType,T2),
	Type=sx::toInt(SType),
	sx::frontchar(T2,'|',Name),
	T=cmp("n",[int(Type),termDomain::str(Name)]),
	!
;
	sx::frontstr(1,S,SType,T2),
	Type=sx::toInt(SType),
	sx::frontchar(T2,'|',Name),
	T=cmp("n",[int(Type),termDomain::str(Name)]),
	!
;
	dlgIF::feil("dtConvNode",[S],[],nill),!,fail.	

dtConvNodeR(S,T) :-
	T=cmp("n",[int(Type),termDomain::str(Name)]),
	SType=toString(Type),
	sx::concat([SType,"|",Name],S),
	!
;
	dlgIF::feil("dtConvNodeR",[],[],T),!,fail.	
% ----------------------------------------------------------------------------------------------
% drwGetCurr
% ----------------------------------------------------------------------------------------------
drwGetCurr(TRW,NodeNo,NodeText,FullText) :-
	FullText=vpiTree::getSelectedNode(TRW),
	sx::sxUnpack(FullText,'{',NTs),
	NTs=[NodeText2,SNodeNo],
	sx::sxStrip(NodeText2,NodeText),
	NodeNo=sx::toInt(SNodeNo),
	!.
% ----------------------------------------------------------------------------------------------
% drwOpenTree
% ----------------------------------------------------------------------------------------------
class predicates
drwOpenTreeX : (wsflags,string,integer,vpiTree::tree,integer,vpiDomains::windowHandle) determ (i,i,i,i,i,o).
clauses
drwOpenTree(Title,Type,Tree,Dir,W) :-
	Type=41, % rule tree
	FLAG=[wsf_TitleBar,wsf_Visible,wsf_Minimize, wsf_Maximize,wsf_ClipChildren,wsf_ClipSiblings,wsf_SizeBorder,wsf_Maximized],
	drwOpenTreeX(FLAG,Title,Type,Tree,Dir,W),
    guix::addWin(Type,Title,W),
	!
;
	FLAG=[wsf_Close,wsf_TitleBar,wsf_Visible,wsf_Minimize, wsf_Maximize,wsf_ClipChildren,wsf_ClipSiblings,wsf_SizeBorder],
	drwOpenTreeX(FLAG,Title,Type,Tree,Dir,W),
    guix::addWin(Type,Title,W),
	!.
		
drwOpenTreeX(FLAG,RT_Name,Type,Tree,Dir,W) :-
	%not(set::member(Type,[41,44,45])),
	Type<>41,
	globalVars::get_i("f_trcm")=1,
	drwCompress(Tree,CTree),
  	FONT=vpi::fontCreate(vpiDomains::ff_Helvetica,[],7),
	TREE_WIN_INFO = vpiTree::tree_WinInfo(FONT,Dir,[]),
	Win=guix::getWinOfClas(1),	      
	R=vpi::winGetClientRect(Win),
	R=rct(Cx,Cy,Chx,Chy),
	My=Cy+20, Mhx=Chx-30, Mhy=Chy-80,
	MR=rct(Cx,My,Mhx,Mhy),
	W=vpiTree::create(w_TopLevel,MR,RT_Name,noMenu,Win,FLAG,tree_event,0,CTree,TREE_WIN_INFO),
	!
;
  	FONT=vpi::fontCreate(vpiDomains::ff_Helvetica,[],7),
	TREE_WIN_INFO = vpiTree::tree_WinInfo(FONT,Dir,[]),
	Win=guix::getWinOfClas(1),	      
	R=vpi::winGetClientRect(Win),
	R=rct(Cx,Cy,Chx,Chy),
	My=Cy+20, Mhx=Chx-30, Mhy=Chy-80,
	MR=rct(Cx,My,Mhx,Mhy),
	W=vpiTree::create(w_TopLevel,MR,RT_Name,noMenu,Win,FLAG,tree_event,0,Tree,TREE_WIN_INFO),
	!.
% ----------------------------------------------------------------------------------------------
% drwCompress
% ----------------------------------------------------------------------------------------------
class predicates
drwCX : (string_list,vpiTree::tree,string_list,vpiTree::tree) determ (i,i,o,o).
drwCXS : (string_list,vpiTree::tree_list,string_list,vpiTree::tree_list) determ (i,i,o,o).
clauses
% parameters:
% 1	in	tree
% 2	out	compressed tree (redundant sub-trees removed)
%
drwCompress(T,Tc) :-
	drwCX([],T,_,Tc),
	!.
drwCX(SL,vpiTree::tree(S,M,Ts,N),SL,vpiTree::tree(Sm,M,[],N)) :-
	setSize(Ts,Nt),
	Nt>0,
	sx::sxUnpack(S,'{',L),
	L=[Name,Suffix],
	set::member(Name,SL),
	sx::concat([Name,"*","{",Suffix],Sm),
	!.
drwCX(SL,vpiTree::tree(S,M,Ts,N),SLm,vpiTree::tree(S,M,Tsm,N)) :-
	sx::sxUnpack(S,'{',L),
	L=[Name,_],
	SLm2=[Name|SL],
	drwCXs(SLm2,Ts,SLm,Tsm),
	!.	
drwCX(SL,vpiTree::tree(S,_,_,_),_,_) :- 
	dlgIF::feil("drwCX",[S|SL],[],nill),!,fail.
drwCXS(SL,[],SL,[]) :-
	!.	
drwCXS(SL,[T|Ts],SLm,[Tm|Tsm]) :-
	drwCX(SL,T,SLm2,Tm),
	drwCXS(SLm2,Ts,SLm,Tsm),
	!.
% ----------------------------------------------------------------------------------------------
% drwGetCR
% ----------------------------------------------------------------------------------------------
drwGetCR(44,S,Rule,RuleK) :-
	sx::searchchar(S,':',P),
	sx::frontstr(P,S,_,S2),
	sx::sxUnpackS(S2," [",L),
	L=[Rule2|_],
	sx::sxStrip(Rule2,Rule),
	dbBase::kbID_F(Rule,RuleK,Clas,_),
	Clas="p3_iIR",
	!
;
	sx::sxUnpackS(S," [",L),
	L=[Rule2,_], 
	sx::sxStrip(Rule2,Rule),
	dbBase::kbID_F(Rule,RuleK,Clas,_),
	Clas="p3_iIR",
	!
;
	sx::sxStrip(S,Rule),
	dbBase::kbID_F(Rule,RuleK,Clas,_),
	Clas="p3_iIR",
	!.
drwGetCR(45,S,Rule,RuleK) :-
	sx::searchchar(S,':',P),
	Q=P-1,
	Q>0,
	sx::frontstr(Q,S,Rule2,_),
	sx::sxStrip(Rule2,Rule),
	dbBase::kbID_F(Rule,RuleK,Clas,_),
	Clas="p3_iIR",
	!.
% ----------------------------------------------------------------------------------------------
% tree_event
% ----------------------------------------------------------------------------------------------
class predicates
tree_event  : vpiDomains::ehandler.
clauses
% 41 SR tree
% 44 IR net
% 45 IR reason
% 50 dep. graph
%
tree_event(Win,e_MouseDbl(_,_,mouse_button_right))=0:-
	drwRightClickDouble(Win),
	!.
tree_event(Win,e_MouseDown(_,_,mouse_button_right))=0:-
	drwRightClickSingle(Win),
	!.
tree_event(Win,e_Menu(id_dir_left,_))=0 :-!,
	guix::getWinClas(Win)<>41,     
	vpiTree::setDirection(Win,vpiTree::tree_DirLeft),
	!.
tree_event(Win,e_Menu(id_dir_right,_))=0 :-!,
	guix::getWinClas(Win)<>41,     
	vpiTree::setDirection(Win,vpiTree::tree_DirRight),
	!.
tree_event(Win,e_Menu(id_dir_up,_))=0 :-!,
	guix::getWinClas(Win)<>41,     
	vpiTree::setDirection(Win,vpiTree::tree_DirUp),
	!.
tree_event(Win,e_Menu(id_dir_down,_))=0 :-!,
	guix::getWinClas(Win)<>41,     
	vpiTree::setDirection(Win,vpiTree::tree_DirDown),
	!.
tree_event(W,e_MouseDbl(_,_,mouse_button_left))=0:-
	guix::getWinClas(W)=41,     
	vpiTree::openCloseNode(W),
	rub::rubEW_Get(W,REW),
	vpi::winSetFocus(REW),
	vpi::winBringToTop(REW),
	!
;	
	vpiTree::openCloseNode(W),
	!.
tree_event(W,e_MouseUp(_,_,_))=0:-
	guix::getWinClas(W)=41,     
	rub::rubEW_Get(W,REW),
	vpi::winSetFocus(REW),
	vpi::winBringToTop(REW),
	!.
tree_event(W,e_Menu(id_file_print,_))=0:-
	guix::getWinClas(W)=41,     
  	S2=vpi::winGetText(W),
	datetime::tsNow(T),
	datetime::tsStr(T,S3),
	sx::sxlJoin([S2,S3]," ",S),
	vpiCommonDialogs::printSetup(9999,_,_,_,_),
	vpiTree::print(W,S),
	rub::rubEW_Get(W,REW),
	vpi::winSetFocus(REW),
	vpi::winBringToTop(REW),
	!
;	
  	S2=vpi::winGetText(W),
	datetime::tsNow(T),
	datetime::tsStr(T,S3),
	sx::sxlJoin([S2,S3]," ",S),
	vpiCommonDialogs::printSetup(9999,_,_,_,_),
	vpiTree::print(W,S),
	!.
tree_event(_,e_GetFocus())=0:-
	TW=guix::getWinOfClas(1),
	vpi::menuEnable(TW,id_dir_left,b_true),
	vpi::menuEnable(TW,id_dir_right,b_true),
	vpi::menuEnable(TW,id_dir_up,b_true),
	vpi::menuEnable(TW,id_dir_down,b_true),
	!.
tree_event(_,e_LoseFocus())=0:-
	TW=guix::getWinOfClas(1),
	vpi::menuEnable(TW,id_dir_left,b_false),
	vpi::menuEnable(TW,id_dir_right,b_false),
	vpi::menuEnable(TW,id_dir_up,b_false),
	vpi::menuEnable(TW,id_dir_down,b_false),
	!.
tree_event(W,e_CloseRequest)=0:-
	guix::getWinClas(W)=41,     
	dlgIF::dlgNote(1,"Close the rule-tree window by closing the rule-tree editor window"),
	rub::rubEW_Get(W,REW),
	vpi::winSetFocus(REW),
	vpi::winBringToTop(REW),
	!.
% ----------------------------------------------------------------------------------------------
% drwGraphComp
% ----------------------------------------------------------------------------------------------
drwGraphComp(S) :-
	sx::frontstr(4,S,F,Func),
	F="[F] ",
	TW=guix::getWinOfClas(1),
    FW=function::new(TW),
    W=FW:getWin(),
	guix::lbtSel(W,idc_syntagm,Func),	
	!
;	
	sx::frontstr(4,S,F,CR),
	F="[C] ",
	dbBase::kbID(CR,K,_,_),
	dbBase::getIrDef(K,_,Loc),
	dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
	dbBase::irType(K,_,Typ),
	inexgui::wcrUpdateRule(Typ,CR,K,Realm),
	!
;	
	sx::sxUnpack(S,'|',L),
	L=[KbId,Realm,N,M],
	Ni=sx::toInt(N),
	M="",
	Loc=dbBase::srLocId(KbId,Realm,Ni,0),
	ruleeditor::rtrTreeDisplayRec(Loc), 
	!
;	
	sx::sxUnpack(S,'|',L),
	L=[KbId,Realm,N,M],
	Ni=sx::toInt(N),
	Mi=sx::toInt(M),
	Loc=dbBase::srLocId(KbId,Realm,Ni,Mi),
	ruleeditor::rtrTreeDisplayRec(Loc), 
	!
;
	not(sx::searchchar(S,'|',_)),
	sx::sxUnpack(S,'[',L),
	L=[ST|_],
	dbBase::stNL_KB(ST,STk),
	sx::sxUnpack(STk,'|',L3),
	L3=[P,E1,E2],
	sentences::frmTabWin(2,P,E1,E2),
	!
;
	not(sx::searchchar(S,'|',_)),
	sx::sxUnpack(S,'[',L),
	L=[ST|_],
	dbBase::stNL_KB(ST,STk),
	sx::sxUnpack(STk,'|',L3),
	L3=[P,E1],
	sentences::frmTabWin(1,P,E1,""),
	!
;
	not(sx::searchchar(S,'|',_)),
	dbBase::stNL_KB(S,STk),
	sx::sxUnpack(STk,'|',L3),
	L3=[P,E1,E2],
	sentences::frmTabWin(2,P,E1,E2),
	!
;
	not(sx::searchchar(S,'|',_)),
	dbBase::stNL_KB(S,STk),
	sx::sxUnpack(STk,'|',L3),
	L3=[P,E1],
	sentences::frmTabWin(1,P,E1,""),
	!
;
	not(sx::searchchar(S,'|',_)),
	dbBase::kbID_F(S,_,Clas,_),
	Clas="p3_iPT",	
	W=guix::getWinOfClas(1),
    Obj=sentencemodel::new(W,1),
    Wd=Obj:getWin(),
    sentencemodel::displayDE(Wd,S),
	!.
% ----------------------------------------------------------------------------------------------
% drwRightClickDouble
% ----------------------------------------------------------------------------------------------
drwRightClickDouble(Win) :-
	guix::getWinClas(Win)=44,
	S=vpiTree::getSelectedNode(Win),
	dtTreeNode(S,Rule,K,_),
	dbBase::getIrDef(K,_,Loc),
	dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
	dbBase::irType(K,_,Typ),
	inexgui::wcrUpdateRule(Typ,Rule,K,Realm),
	!
;
	guix::getWinClas(Win)=45, % IR reasoning
	S=vpiTree::getSelectedNode(Win),
	sx::sxUnpackS(S,"::",L),
	L=[Pr,Sr|_],
	set::member(Pr,["CRIT","WEIGHT","OUT OF CONTEXT","IN CONTEXT"]),
	sx::sxRemoveTail(Sr,"*",Rule),
	dbBase::kbID_F(Rule,K,_,_),
	dbBase::getIrDef(K,_,Loc),
	dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
	dbBase::irType(K,_,Typ),
	inexgui::wcrUpdateRule(Typ,Rule,K,Realm),
	!
;
	guix::getWinClas(Win)=45, % IR reasoning
	S=vpiTree::getSelectedNode(Win),
	sx::sxUnpackS(S,"::",L),
	L=[Sr|_],
	sx::sxRemoveTail(Sr,"*",Rule),
	dbBase::kbID_F(Rule,K,_,_),
	dbBase::getIrDef(K,_,Loc),
	dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
	dbBase::irType(K,_,Typ),
	inexgui::wcrUpdateRule(Typ,Rule,K,Realm),
	!
;
	guix::getWinClas(Win)=41, % sentence rule: open rule component
	drwGetCurr(Win,_,_,FullText),
	rub::rubEW_Get(Win,EW),
	rub::rubNIS_Current(EW,_,T,Nis,_),
	drwRuleComp(T,Nis,FullText),
	!
;
	guix::getWinClas(Win)=50, % dep graph: open graph component
	drwGetCurr(Win,_,S,_),
	drwGraphComp(S),
	!.
drwRightClickSingle(Win) :-
	guix::getWinClas(Win)=44,
	S=vpiTree::getSelectedNode(Win),
	dtTreeNode(S,_,K,_),
  	sentencemodel::dmAnnoDisplayAsst("p3_iIR",K),
	!
;
	guix::getWinClas(Win)=45,
	S=vpiTree::getSelectedNode(Win),
	sx::sxUnpackS(S,"::",L),
	L=[Sr|_],
	sx::sxRemoveTail(Sr,"*",Rule),
	dbBase::kbID_F(Rule,K,_,_),
  	sentencemodel::dmAnnoDisplayAsst("p3_iIR",K),
	!
;
	guix::getWinClas(Win)=50,
	drwGetCurr(Win,_,S,_),
	sx::frontstr(4,S,F,Func),
	F="[F] ",
	dbBase::kbID_F(Func,K,_,_),
  	sentencemodel::dmAnnoDisplayAsst("p3_iFunc",K),
	!
;
	guix::getWinClas(Win)=50,
	drwGetCurr(Win,_,S,_),
	sx::frontstr(4,S,F,CR),
	F="[C] ",
	dbBase::kbID_F(CR,K,_,_),
  	sentencemodel::dmAnnoDisplayAsst("p3_iIR",K),
	!
;	
	guix::getWinClas(Win)=50,
	drwGetCurr(Win,_,S,_),
	sx::sxUnpack(S,'|',L),
	L=[_,_,_,_],
  	sentencemodel::dmAnnoDisplayAsst("p3_iSR",S),
	!
;
	guix::getWinClas(Win)=50,
	drwGetCurr(Win,_,S,_),
	not(sx::searchchar(S,'|',_)),
	sx::searchstring(S," [",P),
	Q=P-1,
	sx::frontstr(Q,S,ST,_),
	dbBase::stNL_KB(ST,STk),
  	sentencemodel::dmAnnoDisplayAsst("p3_iST",STk),
	!	
;
	guix::getWinClas(Win)=50,
	drwGetCurr(Win,_,S,_),
	not(sx::searchchar(S,'|',_)),
	not(sx::searchstring(S," [",_)),
	dbBase::stNL_KB(S,STk),
  	sentencemodel::dmAnnoDisplayAsst("p3_iST",STk),
	!	
;
	guix::getWinClas(Win)=50,
	drwGetCurr(Win,_,S,_),
	not(sx::searchchar(S,'|',_)),
	dbBase::kbID_F(S,K,Clas,_),
	Clas="p3_iPT",	
  	sentencemodel::dmAnnoDisplayAsst("p3_iPT",K),
	!.	
% ----------------------------------------------------------------------------------------------
% drwRuleComp
% ----------------------------------------------------------------------------------------------
drwRuleComp(T,Nis,F) :-
	sx::frontstr(4,T,Fr,Rule),
	Fr="[C] ",
	dbBase::kbID(Rule,K,_,_),
	dbBase::getIrDef(K,_,Loc),
	dbBase::getRecord(Loc,_,_,_,_,_,Realm,_,_,_,_,_),
	dbBase::irType(K,_,Typ),
	inexgui::wcrUpdateRule(Typ,Rule,K,Realm),
	!
;
	sx::frontstr(4,T,Fr,Func),
	Fr="[F] ",
	TW=guix::getWinOfClas(1),
    FW=function::new(TW),
    W=FW:getWin(),
	guix::lbtSel(W,idc_syntagm,Func),	
	!
;
	set::member(10,Nis),
	Nis=[H|_],
	H=21,
	% DE
	dbBase::kbID_F(T,_,Clas,_),
	Clas="p3_iDET",	
	% DE
	TW=guix::getWinOfClas(1),
    Obj=sentencemodel::new(TW,1),
    Wd=Obj:getWin(),
    sentencemodel::displayDE(Wd,T),
	!
;	
	Nis=[H|_],
	H=11,
	% E or P
	sx::sxUnpack(T,'[',L), L=[NL|_],
	dbBase::kbID_F(NL,_,Clas,_),
	Clas="p3_iET",	
	% E
	TW=guix::getWinOfClas(1),
    Obj=sentencemodel::new(TW,1),
    Wd=Obj:getWin(),
    sentencemodel::displayDE(Wd,NL),
	!
;
	Nis=[H|_],
	H=11,
	% E or P
	sx::sxUnpack(T,'[',L), L=[NL|_],
	dbBase::kbID_F(NL,_,Clas,_),
	Clas="p3_iPT",	
	% P
	TW=guix::getWinOfClas(1),
    Obj=sentencemodel::new(TW,1),
    Wd=Obj:getWin(),
    sentencemodel::displayDE(Wd,NL),
	!
;
	Nis=[H|_],
	H=10,
	% ST - open sentence browse window
	sx::sxUnpack(F,'{',L2),
	L2=[STn,_],
	dbBase::stNL_KB(STn,STk),
	sx::sxUnpack(STk,'|',L3),
	L3=[P,E1,E2],
	sentences::frmTabWin(2,P,E1,E2),
	!
;	
	Nis=[H|_],
	H=10,
	% ST - open sentence browse window
	sx::sxUnpack(F,'{',L2),
	L2=[STn,_],
	dbBase::stNL_KB(STn,STk),
	sx::sxUnpack(STk,'|',L3),
	L3=[P,E1],
	sentences::frmTabWin(1,P,E1,""),
	!.
% ----------------------------------------------------------------------------------------------
% setSize
% ----------------------------------------------------------------------------------------------
setSize([],0) :-
    !.
setSize([_|T],N) :-
    setSize(T,M),
    N=M+1,
    !.
% ----------------------------------------------------------------------------------------------
% dtTreeNode
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	CR tree-node string
% 2	out	rule (NL)
% 3	out	rule (KB)
% 4	out	node number
%
dtTreeNode(S,RuleN,RuleK,No) :-
	% prefix::rule::suffix{no
	sx::sxUnpack(S,'{',L),
	L=[S2,No],
	sx::searchstring(S2,"::",P),
	P2=P+1,
	sx::frontstr(P2,S2,_,S3),
	sx::searchstring(S3,"::",P3),
	P4=P3-1,
	sx::frontstr(P4,S3,Rule,_),
	sx::sxRemoveTail(Rule,"*",RuleN),
	dbBase::kbID_F(Rule,RuleK,_,_),
	!	
;
	% prefix::rule{no
	sx::sxUnpack(S,'{',L),
	L=[S2,No],
	sx::searchstring(S2,"::",P),
	P2=P+1,
	sx::frontstr(P2,S2,_,Rule),
	sx::sxRemoveTail(Rule,"*",RuleN),
	dbBase::kbID_F(RuleN,RuleK,_,_),
	!	
;
	% rule :suffix{no
	sx::sxUnpack(S,'{',L),
	L=[S2,No],
	sx::searchstring(S2,"::",P),
	P2=P-1,
	sx::frontstr(P2,S2,Rule,_),
	sx::sxRemoveTail(Rule,"*",RuleN),
	dbBase::kbID_F(RuleN,RuleK,_,_),
	!	
;
	% rule{no
	sx::sxUnpack(S,'{',L),
	L=[Rule,No],
	sx::sxRemoveTail(Rule,"*",RuleN),
	dbBase::kbID_F(RuleN,RuleK,_,_),
	!	
;	
	dlgIF::feil("dtTreeNode",["Unrecognizable inexact rule tree node",S],[],nill),!,fail.	
% ----------------------------------------------------------------------------------------------
% dtTransTree
% ----------------------------------------------------------------------------------------------
class predicates
dtTransVal : (string,string) determ (i,o).
dtTransVals : (string_list,string_list) determ (i,o).
dtTransTrees : (integer,vpiTree::tree_list,vpiTree::tree_list) determ (i,i,o).
dtAdjust : (string,string_list,string) determ (i,i,o).
dtAdj : (string,string) procedure (i,o).
dtTransPars : (sterml,sterml,string_list) determ (i,i,o).
clauses

dtTransTree(NL,vpiTree::tree(S,U,L,N),vpiTree::tree(Name,U,Lt,N)) :-
	NL=0,
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[Rule,Val,MapFrom,atom(Dom),termDomain::str(Tail),_,_,termDomain::str(Prefix),termDomain::str(Suffix),int(N)]),
	nl::conValX(Rule,"p3_dany","",RuleS),
	sx::sxUnquote(RuleS,RuleU),
	nl::conValX(Val,"p3_dany","",ValS),
	nl::conValX(MapFrom,"p3_dany","",MapS),
	dtAdjust(Dom,[Suffix,ValS,MapS],S5),
	Ns=toString(N),
	sx::concat([Prefix,RuleU,S5,Tail,"{",Ns],Name2),
    sx::sxSubstList(Name2,["::*","::::","([])"],["*","::",""],Name),
	dtTransTrees(NL,L,Lt),
	!
;
	NL=1,
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[Rule,Val,MapFrom,atom(Dom),termDomain::str(Tail),_,AParsL,termDomain::str(Prefix),termDomain::str(Suffix),int(N)]),
    termDomain::listBR(AParsL,L2),
    set::size(L2,L2N),
    L2N=0,
	nl::conValX(Rule,"p3_dany","",RuleS),
	sx::sxUnquote(RuleS,RuleU),
	nl::conValX(Val,"p3_dany","",ValS),
	nl::conValX(MapFrom,"p3_dany","",MapS),
	dtAdjust(Dom,[Suffix,ValS,MapS],S5),
	Ns=toString(N),
	sx::concat([Prefix,RuleU,S5,Tail,"{",Ns],Name2),
    sx::sxSubstList(Name2,["::*","::::","([])"],["*","::",""],Name),
	dtTransTrees(NL,L,Lt),
	!
;
	NL=1,
	Term=termDomain::fromTermStringIndirect(S),
	Term=cmp(_,[Rule,Val,MapFrom,atom(Dom),termDomain::str(Tail),ParsL,AParsL,termDomain::str(Prefix),termDomain::str(Suffix),int(N)]),
    termDomain::listBR(AParsL,L2),
    set::size(L2,L2N),
    L2N>0,
    termDomain::listBR(ParsL,L3),
    dtTransPars(L3,L2,L4),
    dbBase::joinNL(L4,Sp),
	nl::conValX(Rule,"p3_dany","",RuleS),
	sx::sxUnquote(RuleS,RuleU),
	nl::conValX(Val,"p3_dany","",ValS),
	nl::conValX(MapFrom,"p3_dany","",MapS),
	dtAdjust(Dom,[Suffix,ValS,MapS],S5),
	dbBase::nlSing("p3_iGivn",Given),
	Ns=toString(N),
	sx::concat([Prefix,RuleU,S5," ",Given," ",Sp," ",Tail,"{",Ns],Name2),
    sx::sxSubstList(Name2,["::*","::::","([])"],["*","::",""],Name),
	dtTransTrees(NL,L,Lt),
	!
;
	dlgIF::feil("dtTransTree",[S],[NL],nill),!,fail.

dtTransTrees(_,[],[]) :-
	!.
dtTransTrees(NL,[H|T],[Ht|Tt]) :-
	dtTransTree(NL,H,Ht),
	dtTransTrees(NL,T,Tt),
	!.

dtTransVals([],[]) :-
	!.
dtTransVals(["str(\"\")"|T],Tt) :-
	dtTransVals(T,Tt),
	!.
dtTransVals([H|T],[Hu|Tt]) :-
	T2=toTerm(H),
	nl::conValX(T2,"p3_dany","",Ht),
	sx::sxUnquote(Ht,Hu),
	dtTransVals(T,Tt),
	!.
	
dtTransVal(H,Hu) :-
	T2=toTerm(H),
	nl::conValX(T2,"p3_dany","",Ht),
	sx::sxUnquote(Ht,Hu),
	!.
dtAdjust(Dom,[Sf,V,M],S) :-
	V="",
	S="",
	!
;
	M=V,
	dtAdjust(Dom,[Sf,V,""],S),
	!
;
	Dom="p3_dcf",
	M="",
	dtAdj(V,Va),
	sx::concat([Sf,"::c=",Va],S),
	!
;	
	Dom="p3_dcf",
	dtAdj(V,Va),
	sx::concat([Sf,"::c=",Va,"(",M,")"],S),
	!
;
	Dom="p3_dmg",
	M="",
	dtAdj(V,Va),
	sx::concat([Sf,"::m=",Va],S),
	!
;	
	Dom="p3_dmg",
	dtAdj(V,Va),
	sx::concat([Sf,"::m=",Va,"(",M,")"],S),
	!
;
	Dom="p3_dpr",
	M="",
	dtAdj(V,Va),
	sx::concat([Sf,"::p=",Va],S),
	!
;	
	Dom="p3_dpr",
	dtAdj(V,Va),
	sx::concat([Sf,"::p=",Va,"(",M,")"],S),
	!
;
	M="",
	sx::concat([Sf,"::",V],S),
	!
;	
	sx::concat([Sf,"::",V,"(",M,")"],S),
	!.
dtAdj(V,Va) :-
	V="-1.000",
	Va="-1.0",
	!
;
	V="1.000",
	Va="1.0",
	!
;
	V="0.000",
	Va="0.0",
	!.
dtAdj(V,Va) :-
	sx::frontchar(V,'-',_),
	L=string::length(V),
	L>5,
	sx::frontstr(5,V,Va,_),
	!
;
	not(sx::frontchar(V,'-',_)),
	L=string::length(V),
	L>4,
	sx::frontstr(4,V,Va,_),
	!
;
	Va=V,
	!.
		
dtTransPars([],[],[]) :-
	!.
dtTransPars([P|Ps],[AP|APs],[S|L]) :-
	P=fstr(D,Dom,F,_),
	nl::conValX(AP,Dom,F,S2),
	dbBase::nlSing(D,S3),
	sx::concat([S3," ",S2],S),
	dtTransPars(Ps,APs,L),
	!
;
	dlgIF::feil("dtTransPars",[],[],cmp("",[P,AP])),!,fail.
end implement p3Tree
