/******************************************************************************

                 Copyright (C) 1999-2003, Prolog Development Center A/S

 FileName:      TREEVIEW.PRO
 Purpose:       Tree View custom control object
 Written by:    Sergey Alexeev
 Version:       1.0
 Build:         002.000316
 Tested with:   Visual Prolog v.5.0 build 533

    @short 

    @detail 

    @package vpi/treeView
    @overview 
    @author PDC
    @version Mimas
    @end  
******************************************************************************/

class p3TreeView
    domains
        itemId = unsigned.

    constants
        wcc_null : itemId  = 0.% it was used also as ::unsigned
    
    constants
        treeviewName : string = "treeView".
    
    
    constants % for Tree View control Flags
        tvs_hasbuttons = 0x0001.
        % Displays plus (+) and minus (-) buttons next to parent items.
        % The user clicks the buttons to expand or collapse a parent item's list of child items.
        % To include buttons with items at the root of the tree view, TVS_LINESATROOT
        % must also be specified.
    
        tvs_haslines = 0x0002.
        % Uses lines to show the hierarchy of items.
    
        tvs_linesatroot = 0x0004.
        % Uses lines to link items at the root of the tree-view control.
        % This value is ignored if TVS_HASLINES is not also specified.
    
        tvs_editlabels = 0x0008.
        % Allows the user to edit the labels of tree-view items.
    
        tvs_disabledragdrop = 0x0010.
        % Prevents the tree-view control from sending TVN_BEGINDRAG notification messages.
    
        tvs_showselalways = 0x0020.
        % Causes a selected item to remain selected when the tree-view
    
    predicates
        moduleInfo : core::classInfo.

    % Register Class
    predicates
	register : ()  procedure ().
	
    domains
        tree_list = tree*.
        tree = tree(
            itemId ItemId,
            itemState StateList,
            vpiDomains::bitmapId Bitmap,
            vpiDomains::bitmapId SelectedBitmap,
            string Text,
            tree_list TreeList).

    domains
        itemState = itemStateFlag*.
        itemStateFlag =
            selected();
            % The item is selected.
            % The appearance of a selected item depends on whether it has the focus and on whether
            % the system colors are used for selection.
            cut();
            % The item is selected as part of a cut and paste operation.
            dropHilited();
            % The item is selected as a drag and drop target.
            bold();
            % The item is bold.
            expanded().
            % The item's list of child items is currently expanded; that is, the child items are visible.
            % This value applies only to parent items.

    domains
        message =
            outofmemory;
            setfocus;
            losefocus;
            return(itemId ItemId);
            click(vpiDomains::pnt Point);
            dblclick(vpiDomains::pnt Point);
            rclick(vpiDomains::pnt Point);
            rdblclick(vpiDomains::pnt Point);
            keydown(integer Key);
            selecting(itemId ItemIdUnselect, itemId ItemIdSelect);
            selected(itemId ItemIdUnselect, itemId ItemIdSelect);
            expanding(itemId ItemId, core::booleanInt CurrentStatus);
            expanded(itemId ItemId, core::booleanInt NewStatus);
            edit(itemId ItemId);
            endEdit(itemId ItemId, string NewText);
            endEditCanceled(itemId ItemId);
            none.

    domains
        callback = (vpiDomains::windowHandle TreeWindow, message What)determ  (i,i).


    % initialization  of Tree View Control
    predicates
        init : (vpiDomains::windowHandle TreeWindow,
            core::integer_list Flags,
            vpiDomains::bitmapId BitMap_UsedToSetBitmapSize,
            callback CallBack)
           procedure (i,i,i,i).

    domains
        insertPlace =
            first();
            last();
            sorted();
            after(itemId PreviousId).
        % Regarding sorted()
        % using sorted() means that the item, tree or tree_list will be inserted sorted()
        % the sorting is not automatcally maintained:
        % Subsequent inserts must also be performed sorted, just like
        % resorting is necessary after label change.

    % inserts a new item in a tree-view control.
    predicates
        insert_item : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId,
            itemId ParentId,
            insertPlace InsertPlace,
            string Text,
            itemState StateList,
            vpiDomains::bitmapId Bitmap,
            vpiDomains::bitmapId SelectedBitmap)
         procedure (i,i,i,i,i,i,i,i).

    predicates
        insert_tree_list : (
            vpiDomains::windowHandle TreeWindow,
            itemId ParentId,
            insertPlace InsertPlace,
            tree_list TreeList)
         procedure (i,i,i,i).

    predicates
        insert_tree : (
            vpiDomains::windowHandle TreeWindow,
            itemId ParentId,
            insertPlace InsertPlace,
            tree Tree)
         procedure (i,i,i,i).

    % removes an item from a tree-view control
    predicates
        delete_item : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
        procedure (i,i).

    % sorts the child items of the specified parent
    % item in a tree-view control
    predicates
        sort_children : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
         procedure (i,i).

    domains
        position =
            nowhere();
            onItemIcon();
            onItemLabel();
            onItemIndent();
            onItemButton();
            onItemRight();
            onItemStateIcon();
            above();
            below();
            toRight();
            toLeft().

    predicates
        get_position : (
            vpiDomains::windowHandle TreeWindow,
            integer X,
            integer Y,
            itemId ItemId,
            position Position)
         procedure (i,i,i,o,o).

    % Retrieves the currently selected
    % item in a tree view control
    predicates
        get_selected : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
         procedure (i,o).

    % retrieves a tree-view items attributes
    predicates
        get_item : (
              vpiDomains::windowHandle TreeWindow,
              itemId ItemId,
              string Text,
              itemState StateList,
              vpiDomains::bitmapId Bitmap,
              vpiDomains::bitmapId SelectedBitmap)
         procedure (i,i,o,o,o,o).

    % Retrieves the parent item of the specified
    % tree view item Returns the handle to the
    % item if successful, or NULL otherwise.
    predicates
        get_parent : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId,
            itemId ParentId)
         procedure (i,i,o).

    % Retrieves the first child item
    % of the specified tree view item
    predicates
        get_child : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId,
            itemId ChildId)
         procedure (i,i,o).

    % Retrieves the next sibling item of a specified
    % item in a tree view control
    predicates
        get_next : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId,
            itemId NextId)
         procedure (i,i,o).

    % Retrieves the previous sibling item of a
    % specified item in a tree view control
    predicates
        get_previous : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId,
            itemId PreviousId)
         procedure (i,i,o).

    % retrives whole tree
    predicates
        get_all : (
            vpiDomains::windowHandle TreeWindow,
            tree_list)
         procedure (i,o).

    % Retrives part of tree, begin from pointed item
    predicates
        get_part : (
            vpiDomains::windowHandle TreeWindow,
            itemId FromItemId,
            tree_list)
         procedure (i,i,o).

    % retrieves a count of the items in a tree-view control
    predicates
        get_count : (
            vpiDomains::windowHandle TreeWindow,
            unsigned Count)
          procedure (i,o).

    % expands or collapses the list of child items
    % associated with the specified parent item
    predicates
        expand : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
         procedure (i,i).

    % selects the specified tree-view item
    predicates
        select : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
         procedure (i,i).

    % removes the selection
    predicates
        unselect : (
            vpiDomains::windowHandle TreeWindow)
         procedure (i).

    % Change Item properties
    predicates
        set_item : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId,
            string Text,
            itemState StateList,
            vpiDomains::bitmapId Bitmap,
            vpiDomains::bitmapId SelectedBitmap)
         procedure (i,i,i,i,i,i).

    % begins in-place editing of the specified items text,
    % replacing the text of the item with a single-line edit
    % control containing the text.
    predicates
        edit_begin : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
         procedure (i,i).

    % Ends the editing of a tree view item's label
    predicates
        edit_end : (
            vpiDomains::windowHandle TreeWindow)
         procedure (i).

    % ensures that a tree-view item is visible, expanding the parent
    % item or scrolling the tree-view control, if necessary.
    predicates
        ensure_visible : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
         procedure (i,i).

    % make ItemId the first visible item (if possible)
    predicates
        firstvisible : (
            vpiDomains::windowHandle TreeWindow,
            itemId ItemId)
         procedure (i,i).

    % retrieves the amount, in pixels,
    % that child items are indented relative
    % to their parent items
    predicates
        get_indent : (
            vpiDomains::windowHandle TreeWindow,
            unsigned Indent)
         procedure (i,o).

    % sets the amount, in pixels,
    % that child items are indented relative
    % to their parent items
    predicates
        set_indent : (
            vpiDomains::windowHandle TreeWindow,
            unsigned Indent)
          procedure (i,i).

    % gets installed subclass handler
    predicates
        tryGetSubclassHandler : (
            vpiDomains::windowHandle TreeWindow,
            vpiDomains::ehandler  Callback)
          determ (i,o).

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Native control (use with care)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Subclass handle parent to tree-control
    predicates
        set_subclassHandler : (
            vpiDomains::windowHandle TreeWindow,
            vpiDomains::ehandler SubclassHandler)
      procedure (i,i).

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Native control (use with care)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Send native event to tree control (see MSDN)
    predicates
         /* Result */ send_native : (
            vpiDomains::windowHandle TreeWindow,
            unsigned Message,
            unsigned Wparam,
            core::integer32 Lparam) -> core::integer32 
         procedure (i,i,i,i).

    predicates
        itemIdExist : (vpiDomains::windowHandle TreeWindow,itemId ItemID) determ(i,i).

    predicates
        badIdentifierException : core::exception.

end class p3TreeView
