/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement p3Rep
open core,termDomain,globalDomains

constants
className = "com/rudstrand/p3rep/p3Rep".
classVersion = "1.0".

class facts
flag : integer := 0. 

class predicates
repAnnotation : (rep,integer,integer,string,string).
headerContext : (rep) determ.
writeSorted : (rep,integer,lst) determ.

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------
% repNonDerived
% -------------------------------------------------------------------------------------
class predicates
repND : (lst).
repNDW : (rep,string_list) determ.
clauses

repNonDerived() :-
    Rep=rep::newReport("Sentence Rules Where No Sentences Are Derived",70),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	List=lst::new(),
	repND(List),
	L=List:getSorted(),
	repNDW(Rep,L),
	Rep:close(),
	!.
	
repND(List) :-
    der::genDerivationGraph,
    dep::get_nd("p3_pDR",T1,N1,T2,N2),
        T1=int(3),
        T2=int(2),
        N1=str(Rule),
        N2=str(St),
        dbBase::stCount(St,N,_),
        N=0,
        dbBase::nlSing(St,Stn),
        sx::sxUnpack(Rule,'|',L),
        L=[_|T],
        sx::sxlJoin(T,"|",Rule2),
        S=string::concat(Stn," (",Rule2,")"),
        List:add(S),
        fail
;
    !.
    
repNDW(_,[]) :-
    !.
repNDW(Rep,[H|T]) :-
    Rep:repWL(0,H),
    repNDW(Rep,T),
    !.
% -------------------------------------------------------------------------------------
% repDeps                                                             
% -------------------------------------------------------------------------------------
class predicates
repDepY : (rep,string_list) determ.
clauses

repDeps(Node,Deps) :-
    Title=string::concat("Dependents of ",Node),
    Rep=rep::newReport(Title,70),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	headerContext(Rep),
	repDepY(Rep,Deps),
	Rep:close(),
	!.
	
repDepY(_,[]) :- 
	!.
repDepY(Rep,[Dep|Deps]) :-
	Rep:repWL(0,Dep),
	repDepY(Rep,Deps),
	!.
% -------------------------------------------------------------------------------------
% integrityProblems                                             
% -------------------------------------------------------------------------------------
class predicates
missingFuncDefs : (rep) determ.
missingFuncDefsX : (rep,string_list) determ.
missingIR_Decs : (rep).
missingIR_Defs : (rep) determ.
missingIR_DefsX : (rep,string_list) determ.
clauses

integrityProblems() :-
    Rep=rep::newReport("Referential Integrity Problems",80),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	header(Rep),
	flag:=1, 
    Rep:repWL(0,"\nMISSING FUNCTION DEFINITIONS:\n"),
    missingFuncDefs(Rep),
    set::setGet(["None.",""],flag,Note),
    Rep:repWL(0,Note),
	flag:=1, 
    Rep:repWL(0,"\nMISSING INEXACT RULE DEFINITIONS:\n"),
    missingIR_Defs(Rep),
    set::setGet(["None.",""],flag,Note2),
    Rep:repWL(0,Note2),
	flag:=1, 
    Rep:repWL(0,"\nMISSING INEXACT RULE DECLARATIONS:\n"),
    missingIR_Decs(Rep),
    set::setGet(["None.",""],flag,Note3),
    Rep:repWL(0,Note3),
	Rep:repWL(0,""),
	Rep:repWL(0,"Note:"),
	Rep:repWL(0,"Missing function definitions in the Pro/3-segment are not reported."),
	Rep:close(),
	!.

missingFuncDefs(Rep) :-
    F1=pop::getSM("f",[]),
    F2=pop::getSM("g",[]),
    F3=pop::getSM("f",["Pro/3"]),
    F4=pop::getSM("g",["Pro/3"]),
    set::conc([F1,F2],F10),
    set::conc([F3,F4],F20),
    set::setRemSub(F20,F10,Fs),
    missingFuncDefsX(Rep,Fs),
    !.

missingFuncDefsX(_,[]) :-
    !.
missingFuncDefsX(Rep,[H|T]) :-
    dbBase::chainCount(H)>0,
    missingFuncDefsX(Rep,T),
    !
;
    dbBase::nlTry(H,Hn),
    Mess=string::concat("Function ",H," (",Hn,")"),
    Rep:repWL(0,Mess),
    flag:=2,
    missingFuncDefsX(Rep,T),
    !.
    
missingIR_Defs(Rep) :-
    IRs=pop::getSM("h",[]),
    missingIR_DefsX(Rep,IRs),
    !.

missingIR_DefsX(_,[]) :-
    !.
missingIR_DefsX(Rep,[H|T]) :-
    dbBase::getIrDefF(H,_,_),
    !
;
    dbBase::nlTry(H,Hn),
    Mess=string::concat("Inexact rule ",H," (",Hn,")"),
    Rep:repWL(0,Mess),
    flag:=2,
    missingIR_DefsX(Rep,T),
    !.
    
missingIR_Decs(Rep) :-
    dbBase::irChain_nd(Chain),
            dbBase::chainTerms(Chain,LT,RT),
                    RT=nill,
                    LT=cmp(_,[atom(Rule)|_]),
                    not(dbBase::smClassDefF(Rule,"h",_,_,_,_,_)),
                    dbBase::nlTry(Rule,RuleN),
                    Mess=string::concat("Inexact rule ",Rule," (",RuleN,")"),
                    Rep:repWL(0,Mess),
                    flag:=2,
                    fail
;
    !.
% -------------------------------------------------------------------------------------
% unrefKnowledge                                                     
% -------------------------------------------------------------------------------------
class predicates
nodesToNL : (sterml,string_list) determ (i,o).
unrefK : (rep,string_list) determ.
filterNodes : (sterml,sterml) determ (i,o).
nodeChain : (integer,string,string) determ (i,i,o).    
clauses

% @@ add input STs from smGraph not existing in dpGraph
%
unrefKnowledge() :-
    Rep=rep::newReport("Unreferenced Knowledge",80),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	header(Rep),
	dep::endNodes("p3_pDP",L),
	filterNodes(L,Lf),
	nodesToNL(Lf,Ln),
    sx::sxlSort(Ln,LnSorted),
	unrefK(Rep,LnSorted),
	Rep:repWL(0,""),
	Rep:repWL(0,"Note:"),
	Rep:repWL(0,"Unreferenced Pro/3 library functions are not included."),
	Rep:repWL(0,"* In public chain and likely referenced in other KBs."),
	Rep:close(),
	!.

filterNodes([],[]) :-
    !.
filterNodes([cmp(H,[int(Type),str(Name)])|Nodes],Ns) :-
    % Pro/3 segment functions 
    Type=19,
    dbBase::smDefF(Name,Clas,_,_,_,_,_),
    set::member(Clas,["p","g","f","h"]),
    Segment=dbBase::smSegment(Name),
    Segment="Pro/3",
    filterNodes(Nodes,Ns),
    !
;
    Type=1,
    dep::hasSTofSG(Name),
    filterNodes(Nodes,Ns),
    !
;
    set::member(Type,[2,5]),
    sx::sxUnpack(Name,'|',L),
    L=[SG|_],
    dep::isDeterminant("p3_pDP",1,SG),
    filterNodes(Nodes,Ns),
    !
;
    nodeChain(Type,Name,Chain),
    dbBase::publicChain(Chain,_),
    NameX=string::concat(Name,"*"),
    filterNodes(Nodes,Ns2),
    Ns=[cmp(H,[int(Type),str(NameX)])|Ns2],
    !
;
    filterNodes(Nodes,Ns2),
    Ns=[cmp(H,[int(Type),str(Name)])|Ns2],
    !.
    
nodesToNL([],[]) :-
       !.
nodesToNL([cmp(_,[int(Type),str(Name)])|Nodes],[NL|NLs]) :-
    Type=1,
    sx::sxRemoveTail(Name,"*",NameX),
    dbBase::nlTry(NameX,NL_Name),
    NL=string::concat("SENTENCE_GROUP            ",NL_Name),
    nodesToNL(Nodes,NLs),
    !
;    
    dep::nodIF_NoNL(NL_Type,Type),
    sx::sxR_Pad(NL_Type," ",26,NL_TypeP),
    sx::sxRemoveTail(Name,"*",NameX),
    dbBase::nlTry(NameX,NL_Name),
    NL=string::concat(NL_TypeP,NL_Name),
    nodesToNL(Nodes,NLs),
    !.
    
unrefK(_,[]) :-
    !.
unrefK(Rep,[H|T]) :-
	Rep:repWL(0,H),
    unrefK(Rep,T),
    !.

nodeChain(Type,Name,Chain) :-
    set::member(Type,[1,2,3,5]),
    sx::sxUnpack(Name,'|',L),
    L=[Chain|_],
    !
;
    Type=19,
    Chain=Name,
    !
;
    Type=6,
    dbExtension::dbuCR_GetTD(Name,Chain,_),
    !.
% -------------------------------------------------------------------------------------
% unrefStrucKnowledge                                                     
% -------------------------------------------------------------------------------------
class predicates
unrefNodes : (termSet,lst).
unrefStrucK : (rep,string_list) determ.
clauses

unrefStrucKnowledge() :-
    Rep=rep::newReport("Unreferenced Structure Objects",80),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	headerContext(Rep),
	Set=pop::getSM_All(),
	List=lst::new(),
	unrefNodes(Set,List),
	L=List:get(),
    sx::sxlSort(L,LSorted),
	Rep:repWL(0,""),
    Rep:repWL(0,"SEGMENT         OBJECT TYPE               NAME"),
	Rep:repWL(0,""),
	unrefStrucK(Rep,LSorted),
	Rep:repWL(0,""),
	Rep:repWL(0,"Note:"),
	Rep:repWL(0,"- Non Pro/3 functions are reported as unreferenced if no definition exists in"),
	Rep:repWL(0,"   the database (even if it possibly exists in another database."),
	Rep:repWL(0,"   (Pro/3 functions are not reported as unreferenced)."),
	Rep:repWL(0,"- Entity types and predicate types are reported as unreferenced if not used"),
	Rep:repWL(0,"   in a sentence or sentence rule."),
	Rep:close(),
	!.

unrefNodes(Set,List) :-
    T=Set:get_nd(),
        T=cmp(_,[str(Type),str(Name),str(Segment)]),
        Type="a",
        not(set::member(Name,["p3_anarg","p3_anser","p3_anser","p3_aext","p3_aips","p3_aipe","p3_aipi","p3_anserxd","p3_aext","p3_aipsxd","p3_aipexd",
                                                "p3_aipixd","p3_aOrdr","p3_aSKNo","p3_aNagF","p3_aNagS","p3_aaarg","p3_adfv"])), % procedure parameter DEs
        No=13,
        not(dep::isDeterminant("p3_pSM",No,Name)),
        dbBase::nlSing(Type,NL_Type),
        sx::sxTrunc(NL_Type,25,NL_Type2),
        sx::sxR_Pad(NL_Type2," ",26,NL_TypeP),
        sx::sxTrunc(Segment,15,Segment2),
        sx::sxR_Pad(Segment2," ",16,SegmentP),
        dbBase::nlSing(Name,NL_Name),
        NL=string::concat(SegmentP,NL_TypeP,NL_Name),
        List:add(NL),
        fail
;
    T=Set:get_nd(),
        T=cmp(_,[str(Type),str(Name),str(Segment)]),
        Type="p",
        No=1,
        not(dep::isDeterminant("p3_pSM",No,Name)),
        not(dep::hasNode("p3_pDP",1,Name)),
        not(dep::hasSTofSG(Name)),
        dbBase::nlSing(Type,NL_Type),
        sx::sxTrunc(NL_Type,25,NL_Type2),
        sx::sxR_Pad(NL_Type2," ",26,NL_TypeP),
        sx::sxTrunc(Segment,15,Segment2),
        sx::sxR_Pad(Segment2," ",16,SegmentP),
        dbBase::nlSing(Name,NL_Name),
        NL=string::concat(SegmentP,NL_TypeP,NL_Name),
        List:add(NL),
        fail
;
    T=Set:get_nd(),
        T=cmp(_,[str(Type),str(Name),str(Segment)]),
        Type="e",
        No=14,
        not(dep::isDeterminant("p3_pSM",No,Name)),
        not(dep::hasET(Name)),
        dbBase::nlSing(Type,NL_Type),
        sx::sxTrunc(NL_Type,25,NL_Type2),
        sx::sxR_Pad(NL_Type2," ",26,NL_TypeP),
        sx::sxTrunc(Segment,15,Segment2),
        sx::sxR_Pad(Segment2," ",16,SegmentP),
        dbBase::nlSing(Name,NL_Name),
        NL=string::concat(SegmentP,NL_TypeP,NL_Name),
        List:add(NL),
        fail
;
    T=Set:get_nd(),
        T=cmp(_,[str(Type),str(Name),str(Segment)]),
        Segment<>"Pro/3",
        set::member(Type,["f","g"]),
        dbBase::chainCount(Name)=0,
        sx::sxTrunc(Segment,15,Segment2),
        sx::sxR_Pad(Segment2," ",16,SegmentP),
        dbBase::nlSing(Name,NL_Name),
        NL=string::concat(SegmentP,"function declaration      ",NL_Name),
        List:add(NL),
        fail
;
    !.
    
unrefStruck(_,[]) :-
    !.
unrefStrucK(Rep,[H|T]) :-
    Rep:repWL(0,H),
    unrefStruck(Rep,T),
    !.
% -------------------------------------------------------------------------------------
% repSent                                                              
% -------------------------------------------------------------------------------------
class predicates
repGetSTs : (string,string_list,string_list) determ (i,i,o).
repSample : (rep,integer,string) determ.
repDetSts : (rep,integer,string,dirgraph) determ.
repSGs : (rep,string_list,string_list) determ (i,i,o).
repSentSGs : (rep,integer,integer,integer,string_list,string_list,integer,integer,integer,integer,dirgraph) determ.
repSentSGsX : (rep,integer,integer,string,string,integer,integer,string,string) determ.	
repSentX : (rep,integer,integer,integer,integer,integer,integer,integer,dirgraph) determ.
repSG_Rules : (rep,integer,integer,string) determ.
repST_Rules : (rep,integer,string) determ.
repSentChainType : (integer,string,string) determ (i,i,o).
genGraph : (integer,dirgraph) determ (i,o).
repDerStat : (integer,rep,string).
clauses

% parameters:
% 1	in	-1/1	incl. rules
% 2	in	-1/1	show sentence groups
% 3	in	-1/1	show kb names
% 4	in	-1/1	show sample sentence
% 5 in  -1/1    show determinant STs
% 6 in  -1/1    show annotations
% 7 in  -1/1    show derivation data
%
repSent(Report,Groups,KBn,Sample,Dets,Anno,Derive) :-
    Kb=dbBase::dbName(),
    File=string::concat(Kb,"_Sentences"),
    Rep=rep::newReport(File,99),
    not(Rep:cancelled),
	vpi::cursorSetWait(),
	headerContext(Rep),
	genGraph(Dets,DG),
	repSentX(Rep,Report,Groups,KBn,Sample,Dets,Anno,Derive,DG),
	Rep:close(),
	!.
	
genGraph(Dets,DG) :-
    Dets=1,
    der::genDerivationGraph(),
  	dep::copy("p3_pDR",DG),
  	DG:subMergeTypeP(int(3)),
  	!
;
  	dep::copy("p3_pDR",DG),
  	!.
       
repSentX(Rep,Report,1,KBn,Sample,Dets,Anno,Derive,DG) :-	
	Context=dbBase::getContext(),
	% all SGs based on rules or actual sentences
	STs=pop::getSTs(Context),
	repSGs(Rep,STs,SG2),
	set::distinct(SG2,SG3),
	sx::sxlSort(SG3,SGs),
	%
	sx::sxR_Pad("1. EXTERNAL SENTENCE GROUPS"," ",84,Ha),
	Haa=string::concat(Ha,"INPUT   DERIVED"),
	Rep:repWL(0,""),
	Rep:repWL(0,Haa),
	repSentSGs(Rep,KBn,2,Report,SGs,STs,Sample,Dets,Anno,Derive,DG),
	%
	sx::sxR_Pad("2. INTERNAL SENTENCE GROUPS"," ",84,Hb),
	Hbb=string::concat(Hb,"INPUT   DERIVED"),
	Rep:repWL(0,""),
	Rep:repWL(0,Hbb),
	repSentSGs(Rep,KBn,0,Report,SGs,STs,Sample,Dets,Anno,Derive,DG),
	%
	sx::sxR_Pad("3. PUBLIC SENTENCE GROUPS"," ",84,Hc),
	Hcc=string::concat(Hc,"INPUT   DERIVED"),
	Rep:repWL(0,""),
	Rep:repWL(0,Hcc),
	repSentSGs(Rep,KBn,1,Report,SGs,STs,Sample,Dets,Anno,Derive,DG),
	!.	
repSentX(Rep,Report,-1,KBn,Sample,Dets,Anno,Derive,DG) :-	
	Context=dbBase::getContext(),
	STs=pop::getSTs(Context),
	dbBase::nlSings(STs,STsn),
	sx::sxL_Pad("INPUT   DERIVED"," ",99,Haa),
	Rep:repWL(0,""),
	Rep:repWL(0,Haa),
	sx::sxlSort(STsn,STsns),
	repSentSTs(Rep,KBn,Report,STsns,Sample,1,Dets,Anno,Derive,DG),
	!.	

repSentSGs(_,_,_,_,[],_,_,_,_,_,_) :- 
	!.
repSentSGs(Rep,KBn,Group,Report,[SGnl|SGs],STs,Sample,Dets,Anno,Derive,DG) :-
	dbBase::kbID(SGnl,SG,_,_),
	repSentChainType(Group,SG,ChainType),
	dbExtension::kbSG_Count(SG,N1,N2),
	Rep:repWL(0,""),
	repSentSGsX(Rep,Anno,KBn,SGnl,SG,N2,N1,"p3_iPT",ChainType),
	repSG_Rules(Rep,Report,Anno,SG),
	Rep:repLine('-'),
	repGetSTs(SG,STs,STsn),
	sx::sxlSort(STsn,STsns),
	repSentSTs(Rep,KBn,Report,STsns,Sample,1,Dets,Anno,Derive,DG),
	repSentSGs(Rep,KBn,Group,Report,SGs,STs,Sample,Dets,Anno,Derive,DG),
	!
;	
	repSentSGs(Rep,KBn,Group,Report,SGs,STs,Sample,Dets,Anno,Derive,DG),
	!.

repSentSGsX(Rep,Anno,1,SGnl,SG,N1,N2,AnnId,Type) :-
    Anno=1,
    N1<0,
	sx::concat([SGnl," [",SG,"] ",Type],S2),
	sx::sxTrunc(S2,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	Rep:repWL(0,S4),
	repAnnotation(Rep,5,79,AnnId,SG),
	!
;    
    Anno=1,
	sx::concat([SGnl," [",SG,"] ",Type],S2),
	sx::sxTrunc(S2,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	sx::fixFormatI(N1,0,1,N1s1),
	sx::sxL_Pad(N1s1," ",10,N1s),
	sx::fixFormatI(N2,0,1,N2s1),
	sx::sxL_Pad(N2s1," ",10,N2s),
	Rep:repWL(0,S4,N1s,N2s),
	repAnnotation(Rep,5,79,AnnId,SG),
	!
;	
    Anno=-1,
    N1<0,
	sx::concat([SGnl," [",SG,"] ",Type],S2),
	sx::sxTrunc(S2,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	Rep:repWL(0,S4),
	!
;    
    Anno=-1,
	sx::concat([SGnl," [",SG,"] ",Type],S2),
	sx::sxTrunc(S2,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	sx::fixFormatI(N1,0,1,N1s1),
	sx::sxL_Pad(N1s1," ",10,N1s),
	sx::fixFormatI(N2,0,1,N2s1),
	sx::sxL_Pad(N2s1," ",10,N2s),
	Rep:repWL(0,S4,N1s,N2s),
	!.
repSentSGsX(Rep,Anno,-1,SGnl,SG,N1,N2,AnnId,_) :-
    Anno=1,
    N1<0,
	sx::sxTrunc(SGnl,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	Rep:repWL(0,S4),
	repAnnotation(Rep,5,79,AnnId,SG),
	!
;
    Anno=1,
	sx::sxTrunc(SGnl,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	sx::fixFormatI(N1,0,1,N1s1),
	sx::sxL_Pad(N1s1," ",10,N1s),
	sx::fixFormatI(N2,0,1,N2s1),
	sx::sxL_Pad(N2s1," ",10,N2s),
	Rep:repWL(0,S4,N1s,N2s),
	repAnnotation(Rep,5,79,AnnId,SG),
	!
;	
    Anno=-1,
    N1<0,
	sx::sxTrunc(SGnl,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	Rep:repWL(0,S4),
	!
;
    Anno=-1,
	sx::sxTrunc(SGnl,78,S3),
	sx::sxR_Pad(S3," ",79,S4),
	sx::fixFormatI(N1,0,1,N1s1),
	sx::sxL_Pad(N1s1," ",10,N1s),
	sx::fixFormatI(N2,0,1,N2s1),
	sx::sxL_Pad(N2s1," ",10,N2s),
	Rep:repWL(0,S4,N1s,N2s),
	!.

% parameters
% 4 inSTs (NL) sorted
% 5 in -1/1 incl. samples
% 6 in -1/1 incl. counts
% 7 in -1/1 incl. determinant STs
% 8 in compressed DR
%
repSentSTs(_,_,_,[],_,_,_,_,_,_) :- 
	!.
repSentSTs(Rep,KBn,Report,[STn|STs],Sample,1,Dets,Anno,Derive,DG) :-
	dbBase::stNL_KB(STn,STk),
	dbBase::stCount(STk,C1,C2),
	repSentSGsX(Rep,Anno,KBn,STn,STk,C2,C1,"p3_iST",""),
	repDerStat(Derive,Rep,STk),
	repSample(Rep,Sample,STk),
	repDetSts(Rep,Dets,STk,DG),
	repST_Rules(Rep,Report,STk),
	repSentSTs(Rep,KBn,Report,STs,Sample,1,Dets,Anno,Derive,DG),
	!.
repSentSTs(Rep,KBn,Report,[STn|STs],Sample,-1,Dets,Anno,Derive,DG) :-
	dbBase::stNL_KB(STn,STk),
	repSentSGsX(Rep,Anno,KBn,STn,STk,-1,-1,"p3_iST",""),
	repDerStat(Derive,Rep,STk),
	repSample(Rep,Sample,STk),
	repDetSts(Rep,Dets,STk,DG),
	repST_Rules(Rep,Report,STk),
	repSentSTs(Rep,KBn,Report,STs,Sample,-1,Dets,Anno,Derive,DG),
	!.

repSG_Rules(_,-1,_,_) :- 
	!.
repSG_Rules(Rep,1,Anno,SG) :-
	dep::get_nd("p3_pDP",A1,A2,B1,B2),
        A1=int(3),
        A2=str(Rule),
        B1=int(1),
        B2=str(SG),
        sx::sxUnpack(Rule,'|',L),
        L=[Chain,Realm,No,SubNo],
        Id=dbBase::dbIdChain(Chain),
        sx::sxlJoin([Id,Realm,No,SubNo],"|",RealmData),
        Rep:repWL(5,"BY RULE:         ",RealmData),
        Anno=1,
        repAnnotation(Rep,14,79,"p3_iSR",RealmData),
        fail
;
	!.
			
repST_Rules(_,-1,_) :- 
	!.
repST_Rules(Rep,1,ST) :-
	dep::get_nd("p3_pDR",A1,A2,B1,B2),
        A1=int(3),
        A2=str(Rule),
        B1=int(2),
        B2=str(ST),
        sx::sxUnpack(Rule,'|',L),
        L=[Chain,Realm,No,SubNo],
        Id=dbBase::dbIdChain(Chain),
        sx::sxlJoin([Id,Realm,No,SubNo],"|",RealmData),
        Rep:repWL(5,"BY RULE:         ",RealmData),
        repAnnotation(Rep,14,79,"p3_iSR",RealmData),
        fail
;
	!.		

repSentChainType(Group,SG,ChainType) :-
	dbBase::chainType(SG,_,Type,_),
	Type=Group,
	ChainType="INT",
	!
;	
	dbBase::chainType(SG,_,Type,Access),
	Type=Group,
	A=Access+1,
	set::setGet(["R","W"],A,S3),
	sx::concat(["PUB/",S3],ChainType),
	!
;	
	dbBase::chainType(SG,DbId,Type,Access),
	Type=Group,
	A=Access+1,
	set::setGet(["R","W"],A,S3),
	sx::concat(["EXT(",DbId,")/",S3],ChainType),
	!.

repSGs(_,[],[]) :-
	!.
repSGs(Rep,[ST|STs],[SGnl|SGs]) :-
	sx::sxUnpack(ST,'|',L),
	L=[SG|_],
	dbBase::nlSing(SG,SGnl),
	repSGs(Rep,STs,SGs),
	!.

repGetSTs(_,[],[]) :-
	!.
repGetSTs(SG,[ST|STsk],[STn|STs]) :-
	sx::sxUnpack(ST,'|',L),
	L=[SG|_],
	dbBase::nlSing(ST,STn),
	repGetSTs(SG,STsk,STs),
	!.
repGetSTs(SG,[_|STsk],STs) :-
	repGetSTs(SG,STsk,STs),
	!.

repSample(_,-1,_) :-
	!.
repSample(Rep,1,ST) :-
    sentenceSample(ST,S1),
	sx::sxQuote(S1,S2),
	Sample=string::concat("EXAMPLE: ",S2),
	Rep:repBlock(Sample,5,79),
	!
;
	!.

repDerStat(DeriveInfo,Rep,St) :-
    DeriveInfo=1,
    der::getDerivationStats(St,N,Secs,_,End,_),
    SecsS=toString(Secs),
    Dur=string::concat("PT",SecsS,"S"),
    xmlGeneric::durationShort(Dur,Durn),
    Ns=toString(N),
    Stat=string::concatList(["Derived ",Ns," sentence(s) during ",DurN," at ",End]),
	Rep:repWL(5,Stat),
	!
;
	!.

repDetSts(_,-1,_,_) :-
	!.
repDetSts(Rep,1,ST,DG) :-
    Preds=DG:preds(int(2),str(ST)),
    Preds:size()>0,
    Rep:repWL(5,"DETERMINANT SENTENCE TYPES:"),
    T=Preds:get_nd(),
        T=cmp(_,[_,str(D)]),
        dbBase::nlSing(D,Dn),
        Rep:repWL(5,Dn),
        fail
;
    !.
% -------------------------------------------------------------------------------------
% repToken                                                                  
% -------------------------------------------------------------------------------------
class predicates
repTokenX : (rep,integer,lst).
getSyno : (string,string,string) procedure (i,i,o).
clauses

% parameters:
% 1 in 1=nl names, 2=nl names with classification
%
repToken(Form) :-
    Kb=dbBase::dbName(),
    File=string::concat(Kb,"_Syntagms"),
	Rep=rep::newReport(File,70),
	not(Rep:cancelled()),
	List=lst::new(),
	repTokenX(Rep,Form,List),
	writeSorted(Rep,0,List),
	Rep:repLine(' '),
	Rep:repWL(0,"* synonym"),
	Rep:close,
    !
;
    !.

repTokenX(Rep,Form,List) :-
    Form=1,
    headerContext(Rep),
	Context=dbBase::getContext(),
	Lng=globalVars::get_s("lang"),
	Rep:repLine(' '),
    dbBase::syntagm_nd(Lng,_,_,_,Type,Tokens,Segment),
        set::member(Segment,Context),
        set::setGet(["t","s"],N2,Type),
        set::setGet(["","*"],N2,Postfix),
        sx::sxlJoin(Tokens," ",S),
        getSyno(Type,S,Main),
        NL=string::concat(S,Postfix,Main),
        List:add(NL),
        fail
;
    Form=2,
    headerContext(Rep),
	Context=dbBase::getContext(),
	Lng=globalVars::get_s("lang"),
	Rep:repLine(' '),
    dbBase::syntagm_nd(Lng,KbName,Clas,Frm,Type,Tokens,Segment),
        set::member(Segment,Context),
        set::setGet(["t","s"],N2,Type),
        set::setGet(["","*"],N2,Postfix),
        dbBase::nlSing(Clas,ClasNL), 
        dbBase::nlSing(Frm,FrmNL), 
        sx::sxlJoin(Tokens," ",S),
        getSyno(Type,S,Main),
        NL=string::concatList([S,PostFix," [",ClasNL,"/",FrmNL," of ",KbName,"]",Main]),
        List:add(NL),
        fail
;
	!.

getSyno("s",S,Main) :-
        dbBase::mainFormOfSyno(S,M),
        Main=string::concat(" --> ",M),
        !.
getSyno(_,_,"") :-
    !.
% -------------------------------------------------------------------------------------------------------------
% repAnnotation
% -------------------------------------------------------------------------------------------------------------
repAnnotation(Rep,Indent,Width,Type,Node) :-
	dbExtension::kbGetAnnoF(Type,Node,Note),
	Rep:repBlock(Note,Indent,Width),
	!
;
	!.	
% -------------------------------------------------------------------------------------------------------------
% writeSorted
% -------------------------------------------------------------------------------------------------------------
class predicates
writeSortedX : (rep,integer,string_list) determ.
clauses

writeSorted(Rep,Indent,List) :-
    L=List:getSorted(),
    writeSortedX(Rep,Indent,L),
    !.
    
writeSortedX(_,_,[]) :-
    !.
writeSortedX(Rep,Indent,[H|T]) :-
    Rep:repWL(Indent,H),
    writeSortedX(Rep,Indent,T),
    !.
% -------------------------------------------------------------------------------------------------------------
% headerContext
% -------------------------------------------------------------------------------------------------------------
class predicates
headerContextX : (rep,string_list) determ.
repTerminology : (rep) determ.
kbName : () -> string determ.
clauses

header(Rep) :-
	Rep:repLine('*'),
	PV=globalVars::get_i("p3v"),
	PVs=toString(PV),
	PR=globalVars::get_i("p3r"),
	PRs=toString(PR),
	sx::sxL_Pad(PRs,"0",2,PRp),
	ReportName=Rep:getName(),
	datetime::tsNow(Tim),
	datetime::tsStr(Tim,Rs),
	Ls=string::concat("Pro/3 v.",PVs,".",PRp),
	Diff=Rep:getWidth()-string::length(Ls)-string::length(Rs),
	Max=Diff-2,
	sx::sxTrunc(ReportName,Max,RepS),
	Diff2=Diff-string::length(RepS),
	AddL=Diff2 div 2,
	AddR=Diff2-AddL,
	sx::sx_ini(AddL,' ',Sl),
	sx::sx_ini(AddR,' ',Sr),
	Rep:repWL(0,Ls,Sl,RepS,Sr,Rs),
	Rep:repLine(' '),
	Name=kbName(),
	Rep:repWL(0,"KB             : ",Name),
	Rep:repLine('*'),
	!.

kbName()=Name:-
    globalVars::get_i("server")=1,
    Name=dbBase::dbName(),
	!
;
    N=dbBase::dbName(),
    Id=dbBase::dbId(),
	Name=string::concat("[",Id,"] ",N),
    !.

headerContext(Rep) :-
	Context=dbBase::getContext(),
	Rep:repLine('*'),
	PV=globalVars::get_i("p3v"),
	PVs=toString(PV),
	PR=globalVars::get_i("p3r"),
	PRs=toString(PR),
	sx::sxL_Pad(PRs,"0",2,PRp),
	ReportName=Rep:getName(),
	datetime::tsNow(Tim),
	datetime::tsStr(Tim,Rs),
	Ls=string::concat("Pro/3 v.",PVs,".",PRp),
	Diff=Rep:getWidth()-string::length(Ls)-string::length(Rs)-string::length(ReportName),
	AddL=Diff div 2,
	AddR=Diff-AddL,
	sx::sx_ini(AddL,' ',Sl),
	sx::sx_ini(AddR,' ',Sr),
	Rep:repWL(0,Ls,Sl,ReportName,Sr,Rs),
	Rep:repLine(' '),
	Name=kbName(),
	Rep:repWL(0,"KB             : ",Name),
	repTerminology(Rep),
	Context=[H|T],
	Rep:repWL(0,"SEGMENT CONTEXT: ",H),
	headerContextX(Rep,T),
	Rep:repLine('*'),
	!
;
	dlgIF::feil("headerContext",[],[],nill),!,fail.	
	
headerContextX(_,[]) :- 
	!.
headerContextX(Rep,[H|T]) :- 
	Rep:repWL(0,"                 ",H),
	headerContextX(Rep,T),
	!.	
repTerminology(Rep) :-
	LangID=globalVars::get_s("lang"),
	dbBase::nlForm(LangID,"p3_iNIL",Lang),
	Rep:repWL(0,"TERMINOLOGY    : ",Lang),
	!.
% -------------------------------------------------------------------------------------------------------------
% sentenceSample
% -------------------------------------------------------------------------------------------------------------
sentenceSample(St,Sample) :-    
    dbBase::stSentence_nd(St,Lt,_),
    dbBase::kbTermSTp(Lt,Sts),
    Sts=St,
	nl::conSentence(cmp(":-",[Lt,atom("true")]),S2),
	Sample=string::concat(S2,"!"),
	!
;
    Sample="",
    !.
end implement p3Rep
