/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement pop
open core,termDomain,globalDomains

constants
className = "com/rudstrand/pop/pop".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% getStTablesNl
% ----------------------------------------------------------------------------------------------
class predicates
getStTabsNl : (lst).
clauses

getStTablesNl()=L :-
    dbBase::kbIsSql,
    List=lst::new(),
    getStTabsNl(List),
    L=List:get(),
    !
;
    L=[],
    !.
    
getStTabsNl(L) :-
    sqlBase::stSentence_nd("x|p3_eSqlTab",Lt,_,_),
        Lt=cmp(_,[cmp(_,[str(TableName),str(St),atom(Type)])]),
        dbBase::nlSing(Type,Typen),
        dbBase::stNL(St,Stn),
        S=string::concatList([Typen,": ",Stn," -> ",TableName]),
        L:add(S),
        fail
;
    !.    
% ----------------------------------------------------------------------------------------------
% getQueries/Menus/Parameters
% ----------------------------------------------------------------------------------------------
class predicates
getQueriesX : (lst,string).
getMenusX : (lst,string).
getParametersX : (lst,string).
packAlts : (sterm,string) determ (i,o).
clauses

getQueries(Lang)=Qs :-
    L=lst::new(),
    getQueriesX(L,Lang),
    Qs=L:get(),
    !.
    
getQueriesX(L,Lang) :-
    dbBase::chainSentences("y",Term),
        Term=cmp(_,[cmp("p3_eKbq",[int(Id),atom(Lng),str(Name),atom(Form),str(Q),int(Menu)])]),
        Lng=Lang,
        dbBase::nlSing(Form,FormNL),
        dbBase::nlSing(Lng,LngNL),
        SId=toString(Id),
        SMenu=toString(Menu),
        sx::sxlJoin([SId,LngNL,Name,FormNL,Q,SMenu],"||",S),
        L:add(S),
        fail
;
    !.

getMenus(Lang)=Qs :-
    L=lst::new(),
    getMenusX(L,Lang),
    Qs=L:get(),
    !.

getMenusX(L,Lang) :-
    dbBase::chainSentences("y",Term),
        Term=cmp(_,[cmp("p3_eKbm",[int(Id),atom(Lng),str(Name),int(Menu)])]),
        Lng=Lang,
        dbBase::nlSing(Lng,LngNL),
        SId=toString(Id),
        SMenu=toString(Menu),
        sx::sxlJoin([SId,LngNL,Name,SMenu],"||",S),
        L:add(S),
        fail
;
    !.

getParameters(Lang)=Qs :-
    L=lst::new(),
    getParametersX(L,Lang),
    Qs=L:get(),
    !.

getParametersX(L,Lang) :-
    dbBase::chainSentences("y",Term),
        Term=cmp(_,[cmp("p3_eKbp",[str(Name),atom(Lng),atom(Dom),str(Title),Alts])]),
        Lng=Lang,
        dbBase::nlSing(Lng,LngNL),
        dbBase::nlSing(Dom,DomNL),
        packAlts(Alts,AltsP),
        sx::sxlJoin([Name,LngNL,DomNL,Title,AltsP],"||",S),
        L:add(S),
        fail
;
    !.
    
packAlts(nill,"") :-
    !.
packAlts(Alts,AltsP) :-
   termDomain::listBR(Alts,L),
   conversion::toStrings(L,Ls),
   sx::sxlJoin(Ls,";;",AltsP),
   !.
% ----------------------------------------------------------------------------------------------
% getCRs
% ----------------------------------------------------------------------------------------------
getCRsNL()=L:-
    Lk=getCRs(),
    dbBase::nlSings(Lk,L),
    !.
getCRsNL(Type)=L:-
    Lk=getCRs(Type),
    dbBase::nlSings(Lk,L),
    !.
    
getCRs()=L:-
    Lst=dbBase::getIrs(),
    L=Lst:get(),
    !.
getCRs("p3_iCR")=L:-
	L2=getCRs("p3_iCand"),
	L3=getCRs("p3_iCor"),
	L4=getCRs("p3_iCnot"),
	L5=getCRs("p3_iCcmb"),
	L6=getCRs("p3_iCbay"),
	set::conc([L2,L3,L4,L5,L6],L),
    !.
getCRs("p3_iFZ")=L:-
	L10=getCRs("p3_iZand"),
	L11=getCRs("p3_iZor"),
	L12=getCRs("p3_iZnot"),
	L13=getCRs("p3_iZowa"),
	set::conc([L10,L11,L12,L13],L),
    !.
getCRs("p3_iSupp")=L:-
	L2=getCRs("p3_iCmap"),
	L3=getCRs("p3_iCswi"),
	L4=getCRs("p3_iCdat"),
	L5=getCRs("p3_iCpar"),
	set::conc([L2,L3,L4,L5],L),
    !.
getCRs("p3_iCdat")=L:-
	L2=getCRs("p3_iCqst"),
	L3=getCRs("p3_iCqry"),
	set::conc([L2,L3],L),
	!.
getCRs("p3_iCswi")=L:-
	L2=getCRs("p3_iCswp"),
	L3=getCRs("p3_iCswc"),
	set::conc([L2,L3],L),
	!.
	
getCRs(SubType)=L:-
    Lst=dbBase::getIrsOfSubtype(SubType),
    L=Lst:get(),
    !.
% ----------------------------------------------------------------------------------------------
% getCRsDomNL
% ----------------------------------------------------------------------------------------------
class predicates
getCRsDomNL_X : (string_list,string_list,string_list) determ (i,i,o).
clauses

% parameters:
% 1	in	cr type or blank if all
% 2	in	1/-1 incl suspended
% 3	out	list of CRs on format: RULENAME_NL :: RULE_TYPE_NL (DOMAIN_NL) suspended rules prefixed with *
%
getCRsDomNL()=L :-
	CRs2=getCRs(),
	getCRsDomNL_X([],CRs2,L),
	!.
getCRsDomNL(Type)=L :-
	Type="p3_iCR",
	CRs2=getCRs(Type),
	getCRsDomNL_X(["p3_dcf","p3_dpr"],CRs2,L),
	!
;
	Type="p3_iFZ",
	CRs2=getCRs(Type),
	getCRsDomNL_X(["p3_dmg"],CRs2,L),
	!
;	
	CRs2=getCRs(Type),
	getCRsDomNL_X([],CRs2,L),
	!.
	
getCRsDomNL_X(_,[],[]) :-
	!.
getCRsDomNL_X([],[H|T],[Hm|Tm]) :-
    sx::sxRemoveHead(H,"*",Hx),
	dbBase::smClassDef(Hx,"h",_,_,Dom,_,_),
    dbBase::getIrDefF(Hx,_,Loc),
	dbBase::getRecord(Loc,_,_,_,Suspended,_,_,_,_,_,_,_),
	Ix=Suspended+2,
	set::setGet(["","","*"],Ix,Prefix),
	dbBase::nlSing(Dom,S2),
	dbBase::nlSing(H,S3),
	dbBase::irType(H,_,Typ),
	dbBase::nlSing(Typ,S4),	
	sx::concat([Prefix,S3," :: ",S4," (",S2,")"],Hm),
	getCRsDomNL_X([],T,Tm),
	!.	
getCRsDomNL_X(Doms,[H|T],L) :-
    sx::sxRemoveHead(H,"*",Hx),
	dbBase::smClassDef(Hx,"h",_,_,Dom,_,_),
    dbBase::getIrDefF(Hx,_,Loc),
	dbBase::getRecord(Loc,_,_,_,Suspended,_,_,_,_,_,_,_),
	Ix=Suspended+2,
	set::setGet(["","","*"],Ix,Prefix),
	set::member(Dom,Doms),
	dbBase::nlSing(Dom,S2),
	dbBase::nlSing(H,S3),
	dbBase::irType(H,_,Typ),
	dbBase::nlSing(Typ,S4),	
	sx::concat([Prefix,S3," :: ",S4," (",S2,")"],Hm),
	getCRsDomNL_X(Doms,T,Tm),
	L=[Hm|Tm],
	!
;
	getCRsDomNL_X(Doms,T,L),
	!.
% ----------------------------------------------------------------------------------------------
% getChains/getChainsNL_WithKB_File/getPubChains/getPubChainsDB/getExtChains
% ----------------------------------------------------------------------------------------------
class predicates
getChainsX : (lst).
getChainsExtDB_X : (chainDB,integer,lst).
getChainsNWK : (lst).
getChainsPubX : (integer,lst).
getChainsPubDB_X : (chainDB,integer,lst).
getChainsExtX : (integer,lst).
addKbId : (string_list,string_list) determ (i,o).    
clauses

getChains()=L:-
    List=lst::new(),
    getChainsX(List),
    L=List:get(),
    !.
getChainsX(List) :-
    dbBase::chains_nd(1,Chain),
        List:add(Chain),
        fail
;
    !.
    
getChainsNL_WithKB_File()=L:-
    List=lst::new(),
    getChainsNWK(List),
    L=List:get(),
    !.
   
getChainsNWK(List) :-
	dbBase::chains_nd(1,Chain),
        dbBase::nlTry(Chain,ChainNL),
        Id=dbBase::dbIdChain(Chain),
        S=string::concatList([Chain," ",ChainNL," [",Id,"]"]),
        List:add(S),
        fail
;
    !.
    
getChainsPub(Rw)=L:-
    List=lst::new(),
    getChainsPubX(Rw,List),
    L=List:get(),
    !.
    
getChainsPubX(Rw,List) :-
    dbBase::kbIsSql,
    !
;    
    Rw=-1,
    dbaBase::pubChain(C,_),
    List:add(C),
    fail
;
    Rw<>-1,
    dbaBase::pubChain(C,Rw),
    List:add(C),
    fail
;
    !.

getChainsPubDB_AccessNL(DB)=L:-
    dbBase::kbIsSql,
    L=[],
    !
;    
    Lr=getChainsPubDB(DB,0),
    Lw=getChainsPubDB(DB,1),
    dbBase::nlSings(Lr,Lrn),
    dbBase::nlSings(Lw,Lwn),
    sx::sxlAddPrefix(Lrn,"READ:  ",Lrp),
    sx::sxlAddPrefix(Lwn,"WRITE: ",Lwp),
    set::conc([Lrp,Lwp],L),
    !.
    
getChainsPubDB(DB,Rw)=L:-
    dbBase::kbIsSql,
    L=[],
    !
;    
    List=lst::new(),
    getChainsPubDB_X(DB,Rw,List),
    L=List:get(),
    !.
    
getChainsPubDB_X(KB,Rw,List) :-
    Rw=-1,
	KB:chain_terms("p3_pIpch",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp(_,[cmp("p3_eChan",[atom(Chain),_])]),
    List:add(Chain),
    fail
;
    Rw=0,
	KB:chain_terms("p3_pIpch",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp(_,[cmp("p3_eChan",[atom(Chain),atom("p3_iRo")])]),
    List:add(Chain),
    fail
;
    Rw=1,
	KB:chain_terms("p3_pIpch",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp(_,[cmp("p3_eChan",[atom(Chain),atom("p3_iRw")])]),
    List:add(Chain),
    fail
;
    !.

getChainsExtDB(DB,Rw)=L:-
    List=lst::new(),
    getChainsExtDB_X(DB,Rw,List),
    L=List:get(),
    !.
    
getChainsExtDB_X(KB,Rw,List) :-
    dbBase::kbIsSql,
    !
;    
    Rw=-1,
	KB:chain_terms("p3_pIxdb",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp(_,[cmp("p3_eChan",[atom(Chain),_])]),
    List:add(Chain),
    fail
;
    Rw=0,
	KB:chain_terms("p3_pIxdb",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp(_,[cmp("p3_eChan",[atom(Chain),atom("p3_iRo")])]),
    List:add(Chain),
    fail
;
    Rw=1,
	KB:chain_terms("p3_pIxdb",Cl,_),
	Cl=clause(Z,_,_,_,_,E,_),
	AZ=math::abs(Z), AZ<10,
	E=cmp(_,[cmp("p3_eChan",[atom(Chain),atom("p3_iRw")])]),
    List:add(Chain),
    fail
;
    !.

getChainsExtAccessNL()=L:-
    dbBase::kbIsSql,
    L=[],
    !
;    
    Lr=getChainsExt(0),
    Lw=getChainsExt(1),
    dbBase::nlSings(Lr,Lrn),
    dbBase::nlSings(Lw,Lwn),
    sx::sxlAddPrefix(Lrn,"READ:  ",Lrp),
    sx::sxlAddPrefix(Lwn,"WRITE: ",Lwp),
    set::conc([Lrp,Lwp],L2),
    addKbId(L2,L),
    !.
    
getChainsExt(Rw)=L:-
    List=lst::new(),
    getChainsExtX(Rw,List),
    L=List:get(),
    !.
    
getChainsExtX(Rw,List) :-
    dbBase::kbIsSql,
    !
;    
    Rw=-1,
    dbaBase::extChain(C,_,_),
    List:add(C),
    fail
;
    Rw<>-1,
    dbaBase::extChain(C,_,Rw),
    List:add(C),
    fail
;
    !.

addKbId([],[]) :-
    !.
addKbId([C|Cs],[Cm|Cms]) :-
    sx::frontstr(7,C,_,Cn),
    dbBase::kbID(Cn,Ck,_,_),
    dbaBase::extChain(Ck,DB,_),
    dbaBase::database(DB,_,Id,_),
    Cm=string::concat(Id," ",C),
    addKbId(Cs,Cms),
    !.
% ----------------------------------------------------------------------------------------------
% getDBs/DBs_Info        
% ----------------------------------------------------------------------------------------------
class predicates
getDBs_X : (integer,string_list) procedure (i,o).
getDB_Infos_X : (integer,string_list) procedure (i,o).
clauses

getDBs()=L:-
    dbBase::kbIsSql,
    !,L=["X"],
    !
;
    getDBs_X(0,L),
    !.
    
getDBs_X(N,L) :-
    dbaBase::database(_,_,Id,N),
    Next=N+1,
    getDBs_X(Next,T),
    L=[Id|T],
    !
;
    L=[],
    !.
    
getDB_Infos()=L:-
    dbBase::kbIsSql,
    !, D=sqlBase::getDb(),
    K=sqlBase::getKb(),
    S=string::concat(D,".",K,".sql [X]"),
    L=[S],
    !
;    
    getDB_Infos_X(0,L),
    !.
    
getDB_Infos_X(N,L) :-
    dbaBase::database(_,File,Id,N),
    H=string::concat(File," [",Id,"]"),
    Next=N+1,
    getDB_Infos_X(Next,T),
    L=[H|T],
    !
;
    L=[],
    !.
% ----------------------------------------------------------------------------------------------
% getNodes/NL                                     
% ----------------------------------------------------------------------------------------------
class predicates
getNodesNL_X : (string_list,string_list) determ (i,o).    
clauses
% returns a list of SM objects of given type
%
% format NODETYPE{NODENAME
% 
/* @@@ not included 15.10.05
v(t(p3_iEngl,  "p3_iFunc", p3_dif,  p3_iNIL, ["FUNCTION"]),"Pro/3")! 
v(t(p3_iEngl,  "p3_iIR",   p3_dif,  p3_iNIL, ["INEXACT_RULE"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iSTO",  p3_dif,  p3_iNIL, ["OPEN_SENTENCE_TYPE"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iAnno", p3_dif,  p3_iNIL, ["ANNOTATION"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iIST",  p3_dif,  p3_iNIL, ["ENTERED_SENTENCE_TYPE"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iDST",  p3_dif,  p3_iNIL, ["DERIVED_SENTENCE_TYPE"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iFncB", p3_dif,  p3_iNIL, ["FUNCTION_DEFINITION"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iInpSer",p3_dif,  p3_iNIL, ["SERIES_INPUT_FACILITY"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iInpClu",p3_dif,  p3_iNIL, ["CLUSTER_INPUT_FACILITY"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iTabLay",p3_dif,  p3_iNIL, ["TABLE_LAYOUT"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iKQry",p3_dif,  p3_iNIL, ["KB_QUERY"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iKQryC",p3_dif,  p3_iNIL, ["KB_CLUSTER_QUERY"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iKQryM",p3_dif,  p3_iNIL, ["KB_QUERY_MENU"]),"Pro/3")!
v(t(p3_iEngl,  "p3_iKQryP",p3_dif,  p3_iNIL, ["KB_QUERY_PARAMETER"]),"Pro/3")!
*/
getNodes()=L:-
    Le=getSM("e",[]),
	sx::sxlAddPrefix(Le,"p3_iET{",Le2),
    Lp=getSM("p",[]),
	sx::sxlAddPrefix(Lp,"p3_iPT{",Lp2),
    La=getSM("a",[]),
	sx::sxlAddPrefix(La,"p3_iDET{",La2),
    Lf=getSM("f",[]),
	sx::sxlAddPrefix(Lf,"p3_iFncD{",Lf2),
    Lg=getSM("g",[]),
	sx::sxlAddPrefix(Lg,"p3_iFncD{",Lg2),
    Lcr=getSM("h"),
	sx::sxlAddPrefix(Lcr,"p3_iIRdef{",Lcr2),
    Lcr3=getCRs(),
	sx::sxlAddPrefix(Lcr3,"p3_iIRdec{",Lcr4),
    Lsr=getSRs(1),
	sx::sxlAddPrefix(Lsr,"p3_iSR{",Lsr2),
    Lr=dbBase::realms(),
	sx::sxlAddPrefix(Lr,"p3_iRelm{",Lr2),
    Lst=getSTs(), 
	sx::sxlAddPrefix(Lst,"p3_iST{",Lst2),
    Ldb=getDBs(),
	sx::sxlAddPrefix(Ldb,"p3_iDB{",Ldb2),
    Lsg=getSegments(),
	sx::sxlAddPrefix(Lsg,"p3_iSegm{",Lsg2),
	set::conc([Le2,Lp2,La2,Lf2,Lg2,Lcr2,Lcr4,Lsr2,Lr2,Lst2,Ldb2,Lsg2],L),
	!.
% format: NODETYP: NODENAME
getNodesNL()=L :-
    Lk=getNodes(),
    getNodesNL_X(Lk,L),
    !.
getNodesNL_X([],[]) :-
    !.
getNodesNL_X([H|T],[Hn|Tn]) :-
    sx::sxUnpack(H,'{',L),
    L=[Tp,Nd],
    set::member(Tp,["p3_iRelm","p3_iDB","p3_iSegm"]),
    dbBase::nlSing(Tp,Tpn),
    Hn=string::concat(Tpn,": ",Nd),
    getNodesNL_X(T,Tn),
    !
;
    sx::sxUnpack(H,'{',L),
    L=[Tp,Nd],
    dbBase::nlSing(Tp,Tpn),
    dbBase::nlTry(Nd,Ndn),
    Hn=string::concat(Tpn,": ",Ndn),
    getNodesNL_X(T,Tn),
    !
;
    Hn=H,
    getNodesNL_X(T,Tn),
    !.
% ----------------------------------------------------------------------------------------------
% getRealms                            
% ----------------------------------------------------------------------------------------------
class predicates
getRealmsWK : (integer,string_list,string_list) determ (i,i,o).
clauses

getRealms()=List :-
    dbBase::kbIsSql,
    !,List=dbBase::realms(),
    !
;    
    dbExtension::multiDb,
    !,L=dbBase::realms(),
    getRealmsWK(1,L,List),
    !
;    
    L=dbBase::realms(),
    getRealmsWK(-1,L,List),
    !.
    
getRealmsWK(_,[],[]) :-
    !.
getRealmsWK(1,[R|Rs],[Rn|Rns]) :-
    sx::sxUnpack(R,'|',L),
    L=[Realm,Id],
    Rn=string::concatList([Id,":",Realm]),
    getRealmsWK(1,Rs,Rns),
    !.
getRealmsWK(-1,[R|Rs],[Rn|Rns]) :-
    sx::sxUnpack(R,'|',L),
    L=[Rn,_],
    getRealmsWK(-1,Rs,Rns),
    !.
% ----------------------------------------------------------------------------------------------
% getNLs
% ----------------------------------------------------------------------------------------------
class predicates
getClassNL : (lst,string,string,string_list,string_list).
getNL : (lst,string,string_list,string_list).
clauses

getClassNLs(Clas,Lang,Synos,Context)=L :-
    Synos=1,
    List=lst::new(),
    getClassNL(List,Clas,Lang,["t","s"],Context),
    L=List:get(),
    !
;    
    List=lst::new(),
    getClassNL(List,Clas,Lang,["t"],Context),
    L=List:get(),
    !.
    
getNLs(Lang,Synos,Context)=L :-
    Synos=1,
    List=lst::new(),
    getNL(List,Lang,["t","s"],Context),
    L=List:get(),
    !
;    
    List=lst::new(),
    getNL(List,Lang,["t"],Context),
    L=List:get(),
    !.
    
getNLs(Lang,Synos)=L :-
    Synos=1,
    List=lst::new(),
    getNL(List,Lang,["t","s"],[]),
    L=List:get(),
    !
;    
    List=lst::new(),
    getNL(List,Lang,["t"],[]),
    L=List:get(),
    !.
    
getNL(List,Lang,Ents,Context) :-
    Context=[],
    dbBase::chainSentences("v",LT),
        LT=cmp(_,[cmp(Typ,[atom(Lng),_,_,_,Tokens]),_]),
        Lng=Lang,
        set::member(Typ,Ents),
        termDomain::list_conc(Tokens,NL),
        List:add(NL),
        fail
;
    not(Context=[]),
    dbBase::chainSentences("v",LT),
        LT=cmp(_,[cmp(Typ,[atom(Lng),_,_,_,Tokens]),str(Segment)]),
        Lng=Lang,
        set::member(Typ,Ents),
        set::member(Segment,Context),
        termDomain::list_conc(Tokens,NL),
        List:add(NL),
        fail
;
	!.

getClassNL(List,Clas,Lang,Ents,Context) :-
    Context=[],
    dbBase::chainSentences("v",LT),
        LT=cmp(_,[cmp(Typ,[atom(Lng),_,atom(Cls),_,Tokens]),_]),
        Lng=Lang,
        Cls=Clas,
        set::member(Typ,Ents),
        termDomain::list_conc(Tokens,NL),
        List:add(NL),
        fail
;
    not(Context=[]),
    dbBase::chainSentences("v",LT),
	LT=cmp(_,[cmp(Typ,[atom(Lng),_,atom(Cls),_,Tokens]),str(Segment)]),
        Lng=Lang,
        Cls=Clas,
        set::member(Typ,Ents),
        set::member(Segment,Context),
        termDomain::list_conc(Tokens,NL),
        List:add(NL),
        fail
;
	!.
% ----------------------------------------------------------------------------------------------
% getTerminology/Identifiers
% ----------------------------------------------------------------------------------------------
class predicates
getTerm : (lst,string,string_list,string_list).
getKB_Names : (lst,string,string_list,string_list).
getId : (lst,string,string_list,string_list).
getSyntaxClassesNL_X : (string,lst,lst).    
clauses

getTerminology(Lang,Synos)=L :-
    Synos=1,
    List=lst::new(),
    getTerm(List,Lang,["t","s"],[]),
    L=List:get(),
    !
;    
    List=lst::new(),
    getTerm(List,Lang,["t"],[]),
    L=List:get(),
    !.
    
getTerminology(Lang,Synos,Context)=L :-
    Synos=1,
    List=lst::new(),
    getTerm(List,Lang,["t","s"],Context),
    L=List:get(),
    !
;    
    List=lst::new(),
    getTerm(List,Lang,["t"],Context),
    L=List:get(),
    !.
    
getKB_NameInfo(Lang,Synos,Context)=L :-
    Synos=1,
    List=lst::new(),
    getKB_Names(List,Lang,["t","s"],Context),
    L=List:get(),
    !
;    
    List=lst::new(),
    getKB_Names(List,Lang,["t"],Context),
    L=List:get(),
    !.
    
getIdentifiers(Lang,Synos)=L :-
    Synos=1,
    List=lst::new(),
    getId(List,Lang,["t","s"],[]),
    L=List:get(),
    !
;    
    List=lst::new(),
    getId(List,Lang,["t"],[]),
    L=List:get(),
    !.
    
getIdentifiers(Lang,Synos,Context)=L :-
    Synos=1,
    List=lst::new(),
    getId(List,Lang,["t","s"],Context),
    L=List:get(),
    !
;    
    List=lst::new(),
    getId(List,Lang,["t"],Context),
    L=List:get(),
    !.
    
getTerm(List,Lang,Syn,Context) :-
    Context=[],
    dbBase::chainSentences("v",LT),
        LT=cmp(_,[cmp(Typ,[atom(Lng),str(Id),_,_,_]),_]),
        Lng=Lang,
        set::member(Typ,Syn),
        List:add(Id),
        fail
;
    not(Context=[]),
    dbBase::chainSentences("v",LT),
        LT=cmp(_,[cmp(Typ,[atom(Lng),str(Id),_,_,_]),str(Segment)]),
        Lng=Lang,
        set::member(Typ,Syn),
        set::member(Segment,Context),
        List:add(Id),
        fail
;
	!.

getKB_Names(List,Lang,Syn,Context) :-
    Context=[],
    dbBase::chainSentences("v",LT,Ref),
        LT=cmp("v",[cmp(Typ,[atom(Lng),str(Rep),atom(Clas),atom(Form),Tokens]),str(Segment)]),
        Lng=Lang,
        set::member(Typ,Syn),
        termDomain::list_conc(Tokens,NL),
        dbBase::nlSing(Clas,C),
        dbBase::nlSing(Form,F),
        SRef=dbBase::dbRef(Ref),
        set::setGet(["t","s"],Ix,Typ),
        set::setGet(["","(synonym)"],Ix,St),
        sx::concat([Rep,"  ",NL," ",St,"  ",C,"  ",F,"  ",Segment,"  [",SRef,"]"],S),
        List:add(S),
        fail
;
    not(Context=[]),
    dbBase::chainSentences("v",LT,Ref),
        LT=cmp("v",[cmp(Typ,[atom(Lng),str(Rep),atom(Clas),atom(Form),Tokens]),str(Segment)]),
        Lng=Lang,
        set::member(Typ,Syn),
        set::member(Segment,Context),
        termDomain::list_conc(Tokens,NL),
        dbBase::nlSing(Clas,C),
        dbBase::nlSing(Form,F),
        SRef=dbBase::dbRef(Ref),
        set::setGet(["t","s"],Ix,Typ),
        set::setGet(["","(synonym)"],Ix,St),
        sx::concat([Rep,"  ",NL," ",St,"  ",C,"  ",F,"  ",Segment,"  [",SRef,"]"],S),
        List:add(S),
        fail
;
	!.

getId(List,Lang,Syn,Context) :-
    Context=[],
    dbBase::chainSentences("v",LT),
        LT=cmp(_,[cmp(Typ,[atom(Lng),str(Id),atom(Clas),_,_]),_]),
        Lng=Lang,
        Clas="p3_dif",
        set::member(Typ,Syn),
        List:add(Id),
        fail
;
    not(Context=[]),
    dbBase::chainSentences("v",LT),
        LT=cmp(_,[cmp(Typ,[atom(Lng),str(Id),atom(Clas),_,_]),str(Segment)]),
        Lng=Lang,
        Clas="p3_dif",
        set::member(Typ,Syn),
        set::member(Segment,Context),
        List:add(Id),
        fail
;
	!.

getSyntaxClassesNL(Lang,Cls,Frms) :-
    List1=lst::new(),
    List2=lst::new(),
    getSyntaxClassesNL_X(Lang,List1,List2),
    Cls=List1:get(),
    Frms=List2:get(),
    !.

getSyntaxClassesNL_X(Lang,List1,List2) :-
    dbBase::chainSentences("v",LT),
	LT=cmp(_,[cmp(_,[atom(Lng),_,atom(Clas),atom(Form),_]),_]),
        Lng=Lang,
        dbBase::nlSing(Clas,C),
        dbBase::nlSing(Form,F),
        List1:addUnique(C),
        List2:addUnique(F),
        fail
;
	!.
% ----------------------------------------------------------------------------------------------
% getSM                        
% ----------------------------------------------------------------------------------------------
class predicates
getSM_X : (lst,string,string_list).
getSM_AllX : (termSet,string_list).	
clauses

% returns a list of SM objects of given type and in given segment(s)
% SM_Type   the name of object type
% Segment   list of segments (all if [])
%
getSM_NL(SM_Type,Context)=L:-
    Lk=getSM(SM_Type,Context),
    dbBase::nlSings(Lk,L),
    !.
    
getSM(SM_Type,Context)=L:-
    List=lst::new(),
    getSM_X(List,SM_Type,Context),
    L=List:get(),
    !.

% returns a list of SM objects of given type and in current context
% SM_Type   the name of object type
%
getSM_NL(SM_Type)=L:-
    Lk=getSM(SM_Type),
    dbBase::nlSings(Lk,L),
    !.
    
getSM(SM_Type)=L:-
    Context=dbBase::getContext(),
    List=lst::new(),
    getSM_X(List,SM_Type,Context),
    L=List:get(),
    !.

getSM_X(List,SM_Type,Context) :-
    Context=[],
    cache::smCacheIsLoaded(),
    cache::getSm_nd(Name,SM_Type,_,_,_,_,_,_),
        List:add(Name),
        fail
;
    not(Context=[]),
    cache::smCacheIsLoaded(),
    cache::getSm_nd(Name,SM_Type,_,_,_,_,_,Segment),
        set::member(Segment,Context),
        List:add(Name),
        fail
;
    Context=[],
    not(cache::smCacheIsLoaded()),
    dbBase::chainSentences("q",LT),
        LT=cmp(_,[cmp(SM_Type,[atom(Name)|_]),_]),
        List:add(Name),
        fail
;
    not(Context=[]),
    not(cache::smCacheIsLoaded()),
    dbBase::chainSentences("q",LT),
        LT=cmp(_,[cmp(SM_Type,[atom(Name)|_]),str(Segment)]),
        set::member(Segment,Context),
        List:add(Name),
        fail
;
	!.

getSM_All()=Set :-
    Context=dbBase::getContext(),
    Set=getSM_All(Context),
    !.
    
getSM_All(Context)=Set :-
    Set=termSet::new(),
    getSM_AllX(Set,Context),
    !.
    
getSM_AllX(Set,Context) :-
    Context=[],
    cache::smCacheIsLoaded(),
    cache::getSm_nd(Name,SM_Type,_,_,_,_,_,Segment),
        Set:add(cmp("",[str(SM_Type),str(Name),str(Segment)])),
        fail
;
    not(Context=[]),
    cache::smCacheIsLoaded(),
    cache::getSm_nd(Name,SM_Type,_,_,_,_,_,Segment),
        set::member(Segment,Context),
        Set:add(cmp("",[str(SM_Type),str(Name),str(Segment)])),
        fail
;
    Context=[],
    dbBase::chainSentences("q",LT),
        LT=cmp(_,[cmp(Type,[atom(Name)|_]),str(Segment)]),
        Set:add(cmp("",[str(Type),str(Name),str(Segment)])),
        fail
;
    not(Context=[]),
    dbBase::chainSentences("q",LT),
        LT=cmp(_,[cmp(Type,[atom(Name)|_]),str(Segment)]),
        set::member(Segment,Context),
        Set:add(cmp("",[str(Type),str(Name),str(Segment)])),
        fail
;
	!.
% ----------------------------------------------------------------------------------------------
% getSegments                                    
% ----------------------------------------------------------------------------------------------
class predicates
getSegmentsX : (lst).
clauses

getSegments()=L:-
    List=lst::new(),
    getSegmentsX(List),
    L=List:get(),
    !.
getSegmentsX(List) :-
	dbBase::chainSentences("v",LT),
	LT=cmp(_,[_,str(Segment)]),
	List:addUnique(Segment),
	fail
;
	dbBase::chainSentences("q",LT),
	LT=cmp(_,[_,str(Segment)]),
	List:addUnique(Segment),
	fail
;
	!.
% ----------------------------------------------------------------------------------------------
% getSRs                               
% ----------------------------------------------------------------------------------------------
class predicates
getSRsDetails : (dbLoc_list,string_list) determ (i,o).
clauses

getSRs(InclSusp)=L :-
    InclSusp=1,
    !,dbBase::recBasetypeLocsAll(3,L2),
    getSRsDetails(L2,L),
    !
;    
    !,dbBase::recBasetypeLocs(3,L2),
    getSRsDetails(L2,L),
    !.
    
getSRsDetails([],[]) :-
    !.
getSRsDetails([Loc|Locs],[D|Ds]) :-
    dbBase::getRecord(Loc,_,_,_,_,_,Realm,No,SubNo,_,LT,_),
	NoS=toString(No),
	SNoS=toString(SubNo),
	LT=cmp(Chain,_),
	sx::sxlJoin([Chain,Realm,NoS,SNoS],"|",D),
    getSRsDetails(Locs,Ds),
    !.
% ----------------------------------------------------------------------------------------------
% getSTs         
% ----------------------------------------------------------------------------------------------
class predicates
remSTs : (string_list,string,string_list) determ (i,i,o).
clauses

getSTs()=L:-
    Context=dbBase::getContext(),
    L=getSTs(Context),
	!.

getSTs(Context)=L:-
    L2=dbBase::stList(),
    PredsInContext=getSM("p",Context),
    filterSTs(L2,PredsInContext,L),
	!.

getSTs(SG,Context)=L:-
    STs=getSTs(Context),
    remSTs(STs,SG,L),
    !.
    
getSentenceSTs()=L:-
    Context=dbBase::getContext(),
    L=getSentenceSTs(Context),
    !.
    
getSentenceSTs(Context)=L:-
    Context=[],
    STs2=dbBase::stBasetypes(1),
    STs3=dbBase::stBasetypes(2),
	set::conc([STs2,STs3],STs),
	set::distinct(STs,L),
	!
;
    L2=getSentenceSTs([]),
    PredsInContext=getSM("p",Context),
    filterSTs(L2,PredsInContext,L),
	!.
getSentenceSTs(SG,Context)=L:-
    L2=getSentenceSTs(Context),
    remSTs(L2,SG,L),
    !.
    
getRuleSTs()=L:-
    Context=dbBase::getContext(),
    L=getRuleSTs(Context),
    !.
    
getRuleSTs(Context)=L:-   
    Context=[],
    L=dbBase::stBasetypes(3),
    !
;
    L2=getRuleSTs([]),
    PredsInContext=getSM("p",Context),
    filterSTs(L2,PredsInContext,L),
	!.
	
% parameters:
% 1 in  list of all STs
% 2 in SG
% 3 out list of STs in SG
%
remSTs([],_,[]) :-
    !.
remSTs([H|T],S,STs) :-
    sx::sxUnpack(H,'|',L),
    L=[S|_],
    remSTs(T,S,Ts),
    STs=[H|Ts],
    !
;
    remSTs(T,S,STs),
    !.

filterSTs([],_,[]) :-
    !.
filterSTs([H|T],SGs,STs) :-
    sx::sxUnpack(H,'|',L),
    L=[SG|_],
    set::member(SG,SGs),
    filterSTs(T,SGs,Ts),
    STs=[H|Ts],
    !
;
    filterSTs(T,SGs,STs),
    !.
% ----------------------------------------------------------------------------------------------
% getBinaryPreds/UnaryPreds                                              
% ----------------------------------------------------------------------------------------------
class predicates
getP_Sel : (string_list,integer,string_list) procedure (i,i,o).
clauses
getBinaryPreds()=L:-
    A=getSM("p"),
    getP_Sel(A,2,L),
    !.
getBinaryPreds(Context)=L:-
    A=getSM("p",Context),
    getP_Sel(A,2,L),
    !.
getUnaryPreds()=L:-
    A=getSM("p"),
    getP_Sel(A,1,L),
    !.
getUnaryPreds(Context)=L:-
    A=getSM("p",Context),
    getP_Sel(A,1,L),
    !.
    
getP_Sel([],_,[]) :-
    !.
getP_Sel([H|T],Card,L) :-
    dbBase::smClassDef(H,"p",Crd,_,_,_,_),
    Crd=Card,
    getP_Sel(T,Card,Tm),
    L=[H|Tm],
    !
;
    getP_Sel(T,Card,L),
    !.
% ----------------------------------------------------------------------------------------------
% getTerminologies                                                    
% ----------------------------------------------------------------------------------------------
class predicates
getTerminologiesX : (lst).
clauses

getTerminologies()=L:-
    List=lst::new(),
    getTerminologiesX(List),
    L=List:get(),
    !.
getTerminologiesX(List) :-
    dbBase::chainSentences("x",LT),
    LT=cmp(_,[cmp("p3_eTerm",[atom(T)])]),
	List:add(T),
	fail
;
	!.
% ----------------------------------------------------------------------------------------------
% getAnno                                              
% ----------------------------------------------------------------------------------------------
class predicates
getAnnoX : (lst).
getAnnoNL_X : (lst).
annoNameNL : (string,string,string) procedure (i,i,o).
clauses
% format NODECLASSif{NODENAMEif
%
getAnno()=L:-
    List=lst::new(),
    getAnnoX(List),
    L=List:get(),
    !.

% format NODECLASS : NODENAME
%
getAnnoNL()=L:-
    List=lst::new(),
    getAnnoNL_X(List),
    L=List:get(),
    !.

getAnnoX(List) :-
    dbBase::chainSentences("p3_pIntd",LT),
        LT=cmp(_,[cmp("p3_eAnno",[atom(NC),str(NN),_])]),
        sx::sxlJoin([NC,NN],"{",S),
        List:add(S),
        fail
;
	!.

getAnnoNL_X(List) :-
    dbBase::chainSentences("p3_pIntd",LT),
        LT=cmp(_,[cmp("p3_eAnno",[atom(NC),str(NN),_])]),
        dbBase::nlSing(NC,NCn),
        annoNameNL(NC,NN,NNn),
        S=string::concat(NCn,": ",NNn),
        List:add(S),
        fail
;
	!.

annoNameNL(Type,K,NL) :-  	
    set::member(Type,["p3_iRelm","p3_iDB","p3_iSegm","p3_iSR"]),
    NL=K,
    !
;
    dbBase::nlTry(K,NL),
    !.
% ----------------------------------------------------------------------------------------------
% inContext
% ----------------------------------------------------------------------------------------------
inContext(_,_,[]) :-
    !.
inContext(K,SM_Type,Context) :-
    dbBase::chainSentences("q",LT),
	LT=cmp(_,[cmp(Type,[atom(Name)|_]),str(Segment)]),
	Type=SM_Type,
	Name=K,
	set::member(Segment,Context),
    !.
end implement pop
