/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement pr
open core,termDomain

constants
className = "com/rudstrand/pr/pr".
classVersion = "1.0".
quote = "\"".

class predicates
% ------------------------------------------------------------------------------------------------------------
prValueList : (sterm,string) determ (i,o).
prExpr : (sterm,string) determ (i,o).
prExprs : (sterml,string_list) determ (i,o).
prFreeVars : (sterm,string_list) determ (i,o).
prFuncFormalPars : (sterml,string_list) determ (i,o).
prPredCond : (string,sterm,string) determ (i,i,o).
prValFindVar : (string,sterm,string) determ (i,i,o).
prPredCondArgs : (sterml,string_list,string_list,string_list) determ (i,i,o,o).
prDEs : (sterml,sterml,sterml) determ (i,o,o).
prDE_Defs : (sterml,sterml,string_list) determ (i,i,o).
prDE_Pars : (sterml,string_list,string_list) determ (i,i,o).
prDERF : (sterm,string) determ (i,o).
prExists : (sterm,string) determ (i,o).
prFuncActPars : (sterml,string_list) determ (i,o).
prFuncBody : (sterml,string_list) determ (i,o).
prFuncCall : (sterm,string) determ (i,o).
prFuncStatement : (sterm,string) determ (i,o).
prObjConds : (sterml,string) determ (i,o).
prSSS : (sterm,string) determ (i,o).
prUsedDEs : (sterml,sterml,string_list) determ (i,i,o).
% ST-->PR
prSentenceCondition : (integer,sterm,string) determ (i,i,o).
prRemoveDummyConds : (sterm,sterm) procedure (i,o).

clauses
classInfo(className, classVersion).
% ------------------------------------------------------------------------------------------------------------
% prExpr;prExprs                                             
% ------------------------------------------------------------------------------------------------------------
prExprs([],[]) :-
	!.
prExprs([E|Es],[S|Ss]) :-
	prExpr(E,S),
	prExprs(Es,Ss),
	!.
		
prExpr(cmp("like",[L,R]),S) :-
	prExpr(L,S1),
	prExpr(R,S2),
	sx::concat(["like(",S1,",",S2,")"],S),
	!.	
prExpr(cmp("unlike",[L,R]),S) :-
	prExpr(L,S1),
	prExpr(R,S2),
	sx::concat(["unlike(",S1,",",S2,")"],S),
	!.	
prExpr(cmp(Opr,Params),S) :-
	Params=[_,_|_],	
	opr::opr_any(Opr,2),
	% infix operator with two or more operands
	prExprs(Params,SS2),
	sx::sxFrame(Opr," ",SOpr),
	sx::sxlJoin(SS2,SOpr,S3),
	sx::sxFrame(S3,"(",")",S),
	!.
prExpr(cmp(Opr,[Opn]),S) :-
	opr::opr_any(Opr,1),
	prExpr(Opn,S1),
	sx::concat(["(",Opr,S1,")"],S),
	!.	
prExpr(Opn,S) :- 
	prValue(Opn,S),
	!.
% ----------------------------------------------------------------------------------------------
% prFreeVars
% ----------------------------------------------------------------------------------------------
class predicates
prL_Vars : (sterm,string_list) procedure (i,o).
prL_VarsX : (sterml,string_list) determ (i,o).
prR_Vars : (sterm,string_list) procedure (i,o).
prR_VarsX : (sterml,string_list) determ (i,o).
prVars : (sterm,string_list) determ (i,o).
prVarsX : (sterml,string_list) determ (i,o).
clauses
%
% finds unassigned variables in the given predicate condition
%
prFreeVars(Term,FreeVars) :-
			tt::tt_term(2,"prFreeVars/e",Term),
	prL_Vars(Term,LVars),
	prR_Vars(Term,RVars),
	set::setRemSub(LVars,RVars,FreeVars),
			tt::tt_strings(2,"prFreeVars/x",FreeVars),!.
prL_Vars(cmp(Opr,[fvar(Var,_),_]),[Var]) :-
	opr::oprEq(Opr),!.	
prL_Vars(cmp(H,T),Vars) :-
	not(opr::oprEq(H)),
	prL_VarsX(T,Vars),!.
prL_Vars(_,[]) :- !.
prL_VarsX([],[]) :- !.
prL_VarsX([H|T],Vars) :-
	prL_Vars(H,Vars1),
	prL_VarsX(T,Vars2),
	set::conc([Vars1,Vars2],Vars),!.
prR_Vars(cmp(Opr,[_,RS]),Vars) :-
	opr::oprEq(Opr),
	prVars(RS,Vars),!.	
prR_Vars(cmp(H,T),Vars) :-
	not(opr::oprEq(H)),
	prR_VarsX(T,Vars),!.
prR_Vars(_,[]) :- !.
prR_VarsX([],[]) :- !.
prR_VarsX([H|T],Vars) :-
	prR_Vars(H,Vars1),
	prR_VarsX(T,Vars2),
	set::conc([Vars1,Vars2],Vars),!.
	
prVars(fvar(Var,_),[Var]) :- !.
prVars(cmp(_,T),Vars) :-
	prVarsX(T,Vars),!.
	
prVarsX([],[]) :- !.
prVarsX([H|T],Vars) :-
	prVars(H,Vars1),
	prVarsX(T,Vars2),
	set::conc([Vars1,Vars2],Vars),!.
% ------------------------------------------------------------------------------------------------------------
% prFunction
% ------------------------------------------------------------------------------------------------------------
% translates a function 3BQ tree into a PR
%
prFunction(Tree,S) :-
	Tree=cmp("FDEF",
	        [cmp("REL_TYPE",
	            [cmp("REP",[atom(Rep)]),
	             _,
	             _]),
	         cmp("APARS",Pars),
	         cmp("BODY",Body)]),
	sx::sxSetFirstUpper(Rep,FV),
	FVar=string::concat(FV,"__"),
	struc::treSubstString(cmp("BODY",Body),"P3_iThis__",FVar,Body2),
	Body2=cmp(_,Body3),
	prFuncBody(Body3,Statements),
	sx::sxlJoin(Statements,",",S1),
	%
	prFuncFormalPars(Pars,ParsS),
	sx::sxlJoin(ParsS,",",S2),
	sx::concat([Rep,"(",S2,"):-",S1],S),!
;
	dlgIF::feil("prFunction",["PR generation failed"],[],nill),
	tt::tt_term(1,"Error TBQ tree",Tree),!,fail.
% -----------------------------------------------------------------------------------------------------------
% prTbqQuery
% -----------------------------------------------------------------------------------------------------------
% translates a 3BQ tree into a PR (ref. prTbqQuery which translates a syntax tree directly)
%
prTbqQuery(Tree,S,"fq") :-
	prFuncCall(Tree,S),
	!.
prTbqQuery(Tree,S,"dq") :-
	Tree=cmp("NDERF",_),
	prDERF(Tree,S),
	!
;	
	Tree=cmp("DDERF",_),
	prDERF(Tree,S),
	!.
prTbqQuery(Tree,S,"sq") :-
	Tree=cmp("#S",_),
	prSSS(Tree,S),
	!.
prTbqQuery(Tree,S,"xq") :-
	Tree=cmp("EX",_),
	prExists(Tree,S),
	!.	 
prTbqQuery(Tree,_,_) :-
	dlgIF::feil("prTbqQuery",["PR generation failed"],[],nill),
	tt::tt_term(1,"Error TBQ tree",Tree),!,fail.
% -----------------------------------------------------------------------------------------------------------
% prPredicate
% -----------------------------------------------------------------------------------------------------------
class predicates
prEntity : (sterm,string) determ (i,o).
clauses

% gen. prolog for unary and binary relation facts with or without attributes
%
prPredicate(cmp("REL_COND",
               [cmp("REL_TYPE",
                   [cmp(_,[atom(Rep)]),
                    cmp(_,[int(1)]),
                    cmp(_,DEs)]),
                Ea,
                cmp("ATTR_CONDS",Vs)]),S) :-
	prEntity(Ea,S1),
	prDE_Defs(DEs,Vs,S2),
	sx::sxlJoin(S2,",",S3),
	sx::concat([Rep,"(",S1,",",S3,")"],S),!.
prPredicate(cmp("REL_COND",
               [cmp("REL_TYPE",
                   [cmp(_,[atom(Rep)]),
                    cmp(_,[int(1)])]),
                Ea]),S) :-
	prEntity(Ea,S1),
	sx::concat([Rep,"(",S1,")"],S),!.
prPredicate(cmp("REL_COND",
               [cmp("REL_TYPE",
                   [cmp(_,[atom(Rep)]),
                    cmp(_,[int(2)]),
                    cmp(_,DEs)]),
                Ea,
                Eb,
                cmp("ATTR_CONDS",Vs)]),S) :-
	prEntity(Ea,S1),
	prEntity(Eb,S4),
	prDE_Defs(DEs,Vs,S2),
	sx::sxlJoin(S2,",",S3),
	sx::concat([Rep,"(",S1,",",S4,",",S3,")"],S),!.
prPredicate(cmp("REL_COND",
               [cmp("REL_TYPE",
                   [cmp(_,[atom(Rep)]),
                    cmp(_,[int(2)])]),
                Ea,Eb]),S) :-
	prEntity(Ea,S1),
	prEntity(Eb,S2),
	sx::concat([Rep,"(",S1,",",S2,")"],S),!.
prEntity(cmp("ENT_ARG",
          [cmp("ENT_TYPE",
              [cmp(_,[atom(Rep)]),
               cmp(_,DEs)]),
           cmp("ATTR_CONDS",Vs)]),S) :-
	prDE_Defs(DEs,Vs,S1),
	sx::sxlJoin(S1,",",S2),
	sx::concat([Rep,"(",S2,")"],S),!.	             
% -----------------------------------------------------------------------------------------------------------
% prSynTreeQuery
% -----------------------------------------------------------------------------------------------------------
% translates a syntax tree query to PR (ref. prTbqQuery which translates 3BQ-tree)
%
prSynTreeQuery(N,ST,PR) :-
    prExists(ST,PR),
	!
;
	prSentenceQuery(N,ST,PR),
	!.
% -----------------------------------------------------------------------------------------------------------
% prRemoveDummyConds
% -----------------------------------------------------------------------------------------------------------
class predicates
prRemDC : (sterml,sterml) procedure (i,o).
clauses
prRemoveDummyConds(cmp(H,T),cmp(H,Tm)) :-
	prRemDC(T,Tm),
	!.
prRemoveDummyConds(C,C) :-
	!.
	
prRemDC([],[]) :-
	!.
prRemDC([H|T],Tm) :-
	H=cmp(Op,[fvar(_,_),fvar(Dummy,_)]),
	opr::oprEq(Op),
	sx::frontstr(6,Dummy,F,_),
	F="Dummy_",
	prRemDC(T,Tm),
	!.
prRemDC([cmp(C,Ct)|T],[cmp(C,Cm)|Tm]) :-
	prRemDC(Ct,Cm),
	prRemDC(T,Tm),
	!.
prRemDC(C,C) :-
	!.
% -----------------------------------------------------------------------------------------------------------
% prSentenceQuery
% -----------------------------------------------------------------------------------------------------------
prSentenceQuery(N,cmp("&",[C1,C2]),S) :-
	prSentenceQuery(N,C1,S1),
	M=N+1,
	prSentenceQuery(M,C2,S2),
	sx::concat([S1,",",S2],S),
	!.
prSentenceQuery(N,T,S) :-
    prRemoveDummyConds(T,Tr), % added 1.2.6 
	prSentenceCondition(N,Tr,S),
	!
;
	prExpr(T,S),
	!.		
% -------------------------------------------------------------------------
% prValue                                                               
% -------------------------------------------------------------------------
class predicates
prValueL : (sterml,string) determ (i,o).
str_fix : (string,real) procedure (o,i).
clauses

% translates from 3BQ tree node value format to PR string value format
%
prValue(atom(A),A) :- 
	!.
prValue(fstr(S,"p3_dif",_,_),S) :- 
	!.
prValue(fstr(S,"p3_dnu",_,_),V) :-
	sx::sxRemChar(S,',',V),
	!.
prValue(fstr(S,"p3_din",_,_),V) :-
	sx::sxRemChar(S,',',V),
	!.
prValue(fstr(S1,_,_,_),S) :- 
	sx::sxQuote(S1,S),
	!.
prValue(fvar(V,_),V)  :- 
	!.
prValue(fint(I,_),S)  :- 
	I>=0,
	S=toString(I),
	!
;
	S1=toString(I),
	sx::concat(["(",S1,")"],S),
	!.
prValue(ffix(F,_),S)  :- 
	F>=0,
	str_fix(S,F),
	!
;
	str_fix(S1,F),
	sx::concat(["(",S1,")"],S),
	!.	
prValue(str(V),S) :-
	sx::sxQuote(V,S),
	!.	
prValue(var(V),V)  :- 
	!.
prValue(int(I),S)  :- 
	I>=0,
	S=toString(I),
	!
;
	S1=toString(I),
	sx::concat(["(",S1,")"],S),
	!.
prValue(fix(F),S)  :- 
	F>=0,
	S=toString(F),
	!
;
	S1=toString(F),
	sx::concat(["(",S1,")"],S),
	!.	
prValue(cmp(Range,[RangeType,Dom,Low,High]),S)  :-
   set::member(Range,["insidex","outsidex"]),
   prValue(Low,S1),
   prValue(High,S2),
   prValue(RangeType,S3),
   prValue(Dom,S4),
   S=string::concatList([Range,"(",S3,",",S4,",",S1,",",S2,")"]),
   !.
   
prValue(cmp(ListDom,List),S) :-
	sx::sxTail(2,ListDom,_,T),
	T="_l",
	prValueL(List,S1),
	sx::concat(["[",S1,"]"],S),!.
prValue(T,S) :- 
	prFuncCall(T,S),!.		
prValue(T,_) :- 
	dlgIF::feil("prValue",[],[],T),!,fail.
prValueL([],"") :- 
	!.
prValueL([H],S) :- 
	prValue(H,S),!.
prValueL([H|T],S) :- 
	prValue(H,S1), 
	prValueL(T,S2), 
	sx::concat([S1,",",S2],S),!.
	
str_fix(S,F) :-
    S=toString(F),
    sx::searchchar(S,'e',_),
    !
;    
	S=toString(F),
	sx::searchchar(S,'.',_),
	!
;
	S1=toString(F),
	S=string::concat(S1,".0"),
	!.
% ----------------------------------------------------------------------------------------------
% prValueList
% ----------------------------------------------------------------------------------------------
class predicates
prValueListX : (sterm,string_list) determ (i,o).
clauses
prValueList(L,S) :-
	prValueListX(L,Ls),
	sx::sxlJoin(Ls,",",S2),
	sx::concat(["[",S2,"]"],S),
	!.
	
prValueListX(nill,[]) :-
	!.
prValueListX(list(V,L),[Vs|Ls]) :-
	prValue(V,Vs),
	prValueListX(L,Ls),
	!.
% ----------------------------------------------------------------------------------------------
% prValFindVar
% ----------------------------------------------------------------------------------------------
class predicates
prValFindVarX : (string,sterml,string) determ (i,i,o).
clauses
%
% finds the DE-var corresponding to DE (par#1) in par#2 (returned as par#3)
%
prValFindVar(DE,fstr(DE_Var,_,_,_),DE_Var) :-
	vars::varDE_Var(DE_Var,DEx),
	DE=DEx,!.
prValFindVar(DE,cmp(_,L),DE_Var) :-
	prValFindVarX(DE,L,DE_Var),!.
prValFindVarX(DE,[H|T],DE_Var) :-
	prValFindVar(DE,H,DE_Var),!
;	
	prValFindVarX(DE,T,DE_Var),!.
% -----------------------------------------------------------------------------------------------------------
%                                 LOCAL class predicates                          
% -----------------------------------------------------------------------------------------------------------
% prDEs
% -----------------------------------------------------------------------------------------------------------
class predicates
prPredCondArg : (sterm,sterml,sterml) determ (i,o,o).
clauses
% parameters:
% 1	in	argument-part of sentence condition
% 2	out	DEs in p#1
% 3	out	conditions in p#1
%
prDEs([],[],[]):-
	!.
prDEs([var(ArgVar)|Args],[var(ArgVar)|DEs],Conds) :-
	prDEs(Args,DEs,Conds),
	!.
prDEs([Arg|Args],DEs,Conds) :-
	prPredCondArg(Arg,DEs2,Conds2),
	prDEs(Args,DEs3,Conds3),
	set::conc([DEs2,DEs3],DEs),
	set::conc([Conds2,Conds3],Conds),
	!.
prPredCondArg(cmp("ATTR_CONDS",Conds),[],Conds):-
	!.
prPredCondArg(Arg,DEs,Conds):-
	Arg=cmp("ENT_ARG",
	       [cmp("ENT_VAR",
                   [cmp("REP",_),
                    cmp("ATTR_VARS",DEs)])]),
        Conds=[],
	!
;	
	Arg=cmp("ENT_ARG",
	       [cmp("ENT_VAR",
                   [cmp("REP",_),
                    cmp("ATTR_VARS",DEs)])
               |DE_Conds]),
        DE_Conds=[DE_Cond],
	prPredCondArg(DE_Cond,_,Conds),
	!.
% -----------------------------------------------------------------------------------------------------------
% prDE_Defs
% -----------------------------------------------------------------------------------------------------------
class predicates
prDE_Get : (string,sterml,sterm) determ (i,i,o).
clauses

prDE_Defs([],_,[]) :- 
	!.
prDE_Defs([fstr(DE,_,_,_)|DEs],Conds,[S1|S]) :- 
	prDE_Get(DE,Conds,V),
	prValue(V,S1),
	prDE_Defs(DEs,Conds,S),
	!.
prDE_Get(DE,[cmp(Eq,[fvar(DE,_),V])|_],V) :-
	opr::oprEq(Eq),
	!.
prDE_Get(DE,[_|Conds],V) :-
	prDE_Get(DE,Conds,V),
	!.
prDE_Get(DE,_,_) :-
	dlgIF::feil("prDE_Get",[DE],[],nill),!,fail.
% -----------------------------------------------------------------------------------------------------------
% prDE_Pars
% -----------------------------------------------------------------------------------------------------------
prDE_Pars([],_,[]) :-
	!.
prDE_Pars([H|DEs],UsedDEs,[S|SArg]) :-
	H=fstr(DE,_,_,_),	
	set::member(DE,UsedDEs),
	S=DE,
	prDE_Pars(DEs,UsedDEs,SArg),
	!
;
	H=fstr(_,_,_,_),	
	S="_",
	prDE_Pars(DEs,UsedDEs,SArg),
	!
;
	H=cmp(_,MDEs),
	prDE_Pars(MDEs,UsedDEs,ML),
	set::setRemSub(["_"],ML,MLR),
	set::size(MLR,N),
	N=0,
	S="_",
	prDE_Pars(DEs,UsedDEs,SArg),
	!.
% ------------------------------------------------------------------------------------------------------------
% prDERF                                            
% ------------------------------------------------------------------------------------------------------------
prDERF(cmp("DDERF",[fstr(DERF,_,_,_),PredCond]),S) :-
	vars::varGen(V),
	prPredCond(DERF,PredCond,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("e",T1),
	sx::concat(["p3_iDdrf(",V,",",quote,DERF,quote,",",quote,S2,quote,")"],S),
	!.
prDERF(cmp("NDERF",[fstr(DERF,_,_,_),PredCond]),S) :-
	vars::varGen(V),
	prPredCond(DERF,PredCond,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("n",T1),
	sx::concat(["p3_iNdrf(",V,",",quote,DERF,quote,",",quote,S2,quote,")"],S),
	!.
% ------------------------------------------------------------------------------------------------------------
% prExists                         
% ------------------------------------------------------------------------------------------------------------
prExists(cmp("EX",[PredCond]),S) :-
	prPredCond("",PredCond,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("x",T1),
	sx::concat(["p3_iExis(",quote,S2,quote,")"],S),
	!.
% ------------------------------------------------------------------------------------------------------------
% prSSS                                     
% ------------------------------------------------------------------------------------------------------------
prSSS(cmp("#S",[PredCond]),S) :-
	vars::varGen(V),
	prPredCond("",PredCond,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("s",T1),
	sx::concat(["p3_iSSS(",V,",",quote,S2,quote,")"],S),
	!.
% ------------------------------------------------------------------------------------------------------------
% prFuncActPars
% ------------------------------------------------------------------------------------------------------------
prFuncActPars([],[]) :- 
	!.
prFuncActPars([Par|Pars],[V|Vs]) :-
	prValue(Par,V),
	prFuncActPars(Pars,Vs),
	!.
% ------------------------------------------------------------------------------------------------------------
% prFuncBody
% ------------------------------------------------------------------------------------------------------------
% translates the body of a function predicate body.
%
prFuncBody([],[]) :- 
	!.
prFuncBody([Statement|Statements],[S|Ss]) :-
	prFuncStatement(Statement,S),
	prFuncBody(Statements,Ss),
	!.	
% ------------------------------------------------------------------------------------------------------------
% prFuncCall
% ------------------------------------------------------------------------------------------------------------
prFuncCall(T,S) :-
	T=cmp("DFNC",Def),
	Def=[atom(Func),_|Pars],
	vars::varGen(V),
	prFuncActPars(Pars,S1),
	sx::sxlJoin([V|S1],",",S2),
	sx::concat([Func,"(p3_iDtFn,",S2,")"],S),!
;	
	T=cmp("NFNC",Def),
	Def=[atom(Func),_|Pars],
	vars::varGen(V),
	prFuncActPars(Pars,S1),
	sx::sxlJoin([V|S1],",",S2),
	sx::concat([Func,"(",S2,")"],S),!.
% ------------------------------------------------------------------------------------------------------------
% prFuncFormalPars
% ------------------------------------------------------------------------------------------------------------
prFuncFormalPars([],[]) :- 
	!.
prFuncFormalPars([fvar(V,_)|Pars],[V|Vs]) :-
	prFuncFormalPars(Pars,Vs),!.
% ------------------------------------------------------------------------------------------------------------
% prFuncStatement
% ------------------------------------------------------------------------------------------------------------
% statement in function.
% note that function calls and derfs only can occurr as individual is-statements
%
prFuncStatement(atom("p3_iNdFn"),"") :- 
    !.
prFuncStatement(cmp(Opr,[V1,V2]),S) :-
	vars::varStr(V1,V1s),
	vars::varStr(V2,V2s),
	opr::opr_cmpass(Opr),
	sx::concat([V1s," ",Opr," ",V2s],S),
	!.
prFuncStatement(cmp(Opr,[V,cmp("NFNC",[atom(Func),_|Pars])]),S) :-
	opr::oprEq(Opr),
	vars::varStr(V,Vs),
	prFuncActPars(Pars,S1),
	sx::sxlJoin([Vs|S1],",",S2),
	sx::concat([Func,"(",S2,")"],S),
	!.
prFuncStatement(cmp(Opr,[V,cmp("DFNC",[atom(Func),_|Pars])]),S) :-
	opr::oprEq(Opr),
	vars::varStr(V,Vs),
	prFuncActPars(Pars,S1),
	sx::sxlJoin([Vs|S1],",",S2),
	sx::concat([Func,"(p3_iDtFn,",S2,")"],S),
	!.
prFuncStatement(cmp(Opr,[V,cmp("DDERF",[fstr(DE_Var,_,_,_),RelVar])]),S) :-
	opr::oprEq(Opr),
	vars::varStr(V,Vs),
	prPredCond(DE_Var,RelVar,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("e",T1),
	sx::concat(["p3_iDdrf(",Vs,",",quote,DE_Var,quote,",",quote,S2,quote,")"],S),
	!.
prFuncStatement(cmp(Opr,[V,cmp("NDERF",[fstr(DE_Var,_,_,_),RelVar])]),S) :-
	opr::oprEq(Opr),
	vars::varStr(V,Vs),
	prPredCond(DE_Var,RelVar,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("n",T1),
	sx::concat(["p3_iNdrf(",Vs,",",quote,DE_Var,quote,",",quote,S2,quote,")"],S),
	!.
prFuncStatement(cmp(Opr,[V,cmp("#S",[PredCond])]),S) :-
	opr::oprEq(Opr),
	vars::varStr(V,Vs),
	prPredCond("",PredCond,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("s",T1),
	sx::concat(["p3_iSSS(",Vs,",",quote,S2,quote,")"],S),
	!.
prFuncStatement(cmp("EX",[PredCond]),S) :-
	prPredCond("",PredCond,S1),
	transpr::proPR_Term(S1,T1),
	S2=termDomain::toTermStringIndirect("x",T1),
	sx::concat(["p3_iExis(",quote,S2,quote,")"],S),
	!.
prFuncStatement(cmp("FALSE",[St]),S) :-
	prFuncStatement(St,S1),
	sx::concat(["not(",S1,")"],S),
	!.
prFuncStatement(cmp("like",[V,T]),S) :-
	vars::varStr(V,Vs),
	prExpr(T,S1),
	sx::concat(["like(",Vs,",",S1,")"],S),
	!.
prFuncStatement(cmp("unlike",[V,T]),S) :-
	vars::varStr(V,Vs),
	prExpr(T,S1),
	sx::concat(["unlike(",Vs,",",S1,")"],S),
	!.
prFuncStatement(cmp(Opr,[V,T]),S) :-
	vars::varStr(V,Vs),
	prExpr(T,S1),
	sx::concat([Vs," ",Opr," ",S1],S),
	!.
prFuncStatement(T,_) :- 
    dlgIF::feil("prFuncStatement",[],[],T),!,fail.
% ------------------------------------------------------------------------------------------------------------
% prObjConds                                               
% ------------------------------------------------------------------------------------------------------------
prObjConds([C],S) :- 
	prFuncStatement(C,S),!
;	
	prExpr(C,S),!.
prObjConds([Exp|T],S) :- 
	prExpr(Exp,S1),
	prObjConds(T,S2),
	sx::concat([S1,",",S2],S),!.
% -----------------------------------------------------------------------------------------------------------
% prPredCond
% -----------------------------------------------------------------------------------------------------------
prPredCond(DE,C,S) :- 
	C=cmp("REL_COND",
	     [cmp("REL_VAR",
	         [cmp("REP",[atom(P)]),
	          _,
	          cmp("ATTR_VARS",PDEs)])
	     |Args]),
	prDEs(Args,DEs2,Conds),
	set::conc([PDEs,DEs2],DEs),
	prUsedDEs(DEs,Conds,UsedDEs2),
	UsedDEs=[DE|UsedDEs2],	
	prDE_Pars(PDEs,UsedDEs,SDEs),
	prPredCondArgs(Args,UsedDEs,SArgs,SConds),
	set::conc([SArgs,SDEs],S2),
	sx::sxlJoin(S2,",",S3),
	sx::concat([P,"(",S3,")"],S4),
	sx::sxlJoin([S4|SConds],",",S),
	!
;
	C=cmp("REL_COND",
	     [cmp("REL_VAR",
	         [cmp("REP",[atom(P)]),
	          _])
	     |Args]),
	prDEs(Args,DEs,Conds),
	prUsedDEs(DEs,Conds,UsedDEs2),	
	UsedDEs=[DE|UsedDEs2],	
	prPredCondArgs(Args,UsedDEs,SArgs,SConds),
	sx::sxlJoin(SArgs,",",S3),
	sx::concat([P,"(",S3,")"],S4),
	sx::sxlJoin([S4|SConds],",",S),
	!.
% -----------------------------------------------------------------------------------------------------------
% prPredCondArgs
% -----------------------------------------------------------------------------------------------------------
class predicates
prArg : (sterm,string_list,string,string) determ (i,i,o,o).
clauses
prPredCondArgs([],_,[],[]):-
	!.
prPredCondArgs([Arg|Args],UsedDEs,[SArg2|SArgs],[SCond2|SConds]) :-
	prArg(Arg,UsedDEs,SArg2,SCond2),
	prPredCondArgs(Args,UsedDEs,SArgs,SConds),
	!.
	
prArg(var(ArgVar),_,ArgVar,""):-
	!.
prArg(cmp("ATTR_CONDS",Conds),_,"",S):-
	prObjConds(Conds,S),
	!.
prArg(Arg,UsedDEs,SArg,SCond):-
	Arg=cmp("ENT_ARG",
	       [cmp("ENT_VAR",
                   [cmp("REP",[atom(E)]),
                    cmp("ATTR_VARS",DEs)])
               |DE_Conds]),
        prPredCondArgs(DE_Conds,UsedDEs,_,SCond2),
	prDE_Pars(DEs,UsedDEs,SDEs),
	sx::sxlJoin(SDEs,",",S),
	sx::concat([E,"(",S,")"],SArg),
	sx::sxlJoin(SCond2,",",SCond),
	!.
% -----------------------------------------------------------------------------------------------------------
% prUsedDEs
% -----------------------------------------------------------------------------------------------------------
prUsedDEs([],_,[]) :-
	!.
prUsedDEs([fstr(DE,Dom,_,_)|DEs],Conds,[DE|UsedDEs]) :-
	struc::treHasNode(cmp("",Conds),fvar(DE,Dom)),	
	prUsedDEs(DEs,Conds,UsedDEs),
	!.
prUsedDEs([_|DEs],Conds,UsedDEs) :-
	prUsedDEs(DEs,Conds,UsedDEs),
	!.
% -----------------------------------------------------------------------------------------------------------
% prSentenceCondition
% -----------------------------------------------------------------------------------------------------------
class predicates
prArgument : (integer,sterm,string,string) determ (i,i,o,o).
prArgVar : (integer,string,sterml,string) determ (i,i,i,o).
prArgDEs : (integer,string,sterml,string_list) determ (i,i,i,o).
prCond : (sterm,string) determ (i,o).	
prConds : (sterml,string_list) determ (i,o).	
prReplaceDEs : (integer,string,sterml,sterml,sterml) determ (i,i,i,i,o).
clauses

% ST --> PR
%
prSentenceCondition(N,ST,PR) :- 
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",
	          [cmp("REP",[atom(P)]),
	           cmp("CARD",[int(2)])
	          |DEs]),
	       ArgA,
	       ArgB
	      |Conds]),
	Na=N*10+1,
	Nb=Na+1,
	prArgument(Na,ArgA,Sea,Cea),
	prArgument(Nb,ArgB,Seb,Ceb),      
	Np=N*10,
	prReplaceDEs(Np,P,DEs,Conds,CondsM),
	prConds(CondsM,Cp),
	prArgDEs(Np,P,DEs,L),
	sx::sxlJoin([Sea,Seb|L],",",S2),
	sx::concat([P,"(",S2,")"],S3),
	sx::sxlJoin([S3,Cea,Ceb|Cp],",",PR),
	!
;
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",
	          [cmp("REP",[atom(P)]),
	           cmp("CARD",[int(1)])
	          |DEs]),
	       ArgA
	      |Conds]),
	Na=N*10+1,
	prArgument(Na,ArgA,Sea,Cea),
	Np=N*10,
	prReplaceDEs(Np,P,DEs,Conds,CondsM),
	prConds(CondsM,Cp),
	prArgDEs(Np,P,DEs,L),
	sx::sxlJoin([Sea|L],",",S2),
	sx::concat([P,"(",S2,")"],S3),
	sx::sxlJoin([S3,Cea|Cp],",",PR),
	!.
	
prArgument(_,var(V),V,"") :-
	!.
prArgument(N,C,Sv,Sc) :-
        C=cmp("ENT_ARG",
             [cmp("ENT_TYPE",
                 [cmp("REP",[atom(K)]),
                  cmp("ATTRS",DEs)])
             |Conds]),
	prArgVar(N,K,DEs,Sv),
	prReplaceDEs(N,K,DEs,Conds,CondsM),	
	prConds(CondsM,L),
	sx::sxlJoin(L,",",Sc),
	!.
prArgVar(N,K,DEs,S) :-	
	prArgDEs(N,K,DEs,L),
	sx::sxlJoin(L,",",S2),
	sx::concat([K,"(",S2,")"],S),
	!.
		      
prArgDEs(_,_,[],[]) :-
	!.
prArgDEs(N,K,[cmp("ATTRS",DEs)],Vs) :-
	prArgDEs(N,K,DEs,Vs),
	!.
prArgDEs(N,K,[fstr(D,_,_,_)|DEs],[V|Vs]) :-
	vars::varCompsR(V,K,N,D),
	prArgDEs(N,K,DEs,Vs),
	!.      

prConds([],[]) :-
	!.
prConds([C|Cs],[S|T]) :-
	prCond(C,S),
	prConds(Cs,T),
	!.

prCond(C,PR) :-
	prFuncCall(C,PR),
	!
;
	prExpr(C,PR),
	!.

prReplaceDEs(_,_,[],Conds,Conds) :-
	!.
prReplaceDEs(N,K,[cmp("ATTRS",DEs)],Conds,CondsM) :-
	prReplaceDEs(N,K,DEs,Conds,CondsM),
	!.
prReplaceDEs(N,K,DEs,[cmp("ATTR_CONDS",Conds)],CondsM) :-
	prReplaceDEs(N,K,DEs,Conds,CondsM),
	!.
prReplaceDEs(N,K,[fstr(DE,Dom,_,_)|DEs],Conds,CondsM) :-
	vars::varCompsR(V,K,N,DE),
	struc::treReplace(cmp("",Conds),fvar(DE,Dom),var(V),T),
	T=cmp("",Conds2),
	prReplaceDEs(N,K,DEs,Conds2,CondsM),
	!.
end implement pr
