/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement proc
open core,resourceIdentifiers,termDomain,globalDomains

constants
className = "com/rudstrand/proc/proc".
classVersion = "1.0".

class predicates
addStream : (odbcConnection,string,string,string,string,string,count).
adjustList : (string_list,string,string_list) determ (i,i,o).
preProcess : (string,string) determ (i,o).
pie3_Exec : (integer,sterm) determ.
pro_type : (sterm,string) procedure (i,o).
runQueriesNL : (string) determ.
runQueriesNL : (outFile,string) determ.
runQueriesPR : (string) determ.
runQueriesPR : (outFile,string) determ.
procXml : (integer,integer,string,string File) determ.
proQueryResult : (integer,sterm,integer,string) determ.
proQueryResult : (odbcConnection,string,integer,sterm,integer,string) determ.
proQueryResult_F : (outFile,string,integer,sterm,integer) determ.
proQuerySolGen : (integer,integer,sterm) determ.
proQuerySolGen : (odbcConnection,string,integer,integer,sterm) determ.
proQuerySolGen_F : (outFile,string,integer,integer,sterm) determ.
proConstructSol : (outfile,string,integer,integer,sterm).
proConstructSolXML : (outfile,string,integer,integer,sterm).
proConstructSolODBC : (odbcConnection,string,integer,integer,sterm) determ.
proc3dl : (string) determ.
proc3 : (integer,integer,string,string,string,string) determ.
delScan : (string,string,string,sterml,sterml,sterml,string,fileSeg,string,count) determ.
delInput : (string_list,string,string,fileSeg,count).
delGenPR : (string,string,string,sterml,sterml,sterml,string_list,string) determ (i,i,i,i,i,i,i,o).
delReadVals : (sterml,string_list,string_list) determ (i,i,o).
writeQuery : (outFile,string).

createSQL_Table : (odbcConnection,integer,integer,integer,string,integer,integer,string_list,string,integer) procedure (i,i,i,i,o,o,o,o,o,o).
popSQL_Table : (odbcConnection,integer,integer,string,integer,integer,string_list) determ.
writeSqlLines : (odbcConnection,integer,string_list) determ.

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% processFiles
% ----------------------------------------------------------------------------------------------
% processes a set of input files of any type
%
processFiles([]) :-
    !.
processFiles([F|Fs]) :-
    Ext=fileName::getExtension(F),
    processFile(0,Ext,F,""),
    processFiles(Fs),
    !.
    
% parameters:
% 1 in 0:trans. to IF and store in KB, >0:store in knowledge stream (2=sentence,3=SR,4=FD,6=IR)
% 2 in file extension
% 3 in complete file path/name
% 4 in sql qualified table name (only if p#1 is 2)
%
processFile(K_Type,Ext,F,Table) :-
    LExt=string::toLowercase(Ext),
    LExt="xml",
    procXml(10,K_Type,Table,F),
    Mess=string::concat("Processed file ",F,"!"),
    dlgIF::msgS(Mess),
    !
;
    LExt=string::toLowercase(Ext),
    LExt="3dl",
    proc3dl(F),
    Mess=string::concat("Processed file ",F,"!"),
    dlgIF::msgS(Mess),
    !
;
    LExt=string::toLowercase(Ext),
    LExt="3nl",
    preProcess(F,Ft),
    proc3(1,K_Type,LExt,F,Ft,Table),
    Mess=string::concat("Processed file ",F,"!"),
    dlgIF::msgS(Mess),
    !
;
    LExt=string::toLowercase(Ext),
    LExt="3pr",
    preProcess(F,Ft),
    proc3(9,K_Type,LExt,F,Ft,Table),
    Mess=string::concat("Processed file ",F,"!"),
    dlgIF::msgS(Mess),
    !
;
    Mess=string::concat("Cannot process files of type: ",Ext,"!"),
    dlgIF::dlgNote(1,Mess),
    !.
% ----------------------------------------------------------------------------------------------
% processWin/Wins
% ----------------------------------------------------------------------------------------------
processWins() :-
    W=guix::getWinOfClas(0),
    processWin(W),
    fail
;
    !.
% processes an input window of any type excl. 3dl, but incl. queries
%
processWin(W) :-
	N=globalVars::get_i("error_flag"),
	M=globalVars::get_i("quit_flag"),
	L=M+N,
	L>0,
	dlgIF::dlgNote(1,"Cannot process knowledge while Error or Quit-flag is ON. Reset flags ([X]-button in the toolbar) or exit and re-start Pro/3!"),
	!
;
    % PR queries from editor win - one output file
    Type=guix::getWinClas(W),
    Type=0,
    Opt=globalVars::get_i("f_outq"),
    set::member(Opt,[-1,-2,-3,-4,-9,-10,-11]),
    S=vpiEditor::getText(W),
    S2=string::trim(S),
	sx::sxStripComments(S2,S3),
    sx::sxTail(1,S3,_,Tail),
    Tail="?",
	FileName=guix::getWinName(W),
	FType=filename::getExtension(FileName),
	LType=string::toLowerCase(FType),
	LType="3pr",
	Name=filename::getName(FileName),
    P=globalVars::get_s("std_path"),
    fioExt::file_type(Opt,Ext),
    QName=string::concat(P,"\\Q_",Name,".",Ext),
    File=outFile::new(QName,Opt),
	runQueriesPR(File,S3),
	File:close(),
	editorWindow::openFiles([QName]),
	!
;
    % PR queries from editor win - several output files
    Type=guix::getWinClas(W),
    Type=0,
    S=vpiEditor::getText(W),
    S2=string::trim(S),
	sx::sxStripComments(S2,S3),
    sx::sxTail(1,S3,_,Tail),
    Tail="?",
	FileName=guix::getWinName(W),
	FType=filename::getExtension(FileName),
	LType=string::toLowerCase(FType),
	LType="3pr",
	runQueriesPR(S3),
	!
;
    % NL queries from editor win - one output file
    Type=guix::getWinClas(W),
    Type=0,
    Opt=globalVars::get_i("f_outq"),
    set::member(Opt,[-1,-2,-3,-4,-9,-10,-11]),
    S=vpiEditor::getText(W),
    S2=string::trim(S),
	sx::sxStripComments(S2,S3),
    sx::sxTail(1,S3,_,Tail),
    Tail="?",
	FileName=guix::getWinName(W),
	FType=filename::getExtension(FileName),
	LType=string::toLowerCase(FType),
	LType="3nl",
	Name=filename::getName(FileName),
    P=globalVars::get_s("std_path"),
    fioExt::file_type(Opt,Ext),
    QName=string::concat(P,"\\Q_",Name,".",Ext),
    File=outFile::new(QName,Opt),
	runQueriesNL(File,S3),
	File:close(),
	editorWindow::openFiles([QName]),
	!
;
    % NL queries from editor win - several output files
    Type=guix::getWinClas(W),
    Type=0,
    S=vpiEditor::getText(W),
    S2=string::trim(S),
	sx::sxStripComments(S2,S3),
    sx::sxTail(1,S3,_,Tail),
    Tail="?",
	FileName=guix::getWinName(W),
	FType=filename::getExtension(FileName),
	LType=string::toLowerCase(FType),
	LType="3nl",
	runQueriesNL(S3),
	!
;
    % xml
  	guix::getWinClas(W)=0,
  	FileName=guix::getWinName(W),
	FileName<>"",
	fio::file_name(FileName,_,Name,Type),
	LType=string::toLowerCase(Type),
	LType="xml",
	!,
	S=vpiEditor::getText(W),
	sx::concat([Name,".",Type],Realm),
    dep::massSentenceUpdateStart,
	xmlP3::xmlProcess(Realm,S),
    dep::massSentenceUpdateEnd,
	!
;	
    % NL/PR
    Type=guix::getWinClas(W),
    Type=0,
	FileName=guix::getWinName(W),
	fio::file_name(FileName,_,Name,FType),
	sx::concat([Name,".",FType],Realm),
	dbBase::realmExists(Realm,Files),
	set::size(Files,N),
	N>0,
	!,
	proCheckRealm(Files,Realm),
	processWin(W),
	!
;	
    % NL/PR
    Type=guix::getWinClas(W),
    Type=0,
	FileName=guix::getWinName(W),
	fio::file_name(FileName,_,Name,FType),
	LType=string::toLowerCase(FType),
	sx::concat([Name,".",FType],Realm),
	FRealm=string::toUpperCase(Realm),
    S=vpiEditor::getText(W),
    S2=string::trim(S),
	sx::sxStripComments(S2,S3),
    sx::sxUnpack(S3,'!',L),
    adjustList(L,"!",La),
    dep::massSentenceUpdateStart,
    C=count::new(0),
	proText(LType,FRealm,La,C),
    dep::massSentenceUpdateEnd,
	!
;
    % assumes (NL) query from asst. window (one query only)
    globalVars::get_i("error_flag")=0,
    globalVars::get_i("quit_flag")=0,
    AW=guix::getWinOfClas(4),
	guix::lbtGet(AW,idc_query,S),
	sx::sxStrip(S,S2),
	runQueryNL(S2),
	guix::lbxGetAll(AW,idc_query,L),
	set::distinct([S|L],Ld),
	set::trunc(30,Ld,Ldt),
	guix::lbxPop(AW,idc_query,Ldt),
	guix::lbtSelectFirst(AW,idc_query),
	!.
% -------------------------------------------------------------------------------------------------------------
% showResponse
% -------------------------------------------------------------------------------------------------------------
% parameters:
% 1 in  -1=query-set; 1=single query
% 2 in  outfile
%
showResponse(Single,File) :-
    Single=-1,
    N=File:getPar(),
    set::member(N,[-3,-4]),
    Name=File:getName(),
    File:close(),
    editorWindow::openFiles(-1,[Name]),
    !
;
    Single=1,
    N=File:getPar(),
    set::member(N,[3,4]),
    Name=File:getName(),
    File:close(),
    editorWindow::openFiles(-1,[Name]),
    !
;
    Single=-1,
    N=File:getPar(),
    set::member(N,[-1,-2,-9,-10,-11]),
    Name=File:getName(),
    File:close(),
    editorWindow::openFiles([Name]),
    !
;
    Single=1,
    N=File:getPar(),
    set::member(N,[1,2,9]),
    Name=File:getName(),
    File:close(),
    editorWindow::openFiles([Name]),
    !
;
    Name=File:getName(),
    File:close(),
    file::delete(Name),
    !.
% -------------------------------------------------------------------------------------------------------------
% menuQueryRun
% -------------------------------------------------------------------------------------------------------------
class predicates
menuQueryRun2 : (string,string,string) determ.
queryRun : (string,string) determ.
querySetPar : (string,string,string,string) procedure (i,i,i,o).
queryGetPar : (string,string,string,string) procedure (i,i,i,o).
queryCluster : (string,string,integer_list,string_list,string_list) determ.
setPars : (string,string,string_list,string_list) procedure (i,i,i,o).		
substPars : (string,string_list,string_list,string) determ (i,i,i,o).    
clauses

% runs one query-meny query or cluster
%
menuQueryRun(MenuId) :- 
	dbExtension::kbQuery(MenuId,Name,Frmat,Query,_),	
	Query2=sx::replaceAllC(Query,'`','"'),
	menuQueryRun2(Name,Frmat,Query2),
	!
;
	globalVars::get_i("quit_flag")=1,
	dlgIF::dlgNote(0,"QUIT has been requested!"),
	!
;
	dbExtension::kbQuery(MenuId,_,_,_,_),
	S1=toString(MenuId),
	Mess=string::concat("Unable to process query '",S1,"'!"),
	dlgIF::dlgNote(1,Mess),
	!
;	
	S1=toString(MenuId),
	Mess=string::concat("'",S1,"' is not a known query ID!"),
	dlgIF::dlgNote(1,Mess),
	!.

menuQueryRun2(Name,Frmat,Query) :-
    sx::searchchar(Query,'[',P1),
    sx::searchchar(Query,']',P2),
    P1<P2, % has parameters
	querySetPar(Name,Frmat,Query,Query2),
	globalVars::get_i("quit_flag")=0,
	queryRun(Frmat,Query2),
    !
;    
    sx::sxUnpack(Query,';',QNos),
    checkClusteredQueries(QNos,Nos,Pars),
    setPars(Name,Frmat,Pars,Vals),
	globalVars::get_i("quit_flag")=0,
    queryCluster(Name,Frmat,Nos,Pars,Vals),
    !  
;
	globalVars::get_i("quit_flag")=0,
    queryRun(Frmat,Query),
    !.

queryRun("p3_iPR",Query) :- 
    runQueryPR(Query),
    !.
queryRun("p3_iNL",Query) :- 
    runQueryNL(Query),
    !.

% prompts the KE for query parameters (if any) and inserts the responses into the query
%	
querySetPar(Name,Form,Qc,Qr) :-
    sx::sxGetInBrack(Qc,'[',']',Par),
    queryGetPar(Name,Form,Par,APar),
    globalVars::get_i("quit_flag")=0,
    P=string::concat("[",Par,"]"),
	Q2=sx::replaceAll(Qc,P,APar),
    querySetPar(Name,Form,Q2,Qr),
    !
;
    Qr=Qc,
    !.

queryGetPar(Name,Form,Par,APar) :- 
	dbExtension::kbQueryPar(Par,Dom,Title,Alts,_),
    Alts=[],   
	W=guix::getWinOfClas(1),
    Dlg=queryparameter::new(),
    dbBase::nlSing(Dom,DomNL),
    dbBase::nlSing(Form,FormNL),
    Dlg:show(W,Name,DomNL,FormNL,Title),
    globalVars::get_i("quit_flag")=0,
    Vals=globalVars::popRetPars(),
    Vals=[APar],
    !
;    
	dbExtension::kbQueryPar(Par,Dom,Title,Alts,_),
    set::size(Alts,N),
    N>0, 
	W=guix::getWinOfClas(1),
    Dlg=querylistparameter::new(),
    dbBase::nlSing(Dom,DomNL),
    dbBase::nlSing(Form,FormNL),
    Dlg:show(W,Name,DomNL,FormNL,Title,Alts),
    globalVars::get_i("quit_flag")=0,
    Vals=globalVars::popRetPars(),
    Vals=[APar],
    !  
;
    APar="",
	!.

queryCluster(_,_,[],_,_) :-
    !.
queryCluster(Name,Frmat,[N|Ns],Pars,Vals) :-
	dbExtension::kbQuery(N,_,_,Query,_),
	Query2=sx::replaceAllC(Query,'`','"'),
	substPars(Query2,Pars,Vals,Q),
	queryRun(Frmat,Q),
	queryCluster(Name,Frmat,Ns,Pars,Vals),
	!.

setPars(_,_,[],[]) :-
    !.
setPars(Name,Form,[Par|Pars],Vals) :-
    queryGetPar(Name,Form,Par,Val),
    globalVars::get_i("quit_flag")=0, 
    setPars(Name,Form,Pars,Vals2),
    Vals=[Val|Vals2],
    !
;
    Vals=[],
    !.

substPars(Q,[],_,Q) :-
    !.
substPars(Q,[P|Ps],[V|Vs],Qp) :-
    S=string::concat("[",P,"]"),
	Q2=sx::replaceAll(Q,S,V),
    substPars(Q2,Ps,Vs,Qp),
    !.
% -------------------------------------------------------------------------------------------------------------
% checkClusteredQueries
% -------------------------------------------------------------------------------------------------------------
checkClusteredQueries([],[],[]) :-
    !.
checkClusteredQueries([M|Ms],[N|Ns],Pars) :-
    N=sx::toInt(M),
	dbExtension::kbQuery(N,_,_,Query,_),	
	getPars(Query,"",_,Pars2),
	checkClusteredQueries(Ms,Ns,Pars3),
	set::conc([Pars2,Pars3],Pars4),
	set::distinct(Pars4,Pars),
	!.
% -------------------------------------------------------------------------------------------------------------
% getPars
% -------------------------------------------------------------------------------------------------------------
getPars(Q,Replacement,Qrep,Pars) :-
    sx::searchchar(Q,'[',P1),
    sx::searchchar(Q,']',P2),
    P1<P2, % has parameters
    sx::sxGetInBrack(Q,'[',']',Par),
    S=string::concat("[",Par,"]"),
	Q2=sx::replaceAll(Q,S,Replacement),
    getPars(Q2,Replacement,Qrep,Pars2),
    Pars=[Par|Pars2],
    !
;
    Pars=[],
    Qrep=Q,
    !.	
% -------------------------------------------------------------------------------------------------------------
% runQueriesNL/PR
% -------------------------------------------------------------------------------------------------------------
class predicates
runQsNL : (count,string_list) determ.
runQsNL : (count,outFile,string_list) determ.
runQsPR : (count,string_list) determ.
runQsPR : (count,outFile,string_list) determ.
clauses

runQueriesNL(S) :-
    sx::sxUnpack(S,'?',Qs),
    adjustList(Qs,"?",Qsa),
    Qsa=[],
    runQueryNL(""),
    !
;
    sx::sxUnpack(S,'?',Qs),
    adjustList(Qs,"?",Qsa),
    Cnt=count::newCounter(0),
    runQsNL(Cnt,Qsa),
    !.   
runQueriesNL(File,S) :-
    sx::sxUnpack(S,'?',Qs),
    adjustList(Qs,"?",Qsa),
    Cnt=count::newCounter(0),
    runQsNL(Cnt,File,Qsa),
    !.

runQueriesPR(S) :-
    sx::sxUnpack(S,'?',Qs),
    adjustList(Qs,"?",Qsa),
    Qsa=[],
    runQueryNL(""), % no difference NL-PR
    !
;
    sx::sxUnpack(S,'?',Qs),
    adjustList(Qs,"?",Qsa),
    Cnt=count::newCounter(0),
    runQsPR(Cnt,Qsa),
    !.
runQueriesPR(File,S) :-
    sx::sxUnpack(S,'?',Qs),
    adjustList(Qs,"?",Qsa),
    Cnt=count::newCounter(0),
    runQsPR(Cnt,File,Qsa),
    !.

runQsNL(_,[]) :-
    !.
runQsNL(Cnt,[Q|Qs]) :-
                Cnt:inc(No),
				Sn=toString(No),
				sx::concat(["processes query #",Sn," '",Q,"'..."],M),
				dlgIF::msgS(M),
    sx::sxGiveTail(Q,"?",Qt),
    runQueryNL(Qt),
    runQsNL(Cnt,Qs),
    !.
runQsNL(_,_,[]) :-
    !.
runQsNL(Cnt,File,[Q|Qs]) :-
                Cnt:inc(No),
				Sn=toString(No),
				sx::concat(["processes query #",Sn," '",Q,"'..."],M),
				dlgIF::msgS(M),
    sx::sxGiveTail(Q,"?",Qt),
    runQueryNL_F(File,Qt),
    runQsNL(Cnt,File,Qs),
    !.
    
runQsPR(_,[]) :-
    !.
runQsPR(Cnt,[Q|Qs]) :-
                Cnt:inc(No),
				Sn=toString(No),
				sx::concat(["processes query #",Sn," '",Q,"'..."],M),
				dlgIF::msgS(M),
    runQueryPR(Q),
    runQsPR(Cnt,Qs),
    !.
runQsPR(_,_,[]) :-
    !.
runQsPR(Cnt,File,[Q|Qs]) :-
                Cnt:inc(No),
				Sn=toString(No),
				sx::concat(["processes query #",Sn," '",Q,"'..."],M),
				dlgIF::msgS(M),
    runQueryPR_F(File,Q),
    runQsPR(Cnt,File,Qs),
    !.
% 1. removes blank strings
% 2. concatenates adjacent strings A and B if B starts with C (|C|=1)
%    
adjustList([],_,[]) :-
    !.
adjustList([F,G],_,[F]) :-
    F<>"",
    G="",
    !.
adjustList([F,G,H],C,[Fa]) :-
    F<>"",
    G="",
    H="",
    Fa=string::concat(F,C),
    !.
adjustList([F,G,H|T],C,L) :-
    F<>"",
    G="",
    H<>"",
    Fa=string::concat(F,C,H),
    adjustList([Fa|T],C,L),
    !.
adjustList([H|T],C,L) :-
    string::trim(H)="",
    adjustList(T,C,L),
    !.
adjustList([H|T],C,[H|L]) :-
    adjustList(T,C,L),
    !.
% -------------------------------------------------------------------------------------------------------------
% writeQuery
% -------------------------------------------------------------------------------------------------------------
writeQuery(File,Q) :-
    P=File:getPar(),
    AP=math::abs(P),
    AP=2,
    File:nl,
    File:writenl(Q),
    File:nl,
    !
;
    !.
% -------------------------------------------------------------------------------------------------------------
% runQueryPR
% -------------------------------------------------------------------------------------------------------------
% runs PR query w/out ./? terminator
%	
runQueryPR(Query) :-
    set::member(Query,["","?"]),
	runQueryNL_Type(nill,"?"), % no difference NL-PR
	!
;    
	Qp=string::trim(Query),
	Qt=string::concat(Qp,"."),
	transpr::proPR_Term(Qt,Term),
				tt::tt_term(2,"query (IF)",Term),
	processQuery("eq",Term,Qp),
	!.	

runQueryPR(ODBC,Query) :-
	Qp=string::trim(Query),
	Qt=string::concat(Qp,"."),
	transpr::proPR_Term(Qt,Term),
				tt::tt_term(2,"query (IF)",Term),
	processQuery(ODBC,"eq",Term,Qp),
	!.	

runQueryPR_F(File,Query) :-
	Qp=string::trim(Query),
	File:setQuery(Qp),
	Qt=string::concat(Qp,"."),
	transpr::proPR_Term(Qt,Term),
				tt::tt_term(2,"query (IF)",Term),
    writeQuery(File,Qt),
	processQuery_F(File,"eq",Term,Qp),
	!.	
% -------------------------------------------------------------------------------------------------------------
% runQueryNL
% -------------------------------------------------------------------------------------------------------------
class predicates
runQueryNL_Type : (sterm,string) determ.
runQueryNL_Type : (odbcConnection,sterm,string) determ.
runQueryNL_Type_F : (outFile,sterm,string) determ.
clauses

% runs NL query with/without terminator ?
%
runQueryNL(Query) :-
    set::member(Query,["","?"]),
	runQueryNL_Type(nill,"?"),
	!
;    
	Qp=string::trim(Query),
	sx::sxGiveTail(Qp,"?",QpT),
	transnl::parSynTree(QpT,SynTr),
	not(SynTr=nill),
	runQueryNL_Type(SynTr,Query),
	!.	

runQueryNL(ODBC,Query) :-
    set::member(Query,["","?"]),
	runQueryNL_Type(ODBC,nill,"?"),
	!
;
	Qp=string::trim(Query),
	sx::sxGiveTail(Qp,"?",QpT),
	transnl::parSynTree(QpT,SynTr),
	not(SynTr=nill),
	runQueryNL_Type(ODBC,SynTr,Query),
	!.	

runQueryNL_F(File,Query) :-
	Qp=string::trim(Query),
	sx::sxGiveTail(Qp,"?",QpT),
	File:setQuery(QpT),
	transnl::parSynTree(QpT,SynTr),
    writeQuery(File,QpT),
	runQueryNL_Type_F(File,SynTr,Query),
	!.	

runQueryNL_Type(SynTr,Query) :-
	SynTr=cmp("ABOUT",[T]),
	% all about query
	List=lst::new(),
	allabout::fetchRecords(List,"",T,0,1,1,0,0,0,0,0,0,0,-1),	
	allabout::reportRecords(0,1,List,"",T),
	!
;
	SynTr=cmp(QT,_),
	set::member(QT,["DFNC","NFNC","NDERF","DDERF","#S","EX"]),
	% single-value query - TBQ is still used
				dlgIF::msgStatus("3BQ-tree construction"),
	tbq::tbqQuery(SynTr,TBQ_tree),
				tt::tt_header(1,"Query (TBQ)"),
				tt::tt_term(1,"",TBQ_tree),
	pr::prTbqQuery(TBQ_Tree,PR,Type),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery(Type,Term,Query),
				dlgIF::msgStatus(""),
    !
;
	SynTr=cmp(QT,_),
	not(set::member(QT,["DFNC","NFNC","NDERF","DDERF","#S","EX"])),
	pr::prSynTreeQuery(0,SynTr,PR),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery("eq",Term,Query),
				dlgIF::msgStatus(""),
	!.

runQueryNL_Type(ODBC,SynTr,Query) :-
	SynTr=cmp("ABOUT",[T]),
	% all about query
	List=lst::new(),
	allabout::fetchRecords(List,"",T,0,1,1,0,0,0,0,0,0,0,-1),
	Opt=globalVars::get_i("f_outq"),
	AOpt=math::abs(Opt),	
	allabout::reportRecordsODBC(ODBC,AOpt,List),
	!
;
	SynTr=cmp(QT,_),
	set::member(QT,["DFNC","NFNC","NDERF","DDERF","#S"]),
	% single-value query - TBQ is still used
				dlgIF::msgStatus("3BQ-tree construction"),
	tbq::tbqQuery(SynTr,TBQ_tree),
				tt::tt_header(1,"Query (TBQ)"),
				tt::tt_term(1,"",TBQ_tree),
	pr::prTbqQuery(TBQ_Tree,PR,Type),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery(ODBC,Type,Term,Query),
				dlgIF::msgStatus(""),
    !
;
	SynTr=cmp(QT,_),
	QT="EX",
	pr::prSynTreeQuery(0,SynTr,PR),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery(ODBC,"xq",Term,Query),
				dlgIF::msgStatus(""),
    !
;
	SynTr=cmp(QT,_),
	not(set::member(QT,["DFNC","NFNC","NDERF","DDERF","#S","EX"])),
	pr::prSynTreeQuery(0,SynTr,PR),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery(ODBC,"eq",Term,Query),
				dlgIF::msgStatus(""),
	!.

runQueryNL_Type_F(File,SynTr,Query) :-
	SynTr=cmp("ABOUT",[T]),
	% all about query
	List=lst::new(),
	allabout::fetchRecords(List,"",T,0,1,1,0,0,0,0,0,0,0,-1),	
	allabout::reportRecords(File,1,List),
				dlgIF::msgStatus(""),
    !
;
	SynTr=cmp(QT,_),
	set::member(QT,["DFNC","NFNC","NDERF","DDERF","#S"]),
	% single-value query - TBQ is still used
				dlgIF::msgStatus("3BQ-tree construction"),
	tbq::tbqQuery(SynTr,TBQ_tree),
				tt::tt_header(1,"Query (TBQ)"),
				tt::tt_term(1,"",TBQ_tree),
	pr::prTbqQuery(TBQ_Tree,PR,Type),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery_F(File,Type,Term,Query),
				dlgIF::msgStatus(""),
    !
;
	SynTr=cmp(QT,_),
	QT="EX",
	pr::prSynTreeQuery(0,SynTr,PR),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery_F(File,"xq",Term,Query),
				dlgIF::msgStatus(""),
    !
;
	SynTr=cmp(QT,_),
	not(set::member(QT,["DFNC","NFNC","NDERF","DDERF","#S","EX"])),
	pr::prSynTreeQuery(0,SynTr,PR),
				tt::tt_s(2,"Query (PR)",PR),
	transpr::proPR_Term(PR,Term),
				tt::tt_term(2,"Query (IF)",Term),
	processQuery_F(File,"eq",Term,Query),
				dlgIF::msgStatus(""),
	!.
% -------------------------------------------------------------------------------------------------------------
% processQuery
% -------------------------------------------------------------------------------------------------------------
class predicates
showResp : (string,integer).
clauses

processQuery(Type,Term,Query) :-
	not(set::member(Type,["fq","xq","dq","sq"])),
	Opt=globalVars::get_i("f_outq"),
    set::member(Opt,[0,8,-8]),
	sol::solSetNew(Term,SolRef),
	pie3_Exec(SolRef,Term),
	proQueryResult(Opt,Term,SolRef,Query),
	!
;
	set::member(Type,["fq","xq","dq","sq"]),
	Opt=2,
	sol::solSetNew(Term,SolRef),
	pie3_Exec(SolRef,Term),
	P=globalVars::get_s("std_path"),
    fioExt::file_type(Opt,FType),
    FileName=string::concat(P,"\\Dummy.",FType),
	File=outFile::new(FileName,Opt,Query),
	proQueryResult_F(File,Type,Opt,Term,SolRef),
	File:close(),
	showResp(FileName,Opt),
	!
;
	Opt=globalVars::get_i("f_outq"),
    dbBase::queryStF(Term,STk),
	sol::solSetNew(Term,SolRef),
	pie3_Exec(SolRef,Term),
	P=globalVars::get_s("std_path"),
    dbBase::stNL(STk,ST),
    sx::makeIdentifier(1,3,2,ST,64,"",Name),
    fioExt::file_type(Opt,FType),
    FileName=string::concat(P,"\\",Name,".",FType),
	File=outFile::new(FileName,Opt,Query),
	proQueryResult_F(File,Type,Opt,Term,SolRef),
	File:close(),
	showResp(FileName,Opt),
	!
;
	Opt=globalVars::get_i("f_outq"),
    dbBase::querySgNl(Term,Sg),
    sx::makeIdentifier(1,3,2,Sg,64,"",Name),
	sol::solSetNew(Term,SolRef),
	pie3_Exec(SolRef,Term),
	P=globalVars::get_s("std_path"),
    fioExt::file_type(Opt,FType),
    FileName=string::concat(P,"\\",Name,".",FType),
	File=outFile::new(FileName,Opt,Query),
	proQueryResult_F(File,Type,Opt,Term,SolRef),
	File:close(),
	showResp(FileName,Opt),
	!.
	
processQuery(ODBC,Type,Term,Query) :-
	Opt=globalVars::get_i("f_outq"),
	sol::solSetNew(Term,SolRef),
	pie3_Exec(SolRef,Term),
	proQueryResult(ODBC,Type,Opt,Term,SolRef,Query),
	!.
	
processQuery_F(File,Type,Term,Query) :-
	set::member(Type,["fq","xq","dq","sq"]),
	File:setQuery(Query),
	sol::solSetNew(Term,SolRef),
	pie3_Exec(SolRef,Term),
	proQueryResult_F(File,Type,2,Term,SolRef),
	!
;	
	File:setQuery(Query),
	sol::solSetNew(Term,SolRef),
	pie3_Exec(SolRef,Term),
	Opt=File:getPar(),
	proQueryResult_F(File,Type,Opt,Term,SolRef),
	!.
	
showResp(Name,Opt) :-
   AOpt=math::abs(Opt),
   set::member(AOpt,[1,2,9]),
   editorWindow::openFiles([Name]),
   !
;
   AOpt=math::abs(Opt),
   set::member(AOpt,[3,4]),
   editorWindow::openFiles(-1,[Name]),
   !
;
   Opt<>0,
   Mess=string::concat("Output file ",Name," generated!"),
   dlgIF::dlgNote(0,Mess), 
   !
;
   !.
% -------------------------------------------------------------------------------------------------------------
% proQueryResult
% -------------------------------------------------------------------------------------------------------------
class predicates
proQueryText : (outFile,integer,string).
clauses	

proQueryResult(Opt,Term,SolRef,Query) :-
	sol::solSetSize(SolRef,N),
	N=0,
	Mess=string::concatList(["Query  [",Query,"] has no solutions!"]),
	dlgIF::msgS(Mess),
	!
;
	Opt=0,
	sol::solSetSize(SolRef,N),
	N>0,
	S=toString(N),
	S1=string::concat(S," solution(s) generated [",Query,"]"),
	dlgIF::msgS(S1),
	!
;	
    AOpt=math::abs(Opt),
	AOpt=8,
	sol::solSetSize(SolRef,N),
	N>0,
	Ns=toString(N),
	Mess=string::concatList(["Query [",Query,"] has ",Ns," solutions!"]),
	dlgIF::msgS(Mess),
	proQuerySolGen(SolRef,Opt,Term),
    !.
    
proQueryResult(ODBC,Type,Opt,Term,SolRef,_) :-
	sol::solSetSize(SolRef,N),
	N=0,
	!
;	
	proQuerySolGen(ODBC,Type,SolRef,Opt,Term),
    !.
    
proQueryResult_F(File,Type,Opt,Term,SolRef) :-
	sol::solSetSize(SolRef,N),
	N=0,
    AOpt=math::abs(Opt),
	AOpt=2, 
	Query=File:getQuery(),
    SSolRef=toString(SolRef),
	Mess=string::concatList(["Query '",Query,"' [",SSolRef,"] has no solutions!"]),
	dlgIF::msgS(Mess),
    proQueryText(File,AOpt,Query),
    File:writenl("NO SOLUTIONS"),
    !
;
	sol::solSetSize(SolRef,N),
	N=0,
    AOpt=math::abs(Opt),
	AOpt<>2,
	Query=File:getQuery(),
    SSolRef=toString(SolRef),
	Mess=string::concatList(["Query '",Query,"' [",SSolRef,"] has no solutions!"]),
	dlgIF::msgS(Mess),
    !
;	
	sol::solSetSize(SolRef,N),
	N>0,
	Ns=toString(N),
    Query=File:getQuery(),
    SSolRef=toString(SolRef),
	Mess=string::concatList(["Query '",Query,"' [",SSolRef,"] has ",Ns," solutions!"]),
	dlgIF::msgS(Mess),
    AOpt=math::abs(Opt),
	proQueryText(File,AOpt,Query),
	proQuerySolGen_F(File,Type,SolRef,Opt,Term),
	!.
    
proQueryText(File,Opt,S) :- 
    Opt=2,
    File:writenl(S),
    !
;
    !.
% -------------------------------------------------------------------------------------------------------------
% proQuerySolGen
% -------------------------------------------------------------------------------------------------------------
class predicates
proQuerySolGenX : (integer,integer,sterm) determ.
proQuerySolGenX : (odbcConnection,string,integer,integer,sterm) determ.
proQuerySolGenX_F : (outfile,string,integer,integer,sterm) determ.
proOpenTable : (integer,integer) determ.
writeTables : (outFile,integer,integer,integer) determ.  
clauses

proQuerySolGen(SolRef,Opt,cmp(",",[ObjTemplate|_])) :- proQuerySolGenX(SolRef,Opt,ObjTemplate),!.
proQuerySolGen(SolRef,Opt,ObjTemplate) :- proQuerySolGenX(SolRef,Opt,ObjTemplate),!.

proQuerySolGen(ODBC,Type,SolRef,Opt,cmp(",",[ObjTemplate|_])) :- proQuerySolGenX(ODBC,Type,SolRef,Opt,ObjTemplate),!.
proQuerySolGen(ODBC,Type,SolRef,Opt,ObjTemplate) :- proQuerySolGenX(ODBC,Type,SolRef,Opt,ObjTemplate),!.

proQuerySolGen_F(File,Type,SolRef,Opt,cmp(",",[ObjTemplate|_])) :- proQuerySolGenX_F(File,Type,SolRef,Opt,ObjTemplate),!.
proQuerySolGen_F(File,Type,SolRef,Opt,ObjTemplate) :- proQuerySolGenX_F(File,Type,SolRef,Opt,ObjTemplate),!.

proQuerySolGenX(SolRef,_,ObjTemplate) :- 
	table::tabGen(1,SolRef,ObjTemplate),
	stb::stbM_Get(SolRef,NTab,_,_,_,_),
			S=toString(NTab),
			sx::concat([" ",S," table(s) generated"],S1),
			dlgIF::msgA(S1),
			tt::tt_table(2,"all solution tables",SolRef),
	proOpenTable(SolRef,NTab),
	!.

proQuerySolGenX(ODBC,Type,SolRef,Opt,ObjTemplate) :- 
    Opt=-8,
	table::tabGen(1,SolRef,ObjTemplate),
	stb::stbM_Get(SolRef,NTab,_,_,_,_),
	proOpenTableODBC(ODBC,SolRef,NTab,0),
	sol::solSetSize(SolRef,N),
	globalVars::add("request_sols",N,_),
	!
;
    Opt<>-8,
    proConstructSolODBC(ODBC,Type,SolRef,Opt,ObjTemplate),
    !.

proQuerySolGenX_F(File,Type,SolRef,Opt,ObjTemplate) :- 
	AOpt=math::abs(Opt),
	set::member(AOpt,[3,4,5,6]),
	table::tabGen(1,SolRef,ObjTemplate),
			stb::stbM_Get(SolRef,NTab,_,_,_,_),
			S=toString(NTab),
			sx::concat([" ",S," table(s) generated"],S1),
			dlgIF::msgA(S1),
			tt::tt_table(2,"all solution tables",SolRef),
	writeTables(File,Opt,SolRef,NTab),
	!		
;
	AOpt=math::abs(Opt),
	set::member(AOpt,[10,11]),
	proConstructSolXML(File,Type,SolRef,AOpt,ObjTemplate),
	!
;
	AOpt=math::abs(Opt),
	set::member(AOpt,[1,2,9]),
	proConstructSol(File,Type,SolRef,AOpt,ObjTemplate),
	!.

proOpenTable(_,0) :-
   !.
proOpenTable(SolRef,Table) :- 
    stb::openTableWindow(SolRef,Table),
    Next=Table-1,
    proOpenTable(SolRef,Next),
    !.

% parameters:
% 1 in ODBC handle
% 2 in sol.ref.
% 3 in Pro/3 table#
% 4 in =0: use par#3 as SQL Table#, else: use this
%
proOpenTableODBC(_,_,0,_) :-
   !.
proOpenTableODBC(ODBC,SolRef,Table,SqlTable) :- 
    set::setMax([Table,SqlTable],Tab),
    STable=toString(Tab),
    table::tabSortIfRequired(SolRef,Table), 
    createSQL_Table(ODBC,SolRef,Table,Tab,TableName,NRows,NCols,Domains,Title,Width),
    NRows>0,
    NCols>0,
    popSQL_Table(ODBC,SolRef,Table,TableName,NRows,NCols,Domains),
    R=globalVars::get_i("request_no"),
    Rs=toString(R),
    SR=globalVars::get_i("request_subno"),
    SRs=toString(SR),
    SNCols=toString(NCols),
    SNRows=toString(NRows),
    SWidth=toString(Width),
    S=string::concatList(["insert into p3bridge.p3response_table (request_no,sub_no,table_no,table_name,title,no_of_cols,width,no_of_rows) values (",
                                    Rs,",",SRs,",",STable,",'",TableName,"','",Title,"',",SNCols,",",SWidth,",",SNRows,")"]),
    sqlTools::executeDirect(ODBC,S),
    Next=Table-1,
    proOpenTableODBC(ODBC,SolRef,Next,SqlTable),
    !
;    
    Next=Table-1,
    proOpenTableODBC(ODBC,SolRef,Next,SqlTable),
    !.
  
writeTables(_,_,_,0) :-
    !.
writeTables(File,Opt,SolRef,Table) :-    
	stb::stbH_Get(SolRef,Table,_,_,STk,_,_,_),
	dbBase::stNL(STk,ST),
    MOpt=-Opt,
    table::tabEditor(File,MOpt,ST,SolRef,Table,1,999999,1,1,-1),
    Next=Table-1,
    writeTables(File,Opt,SolRef,Next),
    !.
% ----------------------------------------------------------------------------------------------
% proConstructSol                                    
% ----------------------------------------------------------------------------------------------
proConstructSol(File,Type,SolRef,Opt,Term) :-
    Type="sq",
	% sss query
	nl::conSol("sq",SolRef,Opt,1,Term,S),
	File:writenl(S),
	!
;
    Type<>"sq",
	sol::forallsol(SolRef,SolNo),
	   nl::conSol(Type,SolRef,Opt,SolNo,Term,S),
	   File:writenl(S),
	   fail
;
	!.
% ----------------------------------------------------------------------------------------------
% proConstructSolXML                             
% ----------------------------------------------------------------------------------------------
proConstructSolXML(File,Type,SolRef,Opt,TermTemplate) :-
    set::setGet([10,11],Ix,Opt),
    set::setGet(["NL","PR"],Ix,XOpt),
    xmlP3::xmlKnowA(Header),
    File:writenl(Header),
	sol::forallsol(SolRef,SolNo),
	   nl::conSol(Type,SolRef,9,SolNo,TermTemplate,S),
	   sx::sxRemoveTail(S,"!",Sr),
	   transpr::proPR_Term(Sr,Term),
       xmlP3::xmlFromSentence(XOpt,"QUERY",0,Term,X),
	   File:writenl(X),
	   fail
;
    xmlP3::xmlKnowB(Footer),
    File:writenl(Footer),
	!.
% ----------------------------------------------------------------------------------------------
% proConstructSolODBC                                    
% ----------------------------------------------------------------------------------------------
% constructs and inserts solutions (non-table format)
%
proConstructSolODBC(ODBC,Type,SolRef,Opt,Term) :-
    Ix=math::abs(Opt)-9,
    set::member(Ix,[1,2]),
    AOpt=math::abs(Opt),
    set::member(AOpt,[10,11]),
    set::setGet(["NL","PR"],Ix,XOpt),
    R=globalVars::get_i("request_no"),
    Rs=toString(R),
    SR=globalVars::get_i("request_subno"),
    SRs=toString(SR),
    xmlP3::xmlKnowA(Header),
	sol::forallsol(SolRef,SolNo),
	   nl::conSol(Type,SolRef,9,SolNo,Term,S),
	   sx::sxRemoveTail(S,"!",Srt),
	   transpr::proPR_Term(Srt,XTerm),
       xmlP3::xmlFromSentence(XOpt,"QUERY",0,XTerm,X),
       Ns=toString(SolNo),
       Sq=string::concatList(["insert into p3bridge.p3response (request_no,sub_no,line_no,line) values (",Rs,",",SRs,",",Ns,",'",Header,X,"')"]),
       sqlTools::executeDirect(ODBC,Sq),
	   fail
;
    AOpt=math::abs(Opt),
    set::member(AOpt,[1,9]),
    R=globalVars::get_i("request_no"),
    Rs=toString(R),
    SR=globalVars::get_i("request_subno"),
    SRs=toString(SR),
	sol::forallsol(SolRef,SolNo),
        Ns=toString(SolNo),
        nl::conSol(Type,SolRef,AOpt,SolNo,Term,S2),
        S=string::concatList(["insert into p3bridge.p3response (request_no,sub_no,line_no,line) values (",Rs,",",SRs,",",Ns,",'",S2,"')"]),
        sqlTools::executeDirect(ODBC,S),
	    fail
;
	sol::solSetSize(SolRef,N),
	globalVars::add("request_sols",N,_),
	!.
% -------------------------------------------------------------------------------------------------------------
% writeSqlLines
% -------------------------------------------------------------------------------------------------------------
writeSqlLines(_,N,[]) :-
    M=N-1,
	globalVars::add("request_sols",M,_),
    !.
writeSqlLines(ODBC,N,[H|T]) :-
    R=globalVars::get_i("request_no"),
    Rs=toString(R),
    SR=globalVars::get_i("request_subno"),
    SRs=toString(SR),
    Ns=toString(N),
    S=string::concatList(["insert into p3bridge.p3response (request_no,sub_no,line_no,line) values (",Rs,",",SRs,",",Ns,",'",H,"')"]),
    sqlTools::executeDirect(ODBC,S),
    Next=N+1,
    writeSqlLines(ODBC,Next,T),
    !.
% -------------------------------------------------------------------------------------------------------------
% proText
% -------------------------------------------------------------------------------------------------------------
class predicates
processTerm : (string,string,sterm,integer).
clauses

% processes (translates and stores) a list of input texts (text=sentence/func.def./sent.rule)
%
proText(_,_,[],_) :-
    !.
proText(Type,Realm,[""|T],C) :- 
    proText(Type,Realm,T,C),
    !.
proText(Type,Realm,[Text|Texts],C) :-
    globalVars::get_i("quit_flag")=1,
    !
;    
    Type="3pr",
                C:inc(N),
				Sn=toString(N),
				sx::concat(["processes realm ",Realm," #",Sn],M),
				dlgIF::msgS(M),
	Text2=string::trim(Text),
	PRt=string::concat(Text2,"."),
	transpr::proPR_Term(PRt,Term),
    pro_type(Term,TermType),
				tt::tt_s(2,"Knowledge type",TermType),
				tt::tt_term(2,"Knowledge (IF)",Term),
     processTerm(TermType,Realm,Term,N),
     proText(Type,Realm,Texts,C),
	!
;
    Type="3nl",
                C:inc(N),
				Sn=toString(N),
				sx::concat(["processes realm ",Realm," #",Sn],M),
				dlgIF::msgS(M),
	Text2=string::trim(Text),
	NLt=string::concat(Text2,"!"),
	transnl::parTranslate(NLt,Term,TermType),
				tt::tt_s(2,"Knowledge type",TermType),
				tt::tt_term(2,"Knowledge (IF)",Term),
     processTerm(TermType,Realm,Term,N),
     proText(Type,Realm,Texts,C),
	!.

processTerm(Type,Realm,Term,No) :-
    % insert fact
	Type="s",
	dbBase::recordInsert(2,Realm,No,0,Term,atom("true"),_),
	!
;
    % insert SR
	Type="r",
	Sign=globalVars::get_i("rulsta"),
	RTyp=3*Sign,
	Term=cmp(":-",[LT,RT]),
	dbBase::recordInsert(RTyp,Realm,No,0,LT,RT,_),
	!
;
    % insert function
	Type="f",
	Term=cmp(":-",[LT,RT]),
	dbBase::recordInsert(4,Realm,No,0,LT,RT,_),
	!
;
    SNo=toString(No),
    Mess=string::concatList(["Knowledge assert of type: ",Type," and realm: ",Realm,"|",SNo," failed!"]),
    dlgIF::msgS(Mess),
	!.
% -------------------------------------------------------------------------------------------------------------
% proCheckRealm
% -------------------------------------------------------------------------------------------------------------
% deletes given realm in given KB files. fails only if KE rejects deleting an existing realm
% parameters:
% 1	in	KB files where given realm (par#2) exists
% 2	in	realm
%
proCheckRealm([],_) :-
	!.
proCheckRealm(Files,Realm) :-
	globalVars::get_i("all_flag")=1,
	vpi::cursorSetWait(),
			dlgIF::msgStatus("deleting realm"),
	RealmU=string::toUpperCase(Realm),	
	dbBase::realmKbDelete(Files,RealmU),
	dbBase::realmKbDelete(Files,RealmU),
			sx::concat(["Deleted realm '",RealmU,"'"],Mess2),
			dlgIF::msgS(Mess2),
			dlgIF::msgStatus(""),
	!
;	
	RealmU=string::toUpperCase(Realm),	
	dbBase::joinNL(Files,S2),
	sx::concat(["Realm ",RealmU," exists already in KBs '",S2,"'"],Mess1),
	Reply=vpiCommonDialogs::ask(Mess1,"Do you want to replace this realm?",["Yes","Yes ALL","No"]),
	Reply<>vpiCommonDialogs::resp_3,
	set::setGet([0,1],Flag,Reply),
	set::setGet([-1,1],Flag,FlagX),
	globalVars::set_i("all_flag",FlagX),
	vpi::cursorSetWait(),
			dlgIF::msgStatus("deleting realm"),
	dbBase::realmKbDelete(Files,RealmU),
			sx::concat(["Deleted realm '",RealmU,"'"],Mess2),
			dlgIF::msgS(Mess2),
			dlgIF::msgStatus(""),
	!.
% ----------------------------------------------------------------------------------------------
% pro_type
% ----------------------------------------------------------------------------------------------
class predicates
containVariables : (sterm) determ.
containVariablesX : (sterml) determ.
clauses
pro_type(TPro,"r") :-
    containVariables(TPro),
    !.
pro_type(_,"s") :- 
    !.
% containVariables is used to determine if the wff is a fact or a query.
% var("_") is not considered as a variable in this regard since it is used
% to represent facts with an "any-argument" 
%
containVariables(var("_")) :- !,fail.
containVariables(var(_)) :- !.
containVariables(cmp(_,[H|T])) :- 
    containVariables(H),
    !
;
	containVariablesX(T),
	!.
containVariables(list(T1,T2)) :-		
	containVariables(T1),
	!
;
	containVariables(T2),
	!.
containVariablesX([H|T]) :-
    containVariables(H),
    !
;
    containVariablesX(T),
    !.
% ----------------------------------------------------------------------------------------------
% pie3_IF
% ----------------------------------------------------------------------------------------------
pie3_IF(Query,SolRef) :-
/* no more re-use 27.5.5
	globalVars::get_i("ssreus")=1,
	sol::solSetExists(Query,SolRef),
	globalVars::incr("piecal",_,_),
		tt::tt_pie(2,0,SolRef),
    !
;	
*/
	sol::solSetNew(Query,SolRef),
	pie3_Exec(SolRef,Query),
	!.
% ---------------------------------------------------------------------------------------------
% pie3_InternalNew
% ----------------------------------------------------------------------------------------------
class predicates
pie3_IF_InternalNew : (sterm,string_list,integer) determ (i,i,o).
clauses
pie3_InternalNew(Query,RelComps,SolRef) :-
	pie3_IF_InternalNew(Query,RelComps,SolRef),
	!
;
	dlgIF::feil("pie3_Internal",RelComps,[],Query),!,fail.	
pie3_IF_InternalNew(Query,RelComps,SolRef) :-
/* no more re-use 27.5.5
	globalVars::get_i("ssreus")=1,
	sol::solSetExists(Query,SolRef),
	sol::exSolSet(SolRef,RelComps),
	globalVars::incr("piecal",_,_),
		tt::tt_pie(2,0,SolRef),
    !
;	
*/
	sol::solSetNew(Query,SolRef),
	sol::dropSolSet(SolRef),
	sol::addSolSet(SolRef,RelComps),
	pie3_Exec(SolRef,Query),
	!.
% ----------------------------------------------------------------------------------------------
% pie3_Exec
% ----------------------------------------------------------------------------------------------
class predicates
pie3_ResetCounters : (integer) procedure (i).
pie3_Result : (integer,integer).		
pie3_Start : (integer,sterm,integer).
clauses

pie3_Exec(SolRef,Query) :-
	globalVars::incr("piercl",_,Rec),
	pie3_Start(SolRef,Query,Rec),
	%trap(pie::pie(SolRef,Query),ExitCode,pie::errorhnd(ExitCode)),
    profileTime::start("pie"),
	pie::pie(SolRef,Query),
	profileTime::stop("pie"),
			dlgIF::msgStatus(""),
	sol::solSetEnumerate(SolRef),
	globalVars::decr("piercl",_,_),
		tt::tt_cc,
	pie3_Result(SolRef,Rec),
	!.
pie3_Exec(SolRef,_) :- 
	globalVars::decr("piercl",_,_),
		tt::tt_cc,
		Rec=globalVars::get_i("piercl")+1,
		SRec=toString(Rec),
		SSolRef=toString(SolRef),
		Rus=globalVars::get_i("piecal"),
		SRus=toString(Rus),
		sx::concat(["PIEX error at execution no ",SSolRef," at recursion level ",SRec," #calls=",SRus],S1),
		tt::tt_ss(1,S1),
		dlgIF::msgS(S1),
	sol::solSetSizeSet(SolRef,-1),
	!
;
	dlgIF::feil("pie3_Exec",[],[],nill),!,fail.
			
pie3_Start(SolRef,Query,1) :-
	pie3_ResetCounters(1),
	ir::crtRetract,
	SSolRef=toString(SolRef),
	sx::concat(["PIEX entry no ",SSolRef," at recursion level 1"],S1),
	tt::tt_pie(2,-1,SolRef),
	tt::tt_ss(2,S1),
	tt::tt_term(1,"Executing",Query),
	dlgIF::msgS(S1),
	!.
pie3_Start(SolRef,Query,Rec) :-
	tt::getLevel()>1,
	SRec=toString(Rec),
	SSolRef=toString(SolRef),
	sx::concat(["PIEX entry no ",SSolRef," at recursion level ",SRec],S1),
	tt::tt_pie(2,-1,SolRef),
	tt::tt_ss(2,S1),
	tt::tt_term(2,"executing",Query),
	dlgIF::msgS(S1),
	!
;
	!.
pie3_ResetCounters(1) :-
	globalVars::set_i("solcnt0",0),
	globalVars::set_i("solcnt1",0),
	globalVars::set_i("solcnt2",0),
	globalVars::set_i("solcnt3",0),
	globalVars::set_i("solcnt4",0),!.
pie3_ResetCounters(2) :-
	globalVars::set_i("solcnt1",0),
	globalVars::set_i("solcnt2",0),
	globalVars::set_i("solcnt3",0),
	globalVars::set_i("solcnt4",0),!.
pie3_ResetCounters(3) :-
	globalVars::set_i("solcnt2",0),
	globalVars::set_i("solcnt3",0),
	globalVars::set_i("solcnt4",0),!.
pie3_ResetCounters(4) :-
	globalVars::set_i("solcnt3",0),
	globalVars::set_i("solcnt4",0),!.

pie3_ResetCounters(_) :-
	globalVars::set_i("solcnt4",0),!.
pie3_Result(SolRef,1) :-		
	sol::solSetSize(SolRef,N),
	Sn=toString(N),
	SSolRef=toString(SolRef),
	Rus=globalVars::get_i("piecal"),
	SRus=toString(Rus),
	sx::concat(["PIEX exit  no ",SSolRef," at recursion level 1 #solutions=",Sn," #calls=",SRus],S2),
	S3=string::concat(S2,"\n"),
	tt::tt_ss(1,S3),
	tt::tt_pie(1,1,SolRef),
	dlgIF::msgS(S2),
	!.
pie3_Result(SolRef,Rec) :-		
	tt::getLevel()>1,
	SRec=toString(Rec),
	sol::solSetSize(SolRef,N),
	Sn=toString(N),
	SSolRef=toString(SolRef),
	Rus=globalVars::get_i("piecal"),
	SRus=toString(Rus),
	sx::concat(["PIEX exit  no ",SSolRef," at recursion level ",SRec," #solutions=",Sn," #calls=",SRus],S2),
	tt::tt_ss(2,S2),
	tt::tt_pie(1,1,SolRef),
	dlgIF::msgS(S2),
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
%  pie3_Syn
% ----------------------------------------------------------------------------------------------
% this predicate is used only when the inference engine is invoked from within the 
% inference engine
%
pie3_Syn(SynTree,SolRef) :-
	L=globalVars::get_i("piercl"),
	L<1,
	dlgIF::feil("pie3_Syn",[],[L],SynTree),!,fail
;
	SynTree=cmp(QType,_),
	set::member(QType,["DFNC","NFNC","NDERF","DDERF","#S","EX"]),
	% single-value query - TBQ is used
	tbq::tbqQuery(SynTree,TBQ_Tree),
	pr::prTbqQuery(TBQ_Tree,PR,_),
	transpr::proPR_Term(PR,PR_Term),
	pie3_IF(PR_Term,SolRef),
	!
;
	SynTree=cmp(QType,_),
	not(set::member(QType,["DFNC","NFNC","NDERF","DDERF","#S","EX"])),
	% sentence query - TBQ is no longer used
	pr::prSentenceQuery(0,SynTree,PR),
	transpr::proPR_Term(PR,PR_Term),
	pie3_IF(PR_Term,SolRef),
	!
;
	dlgIF::feil("pie3_Syn",[],[],SynTree),!,fail.	
% ----------------------------------------------------------------------------------------------
%  procXml
% ----------------------------------------------------------------------------------------------
class predicates
procSegments : (integer,integer,string,string,fileSeg,count,count,count,count) determ.
procSeg : (string,string,count,count,count,count) determ.
giveMess : (integer,integer,integer,integer,integer).
clauses

procXml(Format,K_Type,Table,F) :-
    Csn=count::newCounter(0),
    Csr=count::newCounter(0),
    Ccr=count::newCounter(0),
    Cfn=count::newCounter(0),
    Realm=fioExt::makeRealmName(F),
    IType=globalVars::get_i("uc_xml")+2,
    set::setGet([8,0,16],IType,Type),
    File2=fileSeg::openRead(F,"<sentence ","</sentence>",1),
    File2:openReadFile,
    procSegments(Format,K_Type,Table,Realm,File2,Csn,Csr,Ccr,Cfn),
    N1=Csn:get()+Csr:get()+Ccr:get()+Cfn:get(),
    File2:close(),
    File3=fileSeg::openRead(F,"<sentenceRule ","</sentenceRule>",1),
    File3:openReadFile,
    procSegments(Format,K_Type,Table,Realm,File3,Csn,Csr,Ccr,Cfn),
    N2=Csn:get()+Csr:get()+Ccr:get()+Cfn:get()-N1,
    File3:close(),
    File4=fileSeg::openRead(F,"<inexactRule ","</inexactRule>",1),
    File4:openReadFile,
    procSegments(Format,K_Type,Table,Realm,File4,Csn,Csr,Ccr,Cfn),
    N3=Csn:get()+Csr:get()+Ccr:get()+Cfn:get()-N1-N2,
    File4:close(),
    File5=fileSeg::openRead(F,"<function ","</function>",1),
    File5:openReadFile,
    procSegments(Format,K_Type,Table,Realm,File5,Csn,Csr,Ccr,Cfn),
    N4=Csn:get()+Csr:get()+Ccr:get()+Cfn:get()-N1-N2-N3,
    File5:close(),
    giveMess(Type,N1,N2,N3,N4),
    !.

giveMess(Type,N1,N2,N3,N4) :-
    Type=0,
    Ssn=toString(N1),
    Ssr=toString(N2),
    Scr=toString(N3),
    Sfn=toString(N4),
    Msn=string::concat("Stored ",Ssn," sentences!"), 
    dlgIF::msgS(Msn),
    Msr=string::concat("Stored ",Ssr," sentence rules!"), 
    dlgIF::msgS(Msr),
    Mcr=string::concat("Stored ",Scr," inexact rules!"), 
    dlgIF::msgS(Mcr),
    Mfn=string::concat("Stored ",Sfn," functions!"), 
    dlgIF::msgS(Mfn),
    !
;
    !.
   
procSegments(Format,Type,Table,Realm,File,Csn,Csr,Ccr,Cfn) :- 
    Type=0,
    programControl::repeat,
        X=File:readNext(),
        procSeg(Realm,X,Csn,Csr,Ccr,Cfn),
        File:fileEnd(),
        !
;        
    Type<>0,
    Sf=toString(Format),
    St=toString(Type),
    Con=sqlBase::getConnection(),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    programControl::repeat,
        X=File:readNext(),
        sx::sxStrip(X,Xs),
        addStream(Con,Table,Time,Sf,St,Xs,Csn),
        File:fileEnd(),
        N=Csn:get(),
        Ns=toString(N),
        Mess=string::concat("Stored ",Ns," rows!"), 
        dlgIF::msgS(Mess),
        !.
    
procSeg(Realm,X,Csn,Csr,Ccr,Cfn) :-
    X="",
    !
;
    Xt=xmlTree::newTree(X),
    T=Xt:get(),
    xmlP3::xmlProcessTrees(Realm,[T],Nsn,Nsr,Ncr,Nfn),
    Csn:add(Nsn,_),
    Csr:add(Nsr,_),
    Ccr:add(Ncr,_),
    Cfn:add(Nfn,_),
    !.
% ----------------------------------------------------------------------------------------------
% delRealm
% ----------------------------------------------------------------------------------------------
% filters out files corresponding to existing realms which KE rejects to delete
%
delRealm([],[]) :- 
	!.	
delRealm([File|Files],[File|Fs]) :-
    T=fileName::getExtension(File),
	Tl=string::toLowercase(T),
	set::member(Tl,["3dl","3nl","3pr"]),
    Realm=fioExt::makeRealmName(File),
	dbBase::realmExists(Realm,KBs),
	set::size(KBs,NFiles),
	NFiles>0,
	proCheckRealm(KBs,Realm),
	delRealm(Files,Fs),
	!.
delRealm([File|Files],Fs) :-
    T=fileName::getExtension(File),
	Tl=string::toLowercase(T),
	set::member(Tl,["3dl","3nl","3pr"]),
    Realm=fioExt::makeRealmName(File),
	dbBase::realmExists(Realm,KBs),
	set::size(KBs,NFiles),
	NFiles>0,
	delRealm(Files,Fs),
	!.
delRealm([File|Files],[File|Fs]) :-
	delRealm(Files,Fs),
	!.
% ----------------------------------------------------------------------------------------------
% proc3dl
% ----------------------------------------------------------------------------------------------
proc3dl(F) :-
    Realm=fioExt::makeRealmName(F),
    C=count::newCounter(0),
    /*
    IType=globalVars::get_i("uc_3dl")+2,
    set::setGet([8,0,16],IType,Type),
    */
    File=fileSeg::openRead(F,"\n",-1),
    File:openReadFile,
    S1=File:readNext(),
    Delimiter=string::trim(S1),
    S2=File:readNext(),
    ST=string::trim(S2),
                sx::concat(["Processing ",Realm],Mess),
                dlgIF::msgS(Mess),
	sx::sxUnpackS(ST,Delimiter,L),
	delInput(L,Delimiter,Realm,File,C),	
    File:close(),
    N=C:get(),
    Sn=toString(N),
    Mfn=string::concat("Stored ",Sn," sentences!"), 
    dlgIF::msgS(Mfn),
    !.

delInput([Px,Eax,Ebx],D,Realm,File,C) :-
	sx::sxStrip(Px,P),
	sx::sxStrip(Eax,Ea),
	sx::sxStrip(Ebx,Eb),
	dbBase::kbID(P,Pk,_,_),
	dbBase::kbID(Ea,Eak,_,_),
	dbBase::kbID(Eb,Ebk,_,_),
	dbBase::smClassDefTree(Pk,"p",PDef),
	PDef=cmp(_,[_,_,cmp("ATTRS",PDEs)]),
	dbBase::smClassDefTree(Eak,"e",EaDef),
	EaDef=cmp("ENT_TYPE",[_,cmp("ATTRS",EaDEs)]),
	dbBase::smClassDefTree(Ebk,"e",EbDef),
	EbDef=cmp("ENT_TYPE",[_,cmp("ATTRS",EbDEs)]),
	delScan(Pk,Eak,Ebk,PDEs,EaDEs,EbDEs,D,File,Realm,C),
	!
;
	sx::sxStrip(Px,P),
	sx::sxStrip(Eax,Ea),
	sx::sxStrip(Ebx,Eb),
	dbBase::kbID(P,Pk,_,_),
	dbBase::kbID(Ea,Eak,_,_),
	dbBase::kbID(Eb,Ebk,_,_),
	dbBase::smClassDefTree(Eak,"e",EaDef),
	EaDef=cmp("ENT_TYPE",[_,cmp("ATTRS",EaDEs)]),
	dbBase::smClassDefTree(Ebk,"e",EbDef),
	EbDef=cmp("ENT_TYPE",[_,cmp("ATTRS",EbDEs)]),
	delScan(Pk,Eak,Ebk,[],EaDEs,EbDEs,D,File,Realm,C),
	!.
delInput([Px,Eax],D,Realm,File,C) :-
	sx::sxStrip(Px,P),
	sx::sxStrip(Eax,Ea),
	dbBase::kbID(P,Pk,_,_),
	dbBase::kbID(Ea,Eak,_,_),
	dbBase::smClassDefTree(Pk,"p",PDef),
	PDef=cmp(_,[_,_,cmp("ATTRS",PDEs)]),
	dbBase::smClassDefTree(Eak,"e",EaDef),
	EaDef=cmp("ENT_TYPE",[_,cmp("ATTRS",EaDEs)]),
	delScan(Pk,Eak,"",PDEs,EaDEs,[],D,File,Realm,C),
	!
;
	sx::sxStrip(Px,P),
	sx::sxStrip(Eax,Ea),
	dbBase::kbID(P,Pk,_,_),
	dbBase::kbID(Ea,Eak,_,_),
	dbBase::smClassDefTree(Eak,"e",EaDef),
	EaDef=cmp("ENT_TYPE",[_,cmp("ATTRS",EaDEs)]),
	delScan(Pk,Eak,"",[],EaDEs,[],D,File,Realm,C),
	!.
delInput(L,_,Realm,_,_) :-
	dlgIF::feil("delInput",["Error in delimited file header record",Realm|L],[],nill),!.
% ----------------------------------------------------------------------------------------------
% proc3
% ----------------------------------------------------------------------------------------------
class predicates
txtInput : (integer,integer,string,string,fileSeg,count,string) determ.
clauses

proc3(K_Format,K_Type,FType,F,Ft,Table) :-
    Realm=fioExt::makeRealmName(F),
    C=count::newCounter(0),
    File=fileSeg::openRead(Ft,"!",-1),
    File:openReadFile,
                sx::concat(["Processing ",Realm],Mess),
                dlgIF::msgS(Mess),
	txtInput(K_Format,K_Type,FType,Realm,File,C,Table),	
    File:close(),
    N=C:get(),
    Sn=toString(N),
    Mfn=string::concat("Stored ",Sn," records!"), 
    dlgIF::msgS(Mfn),
    !.

txtInput(K_Format,K_Type,FType,Realm,File,C,Table) :-
    K_Type=0,
	programControl::repeat,	
        Sr=File:readNext(),
        sx::sxStrip(Sr,S),
	   			tt::tt_ss(2,S),
       proText(FType,Realm,[S],C),
	   delEnd(File),
    !
;
    K_Type<>0,
    Sf=toString(K_Format),
    St=toString(K_Type),
    Con=sqlBase::getConnection(),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
	programControl::repeat,	
        S1=File:readNext(),
        sx::sxStrip(S1,S2),
        addStream(Con,Table,Time,Sf,St,S2,C),
	    delEnd(File),
    !.	   
% ----------------------------------------------------------------------------------------------
% addStream
% ----------------------------------------------------------------------------------------------
addStream(Con,Table,Time,Format,Type,Knowledge,C) :-
    Knowledge<>"",
    Sql=string::concatList(["insert ",Table," (k_format,k_type,knowledge,created,processed,kb,source) values (",Format,",",Type,",'",Knowledge,"','",Time,"',0,'','')"]),
    sqlTools::executeDirect(Con,Sql),
    C:inc(_),
    !
;
    !.
% ----------------------------------------------------------------------------------------------
% delScan
% ----------------------------------------------------------------------------------------------
class predicates
delEnd : (fileSeg) determ.
delScanMess : (integer).
delScanX : (string,string,string,string,sterml,sterml,sterml,string,sterml,count,string) determ.
clauses

delScan(P,Ea,Eb,PDEs,EaDEs,EbDEs,D,File,Realm,C) :-
	set::conc([EaDEs,EbDEs,PDEs],DEs),
	programControl::repeat,	
        Sr=File:readNext(),
        sx::sxStrip(Sr,S),
	   			tt::tt_ss(2,S),
	   delScanX(S,P,Ea,Eb,PDEs,EaDEs,EbDEs,D,DEs,C,Realm),
	   delEnd(File),
	!.
delEnd(File) :-	
	globalVars::get_i("quit_flag")=1,
	!
;	
	globalVars::get_i("error_flag")=1,
	!
;	
    File:fileEnd(),
	!.	
delScanX(_,_,_,_,_,_,_,_,_,_,_) :-
	globalVars::get_i("quit_flag")=1,
	!
;	
	globalVars::get_i("error_flag")=1,
	!.
delScanX("",_,_,_,_,_,_,_,_,_,_) :-
	!.
delScanX(S,P,Ea,Eb,PDEs,EaDEs,EbDEs,D,DEs,C,Realm) :-
	sx::sxUnpackS(S,D,SVals),
	sx::sxlTrim(SVals,SVals2),
	delReadVals(DEs,SVals2,PVals),
	delGenPR(P,Ea,Eb,PDEs,EaDEs,EbDEs,PVals,PR),
	transpr::proPR_Term(PR,Term),
	C:inc(SentNo),
    delScanMess(SentNo),
	dbBase::recordInsert(2,Realm,SentNo,0,Term,atom("true"),_),
	!.
	
delScanMess(N) :-
	R=N mod 10,
	R=0,
	S=toString(N),
	sx::concat([S," sentences processed."],Mess),
	dlgIF::msgStatus(Mess),
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
% delGenPR
% ----------------------------------------------------------------------------------------------
class predicates
delGenPR_X : (sterml,string_list,string_list,string_list) determ (i,i,o,o).	
clauses

delGenPR(P,Ea,_,[],EaDEs,[],PVals,PR) :-
	delGenPR_X(EaDEs,PVals,_,La),
	sx::sxlJoin(La,",",Sa),
	sx::concat([P,"(",Ea,"(",Sa,"))."],PR),
	!.
delGenPR(P,Ea,Eb,[],EaDEs,EbDEs,PVals,PR) :-
	delGenPR_X(EaDEs,PVals,P2,La),
	delGenPR_X(EbDEs,P2,_,Lb),
	sx::sxlJoin(La,",",Sa),
	sx::sxlJoin(Lb,",",Sb),
	sx::concat([P,"(",Ea,"(",Sa,"),",Eb,"(",Sb,"))."],PR),
	!.
delGenPR(P,Ea,_,PDEs,EaDEs,[],PVals,PR) :-
	delGenPR_X(EaDEs,PVals,P2,La),
	delGenPR_X(PDEs,P2,_,Lp),
	sx::sxlJoin(La,",",Sa),
	sx::sxlJoin(Lp,",",Sp),
	sx::concat([P,"(",Ea,"(",Sa,"),",Sp,")."],PR),
	!.
delGenPR(P,Ea,Eb,PDEs,EaDEs,EbDEs,PVals,PR) :-
	delGenPR_X(EaDEs,PVals,P2,La),
	delGenPR_X(EbDEs,P2,P3,Lb),
	delGenPR_X(PDEs,P3,_,Lp),
	sx::sxlJoin(La,",",Sa),
	sx::sxlJoin(Lb,",",Sb),
	sx::sxlJoin(Lp,",",Sp),
	sx::concat([P,"(",Ea,"(",Sa,"),",Eb,"(",Sb,"),",Sp,")."],PR),
	!.
	
delGenPR_X([],P,P,[]) :-
	!.
delGenPR_X([_|Ds],[V|Vs],Vr,[V|Vm]) :-
	delGenPR_X(Ds,Vs,Vr,Vm),
	!.
% ----------------------------------------------------------------------------------------------
% delReadVals
% ----------------------------------------------------------------------------------------------
delReadVals([],V,[]) :-
	set::size(V,N),
	N>0,
	dlgIF::feil("delReadVals",["Too many fields in record"|V],[],nill),!,fail
;
	!.	
delReadVals([fstr(_,Dom,_,_)|DEs],[SVal|SVals],[PVal|PVals]) :-
	syn::sbGetDE_Val(Dom,SVal,PVal),
	delReadVals(DEs,SVals,PVals),
	!
;
	dlgIF::feil("delReadVals",["Field error",Dom,SVal],[],nill),!,fail.	
% ----------------------------------------------------------------------------------------------
% preProcess
% ----------------------------------------------------------------------------------------------
preProcess(F,Fr) :-
    Fr=filename::setExtension(F,"tmp"),
    Fs=fileSeg::openWrite(F,Fr,"/*","*/",1),
    Fs:openWriteFile,
    Fs:stripInv,
    Fs:close,
    !.
% ----------------------------------------------------------------------------------------------
% loadStandard
% ----------------------------------------------------------------------------------------------
loadStandardConf () :-
	Reply=vpiCommonDialogs::ask("Pro/3 standard knowledge files","The standard knowledge files will now be loaded. You normally load ALL files (except when you are creating a multi-database configuration and possibly need only the minimum).",["All","All ex. functions","Minimum"]),
	loadStandard(Reply),
	!.
	
loadStandard(0) :-
	P=globalVars::get_s("exe_path"),
	Fa=string::concat(P,"\\std\\PRO3-B-ANN.3pr"),
	fioExt::filCheckFile(1,"Pro/3 system file",Fa,Far),
	fio::file_name(Far,P2,_,_),
	Fb=string::concat(P2,"PRO3-B-BASE.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fb,Fbr),
	Fb2=string::concat(P2,"PRO3-B-CRT.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fb2,Fbr2),
	Fc=string::concat(P2,"PRO3-B-FNS.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fc,Fcr),
	Fd=string::concat(P2,"PRO3-B-META.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fd,Fdr),
	Fd2=string::concat(P2,"PRO3-E-CRT.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fd2,Fdr2),
	Fd3=string::concat(P2,"PRO3-EF-DATE.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fd3,Fdr3),
	Fd4=string::concat(P2,"PRO3-EF-MATH.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fd4,Fdr4),
	Fd5=string::concat(P2,"PRO3-EF-MISC.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fd5,Fdr5),
	Fe=string::concat(P2,"PRO3-E-FNT.3pr"),fioExt::filCheckFile(1,"Pro/3 system file",Fe,Fer),
	Ff=string::concat(P2,"PRO3-E-LEX.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Ff,Ffr),
	Fg=string::concat(P2,"PRO3-E-Q1.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fg,Fgr),
	Fh=string::concat(P2,"PRO3-E-TERM.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fh,Fhr),
	Fh2=string::concat(P2,"PRO3-F-DATE.3nl"), fioExt::filCheckFile(1,"Pro/3 system file",Fh2,Fhr2),
	Fh3=string::concat(P2,"PRO3-F-MATH.3nl"), fioExt::filCheckFile(1,"Pro/3 system file",Fh3,Fhr3),
	Fh4=string::concat(P2,"PRO3-F-MISC.3nl"), fioExt::filCheckFile(1,"Pro/3 system file",Fh4,Fhr4),
	IntCheck=dep::getIntegrityCheck(),
    dep::setIntegrityCheck(-1),
    dep::setSmGraphUpdate(-1),
	vpi::cursorSetWait(),
	cache::removeTm,
	cache::removeSm,
	dbBase::realmDelete("PRO3-B-ANN.3PR"),
	dbBase::realmDelete("PRO3-F-MISC.3NL"),
	dbBase::realmDelete("PRO3-F-MATH.3NL"),
	dbBase::realmDelete("PRO3-F-DATE.3NL"),
	dbBase::realmDelete("PRO3-E-Q1.3PR"),
	dbBase::realmDelete("PRO3-E-TERM.3PR"),
	dbBase::realmDelete("PRO3-E-FNT.3PR"),
	dbBase::realmDelete("PRO3-EF-MISC.3PR"),
	dbBase::realmDelete("PRO3-EF-MATH.3PR"),
	dbBase::realmDelete("PRO3-EF-DATE.3PR"),
	dbBase::realmDelete("PRO3-E-CRT.3PR"),
	dbBase::realmDelete("PRO3-E-LEX.3PR"),
	dbBase::realmDelete("PRO3-B-FNS.3PR"),
	dbBase::realmDelete("PRO3-B-CRT.3PR"),
	dbBase::realmDelete("PRO3-B-META.3PR"),
	dbBase::realmDelete("PRO3-B-BASE.3PR"),
	globalVars::set_s("lang","p3_iEngl"),
	vpi::cursorSetWait(),
	processFiles([Fdr,Ffr,Fbr,Fbr2,Fcr,Fdr2,Fdr3,Fdr4,Fdr5,Fer,Fgr,Fhr]),
    cache::loadTm("p3_iEngl"),
    cache::loadSm,
	processFiles([Fhr2,Fhr3,Fhr4,Far]),
    cache::loadSm,
    dep::setIntegrityCheck(IntCheck),
    dep::setSmGraphUpdate(1),
    dep::depSMM_Regen(),
	!.
loadStandard(1) :-
	P=globalVars::get_s("exe_path"),
	Fa=string::concat(P,"\\std\\PRO3-B-ANN.3pr"),
	fioExt::filCheckFile(1,"Pro/3 system file",Fa,Far),
	fio::file_name(Far,P2,_,_),
	Fb=string::concat(P2,"PRO3-B-BASE.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fb,Fbr),
	Fb2=string::concat(P2,"PRO3-B-CRT.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fb2,Fbr2),
	Fc=string::concat(P2,"PRO3-B-FNS.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fc,Fcr),
	Fd=string::concat(P2,"PRO3-B-META.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fd,Fdr),
	Fd2=string::concat(P2,"PRO3-E-CRT.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fd2,Fdr2),
	Fe=string::concat(P2,"PRO3-E-FNT.3pr"),fioExt::filCheckFile(1,"Pro/3 system file",Fe,Fer),
	Ff=string::concat(P2,"PRO3-E-LEX.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Ff,Ffr),
	Fg=string::concat(P2,"PRO3-E-Q1.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fg,Fgr),
	Fh=string::concat(P2,"PRO3-E-TERM.3pr"), fioExt::filCheckFile(1,"Pro/3 system file",Fh,Fhr),
	IntCheck=dep::getIntegrityCheck(),
    dep::setIntegrityCheck(-1),
    dep::setSmGraphUpdate(-1),
	vpi::cursorSetWait(),
	cache::removeTm,
	cache::removeSm,
	dbBase::realmDelete("PRO3-B-ANN.3PR"),
	dbBase::realmDelete("PRO3-E-Q1.3PR"),
	dbBase::realmDelete("PRO3-E-TERM.3PR"),
	dbBase::realmDelete("PRO3-E-FNT.3PR"),
	dbBase::realmDelete("PRO3-E-CRT.3PR"),
	dbBase::realmDelete("PRO3-E-LEX.3PR"),
	dbBase::realmDelete("PRO3-B-FNS.3PR"),
	dbBase::realmDelete("PRO3-B-CRT.3PR"),
	dbBase::realmDelete("PRO3-B-META.3PR"),
	dbBase::realmDelete("PRO3-B-BASE.3PR"),
	globalVars::set_s("lang","p3_iEngl"),
	vpi::cursorSetWait(),
	processFiles([Fdr,Ffr,Fbr,Fbr2,Fcr,Fdr2,Fer,Fgr,Fhr]),
    cache::loadTm("p3_iEngl"),
    cache::loadSm,
	processFiles([Far]),
    cache::loadSm,
    dep::setIntegrityCheck(IntCheck),
    dep::setSmGraphUpdate(1),
    dep::depSMM_Regen(),
	!.
loadStandard(2) :-
	vpi::cursorSetWait(),
	P=globalVars::get_s("exe_path"),
	Fa=string::concat(P,"\\std\\PRO3-B-BASE.3pr"),
	fioExt::filCheckFile(1,"Pro/3 system file",Fa,Far),
	fio::file_name(Far,P2,_,_),
	Fh=string::concat(P2,"PRO3-E-TERM.3pr"),
	fioExt::filCheckFile(1,"Pro/3 system file",Fh,Fhr),
	dbBase::realmDelete("PRO3-B-BASE.3PR"),
	dbBase::realmDelete("PRO3-E-TERM.3PR"),
	processFiles([Far,Fhr]),
	!.
% ----------------------------------------------------------------------------------------------
% createSQL_Table
% ----------------------------------------------------------------------------------------------
class predicates
createSQL_Columns : (odbcConnection,integer,integer,integer,integer,integer,string_list,string_list,string_list,integer) determ (i,i,i,i,i,i,i,o,o,o).
clauses

% creates an sql table corresponding to the given pro/3 table.
% table name: 'rn_m_xxx' where n=request_no, M=subno and xxx is derived from the ST
% column names: derived from NL_name and postfixed with _n where n is argument# if required
%
createSQL_Table(ODBC,SolRef,Table,SqlTable,TableName,NRows,NCols,Domains,STn,Width) :-
    stb::stbH_Get(SolRef,Table,NCols,NRows,ST,_,_,_),
    NRows>0,
    dbBase::nlSing(ST,STn),
    sqlBase::getParameters(Max,_,_,Blank,Case,_,_,_),
    Maxx=Max-12,
    sx::makeIdentifier(Case,Case,Case,STn,Maxx,Blank,STn1),
    sqlTools::makeSqlName(STn1,Max,STn2),
    R=globalVars::get_i("request_no"),
    Rs=toString(R),
    SR=globalVars::get_i("request_subno"),
    SRs=toString(SR),
    TableName=string::concat("r",Rs,"_",SRs,"_",STn2),
    createSQL_Columns(ODBC,SolRef,Table,SqlTable,1,NCols,[],CL,Domains,Width),
    sx::sxlJoin(CL,",",Cs),
    S=string::concatList(["create table ",TableName," (n int,",Cs,")"]),
    sqlTools::executeDirect(ODBC,S),
    !
;
    TableName="",
    NRows=0,
    NCols=0,
    Domains=[],
    STn="",
    Width=0,
    !.
    
createSQL_Columns(_,_,_,_,N,NCols,_,[],[],0) :-
    N>NCols,
    !.
createSQL_Columns(ODBC,SolRef,Table,SqlTable,ColNo,NCols,Created,[C|Cs],[Domain|Domains],W) :-
    stb::stbD_GetCol(SolRef,Table,ColNo,ArgNo,_,_,DEn,Domain,Form,Width,_,_,_),
    sqlBase::getParameters(Max,_,_,Blank,Case,_,_,_),
    Maxx=Max-2,
    sx::makeIdentifier(Case,Case,Case,DEn,Maxx,Blank,DEn1),
    sqlTools::makeSqlName(DEn1,Max,DEn2),
    set::member(DEn2,Created),
    Q=toString(ArgNo),
    DEn3=string::concat("A",Q,DEn2),
    sqlTools::makeSqlName(DEn3,Max,DEn4),
    sqlSt::colType(Width,Form,Domain,SQL_Type),
    C=string::concat(DEn4," ",SQL_Type),
    dbBase::nlSing(Domain,NDomain),
    SWidth=toString(Width),
    R=globalVars::get_i("request_no"),
    Rs=toString(R),
    SR=globalVars::get_i("request_subno"),
    SRs=toString(SR),
    SColNo=toString(ColNo),
    STable=toString(SqlTable),
    S=string::concatList(["insert into p3bridge.p3response_column (request_no,sub_no,table_no,column_no,width,domain,title) values (",
                                                                               Rs,",",SRs,",",STable,",",SColNo,",",SWidth,",'",NDomain,"','",SQL_Type,"','",DEn,"','",DEn4,"')"]),
    sqlTools::executeDirect(ODBC,S),
    Next=ColNo+1,
    createSQL_Columns(ODBC,SolRef,Table,SqlTable,Next,NCols,[C|Created],Cs,Domains,W2),
    W=Width+W2,
    !
;    
    stb::stbD_GetCol(SolRef,Table,ColNo,_,_,_,DEn,Domain,Form,Width,_,_,_),
    sqlBase::getParameters(Max,_,_,Blank,Case,_,_,_),
    Maxx=Max-2,
    sx::makeIdentifier(Case,Case,Case,DEn,Maxx,Blank,DEn1),
    sqlTools::makeSqlName(DEn1,Max,DEn2),
    sqlSt::colType(Width,Form,Domain,SQL_Type),
    C=string::concat(DEn2," ",SQL_Type),
    dbBase::nlSing(Domain,NDomain),
    SWidth=toString(Width),
    R=globalVars::get_i("request_no"),
    Rs=toString(R),
    SR=globalVars::get_i("request_subno"),
    SRs=toString(SR),
    SColNo=toString(ColNo),
    STable=toString(SqlTable),
    S=string::concatList(["insert into p3bridge.p3response_column (request_no,sub_no,table_no,column_no,width,domain,sql_type,title,column_name) values (",
                                                                               Rs,",",SRs,",",STable,",",SColNo,",",SWidth,",'",NDomain,"','",SQL_Type,"','",DEn,"','",DEn2,"')"]),
    sqlTools::executeDirect(ODBC,S),
    Next=ColNo+1,
    createSQL_Columns(ODBC,SolRef,Table,SqlTable,Next,NCols,[C|Created],Cs,Domains,W2),
    W=Width+W2,
    !.
% ----------------------------------------------------------------------------------------------
% popSQL_Table
% ----------------------------------------------------------------------------------------------
class predicates
makeSqlRow : (odbcConnection,integer,integer,integer,integer,integer,string_list,string_list) determ (i,i,i,i,i,i,i,o).
adjustSqlValue : (string,string,string) procedure (i,i,o).
clauses

popSQL_Table(ODBC,SolRef,Table,TableName,NRows,NCols,Domains) :-
    stb::stbIsSorted(SolRef,Table,Sorted),
    Sorted=0,
	stb::foralltbl(SolRef,Table,Np),
        makeSqlRow(ODBC,SolRef,Table,Np,1,NCols,Domains,Values),
        sx::sxlJoin(Values,",",Sv),
        Ns=toString(Np),
        S=string::concatList(["insert into ",TableName," values (",Ns,",",Sv,")"]),
        sqlTools::executeDirect(ODBC,S),
        Np=NRows,
        !
;    
    stb::stbIsSorted(SolRef,Table,Sorted),
    Sorted<>0,
	stb::foralltblsorted(SolRef,Table,N,Np),
        makeSqlRow(ODBC,SolRef,Table,Np,1,NCols,Domains,Values),
        sx::sxlJoin(Values,",",Sv),
        Ns=toString(N),
        S=string::concatList(["insert into ",TableName," values (",Ns,",",Sv,")"]),
        sqlTools::executeDirect(ODBC,S),
        N=NRows,
        !.
	
makeSqlRow(_,_,_,_,Col,NCols,_,[]) :-
    Col>NCols,
    !.
makeSqlRow(ODBC,SolRef,Table,Row,Col,NCols,[Dom|Doms],[V|Values]) :-
  	stb::stbCellGet(SolRef,Table,Row,Col,Vn,_),
    adjustSqlValue(Dom,Vn,V),  	
    Next=Col+1,
    makeSqlRow(ODBC,SolRef,Table,Row,Next,NCols,Doms,Values),
    !.

adjustSqlValue(Dom,V,Val) :-
    set::member(Dom,["p3_dxd","p3_dxt","p3_dxdt"]),
    Val=sqlTools::xmlToDatetime(V),
    !
;
    dom::domNum(Dom),
    sx::sxTail(1,V,F,T),
    T="%",
    R=sx::toReal(F),
    Ra=R/100.0,
    Val=toString(Ra),
    !
;
    dom::domNum(Dom),
    sx::sxRemChar(V,',',Val),
    !
;
    sx::sxUnquote(V,V2),
    Val=string::concat("'",V2,"'"),
    !.
% ----------------------------------------------------------------------------------------------
% processQuestion
% ----------------------------------------------------------------------------------------------
storeAlts(_,_,_,_,[]) :-
       !.
storeAlts(ODBC,Rs,Ms,N,[A|Alts]) :-
    Ns=toString(N),   
    S=string::concatList(["insert into p3bridge.p3answer (request_no,question_no,alternative,answer) values (",Rs,",",Ms,",",Ns,",'",A,"')"]),
    sqlTools::executeDirect(ODBC,S),   
    Next=N+1, 
    storeAlts(ODBC,Rs,Ms,Next,Alts),
    !.    
% ----------------------------------------------------------------------------------------------
% processQuestion
% ----------------------------------------------------------------------------------------------
class predicates
getAnswer : (odbcConnection,string,string,string) determ (i,i,i,o).
checkAnswer : (count,odbcConnection,string,string,string) determ (i,i,i,i,o).
clauses

processQuestion(Rule,Question,Dom,AltsNL,Def,Answer) :-
    set::size(AltsNL,N),
    N>0,
    Ns=toString(N),
	R=globalVars::get_i("request_no"),
	Rs=toString(R),
	dbExtension::kbGetAnnotationBlank("p3_iIR",Rule,Anno),
	dbBase::nlSing(Rule,RuleNL),
	ODBC=sqlBase::getConnection(),
    S=string::concatList(["insert into p3bridge.p3question (request_no,question_no,question_type,question,status,default_answer,question_help,rule,domain) values (",
                                    Rs,",0,",Ns,",'",Question,"',1,'",Def,"','",Anno,"','",RuleNL,"','",Dom,"')"]),
    sqlTools::executeDirect(ODBC,S),
    storeAlts(ODBC,Rs,"0",1,AltsNL),
    getAnswer(ODBC,Rs,Def,Answer),
	!
;
    set::size(AltsNL,N),
    N=0,
	R=globalVars::get_i("request_no"),
	Rs=toString(R),
	dbExtension::kbGetAnnotationBlank("p3_iIR",Rule,Anno),
	dbBase::nlSing(Rule,RuleNL),
	ODBC=sqlBase::getConnection(),
    S=string::concatList(["insert into p3bridge.p3question (request_no,question_no,question_type,question,status,default_answer,question_help,rule,domain) values (",
                                    Rs,",0,0,'",Question,"',1,'",Def,"','",Anno,"','",RuleNL,"','",Dom,"')"]),
    sqlTools::executeDirect(ODBC,S),
    getAnswer(ODBC,Rs,Def,Answer),
	!.

getAnswer(ODBC,Rs,Def,S) :-
    Counter=count::new(2),
    programControl::sleep(2000),
    programControl::repeat,
        programControl::sleep(1000),
        Counter:add(1,_),
        checkAnswer(Counter,ODBC,Rs,Def,S),
        !.
        
checkAnswer(Counter,ODBC,Rs,Def,Ans) :-
    S=string::concat("select request_status from p3bridge.p3request where request_no=",Rs),
    Sql=sqlSelect::new(ODBC,S),
    Status=Sql:getI(1,1),
    Status<0,
    % user break or other error 
    ir::setBreak(1),
    Ans=Def,
    !
;
    Counter:get()>30, % time-out when waiting for question answer
    ir::setBreak(2),
    S5=string::concat("update p3bridge.p3request set request_status=-12 where request_no=",Rs),
    sqlTools::executeDirect(ODBC,S5),
    Ans=Def,
    !
;    
    S=string::concat("select status,answer from p3bridge.p3question where status>=3 and question_no=0 and request_no=",Rs),
    Sql=sqlSelect::new(ODBC,S),
    %Status=Sql:getI(1,1),
    S2=Sql:getS(1,2),
    Ans=string::trim(S2),
    %checkBreak(Status,Ans2,Def,Ans),
    S3=string::concat("delete from p3bridge.p3question where question_no=0 and request_no=",Rs),
    sqlTools::executeDirect(ODBC,S3),
    S4=string::concat("delete from p3bridge.p3answer where question_no=0 and request_no=",Rs),
    sqlTools::executeDirect(ODBC,S4),
    !.
/*    
checkBreak : (integer,string,string,string) procedure (i,i,i,o).
checkBreak(Status,AnsG,Def,Ans) :-
    Status=4,
    ir::setBreak(1),
    % request_status has been updated by client
    Ans=Def,
    !
;
    Ans=AnsG,
    !.
*/    
end implement proc
