/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement pscan
open core,pie_scan

constants
   className = "com/rudstrand/pscan/pscan".
   classVersion = "1.0".

clauses
classInfo(className, classVersion).

class predicates
pscFixOther : (tokl,tokl) procedure (i,o).
pscFixNumComma : (string,string) procedure (i,o).
clauses  

pscNL(S,T) :-
    St=string::trim(S),
	pscFixNumComma(St,Sc),
	pie_scan::tokl(0,Sc,Tm),
	pscFixOther(Tm,T),
	!
;
	pfeil::syserr("pscNL","NL scanning error",[S]),!,fail.	

pscPR(S,T) :-
	pie_scan::tokl(0,S,T),
	!
;
	pfeil::syserr("pscPR","PR scanning error",[S]),!,fail.	

% replaces sub-strings of the type '<digit><comma><digit>' with '<digit><digit>'
%
pscFixNumComma(S,T) :-
	L=string::length(S),
	P=string::searchchar(S,',',string::caseSensitive()),
	P>0,
	P<(L-1),
	P1=P-1,
	P2=P+1,
	C1=string::subchar(S,P1),
	sx::isDigit(C1),
	C2=string::subchar(S,P2),
	sx::isDigit(C2),
    string::front(S,P2,_,T2),
    string::front(S,P,Front,_),
	pscFixNumComma(T2,Tail),
	T=string::concat(Front,Tail),
	!
;
	P=string::searchchar(S,',',string::caseSensitive()),
	P2=P+1,
	string::front(S,P2,Front,T2),
	pscFixNumComma(T2,Tail),
	T=string::concat(Front,Tail),
	!	
;
	T=S,
	!.	
	
pscFixOther([],[]) :- 
	!.
pscFixOther([t(atom("=*"),P),t(var(V),Q)|T],[t(atom(Opr),P)|Tf]) :-
	P=Q-2,
	set::member(V,["IN","NU","LI","IF","ID","IT","SY","KB","INC","DEC","PR","CF","MG","XD","XT","XDT","XDR"]),
	Opr=string::concat("=*",V),
	pscFixOther(T,Tf),
	!.
pscFixOther([t(atom("="),P),t(var(V),Q)|T],[t(atom(Opr),P)|Tf]) :-
	P=Q-1,
	set::member(V,["IN","NU","LI","IF","ID","IT","SY","KB","INC","DEC","PR","CF","MG","XD","XT","XDT","XDR"]),
	Opr=string::concat("=",V),
	pscFixOther(T,Tf),
	!.
pscFixOther([t(atom("=-"),P)|T],[t(atom("="),P),t(atom("-"),PP)|Tf]) :-
	PP=P+1,
	pscFixOther(T,Tf),
	!.
pscFixOther([t(atom("=+"),P)|T],[t(atom("="),P),t(atom("+"),PP)|Tf]) :-
	PP=P+1,
	pscFixOther(T,Tf),
	!.
pscFixOther([t(int(I),P),t(atom("%"),_)|T],[t(fix(Fp),P)|Tf]) :-
    F=I,
    sx::movePoint(-2,F,Fp),
	pscFixOther(T,Tf),
	!.
pscFixOther([t(fix(F),P),t(atom("%"),_)|T],[t(fix(Fp),P)|Tf]) :-
    sx::movePoint(-2,F,Fp),
	pscFixOther(T,Tf),
	!.
pscFixOther([H|T],[H|Tf]) :- 
	pscFixOther(T,Tf),
	!.
% --------------------------------------------------------------------------
% str_tokens                                                              
% --------------------------------------------------------------------------
str_tokens([],[]) :- !.
str_tokens([t(var(S),_)|T],[termDomain::str(S)|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(gvar(S),_)|T],[termDomain::str(S)|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(atom(S),_)|T],[termDomain::str(S)|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(chr(C),_)|T],[termDomain::str(S)|Tm]) :- S=string::charToString(C), str_tokens(T,Tm),!.
str_tokens([t(int(I),_)|T],[termDomain::str(S)|Tm]) :- S=toString(I), str_tokens(T,Tm),!.
str_tokens([t(fix(F),_)|T],[termDomain::str(S)|Tm]) :- S=toString(F), str_tokens(T,Tm),!.
str_tokens([t(str(S),_)|T],[termDomain::str(S)|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(lbrack,_)|T],[termDomain::str("[")|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(rbrack,_)|T],[termDomain::str("]")|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(lpar,_)|T],[termDomain::str("(")|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(rpar,_)|T],[termDomain::str(")")|Tm]) :- str_tokens(T,Tm),!.
%27.10.05 str_tokens([t(lcurly,_)|T],[termDomain::str("{")|Tm]) :- str_tokens(T,Tm),!.
%27.10.05 str_tokens([t(rcurly,_)|T],[termDomain::str("}")|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(comma,_)|T],[termDomain::str(",")|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(bar,_)|T],[termDomain::str("|")|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(dot,_)|T],[termDomain::str(".")|Tm]) :- str_tokens(T,Tm),!.
str_tokens([t(T,_)|_],_) :- 
	S1=toString(T),
	pfeil::syserr("str_tokens","Unrecognized token",[S1]),
	!,fail.	
end implement pscan
