/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement rn
open core,termDomain

constants
className = "com/rudstrand/rn/rn".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

class predicates
rnlArg : (string_list,integer,string,sterml,sterml,sterm) determ (i,i,i,i,i,o).	
rnlConditions : (string_list,sterml,sterml) determ (i,i,o).
rnlDEs : (integer,integer,sterm,string_list) procedure (i,o,i,o).
rnlDE_Vars : (string_list,string_list) determ (i,o).
rnlGetDE_ByProxy : (string,string_list,string) determ (i,i,o).
rnlGetProxy : (string,string,string_list,string) determ (i,i,i,o).
rnlMarkProxy : (string_list,string,string_list) determ (i,i,o).
rnlNonProxy : (sterm,string_list,string_list) procedure (i,i,o).
rnlOper : (string,string) procedure (i,o).
rnlProxy : (integer,integer,sterm,string_list,string_list) procedure (i,o,i,i,o).
rnlRemoveDummyConds : (sterm,sterm) procedure (i,o).                       
rnlRSide : (integer,integer,string_list,sterm,sterm,string) determ (i,o,i,i,o,o).
rnlRT : (integer,integer,string_list,string_list,sterm,sterm,string_list) determ (i,o,i,i,i,o,o).
rnlSentCond : (integer,integer,string_list,sterm,sterm) determ (i,o,i,i,o).
rnlVarCond : (integer,integer,string_list,string_list,sterm,sterm,string_list) determ (i,o,i,i,i,o,o).
clauses
% ----------------------------------------------------------------------------------------------
% rnlST_RT
% ----------------------------------------------------------------------------------------------
% translates from syntax tree to rule tree
% parameters:
% 1	in	syntax tree to be translated to rule-tree
% 2	out	rule-tree
% 3	out	rule-type
% 4	out	highest use argument suffix
% 5	out	DEs for the tree DE listbox
%
rnlST_RT(ST,Tree,SType,DE_Vars) :-
	ST=cmp("=>",[Condition,Conclusion]),
	SType="1",
	% get DE lists - element format: <argument_no> | <entity_NL> | <DE_NL> | <DE_KB>
	rnlDEs(0,_,Conclusion,DEs2),
	rnlDEs(1,_,Condition,DEs3),
	set::conc([DEs2,DEs3],DEs),
	% get DE lists with proxy info - element format: <argument_no> | <entity_NL> | <DE_NL> | <DE_KB> [ | <proxy_var> ]
	rnlProxy(1,_,Condition,DEs,DEsP2),
	rnlNonProxy(Condition,DEsP2,DEsP),	
				tt::tt_strings(3,"Rule DE proxies",DEsP),
	rnlRT(0,_,DEsP,[],Conclusion,Conc,Vars2),
	rnlRT(1,_,DEsP,Vars2,Condition,Cond,Vars),
				tt::tt_strings(3,"Rule-tree variables",Vars),
	rnlDE_Vars(DEsP,DE_Vars),
	Tree2=cmp("=>",[Cond,Conc]),
	rnlRemoveDummyConds(Tree2,Rule3),
				tt::tt_term(3,"Rule-tree (unlinearized)",Rule3),
	struc::treAndLinearize(Rule3,Tree),
				tt::tt_term(2,"Rule-tree",Tree),
	!.
% ----------------------------------------------------------------------------------------------
% rnlArg
% ----------------------------------------------------------------------------------------------
class predicates
rnlElimDummy : (string,string_list,sterml,sterml) determ (i,i,i,o).
rnlDE_Conds : (string,string,string_list,sterml,sterml,sterml) determ (i,i,i,i,i,o).	
rnlGenCond : (string_list,string,string,integer,sterml,sterm) determ (i,i,i,i,i,o).
rnlGetConds : (string,sterml,sterml) determ (i,i,o).
clauses
% constructs the rule sub-tree corresponding to a syntax tree predicate condition argument
% parameters:
% 1	in	DE list incl. proxy marks
% 2	in	argument unique no
% 3	in	argument (KB)
% 4	in 	argument's DEs
% 5	in	conditions for this argument
% 6	out	argument rule tree
%
/* target arg format:
          eNational[11]
               ENational_11dNationality
                    =
                         iDanish {atom}
*/                        
rnlArg(Ds,N,A,DEs,Conds,Rule) :-
	Ns=toString(N),
	sx::concat([A,"[",Ns,"]"],Ab),
	rnlDE_Conds(Ns,A,Ds,DEs,Conds,Dcs),
	Rule=cmp(Ab,Dcs),
	!.	
	
rnlDE_Conds(_,_,_,[],_,[]) :-
	!.
rnlDE_Conds(N,A,Ds,[fstr(DE,_,_,_)|DEs],Conds,[Dc|Dcs]) :-
	rnlGetConds(DE,Conds,Cs),
	rnlElimDummy(N,Ds,Cs,Csr),
	Ni=sx::toInt(N),
	rnlGenCond(Ds,DE,A,Ni,Csr,Dc),
	rnlDE_Conds(N,A,Ds,DEs,Conds,Dcs),
	!.	
% gets the conditions for given DE
%
rnlGetConds(_,[],[]) :-
	!.
rnlGetConds(DE,[cmp("ATTR_CONDS",Conds)],Cs) :-
	rnlGetConds(DE,Conds,Cs),
	!.
rnlGetConds(DE,[Cond|Conds],[Cond|Cs]) :-
	Cond=cmp(_,[fstr(DE,_,_,_),_]),
	rnlGetConds(DE,Conds,Cs),
	!.
rnlGetConds(DE,[Cond|Conds],[Cond|Cs]) :-
	Cond=cmp(_,[fvar(DE,_),_]), 
	rnlGetConds(DE,Conds,Cs),
	!.
rnlGetConds(DE,[_|Conds],Cs) :-
	rnlGetConds(DE,Conds,Cs),
	!.
% eliminates assignments between a DE and its proxy (since the proxy is not used in expressions)
%
rnlElimDummy(_,_,[],[]) :-
	!.
rnlElimDummy(N,Ds,[Cond|Conds],CondsM) :-
	Cond=cmp(Opr,[fvar(DE,_),fvar(V,_)]),
	opr::oprEq(Opr),
	rnlGetProxy(N,DE,Ds,Proxy), % DE has a proxy
	V=Proxy,
	rnlGetDE_ByProxy(Proxy,Ds,DE2),
	vars::varComps(DE2,_,PN,PD),
	PNs=toString(PN),
	N=PNs,
	DE=PD, % the proxy is genuine (and not a reference)
	rnlElimDummy(N,Ds,Conds,CondsM),
	!.		
rnlElimDummy(N,Ds,[Cond|Conds],[Cond|CondsM]) :-
	rnlElimDummy(N,Ds,Conds,CondsM),
	!.		
rnlGenCond(_,DE,A,N,[],var(DV)) :-
	vars::varCompsR(DV,A,N,DE),
	!.
rnlGenCond(Ds,DE,A,N,[Cond],C) :-
	rnlConditions(Ds,[Cond],CondM),
	vars::varCompsR(DV,A,N,DE),
	C=cmp(DV,CondM),
	!.
rnlGenCond(Ds,DE,A,N,Conds,C) :-
	rnlConditions(Ds,Conds,CondsM),
	vars::varCompsR(DV,A,N,DE),
	C=cmp(DV,[cmp("&",CondsM)]),
	!.
% ----------------------------------------------------------------------------------------------
% rnlCondition/s
% ----------------------------------------------------------------------------------------------
class predicates
rnlCondition : (string_list,sterm,sterm) determ (i,i,o).
clauses
rnlConditions(_,[],[]) :-
	!.
rnlConditions(Ds,[Cond|Conds],[C|Cs]) :-
	rnlCondition(Ds,Cond,C),
	rnlConditions(Ds,Conds,Cs),
	!.
rnlCondition(Ds,cmp(Opr,[_,RSide]),cmp(Oper,[RS])) :-
	rnlOper(Opr,Oper),
	rnlRSide(1,_,Ds,RSide,RS,_),
	!.
% ----------------------------------------------------------------------------------------------
% rnlOper
% ----------------------------------------------------------------------------------------------
rnlOper(Opr,OprM) :-
	sx::frontstr(2,Opr,F,_),
	F="is",
	OprM="=",
	!
;
	sx::frontstr(3,Opr,F,_),
	F="nis",
	OprM="=*",
	!
;
	OprM=Opr,
	!.
% ----------------------------------------------------------------------------------------------
% rnlDEs
% ----------------------------------------------------------------------------------------------
class predicates
rnlNLs : (integer,string,sterml,string_list) determ (i,i,i,o).
clauses
% generates the DE list: E|No|D|Dk
%
rnlDEs(N,M,cmp("&",[C2,C3]),DEs) :-
	rnlDEs(N,N2,C2,DEs2),
	rnlDEs(N2,M,C3,DEs3),
	set::conc([DEs2,DEs3],DEs),
	!.
rnlDEs(N,M,T,DEs) :-
	T=cmp("FALSE",[SC]),
	rnlDEs(N,M,SC,DEs),
	!
;
	T=cmp(_,[_,cmp("#S",[SC])]),
	rnlDEs(N,M,SC,DEs),
	!.
rnlDEs(N,M,ST,DEs) :-
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",
	          [cmp("REP",[atom(P)]),
	           _,
	           cmp("ATTRS",PDEs)]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",
	              [cmp("REP",[atom(Ea)]),
	               cmp("ATTRS",PEas)])
	          |_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",
	              [cmp("REP",[atom(Eb)]),
	               cmp("ATTRS",PEbs)])
	          |_])
	      |_]),
	PNo=N*10,	               
	dbBase::nlSing(P,Pn),
	rnlNLs(PNo,Pn,PDEs,Dp),				           
	EaNo=N*10+1,	               
	dbBase::nlSing(Ea,Ean),
	rnlNLs(EaNo,Ean,PEas,Dea),				           
	EbNo=N*10+2,	               
	dbBase::nlSing(Eb,Ebn),
	rnlNLs(EbNo,Ebn,PEbs,Deb),
	set::conc([Dp,Dea,Deb],DEs),
	M=N+1,
	! 
;	
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",
	          [cmp("REP",[atom(P)]),
	           _,
	           cmp("ATTRS",PDEs)]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",
	              [cmp("REP",[atom(Ea)]),
	               cmp("ATTRS",PEas)])
	          |_])
	      |_]),
	PNo=N*10,	               
	dbBase::nlSing(P,Pn),
	rnlNLs(PNo,Pn,PDEs,Dp),				           
	EaNo=N*10+1,	               
	dbBase::nlSing(Ea,Ean),
	rnlNLs(EaNo,Ean,PEas,Dea),				           
	set::conc([Dp,Dea],DEs),
	M=N+1,
	! 
;					           
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",[_,_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",
	              [cmp("REP",[atom(Ea)]),
	               cmp("ATTRS",PEas)])|_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",
	              [cmp("REP",[atom(Eb)]),
	               cmp("ATTRS",PEbs)])|_])|_]),
	EaNo=N*10+1,	               
	dbBase::nlSing(Ea,Ean),
	rnlNLs(EaNo,Ean,PEas,Dea),				           
	EbNo=N*10+2,	               
	dbBase::nlSing(Eb,Ebn),
	rnlNLs(EbNo,Ebn,PEbs,Deb),
	set::conc([Dea,Deb],DEs),
	M=N+1,
	! 
;					           
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",[_,_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",
	              [cmp("REP",[atom(Ea)]),
	               cmp("ATTRS",PEas)])
	          |_])]),
	EaNo=N*10+1,	               
	dbBase::nlSing(Ea,Ean),
	rnlNLs(EaNo,Ean,PEas,DEs),
	M=N+1,
	!. 
rnlDEs(N,N,_,[]) :-
	!.
rnlNLs(_,_,[],[]) :-
	!.
rnlNLs(N,E,[DD|Dks],[DE|DEs]) :-
	Ns=toString(N),
	DD=fstr(Dk,_,_,_),
	dbBase::nlSing(Dk,D),
	sx::sxlJoin([Ns,E,D,Dk],"|",DE),
	rnlNLs(N,E,Dks,DEs),
	!.
% ----------------------------------------------------------------------------------------------
% rnlDE_Vars
% ----------------------------------------------------------------------------------------------
% parameters
% 1	in	DE list, format: <argument_no> | <entity_NL> | <DE_NL> | <DE_KB> [ | <proxy_var> ]
% 2	out	DE variable list
%
rnlDE_Vars([],[]) :-
	!.
rnlDE_Vars([D|Ds],[DV|DVs]) :-
	sx::sxUnpack(D,'|',L),
	L=[N,En,_,K|_],
	dbBase::kbID(En,E,_,_),
	Ni=sx::toInt(N),
	vars::varCompsR(DV,E,Ni,K),
	rnlDE_Vars(Ds,DVs),
	!.
rnlDE_Vars(Ds,_) :-
	dlgIF::feil("rnlDE_Vars",Ds,[],nill),!,fail.	
% ----------------------------------------------------------------------------------------------
% rnlGetProxy
% ----------------------------------------------------------------------------------------------
% returns a DE's proxy (fails if none)
%
rnlGetProxy(N,DE,[D|Ds],Proxy) :-
	sx::sxUnpack(D,'|',L),
	L=[N,_,_,DE,Proxy],
	!
;	
	rnlGetProxy(N,DE,Ds,Proxy),
	!.
% ----------------------------------------------------------------------------------------------
% rnlGetDE_ByProxy
% ----------------------------------------------------------------------------------------------
% returns a proxy's DE (fails if none)
% <argument_no> | <entity_NL> | <DE_NL> | <DE_KB> [ | <proxy_var> ]
% 
rnlGetDE_ByProxy(Proxy,[D|Ds],DE_Var) :-
	sx::sxUnpack(D,'|',L),
	L=[N,A,_,Dk,Proxy],
	dbBase::kbID(A,Ak,_,_),
	Ni=sx::toInt(N),
	vars::varCompsR(DE_Var,Ak,Ni,Dk),
	!
;	
	rnlGetDE_ByProxy(Proxy,Ds,DE_Var),
	!.
% ----------------------------------------------------------------------------------------------
% rnlMarkProxy
% ----------------------------------------------------------------------------------------------
% marks all proxy variables in the DE-list as ineligible proxies (with a leading '*').
%
rnlMarkProxy([D|DEs],Proxy,[Dm|DEs]) :-
	sx::sxUnpack(D,'|',L),
	L=[N,E,DE,Dk,Proxy],
	ProxyM=string::concat("*",Proxy),
	sx::sxlJoin([N,E,DE,Dk,ProxyM],"|",Dm),
	!.
rnlMarkProxy([D|DEs],Proxy,[D|DEsM]) :-
	rnlMarkProxy(DEs,Proxy,DEsM),
	!.
% ----------------------------------------------------------------------------------------------
% rnlNonProxy
% ----------------------------------------------------------------------------------------------
class predicates
rnlNonProxies : (sterml,string_list,string_list) determ (i,i,o).
rnlRemProx : (string,string_list,string_list) determ (i,i,o).
clauses
rnlNonProxy(cmp("&",L),P,Pm) :-
	rnlNonProxies(L,P,Pm),
	!.
rnlNonProxy(cmp(Opr,[fvar(V,_),_]),P,Pm) :-
	% variables occurring on 'left sides' are not proxies
	opr::opr_any(Opr,_),
	rnlRemProx(V,P,Pm),
	!.
rnlNonProxy(_,P,P) :-
	!.
	
rnlNonProxies([],P,P) :-
	!.
rnlNonProxies([H|T],P,Pm) :-
	rnlNonProxy(H,P,P2),
	rnlNonProxies(T,P2,Pm),
	!.	
	
rnlRemProx(V,[P|Ps],[Pm|Ps]) :-
	sx::sxUnpack(P,'|',L),
	L=[A,B,C,D,Pr],
	V=Pr,
	sx::sxlJoin([A,B,C,D],"|",Pm),
	!.
rnlRemProx(V,[P|Ps],[P|Pms]) :-
	rnlRemProx(V,Ps,Pms),
	!.
% ----------------------------------------------------------------------------------------------
% rnlProxy
% ----------------------------------------------------------------------------------------------
class predicates
rnlProx : (string,sterml,string_list,string_list) determ (i,i,i,o).
rnlProx_Subst : (string,string_list,string,string,string_list) determ (i,i,i,i,o).
clauses
% a proxy variable is a variable that represents a DE. it must be assigned to a DE in its first occurrence in the rule.
% it can when referenced in 'right sides' further down in the rule be replaced with its DE.
%
% The purpose of this predicate is to mark DEs with their corresponding potential proxy variables in the DE-list.
% parameters:
% 1	in	next SC number
% 2	out	do.
% 3	rule
% 4	in	DE list
% 5	out	do.
%
% DEs with proxies are changed from E|No|D|Dk to E|No|D|Dk|Proxy.
% 
rnlProxy(N,M,cmp("&",[C2,C3]),DEs,DEsP) :-
	rnlProxy(N,N2,C2,DEs,DEs2),
	rnlProxy(N2,M,C3,DEs2,DEsP),
	!.
rnlProxy(N,M,T,DEs,DEsP) :-
	T=cmp("FALSE",[SC]),
	rnlProxy(N,M,SC,DEs,DEsP),
	!
;
	T=cmp(_,[_,cmp("#S",[SC])]),
	rnlProxy(N,M,SC,DEs,DEsP),
	!.
rnlProxy(N,M,ST,DEs,DEsP) :-
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",[_,_,_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",[_,_])
	          |Eac]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",[_,_])
	          |Ebc])
	      |Pc]),
	PNo=N*10,
	Ps=toString(PNo),	               
	rnlProx(Ps,Pc,DEs,DEs2),
	EaNo=N*10+1,
	Eas=toString(EaNo),	               
	rnlProx(Eas,Eac,DEs2,DEs3),
	EbNo=N*10+2,
	Ebs=toString(EbNo),	               
	rnlProx(Ebs,Ebc,DEs3,DEsP),
	M=N+1,
	!
;
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",[_,_,_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",[_,_])
	          |Eac])
	      |Pc]),
	PNo=N*10,
	Ps=toString(PNo),	               
	rnlProx(Ps,Pc,DEs,DEs2),
	EaNo=N*10+1,
	Eas=toString(EaNo),	               
	rnlProx(Eas,Eac,DEs2,DEsP),
	M=N+1,
	!
;
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",[_,_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",[_,_])
	          |Eac]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",[_,_])
	          |Ebc])]),
	EaNo=N*10+1,
	Eas=toString(EaNo),	               
	rnlProx(Eas,Eac,DEs,DEs3),
	EbNo=N*10+2,
	Ebs=toString(EbNo),	               
	rnlProx(Ebs,Ebc,DEs3,DEsP),
	M=N+1,
	!
;
	ST=cmp("REL_COND",
	      [cmp("REL_TYPE",[_,_]),
	       cmp("ENT_ARG",
	          [cmp("ENT_TYPE",[_,_])
	          |Eac])]),
	EaNo=N*10+1,
	Eas=toString(EaNo),	               
	rnlProx(Eas,Eac,DEs,DEsP),
	M=N+1,
	!.
rnlProxy(N,N,_,DEs,DEs) :-
	!.
rnlProx(_,[],DEs,DEs) :-
	!.
rnlProx(N,[cmp("ATTR_CONDS",Conds)],DEs,DEsP) :-	
	rnlProx(N,Conds,DEs,DEsP),
	!.
rnlProx(N,[cmp(Opr,[fvar(DE,_),fvar(Proxy,_)])|Conds],DEs,DEsP) :-
	% Proxy is a proxy variable if unreferenced
	opr::oprEq(Opr),
	rnlProx_Subst(N,DEs,DE,Proxy,DEs2),
	rnlProx(N,Conds,DEs2,DEsP),
	!.
rnlProx(N,[_|Conds],DEs,DEsP) :-
	rnlProx(N,Conds,DEs,DEsP),
	!.
rnlProx_Subst(N,[],DE,Proxy,_) :-
	dlgIF::feil("rnlProx_Subst",["DE not found",DE,Proxy,N],[],nill),!,fail.
rnlProx_Subst(_,[DE|_],_,Proxy,_) :-
	sx::sxUnpack(DE,'|',L),
	L=[_,_,_,_,P],
	P=Proxy,
	% already defined as proxy higher up in the tree
	!,fail.
rnlProx_Subst(N,[DE|DEs],Dk,Proxy,[DEp|DEs]) :-
	sx::sxUnpack(DE,'|',L),
	L=[N,E,Dn,Dk],
	sx::sxlJoin([N,E,Dn,Dk,Proxy],"|",DEp),
	!.
rnlProx_Subst(N,[DE|DEs],Dk,Proxy,[DE|DEsP]) :-
	rnlProx_Subst(N,DEs,Dk,Proxy,DEsP),
	!.
% ----------------------------------------------------------------------------------------------
% rnlRSide
% ----------------------------------------------------------------------------------------------
class predicates
rnlParams : (string_list,sterml,sterml) determ (i,i,o).
rnlRSides : (string_list,sterml,sterml) determ (i,i,o).
clauses
rnlRSide(N,N,_,atom(A),atom(A),"p3_dif") :-
	!.
rnlRSide(N,N,_,fstr(A,"p3_dli",_,_),str(A),"p3_dli") :-
	!.
rnlRSide(N,N,_,fstr(A,"p3_dif",_,_),atom(A),"p3_dif") :-
	!.
rnlRSide(N,N,_,fstr(A,"p3_dsy",_,_),str(A),"p3_dsy") :-
	!.
rnlRSide(N,N,_,fint(I,_),int(I),"p3_din") :-
	!.
rnlRSide(N,N,_,ffix(F,_),fix(F),"p3_dnu") :-
	!.
rnlRSide(N,N,Ds,fvar(Proxy,_),var(V),"p3_dany") :-
	rnlGetDE_ByProxy(Proxy,Ds,V),
	!
;
	V=Proxy,
	!.	
rnlRSide(N,M,Ds,cmp("#S",[SentCond]),cmp("p3_iSSS",[SentCondR]),"p3_din") :-
	rnlSentCond(N,M,Ds,SentCond,SentCondR),
	!.
rnlRSide(N,N,Ds,cmp(FT,[atom(F),_|Pars]),cmp(F,ParsM),Dom) :-
	set::member(FT,["DFNC","NFNC"]),
	dbBase::smDef(F,Clas,_,_,Dom,_,_),
	set::member(Clas,["f","g"]),
	rnlParams(Ds,Pars,ParsM),
	!.
rnlRSide(N,N,Ds,cmp(Opr,Ops),cmp(Opr,OpsM),"p3_dany") :-
	rnlRSides(Ds,Ops,OpsM),
	!.		
rnlRSides(_,[],[]) :-
	!.
rnlRSides(Ds,[R|Rs],[M|Ms]) :-
	rnlRSide(1,_,Ds,R,M,_),
	rnlRSides(Ds,Rs,Ms),
	!.
rnlParams(_,[],[]) :-
	!.
rnlParams(Ds,[P|Pars],[PM|ParsM]) :-
	P=cmp(_,[Par]),
	rnlRSide(1,_,Ds,Par,PM,_),
	rnlParams(Ds,Pars,ParsM),
	!.
% ----------------------------------------------------------------------------------------------
% rnlRemoveDummyConds
% ----------------------------------------------------------------------------------------------
/*
Replaces dummy DE assignments on the format: 
               lot[11]
                    lot number
                         = i
                              lot number[11] {var}
into:
               lot[11]
                    lot number {var}
*/      
class predicates
rnlRDCx : (sterml,sterml) determ (i,o).
rnlRDC_Rem : (string,string,sterml,sterml) determ (i,i,i,o).
clauses
rnlRemoveDummyConds(Tc,Tr) :-
	Tc=cmp(E,DEs),
	sx::sxUnpack(E,'[',L),
	L=[En,_],
	sx::sxGetInBrack(E,'[',']',SNo),
	rnlRDC_Rem(En,SNo,DEs,DEx),
	Tr=cmp(E,DEx),
	!.
rnlRemoveDummyConds(cmp(H,B),cmp(H,Br)) :-
	rnlRDCx(B,Br),
	!.
rnlRemoveDummyConds(C,C) :-
	!.
rnlRDC_Rem(_,_,[],[]) :-
	!.
rnlRDC_Rem(En,SNo,[DE|DEs],[var(D)|DEsx]) :-
	DE=cmp(D,[cmp(Opr,[var(V)])]),
	sx::frontchar(Opr,'=',_),
	sx::concat([D,"[",SNo,"]"],W),
	V=W,
	rnlRDC_Rem(En,SNo,DEs,DEsx),
	!.
rnlRDC_Rem(En,SNo,[DE|DEs],[DE|DEsx]) :-
	rnlRDC_Rem(En,SNo,DEs,DEsx),
	!.
rnlRDCx([],[]) :-
	!.
rnlRDCx([H|T],[Hx|Tx]) :-
	rnlRemoveDummyConds(H,Hx),
	rnlRDCx(T,Tx),
	!.		
% ----------------------------------------------------------------------------------------------
% rnlRT
% ----------------------------------------------------------------------------------------------
% generates the rule tree from the syntax tree
%
rnlRT(N,M,DEs,Vars,cmp("&",[C2,C3]),cmp("&",[C2r,C3r]),VarsM) :-
	rnlRT(N,N2,DEs,Vars,C2,C2r,Vars2),
	rnlRT(N2,M,DEs,Vars2,C3,C3r,VarsM),
	!.
rnlRT(N,N,_,Vars,atom("true"),atom("true"),Vars) :-
	!.
rnlRT(N,M,DEs,Vars,cmp("FALSE",[Syn]),cmp("not",[Rule]),Vars) :-
	rnlSentCond(N,M,DEs,Syn,Rule),
	!.
rnlRT(N,M,DEs,Vars,cmp("FALSE",[Syn]),cmp("not",[Rule]),VarsM) :-
	rnlVarCond(N,M,DEs,Vars,Syn,Rule,VarsM),
	!.
rnlRT(N,M,DEs,Vars,Syn,Rule,Vars) :-
	rnlSentCond(N,M,DEs,Syn,Rule),
	!.
rnlRT(N,M,DEs,Vars,Syn,Rule,VarsM) :-
	rnlVarCond(N,M,DEs,Vars,Syn,Rule,VarsM),
	!.
% ----------------------------------------------------------------------------------------------
% rnlSentCond
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	argument no
%
rnlSentCond(N,M,DEs,Syn,Rule) :-
/* target rule format:
     pLivesOn|eNational|eLot
          eNational[11]
               ENational_11dNationality
                    =
                         iDanish {atom}
          eLot[12]
               ELot_12dLotNumber {var}             
or
     pLivesOn|eNational|eLot
          eNational[11]
               ENational_11dNationality
                    =
                         iDanish {atom}
          X {Var}
*/
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	            cmp("CARD",[int(2)]),
	            cmp("ATTRS",PDEs)]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Ea)]),
	                cmp("ATTRS",EaDEs)])
	           |Eac]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Eb)]),
	                cmp("ATTRS",EbDEs)])
	           |Ebc])
	       |Pc]),
	set::member(R,["REL_COND","REL_FACT"]),
	PNo=N*10,	               
	rnlArg(DEs,PNo,P,PDEs,Pc,Pr),
	EaNo=N*10+1,
	rnlArg(DEs,EaNo,Ea,EaDEs,Eac,Ear),
	EbNo=N*10+2,
	rnlArg(DEs,EbNo,Eb,EbDEs,Ebc,Ebr),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[Ear,Pr,Ebr]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	            cmp("CARD",[int(1)]),
	            cmp("ATTRS",PDEs)]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Ea)]),
	                cmp("ATTRS",EaDEs)])
	           |Eac])
	       |Pc]),
	set::member(R,["REL_COND","REL_FACT"]),
	PNo=N*10,	               
	rnlArg(DEs,PNo,P,PDEs,Pc,Pr),
	EaNo=N*10+1,
	rnlArg(DEs,EaNo,Ea,EaDEs,Eac,Ear),
	sx::sxlJoin([P,Ea],"|",STk),
	Rule=cmp(STk,[Ear,Pr]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(2)])]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Ea)]),
	                cmp("ATTRS",EaDEs)])
	           |Eac]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Eb)]),
	                cmp("ATTRS",EbDEs)])
	           |Ebc])]),
	set::member(R,["REL_COND","REL_FACT"]),
	EaNo=N*10+1,
	rnlArg(DEs,EaNo,Ea,EaDEs,Eac,Ear),
	EbNo=N*10+2,
	rnlArg(DEs,EbNo,Eb,EbDEs,Ebc,Ebr),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[Ear,Ebr]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(1)])]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Ea)]),
	                cmp("ATTRS",EaDEs)])
	           |Eac])]),
	set::member(R,["REL_COND","REL_FACT"]),
	EaNo=N*10+1,
	rnlArg(DEs,EaNo,Ea,EaDEs,Eac,Ear),
	sx::sxlJoin([P,Ea],"|",STk),
	Rule=cmp(STk,[Ear]),	  
	M=N+1,
	!
; % varieties with argument variables:	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(2)]),
	            cmp("ATTRS",PDEs)]),
	        var(Ea),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Eb)]),
	                cmp("ATTRS",EbDEs)])
	           |Ebc])
	       |Pc]),
	set::member(R,["REL_COND","REL_FACT"]),
	PNo=N*10,	               
	rnlArg(DEs,PNo,P,PDEs,Pc,Pr),
	EbNo=N*10+2,
	rnlArg(DEs,EbNo,Eb,EbDEs,Ebc,Ebr),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[var(Ea),Pr,Ebr]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(2)]),
	            cmp("ATTRS",PDEs)]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Ea)]),
	                cmp("ATTRS",EaDEs)])
	           |Eac]),
	        var(Eb)
	       |Pc]),
	set::member(R,["REL_COND","REL_FACT"]),
	PNo=N*10,	               
	rnlArg(DEs,PNo,P,PDEs,Pc,Pr),
	EaNo=N*10+1,
	rnlArg(DEs,EaNo,Ea,EaDEs,Eac,Ear),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[Ear,Pr,var(Eb)]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(2)]),
	            cmp("ATTRS",PDEs)]),
	        var(Ea),
	        var(Eb)
	       |Pc]),
	set::member(R,["REL_COND","REL_FACT"]),
	PNo=N*10,	               
	rnlArg(DEs,PNo,P,PDEs,Pc,Pr),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[var(Ea),Pr,var(Eb)]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(1)]),
	            cmp("ATTRS",PDEs)]),
	        var(Ea)
	       |Pc]),
	set::member(R,["REL_COND","REL_FACT"]),
	PNo=N*10,	               
	rnlArg(DEs,PNo,P,PDEs,Pc,Pr),
	sx::sxlJoin([P,Ea],"|",STk),
	Rule=cmp(STk,[var(Ea),Pr]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(2)])]),
	        var(Ea),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Eb)]),
	                cmp("ATTRS",EbDEs)])
	           |Ebc])]),
	set::member(R,["REL_COND","REL_FACT"]),
	EbNo=N*10+2,
	rnlArg(DEs,EbNo,Eb,EbDEs,Ebc,Ebr),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[var(Ea),Ebr]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(2)])]),
	        cmp("ENT_ARG",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(Ea)]),
	                cmp("ATTRS",EaDEs)])
	           |Eac]),
	         var(Eb)]),
	set::member(R,["REL_COND","REL_FACT"]),
	EaNo=N*10+1,
	rnlArg(DEs,EaNo,Ea,EaDEs,Eac,Ear),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[Ear,var(Eb)]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(2)])]),
	         var(Ea),
	         var(Eb)]),
	set::member(R,["REL_COND","REL_FACT"]),
	sx::sxlJoin([P,Ea,Eb],"|",STk),
	Rule=cmp(STk,[var(Ea),var(Eb)]),	  
	M=N+1,
	!
;	
	Syn=cmp(R,
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(P)]),
	           	cmp("CARD",[int(1)])]),
	        var(Ea)]),
	set::member(R,["REL_COND","REL_FACT"]),
	sx::sxlJoin([P,Ea],"|",STk),
	Rule=cmp(STk,[var(Ea)]),	  
	M=N+1,
	!.
% ----------------------------------------------------------------------------------------------
% rnlVarCond
% ----------------------------------------------------------------------------------------------
class predicates
rnlVarCondDom : (string_list,string,string,string,string_list) determ (i,i,i,o,o).
rnlVarExists : (string,string_list,string) determ (i,i,o).
clauses
rnlVarCond(N,M,DEs,Vars,cmp(Opr,[fvar(Var,_),RSide]),cmp(W,[cmp(Opr,[RSideM])]),VarsM) :-
	opr::oprEq(Opr),
	rnlRSide(N,M,DEs,RSide,RSideM,RDom),
	rnlVarCondDom(Vars,Var,RDom,W,VarsM),
	!
;		
	opr::oprEq(Opr),
	rnlRSide(N,M,DEs,RSide,RSideM,RDom),
	rnlVarCondDom(Vars,Var,RDom,W,VarsM),
	!.		
rnlVarCond(N,M,DEs,Vars,cmp(Opr,[fvar(Var,_),RSide]),cmp(Var,[cmp(Opr,[RSideM])]),Vars) :-
	rnlRSide(N,M,DEs,RSide,RSideM,_),
	!.
rnlVarCondDom(Vars,Var,RDom,Var,Vars) :-
	rnlVarExists(Var,Vars,Dom),
	!,dom::domCompat(Dom,RDom),
	!.
rnlVarCondDom(Vars,Var,RDom,W,[W|Vars]) :-
	dom::domShortLong(SDom,RDom),
	sx::concat([Var," /",SDom],W),
	!.
rnlVarExists(Var,[V|Vs],Dom) :-
	sx::sxUnpackS(V," /",L),
	L=[W,SDom],
	Var=W,
	dom::domShortLong(SDom,Dom),
	!
;
	rnlVarExists(Var,Vs,Dom),
	!.	
end implement rn
