/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement rt
open core,termDomain

constants
className = "com/rudstrand/rt/rt".
classVersion = "1.0".

class predicates
rcnCondition : (sterm,sterml,sterm,string_list,string_list) determ (i,i,o,i,o).
rcnExpression : (sterm,sterm) determ (i,o).
rcnFixRemove : (string_list,string_list,sterm,sterm,string_list,sterm,sterm) determ (i,i,i,i,o,o,o).
rcnGenComps : (sterml,sterm,sterml) determ (i,i,o).
rcnList : (sterm,sterml) determ (i,o).
rcnLiteral : (sterm,sterm) determ (i,o).
rcnMakePars : (string,string,string_list,sterml,sterml) determ (i,i,i,i,o).
rcnMess : (string) procedure (i).
rcnComplexCondition : (sterm,sterml,sterm,string_list,string_list) determ (i,i,o,i,o).
rcnGetDE_Conditions : (sterm,sterml) procedure (i,o).
rcnSimpleLiteral : (sterm,sterm) determ (i,o).
rcnST_Variable : (sterm,sterml,sterm) determ (i,i,o).
rcnVariableCondition : (sterm,sterm,string_list,string_list) determ (i,o,i,o).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% rcnTree
% ----------------------------------------------------------------------------------------------
%
% translates a sentence rule from IF-format to RT/internal-format
% parameters:
% 1	in	record type (or 0 if new rule)
% 2	in	term to be translated to rule-tree
% 3	out	rule-tree
% 4	out	rule-type
% 5	out	DEs for the tree DE listbox
%
rcnTree(RT,Rule,Tree,SType,DEs) :-
	dbBase::recordType(RT,ART,_,_),
	set::member(ART,[3,5]),
	Rule=cmp(":-",[Conc,Cond]),
	Cond=cmp(Proc,[str(Q),_,str(A)|ImpDEs]),
	rub::rubRuleType(_,TypeNo,Proc),
	SType=toString(TypeNo),
	!,
	Query=termDomain::fromTermStringIndirect(Q),
	Ass=termDomain::fromTermStringIndirect(A),
	rcnGenComps(ImpDEs,Ass,DE_Ass),
	rcnGetDE_Conditions(Query,QueryAss),
	rcnComplexCondition(Query,QueryAss,Cond2,[],_),
	rcnST_Variable(Conc,DE_Ass,Conc2),
	Rule2=cmp(":-",[Conc2,Cond2]),
				tt::tt_term(3,"Rule-tree (pre-clean)",Rule2),
	rcnCleanRule(1,Proc,Rule2,Rule3,DEs),
	struc::treAndLinearize(Rule3,Tree),
				tt::tt_term(2,"Rule-tree",Tree),
	!
;
	dbBase::recordType(RT,ART,_,_),
	set::member(ART,[3,5]),
	SType="1",
	Rule=cmp(":-",[Conc,Cond]), % implication rule
				tt::tt_term(2,"Rule (KB)",Rule),
	rcnGetDE_Conditions(Cond,DE_Ass),
	rcnComplexCondition(Cond,DE_Ass,Cond2,[],_),
	rcnST_Variable(Conc,DE_Ass,Conc2),
	Rule2=cmp(":-",[Conc2,Cond2]),
				tt::tt_term(3,"Rule-tree (pre-clean)",Rule2),
	rcnCleanRule(1,"p3_iRimp",Rule2,Rule3,DEs),
	struc::treAndLinearize(Rule3,Tree),
				tt::tt_term(2,"Rule-tree",Tree),
	!.
% ----------------------------------------------------------------------------------------------
% rcnCleanRule
% ----------------------------------------------------------------------------------------------
rcnCleanRule(RemNILs,Proc,Rule,RuleM,DEs) :-
	RemNILs=1,
	Rule=cmp(_,[Conc2,Cond2]),
	rcnFixRule(-1,Cond2,Cond3,RemovedDEs2,CondDEs),
	rcnFixRule(1,Conc2,Conc3,RemovedDEs3,ConcDEs),
	set::conc([RemovedDEs2,RemovedDEs3],RemovedDEs),
	set::conc([CondDEs,ConcDEs],DEs2),
	rcnFixRemove(RemovedDEs,DEs2,Cond3,Conc3,DEs,Cond5,Conc5),
	struc::treElimDummyLogOpr(Cond5,Condition),
	struc::treElimDummyLogOpr(Conc5,Conclusion),
	rub::rubRuleType(_,TypNo,Proc),	         
	rub::rubRuleType(Typ,TypNo,_),
	RuleM=cmp(Typ,[Condition,Conclusion]),
				tt::tt_term(2,"Rule-tree (unlinearized)",RuleM),
	!
;
	RemNILs=-1,
	Rule=cmp(_,[Conc2,Cond2]),
	rcnFixRule(-1,Cond2,Cond3,RemovedDEs2,CondDEs),
	rcnFixRule(1,Conc2,Conc3,RemovedDEs3,ConcDEs),
	set::conc([RemovedDEs2,RemovedDEs3],RemovedDEs),
	set::conc([CondDEs,ConcDEs],DEs2),
	rcnFixRemove(RemovedDEs,DEs2,Cond3,Conc3,DEs,Cond5,Conc5),
	struc::treElimDummyLogOpr(Cond5,Condition),
	struc::treElimDummyLogOpr(Conc5,Conclusion),
	rub::rubRuleType(_,TypNo,Proc),	         
	rub::rubRuleType(Typ,TypNo,_),
	RuleM=cmp(Typ,[Condition,Conclusion]),
				tt::tt_term(2,"Rule-tree (unlinearized)",RuleM),
	!.
class predicates
rcnArgumentVariable : (string,sterm,sterml,sterm) determ (i,i,i,o).
rcnD_Variables : (sterml,sterml,sterml) determ (i,i,o).
rcnD_Variable : (sterm,sterml,sterm) determ (i,i,o).	
rcnDE_Comps : (sterm,sterml,sterml) determ (i,i,o).
rcnFixExp : (sterml,sterml) determ (i,o).
clauses
rcnArgumentVariable(PT,Arg,Ass,ArgM) :-
	% OpenArgumentVariable
	Arg=var(V),
	not(vars::varComps(V,_,_,_)), % to avoid mixing up an argument variable with a predicate's DE variable
	ArgM=Arg,
	!
;
	% EntityArgument
	Arg=cmp(ET,D_Vars),
	rcnGetArgNo(Arg,No),
	SNo=toString(No),
	rcnD_Variables(D_Vars,Ass,D_VarsM),
	sx::concat([ET,"[",SNo,"]"],ETNO),
	ArgM=cmp(ETNO,D_VarsM),
	!
;
	dlgIF::feil("rcnArgumentVariable",["Error in sentence rule record format",PT],[],Arg),!,fail.
rcnD_Variables([],_,[]) :-
	!.
rcnD_Variables([V|Vs],Ass,[Vm|Vms]) :-
	rcnD_Variable(V,Ass,Vm),
	rcnD_Variables(Vs,Ass,Vms),
	!.
rcnD_Variable(V,Ass,Vm) :-
	V=var(_),
	rcnDE_Comps(V,Ass,L),
	set::size(L,N),
	N=0,
	Vm=V,
	!
;
	V=var(DE),
	rcnDE_Comps(V,Ass,L),
	L=[A],
	A=cmp(Opr,[_,Exp]),
	opr::oprIntExt(Opr,Op),
	Vm=cmp(DE,[cmp(Op,[Exp])]),
	!
;
	V=var(DE),
	rcnDE_Comps(V,Ass,L),
	set::size(L,N),
	N>1,
	rcnFixExp(L,Lm),
	Vm=cmp(DE,[cmp("&",Lm)]),
	!
;
	dlgIF::feil("rcnD_Variable",["Error in sentence rule record format"],[],V),!,fail.
rcnDE_Comps(_,[],[]) :-
	!.
rcnDE_Comps(DE,[A|As],[cmp(Op,[DE,ExpM])|Vm]) :-
	A=cmp(Opr,[DE,Exp]),
	rcnExpression(Exp,ExpM),
	opr::oprIntExt(Opr,Op),
	rcnDE_Comps(DE,As,Vm),
	!.
rcnDE_Comps(DE,[_|As],Vm) :-
	rcnDE_Comps(DE,As,Vm),
	!.
rcnFixExp([],[]) :-
	!.
rcnFixExp([E|Es],[M|Ms]) :-
	E=cmp(Opr,[_,Ex]),
	opr::oprIntExt(Opr,Op),
	M=cmp(Op,[Ex]),
	rcnFixExp(Es,Ms),
	!.
% ----------------------------------------------------------------------------------------------
% rcnCondition
% ----------------------------------------------------------------------------------------------
rcnCondition(Cond,Ass,CondM,Vars,VarsM) :-
	% SentenceCountCondition
	Cond=cmp("p3_iSSS",[var(V),str(S)]),
    T=termDomain::fromTermStringIndirect(S),
	rcnGetDE_Conditions(T,DE_Ass),
	rcnComplexCondition(T,DE_Ass,T2,Vars,Vars2),
	W=string::concat(V," /IN"),
	CondM=cmp(W,[cmp("=",[cmp("p3_iSSS",[T2])])]), 
	VarsM=[W|Vars2],
	!
;
	Cond=cmp("not",[NT_Cond]),
	rcnGetDE_Conditions(NT_Cond,DE_Ass),
	rcnComplexCondition(NT_Cond,DE_Ass,T2,Vars,VarsM),
	CondM=cmp("not",[T2]),
	!
;
	rcnST_Variable(Cond,Ass,CondM2),
	VarsM=Vars,
	CondM=CondM2,
	!
;
	rcnVariableCondition(Cond,CondM,Vars,VarsM),
	!.
% ----------------------------------------------------------------------------------------------
% rcnExpression
% ----------------------------------------------------------------------------------------------
rcnExpression(Exp,ExpM) :-
	Exp=cmp(Opr,[E2,E3]),
	opr::opr_any(Opr,_),
	rcnExpression(E2,E2M),
	rcnExpression(E3,E3M),
	ExpM=cmp(Opr,[E2M,E3M]),
	!
;
	Exp=cmp(Opr,[E]),
	opr::opr_any(Opr,_),
	rcnExpression(E,EM),
	ExpM=cmp(Opr,[EM]),
	!
;
	rcnSimpleExpression(Exp,ExpM),
	!.
% ----------------------------------------------------------------------------------------------
% rcnFixRule
% ----------------------------------------------------------------------------------------------
class predicates
rcnFix : (integer,string,integer,sterml,sterml,sterml,string_list,string_list) determ (i,i,i,i,i,o,o,o).
rcnFixNewDE : (integer,string,sterm) determ (i,i,o).
rcnFixReorder : (string,sterml,sterml) determ (i,i,o).
rcnFixRules : (integer,sterml,sterml,string_list,string_list) determ (i,i,o,o,o).
rcnFixNewMDEs : (sterml,sterml) determ (i,o).
clauses
% validates and if possible repairs a rule record retrieved from the KB.
% - adds DE components for new DEs
% - reorders DE components as required
% - removes invalid (unknown) DE components
% - adjusts domain-specific assignments for DEs
%
% the entire operation is in IF 
% parameters:
% 1	in	-1/1 par#2 is rule conclusion
% 2	in	part of rule
% 3	out	par#3 repaired
% 4	out	removed DEs
% 5	out	DEs in rule component (not including rule's in NOT/true-branches)
% 
rcnFixRule(Conclusion,T,Tm,Removed,AllDEs) :-
	T=cmp(H,_),
	sx::frontstr(4,H,F,_),
	set::member(F,["[F] ","[P] ","[C] "]),
	Tm=T,
	Removed=[],
	AllDEs=[],
	!
;
	T=cmp(Arg,DCs),
	sx::sxUnpack(Arg,'[',L2),
	L2=[Ak,_],
	sx::sxGetInBrack(Arg,'[',']',SNo),
	No=sx::toInt(SNo),
	dbBase::nlTry(Ak,A),
	dbBase::smClassDefF(Ak,"e",_,_,_,_,_),
	dbBase::kbEntityDEs(Ak,_,DEs),
	rcnFix(Conclusion,A,No,DEs,DCs,DCm,Removed,AllDEs),
	Tm=cmp(Arg,DCm),
	!
;
	T=cmp(Arg,DCs),
	sx::sxUnpack(Arg,'[',L2),
	L2=[Ak,_],
	dbBase::smClassDefF(Ak,"p",_,_,_,_,_),
	sx::sxGetInBrack(Arg,'[',']',SNo),
	No=sx::toInt(SNo),
	dbBase::nlTry(Ak,A),
	dbBase::kbPredDEs(Ak,_,DEs),
	rcnFix(Conclusion,A,No,DEs,DCs,DCm,Removed,AllDEs),
	Tm=cmp(Arg,DCm),
	!
;
	T=cmp(H,B),
	rcnFixRules(Conclusion,B,Bm,Removed,AllDEs),
	Tm=cmp(H,Bm),
	!
;
	Tm=T,
	Removed=[],
	AllDEs=[],
	!.
rcnFixRules(_,[],[],[],[]) :-
	!.
rcnFixRules(Conclusion,[H|T],[Hm|Tm],Removed,DEs) :-
	rcnFixRule(Conclusion,H,Hm,Rem2,DEs2),
	rcnFixRules(Conclusion,T,Tm,Rem3,DEs3),
	set::conc([Rem2,Rem3],Removed),
	set::conc([DEs2,DEs3],DEs),
	!.
% parameters:
% 1	in	1/-1 rule conclusion
% 2	in	argument (NL)
% 3	in	argument no
% 4	in	DEs (from def.)
% 5	in	DE conditions from rule
% 6	out	updated par#6
% 7	out	removed DEs (NL)
% 8	out	added DEs (NL)
%
rcnFix(_,_,_,[],[],[],[],[]) :-
	!.
rcnFix(Conclusion,A,ArgNo,[],[DC|DCs],[],[DV|Rs],[]) :-
	DC=cmp(Vk,_),
	dbBase::nlTry(Vk,V),
	rcnFix(Conclusion,A,ArgNo,[],DCs,_,Rs,_),
	sx::concat(["Data element '",V,"' has been removed from argument '",A,"'"],Mess),
	dbBase::kbID(A,Ak,_,_),
	vars:: varCompsR(DV,Ak,ArgNo,Vk),
	rcnMess(Mess),
	!
;
	DC=var(Vk),
	rcnFix(Conclusion,A,ArgNo,[],DCs,_,Rs,_),
	dbBase::nlTry(Vk,V),
	sx::concat(["Data element '",V,"' has been removed from argument '",A,"'"],Mess),
	dbBase::kbID(A,Ak,_,_),
	vars:: varCompsR(DV,Ak,ArgNo,Vk),
	rcnMess(Mess),
	!.
rcnFix(Conclusion,A,ArgNo,[fstr(DE,Dom,F,W)|DEs],DCs,NewDCs,Removed,New) :-
	DCs=[cmp(DV,_)|_],
	vars:: varComps(DV,_,_,Vk),
	DE<>Vk,
	rcnFixReorder(DE,DCs,DCm),
	rcnFix(Conclusion,A,ArgNo,[fstr(DE,Dom,F,W)|DEs],DCm,NewDCs,Removed,New),
	dbBase::nlTry(Vk,V),
	sx::concat(["Data element '",V,"' has changed position in argument '",A,"'"],Mess),
	rcnMess(Mess),
	!	
;	
	DCs=[var(DV)|_],
	vars:: varComps(DV,_,_,Vk),
	DE<>Vk,
	rcnFixReorder(DE,DCs,DCm),
	rcnFix(Conclusion,A,ArgNo,[fstr(DE,Dom,F,W)|DEs],DCm,NewDCs,Removed,New),
	dbBase::nlTry(Vk,V),
	sx::concat(["Data element '",V,"' has changed position in argument '",A,"'"],Mess),
	rcnMess(Mess),
	!	
;
	DCs=[cmp(DV,_)|_],
	vars:: varComps(DV,_,_,Vk),
	DE<>Vk,
	% new DE 
	dbBase::nlTry(DE,Dn),
	sx::concat(["New data element '",Dn,"' in argument '",A,"'"],Mess),
	rcnMess(Mess),
	rcnFix(Conclusion,A,ArgNo,DEs,DCs,NewDCs2,Removed,Ns),
	dbBase::kbID(A,Ak,_,_),
	vars:: varCompsR(DEV,Ak,ArgNo,DE),
	rcnFixNewDE(Conclusion,DEV,NewDE),
	NewDCs=[NewDE|NewDCs2],
	New=[DEV|Ns],
	!	
;
	DCs=[var(DV)|_],
	vars:: varComps(DV,_,_,Vk),
	DE<>Vk,
	% new DE
	dbBase::nlTry(DE,Dn),
	sx::concat(["New data element '",Dn,"' in argument '",A,"'"],Mess),
	rcnMess(Mess),
	rcnFix(Conclusion,A,ArgNo,DEs,DCs,NewDCs2,Removed,Ns),
	dbBase::kbID(A,Ak,_,_),
	vars:: varCompsR(DEV,Ak,ArgNo,DE),
	rcnFixNewDE(Conclusion,DEV,NewDE),
	NewDCs=[NewDE|NewDCs2],
	New=[DEV|Ns],
	!	
;
	DCs=[var(DV)|_],
	vars:: varComps(DV,_,_,Vk),
	DE<>Vk,
	% new DE
	dbBase::nlTry(DE,Dn),
	sx::concat(["New data element '",Dn,"' in argument '",A,"'"],Mess),
	rcnMess(Mess),
	rcnFix(Conclusion,A,ArgNo,DEs,DCs,NewDCs2,Removed,Ns),
	dbBase::kbID(A,Ak,_,_),
	vars:: varCompsR(DEV,Ak,ArgNo,DE),
	rcnFixNewDE(Conclusion,DEV,NewDE),
	NewDCs=[NewDE|NewDCs2],
	New=[DEV|Ns],
	!	
;
	DCs=[],
	% new DE
	dbBase::nlTry(DE,Dn),
	sx::concat(["New data element '",Dn,"' in argument '",A,"'"],Mess),
	rcnMess(Mess),
	dbBase::kbID(A,Ak,_,_),
	vars:: varCompsR(DEV,Ak,ArgNo,DE),
	rcnFixNewDE(Conclusion,DEV,NewDE),
	NewDCs=[NewDE],
	New=[DEV],
	Removed=[],
	!	
;
	DCs=[cmp(DV,[cmp(Opr,B)])|T],
	opr::oprEq(Opr),
	vars:: varComps(DV,_,_,Vk),
	DE=Vk,
	opr::oprIntExt(Opr,Op),
	opr::oprDom(Dk,E,Op),
	set::member(E,["=","=*"]),
	Dk<>Dom,
	% changed assignment domain
	dbBase::nlSing(DE,Dn),
	sx::concat(["Data element '",Dn,"' in argument '",A,"' has changed domain"],Mess),
	rcnMess(Mess),
	rcnFix(Conclusion,A,ArgNo,DEs,T,NewDCs2,Removed,Ns),
	NewDCs=[cmp(DV,[cmp(E,B)])|NewDCs2],
	New=[DV|Ns],
	!
;
	DCs=[DC|Ds],
	rcnFix(Conclusion,A,ArgNo,DEs,Ds,NewDs,Removed,Ns),
	dbBase::kbID(A,Ak,_,_),
	vars:: varCompsR(DV,Ak,ArgNo,DE),
	NewDCs=[DC|NewDs], % no change
	New=[DV|Ns],
	!.
% parameters:
% 1	in	DE
% 2	in	list of DE components
% 3	out	reordered par#2 such that component corresponding to DE is first in list
%
% fails if no component corresponding to DE
%
rcnFixReorder(DE,[D|Ds],DM) :-
	D=var(V),
	DE=V,
	DM=[D|Ds],
	!
;
	D=cmp(V,_),
	DE=V,
	DM=[D|Ds],
	!
;
	rcnFixReorder(DE,Ds,DM2),
	DM2=[A|B],
	DM=[A,D|B],
	!.
	
% adds a new DE component to an argument
% parameters:
% 1	in	1/-1 rule conclusion
% 2	in	new DE (NL)
% 3	in	do. domain
% 4	out	new DE rule component corr. to par#3
%	
rcnFixNewDE(Conclusion,DE,NewDE) :-
	Conclusion=1,
	NewDE=cmp(DE,[cmp("=",[nill])]),
	!
;
	Conclusion=-1,
	NewDE=var(DE),
	!.
rcnFixNewMDEs([],[]) :-
	!.
rcnFixNewMDEs([var(M)|Ms],[A|As]) :-
	A=cmp(M,[cmp("=",[nill])]),
	rcnFixNewMDEs(Ms,As),
	!.
% ----------------------------------------------------------------------------------------------
% rcnFixRemove
% ----------------------------------------------------------------------------------------------
class predicates
rcnFixRem : (sterm,string_list,sterm) procedure (i,i,o).
rcnFixRems : (sterml,string_list,sterml) determ (i,i,o).	
clauses
% parameters:
% 1	in	removed DEs (replace with nill in all expressions
% 2	in	all DEs in rule
% 3	in	condition
% 4	in	conclusion
% 5	out =par#2 (after rem. of complex DEs)
% 6	out	adjusted condition
% 7	out	adjusted conclusion
%
rcnFixRemove(RemovedDEs,DEs,Cond,Conc,DEs,CondM,ConcM) :-
	rcnFixRem(cmp("",[Cond,Conc]),RemovedDEs,Tm),
	Tm=cmp(_,[CondM,ConcM]),
	!.
	
rcnFixRems([],_,[]) :-
	!.
rcnFixRems([H|T],RDEs,[Hm|Tm]) :-
	rcnFixRem(H,RDEs,Hm),
	rcnFixRems(T,RDEs,Tm),
	!.

rcnFixRem(T,RDEs,Tm) :-
	T=cmp(H,L),
	rcnFixRems(L,RDEs,Lm),
	Tm=cmp(H,Lm),
	!
;
	Tm=T,
	!.
% ---------------------------------------------------------------------------------------------_
% rcnGenComps
% ----------------------------------------------------------------------------------------------
% generates the list of conclusion DE assignments for set rules 
% parameters:
% 1	in	list of conclusions DEs <DE/MDE_Variables>
% 2	in	list-structure of assignments
% 3	out	DE assignments 
%
rcnGenComps([],nill,[]) :-
	!.
rcnGenComps([DE|DEs],list(A,As),[C|Cs]) :-
	DE=var(V),
	vars:: varComps(V,_,_,_),
	C=cmp("=",[var(V),A]),
	rcnGenComps(DEs,As,Cs),
	!.
% ---------------------------------------------------------------------------------------------_
% rcnGetArgNo
% ----------------------------------------------------------------------------------------------
rcnGetArgNo(Arg,No) :-
	Arg=cmp(_,[var(V)|_]),
	vars:: varComps(V,_,No,_),
	!
;
	dlgIF::feil("rncGetArgNo",[],[],Arg),!,fail.
% ---------------------------------------------------------------------------------------------_
% rcnList
% ----------------------------------------------------------------------------------------------
rcnList(nill,[]) :-
	!.
rcnList(list(H,T),[Hm|Tm]) :-
	rcnSimpleExpression(H,Hm),
	rcnList(T,Tm),
	!.
%% ---------------------------------------------------------------------------------------------_
% rcnLiteral
% ----------------------------------------------------------------------------------------------
class predicates
rcnListLiteral : (sterm,sterml) determ (i,o).
clauses

rcnLiteral(L,Lm) :-
	rcnSimpleLiteral(L,Lm),
	!
;
	rcnListLiteral(L,Ls),
	Lm=cmp("[]",Ls),
	!.
	
rcnListLiteral(nill,[]) :-
	!.
rcnListLiteral(list(H,T),[Hm|Tm]) :-
	rcnSimpleLiteral(H,Hm),
	rcnListLiteral(T,Tm),
	!.
% ---------------------------------------------------------------------------------------------_
% rcnMakePars
% ----------------------------------------------------------------------------------------------
rcnMakePars(_,_,[],[],[]) :-
	!.
rcnMakePars(Typ,Func,DEs,[],FA) :-
	dbBase::nlSing(Func,Fn),
	sx::concat(["Formal parameter has been added in call to ",Typ," '",Fn,"'"],Mess),
	rcnMess(Mess),
	rcnMakePars(Typ,Func,DEs,[nill],FA),
	!.
rcnMakePars(Typ,Func,[],_,[]) :-
	dbBase::nlSing(Func,Fn),
	sx::concat(["Actual parameter(s) have been removed in call to ",Typ," '",Fn,"'"],Mess),
	rcnMess(Mess),
	!.
rcnMakePars(Typ,Func,[DE|DEs],[Par|Pars],[FA|FAS]) :-
	rcnExpression(Par,ParM),
	FA=cmp(DE,[cmp("=",[ParM])]),
	rcnMakePars(Typ,Func,DEs,Pars,FAS),
	!.	
% ---------------------------------------------------------------------------------------------_
% rcnMess
% ----------------------------------------------------------------------------------------------
rcnMess(S) :-
	globalVars::get_i("inter")=1,
	dlgIF::dlgNote(0,S),
	!
;
	dlgIF::msgS(S),
	!.
% ----------------------------------------------------------------------------------------------
% rcnComplexCondition
% ----------------------------------------------------------------------------------------------
rcnComplexCondition(Cond,Ass,CondM,Vars,VarsM) :-
	Cond=cmp(",",[C2,C3]),
	rcnComplexCondition(C2,Ass,C2M,Vars,Vars2),
	rcnComplexCondition(C3,Ass,C3M,Vars2,VarsM),
	CondM=cmp("&",[C2M,C3M]),
	!
;
	rcnCondition(Cond,Ass,CondM,Vars,VarsM),
	!.
% ----------------------------------------------------------------------------------------------
% rcnGetDE_Conditions
% ----------------------------------------------------------------------------------------------
rcnGetDE_Conditions(Cond,Ass) :-
	Cond=cmp(",",[C2,C3]),
	rcnGetDE_Conditions(C2,Ass2),
	rcnGetDE_Conditions(C3,Ass3),
	set::conc([Ass2,Ass3],Ass),	
	!.
rcnGetDE_Conditions(Cond,[Cond]) :-
	Cond=cmp(_,[Var,_]),
	Var=var(V),
	vars:: varComps(V,_,_,_),
	!.		
rcnGetDE_Conditions(_,[]) :-
	!.		
% ----------------------------------------------------------------------------------------------
% rcnSimpleExpression
% ----------------------------------------------------------------------------------------------
rcnSimpleExpression(E,Em) :-
	rcnList(E,L),
	Em=cmp("[]",L),
	!.
rcnSimpleExpression(nill,nill) :-
	!.
rcnSimpleExpression(Exp,ExpM) :-
	rcnLiteral(Exp,ExpM),
	!
;
	% Variable
	Exp=var(_),
	ExpM=Exp,
	!
;
	% CR call
	Exp=cmp("p3_iIR",[atom(F)]),
	FF=string::concat("[C] ",F),
	ExpM=atom(FF),
	!
;
	% CR call
	Exp=cmp("p3_iIR",[atom(Rule)|Pars]),
	dbBase::smClassDef(Rule,"h",_,DEs,_,_,_),
	rcnMakePars("certainty rule",Rule,DEs,Pars,ParsM),
	FF=string::concat("[C] ",Rule),
	ExpM=cmp(FF,ParsM),
	!
;
	% DeterministicFunctionCall - no parameters
	Exp=cmp(F,[atom("p3_iDtFn"),var(_)]),
	dbBase::smClassDefF(F,"f",_,DEs,_,_,_),
	DEs=[],
	FF=string::concat("[F] ",F),
	ExpM=atom(FF),
	!
;
	% NondeterministicFunctionCall - no parameters
	Exp=cmp(F,[var(_)]),
	dbBase::smClassDefF(F,"g",_,DEs,_,_,_),
	DEs=[],
	FF=string::concat("[F] ",F),
	ExpM=atom(FF),  
	!
;
	% DeterministicFunctionCall
	Exp=cmp(F,[atom("p3_iDtFn"),var(_)|Pars]),
	dbBase::smClassDefF(F,"f",_,DEs,_,_,_),
	rcnMakePars("function",F,DEs,Pars,ParsM),
	FF=string::concat("[F] ",F),
	ExpM=cmp(FF,ParsM),
	!
;
	% NondeterministicFunctionCall
	Exp=cmp(F,[var(_)|Pars]),
	dbBase::smClassDefF(F,"g",_,DEs,_,_,_),
	rcnMakePars("function",F,DEs,Pars,ParsM),
	FF=string::concat("[F] ",F),
	ExpM=cmp(FF,ParsM),
	!
;
	% ProcedureCall
	Exp=atom(F),
	dbBase::smClassDefF(F,"i",_,_,_,_,_),
	FF=string::concat("[P] ",F),
	ExpM=atom(FF),
	!
;
	% ProcedureCall
	Exp=cmp(F,Pars),
	dbBase::smClassDefF(F,"i",_,DEs,_,_,_),
	rcnMakePars("procedure",F,DEs,Pars,ParsM),
	FF=string::concat("[P] ",F),
	ExpM=cmp(FF,ParsM),
	!.
% ---------------------------------------------------------------------------------------------_
% rcnSimpleLiteral
% ----------------------------------------------------------------------------------------------
rcnSimpleLiteral(atom(V),atom(V)) :-
	!.
rcnSimpleLiteral(str(V),str(V)) :-
	!.
rcnSimpleLiteral(int(V),int(V)) :-
	!.
rcnSimpleLiteral(fix(V),fix(V)) :-
	!.
% ----------------------------------------------------------------------------------------------
% rcnST_Variable
% ----------------------------------------------------------------------------------------------
class predicates
rcnMakeST : (string,sterm,sterm,string) determ (i,i,i,o).	
clauses

rcnST_Variable(Cond,Ass,CondM) :-
	Cond=cmp(PT,[Arg|_]),
	dbBase::smClassDefF(PT,"p",Card,DEs,_,_,_),
	Card=1,
	set::size(DEs,N),
	N=0,
	rcnArgumentVariable(PT,Arg,Ass,ArgM),
	rcnMakeST(PT,Arg,nill,ST),
	CondM=cmp(ST,[ArgM]),
	!
;
	Cond=cmp(PT,[Arg|D_Variables]),
	dbBase::smClassDefF(PT,"p",Card,DEs,_,_,_),
	Card=1,
	set::size(DEs,N),
	N>0,
	rcnArgumentVariable(PT,Arg,Ass,ArgM),
	rcnD_Variables(D_Variables,Ass,D_VariablesM),
	rcnGetArgNo(cmp("",D_Variables),No),
	SNo=toString(No),
	sx::concat([PT,"[",SNo,"]"],PTNO),
	rcnMakeST(PT,Arg,nill,ST),
	CondM=cmp(ST,[ArgM,cmp(PTNO,D_VariablesM)]),
	!
;
	Cond=cmp(PT,[Arg2,Arg3|_]),
	dbBase::smClassDefF(PT,"p",Card,DEs,_,_,_),
	Card=2,
	set::size(DEs,N),
	N=0,
	rcnArgumentVariable(PT,Arg2,Ass,Arg2M),
	rcnArgumentVariable(PT,Arg3,Ass,Arg3M),
	rcnMakeST(PT,Arg2,Arg3,ST),
	CondM=cmp(ST,[Arg2M,Arg3M]),
	!
;
	Cond=cmp(PT,[Arg2,Arg3|D_Variables]),
	dbBase::smClassDefF(PT,"p",Card,DEs,_,_,_),
	Card=2,
	set::size(DEs,N),
	N>0,
	rcnArgumentVariable(PT,Arg2,Ass,Arg2M),
	rcnArgumentVariable(PT,Arg3,Ass,Arg3M),
	rcnD_Variables(D_Variables,Ass,D_VariablesM),
	rcnGetArgNo(cmp("",D_Variables),No),
	SNo=toString(No),
	sx::concat([PT,"[",SNo,"]"],PTNO),
	rcnMakeST(PT,Arg2,Arg3,ST),
	CondM=cmp(ST,[Arg2M,cmp(PTNO,D_VariablesM),Arg3M]),
	!.
rcnMakeST(PT,A2,A3,ST) :-
	A3=nill,
	A2=var(V),
	sx::sxlJoin([PT,V],"|",ST),
	!
;
	A3=nill,
	A2=cmp(ET,_),
	sx::sxlJoin([PT,ET],"|",ST),
	!
;	
	A2=var(V),
	A3=var(W),
	sx::sxlJoin([PT,V,W],"|",ST),
	!
;
	A2=var(V),
	A3=cmp(ET,_),
	sx::sxlJoin([PT,V,ET],"|",ST),
	!
;
	A2=cmp(ET,_),
	A3=var(W),
	sx::sxlJoin([PT,ET,W],"|",ST),
	!
;
	A2=cmp(E2,_),
	A3=cmp(E3,_),
	sx::sxlJoin([PT,E2,E3],"|",ST),
	!.
% ----------------------------------------------------------------------------------------------
% rcnVariableCondition
% ----------------------------------------------------------------------------------------------
class predicates
rcnCheckVar : (string,string,string,string_list,string_list) determ (i,i,o,i,o).
rcnVarExists : (string,string_list) determ (i,i).
clauses
rcnVariableCondition(Cond,CondM,Vars,VarsM) :-
	% remove conditions involving DEs as these are moved to the sentence-conditions
	Cond=cmp(_,[var(DE),_]),
	vars:: varComps(DE,_,_,_),
	CondM=nill,
	VarsM=Vars,
	!
;
	Cond=cmp(Opr,[var(V),Exp]),
	rcnCheckVar(V,Opr,W,Vars,VarsM),
	rcnExpression(Exp,ExpM),
	opr::oprIntExt(Opr,Op),
	CondM=cmp(W,[cmp(Op,[ExpM])]),
	VarsM=[W|Vars],
	!
;
	Cond=cmp(Opr,[var(V),Exp]),
	opr::oprIntExt(Opr,Op),
	rcnExpression(Exp,ExpM),
	CondM=cmp(V,[cmp(Op,[ExpM])]),
	VarsM=Vars,
	!
;
	dlgIF::feil("rcnVariableCondition",[],[],Cond),!,fail.
rcnCheckVar(V,_,V,Vars,Vars) :-
	rcnVarExists(V,Vars),
	!.
rcnCheckVar(V,Opr,W,Vars,[W|Vars]) :-
	opr::oprDomF(Dom,_,Opr),
	sx::frontstr(4,Dom,_,DomT),
	DomU=string::toUpperCase(DomT),
	sx::concat([V," /",DomU],W),
	!.
rcnVarExists(Var,[V|Vs]) :-
	sx::sxUnpackS(V," /",L),
	L=[H,_],
	Var=H,
	!
;
	rcnVarExists(Var,Vs),
	!.
end implement rt
