/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement rub
open core,termDomain,resourceIdentifiers
constants
className = "com/rudstrand/rub/rub".
classVersion = "1.0".
class predicates
rubGetNewArgNos : (string,integer,sterm,integer_list,integer_list,integer) procedure (i,i,i,o,o,o).
rubGetNodeText : (string_list,string,string) determ (i,i,o).
rubNIS : (integer,integer,sterm,integer_list,integer_list,string_list,string_list) determ (i,i,i,i,o,i,o).
rubRenumber : (string_list,integer_list,integer_list,sterm,string_list,sterm) procedure (i,i,i,i,o,o).
clauses
classInfo(className, classVersion).
% ---------------------------------------------------------------------------
% rubCountNodes
% ---------------------------------------------------------------------------
class predicates
rubCountNodesX : (sterml,integer) determ (i,o).
clauses
rubCountNodes(cmp(_,L),N) :-
	rubCountNodesX(L,M),
	N=M+1,!.
rubCountNodes(_,1) :-
	!.
	
rubCountNodesX([],0) :-
	!.
rubCountNodesX([H|T],N) :-
	rubCountNodes(H,L),
	rubCountNodesX(T,M),
	N=L+M,!.		
% ----------------------------------------------------------------------------------------------
% rubDE_Add
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	-1/1 implode/explode
% 2	in	rule editor window
% 3	in	argument no
% 4	in	argument (NL)
% 5	in	argument's DEs (not expanded) (IF)
% 6	out	argument's DE variables
%
% DEs are added to the rule editor's DE list control. 
% ARGUMENT | NO | DE | DE_VAR | DENL[NO] or 
%
rubDE_Add(_,_,_,_,[],[]) :-
	!.
rubDE_Add(-1,REW,SNo,E,[DE|DEs],[V|Vs]) :-
	dbBase::nlSing(DE,DEnl),
	No=sx::toInt(SNo),
	vars::varCompsR(V,E,No,DE),
	sx::concat([DEnl,"[",SNo,"]"],DEnlNo),
	sx::sxlJoin([E,SNo,DE,V,DEnlNo],"|",DElist),
	guix::lbxAdd(REW,idc_des,DElist),
	rubDE_Add(-1,REW,SNo,E,DEs,Vs),
	!.
rubDE_Add(1,REW,SNo,E,[DE|DEs],Vs) :-
	dbBase::smClassDefF(DE,"a",_,_,_,_,_),	
	dbBase::nlSing(DE,DEnl),
	No=sx::toInt(SNo),
	vars::varCompsR(V,E,No,DE),
	sx::concat([DEnl,"[",SNo,"]"],DEnlNo),
	sx::sxlJoin([E,SNo,DE,V,DEnlNo],"|",DElist),
	guix::lbxAdd(REW,idc_des,DElist),
	rubDE_Add(1,REW,SNo,E,DEs,Vs2),
	Vs=[V|Vs2],
	!.
% ---------------------------------------------------------------------------
% rubDE_Adjust
% ---------------------------------------------------------------------------
% parameters:
% 1	in	DE-vars, format: DE_VAR 
% 2	out	DE-vars, format: ARGUMENT | NO | DE | DE_VAR | DENL[NO] or 
% 3	out	do. but only conclusion DEs
%
rubDE_Adjust([],[],[]) :-
	!.
rubDE_Adjust([H|T],[Hm|Tm],[Hm|Cm]) :-
	vars::varComps(H,A,N,D),
	N<10,
	dbBase::nlSing(D,Dn),
	Ns=toString(N),
	sx::concat([Dn,"[",Ns,"]"],Dx),
	sx::sxlJoin([A,Ns,D,H,Dx],"|",Hm),
	rubDE_Adjust(T,Tm,Cm),
	!.
rubDE_Adjust([H|T],[Hm|Tm],C) :-
	vars::varComps(H,A,N,D),
	N>=10,
	dbBase::nlSing(D,Dn),
	Ns=toString(N),
	sx::concat([Dn,"[",Ns,"]"],Dx),
	sx::sxlJoin([A,Ns,D,H,Dx],"|",Hm),
	rubDE_Adjust(T,Tm,C),
	!
;
	dlgIF::feil("rubDE_Adjust",[H],[],nill),!,fail.
% ---------------------------------------------------------------------------
% rubDEs
% ---------------------------------------------------------------------------
class predicates
rubDEsX : (integer,integer,string_list,string_list) determ (i,i,i,o).	
rubGetHigh : (integer,integer,sterm,integer,integer) procedure (i,i,i,i,o).
rubGetHighX : (integer,integer,sterml,integer,integer) determ (i,i,i,i,o).
clauses

% returns DEs to be used for populating editor insert box
%
rubDEs(Cond,Type,REW,DEs) :-
	Cond=2,
	set::member(Type,[21,22]),
	% conclusion DEs can be included
	guix::lbxGetAll(REW,idc_des,LDEs),
	rubDEsX(Type,9999,LDEs,DEs),
	!
;
	guix::lbxGetAll(REW,idc_des,DsAll),
	guix::lbxGetAll(REW,idc_des_conc,Ds2),
	set::setRemSub(Ds2,DsAll,LDEs),
	rubDEsX(Type,9999,LDEs,DEs),
	!.	

rubDEsX(_,_,[],[]) :-
	!.
rubDEsX(4,High,[LDE|LDEs],[DE|DEs]) :-
	sx::sxGetInBrack(LDE,'[',']',SNo),
	No=sx::toInt(SNo),
	No>=10,
	sx::sxUnpack(LDE,'|',L),
	L=[_,_,_,_,DE],
	rubDEsX(4,High,LDEs,DEs),
	!
;
	sx::sxGetInBrack(LDE,'[',']',SNo),
	No=sx::toInt(SNo),
	No>=10,
	sx::sxUnpack(LDE,'|',L),
	L=[_,_,_,_,_,DE],
	rubDEsX(4,High,LDEs,DEs),
	!.
rubDEsX(Type,High,[LDE|LDEs],[DE|DEs]) :-
	Type<>4,
	sx::sxGetInBrack(LDE,'[',']',SNo),
	No=sx::toInt(SNo),
	No<=High,
	sx::sxUnpack(LDE,'|',L),
	L=[_,_,_,_,DE],
	rubDEsX(Type,High,LDEs,DEs),
	!
;
	Type<>4,
	sx::sxGetInBrack(LDE,'[',']',SNo),
	No=sx::toInt(SNo),
	No<=High,
	sx::sxUnpack(LDE,'|',L),
	L=[_,_,_,_,_,DE],
	rubDEsX(Type,High,LDEs,DEs),
	!.
rubDEsX(Type,High,[_|LDEs],DEs) :-
	rubDEsX(Type,High,LDEs,DEs),
	!.

rubGetHigh(No,Max,_,High,High) :-
	No>Max,
	!.	
rubGetHigh(No,Max,cmp(LogOpr,Conds),High,Higher) :-
	set::member(LogOpr,["&","v"]),
	Next=No+1,
	rubGetHighX(Next,Max,Conds,High,Higher),
	!.
rubGetHigh(_,_,cmp(STb,[cmp(Eb,_)|_]),High,Higher) :-
	sx::sxUnpack(STb,'{',_), 
	sx::sxGetInBrack(Eb,'[',']',SNo),
	High2=sx::toInt(SNo),
	set::setMax([High,High2],Higher),
	!.
rubGetHigh(_,_,_,High,High) :-
	!.		
	
rubGetHighX(_,_,[],High,High) :-
	!.
rubGetHighX(No,Max,[Cond|Conds],High,Higher) :-
	rubGetHigh(No,Max,Cond,High,High2), 
	Next=No+1,
	rubGetHighX(Next,Max,Conds,High2,Higher),
	!.
% ---------------------------------------------------------------------------
% rubDE_VarDom
% ---------------------------------------------------------------------------
rubDE_VarDom(V,Dom) :-
	vars::varComps(V,_,_,DE),
	dbBase::smClassDefF(DE,"a",_,_,Dom,_,_),
	!.
% ---------------------------------------------------------------------------
% rubGetDV
% ---------------------------------------------------------------------------
class predicates
rubGetDVsX : (sterml,string_list,string_list,string_list) determ (i,o,o,o).	
rubGetDVs_DEs : (sterml,string_list) determ (i,o).
rubGetDVs_DE : (sterm,string_list) determ (i,o).	
rubGetDV_CheckArg : (string,string_list) procedure (i,o).
clauses
% parameters:
% 1	in	rule-tree
% 2	out	DE-variables in tree
% 3	out	simple variables in tree (including /DD where domain-indicator is used)
% 4	out	argument variables in tree
%
rubGetDVs(T,DVs,Vars,ArgVars) :-
	% ST
	T=cmp(ST,Args),
	sx::sxUnpack(ST,'{',L),
	L=[STp|_],
	sx::sxUnpack(STp,'|',L2),
	L2=[Pk,Arg],
	dbBase::smClassDefF(Pk,"p",_,_,_,_,_),	
	rubGetDV_CheckArg(Arg,ArgVars),
	rubGetDVsX(Args,DVs,Vars,_),
	!
;
	% ST
	T=cmp(ST,Args),
	sx::sxUnpack(ST,'{',L),
	L=[STp|_],
	sx::sxUnpack(STp,'|',L2),
	L2=[Pk,Arg2,Arg3],
	dbBase::smClassDefF(Pk,"p",_,_,_,_,_),	
	rubGetDV_CheckArg(Arg2,ArgVars2),
	rubGetDV_CheckArg(Arg3,ArgVars3),
	rubGetDVsX(Args,DVs,Vars,_),
	set::conc([ArgVars2,ArgVars3],ArgVars),
	!
;
	% argument
	T=cmp(A,DTs),
	sx::sxGetInBrack(A,'[',']',_),
	rubGetDVs_DEs(DTs,DVs),
	Vars=[],
	ArgVars=[],
	!
;
	% sentence count
	T=cmp(V,[cmp(Op,[cmp(H,[SSS])])]),
	sx::frontstr(7,H,F,_),
	F="p3_iSSS",
	sx::sxUnpack(Op,'{',L2),
	L2=[Op2|_],
	opr::oprEq(Op2),
	sx::sxUnpack(V,'{',L3),
	L3=[Var|_],
	rubIsVar(Var,_,_),
	Vars=[Var],
	rubGetDVs(SSS,DVs,_,_),
	ArgVars=[],
	!
;
	% simple var (definition)
	T=cmp(V,[cmp(Op,_)]),
	sx::searchstring(V," /",_),
	sx::sxUnpack(Op,'{',L2),
	L2=[Op2|_],
	opr::oprEq(Op2),
	sx::sxUnpack(V,'{',L3),
	L3=[Var|_],
	rubIsVar(Var,_,_),
	DVs=[],
	Vars=[Var],
	ArgVars=[],
	!
;
	% simple var (not definition)
	T=cmp(V,[cmp(Op,_)]),
	sx::sxUnpack(Op,'{',L2),
	L2=[Op2|_],
	opr::opr_cmpass(Op2),
	sx::sxUnpack(V,'{',L3),
	L3=[Var|_],
	rubIsVar(Var,_,_),
	DVs=[],
	Vars=[Var],
	ArgVars=[],
	!
;
	T=cmp(_,L),
	rubGetDVsX(L,DVs,Vars,ArgVars),
	!
;
	DVs=[],
	Vars=[],
	ArgVars=[],
	!.
rubGetDV_CheckArg(Arg,ArgVars) :-
	string::front(Arg,1,F,_),
	Fu=string::toUpperCase(F),
	Fu=F, % = argument variable
	ArgVars=[Arg],
	!
;
	ArgVars=[],
	!.
rubGetDVsX([],[],[],[]) :-
	!.
rubGetDVsX([H|T],DEs,Vars,ArgVars) :-
	rubGetDVs(H,DEs2,Vars2,ArgVars2),
	rubGetDVsX(T,DEs3,Vars3,ArgVars3),
	set::conc([DEs2,DEs3],DEs),
	set::conc([Vars2,Vars3],Vars),
	set::conc([ArgVars2,ArgVars3],ArgVars),
	!.
rubGetDVs_DEs([],[]) :-
	!.
rubGetDVs_DEs([H|T],DVs) :-
	rubGetDVs_DE(H,DVs2),
	rubGetDVs_DEs(T,DVs3),
	set::conc([DVs2,DVs3],DVs),
	!.
rubGetDVs_DE(DC,DVs) :-
	DC=var(DV),
	sx::sxUnpack(DV,'{',L),
	L=[DV2|_],
	DVs=[DV2],
	!
;	
	DC=cmp(DV,_),
	sx::sxUnpack(DV,'{',L),
	L=[DV2|_],
	DVs=[DV2],
	!
;
	dlgIF::feil("rubGetDVs_DE",[],[],DC),!,fail.
% ---------------------------------------------------------------------------
% rubDummyAss
% ---------------------------------------------------------------------------
% generates assignments to nill for all DEs in the conclusion in a new rule
%
rubDummyAss(_,[],[]) :-
	!.
rubDummyAss(Exp,[DE|DEs],[cmp(DE,[cmp(Opr,[nill])])|Exs]) :-
	dbBase::kbID(DE,Dk,_,_),
	dbBase::smClassDef(Dk,"a",_,_,Dom,_,_),		
	dom::domShortLong(DomS,Dom),
	sx::concat(["=",DomS],Opr),
	rubDummyAss(Exp,DEs,Exs),
	!.
% ------------------------------------------------------------------------------------------------------------
% rubElimDEs
% ------------------------------------------------------------------------------------------------------------
class predicates
rubElimDEsX : (sterml,string_list,string_list) determ (i,i,o).	
rubED_Args : (sterml,string_list,string_list) determ (i,i,o).	
rubED_GetDEs : (string,string_list,string_list) determ (i,i,o).	
clauses
% eliminates DEs from par#2 which are not referred in the rule (because of OR elimination)
%
rubElimDEs(cmp(RT,B),DEsC,DEsR) :-
	rubRuleType(RT,_,_),
	rubElimDEsX(B,DEsC,DEsR),
	!.
rubElimDEs(cmp(LO,B),DEsC,DEsR) :-
	set::member(LO,["&","v","not"]),
	rubElimDEsX(B,DEsC,DEsR),
	!.
rubElimDEs(cmp("#",B),DEsC,DEsR) :-
	rubElimDEsX(B,DEsC,DEsR),
	!.
rubElimDEs(cmp(STb,Args),DEsC,DEsR) :-
	sx::sxUnpack(STb,'{',_), 
	rubED_Args(Args,DEsC,DEsR),
	!.	
rubElimDEs(_,_,[]) :-
	!.
	
rubElimDEsX([],_,[]) :-
	!.
rubElimDEsX([H|T],DEsC,DEsR) :-
	rubElimDEs(H,DEsC,DEs2),
	rubElimDEsX(T,DEsC,DEs3),
	set::conc([DEs2,DEs3],DEsR),
	!.
rubED_Args([],_,[]) :-	
	!.
rubED_Args([cmp(Eb,_)|Args],DEsC,DEsR) :-	
	sx::sxGetInBrack(Eb,'[',']',SNo),
	sx::sxUnpack(Eb,'[',L),
	L=[E|_],
	sx::concat([SNo,"|",E,"|"],DEfr),
	rubED_GetDEs(DEfr,DEsC,DEs2),
	rubED_Args(Args,DEsC,DEs3),
	set::conc([DEs2,DEs3],DEsR),
	!.
rubED_Args([var(_)|Args],DEsC,DEsR) :-	
	rubED_Args(Args,DEsC,DEsR),
	!.
	
rubED_GetDEs(_,[],[]) :-
	!.
rubED_GetDEs(F,[DE|DEs],[DE|DEsR]) :-
	sx::searchstring(DE,F,P),
	P=1,
	rubED_GetDEs(F,DEs,DEsR),
	!.
rubED_GetDEs(F,[_|DEs],DEsR) :-
	rubED_GetDEs(F,DEs,DEsR),
	!.
% ---------------------------------------------------------------------------
% rubFirstNIL
% ---------------------------------------------------------------------------
class predicates
rubFNX : (sterm,integer,integer) determ (i,i,o).
rubFNXS : (sterml,integer,integer) determ (i,i,o).
clauses
% returns the number of the first NIL-node >= No  (or 1 if no nill)
%
rubFirstNIL(Rule,No,N) :-
	rubFNX(Rule,No,N),
	!
;
	N=1,!.	
rubFNX(Node,Min,N) :-
	rubIsNill(Node),
	rubNodeNo(Node,N),
	N>=Min,
	!.	
rubFNX(cmp(_,B),Min,F) :-
	rubFNXS(B,Min,F),
	!.
rubFNXS([H|T],Min,F) :-
	rubFNX(H,Min,F),
	!
;
	rubFNXS(T,Min,F),
	!.

% returns lists of DEs (IF) and adds DEs to the windows DE list control 
% parameters:
% 1	in	rule window
% 2	in	1/2: unary/binary ST
% 3	in	pred
% 4	in	1st ent or blank
% 5	in	2nd ent or blank
% 6	in	argument no
% 7	in	list of DE vars corr. to par#4 
% 8	out	list of DE vars corr. to par#5
% 9 	out	list of DE vars corr. to par#6 
%
rubGetDEsK(REW,1,Pred,"",_,No,PP_DEs,[],[]) :-
	SNo=toString(No),
	dbBase::smClassDef(Pred,"p",_,P_DEs,_,_,_),	
	rubDE_Add(-1,REW,SNo,Pred,P_DEs,PP_DEs),
	!.
rubGetDEsK(REW,1,Pred,Ent1,_,No,PP_DEs,E1P_DEs,[]) :-
	SNo=toString(No),
	dbBase::smClassDef(Pred,"p",_,P_DEs,_,_,_),	
	rubDE_Add(-1,REW,SNo,Pred,P_DEs,PP_DEs),
	No1=No+1,
	SNo1=toString(No1),
	dbBase::smClassDef(Ent1,"e",_,E1_Des,_,_,_),		
	rubDE_Add(-1,REW,SNo1,Ent1,E1_DEs,E1P_DEs),
	!.
rubGetDEsK(REW,2,Pred,"","",No,PP_DEs,[],[]) :-
	SNo=toString(No),
	dbBase::smClassDef(Pred,"p",_,P_DEs,_,_,_),	
	rubDE_Add(-1,REW,SNo,Pred,P_DEs,PP_DEs),
	!.
rubGetDEsK(REW,2,Pred,"",Ent2,No,PP_DEs,[],E2P_DEs) :-
	SNo=toString(No),
	dbBase::smClassDef(Pred,"p",_,P_DEs,_,_,_),	
	rubDE_Add(-1,REW,SNo,Pred,P_DEs,PP_DEs),
	No2=No+1,
	SNo2=toString(No2),
	dbBase::smClassDef(Ent2,"e",_,E2_Des,_,_,_),		
	rubDE_Add(-1,REW,SNo2,Ent2,E2_DEs,E2P_DEs),
	!.
rubGetDEsK(REW,2,Pred,Ent1,"",No,PP_DEs,E1P_DEs,[]) :-
	SNo=toString(No),
	dbBase::smClassDef(Pred,"p",_,P_DEs,_,_,_),	
	rubDE_Add(-1,REW,SNo,Pred,P_DEs,PP_DEs),
	No1=No+1,
	SNo1=toString(No1),
	dbBase::smClassDef(Ent1,"e",_,E1_Des,_,_,_),		
	rubDE_Add(-1,REW,SNo1,Ent1,E1_DEs,E1P_DEs),
	!.
rubGetDEsK(REW,2,Pred,Ent1,Ent2,No,PP_DEs,E1P_DEs,E2P_DEs) :-
	SNo=toString(No),
	dbBase::smClassDef(Pred,"p",_,P_DEs,_,_,_),	
	rubDE_Add(-1,REW,SNo,Pred,P_DEs,PP_DEs),
	No1=No+1,
	SNo1=toString(No1),
	dbBase::smClassDef(Ent1,"e",_,E1_Des,_,_,_),		
	rubDE_Add(-1,REW,SNo1,Ent1,E1_DEs,E1P_DEs),
	No2=No1+1,
	SNo2=toString(No2),
	dbBase::smClassDef(Ent2,"e",_,E2_Des,_,_,_),		
	rubDE_Add(-1,REW,SNo2,Ent2,E2_DEs,E2P_DEs),
	!.
% ----------------------------------------------------------------------------------------------
% rubGetNewArgNos
% ----------------------------------------------------------------------------------------------
class predicates
rubGetNewArgNosX : (string,integer,sterml,integer_list,integer_list,integer) determ (i,i,i,o,o,o).
clauses
% parameters:
% 4	out	old/existing sentence argument numbers (0,1,2,..)
% 5	out	corresponding new sentence argument numbers (0,2,3,..)
%
rubGetNewArgNos(Typ,No,cmp(LogOpr,Conds),Olds,News,Next) :-
	sx::sxUnpack(LogOpr,'{',L),
	L=[Op|_],
	set::member(Op,["&","v","not","p3_iSSS"]),
	rubGetNewArgNosX(Typ,No,Conds,Olds,News,Next),
	!.
rubGetNewArgNos(Typ,No,cmp(_,[cmp(_,[cmp(SSS,Conds)])]),Olds,News,Next) :-
	sx::sxUnpack(SSS,'{',L),
	L=[Op|_],
	Op="p3_iSSS",
	rubGetNewArgNosX(Typ,No,Conds,Olds,News,Next),
	!.
rubGetNewArgNos(_,No,cmp(ST,[cmp(Eb,_)|_]),[Old],[No],Next) :-
	sx::sxUnpack(ST,'{',L2),
	L2=[STp|_],
	sx::sxUnpack(STp,'|',L3),
	L3=[P|_],
	dbBase::smClassDefF(P,"p",_,_,_,_,_),	
	sx::sxGetInBrack(Eb,'[',']',SOldNo),
	OldNo=sx::toInt(SOldNo),
	Old=OldNo div 10,
	Next=No+1,
	!.
rubGetNewArgNos(_,No,cmp(ST,[var(_),cmp(Eb,_)|_]),[Old],[No],Next) :-
	sx::sxUnpack(ST,'{',L2),
	L2=[STp|_],
	sx::sxUnpack(STp,'|',L3),
	L3=[P|_],
	dbBase::smClassDefF(P,"p",_,_,_,_,_),	
	sx::sxGetInBrack(Eb,'[',']',SOldNo),
	OldNo=sx::toInt(SOldNo),
	Old=OldNo div 10,
	Next=No+1,
	!.
rubGetNewArgNos(_,No,_,[0],[0],No) :-
	!.		
	
rubGetNewArgNosX(_,No,[],[],[],No) :-
	!.
rubGetNewArgNosX("4",No,[Cond|Conds],Olds,News,Next) :-
	% generalization rule: same argument numbers are used in the OR-branches
	rubGetNewArgNos("4",No,Cond,Old,New,_), 
	rubGetNewArgNosX("4",No,Conds,OldsX,NewsX,Next),
	set::conc([Old,OldsX],Olds),
	set::conc([New,NewsX],News),
	!.
rubGetNewArgNosX(Typ,No,[Cond|Conds],Olds,News,Next) :-
	rubGetNewArgNos(Typ,No,Cond,Old,New,Next2), 
	rubGetNewArgNosX(Typ,Next2,Conds,OldsX,NewsX,Next),
	set::conc([Old,OldsX],Olds),
	set::conc([New,NewsX],News),
	!.
/*
rubGetNewArgNosX(Typ,No,[_|Conds],Olds,News,Next) :-
	rubGetNewArgNosX(Typ,No,Conds,Olds,News,Next),
	!.
*/
% ----------------------------------------------------------------------------------------------
% rubGetNodeText
% ----------------------------------------------------------------------------------------------
% gets the full text of the node matching the given text tail
% parameters:
% 1	in	list of all node full texts
% 2	in	tail to be matched
% 3	out	full text of matching node
%
rubGetNodeText([H|T],Tail,Name) :-
	sx::searchstring(H,Tail,P),
	Hn=string::length(H),
	Tn=string::length(Tail),
	Hn=(P-1+Tn),
	Name=H,!
;
	rubGetNodeText(T,Tail,Name),!.
% ----------------------------------------------------------------------------------------------
% rubIsNill
% ----------------------------------------------------------------------------------------------
rubIsNill(C) :-
	C=atom(S),
	sx::frontstr(2,S,F,_),
	F="_{",
	!.
% ----------------------------------------------------------------------------------------------
% rubListHead
% ----------------------------------------------------------------------------------------------
rubListHead(H,DS,DL) :-
	sx::frontstr(1,H,F,R2),
	F="[",
	sx::sxTail(1,R2,DS,R3),
	R3="]",
	dom::domShortLong(DS,DL),
	!.
% ----------------------------------------------------------------------------------------------
% rubNodeGetCurr
% ----------------------------------------------------------------------------------------------
rubNodeGetCurr(REW,NodeNo,NodeText,ParentNo,ParentText) :-
	rubTW_Get(REW,TRW),
	p3Tree::drwGetCurr(TRW,NodeNo,NodeText,_),
	PTextX=vpiTree::getParentNode(TRW),
	sx::sxUnpack(PTextX,'{',PTs),
	PTs=[ParentText,SParentNo],
	ParentNo=sx::toInt(SParentNo),
	!
;
	rubTW_Get(REW,TRW),
	p3Tree::drwGetCurr(TRW,NodeNo,NodeText,_),
	ParentNo=0,
	ParentText="",
	!.	
% ---------------------------------------------------------------------------
% rubNodeNo
% ---------------------------------------------------------------------------
class predicates
rubNodeNoX : (string,integer) determ (i,o).
clauses
% returns a node's number
%
rubNodeNo(C,N) :-
	C=cmp(H,_),
	rubNodeNoX(H,N),
	!
;
	C=str(H),
	rubNodeNoX(H,N),
	!
;
	C=atom(H),
	rubNodeNoX(H,N),
	!
;
	C=var(H),
	rubNodeNoX(H,N),
	!
;
	C=fint(_,H),
	N=sx::toInt(H),
	!
;
	C=ffix(_,H),
	N=sx::toInt(H),
	!
;
	dlgIF::feil("rubNodeNo",["Invalid rule-tree node"],[],C),!,fail.
	
rubNodeNoX(H,N) :-
	sx::sxUnpack(H,'}',L),
	L=[_,SNo],
	N=sx::toInt(SNo),
	!.
% ---------------------------------------------------------------------------
% rubNIS
% ---------------------------------------------------------------------------
/*
node classification for current node: 
-------------------------------------
1=condition main branch (as parent only)
2=conclusion main branch (as parent only)
3=log opr (rule cond level)
10=sent cond 
11=arg 
12=arg var
13=not
19=complex data element type (sent cond level - only exploded)
20=var cond
21=DE (sent cond level incl. MDE and imploded CDE)
29=log opr (DE / var cond level)
31=oprass
32=oprcmp
40=oprexp (e.g. arithmetic)
41=lit
42=var (rside level - incl. DEs)
43=function/CR/procedure call
44=#
45=list head
99=nill
*/
class predicates
rubNISx : (integer,integer,sterml,integer_list,integer_list,string_list,string_list) determ (i,i,i,i,o,i,o).
clauses

rubNIS_Trace(W,N,TNo,R,TypesC,Types,NamesC,Names) :-
	tt::getLevel()>3,
	rubNIS(N,TNo,R,TypesC,Types,NamesC,Names),
	sx::sxl_ilR(St2,Types),
	sx::sxlJoin(St2,"|",St),
	sx::sxlJoin(Names,"||",Sn),
	guix::txtSet(W,idc_nis,St),
	guix::txtSet(W,idc_nas,Sn),
	sx::sxlJoin(St2,"-",S2),
	dlgIF::msgS(S2),
	dlgIF::msgS(Sn),
	!
;
	tt::getLevel()>3,
	dlgIF::msgS("NIS failed"),
	Types=[],
	Names=[],
	guix::txtSet(W,idc_nis,""),
	guix::txtSet(W,idc_nas,""),
	!
;	
	rubNIS(N,TNo,R,TypesC,Types,NamesC,Names),
	sx::sxl_ilR(St2,Types),
	sx::sxlJoin(St2,"|",St),
	sx::sxlJoin(Names,"||",Sn),
	guix::txtSet(W,idc_nis,St),
	guix::txtSet(W,idc_nas,Sn),
	!
;
	Types=[],
	Names=[],
	guix::txtSet(W,idc_nis,""),
	guix::txtSet(W,idc_nas,""),
	!.

% parameters:
% 1	in	node no. for head root of given sub-tree (1 for rule root)
% 2	in	target node no (must be >1)
% 3	in	rule sub-tree
% 4	out	node type list [target node,target node's father,...,root] 
%               see doc.pro for node type codes
% 5	out	node name list corr. to par#4
%
rubNIS(_,1,_,_,[],_,[]) :-
	!.
rubNIS(N,TNo,_,Types,Types,Names,Names) :-
	N>TNo,
	!.
rubNIS(1,TNo,cmp(RuleType,[Cond,Conc]),_,Types,_,Names) :-
	rubCountNodes(Cond,L),
	Last=L+1,
	TNo<=Last,
	sx::sxUnpack(RuleType,'{',L2),
	L2=[RT|_],
	!, rubNIS(2,TNo,Cond,[1],Types,[RT],Names), % condition
	!
;
	rubCountNodes(Cond,L),
	RootNo=L+2,
	sx::sxUnpack(RuleType,'{',L2),
	L2=[RT|_],
	rubNIS(RootNo,TNo,Conc,[2],Types,[RT],Names), % conclusion
	!.
rubNIS(N,N,Node,TypesC,[99|TypesC],NamesC,["nill"|NamesC]) :-
	rubIsNill(Node),
	!.
% 1/2 rule condition or conclusion can have nodes:
% log opr (non-terminal)
% var cond (non-terminal)
% sent cond (non-terminal)
% not (non-terminal)
rubNIS(N,TNo,cmp(Hx,R),[C],Types,NamesC,Names) :-
	C<3,
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	Types=[3,C],
	Names=[H|NamesC],
	!
;
	C<3,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[3,C],Types,[H|NamesC],Names),
	!
;
	C<3,
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	H="not",
	Types=[13,C],
	Names=[H|NamesC],
	!
;
	C<3,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	H="not",
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[13,C],Types,[H|NamesC],Names),
	!
;
	C<3,
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubIsVar(H,_,_),
	Types=[20,C],
	Names=[H|NamesC],
	!
;
	C<3,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubIsVar(H,_,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[20,C],Types,[H|NamesC],Names),
	!
;	
	C<3,
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[10,C],
	Names=[H|NamesC],
	!
;
	C<3,
	rubCountNodes(cmp(Hx,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubNISx(Next,TNo,R,[10,C],Types,[H|NamesC],Names),
	!.
% 3 log opr (rule cond level) can have node:
% log opr  (non-terminal)
% sent cond (non-terminal)
% var cond (non-terminal)
% not (non-terminal)
rubNIS(N,TNo,cmp(Hx,R),[3|TypesC],Types,NamesC,Names) :-
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	Types=[3,3|TypesC], 
	Names=[H|NamesC],
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[3,3|TypesC],Types,[H|NamesC],Names),
	!
;
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	H="not",
	Types=[13,3|TypesC],
	Names=[H|NamesC],
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	H="not",
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[13,3|TypesC],Types,[H|NamesC],Names),
	!
;
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubIsVar(H,_,_),
	Types=[20,3|TypesC],
	Names=[H|NamesC],
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubIsVar(H,_,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[20,3|TypesC],Types,[H|NamesC],Names),
	!
;	
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[10,3|TypesC],
	Names=[H|NamesC],
	!
;
	rubCountNodes(cmp(Hx,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubNISx(Next,TNo,R,[10,3|TypesC],Types,[H|NamesC],Names),
	!.
% 10 sent cond can have nodes:
% arg var (terminal)
% arg (non-terminal)
rubNIS(N,TNo,var(Hx),[10|TypesC],Types,NamesC,Names) :-
	N=TNo,
	Types=[12,10|TypesC],
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Names=[H|NamesC],
	!.
rubNIS(N,TNo,cmp(Hx,R),[10|TypesC],Types,NamesC,Names) :-
	N=TNo,
	Types=[11,10|TypesC],
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Names=[H|NamesC],
	!
;
	rubCountNodes(cmp(Hx,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubNISx(Next,TNo,R,[11,10|TypesC],Types,[H|NamesC],Names),
	!.
% 11 arg can have nodes:
% DE (terminal or non-terminal)
rubNIS(N,TNo,var(Hx),[11|TypesC],Types,NamesC,Names) :-
	N=TNo,
	Types=[21,11|TypesC],
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Names=[H|NamesC],
	!.
rubNIS(N,TNo,cmp(Hx,R),[11|TypesC],Types,NamesC,Names) :-
	N=TNo,
	Types=[21,11|TypesC],
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Names=[H|NamesC],
	!
;
	rubCountNodes(cmp(Hx,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubNISx(Next,TNo,R,[21,11|TypesC],Types,[H|NamesC],Names),
	!.
% 12 arg var can not have nodes
%
% 13 not can have node:
% sent cond (non-terminal)
rubNIS(N,TNo,cmp(Hx,R),[13|TypesC],Types,NamesC,Names) :-
	N=TNo,
	Types=[10,13|TypesC],
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Names=[H|NamesC],
	!
;
	rubCountNodes(cmp(Hx,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubNISx(Next,TNo,R,[10,13|TypesC],Types,[H|NamesC],Names),
	!.
rubNIS(N,TNo,cmp(Hx,R),[20|TypesC],Types,NamesC,Names) :-
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[29,20|TypesC],Types,[H|NamesC],Names),
	!
;	
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprEq(H),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[31,20|TypesC],Types,[H|NamesC],Names),
	!
;	
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprCmp(H),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[32,20|TypesC],Types,[H|NamesC],Names),
	!.
% 21 DE can have nodes:
% DE [MDE] (terminal or non-terminal)
% log opr (non-terminal)
% oprass (non-terminal)
% oprcmp (non-terminal)
rubNIS(N,N,var(Hx),[21|TypesC],Types,NamesC,Names) :-
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[21,21|TypesC],
	Names=[H|NamesC],
	!.
rubNIS(N,TNo,cmp(Hx,R),[21|TypesC],Types,NamesC,Names) :-
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubIsVar(H,_,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[21,21|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[29,21|TypesC],Types,[H|NamesC],Names),
	!
;	
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprEq(H),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[31,21|TypesC],Types,[H|NamesC],Names),
	!
;	
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprCmp(H),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[32,21|TypesC],Types,[H|NamesC],Names),
	!.
% 29 log opr (DE/var cond level) can have node:
% log opr (non-terminal)
% oprass (non-terminal)
% oprcmp (non-terminal)
rubNIS(N,TNo,cmp(Hx,R),[29|TypesC],Types,NamesC,Names) :-
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	Types=[29,29|TypesC], 
	Names=[H|NamesC],
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	set::member(H,["&","v"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[29,29|TypesC],Types,[H|NamesC],Names),
	!
;
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprEq(H),
	Types=[31,29|TypesC],
	Names=[H|NamesC],
	!
;
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprCmp(H),
	Types=[32,29|TypesC],
	Names=[H|NamesC],
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprEq(H),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[31,29|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::oprCmp(H),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[32,29|TypesC],Types,[H|NamesC],Names),
	!.
% 31 oprass can have node:
% lit (terminal)
% var (terminal)
% call (terminal or non-terminal)
% proc (terminal or non-terminal)
% oprexp (non-terminal)
% sentence count (non-terminal)
% list head (terminal or non-terminal)
rubNIS(N,TNo,R,[31|TypesC],Types,NamesC,Names) :-
	R=str(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	N=TNo,
	Types=[41,31|TypesC], 
	Names=[H|NamesC],
	!
;
	R=fint(H,_),
	N=TNo,
	Hs=toString(H),
	Types=[41,31|TypesC], 
	Names=[Hs|NamesC],
	!
;
	R=ffix(H,_),
	N=TNo,
	Hs=toString(H),
	Types=[41,31|TypesC], 
	Names=[Hs|NamesC],
	!
;
	R=var(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	N=TNo,
	Types=[42,31|TypesC], 
	Names=[H|NamesC],
	!
;
	R=atom(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	N=TNo,
	sx::frontstr(3,H,C,_),
	set::member(C,["[F]","[C]","[P]"]),
	Types=[43,31|TypesC],
	Names=[H|NamesC],
	!
;
	R=atom(Px),
	sx::sxUnpack(Px,'{',L),
	L=[P,_],
	N=TNo,
	Types=[41,31|TypesC],
	Names=[P|NamesC],
	!
;
	R=str(Px),
	sx::sxUnpack(Px,'{',L),
	L=["[]",_],
	N=TNo,
	Types=[45,31|TypesC],
	Names=["[]"|NamesC],
	!
;
	R=str(Px),
	sx::sxUnpack(Px,'{',L),
	L=[ListHead,_],
	sx::sxGetInBrack(ListHead,'[',']',D),
	dom::domShortLongF(D,_),
	N=TNo,
	Types=[45,31|TypesC],
	Names=[ListHead|NamesC],
	!.
rubNIS(N,TNo,cmp(Hx,R),[31|TypesC],Types,NamesC,Names) :-
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	H="p3_iSSS",
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[44,31|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	sx::frontstr(3,H,C,_),
	set::member(C,["[F]","[C]","[P]"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[43,31|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::opr_expr(H,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[40,31|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::opr_expr(H,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[40,31|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=["[]",_],
	rubCountNodes(cmp("[]",R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[45,31|TypesC],Types,["[]"|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[ListHead,_],
	sx::sxGetInBrack(ListHead,'[',']',D),
	dom::domShortLongF(D,_),
	rubCountNodes(cmp(ListHead,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[45,31|TypesC],Types,[ListHead|NamesC],Names),
	!.
% 32 oprcmp can have node:
% lit (terminal)
% var (terminal)
% oprexp (non-terminal)
rubNIS(N,N,R,[32|TypesC],Types,NamesC,Names) :-
	R=str(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[41,32|TypesC], 
	Names=[H|NamesC],
	!
;
	R=fint(H,_),
	Types=[41,32|TypesC], 
	Hs=toString(H),
	Names=[Hs|NamesC],
	!
;
	R=ffix(H,_),
	Types=[41,32|TypesC], 
	Hs=toString(H),
	Names=[Hs|NamesC],
	!
;
	R=atom(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[41,32|TypesC], 
	Names=[H|NamesC],
	!
;
	R=var(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[42,32|TypesC], 
	Names=[H|NamesC],
	!.
rubNIS(N,TNo,cmp(Hx,R),[32|TypesC],Types,NamesC,Names) :-
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::opr_expr(H,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[40,31|TypesC],Types,[H|NamesC],Names),
	!.
% 40 oprexp can have nodes:
% lit (terminal)
% var (terminal)
% oprexp (non-terminal)
% proc call
rubNIS(N,N,R,[40|TypesC],Types,NamesC,Names) :-
	R=str(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[41,40|TypesC], 
	Names=[H|NamesC],
	!
;
	R=fint(H,_),
	Hs=toString(H),
	Types=[41,40|TypesC], 
	Names=[Hs|NamesC],
	!
;
	R=ffix(H,_),
	Hs=toString(H),
	Types=[41,40|TypesC], 
	Names=[Hs|NamesC],
	!
;
	R=atom(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[41,40|TypesC], 
	Names=[H|NamesC],
	!
;
	R=var(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[42,40|TypesC], 
	Names=[H|NamesC],
	!
;
	R=atom(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	sx::frontstr(3,H,C,_),
	set::member(C,["[F]","[C]","[P]"]),
	Types=[43,40|TypesC],
	Names=[H|NamesC],
	!.
rubNIS(N,TNo,cmp(Hx,R),[40|TypesC],Types,NamesC,Names) :-
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::opr_expr(H,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[40,40|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	sx::frontstr(3,H,C,_),
	set::member(C,["[F]","[C]","[P]"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[43,40|TypesC],Types,[H|NamesC],Names),
	!.
% 43 function/CR/procedure call can have nodes:
% DE (non-terminal)
rubNIS(N,TNo,cmp(Hx,R),[43|TypesC],Types,NamesC,Names) :-
	N=TNo,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	Types=[21,43|TypesC],
	Names=[H|NamesC],
	!
;
	rubCountNodes(cmp(Hx,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubNISx(Next,TNo,R,[21,43|TypesC],Types,[H|NamesC],Names),
	!.
% 44 sentence count
% sent cond (non-terminal)
rubNIS(N,TNo,cmp(Hx,R),[44|TypesC],Types,NamesC,Names) :-
	rubCountNodes(cmp(Hx,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	rubNISx(Next,TNo,R,[10,44|TypesC],Types,[H|NamesC],Names),
	!.
% 45 list head can have node:
% lit (terminal)
% var (terminal)
% call (terminal or non-terminal)
% proc (terminal or non-terminal)
% oprexp (non-terminal)
% sentence count (non-terminal)
rubNIS(N,TNo,R,[45|TypesC],Types,NamesC,Names) :-
	R=str(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	N=TNo,
	Types=[41,45|TypesC], 
	Names=[H|NamesC],
	!
;
	R=fint(H,_),
	N=TNo,
	Hs=toString(H),
	Types=[41,45|TypesC], 
	Names=[Hs|NamesC],
	!
;
	R=ffix(H,_),
	N=TNo,
	Hs=toString(H),
	Types=[41,45|TypesC], 
	Names=[Hs|NamesC],
	!
;
	R=var(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	N=TNo,
	Types=[42,45|TypesC], 
	Names=[H|NamesC],
	!
;
	R=atom(Hx),
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	N=TNo,
	sx::frontstr(3,H,C,_),
	set::member(C,["[F]","[C]","[P]"]),
	Types=[43,45|TypesC],
	Names=[H|NamesC],
	!
;
	R=atom(Px),
	sx::sxUnpack(Px,'{',L),
	L=[P,_],
	N=TNo,
	Types=[41,45|TypesC],
	Names=[P|NamesC],
	!.
rubNIS(N,TNo,cmp(Hx,R),[45|TypesC],Types,NamesC,Names) :-
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	H="p3_iSSS",
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[44,45|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	sx::frontstr(3,H,C,_),
	set::member(C,["[F]","[C]","[P]"]),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[43,45|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::opr_expr(H,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[40,45|TypesC],Types,[H|NamesC],Names),
	!
;
	sx::sxUnpack(Hx,'{',L),
	L=[H,_],
	opr::opr_expr(H,_),
	rubCountNodes(cmp(H,R),Ln),
	Last=Ln+N,
	TNo<=Last,
	Next=N+1,
	rubNISx(Next,TNo,R,[40,45|TypesC],Types,[H|NamesC],Names),
	!.
% -----	
rubNISx(N,TNo,[T|Ts],TypesC,Types,NamesC,Names) :-
	rubCountNodes(T,Ln),
	Last=N+Ln-1,
	TNo<=Last,
	!, rubNIS(N,TNo,T,TypesC,Types,NamesC,Names),
	!	
;
	rubCountNodes(T,Ln),
	RootNo=Ln+N,
	!, rubNISx(RootNo,TNo,Ts,TypesC,Types,NamesC,Names),
	!.	
% ---------------------------------------------------------------------------
% rubIsVar
% ---------------------------------------------------------------------------
rubIsVar(V,W,Dom) :-
	sx::searchstring(V," /",_),
	not(vars::varComps(V,_,_,_)),
	not(vars::varFuncParR(_,_,V)),
	sx::sxUnpackS(V," /",L),
	L=[W,SDom],
	dom::domShortLongF(SDom,Dom),
	string::isName(W),
	string::front(V,1,F,_),
	Fu=string::toUpperCase(F),
	Fu=F,
	!
;
	not(sx::searchstring(V," /",_)),
	not(vars::varComps(V,_,_,_)),
	not(vars::varFuncParR(_,_,V)),
	string::isName(V),
	string::front(V,1,F,_),
	Fu=string::toUpperCase(F),
	Fu=F,
	W=V,
	Dom="",
	!.
% ----------------------------------------------------------------------------------------------
% rubNL
% ----------------------------------------------------------------------------------------------
class predicates
rubNLXs : (sterml,integer,sterml) determ (i,i,o).
rubNLX : (sterm,integer,sterm) determ (i,i,o).
clauses
% parameters:
% 1	in	rule-tree IF
% 2	out	rule-tree NL where
%		textual nodes (cmp head/var/str/atom) have format KBNAME{NLNAME}0 and ffix/ffint for fix/int
%
rubNL(RuleIF,RuleNL) :-
	rubNLX(RuleIF,-1,RuleNL),
	!.
% parameters:
% 2	1=parent node is a sentence component argument
%	2=do.		function/procedure/CR
%	3=parent node is CDE 
%	-1=else
%
rubNLX(nill,_,atom("_{_}0")) :-
	!.
rubNLX(int(A),_,fint(A,"0")) :-
	!.
rubNLX(fix(A),_,ffix(A,"0")) :-
	!.
rubNLX(atom(A),_,atom(NL)) :-
	sx::frontstr(4,A,F,T),
	set::member(F,["[F] ","[P] ","[C] "]),
	dbBase::nlSing(T,Tn),
	sx::concat([A,"{",F,Tn,"}0"],NL),
	!
;
	dbBase::nlSing(A,N),
	sx::concat([A,"{",N,"}0"],NL),
	!.
rubNLX(var(V),Parent,var(NL)) :-
	Parent=1,
	vars::varComps(V,_,_,D),
	dbBase::nlSing(D,N),
	sx::concat([V,"{",N,"}0"],NL),
	!
;
	Parent=-1,
	vars::varComps(V,_,No,D),
	dbBase::nlSing(D,Dn),
	SNo=toString(No),
	sx::concat([V,"{",Dn,"[",SNo,"]}0"],NL),
	!
;
	sx::concat([V,"{",V,"}0"],NL),
	!.
rubNLX(str(A),_,str(NL)) :-
	sx::sxQuote(A,N),
	sx::concat([A,"{",N,"}0"],NL),
	!.
rubNLX(cmp(H,L),Parent,cmp(NL,Lm)) :-
	H="[]",
	sx::concat([H,"{",H,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!
;
	Parent=2,
	dbBase::nlSing(H,N),
	sx::concat([H,"{",N,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!
;
	H="p3_iSSS",
	dbBase::nlSing(H,N),
	sx::concat([H,"{",N,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!
;
	sx::frontstr(4,H,F,T),
	set::member(F,["[F] ","[P] ","[C] "]),
	dbBase::nlSing(T,Tn),
	sx::concat([H,"{",F,Tn,"}0"],NL),
	rubNLXs(L,2,Lm),
	!
;
	sx::searchstring(H,"|",_),
	dbBase::stNL_F(H,N),
	sx::concat([H,"{",N,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!
;
	sx::searchstring(H,"|",_),
	sx::sxUnpack(H,'|',ST),
	ST=[P,E2,E3],
	dbBase::nlTry(P,Pn),
	dbBase::nlTry(E2,E2n),
	dbBase::nlTry(E3,E3n),
	sx::concat([H,"{",E2n," ",Pn," ",E3n,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!
;
	sx::searchstring(H,"|",_),
	sx::sxUnpack(H,'|',ST),
	ST=[P,E2],
	dbBase::nlTry(P,Pn),
	dbBase::nlTry(E2,E2n),
	sx::concat([H,"{",E2n," ",Pn,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!
;
	sx::searchstring(H,"[",_),
	sx::sxUnpack(H,'[',AB),
	AB=[A,B],
	dbBase::nlSing(A,An),
	sx::concat([H,"{",An,"[",B,"}0"],NL),
	rubNLXs(L,1,Lm),
	!
;
	vars::varComps(H,_,_,D),
	dbBase::nlSing(D,N),
	sx::concat([H,"{",N,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!
;
	opr::oprIntExt(H,N),
	sx::concat([H,"{",N,"}0"],NL),
	rubNLXs(L,-1,Lm),
	!.
rubNLXs([],_,[]) :-
	!.
rubNLXs([H|L],Parent,[Ht|Lt]) :-
	rubNLX(H,Parent,Ht),
	rubNLXs(L,Parent,Lt),
	!.
% ---------------------------------------------------------------------------
% rubNIS_Current
% ---------------------------------------------------------------------------
rubNIS_Current(REW,No,T,Nis,Nas) :-
	guix::txtGet(REW,idc_nis,Snis), 
	sx::sxUnpack(Snis,'|',Lnis),
	sx::sxl_il(Lnis,Nis),
	guix::txtGet(REW,idc_nas,Snas),
	sx::sxUnpackS(Snas,"||",Nas),
	guix::txtGetI(REW,idc_nodeno,No), 
	guix::txtGet(REW,idc_nodetext,T), 
	!.
% ---------------------------------------------------------------------------
% rubNodeDelete
% ---------------------------------------------------------------------------
class predicates
rubDeleteReplace : (sterm,integer_list,string_list,sterm) determ (i,i,i,o).
clauses
% deletes a sub-tree from the rule-tree (if permissible). 
% all deleted nodes except nill, are replaced by nill
% deleted nill's are replaced by atom("__") for later cleanup.
% no clean-up of dummy nill's or redundant logical operators are carried out.
% no clean-up of resulting invalid references.
% 
rubNodeDelete(Rule,No,Nis,Nas,RuleD) :-
	rubSubGet(Rule,No,Sub),
	rubDeleteReplace(Sub,Nis,Nas,Rep),
	rubSubRep(Rule,No,Rep,RuleD),
	!.
rubDeleteReplace(C,Nis,Nas,Cd) :- 
	rubIsNill(C),
	set::member(2,Nis),
	% in conclusion
	Nis=[_,P|_],
	P<>31,
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("__{__}",SNo),
	Cd=atom(A),
	!
;
	rubIsNill(C),
	set::member(2,Nis),
	% not permissible
	!,vpi::alarm(vpiDomains::mesbox_iconerror),fail
;
	rubIsNill(C),
	Nis=[_,P|_],
	set::member(P,[3]),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("__{__}",SNo),
	Cd=atom(A),
	!
;
	rubIsNill(C),
	!,vpi::alarm(vpiDomains::mesbox_iconerror),fail
;
	% right-side literal
	Nis=[42|_],
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	% literal
	Nis=[41|_],
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	% a log.oper under DE in cond
	Nas=[H|_],
	set::member(H,["&","v"]),
	Nis=[_,21,11|_],
	set::member(1,Nis),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("__{__}",SNo),
	Cd=atom(A),
	!
;
	Nas=[H|_],
	set::member(H,["[]","&","v","#","not","true"]),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	Nas=[_,H|_],
	set::member(H,["[]","&","v","#","not","true"]),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	% a call
	Nas=[H|_],
	sx::frontstr(4,H,Hf,_),
	set::member(Hf,["[F] ","[P] ","[C] "]),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	% an oper under DE in cond
	Nas=[H|_],
	opr::opr_any(H,_),
	Nis=[_,21,11|_],
	set::member(1,Nis),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("__{__}",SNo),
	Cd=atom(A),
	!
;
	% an oper under Var in cond
	Nas=[H|_],
	opr::opr_any(H,_),
	Nis=[_,20|_],
	set::member(1,Nis),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	% a compar oper and everything under
	Nas=[H|_],
	opr::oprCmp(H),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	% an arith oper and everything under
	Nas=[H|_],
	opr::opr_expr(H,_),
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	% the entire condition
	Nis=[_,1],
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cd=atom(A),
	!
;
	C=Cd,
	vpi::alarm(vpiDomains::mesbox_iconerror),!,fail.
% ---------------------------------------------------------------------------
% rubNodeGet
% ---------------------------------------------------------------------------
class predicates
rubNodeGetX : (sterm,integer,integer,sterm) determ (i,i,i,o).
rubNodeGetY : (sterml,integer,integer,sterm) determ (i,i,i,o).
clauses
rubNodeGet(Rule,No,Node) :-
	rubNodeGetX(Rule,1,No,Node),!
;
	dlgIF::feil("rubNodeGet",[],[No],Rule),!,fail.	
	
rubNodeGetX(C,N,N,C) :- 
	!.
rubNodeGetX(cmp(_,B),N,No,Node) :-
	Next=N+1,
	rubNodeGetY(B,Next,No,Node),!.
	
rubNodeGetY([H|T],N,No,Node) :-
	rubNodeGetX(H,N,No,Node),!
;
	rubCountNodes(H,L),
	Next=N+L,
	rubNodeGetY(T,Next,No,Node),!.
% ----------------------------------------------------------------------------------------------
% rubOprDom
% ----------------------------------------------------------------------------------------------
rubOprDom(Opr,Dom,OprDom) :-
	opr::oprEq(Opr),
	dom::domShortLong(Doms,Dom),
	sx::concat([Opr,Doms],OprDom),
	!	
;
	OprDom=Opr,
	!.	
% ----------------------------------------------------------------------------------------------
% rubOprType
% ----------------------------------------------------------------------------------------------
rubOprType(":",1,3,3) :- !.
rubOprType("-",1,1,2) :- !.
rubOprType("^",1,2,2) :- !.
rubOprType("/",1,2,2) :- !.
rubOprType("mod",1,2,2) :- !.
rubOprType("div",1,2,2) :- !.
rubOprType("+",2,1,2) :- !.
rubOprType("*",2,2,2) :- !.
% ----------------------------------------------------------------------------------------------
% rubRuleType
% ----------------------------------------------------------------------------------------------
rubRuleType("=>",1,"p3_iRimp") :- !.
rubRuleType("STAT=>",21,"p3_iRsta") :- !.
rubRuleType("CR=>",22,"p3_iRcr") :- !.
rubRuleType("GEN=>",4,"p3_iRgen") :- !.
rubRuleType("UN=>",61,"p3_iRun") :- !.
rubRuleType("MIN=>",62,"p3_iRmin") :- !.
rubRuleType("MAX=>",63,"p3_iRmax") :- !.
rubRuleType("ACC=>",64,"p3_iRacc") :- !.
rubRuleType("IPnum=>",65,"p3_iRip") :- !.
rubRuleType("IPdat=>",68,"p3_iRipxd") :- !.
rubRuleType("RANK=>",66,"p3_iRrnk") :- !.
rubRuleType("CORR=>",67,"p3_iRcor") :- !.
rubRuleTypeSet(R) :- set::member(R,["p3_iRsta","p3_iRcr"]),!.
rubRuleTypeSel(R) :- set::member(R,["p3_iRmin","p3_iRmax","p3_iRun","p3_iRip","p3_iRipxd","p3_iRacc","p3_iRrnk","p3_iRcor"]),!.
% ----------------------------------------------------------------------------------------------
% rubSC_Adj
% ----------------------------------------------------------------------------------------------
rubSC_Adj(SC,SC_Adj) :-
	SC=cmp(ST,
	      [Ea,
	       cmp(_,[]),
	       Eb]),
	SC_Adj=cmp(ST,
	          [Ea,
	           Eb]),!
;
	SC=cmp(ST,
	      [Ea,
	       cmp(_,[])]),
	SC_Adj=cmp(ST,
	          [Ea]),!.
rubSC_Adj(SC,SC) :-
	!.		       
% ----------------------------------------------------------------------------------------------
% rubNodeSetCurr
% ----------------------------------------------------------------------------------------------
% sets given node (No) as current
%
rubNodeSetCurr(W,No) :-
	L=vpiTree::getAllNodes(W),
	SNo=toString(No),
	sx::frontchar(Tail,'{',SNo),
	rubGetNodeText(L,Tail,Name),
	vpiTree::setSelectedNode(W,Name),
	!.
% ---------------------------------------------------------------------------
% rubNodeText
% ---------------------------------------------------------------------------
rubNodeText(REW,No,T) :-
	rubTW_Get(REW,TRW),
	Nodes=vpiTree::getAllNodes(TRW),
	SNo=toString(No),
	sx::frontchar(Tail,'{',SNo),
	rubGetNodeText(Nodes,Tail,Name),
	sx::sxUnpack(Name,'{',L), L=[T|_],!.
% ----------------------------------------------------------------------------------------------
% rubRenumber
% ----------------------------------------------------------------------------------------------
class predicates
rubInvertList : (string_list,string_list) determ (i,o).
rubRenumberLists : (integer_list,integer_list,string_list,string_list,string_list,string_list) determ (i,i,o,o,o,o).
rubRenumberDEs : (string_list,string_list,string_list,string_list,string_list,string_list) determ (i,i,i,o,o,o).
clauses
rubRenumber(DEs,Olds,News,Rule,DEs,Rule) :-
	set::setEq(Olds,News),
	!.
rubRenumber(DEs,Olds,News,Rule,DEsM,RuleM) :-
	rubRenumberLists(Olds,News,Lo,Lo2,Ln,Ln2),
	rubRenumberDEs(DEs,Lo2,Ln2,DEs2,Vs,Vsu),
	set::conc([Lo,Vs],La),
	set::conc([Ln,Vsu],Lb),
	struc::treSubstList(Rule,La,Lb,Rule2),
	rubInvertList(Ln,Lr),
	rubInvertList(Ln2,Lr2),
	rubRenumberDEs(DEs2,Ln2,Lr2,DEsM,_,Vsu2),
	set::conc([Lr,Vsu2],Lc),
	struc::treSubstList(Rule2,Lb,Lc,RuleM),
	!.
rubRenumber(DEs,_,_,Rule,DEs,Rule) :-
	!.
	
rubRenumberLists([],[],[],[],[],[]) :-
	!.
rubRenumberLists([Old|Olds],[New|News],[Old0,Old1,Old2|OldsM],[Old0,Old1,Old2,Old3,Old4,Old5,Old6,Old7,Old8|OldsM2],[New0,New1,New2|NewsM],[New0,New1,New2,New3,New4,New5,New6,New7,New8|NewsM2]) :-
	Old>0,
	Old<>New,
	SOld=toString(Old),
	SNew=toString(New),
	sx::concat(["[",SOld,"0]"],Old0),
	sx::concat(["[",SOld,"1]"],Old1),
	sx::concat(["[",SOld,"2]"],Old2),
	sx::concat(["|",SOld,"0|"],Old3),
	sx::concat(["|",SOld,"1|"],Old4),
	sx::concat(["|",SOld,"2|"],Old5),
	sx::concat(["_",SOld,"0"],Old6),
	sx::concat(["_",SOld,"1"],Old7),
	sx::concat(["_",SOld,"2"],Old8),
	sx::concat(["[-",SNew,"0]"],New0),
	sx::concat(["[-",SNew,"1]"],New1),
	sx::concat(["[-",SNew,"2]"],New2),
	sx::concat(["|-",SNew,"0|"],New3),
	sx::concat(["|-",SNew,"1|"],New4),
	sx::concat(["|-",SNew,"2|"],New5),
	sx::concat(["_",SNew,"0"],New6),
	sx::concat(["_",SNew,"1"],New7),
	sx::concat(["_",SNew,"2"],New8),
	rubRenumberLists(Olds,News,OldsM,OldsM2,NewsM,NewsM2),
	!.
rubRenumberLists([_|Olds],[_|News],OldsM,OldsM2,NewsM,NewsM2) :-
	rubRenumberLists(Olds,News,OldsM,OldsM2,NewsM,NewsM2),
	!.
rubRenumberDEs([],_,_,[],[],[]) :-
	!.
rubRenumberDEs([D|Ds],L,Ls,[Dm|Dsm],[V|Vs],[Vm|Vms]) :-
	sx::sxUnpack(D,'|',L2),
	L2=[_,_,_,V,_],
	sx::sxSubstList(D,L,Ls,Dm),
	D<>Dm,
	vars::varComps(V,E2,_,D2),
	sx::sxUnpack(Dm,'|',L2m),
	L2m=[_,Nm,_,_,_],
	Nmi=sx::toInt(Nm),
	vars::varCompsR(Vm,E2,Nmi,D2),
	rubRenumberDEs(Ds,L,Ls,Dsm,Vs,Vms),
	!.
rubRenumberDEs([D|Ds],L,Ls,[D|Dsm],Vs,Vms) :-
	rubRenumberDEs(Ds,L,Ls,Dsm,Vs,Vms),
	!.
		
rubInvertList([],[]) :-
	!.
rubInvertList([H|T],[Hm|Tm]) :-
	sx::sxRemChar(H,'-',Hm),
	rubInvertList(T,Tm),
	!.
% ----------------------------------------------------------------------------------------------
% rubSimpleVar
% ----------------------------------------------------------------------------------------------
% succeeds if a valid simple Var is selected
%
rubSimpleVar(Var,Dom,Vars,V) :-
	sx::searchstring(Var," /",_),
	sx::sxUnpackS(Var," /",L),
	L=[W,_],
	rubVarExists(W,Vars,V),
	Var=V,
	!
;
	sx::searchstring(Var," /",_),
	sx::sxUnpackS(Var," /",L),
	L=[W,D],
	dom::domShortLongF(D,_),
	not(rubVarExists(W,Vars,_)),
	rubIsVar(Var,_,_),
	V=Var,
	!
;
	not(sx::searchstring(Var," /",_)),
	rubVarExists(Var,Vars,V),
	Var=V,
	!
;	
	not(sx::searchstring(Var," /",_)),
	not(rubVarExists(Var,Vars,_)),
	dom::domShortLongF(SDom,Dom),
	sx::concat([Var," /",SDom],V),
	!.
% ----------------------------------------------------------------------------------------------
% rubST_Nodes
% ----------------------------------------------------------------------------------------------
class predicates
rubSTN : (string,sterml,sterml) determ (i,i,o).
clauses
% returns a list of the DEs in an SC (in order corresponding to the order in the KB predicate)
% 13.4.99
rubST_NodeDEs(cmp(_,[cmp(Ea,Eas)]),DEs) :-
	sx::sxGetInBrack(Ea,'[',']',SNa),
	rubSTN(SNa,Eas,DEs),!.
rubST_NodeDEs(cmp(_,[cmp(Ea,Eas),cmp(Eb,Ebs)]),DEs) :-
	sx::sxGetInBrack(Ea,'[',']',SNa),
	rubSTN(SNa,Eas,DEs2),
	sx::sxGetInBrack(Eb,'[',']',SNb),
	rubSTN(SNb,Ebs,DEs3),
	set::conc([DEs2,DEs3],DEs),!.
rubST_NodeDEs(cmp(_,[cmp(Ea,Eas),cmp(P,Ps),cmp(Eb,Ebs)]),DEs) :-
	sx::sxGetInBrack(Ea,'[',']',SNa),
	rubSTN(SNa,Eas,DEs2),
	sx::sxGetInBrack(P,'[',']',SNp),
	rubSTN(SNp,Ps,DEs4),
	sx::sxGetInBrack(Eb,'[',']',SNb),
	rubSTN(SNb,Ebs,DEs3),
	set::conc([DEs2,DEs3,DEs4],DEs),!.
rubSTN(_,[],[]) :-
	!.
rubSTN(SNo,[D|Ds],[var(DE)|DEs]) :-
	D=var(V),
	sx::concat([V,"[",SNo,"]"],DE),
	rubSTN(SNo,Ds,DEs),!
;	
	D=cmp(V,_),
	sx::concat([V,"[",SNo,"]"],DE),
	rubSTN(SNo,Ds,DEs),!.
% ---------------------------------------------------------------------------
% rubSetNo
% ---------------------------------------------------------------------------
class predicates
rubSetNoX : (string,integer,string) determ (i,i,o).
clauses
% resets a node's number
%
rubSetNo(C,N,Cm) :-
	C=var(V),
	rubSetNoX(V,N,Vm),
	Cm=var(Vm),
	!
;
	C=str(V),
	rubSetNoX(V,N,Vm),
	Cm=str(Vm),
	!
;
	C=atom(V),
	rubSetNoX(V,N,Vm),
	Cm=atom(Vm),
	!.
	
rubSetNoX(S,N,T) :-
	sx::sxUnpack(S,'}',L),
	L=[H|_],
	Ns=toString(N),
	sx::concat([H,"}",Ns],T),
	!.
% ---------------------------------------------------------------------------
% rubSubGet
% ---------------------------------------------------------------------------
class predicates
rubSubGetX : (sterml,integer,sterm) determ (i,i,o).
clauses
% parameters:
% 1	in	rule-tree
% 2	in	node no
% 3	out	sub-tree with node no
%
rubSubGet(T,N,S) :-
	rubNodeNo(T,M),
	M=N,
	S=T,
	!
;
	T=cmp(_,L),
	rubSubGetX(L,N,S),
	!.
	
rubSubGetX([H|T],N,S) :-
	rubSubGet(H,N,S),
	!
;
	rubSubGetX(T,N,S),
	!.
% ---------------------------------------------------------------------------
% rubSubRep
% ---------------------------------------------------------------------------
class predicates
rubSubRepX : (sterml,integer,sterm,sterml) determ (i,i,i,o).
clauses
% parameters:
% 1	in	rule-tree
% 2	in	node no
% 3	in	replacement for node no
% 4	out	replaced rule-tree
%
rubSubRep(T,N,Rep,Tm) :-
	rubNodeNo(T,M),
	M=N,
	Tm=Rep,
	!
;
	T=cmp(H,L),
	rubSubRepX(L,N,Rep,Lm),
	Tm=cmp(H,Lm),
	!.
	
rubSubRepX([H|T],N,Rep,[Hm|T]) :-
	rubSubRep(H,N,Rep,Hm),
	!.
rubSubRepX([H|T],N,Rep,[H|Tm]) :-
	rubSubRepX(T,N,Rep,Tm),
	!.
% ----------------------------------------------------------------------------------------------
% rubTreeClean
% ----------------------------------------------------------------------------------------------
class predicates
rubRemDom : (string_list,string_list,string_list) determ (i,o,o).
rubTC_Del : (sterm,sterm) procedure (i,o).
rubTC_DelX : (sterml,sterml) determ (i,o).
rubTC_Compr : (sterm,sterm) procedure (i,o).
rubTC_ComprX : (sterml,sterml) determ (i,o).
rubTC_Dum : (sterm,sterm) procedure (i,o).
rubTC_Dums : (sterml,sterml) determ (i,o).
rubTC_Noref : (sterm,string_list,string_list,sterm) procedure (i,i,i,o).	
rubTC_NorefX : (sterml,string_list,string_list,sterml) determ (i,i,i,o).	
clauses
% adjusts the rule-tree after a change:
% - deletes all atom("__") nodes
% - deletes dummy operators
% - removes duplicate DEs
% - renumbers SC arguments
% - removes references to deleted DVs and simple vars
% - makes the first occurrence of a simple variable its definition (i.e. with domain indicator and assignment operator),
%   while subsequent left-side occurrences (if any) are made references without domain indicator and using comparison operator.
%
rubTreeClean(Renumber,REW,Rule,RuleM) :-
	Renumber=1,
	rubGetDVs(Rule,DEs,Vars2,ArgVars), 				% gets all DVs and simple vars
	rubRemDom(Vars2,Vars,VarsND),
	set::setRemSub(Vars,VarsND,VarsUndef),
	set::conc([DEs,Vars,ArgVars],AllVars),
	rubTC_Noref(Rule,AllVars,VarsUndef,Rule2), 			% removes references to deleted DVs and simple vars
	rubTC_Del(Rule2,Rule2b), 				% removes deleted nodes
	rubTC_Compr(Rule2b,Rule3), 				% removes deleted nodes
	rubTC_Dum(Rule3,Rule4), 				% removes dummy operator-expressions
	Rule4=cmp(_,[Conds,_]),
  	guix::txtGet(REW,idc_ruletype,SType),
	rubGetNewArgNos(SType,1,Conds,Olds,News,_),
	rubDE_Adjust(DEs,DEsAll,_), 				% adjusts DVs to list-format
	rubRenumber(DEsAll,Olds,News,Rule4,DEsAllM,RuleM), 	% renumbers the arguments
	guix::lbxPop(REW,idc_des,DEsAllM),
	!
;
	Renumber=-1,
	rubGetDVs(Rule,DEs,Vars2,ArgVars), 				% gets all DVs and simple vars
	rubRemDom(Vars2,Vars,VarsND),
	set::setRemSub(Vars,VarsND,VarsUndef),
	set::conc([DEs,Vars,ArgVars],AllVars),
	rubTC_Noref(Rule,AllVars,VarsUndef,Rule2), 			% removes references to deleted DVs and simple vars
	rubTC_Del(Rule2,Rule2b), 				% removes deleted nodes
	rubTC_Compr(Rule2b,Rule3), 				% removes deleted nodes
	rubTC_Dum(Rule3,RuleM), 				% removes dummy operator-expressions
	rubDE_Adjust(DEs,DEsAll,_), 				% adjusts DVs to list-format
	guix::lbxPop(REW,idc_des,DEsAll),
	!.
% 1	in	all simple vars in rule
% 2	out	all defined simple vars in rule without domain indicator
% 3	out	all simple vars without domain indicator
% 
rubRemDom([],[],[]) :-
	!.
rubRemDom([V|Vs],[W|Ws],Us) :-
	sx::searchstring(V," /",_),
	sx::sxUnpackS(V," /",L),
	L=[W|_],
	rubRemDom(Vs,Ws,Us),
	!.
rubRemDom([V|Vs],Ws,[V|Us]) :-
	rubRemDom(Vs,Ws,Us),
	!.	
% deletes all invalid DV and simple var references
%
rubTC_Noref(C,Vars,Undef,Cm) :-
	C=var(V),
	sx::sxUnpack(V,'{',L),
	L=[Var|_],
	set::member(Var,Vars),
	% known variable
	Cm=C,
	!
;
	C=var(_),
	% unknown variable
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cm=atom(A),
	!
;
	C=cmp(V,_),
	sx::sxUnpack(V,'{',L),
	L=[Var|_],
	set::member(Var,Undef),
	% undefined variable
	rubNodeNo(C,No),
	SNo=toString(No),
	A=string::concat("_{_}",SNo),
	Cm=atom(A),
	!
;
	C=cmp(H,L),
	rubTC_NorefX(L,Vars,Undef,Lm),
	Cm=cmp(H,Lm),
	!
;
	Cm=C,
	!.
rubTC_NorefX([],_,_,[]) :-
	!.
rubTC_NorefX([H|T],Vars,Undef,[Hm|Tm]) :-
	rubTC_Noref(H,Vars,Undef,Hm),
	rubTC_NorefX(T,Vars,Undef,Tm),
	!.
	
% deletes all atom("__{__}n") nodes (deleted nodes)
%	
rubTC_Del(cmp(H,B),cmp(H,Bm)) :-	
	rubTC_DelX(B,Bm),!.
rubTC_Del(C,C) :-	
	!.
	
rubTC_DelX([],[]) :- 
	!.
rubTC_DelX([atom(A)|T],Tm) :-
	sx::frontstr(3,A,F,_),
	F="__{",
	rubTC_DelX(T,Tm),
	!.	
rubTC_DelX([H|T],[Hm|Tm]) :-
	rubTC_Del(H,Hm),
	rubTC_DelX(T,Tm),
	!.	
% changes cmp(X,[]) to var(X)
%	
rubTC_Compr(cmp(ListHead,[]),cmp(ListHead,[])) :-	
	sx::frontstr(2,ListHead,F,_),
	F="[]",
	!.
rubTC_Compr(cmp(H,[]),var(H)) :-	
	!.
rubTC_Compr(cmp(H,B),cmp(H,Bm)) :-	
	rubTC_ComprX(B,Bm),!.
rubTC_Compr(C,C) :-	
	!.
	
rubTC_ComprX([],[]) :- 
	!.
rubTC_ComprX([H|T],[Hm|Tm]) :-
	rubTC_Compr(H,Hm),
	rubTC_ComprX(T,Tm),
	!.	
% deletes all dummy operators
%
rubTC_Dum(cmp(Opr,[B]),B) :-	
	sx::sxUnpack(Opr,'{',L),
	L=[Op|_],
	set::member(Op,["&","v","+","*"]),!.
rubTC_Dum(cmp(H,T),cmp(H,Tm)) :-
	rubTC_Dums(T,Tm),!.
rubTC_Dum(C,C) :-
	!.		
rubTC_Dums([],[]) :-
	!.
rubTC_Dums([H|T],[Hm|Tm]) :-
	rubTC_Dum(H,Hm),
	rubTC_Dums(T,Tm),!.
% ----------------------------------------------------------------------------------------------
% rubTreeGetCurr
% ----------------------------------------------------------------------------------------------
class predicates
rubTreeGetCurrX : (integer,integer,sterm,sterm) determ (i,i,i,o).
rubTreeGetCurrY : (integer,integer,sterml,sterm) determ (i,i,i,o).
clauses
rubTreeGetCurr(Rule,REW,SubTree) :-
	guix::txtGet(REW,idc_nodeno,SNo),
	No=sx::toInt(SNo),
	rubTreeGetCurrX(1,No,Rule,SubTree),
	!.
	
rubTreeGetCurrX(N,N,T,T) :-
	!.
rubTreeGetCurrX(N,No,cmp(_,L),S) :-	
	Next=N+1,	
	rubTreeGetCurrY(Next,No,L,S),!.
rubTreeGetCurrX(N,No,cmp(_,L),S) :-	
	Next=N+1,	
	rubTreeGetCurrY(Next,No,L,S),!.
rubTreeGetCurrY(N,No,[H|_],S) :-
	rubTreeGetCurrX(N,No,H,S),!.
rubTreeGetCurrY(N,No,[H|T],S) :-
	rubCountNodes(H,L),
	Next=N+L,
	rubTreeGetCurrY(Next,No,T,S),!.	
% ----------------------------------------------------------------------------------------------
% rubTW_Get;rubEW_Get
% ----------------------------------------------------------------------------------------------
rubTW_Get(EW,TW) :- guix::getWin(EW,No), N=No+1, guix::getWin(TW,N),!. % rule ed win -> tree win
rubEW_Get(TW,EW) :- guix::getWin(TW,No), N=No-1, guix::getWin(EW,N),!.  % tree win -> rule ed win
% ----------------------------------------------------------------------------------------------
% rubValidate
% ----------------------------------------------------------------------------------------------
class predicates
rubVal2 : (sterm) determ (i).
clauses
rubValidate(Rule,RuleV) :-
	struc::treElimDummyNIL(Rule,Rule2),
	struc::treElimDummyLogOpr(Rule2,RuleV),
	rubVal2(RuleV),
	%rubVariableDoms(RuleV,DVars),
	%rubCheckDomCompat(DVars,RuleV),
	!.
rubVal2(Rule) :-
	struc::treHasNIL(Rule),
	dlgIF::feil("rubVal2",["Rule contains empty element(s)"],[],nill),
	!,fail.	
rubVal2(_) :-
	!.	
% ---------------------------------------------------------------------------
% rubValidateInsNIL
% ---------------------------------------------------------------------------
rubValidateInsNIL(Nis) :- 
	Nis=[_,P|_],
	set::member(P,[3,29,45]),
	!.
% ---------------------------------------------------------------------------
% rubVarHasRef
% ---------------------------------------------------------------------------
class predicates
rubVarHasRefX : (string,sterml) determ (i,i).
clauses
rubVarHasRef(V,var(V)) :-
	!.
rubVarHasRef(V,cmp(_,L)) :-
	rubVarHasRefX(V,L),!.
	
rubVarHasRefX(V,[H|T]) :-
	rubVarHasRef(V,H),!
;
	rubVarHasRefX(V,T),!.
% ---------------------------------------------------------------------------
% rubDE_Var
% ---------------------------------------------------------------------------
% parameters:
% 1	in	argument 
% 2	in	suffix no
% 3	in	DE
% 4	out	DE variable
%
rubDE_Var(E,No,DE,Var) :-
    string::front(E,1,First,Tail),
    UFirst=string::toUpperCase(First),
	SNo=toString(No),
	sx::concat([UFirst,Tail,"_",SNo,DE],Var),
	!.
% ---------------------------------------------------------------------------
% rubVarExists
% ---------------------------------------------------------------------------
rubVarExists(V,[Var|Vars],Vr) :-
	sx::sxUnpackS(Var," /",L),
	L=[W|_],
	V=W,
	Vr=Var,
	!
;
	rubVarExists(V,Vars,Vr),
	!.
% ---------------------------------------------------------------------------
% rubVariables
% ---------------------------------------------------------------------------
class predicates
rubVariablesX : (sterm,integer,integer,string_list) procedure (i,i,i,o).
rubVariablesY : (sterml,integer,integer,string_list) determ (i,i,i,o).
clauses
%
% returns a list of all variables (without domain indicator) in the tree up to and including 
% given node no (node no 0 gives a list of all variables). 
%
rubVariables(cmp(_,[Cond,_]),0,Vars) :-
	rubVariablesX(Cond,2,999999,Vars2),
	set::distinct(Vars2,Vars),
	!.
rubVariables(cmp(_,[Cond,_]),Max,Vars) :-
	N=Max-1,
	rubVariablesX(Cond,2,N,Vars2),
	set::distinct(Vars2,Vars),
	!.
	
rubVariablesX(_,N,Max,[]) :-
	N>Max,
	!.
rubVariablesX(cmp(H,L),N,Max,Vars) :-
	sx::sxUnpack(H,'{',T),
	T=[LogOpr,_],
	set::member(LogOpr,["&","v"]),
	Next=N+1,
	rubVariablesY(L,Next,Max,Vars),
	!.
rubVariablesX(cmp(H,_),_,_,[W]) :-
	sx::sxUnpack(H,'{',T),
	T=[V,_],
	rubIsVar(V,W,_),
	!.
rubVariablesX(_,_,_,[]) :-
	!.
	
rubVariablesY(_,N,Max,[]) :-
	N>Max,
	!.
rubVariablesY([],_,_,[]) :-
	!.
rubVariablesY([H|T],N,Max,Vars) :-
	rubVariablesX(H,N,Max,Vars2),
	rubCountNodes(H,L),
	Next=N+L,
	rubVariablesY(T,Next,Max,Vars3),
	set::conc([Vars2,Vars3],Vars),
	!.	
end implement rub
