/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement ruled
open core,termDomain,resourceIdentifiers,globalDomains

constants
className = "com/rudstrand/ruled/ruled".
classVersion = "1.0".

class predicates
rtrInsIF : (integer,string,integer,integer,string_list) determ (i,i,i,i,i).

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% rtrAddDEs
% ----------------------------------------------------------------------------------------------
rtrAddDEs(REW,T) :-
	rt::rcnFixRule(-1,T,_,_,DEs),
  	rub::rubDE_Adjust(DEs,DEs2,_),
	guix::lbxGetAll(REW,idc_des,L),
	set::conc([DEs2,L],L2),
	set::distinct(L2,L3),
	guix::lbxPop(REW,idc_des,L3),
	!.
% ----------------------------------------------------------------------------------------------
% rtrCheckReplace
% ----------------------------------------------------------------------------------------------
rtrCheckReplace(Realm,SNo,SSubNo) :-
	No=sx::toInt(SNo),
	SubNo=sx::toInt(SSubNo),
    L=dbBase::getSr(Realm,No,SubNo),
    set::size(L,N),
    N>0,
	!,sx::concat(["Sentence rule record ",Realm,"|",SNo,"|",SSubNo," exists already"],S1),
	Reply=vpiCommonDialogs::ask(S1,"Do you want to replace?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	dbBase::srDelete(Realm,No,SubNo),
	!
;
	!.
% ----------------------------------------------------------------------------------------------
% rtrCompDom
% ----------------------------------------------------------------------------------------------
% returns currently selected domain
%
rtrCompDom(REW,Dom) :-
	guix::lbtGet(REW,idc_dom,DomNL),
	dbBase::kbID(DomNL,Dom,_,_),
	!.
% ----------------------------------------------------------------------------------------------
% rtrGetDE_Var
% ----------------------------------------------------------------------------------------------
% ARGUMENT | NO | DE | DE_VAR | DENL[NO] or 
%
rtrGetDE_Var(Vn,[],_) :-
	dlgIF::feil("rtrGetDE_Var",["Cannot find DE variable",Vn],[],nill),!,fail.
rtrGetDE_Var(Vn,[DE|DEs],V) :-
	sx::sxUnpack(DE,'|',L),
	L=[_,_,_,V,Dn],
	Vn=Dn,
	!
/*
;
	sx::sxUnpack(DE,'|',L),
	L=[_,_,_,_,V,Dn],
	Vn=Dn,
	!
*/	
;
	rtrGetDE_Var(Vn,DEs,V),
	!.
% ----------------------------------------------------------------------------------------------
% rtrIF
% ----------------------------------------------------------------------------------------------
class predicates
rtrIFs : (sterml,sterml) determ (i,o).
clauses
% removes NL from a rule tree, such that textual information is changed from kk..k{nn..n (or kk..k{nn..n}99) to kk..k 
% where kk..k is a KB name while nn..n is the corresponding syntagm
% 
rtrIF(C,nill) :-
	rub::rubIsNill(C),
	!.
rtrIF(fint(A,_),int(A)) :-
	!.
rtrIF(ffix(A,_),fix(A)) :-
	!.
rtrIF(atom(N),atom(K)) :-
	sx::sxUnpack(N,'{',L),
	L=[K,_],
	!.
rtrIF(var(N),var(K)) :-
	sx::sxUnpack(N,'{',L),
	L=[K,_],
	!.
rtrIF(str(N),str(K)) :-
	sx::sxUnpack(N,'{',L),
	L=[K,_],
	!.
rtrIF(cmp(H,T),cmp(K,Tk)) :-
	sx::sxUnpack(H,'{',L),
	L=[K,_],
	rtrIFs(T,Tk),
	!
;
	K=H,	
	rtrIFs(T,Tk),
	!.
rtrIFs([],[]) :-
	!.
rtrIFs([H|L],[Ht|Lt]) :-
	rtrIF(H,Ht),
	rtrIFs(L,Lt),
	!.
% ----------------------------------------------------------------------------------------------
% rtrInsIF
% ----------------------------------------------------------------------------------------------
/* 16.6.6
class predicates
rtrInsIFx : (integer,string,integer,integer,string_list) determ (i,i,i,i,i).
clauses

rtrInsIF(RuleType,Realm,No,SubNo,PRs) :-	
	set::size(PRs,N),
	N=1,
	rtrInsIFx(RuleType,Realm,No,SubNo,PRs),
	!
;
	rtrInsIFx(RuleType,Realm,No,SubNo,PRs),
	!.
*/
rtrInsIF(_,_,_,_,[]) :-	
	!.
rtrInsIF(RuleType,Realm,No,SubNo,[PRm|PRs]) :-
	PR=string::concat(PRm,"."),
	transpr::proPR_Term(PR,PR_Term),
                    tt::tt_term(1,"(Sub-)Rule",PR_Term),
	Status=globalVars::get_i("rulsta"),
	RecType=Status*3,
	PR_Term=cmp(_,[LT,RT]),
    dbBase::recordInsert(RecType,Realm,No,SubNo,LT,RT),
                    SNo=toString(No),
                    SSubNo=toString(SubNo),
                    sx::concat(["Rule ",Realm,"|",SNo,"|",SSubNo," is stored in the KB."],Mess),
                    dlgIF::msgS(Mess),
	Next=SubNo+1,
	rtrInsIF(RuleType,Realm,No,Next,PRs),
	!.
% ----------------------------------------------------------------------------------------------
% rtrMakeTree
% ----------------------------------------------------------------------------------------------
class predicates
rtrMakeTreeXs : (sterml,integer,integer,vpiTree::tree_list,sterml) determ (i,i,o,o,o).
rtrMakeTreeX : (sterm,integer,integer,vpiTree::tree,sterm) determ (i,i,o,o,o).
rtrNodeName : (string,integer,string) procedure (i,i,o).
clauses
% converts from a sterm-tree (rule tree) to a tree-package tree (both NL format)
% parameters:
% 1	in	rule-tree NL
% 2	out	tree (corr. to par#1)
% 3	out	rule-tree NL renumbered	
%
rtrMakeTree(Rule,Tree,RuleM) :-
	rtrMakeTreeX(Rule,1,_,Tree,RuleM),
	!.
rtrMakeTreeX(C,N,Nr,vpiTree::tree(S,vpiTree::unmarked,[],1),Cm) :-
	rub::rubIsNill(C),	
	rtrNodeName("_",N,S),
	rub::rubSetNo(C,N,Cm),
	Nr=N+1,
	!.
rtrMakeTreeX(ffix(V,_),N,Nr,vpiTree::tree(S,vpiTree::unmarked,[],1),ffix(V,Ns)) :-
	Vs=toString(V),
	rtrNodeName(Vs,N,S),
	Nr=N+1,
	Ns=toString(N),
	!.
rtrMakeTreeX(fint(V,_),N,Nr,vpiTree::tree(S,vpiTree::unmarked,[],1),fint(V,Ns)) :-
	Vs=toString(V),
	rtrNodeName(Vs,N,S),
	Nr=N+1,
	Ns=toString(N),
	!.
rtrMakeTreeX(atom(A),N,Nr,vpiTree::tree(S,vpiTree::unmarked,[],1),Cm) :-
	rtrNodeName(A,N,S),
	rub::rubSetNo(atom(A),N,Cm),
	Nr=N+1,
	!.
rtrMakeTreeX(var(V),N,Nr,vpiTree::tree(S,vpiTree::unmarked,[],1),Cm) :-
	rtrNodeName(V,N,S),
	rub::rubSetNo(var(V),N,Cm),
	Nr=N+1,
	!.
rtrMakeTreeX(str(A),N,Nr,vpiTree::tree(S,vpiTree::unmarked,[],1),Cm) :-
	rtrNodeName(A,N,S),
	rub::rubSetNo(str(A),N,Cm),
	Nr=N+1,
	!.
rtrMakeTreeX(C,N,Nr,vpiTree::tree(S,vpiTree::unmarked,Lt,1),Cm) :-
	C=cmp(H,L),
	sx::sxUnpack(H,'}',H2),
	H2=[H3|_],
	rtrNodeName(H,N,S),
	Ns=toString(N),
	sx::concat([H3,"}",Ns],Hm),
	Next=N+1,
	rtrMakeTreeXs(L,Next,Nr,Lt,Lm),
	Cm=cmp(Hm,Lm),
	!.
rtrMakeTreeX(C,_,_,_,_) :-
	dlgIF::feil("rtrMakeTreeX",[],[],C),!,fail.
rtrMakeTreeXs([],N,N,[],[]) :-
	!.
rtrMakeTreeXs([H|L],N,Nr,[Ht|Lt],[Hm|Lm]) :-
	rtrMakeTreeX(H,N,N2,Ht,Hm),
	rtrMakeTreeXs(L,N2,Nr,Lt,Lm),
	!.
rtrNodeName(A,N,S) :-
	sx::sxGetInBrack(A,'{','}',NL),
	Ns=toString(N),
	sx::concat([NL,"{",Ns],S),
	!
;
	Ns=toString(N),
	sx::concat([A,"{",Ns],S),
	!.
% ----------------------------------------------------------------------------------------------
% rtrSaveXML
% ----------------------------------------------------------------------------------------------
class predicates
rtrSaveXML_X : (integer,string).
clauses
% parameters:
% 2	in	-1/1 open text window (after saving)
%
rtrSaveXML(Rule,Win,OpWin) :-
	vpi::cursorSetWait(),
  	guix::txtGet(Win,idc_realm,Realm),
  	guix::txtGet(Win,idc_sno,SNo),
	rtrIF(Rule,Rule2),
	Rule2=cmp(Typ,_),
	rub::rubRuleType(Typ,Type,_),
	SType=toString(Type),
	rub::rubTreeClean(-1,Win,Rule2,Rule3),	
				tt::tt_term(2,"Saved rule",Rule3),
	xmlP3::xmlSentenceRule(SType,Realm,SNo,"0",Rule3,X3),
	xmlP3::xmlKnowledge(X3,X4),	
	Path=globalVars::get_s("std_path"),
	sx::concat([Path,"\\",Realm,"-",SNo,".xml"],OutFile),
	file::writeString(OutFile,X4),	
	rtrSaveXML_X(OpWin,OutFile),
	!.
	
rtrSaveXML_X(OpWin,OutFile) :-
	OpWin=1,
	editorWindow::openFiles([OutFile]),
	!
;
	!.
% ----------------------------------------------------------------------------------------------
% isAcyclic
% ----------------------------------------------------------------------------------------------
isAcyclic(Rule,Win) :-
	rtrIF(Rule,Rule2),
	transrt::rlgPR(Rule2,PRs),
	set::size(PRs,N),
	guix::txtGetI(Win,idc_sbno,Sub),
	rtrSubNo(N,Sub,SubNo),
	guix::txtGet(Win,idc_realm,Realm),
	guix::txtGetI(Win,idc_sno,No),
	PRs=[PRm|_],
	PR=string::concat(PRm,"."),
	transpr::proPR_Term(PR,PR_Term),
	PR_Term=cmp(_,[LT,RT]),
	LT=cmp(PID,_),	
	SSent=toString(No),
	SSub=toString(SubNo),
	sx::sxlJoin([PID,Realm,SSent,SSub],"|",RuleName),
	dbExtension::updateDependencies(1,3,LT,RT,RuleName,Acyclic),
	dbExtension::updateDependencies(3,3,LT,RT,RuleName,_), % remove from graph since this was only a check for acyclicity
	Acyclic=0,
    dlgIF::dlgNote(0,"Rule has cyclic knowledge dependencies and cannot be stored!"),
    !,fail
;    
    !.
% ----------------------------------------------------------------------------------------------
% rtrSaveKB
% ----------------------------------------------------------------------------------------------
class predicates
rtrSubNo : (integer,integer,integer) procedure (i,i,o).
clauses

rtrSaveKB(Rule,Win) :-	
			tt::tt_term(2,"Rule (editor)",Rule),
	rtrIF(Rule,Rule2),
	transrt::rlgPR(Rule2,PRs),
	set::size(PRs,N),
	guix::txtGetI(Win,idc_sbno,Sub),
	rtrSubNo(N,Sub,SubNo),
	guix::txtGet(Win,idc_realm,Realm),
	guix::txtGetI(Win,idc_sno,No),
  	guix::txtGet(Win,idc_ruletype,SType),
	RuleType=sx::toInt(SType),
	rtrInsIF(RuleType,Realm,No,SubNo,PRs),
	vpi::cursorSet(Win,vpiDomains::cursor_Arrow),
			globalVars::get_i("error_flag")=0,
			Ns=toString(N),
			sx::concat(["Rule is stored in ",Ns," KB records."],Mess),
	dlgIF::dlgNote(0,Mess),
	!
;
    dlgIF::dlgNote(1,"Rule NOT stored - existing rule with same name (if any) has been deleted!"),
    !.
	
rtrSubNo(_,N,N) :- N>0,!.
rtrSubNo(1,_,0) :- !.
rtrSubNo(_,_,1) :- !.
% ----------------------------------------------------------------------------------------------
% rtrPopIds
% ----------------------------------------------------------------------------------------------
rtrPopIds(IfsSn,IfsK) :-
	globalVars::get_i("idfpop")=-1, % populate with all syntagms
	Lang=globalVars::get_s("lang"),
	IfsA=pop::getIdentifiers(Lang,1,[]),
	IfsS=pop::getIdentifiers(Lang,1,["Pro/3"]),
	set::setRemSub(IfsS,IfsA,IfsK2),
	dbBase::nlSings(IfsK2,IfsK),
	dbBase::nlSings(IfsS,IfsSn),
	!
;	
	Lang=globalVars::get_s("lang"),
	IfsA=pop::getIdentifiers(Lang,1,[]),
	IfsS=pop::getIdentifiers(Lang,1,["Pro/3"]),
	set::setRemSub(IfsS,IfsA,IfsK2),
	dbBase::nlSings(IfsK2,IfsK),
	dbBase::nlSings(IfsS,IfsSn),
	!.
% ----------------------------------------------------------------------------------------------
% redGetST
% ---------------------------------------------------------------------------
redGetST(REW,Card,Pk,Eak,Ebk,Va,Vb) :-
	% E - P - E
	guix::txtGet(REW,idc_ledit,ST),
	guix::lbxGetAll(REW,idc_unary_preds,AllST),
	set::member(ST,AllST),
	dbBase::stNL_KB(ST,K),
	dbBase::stComps(K,Card,Pk,Eak,Ebk),
	dbBase::smClassDef(Pk,"p",Card2,_,_,_,_),		
	Card=Card2,
	Va="",
	Vb="",
	!
;
	guix::txtGet(REW,idc_ledit,ST),
	guix::lbxGetAll(REW,idc_unary_preds,AllST),
	set::member(ST,AllST),
	dbBase::stNL_KB(ST,K),
	dbBase::stComps(K,Card,Pk,Eak,Ebk),
	dbBase::smClassDef(Pk,"p",Card2,_,_,_,_),		
	Card<>Card2,
	dlgIF::dlgNote(0,"Mismatch between the predicate type's definition and the number of arguments"),
	!,fail
;	
	% E - P or E - P - E
	guix::txtGet(REW,idc_ledit,ST),
	dbBase::stNL_KB(ST,K),
	dbBase::stComps(K,Card,Pk,Eak,Ebk),
	dbBase::smClassDef(Pk,"p",Card2,_,_,_,_),		
	Card=Card2,
	Va="",
	Vb="",
	!
;
	guix::txtGet(REW,idc_ledit,ST),
	dbBase::stNL_KB(ST,K),
	dbBase::stComps(K,Card,Pk,Eak,Ebk),
	dbBase::smClassDef(Pk,"p",Card2,_,_,_,_),		
	Card<>Card2,
	dlgIF::dlgNote(0,"Mismatch between the predicate type's definition and the number of arguments"),
	!,fail
;
	% _ - P - _
	guix::txtGet(REW,idc_ledit,ST),
	transnl::scnNL(ST,Ws),
	Ws=[_,pie_scan::w(_,Pk,"p3_iPT",_),_],
	dbBase::smClassDef(Pk,"p",Card,_,_,_,_),		
	Card=1,
	dlgIF::dlgNote(0,"Mismatch between the predicate type's definition and the number of arguments"),
	!,fail
;
	% _ - P
	guix::txtGet(REW,idc_ledit,ST),
	transnl::scnNL(ST,Ws),
	Ws=[_,pie_scan::w(_,Pk,"p3_iPT",_)],
	dbBase::smClassDef(Pk,"p",Card,_,_,_,_),		
	Card=2,
	dlgIF::dlgNote(0,"Mismatch between the predicate type's definition and the number of arguments"),
	!,fail
;
	% V - P - E
	guix::txtGet(REW,idc_ledit,ST),
	transnl::scnNL(ST,Ws),
	Ws=[pie_scan::w(_,Va,"p3_dvar",_),pie_scan::w(_,Pk,"p3_iPT",_),pie_scan::w(_,Ebk,"p3_iET",_)],
	dbBase::smClassDef(Pk,"p",Card,_,_,_,_),		
	Card=2,
	Vb="",
	Eak="",
	!
;
	% E - P - V
	guix::txtGet(REW,idc_ledit,ST),
	transnl::scnNL(ST,Ws),
	Ws=[pie_scan::w(_,Eak,"p3_iET",_),pie_scan::w(_,Pk,"p3_iPT",_),pie_scan::w(_,Vb,"p3_dvar",_)],
	Card=2,
	Va="",
	Ebk="",
	!
;
	% V - P - V
	guix::txtGet(REW,idc_ledit,ST),
	transnl::scnNL(ST,Ws),
	Ws=[pie_scan::w(_,Va,"p3_dvar",_),pie_scan::w(_,Pk,"p3_iPT",_),pie_scan::w(_,Vb,"p3_dvar",_)],
	Card=2,
	Eak="",
	Ebk="",
	!
;
	% V - P
	guix::txtGet(REW,idc_ledit,ST),
	transnl::scnNL(ST,Ws),
	Ws=[pie_scan::w(_,Va,"p3_dvar",_),pie_scan::w(_,Pk,"p3_iPT",_)],
	Card=1,
	Eak="",
	Ebk="",
	Vb="",
	!.
% ----------------------------------------------------------------------------------------------
% rdgRuleTypeDisable
% ----------------------------------------------------------------------------------------------
/*
ref. rubRuleType
*/
rdgRuleTypeDisable(REW,N) :-
	set::member(N,[1,4,7]),
	guix::ctlInvisible(REW,[idc_ctrl,idc_procedure]),
	!
;
	N=22,
	guix::ctlInvisible(REW,[idc_procedure]),
	!
;
	N=21,
	Procs=["p3_rCnt","p3_rSum","p3_rAvg","p3_rMea","p3_rSdv","p3_rMin","p3_rMax","p3_rStv","p3_rEtv","p3_rStim","p3_rEtim","p3_rMit","p3_rMat","p3_rTS"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=61,
	Procs=["p3_rUn","p3_rAssg"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=62,
	Procs=["p3_rMin","p3_rAssg"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=63,
	Procs=["p3_rMax","p3_rAssg"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=64,
	Procs=["p3_rAcc"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=65,
	Procs=["p3_rCip","p3_rLow","p3_rMnVa","p3_rUpp","p3_rDf"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=66,
	Procs=["p3_rRank","p3_rRnks","p3_rSort","p3_rAssg"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=67,
	Procs=["p3_rCorr","p3_rAssg"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	N=68,
	Procs=["p3_rCipxd","p3_rLow","p3_rMnVa","p3_rUpp","p3_rDf"],
	dbBase::nlSings(Procs,L),
	guix::lbxPop(REW,idc_procs,L),
	guix::ctlInvisible(REW,[idc_ctrl]),
	!
;
	dlgIF::feil("rdgRuleTypeDisable",["Unknown rule type"],[N],nill),!,fail.
end implement ruled
