/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement sdb
open core,termDomain,resourceIdentifiers

class facts

sysdb : chainDB := erroneous. % system db
btrsel : (char Ix,string BTreeName,chainDB::bt_selector Selector). % Ix - see dbaBTC for codes

constants

className = "com/rudstrand/sdb/sdb".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% --------------------------------------------------------------------------
% get
% --------------------------------------------------------------------------
get()=KB :- KB=sysdb, !.
% --------------------------------------------------------------------------
% sdbClose
% --------------------------------------------------------------------------
class predicates
sdbCloseSIx : ().
clauses
% closes and destroys the database
%
sdbClose () :-
	globalVars::get_i("sdbopen")=0,
	!
;
	globalVars::get_i("sdbopen")=1,
	sdbCloseSIx,
	sysdb:db_close,
	globalVars::set_i("sdbopen",0),
		dlgIF::msgS("SDB deleted."),!
;
	globalVars::get_i("sdbopen")=2,
	sdbCloseSIx,
	sysdb:db_close,
	globalVars::set_i("sdbopen",0),
		dlgIF::msgS("SDB deleted."),!
;
	dlgIF::feil("sdbClose",[],[],nill),!,fail.	

sdbCloseSIx () :-
	btrsel('S',Chain,SN),
	sysdb:bt_close(SN),
	sysdb:bt_delete(Chain),
	retract(btrsel('S',Chain,SN)),
	fail
;
	!.					
% --------------------------------------------------------------------------
% getIx/dropIx
% --------------------------------------------------------------------------
getIx(Name)=Selector:- btrsel('S',Name,Selector),!.
dropIx(Name) :- retractall(btrsel(_,Name,_)),!.
dropIx(Name,Selector) :- retractall(btrsel('S',Name,Selector)),!.
putIx(Name,Selector) :- assert(btrsel('S',Name,Selector)),!.
% --------------------------------------------------------------------------
% sdbOpen
% --------------------------------------------------------------------------
class predicates
sdbOpenX : () determ.
clauses
% opens (creates, opens and initializes) the database
%
sdbOpen () :-
	globalVars::get_i("sdbopen")>0,
	!
;	
	globalVars::get_i("sdbram")=2,
	sysdb:=chainDB::db_create("SDB",chainDB::in_memory),
	globalVars::set_i("sdbopen",1),
	sdbOpenX,
	!
;
	globalVars::get_i("sdbram")=1,
	Reply=vpiCommonDialogs::ask("Dynamic loading of work DB","Do you want to locate the work DB in RAM?",["Yes","No"]),
	Reply=vpiCommonDialogs::resp_default,
	sysdb:=chainDB::db_create("SDB",chainDB::in_memory),
	globalVars::set_i("sdbopen",1),
	sdbOpenX,
		dlgIF::msgS("SDB initialized in RAM."),
	!	
;	
	sysdb:=chainDB::db_create("SDB",chainDB::in_file),
	globalVars::set_i("sdbopen",2),
	sdbOpenX,
		dlgIF::msgS("SDB initialized on disk."),
	!.
sdbOpenX () :-
	sysdb:db_reuserefs(1),
	% create standard indexes
	sysdb:bt_create("H1",H1,8,4,0),
	sysdb:bt_create("V1",V1,14,4,0),
	L1=globalVars::get_i("max_kbn"),
	LV2=L1*2+12,
	sysdb:bt_create("V2",V2,LV2,4,0),
	sysdb:bt_create("V3",V3,8,4,1),
	sysdb:bt_create("TM1",TM1,8,4,0),
	sysdb:bt_create("TH1",TH1,14,4,0),
	LTH2=L1*3+11,
	sysdb:bt_create("TH2",TH2,LTH2,4,0),
	sysdb:bt_create("TA1",TA1,20,4,0),
	sysdb:bt_create("TD1",TD1,26,4,0),
	sysdb:bt_create("TD2",TD2,20,4,0),
	%
	retractall(btrsel('S',_,_)),
	assert(btrsel('S',"H1",H1)),
	assert(btrsel('S',"V1",V1)),
	assert(btrsel('S',"V2",V2)),
	assert(btrsel('S',"V3",V3)),
	assert(btrsel('S',"TM1",TM1)),
	assert(btrsel('S',"TH1",TH1)),
	assert(btrsel('S',"TH2",TH2)),
	assert(btrsel('S',"TA1",TA1)),
	assert(btrsel('S',"TD1",TD1)),
	assert(btrsel('S',"TD2",TD2)),
	sdbSetStatus,
	!		
;
	dlgIF::feil("sdbOpenX",[],[],nill),!,fail.	
% ------------------------------------------------------------------------------------------------------------
% sdbSetStatus
% ------------------------------------------------------------------------------------------------------------
sdbSetStatus () :-
	globalVars::get_i("knowopen")>0,
	/*
    tt::freeMemory(Af,Bf,Cf,Df,MAf,MBf,MCf,MDf),
    sx::fixFormatR(Af,1,0,As),
    sx::fixFormatR(Bf,1,0,Bs),
    sx::fixFormatR(Cf,1,0,Cs),
    sx::fixFormatR(Df,1,0,Ds),
    sx::fixFormatR(MAf,1,0,MAs),
    sx::fixFormatR(MBf,1,0,MBs),
    sx::fixFormatR(MCf,1,0,MCs),
    sx::fixFormatR(MDf,1,0,MDs),
	A=string::concat("Global: ",As,"% (",MAs,")"), 
	B=string::concat("Heap: ",Bs,"% (",MBs,")"), 
	C=string::concat("Stack: ",Cs,"% (",MCs,")"), 
	D=string::concat("Trail: ",Ds,"%  (",MDs,")"),
	W=guix::getWinOfClas(4),
	guix::lbxPop(W,idc_flags,["Free memory:",A,B,C,D]), 
	*/
	dlgIF::dbaStatuses,
	!
;
	dlgIF::dbaStatuses,
	!.	
% ------------------------------------------------------------------------------------------------------------
% tabName
% ------------------------------------------------------------------------------------------------------------
tabName(SolRef,Table,Name) :-
	S1=toString(SolRef),
	S2=toString(Table),
	sx::concat(["S",S1,"T",S2],Name),
	!.
end implement sdb
