/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement sqlBase
open core,termDomain,globalDomains

constants
className = "com/rudstrand/sqlBase/sqlBase".
classVersion = "1.0".

class facts
con : odbcConnection:=erroneous.
opened : integer:=0. % connection opened 0/1
db : string:="". % SQL db name - blank if none open
kb : string:="". % KB name - blank if none open
mainTable : string:="".
sqlVersion : integer := 0. % KB  version
started : integer := 0. % 0/1 SQL transaction start
chainRef : (string Chain,integer LastCount,string TableName) nondeterm. % chainRef exists only if the chain exists (has records)


parTable : string:="".
% stored in parameter table (see schema):
chain_prefix : string := "c_". 
st_prefix : string := "s_". 
name_case : integer := 2. 
use_underscore : string := "_".
max_length : integer := 64. 
st_db : string := "". 
term_length : integer := 10000. 
term_format : integer := 1.

class predicates
checkLength : (string) determ.
connectionError : () determ.
createMainTable : (string).   
createParTable : (string).   
createChainTableIfNotExists : (string) determ.
createSysTables : (string).
dbError : () determ.
deleteSelectedChainRecords : (sqlSelect,string Chain,integer N).
dropChainTable : (string) determ.
insertParTable : () determ.
validateChainRecord : (sqlSelect,string,string) determ.

clauses
classInfo(className, classVersion).
% -------------------------------------------------------------------------------------------------------------
% 
% -------------------------------------------------------------------------------------------------------------
getMainTable()=T :- T=mainTable,!.

setMaxLength(Max) :- 
    term_length:=Max,
    !.
getMaxLength()=Max :- Max=term_length,!.
% -------------------------------------------------------------------------------------------------------------
% deleteKb
% -------------------------------------------------------------------------------------------------------------
deleteKb() :-
    Sql1=string::concat("drop table if exists ",mainTable),
    sqlTools::executeDirect(con,Sql1),
    Sql2=string::concat("drop table if exists ",parTable),
    sqlTools::executeDirect(con,Sql2),
    chainRef(_,_,Table),
        Sql3=string::concat("drop table if exists ",Table),
        sqlTools::executeDirect(con,Sql3),
        fail
;
    retractall(chainRef(_,_,_)),
    !.
% -------------------------------------------------------------------------------------------------------------
% connectionError
% -------------------------------------------------------------------------------------------------------------
connectionError() :-
    opened=0,
    dlgIF::dlgNote(3,"No SQL connection open!"),
    !
;
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% hasConnection
% -------------------------------------------------------------------------------------------------------------
hasConnection() :-
    opened=1,
    !.
% -------------------------------------------------------------------------------------------------------------
% get/setParameters
% -------------------------------------------------------------------------------------------------------------
setParameters(Max,Chain,ST,Usc,Case,StDb,TermFormat,TermLength) :-
    max_length:=Max,
    chain_prefix:=Chain,
    st_prefix:=ST,
    use_underscore:=Usc,
    name_case:=Case,
    st_db:=StDb,
    term_format:=TermFormat,
    term_length:=TermLength,
    !.
    
getParameters(Max,Chain,ST,Usc,Case,StDb,TermFormat,TermLength) :-
    st_db="",
    StDb=db,
    Max=max_length,
    Chain=chain_prefix,
    ST=st_prefix,
    Usc=use_underscore,
    Case=name_case,
    TermFormat=term_format,
    TermLength=term_length,
    !
;    
    StDb=st_db,
    Max=max_length,
    Chain=chain_prefix,
    ST=st_prefix,
    Usc=use_underscore,
    Case=name_case,
    TermFormat=term_format,
    TermLength=term_length,
    !.
% -------------------------------------------------------------------------------------------------------------
% createDb/Kb
% -------------------------------------------------------------------------------------------------------------
createDb(Name) :-
     connectionError,!,fail
;    
    db<>"",
                        Mess="SQL database is already open!",
                        dlgIF::msgS(Mess),
    !,fail
;    
    Sql1=string::concat("drop database if exists ",Name),
    sqlTools::executeDirect(con,Sql1),
    Sql2=string::concat("create database ",Name),
    sqlTools::executeDirect(con,Sql2),
    Sql3=string::concat("use ",Name),
    sqlTools::executeDirect(con,Sql3),
    db:=Name,
                        Mess=string::concat("Created and opened SQL database ",Name,"!"),
                        dlgIF::msgS(Mess),
    !.
    
createKb(KB) :-
    dbError,!,fail
;   
    createSysTables(KB),
    insertParTable,
    kb:=KB,
                        Mess=string::concat("KB ",KB," created in SQL database ",db,"!"),
                        dlgIF::msgS(Mess),
    !.

createSysTables(KB) :-    
    createMainTable(KB),
    createParTable(KB),
    !.
 
createMainTable(KB) :-    
    mainTable:=string::concat(db,".p3kb_",KB),
    Sql1=string::concat("create table if not exists ",mainTable," "),
    Sql2=" (record_type tinyint not null,base_type tinyint unsigned not null,active tinyint unsigned not null,suspended tinyint unsigned not null,record_date datetime not null,realm varchar(128) not null,recno int unsigned not null,subno int unsigned not null,st varchar(255),chain varchar(128) not null,ref int unsigned not null auto_increment,primary key (chain,ref),index (base_type),index (active),index (suspended),index (realm),index (realm,recno,subno),index (st),index (chain),index (ref))",
    Sql3=string::concat(Sql1,Sql2),
    sqlTools::executeDirect(con,Sql3),
    !.
createParTable(KB) :-    
    parTable:=string::concat(db,".p3pm_",KB),
    Sql11=string::concat("create table if not exists ",parTable),
    Sql12=" (version int not null,created datetime not null,chain_prefix varchar(32) not null,st_prefix varchar(32) not null,name_case tinyint unsigned not null,use_underscore char(1) not null,max_length tinyint unsigned not null,st_db varchar(64) not null,term_format tinyint unsigned not null,term_length int unsigned not null)",
    Sql13=string::concat(Sql11,Sql12),
    sqlTools::executeDirect(con,Sql13),
    !.
    
insertParTable() :-    
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    MaxS=toString(max_length),
    CaseS=toString(name_case),
    dbBase::getProgramBuild(VR,_,_,_,_),
    Version=toString(VR),
    TermFormat=toString(term_format),
    TermLength=toString(term_length),
    Sql21=string::concatList(["insert ",parTable," values (",Version,",'",Time,"','",chain_prefix,"','",st_prefix,"',",CaseS,",'",use_underscore,"',",MaxS,",'",st_db,"',",TermFormat,",",TermLength,")"]),
    sqlTools::executeDirect(con,Sql21),
    !.
% -------------------------------------------------------------------------------------------------------------
% dbError
% -------------------------------------------------------------------------------------------------------------
dbError() :-
    connectionError,
    !
;    
    db="",
    dlgIF::dlgNote(3,"No SQL database open!"),
    !
;    
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% getConnection
% -------------------------------------------------------------------------------------------------------------
getConnection()=Con :-
    connectionError,
    !,fail
;    
    Con=con,
    !.
% -------------------------------------------------------------------------------------------------------------
% getDb/Kb/Term
% -------------------------------------------------------------------------------------------------------------
getDb()=Db:- 
    dbBase::kbIsSql,
    db<>"", Db=db,
    !
;    
    not(dbBase::kbIsSql),
    Db="",
    !.

getKb()=Kb:- 
    dbBase::kbIsSql,
    kb<>"", Kb=kb,
    !
;    
    not(dbBase::kbIsSql),
    Kb=dbBase::dbName(),
    !.

getDbKb()=DbKb :- 
    dbBase::kbIsSql,
    db<>"", kb<>"", DbKb=string::concat(db,".",kb),
    !
;
    not(dbBase::kbIsSql),
    DbKb=dbBase::dbName(),
    !.
% -------------------------------------------------------------------------------------------------------------
% names
% -------------------------------------------------------------------------------------------------------------
names(FullName,DB,KB,Path) :-
    Path2=filename::getPath(FullName),
    sx::sxRemoveTail(Path2,"\\",Path3),
    Path3="",
    sx::sxUnpack(FullName,'.',L),
    L=[DB,KB,Type],   
    string::toLowercase(Type)="sql",
    Path=globalVars::get_s("exe_path"),
    !
;    
    Path2=filename::getPath(FullName),
    sx::sxRemoveTail(Path2,"\\",Path),
    Path<>"",
    N=string::length(Path2),
    sx::frontstr(N,FullName,_,T),
    sx::sxUnpack(T,'.',L),
    L=[DB,KB,Type],   
    string::toLowercase(Type)="sql",
    !.    
% -------------------------------------------------------------------------------------------------------------
% openDb
% -------------------------------------------------------------------------------------------------------------
openDb(DB) :-
    connectionError,!,fail
;
    Sql2=string::concat("create database if not exists ",DB),
    sqlTools::executeDirect(con,Sql2),
    Sql3=string::concat("use ",DB),
    sqlTools::executeDirect(con,Sql3),
    db:=DB,
                        Mess=string::concat("Using SQL database ",DB,"!"),
                        dlgIF::msgS(Mess),
    !.
% -------------------------------------------------------------------------------------------------------------
% closeKb
% -------------------------------------------------------------------------------------------------------------
closeKb() :-
    Sql1=string::concat("delete from ",parTable),
    sqlTools::executeDirect(con,Sql1),
    insertParTable,
    dlgIF::msgS("Updated KB-parameter table!"),
    db:="",
    kb:="",
    retractall(chainRef(_,_,_)),
    dep::initGraphs,
    !.
% -------------------------------------------------------------------------------------------------------------
% openKb
% -------------------------------------------------------------------------------------------------------------
class predicates
checkVersion : ().
clauses

openKb(KB) :-
    dbError,!,fail
;    
    _=existKb(KB),
    kb:=KB,
    mainTable:=string::concat(db,".p3kb_",KB),
    parTable:=string::concat(db,".p3pm_",KB),
    Sql=string::concat("select version,chain_prefix,st_prefix,use_underscore,max_length,name_case,st_db,term_format,term_length from ",parTable),
    Select=sqlSelect::new(con,Sql),
    Select:getNRows()>0,
    sqlVersion:=Select:getI(1,1),
    chain_prefix:=Select:getS(1,2),
    st_prefix:=Select:getS(1,3),
    use_underscore:=Select:getS(1,4),
    max_length:=Select:getI(1,5),
    name_case:=Select:getI(1,6),
    st_db:=Select:getS(1,7),
    term_format:=Select:getI(1,8),
    term_length:=Select:getI(1,9),
    setRef,
                        Mess=string::concat("KB ",KB," opened in SQL database ",db,"!"),
                        dlgIF::msgS(Mess),
    checkVersion,                        
    !
;
    Mess=string::concat("SQL KB ",db,".",KB," does not exist!"),
    dlgIF::dlgNote(1,Mess),
    !,fail.   
    
checkVersion() :-
    sqlVersion>199,
    sqlVersion<506,
    dlgIF::dlgNote(1,"Your pre-version 5.6 SQL-type knowledge  base needs two additional columns in the parameter-table. Ref. Pro/3 upgrade information for details!"),
    !
;
    !.
% -------------------------------------------------------------------------------------------------------------
% openConnection;closeConnection
% -------------------------------------------------------------------------------------------------------------
openConnection() :-
    opened=1,
    !
;    
    con:=odbcConnection::new(32000),
    TaskWindow=vpi::getTaskWindow(),
    ConStr=con:driverConnect(TaskWindow,"",odbc_native::sql_driver_complete), % opens XPs ODBC dialog
    opened:=1,
                        tt::tt_s(1,"ODBC connection opened ",ConStr),
    !.
    
closeConnection() :- 
    opened=0,
    !
;    
    connectionError,!,fail
;    
    con:disconnect,
    opened:=0,
    !.
% -------------------------------------------------------------------------------------------------------------
% recordCount
% -------------------------------------------------------------------------------------------------------------
recordCount()=N :-
    Sql=string::concat("select count(*) from ",mainTable),
    Select=sqlSelect::new(con,Sql),
    Select:getNRows()=1,
    N=Select:getI(1,1),
    !.
% -------------------------------------------------------------------------------------------------------------
% currentRef
% -------------------------------------------------------------------------------------------------------------
currentRef(Chain)=Ref :-
    chainRef(Chain,Ref,_),
    !
;
    Ref=0,
    !.
% -------------------------------------------------------------------------------------------------------------
% getRef
% -------------------------------------------------------------------------------------------------------------
getRef(Chain)=Ref :-
    chainRef(Chain,N,Table),
    !,Ref=N+1,
    retract(chainRef(Chain,N,Table)),
    assert(chainRef(Chain,Ref,Table)),
    !
;    
    dropChainTable(Chain),
    createChainTableIfNotExists(Chain),
    Ref=1,
    Table=chainTable(Chain),
    assert(chainRef(Chain,Ref,Table)),
    Mess=string::concat("Created chain-table ",Table,"!"),
    dlgIF::msgS(Mess),
    !.
% -------------------------------------------------------------------------------------------------------------
% resetTable
% -------------------------------------------------------------------------------------------------------------
resetTable(Table) :-
    Sql=string::concatList(["delete from ",Table]),
    sqlTools::executeDirect(con,Sql),
    !.
% -------------------------------------------------------------------------------------------------------------
% createChainTableIfNotExists
% -------------------------------------------------------------------------------------------------------------
createChainTableIfNotExists(Chain) :-
    Table=chainTable(Chain),
    Sql=string::concatList(["create table if not exists ",Table," (ref int unsigned not null primary key,base_type tinyint unsigned not null,st varchar(255),ie tinyint unsigned not null,lterm text not null,rterm text not null,index (ie),index (base_type),index (st))"]),
    sqlTools::executeDirect(con,Sql),
    !.
% -------------------------------------------------------------------------------------------------------------
% initRef
% -------------------------------------------------------------------------------------------------------------
initRef() :- retractall(chainRef(_,_,_)),!.
% -------------------------------------------------------------------------------------------------------------
% setRef
% -------------------------------------------------------------------------------------------------------------
class predicates
deleteMissingChains : (string_list) determ.    
setRefX : (lst) determ.
clauses

setRef() :-
    MissingChains=lst::new(1),
    setRefX(MissingChains),
    !.

setRefX(MissingChains) :-
    retractall(chainRef(_,_,_)),
    MissingChains:deleteAll,
    Sql1=string::concat("select distinct chain from ",mainTable),
    Select=sqlSelect::new(con,Sql1),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Nc,1),
        createChainTableIfNotExists(Chain),
        Table=chainTable(Chain),
        Sql2=string::concat("select max(ref) from ",Table),
        Select2=sqlSelect::new(con,Sql2),
        MissingChains:add(Chain),
        Select2:getNRows()=1,
        M=Select2:getI(1,1),
        assert(chainRef(Chain,M,Table)),
        MissingChains:delete(Chain),
        fail
;
    L=MissingChains:get(),
    L=[],
    !
;    
    L=MissingChains:get(),
    dbBase::joinNL(L,Ln),
    Mess=string::concat("The following chain-tables are missing: ",Ln,". ","Data from these chains must be deleted from the KB-table!"),
    W=vpi::getTaskWindow(),
    Reply=vpiCommonDialogs::ask(W,"Please confirm deletion",Mess,["Yes","No"]),
    Reply=vpiCommonDialogs::resp_default,
    deleteMissingChains(L),
    !
;    
    L=MissingChains:get(),
    set::size(L,N),
    N>0,
    dlgIF::dlgNote(2,"The knowledge base cannot work properly until the missing chain errors have been resolved!"),
    !.
 
deleteMissingChains([]) :-
    dlgIF::dlgNote(0,"Missing chains have been deleted from KB-table. You should now regenerate all graphs. The KB could be semantically incomplete!"),
    !.
deleteMissingChains([Chain|T]) :-
    Sql=string::concat("delete from ",mainTable," where chain='",Chain,"'"),
    sqlTools::executeDirect(con,Sql),
    retractall(chainRef(Chain,_,_)),
    deleteMissingChains(T),
    !.
% -------------------------------------------------------------------------------------------------------------
% chain...
% -------------------------------------------------------------------------------------------------------------
chains(Chain) :- chainRef(Chain,_,_).
chains(Chain,MaxRef,Table) :- chainRef(Chain,MaxRef,Table).

chainRefsAll(Chain)=Ref :-
    Sql=string::concat("select  ref from ",mainTable," where chain='",Chain,"'"),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Ref=Select:getI(Row,1).

chainBasetypeRefsAll(Basetype,Chain,Ref) :-
    BasetypeS=toString(Basetype),
    Sql=string::concat("select  chain,ref from ",mainTable," where chain='",Chain,"' and base_type=",BasetypeS),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Row,1),
        Ref=Select:getI(Row,1).

chainBasetypeRefs(Basetype,Chain,Ref) :-
    BasetypeS=toString(Basetype),
    Sql=string::concatList(["select  ref from ",mainTable," where chain='",Chain,"' and base_type=",BasetypeS," and suspended=0"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Row,1),
        Ref=Select:getI(Row,1).
% -------------------------------------------------------------------------------------------------------------
% basetype../type..
% -------------------------------------------------------------------------------------------------------------
class predicates
basetypeRefListX : (integer,lst).
basetypeRefListAllX : (integer,lst).
clauses

basetypeRefsAll(Basetype,Chain,Ref) :-
    BasetypeS=toString(Basetype),
    Sql=string::concat("select  chain,ref from ",mainTable," where base_type=",BasetypeS),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Row,1),
        Ref=Select:getI(Row,2).

basetypeRefs(Basetype,Chain,Ref) :-
    BasetypeS=toString(Basetype),
    Sql=string::concat("select  chain,ref from ",mainTable," where base_type=",BasetypeS," and suspended=0"),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Row,1),
        Ref=Select:getI(Row,2).

basetypeRefListAll(Basetype)=L:-
	L=lst::new(-1),
	basetypeRefListAllX(Basetype,L),
	!.

basetypeRefListAllX(Basetype,L) :-
    basetypeRefsAll(Basetype,Chain,Ref),
        SRef=toString(Ref),
        S=string::concat(SRef,"|",Chain),
        L:add(S),
        fail
;
    !.

basetypeRefList(Basetype)=L:-
	L=lst::new(-1),
	basetypeRefListX(Basetype,L),
	!.

basetypeRefListX(Basetype,L) :-
    basetypeRefs(Basetype,Chain,Ref),
        SRef=toString(Ref),
        S=string::concat(SRef,"|",Chain),
        L:add(S),
        fail
;
    !.

typeRefs(Type,Chain,Ref) :-
    TypeS=toString(Type),
    Sql=string::concat("select  chain,ref from ",mainTable," where record_type=",TypeS),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Row,1),
        Ref=Select:getI(Row,2).
% -------------------------------------------------------------------------------------------------------------
% recTypeLocs
% -------------------------------------------------------------------------------------------------------------
class predicates
recTypeLocsX : (lst,integer).
recBasetypeLocsX : (lst,integer).
recBasetypeLocsAllX : (lst,integer).
clauses

recTypeLocs(Type,Locs) :-
    L=lst::new(-1),
    recTypeLocsX(L,Type),
    L2=L:get(),
    dbBase::toLocs(L2,Locs),
    !.
    
recTypeLocsX(L,Type) :-
    typeRefs(Type,Chain,Ref),
        S=toString(sloc(Chain,Ref)),
        L:add(S),
        fail
;
    !.

recBasetypeLocsAll(Type,Locs) :-
    L=lst::new(-1),
    recBasetypeLocsAllX(L,Type),
    L2=L:get(),
    dbBase::toLocs(L2,Locs),
    !.
    
recBasetypeLocsAllX(L,Type) :-
    basetypeRefsAll(Type,Chain,Ref),
        S=toString(sloc(Chain,Ref)),
        L:add(S),
        fail
;
    !.

recBasetypeLocs(Type,Locs) :-
    L=lst::new(-1),
    recBasetypeLocsX(L,Type),
    L2=L:get(),
    dbBase::toLocs(L2,Locs),
    !.
    
recBasetypeLocsX(L,Type) :-
    basetypeRefs(Type,Chain,Ref),
        S=toString(sloc(Chain,Ref)),
        L:add(S),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% chain...Direct
% -------------------------------------------------------------------------------------------------------------
chainSentencesDirect(Chain,LT) :- chainTermsDirect(Chain,LT,RT), RT=atom("true").
chainSentencesDirect(Chain,LT,Ref) :- chainTermsDirect(Chain,LT,RT,Ref), RT=atom("true").
chainIRsDirect(Chain,LT) :- chainTermsDirect(Chain,LT,_).   
chainIRsDirectAll(Chain,LT) :- chainTermsDirectAll(Chain,LT,_).
    
chainTermsDirect(Chain,LT,RT) :-
    chainRef(Chain,_,Table),
    Sql=string::concat("select  lterm,rterm from ",Table," where ie=1"),
    Select=sqlSelect::new(con,Sql,term_length),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        LTs=Select:getS(Row,1),
        LT=sqlToTerm(LTs),
        RTs=Select:getS(Row,2),
        RT=sqlToTerm(RTs).

chainTermsDirect(Chain,LT,RT,Ref) :-
    chainRef(Chain,_,Table),
    Sql=string::concat("select  lterm,rterm,ref from ",Table," where ie=1"),
    Select=sqlSelect::new(con,Sql,term_length),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        LTs=Select:getS(Row,1),
        LT=sqlToTerm(LTs),
        RTs=Select:getS(Row,2),
        RT=sqlToTerm(RTs),
        Ref=Select:getI(Row,3).

chainTermsDirectAll(Chain,LT,RT) :-
    chainRef(Chain,_,Table),
    Sql=string::concat("select  lterm,rterm from ",Table),
    Select=sqlSelect::new(con,Sql,term_length),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        LTs=Select:getS(Row,1),
        LT=sqlToTerm(LTs),
        RTs=Select:getS(Row,2),
        RT=sqlToTerm(RTs).
 
chainTermsDirectAll(Chain,LT,RT,Ref) :-
    chainRef(Chain,_,Table),
    Sql=string::concat("select  lterm,rterm,ref from ",Table),
    Select=sqlSelect::new(con,Sql,term_length),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        LTs=Select:getS(Row,1),
        LT=sqlToTerm(LTs),
        RTs=Select:getS(Row,2),
        RT=sqlToTerm(RTs),
        Ref=Select:getI(Row,3).
% -------------------------------------------------------------------------------------------------------------
% getMeta
% -------------------------------------------------------------------------------------------------------------
getMeta(Ref,Chain,RT,BT,Active,Suspended,Time,Realm,No,SubNo,ST) :-
    SRef=toString(Ref),
    getMetaS(SRef,Chain,RT,BT,Active,Suspended,Time,Realm,No,SubNo,ST),
    !.
    
getMetaS(SRef,Chain,RT,BT,Active,Suspended,Time,Realm,No,SubNo,ST) :-
    Sql=string::concatList(["select  record_type,base_type,active,suspended,record_date,realm,recno,subno,st from ",mainTable," where ref=",SRef," and chain='",Chain,"'"]),
    Select=sqlSelect::new(con,Sql),
    Select:getNRows()>0,
    RT=Select:getI(1,1),
    BT=Select:getI(1,2),
    Active=Select:getI(1,3),
    Suspended=Select:getI(1,4),
    Time=Select:getDT(1,5),
    Realm=Select:getS(1,6),
    No=Select:getI(1,7),
    SubNo=Select:getI(1,8),
    ST=Select:getS(1,9),
    !.
% -------------------------------------------------------------------------------------------------------------
% getRecord
% -------------------------------------------------------------------------------------------------------------
getRecord(Ref,Chain,RT,BT,Active,Suspended,Time,Realm,No,SubNo,ST,LTerm,RTerm) :-
    chainRef(Chain,_,Table),
    SRef=toString(Ref),
    getMetaS(SRef,Chain,RT,BT,Active,Suspended,Time,Realm,No,SubNo,ST),
    Sql=string::concatList(["select  lterm,rterm from ",Table," where ref=",SRef]),
    Select=sqlSelect::new(con,Sql,term_length),
    Select:getNRows()>0,
    LTermS=Select:getS(1,1),
    RTermS=Select:getS(1,2),
    LTerm=sqlToTerm(LTerms),
    RTerm=sqlToTerm(RTerms),
    !
;
    SRef=toString(Ref),
    Mess=string::concat("Failed to retrieve record: ",Chain,"|",SRef),
    dlgIF::dlgNote(3,Mess),
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% getRuleRecord
% -------------------------------------------------------------------------------------------------------------
getRuleRecordChain(Realm,No,SubNo,Chain,Ref) :-
    SNo=toString(No),
    Sub=toString(SubNo),
    Sql=string::concatList(["select  ref,chain from ",mainTable," where realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
    Select=sqlSelect::new(con,Sql),
    Select:getNRows()>0,
    Ref=Select:getI(1,1),
    Chain=Select:getS(1,2),
    !.

getRuleRecord(Chain,Realm,No,SubNo,Ref) :-
    SNo=toString(No),
    Sub=toString(SubNo),
    Sql=string::concatList(["select  ref from ",mainTable," where chain='",Chain,"' and realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
    Select=sqlSelect::new(con,Sql),
    Select:getNRows()>0,
    Ref=Select:getI(1,1),
    !.
% -------------------------------------------------------------------------------------------------------------
% getIrDef
% -------------------------------------------------------------------------------------------------------------
getIrDef(Rule,Body,Chain,Ref) :-
    getIrDefF(Rule,Body,Chain,Ref),
    !
;
    Mess=string::concat("Unknown inexact rule definition: ",Rule),
    dlgIF::dlgNote(3,Mess),
    !,fail.

getIrDefF(Rule,Body,Chain,Ref) :-
    Sql1=string::concatList(["select  ref,chain from ",mainTable," where base_type=6"]),
    Select1=sqlSelect::new(con,Sql1),
    N=Select1:getNRows(),
    N>0,
    M=programControl::sequence_nd(1,N,1),
        Ref=Select1:getI(M,1),
        RefS=toString(Ref),
        Chain=Select1:getS(M,2),
        chainRef(Chain,_,Table),
        Sql2=string::concatList(["select  lterm from ",Table," where ref='",RefS,"'"]),
        Select2=sqlSelect::new(con,Sql2,term_length),
        N2=Select2:getNRows(),
        N2>0,
        BodyS=Select2:getS(1,1),
        Body=sqlToTerm(BodyS),
        Body=cmp(_,[atom(Rule)|_]),
    !.
% -------------------------------------------------------------------------------------------------------------
% realms
% -------------------------------------------------------------------------------------------------------------
class predicates
realmsX : (lst).
clauses

realms()=List :-
    L=lst::new(-1),
    realmsX(L),
    List=L:get(),
    !.
    
realmsX(L) :-    
    Sql=string::concatList(["select distinct realm from ",mainTable," where suspended=0"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Realm=Select:getS(Row,1),
        L:add(Realm),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% stList
% -------------------------------------------------------------------------------------------------------------
class predicates
stListX : (lst).
clauses

stList()=List :-
    L=lst::new(-1),
    stListX(L),
    List=L:get(),
    !.

stListX(L) :-
    Sql=string::concatList(["select distinct st from ",mainTable," where suspended=0"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        ST=Select:getS(Row,1),
        ST<>"",
        L:add(ST),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% stBasetypes
% -------------------------------------------------------------------------------------------------------------
class predicates
stBasetypesX : (integer,lst).
clauses

stBasetypes(Basetype)=List :-
    L=lst::new(-1),
    stBasetypesX(Basetype,L),
    List=L:get(),
    !.

stBasetypesX(Basetype,L) :-
    BasetypeS=toString(Basetype),
    Sql=string::concatList(["select distinct st from ",mainTable," where base_type=",BasetypeS," and suspended=0"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        ST=Select:getS(Row,1),
        ST<>"",
        L:add(ST),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% getSrs
% -------------------------------------------------------------------------------------------------------------
class predicates
getSrsX : (integer,sqlSelect,dbLoc_list) determ (i,i,o).
clauses

getSrs(Realm,No,Sub)=L :-
    Sub=0,
    NoS=toString(No),
    Sql=string::concatList(["select chain,ref from ",mainTable," where realm='",Realm,"' and recno=",NoS]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    getSrsX(N,Select,L),
    !
;    
    Sub<>0,
    NoS=toString(No),
    SubS=toString(Sub),
    Sql=string::concatList(["select chain,ref from ",mainTable," where realm='",Realm,"' and recno=",NoS," and subno=",SubS]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    getSrsX(N,Select,L),
    !.
    
getSrsX(0,_,[]) :-
    !.
getSrsX(N,Select,[H|T]) :-
    Chain=Select:getS(N,1),
    Ref=Select:getI(N,2),
    H=sloc(Chain,Ref),
    Next=N-1,
    getSrsX(Next,Select,T),
    !.
% -------------------------------------------------------------------------------------------------------------
% getST_Count
% -------------------------------------------------------------------------------------------------------------
getST_Count(ST)=N :-
    Sql=string::concatList(["select count(*) from ",mainTable," where st='",ST,"' and suspended=0"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getI(1,1),
    !.

getST_Count(ST,Basetype)=N :-
    BasetypeS=toString(Basetype),
    Sql=string::concatList(["select count(*) from ",mainTable," where st='",ST,"' and base_type=",BasetypeS," and suspended=0"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getI(1,1),
    !.    
% -------------------------------------------------------------------------------------------------------------
% sentenceTypes
% -------------------------------------------------------------------------------------------------------------
sentenceTypes(ST) :-
    Sql=string::concatList(["select distinct st from ",mainTable," where suspended=0"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        ST=Select:getS(Row,1).
% -------------------------------------------------------------------------------------------------------------
% checkIE
% -------------------------------------------------------------------------------------------------------------
class predicates
checkIEx : (integer,integer,integer,integer) procedure (i,i,i,o).
clauses

checkIE(RecType,IE) :-
    dbBase::recordType(RecType,BT,Act,Susp),
    checkIEx(BT,Act,Susp,IE),
    !.
    
checkIEx(BT,Act,Susp,1) :-
    BT<3,
    !
;
    BT=4,
    Susp<>1,
    !
;
    BT<6,
    Susp<>1,
    Act=1,
    !.
checkIEx(_,_,_,0) :- 
    !.
% -------------------------------------------------------------------------------------------------------------
% recordInsertOutright
% -------------------------------------------------------------------------------------------------------------
recordInsertOutright(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref) :-
    LTerm=cmp(Chain,_),
    Ref=getRef(Chain),
    SRef=toString(Ref),
    chainRef(Chain,_,Table),
    checkIE(Rectype,IE),
    LT=sqlToString(LTerm),
    checkLength(LT),
    RT=sqlToString(RTerm),
    checkLength(RT),
	dbBase::recordType(RecType,Basetype,ActiveX,SuspX),
    jmod::swapI(SuspX,-1,0,Susp),
    jmod::swapI(ActiveX,-1,0,Active),
    BasetypeS=toString(Basetype),
    dbBase::kbTermST_proc(LTerm,ST),
    IEs=toString(IE),
    Sql1=string::concatList(["insert ",Table," values (",SRef,",",BasetypeS,",'",ST,"',",IEs,",'",LT,"','",RT,"')"]),
    start,
    sqlTools::executeDirect(con,Sql1),
    ActiveS=toString(Active),
    SuspS=toString(Susp),
    RTs=toString(RecType),
    SubNos=toString(SubNo),
    SentNos=toString(SentNo),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    URealm=string::toUppercase(Realm),
    Sql2=string::concatList(["insert ",mainTable," values (",RTs,",",BasetypeS,",",ActiveS,",",SuspS,",'",Time,"','",URealm,"',",SentNos,",",SubNos,",'",ST,"','",Chain,"',",SRef,")"]), 
    sqlTools::executeDirect(con,Sql2),
    commit,
    !
;
    rollback,
    S1=toString(SentNo),
    S2=toString(SubNo),
    Mess=string::concatList(["Record ",Realm,"|",S1,"|",S2," insertion failed!"]),
    dlgIF::dlgNote(3,Mess),!,fail.
% -------------------------------------------------------------------------------------------------------------
% start/commit/rollback
% -------------------------------------------------------------------------------------------------------------
start() :- 
    started=1, 
    dlgIF::dlgNote(3,"Invalid SQL transaction nesting (double-start)!"),
    !
;
    sqlTools::executeDirect(con,"START TRANSACTION"),
    started:=1,
    !.
    
commit() :-
    started=0,
    dlgIF::dlgNote(3,"Invalid SQL transaction nesting (double start)!"),
    !
;
    sqlTools::executeDirect(con,"COMMIT"),
    started:=0,
    !.
    
rollback() :-
    started=0,
    !
;
    sqlTools::executeDirect(con,"ROLLBACK"),
    started:=0,
    !.    
% -------------------------------------------------------------------------------------------------------------
% recordInsert
% -------------------------------------------------------------------------------------------------------------
class predicates
invalidSyntagm : (string,string,string,string,string,string,string,integer) determ.
otherRecInsChecks : (integer,string,integer,integer,sterm,sterm) determ.
clauses

recordInsert(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref) :-
    not(dbBase::flagsClear),
    !,fail
;	
    % check: DUPLICATE SENTENCE
	globalVars::get_i("dup_chk")=1,
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType>=1,
	BaseType<=2,
	sx::frontstr(4,Realm,File,_),
	not(set::member(File,["PRO3","$SYS","$IRR","$GRAPH"])),
    _=existSentence(LTerm),
    nl::cprTerm(LTerm,S),
    Mess=string::concat("Filtered duplicate sentence: ",S,"!"),
    dlgIF::msgS(Mess),
    Ref=0,
	!
;
    % derived sentences can be inserted without checks
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=1,
	RealmU=string::toUpperCase(Realm),
	recordInsertOutright(RecType,RealmU,SentNo,SubNo,LTerm,RTerm,Ref),
    !
; 
    % check: BLANK REALM
	Realm="",
	dlgIF::dlgNote(1,"Cannot insert record with blank realm name!"),
	!,fail
;
    % check: INVALID RECORD-TYPE
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType>7,
	SRecType=toString(RecType),
	Mess=string::concat("Cannot insert record with type: ",SRecType), 
	dlgIF::dlgNote(3,Mess),
	!,fail
; 
    % check: OTHER
    globalVars::get_i("oth_chk")=1,
	dbBase::recordType(RecType,BaseType,_,_),
    otherRecInsChecks(BaseType,Realm,SentNo,SubNo,LTerm,RTerm),
    !,fail
; 
	% insert non-$GRAPH and non-$SYS sentence record
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2, 
	Realm<>"$SYS",
	Realm<>"$GRAPH",
	!,
	RealmU=string::toUppercase(Realm),
	recordInsertOutright(RecType,RealmU,SentNo,SubNo,LTerm,RTerm,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dep::smGraphUpdate(1,RecType,LTerm),
	dbExtension::updateDependencies(1,RecType,LTerm,RTerm,"",_),
	!
;
    % insert $GRAPH sentences
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2,
	Realm="$GRAPH",
	!,
	recordInsertOutright(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref),
	% no need to mark db as updated as this is done separately for graph changes
    !
;
    % insert $SYS sentences
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=2,
	Realm="$SYS",
	!,
	recordInsertOutright(RecType,Realm,SentNo,SubNo,LTerm,RTerm,Ref),
    !
;
	% insert SR
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=3,
	LTerm=cmp(PID,_),	
	SSent=toString(SentNo),
	SSub=toString(SubNo),
	sx::sxlJoin([PID,Realm,SSent,SSub],"|",Rule),
	dbExtension::updateDependencies(1,RecType,LTerm,RTerm,Rule,Acyclic),
	Acyclic=0,
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,Rule,_),
	Mess=string::concat("Assert of sentence rule: ",Rule," failed due to cycle!"),
	dlgIF::msgS(Mess),
	!,fail
;
	% insert SR
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=3,
	RealmU=string::toUppercase(Realm),
	recordInsertOutright(RecType,RealmU,SentNo,SubNo,LTerm,RTerm,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	% insert derived SR
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=5,
	RealmU=string::toUppercase(Realm),
	recordInsertOutright(RecType,RealmU,SentNo,SubNo,LTerm,RTerm,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	% insert function def
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	dbExtension::updateDependencies(1,RecType,LTerm,RTerm,"",Acyclic),
	Acyclic=0,
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,"",_),
	LTerm=cmp(F,_),
	Mess=string::concat("Assert of function definition: ",F," failed due to cycle!"),
	dlgIF::msgS(Mess),
	!,fail
;
	% insert function def
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=4,
	RealmU=string::toUppercase(Realm),
	recordInsertOutright(RecType,RealmU,SentNo,SubNo,LTerm,RTerm,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	% insert inexact rule
	LTerm=cmp(_,[atom(Rule)|_]),	
	dbExtension::updateDependencies(1,RecType,LTerm,nill,Rule,Acyclic),
	Acyclic=1,
	dbExtension::updateDependencies(3,RecType,LTerm,nill,Rule,_),
	Mess=string::concat("Assert of inexact rule: ",Rule," failed due to cycle!"),
	dlgIF::msgS(Mess),
	!,fail
;
	% insert inexact rule
	dbBase::recordType(RecType,BaseType,_,_),
	BaseType=6,
	RealmU=string::toUppercase(Realm),
	recordInsertOutright(RecType,RealmU,SentNo,SubNo,LTerm,RTerm,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
    SRec=toString(RecType),
    S1=toString(SentNo),
    S2=toString(SubNo),
    Mess=string::concatList(["Record type: ",SRec," realm: ",Realm,"|",S1,"|",S2," NOT inserted!"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.	        

% performs other record-insert checks
%
otherRecInsChecks(BaseType,Realm,SentNo,SubNo,LTerm,RTerm) :-
	BaseType=2, 
	Realm="PRO3-B-BASE.3PR",
	LTerm=cmp("x",[cmp("p3_eVers",[int(MV),str("PRO/3 SYSTEM SENTENCES VERSION")])]),
	dbBase::getProgramBuild(PV,_,_,_,_),
	PV<>MV,
	PVF=PV/100.0,
	MVF=MV/100.0,
	sx::fixFormatR(PVF,2,0,PVs),
	sx::fixFormatR(MVF,2,0,MVs),
	sx::concat(["You are storing PRO3-B-BASE.3PR v.",MVs," under PRO3.EXE v.",PVs,". This can cause errors. Compatibility issues are documented under help page 'PRO3.EXE'. You can get updated files from www.ligaya.net/p3!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!
;
	BaseType=2, 
	LTerm=cmp("v",[cmp(T,[atom(Lng),str(Rep),atom(Clas),atom(Form),_]),_]),
	invalidSyntagm(Realm,T,Lng,Rep,Clas,Form,Realm,SentNo),
	!
;
	BaseType=2, 
	not(RTerm=atom("true")),
	dlgIF::dlgNote(1,"Attempt to insert a type-2 record with invalid right-side"),
	!
;
	BaseType=2, 
	LTerm=cmp("p3_pIpch",[cmp("p3_eChan",[atom(Chain),_])]),
	set::member(Chain,["x","z","p3_pDP","p3_pDR","p3_pSM","p3_pIpch","p3_pIxdb"]),
	Mess=string::concat("Attempt to define system chain as public",Chain,"!"),
	dlgIF::dlgNote(1,Mess),
	!
;
    SentNo<0,
	dlgIF::dlgNote(1,"Attempt to insert a record with negative record number!"),
	!
;
    SubNo<0,
	dlgIF::dlgNote(1,"Attempt to insert a record with negative record sub-number!"),
	!.

% succeeds if the given syntagm/synonym is INVALID
% fails if VALID
%
invalidSyntagm(Realm,T,Lng,Rep,Clas,Form,Realm,No) :-
    _=string::search("+-*/^=><,;:()[]{}",Rep),
    !,fail
; 
    set::member(Rep,["...","<=",">=","<>"]),
    !,fail
;
	not(string::isName(Rep)),
	sx::concat(["'",Rep,"' is not a valid KB name"],Mess),
	dlgIF::dlgNote(1,Mess),
	!	
;
	sx::frontstr(1,Rep,F,_),
    Fu=string::toUppercase(F),
	Fu=F,
	sx::concat(["'",Rep,"' is not a valid KB name, which must start with a lower case letter"],Mess),
	dlgIF::dlgNote(1,Mess),
	!	
;
	sx::frontchar(Rep,F,_),
	not(sx::isLetter(F)),
	sx::concat(["'",Rep,"' is not a valid KB name, which must start with a lower case letter"],Mess),
	dlgIF::dlgNote(1,Mess),
	!	
;
	sx::frontstr(4,Realm,F1,_),
	F1<>"PRO3",
	dbBase::dbaReservedKB_Id(Rep),
	!
;
	T="s",
	_=nlClassFormF(Rep,Clas,Form),
	!,fail % ok - the main form exists
;
	T="s",	
	SNo=toString(No),
	Mess=string::concatList(["Invalid synonym for (",Rep,") in sentence ",SNo," in realm ",Realm,"!"]),
	dlgIF::dlgNote(1,Mess),
	!
;	
	sx::frontstr(4,Realm,F1,_),
	F1<>"PRO3",
	not(dbBase::isValidId(Rep)),
	SNo=toString(No),
	Mess=string::concatList(["Invalid KB-name (",Rep,") in sentence ",SNo," in realm ",Realm,"!"]),
	dlgIF::dlgNote(1,Mess),
	!
;
	sx::sx_rev(Rep,S1),
	sx::searchchar(S1,'_',P),
	N=P-1,
	sx::frontstr(N,S1,F,_),
	string::hasDecimalDigits(F),
	SNo=toString(No),
	sx::concat(["KB name '",Rep,"' in sentence ",SNo," in realm ",Realm," ends with '_digit(s)'"],Mess),
	dlgIF::dlgNote(1,Mess),
	!
;
    _=nlClassFormF(Rep,Clas,Form),
    Lng=globalVars::get_s("lang"),
	Mess=string::concatList(["Syntagm ",Lng,"|",Rep,"|",Clas,"|",Form," exists already!"]),
	dlgIF::dlgNote(1,Mess),
	!
;
    nlSyntagmF(Rep,ClassX,_,_),
    ClassX<>Clas,
	Mess=string::concatList(["Syntagm ",Lng,"|",Rep," exists already as class ",ClassX]),
	dlgIF::dlgNote(1,Mess),
	!.
% -------------------------------------------------------------------------------------------------------------
% chainCount
% -------------------------------------------------------------------------------------------------------------
chainCount(Chain)=N :-
    Sql1=string::concat("select count(*) from ",mainTable," where chain='",Chain,"'"),
    Select=sqlSelect::new(con,Sql1),
    Select:getNRows()>0,
    N=Select:getI(1,1),
    !
;
    Mess=string::concat("Unable to get count for chain: ",Chain,"!"),
    dlgIF::dlgNote(3,Mess),
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% chainTypeCount
% -------------------------------------------------------------------------------------------------------------
chainTypeCount(BaseType,Chain)=N :-
    SType=toString(BaseType),
    Sql1=string::concatList(["select count(*) from ",mainTable," where chain='",Chain,"' and base_type=",SType]),
    Select=sqlSelect::new(con,Sql1),
    Select:getNRows()>0,
    N=Select:getI(1,1),
    !
;
    Mess=string::concat("Unable to get type-count for chain: ",Chain,"!"),
    dlgIF::dlgNote(3,Mess),
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% realmCount
% -------------------------------------------------------------------------------------------------------------
realmCount(Realm)=N :-
    Sql1=string::concat("select count(*) from ",mainTable," where realm='",Realm,"'"),
    Select=sqlSelect::new(con,Sql1),
    Select:getNRows()>0,
    N=Select:getI(1,1),
    !
;
    Mess=string::concat("Unable to get count for realm: ",Realm,"!"),
    dlgIF::dlgNote(3,Mess),
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% realmRef...
% -------------------------------------------------------------------------------------------------------------
class predicates
realmRefListAllX : (lst,string).
realmBasetypeRefListAllX : (lst,string,integer).
clauses

realmRefListAll(Realm)=L :-
    L=lst::new(-1),
    realmRefListAllX(L,Realm),
    !.
    
realmRefListAllX(L,Realm) :-    
    realmRefsAll(Realm,Chain,Ref),
        RefS=toString(Ref),
        S=string::concat(RefS,"|",Chain),
        L:add(S),
        fail
;
    !.
    
realmBasetypeRefListAll(Realm,Basetype)=L :-
    L=lst::new(-1),
    realmBasetypeRefListAllX(L,Realm,Basetype),
    !.
    
realmBasetypeRefListAllX(L,Realm,Basetype) :-    
    realmBasetypeRefsAll(Realm,Basetype,Chain,Ref),
        RefS=toString(Ref),
        S=string::concat(RefS,"|",Chain),
        L:add(S),
        fail
;
    !.
    
realmRefsAll(Realm,Chain,Ref) :-
    Sql=string::concat("select  chain,ref from ",mainTable," where realm='",Realm,"'"),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Row,1),
        Ref=Select:getI(Row,2).

realmBasetypeRefsAll(Realm,Basetype,Chain,Ref) :-
    BasetypeS=toString(Basetype),
    Sql=string::concat("select  chain,ref from ",mainTable," where realm='",Realm,"' and base_type=",BasetypeS),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Row,1),
        Ref=Select:getI(Row,2).
% -------------------------------------------------------------------------------------------------------------
% checkForEmptyChain
% -------------------------------------------------------------------------------------------------------------
checkForEmptyChain(Chain) :-
    chainCount(Chain)=0,
    Table=chainTable(Chain),
    Sql1=string::concat("drop table if exists ",Table),
    sqlTools::executeDirect(con,Sql1),
    retractall(chainRef(Chain,_,_)),
    !
;
    !.
% -------------------------------------------------------------------------------------------------------------
% chainDelete/Outright
% -------------------------------------------------------------------------------------------------------------
class predicates
chainDeleteSentencesX : (integer,string,count).
clauses

chainDelete(Chain) :-
    Ref=chainRefsAll(Chain),
        recordDelete(Chain,Ref),
        fail
;
    Table=chainTable(Chain),
    Sql1=string::concat("drop table if exists ",Table),
    sqlTools::executeDirect(con,Sql1),
    retractall(chainRef(Chain,_,_)),
    !.

chainDeleteSentences(Type,Chain,N) :-
    C=count::new(),
    chainDeleteSentencesX(Type,Chain,C),
    N=C:get(),
    !.
    
chainDeleteSentencesX(Type,Chain,C) :-
    Type=0,
    chainBasetypeRefs(1,Chain,Ref),
        recordDelete(Chain,Ref),
        C:inc(_),
        fail
;        
    Type=0,
    chainBasetypeRefs(2,Chain,Ref),
        recordDelete(Chain,Ref),
        C:inc(_),
        fail
;        
    Type=1,
    chainBasetypeRefs(1,Chain,Ref),
        recordDelete(Chain,Ref),
        C:inc(_),
        fail
;        
    Type=2,
    chainBasetypeRefs(2,Chain,Ref),
        recordDelete(Chain,Ref),
        C:inc(_),
        fail
;
    !.
    
    
chainDeleteOutright(Chain) :-
    dropChainTable(Chain),
    Sql2=string::concat("delete from ",mainTable," where chain='",Chain,"'"),
    sqlTools::executeDirect(con,Sql2),
    !.
% -------------------------------------------------------------------------------------------------------------
% dropChainTable
% -------------------------------------------------------------------------------------------------------------
dropChainTable(Chain) :-    
    Table=chainTable(Chain),
    Sql1=string::concat("drop table if exists ",Table),
    sqlTools::executeDirect(con,Sql1),
    retractall(chainRef(Chain,_,_)),
    !.
% -------------------------------------------------------------------------------------------------------------
% deleteBasetypeOutright
% -------------------------------------------------------------------------------------------------------------
deleteBasetypeOutright(Basetype) :-
    BasetypeS=toString(Basetype),
    Sql1=string::concat("select distinct chain from ",mainTable," where base_type=",BasetypeS),
    Select=sqlSelect::new(con,Sql1),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Chain=Select:getS(Nc,1),
        Table=chainTable(Chain),
        Sql2=string::concat("delete from ",Table," where base_type=",BasetypeS),
        sqlTools::executeDirect(con,Sql2),
        checkForEmptyChain(Chain),
        fail
;        
    BasetypeS=toString(Basetype),
    Sql1=string::concat("delete from ",mainTable," where base_type=",BasetypeS),
    sqlTools::executeDirect(con,Sql1),
    !.
% -------------------------------------------------------------------------------------------------------------
% deleteSelectedChainRecords
% -------------------------------------------------------------------------------------------------------------
% the basis is a select from a chain where ref is the first selected column
% 
deleteSelectedChainRecords(Select,Chain,N) :-    
    M=programControl::sequence_nd(1,N,1),
        Ref=Select:getI(M,1),
        recordDelete(Chain,Ref),
        fail
;    
    checkForEmptyChain(Chain),
    !.
% -------------------------------------------------------------------------------------------------------------
% recordDelete
% -------------------------------------------------------------------------------------------------------------
class predicates
recordDeleteX : (string,integer,integer,integer,integer,integer,integer,string,sterm,sterm) determ.
clauses

recordDelete(Chain,Ref) :-
    getRecord(Ref,Chain,RT,BT,_,Suspended,_,Realm,No,SubNo,_,LTerm,RTerm),
    recordDeleteX(Chain,Ref,RT,BT,Suspended,No,SubNo,Realm,LTerm,RTerm),
    !.

recordDeleteX(Chain,Ref,RecType,BaseType,Susp,SentNo,SubNo,Realm,LTerm,RTerm) :-
    not(dbBase::flagsClear),
    !
;    
	% derived fact
	BaseType=1,
	recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	% suspended
	Susp=1,
	recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	BaseType=2,
	% fact
	Realm<>"$SYS",
	Realm<>"$GRAPH",
	!,dep::smGraphCheckDelete(BaseType,LTerm,RTerm),
	LTerm=cmp(Chain,_),
	recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,"",_),
	dep::smGraphUpdate(3,RecType,LTerm),
	!
;
	BaseType=2,
	Realm="$GRAPH",
	recordDeleteOutright(Chain,Ref),
	!
;
	BaseType=2,
	Realm="$SYS",
	!,recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	!
;
	BaseType=4,
	% function body
	!,recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbExtension::updateDependencies(3,RecType,LTerm,atom("true"),"",_),
	!
;
	BaseType=6,
	% certainty rule
	LTerm=cmp(_,[atom(Rule)|_]),
	!,recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbExtension::updateDependencies(3,RecType,LTerm,nill,Rule,_),
	dep::smGraphUpdate(3,RecType,LTerm),
	!
;
	% rule
	BaseType=3,
	SSent=toString(SentNo),
	SSub=toString(SubNo),
	sx::sxlJoin([Chain,Realm,SSent,SSub],"|",Rule),
	recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,Rule,_),
	!
;
	% rule
	BaseType=5,
	LTerm=cmp(Chain,_),
	SSent=toString(SentNo),
	SSub=toString(SubNo),
	sx::sxlJoin([Chain,Realm,SSent,SSub],"|",Rule),
	recordDeleteOutright(Chain,Ref),
	dbBase::chainHasChangedTerm(LTerm),
	dbExtension::updateDependencies(3,RecType,LTerm,RTerm,Rule,_),
	!
;	
    SRef=toString(Ref),
    Mess=string::concat("Failed to delete record: ",Chain," ",SRef),
	dlgIF::dlgNote(3,Mess),
	!,fail.
% -------------------------------------------------------------------------------------------------------------
% recordDeleteOutrigt
% -------------------------------------------------------------------------------------------------------------
recordDeleteOutright(Chain,Ref) :-
    chainRef(Chain,_,Table),
    SRef=toString(Ref),
    Sql1=string::concat("delete from ",mainTable," where chain='",Chain,"' and ref=",SRef),
    start,
    sqlTools::executeDirect(con,Sql1),
    Sql2=string::concat("delete from ",Table," where ref=",SRef),
    sqlTools::executeDirect(con,Sql2),
    commit,
    checkForEmptyChain(Chain),
    !
;
    rollback,
    RefS=toString(Ref),
    Mess=string::concat("Record ",Chain,"|",RefS," deletion failed!"),
    dlgIF::dlgNote(2,Mess),
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% recordDeletes
% -------------------------------------------------------------------------------------------------------------
recordDeletes([]) :-
    !.
recordDeletes([sloc(Chain,Ref)|Locs]) :-
    recordDelete(Chain,Ref),
    recordDeletes(Locs),
    !.
% -------------------------------------------------------------------------------------------------------------
% stTypeDelete
% -------------------------------------------------------------------------------------------------------------
stTypeDelete(Basetype,ST,N) :-
    % empty chain
    sx::sxUnpack(ST,'|',L),
    L=[Chain|_],
    not(chainRef(Chain,_,_)),
    N=0,
    !
;
    % no records
    BasetypeS=toString(Basetype),
    Sql0=string::concat("select count(*) from ",mainTable," where st='",ST,"' and base_type=",BasetypeS),
    Select=sqlSelect::new(con,Sql0),
    Select:getI(1,1)=0,
    N=0,
    !
;
    % derived sentences
    Basetype=1,
    sx::sxUnpack(ST,'|',L),
    L=[Chain|_],
    Sql0=string::concat("select count(*) from ",mainTable," where st='",ST,"' and base_type=1"),
    Select0=sqlSelect::new(con,Sql0),
    N=Select0:getI(1,1),
    Sql3=string::concat("select distinct chain from ",mainTable," where st='",ST,"' and base_type=1"),
    Select3=sqlSelect::new(con,Sql3),
    M=Select3:getNRows(),
    Sql1=string::concat("delete from ",mainTable," where st='",ST,"' and base_type=1"),
    sqlTools::executeDirect(con,Sql1),  
    Nc=programControl::sequence_nd(1,M,1), 
        Chain=Select3:getS(Nc,1),
        Table=chainTable(Chain),
        Sql2=string::concat("delete from ",Table," where st='",ST,"' and base_type=1"),
        sqlTools::executeDirect(con,Sql2),  
        checkForEmptyChain(Chain),
        Nc=M,
        !
;    
    % other records types 
    Basetype<>1,
    sx::sxUnpack(ST,'|',L),
    L=[Chain|_],
    BasetypeS=toString(Basetype),
    Sql0=string::concat("select count(*) from ",mainTable," where st='",ST,"' and base_type=",BasetypeS),
    Select0=sqlSelect::new(con,Sql0),
    N=Select0:getI(1,1),
    Sql3=string::concat("select distinct chain from ",mainTable," where st='",ST,"' and base_type=",BasetypeS),
    Select3=sqlSelect::new(con,Sql3),
    M=Select3:getNRows(),
    dep::sentenceChanged(3,ST),
    dep::massSentenceUpdateStart,
    Nc=programControl::sequence_nd(1,M,1), 
        Chain=Select3:getS(Nc,1),  
        cache::check(Chain),
        Table=chainTable(Chain),
        Sql4=string::concat("select ref from ",Table," where st='",ST,"' and base_type=",BasetypeS),
        Select4=sqlSelect::new(con,Sql4),
        Nic=Select4:getNRows(),
        deleteSelectedChainRecords(Select4,Chain,Nic),    
        Nc=M,
    dep::massSentenceUpdateEnd,
    !.   
% -------------------------------------------------------------------------------------------------------------
% realNoCount/SubnoCount
% -------------------------------------------------------------------------------------------------------------
realmNoCount(Realm,No)=N :-
    SNo=toString(No),
    Sql=string::concatList(["select count(*) from ",mainTable," where realm='",Realm,"' and recno=",SNo]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getI(1,1),
    !.

realmSubnoCount(Realm,No,Subno)=N :-
    SNo=toString(No),
    Sub=toString(Subno),
    Sql=string::concatList(["select count(*) from ",mainTable," where realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getI(1,1),
    !.
% -------------------------------------------------------------------------------------------------------------
% realmDeleteNo/Subno
% -------------------------------------------------------------------------------------------------------------
realmDeleteNo(Realm,No) :-
    % no records
    SNo=toString(No),
    Sql=string::concatList(["select count(*) from ",mainTable," where realm='",Realm,"' and recno=",SNo]),
    Select=sqlSelect::new(con,Sql),
    Select:getI(1,1)=0,
    !
;
    SNo=toString(No),
    Sql1=string::concatList(["select count(*) from ",mainTable," where realm='",Realm,"' and recno=",SNo]),
    Select1=sqlSelect::new(con,Sql1),
    %N=Select1:getI(1,1),
    Sql2=string::concatList(["select distinct chain from ",mainTable," where realm='",Realm,"' and recno=",SNo]),
    Select2=sqlSelect::new(con,Sql2),
    M=Select2:getNRows(),
    dep::massSentenceUpdateStart,
    Nc=programControl::sequence_nd(1,M,1), 
        Chain=Select2:getS(Nc,1),  
        cache::check(Chain),
        Sql3=string::concatList(["select ref from ",mainTable," where chain='",Chain,"' and realm='",Realm,"' and recno=",SNo]),
        Select3=sqlSelect::new(con,Sql3),
        Nic=Select1:getNRows(),
        deleteSelectedChainRecords(Select3,Chain,Nic),    
        Nc=M,
    dep::massSentenceUpdateEnd,
    Sql4=string::concatList(["delete from ",mainTable," where realm='",Realm,"' and recno=",SNo]),
    sqlTools::executeDirect(con,Sql4),  
    !.   

realmDeleteSubno(Realm,No,Subno) :-
    % no records
    SNo=toString(No),
    Sub=toString(Subno),
    Sql=string::concatList(["select count(*) from ",mainTable," where realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
    Select=sqlSelect::new(con,Sql),
    Select:getI(1,1)=0,
    !
;
    SNo=toString(No),
    Sub=toString(Subno),
    Sql1=string::concatList(["select count(*) from ",mainTable," where realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
    Select1=sqlSelect::new(con,Sql1),
    %N=Select1:getI(1,1),
    Sql2=string::concatList(["select distinct chain from ",mainTable," where realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
    Select2=sqlSelect::new(con,Sql2),
    M=Select2:getNRows(),
    dep::massSentenceUpdateStart,
    Nc=programControl::sequence_nd(1,M,1), 
        Chain=Select2:getS(Nc,1),  
        cache::check(Chain),
        Sql3=string::concatList(["select ref from ",mainTable," where chain='",Chain,"' and realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
        Select3=sqlSelect::new(con,Sql3),
        Nic=Select1:getNRows(),
        deleteSelectedChainRecords(Select3,Chain,Nic),    
        Nc=M,
    dep::massSentenceUpdateEnd,
    Sql4=string::concatList(["delete from ",mainTable," where realm='",Realm,"' and recno=",SNo," and subno=",Sub]),
    sqlTools::executeDirect(con,Sql4),  
    !.   
% -------------------------------------------------------------------------------------------------------------
% realmDelete
% -------------------------------------------------------------------------------------------------------------
realmDelete(Realm,N) :-
    % no records
    Sql0=string::concat("select count(*) from ",mainTable," where realm='",Realm,"'"),
    Select=sqlSelect::new(con,Sql0),
    Select:getI(1,1)=0,
    N=0,
    !
;
    Sql1=string::concat("select count(*) from ",mainTable," where realm='",Realm,"'"),
    Select1=sqlSelect::new(con,Sql1),
    N=Select1:getI(1,1),
    Sql2=string::concat("select distinct chain from ",mainTable," where realm='",Realm,"'"),
    Select2=sqlSelect::new(con,Sql2),
    M=Select2:getNRows(),
    dep::massSentenceUpdateStart,
    Nc=programControl::sequence_nd(1,M,1), 
        Chain=Select2:getS(Nc,1),  
        cache::check(Chain), 
        Sql3=string::concatList(["select ref from ",mainTable," where chain='",Chain,"' and realm='",Realm,"'"]),
        Select3=sqlSelect::new(con,Sql3),
        Nic=Select3:getNRows(),
        deleteSelectedChainRecords(Select3,Chain,Nic),    
        Nc=M,
    dep::massSentenceUpdateEnd,
    Sql4=string::concat("delete from ",mainTable," where realm='",Realm,"'"),
    sqlTools::executeDirect(con,Sql4),  
    !.   
% -----------------------------------------------------------------------------------------------
% nlForm/F
% -----------------------------------------------------------------------------------------------
% sql-variety of kbNL_Ref
%
nlForm(Rep,Form)=NL:-
    NL=nlFormF(Rep,Form),
    !
;    
	Lng=globalVars::get_s("lang"),
	Mess=string::concatList(["No NL-form known for '",Lng,"|",Rep,"|",Form,"'"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.
	
nlFormF(Rep,Form)=NL :-
	Lng=globalVars::get_s("lang"),
    chainSentencesDirect("v",LT),
        LT=cmp(_,[cmp("t",[atom(Ln),str(Rp),_,atom(Fm),Tokens]),_]),
        Ln=Lng,
        Rp=Rep,
        Fm=Form,
        termDomain::list_conc(Tokens,NL),
        !.
% -----------------------------------------------------------------------------------------------
% nlClassForm/F
% -----------------------------------------------------------------------------------------------
% sql-variety of kbNL_Ref
%
nlClassForm(Rep,Clas,Form)=NL:-
    NL=nlClassFormF(Rep,Clas,Form),
    !
;    
	Lng=globalVars::get_s("lang"),
	Mess=string::concatList(["No NL-form known for '",Lng,"|",Rep,"|",Clas,"|",Form,"'"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.
	
nlClassFormF(Rep,Clas,Form)=NL:-	
	Lng=globalVars::get_s("lang"),
    chainSentencesDirect("v",LT),
    LT=cmp(_,[cmp("t",[atom(Ln),str(Rp),atom(Cl),atom(Fm),Tokens]),_]),
	Ln=Lng,
	Rp=Rep,
	Cl=Clas,
	Fm=Form,
	termDomain::list_conc(Tokens,NL),
	!.
% -----------------------------------------------------------------------------------------------
% nlST_F 
% -----------------------------------------------------------------------------------------------
nlST_F(Rep)=NL:-
    sx::sxUnpack(Rep,'|',L),
    L=[P,E1,E2],
    N1=nlFormF(P,"p3_iSnu"),
    N2=nlFormF(E1,"p3_iSnu"),
    N3=nlFormF(E2,"p3_iSnu"),
    NL=string::concat(N2," ",N1," ",N3),
    !
;    
    sx::sxUnpack(Rep,'|',L),
    L=[P,E1],
    N1=nlFormF(P,"p3_iSnu"),
    N2=nlFormF(E1,"p3_iSnu"),
    NL=string::concat(N2," ",N1),
    !.
% -----------------------------------------------------------------------------------------------
% nlSyntagm/F
% -----------------------------------------------------------------------------------------------
nlSyntagm(Rep,Clas,Form,NL) :-
    nlSyntagmF(Rep,Clas,Form,NL),
    !
;    
	Lng=globalVars::get_s("lang"),
	Mess=string::concatList(["Syntagm: ",Lng,"|",Rep," does not exist!"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.
	
nlSyntagmF(Rep,Clas,Form,NL) :-	
	Lng=globalVars::get_s("lang"),
    chainSentencesDirect("v",LT),
    LT=cmp(_,[cmp("t",[atom(Ln),str(Rp),atom(Clas),atom(Form),Tokens]),_]),
	Ln=Lng,
	Rp=Rep,
	termDomain::list_conc(Tokens,NL),
	!.
% -----------------------------------------------------------------------------------------------
% nlAny/F 
% -----------------------------------------------------------------------------------------------
nlAny(Rep)=NL:-
    NL=nlAnyF(Rep),
    !
;    
	Lng=globalVars::get_s("lang"),
	Mess=string::concatList(["No NL-form known for '",Lng,"|",Rep,"'"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.
	
nlAnyF(Rep)=NL:-	
	Lng=globalVars::get_s("lang"),
    chainSentencesDirect("v",LT),
    LT=cmp(_,[cmp("t",[atom(Ln),str(Rp),_,_,Tokens]),_]),
	Ln=Lng,
	Rp=Rep,
	termDomain::list_conc(Tokens,NL),
	!.
% -----------------------------------------------------------------------------------------------
% nlTry
% -----------------------------------------------------------------------------------------------
nlTry(Rep)=NL :-
    NL=nlAnyF(Rep),
    !
;
    NL=Rep,
    !.
% -----------------------------------------------------------------------------------------------
% getSyntagmSets
% -----------------------------------------------------------------------------------------------
class predicates
getSyntagmSet : (string,termSet,sterml) procedure (i,i,i).
getSyntagmSetC : (string,termSet,sterml) procedure (i,i,i).
matchTokens : (sterml,string_list,sterm) determ (i,i,o).
clauses

% parameters:
% 1 in first token in input string 
% 2 in str-list of tokens following the first token
% 3 out syntagm sets matching par 1/2 (LTERMS of the corresponding vocabulary records)
%
getSyntagmSet(Tokens,SynSets) :-
    Lng=globalVars::get_s("lang"),
    Lng=cache::tmCacheIsLoaded(),
    Set=termSet::new(),
    getSyntagmSetC(Lng,Set,Tokens),
    SynSets=Set:get(),
    !
;
    not(_=cache::tmCacheIsLoaded()),
    Lng=globalVars::get_s("lang"),
    Set=termSet::new(),
    getSyntagmSet(Lng,Set,Tokens),
    SynSets=Set:get(),
    !.
    
getSyntagmSetC(Lang,Set,Tokens) :-
    cache::getSyntagms_nd(KbName,Clas,Form,Type,Toks,_,_),
        matchTokens(Tokens,Toks,ToksTerm),
        Syntagm=cmp("v",[cmp(Type,[atom(Lang),str(KbName),atom(Clas),atom(Form),ToksTerm]),str("Pro/3")]),
        Set:add(Syntagm),
        fail
;
    !.

getSyntagmSet(Lang,Set,Tokens) :-
    dbBase::syntagm_nd(Lang,KbName,Clas,Form,Type,Toks,Segment),
        matchTokens(Tokens,Toks,ToksTerm),
        Syntagm=cmp("v",[cmp(Type,[atom(Lang),str(KbName),atom(Clas),atom(Form),ToksTerm]),str(Segment)]),
        Set:add(Syntagm),
        fail
;
    !.

matchTokens(_,[],nill) :-
    !.
matchTokens([str(H)|Tail],[H|T],list(str(H),Rest)) :-
    matchTokens(Tail,T,Rest),
    !.
% -------------------------------------------------------------------------------------------------------------
% sentenceDelete
% -------------------------------------------------------------------------------------------------------------
sentenceDelete(Lt) :-
    Lt=cmp(Chain,_),
    LTerm=sqlToString(Lt),
    Table=chainTable(Chain),
    Sql=string::concatList(["select ref from ",Table," where lterm='",LTerm,"'"]), 
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    N>0,
    Ref=Select:getI(1,1),
    recordDelete(Chain,Ref),
    !.
% -------------------------------------------------------------------------------------------------------------
% exist...
% -------------------------------------------------------------------------------------------------------------
existChain(Chain) :- chainRef(Chain,_,_),!.
%
% returns the number of records in the KB if it exists, else 0.
% note that empty system tables are created if the KB does not exist
%
existKb(KB)=N :-
    dbError,!,fail
;    
    MainTable=string::concat(db,".p3kb_",KB),
    createSysTables(KB), % creates empty sys tables only if these do not exist, thus preventing error in the select/count statement under
    Sql1=string::concat("select count(*) from ",MainTable),
    SC=sqlSelect::new(con,Sql1),
    N=SC:getI(1,1),
    !.

existSentence(LT)=N :-
    LT=cmp(Chain,_),
    chainRef(Chain,_,_),
    Table=chainTable(Chain),
    LTerm=sqlToString(LT),
    Sql=string::concatList(["select count(*) from ",Table," where lterm='",LTerm,"'"]), 
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    N>0,
    Select:getI(1,1)>0,
    !.

existSt(ST)=N :-   
    Sql=string::concatList(["select count(*) from ",mainTable," where st='",ST,"'"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    N>0,
    Select:getI(1,1)>0,
    !.

existStI(ST)=N :-   
    Sql=string::concatList(["select count(*) from ",mainTable," where st='",ST,"' and base_type=2"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    N>0,
    Select:getI(1,1)>0,
    !.

existStD(ST)=N :-   
    Sql=string::concatList(["select count(*) from ",mainTable," where st='",ST,"' and base_type=1"]),
    Select=sqlSelect::new(con,Sql),
    N=Select:getNRows(),
    N>0,
    Select:getI(1,1)>0,
    !.    
% -------------------------------------------------------------------------------------------------------------
% typeCount
% -------------------------------------------------------------------------------------------------------------
typeCount(RT)=N :-
    RTs=toString(RT),
    Sql=string::concatList(["select count(*) from ",mainTable," where record_type=",RTs]),
    Select=sqlSelect::new(con,Sql),
    Select:getNRows()>0,
    N=Select:getI(1,1),
    !.        
% -------------------------------------------------------------------------------------------------------------
% stSentence_nd
% -------------------------------------------------------------------------------------------------------------
stSentence_nd(St,Lt,Chain,Ref) :-
    Sql1=string::concat("select chain,ref from ",mainTable," where base_type < 3 and st='",St,"'"),
    Select1=sqlSelect::new(con,Sql1),
    N1=Select1:getNRows(),
    N=programControl::sequence_nd(1,N1,1),
        Chain=Select1:getS(N,1),
        Table=chainTable(Chain),
        Ref=Select1:getI(N,2),
        RefS=toString(Ref),
        Sql2=string::concat("select lterm from ",Table," where ref=",RefS),
        Select2=sqlSelect::new(con,Sql2,term_length),
        validateChainRecord(Select2,Chain,RefS),
        Lts=Select2:getS(1,1),
        Lt=sqlToTerm(Lts).
% -------------------------------------------------------------------------------------------------------------
% validateChainRecord
% -------------------------------------------------------------------------------------------------------------
validateChainRecord(Select,Chain,Ref) :-
    Select:getNRows()=1,
    !
;
    Select:getNRows()=0,
    Mess=string::concat("Record ",Ref," missing from chain-table: ",Chain,"!"),
    dlgIF::dlgNote(2,Mess),
    !,fail
;
    M=Select:getNRows(),
    Ms=toString(M),
    Mess=string::concatList(["Record ",Ref," has ",Ms," records in chain-table: ",Chain,"!"]),
    dlgIF::dlgNote(2,Mess),
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% recordSuspend
% -------------------------------------------------------------------------------------------------------------
recordSuspend(Chain,Ref) :-
    getRecord(Ref,Chain,RT,BT,_,Suspended,_,Realm,SentNo,SubNo,_,LTerm,RTerm),
    Suspended=0,
    BT>=3,
    BT<=6,
    recordDelete(Chain,Ref),
    NewRT=RT*10,
    recordInsertOutright(NewRT,Realm,SentNo,SubNo,LTerm,RTerm,_),
    !
;    
    getRecord(Ref,Chain,_,BT,_,Suspended,_,_,_,_,_,_,_),
    Suspended=0,
    BT<3,
    BT>6,
    BTs=toString(BT),
    Mess=string::concat("Attempt to suspend base type ",BTs," record!"),
    dlgIF::dlgNote(3,Mess),!,fail
;    
    !.
% -------------------------------------------------------------------------------------------------------------
% recordSetType
% -------------------------------------------------------------------------------------------------------------
recordSetType(Chain,Ref,Type) :-
    getRecord(Ref,Chain,_,_,_,_,_,Realm,N,M,_,LTerm,RTerm),
    recordDelete(Chain,Ref),
    recordInsert(Type,Realm,N,M,LTerm,RTerm,_),
    !.
% -------------------------------------------------------------------------------------------------------------
% realmRename
% -------------------------------------------------------------------------------------------------------------
realmRename(From,To,N) :-
    Sql1=string::concat("select count(*) from ",mainTable," where realm='",From,"'"),
    Select1=sqlSelect::new(con,Sql1),
    N=Select1:getNRows(),
    Sql2=string::concatList(["update ",mainTable," set realm='",To,"' where realm='",From,"'"]),
    sqlTools::executeDirect(con,Sql2),
    dep::realmRename(From,To),
    !.
% -------------------------------------------------------------------------------------------------------------
% chainTable
% -------------------------------------------------------------------------------------------------------------
class predicates
nameCase : (integer,string,string) determ (i,i,o).
clauses

chainTable(Chain)=Table :-
    chainRef(Chain,_,Table),
    !
;
    T1=string::concat(chain_prefix,Chain),
    sx::sxTrunc(T1,max_length,T2),
    nameCase(name_case,T2,Table),
    chainRef(Chain2,_,Table),
    Chain2<>Chain,
    Mess=string::concatList(["Chain-table name for chains ",Chain," and ",Chain2," are identical (",Table,")!"]),
    dlgIF::dlgNote(2,Mess),!,fail
;    
    T1=string::concat(db,".",chain_prefix,Chain),
    sx::sxTrunc(T1,max_length,T2),
    nameCase(name_case,T2,Table),
    !.

nameCase(1,T,T) :- !.
nameCase(2,T,S) :- S=string::toLowercase(T),!.    
nameCase(3,T,S) :- S=string::toUppercase(T),!.       
% -------------------------------------------------------------------------------------------------------------
% checkLength
% -------------------------------------------------------------------------------------------------------------
checkLength(S) :-
    string::length(S)<=term_length,
    !
;
    M=string::length(S),
    term_length<M,
    Ns=toString(term_length),
    Ms=toString(M),
    Mess=string::concat("Term-string with length ",Ms," exceeds max. ",Ns,"!"),
    vpiCommonDialogs::ask(Mess,"Do you want to increase maximum?",["Yes","No"])=vpiCommonDialogs::resp_default,
    term_length:=M,
    !.
% -------------------------------------------------------------------------------------------------------------
% to/from term-strings
% -------------------------------------------------------------------------------------------------------------
sqlToTerm(S)=T :-
    term_format=1,
    T=termDomain::fromTermString(S),
    !
;
    T=toTerm(S),
    !.
    
 sqlToString(T)=S :-
     term_format=1,
     S=termDomain::toTermString(T),
     !
;
    S=toString(T),
    !.
end implement sqlBase
