/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement sqlSt
open core,termDomain,globalDomains

constants
className = "com/rudstrand/sqlSt/sqlSt".
classVersion = "1.0".

class predicates
stTablePop : (odbcConnection,string,integer,string_list,string).
createSqlTable : (string St,string Table,integer Format,string_list Domains) determ (i,i,i,o). % creates 'create table...' sql for ST-table
createOpenShare : (string,string) determ.

clauses
classInfo(className, classVersion).

% -------------------------------------------------------------------------------------------------------------
% *********************** GENERAL PUB NET PREDICATES ***************************
% -------------------------------------------------------------------------------------------------------------

% -------------------------------------------------------------------------------------------------------------
% createDb
% -------------------------------------------------------------------------------------------------------------
createDb() :-
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,"drop database if exists p3publications"),
    sqlTools::executeDirect(Con,"create database p3publications"),
    sqlTools::executeDirect(Con,"use p3publications"),
    sqlTools::executeDirect(Con,"create table p3publications.pub_net (name varchar(64) not null primary key,share_format tinyint not null,share_table varchar(128),default_db varchar(128) not null,updated datetime not null)"),
    sqlTools::executeDirect(Con,"create table p3publications.publication (net varchar(64) not null,st varchar(128) not null,table_name varchar(64) not null,st_format tinyint not null,mode tinyint unsigned not null,kb varchar(128) not null,last_pub datetime not null,primary key (net,st,table_name,kb))"),
    sqlTools::executeDirect(Con,"create table p3publications.subscription (net varchar(64) not null,st varchar(128) not null,mode tinyint unsigned not null,kb varchar(128) not null,last_sub datetime not null,primary key (net,st,kb))"),
    sqlTools::executeDirect(Con,"create table p3publications.cluster (cluster_name varchar(64) not null,net_name varchar(64) not null,knowledge_stream varchar(128))"),
    sqlTools::executeDirect(Con,"create table p3publications.cluster_member (cluster_name varchar(64) not null,kb varchar(128) not null,member_no int not null,last_share datetime not null,last_derivation datetime not null,complete tinyint unsigned not null)"),
    !.
% -------------------------------------------------------------------------------------------------------------
% createDbIfnotexists
% -------------------------------------------------------------------------------------------------------------
createDbIfnotexists() :-
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,"create database if not exists p3publications"),
    sqlTools::executeDirect(Con,"create table if not exists p3publications.pub_net (name varchar(64) not null primary key,share_format tinyint not null,share_table varchar(128),default_db varchar(128) not null,updated datetime not null)"),
    sqlTools::executeDirect(Con,"create table if not exists p3publications.publication (net varchar(64) not null,st varchar(128) not null,table_name varchar(64) not null,st_format tinyint not null,mode tinyint unsigned not null,kb varchar(128) not null,last_pub datetime not null,primary key (net,st,table_name,kb))"),
    sqlTools::executeDirect(Con,"create table if not exists p3publications.subscription (net varchar(64) not null,st varchar(128) not null,mode tinyint unsigned not null,kb varchar(128) not null,last_sub datetime not null,primary key (net,st,kb))"),
    sqlTools::executeDirect(Con,"create table if not exists p3publications.cluster (cluster_name varchar(64) not null,net_name varchar(64) not null,knowledge_stream varchar(128))"),
    sqlTools::executeDirect(Con,"create table if not exists p3publications.cluster_member (cluster_name varchar(64) not null,kb varchar(128) not null,member_no int not null,last_share datetime not null,last_derivation datetime not null,complete tinyint unsigned not null)"),
    !.
% -------------------------------------------------------------------------------------------------------------
% createNet
% -------------------------------------------------------------------------------------------------------------
createNet(Name,DefDb) :-
    Con=sqlBase::getConnection(),
    Sql=string::concat("select name from p3publications.pub_net where name='",Name,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("Publishing net: ",Name," exists already!"),
    dlgIF::dlgNote(1,Mess),
    fail
;    
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["insert p3publications.pub_net (name,share_format,default_db,updated) values ('",Name,"',0,'",DefDb,"','",Time,"')"]),
    sqlTools::executeDirect(Con,Sql),    
    !.

createNet(Name,Format,Table,DefDb) :-
    not(_=string::searchchar(Table,'.',string::caseSensitive())),
    Tab=string::concat(DefDb,".",Table),
    createNet(Name,Format,Tab,DefDb),
    !.
createNet(Name,Format,Table,DefDb) :-
    Con=sqlBase::getConnection(),
    Sql=string::concat("select name from p3publications.pub_net where name='",Name,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("Publishing net: ",Name," exists already!"),
    dlgIF::dlgNote(1,Mess),
    fail
;    
    not(set::member(Format,[1,9,10,11])),
    S=toString(Format),
    Mess=string::concat("Invalid share format: ",S,"!"),
    dlgIF::dlgNote(1,Mess),
    fail
;  
    DefDb="",
    DefDb2=sqlBase::getDb(),  
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    S=toString(Format),
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["insert p3publications.pub_net (name,share_format,share_table,default_db,updated) values ('",Name,"',",S,",'",Table,"','",DefDb2,"','",Time,"')"]),
    sqlTools::executeDirect(Con,Sql), 
    createOpenShare(Name,Table),
    !
;    
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    S=toString(Format),
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["insert p3publications.pub_net (name,share_format,share_table,default_db,updated) values ('",Name,"',",S,",'",Table,"','",DefDb,"','",Time,"')"]),
    sqlTools::executeDirect(Con,Sql),    
    createOpenShare(Name,Table),
    !.
% -------------------------------------------------------------------------------------------------------------
% createOpenShare
% -------------------------------------------------------------------------------------------------------------
createOpenShare(Net,Table) :-
    fullTableName(Net,Table,Qt),
    Con=sqlBase::getConnection(),
    Sql1=string::concatList(["drop table if exists ",Qt]),
    sqlTools::executeDirect(Con,Sql1),
    Sql2=string::concatList(["create table ",Qt," (sentence text not null,created datetime not null)"]),
    sqlTools::executeDirect(Con,Sql2),
    !.
% -------------------------------------------------------------------------------------------------------------
% defaultDb
% -------------------------------------------------------------------------------------------------------------
defaultDb(Net)=Db :-
    Con=sqlBase::getConnection(),
    Sql=string::concat("select default_db from p3publications.pub_net where name='",Net,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Db=Select:getS(1,1),
    Db<>"",
    !.
% -------------------------------------------------------------------------------------------------------------
% deleteNet
% -------------------------------------------------------------------------------------------------------------
deleteNet(Name) :-
    Con=sqlBase::getConnection(),
    Sql1=string::concat("delete from p3publications.pub_net where name='",Name,"'"),
    sqlTools::executeDirect(Con,Sql1),
    Sql2=string::concat("delete from p3publications.publication where net='",Name,"'"),
    sqlTools::executeDirect(Con,Sql2),
    Sql3=string::concat("delete from p3publications.subscription where net='",Name,"'"),
    sqlTools::executeDirect(Con,Sql3),
    !.
% -------------------------------------------------------------------------------------------------------------
% getNet
% -------------------------------------------------------------------------------------------------------------
class predicates
getNetX : (string,integer,string,lst,string) determ (i,o,o,i,o).
clauses

getNet(Net,OpenShare,ShareTable,Pubs,DefDb) :-
    Con=sqlBase::getConnection(),
    Sql=string::concat("select name from p3publications.pub_net where name='",Net,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N=0,
    % unknown net
    !,fail
;   
    Lp=lst::new(),
    getNetX(Net,OpenShare,ShareTable,Lp,DefDb),
    Pubs=Lp:get(),
    !.
    
getNetX(Net,OpenShare,ShareTable,Lp,DefDb) :-    
    Con=sqlBase::getConnection(),
    Sql=string::concat("select share_format,share_table,default_db from p3publications.pub_net where name='",Net,"' and share_format>0"),
    Select=sqlSelect::new(Con,Sql),
    Select:getNRows()=1,
    OpenShare=Select:getI(1,1),
    ShareTable=Select:getS(1,2),
    DefDb=Select:getS(1,3),
    !
;    
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select default_db from p3publications.pub_net where name='",Net,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    Select1:getNRows()=1,
    DefDb=Select1:getS(1,1),
    Sql=string::concat("select st,st_format,table_name,mode,kb from p3publications.publication where net='",Net,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        St=Select:getS(Nc,1),
        Format=Select:getI(Nc,2),
        Table=Select:getS(Nc,3),
        Mode=Select:getI(Nc,4),
        Owner=Select:getS(Nc,5),
        modeCode(ModeN,Mode),
        dbBase::nlSing(ModeN,ModeNl),
        formatCode(FormN,Format),
        S=sx::pack([St,FormN,ModeNl,Table,Owner],"|"),
        Lp:add(S),
        fail
;  
    Con=sqlBase::getConnection(),
    Sql1=string::concat("select default_db from p3publications.pub_net where name='",Net,"'"),
    Select1=sqlSelect::new(Con,Sql1),
    Select1:getNRows()=1,
    DefDb=Select1:getS(1,1),
    OpenShare=0,
    ShareTable="",
    !.    
% -------------------------------------------------------------------------------------------------------------
% getNets
% -------------------------------------------------------------------------------------------------------------
class predicates
getNetsX : (lst).
clauses

getNets()=Nets :-
    L=lst::new(),
    getNetsX(L),
    Nets=L:get(),
    !.
    
getNetsX(L) :-    
    Con=sqlBase::getConnection(),
    Select=sqlSelect::new(Con,"select distinct name from p3publications.pub_net"),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Name=Select:getS(Nc,1),
        L:add(Name),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% getPub
% -------------------------------------------------------------------------------------------------------------
getPub(Net,St,Table,Format,Mode) :-
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["select st_format,table_name,mode from p3publications.publication where net='",Net,"' and st='",St,"'"]),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Format=Select:getI(Nc,1),
        Table=Select:getS(Nc,2),
        Mode=Select:getI(Nc,3),
    !.

getPub(Kb,Net,St,Table,Format,Mode) :-
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["select st_format,table_name,mode from p3publications.publication where net='",Net,"' and kb='",Kb,"' and st='",St,"'"]),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Format=Select:getI(Nc,1),
        Table=Select:getS(Nc,2),
        Mode=Select:getI(Nc,3),
    !.
% -------------------------------------------------------------------------------------------------------------
% getSub
% -------------------------------------------------------------------------------------------------------------
getSub(Net,St,Table,Format,Mode) :-
    getSubMode(Net,St,Mode),
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["select table_name,st_format from p3publications.publication where net='",Net,"' and st='",St,"'"]),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    Nc=programControl::sequence_nd(1,N,1),
        Table2=Select:getS(Nc,1),
        fullTableName(Net,Table2,Table),
        Format=Select:getI(Nc,2),
    !.
% -------------------------------------------------------------------------------------------------------------
% getShareTable
% -------------------------------------------------------------------------------------------------------------
getShareTable(Net,Table,Format) :-
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["select share_table,share_format from p3publications.pub_net where name='",Net,"' and share_format>0"]),
    Select=sqlSelect::new(Con,Sql),
    Select:getNRows()>0,
    Table=Select:getS(1,1),
    Format=Select:getI(1,2),
    !.
% -------------------------------------------------------------------------------------------------------------
% modeCode/formatCode
% -------------------------------------------------------------------------------------------------------------
modeCode("p3_iIncre",1).
modeCode("p3_iRepla",2).

formatCode("NL",1).
formatCode("SQL",8).
formatCode("PR",9).
formatCode("XML/NL",10).
formatCode("XML/PR",11).

% -------------------------------------------------------------------------------------------------------------
% ************************* KB LOCAL PUB NET USAGE PREDICATES ******************
% -------------------------------------------------------------------------------------------------------------

% -------------------------------------------------------------------------------------------------------------
% addPub
% -------------------------------------------------------------------------------------------------------------
addPub(Net,St,Table,Format,Mode,PubS) :-
    not(getNet(Net,_,_,_,_)),
    Mess=string::concat("Unknown publishing network: ",Net,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Sql=string::concat("select share_format from p3publications.pub_net where name='",Net,"' and share_format>0"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("Net: ",Net," is an open share!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Mode=2, % replacement
    Con=sqlBase::getConnection(),
    Sql=string::concat("select st from p3publications.publication where net='",Net,"' and st='",St,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("Cannot add replacement-type publication when there already is a publication of sentence type: ",St,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Sql=string::concat("select st from p3publications.publication where net='",Net,"' and st='",St,"' and mode=2"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("Cannot add publication when there already is a replacement-type publication of sentence type: ",St,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    F=toString(Format),
    Sql=string::concat("select st_format from p3publications.publication where net='",Net,"' and st='",St,"' and st_format<>",F),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("An existing publication with a different format exists alrady for sentence type: ",St,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["select table_name from p3publications.publication where net='",Net,"' and table_name='",Table,"' and st<>'",St,"'"]),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("Table name: ",Table," is being used for another sentenxce type!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Form=toString(Format),
    Use=toString(Mode),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    DbKb=sqlBase::getDbKb(),
    Sql=string::concatList(["insert p3publications.publication (net,st,table_name,st_format,mode,kb,last_pub) values ('",Net,"','",St,"','",Table,"',",Form,",",Use,",'",DbKb,"','",Time,"')"]),            
    sqlTools::executeDirect(Con,Sql),
    set::setGet(["p3_iIncre","p3_iRepla"],Mode,ModeK),
    dbBase::nlSing(ModeK,ModeN),
    formatCode(FormN,Format),
    PubS=sx::pack([St,FormN,ModeN,Table],"|"),
    !.
% -------------------------------------------------------------------------------------------------------------
% addSub
% -------------------------------------------------------------------------------------------------------------
addSub(Net,St,Mode) :-
    not(getNet(Net,_,_,_,_)),
    Mess=string::concat("Unknown publishing network: ",Net,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Sql=string::concat("select st_format from p3publications.publication where net='",Net,"' and st='",St,"'"),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N=0,
    Mess=string::concat("Unknown sentence type publication: ",St,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql=string::concatList(["select kb from p3publications.subscription where net='",Net,"' and st='",St,"' and kb='",DbKb,"'"]),            
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mess=string::concat("Subscription to this sentence type: ",St," exists already!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Use=toString(Mode),
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    DbKb=sqlBase::getDbKb(),
    Sql=string::concatList(["insert p3publications.subscription (net,st,mode,kb,last_sub) values ('",Net,"','",St,"',",Use,",'",DbKb,"','",Time,"')"]),            
    sqlTools::executeDirect(Con,Sql),
    !.
% -------------------------------------------------------------------------------------------------------------
% getSubMode
% -------------------------------------------------------------------------------------------------------------
getSubMode(Net,St,Mode) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql=string::concatList(["select mode from p3publications.subscription where net='",Net,"' and st='",St,"' and kb='",DbKb,"'"]),  
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    N>0,
    Mode=Select:getI(1,1),          
    !.    
% -------------------------------------------------------------------------------------------------------------
% remPub
% -------------------------------------------------------------------------------------------------------------
remPub(Kb,Net,St) :-
    not(getNet(Net,_,_,_,_)),
    Mess=string::concat("Unknown publishing network: ",Net,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Sql=string::concatList(["delete from p3publications.publication where net='",Net,"' and st='",St,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql),
    !.
% -------------------------------------------------------------------------------------------------------------
% remSub
% -------------------------------------------------------------------------------------------------------------
remSub(Net,St) :-
    not(getNet(Net,_,_,_,_)),
    Mess=string::concat("Unknown publishing network: ",Net,"!"),
    dlgIF::dlgNote(0,Mess),
    !,fail
;    
    Con=sqlBase::getConnection(),
    Kb=sqlBase::getDbKb(),
    Sql=string::concatList(["delete from p3publications.subscription where net='",Net,"' and st='",St,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql),
    !.

% -------------------------------------------------------------------------------------------------------------
% ************************* KB LOCAL PUB NET USAGE PREDICATES ******************
% -------------------------------------------------------------------------------------------------------------

% -------------------------------------------------------------------------------------------------------------
% deleteKbNet
% -------------------------------------------------------------------------------------------------------------
deleteKbNet(Net) :-
    Con=sqlBase::getConnection(),
    Kb=sqlBase::getDbKb(),
    Sql=string::concatList(["delete from p3publications.subscription where net='",Net,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql),
    Sql2=string::concatList(["delete from p3publications.publication where net='",Net,"' and kb='",Kb,"'"]),
    sqlTools::executeDirect(Con,Sql2),
    !.
% -------------------------------------------------------------------------------------------------------------
% getKbNetUse
% -------------------------------------------------------------------------------------------------------------
class predicates
getOpenPub : (string,integer) procedure (i,o).
getOpenSub : (string,integer) procedure (i,o).
getNetUseX : (string,lst,lst).
clauses

getKbNetUse(Net,OpenPub,OpenSub,Pubs,Subs) :-
    Lp=lst::new(),
    Ls=lst::new(),
    getOpenPub(Net,OpenPub),
    getOpenSub(Net,OpenSub),
    getNetUseX(Net,Lp,Ls),
    Pubs=Lp:get(),
    Subs=Ls:get(),
    !.
    
getOpenPub(Net,Open) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select mode from p3publications.publication where net='",Net,"' and kb='",DbKb,"' and st=''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    N>0,
    Open=Select:getI(1,1),
    !
 ;
    Open=0,
    !.

getOpenSub(Net,Open) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select mode from p3publications.subscription where net='",Net,"' and kb='",DbKb,"' and st=''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    N>0,
    Open=Select:getI(1,1),
    !
 ;
    Open=0,
    !.

getNetUseX(Net,Lp,Ls) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st from p3publications.subscription where net='",Net,"' and kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        St=Select:getS(Row,1),
        not(sqlSt::getPub(Net,St,_,_,_)),
        Mess=string::concat("Subscription to non-existing publication: ",St,"!"),
        vpiCommonDialogs::resp_default=vpiCommonDialogs::ask("Delete subscription?",Mess,["Yes","No"]),
        Sql2=string::concatList(["delete from p3publications.subscription where net='",Net,"' and kb='",DbKb,"' and st='",St,"'"]),
        sqlTools::executeDirect(Con,Sql2),
        fail
;
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st,mode from p3publications.subscription where net='",Net,"' and kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        St=Select:getS(Row,1),
        Mode=Select:getI(Row,2),
        modeCode(ModeK,Mode),
        dbBase::nlSing(ModeK,ModeN),
        sqlSt::getPub(Net,St,Table,Format,_),
        formatCode(FormatN,Format),
        S=sx::pack([St,FormatN,ModeN,Table],"|"),
        Ls:add(S),
        fail
;
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st,st_format,mode,table_name from p3publications.publication where net='",Net,"' and kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        St=Select:getS(Row,1),
        Format=Select:getI(Row,2),
        Mode=Select:getI(Row,3),
        Table=Select:getS(Row,4),
        formatCode(FormatN,Format),
        modeCode(ModeK,Mode),
        dbBase::nlSing(ModeK,ModeN),
        S=sx::pack([St,FormatN,ModeN,Table],"|"),
        Lp:add(S),
        fail
 ;
    !.
       
% -------------------------------------------------------------------------------------------------------------
% *********************** PROCESS SUB AND PUB ****** ***************************
% -------------------------------------------------------------------------------------------------------------
% -------------------------------------------------------------------------------------------------------------
% publishOpen
% -------------------------------------------------------------------------------------------------------------
class predicates
publishOpenSentence : (odbcConnection,integer,string,sterm,string) determ.
publishOpenX : (string,integer,integer).
getOpenShareTable : (string,string,integer) determ (i,o,o).
shareSentenceExists : (odbcConnection,string,string) determ.
clauses

publishOpen(Net) :-
    Kb=sqlBase::getDbKb(),
    getPub(Kb,Net,"",_,_,Mode),
    getOpenShareTable(Net,Table,Format),
    publishOpenX(Table,Format,Mode),
    !.
    
publishOpenX(Table,Format,Mode) :-
    Mode=2,
    sqlBase::resetTable(Table),
    publishOpenX(Table,Format,1),
    !
;
    Mode=1,
    Con=sqlBase::getConnection(),
    sqlBase::chains(Chain),
        not(set::member(Chain,["v","q","r","x","y","p3_pDR","p3_pDP","p3_pSM","p3_pIntd","p3_pIRinp"])),
        dbBase::chainSentences(Chain,Lt),
            datetime::tsNow(Ts),
            datetime::tsSQL(Ts,Time),
            publishOpenSentence(Con,Format,Table,Lt,Time),
            fail
;
    !.

publishOpenSentence(Con,Format,TableName,Lt,Time) :-
    Format=1,
    nl::conSentence(Lt,S),
    not(shareSentenceExists(Con,TableName,S)),
    Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
    sqlTools::executeDirect(Con,Sql),
    !
;
    Format=9,
    nl::cprTerm(Lt,S),
    not(shareSentenceExists(Con,TableName,S)),
    Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
    sqlTools::executeDirect(Con,Sql),
    !
;
    Format=10,
    nl::cprTerm(Lt,Pr),
    transpr::proPR_Term(Pr,Term),
    xmlP3::xmlFromSentence("NL","PUBLICATIONS",0,Term,S),
    not(shareSentenceExists(Con,TableName,S)),
    Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
    sqlTools::executeDirect(Con,Sql),
    !
;
    Format=11,
    nl::cprTerm(Lt,Pr),
    transpr::proPR_Term(Pr,Term),
    xmlP3::xmlFromSentence("PR","PUBLICATIONS",0,Term,S),
    not(shareSentenceExists(Con,TableName,S)),
    Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
    sqlTools::executeDirect(Con,Sql),
    !.
    
shareSentenceExists(Con,Table,S) :-
    Con=sqlBase::getConnection(),
    Sql1=string::concatList(["select created from ",Table," where sentence='",S,"'"]),
    Select1=sqlSelect::new(Con,Sql1),
    Select1:getNRows()>0,
    !.
    
getOpenShareTable(Net,Table,Format) :-
    Con=sqlBase::getConnection(),
    Sql1=string::concatList(["select share_table,share_format from p3publications.pub_net where name='",Net,"'"]),  
    Select1=sqlSelect::new(Con,Sql1),
    Table=Select1:getS(1,1),
    Format=Select1:getI(1,2),
    !.
% -------------------------------------------------------------------------------------------------------------
% publishAll
% -------------------------------------------------------------------------------------------------------------
publishAll() :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st,st_format,mode,table,net from p3publications.publication where kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        St=Select:getS(Row,1),
        Format=Select:getI(Row,2),
        Mode=Select:getI(Row,3),
        Table=Select:getS(Row,4),
        Net=Select:getS(Row,5),
        publishSt(St,Net,Table,Format,Mode),
        Mess=string::concatList(["Published '",St,"' sentence type to table ",Table,"!"]),
        dlgIF::msgS(Mess),
        fail
;
    !.
    
publishAll(Net) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st,st_format,mode,table from p3publications.publication where net='",Net," and kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        St=Select:getS(Row,1),
        Format=Select:getI(Row,2),
        Mode=Select:getI(Row,3),
        Table=Select:getS(Row,4),
        publishSt(St,Net,Table,Format,Mode),
        Mess=string::concatList(["Published '",St,"' sentence type to table ",Table,"!"]),
        dlgIF::msgS(Mess),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% publishSt
% -------------------------------------------------------------------------------------------------------------
publishSt(St) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st_format,mode,table_name,net from p3publications.publication where kb='",DbKb,"' and st='",St,"'"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Format=Select:getI(Row,1),
        Mode=Select:getI(Row,2),
        Table=Select:getS(Row,3),
        Net=Select:getS(Row,4),
        publishSt(St,Net,Table,Format,Mode),
        !
;
    !.
    
publishSt(St,Net,Table,Format,Mode) :-
    Mode=2,
    fullTableName(Net,Table,FullName),
    Con=sqlBase::getConnection(),
    Sql1=string::concat("drop table if exists ",FullName),
    sqlTools::executeDirect(Con,Sql1),   
    dbBase::stNL_KB(St,Stk),
    createSqlTable(Stk,FullName,Format,Domains),
    stTablePop(Con,Stk,Format,Domains,FullName),
    !
;
    %@@@missing
    !.
% -------------------------------------------------------------------------------------------------------------
% subscribeOpen
% -------------------------------------------------------------------------------------------------------------
class predicates
subscribeOpenX : (string,integer,integer) determ.
clauses

subscribeOpen(Net) :-
    sqlSt::getShareTable(Net,Table,Format),
    getSubMode(Net,"",Mode),
    subscribeOpenX(Table,Format,Mode),
    !.
    
subscribeOpenX(Table,Format,Mode) :-
    Mode=2,
    dbBase::chains_nd(1,Chain),
        not(set::member(Chain,["q","v","r","x","y","p3_pDR","p3_pDP","p3_pSM","p3_pIntd","p3_pIRinp"])),
        dbBase::chainDeleteSentences(0,Chain,_),
        fail
;
    Mode=2,
    subscribeOpenX(Table,Format,1),
    !
;
    Mode=1,
    Con=sqlBase::getConnection(),
    Sql=string::concat("select sentence from ",Table),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    getRecs(N,Select,Format),
    !.
% -------------------------------------------------------------------------------------------------------------
% hasSubscriptions
% -------------------------------------------------------------------------------------------------------------
hasSubscriptions() :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st,mode,net from p3publications.subscription where kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    N>0,
    !.
% -------------------------------------------------------------------------------------------------------------
% subscribeAll
% -------------------------------------------------------------------------------------------------------------
subscribeAll() :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st,mode,net from p3publications.subscription where kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        St=Select:getS(Row,1),
        Mode=Select:getI(Row,2),
        Net=Select:getS(Row,3),
        getSub(Net,St,Table,Format,_),
        subscribeSt(St,Net,Table,Format,Mode),
        Mess=string::concatList(["Subscribed '",St,"' sentence type from table ",Table,"!"]),
        dlgIF::msgS(Mess),
        fail
;
    !.

subscribeAll(Net) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select st,mode from p3publications.subscription where net='",Net," and kb='",DbKb,"' and st<>''"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        St=Select:getS(Row,1),
        Mode=Select:getI(Row,2),
        getSub(Net,St,Table,Format,_),
        subscribeSt(St,Net,Table,Format,Mode),
        Mess=string::concatList(["Subscribed '",St,"' sentence type from table ",Table,"!"]),
        dlgIF::msgS(Mess),
        fail
;
    !.
% -------------------------------------------------------------------------------------------------------------
% subscribeSt
% -------------------------------------------------------------------------------------------------------------
class predicates
getArgs : (string_list,integer,sqlSelect,string_list) determ.
getArg : (integer,sqlSelect,integer,integer,integer,string_list,sterml) determ (i,i,i,i,i,i,o).
getCol : (sqlSelect,integer,integer,string,string) determ (i,i,i,i,o). 
getRecs : (integer,sqlSelect,integer).  
clauses

subscribeSt(St) :-
    Con=sqlBase::getConnection(),
    DbKb=sqlBase::getDbKb(),
    Sql1=string::concatList(["select mode,net from p3publications.subscription where kb='",DbKb,"' and st='",St,"'"]),  
    Select=sqlSelect::new(Con,Sql1),
    N=Select:getNRows(),
    Row=programControl::sequence_nd(1,N,1),
        Mode=Select:getI(Row,1),
        Net=Select:getS(Row,2),
            Mess=string::concat("Subscribing to ",St,"..."),
            dlgIF::msgStatus(Mess),
        getSub(Net,St,Table,Format,_),
        subscribeSt(St,Net,Table,Format,Mode),
        !.

subscribeSt(St,Net,Table,Format,Mode) :-
    Mode=2,
    Format<>8,
    Con=sqlBase::getConnection(),
    fullTableName(Net,Table,FullName),
    Sql=string::concat("select sentence from ",FullName),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    getRecs(N,Select,Format),
    !
;
    Mode=2,
    Format=8,
    Con=sqlBase::getConnection(),
    fullTableName(Net,Table,FullName),
    Sql=string::concat("select * from ",FullName),
    dbBase::stNL_KB(St,Stk),
    Select=sqlSelect::new(Con,Sql),
    N=Select:getNRows(),
    dbBase::smStDes(Stk,Des),
    dbBase::smStDesSql(Stk,DesSql),
    colDefs(Des,DesSql,Domains,_),
    dbBase::stTypeDelete(2,Stk,_),
    sx::sxUnpack(Stk,'|',L),
    getArgs(L,N,Select,Domains),
    !
;
    %@@@missing
    !.

getRecs(N,Select,Format) :-
    Row=programControl::sequence_nd(1,N,1),
        S=Select:getS(Row,1),
        transSentence(Format,S,Lt),
        dbBase::recordInsert(2,"$SUBSCRIPTION",0,0,Lt,atom("true"),_),
        fail
;
    !.

getArgs([P,Ea,Eb],N,Select,Domains) :-
    dbBase::smClassDef(Ea,"e",_,_,_,_,Np),
    First=1,
    Last=First+Np-1,
    dbBase::smClassDef(Eb,"e",_,_,_,_,Nea),
    First2=Last+1,
    Last2=First2+Nea-1,
    dbBase::smClassDef(P,"p",_,_,_,_,Neb),
    First3=Last2+1,
    Last3=First3+Neb-1,
    Row=programControl::sequence_nd(1,N,1),
        getArg(Row,Select,1,First,Last,Domains,Vsa),
        getArg(Row,Select,1,First2,Last2,Domains,Vsb),
        getArg(Row,Select,1,First3,Last3,Domains,Vsp),
        Lt=cmp(P,[cmp(Ea,Vsa),cmp(Eb,Vsb)|Vsp]),
        dbBase::recordInsert(2,"$SUBSCRIPTION",0,0,Lt,atom("true"),_),
        fail
;
    !.
getArgs([P,Ea],N,Select,Domains) :-
    dbBase::smClassDef(Ea,"e",_,_,_,_,Np),
    First=1,
    Last=First+Np-1,
    dbBase::smClassDef(P,"p",_,_,_,_,Nea),
    First2=Last+1,
    Last2=First2+Nea-1,
    Row=programControl::sequence_nd(1,N,1),
        getArg(Row,Select,1,First,Last,Domains,Vsa),
        getArg(Row,Select,1,First2,Last2,Domains,Vsp),
        Lt=cmp(P,[cmp(Ea,Vsa)|Vsp]),
        dbBase::recordInsert(2,"$SUBSCRIPTION",0,0,Lt,atom("true"),_),
        fail
;
    !.

getArg(_,_,Q,_,Last,_,[]) :-
    Q>Last,
    !.
getArg(Row,Select,Q,First,Last,[_|Doms],Vs) :-
    Q<First,
    M=Q+1,
    getArg(Row,Select,M,First,Last,Doms,Vs),
    !.
getArg(Row,Select,Q,First,Last,[Dom|Doms],[V|Vs]) :-
    getCol(Select,Row,Q,Dom,S),
    syn::sbGetDE_ValIF(Dom,S,V),
    M=Q+1,
    getArg(Row,Select,M,First,Last,Doms,Vs),
    !.    
    
getCol(Select,Row,Col,Dom,S) :-
    dom::domInt(Dom),
    Si=Select:getI(Row,Col),
    S=toString(Si),
    !
;    
    dom::domNum(Dom),
    Si=Select:getN(Row,Col),
    S=toString(Si),
    !
;
    dom::domXmlDt(Dom),
    S=Select:getDT(Row,Col),
    !
;    
    S=Select:getS(Row,Col),
    !.

% -------------------------------------------------------------------------------------------------------------
% ******************************* MISC  *****************************************
% -------------------------------------------------------------------------------------------------------------

% -------------------------------------------------------------------------------------------------------------
% transSentence
% -------------------------------------------------------------------------------------------------------------
transSentence(Format,S,Lt) :-
    Format=1,
    sx::sxGiveTail(S,"!",S2),
    transnl::parTranslate(S2,Lt,TermType),
    TermType="s",
    !
;    
    Format=9,
    transpr::proPR_Term(S,Lt),
    !
;
    set::member(Format,[10,11]),
    Xt=xmlTree::newTree(S),
    X=Xt:get(),
    xmlP3::xmlToSentence(X,Lt,_,_),
    !
;
    tt::tt_i(1,"Sentence format",Format),
    tt::tt_s(1,"Sentence not translatable",S),
    Mess=string::concat("Failed to translate sentence with format: ",S,"!"),
    dlgIF::dlgNote(1,Mess),
    !,fail.
% -------------------------------------------------------------------------------------------------------------
% createSqlTable
% -------------------------------------------------------------------------------------------------------------
class predicates
colDefs : (string_list,string_list,string_list,string_list) determ (i,i,o,o).
clauses

createSqlTable(Stk,Table,Format,Domains) :-
    Format=8,
    dbBase::smStDes(Stk,Des),
    dbBase::smStDesSql(Stk,DesSql),
    colDefs(Des,DesSql,Domains,ColDefs),
    sx::sxlJoin(ColDefs,",",S),
    Sql=string::concatList(["create table ",Table," (",S,",created datetime not null)"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,Sql),
    !
;    
    Format<>8,
    Sql=string::concatList(["create table ",Table," (sentence text not null,created datetime not null)"]),
    Con=sqlBase::getConnection(),
    sqlTools::executeDirect(Con,Sql),
    Domains=[],
    !.
% -------------------------------------------------------------------------------------------------------------
% fullTableName
% -------------------------------------------------------------------------------------------------------------
fullTableName(Net,Name,FullName) :-
    _=string::search(Name,"."),
    FullName=Name,
    !
;    
    StDb=defaultDb(Net),
    StDb<>"",
    FullName=string::concat(StDb,".",Name),
    !
;
    Db=sqlBase::getDb(),
    Db<>"",
    FullName=string::concat(Db,".",Name),
    !
;
    Name=FullName,
    !.
% -------------------------------------------------------------------------------------------------------------
% stTableName
% -------------------------------------------------------------------------------------------------------------
stTableName(St)=TableName :-
    dbBase::nlSing(St,Stn),
    sqlBase::getParameters(Max,_,StPrefix,Blank,Case,StDb,_,_),
    N=string::length(StPrefix),
    Maxx=Max-N,
    sx::makeIdentifier(Case,Case,Case,Stn,Maxx,Blank,Name1),
    sqlTools::makeSqlName(Name1,Max,Name2),
    TableName=string::concat(StDb,".",StPrefix,Name2),
    !.
% -------------------------------------------------------------------------------------------------------------
% colDefs
% -------------------------------------------------------------------------------------------------------------
colDefs([],_,[],[]) :-
    !.
colDefs([H|T],[Name|Names],[Dom|Doms],[Hd|Td]) :-
    sx::sxUnpack(H,'|',L),
    L=[_,_,M],
    dbBase::smClassDef(M,"a",_,_,Dom,Form,Width),
    colType(Width,Form,Dom,SqlType),
    Hd=string::concat(Name," ",SqlType),
    colDefs(T,Names,Doms,Td),
    !
;    
    sx::sxUnpack(H,'|',L),
    L=[_,M],
    dbBase::smClassDefF(M,"a",_,_,Dom,Form,Width),
    colType(Width,Form,Dom,SqlType),
    Hd=string::concat(Name," ",SqlType),
    colDefs(T,Names,Doms,Td),
    !
;      
    sx::sxUnpack(H,'|',L),
    L=[M],
    dbBase::smClassDefF(M,"a",_,_,Dom,Form,Width),
    colType(Width,Form,Dom,SqlType),
    Hd=string::concat(Name," ",SqlType),
    colDefs(T,Names,Doms,Td),
    !.  
% -------------------------------------------------------------------------------------------------------------
% colType
% -------------------------------------------------------------------------------------------------------------
colType(Width,Form,Domain,SQL_Type) :-
    set::member(Domain,["p3_dxd","p3_dxt","p3_dxdt"]),
    SQL_Type="datetime",
    !
;
    dom::domInt(Domain),
    SQL_Type="int",
    !
;    
    dom::domFix(Domain),
    P=string::search(Form,"."),
    PP=P+1,
    LL=string::length(Form)-2,
    SDec=string::subString(Form,PP,LL),
    Dec=sx::toInt(SDec)+2,
    SDec2=toString(Dec),
    Width2=Width+1,
    SWidth=toString(Width2),
    SQL_Type=string::concat("decimal(",SWidth,",",SDec2,")"),
    !
;    
    dom::domFix(Domain),
    P=string::search(Form,"."),
    PP=P+1,
    LL=string::length(Form)-1,
    SDec=string::subString(Form,PP,LL),
    SWidth=toString(Width),
    SQL_Type=string::concat("decimal(",SWidth,",",SDec,")"),
    !
;    
    dom::domFix(Domain),
    _=string::search(Form,"%"),
    Width2=Width+1,
    SWidth=toString(Width2),
    SQL_Type=string::concat("decimal(",SWidth,",2)"),
    !
;    
    dom::domFix(Domain),
    SWidth=toString(Width),
    SQL_Type=string::concat("decimal(",SWidth,",0)"),
    !
;    
    SQL_Type="text",
    !.
% -------------------------------------------------------------------------------------------------------------
% stTablePop
% -------------------------------------------------------------------------------------------------------------
class predicates
rowCols : (string_list,sterml,string_list) determ (i,i,o).
clauses

stTablePop(Con,St,Format,Domains,TableName) :-
    Format=1,
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    dbBase::stSentence_nd(St,Lt,_),
        nl::conSentence(Lt,S),
        Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
        sqlTools::executeDirect(Con,Sql),
        fail
;
    Format=9,
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    dbBase::stSentence_nd(St,Lt,_),
        nl::cprTerm(Lt,S),
        Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
        sqlTools::executeDirect(Con,Sql),
        fail
;
    Format=10,
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    dbBase::stSentence_nd(St,Lt,_),
        nl::cprTerm(Lt,Pr),
        transpr::proPR_Term(Pr,Term),
        xmlP3::xmlFromSentence("NL","PUBLICATIONS",0,Term,S),
        Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
        sqlTools::executeDirect(Con,Sql),
        fail
;
    Format=11,
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    dbBase::stSentence_nd(St,Lt,_),
        nl::cprTerm(Lt,Pr),
        transpr::proPR_Term(Pr,Term),
        xmlP3::xmlFromSentence("PR","PUBLICATIONS",0,Term,S),
        Sql=string::concatList(["insert into ",TableName," values ('",S,"','",Time,"')"]),
        sqlTools::executeDirect(Con,Sql),
        fail
;
    Format=8,
    sx::sxUnpack(St,'|',L1),
    L1=[_,_,_],
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    dbBase::stSentence_nd(St,Lt,_),
        Lt=cmp(_,[cmp(_,V1),cmp(_,V2)|V3]),
        set::conc([V1,V2,V3],Vs),
        rowCols(Domains,Vs,L2),
        sx::sxlJoin(L2,",",S),
        Sql=string::concatList(["insert into ",TableName," values (",S,",'",Time,"')"]),
        sqlTools::executeDirect(Con,Sql),
        fail
;
    Format=8,
    sx::sxUnpack(St,'|',L1),
    L1=[_,_],
	datetime::tsNow(Ts),
    datetime::tsSQL(Ts,Time),
    dbBase::stSentence_nd(St,Lt,_),
        Lt=cmp(_,[cmp(_,V1)|V3]),
        set::conc([V1,V3],Vs),
        rowCols(Domains,Vs,L2),
        sx::sxlJoin(L2,",",S),
        Sql=string::concatList(["insert into ",TableName," values (",S,",'",Time,"')"]),
        sqlTools::executeDirect(Con,Sql),
        fail
;
    !.
    
rowCols(_,[],[]) :-
    !.
rowCols([Dom|Doms],[V|Vs],[N|Ns]) :-
    dom::domNum(Dom),
    nl::conValX(V,Dom,"",N),
    rowCols(Doms,Vs,Ns),
    !
;
    nl::conValX(V,Dom,"",S),
    sx::sxUnquote(S,Su),
    N=string::concat("'",Su,"'"),
    rowCols(Doms,Vs,Ns),
    !.
end implement sqlSt
