/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
/*
chain "TM" (one record per sol. set)
      domain STERM format: cmp("TM",
                              [int(SOLREF),
                               int(NTABLES),
                               str(PREDKB),
                               str(PREDNL),
                               int(NENTARGS),
                               int(NARGS)]
      indexes: TM1:SOLREF (un)
	
chain "TH" (one record per table)
	domain STERM format: cmp("TH"
	                        [int(SOLREF),
	                         int(TABNO),
	                         int(NCOLS),
	                         int(NROWS),
	                         str(ST),
	                         int(SORTED)        1=table is already sorted, 0=not (yet) sorted
	                         int(ASCENDING)  1=ascending, -1=descending (only relevant of SORTED=1)
	                         int(WRAPPED)]     1=column headers are wrapped, else -1
	indexes: TH1:SOLREF|TABNO
	              TH2:SOLREF|ST
	
chain "TA" (one record per table argument)
	domain STERM format: cmp("TA"
	                        [int(SOLREF),
	                         int(TABNO),
	                         int(ARGNO),
	                         str(ARGKB),
	                         str(ARGNL),
	                         int(NDES)]
	indexes: TA1:SOLREF|TABNO|ARGNO
		                          
chain "TD" (one record per table column)
	domain STERM format: cmp("TD"
	                        [int(SOLREF),
	                         int(TABNO),
	                         int(ARGNO),
	                         int(DENO),
	                         int(COLNO),
	                         str(DEKB),
	                         str(DENL),
	                         str(DOMAIN),
	                         str(FORMAT),
	                         int(WIDTH),
	                         int(PIXELS),
	                         int(DISPLAY),    1=column id displayed, else -1
	                         int(SORTNO)]    0=not sort key, else 1 or 2 .. or NCOLS
	indexes: TD1:SOLREF|TABNO|ARGNO|DENO
		         TD2:SOLREF|TABNO|COLNO
                               	       
chain "Sn..Tm.." (one record per table cell - one chain per table)
	domain STERM format: cmp("T",
	                        [int(ROWNO),
	                         int(COLNO),
	                         str(VALNL),
	                         VAL])
	indexes: Sn..Tm.. ROWNO|COLNO
if a table is sorted:
---------------------
chain "USn..Tm.." (one record per table row)
	domain STERM format cmp("",[int(PHYSROW),int(LOGROW)])
	indexes: USn..Tm.. VALNLkVALNLm (the values are padded with blanks according to format)
	             XSn..Tm.. LOGROW
	         
	To find the physical row no corresponding to the row logical sequence no N:
	stbGetRowL_P
*/
implement stb
open core,termDomain,globalDomains
constants
    className = "com/rudstrand/stb/stb".
    classVersion = "1.0".
clauses
classInfo(className, classVersion).
class predicates
stbCreateIx : (integer_list,integer,integer) determ.
stbDeleteTables : (integer,integer) determ.
stbSortDelete : (integer,integer,integer) determ.
tbwD_Headers : (integer,integer,integer,integer,string_list,integer_list,integer_list,integer_list) procedure (i,i,i,i,o,o,o,o).
clauses
% -----------------------------------------------------------------------------------------------
% foralltable
% -----------------------------------------------------------------------------------------------
class predicates
foralltableX : (integer,integer,integer,integer,integer) nondeterm (i,i,i,o,o).
clauses
% returns the table row numbers non-deterministically
%
foralltable(SolRef,Table,LRowNo,PRowNo) :-
	stbIsSorted(SolRef,Table,Sorted),
	foralltableX(Sorted,SolRef,Table,LRowNo,PRowNo).
	
foralltableX(0,SolRef,Table,LRowNo,PRowNo) :- 
	foralltbl(SolRef,Table,LRowNo),
	PRowNo=LRowNo.
foralltableX(1,SolRef,Table,LRowNo,PRowNo) :-
	foralltblsorted(SolRef,Table,LRowNo,PRowNo).
foralltableX(-1,SolRef,Table,LRowNo,PRowNo) :- 
	foralltblsorteddec(SolRef,Table,LRowNo,PRowNo).
% -----------------------------------------------------------------------------------------------
% foralltbl
% -----------------------------------------------------------------------------------------------
% returns the table row numbers non-deterministically - can handle (i.e. skip) logically deleted rows
%
foralltbl(SolRef,Table,RowNo) :-
	stbH_Get(SolRef,Table,_,NRows,_,_,_,_),
	RowNo=programControl::sequence_nd(1,NRows,1),
    stbCellGetF(SolRef,Table,RowNo,1,_,_).
% -----------------------------------------------------------------------------------------------
% foralltblsorted/dec
% -----------------------------------------------------------------------------------------------
% returns the table row numbers non-deterministically
%
foralltblsorted(SolRef,Table,LogNo,PhysNo) :-
	sdb::tabName(SolRef,Table,TabName),
	XN=string::concat("X",TabName),
	XH=sdb::getIx(XN),
	stbH_Get(SolRef,Table,_,NRows,_,_,_,_),
	LogNo=programControl::sequence_nd(1,NRows,1),
	   stbGetRowL_P_F(XH,LogNo,PhysNo).

% returns the table row numbers non-deterministically (descending)
%
foralltblsorteddec(SolRef,Table,LogNo,PhysNo) :-
	sdb::tabName(SolRef,Table,TabName),
	XN=string::concat("X",TabName),
	XH=sdb::getIx(XN),
	stbH_Get(SolRef,Table,_,NRows,_,_,_,_),
	Log=programControl::sequence_nd(NRows,1,-1),
	   stbGetRowL_P_F(XH,Log,PhysNo),
	   LogNo=NRows+1-Log.
% --------------------------------------------------------------------------
% stbA_Get
% --------------------------------------------------------------------------
stbA_Get(SolRef,Table,ArgNo,ArgKB,ArgNL,NDEs) :-
    SDB=sdb::get(),
    TA1=sdb::getIx("TA1"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ArgNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
	SDB:key_search(TA1,Key,Ref),
	SDB:ref_term(Ref,ARec),
	ARec=cmp(_,[_,_,_,str(ArgKB),str(ArgNL),int(NDEs)]),!.
stbA_Get(SolRef,Table,ArgNo,"","",0) :-
	dlgIF::feil("stbA_Get",[],[SolRef,Table,ArgNo],nill),!.	
% --------------------------------------------------------------------------
% stbA_Insert
% --------------------------------------------------------------------------
stbA_Insert(SolRef,TabNo,ArgNo,ArgKB,ArgNL,NDEs) :-
    SDB=sdb::get(),
    TA1=sdb::getIx("TA1"),
	S1=toString(SolRef),
	S2=toString(TabNo),
	S3=toString(ArgNo),
	ARec=cmp("TA",[int(SolRef),int(TabNo),int(ArgNo),str(ArgKB),str(ArgNL),int(NDEs)]),
	SDB:chain_insertz("TA",ARec,Ref),
	TA1=sdb::getIx("TA1"),
	sx::concat([S1,"|",S2,"|",S3],Key),
	SDB:key_insert(TA1,Key,Ref),
	!
;
	dlgIF::feil("stbA_Insert",[ArgKB,ArgNL],[SolRef,TabNo,ArgNo,NDEs],nill),!,fail.	
% --------------------------------------------------------------------------
% stbCellGet/F
% --------------------------------------------------------------------------
stbCellGetF(SolRef,Table,RowNo,ColNo,ValNL,Val) :-
	sdb::tabName(SolRef,Table,Chain),
	IX=sdb::getIx(Chain),
	S3=toString(RowNo),
	S4=toString(ColNo),
	sx::concat([S3,"|",S4],Key),
    SDB=sdb::get(),
	SDB:key_search(Ix,Key,Ref),
	SDB:ref_term(Ref,Cell),
	Cell=cmp(_,[_,_,str(ValNL),Val]),
	!.
stbCellGet(SolRef,Table,RowNo,ColNo,ValNL,Val) :-
    stbCellGetF(SolRef,Table,RowNo,ColNo,ValNL,Val),
    !.    
stbCellGet(SolRef,Table,RowNo,ColNo,"",nill) :-
	S=toString(RowNo),
	dlgIF::feil("stbCellGet",[S],[SolRef,Table,ColNo],nill),
	!.	
% --------------------------------------------------------------------------
% stbCellInsert
% --------------------------------------------------------------------------
stbCellInsert(SolRef,Table,RowNo,ColNo,ValNL,Val) :-
	sdb::tabName(SolRef,Table,Chain),
    SDB=sdb::get(),
	SDB:chain_first(Chain,_),
	Rec=cmp("T",[int(RowNo),int(ColNo),str(ValNL),Val]),
	SDB:chain_insertz(Chain,Rec,Ref),
	Ix=sdb::getIx(Chain),
	S3=toString(RowNo),
	S4=toString(ColNo),
	sx::concat([S3,"|",S4],Key),
	SDB:key_insert(Ix,Key,Ref),
	!
;		
	sdb::tabName(SolRef,Table,Chain),
    SDB=sdb::get(),
	not(SDB:chain_first(Chain,_)),
	Rec=cmp("T",[int(RowNo),int(ColNo),str(ValNL),Val]),
	SDB:chain_insertz(Chain,Rec,Ref),
	% creates a table index called SnTm, where n is a solution ref. and m is a table no
	SDB:bt_create(Chain,Ix,15,4,0),
	sdb::dropIx(Chain),
	sdb::putIx(Chain,Ix),
	S3=toString(RowNo),
	S4=toString(ColNo),
	sx::concat([S3,"|",S4],Key),
	SDB:key_insert(Ix,Key,Ref),
	!
;
	S=toString(RowNo),
	dlgIF::feil("stbCellInsert",[S,ValNL],[SolRef,Table,ColNo],nill),!,fail.	
% --------------------------------------------------------------------------
% stbD_GetArg
% --------------------------------------------------------------------------
stbD_GetArg(SolRef,Table,ArgNo,DE_No,ColNo,DE_KB,DE_NL,Domain,Form,Width,Pixels,Display,Sorted) :-
	TD1=sdb::getIx("TD1"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ArgNo),
	S4=toString(DE_No),
	sx::concat([S1,"|",S2,"|",S3,"|",S4],Key),
    SDB=sdb::get(),
	SDB:key_search(TD1,Key,Ref),
	SDB:ref_term(Ref,DRec),
	DRec=cmp(_,[_,_,_,_,int(ColNo),str(DE_KB),str(DE_NL),str(Domain),str(Form),int(Width),int(Pixels),int(Display),int(Sorted)]),!
;
	dlgIF::feil("stbD_GetArg",[],[SolRef,Table,ArgNo],nill),!,fail.	
% --------------------------------------------------------------------------
% stbD_GetCol
% --------------------------------------------------------------------------
stbD_GetCol(SolRef,Table,ColNo,ArgNo,DE_No,DE_KB,DE_NL,Domain,Form,Width,Pixels,Display,Sorted) :-
	TD2=sdb::getIx("TD2"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
    SDB=sdb::get(),
	SDB:key_search(TD2,Key,Ref),
	SDB:ref_term(Ref,DRec),
	DRec=cmp(_,[_,_,int(ArgNo),int(DE_No),_,str(DE_KB),str(DE_NL),str(Domain),
	            str(Form),int(Width),int(Pixels),int(Display),int(Sorted)]),
	!
;
	dlgIF::feil("stbD_GetCol",[],[SolRef,Table,ColNo],nill),!,fail.	
% --------------------------------------------------------------------------
% stbD_GetColWidths
% --------------------------------------------------------------------------
class predicates
stb_GWC : (integer,integer,integer,integer,integer_list,integer_list) determ (i,i,i,i,o,o).
clauses
stbD_GetColWidths(SolRef,Table,Positions,Pixels) :-
	stbH_Get(SolRef,Table,NCols,_,_,_,_,_),
	stb_GWC(SolRef,Table,1,NCols,Positions,Pixels),
	!.
stb_GWC(_,_,NCol,NCols,[],[]) :-
	NCol>NCols,
	!.
stb_GWC(SolRef,Table,NCol,NCols,[Pos|Positions],[Pix|Pixs]) :-
	stbD_GetCol(SolRef,Table,NCol,_,_,_,_,_,_,Pos,Pix,_,_),
	Next=NCol+1,
	stb_GWC(SolRef,Table,Next,NCols,Positions,Pixs),
	!.
% --------------------------------------------------------------------------
% stbD_Insert
% --------------------------------------------------------------------------
stbD_Insert(SolRef,TabNo,ArgNo,DE_No,ColNo,DE_KB,DE_NL,Domain,Form,Width,Pixels) :-
	DRec=cmp("TD",[int(SolRef),int(TabNo),int(ArgNo),int(DE_No),int(ColNo),str(DE_KB),str(DE_NL),str(Domain),str(Form),int(Width),int(Pixels),int(1),int(0)]),
    SDB=sdb::get(),
	SDB:chain_insertz("TD",DRec,Ref),
	TD1=sdb::getIx("TD1"),
	S1=toString(SolRef),
	S2=toString(TabNo),
	S3=toString(ArgNo),
	S4=toString(DE_No),
	sx::concat([S1,"|",S2,"|",S3,"|",S4],Key1),
	SDB:key_insert(TD1,Key1,Ref),
	TD2=sdb::getIx("TD2"),
	S5=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S5],Key2),
	SDB:key_insert(TD2,Key2,Ref),!	
;
	dlgIF::feil("stbD_Insert",[DE_KB,DE_NL,Domain,Form],[SolRef,TabNo,ArgNo,DE_No,ColNo,Width,Pixels],nill),!,fail.	
% --------------------------------------------------------------------------
% stbD_SetDE_NL
% --------------------------------------------------------------------------
stbD_SetDE_NL(SolRef,Table,ColNo,DE_NL) :-
	TD2=sdb::getIx("TD2"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
    SDB=sdb::get(),
	SDB:key_search(TD2,Key,Ref),
	SDB:ref_term(Ref,DRec),
	DRec=cmp("TD",[_,_,ArgNo,DE_No,_,DE_KB,_,Domain,Form,Width,Pixels,Display,Sort]),
	NRec=cmp("TD",[int(SolRef),int(Table),ArgNo,DE_No,int(ColNo),DE_KB,str(DE_NL),Domain,Form,Width,Pixels,Display,Sort]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("stbD_SetDE_NL",[DE_NL],[SolRef,Table,ColNo],nill),!,fail.	
% --------------------------------------------------------------------------
% stbD_SetSort/V
% --------------------------------------------------------------------------
% parameters:
% 3	in 	sort-comp#
%
stbD_SetSort(SolRef,Table,ColNo,SortNo) :-
	TD2=sdb::getIx("TD2"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
    SDB=sdb::get(),
	SDB:key_search(TD2,Key,Ref),
	SDB:ref_term(Ref,DRec),
	DRec=cmp("TD",[_,_,ArgNo,DE_No,_,DE_KB,DE_NL,Domain,Form,Width,Pix,Display,_]),
	NRec=cmp("TD",[int(SolRef),int(Table),ArgNo,DE_No,int(ColNo),DE_KB,DE_NL,Domain,Form,Width,Pix,Display,int(SortNo)]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("stbD_SetSort",[],[SolRef,Table,ColNo,SortNo],nill),!,fail.	
stbD_SetSortV(SolRef,Table,V,ColNo,SortNo) :-
	TD2=sdb::getIx("TD2"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
    SDB=sdb::get(),
	SDB:key_search(TD2,Key,Ref),
	SDB:ref_term(Ref,DRec),
	DRec=cmp("TD",[_,_,ArgNo,DE_No,_,DE_KB,DE_NL,Domain,Form,Width,Pix,Display,_]),
	DE_KB=str(Vd),
	Vd=V,
	NRec=cmp("TD",[int(SolRef),int(Table),ArgNo,DE_No,int(ColNo),DE_KB,DE_NL,Domain,Form,Width,Pix,Display,int(SortNo)]),
	SDB:term_replace(Ref,NRec),
	!
;
	Next=ColNo+1,
	TD2=sdb::getIx("TD2"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
	SDB=sdb::get(),
	SDB:key_search(TD2,Key,_),
	stbD_SetSortV(SolRef,Table,V,Next,SortNo),
	!
;
	dlgIF::feil("stbD_SetSortV",[V],[SolRef,Table,SortNo],nill),!,fail.	
% --------------------------------------------------------------------------
% stbDeleteAll
% --------------------------------------------------------------------------
% deletes all tables for all solution sets
%
stbDeleteAll () :-
    SDB=sdb::get(),
	SDB:chain_terms("TM",MRec,_),
	MRec=cmp(_,[int(SolRef),int(NTables)|_]),
	stbDeleteTables(SolRef,NTables),
	fail
;
	!.	
% --------------------------------------------------------------------------
% stbDeleteSolAll
% --------------------------------------------------------------------------
% deletes all tables for a solution set
%
stbDeleteSolAll(SolRef) :-
	stbM_GetF(SolRef,NTables,_,_,_,_),
	stbDeleteTables(SolRef,NTables),!
;
	!.	
% --------------------------------------------------------------------------
% stbDeleteTable
% --------------------------------------------------------------------------
class predicates
stbDelTabArgs : (integer,integer,integer) determ.
stbDelTabCols : (integer,integer,integer) determ.
clauses
% deletes a table 
%
stbDeleteTable(SolRef,Table) :-
    SDB=sdb::get(),
	stbH_Get(SolRef,Table,NCols,_,ST,Sorted,_,_),
    %	
	TM1=sdb::getIx("TM1"),
	S1=toString(SolRef),
	SDB:key_search(TM1,S1,Ref1),
	SDB:ref_term(Ref1,MRec),
	MRec=cmp(_,[_,_,_,_,_,int(NArgs)]),
	%
	TH1=sdb::getIx("TH1"),
	S2=toString(Table),
	sx::concat([S1,"|",S2],Key),
	SDB:key_search(TH1,Key,Ref2),
	SDB:key_delete(TH1,Key,Ref2),
	%
	TH2=sdb::getIx("TH2"),
	sx::concat([S1,"|",ST],KeyTH2),
	SDB:key_search(TH2,KeyTH2,Ref3),
	SDB:key_delete(TH2,KeyTH2,Ref3),
	%
	SDB:term_delete("TH",Ref2),
	%
	stbDelTabArgs(SolRef,Table,NArgs),
	stbDelTabCols(SolRef,Table,NCols),	
	%
	sdb::tabName(SolRef,Table,Chain),
	Ix=sdb::getIx(Chain),
	SDB:bt_close(Ix),
	SDB:bt_delete(Chain),
	SDB:chain_delete(Chain),
	stbM_DecrNTables(SolRef),
			tt::tt_s(2,"deleted table",Chain),
	stbSortDelete(Sorted,SolRef,Table),
	!
;
	!.
		
stbDelTabArgs(_,_,0) :- 
	!.
stbDelTabArgs(SolRef,Table,ArgNo) :-
    SDB=sdb::get(),
	TA1=sdb::getIx("TA1"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ArgNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
	SDB:key_search(TA1,Key,Ref),
	SDB:key_delete(TA1,Key,Ref),
	SDB:term_delete("TA",Ref),
	Next=ArgNo-1,
	stbDelTabArgs(SolRef,Table,Next),
	!
;
	dlgIF::feil("stbDelTabArgs",[],[SolRef,Table,ArgNo],nill),!,fail.	
stbDelTabCols(_,_,0) :- !.	
stbDelTabCols(SolRef,Table,ColNo) :-	
	TD2=sdb::getIx("TD2"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
	SDB=sdb::get(),
	SDB:key_search(TD2,Key,Ref),
	SDB:key_delete(TD2,Key,Ref),
	SDB:ref_term(Ref,DRec),
	DRec=cmp(_,[_,_,int(ArgNo),int(DE_No)|_]),
	SDB:term_delete("TD",Ref),
	%
	TD1=sdb::getIx("TD1"),
	S4=toString(ArgNo),
	S5=toString(DE_No),
	sx::concat([S1,"|",S2,"|",S4,"|",S5],Key2),
	SDB:key_search(TD1,Key2,Ref2),
	SDB:key_delete(TD1,Key2,Ref2),
	%
	Next=ColNo-1,	
	stbDelTabCols(SolRef,Table,Next),!
;
	dlgIF::feil("stbDelTabCols",[],[SolRef,Table,ColNo],nill),!,fail.	
% --------------------------------------------------------------------------
% stbDeleteTables
% --------------------------------------------------------------------------
% deletes tables 1 to N in a solution set
%
stbDeleteTables(SolRef,0) :- 
	TM1=sdb::getIx("TM1"),
	S1=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(TM1,S1,Ref),
	SDB:key_delete(TM1,S1,Ref),
	SDB:term_delete("TM",Ref),
	!
;
	dlgIF::feil("stbDeleteTables",[],[SolRef],nill),!,fail.	
stbDeleteTables(SolRef,Table) :-
	stbDeleteTable(SolRef,Table),
	Next=Table-1,
	stbDeleteTables(SolRef,Next),
	!.
% --------------------------------------------------------------------------
% stbDisplaySet
% --------------------------------------------------------------------------
stbDisplaySet(SolRef,Table,ColNo,Width,Pixels,Display,Sort) :-
	TD2=sdb::getIx("TD2"),
	S1=toString(SolRef),
	S2=toString(Table),
	S3=toString(ColNo),
	sx::concat([S1,"|",S2,"|",S3],Key),
	SDB=sdb::get(),
	SDB:key_search(TD2,Key,Ref),
	SDB:ref_term(Ref,DRec),
	DRec=cmp(_,[_,_,ArgNo,DE_No,_,DE_KB,DE_NL,Domain,Form,_,_,_,_]),
	NRec=cmp("TD",[int(SolRef),int(Table),ArgNo,DE_No,int(ColNo),DE_KB,DE_NL,Domain,Form,int(Width),int(Pixels),int(Display),int(Sort)]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("stbDisplaySet",[],[SolRef,Table,ColNo,Width,Pixels,Display,Sort],nill),!,fail.	
% --------------------------------------------------------------------------
% stbGetRowL_P/Z; stbGetRowL_P_F
% --------------------------------------------------------------------------
% parameters:
% 1	in	index handle for access table
% 2	in	logical row no
% 3	out	physical row no
%
stbGetRowL_P(XH,LRow,PRow) :-
	stbGetRowL_P_F(XH,LRow,PRow),
	!
;	
	dlgIF::feil("stbGetRowL_P",[],[LRow],nill),!,fail.
stbGetRowL_P_F(XH,LRow,PRow) :-
	Key=toString(LRow),
	SDB=sdb::get(),
	SDB:key_search(XH,Key,Ref),
	SDB:ref_term(Ref,Rec),
	Rec=cmp(_,[int(PRow),_]),
	!.
% returns physical row 0 for logical row 0 (none are existing)
%
stbGetRowL_PZ(_,0,0) :-
	!.
stbGetRowL_PZ(XH,L,P) :-
	stbGetRowL_P(XH,L,P),
	!.
% --------------------------------------------------------------------------
% stbH_Get
% --------------------------------------------------------------------------
stbH_Get(SolRef,Table,NCols,NRows,ST,Sorted,Asc,Wrap) :-
	TH1=sdb::getIx("TH1"),
	S1=toString(SolRef),
	S2=toString(Table),
	sx::concat([S1,"|",S2],Key),
	SDB=sdb::get(),
	SDB:key_search(TH1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[_,_,int(NCols),int(NRows),str(ST),int(Sorted),int(Asc),int(Wrap)]),
	!.
stbH_Get(SolRef,Table,0,0,"",0,1,0) :-
	dlgIF::feil("stbH_Get",[],[SolRef,Table],nill),!.
% --------------------------------------------------------------------------
% stbH_IncRows
% --------------------------------------------------------------------------
stbH_IncRows(SolRef,Table,IncRow) :-
	SDB=sdb::get(),
	TH1=sdb::getIx("TH1"),
	S1=toString(SolRef),
	S2=toString(Table),
	sx::concat([S1,"|",S2],Key),
	SDB:key_search(TH1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp("TH",[int(SolRef),int(Table),NCol,int(NRow)|Tail]),
	IncRow=NRow+1,
	HNew=cmp("TH",[int(SolRef),int(Table),NCol,int(IncRow)|Tail]),
	SDB:term_replace(Ref,HNew),
	!
;
	dlgIF::feil("stbH_IncRows",[],[SolRef,Table],nill),!,fail.	
% --------------------------------------------------------------------------
% stbH_Insert
% --------------------------------------------------------------------------
stbH_Insert(SolRef,Table,NCols,NRows,ST) :-
	HRec=cmp("TH",[int(SolRef),int(Table),int(NCols),int(NRows),str(ST),int(0),int(1),int(-1)]),
	SDB=sdb::get(),
	SDB:chain_insertz("TH",HRec,Ref),
	TH1=sdb::getIx("TH1"),
	S1=toString(SolRef),
	S2=toString(Table),
	sx::concat([S1,"|",S2],Key),
	SDB:key_insert(TH1,Key,Ref),
	TH2=sdb::getIx("TH2"),
	sx::concat([S1,"|",ST],Key2),
	SDB:key_insert(TH2,Key2,Ref),
			tt::tt_s(2,"initialized table",Key),!
;
	dlgIF::feil("stbH_Insert",[ST],[SolRef,Table,NCols],nill),!,fail.	
% --------------------------------------------------------------------------
% stbH_SetNRows
% --------------------------------------------------------------------------
stbH_SetNRows(SolRef,Table,NSol) :-
	SDB=sdb::get(),
	TH1=sdb::getIx("TH1"),
	S1=toString(SolRef),
	S2=toString(Table),
	sx::concat([S1,"|",S2],Key),
	SDB:key_search(TH1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp("TH",[int(SolRef),int(Table),NCol,_,ST,Sorted,Asc,Wrap]),
	HNew=cmp("TH",[int(SolRef),int(Table),NCol,int(NSol),ST,Sorted,Asc,Wrap]),
	SDB:term_replace(Ref,HNew),
	!
;
	dlgIF::feil("stbH_SetNRows",[],[SolRef,Table],nill),!,fail.	
% --------------------------------------------------------------------------
% stbM_DecrNTables
% --------------------------------------------------------------------------
stbM_DecrNTables(SolRef) :-	
	TM1=sdb::getIx("TM1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(TM1,Key,Ref),
	SDB:ref_term(Ref,MRec),
	MRec=cmp(C,[F,int(M)|Tail]),
	N=M-1,
	NRec=cmp(C,[F,int(N)|Tail]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("stbM_DecrNTables",[],[SolRef],nill),!.	
% --------------------------------------------------------------------------
% stbM_Get; stbM_GetF
% --------------------------------------------------------------------------
stbM_GetF(SolRef,NTables,PredKB,PredNL,NEntArgs,NArgs) :-	
	TM1=sdb::getIx("TM1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(TM1,Key,Ref),
	SDB:ref_term(Ref,MRec),
	MRec=cmp(_,[_,int(NTables),str(PredKB),str(PredNL),int(NEntArgs),int(NArgs)]),
	!.
stbM_Get(SolRef,NTables,PredKB,PredNL,NEntArgs,NArgs) :-	
	stbM_GetF(SolRef,NTables,PredKB,PredNL,NEntArgs,NArgs),
	!.	
stbM_Get(SolRef,0,"","",0,0) :-	
	dlgIF::feil("stbM_Get",[],[SolRef],nill),
	!.	
% --------------------------------------------------------------------------
% stbM_Insert
% --------------------------------------------------------------------------
stbM_Insert(SolRef,NTables,PredKB,PredNL,NEntArgs,NArgs) :-
	MRec=cmp("TM",[int(SolRef),int(NTables),str(PredKB),str(PredNL),int(NEntArgs),int(NArgs)]),
	SDB=sdb::get(),
	SDB:chain_insertz("TM",MRec,Ref),
	TM1=sdb::getIx("TM1"),
	SSolRef=toString(SolRef),
	SDB:key_insert(TM1,SSolRef,Ref),
			tt::tt_s(2,"solution set's table",SSolRef),
	!
;
	dlgIF::feil("stbM_Insert",[PredKB,PredNL],[SolRef,NTables,NEntArgs,NArgs],nill),!,fail.	
% --------------------------------------------------------------------------
% stbM_SetNTables
% --------------------------------------------------------------------------
stbM_SetNTables(SolRef,NTables) :-	
	TM1=sdb::getIx("TM1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(TM1,Key,Ref),
	SDB:ref_term(Ref,MRec),
	MRec=cmp(C,[F,_|Tail]),
	NRec=cmp(C,[F,int(NTables)|Tail]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("stbM_SetNTables",[],[SolRef,NTables],nill),!,fail.	
% --------------------------------------------------------------------------
% stbRecord
% --------------------------------------------------------------------------
class predicates
stbRecX : (integer,integer,string,string,string,integer,integer,integer,sterm) determ (i,i,i,i,i,i,i,i,o).
stbRecGetVals : (integer,integer,integer,integer,integer,sterml) determ (i,i,i,i,i,o).
clauses

% converts a table row into KB record format (actually the "left term" field)
%	
stbRecord(SolRef,Table,Row,Rec) :-
	stbM_Get(SolRef,_,P,_,_,NArgs),
	stbH_Get(SolRef,Table,_,_,ST,_,_,_),
	dbBase::stComps(ST,Card,_,E1,E2),
	stbRecX(Card,NArgs,P,E1,E2,SolRef,Table,Row,Rec),
	!
;
	dlgIF::feil("stbRecord",[],[SolRef,Table],nill),!,fail.
	
stbRecX(1,1,P,E1,_,SolRef,Table,RowNo,Rec) :-
	stbA_Get(SolRef,Table,1,_,_,NDEs),
	stbRecGetVals(SolRef,Table,RowNo,1,NDEs,Vals),
	Rec=cmp(P,[cmp(E1,Vals)]),
	!.			
stbRecX(1,2,P,E1,_,SolRef,Table,RowNo,Rec) :-
	stbA_Get(SolRef,Table,1,_,_,NDEs1),
	stbRecGetVals(SolRef,Table,RowNo,1,NDEs1,Vals1),
	stbA_Get(SolRef,Table,2,_,_,NDEs2),
	F=NDEs1+1,
	L=NDEs1+NDEs2,
	stbRecGetVals(SolRef,Table,RowNo,F,L,Vals2),
	Rec=cmp(P,[cmp(E1,Vals1)|Vals2]),
	!.			
stbRecX(2,2,P,E1,E2,SolRef,Table,RowNo,Rec) :-
	stbA_Get(SolRef,Table,1,_,_,NDEs1),
	stbRecGetVals(SolRef,Table,RowNo,1,NDEs1,Vals1),
	stbA_Get(SolRef,Table,2,_,_,NDEs2),
	F=NDEs1+1,
	L=NDEs1+NDEs2,
	stbRecGetVals(SolRef,Table,RowNo,F,L,Vals2),
	Rec=cmp(P,[cmp(E1,Vals1),cmp(E2,Vals2)]),
	!.			
stbRecX(2,3,P,E1,E2,SolRef,Table,RowNo,Rec) :-
	stbA_Get(SolRef,Table,1,_,_,NDEs1),
	stbRecGetVals(SolRef,Table,RowNo,1,NDEs1,Vals1),
	stbA_Get(SolRef,Table,2,_,_,NDEs2),
	F=NDEs1+1,
	L=NDEs1+NDEs2,
	stbRecGetVals(SolRef,Table,RowNo,F,L,Vals2),
	stbA_Get(SolRef,Table,3,_,_,NDEs3),
	F2=L+1,
	L2=L+NDEs3,
	stbRecGetVals(SolRef,Table,RowNo,F2,L2,Vals3),
	Rec=cmp(P,[cmp(E1,Vals1),cmp(E2,Vals2)|Vals3]),
	!.			
stbRecGetVals(_,_,_,ColNo,NCols,[]) :- ColNo>NCols,
	!.
stbRecGetVals(SolRef,Table,RowNo,ColNo,NCols,[Val|Vals]) :-
	stbCellGet(SolRef,Table,RowNo,ColNo,_,Val),
	Next=ColNo+1,
	stbRecGetVals(SolRef,Table,RowNo,Next,NCols,Vals),
	!.
% --------------------------------------------------------------------------
% stbRef
% --------------------------------------------------------------------------
% gets the db location for a table row.
% fails if the table row does not correspond to a dba record.
%
stbRef(SolRef,Table,Row,Loc) :-
	stbRecord(SolRef,Table,Row,Rec),
	stbH_Get(SolRef,Table,_,_,ST,_,_,_),
	sx::sxUnpack(ST,'|',L),
	L=[P|_],
    dbBase::chainSentences(P,LT,Loc),
        Rec=LT,
        !.
% --------------------------------------------------------------------------
% stbSort
% --------------------------------------------------------------------------
class predicates
stbSortGet1 : (integer,integer,integer,integer,integer_list) determ (i,i,i,i,o).
stbSortGet2 : (integer,integer,integer_list,integer_list) determ (i,i,i,o).
clauses

% checks if sort keys are defined and replaces old index (if exists) with a new index
% based on the defined sort keys.
%
stbSort(SolRef,Table) :-
	stbH_Get(SolRef,Table,NCols,_,_,Sorted,_,_),
	stbSortDelete(Sorted,SolRef,Table),
	stbSortGet1(SolRef,Table,1,NCols,Cols1),
	stbSortGet2(1,NCols,Cols1,Cols),
	stbCreateIx(Cols,SolRef,Table),
	stbH_SetSort(SolRef,Table,1),
	!.
stbSortGet1(_,_,ColNo,NCols,[]) :- 
	ColNo>NCols,
	!.	
stbSortGet1(SolRef,Table,ColNo,NCols,[Sort|Cols]) :-	
	stbD_GetCol(SolRef,Table,ColNo,_,_,_,_,_,_,_,_,_,Sort),
	Next=ColNo+1,
	stbSortGet1(SolRef,Table,Next,NCols,Cols),
	!.
	
stbSortGet2(KeyNo,NCols,_,[]) :- 
	KeyNo>NCols,
	!.
stbSortGet2(KeyNo,NCols,Keys,[IndexNo|Ixs]) :-
	set::setGet(Keys,IndexNo,KeyNo),
	Next=KeyNo+1,
	stbSortGet2(Next,NCols,Keys,Ixs),
	!.
stbSortGet2(KeyNo,NCols,Keys,Ixs) :-
	Next=KeyNo+1,
	stbSortGet2(Next,NCols,Keys,Ixs),
	!.
% --------------------------------------------------------------------------
% deleteRow
% --------------------------------------------------------------------------
% the purpose of this delete is to make forall-type of row fetching to recognize that the row has been deleted (logically)
%
deleteRow(SolRef,Table,Row) :-
    stbIsSorted(SolRef,Table,Sorted),
    Sorted=0,
    % delete column 1 cell (only)
	sdb::tabName(SolRef,Table,Chain),
	IX=sdb::getIx(Chain),
	S=toString(Row),
	Key=string::concat(S,"|1"),
    DB=sdb::get(),
	DB:key_search(Ix,Key,Ref),
	DB:key_delete(Ix,Key,Ref),
	!
;
    stbSortDelRow(SolRef,Table,Row),
    !.
% --------------------------------------------------------------------------
% stbSortDelRow
% --------------------------------------------------------------------------
% deletes a table row in a sorted table (the associated index and sort table
% row are also deleted)
%
stbSortDelRow(SolRef,Table,LogRow) :-
	% deletes the sort table index
	sdb::tabName(SolRef,Table,TabName),
	Ix=string::concat("X",TabName),
	XI=sdb::getIx(Ix),
    Key=toString(LogRow),
    SDB=sdb::get(),
	SDB:key_search(XI,Key,Ref),
	SDB:key_delete(XI,Key,Ref),
	! % note that the sort-table row as well as the table row are NOT deleted
;
	dlgIF::feil("stbSortDelRow",[],[SolRef,Table],nill),!,fail.	
% --------------------------------------------------------------------------
% stbIsSorted
% --------------------------------------------------------------------------
% parameters:
% 3	out	-1=desc, 0=not sorted, 1=asc
stbIsSorted(SolRef,Table,Sorted) :-
	stbH_Get(SolRef,Table,_,_,_,Sort,Asc,_),
	Sorted=Sort*Asc,
	!.
% --------------------------------------------------------------------------
% stbST
% --------------------------------------------------------------------------
% returns the table no given ST
% fails if no such table exists
%
stbST(SolRef,ST,Table) :-
	TH2=sdb::getIx("TH2"),
	S1=toString(SolRef),
	sx::concat([S1,"|",ST],Key),
	SDB=sdb::get(),
	SDB:key_search(TH2,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[_,int(Table)|_]),
	!.
% -------------------------------------------------------------------------------------------------------------
% stbCreateIx
% -------------------------------------------------------------------------------------------------------------
class predicates
stbCL_Next : (chainDB::bt_selector,integer,integer) determ.
stbCreateLogSeq : (string,chainDB::bt_selector,chainDB::bt_selector,count,integer) determ.
stbCreateLogSeqNext : (chainDB::bt_selector,chainDB::bt_selector,count,integer) determ .
stbInsKey : (integer,integer,integer,integer_list,integer_list,string,chainDB::bt_selector,string_list) determ.
stbKey : (integer,integer,integer,integer_list,integer_list,string_list,string) determ (i,i,i,i,i,i,o).	
stbKeyAdj : (string,integer,integer) procedure (i,i,o).	
stbKeyCL : (integer,integer,integer,string,integer) determ.
stbKeyCL_Tot : (integer,integer,integer_list,integer) determ.
stbKeyLength : (integer_list,integer,integer,integer_list,integer,string_list) determ (i,i,i,o,o,o).
stbKeyNum : (integer,integer,integer,sterm,integer,string,string) determ (i,i,i,i,i,i,o).	
clauses

% creates a user defined table sorting index IF the first parameter is a non-empty list of
% table column no(s) (which are the sort keys).
% 
stbCreateIx([],SolRef,Table) :- 
	stbH_SetSort(SolRef,Table,0),
	!.
stbCreateIx(Keys,SolRef,Table) :-
	stbKeyLength(Keys,SolRef,Table,Widths,Width,Formats),
	stbKeyCL_Tot(SolRef,Table,Keys,Width),
	sdb::tabName(SolRef,Table,Name),
	Ix=string::concat("U",Name),
	% creates a table sort-index called USnTm, where n is a solution ref. and m is a table no
	SDB=sdb::get(),
	SDB:bt_create(Ix,UI,Width,4,1),
	sdb::dropIx(Ix),
	sdb::putIx(Ix,UI),
	stbH_Get(SolRef,Table,_,NRows,_,_,_,_),
	Row=programControl::sequence_nd(1,NRows,1),
       globalVars::get_i("error_flag")=0,
	   stbInsKey(SolRef,Table,Row,Keys,Widths,Ix,UI,Formats),
	   Row=NRows,
	Ix2=string::concat("X",Name),
	% creates a table sort-index called XSnTm, where n is a solution ref. and m is a table no
	SDB:bt_create(Ix2,XI,7,4,1),
	sdb::dropIx(Ix2),
	sdb::putIx(Ix2,XI),
	Cnt2=count::newCounter(0),
	stbCreateLogSeq(Ix,UI,XI,Cnt2,NRows),	   
	stbH_SetSort(SolRef,Table,1),
	!
;
    S1=toString(SolRef),
    S2=toString(Table),
    Mess=string::concat("Error in sorting key creation for table: ",S1,"/",S2,"!"),
	dlgIF::dlgNote(3,Mess),!,fail.

% parameters:
% 1	in	key table columns
% 2	in	solution ref.
% 3	in	table no
% 4	out	length of key components (corr. to par#1)
%		negative values flag numeric key components
% 5	out	total length of key
% 6	out	key component formats 
%		
stbKeyLength([],_,_,[],0,[]) :- 
	!.
stbKeyLength([Col|Cols],SolRef,Table,[Wa|Ws],Width,[Form|Forms]) :-
	stbD_GetCol(SolRef,Table,Col,_,_,DEk,_,Domain,Form,_,_,_,_),
	dbBase::smClassDefF(DEk,"a",_,_,_,_,W),
	stbKeyAdj(Domain,W,Wa),
	stbKeyLength(Cols,SolRef,Table,Ws,Wsum,Forms),
	Waa=math::abs(Wa),
	Width=Waa+Wsum,
	!
;
	stbD_GetCol(SolRef,Table,Col,_,_,_,_,Domain,Form,W,_,_,_),
	stbKeyAdj(Domain,W,Wa),
	stbKeyLength(Cols,SolRef,Table,Ws,Wsum,Forms),
	Waa=math::abs(Wa),
	Width=Waa+Wsum,
	!.
	
% inserts a key (to be built from all components)
%
stbInsKey(SolRef,Table,Row,Keys,Ws,Chain,UI,Formats) :-
	stbKey(SolRef,Table,Row,Keys,Ws,Formats,S),
	SDB=sdb::get(),
	SDB:chain_insertz(Chain,cmp("",[int(Row),int(0)]),Ref),
	SDB:key_insert(UI,S,Ref),
	!.

% builds a key from the components
%
stbKey(_,_,_,[],_,_,"") :- 
	!.
stbKey(SolRef,Table,Row,[Key|Keys],[W|Ws],[Form|Formats],S) :-
	W>0,
	stbCellGet(SolRef,Table,Row,Key,ValNL,_),
	sx::sxR_Pad(ValNL," ",W,KeyVal),	
	stbKeyCL(SolRef,Table,Key,KeyVal,W),
	stbKey(SolRef,Table,Row,Keys,Ws,Formats,KeyVals),
	S=string::concat(KeyVal,KeyVals),
	!
;
	W<0,
	Wa=-W,
	stbCellGet(SolRef,Table,Row,Key,_,ValIF),
	stbKeyNum(SolRef,Table,Key,ValIF,Wa,Form,KeyVal),
	stbKey(SolRef,Table,Row,Keys,Ws,Formats,KeyVals),
	S=string::concat(KeyVal,KeyVals),
	!.
	
% parameters:
% 1	in	key component domain
% 2	in	key width according to format
% 3	out	adjusted key width (negative value flags numeric key)
%	
stbKeyAdj(Domain,W,Wa) :-
	dom::domInt(Domain),
	Wa=-W-1,
	!
;
	dom::domFix(Domain),
	Wa=-W-1,
	!
;
	Wa=W,
	!.	

stbCreateLogSeq(Chain,UI,XI,Cnt,NRows) :-
    SDB=sdb::get(),
	SDB:chain_first(Chain,_),
	SDB:key_first(UI,_),
	stbCreateLogSeqNext(UI,XI,Cnt,NRows),
	!.
	
stbCreateLogSeqNext(UI,XI,Cnt,NRows) :-
    SDB=sdb::get(),
	programControl::repeat,
	   SDB:key_current(UI,_,Ref),
  	   SDB:ref_term(Ref,Rec),
	   Rec=cmp(_,[int(PRow),_]),
	   Cnt:inc(LRow),
	   NRec=cmp("",[int(PRow),int(LRow)]),
	   SDB:term_replace(Ref,NRec),
	   Key=toString(LRow),
	   SDB:key_insert(XI,Key,Ref),
	   stbCL_Next(UI,LRow,NRows),
	   !
;
	dlgIF::dlgNote(3,"Other error when creating sorting key!"),!,fail.
		
stbCL_Next(_,N,N) :- 
    !.
stbCL_Next(UI,_,_) :- 
    SDB=sdb::get(),
    SDB:key_next(UI,_),!,fail.

% to facilitate correct sorting of signed numerics, a leading 1 is added to
% positive numerics, while negative numerics will have a leading 0 followed by the 10-
% complement of the number. note that integer width will be reduced to 11 (if it is bigger) since
% integers can max be 10 (+ the sign indicator)
%
stbKeyNum(SolRef,Table,Key,int(V),W,F,S) :-
	W>11,
	stbKeyNum(SolRef,Table,Key,int(V),11,F,S),!.
stbKeyNum(SolRef,Table,Key,int(V),W,_,S) :-
	V>=0,
	Vs=toString(V),
	W1=W-1,
	sx::sxL_Pad(Vs,"0",W1,S1),
	S=string::concat("1",S1),
	stbKeyCL(SolRef,Table,Key,S,W),
	!
;			
	V<0,
	W1=W-1,
	Va=-V,
	Sa=toString(Va),
	sx::sxL_Pad(Sa,"0",W1,S1),
	sx::sx10C(S1,S2),
	S=string::concat("0",S2),
	stbKeyCL(SolRef,Table,Key,S,W),
	!.
stbKeyNum(SolRef,Table,Key,fix(V),W,Form,S) :-
	V>=0,
	sx::conFormatF(V,Form,Vs),
	W1=W-1,
	sx::sxL_Pad(Vs,"0",W1,S1),
	S=string::concat("1",S1),
	stbKeyCL(SolRef,Table,Key,S,W),
	!
;		
	V<0,
	Va=-V,
	sx::conFormatF(Va,Form,Vs),
	W1=W-1,
	sx::sxL_Pad(Vs,"0",W1,S1),
	sx::sx10C(S1,S2),
	S=string::concat("0",S2),
	stbKeyCL(SolRef,Table,Key,S,W),
	!.
stbKeyNum(SolRef,Table,Key,_,_,_,_) :- 
	stbD_GetCol(SolRef,Table,Key,_,_,_,DE_NL,_,_,_,_,_,_),
	DE=sx::replaceAll(DE_NL,"|",""),
	Mess=string::concatList(["Error when creating sorting key involving data element: ",DE,"!"]),
	dlgIF::dlgNote(3,Mess),!,fail.

stbKeyCL_Tot(_,_,_,Width) :-
	Width<256,
	!
;
	Ws=toString(Width),
	Mess=string::concatList(["Sorting key total length exceeds 255 characters: ",Ws,"!"]),
	dlgIF::dlgNote(3,Mess),!,fail.

stbKeyCL(SolRef,Table,Key,S,W) :-
	string::length(S)<=W,
	!
;	
	stbD_GetCol(SolRef,Table,Key,_,_,_,DE_NL,_,_,_,_,_,_),
	Ws=toString(W),
	DE=sx::replaceAll(DE_NL,"-|",""),
	Mess=string::concatList(["Data element '",DE,"' used in sorting has actual value in excess of format length (",Ws,"):'",S,"'"]),
	dlgIF::dlgNote(3,Mess),!,fail.
% --------------------------------------------------------------------------
% stbH_SetSort
% --------------------------------------------------------------------------
stbH_SetSort(SolRef,Table,Sorted) :-
    SDB=sdb::get(),
	TH1=sdb::getIx("TH1"),
	S1=toString(SolRef),
	S2=toString(Table),
	sx::concat([S1,"|",S2],Key),
	SDB:key_search(TH1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp("TH",[int(SolRef),int(Table),NCol,NSol,ST,_,Asc,Wrap]),
	HNew=cmp("TH",[int(SolRef),int(Table),NCol,NSol,ST,int(Sorted),Asc,Wrap]),
	SDB:term_replace(Ref,HNew),
	!
;
	dlgIF::feil("stbH_SetSort",[],[SolRef,Table,Sorted],nill),!,fail.	
% --------------------------------------------------------------------------
% stbH_SetSortOrder
% --------------------------------------------------------------------------
stbH_SetSortOrder(SolRef,Table,Asc) :-
    SDB=sdb::get(),
	TH1=sdb::getIx("TH1"),
	S1=toString(SolRef),
	S2=toString(Table),
	sx::concat([S1,"|",S2],Key),
	SDB:key_search(TH1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp("TH",[int(SolRef),int(Table),NCol,NSol,ST,Sorted,_,Wrap]),
	HNew=cmp("TH",[int(SolRef),int(Table),NCol,NSol,ST,Sorted,int(Asc),Wrap]),
	SDB:term_replace(Ref,HNew),
	!
;
	dlgIF::feil("stbH_SetSortOrder",[],[SolRef,Table,Asc],nill),!,fail.	
% --------------------------------------------------------------------------
% stbSortDelete
% --------------------------------------------------------------------------
% deletes a KE-defined table index table and chain, i.e. the sort-extension of the basic table
%
stbSortDelete(0,_,_) :- 
	!.
stbSortDelete(_,SolRef,Table) :-
	sdb::tabName(SolRef,Table,TabName),
	Chain=string::concat("U",TabName),
	SDB=sdb::get(),
	SDB:chain_first(Chain,_),
	% sort-table has at least one row
	UXN=string::concat("U",TabName),
	UXH=sdb::getIx(UXN),
	SDB:bt_close(UXH),	
	SDB:bt_delete(UXN),
	sdb::dropIx(UXN,UXH),
	%
	XXN=string::concat("X",TabName),
	XXH=sdb::getIx(XXN),
	SDB:bt_close(XXH),	
	SDB:bt_delete(XXN),
	sdb::dropIx(XXN,XXH),
	SDB:chain_delete(Chain),
		tt::tt_s(2,"deleted sort-table",TabName),
	!.
% --------------------------------------------------------------------------
% reOpen
% --------------------------------------------------------------------------
reOpen(SolRef,TabNo) :-
    guix::getWinAll(Clas,_,W,SolRef,TabNo,_),
    Clas=203,
    _=vpi::winSendEvent(W,vpiDomains::e_CloseRequest()),
    !,
    stbH_Get(SolRef,TabNo,_,_,ST,_,_,_),
    stbDeleteTable(SolRef,TabNo),
    table::generate(SolRef,ST),
    openTableWindow(SolRef,TabNo),
    !
;
    !.
% --------------------------------------------------------------------------
% openWindow
% --------------------------------------------------------------------------
openTableWindow(SolRef,Table) :-
    table::tabSortIfRequired(SolRef,Table), 
	stbH_Get(SolRef,Table,NCols,NRows,STk,_,_,_),
	dbBase::stNL(STk,ST),
	tbwD_Headers(SolRef,Table,1,NCols,ColTitles,Widths,Aligns,ReadOnly),
    W=guix::getWinOfClas(1),
    Wt=tedP3::createTableWinBuff(W,ST,NRows,ReadOnly,ColTitles,Widths,Aligns,[],SolRef,Table,NCols),
    guix::getWinAll(Clas,_,Wt,_,_,N),
    guix::setWin(Clas,ST,Wt,SolRef,Table,N),
    !.
% ---------------------------------------------------------------------------------------------------------------
% tbwD_Headers
% ---------------------------------------------------------------------------------------------------------------
% returns lists of table data
%
tbwD_Headers(SolRef,Table,N,NCols,[NL|NLs],[W|Ws],[1|As],[1|Rs]) :-
	N<=NCols,
	stbD_GetCol(SolRef,Table,N,_,_,_,NL,Dom,_,_,W,_,_),
	dom::domNum(Dom),
	Next=N+1,
	tbwD_Headers(SolRef,Table,Next,NCols,NLs,Ws,As,Rs),
	!.
tbwD_Headers(SolRef,Table,N,NCols,[NL|NLs],[W|Ws],[-1|As],[1|Rs]) :-
	N<=NCols,
	stbD_GetCol(SolRef,Table,N,_,_,_,NL,_,_,_,W,_,_),
	Next=N+1,
	tbwD_Headers(SolRef,Table,Next,NCols,NLs,Ws,As,Rs),
	!.
tbwD_Headers(_,_,_,_,[],[],[],[]) :- 
	!.
% ---------------------------------------------------------------------------------------------------------------
% makeData
% ---------------------------------------------------------------------------------------------------------------
class predicates
tbwPopX : (integer SolRef,integer Table,integer Sorted,integer Row,integer LastRow,integer Increment,chainDB::bt_selector XH,integer NCols,vpiTableed::tableedit_DATA Data) determ (i,i,i,i,i,i,i,i,o).
tbwRow : (integer Sorted,chainDB::bt_selector XI,integer Row,integer PRow) procedure (i,i,i,o).
tbwPopCols : (integer SolRef,integer Table,integer TRow,integer CellCol,integer NCols,vpiTableed::tableedit_COLUMNDATA) determ (i,i,i,i,i,o).
popBlanks : (integer,integer,vpiTableed::tableedit_COLUMNDATA) determ (i,i,o).
clauses

% returns tableed data corresponding to table row F..L
%
makeData(SolRef,Table,Sorted,F,L,NCols,Data) :-
	Sorted=0,
	tbwPopX(SolRef,Table,Sorted,F,L,1,0,NCols,Data),
	!
;
	Sorted=1,
    stbH_Get(SolRef,Table,_,_,_,_,Asc,_),
    Asc=1,
	sdb::tabName(SolRef,Table,TabName),
	XN=string::concat("X",TabName),
	XH=sdb::getIx(XN),
	tbwPopX(SolRef,Table,Sorted,F,L,1,XH,NCols,Data),
	!
;
	Sorted=1,
    stbH_Get(SolRef,Table,_,NRows,_,_,Asc,_),
    Asc=-1,
	sdb::tabName(SolRef,Table,TabName),
	XN=string::concat("X",TabName),
	XH=sdb::getIx(XN),
	Fa=NRows+1-F,
	La=NRows+1-L,
	tbwPopX(SolRef,Table,Sorted,Fa,La,-1,XH,NCols,Data),
	!
;
	dlgIF::feil("tbwPop",[],[SolRef,Table,Sorted,F,L],nill),!,Data=[].
tbwPopX(SolRef,Table,Sorted,Row,LastRow,Inc,XH,NCols,Data) :-
	Row=LastRow,
	tbwRow(Sorted,XH,Row,PRow),
	tbwPopCols(SolRef,Table,PRow,1,NCols,RowData),
	Data=[RowData],
	!
;
	tbwRow(Sorted,XH,Row,PRow),
	tbwPopCols(SolRef,Table,PRow,1,NCols,RowData),
	Next=Row+Inc,
	tbwPopX(SolRef,Table,Sorted,Next,LastRow,Inc,XH,NCols,Rs),
	Data=[RowData|Rs],
	!.
tbwPopCols(_,_,_,CellCol,NCols,[]) :-
    CellCol>NCols,
	!.
tbwPopCols(SolRef,Table,TRow,CellCol,NCols,[CellData|Cs]) :-
    tedP3::physicalRowIsDeleted(SolRef,Table,TRow),
    ted::makeCell(nill,"",CellData),
	Next=CellCol+1,
	popBlanks(Next,NCols,Cs),
    !
;    
    stbCellGetF(SolRef,Table,TRow,CellCol,ValNL,Val),
    ted::makeCell(Val,ValNL,CellData),
	Next=CellCol+1,
	tbwPopCols(SolRef,Table,TRow,Next,NCols,Cs),
	!.
/*
;
    % deleted row
    ted::makeCell(nill,"",CellData),
	Next=CellCol+1,
	popBlanks(Next,NCols,Cs),
	!.
*/
	
popBlanks(Col,NCols,[]) :-
    Col>NCols,
    !.
popBlanks(Col,NCols,[Cell|Cs]) :-
    ted::makeCell(nill,"",Cell),
    Next=Col+1,
    popBlanks(Next,NCols,Cs),
    !.
    
tbwRow(Sorted,XI,Row,PRow) :-	
	Sorted=1,
	stbGetRowL_P(XI,Row,PRow),
	!
;	
	PRow=Row,
	!.
end implement stb