/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement tbq
open core,termDomain

constants
className = "com/rudstrand/tbq/tbq".
classVersion = "1.0".

class predicates
deVar : (string,sterm,sterm) determ (i,i,o).
tbqDERF : (sterm,sterm) determ (i,o).
tbqEntArg : (sterm,sterml) determ (i,o).
tbqExpression : (sterm,sterm) procedure (i,o).
tbqFuncBody : (sterml,sterml) determ (i,o).
tbqFuncCall : (sterm,sterm) determ (i,o).
tbqMakeDE_Vars : (string,sterml,sterml) determ (i,i,o).
tbqMakeCondVars : (string,string,sterml,sterml) determ (i,i,i,o).
tbqOrderDE_Vals : (sterml,sterml,sterml) determ (i,i,o).
tbqPredCond : (sterm,sterm) determ (i,o).
tbqRemoveDummyACs : (sterml,sterml) determ (i,o).
tbqRemoveDummyAC : (sterm,sterm) procedure (i,o).
tbqSSS : (sterm,sterm) determ (i,o).
tbqFuncStatement : (sterm,sterm) determ (i,o).

clauses
classInfo(className, classVersion).
% ------------------------------------------------------------------------------------------------------------
% tbqFunction
% ------------------------------------------------------------------------------------------------------------
tbqFunction(Syn,Tbq) :-
	Syn=cmp("FDEF",
	       [cmp("REL_TYPE",
	           [cmp("REP",[atom(Rep)]),
	            cmp("DOM",[atom(Dom)]),
	            cmp("FPAR",FPars)]),
	        cmp("BODY",Body)]),
	sx::frontstr(1,Rep,F,T),
	Uf=string::toUpperCase(F),
	RetPar=string::concat(Uf,T),
	tbqFuncBody(Body,Body5),
	tbqRemoveDummyACs(Body5,Body6),
	struc::treElimNill(cmp("",Body6),Body7),
	Body7=cmp("",Body8),
	vars::varFuncPars(Rep,Pars),
	Tbq=cmp("FDEF",
	        [cmp("REL_TYPE",
	            [cmp("REP",[atom(Rep)]),
	             cmp("DOM",[atom(Dom)]),
	             cmp("FPAR",FPars)]),
	         cmp("APARS",[fvar(RetPar,Dom)|Pars]),
	         cmp("BODY",Body8)]),
	!
;
	dlgIF::feil("tbqFunction",["TBQ construction failed"],[],nill),
	tt::tt_term(1,"Error syntax tree",Syn),
	!,fail.
% ------------------------------------------------------------------------------------------------------------
% tbqQuery
% ------------------------------------------------------------------------------------------------------------
tbqQuery(Syn,Tbq) :-
	Syn=cmp("DFNC",_),
	tbqFuncCall(Syn,Tbq2),
	tbqRemoveDummyAC(Tbq2,Tbq3),
	struc::treElimNill(Tbq3,Tbq),
	!
;
	Syn=cmp("NFNC",_),
	tbqFuncCall(Syn,Tbq2),
	tbqRemoveDummyAC(Tbq2,Tbq3),
	struc::treElimNill(Tbq3,Tbq),
	!.
tbqQuery(Syn,Tbq) :-
	Syn=cmp("NDERF",_),
	tbqDERF(Syn,Tbq2),
	tbqRemoveDummyAC(Tbq2,Tbq3),
	struc::treElimNill(Tbq3,Tbq),
	!
;
	Syn=cmp("DDERF",_),
	tbqDERF(Syn,Tbq2),
	tbqRemoveDummyAC(Tbq2,Tbq3),
	struc::treElimNill(Tbq3,Tbq),
	!.
tbqQuery(Syn,Tbq) :-
	Syn=cmp("#S",_),
	tbqSSS(Syn,Tbq2),
	tbqRemoveDummyAC(Tbq2,Tbq3),
	struc::treElimNill(Tbq3,Tbq),
	!.
tbqQuery(Syn,cmp("EX",[Tbq])) :-
	Syn=cmp("EX",[PredCond]),
	tbqPredCond(PredCond,Tbq2),
	tbqRemoveDummyAC(Tbq2,Tbq3),
	struc::treElimNill(Tbq3,Tbq),
	!.
tbqQuery(Syn,_) :- 
	dlgIF::feil("tbqQuery",["TBQ construction failed"],[],nill),
	tt::tt_term(1,"Error syntax tree",Syn),
	!,fail.
% ------------------------------------------------------------------------------------------------------------
%                                LOCAL class predicates                            
% ------------------------------------------------------------------------------------------------------------
% deVar
% ------------------------------------------------------------------------------------------------------------
class predicates
getDE_Var : (string,sterm,sterm) determ (i,i,o).
getDE_Vars : (string,sterml,sterm) determ (i,i,o).
argSubTree : (string,sterm,sterm) determ (i,i,o).	
argNoSubTree : (integer,sterm,sterm) determ (i,i,o).	
clauses

% parameters:
% 1	in	DE referenced (IF) formats:
%        DE
%        ET.DE
%        n.DE
%        ET: entity type or predicate type; n: 0/1/2 to signify argument number
% 2	in	a tbq sub-tree where a DE-variable corresponding to par#1 is defined
% 3	out	the DE-var corresponding to par#1
% fails if no DE-var definition is found in par#2
%
deVar(D,Tree,Var) :-
    % n.DE
    sx::sxUnpack(D,'.',L),
    L=[Arg,DE],
    Dig=string::subString(Arg,0,1),
    sx::isDigitS(Dig),
    !,
    ArgNo=sx::toInt(Arg),
    argNoSubTree(ArgNo,Tree,SubTree),
	getDE_Var(DE,SubTree,Var),
	!
;
    % ET.DE
    sx::sxUnpack(D,'.',L),
    L=[Arg,DE],
    argSubTree(Arg,Tree,SubTree),
    !,
	getDE_Var(DE,SubTree,Var),
	!
;
    % DE
	getDE_Var(D,Tree,Var),
	!
;
    Mess=string::concat("Cannot recognize data element '",D,"' in data element reference expression (tbq::deVar)!"),
    dlgIF::dlgNote(1,Mess),!,fail.
    
getDE_Var(DE,cmp(_,DE_Vars),DE_Var) :- 
	getDE_Vars(DE,DE_Vars,DE_Var),
	!.

getDE_Vars(DE,[H|T],DE_Var) :- 
	H=fstr(DEV,Dom,Frm,L),
	vars::varComps(DEV,_,_,DEx),
	DEx=DE,
	DE_Var=fstr(DEV,Dom,Frm,L),
	!
;
	H=cmp(_,L),
	getDE_Vars(DE,L,DE_Var),
	!
;
	getDE_Vars(DE,T,DE_Var),
	!.	

argSubTree(Arg,cmp(_,[cmp(_,[A0|_]),cmp(_,[A1|_]),cmp(_,[A2|_])]),Tree) :-
    A0=cmp(_,[atom(Arg)|_]),
    Tree=A0,
    !
;
    A1=cmp(_,[cmp(_,[atom(Arg)|_])|_]),
    Tree=A1,
    !
;    
    A2=cmp(_,[cmp(_,[atom(Arg)|_])|_]),
    Tree=A2,
    !.    
argSubTree(Arg,cmp(_,[cmp(_,[A0|_]),cmp(_,[A1|_])]),Tree) :-
    A0=cmp(_,[atom(Arg)|_]),
    Tree=A0,
    !
;
    A1=cmp(_,[cmp(_,[atom(Arg)|_])|_]),
    Tree=A1,
    !.    

argNoSubTree(0,cmp(_,[Tree|_]),Tree) :-
    !.
argNoSubTree(1,cmp(_,[_,Tree|_]),Tree) :-
    !.    
argNoSubTree(2,cmp(_,[_,_,Tree]),Tree) :-
    !.    
% ------------------------------------------------------------------------------------------------------------
% tbqDERF
% ------------------------------------------------------------------------------------------------------------
tbqDERF(cmp(Derf,[atom(DE),RelCond]),cmp(Derf,[DE_Var,RelCondM])) :-
	tbqPredCond(RelCond,RelCondM),
	deVar(DE,RelCondM,DE_Var),
	!.
% ------------------------------------------------------------------------------------------------------------
% tbqEntArg
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	entity argument in syntax tree
% 2	out	entity condition in 3bq tree
%
tbqEntArg(var(V),[var(V)]) :- 
	!.
tbqEntArg(Arg,Cond) :-
	% unparameterized entity type
	Arg=cmp("ENT_TYPE",
	       [cmp("REP",[atom(EntRep)]),
	        cmp("ATTRS",Atts)]),
	vars::varGenPrefix(EntRep,Prefix),
	tbqMakeDE_Vars(Prefix,Atts,VAtts),
	C=cmp("ENT_VAR",
             [cmp("REP",[atom(EntRep)]),
              cmp("ATTR_VARS",VAtts)]),
	Cond=[C],
	!
;
	% parameterized entity type
	Arg=cmp("PAR_ENT",
	       [cmp("ENT_VAR",
	           [cmp("ENT_TYPE",
	               [cmp("REP",[atom(EntRep)]),
	                cmp("ATTRS",Atts)]),
	            cmp("ATTR_VARS",AttrVals)])]),
	vars::varGenPrefix(EntRep,Prefix),
	tbqMakeDE_Vars(Prefix,Atts,VAtts),
	tbqOrderDE_Vals(Atts,AttrVals,AttrVals2),
	tbqMakeCondVars(EntRep,Prefix,AttrVals2,VAttrVals),
        Cond=[cmp("ENT_ARG",
                 [cmp("ENT_VAR",
                     [cmp("REP",[atom(EntRep)]),
                      cmp("ATTR_VARS",VAtts)]),
                  cmp("ATTR_CONDS",VAttrVals)])],
        !
;
	% entity value
	Arg=cmp("ENT",
	       [cmp("ENT_TYPE",
	           [cmp(R,[atom(EntRep)]),
	            cmp("ATTRS",Atts)]),
	        cmp("ATTR_VALS",AttConds)]),
	tbqOrderDE_Vals(Atts,AttConds,AttConds2),
	vars::varGenPrefix(EntRep,Prefix),
	tbqMakeDE_Vars(Prefix,Atts,VAtts),
	tbqMakeCondVars(EntRep,Prefix,AttConds2,VAttConds),
        Cond=[cmp("ENT_ARG",
                 [cmp("ENT_VAR",
                     [cmp(R,[atom(EntRep)]),
                      cmp("ATTR_VARS",VAtts)]),
                  cmp("ATTR_CONDS",VAttConds)])],
        !
;
	% entity condition 
	Arg=cmp("ENT_ARG",
	       [cmp("ENT_TYPE",
	           [cmp(R,[atom(EntRep)]),
	            cmp("ATTRS",Atts)]),
	        cmp("ATTR_CONDS",AttConds)]),
	tbqOrderDE_Vals(Atts,AttConds,AttConds2),
	vars::varGenPrefix(EntRep,Prefix),
	tbqMakeDE_Vars(Prefix,Atts,VAtts),
	tbqMakeCondVars(EntRep,Prefix,AttConds2,VAttConds),
        Cond=[cmp("ENT_ARG",
                 [cmp("ENT_VAR",
                     [cmp(R,[atom(EntRep)]),
                      cmp("ATTR_VARS",VAtts)]),
                  cmp("ATTR_CONDS",VAttConds)])],!
;
	% entity dummy condition 
	Arg=cmp("ENT_ARG",
	       [cmp("ENT_TYPE",
	           [cmp(R,[atom(EntRep)]),
	            cmp("ATTRS",Atts)])]),
	vars::varGenPrefix(EntRep,Prefix),
	tbqMakeDE_Vars(Prefix,Atts,VAtts),
        Cond=[cmp("ENT_ARG",
                 [cmp("ENT_VAR",
                     [cmp(R,[atom(EntRep)]),
                      cmp("ATTR_VARS",VAtts)])])],!.
% ------------------------------------------------------------------------------------------------------------
% tbqExpression
% ------------------------------------------------------------------------------------------------------------
tbqExpression(cmp(Opr,[Opn1]),cmp(Opr,[M1])) :-
	opr::opr_expr(Opr,1),
	tbqExpression(Opn1,M1),
	!.
tbqExpression(cmp(Opr,[Opn1,Opn2]),cmp(Opr,[M1,M2])) :-
	opr::opr_expr(Opr,2),
	tbqExpression(Opn1,M1),
	tbqExpression(Opn2,M2),
	!.
tbqExpression(Exp,ExpM) :- 
	tbqFuncCall(Exp,ExpM),
	!.
tbqExpression(Exp,ExpM) :- 
	tbqDERF(Exp,ExpM),
	!.
tbqExpression(Exp,ExpM) :- 
	tbqSSS(Exp,ExpM),
	!.
tbqExpression(Exp,Exp) :- 
	!.
% ------------------------------------------------------------------------------------------------------------
% tbqFuncBody
% ------------------------------------------------------------------------------------------------------------
tbqFuncBody([],[]) :- 
	!.
tbqFuncBody([atom("p3_iNdFn")|T],Tm) :- 
	tbqFuncBody(T,Tm),
	!.
tbqFuncBody([H|T],[Hm|Tm]) :-
	tbqFuncStatement(H,Hm),
	tbqFuncBody(T,Tm),
	!.
% ------------------------------------------------------------------------------------------------------------
% tbqFuncCall
% ------------------------------------------------------------------------------------------------------------
class predicates
tbqOrderPars : (sterml,sterml,sterml) determ (i,i,o).	
tbqFindPar : (string,sterml,sterml,sterm) determ (i,i,o,o).	
clauses
tbqFuncCall(cmp(FTyp,[atom(F),RDom|Pars]),cmp(FTyp,[atom(F),RDom|OPars])) :-
	set::member(FTyp,["DFNC","NFNC"]),
	dbBase::smDefF(F,Clas,_,DEs,_,_,_),
	set::member(Clas,["f","g"]),
	dbBase::smDeDefs(DEs,Defs),
	tbqOrderPars(Defs,Pars,OPars),
	!.
tbqOrderPars([],[],[]) :-
	!.
tbqOrderPars([fstr(DE,_,_,_)|DEs],Pars,[OPar|OPars]) :-
	tbqFindPar(DE,Pars,ParsRem,OPar),
	tbqOrderPars(DEs,ParsRem,OPars),!.
tbqOrderPars([],L,_) :-
	dlgIF::feil("tbqOrderPars",["Excess parameter(s) in function call"],[],nill),
	tt::tt_terms(1,"feil",L),!,fail.
tbqFindPar(DE,[],_,_) :-
	sx::concat(["Missing parameter '",DE,"' in function call"],Mess),
	dlgIF::feil("tbqFindPar",[Mess],[],nill),!,fail.
tbqFindPar(DE,[Par|Pars],Pars,Val) :-
	Par=cmp(Opr,[fvar(DE,_),Val]),
	opr::oprEq(Opr),!.
tbqFindPar(DE,[Par|Pars],[Par|RemPars],Val) :-
	tbqFindPar(DE,Pars,RemPars,Val),!.
% ------------------------------------------------------------------------------------------------------------
% tbqFuncStatement
% ------------------------------------------------------------------------------------------------------------
tbqFuncStatement(cmp("EX",[PredCond]),cmp("EX",[PredCondM])) :- 
	tbqPredCond(PredCond,PredCondM),
	!.
tbqFuncStatement(cmp("FALSE",[St]),cmp("FALSE",[Stm])) :- 
	tbqFuncStatement(St,Stm),
	!.
tbqFuncStatement(cmp("TRACE",[Dom,Exp]),cmp("TRACE",[Dom,ExpM])) :- 
	tbqExpression(Exp,ExpM),
	!.
tbqFuncStatement(cmp(Opr,[Opn1,Opn2]),cmp(Opr,[Opn1,Opn2m])) :- 
	opr::opr_cmpass(Opr),
	tbqExpression(Opn2,Opn2m),
	!.
% ------------------------------------------------------------------------------------------------------------
% tbqMakeDE_Vars
% ------------------------------------------------------------------------------------------------------------
% adds entity prefix to front of attribute names in entity definitions so that they can serve as variables during pie execution.
%
tbqMakeDE_Vars(_,[],[]) :- 
	!.
tbqMakeDE_Vars(Prefix,[fstr(DE,Dom,Frm,W)|T],[fstr(VDE,Dom,Frm,W)|Tm]) :- 
	VDE=string::concat(Prefix,DE),
	tbqMakeDE_Vars(Prefix,T,Tm),
	!.
% ---------------------------------------------------------------------------
% tbqMakeCondVars
% ---------------------------------------------------------------------------
class predicates
tbqMCV_1 : (string,string,sterm,sterm) procedure (i,i,i,o).
tbqMCV_2 : (string,string,sterml,sterml) determ (i,i,i,o).                         
clauses
%
% adds entity prefix to front of attribute names in an attribute condition 
% list, so that they can serve as variables during pie execution.
% parameters:
% 1	in	prefix
% 2	in	attribute conditions (in 'data model order') from syntax tree
% 3	in	variable'ized attribute conditions
%
tbqMakeCondVars(_,_,[],[]) :- !.
tbqMakeCondVars(Ent,Prefix,[cmp(Opr,[fvar(Att,D),RSide])|T],[cmp(Opr,[fvar(VAtt,D),RSideV])|Tm]) :-
	string::isname(Att),  
	VAtt=string::concat(Prefix,Att),
	tbqMCV_1(Ent,Prefix,RSide,RSideV),
	tbqMakeCondVars(Ent,Prefix,T,Tm),!.
	
tbqMCV_1(_,_,cmp("#S",[PredCond]),cmp("#S",[VPredCond])) :-
	tbqPredCond(PredCond,VPredCond),!.
tbqMCV_1(_,_,cmp("DDERF",[DE,PredCond]),cmp("DDERF",[DE,VPredCond])) :-
	tbqPredCond(PredCond,VPredCond),!.
tbqMCV_1(_,_,cmp("NDERF",[DE,PredCond]),cmp("NDERF",[DE,VPredCond])) :-
	tbqPredCond(PredCond,VPredCond),!.
tbqMCV_1(Ent,Prefix,cmp(H,T),cmp(H,Tv)) :- 
	tbqMCV_2(Ent,Prefix,T,Tv),!.
tbqMCV_1(Ent,Prefix,fvar(Att,D),fvar(Var,D)) :-
	sx::searchchar(Att,'.',P),
	N=P-1,
	sx::frontstr(N,Att,Pf,_),
	Pf=Ent,
	sx::frontstr(P,Att,_,AttM),
	Var=string::concat(Prefix,AttM),
	!.
tbqMCV_1(_,_,T,T) :- !.
	
tbqMCV_2(_,_,[],[]) :- !.
tbqMCV_2(Ent,Prefix,[A|As],[V|Vs]) :-
	tbqMCV_1(Ent,Prefix,A,V),
	tbqMCV_2(Ent,Prefix,As,Vs),!. 
% ------------------------------------------------------------------------------------------------------------
% tbqOrderDE_Vals
% ------------------------------------------------------------------------------------------------------------
class predicates
combine_attvals : (sterml,sterml,sterml) determ (i,i,o).
tbqPickAtts : (sterm,sterml,sterml) determ (i,i,o).
clauses

% orders the attribute/value pairs (2. par) to the sequence of the formal
% attribute list (1. par).
% 9.1.96
%

tbqOrderDE_Vals([],_,[]) :- 
    !.
tbqOrderDE_Vals([Att|Atts],AttVals,PickVals) :-
	tbqPickAtts(Att,AttVals,PickVals1),
	tbqOrderDE_Vals(Atts,AttVals,PickVals2),
	combine_attvals(PickVals1,PickVals2,PickVals),
	!.
% puts the first list in front of the second list (lists may be reordered individually)
%
combine_attvals([],S,S) :- 
	!.
combine_attvals([H|T],S,C) :- 
	combine_attvals(T,[H|S],C),
	!.	
	
tbqPickAtts(_,[],[]) :- 
	!.
tbqPickAtts(fstr(Att,D,F,W),[cmp(BOpr,[fvar(Att,D2),Opn])|Vals],[cmp(BOpr,[fvar(Att,D2),Opn])|ValsM]) :-
	opr::opr_any(BOpr,2),
	tbqPickAtts(fstr(Att,D,F,W),Vals,ValsM),
	!
;
	dlgIF::feil("tbqPickAtts",["Expected binary operator",BOpr,Att],[],nill),!,fail.	
tbqPickAtts(fstr(Att,D,F,W),[_|Vals],ValsM) :- 
	tbqPickAtts(fstr(Att,D,F,W),Vals,ValsM),
	!.
% ------------------------------------------------------------------------------------------------------------
% tbqPredCond                                    
% ------------------------------------------------------------------------------------------------------------
tbqPredCond(RelCond,Cond) :-
	% UR
	RelCond=cmp("REL_COND",
	           [cmp("REL_TYPE",
	               [cmp("REP",[atom(Rep)]),
	                cmp("CARD",[int(1)])]),
	            EntArg1]),            
	tbqEntArg(EntArg1,EntCond1),
        Cond=cmp("REL_COND",
                [cmp("REL_VAR",
                    [cmp("REP",[atom(Rep)]),
	             cmp("CARD",[int(1)])])
	        |EntCond1]),!
;                
	% URA w/ attribute conditions
	RelCond=cmp("REL_COND",
	           [cmp("REL_TYPE",
	               [cmp("REP",[atom(Rep)]),
	                cmp("CARD",[int(1)]),
	                cmp("ATTRS",Attrs)]),
	            EntArg1,
	            cmp("ATTR_CONDS",Conds)]),  
	tbqOrderDE_Vals(Attrs,Conds,AttConds),
	vars::varGenPrefix(Rep,Prefix),
	tbqMakeDE_Vars(Prefix,Attrs,VAtts),
	tbqMakeCondVars(Rep,Prefix,AttConds,VAttConds),
	tbqEntArg(EntArg1,EntCond1),
        set::conc([EntCond1,[cmp("ATTR_CONDS",VAttConds)]],L),
        Cond=cmp("REL_COND",
                [cmp("REL_VAR",
                    [cmp("REP",[atom(Rep)]),
	             cmp("CARD",[int(1)]),
	             cmp("ATTR_VARS",VAtts)])
	        |L]),!
;                
	% URA w/out attributes
	RelCond=cmp("REL_COND",
	           [cmp("REL_TYPE",
	               [cmp("REP",[atom(Rep)]),
	                cmp("CARD",[int(1)]),
	                cmp("ATTRS",Attrs)]),
	            EntArg1]),            
	vars::varGenPrefix(Rep,Prefix),
	tbqMakeDE_Vars(Prefix,Attrs,VAtts),
	tbqEntArg(EntArg1,EntCond1),
        Cond=cmp("REL_COND",
                [cmp("REL_VAR",
                    [cmp("REP",[atom(Rep)]),
	             cmp("CARD",[int(1)]),
	             cmp("ATTR_VARS",VAtts)])
	        |EntCond1]),!
;                
	% BR
	RelCond=cmp("REL_COND",
	           [cmp("REL_TYPE",
	               [cmp("REP",[atom(Rep)]),
	                cmp("CARD",[int(2)])]),
	            EntArg1,
	            EntArg2]),            
	tbqEntArg(EntArg1,EntCond1),
	tbqEntArg(EntArg2,EntCond2),
        set::conc([EntCond1,EntCond2],L),
        Cond=cmp("REL_COND",
                [cmp("REL_VAR",
                    [cmp("REP",[atom(Rep)]),
	             cmp("CARD",[int(2)])])
	        |L]),!
;                
	% BRA w/ attribute conditions
	RelCond=cmp("REL_COND",
	           [cmp("REL_TYPE",
	               [cmp("REP",[atom(Rep)]),
	                cmp("CARD",[int(2)]),
	                cmp("ATTRS",Attrs)]),
	            EntArg1,
	            EntArg2,
	            cmp("ATTR_CONDS",Conds)]),            
	tbqOrderDE_Vals(Attrs,Conds,AttConds),
	vars::varGenPrefix(Rep,Prefix),
	tbqMakeDE_Vars(Prefix,Attrs,VAtts),
	tbqMakeCondVars(Rep,Prefix,AttConds,VAttConds),
	tbqEntArg(EntArg1,EntCond1),
	tbqEntArg(EntArg2,EntCond2),
        set::conc([EntCond1,EntCond2,[cmp("ATTR_CONDS",VAttConds)]],L),
        Cond=cmp("REL_COND",
                [cmp("REL_VAR",
                    [cmp("REP",[atom(Rep)]),
	             cmp("CARD",[int(2)]),
	             cmp("ATTR_VARS",VAtts)])
	        |L]),!
;                
	% BRA w/out attributes
	RelCond=cmp("REL_COND",
	           [cmp("REL_TYPE",
	               [cmp("REP",[atom(Rep)]),
	                cmp("CARD",[int(2)]),
	                cmp("ATTRS",Attrs)]),
	            EntArg1,
	            EntArg2]),            
	vars::varGenPrefix(Rep,Prefix),
	tbqMakeDE_Vars(Prefix,Attrs,VAtts),
	tbqEntArg(EntArg1,EntCond1),
	tbqEntArg(EntArg2,EntCond2),
        set::conc([EntCond1,EntCond2],L),
        Cond=cmp("REL_COND",
                [cmp("REL_VAR",
                    [cmp("REP",[atom(Rep)]),
	             cmp("CARD",[int(2)]),
	             cmp("ATTR_VARS",VAtts)])
	        |L]),!.
% ------------------------------------------------------------------------------------------------------------
% tbqRemoveDummyAC/s
% ------------------------------------------------------------------------------------------------------------
class predicates
tbqRACs : (sterml,sterml) determ (i,o).
clauses

tbqRemoveDummyACs([],[]) :-
	!.
tbqRemoveDummyACs([H|T],[Hm|Tm]) :-
	tbqRemoveDummyAC(H,Hm),
	tbqRemoveDummyACs(T,Tm),
	!.
tbqRemoveDummyAC(cmp("ATTR_CONDS",ACs),cmp("ATTR_CONDS",RACs)) :-
	tbqRACs(ACs,RACs),
	set::size(RACs,N),
	N>0,
	!.	
tbqRemoveDummyAC(cmp("ATTR_CONDS",ACs),nill) :-
	tbqRACs(ACs,RACs),
	set::size(RACs,N),
	N=0,
	!.	
tbqRemoveDummyAC(cmp(H,L),cmp(H,Lm)) :-
	tbqRemoveDummyACs(L,Lm),
	!.
tbqRemoveDummyAC(T,T) :-
	!.

tbqRACs([],[]) :-
	!.
tbqRACs([H|T],Tm) :-
	H=cmp(Opr,[_,fvar(V,_)]),
	opr::oprEq(Opr),
	sx::frontstr(6,V,F,_),
	F="Dummy_",
	tbqRACs(T,Tm),
	!.
tbqRACs([H|T],[H|Tm]) :-	
	tbqRACs(T,Tm),
	!.
% ------------------------------------------------------------------------------------------------------------
% tbqSSS
% ------------------------------------------------------------------------------------------------------------
tbqSSS(cmp("#S",[RelCond]),cmp("#S",[RelCondM])) :-
	tbqPredCond(RelCond,RelCondM),
	!.
end implement tbq
