/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement tedP3
open core

constants
className = "com/rudstrand/tedP3/tedP3".
classVersion = "1.0".

class facts
twh : (string_list) determ. % headers
tww : (integer_list) determ. % pixel widths
twa : (integer_list) determ. % alignments (-1:L, 0:C, 1:R)
twr : (integer_list) determ. % read-only flags (0/1)
twb : (string_list) determ. % buttons (possibly empty)
deletedRow : (integer SolNo,integer TableNo,integer RowNo) nondeterm.
tedCount : integer := 0.
solNo : integer :=0.
table : integer :=0.
nRows : integer :=0.

class predicates
buffAnswerFunc : vpiTableed::tableedit_ANSWERDATA_FUNC.               
tab_Callback : vpiTableed::tableedit_CALLBACK .
tableedbuff_class_handler : vpiDomains::ehandler.
tedbuff_window_eh : vpiDomains::ehandler.
createButtons : (vpiDomains::windowHandle,vpiDomains::font,integer,integer,string_list) determ (i,i,i,i,i).
rectangle : (integer,integer,integer_list,string_list,vpiDomains::rct) determ (i,i,i,i,o).
getCols : () -> vpiTableed::tableedit_COLUMNLIST determ.

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% createTableWinBuff
% ----------------------------------------------------------------------------------------------
% buffered
%
createTableWinBuff(TW,Title,Rows,ReadOnly,Headers,Widths,Alignments,Buttons,SolNo,Table,NCols)=W :-
    retractall(twr(_)),
    retractall(twh(_)),
    retractall(tww(_)),
    retractall(twa(_)),
    retractall(twb(_)),
    retractall(deletedRow(SolNo,Table,_)),
    assert(twr(ReadOnly)),
    assert(twh(Headers)),
    assert(tww(Widths)),
    assert(twa(Alignments)),
    assert(twb(Buttons)),
    solNo:=SolNo,
    table:=Table,
    nRows:=Rows,
    N=tedCount,
    M=N+1,
    tedCount:=M,
    Ns=toString(M),
    TableClass=string::concat("tableedbuff_class",Ns),
    vpi::classCreate(TableClass,tableedbuff_class_handler),
    MaxRow=globalVars::get_i("maxrow"),
    set::setMin([MaxRow,Rows],TableRows),   
    rectangle(1,TableRows,Widths,Buttons,R),
    W=vpi::winCreate(vpiDomains::w_TopLevel,R,Title,vpiDomains::noMenu(),TW,
        [vpiDomains::wsf_TitleBar,vpiDomains::wsf_Minimize,vpiDomains::wsf_BlackRect,vpiDomains::wsf_Close,
         vpiDomains::wsf_ClipSiblings,vpiDomains::wsf_SizeBorder],tedbuff_window_eh,0),
    C=guix::succ(W),
   	Font=vpi::fontCreate(vpiDomains::ff_Helvetica,[],8),
  	vpi::winSetFont(W,Font),
  	vpi::winSetFont(C,Font),
  	ted::setColProp(C,NCols,-1),
    !.

% callback pop-up editor in table window
%
tab_Callback(Win,Event):-
    % delete sentence
	Event=vpiTableed::delete_row(Row,_),
	guix::getWinAll(_,_,Win,SolRef,TabNo,_),
	logicalRowIsDeleted(SolRef,TabNo,Row), % deleted row - ignore
    !
;
	Event=vpiTableed::delete_row(Row,_),
	guix::getWinAll(_,_,Win,SolRef,TabNo,_),
    sentence::fetchRow(SolRef,TabNo,Row,_,_,_,RT),
    RT=1,
    dlgIF::dlgNote(1,"Cannot delete derived sentences!"),
	!
;
	Event=vpiTableed::delete_row(Row,_),
	guix::getWinAll(_,_,Win,SolRef,TabNo,_),
    stb::stbIsSorted(SolRef,TabNo,Sorted),
    Sorted=0,
    sentence::fetchRow(SolRef,TabNo,Row,_,_,Loc,RT),
    RT=2,
    dbBase::recordDelete(Loc),
    assert(deletedRow(SolRef,TabNo,Row)),
	!
;
	Event=vpiTableed::delete_row(Row,_),
	guix::getWinAll(_,_,Win,SolRef,TabNo,_),
    stb::stbIsSorted(SolRef,TabNo,Sorted),
    Sorted<>0,
    sentence::fetchRow(SolRef,TabNo,Row,_,_,Loc,RT),
    RT=2,
    dbBase::recordDelete(Loc),
	sdb::tabName(SolRef,TabNo,TabName),
	Ix=string::concat("X",TabName),
	XI=sdb::getIx(Ix),
    stb::stbGetRowL_P(XI,Row,PRow),
    assert(deletedRow(SolRef,TabNo,PRow)),
	!
;
    % open in sentence window
	Event=vpiTableed::copy(Row,_,_,_),
	guix::getWinAll(_,_,Win,SolRef,TabNo,_),
	logicalRowIsDeleted(SolRef,TabNo,Row), % deleted row - ignore
    !
;
	Event=vpiTableed::copy(Row,_,_,_),
	guix::getWinAll(_,_,Win,SolRef,TabNo,_),
	sentence::displayRow(SolRef,TabNo,Row),
	!
;	
    % open blank window
	Event=vpiTableed::insert_row(_,_),
	guix::getWinAll(_,_,Win,SolRef,TabNo,_),
    sentence::openWin(SolRef,TabNo,-1),
	!.
% ----------------------------------------------------------------------------------------------
% createButtons
% ----------------------------------------------------------------------------------------------
createButtons(_,_,_,_,[]) :-
    !.
createButtons(W,Font,No,Y,[B|Bs]) :-
    X=(No-1)*100,
	X2=X+99,
	Y2=Y+19,
	R=vpiDomains::rct(X,Y,X2,Y2),
	Id=20000+No,
	C=vpiDomains::ctl(vpiDomains::wdef(vpiDomains::wc_PushButton,R,B,vpiDomains::u_Pixels),Id,[vpiDomains::wsf_Group]),
    Cw=vpi::winCreateDynControl([C],W),
  	vpi::winSetFont(Cw,Font),
	Next=No+1,
	createButtons(W,Font,Next,Y,Bs),
	!.
% ----------------------------------------------------------------------------------------------
% bufferAnswerFunc
% ----------------------------------------------------------------------------------------------
% retrieving data from Pro/3 table
%
buffAnswerFunc(Win,FromRow,ToRow,Data):-
    guix::getWinAll(_,_,Win,SolRef,Table,_), 
	stb::stbH_Get(SolRef,Table,NCols,NRows,_,Sorted,_,_),
	FromRow<=ToRow,
	FromRow<=NRows,
	set::setMin([ToRow,NRows],TRow),
    stb::makeData(SolRef,Table,Sorted,FromRow,TRow,NCols,Data),
    !. 
% ----------------------------------------------------------------------------------------------
% eventhandler for non-buffered table window with button handling
% ----------------------------------------------------------------------------------------------
ted_window_eh(W,vpiDomains::e_Create(_))=0 :-
    R=vpi::winGetClientRect(W),
    Buttons=ted::getButtons(),
    Buttons=[],
    Def=vpiDomains::wdef(vpiDomains::wc_Custom,R,"",vpiDomains::u_pixels),
    N=ted::getCount(),
    Ns=toString(N),
    TableClass=string::concat("tableed_class",Ns),
    Id=N+30000,
    DefL=[vpiDomains::customctl(Def,TableClass,Id,[])],
    C=vpi::winCreateDynControl(DefL,W),
    guix::addWin(203,"",W,solNo,table),
    guix::addWin(204,"",C,solNo,table),
    !
;    
    Rm=vpi::winGetClientRect(W),
    Rm=vpiDomains::rct(X,Y,X2,Y2),
    Y2m=Y2-20,
    R=vpiDomains::rct(X,Y,X2,Y2m),
    Buttons=ted::getButtons(),
    Y2b=Y2m+1,
   	Font=vpi::fontCreate(vpiDomains::ff_Helvetica,[],8),
    createButtons(W,Font,1,Y2b,Buttons),
    Def=vpiDomains::wdef(vpiDomains::wc_Custom,R,"",vpiDomains::u_pixels),
    N=ted::getCount(),
    Ns=toString(N),
    TableClass=string::concat("tableed_class",Ns),
    Id=N+30000,
    DefL=[vpiDomains::customctl(Def,TableClass,Id,[])],
    C=vpi::winCreateDynControl(DefL,W),
    guix::addWin(203,"",W,solNo,table),
    guix::addWin(204,"",C,solNo,table),
    !.
ted_window_eh(W,vpiDomains::e_CloseRequest())=0 :-
    Wc=guix::succ(W),
    guix::dropWin(W),
    guix::dropWin(Wc),
    vpi::winDestroy(W),
    !;!.
ted_window_eh(W,vpiDomains::e_Control(Button,_,_,_))=0 :-
    tedButton::button(W,Button),
    !.
% ----------------------------------------------------------------------------------------------
% eventhandler for buffered table window with button handling
% ----------------------------------------------------------------------------------------------
tedbuff_window_eh(W,vpiDomains::e_Create(_))=0 :-
    R=vpi::winGetClientRect(W),
    twb(Buttons),
    Buttons=[],
    Def=vpiDomains::wdef(vpiDomains::wc_Custom,R,"",vpiDomains::u_pixels),
    N=tedCount,
    Ns=toString(N),
    TableClass=string::concat("tableedbuff_class",Ns),
    Id=N+30000,
    DefL=[vpiDomains::customctl(Def,TableClass,Id,[])],
    C=vpi::winCreateDynControl(DefL,W),
    guix::addWin(203,"",W,solNo,table),
    guix::addWin(204,"",C,solNo,table),
    !
;    
    Rm=vpi::winGetClientRect(W),
    Rm=vpiDomains::rct(X,Y,X2,Y2),
    Y2m=Y2-20,
    R=vpiDomains::rct(X,Y,X2,Y2m),
    twb(Buttons),
    Y2b=Y2m+1,
   	Font=vpi::fontCreate(vpiDomains::ff_Helvetica,[],8),
    createButtons(W,Font,1,Y2b,Buttons),
    Def=vpiDomains::wdef(vpiDomains::wc_Custom,R,"",vpiDomains::u_pixels),
    N=tedCount,
    Ns=toString(N),
    TableClass=string::concat("tableedbuff_class",Ns),
    Id=N+30000,
    DefL=[vpiDomains::customctl(Def,TableClass,Id,[])],
    C=vpi::winCreateDynControl(DefL,W),
    guix::addWin(203,"",W,solNo,table),
    guix::addWin(204,"",C,solNo,table),
    !.
tedbuff_window_eh(W,vpiDomains::e_CloseRequest())=0 :-
    Wc=guix::succ(W),
    guix::dropWin(W),
    guix::dropWin(Wc),
    vpi::winDestroy(W),
    !;!.
tedbuff_window_eh(W,vpiDomains::e_Control(Button,_,_,_))=0 :-
    tedButton::button(W,Button),
    !.
% ----------------------------------------------------------------------------------------------
% 
% ----------------------------------------------------------------------------------------------
tableedbuff_class_handler(W,vpiDomains::e_Create(_))=0 :-
    Cols=getCols(),
    vpiTableed::tableedit_init(W,[vpiTableed::callback(tab_Callback),vpiTableed::bufferingmode(vpiTableed::buffering(buffAnswerFunc,30,nRows))],Cols,[]),
    !.
tableedbuff_class_handler(W,Event)=0 :-
    vpiTableed::tableedit_handleevent(W,Event),
    !.

getCols()=Cols:-
    twr(ReadOnly),
    twh(Headers),
    tww(Widths),
    twa(Alignments),
    ted::getColsX(ReadOnly,Headers,Widths,Alignments,Cols),
    !.
% ----------------------------------------------------------------------------------------------
% rectangle
% ----------------------------------------------------------------------------------------------
rectangle(Cascade,Rows,Widths,Buttons,R) :-
    editorWindow::getMax(Mx,_),
    guix::getPixelAdjust(_,Dy),
    Ay=((Rows*18+37)*Dy) div 100,
    Rx=set::sum(Widths)+70,
    Ax=set::min([Mx,Rx]),
    Re=editorWindow::getEditRect(Cascade),
    Re=vpiDomains::rct(X,Y,_,_),
    set::size(Buttons,N),
    set::setMin([N,1],Nm), 
    Hx=X+Ax,
    Hy=Y+Ay+Nm*20,
    R=vpiDomains::rct(X,Y,Hx,Hy),
    !.
% ----------------------------------------------------------------------------------------------
% logical/physicalRowIsDeleted
% ----------------------------------------------------------------------------------------------
logicalRowIsDeleted(SolNo,TabNo,RowNo) :- 
    stb::stbIsSorted(SolNo,TabNo,Sorted),
    Sorted=0,
    deletedRow(SolNo,TabNo,RowNo),
    !
;
    stb::stbIsSorted(SolNo,TabNo,Sorted),
    Sorted<>0,
	sdb::tabName(SolNo,TabNo,TabName),
	Ix=string::concat("X",TabName),
	XI=sdb::getIx(Ix),
    stb::stbGetRowL_P(XI,RowNo,PRow),
    deletedRow(SolNo,TabNo,PRow),
    !.

physicalRowIsDeleted(SolNo,TabNo,RowNo) :- 
    deletedRow(SolNo,TabNo,RowNo),
    !.
end implement tedP3
