/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement transnl
open core,termDomain,globalDomains

constants
className = "com/rudstrand/transnl/transnl".
classVersion = "1.0".

clauses
classInfo(className, classVersion).

class predicates
parChkDoubles : (pie_scan::wrds) determ.
parChkOther : (pie_scan::wrds) determ.
scnSimplify : (pie_scan::wrds,pie_scan::wrds) determ (i,o).
scnTextF : (string,string,string,string) determ (i,i,o,o).
scnTextBlockF : (string,string,string_list,string) determ (i,i,o,o).
scnChkDoubles : (pie_scan::tokl).	
scnEnumWords : (integer,pie_scan::wrds,pie_scan::wrds) determ (i,i,o).

clauses
% ------------------------------------------------------------------------------------------------------------
%                                GLOBAL class predicates                            
% ----------------------------------------------------------------------------------------------
% parTranslate	
% ----------------------------------------------------------------------------------------------
class predicates
parTerm : (sterm SyntaxTree,sterm Term,string TermType) determ (i,o,o).
clauses

% translates one sentence, sentence rule, function definition and query? from NL to IF. must be terminated by !/?.
%
parTranslate(NL,Term,TermType) :-
	parSynTree(NL,SynTr),
	parTerm(SynTr,Term,TermType),
	!
;
    Mess=string::concat("Translation of NL '",NL,"' failed!"),
    dlgIF::dlgNote(1,Mess),!,fail.

parTerm(SynTr,PR_Term,TermType) :-
	SynTr=cmp("FDEF",_),
	TermType="f",
	% function
				dlgIF::msgStatus("3BQ-tree construction"),
	tbq::tbqFunction(SynTr,TBQ_tree),
				tt::tt_header(1,"3BQ TREE"),
				tt::tt_term(1,"",TBQ_tree),
	pr::prFunction(TBQ_Tree,PR),
				tt::tt_header(1,"PR FORMAT"),
				tt::tt_s(1,"Function (PR)",PR),
	transpr::proPR_Term(PR,PR_Term),
				tt::tt_header(1,"INTERNAL FORMAT"),
				tt::tt_term(1,"Function",PR_Term),
				dlgIF::msgStatus(""),!
;
	SynTr=cmp("REL_FACT",B),
	TermType="s",
	% sentence
	SynTrM=cmp("REL_COND",B),
				dlgIF::msgStatus("PR generation"),
	pr::prPredicate(SynTrM,PR),				
				tt::tt_s(2,"Sentence (PR)",PR),	
				dlgIF::msgStatus("IF generation"),
	transpr::proPR_Term(PR,PR_Term),
				dlgIF::msgStatus(""),!
;
	% rule	
				dlgIF::msgStatus("rule-tree generation"),
	TermType="r",
	rn::rnlST_RT(SynTr,Rule,SType,_),
				tt::tt_s(2,"Rule type",SType), 
				tt::tt_term(2,"Rule tree",Rule),
	rub::rubValidate(Rule,RuleV),
				tt::tt_term(1,"Rule (validated)",RuleV),	
				dlgIF::msgStatus("PR generation"),
	transrt::rlgPR(RuleV,PRs),
	sx::sxlJoin(PRs,". ",PRm),
	PR=string::concat(PRm,"."),
				tt::tt_header(1,"PR FORMAT"),
				tt::tt_strings(1,"Rule's PRs",PRs),	
				dlgIF::msgStatus("IF generation"),
	transpr::proPR_Term(PR,PR_Term),
				tt::tt_header(1,"INTERNAL FORMAT"),
				tt::tt_term(1,"Rule",PR_Term),
				dlgIF::msgStatus(""),
    !.
% -------------------------------------------------------------------------------------------------------------
% parCompSynTree
% -------------------------------------------------------------------------------------------------------------
% parameteres:
% 1	in	NL text representing a syntax object (without termination character)
% 2	out	syntax tree
%
parCompSynTree(NL,SynTree) :-
	sx::sxStrip(NL,NL_trim),
				tt::tt_s(2,"NL",NL_trim),
	transnl::scnNL(NL_trim,Wrds), % non-deterministic (i.e. many possible scanned word sequences)
	parChkDoubles(Wrds), 
	parChkOther(Wrds),
	syn::synComponentAnalysis(Wrds,SynTree),
				tt::tt_header(2,"SYNTAX TREE"),
				tt::tt_term(2,"",SynTree),
	!.
% -------------------------------------------------------------------------------------------------------------
% parNL_PR
% -------------------------------------------------------------------------------------------------------------
% @@ should be exhanced for sentences etc.
%
parNL_PR(NL,PR) :-
	sx::sxStrip(NL,S2),
	sx::sxTail(1,S2,S3,S4),
	S4="?",
	parSynTree(S3,SynTree),
	tbq::tbqQuery(SynTree,TBQ_Tree),
	pr::prTbqQuery(TBQ_Tree,PR,_),
	!
;
    Mess=string::concat("NL to PR of '",NL,"' failed!"),
	dlgIF::dlgNote(1,Mess),!,fail.
% -------------------------------------------------------------------------------------------------------------
% parSynTree
% -------------------------------------------------------------------------------------------------------------
class predicates
parST : (count,string,sterm) procedure (i,i,o).
clauses
% parameteres:
% 1	in	NL text representing a sentence, function body or query (with termination character)
% 2	out	syntax tree
%
parSynTree(NL,SynTree) :-
    C=count::newCounter(0),
	parST(C,NL,SynTree),
	!.

parST(CInterprets,NL,SynTree) :-
				tt::tt_s(1,"NL",NL),
	sx::sxTail(1,NL,Sent,Term),	
	TermChar=string::subChar(Term,0),			
				dlgIF::msgStatus("Scanning..."),
	scnNL(Sent,Wrds), % non-deterministic (i.e. there could be many possible scanned word sequences)
	globalVars::get_i("quit_flag")=0,
				CInterprets:inc(N),
				Ns=toString(N),
				Mess=string::concat("PARSER OUTPUT #",Ns),
				tt::tt_header(3,Mess),
				tt::tt_wrds(3,"",Wrds),	
	% can give warnings, but never fails			
	parChkDoubles(Wrds), 
	parChkOther(Wrds),
				RS=globalVars::get_i("re_scan"),
				dlgIF::confCont(N,RS,"scans have been done"),
				M=CInterprets:get(),
				Ms=toString(M),
				sx::concat(["Starts syntax analysis of interpretation ",Ms,"."],Mess2),
				dlgIF::msgStatus(Mess2),
				globalVars::set_i("syn_max",0),
	syn::synAnalysis(TermChar,Wrds,SynTree),
				sx::concat(["Syntax analysis of interpretation ",Ms," succeeded."],Mess3),
				dlgIF::msgStatus(""),
				dlgIF::msgS(Mess3),
				tt::tt_header(1,"SYNTAX TREE"),
				tt::tt_term(1,"",SynTree),
	!
;
				CInterprets:get()=0,
				dlgIF::dlgExp(15),
				Mess=string::concatList(["Text '",NL,"' could not be successfully scanned!"]),
				dlgIF::dlgNote(1,Mess),
	SynTree=nill,
	!
;
				dlgIF::dlgExp(16),
				N=CInterprets:get(),
				S=toString(N),
				Mess=string::concatList(["Syntax analysis of NL '",NL,"' given up on/after interpretation #",S,"!"]),
				dlgIF::dlgNote(1,Mess),
	SynTree=nill,
	!.
% -------------------------------------------------------------------------------------------------------------
% parParameters
% -------------------------------------------------------------------------------------------------------------
class predicates
parSP_Step2 : (string,sterml,sterml) nondeterm (i,i,o).
parSP_Step3 : (pie_scan::wrds,sterml,sterml) determ (i,i,o).
clauses
% translates a parameter-list (ie. from a call) to a syntax tree
% 
parParameters(NL,DEs,ParList) :-
	sx::sxStrip(NL,NL_trim),
	parSP_Step2(NL_trim,DEs,ParList),
	!.
parSP_Step2(Sent,DEs,ParList) :-
	scnNL(Sent,Wrds),
	parChkDoubles(Wrds),
	parChkOther(Wrds),
	parSP_Step3(Wrds,DEs,ParList)
;
	!,fail.
parSP_Step3(_,_,[]) :- 
	globalVars::get_i("quit_flag")=1,
	!.
parSP_Step3(T,DEs,ParList) :-
	globalVars::get_i("quit_flag")=0,
		syn::sbTokenOR_Opt([pie_scan::w(0,"p3_iGivn","@","@")],T,T4),
		syn::sbTokenOR_Opt([pie_scan::w(0,"[","@","@")],T4,T5),
	syn::synDE_DefinitionList(1,DEs,T5,T6,ParList),
		syn::sbTokenOR_Opt([pie_scan::w(0,"]","@","@")],T6,Tr),
	Tr=[],
	!.
% ------------------------------------------------------------------------------------------------------------
% parChkDoubles                                                              
% ------------------------------------------------------------------------------------------------------------
% simply warns of possible repeated words (excluding parenthesises) - never fails
%
parChkDoubles([]) :- 
	!.
parChkDoubles([T|Words]) :-
	T=pie_scan::w(_,"(",_,_),
	parChkDoubles(Words),
	!.
parChkDoubles([T|Words]) :-
	T=pie_scan::w(_,")",_,_),
	parChkDoubles(Words),
	!.
parChkDoubles([T,T|Words]) :-
	S=toString(T),
	sx::concat(["*POSSIBLE KE ERROR* Repeated 'word' (NL name) in input: '",S,"'"],Mess),
	dlgIF::msgS(Mess),
	parChkDoubles(Words),
	!.
parChkDoubles([_|Words]) :-
	parChkDoubles(Words),
	!.
% --------------------------------------------------------------------------
% parChkOther
% --------------------------------------------------------------------------
% simply warns of a possibly missing comma - never fails
%
parChkOther([]) :- 
	!.
parChkOther([T1,T2|Tokens]) :-
	T1=pie_scan::w(_,",",_,_),
	T2=pie_scan::w(_,"p3_iIbc",_,_),
	parChkOther(Tokens),
	!.
parChkOther([T|Tokens]) :-
	T=pie_scan::w(_,"p3_iIbc",_,_),
	dbBase::nlSing("p3_iIbc",NL),
	sx::concat(["*POSSIBLE KE ERROR* Expected comma before '",NL,""],Mess),
	dlgIF::msgS(Mess),
	parChkOther(Tokens),
	!.
parChkOther([_|Tokens]) :-
	parChkOther(Tokens),
	!.
% --------------------------------------------------------------------------
% scnNL                                                              
% --------------------------------------------------------------------------
class predicates
check_length : (sterml,integer) determ.
clauses
% parameters:
% 1	in	text to be scanned
% 2	out	list of word-triplets
% always succeeds (unless system error)
%
scnNL(Text,Wrds) :-
			dlgIF::msgStatus("scanning"),
	pscan::pscNL(Text,Tokens),
			tt::tt_tokl(2,"SCANNER OUTPUT",Tokens),
			dlgIF::msgStatus("parsing"),
	scnChkDoubles(Tokens),
	pscan::str_tokens(Tokens,STokens),
	L=globalVars::get_i("max_tok"),
	check_length(STokens,L),
	scnSyntagmSets(Tokens,STokens,SynSets),
			tt::tt_termslist(2,"syntagm sets",SynSets),
	!,			
	nl_tok_scan(Tokens,STokens,SynSets,Wrds1),
	scnSimplify(Wrds1,Wrds2),
	scnEnumWords(1,Wrds2,Wrds)
	.
	
check_length([],_) :- 
	!.
check_length([atom(S)|T],L) :-
	string::length(S)<=L,
	check_length(T,L),
	!
;
	Ls=toString(L),
	Mess=string::concatList(["Token '",S,"' exceeds max. length of ",Ls,"!"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.	
check_length([var(S)|T],L) :-
	string::length(S)<=L,
	check_length(T,L),
	!
;
	Ls=toString(L),
	Mess=string::concatList(["Token '",S,"' exceeds max. length of ",Ls,"!"]),
	dlgIF::dlgNote(1,Mess),
	!,fail.	
check_length([_|T],L) :- 
	check_length(T,L),
	!.	
% --------------------------------------------------------------------------
% scnSimplify                                                           
% --------------------------------------------------------------------------
% 1/removes all 'the' words either by changing the following undetermined form to determined or else
% by deleting it with no change.
% 2/removes 'a'/'an'/'en'
% always succeeds
%
scnSimplify([],[]) :- !.
scnSimplify([pie_scan::w(_,"p3_iThe",_,_),pie_scan::w(_,E,C,"p3_iSnu")|T],[pie_scan::w(0,E,C,"p3_iSnd")|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"p3_iThe",_,_),pie_scan::w(_,E,C,"p3_iPlu")|T],[pie_scan::w(0,E,C,"p3_iPld")|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"p3_iThe",_,_),pie_scan::w(_,E,C,F)|T],[pie_scan::w(0,E,C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"p3_iA",_,_),pie_scan::w(_,E,C,F)|T],[pie_scan::w(0,E,C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,"=",C,F)|T],[pie_scan::w(0,"<>",C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,"<>",C,F)|T],[pie_scan::w(0,"=",C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,">",C,F)|T],[pie_scan::w(0,"<=",C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,">=",C,F)|T],[pie_scan::w(0,"<",C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,"<",C,F)|T],[pie_scan::w(0,">=",C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,"<=",C,F)|T],[pie_scan::w(0,">",C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,"like",C,F)|T],[pie_scan::w(0,"unlike",C,F)|Tm]) :- scnSimplify(T,Tm),!.
scnSimplify([pie_scan::w(_,"not",_,_),pie_scan::w(_,"unlike",C,F)|T],[pie_scan::w(0,"like",C,F)|Tm]) :- scnSimplify(T,Tm),!.
% Norwegian 
scnSimplify([pie_scan::w(_,"p3_iDen",_,_),pie_scan::w(_,E,C,F)|T],[pie_scan::w(0,E,C,F)|Tm]) :- scnSimplify(T,Tm),!.
%
scnSimplify([H|T],[H|Tm]) :- scnSimplify(T,Tm),!.
% --------------------------------------------------------------------------
% nl_tok_scan
% --------------------------------------------------------------------------
class predicates
check_syntagms : (pie_scan::tokl,sterml,sterml_list,sterml,pie_scan::wrd,pie_scan::tokl,sterml,sterml_list) nondeterm (i,i,i,i,o,o,o,o).
match_tokens : (pie_scan::tokl,sterml,sterml_list,sterm,pie_scan::tokl,sterml,sterml_list) determ (i,i,i,i,o,o,o).	
clauses

% scans a token list into a pie_scan::wrd-list
%
nl_tok_scan([],_,_,[]) :- 
	tt::tt_ss(3,"parsed tokens")
	.
nl_tok_scan([pie_scan::t(pie_scan::gvar(V),_)|T],[_|Ts],[_|Sets],[pie_scan::w(0,V,"p3_dvar","#")|W]) :- 
	nl_tok_scan(T,Ts,Sets,W)
	.
nl_tok_scan([pie_scan::t(pie_scan::str(S),_)|T],[_|Ts],[_|Sets],[pie_scan::w(0,S,"p3_dli","")|W]) :- 
	nl_tok_scan(T,Ts,Sets,W)
	.
nl_tok_scan([pie_scan::t(pie_scan::int(I),_)|T],[_|Ts],[_|Sets],[pie_scan::w(0,Si,"p3_din","")|W]) :- 
	Si=toString(I),
	nl_tok_scan(T,Ts,Sets,W)
	.
nl_tok_scan([pie_scan::t(pie_scan::fix(F),_)|T],[_|Ts],[_|Sets],[pie_scan::w(0,Sf,"p3_dnu","")|W]) :- 
	Sf=toString(F),
	nl_tok_scan(T,Ts,Sets,W)
	.
nl_tok_scan([pie_scan::t(_,_)|Tokens],[str(_)|STokens],[Set|Sets],[W|Wrds]) :-
	check_syntagms(Tokens,STokens,Sets,Set,W,TokensR,STokensR,SetsR),
	nl_tok_scan(TokensR,STokensR,SetsR,Wrds)
	.	
nl_tok_scan([pie_scan::t(pie_scan::var(V),_)|T],[_|Ts],[_|Sets],[pie_scan::w(0,V,"p3_dvar","")|W]) :- 
	nl_tok_scan(T,Ts,Sets,W)
	.
check_syntagms(_,_,_,[],_,_,_,_) :- 
	!,fail.
check_syntagms(Tokens,STokens,Sets,[Term|Terms],Wrd,TokensR,STokensR,SetsR) :-
	Lng=globalVars::get_s("lang"),
	Term=cmp("v",[cmp(Ent,[atom(Lng),str(R),atom(C),atom(F),list(_,L)]),_]),
	set::member(Ent,["t","s"]),
	match_tokens(Tokens,STokens,Sets,L,TokensR,STokensR,SetsR),
	Wrd=pie_scan::w(0,R,C,F)
;
	check_syntagms(Tokens,STokens,Sets,Terms,Wrd,TokensR,STokensR,SetsR)
	.	
	
match_tokens(T,S,Sets,nill,T,S,Sets) :- 
	!.
match_tokens([_|Tokens],[STok|STokens],[_|Sets],list(Tok,L),TokensR,STokensR,SetsR) :-
	STok=Tok, 
	match_tokens(Tokens,STokens,Sets,L,TokensR,STokensR,SetsR),
	!.		
% --------------------------------------------------------------------------
% scnSyntagmSets [3kb/sql]                  
% --------------------------------------------------------------------------
class predicates
get_syntagms : (integer,string,sterml,sterml) determ (i,i,i,o).
clauses
% the set of syntagms starting with token (and with a matching tail) is determined for each token 
% in the given list. string-tokens (originating from quoted strings) have empty syntagm-sets. 
% always succeeds.
%
scnSyntagmSets([],[],[]) :- 
	!.
scnSyntagmSets([pie_scan::t(pie_scan::str(S),_)|T1],[_|T2],[[]|Sets]) :- 
			tt::tt_s(3,"token/string",S),
	scnSyntagmSets(T1,T2,Sets),
	!.
scnSyntagmSets([pie_scan::t(pie_scan::gvar(S),_)|T1],[_|T2],[[]|Sets]) :- 
			tt::tt_s(3,"token/#variable",S),
	scnSyntagmSets(T1,T2,Sets),
	!.
scnSyntagmSets([pie_scan::t(pie_scan::atom(_),_)|T1],[str(Key)|T2],[Set|Sets]) :-
			tt::tt_s(3,"token/atom",Key),
	get_syntagms(1,Key,T2,Set),
			tt::tt_terms(3,"syntagms",Set),
	scnSyntagmSets(T1,T2,Sets),
	!.
scnSyntagmSets([_|T1],[str(Key)|T2],[Set|Sets]) :-
			tt::tt_s(3,"token/other",Key),
	get_syntagms(0,Key,T2,Set),
			tt::tt_terms(3,"syntagms",Set),
	scnSyntagmSets(T1,T2,Sets),
	!.
	
% gets the set of syntagms whose token list starts with 'Key' and has a tail which is 
% either empty or which is a front part of Tail
% parameters:
% 1	in	1:report unrecognizable tokens; 0:not
% 2	in	first token 
% 3	in	tail of tokens
% 4	out	syntagm sets (possibly empty)
% never fails	
%
get_syntagms(_,Key,Tail,SynSets) :-
    dbExtension::syntagmSets(Key,Tail,SynSets),
    not(SynSets=[]),
    !.
get_syntagms(1,Key,_,[]) :- 
	tt::tt_s(2,"Undefined token (possible error)",Key),
	!.	
get_syntagms(0,_,_,[]) :- 
	!.
% --------------------------------------------------------------------------
% scnText/Block
% --------------------------------------------------------------------------
scnText(Typ,S,T,Tc) :-
	scnTextF(Typ,S,T,Tc),
	!
;
	Mess=string::concatList(["Text '",S,"' is invalid!"]),
	dlgIF::dlgNote(1,Mess),!,fail.

scnTextF(Typ,S,T,Tc) :-
	Typ="xml",
	sx::sxStrip(S,T),
	Tc="",
	!
;
	set::member(Typ,["3PR","3NL"]),
	sx::sxStripComments(S,S2),
	sx::sxStrip(S2,T),
	string::length(T)>3,
	sx::sxTail(1,T,_,Tc),
	set::member(TC,["!","?"]),
	!
;
	not(set::member(Typ,["XML","3PR","3NL"])),
	Mess=string::concatList(["Invalid knowledge-text type: ",Typ,"!"]),
	dlgIF::dlgNote(1,Mess),!,fail.
	
scnTextBlock(Typ,S,L,Tc) :-
	scnTextBlockF(Typ,S,L,Tc),
	!
;
	Mess=string::concatList(["Invalid knowledge text '",S,"'!"]),
	dlgIF::dlgNote(1,Mess),!,fail.

scnTextBlockF(Typ,S,L,Tc) :-
	Typ="3nl",
	sx::sxStripComments(S,S2),
	sx::sxStrip(S2,S3),
	sx::sxTail(1,S3,_,Tc),
	Tc="?",
	sx::sxUnpack(S3,'?',L),
	!
;	
	Typ="3nl",
	sx::sxStripComments(S,S2),
	sx::sxStrip(S2,S3),
	sx::sxTail(1,S3,_,Tc),
	Tc="!",
	sx::sxUnpack(S3,'!',L),
	!
;
	Typ="3PR",
	sx::sxStripComments(S,S2),
	sx::sxStrip(S2,S3),
	sx::sxTail(1,S3,_,Tc),
	Tc="?",
	sx::sxUnpack(S3,'?',L),
	!
;	
	Typ="3PR",
	sx::sxStripComments(S,S2),
	sx::sxStrip(S2,S3),
	sx::sxTail(1,S3,_,Tc),
	Tc="!",
	sx::sxUnpack(S3,'!',L),
	!
;
	not(set::member(Typ,["3PR","3NL"])),
	dlgIF::feil("scnTextBlock",[Typ,S],[],nill),!,fail.
% --------------------------------------------------------------------------
% scnChkDoubles
% --------------------------------------------------------------------------
scnChkDoubles([]) :- !.
scnChkDoubles([T1|Tokens]) :-
	T1=pie_scan::t(pie_scan::lpar,_),
	scnChkDoubles(Tokens),!.
scnChkDoubles([T1|Tokens]) :-
	T1=pie_scan::t(pie_scan::rpar,_),
	scnChkDoubles(Tokens),!.
scnChkDoubles([T1,T2|Tokens]) :-
	T1=pie_scan::t(T,_),
	T2=pie_scan::t(T,_),
	S1=toString(T),
	sx::concat(["*POSSIBLE KE ERROR* Repeated token '",S1,"'"],Mess),
	dlgIF::msgS(Mess),
	scnChkDoubles(Tokens),!.
scnChkDoubles([_|Tokens]) :-
	scnChkDoubles(Tokens),!.
% --------------------------------------------------------------------------
% scnEnumWords
% --------------------------------------------------------------------------
scnEnumWords(_,[],[]) :- 
	!.
scnEnumWords(N,[pie_scan::w(_,A,B,C)|Wc],[pie_scan::w(N,A,B,C)|Wr]) :-
	M=N+1,
	scnEnumWords(M,Wc,Wr),
	!.
	
% --------------------------------------------------------------------------
% transQuery
% --------------------------------------------------------------------------
class predicates
transQuery2 : (sterm,sterm) determ (i,o).
clauses

transQuery(S,Query,QDom) :-
	sx::sxStrip(S,Q3),
	Q3<>"",
	globalVars::set_i("pre_cnt",0),
	transnl::parSynTree(Q3,Syn),
	transQuery2(Syn,Query),
	queryDom(Query,QDom),
	!
;
	sx::sxStrip(S,Q2),
	Q2<>"",
	dlgIF::dlgNote(1,"Invalid query type"),
	!,fail
;
	dlgIF::dlgNote(1,"No query entered"),
	!,fail.
	
transQuery2(Syn,Query) :-
	Syn=cmp(QT,_),
	set::member(QT,["DFNC","NFNC","NDERF","DDERF","#S","EX"]),
	!,tbq::tbqQuery(Syn,Tbq),
	pr::prTbqQuery(Tbq,PR,_),
	transpr::proPR_Term(PR,Query),
	!
;	
	pr::prSynTreeQuery(0,Syn,PR),
	transpr::proPR_Term(PR,Query),
	!.

queryDom(Q,Dom) :-
	Q=cmp("p3_iDdrf",[_,str(DEV),_]),
	vars::varComps(DEV,_,_,DE),
	dbBase::smClassDefF(DE,"a",_,_,Dom,_,_),
	!
;
	Q=cmp("p3_iSSS",[_,_]),
	Dom="p3_din",
	!
;
	Q=cmp("p3_iExis",[_]),
	Dom="p3_dif",
	!
;
	Q=cmp(Func,_),
	dbBase::smDefF(Func,Clas,_,_,Dom,_,_),
	set::member(Clas,["f","g"]),
	!
;
    % not single-value query
	Dom="p3_dany",
	!.
end implement transnl
