/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement transrt
open core,termDomain

constants
className = "com/rudstrand/transrt/transrt".
classVersion = "1.0".

class predicates
rlgArgument : (sterm,string,string) determ (i,o,o).
rlgAssignmentComparison : (string,sterm,string) determ (i,i,o).
rlgAssignmentComparisons : (string,sterml,string) determ (i,i,o).
rlgCertaintyRuleCall : (sterm,string) determ (i,o).
rlgCondition : (sterm,string) determ (i,o).
rlgD_Components : (string,sterml,string,string) determ (i,i,o,o).
rlgExpression : (string,sterm,string) determ (i,i,o).
rlgFunctionCall : (sterm,string) determ (i,o).
rlgGenDE_Defs : (sterm,integer,sterm,sterml,string_list,sterml) determ (i,i,i,o,o,o).	
rlgLiteral : (string,sterm,string) determ (i,i,o).
rlgMess : (string).
rlgParameters : (sterml,string) determ (i,o).
rlgProcedureCall : (sterm,string) determ (i,o).
rlgSentenceCountCondition : (sterm,string) determ (i,o).
rlgSentenceCondition : (sterm,string,string) determ (i,o,o).
rlgSimpleCondition : (sterm,string) determ (i,o).
rlgSimpleExpression : (string,sterm,string) determ (i,i,o).
rlgVariableCondition : (sterm,string) determ (i,o).

clauses
classInfo(className, classVersion).
% ------------------------------------------------------------------------------------------------------------
% rlgPR
% ------------------------------------------------------------------------------------------------------------
class predicates
rlgPR_ElementaryRule : (sterm,string) determ (i,o).
rlgPRs : (sterml,string_list) determ (i,o).
rlgMakeDE_Vars2 : (string_list,sterml) determ (i,o).
rlgMakeDE_Vars : (string_list,string) determ (i,o).
clauses

% parameters:
% 1	in	SR rule tree (poss. with OR's)
% 2	out	list of PR strings corresponding to OR-free rules of par#1
% 
rlgPR(Rule,PRs) :-
	rub::rubValidate(Rule,Rule2),
	struc::treElimOR(Rule2,Rules),			
			tt::tt_terms(2,"OR-eliminated rules",Rules),				
	rlgPRs(Rules,PRs),
	!.	
	
rlgPRs([],[]) :-
	!.
rlgPRs([Rule|Rules],[PR|PRs]) :-
	rlgPR_ElementaryRule(Rule,PR),
	rlgPRs(Rules,PRs),
	!.
	
% parameters:
% 1	in	OR-free rule tree
% 2	in	DE-cars, format: ARGUMENT | NO | DE | DE_VAR | DENL[NO] or 
% 3	out	rule (PR)
%
rlgPR_ElementaryRule(Rule,PR) :-
	Rule=cmp(Type,[Cond,Conc]),
	rub::rubRuleType(Type,No,_),
	set::member(No,[1,4]),
	!,rlgSentenceCondition(Conc,Conclusion,PR2),
	rlgCondition(Cond,PR3),
	sx::sxlJoin([PR3,PR2],",",S4),
	sx::concat([Conclusion,":-",S4],PR),
	!
;
	Rule=cmp(Type,[Cond,Conc]),
	% the query which generates the condition set
	rlgCondition(Cond,Q1),
	transpr::proPR_Term(Q1,Q2),
	Q3=termDomain::toTermStringIndirect("q",Q2),
	sx::sxQuote(Q3,Query),
	% the list of conclusion DE definitions and DE variables
	rlgSentenceCondition(Conc,Conclusion,Ass2),
	transpr::proPR_Term(Conclusion,Conc2),
	transpr::proPR_Term(Ass2,Ass3),
	rlgGenDE_Defs(Ass3,0,Conc2,_,DEVs2,Ass4),
	% DE variables (1)
	rlgMakeDE_Vars(DEVs2,DE_Vars),
	% DE variables (2)
	sx::sxlJoin(DEVs2,",",DEVs),
	% the list of conclusion assignments corresponding to the DE definitions
	termDomain::listBR(Ass5,Ass4),
	Ass6=termDomain::toTermStringIndirect("a",Ass5),
	sx::sxQuote(Ass6,Ass),
	%
	rub::rubRuleType(Type,_,RuleType),
	sx::concat([RuleType,"(",Query,",",DE_Vars,",",Ass,",",DEVs,")"],CondS), 
	sx::concat([Conclusion,":-",CondS],PR),
	!.
	
rlgMakeDE_Vars(DEVs,Ds) :-
	rlgMakeDE_Vars2(DEVs,D2),
	termDomain::listBR(D3,D2),
	D4=termDomain::toTermStringIndirect("d",D3),
	sx::sxQuote(D4,Ds),
	!.
	
rlgMakeDE_Vars2([],[]) :-
	!.
rlgMakeDE_Vars2([DV|DVs],[DEV|DEVs]) :-
	vars::varComps(DV,_,_,D),
	dbBase::smDef(D,Clas,_,_,A,B,C),
	Clas="a",
	DEV=fstr(DV,A,B,C),
	rlgMakeDE_Vars2(DVs,DEVs),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgArgument
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	argument
% 2	out	argument (ST-)variable (PR)
% 3	out	argument DE conditions (PR)
% 4 par#2 (term)
% 5 par#3 (term)
%
rlgArgument(A,As,Ac) :-
	A=var(V),
	As=V,
	Ac="",
	!
;
	A=cmp(AT,D_Comps),
	sx::sxUnpack(AT,'[',L),
	L=[Ak|_],
	sx::sxGetInBrack(AT,'[',']',SNo),
	rlgD_Components(SNo,D_Comps,Ds,Ac),
	sx::concat([Ak,"(",Ds,")"],As),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgAssignmentComparisons
% ------------------------------------------------------------------------------------------------------------
rlgAssignmentComparisons(_,[],"") :-
	!.
rlgAssignmentComparisons(V,[ACOP|ACOPs],C) :-
	rlgAssignmentComparison(V,ACOP,C2),
	rlgAssignmentComparisons(V,ACOPs,C3),
	sx::sxlJoin([C2,C3],",",C),
	!.
rlgAssignmentComparison(V,ACOP,C) :-
	ACOP=cmp(Opr,[nill]),
	opr::opr_cmpass(Opr),
	sx::concat(["Variable '",V,"' is assigned nill-value - rule cannot be translated!"],Mess),
	rlgMess(Mess),
	!,fail
;
	sx::sxUnpackS(V," /",L),
	L=[W,SDom],
	dom::domShortLongF(SDom,Dom),
	ACOP=cmp(Opr,[Exp]),
	opr::oprEq(Opr),
	opr::oprDom(Dom,Opr,OprD),
	opr::oprIntExt(Op,OprD),
	rlgExpression(Dom,Exp,S),
	sx::concat([W," ",Op," ",S],C),
	!
;
	vars::varComps(V,_,_,DE),
	dbBase::smClassDefF(DE,"a",_,_,Dom,_,_),
	ACOP=cmp(Opr,[Exp]),
	opr::oprEq(Opr),
	opr::oprDom(Dom,Opr,OprD),
	opr::oprIntExt(Op,OprD),
	rlgExpression(Dom,Exp,S),
	sx::concat([V," ",Op," ",S],C),
	!
;
	vars::varFuncParR(_,DE,V),
	dbBase::smClassDef(DE,"a",_,_,Dom,_,_),
	ACOP=cmp(Opr,[Exp]),
	opr::oprEq(Opr),
	opr::oprDom(Dom,Opr,OprD),
	opr::oprIntExt(Op,OprD),
	rlgExpression(Dom,Exp,S),
	sx::concat([V," ",Op," ",S],C),
	!
;
	sx::sxUnpackS(V," /",L),
	L=[W,SDom],
	dom::domShortLongF(SDom,Dom),
	ACOP=cmp(Opr,[Exp]),
	opr::opr_cmpass(Opr),
	opr::oprIntExt(Op,Opr),
	rlgExpression(Dom,Exp,S),
	sx::concat([W," ",Op," ",S],C),
	!
;
	ACOP=cmp(Opr,[Exp]),
	opr::opr_cmpass(Opr),
	opr::oprIntExt(Op,Opr),
	rlgExpression("p3_dany",Exp,S),
	sx::concat([V," ",Op," ",S],C),
	!
;
	dlgIF::feil("rlgAssignmentComparison",[V],[],ACOP),!,fail.
% ------------------------------------------------------------------------------------------------------------
% rlgCertaintyRuleCall
% ------------------------------------------------------------------------------------------------------------
rlgCertaintyRuleCall(E,S) :-
	E=atom(F),
	sx::frontstr(4,F,Fr,K),
	Fr="[C] ",
	sx::concat(["p3_iIR(",K,")"],S),
	!
;
	E=cmp(F,Params),
	sx::frontstr(4,F,Fr,K),
	Fr="[C] ",
	rlgParameters(Params,S2),
	sx::concat(["p3_iIR(",K,",",S2,")"],S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgCondition
% ------------------------------------------------------------------------------------------------------------
class predicates
rlgConditions : (sterml,string) determ (i,o).
clauses

% parameters:
% 1	in	condition tree
% 2	out	condition PR
%
rlgCondition(Cond,PR) :-
	Cond=cmp("&",Conds),
	rlgConditions(Conds,PR),
	!
;
	rlgSimpleCondition(Cond,PR),
	!.
	
rlgConditions([C],PR2) :-
	rlgCondition(C,PR2),
	!.
rlgConditions([C|Cs],PR) :-
	rlgCondition(C,PR2),
	rlgConditions(Cs,PR3),
	sx::sxlJoin([PR2,PR3],",",PR),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgD_Components
% ------------------------------------------------------------------------------------------------------------
class predicates
rlgD_Component : (string,sterm,string,string) determ (i,i,o,o).
rlgDE_Component : (sterm,string,string) determ (i,o,o).
clauses

% parameters:
% 1	in	rule DE list
% 2	in	argument no
% 3	in	data element components (for argument)
% 4	out	argument DE variable list string
% 5	out	do. conditions string
%
rlgD_Components(_,[],"","") :-
	!.
rlgD_Components(SNo,[DC|DCs],S,C) :-
	rlgD_Component(SNo,DC,S2,C2),
	rlgD_Components(SNo,DCs,S3,C3),
	sx::sxlJoin([S2,S3],",",S),
	sx::sxlJoin([C2,C3],",",C),
	!.
	
rlgD_Component(_,DC,S,C) :-
	rlgDE_Component(DC,S,C),
	!.
		
rlgDE_Component(DC,K,C) :-
	DC=var(K),
	vars::varComps(K,_,_,DEk),
	dbBase::smDefF(DEk,Clas,_,_,_,_,_),
	Clas="a",
	C="",
	!
;
	DC=cmp(K,[cmp("&",ACOPs)]),
	vars::varComps(K,_,_,DEk),
	dbBase::smDefF(DEk,Clas,_,_,_,_,_),
	Clas="a",
	rlgAssignmentComparisons(K,ACOPs,C),
	!
;
	DC=cmp(K,[ACOP]),
	vars::varComps(K,_,_,DEk),
	dbBase::smDefF(DEk,Clas,_,_,_,_,_),
	Clas="a",
	rlgAssignmentComparison(K,ACOP,C),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgExpression
% ------------------------------------------------------------------------------------------------------------
rlgExpression(Dom,Exp,S) :-
	Exp=cmp(UnOpr,[Exp]),
	opr::opr_expr(UnOpr,1),
	rlgExpression(Dom,Exp,S2),
	sx::concat([UnOpr,"(",S2,")"],S),
	!
;
	Exp=cmp(BinOpr,[Exp2,Exp3]),
	opr::opr_expr(BinOpr,2),
	rlgExpression(Dom,Exp2,S2),
	rlgExpression(Dom,Exp3,S3),
	sx::concat(["(",S2," ",BinOpr," ",S3,")"],S),
	!
;
	rlgSimpleExpression(Dom,Exp,S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgFunctionCall
% ------------------------------------------------------------------------------------------------------------
rlgFunctionCall(E,S) :-
	E=atom(F),
	sx::frontstr(4,F,Fr,K),
	Fr="[F] ",
	dbBase::smClassDefF(K,"f",_,_,_,_,_),
	vars::varGenP(K,V),
	sx::concat([K,"(p3_iDtFn,",V,")"],S),
	!
;
	E=atom(F),
	sx::frontstr(4,F,Fr,K),
	Fr="[F] ",
	dbBase::smClassDefF(K,"g",_,_,_,_,_),
	vars::varGenP(K,V),
	sx::concat([K,"(",V,")"],S),
	!
;
	E=cmp(F,Params),
	sx::frontstr(4,F,Fr,K),
	Fr="[F] ",
	dbBase::smClassDefF(K,"f",_,_,_,_,_),
	vars::varGenP(K,V),
	rlgParameters(Params,S2),
	sx::concat([K,"(p3_iDtFn,",V,",",S2,")"],S),
	!
;
	E=cmp(F,Params),
	sx::frontstr(4,F,Fr,K),
	Fr="[F] ",
	dbBase::smClassDefF(K,"g",_,_,_,_,_),
	vars::varGenP(K,V),
	rlgParameters(Params,S2),
	sx::concat([K,"(",V,",",S2,")"],S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgGenDE_Defs
% ------------------------------------------------------------------------------------------------------------
class predicates
rlgGetAss : (string,sterm,sterm) determ (i,i,o).
rlgGetAssX : (string,sterml,sterm) determ (i,i,o).
rlgGenDE_DefsX : (sterm,integer,sterml,sterml,string_list,sterml) determ (i,i,i,o,o,o).	
clauses
% parameters:
% 1	in	conclusion DE assignments
% 2	in	argument no for parent argument (or 0 if no parent argumkent)
% 3	in	part of conclusion tree
% 4	out	conclusion DE definition list
% 5	out	conclusion DE variable list
% 6	out	conclusion DE assignment list
%
rlgGenDE_Defs(Assignments,ParentArg,T,Defs,DEVs,Ass) :-
	T=cmp(A,B),
	dbBase::smClassDefF(A,"e",_,_,_,_,_),
	rt::rcnGetArgNo(T,No),
	% argument node
	rlgGenDE_DefsX(Assignments,No,B,Defs,DEVs,Ass),
	!
;
	T=cmp(_,B),
	rlgGenDE_DefsX(Assignments,0,B,Defs,DEVs,Ass),
	!
;
	ParentArg>0,
	T=var(V),
	vars::varComps(V,_,_,K),
	dbBase::smDefTree(K,Clas,Def),
	Clas="a",
	rlgGetAss(V,Assignments,A), 
	Defs=[Def],
	DEVs=[V],
	Ass=[A],
	!
;
	ParentArg=0,
	T=var(V),
	% predicate's DE 
	vars::varComps(V,_,_,K),
	dbBase::smDefTree(K,Clas,Def),
	Clas="a",
	rlgGetAss(V,Assignments,A), 
	Defs=[Def],
	DEVs=[V],
	Ass=[A],
	!
;
	ParentArg>0,
	dlgIF::feil("rlgGenDE_Defs",[],[ParentArg],cmp("",[T,Assignments])),!,fail
;
	Defs=[],
	DEVs=[],
	Ass=[],
	!.
rlgGenDE_DefsX(_,_,[],[],[],[]) :-
	!.
rlgGenDE_DefsX(Assignments,ParentArg,[H|T],Defs,DEVs,Ass) :-
	rlgGenDE_Defs(Assignments,ParentArg,H,Defs2,DEVs2,Ass2),
	rlgGenDE_DefsX(Assignments,ParentArg,T,Defs3,DEVs3,Ass3),
	set::conc([Defs2,Defs3],Defs),
	set::conc([DEVs2,DEVs3],DEVs),
	set::conc([Ass2,Ass3],Ass),
	!.
% returns the expression assigned to V (fails if none)
%
rlgGetAss(V,Ass,A) :-
	Ass=cmp(",",L),
	rlgGetAssX(V,L,A),
	!
;
	Ass=cmp(Opr,[var(V),A]),
	opr::oprEq(Opr),
	!.
	
rlgGetAssX(V,[H|T],A) :-
	rlgGetAss(V,H,A),
	!
;
	rlgGetAssX(V,T,A),
	!.	
% ------------------------------------------------------------------------------------------------------------
% rlgLiteral
% ------------------------------------------------------------------------------------------------------------
class predicates
rlgLiterals : (string,sterml,string) determ (i,i,o).
clauses
rlgLiteral(Dom,E,S) :-
	% ListLiteral
	sx::sxTail(2,Dom,D,T),
	T="_l",
	E=cmp("[]",L),
	rlgLiterals(D,L,S2),
	sx::concat(["[",S2,"]"],S),
	!
;
	% ListLiteral
	sx::sxTail(2,Dom,D,T),
	T="_l",
	E=cmp(ListHead,L),
	sx::sxGetInBrack(ListHead,'[',']',D2),
	dom::domShortLongF(D2,_),
	rlgLiterals(D,L,S2),
	sx::concat(["[",S2,"]"],S),
	!
;
	E=str(S2),
	sx::sxQuote(S2,S),
	!
;
	E=atom(S),
	!
;
	E=int(I),
	S=toString(I),
	!
;
	E=fix(F),
	S=toString(F),
	!
;
	dlgIF::feil("rlgLiteral",["Invalid literal",Dom],[],E),!,fail.
		
rlgLiterals(_,[],"") :-
	!.
rlgLiterals(Dom,[H|T],S) :-
	rlgLiteral(Dom,H,S2),
	rlgLiterals(Dom,T,S3),
	sx::sxlJoin([S2,S3],",",S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgMess
% ------------------------------------------------------------------------------------------------------------
rlgMess(Mess) :-
	dlgIF::dlgNote(0,Mess),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgParameters
% ------------------------------------------------------------------------------------------------------------
class predicates
rlgParameter : (sterm,string) determ (i,o).
clauses
rlgParameters([],"") :-
	!.
rlgParameters([P|Ps],S) :-
	rlgParameter(P,S2),
	rlgParameters(Ps,S3),
	sx::sxlJoin([S2,S3],",",S),
	!.
	
rlgParameter(P,S) :-
	P=cmp(DE,[cmp(_,[Exp])]),
	dbBase::smClassDef(DE,"a",_,_,Dom,_,_),
	rlgExpression(Dom,Exp,S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgProcedureCall
% ------------------------------------------------------------------------------------------------------------
rlgProcedureCall(E,S) :-
	E=atom(F),
	sx::frontstr(4,F,Fr,S),
	Fr="[P] ",
	!
;
	E=cmp(F,Params),
	sx::frontstr(4,F,Fr,K),
	Fr="[P] ",
	rlgParameters(Params,S2),
	sx::concat([K,"(",S2,")"],S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgSentenceCountCondition
% ------------------------------------------------------------------------------------------------------------
rlgSentenceCountCondition(C,S) :-
	C=cmp(V,[cmp(_,[cmp("p3_iSSS",[SC])])]),
	sx::sxUnpackS(V," /",L),
	L=[W,SDom],
	dom::domShortLongF(SDom,Dom),
	dom::domCompat(Dom,"p3_din"),
	rlgSentenceCondition(SC,S2,S3),
	sx::sxlJoin([S2,S3],",",S4),
	transpr::proPR_Term(S4,T),
	S5=termDomain::toTermStringIndirect("s",T),
	sx::sxQuote(S5,S6),
	sx::concat(["p3_iSSS(",W,",",S6,")"],S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgSentenceCondition
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	ST condition
% 2	out	ST variable (PR)
% 3	out	ST DE conditions (PR)
%
rlgSentenceCondition(T,ST,SC) :-
	 T=cmp(STk,[E2,P,E3]),
	 sx::sxUnpack(STk,'|',L),
	 L=[Pk|_],
     dbBase::smClassDefF(Pk,"p",Card,_,_,_,_),
	 Card=2,
	 rlgArgument(E2,E2s,E2c),
	 rlgArgument(E3,E3s,E3c),
	 rlgArgument(P,Ps2,Pc),
	 sx::searchstring(Ps2,"(",N2),
	 sx::frontstr(N2,Ps2,_,Ps3),
	 sx::sxTail(1,Ps3,Ps,_),
	 sx::concat([Pk,"(",E2s,",",E3s,",",Ps,")"],ST),
	 sx::sxlJoin([E2c,E3c,Pc],",",SC),
	 !
;	 
	 T=cmp(STk,[E2,P]),
	 sx::sxUnpack(STk,'|',L),
	 L=[Pk|_],
     dbBase::smClassDefF(Pk,"p",Card,_,_,_,_),
	 Card=1,
	 rlgArgument(E2,E2s,E2c),
	 rlgArgument(P,Ps2,Pc),
	 sx::searchstring(Ps2,"(",N2),
	 sx::frontstr(N2,Ps2,_,Ps3),
	 sx::sxTail(1,Ps3,Ps,_),
	 sx::concat([Pk,"(",E2s,",",Ps,")"],ST),
	 sx::sxlJoin([E2c,Pc],",",SC),
	 !
;	 
	 T=cmp(STk,[E2,E3]),
	 sx::sxUnpack(STk,'|',L),
	 L=[Pk|_],
     dbBase::smClassDefF(Pk,"p",Card,_,_,_,_),
	 Card=2,
	 rlgArgument(E2,E2s,E2c),
	 rlgArgument(E3,E3s,E3c),
	 sx::concat([Pk,"(",E2s,",",E3s,")"],ST),
	 sx::sxlJoin([E2c,E3c],",",SC),
	 !
;	 
	 T=cmp(STk,[E2]),
	 sx::sxUnpack(STk,'|',L),
	 L=[Pk|_],
     dbBase::smClassDefF(Pk,"p",Card,_,_,_,_),
	 Card=1,
	 rlgArgument(E2,E2s,SC),
	 sx::concat([Pk,"(",E2s,")"],ST),
	 !.
% ------------------------------------------------------------------------------------------------------------
% rlgSimpleCondition
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	condition tree
% 2	out	condition PR
%
rlgSimpleCondition(C,PR) :-
	rlgSentenceCondition(C,SC_Var,DE_Conds),
	sx::sxlJoin([SC_Var,DE_Conds],",",PR),
	!
;
	rlgSentenceCountCondition(C,PR),
	!
;
	rlgVariableCondition(C,PR),
	!
;
	C=cmp("not",[SCo]),
	rlgSentenceCondition(SCo,S2,S3),
	sx::sxlJoin([S2,S3],",",S4),
	sx::concat(["not(",S4,")"],PR),
	!
;
	C=cmp("true",[SCo]),
	rlgSentenceCondition(SCo,S2,S3),
	sx::sxlJoin([S2,S3],",",S4),
	sx::concat(["true(",S4,")"],PR),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgSimpleExpression
% ------------------------------------------------------------------------------------------------------------
class predicates
rlgList : (string,sterml,string) determ (i,i,o).
clauses
rlgSimpleExpression(Dom,Exp,S) :-
	% List
	sx::sxTail(2,Dom,Dm,T),
	T="_l",
	Exp=cmp("[]",L),
	rlgList(Dm,L,S2),
	sx::concat(["[",S2,"]"],S),
	!
;
	% List
	sx::sxTail(2,Dom,Dm,T),
	T="_l",
	Exp=cmp(ListHead,L),
	sx::sxGetInBrack(ListHead,'[',']',D),
	dom::domShortLongF(D,_),
	rlgList(Dm,L,S2),
	sx::concat(["[",S2,"]"],S),
	!
;
	Exp=var(S),
	!
;
	rlgFunctionCall(Exp,S),
	!
;
	rlgProcedureCall(Exp,S),
	!
;
	rlgCertaintyRuleCall(Exp,S),
	!
;	
	rlgLiteral(Dom,Exp,S),
	!.
rlgList(_,[],"") :-
	!.
rlgList(Dom,[H|T],S) :-
	rlgSimpleExpression(Dom,H,S2),
	rlgList(Dom,T,S3),
	sx::sxlJoin([S2,S3],",",S),
	!.
% ------------------------------------------------------------------------------------------------------------
% rlgVariableCondition
% ------------------------------------------------------------------------------------------------------------
rlgVariableCondition(C,S) :-
	C=cmp(V,[ACOP]),
	rub::rubIsVar(V,_,_),
	rlgAssignmentComparison(V,ACOP,S),
	!.
end implement transrt
