/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

implement tree
open core,vpiDomains

constants
        tree_frame_hor                  : integer = 3.% it was used also as core::integer_list, ::short
        tree_frame_ver                  : core::integer16 = 3.
        tree_link_width                 : integer = 5.
        tree_skip                       = 2.

        tree_scroll_hor                 : integer = 20.
        tree_scroll_ver                 : integer = 20.

        tree_selected_node_color        = color_Black.
        tree_selected_text_color        = color_White.
        tree_frame_color                = color_Black.
        tree_link_color                 = color_Black.

domains
        tree_node       = node(integer ColLR, integer, integer RowLR, integer ColUD, integer, integer RowUD, integer, integer, integer, selector, core::booleanInt Closed, core::booleanInt In, core::booleanInt Out, tree_LINK, arrow_type);
                          row_mark(integer, integer, integer);
                          closed_mark(selector, integer).
        tree_node_list  = tree_node*.
        tree_link       = begin; end_; cross; line.

class facts- tree_base
        tree_above : (vpiDomains::windowHandle, tree_node).
        tree_below : (vpiDomains::windowHandle, tree_node).
        screen_context : (vpiDomains::windowHandle, integer, integer, integer).
        tree_size : (vpiDomains::windowHandle, integer, integer, integer, integer, integer).
        multiple_selector : (vpiDomains::windowHandle, selector).
        opened_multiple_in_tree_base : (vpiDomains::windowHandle, tree_node).
        opened_multiple_in_closed_base : (vpiDomains::windowHandle, tree_node, selector).
        win_info : (vpiDomains::windowHandle, win_info).
        tree_callback : (vpiDomains::windowHandle, vpiDomains::ehandler).

constants
className = "com/rudstrand/tree/tree".
classVersion = "1.0".

class predicates
tree_base_clean_up : (vpiDomains::windowHandle) procedure(i).
tree_base_copy : (vpiDomains::windowHandle From, vpiDomains::windowHandle To) procedure(i,i).

clauses
classInfo(className, classVersion).
        tree_base_clean_up(Win):-
                retractall(tree_above(Win, _)),
                retractall(tree_below(Win, _)),
                retractall(screen_context(Win, _, _, _)),
                retractall(tree_size(Win, _, _, _, _, _)),
                retractall(multiple_selector(Win, _)),
                retractall(opened_multiple_in_tree_base(Win, _)),
                retractall(opened_multiple_in_closed_base(Win, _, _)),
                retractall(win_info(Win, _)),
                retractall(tree_callback(Win, _)).

        tree_base_copy(From, To):-
                tree_above(From, Node),
                assertz(tree_above(To, Node)), fail.
        tree_base_copy(From, To):-
                tree_below(From, Node),
                assertz(tree_below(To, Node)), fail.
        tree_base_copy(From, To):-
                screen_context(From, XOff, YOff, Dir),
                assert(screen_context(To, XOff, YOff, Dir)), fail.
        tree_base_copy(From, To):-
                tree_size(From, WidLR, HeiLR, WidUD, HeiUD, N),
                assert(tree_size(To, WidLR, HeiLR, WidUD, HeiUD, N)), fail.
        tree_base_copy(From, To):-
                multiple_selector(From, Selector),
                assert(multiple_selector(To, Selector)), fail.
        tree_base_copy(From, To):-
                opened_multiple_in_tree_base(From, Node),
                assert(opened_multiple_in_tree_base(To, Node)), fail.
        tree_base_copy(From, To):-
                opened_multiple_in_closed_base(From, Node, Selector),
                assert(opened_multiple_in_closed_base(To, Node, Selector)), fail.
        tree_base_copy(From, To):-
                win_info(From, TREE_WIN_INFO),
                assert(win_info(To, TREE_WIN_INFO)), fail.
        tree_base_copy(From, To):-
                tree_callback(From, EHANDLER),
                assert(tree_callback(To, EHANDLER)), fail.
        tree_base_copy(_, _).

/******************************************************************************

                SELECTION HANDLING

******************************************************************************/
class facts - selection_base
        selected : (vpiDomains::windowHandle, integer, integer, integer).
        selected_old : (vpiDomains::windowHandle, integer, integer, integer).
        hit_point : (vpiDomains::windowHandle, vpiDomains::pnt).
        invalid_selection : (vpiDomains::windowHandle, vpiDomains::rct).

class predicates
        selection_update : (vpiDomains::windowHandle, core::booleanInt) procedure(i,i).
        selection_set_old : (vpiDomains::windowHandle) procedure(i).
        get_selected_node : (vpiDomains::windowHandle, tree_node) determ anyflow.
        selection_base_clean_up : (vpiDomains::windowHandle) procedure(i).
        selection_base_copy : (vpiDomains::windowHandle From , vpiDomains::windowHandle To) determ(i,i).

clauses
        selection_update(Win, b_false):- !,
                retractall(invalid_selection(Win, _)),
                retractall(selected_old(Win, _, _, _)).
        selection_update(Win, _):-
                selected(Win, I, J, K),
                selected_old(Win, I, J, K), !,
                retractall(invalid_selection(Win, _)),
                retractall(selected_old(Win, _, _, _)).
        selection_update(Win, _):-
                retract(invalid_selection(Win, RCT)),
                vpi::winInvalidate(Win, RCT),
                vpi::winUpdate(Win), fail.
        selection_update(Win, _):-
                retractall(selected_old(Win, _, _, _)).

        selection_set_old(Win):-
                selected(Win, I, J, K), !,
                retractall(selected_old(Win, _, _, _)),
                assert(selected_old(Win, I, J, K)).
        selection_set_old(_).

        get_selected_node(Win, node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Selector, Closed, In, Out, Link, ArrowType)):-
                selected(Win, I, J, K),
                tree_below(Win, node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Selector, Closed, In, Out, Link, ArrowType)), !.
        get_selected_node(Win, node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Selector, Closed, In, Out, Link, ArrowType)):-
                selected(Win, I, J, K),
                tree_above(Win, node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Selector, Closed, In, Out, Link, ArrowType)), !.

        selection_base_clean_up(Win):-
                retractall(selected(Win, _, _, _)),
                retractall(selected_old(Win, _, _, _)),
                retractall(hit_point(Win, _)),
                retractall(invalid_selection(Win, _)).

        selection_base_copy(From, To):-
                selected(From, I, J, K), !,
                assert(selected(To, I, J, K)).

/******************************************************************************

                SEARCH HANDLING

******************************************************************************/
class facts- search_base
        search_sel : (vpiDomains::windowHandle, selector).
        search : (vpiDomains::windowHandle, integer, integer, integer).
        search_rct : (vpiDomains::windowHandle, vpiDomains::rct).

class predicates
        search_found : (integer, integer, integer, integer, integer, integer) determ(i,i,i,i,i,i).
        search_rect : (vpiDomains::windowHandle, integer, integer, core::booleanInt) determ(i,o,o,i).
        search_rect_ : (vpiDomains::rct, vpiDomains::rct, integer, integer) procedure(i,i,o,o).
        search_base_clean_up : (vpiDomains::windowHandle) procedure(i).

clauses
        search_found(I0, _, _, I1, _, _):-
                I0 > I1, !.
        search_found(I, J0, _, I, J1, _):-
                J0 < J1, !.
        search_found(I, J, K0, I, J, K1):-
                K0 > K1, !.

        search_rect(Win, Dx, Dy, Update):-
                C = vpi::winGetClientRect(Win),
                retract(search_rct(Win, R)), !,
                search_rect_(C, R, Dx, Dy),
                selection_update(Win, Update).

        search_rect_(C, R, Dx, Dy):-
                I = vpi::rectIntersect(C, R),
                I = vpidomains::rct(0, 0, 0, 0), !,
                C = vpidomains::rct(Lc, Tc, Rc, Bc),
                R = vpidomains::rct(Lr, Tr, Rr, Br),
                Dx =(Lc - Lr + Rc - Rr) div 2,
                Dy =(Tc - Tr + Bc - Br) div 2.
        search_rect_(C, R, Dx, Dy):-
                U = vpi::rectUnion(C, R),
                U = vpidomains::rct(Lu, Tu, Ru, Bu),
                C = vpidomains::rct(Lc, Tc, Rc, Bc),
                Dx = Lc - Lu + Rc - Ru,
                Dy = Tc - Tu + Bc - Bu.

        search_base_clean_up(Win):-
                retractall(search_sel(Win, _)),
                retractall(search(Win, _, _, _)),
                retractall(search_rct(Win, _)).

/******************************************************************************

                WRITING A TREE

******************************************************************************/
constants
        write_lr        : integer = 0.
        write_ud        : integer = 1.

        write_          : integer = 0.
        write_search    : integer = 1.
        write_searched  : integer = 2.
        write_selected  : integer = 3.

class facts- write_base
        write_mode : (vpiDomains::windowHandle, integer).
        tree_link : (vpiDomains::windowHandle, integer, integer, tree_LINK).
        global_link : (vpiDomains::windowHandle, integer, integer).

class facts- write_check_stop
        write_stop_after_row : (vpiDomains::windowHandle).
        write_stop : (vpiDomains::windowHandle).

class predicates
                write_tree : (vpiDomains::windowHandle) determ(i).
        write_tree_check_stop : (vpiDomains::windowHandle, tree_node, integer, integer) procedure(i,i,i,i).
        write_tree_right : (vpiDomains::windowHandle, integer, integer, integer) procedure(i,i,i,i).
        write_tree_left : (vpiDomains::windowHandle, integer, integer, integer) procedure(i,i,i,i).
        write_tree_up : (vpiDomains::windowHandle, integer, integer, integer) procedure(i,i,i,i).
        write_tree_down : (vpiDomains::windowHandle, integer, integer, integer) procedure(i,i,i,i).
        write_drawing : (vpiDomains::windowHandle, vpiDomains::rct, integer, integer, integer, selector, core::booleanInt, arrow_type) procedure(i,i,i,i,i,i,i,i).
        write_drawing : (vpiDomains::windowHandle, integer, integer, integer, integer, integer, selector, core::booleanInt, arrow_type, vpiDomains::rct, vpiDomains::rct) procedure(i,i,i,i,i,i,i,i,i,i,i).
        write_frame : (vpiDomains::windowHandle, vpiDomains::rct, core::booleanInt, arrow_type) determ(i,i,i,i).
        write_get_pen : (arrow_type, vpiDomains::pen) determ(i,o).
        write_get_pen : (arrow_type, arrow_pen_list, vpiDomains::pen) determ(i,i,o).
        write_link : (vpiDomains::windowHandle, core::booleanInt, tree_link, integer, integer, integer, integer) determ(i,i,i,i,i,i,i).
        write_link : (vpiDomains::windowHandle, core::booleanInt, integer, integer, integer, integer) determ(i,i,i,i,i,i).
        write_link : (vpiDomains::windowHandle, integer, integer, integer) procedure(i,i,i,i).
        write_line : (vpiDomains::windowHandle, integer, integer, integer, integer, integer) determ(i,i,i,i,i,i).
        write_set_mode : (vpiDomains::windowHandle, integer) procedure(i,i).
        write_base_clean_up : (vpiDomains::windowHandle) procedure(i).

clauses
        write_tree(Win):-
                retractall(write_stop_after_row(Win)),
                retractall(write_stop(Win)), fail.
        write_tree(Win):-
                screen_context(Win, XOff, YOff, tree_DirRight), !,
                RCT = vpi::winGetClientRect(Win),
                RCT = vpidomains::rct(_, Top, _, Bottom),
                write_tree_right(Win, XOff, YOff, Bottom),
                write_link(Win, Top, Bottom, write_lr).
        write_tree(Win):-
                screen_context(Win, XOff, YOff, tree_DirLeft), !,
                tree_size(Win, TreeWidth, _, _, _, _), !,
                XOff_ = XOff - TreeWidth,
                RCT = vpi::winGetClientRect(Win),
                RCT = vpidomains::rct(_, Top, _, Bottom),
                write_tree_left(Win, XOff_, YOff, Bottom),
                write_link(Win, Top, Bottom, write_lr).
        write_tree(Win):-
                screen_context(Win, XOff, YOff, tree_DirDown), !,    % AK 08.06.01 FB18759 %tree_DirUp
                RCT = vpi::winGetClientRect(Win),
                RCT = vpidomains::rct(Left, _, Right, _),
                write_tree_Down(Win, XOff, YOff, Right),     % AK 08.06.01 FB18759 %write_tree_Up
                write_link(Win, Left, Right, write_ud).
        write_tree(Win):-
                screen_context(Win, XOff, YOff, tree_DirUp), !,    % AK 08.06.01 FB18759 %tree_DirDown
                tree_size(Win, _, _, _, TreeHeight, _), !,
                YOff_ = YOff - TreeHeight,
                RCT = vpi::winGetClientRect(Win),
                RCT = vpidomains::rct(Left, _, Right, _),
                write_tree_up(Win, XOff, YOff_, Right),     % AK 08.06.01 FB18759 %write_tree_down
                write_link(Win, Left, Right, write_ud).

        write_tree_check_stop(Win, row_mark(_ , _, _), _, _):-
                retract(write_stop_after_row(Win)), !,
                retractall(write_stop(Win)),
                assert(write_stop(Win)).
        write_tree_check_stop(Win, node(_, _, Row, _, _, _, _, _, _, _, _, _, _, _, _), StopRow, write_lr):-
                Row > StopRow, !,
                retractall(write_stop_after_row(Win)),
                assert(write_stop_after_row(Win)).
        write_tree_check_stop(Win, node(_, _, _, _, _, Row, _, _, _, _, _, _, _, _, _), StopRow, write_ud):-
                Row > StopRow, !,
                retractall(write_stop_after_row(Win)),
                assert(write_stop_after_row(Win)).
        write_tree_check_stop(_, _, _, _).

        write_tree_right(Win, XOff, _, _):-
                write_mode(Win, write_),
                global_link(Win, Col, _),
                Col_ = Col - XOff - tree_link_width,
                assert(tree_link(Win, Col_, -1, begin)), fail.
        write_tree_right(Win, XOff, YOff, Bottom):-
                tree_below(Win, NODE),
                StopRow = Bottom + YOff,
                write_tree_check_stop(Win, NODE, StopRow, write_lr),
                NODE = node(Col, Col_, Row, H, H_, _, I, J, K, Drawing, Closed, In, Out, Link, ArrowType),
                % Pro/3
                sx::sxUnpack(Drawing,'{',Ld),
                Ld=[Draw|_],
                %
                Y = Row - YOff,
                L = Col - XOff,
                R = Col_ - XOff,
                T = Y -(H_ - H) div 2,
                B = T + H_ - H,
                RCT = vpidomains::rct(L, T, R, B),
                write_drawing(Win, RCT, I, J, K, Draw, Closed, ArrowType),
                not(write_mode(Win, write_search)),
                RCT = vpidomains::rct(Xb, _, Xc, _),
                Xa = Xb - tree_link_width,
                Xd = Xc + tree_link_width,
                write_link(Win, In, Link, Xa, Xb, Y, write_lr),
                write_link(Win, Out, Xc, Xd, Y, write_lr),
                retract(write_stop(Win)), !.
        write_tree_right(_, _, _, _).

        write_tree_left(Win, XOff, _, _):-
                write_mode(Win, write_),
                global_link(Win, Col, _),
                Col_ = -Col - XOff + tree_link_width,
                assert(tree_link(Win, Col_, -1, begin)), fail.
        write_tree_left(Win, XOff, YOff, Bottom):-
                tree_below(Win, NODE),
                StopRow = Bottom + YOff,
                write_tree_check_stop(Win, NODE, StopRow, write_lr),
                NODE = node(Col, Col_, Row, H, H_, _, I, J, K, Drawing, Closed, In, Out, Link, ArrowType),
                % Pro/3
                sx::sxUnpack(Drawing,'{',Ld),
                Ld=[Draw|_],
                %
                Y = Row - YOff,
                L = -Col_ - XOff,
                R = -Col - XOff,
                T = Y -(H_ - H) div 2,
                B = T + H_ - H,
                RCT = vpidomains::rct(L, T, R, B),
                write_drawing(Win, RCT, I, J, K, Draw, Closed, ArrowType),
                not(write_mode(Win, write_search)),
                RCT = vpidomains::rct(Xc, _, Xb, _),
                Xa = Xb + tree_link_width,
                Xd = Xc - tree_link_width,
                write_link(Win, In, Link, Xa, Xb, Y, write_lr),
                write_link(Win, Out, Xc, Xd, Y, write_lr),
                retract(write_stop(Win)), !.
        write_tree_left(_, _, _, _).

        write_tree_down(Win, _, YOff, _):-  % AK 08.06.01 FB18759 %write_tree_up
                write_mode(Win, write_),
                global_link(Win, _, Col),
                Col_ = Col - YOff - tree_link_width,
                assert(tree_link(Win, Col_, -1, begin)), fail.
        write_tree_down(Win, XOff, YOff, Right):-  % AK 08.06.01 FB18759 %write_tree_up
                tree_below(Win, NODE),
                StopRow = Right + XOff,
                write_tree_check_stop(Win, NODE, StopRow, write_ud),
                NODE = node(H, H_, _, Col, Col_, Row, I, J, K, Drawing, Closed, In, Out, Link, ArrowType),
                % Pro/3
                sx::sxUnpack(Drawing,'{',Ld),
                Ld=[Draw|_],
                %
                X = Row - XOff,
                L = X -(H_ - H) div 2,
                R = L + H_ - H,
                T = Col - YOff,
                B = Col_ - YOff,
                RCT = vpidomains::rct(L, T, R, B),
                write_drawing(Win, RCT, I, J, K, Draw, Closed, ArrowType),
                not(write_mode(Win, write_search)),
                RCT = vpidomains::rct(_, Yb, _, Yc),
                Ya = Yb - tree_link_width,
                Yd = Yc + tree_link_width,
                write_link(Win, In, Link, Ya, Yb, X, write_ud),
                write_link(Win, Out, Yc, Yd, X, write_ud),
                retract(write_stop(Win)), !.
        write_tree_down(_, _, _, _).  % AK 08.06.01 FB18759 %write_tree_up

        write_tree_up(Win, _, YOff, _):-  % AK 08.06.01 FB18759 %write_tree_down
                write_mode(Win, write_),
                global_link(Win, _, Col),
                Col_ = -Col - YOff + tree_link_width,
                assert(tree_link(Win, Col_, -1, begin)), fail.
        write_tree_up(Win, XOff, YOff, Right):-  % AK 08.06.01 FB18759 %write_tree_down
                tree_below(Win, NODE),
                StopRow = Right + XOff,
                write_tree_check_stop(Win, NODE, StopRow, write_ud),
                NODE = node(H, H_, _, Col, Col_, Row, I, J, K, Draw, Closed, In, Out, Link, ArrowType),
                % Pro/3
                sx::sxUnpack(Draw,'{',Ld),
                Ld=[Drw|_],
                %
                X = Row - XOff,
                L = X -(H_ - H) div 2,
                R = L + H_ - H,
                T = -Col_ - YOff,
                B = -Col - YOff,
                RCT = vpidomains::rct(L, T, R, B),
                write_drawing(Win, RCT, I, J, K, Drw, Closed, ArrowType),
                not(write_mode(Win, write_search)),
                RCT = vpidomains::rct(_, Yc, _, Yb),
                Ya = Yb + tree_link_width,
                Yd = Yc - tree_link_width,
                write_link(Win, In, Link, Ya, Yb, X, write_ud),
                write_link(Win, Out, Yc, Yd, X, write_ud),
                retract(write_stop(Win)), !.
        write_tree_up(_, _, _, _).  % AK 08.06.01 FB18759 %write_tree_down

        write_drawing(Win, RCT, I, J, K, Drawing, Closed, ArrowType):- !,
                RCT = vpidomains::rct(L, T, R, B),
                L_ = L + tree_frame_hor,
                T_ = T + tree_frame_ver,
                R_ = R - tree_frame_hor,
                B_ = B - tree_frame_ver,
                RCT_ = vpidomains::rct(L_, T_, R_, B_),
                vpi::winGetFontMetrics(Win, _, Ascent, _),
                Xt = L_,
                Yt = T_ + Ascent,
                write_drawing(Win, Xt, Yt, I, J, K, Drawing, Closed, ArrowType, RCT, RCT_).

        write_drawing(Win, Xt, Yt, I, J, K, Drawing, Closed, ArrowType, RCT, RCT_):-
                write_mode(Win, write_),
                write_frame(Win, RCT, Closed, ArrowType),
                selected(Win, I, J, K), !,
                DRAWTOOLS = vpi::winGetDrawTools(Win),
                vpi::winSetBrush(Win, vpidomains::brush(pat_Solid, tree_selected_node_color)),
                vpi::winSetPen(Win, vpidomains::pen(1, ps_Solid, tree_selected_node_color)),
                vpi::winSetForeColor(Win, tree_selected_text_color),
                vpi::drawRect(Win, RCT_),
                vpi::drawText(Win, Xt, Yt, Drawing),
                vpi::winSetDrawTools(Win, DRAWTOOLS).
        write_drawing(Win, Xt, Yt, _, _, _, Drawing, _, _, _, _):-
                write_mode(Win, write_), !,
                vpi::drawText(Win, Xt, Yt, Drawing).
        write_drawing(Win, _, _, I, J, K, _, _, _, RCT, RCT_):-
                write_mode(Win, write_selected),
                hit_point(Win, PNT),
                vpi::rectPntInside(RCT, PNT), !,
                retractall(selected(Win, _, _, _)),
                assert(selected(Win, I, J, K)),
                assert(invalid_selection(Win, RCT_)).
        write_drawing(Win, _, _, I, J, K, _, _, _, _, RCT_):-
                write_mode(Win, write_selected),
                selected_old(Win, I, J, K), !,
                assert(invalid_selection(Win, RCT_)).
        write_drawing(Win, _, _, I, J, K, _, _, _, _, RCT_):-
                write_mode(Win, write_searched),
                selected_old(Win, I, J, K), !,
                assert(invalid_selection(Win, RCT_)).
        write_drawing(Win, _, _, I, J, K, _, _, _, RCT, RCT_):-
                write_mode(Win, write_search),
                search(Win, I, J, K), !,
                write_set_mode(Win, write_searched),
                retractall(selected(Win, _, _, _)),
                assert(selected(Win, I, J, K)),
                assert(invalid_selection(Win, RCT_)),
                retractall(search_rct(Win, _)),
                assert(search_rct(Win, RCT)).
        write_drawing(Win, _, _, I, J, K, _, _, _, _, RCT_):-
                write_mode(Win, write_search),
                selected_old(Win, I, J, K), !,
                assert(invalid_selection(Win, RCT_)).
        write_drawing(_, _, _, _, _, _, _, _, _, _, _).

        write_frame(Win, RCT, b_true, ArrowType):- !,
                R_hor = 2 * tree_frame_hor,
                R_ver = 2 * tree_frame_ver,
                Pen = vpi::winGetPen(Win),
                write_get_pen(ArrowType, Pen_),
                vpi::winSetPen(Win, Pen_),
                vpi::drawRoundRect(Win, RCT, R_hor, R_ver),
                vpi::winSetPen(Win, Pen).
        write_frame(Win, RCT, b_false, ArrowType):- !,
                Pen = vpi::winGetPen(Win),
                write_get_pen(ArrowType, Pen_),
                vpi::winSetPen(Win, Pen_),
                vpi::drawRect(Win, RCT),
                vpi::winSetPen(Win, Pen).

        write_get_pen(ArrowType, Pen):-
                win_info(_, tree_WinInfo(_, _, List)), !,
                write_get_pen(ArrowType, List, Pen).
        write_get_pen(ArrowType, Pen):-
                win_info(_, tree_WinInfoAdv(_, _, List, _, _)), !,
                write_get_pen(ArrowType, List, Pen).
        write_get_pen(_, vpidomains::pen(1, ps_Solid, tree_frame_color)).

        write_get_pen(_, [], vpidomains::pen(1, ps_Solid, tree_frame_color)):- !.
        write_get_pen(ArrowType, [ tree_ArrowPen(ArrowType, Pen) | _ ], Pen):- !.
        write_get_pen(ArrowType, [ _ | T ], Pen):-
                write_get_pen(ArrowType, T, Pen).

        write_link(Win, _, _, _, _, _, _):-
                write_mode(Win, MODE),
                MODE <> write_, !.
        write_link(Win, b_true, _, ColA, ColB, Row, Dir):-
                write_line(Win, ColA, Row, ColB, Row, Dir), fail.
        write_link(Win, b_true, begin, Col, _, Row, _):- !,
                assert(tree_link(Win, Col, Row, begin)).
        write_link(Win, b_true, end_, Col, _, Row1, Dir):-
                retract(tree_link(Win, Col, Row0, begin)), !,
                write_line(Win, Col, Row0, Col, Row1, Dir).
        write_link(Win, b_true, end_, Col, _, Row, _):- !,
                assert(tree_link(Win, Col, Row, end_)).
        write_link(Win, b_true, cross, Col, _, Row1, Dir):-
                tree_link(Win, Col, Row0, begin), !,
                write_line(Win, Col, Row0, Col, Row1, Dir).
        write_link(Win, b_true, cross, Col, _, Row, _):- !,
                assert(tree_link(Win, Col, Row, end_)),
                assert(tree_link(Win, Col, Row, begin)).
        write_link(_, _, _, _, _, _, _).

        write_link(Win, _, _, _, _, _):-
                write_mode(Win, MODE),
                MODE <> write_, !.
        write_link(Win, b_true, ColC, ColD, Row, Dir):- !,
                write_line(Win, ColC, Row, ColD, Row, Dir).
        write_link(_, _, _, _, _, _).

        write_link(Win, _, _, _):-
                write_mode(Win, MODE),
                MODE <> write_, !.
        write_link(Win, Row1, _, Dir):-
                retract(tree_link(Win, Col, Row0, end_)),
                write_line(Win, Col, Row0, Col, Row1, Dir), fail.
        write_link(Win, _, Row1, Dir):-
                retract(tree_link(Win, Col, Row0, begin)),
                write_line(Win, Col, Row0, Col, Row1, Dir), fail.
        write_link(_, _, _, _).

        write_line(Win, _, _, _, _, _):-
                vpi::winSetPen(Win, vpidomains::pen(1, ps_Solid, tree_link_color)), fail.
        write_line(Win, Col0, Row0, Col1, Row1 , write_lr):- !,
                vpi::drawLine(Win, vpidomains::pnt(Col0, Row0), vpidomains::pnt(Col1 , Row1)).
        write_line(Win, Col0, Row0, Col1, Row1 , write_ud):- !,
                vpi::drawLine(Win, vpidomains::pnt(Row0, Col0), vpidomains::pnt(Row1 , Col1)).

        write_set_mode(Win, MODE):-
                retractall(write_mode(Win, _)),
                assert(write_mode(Win, MODE)), fail.
        write_set_mode(Win, write_selected):- !,
                selection_set_old(Win).
        write_set_mode(Win, write_search):- !,
                selection_set_old(Win).
        write_set_mode(_, _).

        write_base_clean_up(Win):-
                retractall(write_mode(Win, _)),
                retractall(tree_link(Win, _, _, _)),
                retractall(global_link(Win, _, _)).

/******************************************************************************

                predicates TO MOVE TREE BASE

******************************************************************************/
class predicates
        move_base : (vpiDomains::windowHandle) determ(i).
        move_base_lr : (vpiDomains::windowHandle, integer) determ(i,i).
        move_base_ud : (vpiDomains::windowHandle, integer) determ(i,i).
        move_above_lr : (vpiDomains::windowHandle, integer) procedure(i,i).
        move_below_lr : (vpiDomains::windowHandle, integer) determ(i,i).
        move_above_ud : (vpiDomains::windowHandle, integer) procedure(i,i).
        move_below_ud : (vpiDomains::windowHandle, integer) determ anyflow.
        move_global_link : (vpiDomains::windowHandle, tree_node, tree_link, tree_link) determ(i,i,i,i).

clauses
        move_base(Win):-
                screen_context(Win, _, YOff, tree_DirRight), !,
                move_base_lr(Win, YOff).
        move_base(Win):-
                screen_context(Win, _, YOff, tree_DirLeft), !,
                move_base_lr(Win, YOff).
        move_base(Win):-
                screen_context(Win, XOff, _, tree_DirUp), !,
                move_base_ud(Win, XOff).
        move_base(Win):-
                screen_context(Win, XOff, _, tree_DirDown), !,
                move_base_ud(Win, XOff).

        move_base_lr(Win, Row0):-
                tree_above(Win, row_mark(Row1, _, _)),
                Row0 <= Row1, !,
                move_above_lr(Win, Row0).
        move_base_lr(Win, Row0):-
                tree_below(Win, row_mark(Row1, _, _)),
                Row0 > Row1, !,
                move_below_lr(Win, Row0).
        move_base_lr(_, _).

        move_base_ud(Win, Row0):-
                tree_above(Win, row_mark(_, Row1, _)),
                Row0 <= Row1, !,
                move_above_ud(Win, Row0).
        move_base_ud(Win, Row0):-
                tree_below(Win, row_mark(_, Row1, _)),
                Row0 > Row1, !,
                move_below_ud(Win, Row0).
        move_base_ud(_, _).

        move_above_lr(Win, Row0):-
                retract(tree_above(Win, NODE)),
                asserta(tree_below(Win, NODE)),
                move_global_link(Win, NODE, end_, begin),
                NODE = row_mark(Row1, _, _),
                Row0 > Row1, !.
        move_above_lr(_, _).

        move_below_lr(Win, Row0):-
                retract(tree_below(Win, NODE)),
                asserta(tree_above(Win, NODE)),
                move_global_link(Win, NODE, begin, end_),
                NODE = row_mark(Row1, _, _),
                Row0 <= Row1, !,
                retract(tree_above(Win, NODE1)), !,
                asserta(tree_below(Win, NODE1)),
                move_above_lr(Win, Row1).
        move_below_lr(_, _).

        move_above_ud(Win, Row0):-
                retract(tree_above(Win, NODE)),
                asserta(tree_below(Win, NODE)),
                move_global_link(Win, NODE, end_, begin),
                NODE = row_mark(_, Row1, _),
                Row0 > Row1, !.
        move_above_ud(_, _).

        move_below_ud(Win, Row0):-
                retract(tree_below(Win, NODE)),
                asserta(tree_above(Win, NODE)),
                move_global_link(Win, NODE, begin, end_),
                NODE = row_mark(_, Row1, _),
                Row0 <= Row1, !,
                retract(tree_above(Win, NODE1)), !,
                asserta(tree_below(Win, NODE1)),
                move_above_ud(Win, Row1).
        move_below_ud(_, _).

        move_global_link(Win, node(ColLR, _, _, ColUD, _, _, _, _, _, _, _, _, _, A, _), A, _):- !,
                assert(global_link(Win, ColLR, ColUD)).
        move_global_link(Win, node(ColLR, _, _, ColUD, _, _, _, _, _, _, _, _, _, R, _), _, R):- !,
                retract(global_link(Win, ColLR, ColUD)), !.
        move_global_link(_, _, _, _).

/******************************************************************************

                SCROLL predicates

******************************************************************************/
constants
        scroll_lr       : integer = 0.
        scroll_ud       : integer = 1.

class facts- scroll_base
        h_scroll : (vpiDomains::windowHandle, real).
        v_scroll : (vpiDomains::windowHandle, real).

class predicates
        scroll_tree : (vpiDomains::windowHandle, integer, integer) determ(i,i,i).
        scroll_adjust : (vpiDomains::windowHandle, integer, integer, integer) procedure(i,i,i,o).
        scroll_get_bar : (vpiDomains::windowHandle, integer, vpiDomains::scrollType, real) determ(i,i,o,o).
        scroll_update : (vpiDomains::windowHandle, integer, integer, integer) determ(i,i,i,i).
        scroll_update_get_size : (vpiDomains::windowHandle, integer, integer, integer) determ(i,i,o,o).
        scroll_update_get_range : (vpiDomains::windowHandle, integer, integer, integer, integer, real) determ(i,i,i,i,o,o).
        scroll_update_set_scroll : (vpiDomains::windowHandle, integer, real) determ(i,i,i).
        scroll_base_clean_up : (vpiDomains::windowHandle) procedure(i).
        scroll_null : (integer,integer) determ(i,i).        % BB! 13.10.97
        get_ratio : (integer, real) procedure(i,o).

clauses

        scroll_null(0,0):-!.        % BB! 13.10.97

        scroll_tree(Win, Dx, Dy):-
                scroll_adjust(Win, scroll_lr, Dx, Dx_),
                scroll_adjust(Win, scroll_ud, Dy, Dy_),
                not(scroll_null(Dx_,Dy_)),
                retract(screen_context(Win, X, Y, Dir)), !,
                X_ = X - Dx_,
                Y_ = Y - Dy_,
                assert(screen_context(Win, X_, Y_ , Dir)),
                scroll_update(Win, Dir, X_, Y_),
                RCT = vpi::winGetClientRect(Win),
                vpi::winScroll(Win, RCT, Dx_, Dy_),
                vpi::winUpdate(Win).
        scroll_tree(_, _, _).

        scroll_adjust(_, _, 0, 0):-!.
        scroll_adjust(Win, scroll_lr, _, 0):-
                not(h_scroll(Win, _)),!.
        scroll_adjust(Win, scroll_ud, _, 0):-
                not(v_scroll(Win, _)),!.
        scroll_adjust(Win, SDir, Scroll, Pos1):-
                Scroll > 0,
                scroll_get_bar(Win, SDir, SB, Ratio),
                Pos = vpi::winGetScrollPos(Win, SB),
                Pos1 = convert(integer,Pos * Ratio),
                Pos1 <=  Scroll,!.
        scroll_adjust(Win, SDir, Scroll, Scroll1):-
                Scroll < 0,
                scroll_get_bar(Win, SDir, SB, Ratio),
                vpi::winGetScrollRange(Win, SB, _, Range),
                Pos = vpi::winGetScrollPos(Win, SB),
                Scroll1 = convert(integer,(Pos - Range) * Ratio),
                Scroll1 >= Scroll,
                !.
        scroll_adjust(_, _, Scroll, Scroll).

        scroll_get_bar(Win, scroll_lr, sb_Horz, Ratio):-
                h_scroll(Win, Ratio),!.
        scroll_get_bar(Win, scroll_ud, sb_Vert, Ratio):-
                v_scroll(Win, Ratio),!.

        scroll_update(Win, Dir, HPos, VPos):-
                RCT = vpi::winGetClientRect(Win),
                RCT = vpidomains::rct(L, T, R, B),
                SWidth = R - L,
                SHeight = B - T,
                scroll_update_get_size(Win, Dir, TWidth, THeight),
                scroll_update_get_range(Win, scroll_lr, TWidth, SWidth, HRange, HRatio),
                scroll_update_get_range(Win, scroll_ud, THeight, SHeight, VRange, VRatio),
                vpi::winSetScrollRange(Win, sb_Horz, 0, HRange),
                vpi::winSetScrollRange(Win, sb_Vert, 0, VRange),
                HPos1 = convert(integer, HPos / HRatio),
                VPos1 = convert(integer, VPos / VRatio),
                vpi::winSetScrollPos(Win, sb_Horz, HPos1),
                vpi::winSetScrollPos(Win, sb_Vert, VPos1),
                move_base(Win).

        scroll_update_get_size(Win, tree_DirRight, Width, Height):- !,
                tree_size(Win, Width, Height, _, _, _), !.
        scroll_update_get_size(Win, tree_DirLeft, Width, Height):- !,
                tree_size(Win, Width, Height, _, _, _), !.
        scroll_update_get_size(Win, tree_DirUp, Width, Height):- !,
                tree_size(Win, _, _, Width, Height, _), !.
        scroll_update_get_size(Win, tree_DirDown, Width, Height):- !,
                tree_size(Win, _, _, Width, Height, _), !.

        scroll_update_get_range(Win, Dir, T, S, Range1, Ratio):-
                T > S, !,
                Range = T - S,
                get_ratio(Range, Ratio),
                Range1 = convert(integer,Range / Ratio),
                scroll_update_set_scroll(Win, Dir, Ratio).
        scroll_update_get_range(Win, scroll_lr, _, _, 0, 1.0):-!,
                retractall(h_scroll(Win, _)).
        scroll_update_get_range(Win, scroll_ud, _, _, 0, 1.0):-!,
                retractall(v_scroll(Win, _)).

        scroll_update_set_scroll(Win, scroll_lr, Ratio):-
                retract(h_scroll(Win, _)),!,
                assert(h_scroll(Win, Ratio)).
        scroll_update_set_scroll(Win, scroll_lr, Ratio):-!,
                assert(h_scroll(Win, Ratio)).
        scroll_update_set_scroll(Win, scroll_ud, Ratio):-
                retract(v_scroll(Win, _)),!,
                assert(v_scroll(Win, Ratio)).
        scroll_update_set_scroll(Win, scroll_ud, Ratio):-!,
                assert(v_scroll(Win, Ratio)).

        scroll_base_clean_up(Win):-
                retractall(h_scroll(Win, _)),
                retractall(v_scroll(Win, _)).

        get_ratio(Range, 1.0):-
                Range <= 32767,!.
        get_ratio(Range, Ratio):-
                Ratio = Range / 32767,!.

/******************************************************************************

                SIZE HANDLING

******************************************************************************/
class predicates
        resize_tree : (vpiDomains::windowHandle) procedure(i).
        resize_offset : (integer, integer, integer, integer) procedure(i,i,i,o).
        resize_invalidate : (vpiDomains::windowHandle, integer, integer, integer, integer) procedure(i,i,i,i,i).

clauses
        resize_tree(Win):-
                retract(screen_context(Win, X, Y, Dir)),
                scroll_update_get_size(Win, Dir, TWidth, THeight),
                RCT = vpi::winGetClientRect(Win),
                RCT = vpidomains::rct(L, T, R, B),
                SWidth = R - L,
                SHeight = B - T,
                resize_offset(TWidth, SWidth, X, X_),
                resize_offset(THeight, SHeight, Y, Y_),
                assert(screen_context(Win, X_, Y_, Dir)),
                scroll_update(Win, Dir, X_, Y_),
                resize_invalidate(Win, X, X_, Y, Y_), !.
        resize_tree(_).

        resize_offset(T, S, _, 0):-
                S >= T, !.
        resize_offset(_, _, O, 0):-
                O <= 0, !.
        resize_offset(T, S, O, O_):-
                S + O >= T, !,
                O_ = T - S.
        resize_offset(_, _, O, O).

        resize_invalidate(_, X, X, Y, Y):- !.
        resize_invalidate(Win, _, _, _, _):-
                vpi::winInvalidate(Win).

class facts- closed_base
        closed_above : (vpiDomains::windowHandle, tree_node).
        closed_below : (vpiDomains::windowHandle, tree_node).

class predicates
        closed_base_clean_up : (vpiDomains::windowHandle) procedure(i).
        closed_base_copy : (vpiDomains::windowHandle From, vpiDomains::windowHandle To) procedure(i,i).

clauses
        closed_base_clean_up(Win):-
                retractall(closed_above(Win, _)),
                retractall(closed_below(Win, _)).

        closed_base_copy(From, To):-
                closed_above(From, NODE),
                assertz(closed_above(To, NODE)), fail.
        closed_base_copy(From, To):-
                closed_below(From, NODE),
                assertz(closed_below(To, NODE)), fail.
        closed_base_copy(_, _).

/******************************************************************************

                TREE EVENT HANDLER

******************************************************************************/
class predicates
        tree_no_handler : vpiDomains::ehandler.
        tree_user_call : vpiDomains::ehandler.
        tree_handler : vpiDomains::ehandler.

clauses
        tree_no_handler(_, _) = _ :- !, fail.

        tree_user_call(W, Event) = Ret :-
                tree_callback(W, CALLBACK), !,
                Ret = CALLBACK(W, Event).

        tree_handler(W, vpidomains::e_Create(CrData)) = Ret :- !,
                retract(convert_window(Win)),
                tree_base_copy(Win, W),
                tree_base_clean_up(Win),
                selection_base_copy(Win, W),
                selection_base_clean_up(Win),
                retract(win_info(_, I)), !,
                assert(win_info(W, I)),
                I = tree_WinInfo(FONT, _, _),
                vpi::winSetFont(W, FONT),
                resize_tree(W),
                Ret = tree_user_call(W, vpidomains::e_Create(CrData)).

        tree_handler(W, vpidomains::e_Destroy) = Ret :- !,
                retract(tree_callback(W, CALLBACK)), !,
                tree_base_clean_up(W),
                selection_base_clean_up(W),
                search_base_clean_up(W),
                write_base_clean_up(W),
                scroll_base_clean_up(W),
                closed_base_clean_up(W),
                Ret = CALLBACK(W, vpidomains::e_Destroy).

        tree_handler(W, vpidomains::e_Update(UpData)) = 0 :-
                Tree_BackColor = vpi::getAttrVal(attr_color_window),
                vpi::winSetPen(W, vpidomains::pen(1, ps_Solid, Tree_backcolor)),
                vpi::winSetBrush(W, vpidomains::brush(pat_Solid, Tree_backcolor)) ,
                vpi::drawRect(W, UpData),
                _ = tree_user_call(W, vpidomains::e_Update(UpData)), fail.
        tree_handler(W, vpidomains::e_Update(_)) = 0 :- !,
                write_tree(W).

        tree_handler(W, vpidomains::e_EraseBackground) = Ret :- !,
                Ret = tree_user_call(W, vpidomains::e_EraseBackground).

        tree_handler(W, vpidomains::e_CloseRequest) = Ret :- !,
                Ret = tree_user_call(W, vpidomains::e_CloseRequest).

        tree_handler(W, vpidomains::e_EndSession(R)) = Ret :- !,
                Ret = tree_user_call(W, vpidomains::e_EndSession(R)).

        tree_handler(W, vpidomains::e_Size(Width, Height)) = Ret :- !,
                resize_tree(W),
                Ret = tree_user_call(W, vpidomains::e_Size(Width, Height)).

        tree_handler(W, Event) = Ret :-
                Ret = tree_user_call(W, Event), !.

        tree_handler(W, vpidomains::e_MouseDown(PNT, _, _)) = 0 :- !,
                retractall(hit_point(W, _)),
                assert(hit_point(W, PNT)),

                write_set_mode(W, write_selected),
                write_tree(W),
                write_set_mode(W, write_),

                retractall(hit_point(W, _)),
                selection_update(W, b_true).

        tree_handler(W, vpidomains::e_Char(k_Down, _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_VScroll(sc_LineDown, 0)).
        tree_handler(W, vpidomains::e_Char(k_Up, _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_VScroll(sc_LineUp, 0)).
        tree_handler(W, vpidomains::e_Char(k_Right, _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_HScroll(sc_LineDown, 0)).
        tree_handler(W, vpidomains::e_Char(k_Left, _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_HScroll(sc_LineUp, 0)).
        tree_handler(W, vpidomains::e_Char(k_PgUp, _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_VScroll(sc_PageUp, 0)).
        tree_handler(W, vpidomains::e_Char(k_PgDown , _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_VScroll(sc_PageDown, 0)).
        tree_handler(W, vpidomains::e_Char(k_Home, _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_HScroll(sc_PageUp, 0)).
        tree_handler(W, vpidomains::e_Char(k_End, _)) = 0 :-!,
                _ = vpi::winSendEvent(W, vpidomains::e_HScroll(sc_PageDown, 0)).

        tree_handler(W, vpidomains::e_HScroll(sc_LineDown, _)) = 0 :- !,
                scroll_tree(W, -tree_scroll_hor, 0).
        tree_handler(W, vpidomains::e_HScroll(sc_LineUp, _)) = 0 :- !,
                scroll_tree(W, tree_scroll_hor, 0).
        tree_handler(W, vpidomains::e_HScroll(sc_PageDown, _)) = 0 :- !,
                RCT = vpi::winGetClientRect(W),
                RCT = vpidomains::rct(L, _, R, _),
                Shift = L - R,
                scroll_tree(W, Shift, 0).
        tree_handler(W, vpidomains::e_HScroll(sc_PageUp, _)) = 0 :- !,
                RCT = vpi::winGetClientRect(W),
                RCT = vpidomains::rct(L, _, R, _),
                Shift = R - L,
                scroll_tree(W, Shift, 0).
        tree_handler(W, vpidomains::e_HScroll(sc_ThumbTrack, Pos_)) = 0:-
                screen_context(W, Pos, _, _),
                h_scroll(W, Ratio),!,
                Shift = Pos - convert(integer, Pos_ * Ratio),
                scroll_tree(W, Shift, 0).
        tree_handler(W, vpidomains::e_VScroll(sc_LineDown, _)) = 0 :- !,
                scroll_tree(W, 0, -tree_scroll_ver).
        tree_handler(W, vpidomains::e_VScroll(sc_LineUp, _)) = 0 :- !,
                scroll_tree(W, 0, tree_scroll_ver).
        tree_handler(W, vpidomains::e_VScroll(sc_PageDown, _)) = 0 :- !,
                RCT = vpi::winGetClientRect(W),
                RCT = vpidomains::rct(_, T, _, B),
                Shift = T - B,
                scroll_tree(W, 0, Shift).
        tree_handler(W, vpidomains::e_VScroll(sc_PageUp, _)) = 0 :- !,
                RCT = vpi::winGetClientRect(W),
                RCT = vpidomains::rct(_, T, _, B),
                Shift = B - T,
                scroll_tree(W, 0, Shift).
        tree_handler(W, vpidomains::e_VScroll(sc_ThumbTrack, Pos_)) = 0:-
                screen_context(W, _, Pos, _),
                v_scroll(W, Ratio),!,
                Shift = Pos - convert(integer, Pos_ * Ratio),
                scroll_tree(W, 0, Shift).

/******************************************************************************

                PRINT predicates

******************************************************************************/
class facts- print_base
        print_bottom : (integer, integer, integer) determ.

class predicates
        print_tree : (vpiDomains::windowHandle, integer, integer, integer, string, integer, integer, vpiDomains::font) determ anyflow.
        print_tree_page : (vpiDomains::windowHandle, integer, integer, integer, integer, vpiDomains::font) procedure(i,i,i,i,i,i).
        print_tree_node : (vpiDomains::windowHandle, tree_node, integer, integer, integer, integer) determ anyflow.
        print_link : (vpiDomains::windowHandle, integer, integer, core::booleanInt, tree_link, integer, integer, integer) procedure(i,i,i,i,i,i,i,i).
        print_link : (vpiDomains::windowHandle, integer, integer, core::booleanInt, integer, integer, integer) procedure(i,i,i,i,i,i,i).
        print_link : (vpiDomains::windowHandle, integer, integer) procedure(i,i,i).
        print_line : (vpiDomains::windowHandle, integer, integer, integer, integer, integer, integer) procedure(i,i,i,i,i,i,i).
        print_frame : (vpiDomains::windowHandle, vpiDomains::rct, core::booleanInt, integer) determ anyflow.
        print_tree_adjust : (vpiDomains::windowHandle, integer) determ anyflow.
        print_move_closed : (vpiDomains::windowHandle) determ anyflow.
        print_node_info : (vpiDomains::windowHandle, selector, integer, integer, integer, integer, arrow_type) determ anyflow.
        print_get_parent : (vpiDomains::windowHandle, integer, integer, integer, tree_node) nondeterm anyflow.
        print_get_max_parent : (tree_node_list, tree_node) determ anyflow.

clauses
        print_tree(Win, StartX, StartY, HeaderHei, Title, PageNo, PageHeight, Font):-
                retract(screen_context(Win, XOff, YOff, _)), !,
                vpi::printStartPage(Win),
                VRP = convert(integer, vpi::getAttrVal(attr_printer_vres)),
                HRP = convert(integer, vpi::getAttrVal(attr_printer_hres)),
                VRS = convert(integer, vpi::getAttrVal(attr_screen_vres)),
                HRS = convert(integer, vpi::getAttrVal(attr_screen_hres)),
                vpi::winSetMapMode(Win, mm_arbitrary),
                vpi::winSetMapScale(Win, vpidomains::pnt(0, 0), vpidomains::pnt(VRS, HRS), vpidomains::pnt(0, 0), vpidomains::pnt(VRP, HRP)),
                vpi::winSetFont(Win, Font),
                string5x::format(Header, "%s  Page -%d-", Title, PageNo),
                vpi::drawText(Win, StartX, StartY, Header),
                YOff_ = YOff - HeaderHei,
                print_tree_page(Win, XOff, YOff_, HeaderHei, PageHeight, Font),
                print_link(Win, HeaderHei, PageHeight),
                vpi::printEndPage(Win),
                PageNo_ = PageNo + 1,
                print_tree(Win, StartX, StartY, HeaderHei, Title, PageNo_, PageHeight, Font).
        print_tree(_, _, _, _, _, _, _, _).

        print_tree_page(Win, XOff, _, Top, _, Font):-
                VRP = convert(integer, vpi::getAttrVal(attr_printer_vres)),
                HRP = convert(integer, vpi::getAttrVal(attr_printer_hres)),
                VRS = convert(integer, vpi::getAttrVal(attr_screen_vres)),
                HRS = convert(integer, vpi::getAttrVal(attr_screen_hres)),
                vpi::winSetMapMode(Win, mm_arbitrary),
                vpi::winSetMapScale(Win, vpidomains::pnt(0, 0), vpidomains::pnt(VRS, HRS), vpidomains::pnt(0, 0), vpidomains::pnt(VRP, HRP)),
                vpi::winSetFont(Win, Font),
                global_link(Win, Col, _),
                Col_ = Col - XOff - tree_link_width,
                assert(tree_link(Win, Col_, Top, begin)), fail.
        print_tree_page(Win, XOff, YOff, Top, Bottom, _):-
                tree_below(Win, NODE),
                print_tree_node(Win, NODE, XOff, YOff, Top, Bottom), !.
        print_tree_page(_, _, _, _, _, _).

        print_tree_node(Win, row_mark(MarkLR, _, _), XOff, YOff, _, Bottom):-
                MarkLR > Bottom + YOff, !,
                print_bottom(PrintBottom0, _, PrintBottom2),
                YOff_ =(PrintBottom0 + PrintBottom2) div 2,
                 assert(screen_context(Win, XOff, YOff_, tree_DirRight)),
                move_base(Win).
        print_tree_node(_, row_mark(MarkLR2, _, _), _, _, _, _):-
                retract(print_bottom(_, MarkLR0, MarkLR1)),
                assert(print_bottom(MarkLR0, MarkLR1, MarkLR2)), !, fail.
        print_tree_node(Win, NODE, XOff, YOff, Top, _):-
                NODE    = node(Col, Col_, Row, H, H_, _, _, _, _, Drawing, Closed, In, Out, Link, ArrowType),
                Y = Row - YOff,
                L = Col - XOff,
                R = Col_ - XOff,
                T = Y -(H_ - H) div 2,
                B = T + H_ - H,
                RCT = vpidomains::rct(L, T, R, B),
                RCT = vpidomains::rct(Xb, _, Xc, _),
                Xa = Xb - tree_link_width,
                Xd = Xc + tree_link_width,
                print_link(Win, T, Top, In, Link, Xa, Xb, Y),
                print_link(Win, T, Top, Out, Xc, Xd, Y),
                T > Top, !,
                vpi::winGetFontMetrics(Win, _, Ascent, _),
                Xt = L + tree_frame_hor,
                Yt = T + tree_frame_ver + Ascent,
                print_frame(Win, RCT, Closed, ArrowType),
                % Pro/3
                sx::sxUnpack(Drawing,'{',DrawingL), DrawingL=[DrawingX|_],
                vpi::drawText(Win, Xt, Yt, DrawingX), !, fail.

        print_link(Win, T, Top, b_true, _, ColA, ColB, Row):-
                print_line(Win, T, Top, ColA, Row, ColB, Row), fail.
        print_link(Win, _, _, b_true, begin, Col, _, Row):- !,
                assert(tree_link(Win, Col, Row, begin)).
        print_link(Win, T, Top, b_true, end_, Col, _, Row1):-
                retract(tree_link(Win, Col, Row0, begin)), !,
                print_line(Win, T, Top, Col, Row0, Col, Row1).
        print_link(Win, _, _, b_true, end_, Col, _, Row):- !,
                assert(tree_link(Win, Col, Row, end_)).
        print_link(Win, T, Top, b_true, cross, Col, _, Row1):-
                tree_link(Win, Col, Row0, begin), !,
                print_line(Win, T, Top, Col, Row0, Col, Row1).
        print_link(Win, _, _, b_true, cross, Col, _, Row):- !,
                assert(tree_link(Win, Col, Row, end_)),
                assert(tree_link(Win, Col, Row, begin)).
        print_link(_, _, _, _, _, _, _, _).

        print_link(Win, T, Top, b_true, ColC, ColD, Row):- !,
                print_line(Win, T, Top, ColC, Row, ColD, Row).
        print_link(_, _, _, _, _, _, _).

        print_link(Win, Row1, _):-
                retract(tree_link(Win, Col, Row0, end_)),
                print_line(Win, 1, 0, Col, Row0, Col, Row1), fail.
        print_link(Win, _, Row1):-
                retract(tree_link(Win, Col, Row0, begin)),
                print_line(Win, 1, 0, Col, Row0, Col, Row1), fail.
        print_link(_, _, _).

        print_line(_, T, Top, _, _, _, _):-
                T <= Top, !.
        print_line(Win, _, _, _, _, _, _):-
                vpi::winSetPen(Win, vpidomains::pen(1, ps_Solid, tree_link_color)), fail.
        print_line(Win, _, _, Col0, Row0, Col1, Row1):- !,
                vpi::drawLine(Win, vpidomains::pnt(Col0, Row0), vpidomains::pnt(Col1 , Row1)).

        print_frame(Win, RCT, b_true, ArrowType):- !,
                R_hor = 2 * tree_frame_hor,
                R_ver = 2 * tree_frame_ver,
                write_get_pen(ArrowType, Pen),
                vpi::winSetPen(Win, Pen),
                vpi::drawRoundRect(Win, RCT, R_hor, R_ver).
        print_frame(Win, RCT, b_false, ArrowType):- !,
                write_get_pen(ArrowType, Pen),
                vpi::winSetPen(Win, Pen),
                vpi::drawRect(Win, RCT).

        print_tree_adjust(Win, PageWidth):-
                print_move_closed(Win),
                tree_below(Win, node(_, ColLR, _, _, _, _, I, J, _, _, _, _, _, _, _)),
                ColLR + tree_frame_hor > PageWidth, !,
                findall(Node, print_get_parent(Win, PageWidth, I, J, Node), List),
                print_get_max_parent(List, Parent),
                Parent = node(_, ColLR_par, _, _, ColUD_par, _, I_par, J_par, K_par, Selector, _, _, _, _, _),
                convert_tree(Win, tree("", unmarked, [], 0), ColLR_par, ColUD_par, I_par, J_par, _, K_par, Selector, "", convert_close_tree_base),
                move_base(Win),
                print_tree_adjust(Win, PageWidth).
        print_tree_adjust(_, _).

        print_move_closed(Win):-
                retract(closed_above(Win, NODE)),
                asserta(closed_below(Win, NODE)), fail.
        print_move_closed(Win):-
                closed_below(Win, closed_mark(Sel, _)),
                retract(tree_size(Win, TreeWidth, TreeHeight, _, _, N0)), !,
                N1 = N0 + 1,
                N2 = N0 + 2,
                I = N0 + 3,
                N4 = N0 + 4,
                print_node_info(Win, Sel, W, W_, ColUD, ColUD_, ArrowType),
                ColLR = W_ - W + tree_frame_hor,
                Height = ColUD_ - ColUD + 2 * tree_skip,
                RowLR =(I + 1) * Height div 2,
                TreeHeight_ =(N4 + 1) div 2 * Height,
                NODE = node(tree_frame_hor, ColLR, RowLR, ColUD, ColUD_, 0, I, 0, 0, Sel, b_true, b_false, b_false, line, ArrowType),
                assertz(tree_below(Win, row_mark(TreeHeight, 0, N0))),
                assertz(tree_below(Win, row_mark(TreeHeight, 0, N1))),
                assertz(tree_below(Win, row_mark(TreeHeight, 0, N2))),
                assertz(tree_below(Win, row_mark(TreeHeight_, 0, I))),
                assertz(tree_below(Win, NODE)),
                assert(tree_size(Win, TreeWidth, TreeHeight_, 0, 0, N4)),
                convert_tree(Win, tree("", unmarked, [], 0), ColLR, 0, I, 0, _, 0, Sel, "", convert_open),
                move_base(Win),
                print_move_closed(Win).
        print_move_closed(_).

        print_node_info(Win, Sel, ColLR, ColLR_, ColUD, ColUD_, ArrowType):-
                tree_below(Win, node(ColLR, ColLR_, _, ColUD, ColUD_, _, _, _, _, Sel, _, _, _, _, ArrowType)), !.
        print_node_info(Win, Sel, ColLR, ColLR_, ColUD, ColUD_, ArrowType):-
                closed_below(Win, node(ColLR, ColLR_, _, ColUD, ColUD_, _, _, _, _, Sel, _, _, _, _, ArrowType)), !.

        print_get_parent(Win, PageWidth, I, J, NODE):-
                tree_below(Win, NODE),
                NODE = node(_, ColLR, _, _, _, _, I_, J_, _, _, _, _, _, _, _),
                I_ - J_ <= I - J,
                I_ + J_ >= I + J,
                ColLR + tree_frame_hor <= PageWidth.

        print_get_max_parent([ H ], H):- !.
        print_get_max_parent([ node(_, Col0, _, _, _, _, _, _, _, _, _, _, _, _, _) | T ], P):-
                print_get_max_parent(T, P),
                P = node(_, Col1, _, _, _, _, _, _, _, _, _, _, _, _, _),
                Col1 >= Col0, !.
        print_get_max_parent([ H | _ ], H).

/******************************************************************************

                AUXILIARY predicates

******************************************************************************/
constants
        update_scroll   : integer = 0.
        update_full     : integer = 1.

class facts- open_close_info
        closed : (selector).
        opened : (core::string_list).
        opened_checked : (core::string_list).
        selected_route : (core::string_list).
        selected_route_checked : (core::string_list).

class predicates
        close_node : (vpiDomains::windowHandle) determ anyflow.
        open_node : (vpiDomains::windowHandle) determ anyflow.
        close_opened_in_tree_base : (vpiDomains::windowHandle, integer, integer, selector) determ anyflow.
        close_opened_in_closed_base : (vpiDomains::windowHandle, integer, integer, selector) determ anyflow.
        adjust_selected : (vpiDomains::windowHandle, integer, integer, integer, integer, integer) determ anyflow.
        adjust_selected : (vpiDomains::windowHandle, integer, integer, integer, integer) determ anyflow.
        get_update_mode : (integer, core::booleanInt) determ anyflow.
        get_selector : (vpiDomains::windowHandle, selector) nondeterm anyflow.
        get_font : (win_info, vpiDomains::font) procedure(i,o).
        set_context : (vpiDomains::windowHandle, vpiDomains::rct, win_info) procedure(i,i,i).
        get_parent : (vpiDomains::windowHandle, integer, integer, integer, tree_node) determ anyflow.
        parent : (vpiDomains::windowHandle, integer, integer, integer, tree_node) nondeterm anyflow.
        parent_ : (integer, integer, integer, tree_node) determ anyflow.
        parent : (tree_node_list, tree_node) determ anyflow.
        parent__ : (integer, integer, integer, integer) determ anyflow.
        get_hscroll_height : (vpiDomains::windowHandle) -> integer  procedure(i).
        get_vscroll_width : (vpiDomains::windowHandle) -> integer  procedure(i).
        save_closed : (vpiDomains::windowHandle) determ anyflow.
        restore_closed : (vpiDomains::windowHandle) procedure(i).
        restore_closed : (vpiDomains::windowHandle, selector) determ anyflow.
        save_opened : (vpiDomains::windowHandle) procedure(i).
        check_opened_or_selected : (tree, integer) procedure(i,i).
        check_opened_or_selected : (tree, tree, core::string_list) determ anyflow.
        check_opened_or_selected_ : (tree, tree_list, core::string_list) determ anyflow.
        restore_opened_or_selected : (tree, tree, tree, core::string_list, integer) determ anyflow.
        restore_opened_or_selected : (tree, tree_list, tree_list, tree_list, core::string_list, integer) determ anyflow.
        tree_member : (core::string_list, string) determ anyflow.
        is_subpath : (core::string_list, core::string_list) determ anyflow.
        get_subtree : (tree, selector, tree_list) procedure(i,i,o).
        get_subtree_ : (tree_list, selector, tree_list) determ anyflow.
        reverse_list : (tree_list, tree_list, tree_list) determ anyflow.
        reverse_list : (core::string_list, core::string_list, core::string_list) determ anyflow.
        save_selected : (vpiDomains::windowHandle) determ anyflow.
        get_parent_list : (vpiDomains::windowHandle, integer, integer, integer, core::string_list, core::string_list) determ anyflow.
        restore_selected : (vpiDomains::windowHandle) procedure(i).
        search_by_route : (vpiDomains::windowHandle, core::string_list, integer, integer, integer, integer, integer, integer, integer) determ anyflow.
        order_list : (core::string_list, core::string_list, core::string_list) determ anyflow.
        max_selector : (selector, selector, core::string_list, core::string_list, core::string_list) determ anyflow.
        tree_GetHandler : (vpiDomains::windowHandle) -> vpiDomains::ehandler  procedure(i).

    clauses
        close_node(Win):-
                get_selected_node(Win, NODE),
                NODE = node(_, ColLR_sel, _, _, ColUD_sel, _, I_sel, J_sel, K_sel, Selector, b_false, _, b_true, _, _), !,
                I = I_sel - J_sel,
                convert_tree(Win, tree("", unmarked, [], 0), ColLR_sel, ColUD_sel, I_sel, J_sel, _, K_sel, Selector, "", convert_close_tree_base),
                set_selected(Win, I, 0, K_sel, 0, 0, update_full).

        open_node(Win):-
                selected(Win, _, 0, _),
                get_selected_node(Win, NODE),
                NODE =  node(_, ColLR_sel, _, _, ColUD_sel, _, I_sel, 0, K_sel, Selector, b_true, _, b_false, _, _),
                close_opened_in_tree_base(Win, I_sel , K_sel, Selector), !,
                close_opened_in_closed_base(Win, I_sel , K_sel, Selector),
                selected(Win, I_sel_, _ , _), !,
                convert_tree(Win, tree("", unmarked, [], 0), ColLR_sel, ColUD_sel, I_sel_, 0, J_sel, K_sel, Selector, "", convert_open),
                I = I_sel_ + J_sel,
                set_selected(Win, I, J_sel, K_sel, 0, 0, update_full).

        close_opened_in_tree_base(Win, I_sel, K_sel, Selector):-
                opened_multiple_in_tree_base(Win, node(_, ColLR, _, _, ColUD, _, I, J, K, Selector, _, _, _, _, _)), !,
                adjust_selected(Win, I_sel, K_sel, I, J, K),
                convert_tree(Win, tree("", unmarked, [], 0), ColLR, ColUD, I, J, _, K, Selector, "", convert_close_tree_base).
        close_opened_in_tree_base(_, _, _, _).

        close_opened_in_closed_base(Win, _, _, Selector):-
                retract(opened_multiple_in_closed_base(Win, node(_, ColLR, _, _, ColUD, _, I, J, K, Selector, _, _, _, _, _), MarkSel)), !,
                convert_tree(Win, tree("", unmarked, [], 0), ColLR, ColUD, I, J, _, K, Selector, MarkSel, convert_close_closed_base).
        close_opened_in_closed_base(_, _, _, _).

        adjust_selected(Win, I_sel, K_sel, I, J, K):-
                convert_get_node_context(I, J, K, I_sel, 0, K_sel, Context),
                adjust_selected(Win, I_sel, K_sel, J, Context).

        adjust_selected(_, _, _, _, convert_not):- !.
        adjust_selected(Win, I_sel, K_sel, J, convert_single):- !,
                I_sel_ = I_sel - J,
                retractall(selected(Win, _, _, _)),
                assert(selected(Win, I_sel_, 0, K_sel)).
        adjust_selected(Win, I_sel, K_sel, J, convert_double):- !,
                I_sel_ = I_sel - 2 * J,
                retractall(selected(Win, _, _, _)),
                assert(selected(Win, I_sel_, 0, K_sel)).

        get_update_mode(update_scroll, b_true):- !.
        get_update_mode(update_full, b_false):- !.

        get_selector(Win, S):-
                tree_above(Win, node(_, _, _, _, _, _, _, _, _, S, _, _, _, _, _)),
                not(multiple_selector(Win, S)).
        get_selector(Win, S):-
                tree_below(Win, node(_, _, _, _, _, _, _, _, _, S, _, _, _, _, _)),
                not(multiple_selector(Win, S)).
        get_selector(Win, S):-
                closed_above(Win, node(_, _, _, _, _, _, _, _, _, S, _, _, _, _, _)),
                not(multiple_selector(Win, S)).
        get_selector(Win, S):-
                closed_below(Win, node(_, _, _, _, _, _, _, _, _, S, _, _, _, _, _)),
                not(multiple_selector(Win, S)).
        get_selector(Win, S):-
                multiple_selector(Win, S).

        get_font(tree_WinInfo(F, _, _), F):- !.
        get_font(tree_WinInfoAdv(F, _, _, _, _), F):- !.

        set_context(Win, _, tree_WinInfo(F, D, A)):- !,
                assert(win_info(Win, tree_WinInfo(F, D, A))),
                assert(screen_context(Win, 0, 0, D)),
                assert(selected(Win, 0, 0, 0)).
        set_context(Win, _, tree_WinInfoAdv(F, D, A, List, _)):-
                assert(win_info(Win, tree_WinInfo(F, D, A))),
                assert(screen_context(Win, 0, 0, D)),
                tree_size(Win, _, _, _, _, N),
                R = N div 2,
                search_by_route(Win, List, 0, R, R, 0, I, J, K),
                set_selected(Win, I, J, K, 0, 0, update_full),
                !.
        set_context(Win, _, tree_WinInfoAdv(_, _, _, List, _)):-
                reverse_list(List, [], List_),
                List_ = [ Selector | _ ],
                tree_above(Win, node(_, _, _, _, _, _, I, J, K, Selector, _, _, _, _, _)),
                set_selected(Win, I, J, K, 0, 0, update_full),
                !.
        set_context(Win, _, _):-
                retractall(selected(Win, _, _, _)),
                assert(selected(Win, 0, 0, 0)).

        get_parent(Win, I, J, K, Parent):-
                findall(NODE, parent(Win, I, J, K, NODE), List),
                parent(List, Parent).

        parent(Win, I, J, K, NODE):-
                tree_above(Win, NODE),
                parent_(I, J, K, NODE).
        parent(Win, I, J, K, NODE):-
                tree_below(Win, NODE),
                parent_(I, J, K, NODE).

        parent_(I, J, K, node(_, _, _, _, _, _, I, J, K_, _, _, _, _, _, _)):-
                K_ < K, !.
        parent_(I, J, 0, node(_, _, _, _, _, _, I_, J_, _, _, _, _, _, _, _)):-
                I_ - J_ <= I - J,
                I_ + J_ >= I + J,
               (I_ - I) *(I_ - I) +(J_ - J) *(J_ - J) <> 0.

        parent([P], P):- !.
        parent([P0|T], P1):-
                parent(T, P1),
                P0 = node(_, _, _, _, _, _, _, J0, K0, _, _, _, _, _, _),
                P1 = node(_, _, _, _, _, _, _, J1, K1, _, _, _, _, _, _),
                parent__(J0, J1, K0, K1), !.
        parent([P|_], P).

        parent__(J0, J1, _, _):-
                J1 < J0, !.
        parent__(J, J, K0, K1):-
                K1 > K0, !.

        get_hscroll_height(Win) = HScroll_height:-
                h_scroll(Win, _),!,
                HScroll_height = vpi::winGetAttrVal(Win, attr_ctl_horz_sbar_height).
        get_hscroll_height(_) = 0.

        get_vscroll_width(Win) = VScroll_width:-
                v_scroll(Win, _),!,
                VScroll_width = vpi::winGetAttrVal(Win, attr_ctl_vert_sbar_width).
        get_vscroll_width(_) = 0.

        save_closed(Win):-
                retract(opened_multiple_in_closed_base(Win, node(_, ColLR, _, _, ColUD, _, I, J, K, Selector, _, _, _, _, _), MarkSel)), !,
                convert_tree(Win, tree("", unmarked, [], 0), ColLR, ColUD, I, J, _, K, Selector, MarkSel, convert_close_closed_base),
                save_closed(Win).
        save_closed(_):-
                retractall(closed(_)), fail.
        save_closed(Win):-
                closed_above(Win, closed_mark(Selector, _)),
                assert(closed(Selector)), fail.
        save_closed(Win):-
                closed_below(Win, closed_mark(Selector, _)),
                assert(closed(Selector)), fail.
        save_closed(_).

        restore_closed(Win):-
                retract(closed(Selector)),
                restore_closed(Win, Selector), fail.
        restore_closed(_).

        restore_closed(Win, Selector):-
                tree_above(Win, node(_, ColLR_sel, _, _, ColUD_sel, _, I_sel, J_sel, K_sel, Selector, b_false, _, b_true, _, _)), !,
                convert_tree(Win, tree("", unmarked, [], 0), ColLR_sel, ColUD_sel, I_sel, J_sel, _, K_sel, Selector, "", convert_close_tree_base).
        restore_closed(Win, Selector):-
                tree_below(Win, node(_, ColLR_sel, _, _, ColUD_sel, _, I_sel, J_sel, K_sel, Selector, b_false, _, b_true, _, _)), !,
                convert_tree(Win, tree("", unmarked, [], 0), ColLR_sel, ColUD_sel, I_sel, J_sel, _, K_sel, Selector, "", convert_close_tree_base).

        save_opened(Win):-
                opened_multiple_in_tree_base(Win, node(_, _, _, _, _, _, I, J, K, Selector, _, _, _, _, _)),
                get_parent_list(Win, I, J, K, [ Selector ], List),
                reverse_list(List, [], List_),
                assert(opened(List_)), fail.
        save_opened(_).

        check_opened_or_selected(T, b_true):-
                opened(L),
                reverse_list(L, [], L_),
                check_opened_or_selected(T, T, L_),
                assert(opened_checked(L)), fail.
        check_opened_or_selected(T, b_false):-
                selected_route(L),
                reverse_list(L, [], L_),
                check_opened_or_selected(T, T, L_),
                assert(selected_route_checked(L)), fail.
        check_opened_or_selected(_, _).

        check_opened_or_selected(_, tree(Sel, _, _, _), [ Sel ]):- !.
        check_opened_or_selected(T, tree(Sel, _, _, _), [ Sel | L ]):-
                get_subtree(T, Sel, List),
                check_opened_or_selected_(T, List, L).

        check_opened_or_selected_(Tree, [ H | _ ], L):-
                check_opened_or_selected(Tree, H, L), !.
        check_opened_or_selected_(Tree, [ _ | T ], L):-
                check_opened_or_selected_(Tree, T, L).

        restore_opened_or_selected(Tree, tree(Sel, Mark, _, ArrowType), tree(Sel, Mark, List_, ArrowType), L, b_true):-
                not(tree_member(L, Sel)),
		opened_checked(OpPath),
		CurPath = [ Sel | L ],
		is_subpath(CurPath, OpPath), !,
		get_subtree(Tree, Sel, List),
		restore_opened_or_selected(Tree, List, [], List_, CurPath, b_true).
	restore_opened_or_selected(_, tree(Sel, Mark, _, ArrowType), tree(Sel, Mark, [], ArrowType), _, b_true):-
		opened_checked([ Sel | _ ]), !.
	restore_opened_or_selected(Tree, tree(Sel, Mark, _, ArrowType), tree(Sel, Mark, List_, ArrowType), L, b_false):-
		not(tree_member(L, Sel)),
		selected_route_checked(SelPath),
		CurPath = [ Sel | L ],
		is_subpath(CurPath, SelPath), !,
		get_subtree(Tree, Sel, List),
		restore_opened_or_selected(Tree, List, [], List_, CurPath, b_false).
	restore_opened_or_selected(_, tree(Sel, Mark, _, ArrowType), tree(Sel, Mark, [], ArrowType), _, b_false):-
		selected_route_checked([ Sel | _ ]), !.
	restore_opened_or_selected(Tree, tree(Sel, Mark, List, ArrowType), tree(Sel, Mark, List_, ArrowType), L, Mode):-
		restore_opened_or_selected(Tree, List, [], List_, [ Sel | L ], Mode).

	restore_opened_or_selected(_, [], L, L_, _, _):-
		reverse_list(L, [], L_), !.
	restore_opened_or_selected(Tree, [ H | T ], L, L_, SL, Mode):-
		restore_opened_or_selected(Tree, H, H_, SL, Mode),
		restore_opened_or_selected(Tree, T, [ H_ | L ], L_, SL, Mode).

        tree_member([ H | _ ], H):- !.
        tree_member([ _ | T ], H):-
	        tree_member(T, H).

        is_subpath(L, L):- !.
        is_subpath(L, [ _ | T ]):-
	        is_subpath(L, T).

	get_subtree(tree(Sel, _, List, _), Sel, List):-
		List = [ _ | _ ], !.
	get_subtree(tree(_, _, List, _), Sel, List_):-
		get_subtree_(List, Sel, List_), !.
	get_subtree(_, _, []).

	get_subtree_([], _, []):- !.
	get_subtree_([ H | _ ], Sel, L):-
		get_subtree(H, Sel, L),
		L = [ _ | _ ], !.
	get_subtree_([ _ | T ], Sel, L):-
		get_subtree_(T, Sel, L).

	reverse_list([], L, L):- !.
	reverse_list([ H | T ], L, L_):-
		reverse_list(T, [ H | L ], L_).

	save_selected(Win):-
		get_selected_node(Win, node(_, _, _, _, _, _, I, J, K, Selector, _, _, _, _, _)), !,
		get_parent_list(Win, I, J, K, [ Selector ], List),
		retractall(selected_route(_)),
		reverse_list(List, [], List_),
		assert(selected_route(List_)).

	get_parent_list(Win, I, J, K, L0, L1):-
		get_parent(Win, I, J, K, node(_, _, _, _, _, _, I_, J_, K_, Selector, _, _, _, _, _)), !,
		get_parent_list(Win, I_, J_, K_, [ Selector | L0 ], L1).
	get_parent_list(_, _, _, _, L, L).

	restore_selected(Win):-
		retractall(selected(Win, _, _, _)),
		retract(selected_route_checked(List)),
		reverse_list(List, [], List_),
		tree_size(Win, _, _, _, _, N),
		R = N div 2,
		search_by_route(Win, List_, 0, R, R, 0, I, J, K), !,
		assert(selected(Win, I, J, K)).
	restore_selected(Win):-
		assert(selected(Win, 0, 0, 0)).

	search_by_route(_, [], _, I, J, K, I, J, K):- !.
	search_by_route(Win, [ Selector | List ], Col0, I0, J0, _, I2, J2, K2):-
		Col1 = Col0 + tree_link_width,
		tree_above(Win, NODE),
		NODE = node(Col1, Col2, _, _, _, _, I1, J1, K1, Selector, _, _, _, _, _),
		I1 - J1 >= I0 - J0,
		I1 + J1 <= I0 + J0, !,
		Col3 = Col2 + tree_link_width,
		search_by_route(Win, List, Col3, I1, J1, K1, I2, J2, K2).
	search_by_route(Win, [ Selector | List ], Col0, I0, J0, _, I2, J2, K2):-
		Col1 = Col0 + tree_link_width,
		tree_below(Win, NODE),
		NODE = node(Col1, Col2, _, _, _, _, I1, J1, K1, Selector, _, _, _, _, _),
		I1 - J1 >= I0 - J0,
		I1 + J1 <= I0 + J0, !,
		Col3 = Col2 + tree_link_width,
		search_by_route(Win, List, Col3, I1, J1, K1, I2, J2, K2).

	order_list([], L, L):- !.
	order_list([ S | L0 ], L1, L2):-
		max_selector(S, Smax, L0, [], L0_),
		order_list(L0_, [ Smax | L1 ], L2).

	max_selector(S, S, [], L, L):- !.
	max_selector(S1, S, [ S0 | T0 ], L0, L1):-
		S0 <= S1, !,
		max_selector(S1, S, T0, [ S0 | L0 ], L1).
	max_selector(S1, S, [ S0 | T0 ], L0, L1):-
		max_selector(S0, S, T0, [ S1 | L0 ], L1).

	tree_GetHandler(W) = EH :-
		tree_callback(W, EH), !.
	tree_GetHandler(W) = EH :-
		EH = vpi::winGetHandler(W).

/******************************************************************************
		clauses FOR global predicates
******************************************************************************/
    clauses
        create(WinType, RCT, Title, Menu, Parent, Flags, EventHandler,
               CreationData, Tree, TreeWinInfo) = TreeWin:-
            Win = vpi::winCreate(w_Child, RCT, "", noMenu, Parent, [wsf_Invisible],
                                 tree_no_handler, 0),
            get_font(TreeWinInfo, Font),
            vpi::winSetFont(Win, Font),
            tree_base_clean_up(Win),
            convertTree(Win, Tree),
            set_context(Win, RCT, TreeWinInfo),
            assert(tree_callback(Win, EventHandler)),
            assert(convert_window(Win)),
            TreeWin = vpi::winCreate(WinType, RCT, Title, Menu, Parent, Flags, tree_handler, CreationData),
            search_base_clean_up(Win),
            scroll_base_clean_up(Win),
            closed_base_clean_up(Win),
            vpi::winDestroy(Win),
            get_selected(TreeWin, I, J, K),
            set_selected(TreeWin, I, J, K, 0, 0, update_full),
            !.
	create(_, _, _, _, _, _, _, _, _, _) = _:-
            common_exception::raise_Error(classInfo, "create").

    clauses
        createSubclass(W, Tree, TreeWinInfo):-
            EventHandler = tree_GetHandler(W),
            tree_base_clean_up(W),
            selection_base_clean_up(W),
            search_base_clean_up(W),
            scroll_base_clean_up(W),
            closed_base_clean_up(W),
            RCT = vpi::winGetClientRect(W),
            TWin = vpi::getTaskWindow(),
            Win = vpi::winCreate(w_TopLevel, RCT, "", noMenu, TWin, [wsf_Invisible], tree_no_handler, 0),
            get_font(TreeWinInfo, Font),
            vpi::winSetFont(Win, Font),
            tree_base_clean_up(Win),
            convertTree(Win, Tree),
            set_context(Win, RCT, TreeWinInfo),
            assert(tree_callback(Win, EventHandler)),
%            assert(convert_window(Win)),
            vpi::winSetHandler(W, tree_handler),
%            retract(convert_window(Win)),
            tree_base_copy(Win, W),
            tree_base_clean_up(Win),
            write_set_mode(W, write_),
            selection_base_copy(Win, W),
            selection_base_clean_up(Win),
%            retract(win_info(W, Inf)),
%            assert(win_info(W, Inf)),
            vpi::winSetFont(W, Font),
            resize_tree(W),
            search_base_clean_up(Win),
            scroll_base_clean_up(Win),
            closed_base_clean_up(Win),
            vpi::winDestroy(Win),
            get_selected(W, I, J, K),
            set_selected(W, I, J, K, 0, 0, update_full),
            !.
	createSubclass(_, _, _):-
            common_exception::raise_Error(classInfo, "createSubclass").

    clauses
	update(W, Tree):-
            vpi::cursorSetWait(),
            retractall(_, open_close_info),
            save_closed(W),
            save_opened(W),
            save_selected(W),
            retract(win_info(W, TREE_WIN_INFO)),
            retract(tree_callback(W, HANDLER)),
            tree_base_clean_up(W),
            selection_base_clean_up(W),
            scroll_base_clean_up(W),
            closed_base_clean_up(W),
            retractall(_, convert_base),
            convert_register_selector(W, Tree),
            check_opened_or_selected(Tree, b_true),
            restore_opened_or_selected(Tree, Tree, Tree_, [], b_true),
            check_opened_or_selected(Tree_, b_false),
            restore_opened_or_selected(Tree_, Tree_, Tree__, [], b_false),
            tree_base_clean_up(W),
            convertTree(W, Tree__),
            set_context(W, vpidomains::rct(0, 0, 0, 0), TREE_WIN_INFO),
            assert(tree_callback(W, HANDLER)),
            restore_closed(W),
            restore_selected(W),
            retractall(_, open_close_info),
            selected(W, I, J, K),
            set_selected(W, I, J, K, 0, 0, update_full),
            !.
	update(_, _):-
            common_exception::raise_Error(classInfo, "update").

/******************************************************************************
                CONVERT predicates
******************************************************************************/

    constants
        convert_close   : integer = 0.
        convert_not     : integer = 1.
        convert_single  : integer = 2.
        convert_choosen : integer = 3.
        convert_double  : integer = 4.
        convert_close_  : integer = 5.

    constants
	convert                   : integer = 0.
	convert_base_             : integer = 1.
	convert_to_closed_base    : integer = 2.
	convert_open              : integer = 3.
	convert_close_tree_base   : integer = 4.
	convert_close_closed_base : integer = 5.

    class predicates
	convertTree : (vpiDomains::windowHandle, tree) procedure(i,i).
    clauses
        convertTree(Win, Tree):-
            convert_tree(Win, Tree, 0, 0, 0, 0, _, 0, "", "", convert),
            !.
        convertTree(_, _):-
            common_exception::raise_Error(classInfo, "convertTree").

    class facts - convert_base
        convert_window : (vpiDomains::windowHandle).
        tree_convert : (tree_node).
        tree_height : (integer, integer, integer).
        registered_selector : (selector, core::booleanInt).
        max_off_lr : (integer).
        max_off_ud : (integer).

    class predicates
        convert_tree : (vpiDomains::windowHandle TreeWin,tree Tree,integer ColLR, integer ColUD,
                        integer I,integer J,integer NewJ,integer K,selector Sel, selector MarkSel,
                        integer Mode) determ(i,i,i,i,i,i,o,i,i,i,i).

    clauses
        convert_tree(Win, Tree, ColLR_sel, ColUD_sel, I_sel, J_sel, J_sel_, K_sel,
                     Selector, MarkSel, Mode):-
            retractall(_, convert_base),
            convert_begin(Win, Tree, J_sel, Selector, MarkSel, Mode),
            convert_get_J(Win, J_sel_, Mode),
            convert_main(Win, ColLR_sel, ColUD_sel, I_sel, J_sel, J_sel_, K_sel, Selector, Mode),
            convert_get_range(Win, J_sel, N, N_2, Mode),
            convert_for(I, N, N_2),
              convertRow(Win, I, RowLR, RowMarkLR, RowUD, RowMarkUD),
              asserta(tree_above(Win, row_mark(RowMarkLR, RowMarkUD, I))),
              retract(tree_convert(node(ColLR, ColLR_, _, ColUD, ColUD_, _, I, J, K,
                                   Drawing, Closed, In, Out, Link, ArrowType))),
              NODE = node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K,
                          Drawing, Closed, In, Out, Link, ArrowType),
              asserta(tree_above(Win, NODE)),
              convert_multiple(Win, NODE),
              move_global_link(Win, NODE, begin, end_),
              convert_adjust_tree_size(NODE, Mode),
            fail.
        convert_tree(Win, _, _, _, _, _, 0, _, _, _, Mode):-
            Mode <= convert_base_,
            !,
            retract(tree_size(Win, WidLR, HeiLR, _, HeiUD, N)),
            !,
            tree_height(_, _, WidUD),
            !,
            assert(tree_size(Win, WidLR, HeiLR, WidUD, HeiUD, N)),
            retractall(_, convert_base).
	convert_tree(Win, _, _, _, _, J_sel, J_sel_, _, _, _, Mode):-
            convert_get_size(Win, J_sel, J_sel_, N, Mode),
            vpi::winGetFontMetrics(Win, _, Ascent, Descent),
            Height = Ascent + Descent + 2 * tree_frame_ver + 2 * tree_skip,
            HeiLR =(N div 2 + 1) * Height,
            max_off_lr(WidLR),
            max_off_ud(HeiUD),
            tree_height(_, _, WidUD),
            !,
            retractall(tree_size(Win, _, _, _, _, _)),
            assert(tree_size(Win, WidLR, HeiLR, WidUD, HeiUD, N)),
            retractall(_, convert_base).
        convert_tree(_, _, _, _, _, _, 0, _, _, _, _):-
            retractall(_, convert_base).

    class predicates
        convert_begin : (vpiDomains::windowHandle TreeWin,tree Tree,integer J,
                        selector Sel,selector MarkSel,integer Mode) determ(i,i,i,i,i,i).
    clauses
        convert_begin(Win, _, _, _, _, Mode):-
            Mode <> convert_close_closed_base,
            write_base_clean_up(Win),
            write_set_mode(Win, write_),
            fail.
        convert_begin(Win, Tree, _, _, _, convert):-
            !,
            convert_register_selector(Win, Tree),
            convert_tree_base(Win, Tree, 0, Y, 0, 0, MaxOffLR, 0, MaxOffUD, b_false, line, convert),
            N = Y * 2 - 1,
            vpi::winGetFontMetrics(Win, _, Ascent, Descent),
            Height = Ascent + Descent + 2 * tree_frame_ver + 2 * tree_skip,
            TreeHeightLR = Y * Height,
            retractall(tree_size(Win, _, _, _, _, _)),
            assert(tree_size(Win, MaxOffLR, TreeHeightLR, 0, MaxOffUD, N)).
        convert_begin(Win, _, _, _, _, convert_base_):-
            retractall(opened_multiple_in_tree_base(Win, _)),
            retract(tree_above(Win, NODE)),
              asserta(tree_below(Win, NODE)),
            fail.
        convert_begin(Win, _, _, _, _, convert_base_):-
            tree_size(Win, _, _, _, _, N),
            I = N div 2,
            convert_get_child(Win, NODE, I, I, 0),
            convert_node_base(Win, NODE, 0, MaxOffLR, 0, MaxOffUD, convert_base_),
            vpi::winGetFontMetrics(Win, _, Ascent, Descent),
            Height = Ascent + Descent + 2 * tree_frame_ver + 2 * tree_skip,
            TreeHeightLR =(N + 1) div 2 * Height,
            retractall(tree_size(Win, _, _, _, _, _)),
            assert(tree_size(Win, MaxOffLR, TreeHeightLR, 0, MaxOffUD, N)),
            retractall(tree_above(Win, _)),
            retractall(tree_below(Win, _)),
            fail.
        convert_begin(Win, _, _, _, _, convert_base_):-
            retractall(opened_multiple_in_closed_base(Win, _, _)),
            retract(closed_below(Win, NODE)),
              asserta(closed_above(Win, NODE)),
            fail.
        convert_begin(Win, _, _, _, _, convert_base_):-
            convert_set_closed_base_nd(Win,J),
            findall(NODE, convert_get_child(Win, NODE, J, J, tree_link_width), List),
            convert_nodelist_base(Win, List, tree_link_width, _, tree_link_width, _, convert_to_closed_base),
            retractall(tree_below(Win, _)),
            fail.
        convert_begin(_, _, _, _, _, convert_base_):-
            !.
        convert_begin(Win, _, _, _, _, Mode):-
            Mode <> convert_close_closed_base,
            retractall(opened_multiple_in_tree_base(Win, _)),
            assert(max_off_lr(0)),
            assert(max_off_ud(0)),
            retract(tree_above(Win, NODE)),
              asserta(tree_below(Win, NODE)),
            fail.
        convert_begin(Win, _, _, _, _, Mode):-
            Mode <> convert_close_tree_base,
            retract(closed_above(Win, NODE)),
              asserta(closed_below(Win, NODE)),
            fail.
        convert_begin(Win, _, J_sel, Selector, _, Mode):-
            Mode <> convert_open,
            assertz(closed_below(Win, closed_mark(Selector, J_sel))),
            fail.
        convert_begin(Win, _, _, Selector, _, convert_open):-
            !,
            convert_set_closed_base(Win, Selector, _).
        convert_begin(_, _, _, _, _, convert_close_tree_base):-
            !.
        convert_begin(Win, _, J_sel, _, MarkSel, convert_close_closed_base):-
            !,
            convert_set_closed_base(Win, MarkSel, N),
            retract(closed_above(Win, closed_mark(_, _))),
            !,
            N_ = N - J_sel,
            asserta(closed_above(Win, closed_mark(MarkSel, N_))).

    class predicates
        convert_register_selector : (vpiDomains::windowHandle TreeWin,tree Tree) determ(i,i).
        convert_register_selector1 : (vpiDomains::windowHandle TreeWin,tree_list SubTree) determ(i,i).
    clauses

        convert_register_selector(Win, tree(Selector, _, [], _)):-
            registered_selector(Selector, _),
            !,
            convert_register_multiple(Win, Selector).
        convert_register_selector(_, tree(Selector, _, TreeList, _)):-
            registered_selector(Selector, _),
            !,
            convert_out_treelist_base(TreeList, _).
        convert_register_selector(Win, tree(Selector, _, TreeList, _)):-
            retract(registered_selector(Selector , _)),
            !,
            assert(registered_selector(Selector, b_true)),
            convert_register_multiple(Win, Selector),
            convert_register_selector1(Win, TreeList).
        convert_register_selector(Win, tree(Selector, _, TreeList, _)):-
            convert_out_treelist_base(TreeList, Out),
            assert(registered_selector(Selector, Out)),
            convert_register_selector1(Win, TreeList).

        convert_register_selector1(_, []):-
            !.
        convert_register_selector1(Win, [H|T]):-
            convert_register_selector(Win, H),
            convert_register_selector1(Win, T).

    class predicates
        convert_register_multiple : (vpiDomains::windowHandle TreeWin,selector Sel) procedure(i,i).
    clauses
        convert_register_multiple(Win, Selector):-
            multiple_selector(Win, Selector),
            !.
        convert_register_multiple(Win, Selector):-
            assert(multiple_selector(Win, Selector)).

    class predicates
        convert_tree_base : (vpiDomains::windowHandle TreeWin,tree Tree,integer X,integer Y,
                             integer K,integer OffLR,integer MaxOffLR,integer OffUD,integer MaxOffUD,
                             core::booleanInt In,tree_link Link,integer Mode) determ anyflow.
    clauses
        convert_tree_base(Win, tree(Selector, _, List, ArrowType), X, Y, K, OffLR, MaxOffLR, OffUD,
                          MaxOffUD, In, Link, Mode):-
            % Pro/3
            sx::sxUnpack(Selector,'{',Lx),
            Lx=[SelX|_],
            %
            vpi::winGetTextExtent(Win, SelX, -1, Width, _),
            vpi::winGetFontMetrics(Win, _, Ascent, Descent),
            Height = Ascent + Descent,
            OffLR1 = OffLR + tree_link_width,
            OffUD1 = OffUD + tree_link_width,
            OffLR2 = OffLR1 + Width + 2 * tree_frame_hor,
            OffUD2 = OffUD1 + Height + 2 * tree_frame_ver,
            OffLR3 = OffLR2 + tree_link_width,
            OffUD3 = OffUD2 + tree_link_width,
            convert_treelist_base(Win, List, X, Y, K, OffLR3, 0, MaxOffLR, OffUD3, 0, MaxOffUD, b_true, b_true, Mode),
            convert_out_treelist_base(List, Out),
            I = X + Y - 1,
            J = Y - X - 1,
            convert_RegisterSelector(Win, Selector, Out, Closed),
            NODE = node(OffLR1, OffLR2, 0, OffUD1, OffUD2, 0, I, J, K, Selector, Closed, In, Out, Link, ArrowType),
            convertNode(Win, NODE, Mode).

    class predicates
        convert_treelist_base : (vpiDomains::windowHandle TreeWin,tree_list SubTree,integer X,integer Y,
                                 integer K, integer OffLR,integer OffLR1,integer MaxOffLR,integer OffUD,
                                 integer OffUD1,integer MaxOffUD,core::booleanInt In,core::booleanInt,
                                 integer Mode) determ anyflow.
    clauses
        convert_treelist_base(_, [], X0, X1, _, OffLR, 0, OffLR, OffUD, 0, OffUD, _, b_true, _):-
            !,
            X1 = X0 + 1.
        convert_treelist_base(Win, [H], X0, X1, K, OffLR, 0, MaxOffLR, OffUD, 0, MaxOffUD, In, b_true, Mode):-
            !,
            K_ = K + 1,
            convert_tree_base(Win, H, X0, X1, K_, OffLR, MaxOffLR, OffUD, MaxOffUD, In, line, Mode).
        convert_treelist_base(Win, [H|T], X0, X2, K, OffLR, 0, MaxOffLR, OffUD, 0, MaxOffUD, In, b_true, Mode):-
            !,
            convert_tree_base(Win, H, X0, X1, 0, OffLR, OffLR1, OffUD, OffUD1, In, begin, Mode),
            convert_treelist_base(Win, T, X1, X2, K, OffLR, OffLR1, MaxOffLR, OffUD, OffUD1, MaxOffUD, In, b_false, Mode).
        convert_treelist_base(Win, [H], X0, X1, _, OffLR, MaxOffLR0, MaxOffLR1, OffUD, MaxOffUD0, MaxOffUD1, In, b_false, Mode):-
            !,
            convert_tree_base(Win, H, X0, X1, 0, OffLR, OffLR1, OffUD, OffUD1, In, end_, Mode),
            convert_max(MaxOffLR0, OffLR1, MaxOffLR1),
            convert_max(MaxOffUD0, OffUD1, MaxOffUD1).
        convert_treelist_base(Win, [H|T], X0, X2, K, OffLR, MaxOffLR0, MaxOffLR2, OffUD, MaxOffUD0, MaxOffUD2, In, b_false, Mode):-
            !,
            convert_tree_base(Win, H, X0, X1, 0, OffLR, OffLR1, OffUD, OffUD1, In, cross, Mode),
            convert_max(MaxOffLR0, OffLR1, MaxOffLR1),
            convert_max(MaxOffUD0, OffUD1, MaxOffUD1),
            convert_treelist_base(Win, T, X1, X2, K, OffLR, MaxOffLR1, MaxOffLR2, OffUD, MaxOffUD1, MaxOffUD2, In, b_false, Mode).

    class predicates
        convert_max : (integer X,integer Y,integer Max) procedure(i,i,o).
    clauses
        convert_max(X, Y, X):-
            X >= Y,
            !.
        convert_max(_, Y, Y).

    class predicates
        convert_out_treelist_base : (tree_list, core::booleanInt) procedure(i,o).
    clauses
        convert_out_treelist_base([], b_false):-
            !.
        convert_out_treelist_base(_, b_true).

    class predicates
        convert_RegisterSelector : (vpiDomains::windowHandle, selector, core::booleanInt, core::booleanInt) determ anyflow.
    clauses
        convert_RegisterSelector(Win, Selector, _, b_false):-
            not(multiple_selector(Win, Selector)),
            !.
        convert_RegisterSelector(_, Selector, _, b_false):-
            registered_selector(Selector, b_false),
            !.
        convert_RegisterSelector(_, _, b_true, b_false):-
            !.
        convert_RegisterSelector(_, _, b_false, b_true):-
            !.

    class predicates
        convertNode : (vpiDomains::windowHandle TreeWin,tree_node Node,integer Mode) procedure(i,i,i).
    clauses
        convertNode(Win, NODE, convert_to_closed_base):-
            !,
            assertz(closed_below(Win, NODE)).
        convertNode(Win, NODE, convert_close_closed_base):-
            !,
            asserta(closed_above(Win, NODE)).
        convertNode(_, NODE, _):-
            assert(tree_convert(NODE)).

    class predicates
        convert_get_child : (vpiDomains::windowHandle TreeWin,tree_node Node,integer I,integer J,
                             integer Col) nondeterm(i,o,i,i,i).
    clauses
        convert_get_child(Win, NODE, I, J, Col):-
                Col_ = Col + tree_link_width,
                tree_below(Win, NODE),
                  NODE = node(Col_, _, _, _, _, _, I_, J_, _, _, _, _, _, _, _),
                  I_ - J_ >= I - J,
                  I_ + J_ <= I + J.

    class predicates
        convert_node_base : (vpiDomains::windowHandle TreeWin,tree_node Node,integer OffLR,
                             integer MaxOffLR,integer OffUD,integer MaxOffUD,integer Mode) determ(i,i,i,o,i,o,i).
    clauses
        convert_node_base(Win, node(_, Col, _, _, _, _, I, J, K, Selector, Closed, In, Out, Link, ArrowType),
                          OffLR, MaxOffLR, OffUD, MaxOffUD, Mode):-
             % Pro/3
             sx::sxUnpack(Selector,'{',Lx),
             Lx=[SelX|_],
             %
            vpi::winGetTextExtent(Win, SelX, -1, Width, _),
            vpi::winGetFontMetrics(Win, _, Ascent, Descent),
            Height = Ascent + Descent,
            OffLR1 = OffLR + tree_link_width,
            OffUD1 = OffUD + tree_link_width,
            OffLR2 = OffLR1 + Width + 2 * tree_frame_hor,
            OffUD2 = OffUD1 + Height + 2 * tree_frame_ver,
            OffLR3 = OffLR2 + tree_link_width,
            OffUD3 = OffUD2 + tree_link_width,
            Col_ = Col + tree_link_width,
            findall(NODE_, convert_get_child(Win, NODE_, I, J, Col_), List),
            convert_nodelist_base(Win, List, OffLR3, MaxOffLR, OffUD3, MaxOffUD, Mode),
            NODE = node(OffLR1, OffLR2, 0, OffUD1, OffUD2, 0, I, J, K, Selector, Closed, In, Out, Link, ArrowType),
            convert_multiple_for_mode(Win, NODE, Mode),
            convertNode(Win, NODE, Mode).

    class predicates
        convert_multiple_for_mode : (vpiDomains::windowHandle TreeWin,tree_node Node,integer Mode) determ(i,i,i).
    clauses
        convert_multiple_for_mode(_, _, convert_base_):-
            !.
        convert_multiple_for_mode(Win, NODE, convert_to_closed_base):-
            NODE = node(_, _, _, _, _, _, _, _, _, Selector, b_false, _, b_true, _, _),
            multiple_selector(Win, Selector),
            tree_below(Win, closed_mark(MarkSel, _)), !,
            assert(opened_multiple_in_closed_base(Win, NODE, MarkSel)).
        convert_multiple_for_mode(_, _, convert_to_closed_base):-
            !.
        convert_multiple_for_mode(Win, NODE, Mode):-
            Mode <> convert_close_tree_base,
            !,
            closed_above(Win, closed_mark(MarkSel, _)),
            retract(opened_multiple_in_closed_base(Win, NODE, MarkSel)),
            !.
        convert_multiple_for_mode(_, _, _).

    class predicates
        convert_nodelist_base : (vpiDomains::windowHandle TreeWin,tree_node_list NodeList,
                                 integer OffLR,integer MaxOffLR, integer OffUD, integer MaxOffUD,
                                 integer Mode) determ(i,i,i,o,i,o,i).
    clauses
        convert_nodelist_base(_, [], OffLR, OffLR, OffUD, OffUD, _):-
            !.
        convert_nodelist_base(Win, [H|T], OffLR, MaxOffLR2, OffUD, MaxOffUD2, Mode):-
            !,
            convert_node_base(Win, H, OffLR, OffLR1, OffUD, OffUD1, Mode),
            convert_nodelist_base(Win, T, OffLR, MaxOffLR1, OffUD, MaxOffUD1, Mode),
            convert_max(MaxOffLR1, OffLR1, MaxOffLR2),
            convert_max(MaxOffUD1, OffUD1, MaxOffUD2).

    class predicates
        convert_set_closed_base_nd : (vpiDomains::windowHandle TreeWin, integer J) nondeterm(i,o).
    clauses
        convert_set_closed_base_nd(Win, J):-
                retract(closed_above(Win, NODE)),
                asserta(tree_below(Win, NODE)),
                NODE = closed_mark(_, J),
                assertz(closed_below(Win, NODE)).

    class predicates
        convert_set_closed_base : (vpiDomains::windowHandle TreeWin,selector Sel,integer J) determ(i,i,o).
    clauses
        convert_set_closed_base(Win, Selector, J):-
                retract(closed_below(Win, NODE)),
                asserta(closed_above(Win, NODE)),
                NODE = closed_mark(Selector, J),
                !.

    class predicates
        convert_get_J : (vpiDomains::windowHandle TreeWin,integer J,integer Mode) determ(i,o,i).
    clauses
        convert_get_J(Win, J, convert_open):-
            !,
            closed_above(Win, closed_mark(_, J)),
            !.
        convert_get_J(_, 0, _).

    class predicates
        convert_main : (vpiDomains::windowHandle TreeWin,integer ColLR,integer ColUD,
                        integer I,integer J,integer NewJ,integer K,selector Sel,
                        integer Mode) determ(i,i,i,i,i,i,i,i,i).
    clauses
        convert_main(Win, ColLR_sel, ColUD_sel, I_sel, J_sel, _, K_sel, Selector, convert_close_closed_base):-
            retract(closed_below(Win, NODE)),
            !,
            convert_main_closed(Win, ColLR_sel, ColUD_sel, I_sel, J_sel, K_sel, Selector, NODE).
        convert_main(Win, ColLR_sel, ColUD_sel, I_sel, J_sel, J_sel_, K_sel, Selector, Mode):-
            Mode <> convert,
            Mode <> convert_base_,
            retract(tree_below(Win, NODE)),
              convert_closed(Win, NODE, ColLR_sel, ColUD_sel, I_sel, K_sel, Mode),
              convert_get_coord(Win, NODE, J_sel, I, J, K, Mode),
              convert_get_node_context(I_sel, J_sel, K_sel, I, J, K, Context),
              convert_node(Win, NODE, ColLR_sel, ColUD_sel, I_sel, J_sel, J_sel_, K_sel, Selector, Context, Mode),
            fail.
        convert_main(_, _, _, _, _, _, _, _, _).

    class predicates
        convert_main_closed : (vpiDomains::windowHandle TreeWin,integer ColLR,integer ColUD,
                               integer I,integer J,integer K,selector Sel,tree_node Node) determ(i,i,i,i,i,i,i,i).
    clauses
        convert_main_closed(Win, _, _, _, _, _, _, NODE):-
            NODE = closed_mark(MarkSel, J),
            !,
            asserta(closed_above(Win, closed_mark(MarkSel, J))).
        convert_main_closed(Win, ColLR_sel, ColUD_sel, I_sel, J_sel, K_sel, Selector, NODE):-
            NODE = node(_, _, _, _, _, _, I, J, K, _, _, _, _, _, _),
            convert_get_node_context(I_sel, J_sel, K_sel, I, J, K, Context),
            convert_node(Win, NODE, ColLR_sel, ColUD_sel, I_sel, J_sel, 0, K_sel, Selector, Context, convert_close_closed_base),
            convert_main(Win, ColLR_sel, ColUD_sel, I_sel, J_sel, 0, K_sel, Selector, convert_close_closed_base).

    class predicates
        convert_closed : (vpiDomains::windowHandle TreeWin,tree_node Node,integer ColLR,
                          integer ColUD,integer I, integer K, integer Mode) procedure(i,i,i,i,i,i,i).
    clauses
        convert_closed(Win, row_mark(_, _, I_sel), ColLR_sel, ColUD_sel, I_sel, K_sel, convert_open):-
            retract(closed_below(Win, NODE)),
            convert_node_closed(Win, NODE, ColLR_sel, ColUD_sel, I_sel, K_sel),
            !.
        convert_closed(_, _, _, _, _, _, _).

    class predicates
        convert_node_closed : (vpiDomains::windowHandle TreeWin,tree_node Node,integer ColLR,
                               integer ColUD,integer I,integer K) determ(i,i,i,i,i,i).
    clauses
        convert_node_closed(Win, closed_mark(Sel, J), _, _, _, _):-
            !,
            asserta(closed_below(Win, closed_mark(Sel, J))).
        convert_node_closed(Win, NODE, ColLR_sel, ColUD_sel, I_sel, K_sel):-
            NODE = node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, 0, J_sel_, K, Sel, Closed, In, Out, Link, ArrowType),
            closed_above(Win, closed_mark(_, J_sel_)),
            !,
            K_ = K + K_sel,
            ColLR1 = ColLR + ColLR_sel,
            ColUD1 = ColUD + ColUD_sel,
            ColLR_1 = ColLR_ + ColLR_sel,
            ColUD_1 = ColUD_ + ColUD_sel,
            NODE_ = node(ColLR1, ColLR_1, RowLR, ColUD1, ColUD_1, RowUD, I_sel, J_sel_, K_, Sel, Closed, In, Out, Link, ArrowType),
            assert(tree_convert(NODE_)),
            convert_multiple_for_mode(Win, NODE, convert_open),
            fail.
        convert_node_closed(Win, NODE, ColLR_sel, ColUD_sel, I_sel, _):- !,
            NODE = node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Sel, Closed, In, Out, Link, ArrowType),
            I_ = I + I_sel,
            ColLR1 = ColLR + ColLR_sel,
            ColUD1 = ColUD + ColUD_sel,
            ColLR_1 = ColLR_ + ColLR_sel,
            ColUD_1 = ColUD_ + ColUD_sel,
            NODE_ = node(ColLR1, ColLR_1, RowLR, ColUD1, ColUD_1, RowUD, I_, J, K, Sel, Closed, In, Out, Link, ArrowType),
            assert(tree_convert(NODE_)),
            convert_multiple_for_mode(Win, NODE, convert_open),
            fail.

    class predicates
        convert_get_coord : (vpiDomains::windowHandle TreeWin,tree_node Node,integer J_Sel,
                             integer I,integer J,integer K,integer Mode) determ(i,i,i,o,o,o,i).
    clauses
        convert_get_coord(Win, row_mark(_, _, I), J_sel, 0, 0, 0, Mode):-
            !,
            I_ = I - 2 * J_sel - 1,
            convertRow(Win, I_, RowLR, RowMarkLR, RowUD, RowMarkUD),
            asserta(tree_above(Win, row_mark(RowMarkLR, RowMarkUD, I_))),
            retract(tree_convert(node(ColLR, ColLR_, _, ColUD, ColUD_, _, I_, J, K, Drawing, Closed, In, Out, Link, ArrowType))),
            NODE = node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I_, J, K, Drawing, Closed, In, Out, Link, ArrowType),
            asserta(tree_above(Win, NODE)),
            convert_multiple(Win, NODE),
            move_global_link(Win, NODE, begin, end_),
            convert_adjust_tree_size(NODE, Mode),
            fail.
        convert_get_coord(_, NODE, _, I, J, K, _):-
            NODE = node(_, _, _, _, _, _, I, J, K, _, _, _, _, _, _).

    class predicates
        convert_get_node_context : (integer I_Sel,integer J_Sel,integer K_Sel,
                                    integer I, integer J, integer K, integer Context) procedure(i,i,i,i,i,i,o).
        convert_get_node_context1 : (integer, integer, integer, integer) procedure(i,i,i,o).
    clauses
        convert_get_node_context(I_sel, J_sel, K_sel, I_sel, J_sel, K_sel, convert_choosen):-
            !.
        convert_get_node_context(I_sel, J_sel, K_sel, I_sel, J_sel, K, convert_single):-
            K < K_sel,
            !.
        convert_get_node_context(I_sel, J_sel, _, I_sel, J_sel, _, convert_close_):-
            !.
        convert_get_node_context(I_sel, J_sel, _, I, J, _, Context):-
            X_sel = I_sel - J_sel,
            Y_sel = I_sel + J_sel,
            X = I - J,
            Y = I + J + 1,
            convert_get_node_context1(X, X_sel, Y_sel, X_),
            convert_get_node_context1(Y, X_sel, Y_sel, Y_),
            Context = X_ * Y_.

        convert_get_node_context1(A, B, _, 1):-
            A <= B,
            !.
        convert_get_node_context1(A, _, B, 0):-
            A <= B,
            !.
        convert_get_node_context1(_, _, _, 2).

    class predicates
        convert_node : (vpiDomains::windowHandle TreeWin,tree_node Node,integer ColLR,integer ColUD,
                        integer I,integer J,integer NewJ,integer K,selector Sel,integer Context,
                        integer Mode) determ(i,i,i,i,i,i,i,i,i,i,i).
    clauses
        convert_node(Win, NODE, ColLR_sel, ColUD_sel, I_sel, J_sel, _, _, Selector, convert_close, Mode):-
            NODE = node(ColLR, ColLR_, _, ColUD, ColUD_, _, I, J, K, Sel, b_false, In, b_true, Link, ArrowType),
            multiple_selector(Win, Sel), !,
            ColLR1 = ColLR - ColLR_sel,
            ColUD1 = ColUD - ColUD_sel,
            ColLR_1 = ColLR_ - ColLR_sel,
            ColUD_1 = ColUD_ - ColUD_sel,
            I_ = I - I_sel + J_sel,
            NODE_ = node(ColLR1, ColLR_1, 0, ColUD1, ColUD_1, 0, I_, J, K, Sel, b_false, In, b_true, Link, ArrowType),
            assertz(closed_below(Win, NODE_)),
            convert_multiple_for_mode(Win, NODE, Mode),
            assert(opened_multiple_in_closed_base(Win, NODE_, Selector)).
        convert_node(Win, node(ColLR, ColLR_, _, ColUD, ColUD_, _, I, J, K, Sel, Closed, In, Out, Link, ArrowType),
                     ColLR_sel, ColUD_sel, I_sel, J_sel, _, _, _, convert_close, _):-
            !,
            ColLR1 = ColLR - ColLR_sel,
            ColUD1 = ColUD - ColUD_sel,
            ColLR_1 = ColLR_ - ColLR_sel,
            ColUD_1 = ColUD_ - ColUD_sel,
            I_ = I - I_sel + J_sel,
            assertz(closed_below(Win, node(ColLR1, ColLR_1, 0, ColUD1, ColUD_1, 0, I_, J, K, Sel, Closed, In, Out, Link, ArrowType))).
        convert_node(Win, NODE, _, _, _, _, _, _, _, convert_not, Mode):-
            !,
            convertNode(Win, NODE, Mode).
        convert_node(Win, node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Sel, Closed, In, Out, Link, ArrowType),
                     _, _, _, J_sel, J_sel_, _, _, convert_single, Mode):-
            !,
            convert_get_shift(J_sel, J_sel_, Shift, _, _, Mode),
            I_ = I + Shift,
            J_ = J + Shift,
            Node = node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I_, J_, K, Sel, Closed, In, Out, Link, ArrowType),
            convertNode(Win, Node, Mode).
        convert_node(Win, node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Sel, _, In, _, Link, ArrowType),
                     _, _, _, J_sel, J_sel_, _, _, convert_choosen, Mode):-
            !,
            convert_get_shift(J_sel, J_sel_, Shift, Closed, Out, Mode),
            I_ = I + Shift,
            J_ = J + Shift,
            Node = node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I_, J_, K, Sel, Closed, In, Out, Link, ArrowType),
            convertNode(Win, Node, Mode).
        convert_node(Win, node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I, J, K, Sel, Closed, In, Out, Link, ArrowType),
                     _, _, _, J_sel, J_sel_, _, _, convert_double, Mode):-
            !,
            convert_get_shift(J_sel, J_sel_, Shift, _, _, Mode),
            I_ = I + 2 * Shift,
            Node = node(ColLR, ColLR_, RowLR, ColUD, ColUD_, RowUD, I_, J, K, Sel, Closed, In, Out, Link, ArrowType),
            convertNode(Win, Node, Mode).
        convert_node(Win, NODE, ColLR_sel, ColUD_sel, I_sel, J_sel, _, K_sel, Selector, convert_close_, Mode):-
            NODE = node(ColLR, ColLR_, _, ColUD, ColUD_, _, I, J, K, Sel, b_false, In, b_true, Link, ArrowType),
            multiple_selector(Win, Sel),
            !,
            ColLR1 = ColLR - ColLR_sel,
            ColUD1 = ColUD - ColUD_sel,
            ColLR_1 = ColLR_ - ColLR_sel,
            ColUD_1 = ColUD_ - ColUD_sel,
            I_ = I - I_sel + J_sel,
            K_ = K - K_sel,
            NODE_ = node(ColLR1, ColLR_1, 0, ColUD1, ColUD_1, 0, I_, J, K_, Sel, b_false, In, b_true, Link, ArrowType),
            assertz(closed_below(Win, NODE_)),
            convert_multiple_for_mode(Win, NODE, Mode),
            assert(opened_multiple_in_closed_base(Win, NODE_, Selector)).
        convert_node(Win, node(ColLR, ColLR_, _, ColUD, ColUD_, _, I, J, K, Sel, Closed, In, Out, Link, ArrowType),
                     ColLR_sel, ColUD_sel, I_sel, J_sel, _, K_sel, _, convert_close_, _):-
            !,
            ColLR1 = ColLR - ColLR_sel,
            ColUD1 = ColUD - ColUD_sel,
            ColLR_1 = ColLR_ - ColLR_sel,
            ColUD_1 = ColUD_ - ColUD_sel,
            I_ = I - I_sel + J_sel,
            K_ = K - K_sel,
            assertz(closed_below(Win, node(ColLR1, ColLR_1, 0, ColUD1, ColUD_1, 0, I_, J, K_, Sel, Closed, In, Out, Link, ArrowType))).

    class predicates
        convert_get_shift : (integer, integer, integer, core::booleanInt, core::booleanInt, integer) procedure(i,i,o,o,o,i).

    clauses
        convert_get_shift(_, J, J, b_false, b_true, convert_open):-
            !.
        convert_get_shift(J, _, J_, b_true, b_false, _):-
            J_ = -J.

    class predicates
        convert_get_range : (vpiDomains::windowHandle TreeWin,integer J,integer N,integer N_2,
                            integer Mode) procedure(i,i,o,o,i).
    clauses
        convert_get_range(Win, _, 0, N, Mode):-
            Mode <= convert_base_,
            tree_size(Win, _, _, _, _, N),
            !.
        convert_get_range(Win, _, N_1, N_2, convert_open):-
            closed_above(Win, closed_mark(_, J_sel)),
            tree_size(Win, _, _, _, _, N),
            !,
            N_1 = N - 1,
            N_2 = N + 2 * J_sel.
        convert_get_range(Win, J_sel, N_1, N_2, convert_close_tree_base):-
            tree_size(Win, _, _, _, _, N),
            !,
            N_1 = N - 2 * J_sel - 1,
            N_2 = N_1 + 1.
        convert_get_range(_, _, 0, 0, _).

    class predicates
        convert_for : (integer, integer, integer) multi(o,i,i).
    clauses
        convert_for(I, I, _).
        convert_for(I, I0, N):-
            I1 = I0 + 1,
            I1 < N,
            convert_for(I, I1, N).

    class predicates
        convertRow : (vpiDomains::windowHandle TreeWin,integer I,integer RowLR,
                     integer Height,integer RowUD,integer RowMarkUD) determ(i,i,o,o,o,o).
        convert_row_nd : (integer, integer) nondeterm(i,o).
        convert_row : (core::integer_list , integer, integer) determ(i,i,o).
    clauses
        convertRow(_, I, 0, 0, 0, 0):-
            I < 0,
            !,
            fail.
        convertRow(Win, 0, RowLR, Height, RowUD, RowMarkUD):-
            !,
            vpi::winGetFontMetrics(Win, _, Ascent, Descent),
            Height = Ascent + Descent + 2 * tree_frame_ver + 2 * tree_skip,
            RowLR = Height div 2,
            findall(HeightUD , convert_row_nd(0, HeightUD), ListUD),
            convert_row(ListUD, 0, MaxHeightUD),
            RowUD = MaxHeightUD div 2 + tree_frame_hor + tree_skip,
            retractall(tree_height(_, _, _)),
            TreeHeightUD = RowUD + MaxHeightUD div 2 + tree_frame_hor + tree_skip,
            assert(tree_height(MaxHeightUD, RowUD, TreeHeightUD)),
            RowMarkUD = RowUD + MaxHeightUD div 2.
        convertRow(Win, I, RowLR, RowMarkLR, RowUD1, RowMarkUD):-
            vpi::winGetFontMetrics(Win, _, Ascent, Descent),
            Height = Ascent + Descent + 2 * tree_frame_ver + 2 * tree_skip,
            RowLR =(I + 1) * Height div 2,
            RowMarkLR =(I + 2) * Height div 2,
            findall(HeightUD , convert_row_nd(I, HeightUD), ListUD),
            convert_row(ListUD, 0, MaxHeightUD1),
            retract(tree_height(MaxHeightUD0, RowUD0, _)), !,
            convert_max(MaxHeightUD0, MaxHeightUD1, MaxHeightUD2),
            RowUD1 = RowUD0 + MaxHeightUD2 div 2 + tree_frame_hor + tree_skip,
            TreeHeightUD = RowUD1 + MaxHeightUD1 div 2 + tree_frame_hor + tree_skip,
            assert(tree_height(MaxHeightUD1, RowUD1, TreeHeightUD)),
            RowMarkUD = RowUD1 + MaxHeightUD1 div 2.

        convert_row_nd(I, H):-
            tree_convert(node(H_, H__, _, _, _, _, I, _, _, _, _, _, _, _, _)),
            H = H__ - H_ - 2 * tree_frame_hor.

        convert_row([] , H, H):-
            !.
        convert_row([H|T] , MaxH0, MaxH2):-
            convert_max(H, MaxH0, MaxH1),
            convert_row(T, MaxH1, MaxH2).

    class predicates
        convert_multiple : (vpiDomains::windowHandle, tree_node) procedure(i,i).
    clauses
        convert_multiple(Win, NODE):-
            NODE = node(_, _, _, _, _, _, _, _, _, Selector, b_false, _, b_true, _, _),
            multiple_selector(Win, Selector),
            !,
            assert(opened_multiple_in_tree_base(Win, NODE)).
        convert_multiple(_, _).

    class predicates
        convert_adjust_tree_size : (tree_node Node,integer Mode) determ(i,i).
        convert_adjust_tree_size_lr : (integer) determ(i).
        convert_adjust_tree_size_ud : (integer) determ(i).
    clauses
        convert_adjust_tree_size(node(_, ColLR, _, _, ColUD, _, _, _, _, _, _, _, _, _, _), Mode):-
            Mode <> convert,
            Mode <> convert_base_,
            Mode <> convert_close_closed_base,
            !,
            OffLR = ColLR + tree_link_width,
            OffUD = ColUD + tree_link_width,
            convert_adjust_tree_size_lr(OffLR),
            convert_adjust_tree_size_ud(OffUD).
        convert_adjust_tree_size(_, _).

        convert_adjust_tree_size_lr(Off):-
            max_off_lr(MaxOff),
            Off <= MaxOff,
            !.
        convert_adjust_tree_size_lr(Off):-
            retract(max_off_lr(_)),
            !,
            assert(max_off_lr(Off)).

        convert_adjust_tree_size_ud(Off):-
            max_off_ud(MaxOff),
            Off <= MaxOff,
            !.
        convert_adjust_tree_size_ud(Off):-
            retract(max_off_ud(_)),
            !,
            assert(max_off_ud(Off)).

    class predicates
        convert_get_size : (vpiDomains::windowHandle TreeWin,integer J,integer NewJ,
                            integer N,integer Mode) determ(i,i,o,o,i).
    clauses
        convert_get_size(Win, _, J_sel, N_, convert_open):-
            !,
            retract(closed_above(Win, closed_mark(_, J_sel))),
            tree_size(Win, _, _, _, _, N),
            !,
            N_ = N + 2 * J_sel.
        convert_get_size(Win, J_sel, 0, N_, convert_close_tree_base):-
            tree_size(Win, _, _, _, _, N),
            !,
            N_ = N - 2 * J_sel.

/******************************************************************************

******************************************************************************/
    class predicates
        get_selected : (vpiDomains::windowHandle TreeWin, integer I,
                        integer J, integer K) procedure(i,o,o,o).

    clauses
        get_selected(W, I, J, K):-
            selected(W, I, J, K),
            !.
        get_selected(_, 0, 0, 0).

    class predicates
	set_selected : (vpiDomains::windowHandle TreeWin,integer I,integer J,integer K,
                        integer Dx,integer Dy,integer Update) procedure(i,i,i,i,i,i,i).
	set_selected1 : (vpiDomains::windowHandle TreeWin,integer I,integer J,integer K,
                         integer Dx, integer Dy, integer Update) determ(i,i,i,i,i,i,i).
	set_selected2 : (vpiDomains::windowHandle TreeWin,integer X,integer Y,
                         integer Update) determ(i,i,i,i).

    clauses
        set_selected(Win, I, J, K, Dx, Dy, Update):-
            set_selected1(Win, I, J, K, Dx, Dy, Update),
            !.
	set_selected(Win, _, _, _, _, _, _):-
            write_base_clean_up(Win),
            retract(tree_above(Win, Node)),
              asserta(tree_below(Win, Node)),
            fail.
	set_selected(Win, I, J, K, Dx, Dy, Update):-
            set_selected1(Win, I, J, K, Dx, Dy, Update),
            !.
	set_selected(_, _, _, _, _, _, _).

	set_selected1(Win, I, J, K, Dx, Dy, Update):-
            screen_context(Win, X, Y, _),
            !,
            retractall(search(Win, _, _, _)),
            assert(search(Win, I, J, K)),
            write_set_mode(Win, write_search),
            write_tree(Win),
            write_set_mode(Win, write_),
            get_update_mode(Update, Update_),
            search_rect(Win, Dx_, Dy_, Update_),
            X_ = X - Dx_ + Dx,
            Y_ = Y - Dy_ + Dy,
            set_selected2(Win, X_, Y_, Update).

        set_selected2(Win, X_, Y_, update_scroll):-
            screen_context(Win, X, Y, _),
            !,
            Dx = X - X_,
            Dy = Y - Y_,
            scroll_tree(Win, Dx, Dy).
	set_selected2(Win, X_, Y_, update_full):-
            retract(screen_context(Win, _, _, Dir)),
            !,
            assert(screen_context(Win, X_, Y_, Dir)),
            resize_tree(Win),
            vpi::winInvalidate(Win),
            vpi::winUpdate(Win).

    clauses
	getAllNodes(W) = L_ :-
		findall(S, get_selector(W, S), L),
		order_list(L, [], L_), !.
	getAllNodes(_) = [] .

	getSelectedNode(W) = Selector :-
		get_selected_node(W, NODE),
		NODE = node(_, _, _, _, _, _, _, _, _, Selector, _, _, _, _, _), !.
	getSelectedNode(_) = "" :-
                common_exception::raise_Error(classInfo, "getSelectedNode").

	getParentNode(W) = Selector :-
		selected(W, I, J, K), !,
		get_parent(W, I, J, K, NODE),
		NODE = node(_, _, _, _, _, _, _, _, _, Selector, _, _, _, _, _).

	getNodeArrowtype(W) = ArrowType :-
		get_selected_node(W, NODE),
		NODE = node(_, _, _, _, _, _, _, _, _, _, _, _, _, _, ArrowType), !.
	getNodeArrowtype(_) = 0 :-
                common_exception::raise_Error(classInfo, "getNodeArrowtype").

	getWininfo(W) = tree_WinInfoAdv(F, D, A, List, vpidomains::pnt(0, 0)) :-
		get_selected_node(W, node(_, _, _, _, _, _, I, J, K, Selector, _, _, _, _, _)),
		get_parent_list(W, I, J, K, [ Selector ], List),
		win_info(W, tree_WinInfo(F, D, A)), !.
	getWininfo(_) = tree_WinInfo(F, 0, []) :-
		F = uncheckedConvert(vpiDomains::font, 0),
                common_exception::raise_Error(classInfo, "getWininfo").

	getWinrect(W, PW) = vpidomains::rct(L0, T0, R0, B0) :-
		HW = get_hscroll_height(W),
		VW = get_vscroll_width(W),
		RCT = vpi::winGetClientRect(W),
		RCT = vpidomains::rct(L, T, R, B),
		Rn = R + VW, Bn = B + HW,
		PNTLIST = vpi::winMapPoints(W, PW, [vpidomains::pnt(L, T), vpidomains::pnt(Rn, Bn)]),
		PNTLIST = [vpidomains::pnt(L0, T0), vpidomains::pnt(R0, B0)],!.
	getWinrect(_,_) = vpidomains::rct(0,0,0,0):-breakControl5x::errorExit(). % Should never be called !

	getCurrentSearch(W) = Selector :-
		search_sel(W, Selector), !.

	setSelectedNode(W, Selector):-
		searchFirstNode(W, Selector), !.
	setSelectedNode(_, _).

	setDirection(W, Dir):-
		retractall(screen_context(W, _, _, _)),
		assert(screen_context(W, 0, 0, Dir)),
		get_selected(W, I, J, K),
		set_selected(W, I, J, K, 0, 0, update_full), !.
	setDirection(_, _):-
                common_exception::raise_Error(classInfo, "setDirection").

	setFont(W, FONT):-
		vpi::cursorSetWait(),
		retract(win_info(W, tree_WinInfo(_, Dir, ArrPenList))),
		assert(win_info(W, tree_WinInfo(FONT, Dir, ArrPenList))),
		vpi::winSetFont(W, FONT),
		convert_tree(W, tree("", unmarked, [], 0), 0, 0, 0, 0, _, 0, "", "", convert_base_),
		selected(W, I, J, K),
		set_selected(W, I, J, K, 0, 0, update_full), !.
	setFont(_, _):-
                common_exception::raise_Error(classInfo, "setFont").

	openCloseNode(W):-
		open_node(W), !.
	openCloseNode(W):-
		close_node(W), !.
	openCloseNode(_).

	searchFirstNode(W, Selector):-
		retractall(search_sel(W, _)),
		assert(search_sel(W, Selector)),
		write_base_clean_up(W),
		retract(tree_above(W, NODE)),
		asserta(tree_below(W, NODE)), fail.
	searchFirstNode(W, Selector):-
		tree_below(W, node(_, _, _, _, _, _, I, J, K, Selector, _, _, _, _, _)),
		retractall(search(W, _, _, _)),
		assert(search(W, I, J, K)),
		write_set_mode(W, write_search),
		write_tree(W),
		write_set_mode(W, write_),
		search_rect(W, Dx, Dy, b_true), !,
		scroll_tree(W, Dx, Dy).
	searchFirstNode(W, _):-
		write_set_mode(W, write_),!,
		fail.
%		vpiCommonDialogs::note("Search node not found").

	searchNodeAgain(W):-
		search_sel(W, Selector),
		selected(W, I, J, K),
		tree_below(W, node(_, _, _, _, _, _, I_, J_, K_, Selector, _, _, _, _, _)),
		search_found(I_, J_, K_, I, J, K),
		retractall(search(W, _, _, _)),
		assert(search(W, I_, J_, K_)),
		write_set_mode(W, write_search),
		write_tree(W),
		write_set_mode(W, write_),
		search_rect(W, Dx, Dy, b_true),
		scroll_tree(W, Dx, Dy), !.
	searchNodeAgain(_):-
		vpiCommonDialogs::note("Search node not found").

	moveToSelectedNode(W):-
		get_selected(W, I, J, K),
		set_selected(W, I, J, K, 0, 0, update_scroll), !.
	moveToSelectedNode(_).

	moveToParentNode(W):-
		get_selected(W, I, J, K),
		get_parent(W, I, J, K, NODE),
		NODE = node(_, _, _, _, _, _, I_, J_, K_, _, _, _, _, _, _),
		set_selected(W, I_, J_, K_, 0, 0, update_scroll), !.
	moveToParentNode(_).

	print(W, Title):-
		VRP = convert(integer, vpi::getAttrVal(attr_printer_vres)),
		HRP = convert(integer, vpi::getAttrVal(attr_printer_hres)),
		VRS = convert(integer, vpi::getAttrVal(attr_screen_vres)),
		HRS = convert(integer, vpi::getAttrVal(attr_screen_hres)),
		VP = convert(integer, vpi::getAttrVal(attr_printer_height)),
		HP = convert(integer, vpi::getAttrVal(attr_printer_width)),
		PW = vpi::printStartJob(Title),
		vpi::winSetMapMode(PW, mm_arbitrary),
		vpi::winSetMapScale(PW, vpidomains::pnt(0, 0), vpidomains::pnt(VRS, HRS), vpidomains::pnt(0, 0), vpidomains::pnt(VRP, HRP)),
		tree_base_copy(W, PW),
		closed_base_copy(W, PW),
		Font = vpi::winGetFont(W),
		vpi::winSetFont(PW, Font),
		vpi::cursorSetWait(),
		convert_tree(PW, tree("", unmarked, [], 0), 0, 0, 0, 0, _, 0, "", "", convert_base_),
		retractall(screen_context(PW, _, _, _)),
		assert(screen_context(PW, 0, 0, tree_DirRight)),
		move_base(PW),
		write_base_clean_up(PW),
		write_set_mode(PW, write_),
		retractall(_, print_base),
		assert(print_bottom(0, 0, 0)),
		LHRS = convert(core::integer32,HRS),
		LPageWidth =(HP * LHRS) div HRP,
%		HDel = HRS div HRP,	% 12.01.1998 BB!(replace "/" to "div")
		PageWidth = convert(integer,LPageWidth),
		LVRS = convert(core::integer32,VRS),
		LPageHeight =(VP * LVRS) div VRP,
%		VDel = VRS div VRP,	% 12.01.1998 BB!(replace "/" to "div")
		PageHeight = convert(integer,LPageHeight),
		StartX = HRS div 2,
		StartY = VRS div 2,
		vpi::winGetFontMetrics(PW, Leading, Ascent, Descent),
		HeaderHei = Leading + Ascent + Descent + StartY,
		print_tree_adjust(PW, PageWidth),
		print_tree(PW, StartX, StartY, HeaderHei, Title, 1, PageHeight, Font),
		vpi::printEndJob(PW),
		tree_base_clean_up(PW),
		write_base_clean_up(PW),
		retractall(_, print_base), !.
	print(_, _):-
                common_exception::raise_Error(classInfo, "print").
end implement tree
