/*****************************************************************************

                        Copyright (c) Jens H. Holm

******************************************************************************/

class treePack
open core

constants
        tree_DirRight : integer = 0.
        tree_DirLeft  : integer = 1.
        tree_DirDown  : integer = 2.
        tree_DirUp    : integer = 3.


domains
        node_mark      = marked;
                         unmarked.

        selector       = string.
        direction      = integer.
	arrow_type     = integer.

	arrow_pen      = tree_arrowPen(arrow_type, vpiDomains::pen).
	arrow_pen_list = arrow_pen*.

        win_info       = tree_WinInfo(vpiDomains::font, direction, arrow_pen_list);
                         tree_WinInfoAdv(vpiDomains::font, direction, arrow_pen_list,
                                         core::string_list, vpiDomains::pnt).

        tree           = tree(selector, node_mark, tree_list, arrow_type).
        tree_list      = tree*.
        tree_lists    = tree_list*.


predicates
classInfo : core::classInfo.

    predicates
        create : (vpiDomains::wintype WinType,
                  vpiDomains::rct,
                  string Title,
                  vpiDomains::menu Menu,
                  vpiDomains::windowHandle Parent,
                  vpiDomains::wsflags Flags,
                  vpiDomains::ehandler Eventhandler,
                  core::integer32 CreationData,
                  tree Tree,
                  win_info TreeWindowInfo) -> vpiDomains::windowHandle procedure(i,i,i,i,i,i,i,i,i,i).
        %
        % @short Main predicate - creates a tree window with the tree Tree,
        % the path Path and the c-eventhandler EventHandler.
        % @end

    predicates
        createSubclass : (vpiDomains::windowHandle Win,tree Tree,win_info TreeWindowInfo) procedure(i,i,i).
        %
        % @short Creates the tree for the given window
        % @end

    predicates
        getAllNodes : (vpiDomains::windowHandle TreeWin) -> core::string_list procedure(i).
        %
        % @short Collects all tree Selectors
        % @end

    predicates
        getSelectedNode : (vpiDomains::windowHandle TreeWin) -> selector procedure(i).
        %
        % @short Returns the currently selected node's Selector
        % @end

    predicates
        getParentNode : (vpiDomains::windowHandle TreeWin) -> selector determ(i).
        %
        % @short Returns the parent's Selector for the selected node
        % @end

    predicates
        getNodeArrowtype : (vpiDomains::windowHandle TreeWin) -> arrow_type procedure(i).
        %
        % @short Returns the currently selected node's arrow type
        % @end

    predicates
        getWininfo : (vpiDomains::windowHandle TreeWin) -> win_info procedure(i).
        %
        % @short Gets tree window data
        % @end

    predicates
        getWinrect : (vpiDomains::windowHandle TreeWin,
                      vpiDomains::windowHandle OutWindow) -> vpiDomains::rct procedure(i,i).
        %
        % @short Used to obtain relative tree-window coordinates in OutWindow
        % @end

    predicates
        getCurrentSearch : (vpiDomains::windowHandle TreeWin) -> selector determ(i).
        %
        % @short Returns selector used in a current serch, or fails
        % @end

    predicates
        setSelectedNode : (vpiDomains::windowHandle TreeWin,selector Selector) procedure(i,i).
        %
        % @short Moves selection to the first node matching the Selector
        % @end

    predicates
        setDirection : (vpiDomains::windowHandle TreeWin,integer Direction) procedure(i,i).
        %
        % @short Sets new direction for tree drawing
        % @end

    predicates
        setFont : (vpiDomains::windowHandle TreeWin,vpiDomains::font Font) procedure(i,i).
        %
        % @short Sets new font for tree Selectors
        % @end

    predicates
        update : (vpiDomains::windowHandle TreeWin,tree Tree) procedure(i,i).
        %
        % @short Used to change the tree in a window without deleting the window
        % @end

    predicates
        openCloseNode : (vpiDomains::windowHandle TreeWin) procedure(i).
        %
        % @short explodes/implodes subtree from the selected node
        % @end

    predicates
        searchFirstNode : (vpiDomains::windowHandle TreeWin,selector Selector) determ(i,i).
        %
        % @short Selects first node using the Selector
        % @end

    predicates
        searchNodeAgain : (vpiDomains::windowHandle TreeWin) procedure(i).
        %
        % @short Selects next node using the choosen Selector
        % @end

    predicates
        moveToSelectedNode : (vpiDomains::windowHandle TreeWin) procedure(i).
        %
        % @short Makes selected node visible
        % @end

    predicates
        moveToParentNode : (vpiDomains::windowHandle TreeWin) procedure(i).
        %
        % @short Moves selection from child node to the parent node
        % @end

    predicates
        print : (vpiDomains::windowHandle Win,string Title) procedure(i,i).
        %
        % @short Prints a tree
        % @end
end class treePack