/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement tt
open core,termDomain,resourceIdentifiers,globalDomains

class facts

tFile : outputStream := erroneous.
traceLevel : integer := 0. % 1-5
traceFile : string := "". 
traceClose : integer := 0. % -1/1 always close after write
traceStatus : integer := -1. % -1=file not created; 0=created and closed; 1=created and open
freeGlobal : real := 100.0.
freeStack : real := 100.0.
freeHeap : real := 100.0.
freeTrail : real := 100.0.
profileStart : string := "".    
constants

className = "com/rudstrand/tt/tt".
classVersion = "1.0".

class predicates

ttMultiLineW : (integer I,string S) determ.

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% getFile
% ----------------------------------------------------------------------------------------------
getFile()=F :- F=traceFile.
setFile(F) :- F2=filename::setExtension(F,"3tc"), traceFile:=F2,!.
getLevel()=L :- L=traceLevel.
setLevel(L) :- 
    traceLevel:=L,
 	Ls=toString(L),
 	W=guix::getWinOfClas(1),
	vpiToolbar::setValue(W,idt_tracelevel,vpiToolbar::text_value(Ls)),
    !.
    
getClose()=L :- L=traceClose.
setClose(L) :- traceClose:=L.
% ----------------------------------------------------------------------------------------------
% tt_cc
% ----------------------------------------------------------------------------------------------
% temp. closes trace file if this is opted for, else leaves file as is.
%
tt_cc() :-
	traceStatus=1,
	traceClose=1,
   	tFile:close(),
	traceStatus:=0,
	!.
tt_cc() :- 
	!.	
% ----------------------------------------------------------------------------------------------
% tt_close
% ----------------------------------------------------------------------------------------------
% permanently closes trace file. eventual sub-sequent openings will scratch the contents.
%
tt_close() :-
    traceStatus<>-1,
    tt_co,
   	trap(tFile:close(),Error,pfeil::errorHandler("tt_close",Error)),
	traceStatus:=-1,
	!.
tt_close() :- 
    !.
% ----------------------------------------------------------------------------------------------
% tt_co
% ----------------------------------------------------------------------------------------------
% opens trace file if non-existing or temp. closed. always exits with an open file. never fails.
%
tt_co() :-
	traceStatus=1,
	!
;
    traceStatus=0,
   	tFile:=outputStream_file::append(traceFile),
	traceStatus:=1,
	!
;	
    traceStatus=-1,
   	tFile:=outputStream_file::create(traceFile),
	traceStatus:=1,
	!.
% ----------------------------------------------------------------------------------------------
% reset
% -------------------------------------------------------------------------------------------
reset() :-
    tt_close,
    tt_open,
    !.
% ----------------------------------------------------------------------------------------------
% ttHeader
% -------------------------------------------------------------------------------------------
ttHeader(Lev,_,"") :-
	traceLevel>=Lev, 
	tt_co,
	!.
ttHeader(Lev,DataType,Point) :-
	traceLevel>=Lev,
	tt_co,
	S=string::concat("#",DataType," [",Point,"]"),
	wr([S]),
	!.
% ----------------------------------------------------------------------------------------------
% ttMultiLineW
% ----------------------------------------------------------------------------------------------
class predicates
ttMultiLineW1 : (integer I,string S,string T) determ.
ttMultiLine2 : (string S1,string S2,string S3) procedure (i,o,o).
clauses
ttMultiLineW(0,S) :- 
    ttMultiLineW1(0,"",S),
    !.
ttMultiLineW(Ind,S) :-
    sx::sx_ini(Ind,' ',S1),
	ttMultiLineW1(Ind,S1,S),
	!.
	
ttMultiLineW1(_,_,"") :- 
    !.
ttMultiLineW1(Ind,SInd,S) :-
	% rest string fits into one line
	N=string::length(S),
	M=Ind+N,
	M<=95,
	wrl([SInd,S]),!
;
	% indentation too big - write rest
	Ind>70,
	wrl([SInd,S]),
	!
;	
	% rest string does not fit into one line
	L=80-Ind,
	sx::frontstr(L,S,F1,T1),
	sx::frontstr(15,T1,F2,T2),
	ttMultiLine2(F2,F3,T3),
	wrl([SInd,F1,F3]),
	T4=string::concat(T3,T2),
	ttMultiLineW1(Ind,SInd,T4),
	!.		
ttMultiLine2(S,F,T) :-
	sx::searchstring(S,", ",P),
	N=P+1,
	sx::frontstr(N,S,F,T),!
;
	sx::searchstring(S,") ",P),
	N=P+1,
	sx::frontstr(N,S,F,T),!
;
	sx::searchstring(S,"} ",P),
	N=P+1,
	sx::frontstr(N,S,F,T),!
;
	sx::searchstring(S,"& ",P),
	N=P+1,
	sx::frontstr(N,S,F,T),!
;
	sx::searchstring(S,":- ",P),
	N=P+1,
	sx::frontstr(N,S,F,T),!
;
	sx::searchstring(S,",",P),
	sx::frontstr(P,S,F,T),!
;		
	sx::searchstring(S,")",P),
	sx::frontstr(P,S,F,T),!
;		
	sx::searchstring(S,"}",P),
	sx::frontstr(P,S,F,T),!
;		
	sx::searchstring(S,"&",P),
	sx::frontstr(P,S,F,T),!
;		
	sx::searchstring(S,":-",P),
	N=P+1,
	sx::frontstr(N,S,F,T),!
;		
	F=S,
	T="",!.
% ----------------------------------------------------------------------------------------------
% tt_header
% ----------------------------------------------------------------------------------------------
tt_header(Lev,S) :-
	traceLevel>=Lev, 
	tt_co,
	sx::sx_ini(80,'*',Sx),
	wrl([Sx]),
	LPad=40-string::length(S) div 2,
	sx::sxL_Pad(S," ",LPad,Sp),
	wrl([Sp]),
	wrl([Sx]),
	tt_cc,!
;
	!.	
% ----------------------------------------------------------------------------------------------
% startProfile
% ----------------------------------------------------------------------------------------------
startProfile(Loc) :-
    tt::getLevel()=0,
    !
;    
    profileStart<>"",
    Mess=string::concat("Previous profile start at ",profileStart," superseded by new start at ",Loc,"!"),
    dlgIF::msgS(Mess),
    fail
;    
    profileStart:=Loc,
    profileTime::init,
    gcStatistics::enable,
    !.
% ----------------------------------------------------------------------------------------------
% stopProfile
% ----------------------------------------------------------------------------------------------
stopProfile() :- 
    tt::getLevel()=0,
    !
;    
	ttHeader(1,"PROFILER",""),
	wrl([]),
	wrl(["Start Point:",profileStart]),
	wrl([]),
	gcStatistics::getGcCounters(G1,G2,G3),
	G1=unsigned64(L1,H1),
	G2=unsigned64(L2,H2),
	G3=unsigned64(L3,H3),
	SL1=toString(L1),
	SL2=toString(L2),
	SL3=toString(L3),
	SH1=toString(H1),
	SH2=toString(H2),
	SH3=toString(H3),
    S1=string::concat("Total program  =",SH1,SL1),
    S2=string::concat("Total GC      =",SH2,SL2),
    S3=string::concat("Total finalizer=",SH3,SL3),
    wrl(["GC Counters"]),
    wrl([S1]),
    wrl([S2]),
    wrl([S3]),
    wrl([]),
    /*
	gcStatistics::getGcTimes(XG1,XG2,XG3),
	XG1=integer64(XL1,XH1),
	XG2=integer64(XL2,XH2),
	XG3=integer64(XL3,XH3),
	XSL1=toString(XL1),
	XSL2=toString(XL2),
	XSL3=toString(XL3),
	XSH1=toString(XH1),
	XSH2=toString(XH2),
	XSH3=toString(XH3),
    XS1=string::concat("Total program  =",XSH1,XSL1),
    XS2=string::concat("Total GC      =",XSH2,XSL2),
    XS3=string::concat("Total finalizer=",XSH3,XSL3),
    wrl(["GC Times"]),
    wrl([XS1]),
    wrl([XS2]),
    wrl([XS3]),
    wrl([]),
    */
	wrl(["Center                  |     #calls    |gross avg/call | gross total   |  net avg/call |     net total |      GC       "]),
    profileTime::statistics_nd(Center,Ai,Bf,Cf,Df,Ef,Ff,_),
	As=toString(Ai),
    sx::fixFormatR(Bf,7,1,Bs),
    sx::fixFormatR(Cf,4,1,Cs),
    sx::fixFormatR(Df,7,1,Ds),
    sx::fixFormatR(Ef,4,1,Es),
    sx::fixFormatR(Ff,4,1,Fs),
    sx::sxL_Pad(As,"0",15,A),
    sx::sxL_Pad(Bs,"0",15,B),
    sx::sxL_Pad(Cs,"0",15,C),
    sx::sxL_Pad(Ds,"0",15,D),
    sx::sxL_Pad(Es,"0",15,E),
    sx::sxL_Pad(Fs,"0",15,F),
    sx::sxR_Pad(Center," ",24,CenterP),
    sx::sxTrunc(CenterP,24,CenterT),
	sx::sxlJoin([CenterT,A,B,C,D,E,F],"|",T),
    wrl([T]),
    fail
;
    profileStart:="",
	tt_cc,
	!.	
% ----------------------------------------------------------------------------------------------
% memory
% ----------------------------------------------------------------------------------------------
memory(Lev,S) :- 
	ttHeader(Lev,"MEMORY",S),
	freeMemory(Af,Bf,Cf,Df,MAf,MBf,MCf,MDf),
	As=toString(Af),
	Bs=toString(Bf),
	Cs=toString(Cf),
	Ds=toString(Df),
	MAs=toString(MAf),
	MBs=toString(MBf),
	MCs=toString(MCf),
	MDs=toString(MDf),
	wrl([]),
	wrl(["Free global % ",As,"(",MAs,")"]), 
	wrl(["Free heap % ",Bs,"(",MBs,")"]), 
	wrl(["Free stack % ",Cs,"(",MCs,")"]), 
	wrl(["Free trail % ",Ds,"(",MDs,")"]), 
	tt_cc,
	!
;
	!.	

freeMemory(Glob,Heap,Stack,Trail,Fg,Fh,Fs,Ft) :-
	A=memory::getFreeGlobalStack(),
	B=memory::getFreeHeap(),
	C=memory::getFreeStack(),
	D=memory::getFreeTrail(),
	E=memory::getTotalGlobalStack(),
	F=memory::getTotalTrail(),
	G=memory::getUsedStack(),
	H=memory::getUsedHeap(),
	Glob=(A/E)*100.0,
	Heap=(B/(B+H))*100.0,
	Stack=(C/(C+G))*100.0,
	Trail=(D/F)*100.0,
	set::setMin([Glob,freeGlobal],Fg),
	set::setMin([Heap,freeHeap],Fh),
	set::setMin([Stack,freeStack],Fs),
	set::setMin([Trail,freeTrail],Ft),
	freeGlobal:=Fg,
	freeStack:=Fs,
	freeHeap:=Fh,
	freeTrail:=Ft,
    !.
% ----------------------------------------------------------------------------------------------
% tt_i
% ----------------------------------------------------------------------------------------------
tt_i(Lev,S,I) :- 
	ttHeader(Lev,"INTEGER",S),
	Is=toString(I),
	wrl([Is]), 
	tt_cc,
	!
;
	!.	
tt_n(Lev,S,I) :- 
	ttHeader(Lev,"NUMERIC",S),
	Is=toString(I),
	wrl([Is]), 
	tt_cc,
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
% tt_is/iss
% ----------------------------------------------------------------------------------------------	
class predicates
tt_isx : (integer_list L) determ (i).
tt_iss_x : (integer N,integer_list_list IL) determ (i,i).
clauses
tt_is(Lev,S,Is) :- 
	ttHeader(Lev,"INTEGERS",S),
	tt_isx(Is), 
	tt_cc,
	!
;
	!.
tt_isx([]) :- 
    wrl([]),
    !.
tt_isx([H|T]) :- 
    Hs=toString(H),
    wr([Hs,";"]), 
    tt_isx(T),
    !.
tt_iss(Lev,S,T) :-
	ttHeader(Lev,"LIST OF INTEGER LISTS",S),
	wrl([]),
	tt_iss_x(1,T),
	tt_cc,
	!
;
	!.
tt_iss_x(_,[]) :- 
    !.
tt_iss_x(N,[H|T]) :-
	S1=toString(N),
	wrl(["  List ",S1]),
	tt_isx(H),
	M=N+1,
	tt_iss_x(M,T),
	!.
% ----------------------------------------------------------------------------------------------
% tt_pie
% ----------------------------------------------------------------------------------------------
tt_pie(Lev,Ex,N) :-
	Ex=-1,
	traceLevel>-Lev, 
	tt_co,
	Ns=toString(N),
	sx::concat([">>>>> ",Ns," "],S1),
	sx::sxR_Pad(S1,">",60,S),
	wrl([S]),
	tt_cc,
	!
;
	Ex=1,
	traceLevel>=Lev, 
	tt_co,
	Ns=toString(N),
	sx::concat(["<<<<< ",Ns," "],S1),
	sx::sxR_Pad(S1,"<",60,S),
	wrl([S]),
	tt_cc,
	!
;
	Ex=0,
	traceLevel>=Lev, 
	tt_co,
	Ns=toString(N),
	sx::concat(["<><><> ",Ns," is reused <><><><><><>"],S),
	wrl([S]),
	tt_cc,
	!
;
	!.	
% ----------------------------------------------------------------------------------------------
% tt_open
% ----------------------------------------------------------------------------------------------
tt_open() :-
    traceStatus=1,
	Ls=toString(traceLevel),
	W=guix::getWinOfClas(1),  
	vpiToolbar::setValue(W,idt_tracelevel,vpiToolbar::text_value(Ls)),
    !
;    
    traceStatus=0,
    tt_co,
    !
;    
	globalVars::get_s("kb_file")="",
	!,
   	traceFile:="pro3.3tc",
   	tFile:=outputStream_file::create(traceFile),
    traceStatus:=1,
    set::setMax([traceLevel,1],Level),
   	setLevel(Level),
   	tt_cc,
   	!
;
   	tFile:=outputStream_file::create(traceFile),
   	traceStatus:=1,
    set::setMax([traceLevel,1],Level),
   	setLevel(Level),
   	tt_cc,
    !.
% ----------------------------------------------------------------------------------------------
% ttPieS/T
% ----------------------------------------------------------------------------------------------
ttPieS(S1,S2) :-
	tt_co,
    U=string::concat(S1,S2),
    wrx([U,"\n"]),
    tt_cc,
    !;!.

ttPieT(S,T) :-    
	tt_co,
    wrx([S,"\n"]),
	tt_term_x("         ",T),
    tt_cc,
    !;!.
% ----------------------------------------------------------------------------------------------
% tt_s/snl
% ----------------------------------------------------------------------------------------------
class predicates
tt_s_1 : (string) determ.
clauses

tt_s(Lev,S,T) :-
	ttHeader(Lev,"STRING",S),
	tt_s_1(T),
	tt_cc,
	!
;
	!.	
tt_s_1(S) :-
	string::length(S)<40,
	wrl([S]),
	!
;
	wrl([]),
	ttMultiLineW(0,S),
	!.	
% ----------------------------------------------------------------------------------------------
% tt_ss/ssnl
% ----------------------------------------------------------------------------------------------
tt_ss(Lev,S) :- 
	traceLevel>=Lev, 
	tt_co,
	wrl([S]),
	tt_cc,!
;
	!.	
/*
tt_ssnl(Lev,K) :- 
	traceLevel>=Lev, 
    kbNL_Try(K,S),
	tt_co,
	wr(S,"\n"),
	tt_cc,
	!
;
	!.
*/	
% ----------------------------------------------------------------------------------------------
% tt_strings
% ----------------------------------------------------------------------------------------------
class predicates
tt_strings_x : (integer,string_list) determ (i,i).
clauses
tt_strings(Lev,S,T) :-
	ttHeader(Lev,"STRINGS",S),
	wrl([]),
	tt_strings_x(1,T),
	tt_cc,
	!
;
	!.
tt_strings_x(_,[]) :- 
    !.
tt_strings_x(N,[H|T]) :-
	S=toString(N),
	wrl(["     ",S,":",H]),
	M=N+1,
	tt_strings_x(M,T),
	!.
% ----------------------------------------------------------------------------------------------
% tt_stringss
% ----------------------------------------------------------------------------------------------
class predicates
tt_stringss_x : (integer,string_list_list) determ.
clauses
tt_stringss(Lev,S,T) :-
	ttHeader(Lev,"LIST OF STRING LISTS",S),
	wrl([]),
	tt_stringss_x(1,T),
	tt_cc,
	!
;
	!.
tt_stringss_x(_,[]) :- 
    !.
tt_stringss_x(N,[H|T]) :-
	S1=toString(N),
	wrl(["  List ",S1]),
	tt_strings_x(1,H),
	M=N+1,
	tt_stringss_x(M,T),
	!.
% ----------------------------------------------------------------------------------------------
% tt_term
% ----------------------------------------------------------------------------------------------
class predicates
tt_term_2 : (string S,sterml T) determ.
tt_term_3 : (integer I,string S,string T) determ.
tt_term_x : (string Indent,sterm T) determ.
clauses

tt_term(Lev,S,T) :- 
	ttHeader(Lev,"TERM",S),
	wrl([]),
	tt_term_x("",T),
	tt_cc,
	!
;
	!.
	
tt_term_x(Ind,cmp(Head,Tail)) :-
	wrl([Ind,Head]),
	NewInd=string::concat(Ind,"     "),
	tt_term_2(NewInd,Tail),
	!.
tt_term_x(Ind,list(Head,Tail)) :-
    S=toString(Head),
	wrl([Ind,S]),
	tt_term_x(Ind,Tail),
	!.
	
tt_term_x(Ind,nill) :- wrl([Ind,"nill"]),!.
tt_term_x(Ind,str(V)) :- wrl([Ind,V," {str}"]),!.
tt_term_x(Ind,int(V)) :- S=toString(V), wrl([Ind,S," {int}"]),!.
tt_term_x(Ind,fix(V)) :- S=toString(V), wrl([Ind,S," {fix}"]),!.
tt_term_x(Ind,atom(V)) :- wrl([Ind,V," {atom}"]),!.
tt_term_x(Ind,var(V)) :- wrl([Ind,V," {var}"]),!.
tt_term_x(Ind,nill) :- wrl([Ind," {nill}"]),!.
tt_term_x(Ind,ffix(V,D)) :- S=toString(V), wrl([Ind,S," {ffix;",D,"}"]),!.
tt_term_x(Ind,fstr(V,D,F,W)) :- S=toString(W), wrl([Ind,V," {fstr;",D,";",F,";",S,"}"]),!.
tt_term_x(Ind,fint(V,D)) :- S=toString(V), wrl([Ind,S," {fint;",D,"}"]),!.
tt_term_x(Ind,fvar(V,D)) :- wrl([Ind,V," {fvar;",D,"}"]),!.
tt_term_x(Ind,Term) :-	
	S=toString(Term),
	N=string::length(Ind),
	tt_term_3(N,Ind,S),!.
tt_term_3(N,Ind,S) :-
	L=N+string::length(S),
	L<85,	
	wrl([Ind,S]),
	!
;
	ttMultiLineW(N,S),
	!.	
tt_term_2(_,[]) :- 
    !.
tt_term_2(Ind,[H|T]) :-		
	tt_term_x(Ind,H),
	tt_term_2(Ind,T),
	!.
% ----------------------------------------------------------------------------------------------
% tt_terms
% ----------------------------------------------------------------------------------------------
class predicates
tt_terms_x : (integer N,sterml) determ.
clauses
tt_terms(Lev,S,T) :-
	ttHeader(Lev,"TERMS",S),
	wrl([]),
	tt_terms_x(1,T),
	tt_cc,
	!
;
	!.
tt_terms_x(_,[]) :- 
    !.
tt_terms_x(N,[H|T]) :-
    Ns=toString(N),
	wrl(["     Term ",Ns,":"]),
	tt_term_x("     ",H),
	M=N+1,
	tt_terms_x(M,T),
	!.
% ----------------------------------------------------------------------------------------------
% tt_termslist
% ----------------------------------------------------------------------------------------------
class predicates
tt_terms_list_x : (integer,sterml_list) determ.
clauses
tt_termslist(Lev,S,T) :-
	ttHeader(Lev,"TERMSLIST",S),
	wrl([]),
	tt_terms_list_x(1,T),
	tt_cc,!
;
	!.
tt_terms_list_x(_,[]) :- !. 
tt_terms_list_x(N,[H|T]) :- 
    Ns=toString(N),
	wrl(["     Termlist ",Ns,":"]),
	wrl([]),
	tt_terms_x(1,H),
	M=N+1,
	tt_terms_list_x(M,T),!.
% ----------------------------------------------------------------------------------------------
% tt_tokl
% ----------------------------------------------------------------------------------------------
class predicates
tt_tokl_x : (pie_scan::tokl) determ.
clauses
tt_tokl(Lev,S,T) :-
	ttHeader(Lev,"TOKEN LIST",S),
	wrl([]),
	tt_tokl_x(T),
	tt_cc,
	!
;
	!.
tt_tokl_x([]) :- 
    !.
tt_tokl_x([pie_scan::t(Tok,Cur)|T]) :- 
	SCur=toString(Cur),
	STok=toString(Tok),
	sx::concat([SCur," ",STok],S),
	ttMultiLineW(5,S),
	tt_tokl_x(T),
	!.
% ----------------------------------------------------------------------------------------------
% tt_wrd; tt_wrds
% ----------------------------------------------------------------------------------------------
class predicates
ttWrdsX : (pie_scan::wrds).
clauses
tt_wrd(Lev,S,pie_scan::w(N,R,C,F)) :-
	ttHeader(Lev,"WORD",S),
	wrl([]),
	Ns=toString(N),
	wrl(["     ",Ns," ",R," ",C," ",F]),
	wrl([]),
	tt_cc,
	!
;
	!.
tt_wrds(Lev,S,P) :-
	ttHeader(Lev,"WORDS",S),
	wrl([]),
	ttWrdsX(P),
	tt_cc,
	!
;
	!.
ttWrdsX([]) :- 
    !.
ttWrdsX([pie_scan::w(N,R,C,F)|T]) :- 
	Ns=toString(N),
    wrl(["     ",Ns," ",R," ",C," ",F]), 
    wrl([]), 
    ttWrdsX(T),
    !.
% ----------------------------------------------------------------------------------------------
% tt_table
% ----------------------------------------------------------------------------------------------
class predicates
tt_tables : (integer,integer,integer,integer) determ.
tt_table_t : (integer,integer,integer) determ.
tt_table_a : (integer,integer,integer,integer) determ.	
tt_table_h : (integer,integer,integer,integer,integer) determ.
tt_table_c : (integer,integer,integer,integer,integer) determ.
tt_table_contents : (integer,integer,integer,integer,integer) determ.
clauses
tt_table(Lev,S,SolRef) :-
	ttHeader(Lev,"TABLES",S),
	stb::stbM_GetF(SolRef,NTables,PredKB,PredNL,NEntArgs,NArgs),	
	sol::solSetSize(SolRef,NSol),
	SSolRef=toString(SolRef),
	SNSol=toString(NSol),
	SNEntArgs=toString(NEntArgs),
	SNArgs=toString(NArgs),
	SNTables=toString(NTables),
	wrl(["Based on solution set no ",SSolRef," with ",SNSol," solutions"]),
	wrl(["# entity arguments     =",SNEntArgs]),	
	wrl(["# table arguments      =",SNArgs]),	
	wrl(["predicate              =",PredNL," [",PredKB,"]"]),	
	wrl(["# tables               =",SNTables]),wrl([]),
	tt_tables(SolRef,1,NTables,NArgs),
	!
;	
	not(stb::stbM_GetF(SolRef,_,_,_,_,_)),
	SSolRef=toString(SolRef),
	wrl(["no table exists for sol.set ",SSolRef]),
	tt_cc,
	!	
;
	tt_cc,
	!.
tt_tables(_,Table,NTables,_) :- 
	Table>NTables,
	!,fail.
tt_tables(SolRef,Table,NTables,NArgs) :-
	tt_table_t(SolRef,Table,NArgs),
	Next=Table+1,
	tt_tables(SolRef,Next,NTables,NArgs),
	!.
	
% always fails
%
tt_table_t(SolRef,Table,NArgs) :-
	stb::stbH_Get(SolRef,Table,NCols,NRows,ST,Sort,_,_),
	STable=toString(Table),
	SNRows=toString(NRows),
	SNCols=toString(NCols),
	SSort=toString(Sort),
	wrl(["table no =",STable]),
	wrl(["#rows    =",SNRows]),
	wrl(["#columns =",SNCols]),
	wrl(["ST       =",ST]),
	wrl(["Sorted   =",SSort]),
	tt_table_a(SolRef,Table,1,NArgs),
	tt_table_contents(Sort,SolRef,Table,NRows,NCols),!.
tt_table_contents(0,SolRef,Table,NRows,NCols) :-
	% for all rows in table
	Row=programControl::sequence_nd(1,NRows,1),
	   tt_table_c(SolRef,Table,Row,1,NCols),!.
tt_table_contents(1,SolRef,Table,NRows,NCols) :-
	sdb::tabName(SolRef,Table,TabName),
	UI=string::concat("X",TabName),
	XI=sdb::getIx(UI),
	% for all rows in table
	LRow=programControl::sequence_nd(1,NRows,1),
	   stb::stbGetRowL_P_F(XI,LRow,Row),
	   SLRow=toString(LRow),
	   wrl(["     Log.row.no=",SLRow]),		
	   tt_table_c(SolRef,Table,Row,1,NCols),!.
% outputs all arguments - always succeeds
%
tt_table_a(_,_,ArgNo,NArgs) :- ArgNo>NArgs,!.
tt_table_a(SolRef,Table,ArgNo,NArgs) :-
	stb::stbA_Get(SolRef,Table,ArgNo,ArgKB,ArgNL,NDEs),
	SArgNo=toString(ArgNo),
	SNDes=toString(NDEs),
	wrl(["argument#        =",SArgNo]),
	wrl(["argument         =",ArgNL," [",ArgKB,"]"]),
	wrl(["# data elements  =",SNDEs]),
	tt_table_h(SolRef,Table,ArgNo,1,NDEs),
	Next=ArgNo+1,
	tt_table_a(SolRef,Table,Next,NArgs),!.
tt_table_h(_,_,_,DE_No,NDEs) :- DE_No>NDEs,!.
tt_table_h(SolRef,Table,ArgNo,DE_No,NDEs) :-
	stb::stbD_GetArg(SolRef,Table,ArgNo,DE_No,ColNo,DE_KB,DE_NL,Domain,Form,Width,Pixels,Display,Sorted),
	SDE_No=toString(DE_No),
	SColNo=toString(ColNo),
	SWidth=toString(Width),
	SPixels=toString(Pixels),
	SDisplay=toString(Display),
	SSorted=toString(Sorted),
	wrl(["DE#    =",SDE_No]),
	wrl(["DE     =",DE_NL," [",DE_KB,"]"]),
	wrl(["col#   =",SColNo]),
	wrl(["domain =",Domain]),
	wrl(["format =",Form]),
	wrl(["width  =",SWidth]),
	wrl(["pixels =",SPixels]),
	wrl(["display=",SDisplay]),
	wrl(["sort   =",SSorted]),wrl([]),
	Next=DE_No+1,
	tt_table_h(SolRef,Table,ArgNo,Next,NDEs),!.
% outputs all attributes - always fails i.e. when no more attributes
%
tt_table_c(_,_,_,ColNo,NCols) :- 
	ColNo>NCols,wrl([]),
	!,fail.
tt_table_c(SolRef,Table,Row,ColNo,NCols) :-
	stb::stbCellGet(SolRef,Table,Row,ColNo,ValNL,Val),
	SSolRef=toString(SolRef),
	STable=toString(Table),
	SRow=toString(Row),
	SColNo=toString(ColNo),
	Ts=toString(Val),
	wrl(["     ",SSolRef,"/",STable,"/",SRow,"/",SColNo,": ",ValNL,"     ",Ts]),
	Next=ColNo+1,
	tt_table_c(SolRef,Table,Row,Next,NCols),
	!.
% ----------------------------------------------------------------------------------------------
% wr/l
% ----------------------------------------------------------------------------------------------
class predicates
wrx : (string_list) determ.
clauses

wrl(L) :-
    wr(L),
    wr(["\n"]),
    !.
    
wr(_) :-
    traceStatus<>1,
    !.
wr(L) :-
    wrx(L),
    !
;
    !.
    
wrx([]) :-
    !.
wrx([H|T]) :-
    tFile:write(H),
    wrx(T),
    !.    
end implement tt
