/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement vars
open core,termDomain

constants

className = "com/rudstrand/vars/vars".
classVersion = "1.0".

clauses
classInfo(className, classVersion).
% ----------------------------------------------------------------------------------------------
% varDE/_Dom
% ----------------------------------------------------------------------------------------------
varDE(V,DE) :-
	varComps(V,_,_,DE),
	!.
	
varDE_Dom(V,Dom,F,W) :-
	varDE(V,DE),
	dbBase::smClassDef(DE,"a",_,_,Dom,F,W),
	!.
% ---------------------------------------------------------------------------
% varSimple
% ---------------------------------------------------------------------------
varSimple(V) :-
    V<>"",
	First=string::subString(V,0,1),
	sx::isLetterS(First),
	string::isName(V),
	UFirst=string::toUpperCase(First),
	UFirst=First,
	!.
% -----------------------------------------------------------------------------------------------
% varPrefix
% -----------------------------------------------------------------------------------------------
varPrefix(ArgNo,Arg,Prefix) :-
	SNo=toString(ArgNo),
	string::front(Arg,1,F,T),
	Fu=string::toUpperCase(F),
	sx::concat([Fu,T,"_",SNo],Prefix),
	!.
% ----------------------------------------------------------------------------------------------
% varStr
% ----------------------------------------------------------------------------------------------
varStr(var(V),V) :- !.
varStr(fvar(V,_),V) :- !.
% -------------------------------------------------------------------------------------
% varFuncPars
% -------------------------------------------------------------------------------------
class predicates
varFuncParsX : (string,string_list,sterml) determ (i,i,o).
clauses

varFuncPars(Func,Pars) :-
	dbBase::smDef(Func,Type,_,DEs,_,_,_),
	set::member(Type,["f","g","h","i"]),
	sx::frontchar(Func,F,T),
	Fs=string::charToString(F),
	Fu=string::toUpperCase(Fs),
	Prefix=string::concat(Fu,T,"__"),
	vars::varFuncParsX(Prefix,DEs,Pars),
	!.
	
varFuncParsX(_,[],[]) :-
	!.
varFuncParsX(Prefix,[D|Ds],[P|Ps]) :-
	dbBase::smClassDef(D,"a",_,_,Dom,_,_),
	V=string::concat(Prefix,D),
	P=fvar(V,Dom),
	vars::varFuncParsX(Prefix,Ds,Ps),
	!.
% ------------------------------------------------------------------------------------------------------------
% varGenPrefix                            
% ------------------------------------------------------------------------------------------------------------
% generates the entity/predicate prefix.
% parameters:
% 1	in	entity/predicate name 
% 2	out	prefix
%
varGenPrefix(Name,Prefix) :-
	globalVars::incr("pre_cnt",_,N),
	varGenPrefixNo(Name,N,Prefix),
	!.
% -------------------------------------------------------------------------------------
% varGen
% -------------------------------------------------------------------------------------
% generates a pure variable
% (used to use global varno instd of pre_cont)
% 
varGen(Var) :-
	globalVars::incr("pre_cnt",_,Next),
	S=toString(Next),
	Var=string::concat("X",S),
	!.
varGenP(Prefix,Var) :-
	globalVars::incr("pre_cnt",_,Next),
	S=toString(Next),
	sx::frontstr(1,Prefix,F,T),
	Fu=string::toUpperCase(F),
	Var=string::concat(Fu,T,"_",S),
	!.
% ----------------------------------------------------------------------------------------------
% var_atts
% ----------------------------------------------------------------------------------------------
% generates a variable and attribute list:
%
% 1. par 	in	list of variables used for the object's attributes
% 2. par	in	list of the object's attributes
% 3. par	out	combination of 1 and 2 (output variable/attribute is
%                       generated only if the attribute is parameterized) 
%
var_atts([],[],[]) :- 
    !.
var_atts([Var|Vars],[Att|Atts],[AttVar|AttVars]) :-
	Var=var(V),
	AttVar=cmp("VAR",[var(V),Att]),
	var_atts(Vars,Atts,AttVars),
	!.
var_atts([_|Vars],[_|Atts],AttVars) :- 
    var_atts(Vars,Atts,AttVars),
    !.
% -----------------------------------------------------------------------------------------------
% varDE_Var
% -----------------------------------------------------------------------------------------------
% parameters:
% 1	in	DE variable (string)
% 2	out	DE          (string)
%
% attribute-variable format is  E{e|E}*_n+d+
%
varDE_Var(V,DE) :- 
	varComps(V,_,_,DE),
	!.
% -----------------------------------------------------------------------------------------------
% varDE_Vars
% -----------------------------------------------------------------------------------------------
% converts between a list of attribute names and a corresponding list of attribute variables
% 1	prefix string
% 2	DE list - format: fstr(..)
% 3	DE var list - format: var(DE_Var)
%
varDE_Vars(_,[],[]) :- 
	!.
varDE_Vars(Prefix,[fstr(A,_,_,_)|Atts],[var(V)|AttVars]) :-
	V=string::concat(Prefix,A),
	varDE_Vars(Prefix,Atts,AttVars),
	!.

varDE_Vs(_,[],[]) :- 
	!.
varDE_Vs(Prefix,[A|Atts],[V|AttVars]) :-
	V=string::concat(Prefix,A),
	varDE_Vs(Prefix,Atts,AttVars),
	!.
% -----------------------------------------------------------------------------------------------
% varComps
% -----------------------------------------------------------------------------------------------
class predicates
findQualifier : (string_list,string,integer) determ (i,o,o).
clauses

% parameters:
% 1	in	DE var (f.ex. "EComp_12dTicker" or "P3_eIr_12p3_dAs")
% 2	out	entity component
% 3	out	entity no suffix (1-3 digits in SRs, but any number in queries)
% 4	out	DE
% fails if p#1 is not a valid DE var
%
varComps(DV,Ek,No,DE) :-
    sx::sxUnpack(DV,'_',L2),
    L2=[_|T],
    findQualifier(T,Qual,No),
    sx::sxUnpackS(DV,Qual,L3),
    L3=[Ent,DE],
	string::front(Ent,1,First,Te),
	FirstL=string::toLowerCase(First),
	FirstL<>First, % 1st character must be uppercase
	Ek=string::concat(FirstL,Te),
    !.
    
findQualifier([S|L],Qual,N) :-
    sx::getInt(S,N,_),
    Q=toString(N),
    Qual=string::concat("_",Q),
    !
;
    findQualifier(L,Qual,N),
    !.
    
/* 21.2.6 couldn't handle multiple underscores    
varComps(DV,Ek,No,DE) :-
	P=string::searchchar(DV,'_',string::caseSensitive()),
	string::front(DV,P,Ent,T2),
	string::front(T2,1,_,T3),
	sx::getInt(T3,No,DE),
	string::front(Ent,1,First,T4),
	FirstL=string::toLowerCase(First),
	FirstL<>First, % 1st character must be uppercase
	Ek=string::concat(FirstL,T4),
	!.
*/	
varCompsR(DV,A,N,DE) :-
	Ns=toString(N),
	string::frontchar(A,Fc,T),
	F=string::charToString(Fc),
	Fu=string::toUpperCase(F),
	sx::concat([Fu,T,"_",Ns,DE],DV),
	!.
% -------------------------------------------------------------------------------------
% varFuncPar/R
% -------------------------------------------------------------------------------------
% parameters:
% 1	in	function/CR/argument kb-name
% 2	in 	DE kb-name
% 3	out	parameter-variable
%
varFuncPar(Func,DE,Par) :-
	string::frontchar(Func,F,T),
	Fs=string::charToString(F),
	Fu=string::toUpperCase(Fs),
	Par=string::concat(Fu,T,"__",DE),
	!.
varFuncParR(Func,DE,Par) :-
	sx::sxUnpackS(Par,"__",L),
	L=[Fnc,DE],
	string::frontchar(Fnc,F,T),
	Fu=string::charToString(F),
	Fs=string::toLowerCase(Fu),
	Func=string::concat(Fs,T),
	!.
% ----------------------------------------------------------------------------------------------
% varGenDE_Vars;varGenDE_VarsNo
% ----------------------------------------------------------------------------------------------
class predicates
varGenPrefixNo : (string Name,integer N,string Prefix) determ (i,i,o).
clauses
varGenDE_VarsNo(Obj,N,DEs,Vars) :-
	varGenPrefixNo(Obj,N,Pref),
	varDE_Vars(Pref,DEs,Vars),
	!.
varGenPrefixNo(Name,N,Prefix) :-
	Sn=toString(N),
	sx::frontstr(1,Name,F,T),
	Fu=string::toUpperCase(F),
	Prefix=string::concat(Fu,T,"_",Sn),
	!.
	
varGenDE_VarsSNo(Obj,N,DEs,Vars) :-
	varGenPrefixNo(Obj,N,Pref),
	varDE_Vs(Pref,DEs,Vars),
	!.
/*
varGenDE_Vars(Obj,DEs,DE_Vars) :-
	varGenPrefix(Obj,Pref),
	varDE_Vars(Pref,DEs,DE_Vars),
	!
;
	feil("varGenDE_Vars",[Obj],[],cmp("",DEs)),!,fail.	
varGenDE_Vs(Obj,DEs,DE_Vars) :-
	varGenPrefix(Obj,Pref),
	varDE_Vs(Pref,DEs,DE_Vars),
	!
;
	feil("varGenDE_Vs",[Obj|DEs],[],nill),!,fail.	
% generates the entity/predicate prefix.
% parameters:
% 1	in	entity/predicate name 
% 2	out	prefix
%
varGenPrefix(Name,Prefix) :-
	incr(pre_cnt,_,N),
	varGenPrefixNo(Name,N,Prefix),
	!.
*/
end implement vars
