/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement xmlP3
open core,termDomain,globalDomains

constants
className = "com/rudstrand/xmlP3/xmlP3".
classVersion = "1.0".

class predicates
xmlCR_rule : (string,string,sterml,sterm) determ (i,i,i,o).
xmlCR_andRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_combRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_Params : (string,string_list,sterml) determ (i,i,o).
xmlCR_orRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_bayRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_Fuzzy : (string,sterml,sterm) determ (i,i,o).
xmlCR_OWA_FZ : (string,sterml,sterm) determ (i,i,o).
xmlCR_parRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_switchRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_queryRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_questionRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_mapRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_notRule : (string,sterml,sterm) determ (i,i,o).
xmlCR_heading : (string,string,string,string_list,sterm,sterm,sterm,sterm) determ (i,i,i,i,i,i,i,o).
xmlCR_Calls : (string,string,sterml,sterml) determ (i,i,i,o).
xmlCR_Call : (string,string,sterm,sterm) determ (i,i,i,o).
xmlCR_OWA_Sets : (string,string,sterml,sterml) determ (i,i,i,o).
xmlCR_Sets : (string,string,sterml,sterml) determ (i,i,i,o).
xmlCR_Set : (string,string,sterm,sterm) determ (i,i,i,o).
xmlCR_Val : (string,string,sterm,sterm) determ (i,i,i,o).
xmlCR_Context : (string,string,string,sterm,sterm,sterml) determ (i,i,i,i,i,o).
xmlCR_increment : (string,string,sterm,sterm) determ (i,i,i,o).
xmlCR_decrement : (string,string,sterm,sterm) determ (i,i,i,o).
xmlCR_map : (string,string,string,sterm,sterml) determ (i,i,i,i,o).
xmlCR_scCalls : (string,string,string,sterml,sterml) determ (i,i,i,i,o).
xmlToCR_Bays : (string,sterml,sterml) determ (i,i,o).
xmlToCR_Call : (string,sterm,string,sterml) determ (i,i,o,o).
xmlToCR_Calls : (string,sterml,sterml) determ (i,i,o).
xmlToCR_OWA_Sets : (string,sterml,sterml) determ (i,i,o).
xmlToCR_Sets : (string,sterml,sterml) determ (i,i,o).
xmlToCR_Head : (string,sterm,sterm,sterm,sterm,sterm,sterm,sterm) determ (i,i,o,o,o,o,o,o).
xmlToCR_Map : (string,string,string,sterml,sterm) determ (i,i,i,i,o).
xmlToCR_Query : (string,string,sterm,string) determ (i,i,o,o).
xmlToCR_Question : (string,string,sterm) procedure (i,i,o).
xmlToCR_ParCalls : (string,string,sterml,sterml) determ (i,i,i,o).
xmlToCR_Switch : (string,sterml,sterml) determ (i,i,o).
xmlToCR_Val : (string,sterm,sterm) determ (i,i,o).
	
xmlToDE_List : (string,string_list,sterml,sterml) determ (i,i,i,o).
xmlToSentEnt : (string,sterm,sterm) determ (i,i,o).
xmlToSentPred : (string,sterm,string,sterml,integer) determ (i,i,o,o,o).
xmlToVal : (string,string,string,sterm) determ (i,i,i,o).
xmlS_DEs : (string,string_list,sterml,sterml) determ (i,i,i,o).
xmlS_E : (string,sterm,sterm) determ (i,i,o).
xmlS_P : (string,sterm,sterm) determ (i,i,o).
xmlToFunc : (string,sterm,string,sterm,string_list) determ (i,i,i,o,o).
xmlToFuncBodies : (string,string,sterml,sterml,string_list) determ (i,i,i,i,o).
xmlInsFunc : (string,string,string,sterm,string_list) determ (i,i,i,i,i).
xmlTreeFromFunc : (string,string,string,sterm) determ (i,i,i,o).
xmlIdentifier : (string,string,string) determ (i,i,o).
xmlIdentifierR : (string,string,string) determ (i,i,o).
xmlInsSRs : (string,string,integer,string_list).
xmlValue : (string,string,sterm,string) determ (i,i,i,o).

clauses
classInfo(className, classVersion).
% ------------------------------------------------------------------------------------------------------------
% xmlFromCR
% ------------------------------------------------------------------------------------------------------------
% translates CR IF into an XML tree 
% parameters:
% 1	in	PR/NL
% 2	in	realm
% 3	in	cr body
% 4	out	xml text
%
xmlFromCR(Form,Realm,Body,X) :-
	xmlTreeFromCR(Form,Realm,Body,Tree),
	XML=xmlTree::newTreeXML(Tree),
	X=XML:getX(5),
	!
;
	Body=cmp(_,[atom(K)|_]),
	sx::concat(["Translation of inexact rule '",K,"' failed!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail.
% ------------------------------------------------------------------------------------------------------------
% xmlFromFunc
% ------------------------------------------------------------------------------------------------------------
xmlFromFunc(Form,Realm,Func,X) :-	
	xmlTreeFromFunc(Form,Realm,Func,Tree),
	XML=xmlTree::newTreeXML(Tree),
	X=XML:getX(5),
	!
;
	sx::concat(["Translation to XML of function '",Func,"' failed!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail.
% ------------------------------------------------------------------------------------------------------------
% xmlFromSentence
% ------------------------------------------------------------------------------------------------------------
% translates sentence IF into an XML tree 
% parameters:
% 1	in	PR/NL
% 2	in	sentence
% 3	out	xml text
%
xmlFromSentence(Form,Realm,No,Term,X) :-
	xmlTreeFromSentence(Form,Realm,No,Term,Tree),
	XML=xmlTree::newTreeXML(Tree),
	X=XML:getX(5),
	!
;
	dlgIF::feil("xmlFromSentence",[Form],[],Term),
	!,fail.
% ------------------------------------------------------------------------------------------------------------
% xmlFromRule
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlNodeElement : (string,string,string,string,string) procedure (i,i,i,i,o).
xmlTreeElement : (string,string,string,string,string_list,string) procedure (i,i,i,i,i,o).
xmlRules : (string,string,sterml,string) determ (i,i,i,o).
xmlFromRuleX : (string,string,sterm,string) determ (i,i,i,o).
clauses

% translates a rule tree (internal format) into an equivalent indented XML text
% parameters:
% 1	in	indentation string
% 2	in	rule (sub-)tree
% 3	out	xml text
%
xmlFromRule(Realm,Tree,X) :-
	xmlFromRuleX("     ","SentenceRule",Tree,X),
	!
;
	sx::concat(["Translation of sentence rule ",Realm," failed!"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail.

xmlFromRuleX(Ind,Node,Tree,X) :-
	Node="ACOP",
	Tree=cmp(H,[T]),
	opr::opr_cmpass(H),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"Expression",T,X2),
	xmlTreeElement(Ind,Node,H,"",[X2],X),
	!
;
	Node="ACOPs",
	Tree=cmp("&",ACOPs),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"ACOP",ACOPs,X2),
	xmlTreeElement(Ind,Node,"&","",[X2],X),
	!
;
	Node="Argument",
	xmlFromRuleX(Ind,"EP_Argument",Tree,X),
	!
;
	Node="Argument",
	xmlFromRuleX(Ind,"SimpleVariable",Tree,X),
	!
;
	Node="ASOP",
	Tree=cmp(H,[T]),
	opr::oprEq(H),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"Expression",T,X2),
	xmlTreeElement(Ind,Node,H,"",[X2],X),
	!
;
	Node="CertaintyRuleCall",
	Tree=atom(A),
	sx::frontstr(4,A,F,T),
	F="[C] ",
	xmlNodeElement(Ind,Node,A,T,X),
	!
;
	Node="CertaintyRuleCall",
	Tree=cmp(H,T),
	sx::frontstr(4,H,F,K),
	F="[C] ",
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"Parameter",T,X2),
	xmlTreeElement(Ind,Node,H,K,[X2],X),
	!
;
	Node="Conclusion",
	xmlFromRuleX(Ind,"SentenceCondition",Tree,X),
	!
;
	Node="Condition",
	Tree=cmp(H,T),
	set::member(H,["&","v","not","true"]),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"Condition",T,X2),
	xmlTreeElement(Ind,Node,H,"",[X2],X),
	!
;
	Node="Condition",
	Tree=nill,
	xmlNodeElement(Ind,Node,"_","",X),
	!
;
	Node="Condition",
	xmlFromRuleX(Ind,"SentenceCondition",Tree,X),
	!
;
	Node="Condition",
	xmlFromRuleX(Ind,"VariableCondition",Tree,X),
	!
;
	Node="Condition",
	xmlFromRuleX(Ind,"SentenceCountCondition",Tree,X),
	!
;
	Node="D_Component",
	xmlFromRuleX(Ind,"DE_Component",Tree,X),
	!
;
	Node="DE_Component",
	Tree=var(V),
	vars::varComps(V,_,_,D),
	xmlNodeElement(Ind,Node,V,D,X),
	!
;
	Node="DE_Component",
	Tree=cmp(H,[T]),
	vars::varComps(H,_,_,D),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"ACOPs",T,X2),
	xmlTreeElement(Ind,Node,H,D,[X2],X),
	!
;
	Node="DE_Component",
	Tree=cmp(H,[T]),
	vars::varComps(H,_,_,D),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"ACOP",T,X2),
	xmlTreeElement(Ind,Node,H,D,[X2],X),
	!
;
	Node="DE_Variable",
	Tree=var(V),
	vars::varComps(V,_,_,D),
	xmlNodeElement(Ind,Node,V,D,X),
	!
;
	Node="EP_Argument",
	Tree=cmp(H,T),
	sx::sxUnpack(H,'[',L),
	L=[E,_],
	dbBase::smClassDefF(E,"e",_,_,_,_,_),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"D_Component",T,X2),
	xmlTreeElement(Ind,Node,H,E,[X2],X),
	!
;
	Node="EP_Argument",
	Tree=cmp(H,T),
	sx::sxUnpack(H,'[',L),
	L=[E,_],
	dbBase::smClassDefF(E,"p",_,_,_,_,_),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"D_Component",T,X2),
	xmlTreeElement(Ind,Node,H,E,[X2],X),
	!
;
	Node="Expression",
	Tree=cmp(UnOpr,[T]),
	opr::opr_expr(UnOpr,1),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"Expression",T,X2),
	xmlTreeElement(Ind,Node,UnOpr,"",[X2],X),
	!
;
	Node="Expression",
	Tree=cmp(BinOpr,[Ea,Eb]),
	opr::opr_expr(BinOpr,2),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"Expression",Ea,X2),
	xmlFromRuleX(IndM,"Expression",Eb,X3),
	xmlTreeElement(Ind,Node,BinOpr,"",[X2,X3],X),
	!
;
	Node="Expression",
	xmlFromRuleX(Ind,"SimpleExpression",Tree,X),
	!
;
	Node="FunctionCall",
	Tree=atom(A),
	sx::frontstr(4,A,F,T),
	F="[F] ",
	xmlNodeElement(Ind,Node,A,T,X),
	!
;
	Node="FunctionCall",
	Tree=cmp(H,T),
	sx::frontstr(4,H,F,K),
	F="[F] ",
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"Parameter",T,X2),
	xmlTreeElement(Ind,Node,H,K,[X2],X),
	!
;
	Node="List",
	Tree=cmp("[]",ListMembers),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"ListMember",ListMembers,X2),
	xmlTreeElement(Ind,Node,"[]","",[X2],X),
	!
;
	Node="List",
	Tree=cmp("[]",[]),
	xmlNodeElement(Ind,Node,"[]","",X),
	!
;
	Node="List",
	Tree=cmp(ListHead,ListMembers),
	sx::sxGetInBrack(ListHead,'[',']',D),
	dom::domShortLongF(D,_),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"ListMember",ListMembers,X2),
	xmlTreeElement(Ind,Node,ListHead,"",[X2],X),
	!
;
	Node="List",
	Tree=cmp(ListHead,[]),
	sx::sxGetInBrack(ListHead,'[',']',D),
	dom::domShortLongF(D,_),
	xmlNodeElement(Ind,Node,ListHead,"",X),
	!
;
	Node="ListLiteral",
	Tree=cmp("[]",T),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"SimpleLiteral",T,X2),
	xmlTreeElement(Ind,Node,"[]","",[X2],X),
	!
;
	Node="ListLiteral",
	Tree=cmp("[]",[]),
	xmlNodeElement(Ind,Node,"[]","",X),
	!
;
	Node="ListLiteral",
	Tree=cmp(ListHead,T),
	sx::sxGetInBrack(ListHead,'[',']',D),
	dom::domShortLongF(D,_),
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"SimpleLiteral",T,X2),
	xmlTreeElement(Ind,Node,ListHead,"",[X2],X),
	!
;
	Node="ListLiteral",
	Tree=cmp(ListHead,[]),
	sx::sxGetInBrack(ListHead,'[',']',D),
	dom::domShortLongF(D,_),
	xmlNodeElement(Ind,Node,ListHead,"",X),
	!
;
	Node="ListMember",
	xmlFromRuleX(Ind,"SimpleLiteral",Tree,X),
	!
;
	Node="ListMember",
	xmlFromRuleX(Ind,"Variable",Tree,X),
	!
;
	Node="Literal",
	xmlFromRuleX(Ind,"ListLiteral",Tree,X),
	!
;
	Node="Literal",
	xmlFromRuleX(Ind,"SimpleLiteral",Tree,X),
	!
;
	Node="Parameter",
	Tree=cmp(DE,[T]),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"ASOP",T,X2),
	xmlTreeElement(Ind,Node,DE,DE,[X2],X),
	!
;
	Node="ProcedureCall",
	Tree=atom(A),
	sx::frontstr(4,A,F,T),
	F="[P] ",
	xmlNodeElement(Ind,Node,A,T,X),
	!
;
	Node="ProcedureCall",
	Tree=cmp(H,T),
	sx::frontstr(4,H,F,K),
	F="[P] ",
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"Parameter",T,X2),
	xmlTreeElement(Ind,Node,H,K,[X2],X),
	!
;
	Node="SentenceCondition",
	Tree=cmp(H,T),
	sx::sxUnpack(H,'|',L),
	set::size(L,N),
	N>1,
	IndM=string::concat(Ind,"     "),
	xmlRules(IndM,"Argument",T,X2),
	xmlTreeElement(Ind,Node,H,H,[X2],X),
	!
;
	Node="SentenceCountCondition",
	Tree=cmp(H,[cmp(Opr,[cmp("p3_iSSS",[T])])]),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"SentenceCountRightSide_1",cmp(Opr,[cmp("p3_iSSS",[T])]),X2),
	xmlTreeElement(Ind,Node,H,"",[X2],X),
	!
;
	Node="SentenceCountRightSide_1",
	Tree=cmp(H,[T]),
	opr::oprEq(H),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"SentenceCountRightSide_2",T,X2),
	xmlTreeElement(Ind,Node,H,"",[X2],X),
	!
;
	Node="SentenceCountRightSide_2",
	Tree=cmp(H,[T]),
	H="p3_iSSS",
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"SentenceCondition",T,X2),
	xmlTreeElement(Ind,Node,"p3_iSSS","#",[X2],X),
	!
;
	Node="SentenceRule",
	Tree=cmp(H,[Cond,Conc]),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"Condition",Cond,X2),
	xmlFromRuleX(IndM,"Conclusion",Conc,X3),
	xmlTreeElement(Ind,Node,H,"",[X2,X3],X),
	!
;	
	Node="SimpleExpression",
	Tree=nill,
	xmlNodeElement(Ind,Node,"_","",X),
	!
;	
	Node="SimpleExpression",
	xmlFromRuleX(Ind,"Literal",Tree,X),
	!
;	
	Node="SimpleExpression",
	xmlFromRuleX(Ind,"List",Tree,X),
	!
;	
	Node="SimpleExpression",
	xmlFromRuleX(Ind,"Variable",Tree,X),
	!
;	
	Node="SimpleExpression",
	xmlFromRuleX(Ind,"FunctionCall",Tree,X),
	!
;	
	Node="SimpleExpression",
	xmlFromRuleX(Ind,"ProcedureCall",Tree,X),
	!
;	
	Node="SimpleExpression",
	xmlFromRuleX(Ind,"CertaintyRuleCall",Tree,X),
	!
;	
	Node="SimpleLiteral",
	Tree=nill,
	xmlNodeElement(Ind,Node,"_","",X),
	!
;
	Node="SimpleLiteral",
	Tree=int(H),
	S=toString(H),
	xmlNodeElement(Ind,"integer",S,"",X),
	!
;
	Node="SimpleLiteral",
	Tree=fix(H),
	S=toString(H),
	xmlNodeElement(Ind,"number",S,"",X),
	!
;
	Node="SimpleLiteral",
	Tree=atom(H),
	dbBase::nlTry(H,N),
	xmlNodeElement(Ind,"identifier",H,N,X),
	!
;
	Node="SimpleLiteral",
	Tree=str(H),
	xmlNodeElement(Ind,"string",H,"",X),
	!
;
	Node="SimpleVariable",
	Tree=var(H),
	xmlNodeElement(Ind,"variable",H,"",X),
	!
;
	Node="Variable",
	xmlFromRuleX(Ind,"SimpleVariable",Tree,X),
	!
;
	Node="Variable",
	xmlFromRuleX(Ind,"DE_Variable",Tree,X),
	!
;
	Node="Variable",
	xmlFromRuleX(Ind,"MDE_Variable",Tree,X),
	!
;
	Node="VariableCondition",
	Tree=cmp(H,[T]),
	IndM=string::concat(Ind,"     "),
	xmlFromRuleX(IndM,"ACOP",T,X2),
	xmlTreeElement(Ind,Node,H,"",[X2],X),
	!.
xmlRules(_,_,[],"") :-
	!.
xmlRules(Ind,Node,[H|T],X) :-
	xmlFromRuleX(Ind,Node,H,X2),
	xmlRules(Ind,Node,T,X3),
	X=string::concat(X2,X3),
	!.
% ------------------------------------------------------------------------------------------------------------
% scriptWrap
% ------------------------------------------------------------------------------------------------------------
scriptWrap(Format,Xt,Xr) :-
	Version=globalVars::get_i("p3v"), % program version
	Release=globalVars::get_i("p3r"), % program release
	Build=globalVars::get_i("p3b"), % program build
	xmlGeneric::timeNowXML(Time),
	V=toString(Version),
	R=toString(Release),
	B=toString(Build),
	Xr=cmp("pro3script",
	            [cmp("Attributes",
	                    [cmp("version",[str(V)]),
	                     cmp("release",[str(R)]),
	                     cmp("build",[str(B)]),
	                     cmp("createdTime",[str(Time)]),
	                     cmp("author",[str("Pro/3")]),
	                     cmp("format",[str(Format)])])
                |Xt]),
    !.
% ------------------------------------------------------------------------------------------------------------
% xmlKnowledge
% ------------------------------------------------------------------------------------------------------------
% wraps an xml-string into a standard xml-header and the 'pro3Knowledge' tag
%
xmlKnowledge(X,Xr) :-
	xmlKnowA(A),
	xmlKnowB(B),
	sx::concat([A,X,B],Xr),
	!. 
	
xmlKnowA(X) :-
	Version=globalVars::get_i("p3v"), % program version
	Release=globalVars::get_i("p3r"), % program release
	Build=globalVars::get_i("p3b"), % program build
	xmlGeneric::timeNowXML(Time),
	V=toString(Version),
	R=toString(Release),
	B=toString(Build),
	Q=string::charToString('"'),
	sx::concat(["<pro3Knowledge version=",Q,V,Q," release=",Q,R,Q," build=",Q,B,Q," createdTime=",Q,Time,Q," author=",Q,"Pro/3",Q," xmlns=",Q,"http://ligaya.net/xml/",Q," >\n"],X2),
	xmlGeneric::xmlHeader(X2,X),
	!.

xmlKnowB("</pro3Knowledge>\n") :- !.
% ------------------------------------------------------------------------------------------------------------
% xmlProcess
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlProcMess : (integer,integer,integer,integer,string) procedure (i,i,i,i,o).
xmlProcMess2 : (string,integer,string) procedure (i,i,o).
clauses
% loads XML-format knowledge into the KB
% parameters:
% 1	in	default realm name
% 2	in	xml text
%
xmlProcess(Realm,X) :-
	XML=xmlTree::newTree(X),
	Tree=XML:get(),
			tt::tt_term(3,"XML Trees",Tree),
	Tree=cmp(_,[_|Trees]),
	xmlProcessTrees(Realm,Trees,Nsn,Nsr,Ncr,Nfn),
	xmlProcMess(Nsn,Nsr,Ncr,Nfn,Mess),
	dlgIF::dlgNote(0,Mess),
	!
;
	dlgIF::dlgNote(0,"XML loading were NOT completed!"),
	!.
	
xmlProcMess(Nsn,Nsr,Ncr,Nfn,Mess) :-
	xmlProcMess2("sentence(s)",Nsn,S2),
	xmlProcMess2("sentence rule(s)",Nsr,S3),
	xmlProcMess2("inexct rule(s)",Ncr,S5),
	xmlProcMess2("function(s)",Nfn,S6),
	dbBase::joinNL([S2,S3,S5,S6],S7),
	sx::concat([S7," loaded from XML!"],Mess),
	!.
xmlProcMess2(_,0,"") :-
	!.
xmlProcMess2(S,N,T) :-
	S2=toString(N),
	sx::concat([S2," ",S],T),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlSentenceRule
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	sentence rule type number 
% 2	in	realm
% 3	in 	realm no
% 4	in	realm sub no
% 5	in	rule-tree 
% 6	out	xml (node sentenceRule)
%
xmlSentenceRule(RuleType,Realm,No,SubNo,Rule,X) :-
	xmlFromRule(Realm,Rule,X2),
	Q=string::charToString('"'),
	sx::concat(["<sentenceRule ruleType=",Q,RuleType,Q," realm=",Q,Realm,Q," realmNo=",Q,No,Q," realmSubNo=",Q,SubNo,Q," >\n",X2,"</sentenceRule>\n"],X),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlToCR
% ------------------------------------------------------------------------------------------------------------
% parameters
% 1	in	xml-tree corresponding to a CR
% 2	out	CR as a term
%
xmlToCR(X,CR,Def,Realm) :-
	X=cmp(_,[cmp("Attributes",[cmp("format",[str(Form)]),cmp("realm",[str(Realm)])]),Body]),
	xmlToCR_Rule(Form,Body,CR,Def),
	!
;
	dlgIF::feil("xmlToRule",["Unrecognized certainty rule XML tree structure"],[],X),!,fail.
% ------------------------------------------------------------------------------------------------------------
% xmlToRule
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlLabelAdjust : (string,string) procedure (i,o).
xmlMakeNode : (string,string,sterm) determ (i,i,o).	
xmlToRuleX : (sterml,sterml) determ (i,o).
clauses

% parameters:
% 1	in	xml-tree corresponding to a sentence rule
% 2	out	rule-tree (IF)
%
xmlToRule(X,Rule) :- 
	X=cmp("tree",[cmp("Attributes",Atts)|Xs]),
	xmlGeneric::xmlGetAttr("label",Atts,Label),
	xmlLabelAdjust(Label,LabelM),
	xmlToRuleX(Xs,RTs),
	Rule=cmp(LabelM,RTs),
	!
;
	X=cmp("node",[cmp("Attributes",Atts)]),
	xmlGeneric::xmlGetAttr("class",Atts,Clas),
	xmlGeneric::xmlGetAttr("label",Atts,Label),
	xmlLabelAdjust(Label,LabelM),
	xmlMakeNode(Clas,LabelM,Rule),
	!
;
	dlgIF::feil("xmlToRule",["Unrecognized sentence rule XML tree structure"],[],X),!,fail.

xmlToRuleX([],[]) :-
	!.
xmlToRuleX([H|T],[Hm|Tm]) :-
	xmlToRule(H,Hm),
	xmlToRuleX(T,Tm),
	!.

xmlMakeNode(_,"_",nill) :- 
	!.
xmlMakeNode("integer",S,int(V)) :- 
	V=sx::toInt(S),
	!.
xmlMakeNode("number",S,fix(V)) :- 
	V=sx::toReal(S),
	!.
xmlMakeNode("variable",S,var(S)) :- 
	!.
xmlMakeNode("identifier",S,atom(S)) :- 
	!.
xmlMakeNode("string",S,str(S)) :- 
	!.
xmlMakeNode("DE_Component",S,var(S)) :- 
	!.
xmlMakeNode(Clas,S,_) :-
	dlgIF::feil("xmlMakeNode",["Unrecognized rule tree leaf class",Clas,S],[],nill),!,fail.

xmlLabelAdjust(L,Lm) :-
	opr::oprDomF(_,Opr,L),
	Lm=Opr,
	!
;
	Lm=L,
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlToSentence
% ------------------------------------------------------------------------------------------------------------
% parameters
% 1	in	xml-tree corresponding to a sentence
% 2	out	sentence as term
% 2	out	realm
% 2	out	realm no
%
xmlToSentence(X,T,Realm,No) :-
	X=cmp(_,[cmp("Attributes",[cmp("format",[str(Form)]),
	                           cmp("realm",[str(Realm)]),
	                           cmp("realmNo",[str(SNo)])]),
	         P,Ea,Eb]),
	No=sx::toInt(SNo),
	xmlToSentPred(Form,P,Pk,DEs,Card),
	Card=2,
	xmlToSentEnt(Form,Ea,Eak),
	xmlToSentEnt(Form,Eb,Ebk),
	T=cmp(Pk,[Eak,Ebk|DEs]),
	!
;
	X=cmp(_,[cmp("Attributes",[cmp("format",[str(Form)])|_]),P,_,_]),
	xmlToSentPred(Form,P,Pk,_,Card),
	Card=1,
	sx::concat(["Binary predicate '",Pk,"' appears as unary in XML input"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail
;
	X=cmp(_,[cmp("Attributes",[cmp("format",[str(Form)]),
	                           cmp("realm",[str(Realm)]),
	                           cmp("realmNo",[str(SNo)])]),
	         P,Ea]),
	No=sx::toInt(SNo),
	xmlToSentPred(Form,P,Pk,DEs,Card),
	Card=1,
	xmlToSentEnt(Form,Ea,Eak),
	T=cmp(Pk,[Eak|DEs]),
	!
;
	X=cmp(_,[cmp("Attributes",[cmp("format",[str(Form)])|_]),P,_]),
	xmlToSentPred(Form,P,Pk,_,Card),
	Card=2,
	sx::concat(["Unary predicate '",Pk,"' appears as binary in XML input"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail
;
	dlgIF::feil("xmlToSentence",["Unrecognized sentence XML tree structure"],[],X),!,fail.
% ------------------------------------------------------------------------------------------------------------
% xmlTreeFromCR
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	PR/NL
% 2	in	realm
% 3	in	CR term
% 4	out	xml-tree
%
xmlTreeFromCR(Form,Realm,Body,Tree) :-
	Body=cmp(Typ,Content),
	xmlCR_rule(Form,Typ,Content,Tree2),
	Tree=cmp("inexactRule",[
	                         cmp("Attributes",[
	                                          cmp("format",[str(Form)]),
	                                          cmp("realm",[str(Realm)])]),
	                         Tree2]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlTreeFromFunc
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlTreeFromFuncPars : (string,sterml,sterml) determ (i,i,o).
xmlTreeFromFuncBodies : (string,string_list_list,sterml) determ (i,i,o).
xmlTreeFromStats : (string_list,sterml) determ (i,o).	                  
clauses
% parameters:
% 1	in	PR/NL
% 2	in	realm
% 3	in	function
% 4	out	xml-tree
%
xmlTreeFromFunc(Form,Realm,Func,Tree) :-
    dbBase::smDefTree(Func,Clas,Def),
    Segment=dbBase::smSegment(Func),
	Def=cmp(_,[_,
	           cmp("DOM",[atom(Dom)]),
	           cmp("FPAR",DEs)]),
	xmlIdentifier(Form,Func,F),
	set::setGet(["f","g"],Ix,Clas),
	set::setGet(["DETERMINISTIC","NONDETERMINISTIC"],Ix,Det),
	xmlIdentifier(Form,Dom,D),
	xmlTreeFromFuncPars(Form,DEs,Tp),
	dbExtension::functionStatements(Form,Func,BodiesNL), % each body is a list of statement strings (only one string if PR-format)
	xmlTreeFromFuncBodies(Form,BodiesNL,Tb),
	set::conc([Tp,Tb],T),
	Tree=cmp("function",[cmp("Attributes",[cmp("format",[str(Form)]),
	                                       cmp("realm",[str(Realm)]),
	                                       cmp("segment",[str(Segment)])]),
	                     cmp("functionName",[nill,str(F)]),
	                     cmp("determinism",[nill,str(Det)]),
	                     cmp("domain",[nill,str(D)])
	                     |T]),
	!.	                     
xmlTreeFromFuncPars(_,[],[]) :-
	!.
xmlTreeFromFuncPars(Form,[fstr(DE,_,_,_)|DEs],[Tree|Trees]) :-
	xmlIdentifier(Form,DE,D),
	Tree=cmp("parameter",[nill,str(D)]),
	xmlTreeFromFuncPars(Form,DEs,Trees),
	!.
	    
xmlTreeFromFuncBodies(_,[],[]) :-
	!.
xmlTreeFromFuncBodies(Form,[B|Bs],[Tree|Trees]) :-
	xmlTreeFromStats(B,Statements),
	Tree=cmp("functionBody",[nill|Statements]),
	xmlTreeFromFuncBodies(Form,Bs,Trees),
	!.
xmlTreeFromStats([],[]) :-
	!.
xmlTreeFromStats([Stat|Stats],[S|L]) :-
	xmlGeneric::xmlAdjustString(Stat,X),
	S=cmp("statement",[nill,str(X)]),
	xmlTreeFromStats(Stats,L),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlTreeFromSentence
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	NL/PR
% 2	in 	realm
% 3	in	realm no
% 4	in	sentence-term
% 5	out	xml-tree
%
xmlTreeFromSentence(Form,Realm,No,Term,Tree) :-
	SNo=toString(No),
	Term=cmp(P,[Ea,Eb|DEs]),
	Ea=cmp(_,_),
	Eb=cmp(_,_),	
	xmlS_P(Form,cmp(P,DEs),Pt),
	xmlS_E(Form,Ea,Eat),
	xmlS_E(Form,Eb,Ebt),
	Tree=cmp("sentence",[
	                         cmp("Attributes",[
	                                          cmp("format",[str(Form)]),
	                                          cmp("realm",[str(Realm)]),
	                                          cmp("realmNo",[str(SNo)])]),
	                         Pt,Eat,Ebt]),
	!
;
 	SNo=toString(No),
	Term=cmp(P,[Ea|DEs]),
	Ea=cmp(_,_),
	xmlS_P(Form,cmp(P,DEs),Pt),
	xmlS_E(Form,Ea,Eat),
	Tree=cmp("sentence",[
	                         cmp("Attributes",[
	                                          cmp("format",[str(Form)]),
	                                          cmp("realm",[str(Realm)]),
	                                          cmp("realmNo",[str(SNo)])]),
	                         Pt,Eat]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_andRule
% ------------------------------------------------------------------------------------------------------------
xmlCR_andRule(Form,Body,Tree) :-
	Body=[atom(Rule),cmp("p3_iLand",Calls),Priand,Out,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,RDom,_,_),
	xmlCR_heading(Form,Rule,RDom,Params,Priand,Out,Context,T2),
	xmlCR_Calls(Form,Rule,Calls,T3),
	Tree=cmp("andRule",[nill,T2|T3]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_combRule
% ------------------------------------------------------------------------------------------------------------
xmlCR_combRule(Form,Body,Tree) :-
	Body=[atom(Rule),cmp("p3_iComb",Calls),Priand,Out,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,RDom,_,_),
	xmlCR_heading(Form,Rule,RDom,Params,Priand,Out,Context,T2),
	xmlCR_Calls(Form,Rule,Calls,T3),
	Tree=cmp("combinationRule",[nill,T2|T3]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_bayRule
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlCR_BayCalls : (string,string,sterml,sterml) determ (i,i,i,o).
clauses

xmlCR_bayRule(Form,Body,Tree) :-
	Body=[atom(Rule),cmp("p3_iBay",BayCalls),Prior,Out,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,RDom,_,_),
	xmlCR_heading(Form,Rule,RDom,Params,Prior,Out,Context,T2),
	xmlCR_BayCalls(Form,Rule,BayCalls,T3),
	Tree=cmp("bayRule",[nill,
	                    T2
	                    |T3]),
	!.

xmlCR_BayCalls(_,_,[],[]) :-
	!.
xmlCR_BayCalls(Form,CallingRule,[C|Calls],BayCallsT) :-
   	C=cmp(Rule,[Incr,Decr,ParPart]),
   	xmlCR_increment(Form,Rule,Incr,IncT),
   	xmlCR_decrement(Form,Rule,Decr,DecT),
	xmlCR_Call(Form,CallingRule,cmp(Rule,[ParPart]),CallT),
	BayCall=cmp("bayCall",[nill,
	                       IncT,
	                       DecT,
	                       CallT]),
	xmlCR_BayCalls(Form,CallingRule,Calls,CallsT2),
	set::conc([[BayCall],CallsT2],BayCallsT),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_Calls
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlCR_APars : (string,string,sterml,sterml) determ (i,i,i,o).
xmlCR_OWA : (string,string,sterm,sterm) determ (i,i,i,o).
xmlCR_OWA_Set : (string,string,sterm,sterm) determ (i,i,i,o).	
clauses

xmlCR_Call(Form,CallingRule,cmp(Rule,[cmp("p3_iAps",APars)]),Tree) :-
	xmlIdentifier(Form,Rule,RuleX),
	xmlCR_APars(Form,CallingRule,APars,AparTs),
	Tree=cmp("call",[nill,
	                 cmp("rule",[nill,str(RuleX)])
	                 |AParTs]),
	!.
xmlCR_Set(Form,CallingRule,cmp(Rule,[cmp("p3_iAps",APars)]),Tree) :-
	xmlIdentifier(Form,Rule,RuleX),
	xmlCR_APars(Form,CallingRule,APars,AparTs),
	Tree=cmp("combine",[nill,
	                 cmp("set",[nill,str(RuleX)])
	                 |AParTs]),
	!.

xmlCR_OWA_Set(Form,CallingRule,cmp(Rule,[Weight,cmp("p3_iAps",APars)]),Tree) :-
	xmlIdentifier(Form,Rule,RuleX),
	xmlCR_OWA(Form,CallingRule,Weight,WeightX),
	xmlCR_APars(Form,CallingRule,APars,AparTs),
	Tree=cmp("OWA",[nill,
	                WeightX,
	                cmp("set",[nill,str(RuleX)])
	               |AParTs]),
	!.
	
xmlCR_Sets(_,_,[],[]) :-
	!.
xmlCR_Sets(Form,Rule,[Cal|Calls],[Tree|Trees]) :-
	xmlCR_Set(Form,Rule,Cal,Tree),
	xmlCR_Sets(Form,Rule,Calls,Trees),
	!.
	
xmlCR_OWA_Sets(_,_,[],[]) :-
	!.
xmlCR_OWA_Sets(Form,Rule,[Cal|Calls],[Tree|Trees]) :-
	xmlCR_OWA_Set(Form,Rule,Cal,Tree),
	xmlCR_OWA_Sets(Form,Rule,Calls,Trees),
	!.
	
xmlCR_Calls(_,_,[],[]) :-
	!.
xmlCR_Calls(Form,Rule,[Cal|Calls],[Tree|Trees]) :-
	xmlCR_Call(Form,Rule,Cal,Tree),
	xmlCR_Calls(Form,Rule,Calls,Trees),
	!.
	
xmlCR_APars(_,_,[],[]) :-
	!.
xmlCR_APars(Form,Rule,[cmp(_,[atom(D),Val])|APars],[Ta,Tb|Trees]) :-
	dbBase::smClassDefF(D,"a",_,_,Dom,_,_),
	xmlIdentifier(Form,D,Dx),
	Ta=cmp("parameter",[nill,str(Dx)]),
	xmlCR_Val(Form,Dom,Val,Tb),
	xmlCR_APars(Form,Rule,APars,Trees),
	!.
xmlCR_OWA(Form,CallingRule,Weight,WeightX) :-
   	realType(Weight,W),
	Wx=toString(W),
	WeightX=cmp("weight",[nill,str(Wx)]),
	!
;
	xmlCR_Call(Form,CallingRule,Weight,WeightX),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_Context
% ------------------------------------------------------------------------------------------------------------
xmlCR_Context(_,_,_,_,nill,[]) :-
	!.
xmlCR_Context(Form,Rule,Dom,Out,cmp("p3_iCon",ConCalls),[Tree]) :-
	xmlCR_Val(Form,Dom,Out,OutX),
	xmlCR_scCalls(Form,Rule,Dom,ConCalls,ConCallsT),
	Tree=cmp("context",[nill,
	                    cmp("outOfContext",[nill,OutX])
	                    |ConCallsT]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_Fuzzy
% ------------------------------------------------------------------------------------------------------------
xmlCR_Fuzzy(Form,Body,Tree) :-
	Body=[R,cmp(Fuzzy,Calls)],
	R=atom(Rule),
	dbBase::smClassDefF(Rule,"h",_,Params,_,_,_),
	xmlIdentifier(Form,Rule,RuleX),
	T2=cmp("set",[nill,str(RuleX)]),
	xmlCR_Params(Form,Params,T3),
	xmlIdentifier(Form,Fuzzy,FuzzyX),
	T4=cmp("operation",[nill,str(FuzzyX)]),
	xmlCR_Sets(Form,Rule,Calls,T5),
	set::conc([[T2,T4],T3,T5],T),
	Tree=cmp("fuzzySet",[nill|T]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_OWA
% ------------------------------------------------------------------------------------------------------------
xmlCR_OWA_FZ(Form,Body,Tree) :-
	Body=[R,cmp(_,Calls)],
	R=atom(Rule),
	dbBase::smClassDefF(Rule,"h",_,Params,_,_,_),
	xmlIdentifier(Form,Rule,RuleX),
	T2=cmp("set",[nill,str(RuleX)]),
	xmlCR_Params(Form,Params,T3),
	xmlCR_OWA_Sets(Form,Rule,Calls,T5),
	set::conc([[T2],T3,T5],T),
	Tree=cmp("fuzzyOWA_Set",[nill|T]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_heading
% ------------------------------------------------------------------------------------------------------------
xmlCR_heading(Form,Rule,Dom,Params,Pri,Out,Context,Tree) :-
   	realType(Pri,Prior),
	xmlIdentifier(Form,Rule,RuleX),
	RuleT=cmp("rule",[nill,str(RuleX)]),
	xmlIdentifier(Form,Dom,DomX),
	RuleD=cmp("returnDomain",[nill,str(DomX)]),
	xmlCR_Params(Form,Params,ParamsT),
	PriorX=toString(Prior),
	RuleP=cmp("prior",[nill,str(PriorX)]),
	xmlCR_Context(Form,Rule,Dom,Out,Context,ContextT),
	set::conc([[RuleT,RuleD],ParamsT,[RuleP],ContextT],SubTrees),
	Tree=cmp("heading",[nill|SubTrees]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_Inc/Dec
% ------------------------------------------------------------------------------------------------------------
xmlCR_increment(Form,Rule,Inc,Tree) :-
   	realType(Inc,F),
	Fx=toString(F),
	Tree=cmp("increment",[nill,cmp("inc",[nill,str(Fx)])]),
	!
;
	xmlCR_Call(Form,Rule,Inc,Tree2),
	Tree=cmp("increment",[nill,Tree2]),
	!.
xmlCR_decrement(Form,Rule,Dec,Tree) :-
   	realType(Dec,F),
	Fx=toString(F),
	Tree=cmp("decrement",[nill,cmp("dec",[nill,str(Fx)])]),
	!
;
	xmlCR_Call(Form,Rule,Dec,Tree2),
	Tree=cmp("decrement",[nill,Tree2]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_map
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlCR_Mappings : (string,string,string,sterml,sterml,sterml) determ (i,i,i,i,i,o).
clauses

xmlCR_map(_,_,_,nill,[]) :-
	!.
xmlCR_map(Form,DomF,DomT,cmp(MapType,[cmp(_,Froms),cmp(_,Tos)]),[Tree]) :-
	MapType="p3_iNmap",
	xmlCR_Mappings(Form,DomF,DomT,Froms,Tos,Ts),
	Tree=cmp("map",[nill,
	                cmp("mapClass",[nill,str("NUMERIC")])
                        |Ts]),
        !
;      
	MapType="p3_iDmap",
	xmlCR_Mappings(Form,DomF,DomT,Froms,Tos,Ts),
	Tree=cmp("map",[nill,
	                cmp("mapClass",[nill,str("DISCRETE")])
                        |Ts]),
        !.
	
xmlCR_Mappings(_,_,_,[],_,[]) :-
	!.
xmlCR_Mappings(Form,DomF,DomT,[F|Fs],[T|Ts],[Tf,Tt|Trees]) :-
   	xmlCR_Val(Form,DomF,F,Tf),
   	xmlCR_Val(Form,DomT,T,Tt),
	xmlCR_Mappings(Form,DomF,DomT,Fs,Ts,Trees),
	!.	               
% ------------------------------------------------------------------------------------------------------------
% xmlCR_mapRule
% ------------------------------------------------------------------------------------------------------------
xmlCR_mapRule(Form,Body,Tree) :-
	Body=[atom(Rule),Cal,Map,Prior,Out,Def,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,Dom,_,_),
	dbExtension::dbuCR_GetTD(Rule,_,FromDom),
	xmlCR_heading(Form,Rule,Dom,Params,Prior,Out,Context,T2),
	xmlCR_Val(Form,Dom,Def,DefVal),
	xmlCR_Call(Form,Rule,Cal,T3),
	xmlCR_map(Form,FromDom,Dom,Map,T4),
	Tree=cmp("mapRule",[nill,
	                       T2,
	                       cmp("default",[nill,DefVal]),
	                       T3
	                       |T4]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_notRule
% ------------------------------------------------------------------------------------------------------------
xmlCR_notRule(Form,Body,Tree) :-
	Body=[atom(Rule),Cal,Prior,Out,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,Dom,_,_),
	xmlCR_heading(Form,Rule,Dom,Params,Prior,Out,Context,T2),
	xmlCR_Call(Form,Rule,Cal,T3),
	Tree=cmp("notRule",[nill,T2,T3]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_orRule
% ------------------------------------------------------------------------------------------------------------
xmlCR_orRule(Form,Body,Tree) :-
	Body=[atom(Rule),cmp("p3_iLor",Calls),Prior,Out,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,Dom,_,_),
	xmlCR_heading(Form,Rule,Dom,Params,Prior,Out,Context,T2),
	xmlCR_Calls(Form,Rule,Calls,T3),
	Tree=cmp("orRule",[nill,T2|T3]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_Params
% ------------------------------------------------------------------------------------------------------------
xmlCR_Params(_,[],[]) :-
	!.
xmlCR_Params(Form,[D|Params],ParamsT) :-
	xmlIdentifier(Form,D,Dx),
	xmlCR_Params(Form,Params,ParamsT2),
	ParamsT=[cmp("parameter",[nill,str(Dx)])|ParamsT2],
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_queryRule
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlCR_Query : (string,sterm,string) determ (i,i,o).
clauses

xmlCR_queryRule(Form,Body,Tree) :-
	Body=[atom(Rule),cmp(_,[atom(QDom),Query]),Map,Prior,Out,Def,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,Dom,_,_),
	xmlCR_heading(Form,Rule,Dom,Params,Prior,Out,Context,T2),
	xmlCR_Val(Form,Dom,Def,DefX),
	xmlCR_Query(Form,Query,QueryX),
	xmlCR_map(Form,QDom,Dom,Map,T3),
	set::conc([[nill,T2,cmp("default",[nill,DefX]),cmp("query",[nill,str(QueryX)])],T3],Tail),
	Tree=cmp("queryRule",Tail),
	!.
	
xmlCR_Query(Form,Q,Qx) :-
	Form="NL",
	nl::conQuery(Q,Q2),
	xmlGeneric::xmlAdjustString(Q2,Qx),
	!
;
	Form="PR",
    nl::cprTerm(Q,Qx),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_questionRule
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlCR_Question : (string,string,string) procedure (i,i,o).
clauses
xmlCR_questionRule(Form,Body,Tree) :-
	Body=[atom(Rule),cmp(_,[atom(QDom),str(Q)]),Map,Prior,Out,Def,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,RDom,_,_),
	xmlCR_heading(Form,Rule,RDom,Params,Prior,Out,Context,T2),
	xmlIdentifier(Form,QDom,QDomX),
	xmlCR_Val(Form,RDom,Def,DefX),
	xmlCR_map(Form,QDom,RDom,Map,T3),
	xmlCR_Question(Form,Q,Qx),
	set::conc([[nill,T2,cmp("default",[nill,DefX]),cmp("questionDomain",[nill,str(QDomX)]),cmp("question",[nill,str(Qx)])],T3],Tail),
	Tree=cmp("questionRule",Tail),
	!.
	
xmlCR_Question(Form,Q,Qx) :-
	Form="NL",
	irBase::wcrQuestionToNL(Q,Q2),
	xmlGeneric::xmlAdjustString(Q2,Qx),
	!
;
	xmlGeneric::xmlAdjustString(Q,Qx),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_rule
% ------------------------------------------------------------------------------------------------------------
xmlCR_rule(Form,Fuzzy,Body,Tree) :-
	set::member(Fuzzy,["p3_iZnot","p3_iZand","p3_iZor"]),
	xmlCR_Fuzzy(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iZowa",Body,Tree) :-
	xmlCR_OWA_FZ(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCbay",Body,Tree) :-
	xmlCR_bayRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCand",Body,Tree) :-
	xmlCR_andRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCor",Body,Tree) :-
	xmlCR_orRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCnot",Body,Tree) :-
	xmlCR_notRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCand",Body,Tree) :-
	xmlCR_andRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCcmb",Body,Tree) :-
	xmlCR_combRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCswi",Body,Tree) :-
	xmlCR_switchRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCpar",Body,Tree) :-
	xmlCR_parRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCmap",Body,Tree) :-
	xmlCR_mapRule(Form,Body,Tree),
	!.
xmlCR_rule(Form,"p3_iCdat",Body,Tree) :-
	Body=[atom(K)|_],
	dbBase::irType(K,_,Typ),
	Typ="p3_iCqst",
	xmlCR_questionRule(Form,Body,Tree),
	!
;
	Body=[atom(K)|_],
	dbBase::irType(K,_,Typ),
	Typ="p3_iCqry",
	xmlCR_queryRule(Form,Body,Tree),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_parRule
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlCR_ParCalls : (string,string,sterml,sterml) determ (i,i,i,o).
clauses

xmlCR_parRule(Form,Body,Tree) :-
	Body=[atom(Rule),Cal,cmp(_,ParCalls),Prior,Out,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,RDom,_,_),
	xmlCR_heading(Form,Rule,RDom,Params,Prior,Out,Context,T2),
	xmlCR_Call(Form,Rule,Cal,T3),
	T3=cmp(_,CallBodyT),
	xmlCR_ParCalls(Form,Rule,ParCalls,T4),
	Tree=cmp("parameterRule",[nill,
	                          T2,
	                          cmp("ruleCall",CallBodyT)
	                         |T4]),
	!.

xmlCR_ParCalls(_,_,[],[]) :-
	!.
xmlCR_ParCalls(Form,CallingRule,[C|Cs],[VT,CallT|Calls]) :-
   	C=cmp(Rule,[var(V),ParPart]),
	xmlGeneric::xmlAdjustString(V,Vx),
   	VT=cmp("varVal",[nill,str(Vx)]),
	xmlCR_Call(Form,CallingRule,cmp(Rule,[ParPart]),CallT),
	xmlCR_ParCalls(Form,CallingRule,Cs,Calls),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_switchRule
% ------------------------------------------------------------------------------------------------------------
xmlCR_switchRule(Form,Body,Tree) :-
	Body=[atom(Rule),atom(DE),cmp("p3_iSwi",SwitchCalls),Prior,Out,Def,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,Dom,_,_),
	!,xmlCR_heading(Form,Rule,Dom,Params,Prior,Out,Context,T2),
	%dbBase::smClassDefF(DE,"a",_,_,Dom,_,_), seems unnecessary
	xmlIdentifier(Form,DE,Dx),
	xmlCR_Val(Form,Dom,Def,DefX),
	xmlCR_scCalls(Form,Rule,Dom,SwitchCalls,T4),
	Tree=cmp("switchRule",[nill,
	                       T2,
	                       cmp("default",[nill,DefX]),
	                       cmp("criterionParameter",[nill,str(Dx)])
	                       |T4]),
	!
;
	Body=[atom(Rule),Cal,cmp("p3_iSwi",SwitchCalls),Prior,Out,Def,Context],
	dbBase::smClassDefF(Rule,"h",_,Params,Dom,_,_),
	xmlCR_heading(Form,Rule,Dom,Params,Prior,Out,Context,T2),
	xmlCR_Call(Form,Rule,Cal,T3),
	Cal=cmp(K,_),
	dbBase::smClassDefF(K,"h",_,_,KDom,_,_),
	xmlCR_Val(Form,KDom,Def,DefX),
	T3=cmp(_,CallBodyT),
	xmlCR_scCalls(Form,Rule,KDom,SwitchCalls,T4),
	Tree=cmp("switchRule",[nill,
	                       T2,
	                       cmp("default",[nill,DefX]),
	                       cmp("criterionCall",CallBodyT)
	                       |T4]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_Val
% ------------------------------------------------------------------------------------------------------------
% parameters:
% 1	in	NL/PR
% 2	in	value domain
% 3	in	value (fstr/fint/ffix ...)
% 4	out	xml value-element
%
xmlCR_Val(_,_,str(V),Vx) :-
	xmlGeneric::xmlAdjustString(V,Va),
	Vx=cmp("strVal",[nill,str(Va)]),
	!.
xmlCR_Val(Form,_,atom(V),Vx) :-
	Form="NL",
	dbBase::nlSing(V,Vn),
	xmlGeneric::xmlAdjustString(Vn,Va),
	Vx=cmp("idVal",[nill,str(Va)]),
	!
;
	xmlGeneric::xmlAdjustString(V,Va),
	Vx=cmp("idVal",[nill,str(Va)]),
	!.
xmlCR_Val(_,_,fix(V),Vx) :-
	Vs=toString(V),
	Vx=cmp("fixVal",[nill,str(Vs)]),
	!.
xmlCR_Val(_,_,int(V),Vx) :-
	Vs=toString(V),
	Vx=cmp("intVal",[nill,str(Vs)]),
	!.
xmlCR_Val(Form,_,var(V),Vx) :-
	Form="NL",
	vars::varFuncParR(CRk,DEk,V),
	CRk="p3_iThis",
	dbBase::nlSing(CRk,CR),
	dbBase::nlSing(DEk,DE),
	sx::concat([CR," ",DE],Var),	
	xmlGeneric::xmlAdjustString(Var,VarA),
	Vx=cmp("varVal",[nill,str(VarA)]),
	!
;
	Form="NL",
	vars::varFuncParR(CRk,DEk,V),
	dbBase::nlSing(CRk,CR),
	dbBase::nlSing(DEk,DE),
	%dbBase::nlSing("p3_iGeni",Geni),
	%sx::concat([CR,Geni," ",DE],Var),	
	sx::concat([CR,":",DE],Var),	
	xmlGeneric::xmlAdjustString(Var,VarA),
	Vx=cmp("varVal",[nill,str(VarA)]),
	!
;
	xmlGeneric::xmlAdjustString(V,Va),
	Vx=cmp("varVal",[nill,str(Va)]),
	!.
xmlCR_Val(_,Dom,V,_) :-
	dlgIF::feil("xmlCR_Val",["Unrecognized actual parameter value",Dom],[],V),!,fail.
% ------------------------------------------------------------------------------------------------------------
% xmlIdentifier/R
% ------------------------------------------------------------------------------------------------------------
xmlIdentifier("PR",Sk,Sx) :-
	xmlGeneric::xmlAdjustString(Sk,Sx),
	!.
xmlIdentifier("NL",Sk,Sx) :-
	dbBase::nlTry(Sk,Sn),
	xmlGeneric::xmlAdjustString(Sn,Sx),
	!.
	
xmlIdentifierR("PR",X,S) :-
	xmlGeneric::xmlReadjustString(X,S),
	!.
xmlIdentifierR("NL",X,S) :-
	xmlGeneric::xmlReadjustString(X,S2),
	dbBase::kbID(S2,S,_,_),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlInsFunc
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlInsFuncBodies : (string,string,string_list) determ (i,i,i).
clauses

xmlInsFunc(Form,Realm,Func,FuncDef,FuncBodies) :-
	dbBase::smDefF(Func,Clas,_,_,_,_,_,Loc),
	set::member(Clas,["f","g"]),
	dbBase::recordDelete(Loc),
	dbBase::recordInsert(2,Realm,0,0,FuncDef,atom("true")),
	dbBase::chainDelete(Func),
	xmlInsFuncBodies(Form,Realm,FuncBodies),
	!
;
	dbBase::smDefF(Func,Clas,_,_,_,_,_),
	not(set::member(Clas,["f","g"])),
	dbBase::nlTry(Clas,S2),
	sx::concat(["Function name '",Func,"' is already used for '",S2,"' sentence model object and cannot be stored in the KB."],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail
;
	dbBase::recordInsert(2,Realm,0,0,FuncDef,atom("true")),
	dbBase::chainDelete(Func),
	xmlInsFuncBodies(Form,Realm,FuncBodies),
	!.
	
xmlInsFuncBodies(_,_,[]) :-
	!.
xmlInsFuncBodies(Form,Realm,[Body|Bodies]) :-
	Form="NL",
	transnl::parTranslate(Body,Term,_),
	Term=cmp(_,[LT,RT]),
	dbBase::recordInsert(4,Realm,0,0,LT,RT),
	xmlInsFuncBodies(Form,Realm,Bodies),
	!
;
	Form="PR",
	transpr::proPR_Term(Body,Term),
	Term=cmp(_,[LT,RT]),
	dbBase::recordInsert(4,Realm,0,0,LT,RT),
	xmlInsFuncBodies(Form,Realm,Bodies),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlInsSRs
% ------------------------------------------------------------------------------------------------------------
xmlInsSRs(_,_,_,[]) :-
	!.
xmlInsSRs(Realm,SNo,N,[PR|PRs]) :-
	N=0,
	No=sx::toInt(SNo),
	transpr::proPR_Term(PR,Term),
	dbExtension::dbaCheckRuleReplace(Realm,No),
	Sign=globalVars::get_i("rulsta"),
	RecType=Sign*3,
	Term=cmp(_,[LT,RT]),
	dbBase::recordInsert(RecType,Realm,No,N,LT,RT),
	M=N+1,
	xmlInsSRs(Realm,SNo,M,PRs),
	!
;
	N>0,
	No=sx::toInt(SNo),
	transpr::proPR_Term(PR,Term),
	dbExtension::dbaCheckSubRuleReplace(Realm,No,N),
	Sign=globalVars::get_i("rulsta"),
	RecType=Sign*3,
	Term=cmp(_,[LT,RT]),
	dbBase::recordInsert(RecType,Realm,No,N,LT,RT),
	M=N+1,
	xmlInsSRs(Realm,SNo,M,PRs),
	!
;
    !. 
% ------------------------------------------------------------------------------------------------------------
% xmlNode
% ------------------------------------------------------------------------------------------------------------
xmlNodeElement(Ind,Clas,Label,"",X) :-
	xmlGeneric::xmlAdjustString(Clas,Clas2),
	xmlGeneric::xmlAdjustString(Label,Label2),
	sx::sxQuote(Clas2,ClasQ),
	sx::sxQuote(Label2,LabelQ),
	sx::concat([Ind,"<node class=",ClasQ," label=",LabelQ," />\n"],X),
	!.
xmlNodeElement(Ind,Clas,Label,K,X) :-
	xmlGeneric::xmlAdjustString(Clas,Clas2),
	xmlGeneric::xmlAdjustString(Label,Label2),
	sx::sxQuote(Clas2,ClasQ),
	sx::sxQuote(Label2,LabelQ),
	dbBase::nlTry(K,N),
	xmlGeneric::xmlAdjustString(N,N2),
	sx::sxQuote(N2,Nq),
	sx::concat([Ind,"<node class=",ClasQ," label=",LabelQ," nlLabel=",Nq," />\n"],X),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlProcessTrees
% ------------------------------------------------------------------------------------------------------------
xmlProcessTrees(_,[],0,0,0,0) :-
	!.
xmlProcessTrees(DefRealm,[T|Trees],Nsn,Nsr,Ncr,Nfn) :-
	globalVars::get_i("quit_flag")=1,
	!,fail
;
	T=cmp("sentence",_),
	xmlToSentence(T,Term,Realm,No),
			tt::tt_term(3,"Sentence (from XML)",Term),
	dbBase::recordInsert(2,Realm,No,0,Term,atom("true"),_),
	xmlProcessTrees(DefRealm,Trees,Nsn2,Nsr,Ncr,Nfn),
	Nsn=Nsn2+1,
	!
;
	T=cmp("function",[cmp("Attributes",Atts),cmp("functionName",[_,str(Func)])|_]),
			sx::concat(["Loading function '",Func,"'."],Mess),
			dlgIF::msgS(Mess),
	xmlGeneric::xmlGetAttr("format",Atts,Form),
	xmlGeneric::xmlGetAttr("realm",Atts,Realm),
	xmlGeneric::xmlGetAttr("segment",Atts,Segment),
	xmlToFunc(Form,T,Segment,FuncDef,FuncBodies),
	xmlIdentifierR(Form,Func,K),
	xmlInsFunc(Form,Realm,K,FuncDef,FuncBodies),
			dlgIF::msgS("OK!"),
	xmlProcessTrees(DefRealm,Trees,Nsn,Nsr,Ncr,Nfn2),
	Nfn=Nfn2+1,
	!
;
	T=cmp("inexactRule",[cmp("Attributes",Atts),CR]),
	xmlGeneric::xmlGetAttr("format",Atts,Form),
	xmlGeneric::xmlGetAttr("realm",Atts,Realm),
	xmlToCR_Rule(Form,CR,CR_Term,Def),
			tt::tt_term(3,"Inexact rule (from XML)",CR_Term),
			CR_Term=cmp(_,[atom(Rule)|_]),
			sx::concat(["Loading inexact rule ",Realm," '",Rule,"'."],Mess),
			dlgIF::msgS(Mess),
	dbBase::irReplace(Realm,Def,CR_Term),
			dlgIF::msgS("OK!"),
	xmlProcessTrees(DefRealm,Trees,Nsn,Nsr,Ncr2,Nfn),
	Ncr=Ncr2+1,
	!
;
	T=cmp("sentenceRule",[cmp("Attributes",Atts),SR]),
	xmlGeneric::xmlGetAttr("realmSubNo",Atts,SubNo),
	SubNo="0",
	xmlGeneric::xmlGetAttr("realm",Atts,Realm),
	xmlGeneric::xmlGetAttr("realmNo",Atts,No),
			sx::concat(["Loading sentence rule ",Realm,"|",No,"."],Mess),
			dlgIF::msgS(Mess),
	xmlToRule(SR,RuleTree),
			tt::tt_term(3,"Rule-tree",RuleTree),
	transrt::rlgPR(RuleTree,PRs),
	set::size(PRs,N),
	N=1,
	xmlInsSRs(Realm,No,0,PRs),
			dlgIF::msgS("OK!"),
	xmlProcessTrees(DefRealm,Trees,Nsn,Nsr2,Ncr,Nfn),
	Nsr=Nsr2+N,
	!
;
	T=cmp("sentenceRule",[cmp("Attributes",Atts),SR]),
	xmlGeneric::xmlGetAttr("realmSubNo",Atts,SubNo),
	SubNo="0",
	xmlGeneric::xmlGetAttr("realm",Atts,Realm),
	xmlGeneric::xmlGetAttr("realmNo",Atts,No),
			sx::concat(["Loading sentence rules ",Realm,"|",No,"."],Mess),
			dlgIF::msgS(Mess),
	xmlToRule(SR,RuleTree),
			tt::tt_term(3,"Rule-tree (from XML)",RuleTree),
	transrt::rlgPR(RuleTree,PRs),
	set::size(PRs,N),
	N>1,
	xmlInsSRs(Realm,No,1,PRs),
			dlgIF::msgS("OK!"),
	xmlProcessTrees(DefRealm,Trees,Nsn,Nsr2,Ncr,Nfn),
	Nsr=Nsr2+N,
	!
;
	T=cmp("sentenceRule",[cmp("Attributes",Atts),SR]),
	xmlGeneric::xmlGetAttr("realmSubNo",Atts,SubNo),
	SubNo<>"0",
	xmlGeneric::xmlGetAttr("realm",Atts,Realm),
	xmlGeneric::xmlGetAttr("realmNo",Atts,No),
	xmlToRule(SR,RuleTree),
			tt::tt_term(3,"Rule-tree",RuleTree),
	transrt::rlgPR(RuleTree,PRs),
	set::size(PRs,N),
	N>1,
	sx::concat([Realm,"|",No,"|",SubNo],S),
	dlgIF::feil("xmlProcessTrees",[S,"Error sentence rule XML: Sentence rules involving more than one rule record and must have realmSubno=0"],[],nill),!,fail
;
	T=cmp("sentenceRule",[cmp("Attributes",Atts),SR]),
	xmlGeneric::xmlGetAttr("realmSubNo",Atts,SubNo),
	SubNo<>"0",
	xmlGeneric::xmlGetAttr("realm",Atts,Realm),
	xmlGeneric::xmlGetAttr("realmNo",Atts,No),
			sx::concat(["Loading sentence rule ",Realm,"|",No,"|",SubNo,"."],Mess),
			dlgIF::msgS(Mess),
	xmlToRule(SR,RuleTree),
			tt::tt_term(3,"Rule-tree",RuleTree),
	transrt::rlgPR(RuleTree,PRs),
	set::size(PRs,N),
	N=1,
	SNo=sx::toInt(SubNo),
	xmlInsSRs(Realm,No,SNo,PRs),
			dlgIF::msgS("OK!"),
	xmlProcessTrees(DefRealm,Trees,Nsn,Nsr2,Ncr,Nfn),
	Nsr=Nsr2+N,
	!
;	
	T=cmp("sentenceRule",[cmp("Attributes",Atts),_]),
	xmlGeneric::xmlGetAttr("realmSubNo",Atts,SubNo),
	xmlGeneric::xmlGetAttr("realm",Atts,Realm),
	xmlGeneric::xmlGetAttr("realmNo",Atts,No),
	sx::concat([Realm,"|",No,"|",SubNo],S),
	dlgIF::feil("xmlProcessTrees",[S,"Error in sentence rule XML"],[],nill),!,fail
;
	dlgIF::feil("xmlProcessTrees",[DefRealm,"Unrecognized XML"],[],nill),!,fail.
% ------------------------------------------------------------------------------------------------------------
% xmlTree
% ------------------------------------------------------------------------------------------------------------
xmlTreeElement(Ind,Clas,Label,"",L,X) :-
	xmlGeneric::xmlAdjustString(Clas,Clas2),
	xmlGeneric::xmlAdjustString(Label,Label2),
	sx::sxQuote(Clas2,ClasQ),
	sx::sxQuote(Label2,LabelQ),
	sx::concat(L,Ls),
	sx::concat([Ind,"<tree class=",ClasQ," label=",LabelQ,">\n",Ls,Ind,"</tree>\n"],X),
	!.
xmlTreeElement(Ind,Clas,Label,K,L,X) :-
	xmlGeneric::xmlAdjustString(Clas,Clas2),
	xmlGeneric::xmlAdjustString(Label,Label2),
	sx::sxQuote(Clas2,ClasQ),
	sx::sxQuote(Label2,LabelQ),
	dbBase::nlTry(K,N),
	xmlGeneric::xmlAdjustString(N,N2),
	sx::sxQuote(N2,Nq),
	sx::concat(L,Ls),
	sx::concat([Ind,"<tree class=",ClasQ," label=",LabelQ," nlLabel=",Nq,">\n",Ls,Ind,"</tree>\n"],X),
	!.
% ----------------------------------------------------------------------------------------------
% xmlS_DEs
% ----------------------------------------------------------------------------------------------
class predicates
xmlS_DE_L : (string,string,sterm,sterml) determ (i,i,i,o).
clauses
xmlS_DEs(_,[],[],[]) :-
	!.
xmlS_DEs(_,Ds,[],[]) :-
	dlgIF::feil("xmlS_DEs",["Data element/value list mis-match"|Ds],[],nill),fail.
xmlS_DEs(_,[],Vs,[]) :-
	dlgIF::feil("xmlS_DEs",["Data element/value list mis-match"],[],cmp("",Vs)),fail.
xmlS_DEs(Form,[D|Ds],[V|Vs],[Ta,Tb|Ts]) :-
	dbBase::smClassDefF(D,"a",_,_,Dom,_,_),
	xmlIdentifier(Form,D,Dx),
	sx::sxTail(2,Dom,FDom,Tail),
	Tail="_l",
	Ta=cmp("DATA_ELEMENT",[nill,str(Dx)]),
	xmlS_DE_L(Form,FDom,V,Tbs),
	Tb=cmp("LIST",[nill|Tbs]),         
	xmlS_DEs(Form,Ds,Vs,Ts),
	!
;
	dbBase::smClassDefF(D,"a",_,_,Dom,_,_),
	xmlIdentifier(Form,D,Dx),
	xmlValue(Form,Dom,V,Vx),
	Ta=cmp("DATA_ELEMENT",[nill,str(Dx)]),
	Tb=cmp("VALUE",[nill,str(Vx)]),         
	xmlS_DEs(Form,Ds,Vs,Ts),
	!.
xmlS_DEs(Form,Ds,Vs,_) :-
	dlgIF::feil("xmlS_DEs",[Form|Ds],[],cmp("",Vs)),fail.
xmlS_DE_L(_,_,nill,[]) :-
	!.
xmlS_DE_L(Form,Dom,list(H,T),[X|Xs]) :-
	xmlValue(Form,Dom,H,Vx),
	X=cmp("VALUE",[nill,str(Vx)]),         
	xmlS_DE_L(Form,Dom,T,Xs),
	!.	
% ----------------------------------------------------------------------------------------------
% xmlS_E
% ----------------------------------------------------------------------------------------------
xmlS_E(Form,cmp(E,Vs),T) :-
	xmlIdentifier(Form,E,Ex),
	dbBase::smClassDefF(E,"e",_,Ds,_,_,_),
	xmlS_DEs(Form,Ds,Vs,Ts),
	T=cmp("ENTITY",[nill,
	           cmp("NAME",[nill,str(Ex)])
	           |Ts]),
	!.
% ----------------------------------------------------------------------------------------------
% xmlS_P
% ----------------------------------------------------------------------------------------------
xmlS_P(Form,cmp(P,Vs),T) :-
	xmlIdentifier(Form,P,Px),
	dbBase::smClassDefF(P,"p",_,Ds,_,_,_),
	xmlS_DEs(Form,Ds,Vs,Ts),
	T=cmp("PREDICATE",[nill,
	           cmp("NAME",[nill,str(Px)])
	           |Ts]),
	!.
% ------------------------------------------------------------------------------------------------------------
% xmlCR_scCalls
% ------------------------------------------------------------------------------------------------------------
xmlCR_scCalls(_,_,_,[],[]) :-
	!.
xmlCR_scCalls(Form,CallingRule,Dom,[C|Calls],SwitchCallsT) :-
   	C=cmp(Rule,[nill,nill,nill,nill,ParPart]),
	xmlCR_Call(Form,CallingRule,cmp(Rule,[ParPart]),CallT),
	CallT=cmp(_,CallBody),
	SwitchCall=cmp("scCall",[nill,
	                         cmp("defaultCall",CallBody)]),
	xmlCR_scCalls(Form,CallingRule,Dom,Calls,CallsT2),
	set::conc([[SwitchCall],CallsT2],SwitchCallsT),
	!
;
   	C=cmp(Rule,[atom(Opr),V,nill,nill,ParPart]),
   	xmlIdentifier(Form,Opr,OprX),
   	xmlCR_Val(Form,Dom,V,Vx),
	xmlCR_Call(Form,CallingRule,cmp(Rule,[ParPart]),CallT),
	SwitchCall=cmp("scCall",[nill,
	                         cmp("expression",[nill,
	                                           cmp("operator",[nill,str(OprX)]),
	                                           Vx,
	                                           CallT])]),
	xmlCR_scCalls(Form,CallingRule,Dom,Calls,CallsT2),
	set::conc([[SwitchCall],CallsT2],SwitchCallsT),
	!
;
   	C=cmp(Rule,[atom(Opr),V,atom(Opr2),V2,ParPart]),
   	xmlIdentifier(Form,Opr,OprX),
   	xmlCR_Val(Form,Dom,V,Vx),
   	xmlIdentifier(Form,Opr2,OprX2),
   	xmlCR_Val(Form,Dom,V2,Vx2),
	xmlCR_Call(Form,CallingRule,cmp(Rule,[ParPart]),CallT),
	SwitchCall=cmp("scCall",[nill,
	                         cmp("range",[nill,
	                                      cmp("operator",[nill,str(OprX)]),
	                                      Vx,
	                                      cmp("operator",[nill,str(OprX2)]),
	                                      Vx2,
	                                      CallT])]),
	xmlCR_scCalls(Form,CallingRule,Dom,Calls,CallsT2),
	set::conc([[SwitchCall],CallsT2],SwitchCallsT),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_Bays
% ----------------------------------------------------------------------------------------------
class predicates
xmlToCR_Inc : (string,sterm,sterm) determ (i,i,o).
clauses

xmlToCR_Bays(_,[],[]) :-
	!.
xmlToCR_Bays(Form,[B|Bs],[Bay|Bays]) :-
	B=cmp("bayCall",[_,Inc,Dec,Call]),
	xmlToCR_Inc(Form,Inc,IncK),
	xmlToCR_Inc(Form,Dec,DecK),
	xmlToCR_Call(Form,Call,Rule,Params),
	Bay=cmp(Rule,[IncK,DecK,cmp("p3_iAps",Params)]),
	xmlToCR_Bays(Form,Bs,Bays),
	!.

xmlToCR_Inc(Form,Inc,IncK) :-
	Inc=cmp(_,[_,Call]),
	Call=cmp("call",_),
	xmlToCR_Call(Form,Call,Rule,Params),
	IncK=cmp(Rule,[cmp("p3_iAps",Params)]),
	!
;
	Inc=cmp(_,[_,cmp(_,[_,str(In)])]),
	InN=sx::toReal(In),
	IncK=fix(InN),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_Call/s
% ----------------------------------------------------------------------------------------------
class predicates
xmlToCR_Params : (string,sterml,sterml) determ (i,i,o).	
clauses
xmlToCR_OWA_Sets(_,[],[]) :-
	!.
xmlToCR_OWA_Sets(Form,[C|Cs],Cals) :-
	C=cmp(_,[nill,
	         cmp("weight",[nill,str(Ws)]),
	         cmp("set",[nill,str(R)])
	        |Pars]),
	W=sx::toReal(Ws),
	xmlIdentifierR(Form,R,Rule),
	xmlToCR_Params(Form,Pars,Params),
	Cal=cmp(Rule,[fix(W),cmp("p3_iAps",Params)]),
	xmlToCR_OWA_Sets(Form,Cs,Cals2),
	Cals=[Cal|Cals2],
	!
;
	C=cmp(_,[nill,
	         WCal,
	         cmp("set",[nill,str(R)])
	        |Pars]),
	xmlToCR_Call(Form,WCal,CRule,CPars),
	xmlIdentifierR(Form,R,Rule),
	xmlToCR_Params(Form,Pars,Params),
	Cal=cmp(Rule,[cmp(CRule,[cmp("p3_iAps",CPars)]),cmp("p3_iAps",Params)]),
	xmlToCR_OWA_Sets(Form,Cs,Cals2),
	Cals=[Cal|Cals2],
	!
;
	xmlToCR_OWA_Sets(Form,Cs,Cals),
	!.
xmlToCR_Sets(_,[],[]) :-
	!.
xmlToCR_Sets(Form,[C|Cs],Cals) :-
	C=cmp(_,[nill,cmp("set",[nill,str(R)])|Pars]),
	xmlIdentifierR(Form,R,Rule),
	xmlToCR_Params(Form,Pars,Params),
	Cal=cmp(Rule,[cmp("p3_iAps",Params)]),
	xmlToCR_Sets(Form,Cs,Cals2),
	Cals=[Cal|Cals2],
	!
;
	xmlToCR_Sets(Form,Cs,Cals),
	!.
xmlToCR_Calls(_,[],[]) :-
	!.
xmlToCR_Calls(Form,[C|Cs],[Cal|Cals]) :-
	C=cmp(_,[nill,cmp("rule",[nill,str(R)])|Pars]),
	xmlIdentifierR(Form,R,Rule),
	xmlToCR_Params(Form,Pars,Params),
	Cal=cmp(Rule,[cmp("p3_iAps",Params)]),
	xmlToCR_Calls(Form,Cs,Cals),
	!.
xmlToCR_Call(Form,Call,Rule,Params) :-
	Call=cmp(_,[nill,cmp("rule",[nill,str(R)])|Pars]),
	xmlIdentifierR(Form,R,Rule),
	xmlToCR_Params(Form,Pars,Params),
	!.
xmlToCR_Params(_,[],[]) :-
	!.
xmlToCR_Params(Form,[P,V|Ps],[Par|Pars]) :-
	P=cmp("parameter",[nill,str(D)]),
	xmlIdentifierR(Form,D,DE),
	dbBase::smClassDefF(DE,"a",_,_,Dom,_,_),
	xmlToCR_Val(Form,V,Value),
	opr::oprDom(Dom,"is",Opr),
	Par=cmp(Opr,[atom(DE),Value]),
	xmlToCR_Params(Form,Ps,Pars),
	!.	
% ----------------------------------------------------------------------------------------------
% xmlToCR_Head
% ----------------------------------------------------------------------------------------------
class predicates
xmlToCR_Context : (string,sterml,sterm,sterm) determ (i,i,o,o).
xmlToCR_Pars : (string,sterml,sterm) procedure (i,i,o).
xmlToCR_Prior : (sterml,real) determ (i,o).
clauses

xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Ps) :-
	Heading=cmp("heading",[_,
	                       cmp("rule",[_,str(R)]),
	                       cmp("returnDomain",[_,str(D)])
	                       |Pars]),
	xmlIdentifierR(Form,R,Rk),
	Rule=atom(Rk),
	xmlIdentifierR(Form,D,Dk),
	RDom=atom(Dk),
	xmlToCR_Pars(Form,Pars,Ps),
	xmlToCR_Prior(Pars,Pr),
	Prior=fix(Pr),
	xmlToCR_Context(Form,Pars,Context,Out),
	!.

xmlToCR_Pars(_,[],nill) :-
	!.	
xmlToCR_Pars(Form,Head,list(atom(Par),Pars)) :-
	Head=[cmp("parameter",[nill,str(P)])|T],
	xmlIdentifierR(Form,P,Par),
	xmlToCR_Pars(Form,T,Pars),
	!.
xmlToCR_Pars(_,_,nill) :-
	!.

xmlToCR_Prior([],_) :-
	dlgIF::feil("xmlToCR_Prior",["No prior probability found in XML"],[],nill),!,fail.
xmlToCR_Prior([cmp("prior",[nill,str(P)])|_],Prior) :-
        sx::sxRemChar(P,' ',Pr),
	Prior=sx::toReal(Pr),
	!.
xmlToCR_Prior([_|T],Prior) :-
	xmlToCR_Prior(T,Prior),
	!.

xmlToCR_Context(_,[],nill,nill) :-
	!.
xmlToCR_Context(Form,[C|Cs],Context,Out) :-
	C=cmp("context",[nill,cmp("outOfContext",[nill,V])
	                 |ConCalls]),
	xmlToCR_Val(Form,V,Out),
	xmlToCR_Switch(Form,ConCalls,Calls),
	%xmlToCR_Con(Form,ConCalls,Calls),
	Context=cmp("p3_iCon",Calls),
	!
;
	xmlToCR_Context(Form,Cs,Context,Out),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_Map
% ----------------------------------------------------------------------------------------------
class predicates
xmlToCR_Maps : (string,string,string,sterml,sterml,sterml) determ (i,i,i,i,o,o).
clauses

xmlToCR_Map(_,_,_,[],nill) :-
	!.
xmlToCR_Map(Form,FDom,TDom,[Map],MapK) :-
	Map=cmp("map",[nill,
	               cmp("mapClass",[nill,str("DISCRETE")])
	               |Ms]),
	xmlToCR_Maps(Form,FDom,TDom,Ms,Vls,Cfs),
	MapK=cmp("p3_iDmap",[cmp("p3_iVls",Vls),cmp("p3_iCfs",Cfs)]),
	!
;
	Map=cmp("map",[nill,
	               cmp("mapClass",[nill,str("NUMERIC")])
	               |Ms]),
	xmlToCR_Maps(Form,FDom,TDom,Ms,Vls,Cfs),
	MapK=cmp("p3_iNmap",[cmp("p3_iVls",Vls),cmp("p3_iCfs",Cfs)]),
	!.
	
xmlToCR_Maps(_,_,_,[],[],[]) :-
	!.
xmlToCR_Maps(Form,FDom,TDom,[F,T|Ms],[V|Vls],[C|Cfs]) :-
	xmlToCR_Val(Form,F,V),
	xmlToCR_Val(Form,T,C),
	xmlToCR_Maps(Form,FDom,TDom,Ms,Vls,Cfs),
	!. 
% ----------------------------------------------------------------------------------------------
% xmlToCR_Query
% ----------------------------------------------------------------------------------------------
xmlToCR_Query(Form,Q,Query,Dom) :-
	Form="NL",
	xmlGeneric::xmlReadjustString(Q,Q2),
	irBase::wcrTransQuery(Q2,Query,Dom),
	!
;
	Form="PR",
	xmlGeneric::xmlReadjustString(Q,Q2),
	transpr::proPR_Term(Q2,Query),
	transnl::queryDom(Query,Dom),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_Question
% ----------------------------------------------------------------------------------------------
xmlToCR_Question(Form,Q,str(Qk)) :-
	Form="NL",
	xmlGeneric::xmlReadjustString(Q,Q2),
	irBase::wcrQuestionToIF(Q2,Qk),
	!
;
	xmlGeneric::xmlReadjustString(Q,Qk),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_Rule
% ----------------------------------------------------------------------------------------------
xmlToCR_Rule(Form,Body,CR,Def) :-
	Body=cmp("fuzzyOWA_Set",[nill,
	                         cmp("set",[_,str(R)])
	                        |T]),
	xmlIdentifierR(Form,R,Rk),
	Rule=atom(Rk),
	xmlToCR_Pars(Form,T,Ps),
	xmlToCR_OWA_Sets(Form,T,Tk),
	CR=cmp("p3_iZowa",[Rule,cmp("p3_iZowa",Tk)]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,atom("p3_dmg"),Ps]),str(Seg)]),
	!
;
	Body=cmp("fuzzySet",[nill,
	                     cmp("set",[_,str(R)]),
	                     cmp("operation",[_,str(S)])
	                     |T]),
	xmlIdentifierR(Form,R,Rk),
	Rule=atom(Rk),
	xmlIdentifierR(Form,S,SetSubtype),
	xmlToCR_Pars(Form,T,Ps),
	xmlToCR_Sets(Form,T,Tk),
	irBase::wcrSetType(SetSubType,SetType),
	CR=cmp(SetType,[Rule,cmp(SetSubtype,Tk)]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,atom("p3_dmg"),Ps]),str(Seg)]),
	!
;
	Body=cmp("questionRule",[nill,
	                         Heading,
	                         cmp("default",[nill,DefX]),
	                         cmp("questionDomain",[nill,str(QDom)]),
	                         cmp("question",[nill,str(Q)])
	                         |Map]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Val(Form,DefX,DefV),
	xmlIdentifierR(Form,QDom,QDomK),
	xmlToCR_Question(Form,Q,Qk),
	Question=cmp("p3_iQst",[atom(QDomK),Qk]),
	RDom=atom(Dom),
	xmlToCR_Map(Form,QDomK,Dom,Map,MapK),
	CR=cmp("p3_iCdat",[Rule,Question,MapK,Prior,Out,DefV,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("queryRule",[nill,
	                      Heading,
	                      cmp("default",[nill,DefX]),
	                      cmp("query",[nill,str(Q)])
	                      |Map]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Val(Form,DefX,DefV),
	xmlToCR_Query(Form,Q,Qk,QDomK),
	Query=cmp("p3_iQry",[atom(QDomK),Qk]),
	RDom=atom(Dom),
	xmlToCR_Map(Form,QDomK,Dom,Map,MapK),
	CR=cmp("p3_iCdat",[Rule,Query,MapK,Prior,Out,DefV,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("bayRule",[nill,Heading|Bays]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Bays(Form,Bays,BaysK),
	CR=cmp("p3_iCbay",[Rule,cmp("p3_iBay",BaysK),Prior,Out,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("switchRule",[nill,
	                       Heading,
	                       cmp("default",[nill,DefX]),
	                       cmp("criterionParameter",[nill,str(Dx)])
	                       |SCalls]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Val(Form,DefX,DefV),
	xmlIdentifierR(Form,Dx,Dk),
	xmlToCR_Switch(Form,SCalls,SCallsK),
	CR=cmp("p3_iCswi",[Rule,atom(Dk),cmp("p3_iSwi",SCallsK),Prior,Out,DefV,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("switchRule",[nill,
	                       Heading,
	                       cmp("default",[nill,DefX]),
	                       CCall
	                       |SCalls]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Val(Form,DefX,DefV),
	xmlToCR_Call(Form,CCall,CRule,CParams),
	xmlToCR_Switch(Form,SCalls,SCallsK),
	CR=cmp("p3_iCswi",[Rule,cmp(CRule,[cmp("p3_iAps",CParams)]),cmp("p3_iSwi",SCallsK),Prior,Out,DefV,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("parameterRule",[nill,
	                       Heading,
	                       RCall
	                       |PCalls]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Call(Form,RCall,CRule,CParams),
	dbExtension::dbuCR_GetTD(CRule,_,Dom),
	xmlToCR_ParCalls(Form,Dom,PCalls,PCallsK),
	CR=cmp("p3_iCpar",[Rule,cmp(CRule,[cmp("p3_iAps",CParams)]),cmp("p3_iPCal",PCallsK),Prior,Out,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("andRule",[nill,Heading|SCalls]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Calls(Form,SCalls,SCallsK),
	CR=cmp("p3_iCand",[Rule,cmp("p3_iLand",SCallsK),Prior,Out,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("orRule",[nill,Heading|SCalls]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Calls(Form,SCalls,SCallsK),
	CR=cmp("p3_iCor",[Rule,cmp("p3_iLor",SCallsK),Prior,Out,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("notRule",[nill,Heading,SCall]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Call(Form,SCall,NRule,NParams),
	CR=cmp("p3_iCnot",[Rule,cmp(NRule,[cmp("p3_iAps",NParams)]),Prior,Out,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!
;
	Body=cmp("mapRule",[nill,Heading,
	                    cmp("default",[_,DefX]),
	                    MCal
	                    |Map]),
	xmlToCR_Head(Form,Heading,Rule,RDom,Prior,Out,Context,Params),
	xmlToCR_Val(Form,DefX,DefV),
	xmlToCR_Call(Form,MCal,MRule,MParams),
	dbExtension::dbuCR_GetTD(MRule,_,Dom),
	RDom=atom(RDomK),
	xmlToCR_Map(Form,Dom,RDomK,Map,MapK),
	CR=cmp("p3_iCmap",[Rule,cmp(MRule,[cmp("p3_iAps",MParams)]),MapK,Prior,Out,DefV,Context]),
	Seg=dbBase::getCurrentSegment(),
	Def=cmp("q",[cmp("h",[Rule,RDom,Params]),str(Seg)]),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_ParCalls
% ----------------------------------------------------------------------------------------------
xmlToCR_ParCalls(_,_,[],[]) :-
	!.
xmlToCR_ParCalls(Form,Dom,[Var,PCall|SCalls],[PCallK|PCallsK]) :-
	xmlToCR_Val(Form,Var,V),
	xmlToCR_Call(Form,PCall,PRule,Params),
	PCallK=cmp(PRule,[V,cmp("p3_iAps",Params)]),
	xmlToCR_ParCalls(Form,Dom,SCalls,PCallsK),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_Switch
% ----------------------------------------------------------------------------------------------
xmlToCR_Switch(_,[],[]) :-
	!.
xmlToCR_Switch(Form,[SCall|SCalls],[SCallK|SCallsK]) :-
	SCall=cmp("scCall",[_,
	                        cmp("expression",[_,
	                                          cmp("operator",[_,str(OprS)]),
	                                          Value,
	                                          CallTo])]),
	sx::sxRemChar(OprS,' ',Opr2),	                                          	
	xmlIdentifierR(Form,Opr2,OprK),
	xmlToCR_Val(Form,Value,Val),
	xmlToCR_Call(Form,CallTo,CRule,Params),
	SCallK=cmp(CRule,[atom(OprK),Val,nill,nill,cmp("p3_iAps",Params)]),
	xmlToCR_Switch(Form,SCalls,SCallsK),
	!
;      
	SCall=cmp("scCall",[_,
	                        cmp("range",[_,
	                                          cmp("operator",[_,str(OprS)]),
	                                          Value,
	                                          cmp("operator",[_,str(OprS2)]),
	                                          Value2,
	                                          CallTo])]),
	sx::sxRemChar(OprS,' ',Opr2),	                                          	
	xmlIdentifierR(Form,Opr2,OprK),
	xmlToCR_Val(Form,Value,Val),
	sx::sxRemChar(OprS2,' ',Opr3),	                                          	
	xmlIdentifierR(Form,Opr3,OprK2),
	xmlToCR_Val(Form,Value2,Val2),
	xmlToCR_Call(Form,CallTo,CRule,Params),
	SCallK=cmp(CRule,[atom(OprK),Val,atom(OprK2),Val2,cmp("p3_iAps",Params)]),
	xmlToCR_Switch(Form,SCalls,SCallsK),
	!
;
	SCall=cmp("scCall",[_,DefCall]),
	xmlToCR_Call(Form,DefCall,CRule,Params),
	SCallK=cmp(CRule,[nill,nill,nill,nill,cmp("p3_iAps",Params)]),
	xmlToCR_Switch(Form,SCalls,SCallsK),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToCR_Val
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	NL/PR
% 2	in	xml value element
% 3	out	term value element
%
xmlToCR_Val(Form,XVal,Val) :-
	XVal=cmp("intVal",[nill,str(V)]),
	sx::sxRemChar(V,' ',V2),
	Vi=sx::toInt(V2),
	Val=int(Vi),
	!
;
	XVal=cmp("fixVal",[nill,str(V)]),
	sx::sxRemChar(V,' ',V2),
	Vi=sx::toReal(V2),
	Val=fix(Vi),
	!
;
	XVal=cmp("strVal",[nill,str(V)]),
	xmlGeneric::xmlReadjustString(V,Vs),
	Val=str(Vs),
	!
;
	XVal=cmp("idVal",[nill,str(V)]),
	Form="NL",
	xmlGeneric::xmlReadjustString(V,Vs),
	dbBase::kbID(Vs,Vi,_,_),
	Val=atom(Vi),
	!
;
	XVal=cmp("idVal",[nill,str(V)]),
	Form="PR",
	xmlGeneric::xmlReadjustString(V,Vs),
	Val=atom(Vs),
	!
;
	XVal=cmp("varVal",[nill,str(V)]),
	Form="NL",
	xmlGeneric::xmlReadjustString(V,Vs),
	transnl::parCompSynTree(Vs,Tree),
	Tree=fvar(PVar,_),
	Val=var(PVar),
	!
;
	XVal=cmp("varVal",[nill,str(V)]),
	Form="NL",
	xmlGeneric::xmlReadjustString(V,Vs),
	transnl::parCompSynTree(Vs,Tree),
	Tree=var(PVar),
	Val=var(PVar),
	!
;
	XVal=cmp("varVal",[nill,str(V)]),
	Form="NL",
	xmlGeneric::xmlReadjustString(V,Vs),
	Val=var(Vs),
	!
;
	XVal=cmp("varVal",[nill,str(V)]),
	Form="PR",
	xmlGeneric::xmlReadjustString(V,Vs),
	Val=var(Vs),
	!
;
	dlgIF::feil("xmlToCR_Val",["Unrecognized XML value in certainty rule",Form],[],XVal),!,fail.
% ----------------------------------------------------------------------------------------------
% xmlToDE_List
% ----------------------------------------------------------------------------------------------
class predicates
xmlToDE : (string,sterml,string_list,sterml) determ (i,i,o,o).
xmlToDE_L : (string,string,sterml,sterm) determ (i,i,i,o).
xmlToDE_Order : (string_list,string_list,sterml,sterml) determ (i,i,i,o).
clauses
% parameters:
% 1	in	NL/PR
% 2	in	list of DEs (according to def)
% 3	in	DE/value pairs (xml)
% 4	out	list of values (IF) in sequence of p#2
%
xmlToDE_List(Form,DEs,Dx,Vals) :-
	xmlToDE(Form,Dx,Ds,Vs),
	xmlToDE_Order(DEs,Ds,Vs,Vals),
	!.
	
xmlToDE_Order([],[],_,[]) :-
	!.
xmlToDE_Order([],[D|_],_,_) :-
	sx::concat(["Superflous data element '",D,"' in XML"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail.
xmlToDE_Order([DE|DEs],Ds,Vs,[Val|Vals]) :-
	set::setGet(Ds,Ix,DE),
	set::setRem(Ds,Ix,Ds2),
	set::setGet(Vs,Ix,Val),
	set::setRem(Vs,Ix,Vs2),
	xmlToDE_Order(DEs,Ds2,Vs2,Vals),
	!.
xmlToDE_Order([DE|_],_,_,_) :-
	sx::concat(["Missing data element '",DE,"' in XML"],Mess),
	dlgIF::dlgNote(1,Mess),
	!,fail.
xmlToDE(_,[],[],[]) :-
	!.
xmlToDE(Form,[D,V|DVs],[Dk|Dks],[Vk|Vks]) :-
	D=cmp("DATA_ELEMENT",[_,str(Dx)]),
	xmlIdentifierR(Form,Dx,Dk),
	dbBase::smClassDefF(Dk,"a",_,_,Dom,_,_),
	V=cmp("VALUE",[_,str(Vx)]),
	xmlToVal(Form,Dom,Vx,Vk),
	xmlToDE(Form,DVs,Dks,Vks),
	!
;
	D=cmp("DATA_ELEMENT",[_,str(Dx)]),
	xmlIdentifierR(Form,Dx,Dk),
	dbBase::smClassDefF(Dk,"a",_,_,Dom,_,_),
	sx::sxTail(2,Dom,_,T),
	T="_l",
	V=cmp("LIST",[_,str("")]), % empty list
	Vk=nill,	
	xmlToDE(Form,DVs,Dks,Vks),
	!
;
	D=cmp("DATA_ELEMENT",[_,str(Dx)]),
	xmlIdentifierR(Form,Dx,Dk),
	dbBase::smClassDefF(Dk,"a",_,_,Dom,_,_),
	sx::sxTail(2,Dom,FDom,T),
	T="_l",
	V=cmp("LIST",[_|Ls]),
	xmlToDE_L(Form,FDom,Ls,Vk),	
	xmlToDE(Form,DVs,Dks,Vks),
	!
;
	dlgIF::feil("xmlToDE",["Cannot recognize XML for data element/value pair"],[],cmp("",[D,V])),!,fail.
xmlToDE_L(_,_,[],nill) :-
	!.
xmlToDE_L(Form,Dom,[V|Ls],list(Vk,Vs)) :-
	V=cmp("VALUE",[_,str(Vx)]),
	xmlToVal(Form,Dom,Vx,Vk),
	xmlToDE_L(Form,Dom,Ls,Vs),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToFunc
% ----------------------------------------------------------------------------------------------
class predicates
xmlToFuncPars : (string,string,sterml,sterm,sterml) procedure (i,i,i,o,o).
xmlToFunc2 : (string,string,sterml,sterm,string_list) determ (i,i,i,o,o).
clauses
xmlToFunc(Form,X,Segment,FuncDef,FuncBodies) :-
	X=cmp(_,[_,
	         cmp("functionName",[_,str(F)]),
	         cmp("determinism",[_,str(Det)]),
	         cmp("domain",[_,str(D)])
	         |T]),
	xmlIdentifierR(Form,F,Func),
	xmlIdentifierR(Form,D,Dom),
	set::setGet(["DETERMINISTIC","NONDETERMINISTIC"],Ix,Det),
	set::setGet(["f","g"],Ix,Clas),
	xmlToFunc2(Form,Func,T,DEs,FuncBodies),
	FuncDef=cmp("q",[cmp(Clas,[atom(Func),atom(Dom),DEs]),str(Segment)]),
	!
;
	dlgIF::feil("xmlToFunc",["Unrecognized function XML tree structure"],[],X),!,fail.
xmlToFunc2(Form,Func,X,DEs,Bodies) :-
	xmlToFuncPars(Form,Func,X,DEs,Pars),
	xmlToFuncBodies(Form,Func,Pars,X,Bodies),
	!.
	
xmlToFuncPars(_,_,[],nill,[]) :-
	!.	
xmlToFuncPars(Form,Func,Head,list(atom(D),Ds),[var(VPar)|Pars]) :-
	Head=[cmp("parameter",[nill,str(P)])|T],
	xmlIdentifierR(Form,P,Par),
	dbBase::smClassDefF(Par,"a",_,_,D,_,_),
	vars::varFuncPar(Func,D,VPar),
	xmlToFuncPars(Form,Func,T,Ds,Pars),
	!.
xmlToFuncPars(_,_,_,nill,[]) :-
	!.
% ----------------------------------------------------------------------------------------------
% xmlToFuncBodies
% ----------------------------------------------------------------------------------------------
class predicates
xmlToFuncStats : (sterml,string_list) determ (i,o).
clauses
xmlToFuncBodies(_,_,_,[],[]) :-
	!.
xmlToFuncBodies(Form,Func,Pars,[B|Bs],[Body|Bodies]) :-
	Form="NL",
	B=cmp("functionBody",[_|Stats]),
	xmlToFuncStats(Stats,SL),
	sx::sxlJoin(SL,"",Body),
	xmlToFuncBodies(Form,Func,Pars,Bs,Bodies),
	!
;
	Form="PR",
	B=cmp("functionBody",[_|Stats]),
	xmlToFuncStats(Stats,SL),
	SL=[Body],
	xmlToFuncBodies(Form,Func,Pars,Bs,Bodies),
	!.
xmlToFuncBodies(Form,Func,Pars,[cmp("parameter",_)|Bs],Bodies) :-
	xmlToFuncBodies(Form,Func,Pars,Bs,Bodies),
	!.
xmlToFuncStats([],[]) :-
	!.
xmlToFuncStats([Stat|Stats],[St|Sts]) :-
	Stat=cmp("statement",[_,str(Sx)]),
	xmlGeneric::xmlReadjustString(Sx,St),
	xmlToFuncStats(Stats,Sts),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToSentEnt
% ----------------------------------------------------------------------------------------------
xmlToSentEnt(Form,E,Ei) :-
	E=cmp("ENTITY",[_,
	           cmp("NAME",[_,str(Ex)])
	          |Dx]),
	xmlIdentifierR(Form,Ex,Ek),
	dbBase::smClassDefF(Ek,"e",_,DEs,_,_,_),
	xmlToDE_List(Form,DEs,Dx,Vals),
	Ei=cmp(Ek,Vals),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToSentPred
% ----------------------------------------------------------------------------------------------
xmlToSentPred(Form,P,Pk,Vals,Card) :-
	P=cmp("PREDICATE",[_,
	           cmp("NAME",[_,str(Px)])
	          |Dx]),
	xmlIdentifierR(Form,Px,Pk),
	dbBase::smClassDefF(Pk,"p",Card,DEs,_,_,_),
	xmlToDE_List(Form,DEs,Dx,Vals),
	!.
% ----------------------------------------------------------------------------------------------
% xmlToVal
% ----------------------------------------------------------------------------------------------
% parameters:
% 1	in	NL/PR
% 2	in	value domain
% 3	in	value as XML NL/PR string
% 4	out	value as term
%
xmlToVal("NL",Dom,Vx,Vk) :-
	xmlGeneric::xmlReadjustString(Vx,S),
	syn::sbGetDE_Val(Dom,S,PR),
	transpr::proPR_Term(PR,Vk),
	!.
xmlToVal("PR",Dom,Vx,Vk) :-
	dom::domAtom(Dom),
	xmlGeneric::xmlReadjustString(Vx,S),
	Vk=atom(S),
	!
;
	dom::domStr(Dom),
	xmlGeneric::xmlReadjustString(Vx,S),
	Vk=str(S),
	!
;
	dom::domInt(Dom),
	S=sx::toInt(Vx),
	Vk=int(S),
	!
;
	dom::domFix(Dom),
	S=sx::toReal(Vx),
	Vk=fix(S),
	!
;
	dlgIF::feil("xmlToVal",[Dom,Vx],[],nill),!,fail.
% ----------------------------------------------------------------------------------------------
% xmlValue
% ----------------------------------------------------------------------------------------------
class predicates
xmlValueList : (string,sterm,string_list) determ (i,i,o).
clauses
xmlValue(Form,Dom,V,X) :-
	Form="NL",
	nl::conValX(V,Dom,"",S),
	xmlGeneric::xmlAdjustString(S,X),
	!
;
	V=str(S),
	xmlGeneric::xmlAdjustString(S,X),
	!
;
	V=atom(S),
	xmlGeneric::xmlAdjustString(S,X),
	!
;
	V=int(S),
	X=toString(S),
	!
;
	V=fix(S),
	X=toString(S),
	!
;
	V=list(_,_),
	xmlValueList(Dom,V,L),
	sx::sxlJoin(L,",",X2),
	sx::concat(["[",X2,"]"],X3),
	xmlGeneric::xmlAdjustString(X3,X),
	!.
	
xmlValueList(_,nill,[]) :-
	!.
xmlValueList(Dom,list(H,L),[X|Xs]) :-
	xmlValue("PR",Dom,H,X),
	xmlValueList(Dom,L,Xs),
	!.	
end implement xmlP3
