/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
implement xmlTree
open core,termDomain

facts
xmlTr : sterm := nill.

constants
className = "com/rudstrand/lst/lst".
classVersion = "1.0".

predicates
xmlToTree : (string,sterm) determ (i,o). % xml --> xml-tree

clauses
classInfo(className, classVersion).
% ------------------------------------------------------------------------------------------------------------
% CONSTRUCTORS
% ------------------------------------------------------------------------------------------------------------
newTree(S):-
    xmlToTree(S,X),
    xmlTr:=X,
    !
;
    xmlTr:=nill,
    !.
newTreeXML(XML):- xmlTr:=XML.
% ------------------------------------------------------------------------------------------------------------
% get
% ------------------------------------------------------------------------------------------------------------
get()=Tree :- Tree=xmlTr,!.
% ------------------------------------------------------------------------------------------------------------
% getTagValueAfterTag
% ------------------------------------------------------------------------------------------------------------
predicates
getTAT : (sterml,string,string,string,string) determ (i,i,i,i,o).
getTATs : (sterml,string,string) determ (i,i,o).
clauses
getTagValueAfterTag(Tag,TagValue,NextTag)=Value :-
    Root=xmlTr,
    Root=cmp(_,[_|Trees]),
    getTAT(Trees,Tag,TagValue,NextTag,Value),
    !.
    
getTAT([H|T],Tag,TagValue,NextTag,Value) :-
    H=cmp(Tag,[_,str(TagValue)]),
    getTATs(T,NextTag,Value),
    !
;
    H=cmp(_,[_|Subs]),
    getTAT(Subs,Tag,TagValue,NextTag,Value),
    !
;
    getTAT(T,Tag,TagValue,NextTag,Value),
    !.
getTATs([H|T],Tag,Value) :-
    H=cmp(Tag,[_,str(Value)]),
    !
;    
    getTATs(T,Tag,Value),
    !.
% ------------------------------------------------------------------------------------------------------------
% getSubTree
% ------------------------------------------------------------------------------------------------------------
predicates
getSub : (sterm,string,sterm) determ (i,i,o).
getSubs : (sterml,string,sterm) determ (i,i,o).
clauses
getSubTree(Tag)=Tree:-
    Root=xmlTr,
    getSub(Root,Tag,Tree),
    !.
    
getSub(T,Tag,Tree) :-
    T=cmp(Tag,_),
    Tree=T,
    !
;
    T=cmp(_,L),
    getSubs(L,Tag,Tree),
    !.
    
getSubs([H|T],Tag,Tree) :-
    getSub(H,Tag,Tree),
    !
;
    getSubs(T,Tag,Tree),
    !.
% ------------------------------------------------------------------------------------------------------------
% getAllOfTag
% ------------------------------------------------------------------------------------------------------------
predicates
getAllOfTagX : (string,sterm,string_list) determ (i,i,o).
getAllOfTags : (string,sterml,string_list) determ (i,i,o).
clauses
getAllOfTag(Tag)=L:-
    Tree=xmlTr,
    Tree=cmp(Tag,[_,str(S)]),
    L=[S],
    !
;    
    Tree=xmlTr,
    Tree=cmp(Tag,_),
    pfeil::syserr("xmlTr/getAllOfTag","Invalid tag for this predicate",[Tag]),!,fail
;
    Tree=xmlTr,
    Tree=cmp(_,T),
    getAllOfTags(Tag,T,L),
    !
;
    L=[],
    !.
    
getAllOfTagX(Tag,Tree,L) :-
    Tree=cmp(Tag,[_,str(S)]),
    L=[S],
    !
;    
    Tree=cmp(Tag,_),
    pfeil::syserr("xmlTr/getAllOfTagX","Invalid tag for this predicate",[Tag]),!,fail
;
    Tree=cmp(_,T),
    getAllOfTags(Tag,T,L),
    !
;
    L=[],
    !.
    
getAllOfTags(_,[],[]) :-
    !.
getAllOftags(Tag,[H|T],L) :-
     getAllOfTagX(Tag,H,L2),
     getAllOfTags(Tag,T,L3),
     set::conc([L2,L3],L),
     !.
% ------------------------------------------------------------------------------------------------------------
% xmlToTree
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlEmptyTag : (string,string,sterm,string) determ (i,o,o,o).	
xmlOpenTag : (string,string,sterm,string) determ (i,o,o,o).	
xmlParseAttrs : (string,sterml) determ (i,o).	
xmlParseTag : (string,string,sterm) determ (i,o,o).
xmlRemoveComments : (string,string) procedure (i,o).
xmlTag : (string,sterm,string) determ (i,o,o).
xmlTags : (string,sterml,string) procedure (i,o,o).
xmlContents : (string,string,integer,sterml,string) determ (i,i,i,o,o).
clauses
% parameters:
% 1	in	xml text
% 2	out	xml trees
%
xmlToTree(S,Tree) :-
	xmlRemoveComments(S,S2),
	S3=sx::replaceAll(S2,"\n",""),
	xmlTag(S3,Tree,SRest),
	SRest="",
	!.
xmlTag(S,Tree,SRest) :-
	% <Tag ... />	
	xmlEmptyTag(S,Tag,Attrs,S2),
	sx::sxStrip(S2,SRest),
	Tree=cmp(Tag,[Attrs]),
	!
;
	% <Tag...> contents </Tag> ...
	xmlOpenTag(S,Tag,Attrs,S2),
	sx::concat(["</",Tag,">"],CloseTag),
	L=string::length(CloseTag),
	xmlContents(S2,CloseTag,L,Trees,S3),
	sx::sxStrip(S3,SRest),
	Tree=cmp(Tag,[Attrs|Trees]),
	!.
xmlTags(S,[Tree|Trees],T) :-
	xmlTag(S,Tree,S2),
	xmlTags(S2,Trees,T),
	!.
xmlTags(S,[],S) :-
	!.	
	
xmlContents(S,CloseTag,L,Trees,T) :-
	sx::frontstr(L,S,F,S2),
	F=CloseTag,
	sx::sxStrip(S2,T),
	Trees=[str("")],
	!
;
	sx::searchchar(S,'<',P),
	sx::searchstring(S,CloseTag,Q),
	P<Q,
	xmlTags(S,Trees,S2),
	sx::frontstr(L,S2,F,S3),
	F=CloseTag,
	sx::sxStrip(S3,T),
	!
;
	sx::searchstring(S,CloseTag,P),
	Q=P-1,
	sx::frontstr(Q,S,V,_),
	R=P+L-1,
	sx::frontstr(R,S,_,T),
	Trees=[str(V)],
	!.
xmlEmptyTag(S,Tag,Attrs,T) :-
	sx::frontchar(S,'<',S2),
	sx::searchchar(S2,'>',P),
	Q=P-1,
	Q>1,
	sx::searchstring(S2,"/>",P2),
	P2=Q,
	Q2=P2-1,
	sx::frontstr(Q2,S2,STag,_),
	xmlParseTag(STag,Tag,Attrs),
	sx::frontstr(P,S2,_,S3),
	sx::sxStrip(S3,T),
	!.
xmlOpenTag(S,Tag,Attrs,T) :-
	sx::frontchar(S,'<',S2),
	sx::searchchar(S2,'>',P),
	Q=P-1,
	Q>0,
	sx::frontstr(Q,S2,STag,_),
	xmlParseTag(STag,Tag,Attrs),
	sx::frontstr(P,S2,_,S3),
	sx::sxStrip(S3,T),
	!.
xmlParseTag(S,Tag,Attrs) :-
	sx::sxUnpack(S,' ',L),
	L=[Tag|_],
	N=string::length(Tag),
	sx::frontstr(N,S,_,S2),
	sx::sxStrip(S2,S3),
	xmlParseAttrs(S3,Atts),
	Atts=[],
	Attrs=nill,
	!
;
	sx::sxUnpack(S,' ',L),
	L=[Tag|_],
	N=string::length(Tag),
	sx::frontstr(N,S,_,S2),
	sx::sxStrip(S2,S3),
	xmlParseAttrs(S3,Atts),
	Attrs=cmp("Attributes",Atts),
	!.
	
xmlParseAttrs("",[]) :-
	!.
xmlParseAttrs(S,[Attr|Attrs]) :-
	sx::searchchar(S,'=',P2),
	P2>1,
	P3=P2-1,
	sx::frontstr(P3,S,SAttr,SVal),
	sx::sxStrip(SAttr,Att),
	sx::sxGetInBrack(SVal,'"','"',Val),
	Attr=cmp(Att,[str(Val)]),
	sx::searchchar(S,'"',P),
	sx::frontstr(P,S,_,S2),
	sx::searchchar(S2,'"',Q),
	sx::frontstr(Q,S2,_,S3),
	sx::sxStrip(S3,S4),
	xmlParseAttrs(S4,Attrs),
	!.
	
xmlRemoveComments(S,T) :-
	sx::frontstr(2,S,F,_),
	F="<?",
	sx::searchstring(S,"?>",P4),
	P5=P4+1,
	sx::frontstr(P5,S,_,S3),
	xmlRemoveComments(S3,T),
	!
;
	sx::searchstring(S,"<?",P2),
	P3=P2-1,
	sx::frontstr(P3,S,S2,_),
	sx::searchstring(S,"?>",P4),
	P5=P4+1,
	sx::frontstr(P5,S,_,S3),
	S4=string::concat(S2,S3),
	xmlRemoveComments(S4,T),
	!
;
	sx::frontstr(4,S,F,_),
	F="<!--",
	sx::searchstring(S,"-->",P4),
	P5=P4+2,
	sx::frontstr(P5,S,_,S3),
	xmlRemoveComments(S3,T),
	!
;
	sx::searchstring(S,"<!--",P2),
	P3=P2-1,
	sx::frontstr(P3,S,S2,_),
	sx::searchstring(S,"-->",P4),
	P5=P4+2,
	sx::frontstr(P5,S,_,S3),
	S4=string::concat(S2,S3),
	xmlRemoveComments(S4,T),
	!
;
	T=S,
	!.
% ------------------------------------------------------------------------------------------------------------
% getX
% ------------------------------------------------------------------------------------------------------------
class predicates
xmlAtts : (sterml,string) determ (i,o).	
xmlFromTrees : (string,string,sterml,string) determ (i,i,i,o).
xmlFromTreeX : (string,string,sterm,string) determ (i,i,i,o).	
clauses
% parameter:
% 1	in	size of indentation [0,1,..
%
getX(Indent)=S :-
    T=xmlTr,
	sx::sx_ini(Indent,' ',Ind),
	xmlFromTreeX("",Ind,T,S),
	!.
	
xmlFromTreeX(Curr,Ind,T,S) :-
	T=cmp(Simple,[cmp("Attributes",Atts)]),
	xmlAtts(Atts,S3),
	sx::concat([Curr,"<",Simple,S3,"></",Simple,">\n"],S),
	!
;	
	T=cmp(Simple,[cmp("Attributes",Atts),str(S2)]),
	xmlAtts(Atts,S3),
	sx::concat([Curr,"<",Simple,S3,">",S2,"</",Simple,">\n"],S),
	!
;	
	T=cmp(Complex,[cmp("Attributes",Atts)|Elms]),
	NewCurr=string::concat(Curr,Ind),
	xmlFromTrees(NewCurr,Ind,Elms,S2),
	xmlAtts(Atts,S3),
	sx::concat([Curr,"<",Complex,S3,">\n",S2,Curr,"</",Complex,">\n"],S),
	!
;
	T=cmp(Simple,[nill]),
	sx::concat([Curr,"<",Simple,"></",Simple,">\n"],S),
	!
;	
	T=cmp(Simple,[nill,str(S2)]),
	sx::concat([Curr,"<",Simple,">",S2,"</",Simple,">\n"],S),
	!
;	
	T=cmp(Complex,[nill|Elms]),
	NewCurr=string::concat(Curr,Ind),
	xmlFromTrees(NewCurr,Ind,Elms,S2),
	sx::concat([Curr,"<",Complex,">\n",S2,Curr,"</",Complex,">\n"],S),
	!.
xmlFromTrees(_,_,[],"") :-
	!.
xmlFromTrees(Curr,Ind,[H|T],S) :-
	xmlFromTreeX(Curr,Ind,H,S2),
	xmlFromTrees(Curr,Ind,T,S3),
	S=string::concat(S2,S3),
	!.
xmlAtts([],"") :-
	!.
xmlAtts([A|As],S) :-
	A=cmp(Att,[str(V)]),
	Cq=string::charToString('"'),
	sx::concat([" ",Att,"=",Cq,V,Cq],S2),
	xmlAtts(As,S3),
	S=string::concat(S2,S3),
	!.
% ------------------------------------------------------------------------------------------------------------
% removeAmpersands
% ------------------------------------------------------------------------------------------------------------
class predicates
remAmpsX : (sterml,sterml) determ (i,o).
clauses

removeAmpersands(str(S),str(R)) :-
    S1=sx::replaceAll(S,"&lt;","<"),
    S2=sx::replaceAll(S1,"&gt;",">"),
    Q=string::charToString('"'),
    R=sx::replaceAll(S2,"&quot;",Q),
    !.
removeAmpersands(cmp(H,T),cmp(H,R)) :-
    remAmpsX(T,R),
    !.
removeAmpersands(C,C) :-
    !.
    
remAmpsX([],[]) :-
    !.
remAmpsX([H|T],[R|Tr]) :-
    removeAmpersands(H,R),
    remAmpsX(T,Tr),
    !.
end implement xmlTree
